; a c t o r t e s t . s c h
;
; lth@cs.uoregon.edu / August 14, 1995
; $Id$
;
; Test program for actors code.

(require "actors.sch")

; This is the template for an 'account' actor.

(define acct-1 
  (actor (acctno)
    (define balance 0)

    (lambda (msg)
      (case (car msg)
	((withdraw) (set! balance (- balance (cadr msg))))
	((deposit)  (set! balance (+ balance (cadr msg))))
	((query)    (send (cadr msg) (list 'balance acctno balance)))))))

; A little bit more complicated

(define acct-2 
  (actor (acctno)
    (define balance 0)

    (lambda (msg)
      (case (car msg)
	((withdraw) (set! balance (- balance (cadr msg)))
		    (send (caddr msg) '(done)))
	((deposit)  (set! balance (+ balance (cadr msg)))
		    (send (caddr msg) '(done)))
	((query)    (send (cadr msg) (list 'balance acctno balance)))))))

; This is the template for a simple 'receptionist' actor.

(define receptionist-1
  (actor ()
    (lambda (msg)
      (case (car msg)
	((query)   (send (cadr msg) (list 'query *self*)))
	((balance) (display-balance (cadr msg) (caddr msg)))))))

; A slightly more sophisticated receptionist.

(define receptionist-2
  (actor ()
    (lambda (msg)
      (case (car msg)
	((query)    (send (cadr msg) (list 'query *self*)))
	((deposit)  (send (cadr msg) (list 'deposit (caddr msg))))
	((withdraw) (send (cadr msg) (list 'withdraw (caddr msg))))
	((balance)  (display-balance (cadr msg) (caddr msg)))))))

; A more sophisticated receptionist still.
; The emerging pattern of 'send-and-wait-and-revert' could be abstracted.

(define receptionist-3
  (actor ()
    (lambda (msg)
      (case (car msg)
	((query)    
	 (let ((acct (cadr msg))
	       (recv (caddr msg)))
	   (set! *behavior* 
		 (let ((b *behavior*))
		   (lambda (msg)
		     (display-balance (cadr msg) (caddr msg))
		     (set! *behavior* b)
		     (send recv '(done)))))
	   (send acct (list 'query *self*))))
	((deposit withdraw)
	 (let ((acct (cadr msg))
	       (amt  (caddr msg))
	       (recv (cadddr msg)))
	   (set! *behavior*
		 (let ((b *behavior*))
		   (lambda (msg)
		     (set! *behavior* b)
		     (send recv '(done)))))
	   (send acct (list (car msg) amt *self*))))))))

(define (display-balance acctno balance)
  (format #t "Balance on ~a is ~a~%" acctno balance))

; This is the first version of the code. It has a race
; condition: both messages may be delivered to the account
; before any are delivered from the receptionist to the account.

(define (example1)
  (let ((acct         (make-actor acct-2 1234567890))
	(receptionist (make-actor receptionist-1)))
    (send receptionist `(query ,acct))
    (send acct `(deposit ,500))
    (send acct `(withdraw ,1000))
    (send receptionist `(query ,acct))))


; This is the second version of the code. It has a race condition too,
; because messages may not be delivered in the order they are sent.

(define (example2)
  (let ((acct         (make-actor acct-1 1234567890))
	(receptionist (make-actor receptionist-2)))
    (send receptionist `(query ,acct))
    (send receptionist `(deposit ,acct ,500))
    (send receptionist `(withdraw ,acct ,1000))
    (send receptionist `(query ,acct))))
  
; This is the third version of the code. It is correct: it synchronizes
; after each operation, as necessary, by requiring the receptionist
; to report back.

(define (example3)
  (let* ((acct         (make-actor acct-1 1234567890))
	 (receptionist (make-actor receptionist-3))
	 (user         (make-actor u receptionist acct)))
    (send user '(start))))

; This is all very CPS, of course. Should have a macro 'become' here?

(define u
  (actor (receptionist acct)
    (lambda (msg)
      (set! *behavior*
	    (lambda (msg)
	      (set! *behavior*
		    (lambda (msg)
		      (set! *behavior*
			    (lambda (msg)
			      (set! *behavior* 
				    (lambda (msg)
				      (display "DONE!") (newline)))
			      (send receptionist `(query ,acct ,*self*))))
		      (send receptionist `(withdraw ,acct ,1000 ,*self*))))
	      (send receptionist `(deposit ,acct ,500 ,*self*))))
      (send receptionist `(query ,acct ,*self*)))))

; Driver

(define *actor-init* #f)
(define (actor-test n)
  (if (not *actor-init*)
      (begin (begin-tasking)
	     (set! *actor-init* #t)))
  (case n
    ((1) (example1))
    ((2) (example2))
    ((3) (example3))
    ((0) (end-tasking)))
  #t)

; eof
