; p h i l o s . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 25, 1995
; $Id$
;
; The ubiquitos dining philosophers problem, in STk, using threads.
;
; Demonstrates some uses of the synchronization mechanisms.
; Note that Sleep requires Tk, so this won't work with snow.
;
; Exercise: animate this using Tk, a round table, and different colors
;           or icons for different states.

(require "threads.sch")
(require "synch.sch")
(require "sleep.sch")
(require "monitor.sch")

(monitor forks (init grab drop waittime)

  (define v #f)
  (define w #f)

  (define (init n)
    (set! v (make-vector n #f))
    (set! w (make-vector n 0)))

  (define (grab id n t)
    (if (not (vector-ref v n))
        (begin (vector-set! v n id)
	       (vector-set! w n t)
               #t)
	(begin (format #t "philosopher ~a failed to grab ~a!~%" id n)
	       #f)))

  (define (drop id n)
    (if (not (eqv? (vector-ref v n) id))
        (error "Philosopher ~a can't drop ~a!~%" id n)
        (begin (vector-set! v n #f)
	       (vector-set! w n 0))))

  (define (waittime n)
    (vector-ref w n))

;end monitor forks
)

; STk's random procedure is not thread-safe.

(define random
  (let ((r random))
    (atomic-lambda (n)
      (r n))))

(define (make-philosopher pos count)
  (let ((left  pos)
	(right (modulo (+ pos 1) count)))

    (define (grabloop t)
      (if (not (forks.grab pos left t))
	  (grabloop2 t)
	  (if (not (forks.grab pos right t))
	      (begin (forks.drop pos left)
		     (grabloop2 t)))))

    (define (grabloop2 t)
      (sleep (max (forks.waittime left) (forks.waittime right)) 's)
      (grabloop t))

    (define (think)
      (let ((t (random 8)))
	(format #t "Philosopher ~a is thinking...~%" pos)
	(sleep t 's)))

    (define (eat)
      (let ((t (random 3)))
	(format #t "Philosopher ~a is hungry...~%" pos)
	(grabloop t)
	(format #t "Philosopher ~a is eating...~%" pos)
	(sleep t 's)
	(forks.drop pos left)
	(forks.drop pos right)))

    (define (die)
      (format #t "Philosopher ~a ceased to exist...~%" pos))

    (define (loop)
      (think)
      (eat)
      (if (not (zero? (random 50)))
	  (loop)
	  (die)))

    (spawn (lambda ()
	     (format #t "\"I think, therefore I am!\"~%")
	     (loop)))))

; To run: call (begin-tasking) 
;         then (philosophers 5) for five philosophers.

(define (philosophers n)
  (forks.init n)
  (do ((i (- n 1) (- i 1)))
      ((< i 0) #t)
    (make-philosopher i n)))

; eof
