; s i m . s c h
;
; Small simulation program to test tasking stuff.
; lth@cs.uoregon.edu / June 26, 1995
;
; $Id: sim.sch,v 1.2 1995/07/11 18:10:42 lth Exp $
;
; The program simulates a doctors' office where there are some doctors
; treating a bunch of patients. There is a queue of available doctors
; and a queue of patients. The receptionist gets a patient and a doctor
; and sends them off to have the doctor treat the patient, after which
; the doctor makes himself available for more work, and the patient leaves.

(require "simfw.sch")

; Simulation starts here

(define dqueue #f)                ; queue of doctors
(define pqueue #f)                ; queue of patients
(define new-doctor-id #f)         ; stream of doctor IDs
(define new-patient-id #f)        ; stream of patient IDs
(define treatment-time #f)        ; stream of treatement times
(define next-patient-time #f)     ; stream of patient interarrival times

(define (simulate doctors patients)
  (setup-simulation)
  (msgctl 'info)
  (set! dqueue (make-sim-queue))
  (set! pqueue (make-sim-queue))
  (set! new-doctor-id (integers 0))
  (set! new-patient-id (integers 0))
  (set! treatment-time (random-stream 'uniform 10 60))
  (set! next-patient-time (random-stream 'uniform 1 5))
  (create-receptionist)
  (create-doctors doctors)
  (create-patients patients)
  (start-simulation))


; A receptionist is a process which pairs patients with doctors.

(define (create-receptionist)

  (define (receptionist)
    (msg 'debug "[receptionist] waiting for patient.")
    (let ((patient (receive-patient)))
      (msg 'debug "[receptionist] waiting for doctor.")
      (let ((doctor  (receive-doctor)))
	(msg 'debug "[receptionist] doctor ~a / patient ~a." 
	     (doctor.id doctor) patient)
	(treat doctor patient)
	(receptionist))))

  (simspawn receptionist))


; A doctor is a process receiving patients (essentially, an actor).

(define (create-doctors ndoc)

  (define (a-doctor self)
    (define id (doctor.id self))

    (define (loop)
      (msg 'debug "[doctor ~a] waiting for patient." id)
      (let ((p (doctor.receive-patient self))
	    (t (treatment-time)))
	(msg 'info "[doctor ~a] treating ~a for ~a minutes." id p t)
	(sleep-for t)
	(msg 'info "[doctor ~a] patient ~a is well, I'm available." id p)
	(doctor-available self)
	(loop)))

    (loop))

  (define (create-doctor)
    (let ((doc (make-doctor)))
      (simspawn (lambda () (a-doctor doc)))
      doc))

  (do ((i ndoc (- i 1)))
      ((zero? i) #t)
    (doctor-available (create-doctor))))


; Produce a steady stream of patients, npat of them. A patient is a static
; entity, represented simply by a numeric ID.

(define (create-patients patients)

  (define (patproducer npat)
    (if (not (zero? npat))
	(let* ((id (new-patient-id))
	       (t  (next-patient-time)))
	  (msg 'debug "[patproducer] sleeping for ~a minutes." t)
	  (sleep-for t)
	  (msg 'debug "[patproducer] awake, creating patient ~a." id)
	  (qput pqueue id)
	  (patproducer (- npat 1)))))

  (simspawn (lambda () (patproducer patients))))


(define (treat doctor patient) (qput (doctor.queue doctor) patient))
(define (doctor-available doc) (qput dqueue doc))
(define (receive-patient) (qget pqueue))
(define (receive-doctor) (qget dqueue))

(define (make-doctor) (list 'doctor (new-doctor-id) (make-sim-queue)))
(define doctor.id cadr)
(define doctor.queue caddr)
(define (doctor.receive-patient self) (qget (doctor.queue self)))

(provide "sim.sch")

; eof
