; b b u f . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 25, 1995
; $Id: bbuf.sch,v 1.1 1995/07/26 18:29:59 lth Exp lth $
;
; Simple bounded buffer.

(require "synch.sch")

(define (make-bounded-buffer size)
  (let ((v (make-vector (+ size 5) #f)))
    (vector-set! v 0 (make-semaphore 0))    ; elements available
    (vector-set! v 1 (make-semaphore size)) ; slots available
    (vector-set! v 2 5)                     ; next element
    (vector-set! v 3 5)                     ; next slot
    (vector-set! v 4 (make-mutex))          ; controls access
    v))

(define (bbuf.eltsem bbuf) (vector-ref bbuf 0))
(define (bbuf.slotsem bbuf) (vector-ref bbuf 1))
(define (bbuf.nextelt bbuf) (vector-ref bbuf 2))
(define (bbuf.nextslot bbuf) (vector-ref bbuf 3))
(define (bbuf.mutex bbuf) (vector-ref bbuf 4))
(define (bbuf.nextelt! bbuf e) (vector-set! bbuf 2 e))
(define (bbuf.nextslot! bbuf e) (vector-set! bbuf 3 e))

(define (bbuf.next bbuf idx)
  (+ (remainder (+ (- idx 5) 1) (- (vector-length bbuf) 5)) 5))

(define (bounded-buffer-put! bbuf elt)
  (semaphore-down (bbuf.slotsem bbuf))         ; wait for slot
  (call-with-mutex (bbuf.mutex bbuf)
    (lambda ()
      (let ((idx (bbuf.nextslot bbuf)))
	(vector-set! bbuf idx elt)
	(bbuf.nextslot! bbuf (bbuf.next bbuf idx))
	(semaphore-up (bbuf.eltsem bbuf))))))  ; signal element

(define (bounded-buffer-get bbuf)
  (semaphore-down (bbuf.eltsem bbuf))          ; wait for element
  (call-with-mutex (bbuf.mutex bbuf)
    (lambda ()
      (let* ((idx (bbuf.nextelt bbuf))
	     (elt (vector-ref bbuf idx)))
	(bbuf.nextelt! bbuf (bbuf.next bbuf idx))
	(semaphore-up (bbuf.slotsem bbuf))     ; signal slot
	elt))))

(provide "bbuf.sch")

; eof
