; c a l l b a c k . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 26, 1995
; $Id: callback.sch,v 1.1 1995/07/26 18:29:59 lth Exp lth $
;
; Some wrappers to allow closures to be used as tk command handlers.

(define (tk-callback proc)
  (let ((name (tk-remember proc)))
    (format #f "do-tk-callback '~a" name)))

(define (tk-oneshot-callback proc)
  (letrec ((name (tk-remember (lambda args
				(tk-recycle name)
				(apply proc args)))))
    (format #f "do-tk-callback '~a" name)))

(define (do-tk-callback name . args)
  (let ((proc (tk-lookup name)))
    (apply proc args)))

; NOTE: This table is never garbage collected. If you are registering
; a one-time handler, use tk-oneshot-callback rather than tk-callback.

(define tk-table (make-hash-table))

; NOTE: There is a race condition here in a concurrent system.
; This can be fixed by wrapping the body of the lambda in 
; call-without-interrupts.

(define tk-remember 
  (let ((i 0))
    (lambda (obj)
      (let ((n (string->symbol (string-append "__" (number->string i)))))
	(set! i (+ i 1))
	(hash-table-put! tk-table n obj)
	n))))

(define (tk-lookup name)
  (let ((x (hash-table-get tk-table name #f)))
    (if (not x)
	(error "tk-lookup: not found: " name)
	x)))

(define (tk-recycle name)
  (hash-table-remove! tk-table name))


; Garbage collecting the table -- not implemented.

(define (tk-gctwa)
  ; let h be a new hash table
  ; enumerate all widgets. for each widget, get its command
  ;   if the command has the form "do-tk-callback 'xxx" then 
  ;     lookup xxx in tk-table, let V be its value.
  ;     enter xxx and V into h.
  ;   fi
  ; od
  ; replace tk-table with h
  #f)

(provide "callback.sch")

; eof
