; f u t u r e . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 26, 1995
; $Id$
;
; Simple implementation of futures.
; For STk with tasking extensions.
;
; REFERENCES
;   Robert H. Halstead, Jr. "New Ideas in Parallel Lisp: Language Design,
;   Implementation, and Programming Tools.", in "Parallel Lisp: Languages
;   and Systems", US/Japan Workshop on Parallel Lisp, Sendai, Japan, June
;   1989. Proceedings published by Springer-Verlag as Lecture Notes in
;   Computer Science #441, ISBN 3-540-52782-6 and 0-387-52786-6.
;
; This implementation is not as slick as the original, since the original
; has implicit touching. Implementing implicit touching in STk would be
; hard (a lot of work): all the primitives would have to be instrumented. 
; Essentially, each type check would have an extra case.
;
; In our case, a future is simply a combination of a sync variable and
; a thread.
;
; BUGS
; - Macros are not hygienic.
; - Should use objects or records for the data type.
; - You must call 'begin-tasking' before you create any futures.

(require "threads.sch")
(require "synch.sch")

(define-macro (future expr)
  (let ((sync (gensym)))
    `(let ((,sync (make-sync)))
       (spawn (lambda () (sync-set! ,sync ,expr)))
       (vector 'future ,sync))))

(define (future? f)
  (and (vector? f)
       (> (vector-length f) 0)
       (eq? (vector-ref f 0) 'future)))

(define (touch f)
  (if (future? f)
      (sync-ref (vector-ref f 1))
      (error "get-future-value: not a future: " f)))

; eof
