; p c a l l . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / July 26, 1995
; $Id: pcall.sch,v 1.1 1995/07/26 18:29:59 lth Exp lth $
;
; 'Pcall' control construct, based on tasking. 
; For STk with tasking extensions.
;
; Pcall is a parallel-evaluation form of the common procedure call.
; In the expression  (pcall E1 E2 ... En), all of the Ei are evaluated
; in parallel, and the value of E1 (a procedure) is then called with the
; values of E2 ... En as actual parameters.
;
; Pcall makes no sense on a uniprocessor and is provided as an 
; illustration only.

(require "threads.sch")
(require "synch.sch")

; This is a simple implementation, but the resulting code is inefficient,
; creating a task for every expression.
;
;(define-macro (pcall . ops)
;  (let ((name (gensym))
;	 (eops (gensym)))
;    `(let ((,eops (map sync-ref
;		       (list ,@(map (lambda (o)
;				      `(let ((,name (make-sync)))
;					 (spawn (lambda ()
;						  (sync-set! ,name ,o)))
;					 ,name))
;				    ops)))))
;       (apply (car ,eops) (cdr ,eops)))))


; This is a more complicated implementation, but the resulting code is
; pretty efficient. Tasks are spawned only for non-constant expressions,
; and there is no use of apply.

(define-macro (pcall . ops)
  (letrec ((selfev? (lambda (o)
		      (or (number? o)
			  (and (pair? o) (eq? (car o) 'quote))
			  (and (pair? o) (eq? (car o) 'lambda))
			  (string? o)
			  (vector? o)
			  (boolean? o)
			  (char? o)
			  (symbol? o))))
	   (wrap    (lambda (o)
		      (let ((name (gensym)))
			`(let ((,name (make-sync)))
			   (spawn (lambda ()
				    (sync-set! ,name ,o)))
			   ,name))))
	   (unwrap  (lambda (n)
		      `(sync-ref ,(car n)))))
    (let* ((names (let loop ((names '()) (ops ops))
		   (cond ((null? ops)
			  (reverse names))
			 ((selfev? (car ops))
			  (loop names (cdr ops)))
			 (else
			  (loop (cons (list (gensym) (wrap (car ops))) names)
				(cdr ops))))))
	   (evlist (let loop ((names names) (ops ops) (l '()))
		     (cond ((null? ops)
			    (reverse l))
			   ((selfev? (car ops))
			    (loop names
				  (cdr ops)
				  (cons (car ops) l)))
			   (else
			    (loop (cdr names)
				  (cdr ops)
				  (cons (unwrap (car names)) l)))))))
      `(let ,names
	 ,evlist))))

(provide "pcall.sch")

; eof
