; p r o t e c t q . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / June 26, 1995
; $Id: queue2.sch,v 1.1 1995/06/27 18:48:11 lth Exp $
;
; Some more sophisticated queues for multitasking.

(require "queue.sch")
(require "synch.sch")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Blocking protected queue 

(define (make-protected-blocking-queue . rest)
  (define protect (make-mutex))
  (define block   (make-semaphore 0))

  (define q (if (null? rest)
		(make-basic-queue)
		(make-basic-queue (car rest))))

  (define (get)
    (call-with-mutex protect
      (lambda ()
	(mutex-signal-and-semaphore-down protect block)
	(qget q))))

  (define (put item)
    (call-with-mutex protect
      (lambda ()
	(qput q item)
	(semaphore-up block))))

  (define (empty?)
    (call-with-mutex protect
      (lambda ()
	(qempty? q))))

  (define (data)
    (call-with-mutex protect
      (lambda ()
	(qdata q))))

  (define (push item)
    (call-with-mutex protect
      (lambda ()
	(qpush q item)
	(semaphore-up block))))

  (lambda (selector)
    (case selector
      ((get pop) get)
      ((put) put)
      ((push) push)
      ((empty?) empty?)
      ((data) data)
      (($inspect) (list protect block (q '$inspect)))
      (else ???))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Sorted blocking protected queue -- a priority queue.
; Not very efficient (O(n) insert, O(1) remove), but OK for now.
;
; FIXME: 'push' does not sort!

(define (make-protected-blocking-sorted-queue less?)

  (define (insert e head tail)
    (if (null? head)
	(cons e e)
	(let loop ((l head) (prev '()))
	  (cond ((null? l)
		 (set-cdr! tail e)
		 (cons head e))
		((less? (car l) (car e))
		 (loop (cdr l) l))
		((null? prev)
		 (set-cdr! e head)
		 (cons e tail))
		(else
		 (set-cdr! prev e)
		 (set-cdr! e l)
		 (cons head tail))))))

  (make-protected-blocking-queue insert))

(provide "protectq.sch")

; eof