; q u e u e . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / June 27, 1995
; $Id: queue1.sch,v 1.3 1995/07/26 18:29:59 lth Exp lth $
;
; Basic queue package, implementing queues as closure-objects.
;
; This implementation is probably less efficient than a straight
; implementation, but it is amenable to derivation by specializing
; the insert method.
;
; There are also some methods for implementing dequeues, it's a little
; incomplete right now.

; Standard accessor methods.

; Get an element from the queue.
(define (qget q) ((q 'get)))

; Put an element onto the queue.
(define (qput q item) ((q 'put) item))

(define (qpush q item) ((q 'push) item))

(define qpop qget)

; Determine whether the queue is empty or not.
(define (qempty? q) ((q 'empty?)))

; Get a list containing the queue data.
(define (qdata q) ((q 'data)))

; Create a queue. A specialized 'insert' method which takes an element
; in a list, a head pointer, and a tail pointer, and returns a pair of
; the new head pointer and tail pointer, can be given as an extra argument.

(define (make-basic-queue . rest)
  (define head '())
  (define tail '())

  (define (insert e head tail)
    (if (null? head)
	(cons e e)
	(begin (set-cdr! tail e)
	       (cons head e))))

  (define (get)
    (if (null? head)
	(error "queue: get: no data.")
	(let ((e (car head)))
	  (set! head (cdr head))
	  (if (null? head)
	      (set! tail head))
	  e)))

  (define (put item)
    (let ((e (list item)))
      (let ((v (insert e head tail))) 
	(set! head (car v))
	(set! tail (cdr v)))))

  ; FIXME: there is no magic 'insert' method here. It doesn't make sense.

  (define (push item)
    (set! head (cons item head))
    (if (null? tail)
	(set! tail head)))

  (define (empty?)
    (null? head))

  (define (data)
    (append head '()))

  (if (not (null? rest)) (set! insert (car rest)))
  (lambda (selector)
    (case selector
      ((get pop) get)
      ((put) put)
      ((push) push)
      ((empty?) empty?)
      ((data) data)
      (($inspect) (list head tail))
      (else ???))))

(provide "queue.sch")

; eof
