
#include <stdio.h>
#include <stdlib.h>
#include "sage++user.h"

#define MAXTAGS 1000
static char *tag[MAXTAGS];
void initVariantNames(){
   for(int i = 0; i < MAXTAGS; i++) tag[i] = NULL;
#include "../h/tag.h"
}
void printVariantName(int i){
     if((i >= 0 && i < MAXTAGS) && tag[i]) printf("%s", tag[i]);
     else printf("not a known node variant");
}

void
ProjectUnparse(SgProject *project)
{
  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    printf("************** file %d in project ************\n",i);
    f->unparsestdout();
    printf("\n");
  }
}

// This routine creates a main program and inserts it in
// a blank file.
void create_a_main(SgFile *f){
  SgStatement *s = f->firstStatement();
  SgPointerType *p, *q;
  SgType *v;
  if(s == NULL) printf("no first statement\n");
  SgFuncHedrStmt *mane = new SgFuncHedrStmt("main");
  mane->AddArg("argc",*SgTypeInt());  // add int argc
  //p = new SgPointerType(*SgTypeChar());
  //q = new SgPointerType(* ((SgType *)  p));
  v = new SgPointerType(*SgTypeChar());
  q = new SgPointerType(*v);
  // note: new SgPointerType (*p) will not work, but
  // the new constructor SgPointerType(p) will work.
  //  the reason is that the default copy constructor is called
  mane->AddArg("argv", *q); // char ** argv
  s->insertStmtAfter(*mane);
};

void create_a_fun(SgFile *f){
  SgStatement *s = f->firstStatement();
  SgType *p, *q;
  SgExpression *e;
  SgSymbol *sym;
  if(s == NULL) printf("no first statement\n");
  SgFuncHedrStmt *fun = new SgFuncHedrStmt("foo");

  // create the argument: int x
  // fun->AddArg("x", *SgTypeInt());

  // create the argument: const int * x;
  //p = new SgDescriptType(*SgTypeInt(), BIT_CONST);
  //p = new SgPointerType(*p);
  //fun->AddArg("x", *p);

  // create the argument:  int * const x;
  //SgPointerType *z = new SgPointerType(*SgTypeInt());
  //z->setModifierFlag(BIT_CONST);
  //fun->AddArg("x", *z);

  // create the arg:  int & (*f)()
  //q = new SgReferenceType(*SgTypeInt()); 
  //p = new SgFunctionType(*q);
  //q = new SgPointerType(*p);
  //fun->AddArg("f",*q);  

  // create the arg: char * x[][10]
  //q = new SgPointerType(*SgTypeChar());
  //SgArrayType *art = new SgArrayType(*q);
  //art->addDimension(NULL);
  //SgValueExp * tenvalexp = new SgValueExp(10);
  //art->addDimension(tenvalexp);
  //fun->AddArg("x",*art);  

  // create the arg: float & (*x)[15]
  //q = new SgReferenceType(*SgTypeFloat());
  //SgArrayType *ar = new SgArrayType(*q);
  //SgValueExp * ftvalexp = new SgValueExp(15);
  //ar->addDimension(ftvalexp);
  //q = new SgPointerType(*ar);
  //fun->AddArg("x", *q);

  // create the arg: int & (*f[][100])()
  // a 2-D array of function pointers that return references to ints.
  //q = new SgReferenceType(*SgTypeInt()); 
  //p = new SgFunctionType(*q);
  //q = new SgPointerType(*p);
  //SgArrayType *ar = new SgArrayType(*q);
  //ar->addDimension(NULL);
  //SgValueExp * hvalexp = new SgValueExp(100);
  //ar->addDimension(hvalexp);
  //fun->AddArg("f",*ar);  // add int argc

  // create the arg: int & (*f)( int &, char *)
  // a pointer to a function that returns a reference to an int
  // and takes a reference to an int and a pointer to a char as
  // parameters.
  //q = new SgReferenceType(*SgTypeInt()); 
  //p = new SgFunctionType(*q);
  //q = new SgPointerType(*p);
  //SgExpression *exp = fun->AddArg("f",*q);
  //// find the function_op node.
  //while(exp && (exp->variant() != FUNCTION_OP)) exp = exp->lhs();
  //SgFuncPntrExp *fptr = isSgFuncPntrExp(exp);
  //// find the symbol generated by AddArg
  //e = exp; 
  //while(e && !isSgArrayRefExp(e) && !isSgVarRefExp(e)) e = e->lhs();
  //SgSymbol *fname = e->symbol();
  //q = new SgReferenceType(*SgTypeInt());
  //fptr->AddArg(fname, " ", *q);
  //q = new SgPointerType(*SgTypeChar());
  //fptr->AddArg(fname, " ", *q);

  // now add local variables.

  // int *z;  - a pointer to an int.
  p = new SgPointerType(*SgTypeInt());
  sym = new SgVariableSymb("z", *p, *fun);
  sym->declareTheSymbol(*fun);

  // char & (*y)[3];  - a pointer to an array of char refs.
  q = new SgReferenceType(*SgTypeChar());
  SgArrayType *ar = new SgArrayType(*q);
  SgValueExp * tvalexp = new SgValueExp(3);
  ar->addDimension(tvalexp);
  q = new SgPointerType(*ar);
  sym = new SgVariableSymb("y", *q, *fun);
  sym->declareTheSymbol(*fun);

  // int w[] = {1,2};  - a initialized array.
  ar = new SgArrayType(*SgTypeInt());
  ar->addDimension(NULL);
  sym = new SgVariableSymb("w", *ar, *fun);
  SgVarDeclStmt *decl = sym->makeVarDeclStmt();
  SgExpression * one = new SgValueExp(1);
  SgExpression * two = new SgValueExp(2);
  SgExprListExp * ls = new SgExprListExp(*one);
  ls->append(*two);
  SgExpression * in_ls = new SgInitListExp(*ls);
  decl->setInitialValue(0, *in_ls);
  fun->insertStmtAfter(*decl);

  // char f(int);  - a prototype or (MEMBER_FUNC)
  SgFunctionSymb *proto = new SgFunctionSymb(FUNCTION_NAME, "f", *SgTypeChar(), 
                 *fun);
  SgFunctionRefExp *fref = new SgFunctionRefExp(*proto);
  fref->AddArg("", *SgTypeInt());
  ls = new SgExprListExp(*fref);
  SgStatement *prot_st = new SgVarDeclStmt(*ls, *SgTypeChar());
  fun->insertStmtAfter(*prot_st);
   
  // create the decl: const int * M;
  p = new SgDescriptType(*SgTypeInt(), BIT_CONST);
  p = new SgPointerType(*p);
  sym = new SgVariableSymb("M", *p,  *fun);
  sym->declareTheSymbol(*fun);

  //   int x[][2]={1,2,3,4};
  //SgArrayType *ar_type = new SgArrayType(*SgTypeInt());
  //ar_type->addDimension(NULL);
  //SgValueExp * texp = new SgValueExp(2);
  //ar_type->addDimension(texp);
  //SgSymbol * xsymb =new SgVariableSymb("x", *ar_type , *fun);
  //SgExpression *xref = SgMakeDeclExp(xsymb, ar_type );
  //SgExpression * one = new SgValueExp(1);
  //SgExpression * two = new SgValueExp(2);
  //SgExpression * three = new SgValueExp(3);
  //SgExpression * four2 = new SgValueExp(4);
  //SgExprListExp * ls = new SgExprListExp(*one);
  //ls->append(*two);
  //ls->append(*three);
  //ls->append(*four2);
  //SgExpression * in_ls = new SgInitListExp(*ls);
  //SgExpression & aexp = SgAssignOp(*xref, *in_ls);
  //SgExpression * as_ex = & aexp;
  //SgExpression * as_ex_ls = new SgExprListExp(*as_ex);
  //SgStatement * ar_in = new SgVarDeclStmt(*as_ex_ls,*SgTypeInt());
  //fun->insertStmtAfter(*ar_in);

   s->insertStmtAfter(*fun);

};

// This routine creates a class declaration and a variable
// of that class is declared.
// main() {
//    class NewClass;
//    NewClass x(1,2);
//
void create_class(SgFile *f){
   SgStatement *s = f->firstStatement();

   // create a statement of the form
   // class NewClass;
   // and insert it in the top of the program.
   SgSymbol *classname = new SgSymbol(CLASS_NAME, "NewClass");
   SgClassStmt *class_decl = new SgClassStmt(*classname);
   s->insertStmtAfter(*class_decl);

   // now declare a variable of that type as
   // NewClass x;
   SgDerivedType *t = new SgDerivedType(*classname);
   SgSymbol *x = new SgVariableSymb("x", *t);
   SgVarDeclStmt *decl = x->makeVarDeclStmt();
   class_decl->insertStmtAfter(*decl,*s);

   // now modifiy this declaration to 
   // a class object with a constructor as
   // NewClass x(1,2);
   // make the argument list
   SgExpression *e1 = new SgValueExp(1);       // 1
   SgExpression *e2 = new SgValueExp(2);       // 2
   SgExpression *e = new SgExprListExp(*e2);   // (2)
   SgExpression & elopex = SgExprListOp(*e1, *e);
   SgExpression *args = & elopex; // (1, 2)
   SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
   classinit->setRhs(*args); 
   SgExpression *x_ref = new SgVarRefExp(*x);
   classinit->setLhs(*x_ref);
   SgExpression *list = decl->expr(0);   // the list of vars ... x
   list->setLhs(*classinit);
   // 
   // creating a declaration with an initialization.
   //  in this example we take the type from some function
   //  and use that to create another symbol and we will
   //  make a declaration of the form
   //  T  y = fun(0);
   //
   // first we need a function.  let's make one returning type NewClass.
   SgSymbol * fun_symb = new SgFunctionSymb(PROCEDURE_NAME,"fun",
			     *t,*(f->firstStatement()));
  SgValueExp *ev = new SgValueExp(0);
  SgExpression *e_list = new SgExprListExp(*ev);
  SgFunctionCallExp *call = new SgFunctionCallExp(*fun_symb, *e_list);
  // call now points to "fun(0)"
  // now lets create the symbol of same type as the function
  SgSymbol *y = new SgVariableSymb("y", *t);
  SgExpression *y_ref = new SgVarRefExp(*y);
  SgExpression & asopex = SgAssignOp(*y_ref, *call);  
  SgExpression *y_init =  & asopex;  
  SgExpression *y_list = new SgExprListExp(*y_init);   
  SgVarDeclStmt *newdecl = new SgVarDeclStmt(*y_list);
  decl->insertStmtAfter(*newdecl);
}	
	


int main(int argc, char **argv){

  SgProject project("x.proj");
  SgFile file(CSrc, "y.c");  // create a new file
  int c;

  initVariantNames();

  switch (file.languageType()) {
  case CSrc:
        //create_a_fun(&file);
        create_class(&file);
        ProjectUnparse(&project);
        file.saveDepFile("y.dep");
	break;
  case ForSrc:
	printf("only works for C++ programs.\n");
	break;
  }
  return 0;
}
