
#include <stdio.h>
#include <stdlib.h>
#include "sage++user.h"

#define MAXTAGS 1000
static char *tag[MAXTAGS];
void initVariantNames(){
   for(int i = 0; i < MAXTAGS; i++) tag[i] = NULL;
#include "../h/tag.h"
}
void printVariantName(int i){
     if((i >= 0 && i < MAXTAGS) && tag[i]) printf("%s", tag[i]);
     else printf("not a known node variant");
}

void
ProjectUnparse(SgProject *project)
{
  for (int i = 0; i < project->numberOfFiles(); i++) {
    SgFile *f;
    
    f = &(project->file(i));
    printf("************** file %d in project ************\n",i);
    f->unparsestdout();
    printf("\n");
  }
}


void create_a_template_function(SgFile *f){
  SgStatement *s = f->firstStatement();
  SgExpression *e;
  SgStatement *ret;
  // first let's create a template node and insert it.
  // this will be
  //template <class C  > C  foo(C x){
  //     return x;
  //}
  SgTemplateStmt *templ = new SgTemplateStmt(NULL);
  s->insertStmtAfter(*templ);
  // now add some formal parameters
  SgSymbol *classname = new SgSymbol(CLASS_NAME, "C",*templ);
  SgDerivedType *t = new SgDerivedType(*classname);
  templ->AddArg("", *t); // this will be a class type parameter
  // now add a a function in the scope of the template.
  SgFuncHedrStmt *fun = new SgFuncHedrStmt("foo");
  templ->addFunction(*fun);
  fun->setReturnedType(*t);
  e = fun->AddArg("x", *t);
  SgExpression &xref = e->copy();
  ret = new SgReturnStmt(xref);
  fun->insertStmtAfter(*ret, *fun);

  //now let's create a template class
  //template <class D  > class  TmplClass {
  // public:
  //       D x;
  // private:
  //       inline D f(int , D );
  // public:
  //        TmplClass();
  // protected:
  //       TmplClass operator +(int );
  // };

  SgTemplateStmt *templ2 = new SgTemplateStmt(NULL);
  templ->insertStmtAfter(*templ2);
  SgSymbol *classname2 = new SgSymbol(CLASS_NAME, "D", *templ2);
  SgDerivedType *D = new SgDerivedType(*classname2);
  templ2->AddArg("", *D); // this will be a class type parameter
  SgSymbol *TmplClass = new SgSymbol(CLASS_NAME, "TmplClass");
  SgClassStmt *class_decl = new SgClassStmt(*TmplClass);
  templ2->addClass(*class_decl);

  // now add a data member
  // public:
  //       D x;
  //
   SgSymbol *x = new SgFieldSymb("x", *D, *TmplClass);
   SgVarDeclStmt *decl = x->makeVarDeclStmt();
   class_decl->insertStmtAfter(*decl,*class_decl);

  // now add an inline virtual member function
  // private:
  //       inline D f(int , D );
  //
   SgDescriptType *des = new SgDescriptType(*D, BIT_VIRTUAL);
   SgSymbol *mf = new SgMemberFuncSymb("f", *des, *class_decl, PRIVATE_FIELD);
   SgFunctionRefExp *fref = new SgFunctionRefExp(*mf);
   fref->AddArg("", *SgTypeInt());
   fref->AddArg("", *D);
   SgExpression *ls = new SgExprListExp(*fref);
   SgStatement *decl2 = new SgVarDeclStmt(*ls, *D);
   decl2->addDeclSpec(BIT_INLINE);
   decl->insertStmtAfter(*decl2,*class_decl);

  // now add a constructor - constructors are ordinary functions with
  // an implicit type : DEFAULT
  // public:
  //        TmplClass();
  //
   SgSymbol *mf2 = new SgMemberFuncSymb("TmplClass",*SgTypeDefault(), 
                                         *class_decl, PUBLIC_FIELD);
   SgFunctionRefExp *fref2 = new SgFunctionRefExp(*mf2);
   SgExpression *ls2 = new SgExprListExp(*fref2);
   SgStatement *decl3 = new SgVarDeclStmt(*ls2,*SgTypeDefault() ); 
   decl2->insertStmtAfter(*decl3,*class_decl);

  // now add an overloaded operator: TmplClass operator +(int)
  // protected:
  //       TmplClass operator +(int );
  //
   SgDerivedType *tmplclass= new SgDerivedType(*TmplClass);
   SgSymbol *mf3 = new SgMemberFuncSymb("+",*tmplclass, 
                                         *class_decl, PROTECTED_FIELD);
   SgFunctionRefExp *fref3 = new SgFunctionRefExp(*mf3);
   fref3->AddArg("", *SgTypeInt());
   SgExpression *ls3 = new SgExprListExp(*fref3);
   SgStatement *decl4 = new SgVarDeclStmt(*ls3,*tmplclass); 
   decl3->insertStmtAfter(*decl4,*class_decl);

  // now add the difinition of the inline function f
  // should be
  // template <class E> inline E TmplClass<E>::f(int i, E x){
  //          foo<E>(33);
  // }
   SgTemplateStmt *templ3 = new SgTemplateStmt(NULL);
   templ2->insertStmtAfter(*templ3);
   SgSymbol *E = new SgSymbol(CLASS_NAME, "E",*templ);
   SgDerivedType *Etype = new SgDerivedType(*E);
   templ3->AddArg("", *Etype); // this will be a class type parameter
  // now add a function in the scope of the template.
   SgFuncHedrStmt *fun2 = new SgFuncHedrStmt(*mf);
   fun2->addDeclSpec(BIT_INLINE);
   templ3->addFunction(*fun2);
   // to set the return type of a template function we do
   // NOT use fun2->setReturnedType(*Etype);
   // this will modify the return type of the symbol for the function
   // which will screw up the type in the template class.  for
   // this type of template we must modify the type of the 0-th expression
   // in the function header.
   SgExpression *decl_exp = fun2->expr(0);
   decl_exp->setType(Etype);
   fun2->AddArg("i", *SgTypeInt());
   fun2->AddArg("x", *Etype);
  // the hard part is to add the scope parameter list.
  // it goes on the third expression field of the function header.
   SgExpression *tref = new SgTypeRefExp(*Etype);
   SgExpression *targ = new SgExprListExp(*tref);
   fun2->setExpression(2, *targ);
   // now add the foo<E>(33);  statement.
   // first make a function call
   SgExpression *list = new SgExprListExp(SgValueExp(33));
   SgExpression *funcall = new SgFunctionCallExp(fun->name(), *list);
   SgReturnStmt *ret2 = new SgReturnStmt(*funcall);
   // the parameter for the template arg list is
   SgExpression *targ2 = new SgExprListExp(tref->copy());
   funcall->setRhs(targ2);
   fun2->insertStmtAfter(*ret2, *fun2);

   // now add a declaration of an object of the form
   // TmplClass< int > X(100);
   //
   SgDerivedTemplateType *derttype = 
            new SgDerivedTemplateType(NULL, TmplClass);
   SgTypeRefExp *tref2 = new SgTypeRefExp(*SgTypeInt());
   derttype->addArg(tref2);
   SgSymbol *Obj = new SgSymbol(VARIABLE_NAME, "C", derttype, s );
   SgVarDeclStmt *objdecl = Obj->makeVarDeclStmt();
   // make the class constructor
   templ3->insertStmtAfter(*objdecl,*s);
   // now modifiy this declaration to 
   // a class object with a constructor as
   // NewClass C(100);
   // make the argument list
   SgExpression *e100 = new SgValueExp(100);       // 100
   SgExpression *el100 = new SgExprListExp(*e100);   // (100)
   SgExpression *classinit = new SgExpression(CLASSINIT_OP); 
   classinit->setRhs(*el100); 
   SgExpression *c_ref = new SgVarRefExp(*Obj);
   classinit->setLhs(*c_ref);
   SgExpression *elist = objdecl->expr(0);   
   elist->setLhs(*classinit);
  
}
 

int main(int argc, char **argv){

  SgProject project("x.proj");
  SgFile file(CSrc, "y.c");  // create a new file
  int c;

  initVariantNames();

  switch (file.languageType()) {
  case CSrc:
        create_a_template_function(&file);
        ProjectUnparse(&project);
        file.saveDepFile("y.dep");
	break;
  case ForSrc:
	printf("only works for C++ programs.\n");
	break;
  }
  return 0;
}
