#include  "std.h"
#include  <stdio.h>
#include  <strings.h>
#include  "xstdio.h"
#include  "structs.h"
#include  "strs.h"
#include  <sys/types.h>
#include  <sys/socket.h>
#include  <X11/Xlib.h>
#include  <X11/Intrinsic.h>
#include  <X11/Xutil.h>

#include  "Xm/Xm.h"
#include  "Xm/CascadeB.h"
#include  "Xm/FileSB.h"
#include  "Xm/Form.h"
#include  "Xm/Frame.h"
#include  "Xm/MainW.h"
#include  "Xm/PushB.h"
#include  "Xm/RowColumn.h"
#include  "Xm/ScrollBar.h"
#include  "Xm/ScrolledW.h"
#include  "Xm/SelectioB.h"
#include  "Xm/DrawingA.h"
#include  "Xm/Text.h"
#include  "list.h"

#define MAX_PROC 256
#define MAXLINES 2000
#define MAX_COLOR 8
#define MARGIN 5
#define MAXLINESIZE 300
#define MAX_ARGS 20
#define MAX_FILE_NAME 32
#define CANVAS_WIDTH 1000
#define CANVAS_HEIGHT 5000
/***************************************************************************
 * main_window --> form - > frame --> swindow --> canvas --> node[]         *
 *                                                                          *
 ****************************************************************************/

typedef struct {
  char *chars[MAXLINES];
  int  length[MAXLINES];
  int  rbearing[MAXLINES];
  int  color[MAXLINES];  
  int  descent;
  XFontStruct *font;
  GC   gc;
  int  fontheight;
  int  height;
  int  nitems;
  int  top;
} text_data, *text_data_ptr;


static XtResource resources[] = {
  { XtNfont, XtCFont, XtRFontStruct, sizeof (XFontStruct *),
      XtOffset (text_data_ptr, font), XtRString, "Fixed" },
};

extern GC create_gc();
XColor  ave_color[MAX_COLOR];
XtAppContext         app_context;
Widget                app_shell;
Widget                main_window;
Widget                  form;
Widget         command, warn, frame[2];
Widget                        swindow[2];
Widget                        canvas[2];
Widget         source_select_box;
Widget         src_name;   /* Text Widget */



GC GCanvasGC[2];   /* [0] => main window, [1] => drawing canvas */
XGCValues GValuesGC;

XtInputId fileid;

int bpoint_summary[MAX_PROC];

typedef LINKED_LIST IPD_LIST;
typedef LINKED_LIST FILE_SPEC_LIST;

typedef struct file_spec_object {
  char *filename;
  IPD_LIST *ipl;
} FILE_SPEC_OBJ;

typedef struct ipd_object {
  int  line;        /* which line in the file */
  int  color;       /* how should we color it */
  int  bpoint;      /* is breakpoint before or after the line */
  int  proc[MAX_PROC];   /* processes waiting in that breakpoint */
  char *pset;
} IPD_OBJ;

text_data data, hdata;  
FILE_SPEC_LIST *GFileToShow = NULL;
FILE *fp_log = NULL;
static FILE *fp_break_summ = NULL;
FILE_SPEC_OBJ *allocate_file_spec_object(char *fname)
{
  FILE_SPEC_OBJ *fso = (FILE_SPEC_OBJ *) malloc (sizeof(FILE_SPEC_OBJ));
  fso->filename = (char *) malloc (strlen(fname)+1);
  sprintf(fso->filename,"%s",fname);
  fso->ipl = NULL;
  return fso;
}

void add_line_list_to_file_spec_object (FILE_SPEC_OBJ *fso, LINKED_LIST *ll)
{
  fso->ipl = ll;
}


int get_bpoint_summary(int pid)
{
  return bpoint_summary[pid];
}

IPD_OBJ *allocate_ipd_object()
{
  int i;
  IPD_OBJ *ipd = (IPD_OBJ *) malloc (sizeof (IPD_OBJ));
  ipd->line = ipd->color = 0;
  ipd->pset = NULL;
  for (i=0; i<MAX_PROC; i++)
    ipd->proc[i] = -1;
  return ipd;
}

IPD_OBJ *get_element_from_ipd_list (IPD_LIST *ipdl, int n)
{
  IPD_OBJ *ipd =
    (IPD_OBJ *) get_element_from_list ((LINKED_LIST *)ipdl, n);
  return ipd;
}


/*
 *******************************************************************
 * FUNCTION:     compute_color_partition()                         *
 * DESCRIPTION:  Given a list of elements, partition them along    *
 *               color domain using some atribute value. Currently *
 *               uses the line number attribute for doing so.      *
 * NOTES:        Same functionality is needed for scatter plot.    *
 *               should rewrite this as a general purpose routine  *
 *******************************************************************
*/

void compute_color_partition (IPD_LIST *ipl)
{
  IPD_OBJ *tipo;
  IPD_LIST *tipl = ipl;
  int size, color, i;

  if (tipl == NULL)
    return;
  size = get_size_of_list ((LINKED_LIST *) tipl);
  color = 0;
  for (i=0; i<size; i++) {
    tipo = get_element_from_ipd_list (tipl, i);
  
    /* get the first proc */
    tipo->bpoint = get_bpoint_summary(tipo->proc[0]);

    if (tipo->color == 0) {
      IPD_LIST *nipl = tipl->next;
      tipo->color = ++color;
      while (nipl) {
	IPD_OBJ *nipo = get_element_from_ipd_list (nipl, 0);
	if (nipo->line == tipo->line)
	  nipo->color = tipo->color;
	nipl = nipl->next;
      }
    }
  }
}

char *get_file_text (XmString compound)
{
  caddr_t context;
  char *text;
  XmStringCharSet charset;
  XmStringDirection dir;
  Boolean separator;
  char *buf;
  
  XmStringInitContext (&context, compound);
  XmStringGetNextSegment(context, &text, &charset, &dir, &separator);
  buf = (char *) malloc (strlen(text)+2);
  sprintf(buf,"%s", text);
  XtFree(text);
  XmStringFreeContext (context);
  return buf;
}

/********************************
 * QuitCB: Quit the application.*
 ********************************/

void FileQuitCB (Widget w, caddr_t client_data, caddr_t call_data)
{
  XtRemoveInput(fileid);
  XtUnrealizeWidget (app_shell); 
  XtDestroyWidget (app_shell); 
}


/******************************************************************
 * FUNCTION:         handle_exposures()                           *
 * DESCRIPTION:      handles exposure events for both windows     *
 * DESIGN:                                                        *
 ******************************************************************
*/

void handle_exposures 
(Widget w, text_data *data, XmDrawingAreaCallbackStruct *cb)
{
  int yloc = 0, index = data->top;
  Region region;

  region = XCreateRegion();
  XtAddExposureToRegion(cb->event, region);
  
  XSetRegion (XtDisplay(w), data->gc, region);
  XSetForeground 
    (XtDisplay(w), data->gc, ave_color[7].pixel);
  while (index < data->nitems && yloc < data->height) {
    yloc += data->fontheight;
    if (XRectInRegion(region, MARGIN, yloc-data->fontheight,
		      data->rbearing[index], 
		      data->fontheight) != RectangleOut) {
      XSetBackground
	(XtDisplay(w), data->gc, ave_color[data->color[index]].pixel);
      XDrawImageString (XtDisplay(w), XtWindow (w), data->gc,
			MARGIN, yloc, data->chars[index],
			data->length[index]);
    }
    index++;
  }

  XDestroyRegion(region);
}


/*
 ***********************************************************************
 * FUNCTION:         add_identify_file_spec_list()                     *
 * DESCRIPTION:      given a filename, checks whether an entry already *
 *                   exists for that file. If the entry exists then it *
 *                   returns a pointer to the entry, otherwise it      *
 *                   creates the entry before returning the pointer.   *
 ***********************************************************************
*/

FILE_SPEC_OBJ *identify_file_spec_list 
  (FILE_SPEC_LIST **fsl, char *fname, int add)
{
  FILE_SPEC_LIST *tmp = *fsl;
  FILE_SPEC_OBJ  *fso = NULL;
  int size = get_size_of_list ((LINKED_LIST *) tmp);
  int i;
  for (i=0; i<size; i++) {
    fso = (FILE_SPEC_OBJ *) get_element_from_list((LINKED_LIST *) tmp, i);
    if (!StrCmp(fso->filename, fname))
      return fso;
  }
  /* We did not find any entry for the file "fname" */
  if (add) {
    fso = allocate_file_spec_object (fname);
    add_node_to_list ((LINKED_LIST **) fsl, (void *) fso);
    return fso;
  }
  else 
    return NULL;
}
  

void add_node_to_ipd_list (IPD_LIST **ipl, IPD_OBJ *ipo)
{
  add_node_to_list ((LINKED_LIST **)ipl, (void *) ipo);
}


void read_break_summary()
{
  int i = 0, maxid = 0, tmp = 0;
  if (fp_break_summ == NULL)
    fp_break_summ = fopen("ave.break.summary", "r");

  if (fp_break_summ == NULL) {
    printf("Ave expects a file ave.break.summary to be present\n");
    printf("Assuming no breakpoint has been set yet\n");
  }
  else
    fscanf(fp_break_summ, "%d", &maxid);

  while (i < maxid) {
    fscanf(fp_break_summ,"%d %d\n", &i, &tmp);
    bpoint_summary[i] = tmp;
  }
}

void load_file_data (text_data *data, text_data *hdata, char *fname)
{
  IPD_LIST *tipl;
  FILE_SPEC_OBJ *fso;
  int i, dir, ascent, desc;
  XCharStruct char_info;
  FILE *fp_src;
  char buf[MAXLINESIZE];
  int semi_colon_flag = 0;

  /* if no name is specified, load the first file */
  if (fname)
    fso = identify_file_spec_list(&GFileToShow, fname, 0);
  else 
    fso = (FILE_SPEC_OBJ *) get_element_from_list ((LINKED_LIST *) GFileToShow, 0);


  /* initialize color for each line in data & hdata */
  for (i=0; i<MAXLINES; i++)
    data->color[i] = hdata->color[i] = 0;


  /* Handle hdata */
  i = 0;
  hdata->fontheight = 0;
  tipl = (IPD_LIST *) fso->ipl;
  while (tipl) {
    IPD_OBJ *ipo = (IPD_OBJ *) tipl->contents;
    hdata->chars[i] = XtMalloc (strlen(ipo->pset)+ 10); /* Magic Number */
    if (ipo->bpoint == BREAK_AFTER_EVENT) 
      sprintf(hdata->chars[i], "%s After",  ipo->pset);
    else
      sprintf(hdata->chars[i], "%s Before",  ipo->pset);

    hdata->length[i] = strlen(hdata->chars[i]);
    XTextExtents (hdata->font, hdata->chars[i], hdata->length[i], &dir,
		  &ascent, &desc, &char_info);
    hdata->rbearing[i] = char_info.rbearing;
    hdata->color[i] = ipo->color;
    hdata->descent = desc;
    hdata->fontheight = ascent + desc;
    i++;
    tipl = tipl->next;
  }
  hdata->nitems = i;
  hdata->top = 0;


  /* Handle data */

  /* fso->ipl->contents should have the line --> color mapping available */
  tipl = (IPD_LIST *) fso->ipl;
  while (tipl) {
    IPD_OBJ *tipo = (IPD_OBJ *) tipl->contents;
    data->color[tipo->line-1] = tipo->color;
    tipl = tipl->next;
  }

  if ((fp_src=fopen(fso->filename,"r")) == NULL) {
    fprintf(stderr, "Unable to open source file %s\n", fso->filename);
    exit(1);
  }

  /* Handle the src_name */
  XmTextSetString(src_name, " ");
  XmTextSetString(src_name, fso->filename);


  i = 0;
  while ((fgets(buf, MAXLINESIZE, fp_src)) != NULL &&
	 i < MAXLINES) {
    
    int current = i;
    data->chars[i] = XtMalloc (strlen(buf) + 1);
    buf[strlen (buf) - 1] = '\0';
    strcpy (data->chars[i], buf);
    data->length[i] = strlen (data->chars[i]);
    XTextExtents (data->font, data->chars[i], data->length[i], &dir,
		  &ascent, &desc, &char_info);
    data->rbearing[i] = char_info.rbearing;
    data->descent = desc;
    data->fontheight = ascent + desc;
    /* if I do not end in semicolon raise a flag */
    if ((index(buf,';') == NULL) && (strstr (buf, "*/") == NULL))
      semi_colon_flag++;
    else {
      if (data->color[current] != 0) {
	int j;
	for (j=1; j<=semi_colon_flag; j++)
	  data->color[current-j] = data->color[current];
      }
      semi_colon_flag = 0;
    }
    if ((data->length[i] <2) && (data->color[current] != 0)) {
      while(data->length[current] < 2)
	--current;
      data->color[current] = data->color[i];
    }
    i++;
    data->top = 0;
  }
  data->nitems = i;
  data->top = 0;



  fclose (fp_src);

}

void select_regular_file (Widget parent, caddr_t client_data, caddr_t call_data)
{

  int i, n;
  Widget regular_file_select_box;
  Widget regular_file_ok;
  String name = (String) "Select File";
  Arg args[20];
  

  
  n = 0;
  regular_file_select_box = 
    XmCreateFileSelectionDialog(parent, name, args, n);


  XtManageChild (regular_file_select_box);
  return;
  
  /* OK should be handled properly, if a directory show the directory,
     otherwise show the file */

}

    
/*      
 ************************************************************************
 * FUNCTION:      process_ipd_log()                                     *
 * DESCRIPTION:   Reads the file "ave.log" written by the "ipd" session *
 *                and fills the "ipd_list" with the following:          *
 *                  + Files to be displayed                             *
 *                  + Lines to be highlighted                           *
 *                Calls for color binning                               *
 ************************************************************************
*/

void process_ipd_log (XtPointer client_data, int fid, XtInputId *id)
{

  FILE_SPEC_LIST *fsl = NULL, *tfsl = NULL;
  FILE_SPEC_OBJ  *fso = NULL;
  IPD_OBJ  *ipo = NULL;
  char fname[MAX_FILE_NAME];  
  char pname[MAX_PROC_DIGIT]; 
  char line[MAX_PROC_DIGIT];
  char buf[MAXLINESIZE];
  int i = 0;
  int cnt = -1;
  char *x = NULL;
  int frame = 0;
  static int sequence = 0;
  FILE *myfp =  fopen("ave.log", "r");
  
  /* First read the log file produced by the ipd session */
  while (frame <= sequence) {
    if ((fgets(buf,MAXLINESIZE, myfp)) != NULL) {
      buf[strlen(buf)-1] = '\0';
      if (strstr (buf, "frame") != NULL)
	frame++;
    }
    else {
      fclose(myfp);
      return;
    }
  }

  /* if we come here, then frame > sequence */
  fsl = NULL;
  GFileToShow = NULL;

  while ((fgets (buf, MAXLINESIZE, myfp)) != NULL) {
    buf[strlen(buf)-1] = '\0';

    /*****************************************************
     * The format of the lines we are dealing with:      *
     *   **** (0,1,2,:0) ****                            *
     *   aa_csend()    [avelib.c{}#142]                  *
     *   main()    [mdict.c{}#69]                        *
     *****************************************************/


    /* Handle first kind of lines that gives us process numbers */
    
    /* Line contains both '*' and '(' */
    if ( (index(buf, '*') != NULL) && 
	((x = index(buf, '(')) != NULL) ) {
      i = 0;
      ipo = allocate_ipd_object();
      ipo->pset = (char *) malloc (strlen(buf)+1);
      strcpy(ipo->pset, buf);



      /* Until we see the demarkator ':', we read comma separated
	 list of process identifiers */
      x++;
      i = 0;
      cnt = 0;
      while (1) {
	if ((*x == ',') || (*x == ':')) {
	  pname[i] = '\0';
	  ipo->proc[cnt] = atoi(pname);
	  if (*x == ':')
	    break;
	  else {
	    i = 0;
	    ++cnt;
	    x++;
	  }
	}
	else {
	 pname[i] = *x;
	 x++;
	 i++;
       }
      }
    }


    /* Now check to find the line containing "main()". "ipd" has 
       already been allocated. So we do not need to allocate one */

    if (strstr(buf, "main()") != NULL) {
      x = index(buf,'[');
      x++;
      i = 0;
      while ((*x) != '{') {
	fname[i] = *x;
	x++;
	i++;
      }
      fname[i] = '\0';

      fso = identify_file_spec_list (&fsl, fname, 1);
      x = index (buf, '#');
      i = 0;
      x++;
      while ((*x) != ']') {
	line[i] = *x;
	x++;
	i++;
      }
      line[i] = '\0';
      ipo->line = atoi(line);
      add_node_to_ipd_list (&(fso->ipl), ipo); 
    }
  }

  if (fsl)
    GFileToShow = fsl;
  read_break_summary();
  tfsl = fsl;
  while (tfsl) {
    FILE_SPEC_OBJ *fso = get_element_from_list ((LINKED_LIST *) tfsl, 0);
    IPD_LIST *ipl = fso->ipl;
    compute_color_partition(ipl);
    tfsl = tfsl->next;
  }

  if (frame > sequence) {
    sequence = frame;
    load_file_data (&data, &hdata, (char *) NULL);
  }
  fclose (myfp);
}



void source_select_box_cancel()
{
  XtUnmanageChild (source_select_box);
  XtDestroyWidget (source_select_box);
}

void select_file (Widget w, caddr_t client_data, 
		  XmSelectionBoxCallbackStruct *call_data)
{
  char* filename = get_file_text (call_data->value);
  load_file_data (&data, &hdata, filename);
  
}



void SelectFile (Widget w, caddr_t client_data, caddr_t call_data)
{
  int n;
  Widget select_apply, select_ok, select_cancel, select_list;
  FILE_SPEC_LIST *fsl = GFileToShow;
  char *title = "File Selection";
  Arg args[MAX_ARGS];

  /* No source files selected => no selection box created */
  if (fsl == NULL)
    return;

  n = 0;
  source_select_box = 
    XmCreateSelectionDialog (w, XmStringCreateSimple (title), args, n); 

  select_apply = 
    XmSelectionBoxGetChild (source_select_box, XmDIALOG_APPLY_BUTTON);
  XtAddCallback (source_select_box, XmNapplyCallback, select_file, GFileToShow);

  select_cancel = 
    XmSelectionBoxGetChild (source_select_box, XmDIALOG_CANCEL_BUTTON);
  XtAddCallback (source_select_box, XmNcancelCallback, 
		 source_select_box_cancel, NULL);

  select_ok = XmSelectionBoxGetChild (source_select_box, XmDIALOG_OK_BUTTON);
  XtAddCallback (source_select_box, XmNokCallback, select_file, GFileToShow);

  select_list = XmSelectionBoxGetChild (source_select_box, XmDIALOG_LIST);
  
  while (fsl) {
    FILE_SPEC_OBJ *fso = 
      get_element_from_list( (LINKED_LIST *) fsl, 0 );
    XmListAddItem (select_list, XmStringCreateSimple(fso->filename), 0);
    fsl = fsl->next;
  }
  
  XtManageChild (source_select_box);
  return;
}



create_file_window(Widget parent)
{
  Display *display;

  Widget  menu_bar, menu_pane, cascade; 
  Widget  button;
  Arg args[MAX_ARGS];
  int n;
  int i;
  fp_log = fopen("ave.log", "r");
  if (fp_log == NULL) {
    printf("You need to have a file called ave.log in your directory\n");
    return;
  }


  /* Create application shell */
  n = 0;
  app_shell = 
    XtAppCreateShell ("Ave: Source Window", "XMdemos", applicationShellWidgetClass,
		      XtDisplay(parent), NULL, 0);

  XtGetApplicationResources (app_shell, &data, resources, 
			     XtNumber (resources), NULL, 0);
  XtGetApplicationResources (app_shell, &hdata, resources, 
			     XtNumber (resources), NULL, 0);


  /* Create main window */
  n = 0;
  main_window = XmCreateMainWindow (app_shell, "Ave", args, n);
  XtManageChild (main_window);
  

  /* Create menubar in mainwindow */
  n = 0;
  menu_bar = XmCreateMenuBar (main_window, "menu_bar", args, n);
  XtManageChild (menu_bar);


  /* Create "Exit" PulldownMenu */
  n = 0;
  menu_pane = XmCreatePulldownMenu (menu_bar, "menu_pane", args, n);

  n = 0;
  button = XmCreatePushButton (menu_pane, "Quit", args, n);
  XtManageChild (button);
  XtAddCallback (button, XmNactivateCallback, FileQuitCB, NULL);
  
  n = 0;
  XtSetArg (args[n], XmNsubMenuId, menu_pane); n++;
  cascade = XmCreateCascadeButton (menu_bar, "Exit", args, n);
  XtManageChild (cascade);

  n = 0;
  cascade = XmCreateCascadeButton (menu_bar, "File", args, n);
  XtManageChild (cascade);
  XtAddCallback (cascade, XmNactivateCallback, SelectFile, NULL);


  n = 0;
  cascade = XmCreateCascadeButton (menu_bar, "Break", args, n);
  XtManageChild (cascade);
  XtAddCallback (cascade, XmNactivateCallback, SelectFile, NULL);

  n = 0;
  cascade = XmCreateCascadeButton (menu_bar, "Edit", args, n);
  XtManageChild (cascade);
  XtAddCallback (cascade, XmNactivateCallback, select_regular_file, NULL);

  n = 0;
  XtSetArg (args[n], XmNlistVisibleItemCount, 4); n++;
  form = XmCreateForm (main_window, "form", args, n);
  XtManageChild (form);


  /* Create Frame MainWindow and ScrolledWindow in Frame */
  n = 0;
  XtSetArg (args[n], XmNmarginWidth, 2); n++;
  XtSetArg (args[n], XmNmarginHeight, 2); n++;
  XtSetArg (args[n], XmNshadowThickness, 1); n++;
  XtSetArg (args[n], XmNshadowType, XmSHADOW_OUT); n++;
  frame[0] = XmCreateFrame (form, "frame", args, n);
  XtManageChild (frame[0]);

  /* Create text window to display filename */
  n = 0;
  XtSetArg (args[n], XmNeditable, FALSE); n++;
  src_name = XmCreateText (form, "src", args, n);
  XtManageChild (src_name);

  /* Create Frame MainWindow and ScrolledWindow in Frame */
  n = 0;
  XtSetArg (args[n], XmNmarginWidth, 2); n++;
  XtSetArg (args[n], XmNmarginHeight, 2); n++;
  XtSetArg (args[n], XmNshadowThickness, 1); n++;
  XtSetArg (args[n], XmNshadowType, XmSHADOW_OUT); n++;
  frame[1] = XmCreateFrame (form, "frame", args, n);
  XtManageChild (frame[1]);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetValues (frame[0], args, n);


  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNtopWidget, frame[0]); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetValues (src_name, args, n);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNtopWidget, src_name); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetValues (frame[1], args, n);

  n = 0;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
  XtSetArg (args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
  swindow[0] = XmCreateScrolledWindow (frame[0], "swindow", args, n);
  XtManageChild (swindow[0]);

  n = 0;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
  XtSetArg (args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
  swindow[1] = XmCreateScrolledWindow (frame[1], "swindow", args, n);
  XtManageChild (swindow[1]);


  n = 0;
  XtSetArg (args[n], XmNwidth, CANVAS_WIDTH); n++;
  XtSetArg (args[n], XmNheight, 100); n++;
  XtSetArg (args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  canvas[0] = XmCreateDrawingArea (swindow[0], "canvas", args, n);
  XtManageChild(canvas[0]);
  XtAddCallback(canvas[0], XmNexposeCallback, handle_exposures, &hdata);

  hdata.height = 100;
  hdata.gc = create_gc (canvas[0]);
  ave_load_color (canvas[0]);

  n = 0;
  XtSetArg (args[n], XmNwidth, CANVAS_WIDTH); n++;
  XtSetArg (args[n], XmNheight, CANVAS_HEIGHT); n++;
  XtSetArg (args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  canvas[1] = XmCreateDrawingArea (swindow[1], "canvas", args, n);
  XtManageChild(canvas[1]);
  XtAddCallback(canvas[1], XmNexposeCallback, handle_exposures, &data);

  data.height = CANVAS_HEIGHT;
  data.gc = create_gc (canvas[1]);

  ave_load_color (canvas[1]);

  n = 0;
  XtSetArg (args[n], XmNworkWindow, canvas[1]); n++;
  XtSetValues (swindow[1], args, n);
  n = 0;
  XtSetArg (args[n], XmNworkWindow, canvas[0]); n++;
  XtSetValues (swindow[0], args, n);

  XmMainWindowSetAreas (main_window, menu_bar, NULL, NULL, NULL, form);

  fileid = XtAppAddInput (app_context, fileno(fp_log), (XtPointer) XtInputReadMask, 
		 process_ipd_log, NULL);
  XtRealizeWidget (app_shell);

  /* init data color */
  for (i=0; i<MAXLINES; i++) 
    data.color[i] = hdata.color[i] = 0;

}
	

