#include "structs.h"
#include "default.h"
#include "abstract.h"
#define TRUE 1
#define FALSE 0

extern BAR_GRAPH_NODES *GBarGraph;

/* struct to handle mapping of attribute values to color */
/* Notes: Probably should change color1, color2 to int *color */
/* This way, one can achieve color binning at higher dimensions */

typedef struct color_database {
  int value;
  int color1;
  int color2;
  struct color_database *next;
} COLOR_DATABASE;

COLOR_DATABASE *CDB = NULL;


/*
 ************************************************************************
 * FUNCTION(S):   related to the RECT_DATA struct                       *
 *                o allocate_rect_data()                                *
 *                o free_rect_data()                                    *
 * NOTES:         Should change them to generic list.                   *
 *                typedef LINKED_LIST RECT_DATA_LIST                    *
 ************************************************************************
*/

RECT_DATA *allocate_rect_data (int bar, int row, int col, 
			       int width, int height, int color, int value)
{
  RECT_DATA *rd = (RECT_DATA *) malloc (sizeof (RECT_DATA));
  rd->bar_id = bar;
  rd->row_id = row;
  rd->col_id = col;
  rd->width  = width;
  rd->height = height;
  rd->color  = color;
  rd->value  = value;
  rd->next   = NULL;
  return rd;
}

void free_rect_data (RECT_DATA *rd)
{
  RECT_DATA *p, *q;

  p = rd;
  while (p) {
    q = p->next;
    free(p);
    p = q;
  }
  rd = NULL;
  return;
}


/*
 ************************************************************************
 * FUNCTION(S):   related to the BAR_GRAPH_NODES struct                 *
 *                o add_to_bar_graph_node()                             *
 *                o free_bar_graph_node()                               *
 * NOTES:         Should change them to generic list.                   *
 *                typedef LINKED_LIST BAR_GRAPH_NODES                   *
 ************************************************************************
*/


void add_to_bar_graph_node (ABSTRACT_EVENT *ae, ABSTRACT_EVENT_LIST *first,
			   ABSTRACT_EVENT_LIST *last, int cnt, BAR_GRAPH_NODES **bgn)
{
  BAR_GRAPH_NODES *tmp = *bgn;
  BAR_GRAPH_NODES *t_bg = (BAR_GRAPH_NODES *) malloc (sizeof(BAR_GRAPH_NODES));
  t_bg->xae = ae;
  t_bg->first = first;
  t_bg->last  = last;
  t_bg->discrete = t_bg->aggr = NULL;
  t_bg->cnt = cnt;
  t_bg->next = NULL;

  if ((*bgn) == NULL) {
    (*bgn) = t_bg;
    return;
  }

  while (tmp->next) 
    tmp = tmp->next;

  tmp->next = t_bg;
  return;
}

void free_bar_graph_node (BAR_GRAPH_NODES **bg)
{
  BAR_GRAPH_NODES *p, *q;
  p = *bg;

  while (p) {
    q = p->next;
    p->xae = NULL;
    free_abstract_event_list (p->first);
    free_abstract_event_list (p->last);
    free_rect_data (p->discrete);
    free_rect_data (p->aggr);
    p = q;
  }
  *bg = NULL;
  return;
}
    

/*
 ************************************************************************
 * FUNCTION(S):   related to the Category Statistics                    *
 *                o is_in_aggr(): Used to build histogram               *
 *                o calculate_aggr_stat(): 
 *                o fix_layout_category_statistics()                    *
 ************************************************************************
*/


int is_in_aggr (ATTRIBUTE *atr, BAR_GRAPH_NODES *bgn)
{
  RECT_DATA *rd = bgn->aggr;
  int aval = get_int(atr->val_list);

  if (rd == NULL)
    return FALSE_;

  while (rd) {

/*
    if (aval == rd->value)
      return TRUE_;
*/

    if (atr->color == rd->color)
      return TRUE_;

    rd = rd->next;
  }
  return FALSE_;
}

void calculate_aggr_stat (ABSTRACT_EVENT_LIST *first, 
			  ABSTRACT_EVENT_LIST *last,
			  char *attr_name, BAR_GRAPH_NODES *bgn)
{
  ABSTRACT_EVENT_LIST *t1, *t2;
  RECT_DATA *rd, *trd;
  ABSTRACT_EVENT *tmp;
  int cnt = 1;
  t1 = first;
 

  while (t1) {
    ATTRIBUTE *an1;
    int aval;
    tmp = get_element_from_abstract_list (t1, 0);
    an1 = get_attribute_with_name (tmp, attr_name);
    aval = get_int(an1->val_list);
    cnt = 1;
    t2 = t1->next;

    if (is_in_aggr(an1, bgn) == FALSE_) {
      while (t2) {
	ATTRIBUTE *an2;
	tmp = get_element_from_abstract_list (t2, 0);
	an2 = get_attribute_with_name (tmp, attr_name);

	if (an1->color == an2->color)
	  ++cnt;
	if (t2 == last)
	  break;
	t2 = t2->next;
      }
      rd = 
	allocate_rect_data (-1, -1, -1, 1, 
			    cnt, an1->color, aval);
      printf("allocating rectangle: color = %d, count = %d, aval = %d\n",
	     an1->color, cnt, aval);

      trd = bgn->aggr;
      if (trd == NULL) {
	rd->next = NULL;  /* why is this necessary? */
	bgn->aggr = rd;
      }
      else {
	while (trd->next)
	  trd = trd->next;
	trd->next = rd;
      }

    }
    if (t1 == last)
      break;
    t1 = t1->next;
    
  }
    
}  


void fix_layout_category_statistics (BAR_GRAPH_NODES *bgn)
{
  int bar = 0;
  int row, col;

  while (bgn) {
    RECT_DATA *rd = bgn->aggr;
    ++bar;
    row = col = 1;
    while (rd) {
      rd->bar_id = bar;
      rd->row_id = row;
      rd->col_id = col;
      rd = rd->next;
      ++col;
    }
    bgn = bgn->next;
  }
}


/*
 ********************************************************************
 * FUNCTIONS:   related to color database, and mapping of attribute *
 *              values to a color space.                            *
 ********************************************************************
*/

void free_color_database (COLOR_DATABASE **cdb)
{
  COLOR_DATABASE *p, *q;
  p = *cdb;
  while (p) {
    q = p->next;
    free(p);
    p = q;
  }
  (*cdb) = NULL;
  return;
}
  

void add_to_color_database (int color, int value, COLOR_DATABASE **cdb)
{
  COLOR_DATABASE *pcd = (*cdb);
  COLOR_DATABASE *lag = pcd;
  COLOR_DATABASE *cd = (COLOR_DATABASE *) malloc (sizeof (COLOR_DATABASE));
  cd->value = value;
  cd->next = NULL;
  cd->color1 = color;
  cd->color2 = color;
  /* if the list is null */
  if ((*cdb) == NULL) {
    (*cdb) = cd;
    return;
  }

  /* what if the first element is bigger than me */
  if (pcd->value > value) {
    cd->next = pcd;
    (*cdb) = cd;
    return;
  }
  else if ((pcd->value < value) && (pcd->next == NULL)) {
    pcd->next = cd;
    return;
  }  /* very ugly, but I have a paper deadline */

  pcd = pcd->next;
  /* otherwise */
  while (pcd) {
    if ( pcd->value < value) {
      pcd = pcd->next;
      lag = lag->next;
    }
    else {
      cd->next = pcd;
      lag->next = cd;
      return;
    }
  }
  lag->next = cd;
  return;             /* ugliest piece of code I have written */
  
}

void print_color_database (COLOR_DATABASE **cdb)
{
  COLOR_DATABASE *pcd = *cdb;
  while (pcd) {
    printf("value = %d, color1 = %d, color2 = %d\n", 
	   pcd->value, pcd->color1, pcd->color2);
    pcd = pcd->next;
  }
}

  
void scale_color_database (int scale_factor,  COLOR_DATABASE ** cdb)
{
  COLOR_DATABASE *pcd = *cdb;
  int cnt = -1;
  while (pcd) {
    ++cnt;
    pcd->color2 = cnt/scale_factor;
    pcd = pcd->next;
  }
}

int get_new_color (int clr, COLOR_DATABASE **cdb)
{
  COLOR_DATABASE *pcd = *cdb;
  while (pcd) {
    if (pcd->color1 == clr)
      return pcd->color2;
    pcd = pcd->next;
  }

  /* if here something terribly wrong */
  printf("Soft Error: Something rotten in color space\n");
  return 0; /* But graceful degradation */
}





void scale_color (ABSTRACT_EVENT_LIST *ael, char *attr_name, int scale_factor)
{
  ABSTRACT_EVENT_LIST *tmp = ael;


  while (tmp) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (tmp,0);
    ATTRIBUTE *attr_node = get_attribute_with_name (ae, attr_name);
    attr_node->color = attr_node->color/scale_factor;
    tmp = tmp->next;
  }  
}


int get_color_from_discrete (int val, RECT_DATA *rd)
{
  while (rd) {
    if (val == rd->value)
      return rd->color;
    rd = rd->next;
  }
  return 0;
}

void compute_aggr_color (BAR_GRAPH_NODES *bgn)
{
  RECT_DATA *ard;
  RECT_DATA *drd;
  RECT_DATA *trd;

  while (bgn) {
    drd = bgn->discrete;
    ard = bgn->aggr;
    
    while (ard) {
      ard->color = get_color_from_discrete(ard->value, drd);
      ard = ard->next;
    }
    bgn = bgn->next;
  }
}


void scale_color_in_rect (ABSTRACT_EVENT_LIST *ael, char *attr_name, 
			  COLOR_DATABASE **pcd)
{
  ABSTRACT_EVENT_LIST *tmp = ael;
  while (tmp) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (tmp, 0);
    ATTRIBUTE *attr_node = get_attribute_with_name (ae, attr_name);
    attr_node->color = get_new_color (attr_node->color, pcd);
    tmp = tmp->next;
  }
}


map_attribute_to_color (ABSTRACT_EVENT_LIST *clist, char *name)
{
  int csize = get_size_of_abstract_list (clist);
  int i,j;
  int clr = -1;

  for (i=0; i<csize; i++) {
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (clist, i);
    ATTRIBUTE *a = get_attribute_with_name (ae, name);

    INT_LINKED_LIST *aval = a->val_list;
    
    if (a->color == -1) {
      ++clr;
      add_to_color_database (clr, get_int(aval), &CDB);
      a->color = clr;    
    }

    for (j=i+1; j<csize; j++) {
      ABSTRACT_EVENT *cae = get_element_from_abstract_list (clist, j);
      ATTRIBUTE *c = get_attribute_with_name (cae, name);
      INT_LINKED_LIST *cval = c->val_list;

      if (list_equal (aval, cval) == TRUE)
	c->color = a->color;
    }
  }

  /* may need to scale color */
  if (clr > MAX_COLOR) {
    int scale_factor = clr/MAX_COLOR + 1;

    /* first change the color database */
    scale_color_database (scale_factor, &CDB);

    /* now scale the colors in the rect_obj */
    scale_color_in_rect (clist, name, &CDB);
  }

}

/*
 ************************************************************************
 * FUNCTION(S):   related to the Scatter/Line Plot                      *
 *                o is_in_aggr(): Used to build histogram               *
 *                o calculate_aggr_stat(): 
 *                o fix_layout_category_statistics()                    *
 ************************************************************************
*/


void fix_layout_discrete_plot (BAR_GRAPH_NODES *bgn, char *aname)
{
  int bar = 1, row = 1, col = 1;

  while (bgn) {
    int ncnt = get_max_column (bgn->cnt);
    ABSTRACT_EVENT_LIST *alist = bgn->first;
    ABSTRACT_EVENT *ae = NULL;
    int i = -1;

    while (1) {
      ATTRIBUTE *a;
      RECT_DATA *trd, *rd;

      ++i;
      ae = get_element_from_abstract_list(alist, i);
      a = get_attribute_with_name (ae, aname);
      trd = bgn->discrete;

      rd = allocate_rect_data (bar, row, col, 2, 2, a->color, 
			       get_int(a->val_list));
      /* we need to add the RECT_DATA in its proper place */
      /* would become easy once we change RECT_DATA to LINKED_LIST */
      if (bgn->discrete == NULL) {
	rd->next = bgn->discrete;
	bgn->discrete = rd;
      }
      else {
	while (trd->next)
	  trd = trd->next;
	trd->next = rd;
      }

      if (ae == get_element_from_abstract_list (bgn->last,0))
	break;
      if (col == ncnt) {
	col = 0;
	++row;
      }
      ++col;
    }

    ++bar;
    row = col = 1;
    bgn = bgn->next;
  }
}





/*
 ***********************************************************************
 * The show statement is of the form:
 *       with <select> show <attribute-of-node> wrt <someother-node>;  *
 * In displaying the attribute, we need to partition the attribute space
 * into equivalence classes. 
 * 1. <someother-node> becomes the x-axis.
 * 2. There are multiple <attribute-of-node>s under each <someother-node>
 * 3. We need to map the attributes onto a color space.
 * Let us assume there are "n" nodes that satisfy <someother-node>.
 * For simplicity, let us also assume there are "m" node that satisfy
 * the <attribute-of-node> criterion under each of these "n" nodes.
 * So we build a list of "m*n" elements, which gives us the global 
 * attribute space, which is then partitioned into equivalence classes.
 * The equivalence classes are then mapped to a color space.
 * In the real world, the behavior won't be this simple. Each of the 
 * "n" nodes will not have a constant "m" <attribute-of-node> under
 * them -- the number would vary from node to node. So we need to keep 
 * track of the individual <someother-node> boundaries using pointers.
 * In a perfect world, just keeping track of what "m" is would have 
 * sufficed. 
 * NOTE: Other way of doing it is to keep an integer array of "m"s. But
 * it consumes same space.
 ************************************************************************
*/

void show_select_attribute 
  (ABSTRACT_EVENT *context,  char *attr_name, char *node_name, char *x_axis)
{

  /**********************************************************************
   * context = root of the search tree. Default restricts the           *
   *           search to the children of "context".                     *
     xlist   = list of abstract events that form the x-axis of the 
               scatter/line plot. Satisfies the following:
	       1. They are named "x_axis"
	       2. They are descendants of "context"
     clist   = list of abstract events that satisfy the following two 
               conditions:
               1. They are descendants of "context"
	       2. They have the attribute called "attr_name"
	       3. They are named "node_name"
   ***********************************************************************/


  ABSTRACT_EVENT_LIST *xlist = NULL, *clist = NULL;
  int xsize;

  /*****  *****  temporary variables  *****  *****/
  BAR_GRAPH_NODES *bgn;
  int i,j;

  /****  color counter  ****/
  int clr = -1;
    

  /* free previous Drawing Canvas Display data structures */
  free_bar_graph_node (&GBarGraph);
  free_color_database (&CDB);
  set_show_attribute_flag(1);


  /* set the x axis: xlist */
  xlist = get_named_nodes (x_axis, context, 0);
  if (xlist == NULL) {
    printf("Illegal x-axis specification \n");
    return;
  }

  xsize = get_size_of_abstract_list (xlist);

  /* Computing "clist", and "GBarGraph" */
  for (i=0; i<xsize; i++) {
    ABSTRACT_EVENT_LIST *tael, *first, *last;
    ABSTRACT_EVENT *ae = get_element_from_abstract_list (xlist,i);
    int size;


    tael = get_named_nodes (node_name, ae, 0);
    size = get_size_of_abstract_list (tael);

    first = last = tael;
    for (j=0; j<size-1; j++)
      last = last->next;

    add_to_bar_graph_node (ae, first, last, size, &GBarGraph);

    /* Now add "tael" to "clist" */
    abstract_list_concat (&clist, tael);
  }
  

  /* Compute colors for each element of "clist" */
  map_attribute_to_color (clist, attr_name);

  /* Compute local category statistics */
  bgn = GBarGraph;
  while (bgn) {
    calculate_aggr_stat (bgn->first, bgn->last, attr_name, bgn);
    bgn = bgn->next;
  }


  /* fix the layout of the category statistics. */
  fix_layout_category_statistics (GBarGraph);




  /* Fix the layout of the scatter/line plot */
  fix_layout_discrete_plot (GBarGraph, attr_name);

  compute_aggr_color (GBarGraph);
}

    
