#include <stdio.h>
#include "structs.h"
#include "abstract.h"
#include "assert.h"
#include "event.h"
#include "expr.h"
#include "poeg.h"
#include "scan.h"
#include "symtab.h"
#include "strs.h"
#include "display.h"

extern COLLAPSE_NODE *allocate_collapse_node();
EVENT *GEvntPtr;
EVENT *LastGEvntPtr;
int GPopMCast = 0;
int StackNum;
extern char*get_name (ABSTRACT_EVENT *ae);

/* Global variables required by matching routines */
int inOr, inStar, inCnt;
int inChStar, inChCnt;
int InfinitePop;
int inPtChStar; 
void match_init()
{
  inOr = inCnt = inStar = 0;
  inChStar = inChCnt = 0;
  inPtChStar = 0;
  InfinitePop = 0;
  NodeId = 1;
  return;
}



/*
 *****************************************************************
 * FUNCTION(S):  Switching on/off the primitive events making up *
 *               the matched abstract event.                     *
 *               o make_available_matched_events()               *
 * DESCRIPTION:  Ariadne allows the user to issue the following  *
 *               statements.                                     *
 *               ... some description of pt-chain x1 ...         *
 *               match m1 = x1;                                  *
 *               ... some description of pt-chain x2 ...         *
 *               match m2 = x2 with m1;                          *
 *               where m2 is matched against the events that make*
 *               up m1. So before the second match can start, we *
 *               need to make the events of m1 available for     *
 *               matching.                                       *
 *               The statement "restore(m1)" would also call the *
 *               function.                                       *
 *****************************************************************
*/

void make_available_matched_events(ABSTRACT_EVENT *ae) 
{
  if (ae == NULL)
    return;
  
  if (ae->ep)
    Active(ae->ep) = 1;

  make_available_matched_events(ae->next);
  make_available_matched_events(ae->down);
}

/*
 *****************************************************************
 * FUNCTION(S):  Switching on/off the primitive events making up *
 *               the matched abstract event.                     *
 *               o make_unavailable_matched_events()             *
 * DESCRIPTION:  Once a pt-chain is matched against the trace,   *
 *               the events that make up the match, are disabled,*
 *               and become unavailable for further matching. The*
 *               events can be made available by issuing a       *
 *               "restore" statement or matching a new pt-chain  *
 *               in the context of an existing "match variable". *
 *****************************************************************
*/


void make_unavailable_matched_events (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  if (ae->ep)
    Active(ae->ep) = 0;

  make_unavailable_matched_events(ae->next);
  make_unavailable_matched_events(ae->down);
}


/*
 *****************************************************************
 * FUNCTION:     is_oper_at_reachable()                          *
 * DESCRIPTION:  Ariadne allows for "dummy" writes/reads at the  *
 *               boundary processes. For instance, in a SOR on a *
 *               2-D grid, the internal processes communicate    *
 *               with all their NEWS neighbor, whereas those at  *
 *               the boundary may have either 2 or 3 neighbors.  *
 *               Ariadne expects the user to handle it by putting*
 *               dummy sends/receives at the boundary processes. *
 *               ch msg = (W <: @ R :>)^4;                       *
 *               for a process that has only 2 neighbors, and a  *
 *               trace of the form  W W_dummy W W_dummy          *
 *               the first                                       *
 *****************************************************************
*/

int is_oper_at_reachable (TREE_OBJ *co)
{
  co = co->next;
  while (co) {
    if (co->oper == OperPush || co->oper == OperPop)
      co = co->next;
    else if (co->oper == OperAt)
      return 1;
    else
      return 0;
  }
  return 0;
}


  
void add_to_id_list (IDLIST **idl, IDENT *id)
{
  IDLIST *tmp = *idl;
  IDLIST *new = (IDLIST *) malloc(sizeof(IDLIST));
  new->id = id;
  new->next = NULL;

  if (tmp == NULL) {
    *idl = new;
    return;
  }
  
  while (tmp->next)
    tmp = tmp->next;
  tmp->next = new;
  return;
}


free_current_decl (IDLIST *idl)
{
  IDENT *id;
  OBJECT *o;
  DECL *d;
  while (idl) {
    id = idl->id;
    o = id->id_decl->decl_object;
    id->id_decl->decl_object = NULL;
    free(o);
    d = id->id_decl;
    id->id_decl = id->id_decl->decl_next;
    d->decl_next = NULL;
    free(d);
    idl = idl->next;
  }
}

void free_id_list (IDLIST *idl)
{
  IDLIST *i;
  while (idl) {
    idl->id = NULL;
    i = idl;
    idl = idl->next;
    i->next = NULL;
    free(i);
  }
}


int is_event_in_list (char *marker, WRT_LIST *wrt)
{
  if (marker == NULL)
    return 0;

  if (wrt == NULL)
    return 1;

  while (wrt != NULL) {
    if (!StrCmp(marker, wrt->marker))
      return 1;
    wrt = wrt->next;
  }
  return 0;
}

void change_matched_events_to_false(ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  
  if (ae->ep)
    Active(ae->ep) = 0;

  if (ae->next)
    change_matched_events_to_false (ae->next);

  if (ae->down)
    change_matched_events_to_false (ae->down);
}

void free_wrt_list (WRT_LIST *wrt)
{
  WRT_LIST *tmp = wrt;
  if (wrt == NULL)
    return;
  tmp = wrt->next;
  while (tmp) {
    wrt->next = NULL;
    free(wrt->marker);
    wrt = tmp;
    tmp = wrt->next;
  }
  free(wrt->marker);
}

void add_to_wrt_list (WRT_LIST **head, WRT_LIST *wrt)
{
  WRT_LIST *tmp = *head;
  if (tmp == NULL) {
    *head = wrt;
    return;
  }
  
  while (tmp->next)
    tmp = tmp->next;
  tmp->next = wrt;
  return;
}
    
WRT_LIST *replicate_wrt_list (WRT_LIST *wrt)
{
  WRT_LIST *head = NULL;
  while (wrt) {
    WRT_LIST *new = (WRT_LIST *) malloc(sizeof(WRT_LIST));
    new->marker = StrAlloc (StrLen(wrt->marker)+1);
    new->next = NULL;
    sprintf(new->marker, "%s", wrt->marker);
    wrt = wrt->next;
    add_to_wrt_list (&head, new);
  }
  return head;
}
  
WRT_LIST *new_concat_wrt_list (WRT_LIST *w1, WRT_LIST *w2)
{
  WRT_LIST *t1 = replicate_wrt_list (w1);
  WRT_LIST *t2 = replicate_wrt_list (w2);
  
  if (t1 == NULL)
    return t2;
  if (t2 == NULL)
    return t1;

  while (t1->next)
    t1 = t1->next;
  t1->next = t2;
  return t1;
}

int GetStack ()
{
  return StackNum++;
}

void PutStack ()
{
  --StackNum;
  return;
}
  
/************************************************************************
 * GEvntQ points to a global list of event pointers, that may need to   *
 * be restored, if the match fails at any time. "AppendToGlobalQ" will  *
 * add an event pointer to the head of the queue.                       *
 ************************************************************************/

EVENT_PTR_LIST *get_global_q_ptr (int pid)
{
  return GEvntQ[pid];
}

void AddGlobalQ (EVENT *ep, int pid)
{
  EVENT_PTR_LIST *epl = (EVENT_PTR_LIST *) malloc(sizeof(EVENT_PTR_LIST));
  epl->ep = ep;
  epl->next = GEvntQ[pid];
  GEvntQ[pid] = epl;
  return;
}



void DestroyEvPtrList (EVENT_PTR_LIST *tmp)
{
  tmp->ep = NULL;
  tmp->next = NULL;
  free (tmp);
  return;
}


void RestoreGlobalQ (int val, int pid, EVENT_PTR_LIST *sp)
{
  EVENT_PTR_LIST *epl = GEvntQ[pid];
  while ((epl != NULL) && (epl != sp))
    {
      EVENT_PTR_LIST *tmp = epl;
      Active(epl->ep) = val;
      epl = epl->next;
      tmp->ep = NULL;
      tmp->next = NULL;
      free(tmp);
    }
  GEvntQ[pid] = NULL;
  return;
}




EVENT *next_real_active_event (EVENT *ep)
{
  if (ep == NULL)
    return ep;

  if (Active(ep) == 1)
    return ep;

  while ((ep != NULL) && Active(ep) != 1)
    ep = ep->next;
  return ep;
}

EVENT *Pop(int st_num)
{
  EVENT_PTR_LIST *epl;
  EVENT *ep;
  int sp;
  if (st_num < GNumProcs)
    sp = 0;
  else
    sp = st_num - GNumProcs + 1;
  epl  = GEvntStack[sp];
  if (epl == NULL)
    return GEvntPtr;
  ep = epl->ep;
  epl->ep = NULL;
  GEvntStack[sp] = epl->next;
  free(epl);
  return next_real_active_event(ep);
}

void Push (EVENT *ep, int st_num)
{
  EVENT_PTR_LIST *epl = (EVENT_PTR_LIST *) malloc (sizeof(EVENT_PTR_LIST));
  int sp;
  if (st_num < GNumProcs)
    sp = 0;
  else
    sp = st_num - GNumProcs + 1;

  epl->ep = ep;
  epl->next = GEvntStack[sp];
  GEvntStack[sp] = epl;
  return;
}


void AddEventToSoSet (SO_SET *ws, EVENT *ep)
{
  EVENT_PTR_LIST *epl = allocate_event_ptr_list(ep);
  EVENT_PTR_LIST *tmp = ws->epl;
  (ws->size)++;
  if (tmp == NULL)
    ws->epl = epl;
  else
    {
      while (tmp->next)
	tmp = tmp->next;
      tmp->next = epl;
    }
  return;
}
  
SO_SET *allocate_so_set()
{
  SO_SET *ws = (SO_SET *) malloc(sizeof(SO_SET));
  ws->size = 0;
  ws->epl = NULL;
  return ws;
}

SO_SET *ComputeSoSet (EVENT *ep)
{
  SO_SET *ws;
  EVENT     *te;
  int       pos;
  EVENT_PTR_LIST *epl;

  if (SoSet (ep) != NULL)
    return SoSet(ep);

  ws = allocate_so_set();
  pos = HashEventFunc(ep);
  
  epl = (GEventHashTab->buckets)[pos];
  
  while (epl)
    {
      te = epl->ep;

      /******************************************************************
       * If item in the bucket has the same key value, yet it is not me *
       * myself, then add it to my eventset.                            *
       ******************************************************************/
      
      if ( (MsgID(ep) == MsgID(te)) && (te != ep) && 
	  (EventType(ep) != EventType(te)))
	AddEventToSoSet (ws, te);
      
      epl = epl->next;
    }
  
  SoSet (ep) = ws;
  return ws;
}







void add_allocate_collapse_node (ABSTRACT_EVENT *ae)
{
  int my_id;
  COLLAPSE_NODE *cn = allocate_collapse_node();
  cn->down = NULL;
  cn->pae = ae;
  cn->par_id = cn->my_id = 0;
  if (ae->clps) {
    COLLAPSE_NODE *tmp = ae->clps;
    while (tmp->next)
      tmp = tmp->next;
    tmp->next = cn;
  }
  else {
    cn->next  = NULL;
    ae->clps = cn;
  }
  cn->ave = NULL;
  return;
}


  

int is_event_in_abstract_event (EVENT *ep, ABSTRACT_EVENT *ae)
{
  int nxt = 0, dwn = 0;
  if (ae == NULL)
    return 0;

  if (ae->ep == ep)
    return 1;

  dwn = is_event_in_abstract_event (ep, ae->down);
  nxt = is_event_in_abstract_event (ep, ae->next);
  return (dwn || nxt);
}
    


int is_event_in_wlist (EVENT *ep, WRT_LIST *wlist)
{
  if (wlist == NULL || ep == NULL)
    return 1;
  while(wlist) {
    IDENT *id = id_find(wlist->marker);
    ABSTRACT_EVENT *ae;
    if (id == NULL || id->id_decl == NULL) {
      printf("PANIC: %s is not an abstract event\n", id->id_chars);
      return 0;
    }
    ae = (ABSTRACT_EVENT *) id->id_decl->decl_object;
    if (is_event_in_abstract_event(ep,ae) == 1)
      return 1;
    else
      wlist = wlist->next;
  }
  return 0;
}
      

EVENT *NextActiveEventOnProc (int pid, WRT_LIST *wlist)
{
  EVENT *ep = GExecGraph[pid];
  if (ep == NULL)
    return NULL;

  if ((Active(ep) > 0) && (is_event_in_wlist(ep,wlist) == 1))
    return ep;

  while ((ep != NULL) && 
	 ((Active(ep) == 0) || (is_event_in_wlist(ep,wlist)==0)))
    ep = ep->next;
  
  return ep;

}

EVENT *NextActiveEvent (EVENT *ep, WRT_LIST *wlist)
{
  if (ep == NULL)
    return NULL;
  if (Active(ep) > 0 && is_event_in_wlist(ep, wlist))
    return ep;
  
  while ((ep != NULL) && 
	 ((Active(ep) == 0) || (is_event_in_wlist(ep,wlist)== 0)) )
    ep = ep->next;
  return ep;
}

EVENT *get_next_active_event_after_me(EVENT *ep, WRT_LIST *wlist)
{
  if (ep == NULL)
    return NULL;
  return NextActiveEvent (ep->next, wlist);
}
 

char *get_name_of_chain (TREE_OBJ *to)
{
  char *dwn = '\0';
  char *nxt = '\0';
  char *me;
  char tmp[4];

  if (to->down)
    dwn = get_name_of_chain (to->down);
  if (to->next)
    nxt = get_name_of_chain (to->next);
  
  me = StrAlloc (StrLen(dwn) + StrLen(nxt) + StrLen(to->name) + 6);
  
  if (to->oper == OperParen) {
    sprintf(tmp, "(");
    StrCat (me, tmp);
    StrCat (me, dwn);
    sprintf(tmp, ") ");
    StrCat (me, tmp);
    StrCat (me, nxt);
  }
  else if (to->oper == OperAny) {
   sprintf(tmp, "(+");    
   StrCat (me, tmp);
   StrCat (me, dwn);
   sprintf(tmp, "+) ");
   StrCat (me, tmp);
   StrCat (me, nxt);
  }
  else if (to->cnt == 0) {
    StrCat (me,dwn);
    sprintf(tmp, "* ");
    StrCat (me, tmp);
    StrCat (me,nxt);
  }
  else if (to->cnt>1) {
    StrCat (me,dwn);
    sprintf(tmp, "^%d ", to->cnt);
    StrCat (me, tmp);
    StrCat (me, nxt);
  }
  else if (to->cnt == 1) {
    StrCat (me, to->name);
    StrCat (me, " ");
    StrCat (me, nxt);
  }

  return me;
}

char *get_name_of_ptchain (TREE_OBJ *to)
{
  char *dwn = '\0';
  char *nxt = '\0';
  char *me;
  char tmp[4];

  if (to->down)
    dwn = get_name_of_ptchain (to->down);
  if (to->next)
    nxt = get_name_of_ptchain (to->next);
  
  me = StrAlloc (StrLen(dwn) + StrLen(nxt) + StrLen(to->name) + 6);
  me[0] = '\0';
  
  if (to->oper < OperNone) {
    sprintf(tmp, "(");
    StrCat (me, tmp);
    StrCat (me, dwn);
    sprintf(tmp, ") ");
    StrCat (me, tmp);
    StrCat (me, nxt);
  }
  else if (to->oper == OperAny) {
   sprintf(tmp, "(+");    
   StrCat (me, tmp);
   StrCat (me, dwn);
   sprintf(tmp, "+) ");
   StrCat (me, tmp);
   StrCat (me, nxt);
  }
  else if (to->cnt == 0) {
    StrCat (me,dwn);
    sprintf(tmp, "* ");
    StrCat (me, tmp);
    StrCat (me,nxt);
  }
  else if (to->cnt>1) {
    StrCat (me,dwn);
    sprintf(tmp, "^%d ", to->cnt);
    StrCat (me, tmp);
    StrCat (me, nxt);
  }
  else if (to->cnt == 1) {
    StrCat (me, to->name);
    StrCat (me, " ");
    StrCat (me, nxt);
  }

  return me;
}
    

ABSTRACT_EVENT *match_chain_definition (TREE_OBJ *co, int pid,
			 ABSTRACT_EVENT *parent_node, WRT_LIST *pwrt,
			    WRT_LIST *with_list)
{
  ABSTRACT_EVENT *ths_ae = NULL;
  ABSTRACT_EVENT *nxt_ae = NULL;
  ABSTRACT_EVENT *dwn_ae = NULL;
  ABSTRACT_EVENT *tmp_ae;
  IDLIST *idl = NULL;
  int check = 0;
  WRT_LIST *wrt;

  int star_cnt = 0;

  if (co == NULL)
    return NULL;


  wrt = new_concat_wrt_list (pwrt, co->wrt);

  if (co->down) {
    char *tmp_name = get_name_of_chain (co->down);
    char *name = StrAlloc (StrLen(tmp_name) + 10);
    StrCpy  (name, tmp_name);

    if (co->oper == OperAny) {
      TREE_OBJ *tmp = co->down;
      int stack_no = GetStack();
      EVENT_PTR_LIST *sp = get_global_q_ptr (stack_no);
      char *nme = get_name_of_chain (co);

/*      ths_ae = allocate_abstract_event (nme, nme, co->marker, 
					  parent_node); */

      ths_ae = allocate_abstract_event (NULL, NULL, co->marker,  
					  parent_node); 
      while (tmp) {
	TREE_OBJ *nxt = tmp->next;
	      
	tmp->next = NULL;
	Push(GEvntPtr, stack_no);
	dwn_ae = 
	  match_chain_definition (tmp, stack_no,  ths_ae, wrt, with_list);
	if (!does_abstract_event_contain_error(dwn_ae)) {
	  tmp->next = nxt;
	  break;
	}
	RestoreGlobalQ(1, stack_no, sp);
	tmp->next = nxt;
	tmp = tmp->next;
	GEvntPtr = Pop(stack_no);
	if (tmp)
	  ERROR(dwn_ae) = 0;
      }
      Pop(stack_no);
      ths_ae->down = dwn_ae;
      /* set orig and marker */
      ths_ae->orig = StrAlloc(StrLen(dwn_ae->orig) + 1);
      sprintf(ths_ae->orig, "%s", dwn_ae->orig);

      ths_ae->name = StrAlloc(StrLen(dwn_ae->name) + 1);
      sprintf(ths_ae->name, "%s", dwn_ae->name);

      if (does_abstract_event_contain_error(dwn_ae))
	return ths_ae;
      if (GEvntQ[stack_no]) {
	EVENT_PTR_LIST *epl = GEvntQ[stack_no];
	while (epl) {
	  AddGlobalQ (epl->ep, pid);
	  epl = epl->next;
	}
	RestoreGlobalQ(0, stack_no, sp);
	PutStack();
      }
    }
    else if (co->oper == OperParen) {
      char *nme = get_name_of_chain (co);
      tmp_ae = 
	allocate_abstract_event (nme, nme, NULL, 
				  parent_node);
      tmp_ae->dummy = 1;
	  
      dwn_ae = 
	match_chain_definition (co->down, pid,  parent_node, wrt, with_list);
      if (dwn_ae == NULL) {
	free(tmp_ae);
	tmp_ae = NULL;
	return NULL;
      }
      tmp_ae->down = dwn_ae;
      append_to_sibling_abstract_event (&ths_ae, tmp_ae);
      if (does_abstract_event_contain_error(dwn_ae))
	return ths_ae;
    }
    else if (co->oper == OperNone) {
      ABSTRACT_EVENT *h_dwn_ae = NULL;
  
      if (co->cnt == 0) {  /* Kleene Star */
	char orig[80];
	char tmp[60];
	sprintf(orig, "%s*", name);
	ths_ae = 
	  allocate_abstract_event (orig, name, co->marker, 
				    parent_node);
	
	star_cnt = 0;
	inChStar++;

	while (1) {
	  EVENT_PTR_LIST *sp = GEvntQ[pid];
	  dwn_ae = 
	    match_chain_definition(co->down, pid,  ths_ae, wrt, with_list);
	  if (dwn_ae == NULL) {
	    InfinitePop = 0;
	    break;
	  }
	  if (InfinitePop == 1) {
	    InfinitePop = 0;
	    break;
	  }
	  if (does_abstract_event_contain_error(dwn_ae)) {
	    RestoreGlobalQ(1, pid, sp);
	    break;
	  }
	  append_to_sibling_abstract_event(&h_dwn_ae, dwn_ae);
	  ++star_cnt;
	}
	--inChStar;
	if (star_cnt) {
	  sprintf(tmp, "^%d", star_cnt);
	  StrCat (ths_ae->name, tmp);
	  ths_ae->down = h_dwn_ae;
	  ths_ae->cnt = star_cnt;
	}
	else {
	  free(ths_ae);
	  ths_ae = NULL;
	}
      }
      else if (co->cnt > 1) {
	int i;
	char suff[10];
	sprintf(suff, "^%d", co->cnt);
	StrCat (name, suff);
	      
	ths_ae = 
	  allocate_abstract_event (name, name, (char *) NULL, 
				    parent_node);
	inChCnt++;
	for (i=0; i<co->cnt; i++) {
	  dwn_ae = 
	    match_chain_definition (co->down, pid,  ths_ae, wrt, with_list);
	  append_to_sibling_abstract_event (&h_dwn_ae, dwn_ae);
	  
	  if (does_abstract_event_contain_error (dwn_ae)) {
	    ths_ae->error = 1;
	    ths_ae->down = h_dwn_ae;
	    return ths_ae;
	  }
	}
	ths_ae->down = h_dwn_ae;
	ths_ae->cnt = i;
	--inChCnt;
      }
    }
  }
  else  {
    if (co->oper == OperPush) {
      Push (GEvntPtr, pid);
      return 
	match_chain_definition(co->next, pid,  parent_node, wrt, with_list);
    }
    else if (co->oper == OperPop) {
      if (GPopMCast)
	return NULL;
      else {
	/**** Changed April 12, 1995 ****/
	EVENT *tmp = GEvntPtr;
	GEvntPtr = Pop(pid);
	if (tmp == GEvntPtr) {
	  InfinitePop = 1;
	  return NULL;
	}
	else
	  return 
	    match_chain_definition (co->next, pid,  parent_node, wrt, with_list);
      }
    }
    else if (co->oper == OperAt) {
      SO_SET *ws;
      EVENT_PTR_LIST *epl;
      /* We went ahead with the event ptr, so back up to the last event seen */
      GEvntPtr = LastGEvntPtr;

      ws = ComputeSoSet(GEvntPtr);
      epl = ws->epl;
	  
      SoSet(GEvntPtr) = ws;
	      
      if (ws->size != 1)
	GPopMCast = 1;
      while (epl != NULL) {
	GEvntPtr = epl->ep;
	if (epl->next == NULL)
	  GPopMCast = 0;
	nxt_ae = 
	  match_chain_definition(co->next, pid,  parent_node, wrt, with_list);
	append_to_sibling_abstract_event(&ths_ae, nxt_ae);
	if (does_abstract_event_contain_error(nxt_ae)) 
	  return ths_ae;
	epl = epl->next;
      }
      return ths_ae;
    }
		
    if (co->name) {
      if (co->flag) {        /* Primitive Event */
	ths_ae = allocate_abstract_event (co->name, co->name, co->marker, 
					   parent_node);
	GEvntPtr = NextActiveEvent(GEvntPtr, with_list);
	while (GEvntPtr && 
	       is_event_in_list(get_event_name(GEvntPtr), wrt) == 0) 
	  GEvntPtr = get_next_active_event_after_me(GEvntPtr, with_list);

	      
	if (GEvntPtr == NULL) {
	/*  printf("<ch>   Expecting \"%s\", encountered the right cursor.\n",co->name); */
	  ERROR(ths_ae) = 1;
	  return ths_ae;
	}
	else if (!StrCmp(co->name, get_event_name(GEvntPtr)) ||
		 !StrCmp(co->name, GEvntPtr->event_type)) {
	  ths_ae->ep = GEvntPtr;
	  /* before we can check, make all attributes visible */
	  check = filter_interp_def (co->dfl, &idl, ths_ae);
	  if (check == 0) { /* skip this event */
	    free_current_decl(idl);
	    free_id_list(idl);
	    ths_ae->ep = NULL;
	    GEvntPtr = get_next_active_event_after_me(GEvntPtr, with_list);

	    return 
	      match_chain_definition (co, pid,  parent_node, wrt, with_list);
	  }
	  else if (check == -1) {
	    ERROR(ths_ae) = 1;
	    free_current_decl(idl);
	    free_id_list(idl);
	    return ths_ae;
	  }
	  else { /* check == 1 */
	    Active(GEvntPtr) = 2;
	    AddGlobalQ (GEvntPtr, pid);
	    ths_ae->ep = GEvntPtr;
	    LastGEvntPtr = GEvntPtr;
	    GEvntPtr = GEvntPtr->next;
	  }
	}
	else {	      /* ERROR */
	  /* Not necessarily an error. If we found a 
	     dummy event that does not match we get the
	     next event */
	  if (OID(GEvntPtr) == AA_UNDEFINED) {
	    GEvntPtr = GEvntPtr->next;
	    GEvntPtr = NextActiveEvent (GEvntPtr, with_list);
	    return match_chain_definition (co, pid,  parent_node, wrt, with_list);
	  }
	  else {
/*	    printf("<ch> Expecting %s, found %s in process %d\n",  
		   co->name, EventType(GEvntPtr), PID(GEvntPtr)); */
	    ERROR(ths_ae) = 1;
	    return ths_ae;
	  }
	}
      }
      else {
	ABSTRACT_EVENT *tmp;
	IDENT *id = id_find (co->name);
	TREE_OBJ *tco = (TREE_OBJ *) id->id_decl->decl_object;
	
	GEvntPtr = NextActiveEvent(GEvntPtr, with_list);
	ths_ae  = allocate_abstract_event(co->name, co->name, co->marker, 
					   parent_node);
	tmp = match_chain_definition (tco, pid,  ths_ae, wrt, with_list);
	if (tmp == NULL) {
	  free(ths_ae);
	  ths_ae = NULL;
	  return NULL;
	}
	ths_ae->down = tmp;
	if (does_abstract_event_contain_error(tmp)) {
	  ths_ae->error = TRUE;
	  return ths_ae;
	}
      }
    }
  }
    
  if (GEvntPtr && OID(GEvntPtr) == AA_UNDEFINED) {
    if ((co->next) && (is_oper_at_reachable(co))) {
    /* move till you get a Pop */
      while (co->next && co->next->oper != OperPop)
	co = co->next;
    }
  }
/*  if (inChStar || inChCnt)
    nxt_ae = match_chain_definition (co->next, pid,  parent_node, wrt, with_list);
  else
    nxt_ae = match_chain_definition (co->next, pid,  parent_node, wrt, with_list);
*/

  nxt_ae = match_chain_definition (co->next, pid,  parent_node, wrt, with_list);
  
append_to_sibling_abstract_event (&ths_ae, nxt_ae);
  free_wrt_list(wrt);
  free_current_decl(idl);
  free_id_list(idl);
  return ths_ae; 
}


ABSTRACT_EVENT *match_pchain_definition 
(PCH_OBJ *po, ABSTRACT_EVENT *parent_node, WRT_LIST *with_list)
{

  SET_OBJ *so = po->set;
  SET_OBJ *tso = po->set;
  TREE_OBJ  *co = po->chain;
  char *orig = get_name_of_chain (co);
  ABSTRACT_EVENT *ths_ae = NULL;

  ABSTRACT_EVENT *dwn_ae = NULL;
  ABSTRACT_EVENT *nxt_ae;

  SET_OBJ *set_succ = allocate_set();
  char tmp_msg[200];
  EVENT_PTR_LIST *sp;
  int i;
  CORE_SET_OBJ *core;

  if (so->name) {
    IDENT *id = id_find(so->name);
    tso = (SET_OBJ *) id->id_decl->decl_object;
    core = tso->core;
  }
  
  /* check whether core is null or not April 12, 1995 */
  if (core == NULL) {
    printf("Set %s is not defined\n", so->name);
    return NULL;
  }
    
  for (i=0; i<core->size; i++) {
    int pid = get_process_identifier (tso, i);
    char *name = StrAlloc (find_size(i) + 6);
    sprintf(name,"%d", i);

    nxt_ae = allocate_abstract_event(orig, "proc", (char *) NULL, 
				      parent_node);

    GEvntPtr = NextActiveEventOnProc(pid, with_list);
    sp = get_global_q_ptr (pid);
    dwn_ae = match_chain_definition (co, pid,  nxt_ae, NULL, with_list);

    if (dwn_ae == NULL) {
      free(nxt_ae);
      nxt_ae = NULL;
    }
    else if (po->on_kind == OnAll) {
      nxt_ae->down = dwn_ae;
      nxt_ae->proc = pid;
      append_to_sibling_abstract_event (&ths_ae, nxt_ae);
      if (does_abstract_event_contain_error (dwn_ae)) {
	ERROR(nxt_ae) = 1;
	return ths_ae;
      }
    }
    else {  /* onsome */
      if (does_abstract_event_contain_error(dwn_ae)) 
	RestoreGlobalQ(1, pid, sp);
      else {
	nxt_ae->proc = pid;
	nxt_ae->down = dwn_ae;
/*	printf("doing this\n"); */
	append_to_sibling_abstract_event (&ths_ae, nxt_ae);
      }
    }
  }


  if ((nxt_ae != NULL) && (po->on_kind == OnSome) && (ths_ae == NULL)) {
    ERROR (nxt_ae) = 1;
    append_to_sibling_abstract_event (&ths_ae, nxt_ae);
  }

  /* Before returning change tentatives to false */
  if (ths_ae)
    change_matched_events_to_false (ths_ae);
  return ths_ae;
}


ABSTRACT_EVENT *match_ptchain_definition 
(TREE_OBJ *to, ABSTRACT_EVENT *parent_node, WRT_LIST *with_list)
{
  ABSTRACT_EVENT *ths_ae   = NULL;
  ABSTRACT_EVENT *nxt_ae   = NULL;
  ABSTRACT_EVENT *dwn_ae   = NULL;
  ABSTRACT_EVENT *h_dwn_ae = NULL;
  ABSTRACT_EVENT *tmp_ae;

  int star_cnt  = 0;

  if (to == NULL)
    return NULL;


  /*******************************************************************************
   *              Handle the following situations:                               *
   * '{' ptchain '}', '<' ptchain '>',  '(' ptchain ')', ptchain'^'n, ptchain'*' *
   *****************************************************************************/

  if (to->down) {
    char *tmp_name = get_name_of_ptchain (to->down);
    char *name = StrAlloc (StrLen(tmp_name) + 10);
    sprintf(name, "%s", tmp_name);
    printf("%s\n", tmp_name);


    if ((to->oper == OperParallel) || (to->oper == OperPrecede)) {
      char *new_name;
      tmp_ae = allocate_abstract_event (NULL, NULL, NULL,
					 parent_node);
      dwn_ae = match_ptchain_definition (to->down, tmp_ae,  with_list);
      new_name = get_name (dwn_ae);
      tmp_ae->down = dwn_ae;
      tmp_ae->name = StrAlloc (strlen(new_name)+3);
      tmp_ae->orig = StrAlloc (strlen(new_name)+3);
      tmp_ae->marker = StrAlloc (strlen(new_name)+3);
      if (to->oper == OperPrecede) {
	tmp_ae->name[0] = tmp_ae->marker[0] = tmp_ae->orig[0] = '<';
	tmp_ae->name[1] = tmp_ae->marker[1] = tmp_ae->orig[1] = '\0';
      }
      else if (to->oper == OperParallel) {
	tmp_ae->name[0] = tmp_ae->marker[0] = tmp_ae->orig[0] = '{';
	tmp_ae->name[1] = tmp_ae->marker[1] = tmp_ae->orig[1] = '\0';
      }
      StrCat(tmp_ae->name, new_name);
      StrCat(tmp_ae->orig, new_name);
      StrCat(tmp_ae->marker, new_name);
      if (to->oper == OperPrecede) {
	StrCat(tmp_ae->name, ">");
	StrCat(tmp_ae->orig, ">");
	StrCat(tmp_ae->marker, ">");
      }
      else if (to->oper == OperParallel) {
	StrCat(tmp_ae->name, "}");
	StrCat(tmp_ae->orig, "}");
	StrCat(tmp_ae->marker, "}");
      }
	
      append_to_sibling_abstract_event  (&ths_ae, dwn_ae);
      if (does_abstract_event_contain_error(dwn_ae) ||
	  (assert_relationship_between_abstract_events
	   (dwn_ae, to->wrt, to->oper) == FALSE)) {
	ERROR(ths_ae) = 1;
	return ths_ae;
      }
    }
    else if (to->oper == OperAny) {
      TREE_OBJ *tmp = to->down;
      int pid;
      EVENT_PTR_LIST **sp = (EVENT_PTR_LIST **) 
	calloc (GNumProcs, sizeof (EVENT_PTR_LIST  **));

      
      /* tmp_ae = allocate_abstract_event ("(++)", "(++)", to->marker,
					 parent_node); */

      tmp_ae = allocate_abstract_event (NULL, NULL, to->marker,
					 parent_node);
      
      tmp_ae->dummy = 1;
      for (pid=0; pid<GNumProcs; pid++)
	sp[pid] = GEvntQ[pid];
	  
      ++inOr;
	  
      while (tmp) {
	TREE_OBJ *nxt = tmp->next;
	tmp->next = NULL;
	dwn_ae = match_ptchain_definition (tmp, tmp_ae,  with_list);
	if (!does_abstract_event_contain_error(dwn_ae)) {
	  tmp->next = nxt;
	  break;
	}

	for (pid=0; pid<GNumProcs; pid++)
	  RestoreGlobalQ (1, pid, sp[pid]);

	tmp->next = nxt;
	tmp = tmp->next;
	if (tmp)
	  ERROR(dwn_ae) = 0;
      } 
      --inOr;
      tmp_ae->down = dwn_ae;
      /* fix tmp_ae's name, orig etc */
      /* Added Dec 5, 1996 */
      tmp_ae->name = StrAlloc (StrLen(dwn_ae->name)+1);
      sprintf(tmp_ae->name, "%s", dwn_ae->name);
      tmp_ae->orig = StrAlloc (StrLen(dwn_ae->orig)+1);
      sprintf(tmp_ae->orig, "%s", dwn_ae->orig);

      append_to_sibling_abstract_event(&ths_ae, tmp_ae);
      if (does_abstract_event_contain_error(dwn_ae))
	return ths_ae;
    }
    else if (to->oper == OperParen) {
      char *nme = get_name_of_ptchain (to);
      tmp_ae = allocate_abstract_event (nme, nme, NULL,
					 parent_node);
      tmp_ae->dummy = 1;
      dwn_ae = match_ptchain_definition (to->down, parent_node,  with_list);
      tmp_ae->down = dwn_ae;
      append_to_sibling_abstract_event(&ths_ae, tmp_ae);
      if (does_abstract_event_contain_error(dwn_ae))
	return ths_ae;
      if (dwn_ae == NULL)
	return dwn_ae;
    }
    else if (to->oper == OperNone) {
      if (to->cnt == 0) { /* Kleene Star */
	char tmp[80];
	star_cnt = 0;
	inStar++;
	sprintf (tmp,"%s*", name);  /* extra spaces to handle ^n later */
	ths_ae = allocate_abstract_event (tmp, name, (char *) NULL, 
					  parent_node);	
	inPtChStar++;
	while (1) {
	  int pid;
	  EVENT_PTR_LIST **sp = (EVENT_PTR_LIST **)
	    calloc (GNumProcs, sizeof (EVENT_PTR_LIST **));
		    
	  for (pid=0; pid<GNumProcs; pid++)
	    sp[pid] = GEvntQ[pid];

	  dwn_ae = match_ptchain_definition (to->down, ths_ae,  with_list);
	  if (dwn_ae == NULL)
	    break;
	  if (does_abstract_event_contain_error(dwn_ae)) {
	    for (pid=0; pid<GNumProcs; pid++)
	      RestoreGlobalQ (1, pid, sp[pid]);
	    break;
	  }
	  append_to_sibling_abstract_event  (&h_dwn_ae, dwn_ae);
	  star_cnt++;
	}
	inPtChStar--;
	--inStar;
	if (star_cnt) {
	  sprintf(tmp, "%s^%d", ths_ae->name, star_cnt);
	  free(ths_ae->name);
	  ths_ae->name = NULL;
	  ths_ae->name = StrAlloc (strlen(tmp)+1);
	  sprintf(ths_ae->name, "%s", tmp);
	  /* StrCat (ths_ae->name, tmp); */
	  ths_ae->down = h_dwn_ae;
	  ths_ae->dummy = 1;
	  ths_ae->cnt = star_cnt;
	}
	else 
	  ths_ae = NULL;
      }
      else {
	int i;
	char tmp[80];
	sprintf(tmp, "%s^%d",tmp_name,to->cnt);
	printf("%s\n", tmp);
	ths_ae = 
	  allocate_abstract_event (tmp, tmp, (char *) NULL, 
				   parent_node);
	inCnt++;
	
	for (i=0; i<to->cnt; i++) {
	  dwn_ae = match_ptchain_definition(to->down, ths_ae,  with_list);
	  append_to_sibling_abstract_event (&h_dwn_ae, dwn_ae);
	  if (does_abstract_event_contain_error(dwn_ae)) {
	    ERROR(ths_ae) = 1;
	    ths_ae->down = h_dwn_ae;
	    return ths_ae;
	  }
	}
	ths_ae->down = h_dwn_ae;
	ths_ae->cnt = i;
	--inCnt;
      }
    }
  }
  else {    /* If it is ptchain */
    IDENT *id   = id_find(to->name);
    if (to->flag) {
      PCH_OBJ *po = (PCH_OBJ *) id->id_decl->decl_object;
      
      if ((to->marker) == NULL) {
	to->marker = StrAlloc (StrLen (to->name)+1);
	StrCpy (to->marker, to->name);
      }

      ths_ae = allocate_abstract_event (to->name, to->name, to->marker, 
					 parent_node);
      add_allocate_collapse_node (ths_ae);
      dwn_ae = match_pchain_definition(po, ths_ae,  with_list);
      ths_ae->down = dwn_ae;

      if (does_abstract_event_contain_error(dwn_ae)) {
	ERROR(ths_ae) = TRUE;
	return ths_ae;
      }
      if (dwn_ae == NULL)
	return dwn_ae;
	  
    }
    else { 
      TREE_OBJ *tto = (TREE_OBJ *) id->id_decl->decl_object;
      if ((to->marker) == NULL) {
	to->marker = StrAlloc (StrLen (to->name)+1);
	StrCpy (to->marker, to->name);
      }

      ths_ae = 
	allocate_abstract_event(to->name, to->name, to->marker, 
				 parent_node);
      dwn_ae = match_ptchain_definition (tto, ths_ae,  with_list);
      ths_ae->down = dwn_ae;
      if (does_abstract_event_contain_error(dwn_ae)) {
	ERROR(ths_ae) = TRUE;
	return ths_ae;
      }
      if (dwn_ae == NULL)
	return dwn_ae;
    }
  }
  nxt_ae = match_ptchain_definition (to->next, parent_node,  with_list);
  append_to_sibling_abstract_event (&ths_ae, nxt_ae);
  printf("match_ptchain_def()\n");
  return ths_ae;
}
  
void fix_prev_ptr (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  
  if (ae->next) 
    ae->next->prev = ae;

  fix_prev_ptr (ae->next);
  fix_prev_ptr (ae->down);
  return;
}



/*
 **********************************************************************
 * FUNCTION:      process_match_statement()                           *
 * DESCRIPTION:   o wrt != NULL => matching has to be done in the     *
 *                context of a previous match.                        *
 * NOTES:         change so that matching does not involve creating   *
 *                AVE_STR                                             *
 **********************************************************************
*/


void process_match_statement (TREE_OBJ *to, IDENT *mvar, WRT_LIST *wrt)
{
  ABSTRACT_EVENT *oae = NULL, *ae = NULL;  /* initialization of "oae" required */
  ABSTRACT_EVENT *t_ae;
  ABSTRACT_EVENT *c_ae;
  int pid;

  match_init();
  while (wrt) {
    IDENT *id = id_find(wrt->marker);

    if (error_check_for_identifier (id, wrt->marker) == 0)
      return;

    oae = (ABSTRACT_EVENT *) id->id_decl->decl_object;
    make_available_matched_events (oae);
    wrt = wrt->next;
  }

  ae = match_ptchain_definition(to, (ABSTRACT_EVENT *) NULL, NULL);


  t_ae = allocate_abstract_event (mvar->id_chars, mvar->id_chars, 
				  mvar->id_chars, NULL);
  t_ae->down = ae;
  /* ae and its sibling should have t_ae as their parent: 
     Bug fix: June 22, 1995 */

  c_ae = ae;
  while (c_ae) {
    c_ae->parent = t_ae;
    c_ae = c_ae->next;
  }
					  

  add_node_to_name_list (mvar);
  mvar->id_decl->decl_object = (OBJECT *) t_ae;

  /* this line is not really needed for the textual version */
  process_match_tree_display (t_ae);

  if (does_abstract_event_contain_error(ae)) {
/*    sprintf(terse_feedback, "Match failed."); */
    for (pid = 0; pid <GNumProcs; pid++)
      RestoreGlobalQ (TRUE, pid, (EVENT_PTR_LIST *) NULL);
  }
  else {
/*    sprintf(terse_feedback, "Match succeeded."); */
    for (pid = 0; pid < GNumProcs; pid++)
      RestoreGlobalQ (FALSE, pid, (EVENT_PTR_LIST *) NULL);
  }

  while (wrt) {
    IDENT *id = id_find(wrt->marker);

    oae = (ABSTRACT_EVENT *) id->id_decl->decl_object;
    make_unavailable_matched_events (oae);
    wrt = wrt->next;
  }

}


