#include <stdio.h>
#include "structs.h"
#include "scan.h"
#include "abstract.h"
#include "mtree.h"
#include "expr.h"


/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h :       Data Structures                                  *
 *   abstract.h:       get_process_set_of_abstract_event()              *
 *   global.h:         get_total_procs()                                *
 *   event.h:          PID, LTIME etc.                                  *
 *   match.h:          ComputeSoSet()                                   *
 *   scan.h:           id_find()                                        *
 ************************************************************************
*/             


ABSTRACT_EVENT_LIST 
*get_named_nodes (char *name, ABSTRACT_EVENT *context, int i)
{
  ABSTRACT_EVENT_LIST *head = NULL;
  ABSTRACT_EVENT_LIST *nxt = NULL;
  ABSTRACT_EVENT_LIST *dwn = NULL;
  ABSTRACT_EVENT *ae = NULL;

  if (context == NULL)
    return NULL;

  if (!StrCmp (name, context->name) || !StrCmp (name, context->marker) ||
      !StrCmp (name, "event") || 
      (context->ep && !StrCmp (name, context->ep->event_type))) {
    add_node_to_abstract_list (&head, context);
    if (i)
      nxt = get_named_nodes (name, context->next, i);
    abstract_list_concat (&head, nxt);
    return head;
  }
  else {
    dwn = get_named_nodes (name, context->down, 1);
    if (i)
      nxt = get_named_nodes (name, context->next, 1);
    abstract_list_concat (&head, dwn);
    abstract_list_concat (&head, nxt);
    return head;
  }
}
      
    

COLLAPSE_NODE *get_collapse_part 
  (char *name, ABSTRACT_EVENT *context, int i)
{
  COLLAPSE_NODE *cn = NULL;
  COLLAPSE_NODE *cc = NULL;

  if (context == NULL)
    return NULL;
  
  cn = context->clps;

  while (cn) {
    if (!StrCmp(cn->pattern, name))
      return cn->down;
    cn = cn->next;
  }
  
  if (i) {
    cc = get_collapse_part (name, context->next, i);
    if (cc)
      return cc;
  }

  cc = get_collapse_part (name, context->down, 1);
  if (cc)
    return cc;

  return NULL;
}


COLLAPSE_NODE *Get_Collapse_Part (SYN_TREE_LIST *stl, ABSTRACT_EVENT *context)
{
  SYN_TREE_LIST *s1 = stl->st->left;
  SYN_TREE_LIST *s2 = s1->next;
  int which_part;
  int i = 0;
  COLLAPSE_NODE *cc = NULL;

  if (s2 == NULL)
    which_part = 0;
  else
    which_part = get_int (evaluate_int_list (s2, context));

  cc = get_collapse_part (s1->st->name, context, 0);
  while (i != which_part && cc->next) 
    cc = cc->next;
  
  return cc;
}



/*
 ***********************************************************************
 * FUNCTION:     get_all_sibling()                                     *
 * DESCRIPTION:
 * DESIGN:
 * CALLER:
 * NOTES:
 ***********************************************************************
*/
ABSTRACT_EVENT_LIST *get_all_leaves (ABSTRACT_EVENT *ae)
{
  ABSTRACT_EVENT_LIST *ael = NULL;
  ABSTRACT_EVENT_LIST *tmp = NULL;
  ABSTRACT_EVENT_LIST *me = NULL;
  ABSTRACT_EVENT_LIST *siblings = NULL;
  

  if (ae == NULL) 
    return NULL;

  if (ae->down == NULL) /* I am a leaf */
    add_node_to_abstract_list (&ael, ae);
  else {
    int i, len;
    add_node_to_abstract_list (&me, ae->down);
    siblings = get_all_sibling (ae->down);
    abstract_list_concat (&me, siblings);
    len = get_size_of_abstract_list (me);
    for (i=0; i<len; i++) {
      ABSTRACT_EVENT *tae = get_element_from_abstract_list (me, i);
      tmp = get_all_leaves (tae);
      abstract_list_concat (&ael, tmp);
    }
  }
  return ael;
}
    
  
  
  
ABSTRACT_EVENT_LIST *get_all_sibling (ABSTRACT_EVENT *ae)
{
  ABSTRACT_EVENT_LIST *ael = NULL;
  ABSTRACT_EVENT *parent = NULL;
  ABSTRACT_EVENT *sib = NULL;

  if (ae == NULL)
    return NULL;

  parent = ae->parent;
  
  if (parent == NULL)
    return NULL;

  sib = parent->down;

  while (sib) {
    if (sib != ae) 
      add_node_to_abstract_list (&ael, sib);
    sib = sib->next;
  }

  return (ael);
}

/* scope_of_search == limited ==> only the context
                   == unlimited ==> context and down */

ABSTRACT_EVENT_LIST *evaluate_abstract_list 
  (SYN_TREE_LIST *stl, ABSTRACT_EVENT *context)
{
  ABSTRACT_EVENT_LIST *result = NULL;
  IDENT *id = NULL;
  ExprOper oper;

  if (stl == NULL)
    return NULL;

  oper = stl->st->oper;
  if (oper == ApplyNone) {
    
    if ((stl->st->left != NULL) || (stl->st->right != NULL))
      printf("Ignoring extra arguments to null operation\n");
    

    /* The name can suggest two things: we may have reached the name of the 
       abstract or primitive event, or we may need to look for the name 
       in the given context. It should not be attribute name, since it 
       should be handled separately. */

    id = id_find (stl->st->name);
    
    /* Does ident with name exist? */
    if (id == NULL) {
      printf("Ident node in the expression is NULL!\n");
      return NULL;
    }

    /* Is it a primitive event? */
    if (id->id_reserved == 1) {
      result = get_named_nodes (stl->st->name, context, 0);
      return result;
    }
   
    /* Does the ident have current declaration? */
    if (id->id_decl == NULL) {
      /* May be we need to use the context? */
      if (context == NULL) {
	printf("%s has no current declaration! Fatal Error!\n", id->id_chars);
	return NULL;
      }
    }

    if (!StrCmp (id->id_chars, "event") || 
	(id->id_decl && id->id_decl->decl_kind < DeclMatch))
      result = get_named_nodes (stl->st->name, context, 0);
  }
  else if ( (oper == ApplyPart) || (oper == ApplyExpandPart) )
    result = evaluate_part_relationship (stl, context);
  else 
    result = evaluate_tree_selection (stl, context);

  return result;
}



ABSTRACT_EVENT_LIST *evaluate_tree_selection 
  (SYN_TREE_LIST *stl, ABSTRACT_EVENT *context)
{
  SYN_TREE_LIST *s1, *s2;
  ExprOper oper = stl->st->oper;
  ABSTRACT_EVENT_LIST *result = NULL;
  ABSTRACT_EVENT_LIST *siblings = NULL;
  ABSTRACT_EVENT_LIST *leaf = NULL;
  ABSTRACT_EVENT_LIST *first = NULL;
  ABSTRACT_EVENT_LIST *last = NULL;
  ABSTRACT_EVENT *tmp, *child, *parent;
  int which_child = 0;  /* initialization important */
  int i;
  int len;

  if (stl == NULL)
    return NULL;

  /* Is it ApplyExpandChild? */
  if (oper == ApplyExpandChild) {
    ABSTRACT_EVENT_LIST *left = 
      evaluate_abstract_list (stl->st->left, context);
    ABSTRACT_EVENT_LIST *right =
      evaluate_abstract_list (stl->st->right, context);
    ABSTRACT_EVENT *a1 = get_element_from_abstract_list(left,0);
    ABSTRACT_EVENT *a2 = get_element_from_abstract_list(right,0);
    int by = stl->st->value;
    int i;
    while (a1 != a2) {
      for (i=0; i<by; i++)
	a1 = a1->next;
      if (a1 != a2)
	add_node_to_abstract_list (&left, a1);
    }
    abstract_list_concat (&left, right);
    return left;
  }
    
  s1 = stl->st->left;
  s2 = s1->next;
  s1->next = NULL;

  if (stl->st->right)
    printf("Ignoring extra arguments: evaluate_tree_relation()\n");

  result = evaluate_abstract_list (s1, context);
  if (result == NULL)
    return NULL;

  /* result should identify unique abstract event */
/*  if (get_size_of_abstract_list(result) > 1) {
    printf("Warning: Extra arguments ignored.\n");
    truncate_abstract_list (result,1);
  }
*/
  
  switch (oper) {
  case ApplyChild:
    if (s2)
      which_child = get_int(evaluate_int_list(s2, context));
    tmp = get_element_from_abstract_list(result,0);
    child  = tmp->down;

    for (i=0; (i != which_child) && (child->next != NULL);  
	 i++,  child=child->next);

    result = NULL;
    add_node_to_abstract_list (&result, child);
    return result;

  case ApplyParent:
    tmp = get_element_from_abstract_list (result,0);
    parent = tmp->parent;
    
    result = NULL;
    add_node_to_abstract_list (&result, parent);
    return result;

  case ApplySibling:
    tmp = get_element_from_abstract_list (result, 0);
    siblings = get_all_sibling (tmp);
    result = NULL;
    return siblings;

  case ApplyLeft:
    tmp = get_element_from_abstract_list (result, 0);
    if (tmp)
      add_node_to_abstract_list (&first, tmp->prev);
    return first;
 
  case ApplyRight:
    tmp = get_element_from_abstract_list (result, 0);
    if (tmp)
      add_node_to_abstract_list (&first, tmp->next);
    return first;

  case ApplyLeaf:
    tmp = get_element_from_abstract_list (result, 0);
    leaf = get_all_leaves (tmp);
    result = NULL;
    return leaf;

  case ApplyFirst:
    tmp = get_element_from_abstract_list (result, 0);
    if (tmp)
      add_node_to_abstract_list (&first, tmp);
    return first;

  case ApplyLast:
    len = get_size_of_abstract_list (result);
    tmp = get_element_from_abstract_list (result, len-1);
    if (tmp)
      add_node_to_abstract_list (&last, tmp);
    return last;
  }
}


ABSTRACT_EVENT_LIST 
  *evaluate_part_relationship (SYN_TREE_LIST *stl, ABSTRACT_EVENT *context)
{
  ABSTRACT_EVENT_LIST *result = NULL;
  COLLAPSE_NODE *left, *right, *tmp, *cchild;
  int by, i;

  if (stl == NULL)
    return NULL;
  
  left  = Get_Collapse_Part (stl->st->left, context);
  right = Get_Collapse_Part (stl->st->right, context);
  tmp = left;
  by = stl->st->value;
  result = left->ael;
    
  cchild = left;
  while (cchild != right) {
    for (i=0; i<by; i++)
      cchild = cchild->next;
    if (cchild != right)
      abstract_list_concat (&result, cchild->ael);
  }
  abstract_list_concat (&result, right->ael);
  return result;
}





ABSTRACT_EVENT_LIST 
*get_named_nodes_with_filter (char *name, SYN_TREE_LIST *stl, ABSTRACT_EVENT *context, int i) 
{
  ABSTRACT_EVENT_LIST *head = NULL;
  ABSTRACT_EVENT_LIST *nxt = NULL;
  ABSTRACT_EVENT_LIST *dwn = NULL;
  ABSTRACT_EVENT *ae = NULL;
  int filter_result = 1;

  if (context == NULL)
    return NULL;

  /* Add to the conditional !SteCmp (name, "event") */
  if (!StrCmp (name, context->name) || !StrCmp (name, context->marker) ||
      !StrCmp (name, "event") || 
      (context->ep && !StrCmp (name, context->ep->event_type))) {

    /* Now evaluate the filter expression */
    if (stl != NULL) {
      INT_LINKED_LIST *il = evaluate_int_list (stl, context);
       filter_result = get_int (il);
    }
    if (filter_result == 1)
      add_node_to_abstract_list (&head, context);
    if (i)
      nxt = get_named_nodes_with_filter (name, stl, context->next, i);
    abstract_list_concat (&head, nxt);
    return head;
  }
  else {
    dwn = get_named_nodes_with_filter (name, stl, context->down, 1);
    if (i)
      nxt = get_named_nodes_with_filter (name, stl, context->next, 1);
    abstract_list_concat (&head, dwn);
    abstract_list_concat (&head, nxt);
    return head;
  }
}
      
