/*
 ***********************************************************************
 * FILE:         guii.c                                                *
 * PROGRAMMER:   Joydip Kundu                                          *
 * ORGANIZATION: University of Massachusetts, Amherst                  *
 * FUNCTIONS:                                                          *
 *               + main()                                              *
 *               + create_gui_for_ariadne()                            *
 *               + create_drawing_canvas()                             *
 *               + ave_create_cascade_button()                         *
 *                                                                     *
 * NOTES:        It should be possible to clean up this portion of the *
 *               code We have all widgets visible to every function.   *
 *               This is done to keep the declarations contiguous, and *
 *               make the widget hierarchy apparent. One can redesign  *
 *               it, if needed.                                        *
 *               Can be distributed freely. Please send bug report to  *
 *               kundu@cs.umass.edu                                    *
 ***********************************************************************
*/

#include  "std.h"
#include  <stdio.h>
#include  "xstdio.h"
#include  "structs.h"
#include  "scan.h"
#include  "parse.h"
#include  "parse.tab.h"
#include  "scan.h"
#include  "strs.h"
#include  "ave_interf.h"
#include  "math.h"
#include  <sys/types.h>
#include  <sys/socket.h>
#include "abstract.h"
#include "default.h"
#include "global.h"
#include "support.h"
#include  <X11/Xlib.h>

#include  "Xm/Xm.h"
#include  "Xm/BulletinB.h"
#include  "Xm/CascadeB.h"
#include  "Xm/Form.h"
#include  "Xm/Frame.h"
#include  "Xm/Label.h"
#include  "Xm/List.h"
#include  "Xm/MainW.h"
#include  "Xm/MessageB.h"
#include  "Xm/PushB.h"
#include  "Xm/PushBG.h"
#include  "Xm/RowColumn.h"
#include  "Xm/Scale.h"
#include  "Xm/ScrollBar.h"
#include  "Xm/ScrolledW.h"
#include  "Xm/SelectioB.h"
#include  "Xm/Separator.h"
#include  "Xm/DrawingA.h"
#include  "Xm/Text.h"
#include  "win.h"
#include "DrawingCanvas.h"


extern int NodeId;
extern void ave_mark_nodes (ABSTRACT_EVENT *ae);
extern void AveScanInit();
extern BAR_GRAPH_NODES *GBarGraph;
int bar_shift = 0;
int bar_shift_aggr = 0;



/* Widgets for the main GUI component (query) */
XtAppContext         app_context;
Widget               query_app_shell;
Widget               query_main_window;
Widget               query_form;
Widget         query_command, query_warn, query_frame;
Widget                                    query_swindow;
Widget                                    query_canvas;

/* Selection Box Widgets for the main GUI */
Widget  query_sel_box_cb_show, query_sel_box_cb_view;


/* Widgets for the drawing canvas */
Widget   draw_app_shell;
Widget   draw_main_window;
Widget   draw_form, draw_frame, draw_swindow, draw_canvas;


/* Widgets for the file window */



/* Widgets for the message window */

int g_color_collapsed;
int g_color_error;
int g_color_normal;

/* Scatter Plot Global Variables */
int g_max_color;
int g_scatter_plot_x_sep;
int g_scatter_plot_y_sep;
int g_scatter_plot_width;
int g_scatter_plot_height;
int g_scatter_plot_max_x_dim;
int g_scatter_plot_max_y_dim;

/* Line Plot Global Variable */
int g_line_plot_x_sep;

/* Category Statistics Global Variables */
int g_ctg_stat_x_sep;
int g_ctg_stat_y_sep;
int g_ctg_stat_width;
int g_ctg_stat_height;




/* Global variables needed for keeping state information */
int GLineOrScatter = 0;

FILE *GScriptFile;
XColor  ave_color[MAX_COLOR];
int GShowAttribute = 0;
ABSTRACT_EVENT *GSelectAbstractEvent = NULL;  
/* points to the current node selected */ 
COLLAPSE_NODE *GSelectCollapseNode = NULL;
ABSTRACT_EVENT *GMatchToShow;          /* points to current match variable */
char      *GInputBuf;  /* character buffer, from where yylex() reads input */
int       GInputBufCnt;
FILE      *GCommandFile;  /* file where history of commands are stored */
NAME_LIST *GMatchVarName;  /* List of match variables used by ShowCB */
LINE_DATA *GLines2Draw = NULL;
GC GCanvasGC[2];   /* [0] => main window, [1] => drawing canvas */
XGCValues GValuesGC;

void refresh_query_canvas(Widget w, caddr_t client_data, caddr_t call_data);
void refresh_drawing_canvas(Widget w, caddr_t client_data, caddr_t call_data);
void ave_mark_nodes_all (ABSTRACT_EVENT *ae);
void build_slide_clps (SLIDE_COLLAPSE_NODE *scn, XtEventHandler func);
void build_widget_list (ABSTRACT_EVENT *ae, int thru_part);
void change_widget_coord (ABSTRACT_EVENT *ae);
void ave_calc_line (ABSTRACT_EVENT *ae);
void Query_cb_show_match_variable (Widget w, caddr_t client_data, caddr_t call_data);
typedef struct MenuStruct {
  char *name;
  void (*func) ();
  COLLAPSE_NODE *data;
  struct MenuStruct *sub_menu;
  int  n_sub_items;
  char *sub_menu_title;
} VERT_MENU_ITEM;

void draw_attr()
{
}

void AttrCancelCB ()
{
}



/******************************************************************************
 * Currently we support two canvases: one in the main window, and one as a    *
 * secondary canvas. The canvas in the main window is mainly the protectorate *
 * of various pushbuttons, but in order to show the relationship between the  *
 * buttons we need to draw lines between them.                                *
 * The secondary drawing canvas is solely for drawing bar graphs and like. So *
 * it needs to draw lines (axes), rectangles (bar graph) and text.            *
 *                                                                            *
 * The following 6 routines:                                                  *
 *      add_to_line(), ave_draw_line()                                        *
 *      add_to_rect(), ave_draw_rect()                                        *
 *      add_to_str(),  ave_draw_str()                                         *
 * all manipulates or reads the GCanvasData structure.                        *
 *****************************************************************************/


  
  
int get_max_column (int cnt)
{
  if (cnt < 32)
    return 4;
  if (cnt <= 128)
    return 8;
  if (cnt <= 512)
    return 16;
  return 16;
}

void add_to_line_list(AVE_STR *cave, AVE_STR *pave)
{
  LINE_DATA *ld = (LINE_DATA *) malloc (sizeof (LINE_DATA));
  ld->x1 = cave->x;
  ld->y1 = cave->y;
  ld->w1 = cave->w;
  ld->h1 = cave->h;
  ld->x2 = pave->x;
  ld->y2 = pave->y;
  ld->w2 = pave->w;
  ld->h2 = pave->h;

  ld->next = GLines2Draw;
  GLines2Draw = ld;
}

void ave_draw_line (Widget w, GC gc, LINE_DATA *ld)
{
  if (ld == NULL)
    return;
  XSetForeground (XtDisplay(w), gc, ave_color[0].pixel);
  while (ld) {
    XDrawLine (XtDisplay(w), XtWindow(w), gc,
	       (int) (XSCALE_BUTTON*ld->x1+ld->w1), 
	       (int) (YSCALE_BUTTON*ld->y1+20), 
	       (int) (XSCALE_BUTTON*ld->x2+ld->w2), 
	       (int) (YSCALE_BUTTON*ld->y2+20));
    ld =ld->next;
  }
}




/*
 *******************************************************************
 * FUNCTION:     draw_line_plot()                                  *
 * DESCRIPTION:  called if user clicks on the Line Plot button     *
 * DESIGN:       + Clear the drawing canvas                        *
 *               + Walk through the RectData structure, and plot   *
 *******************************************************************
*/

int ave_draw_line_plot (Widget w, GC gc, RECT_DATA *rd, 
			int x_shift, int y_shift)
{
  int x, y;
  int cnt = 0;

  if (rd == NULL)
    return;

  x = x_shift;

  /* Rectangles are drawn with colors associated with the value */
  while (rd) {
    y = y_shift - g_ctg_stat_y_sep - g_ctg_stat_height*rd->value;
    x += g_ctg_stat_x_sep;
    XSetForeground (XtDisplay(w), gc, ave_color[rd->color].pixel);
    XFillRectangle (XtDisplay(w), XtWindow(w), gc, x, y, 4, 4);

    /* Lines are drawn with the 0-th color */
    if (rd->next) {
      int x1 = x + g_ctg_stat_x_sep;
      int y1 = y_shift - g_ctg_stat_y_sep - g_ctg_stat_height*rd->next->value;
      XSetForeground (XtDisplay(w), gc, ave_color[0].pixel);
      XDrawLine (XtDisplay(w), XtWindow(w), gc, x, y+2, x1, y1+2);
    }
    rd = rd->next;
  }
  return x;
}
    

/*
 *****************************************************************
 * FUNCTION:       ave_draw_scatter_plot()                       *
 * DESCRIPTION:    Draws individual scatter plots.               *
 * DESIGN:                                                       *
 * CALLER:                                                       *
 *****************************************************************
*/

int ave_draw_scatter_plot (Widget w, GC gc, RECT_DATA *rd, 
			    int x_shift, int y_shift)
{
  RECT_DATA *trd = rd;
  int count = 0;
  int colors_seen[MAX_COLOR];
  int i;
  int max_col, max_row;
  int width, height;
  int x_sep, y_sep;
  int x,y;

  if (rd == NULL)  /* sanity check */
    return;

  for (i=0; i<g_max_color; i++)
    colors_seen[i] = 0;

  /* count total number of points in the scatter plot */
  while (trd) {
    ++count;
    trd = trd->next;
  }
  
  max_col = get_max_column(count);
  max_row = ceil(((float) count)/((float) max_col));

  x_sep = g_scatter_plot_x_sep;
  y_sep = g_scatter_plot_y_sep;

  if (max_col < SCATTER_COLUMN_THRESHOLD) {
    width = height = SCATTER_DEFAULT_WIDTH;
  }
  else {
    width  = g_scatter_plot_max_y_dim/(2*max_col);
    height = g_scatter_plot_max_x_dim/(2*max_row);

    if (height < width)  /* make it square for aesthetic reasons */
      width = height;  
    else
      height = width;
  }

  while (rd) {
    XSetForeground (XtDisplay(w), gc, ave_color[rd->color].pixel);

    /* if we are seeing this color for the first time, we need to put it
       in the legend at the top of the scatter plot */
    if (colors_seen[rd->color] == 0) { 
      char tmp[10];
      colors_seen[rd->color] = 1;
      sprintf(tmp, "%d", rd->value);
      x = x_shift + LEGEND_X_DEFAULT;
      y = y_shift - LEGEND_Y_DEFAULT - rd->color*LEGEND_COLOR_DEFAULT;
      XFillRectangle (XtDisplay(w), XtWindow(w), gc, x, y,
		      SCATTER_DEFAULT_WIDTH, SCATTER_DEFAULT_WIDTH);
      x += SCATTER_DEFAULT_WIDTH;
      XDrawString (XtDisplay(w), XtWindow(w), gc, x, y,
		   tmp, strlen(tmp));
    }

    x = x_shift + (x_sep+width)*rd->col_id;
    y = y_shift - (y_sep+height)*rd->row_id;

    /* Now draw the actual rectangles making up the scatter plot */

    XFillRectangle(XtDisplay(w), XtWindow(w), gc, x, y, 
		   SCATTER_DEFAULT_WIDTH,
		   SCATTER_DEFAULT_WIDTH);
       
    rd = rd->next;
  }
  return (x_shift+ ((max_col+2)*(width+g_scatter_plot_x_sep)));
}


int ave_draw_category_statistics (Widget w, GC gc, RECT_DATA *rd, 
				   int x_shift, int y_shift)
{
  int cnt = 0;
  if (rd == NULL)   /* sanity check */
    return;

  while (rd) {
    char tmp[64];
    int x = x_shift + (g_ctg_stat_x_sep + g_ctg_stat_width)*rd->col_id;
    int y = y_shift - g_ctg_stat_y_sep - 
      g_ctg_stat_height*rd->height;

    sprintf(tmp, "%d", rd->height);

    XSetForeground (XtDisplay(w), gc, ave_color[rd->color].pixel);
    XFillRectangle (XtDisplay(w), XtWindow(w), gc, x, y,
		    g_ctg_stat_width, 
		    g_ctg_stat_height*rd->height);
    XDrawString (XtDisplay(w), XtWindow(w), gc, x, y-20, tmp, strlen(tmp));
    rd = rd->next;
    ++cnt;
  }
  x_shift += (cnt+1)*g_ctg_stat_width;
  return x_shift;
}


void ave_draw_error_arrow (Widget w, GC lgc, 
		       int origin_x, int origin_y, int scale)
{
  Display  *dpy = XtDisplay(w);
  Drawable dle = XtWindow(w); 

  XDrawLine (dpy, dle, lgc, origin_x, origin_y, 
	     origin_x+10*scale, origin_y+10*scale);
  XDrawLine (dpy, dle, lgc, origin_x, origin_y, 
	     origin_x+10*scale, origin_y-10*scale);
  XDrawLine (dpy, dle, lgc, origin_x+5*scale, origin_y+5*scale, 
	     origin_x+5*scale+20*scale, origin_y+5*scale);

  XDrawLine (dpy, dle, lgc, origin_x+5*scale, origin_y-5*scale, 
	     origin_x+5*scale+20*scale, origin_y-5*scale);

}
  

void ave_calc_line_slide (SLIDE_COLLAPSE_NODE *scn)
{
  add_to_line_list (scn->ave, scn->prev->parent->ave);
  ave_calc_line (scn->next);
}

void ave_calc_line (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;

  if (ae->ave->draw_flag) {
    if (ae->parent)
      add_to_line_list 	(ae->ave, ae->parent->ave);
    if (ae->clps)
      add_to_line_list(ae->clps->ave, ae->ave);
    ave_calc_line (ae->down);
  }
  if (ae->slide_clps)
    ave_calc_line_slide (ae->slide_clps);
  else
    ave_calc_line(ae->next);
}

void xs_create_menu_buttons (char *title, Widget menu, VERT_MENU_ITEM *menulist,
			     int nitems)
{
  Arg args[3];
  int n;
  int i;
  WidgetList buttons;
  int separators = 0;

  buttons = (WidgetList) XtMalloc (nitems * sizeof (Widget));

  /* If a title is given, create Label and Separator widgets */
  if (title) {
    n = 0;
    XtSetArg (args[n], XmNforeground, ave_color[g_color_collapsed].pixel); n++;
    XtCreateManagedWidget (title, xmLabelWidgetClass, menu, args, n);
    XtCreateManagedWidget ("separator", xmSeparatorWidgetClass, menu, args, n);
  }

  /* Create an entry for each item in the menu */
  for (i=0; i<nitems; i++) {
    /* Null name represents a separator */
    if (menulist[i].name == NULL) {
      n = 0;
      XtSetArg (args[n], XmNforeground, ave_color[COLOR_COLLAPSED].pixel); n++;
      XtCreateManagedWidget("separator", xmSeparatorWidgetClass, menu, args, n);
      separators++;
    }

    /* if there is a name and a callback */
    else if (menulist[i].func) {
      n = 0;
      XtSetArg (args[n], XmNforeground, ave_color[COLOR_COLLAPSED].pixel); n++;
      buttons[i-separators] 
	= XtCreateWidget(menulist[i].name, xmPushButtonWidgetClass, 
			 menu, args, n);
      XtAddCallback(buttons[i-separators], XmNactivateCallback, menulist[i].func,
		    menulist[i].data);
    }

    /* if there is no callback and no submenu */
    else if (!menulist[i].sub_menu) {
      n = 0;
      XtSetArg (args[n], XmNforeground, ave_color[COLOR_COLLAPSED].pixel); n++;
      buttons[i-separators] 
	= XtCreateWidget(menulist[i].name, xmLabelWidgetClass, menu, args, n);
    }
    /* if it has a submenu */
    else {
      Widget sub_menu =
	XmCreatePulldownMenu(menu, menulist[i].sub_menu_title, NULL, 0);
      n = 0;
      XtSetArg (args[n], XmNsubMenuId, sub_menu); n++;
      XtSetArg (args[n], XmNforeground, ave_color[COLOR_COLLAPSED].pixel); n++;
      buttons[i-separators] =
	XtCreateWidget (menulist[i].name, xmCascadeButtonWidgetClass, menu,args, n);
      xs_create_menu_buttons(menulist[i].sub_menu_title, sub_menu,
			     menulist[i].sub_menu, menulist[i].n_sub_items);
    }
  }
  XtManageChildren (buttons, nitems-separators);
}


ABSTRACT_EVENT *get_root_of_tree (ABSTRACT_EVENT *ae)
{
  while (ae->parent)
    ae = ae->parent;
  return ae;
}

ABSTRACT_EVENT *get_select_slider_abstract_event 
(ABSTRACT_EVENT *start, int val,  int base)
{
  int i;
  for (i=base; i<val; i++)
    start = start->next;
  return start;
}

void destroy_slide_widget (SLIDE_COLLAPSE_NODE *scn)
{
  if (scn->winfo) {
    XtUnmanageChild (scn->winfo->w);
    XtDestroyWidget (scn->winfo->w);
    free (scn->winfo);
    scn->winfo = NULL;
  }
}

void ave_free_line()
{
  LINE_DATA *ld = GLines2Draw;
  if (ld == NULL)
    return;

  while (ld) {
    LINE_DATA *tld = ld;
    ld = ld->next;
    tld->next = NULL;
    free(tld);
  }
  GLines2Draw = NULL;
}

    
void ButtonSelect (Widget w, caddr_t client_data, caddr_t call_data)
{
  void *tmp;
  ABSTRACT_EVENT *ae;
  IDENT *id = id_find("select");

  XtVaGetValues (w, XmNuserData, &tmp, NULL);
  ae = (ABSTRACT_EVENT *) tmp;
  set_select_abstract_event (ae);
  set_select_collapse_node (NULL);
  id->id_decl->decl_object = (OBJECT *) ae;
}
    

void handle_slider (Widget w, caddr_t client_data, XEvent *event)
{
  int val;
  void *tmp;
  ABSTRACT_EVENT *ae;
  ABSTRACT_EVENT *par;
  SLIDE_COLLAPSE_NODE *scn;
  SLIDE_COLLAPSE_NODE *new = NULL;
  int low;
  int high;
  XtVaGetValues (w, XmNuserData, &tmp, NULL);
  scn = (SLIDE_COLLAPSE_NODE *) tmp;
  low = scn->lo+1;
  high = scn->hi-1;
  XmScaleGetValue (w, &val);

  if ( (val == scn->lo) || (val == scn->hi) )
    return;
  /* Do we need another slider? */
  if ((val>low) && (val<high) && ((high-low) > 2)) {
    new = allocate_slide_collapse_node(scn->hi-val, val, scn->hi );
    ae = get_select_slider_abstract_event (scn->prev->next, val, scn->lo);
    new->prev = ae;
    new->next = scn->next;
    new->ave  = allocate_ave_str ();
    new->ave->draw_flag = 1;
    ae->slide_clps = new;


    scn->hi = val;
    scn->next = ae;
  }
  else if (val == low) {
    int diff = scn->hi - val;
    ae = scn->prev->next;
    scn->prev->slide_clps = NULL;
    if (diff > 1) {
      ae->slide_clps = scn;
      scn->lo = val;
      scn->prev = ae;
    }
    else {
      scn->prev = NULL;
      scn->next = NULL;
      destroy_slide_widget (scn);
      free(scn);
    }
  }
  else if (val == high) {
    int diff = val-scn->lo;
    ae = scn->next->prev;
    if (diff > 1) {
      scn->hi = val;
      scn->next = ae;
    }
    else {
      scn->prev->slide_clps = NULL;
      scn->next = NULL;
      destroy_slide_widget (scn);
      free(scn);
    }
  }

  
  ae->ave->draw_flag = 1;

  ave_mark_nodes (ae);
  build_widget_list(ae, 1);
  par = get_root_of_tree (ae);
  init_coordinate();
  ave_calc_coord(par, 1);
  if (new)
    build_slide_clps (new, (XtEventHandler) handle_slider);
  change_widget_coord (par);
  ave_free_line();
  ave_calc_line (par);
  XClearWindow(XtDisplay(query_canvas), XtWindow(query_canvas));
  ave_draw_line(query_canvas, GCanvasGC[0], GLines2Draw);
  ave_draw_tree (par, 1);
}
  
  

WIDGET_INFO *allocate_winfo ()
{
  WIDGET_INFO *winfo = (WIDGET_INFO *) malloc (sizeof (WIDGET_INFO));
  winfo->height = 30;
  winfo->width = 0;
  return winfo;
}

void build_slide_clps (SLIDE_COLLAPSE_NODE *scn, XtEventHandler func)
{
  Arg args[12];
  int n;

  if (scn == NULL)
    return;

  if (scn->winfo != NULL)
    return;

  scn->winfo = allocate_winfo();
  n = 0;
/*  XtSetArg (args[n], XmNunitType, Xm100TH_MILLIMETERS); n++; */
  XtSetArg (args[n], XmNx, XSCALE_SLIDER*scn->ave->x + XSCALE_SHIFT); n++;
  XtSetArg (args[n], XmNy, YSCALE_SLIDER*scn->ave->y); n++;
  XtSetArg (args[n], XmNwidth, AVE_SLIDER_WIDTH); n++;
  XtSetArg (args[n], XmNshowValue, True); n++;
  XtSetArg (args[n], XmNmaximum, scn->hi); n++;
  XtSetArg (args[n], XmNminimum, scn->lo); n++;
  XtSetArg (args[n], XmNorientation, XmHORIZONTAL); n++;
  XtSetArg (args[n], XmNscaleMultiple, 1); n++; 
  XtSetArg (args[n], XmNprocessingDirection, XmMAX_ON_RIGHT); n++;
  XtSetArg (args[n], XmNtitleString, 
	    XmStringCreateSimple(scn->prev->orig)); n++;
  XtSetArg (args[n], XmNvalue, scn->lo); n++;
  scn->winfo->w = XmCreateScale (query_canvas, scn->prev->orig, args, n);
  XtVaSetValues (scn->winfo->w, XmNuserData, (void *) scn, NULL);
  scn->winfo->width = AVE_SLIDER_WIDTH-10;
  XtAddEventHandler 
    (scn->winfo->w, ButtonPressMask, FALSE, func, NULL); 
    
}

void ExpandVert (Widget w, caddr_t client_data, caddr_t call_data)
{
  Widget expand_vert;
  Widget button[3];
  Arg args[10];
  int n;

  expand_vert = XtCreateManagedWidget ("Choose One", xmRowColumnWidgetClass,
				       w, NULL, 0);
  button[0] = XmCreatePushButton (expand_vert, "Show All", NULL, 0);
  XtManageChild (button[0]);
  XtManageChild (expand_vert);
}

char *get_proc_set (ABSTRACT_EVENT_LIST *ael)
{
  int cnt = 0;
  int i;
  char *set;
  char tmp[MAX_PROC_DIGIT+2];
  ABSTRACT_EVENT_LIST *t_ael = ael;
  while (t_ael) {
    ++cnt;
    t_ael = t_ael->next;
  }
  set = StrAlloc (cnt*MAX_PROC_DIGIT +2*cnt + 3);
  t_ael = ael;
  sprintf (set, " = { ");
  while (t_ael) {
    for (i=0; i<6 && t_ael; i++) {
      ABSTRACT_EVENT *ae = get_element_from_abstract_list (t_ael, 0);
      sprintf(tmp, "%d ", ae->proc);
      strcat (set, tmp);
      t_ael = t_ael->next;
    }
    strcat (set, "\n");
  }
  strcat (set, "}");
  return set;
}
  

      
void ShowVertInfo (Widget w, COLLAPSE_NODE *client_data, caddr_t call_data)
{
  int i, n;
  Widget show_vert_info, show_list;
  Arg args[6];
  COLLAPSE_NODE *cc = client_data->down;

  show_vert_info = 
    XmCreateSelectionDialog (query_app_shell, "Partition Info", NULL, 0);
  show_list = XmSelectionBoxGetChild (show_vert_info, XmDIALOG_LIST);
  while (cc) {
    char *list = get_proc_set (cc->ael);
    XmString new = 
      XmStringConcat (XmStringCreateSimple(cc->pattern),
		      XmStringCreateSimple(list));
    XmListAddItem (show_list,  new, 0);
    cc = cc->next;
  }
  XtManageChild (show_vert_info);
  return;
}





void mark_slide_nodes_all (SLIDE_COLLAPSE_NODE *scn)
{
  scn->ave->draw_flag = 1;
  ave_mark_nodes_all (scn->next);
  return;
}

void ave_mark_nodes_all (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  ae->ave->draw_flag = 1;
  if (ae->slide_clps)
    mark_slide_nodes_all (ae->slide_clps);
  else
    ave_mark_nodes_all (ae->next);

  ave_mark_nodes_all (ae->down);
}



void change_widget_slide_coord (SLIDE_COLLAPSE_NODE *scn)
{
  Arg args[4];
  int n;
  n = 0;
  XtSetArg (args[n], XmNx, XSCALE_SLIDER*scn->ave->x - 0.5*scn->winfo->width +XSCALE_SHIFT); n++;
  XtSetArg (args[n], XmNy, YSCALE_SLIDER*scn->ave->y - 0.5*scn->winfo->height); n++;
  XtSetArg (args[n], XmNmaximum, scn->hi); n++;
  XtSetArg (args[n], XmNminimum, scn->lo); n++;
  XtSetValues (scn->winfo->w, args, n);

  change_widget_coord(scn->next);
}

void change_horiz_clps_coord (COLLAPSE_NODE *cn)
{
  Arg args[2];
  int n;
  
  if (cn == NULL)
    return;
  if (cn->winfo == NULL)
    return;

  n = 0;

  XtSetArg 
    (args[n], XmNx, XSCALE_BUTTON*cn->ave->x - 0.5*cn->winfo->width); n++;
  XtSetArg 
    (args[n], XmNy, YSCALE_BUTTON*cn->ave->y - 0.5*cn->winfo->height);n++;
  XtSetValues (cn->winfo->w, args, n);
  if (cn->ael)
    change_horiz_clps_coord (cn->next);
  change_horiz_clps_coord(cn->down);
}

void change_widget_coord (ABSTRACT_EVENT *ae)
{
  Arg args[2];
  int n;
  if (ae == NULL)
    return;
  
  if (ae->ave->draw_flag == 0)
    return;
  n = 0;
  XtSetArg (args[n], XmNx, XSCALE_BUTTON*ae->ave->x); n++;
  XtSetArg (args[n], XmNy, YSCALE_BUTTON*ae->ave->y); n++;
  XtSetValues (ae->winfo->w, args, n);
  if (ae->slide_clps)
    change_widget_slide_coord (ae->slide_clps);
  else
    change_widget_coord(ae->next);

  if (ae->clps)
    change_horiz_clps_coord (ae->clps);
  change_widget_coord(ae->down);
}

void DrawAll (Widget w, COLLAPSE_NODE *client_data, caddr_t call_data)
{
  ABSTRACT_EVENT *ae = client_data->pae;
  ABSTRACT_EVENT *par = get_root_of_tree (ae);
  ave_mark_nodes_all (ae);
  build_widget_list(ae, 1);
  init_coordinate();
  ave_calc_coord (par, 1);
  change_widget_coord (par);
  ave_free_line();
  ave_calc_line (par);
  ave_draw_line(query_canvas, GCanvasGC[0], GLines2Draw);
  ave_draw_tree (par, 1);
  return;
}

void SelectSome (Widget w, COLLAPSE_NODE *client_data, caddr_t call_data)
{
  ABSTRACT_EVENT *ae = client_data->pae;
  ABSTRACT_EVENT *par = get_root_of_tree (ae);
}

void SelectPart ()
{
}

void Collapsed_cb_select (Widget w, caddr_t client_data, caddr_t call_data)
{
  void *tmp;
  COLLAPSE_NODE *cn;
  IDENT *id = id_find("select");

  XtVaGetValues (w, XmNuserData, &tmp, NULL);
  cn = (COLLAPSE_NODE *) tmp;
  set_select_collapse_node (cn);
  set_select_abstract_event(NULL);
  ShowVertInfo (w, cn, call_data);
}

  
  
  
void build_clps_widget (COLLAPSE_NODE *cn)
{
  Arg args[3];
  int n = 0;
  

  /* Check for Null pointer */
  if (cn == 0)
    return;

  /* Allocate a widget info structure */
  if (cn->winfo == NULL && (cn->ave->draw_flag == 1) ) {
    char str[64];
    cn->winfo = allocate_winfo();
    sprintf(str, "(%d,%d)", cn->total, cn->part);
    /* Set up the collapsed node. Color to be added later */
    n = 0;

    XtSetArg (args[n], XmNx, XSCALE_BUTTON*cn->ave->x - 10); n++;
    XtSetArg (args[n], XmNy, YSCALE_BUTTON*cn->ave->y); n++;
/*    XtSetArg (args[n], XmNborderWidth, 1); n++;  */
    XtSetArg (args[n], XmNforeground, ave_color[g_color_collapsed].pixel); n++;
    cn->winfo->w = 
      XmCreatePushButton (query_canvas, str, args, n);
    XtAddCallback (cn->winfo->w, XmNactivateCallback, 
		   (XtCallbackProc) Collapsed_cb_select, NULL);
    XtVaSetValues (cn->winfo->w, XmNuserData, (void *) cn, NULL);
  }

}

void build_widget_list (ABSTRACT_EVENT *ae, int thru_part)
{
  int n;
  int clr_index;
  Arg args[3];
  Dimension width;
  int x;

  if (ae == NULL)
    return;
  if (ae->error)
    clr_index = g_color_error;
  else
    clr_index = g_color_normal;
  if ((ae->winfo == NULL) && (ae->ave->draw_flag == 1)) {
    ae->winfo = allocate_winfo();
    n = 0;
/*    XtSetArg (args[n], XmNunitType, Xm100TH_MILLIMETERS); n++; */
    XtSetArg (args[n], XmNx, XSCALE_BUTTON*ae->ave->x); n++;
    XtSetArg (args[n], XmNy, YSCALE_BUTTON*ae->ave->y); n++;
    XtSetArg (args[n], XmNforeground, ave_color[clr_index].pixel); n++;
    ae->winfo->w = XmCreatePushButton (query_canvas, ae->name, args, n);
    XtVaSetValues (ae->winfo->w, XmNuserData, (void *) ae, NULL);
    XtAddCallback
      (ae->winfo->w, XmNactivateCallback, (XtCallbackProc) ButtonSelect, NULL);
    XtVaGetValues (ae->winfo->w, XmNwidth, &width, NULL);
/*    printf("width resd %d\n", (int) width);
    ae->winfo->width = 16*strlen(ae->name);   */
    ae->winfo->width = (int) width;

    /* one man's hack is other man's heuristics */
    x = (int) width - ((int)width % 20);  
    
    ae->ave->w = x/2;
  }
  build_clps_widget (ae->clps);
  if (thru_part)
    build_slide_clps(ae->slide_clps, (XtEventHandler) handle_slider);
  build_widget_list (ae->down, 1);
  if (thru_part)
    build_widget_list (ae->next, 1);
}

void destroy_collapse_widget (COLLAPSE_NODE *cn)
{
  if (cn == NULL)
    return;

  if (cn->winfo) {
    XtUnmanageChild (cn->winfo->w);
    XtDestroyWidget (cn->winfo->w);
    free(cn->winfo);
    cn->winfo = NULL;
  }
  return;
}

void destroy_slide_clps_widget (SLIDE_COLLAPSE_NODE *scn)
{
  if (scn == NULL)
    return;

  if (scn->winfo) {
    XtUnmanageChild (scn->winfo->w);
    XtDestroyWidget (scn->winfo->w);
    free(scn->winfo);
    scn->winfo = NULL;
  }
  return;
}

void destroy_widget (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  
  if (ae->winfo) {
    XtUnmanageChild (ae->winfo->w);
    XtDestroyWidget (ae->winfo->w);
    free(ae->winfo);
    ae->winfo = NULL;
  }
  
  destroy_collapse_widget (ae->clps);
  destroy_slide_clps_widget (ae->slide_clps);
  destroy_widget(ae->next);
  destroy_widget(ae->down);
}

void manage_all_widget (ABSTRACT_EVENT *ae)
{
  ave_calc_line (ae);
  ave_draw_line (query_canvas, GCanvasGC[0], GLines2Draw);
  ave_draw_tree (ae, 1);
}
char *get_text (XmString compound)
{
  XmStringContext context;
  char *text;
  XmStringCharSet charset;
  XmStringDirection dir;
  Boolean separator;
  char *buf;
  
  XmStringInitContext (&context, compound);
  XmStringGetNextSegment(context, &text, &charset, &dir, &separator);
  buf = StrAlloc (strlen(text)+2);
  sprintf(buf,"%s", text);
  XtFree(text);
  XmStringFreeContext (context);
  return buf;
}


  
void show_var (Widget w, caddr_t client_data, XmSelectionBoxCallbackStruct *call_data)
{
  char *name = get_text (call_data->value);
  
  IDENT *id;
  ABSTRACT_EVENT *ae;
  id = id_find (name);
  ae = (ABSTRACT_EVENT *) id->id_decl->decl_object;
  if (ae == NULL)
    return;
  
  destroy_widget (GMatchToShow);
  XClearWindow (XtDisplay(query_canvas), XtWindow(query_canvas));
  ave_free_line();
  GMatchToShow = ae;
  build_widget_list (ae, 1);
  manage_all_widget (ae);

}



/*************************************************************************
 * ShowCancelCB: Cancel Button of the slection box associated with the   *
 *               Show button in the main window. Just destroys the selec-*
 *               tion dialog box.                                        *
 ************************************************************************/

void ShowCancelCB ()
{
  XtUnmanageChild (query_sel_box_cb_show);
  XtDestroyWidget (query_sel_box_cb_show);
}

/****************************************************************************
 *   ShowCB: Pressing this should pop up a selection box window, with the   *
 *           currently available match variables. The user can select one   *
 *           of them. The effect is similar to issuing the command:         *
 *             "show <selected match variable>"                             *
 *           in the command window.                                         *
 ***************************************************************************/

void Query_cb_show_match_variable (Widget w, caddr_t client_data, caddr_t call_data)
{
  int i, n;
  Widget match_list;
  Widget show_apply, show_ok, show_cancel, show_list;
  Arg args[6];
  NAME_LIST *nl = GMatchVarName;
  char *title = "Match Vars";

  /* If no match variable exists, then do not create the selection box */
  if (nl == NULL) {
    XmTextSetString (query_warn, "No match variables available");
    /* Print a warning at the warn window */
    return;
  }

  n = 0;
  /*  XtSetArg (args[n], XmNdialogTitle, (XmString *) title); n++; */
  query_sel_box_cb_show = 
    XmCreateSelectionDialog (w, XmStringCreateSimple (title), args, n); 

  show_apply   = XmSelectionBoxGetChild (query_sel_box_cb_show, XmDIALOG_APPLY_BUTTON);
  XtAddCallback 
    (query_sel_box_cb_show, XmNapplyCallback, (XtCallbackProc)show_var, GMatchToShow);

  show_cancel = XmSelectionBoxGetChild (query_sel_box_cb_show, XmDIALOG_CANCEL_BUTTON);
  XtAddCallback 
    (query_sel_box_cb_show, XmNcancelCallback, (XtCallbackProc) ShowCancelCB, NULL);

  show_ok = XmSelectionBoxGetChild (query_sel_box_cb_show, XmDIALOG_OK_BUTTON);
  XtAddCallback 
    (query_sel_box_cb_show, XmNokCallback, (XtCallbackProc) show_var, GMatchToShow);

  show_list = XmSelectionBoxGetChild (query_sel_box_cb_show, XmDIALOG_LIST);
  
  while (nl) {
    XmListAddItem (show_list, XmStringCreateSimple(nl->id->id_chars), 0);
    nl = nl->next;
  }
  
  XtManageChild (query_sel_box_cb_show);
  return;
}



/**************************************************************************
 * CancelCB: Ignore the input string. Just reinitialize the text window.  *
 *************************************************************************/

void Query_cb_cancel  (Widget w, caddr_t client_data, caddr_t call_data)
{
  XmTextSetString(query_command, (char *) NULL);
}

/* Function to draw rectangles in the main canvas. Note that for each node in 
   the ave_canvas, there are 100x100 pixels to draw any rectangles. So we need 
   to scale things appropriately. Thus first calculate the number of rectangles
   that need to be drawn */

void ave_draw_rect_ave (Widget w, GC gc, RECT_DATA *rd, ABSTRACT_EVENT *ae)
{
  RECT_DATA *trd = rd;
  int x_shift = XSCALE_BUTTON*ae->ave->x + 60;
  int y_shift = YSCALE_BUTTON*ae->ave->y;
  int x_sep, y_sep, width, height;
  int cnt = 0;
  int max_col;
  int max_row;
  int x_cnt = 0;

  while (trd) {
    ++cnt;
    trd = trd->next;
  }

/*  max_col = ceil((double) sqrt(cnt)); */
  max_col = get_max_column (cnt);
  max_row = ceil( ((float) cnt)/( (float) max_col));

  /* default separation is 10 pixels, otherwise it is 100/(2*sqrt(cnt)) */
  if (max_col < 5) {
    height = width = 10;
    x_sep = width/2;
    y_sep = height/2;
  }
  else {
    width = 100/(2*max_col);
    height = 90/(2*max_row);

    if (height < width)
      width = height;
    else
      height = width;
    x_sep = width/2;
    y_sep = height/2;
  }

  
  while (rd) {
    XSetForeground (XtDisplay(w), gc, ave_color[rd->color].pixel);
    XFillRectangle (XtDisplay(w), XtWindow(w), gc,
		    x_shift+(x_sep+width)*rd->col_id, y_shift+ (y_sep+height)
		    *rd->row_id, width, height);
    x_cnt++;
    rd = rd->next;
  }
}
    
void render_query_canvas ()
{
  BAR_GRAPH_NODES *bg = GBarGraph;

  while (bg) {
    if (bg->xae->ave->draw_flag)
      ave_draw_rect_ave(query_canvas, GCanvasGC[0], bg->discrete, bg->xae);
    bg = bg->next;
  }
}
    
/* This is to draw attribute in the draw_canvas */
typedef struct location {
  int x;
  int y;
} LOC;

void render_drawing_canvas ()
{
  int discrete_bar_shift = 0;
  int category_bar_shift = 0;
  char *scatter = "Scatter Plot";
  char *stat = "Category Statistics";
  char *line = "Line Plot";

  LOC origin_disc = {30, 300};
  LOC title_disc  = {200, 40};
  LOC yaxis_disc  = {170, 345};
  LOC disc_arr_start   = {200, 340};
  LOC disc_arr_end     = {220, 340};
  LOC disc_arr_left    = {215, 335};
  LOC disc_arr_right   = {215, 345};

  int height_disc =  256;
  int width_disc  =  512;
  
  int x_shift, shift;
  LOC origin_aggr  = {30, 800};
  LOC title_aggr   = {200, 440};  /* changed for paper (50,600) for seq */
  LOC yaxis_aggr   = {170, 845};
  LOC aggr_arr_start   = {200, 840};
  LOC aggr_arr_end     = {220, 840};
  LOC aggr_arr_left    = {215, 835};
  LOC aggr_arr_right   = {215, 845};

  int height_aggr  = 300;
  int width_aggr   = 512;

  Display  *dpy = XtDisplay(draw_canvas);
  Drawable dle = XtWindow(draw_canvas); 
  GC lgc = GCanvasGC[1];

  BAR_GRAPH_NODES *bg = GBarGraph;
  
  XSetForeground (dpy, lgc, ave_color[0].pixel);

  /* draw the scatter plot co-ordinate axes */
  XDrawLine (dpy, dle, lgc, origin_disc.x, origin_disc.y, 
	     origin_disc.x+width_disc, origin_disc.y);
  XDrawLine (dpy, dle, lgc, origin_disc.x, origin_disc.y, 
	     origin_disc.x, origin_disc.y-height_disc);
	     
  /* Draw scatter/line plot title */
  if (GLineOrScatter == 0)
    XDrawString (dpy, dle, lgc, title_disc.x, title_disc.y,
		 scatter, strlen(scatter));
  else
    XDrawString (dpy, dle, lgc, title_disc.x, title_disc.y,
		 line, strlen(line));

  
  /* Draw scatter plot y-arrow */
  XDrawLine (dpy, dle, lgc, disc_arr_start.x, disc_arr_start.y,
	     disc_arr_end.x, disc_arr_end.y);
  XDrawLine (dpy, dle, lgc, disc_arr_end.x, disc_arr_end.y,
	     disc_arr_left.x, disc_arr_left.y);
  XDrawLine (dpy, dle, lgc, disc_arr_end.x, disc_arr_end.y,
	     disc_arr_right.x, disc_arr_right.y);

  /* Draw scatter plot y-title */
  if (bg != NULL)
    XDrawString (dpy, dle, lgc, yaxis_disc.x, yaxis_disc.y,
		 bg->xae->name, strlen(bg->xae->name));


  /* draw the frequency plot co-ordinate axes */
  XDrawLine (dpy, dle, lgc, origin_aggr.x, origin_aggr.y, 
	     origin_aggr.x+width_aggr, origin_aggr.y);
  XDrawLine (dpy, dle, lgc, origin_aggr.x, origin_aggr.y, 
	     origin_aggr.x, origin_aggr.y-height_aggr);
	     
  /* Draw frequency plot title */
  XDrawString (dpy, dle, lgc, title_aggr.x, title_aggr.y,
	       stat, strlen(stat));

  
  /* Draw frequency plot y-arrow */
  XDrawLine (dpy, dle, lgc, aggr_arr_start.x, aggr_arr_start.y,
	     aggr_arr_end.x, aggr_arr_end.y);
  XDrawLine (dpy, dle, lgc, aggr_arr_end.x, aggr_arr_end.y,
	     aggr_arr_left.x, aggr_arr_left.y);
  XDrawLine (dpy, dle, lgc, aggr_arr_end.x, aggr_arr_end.y,
	     aggr_arr_right.x, aggr_arr_right.y);

  /* Draw frequency plot y-title */
  if (bg != NULL)
    XDrawString (dpy, dle, lgc, yaxis_aggr.x, yaxis_aggr.y,
		 bg->xae->name, strlen(bg->xae->name));

  
  x_shift = origin_disc.x;
  shift = 0;
  while (bg) {
    x_shift = origin_disc.x + shift;
    if (GLineOrScatter == 0)
      shift = ave_draw_scatter_plot (draw_canvas, GCanvasGC[1], bg->discrete, 
				     x_shift, origin_disc.y);
    else
      shift = ave_draw_line_plot (draw_canvas, GCanvasGC[1], bg->discrete,
				  x_shift, origin_disc.y);
    bg = bg->next;
  }
  
  
  bg = GBarGraph;
  x_shift = origin_aggr.x;
  shift = 0;
  while (bg) {
    x_shift = origin_aggr.x + shift;
    shift = ave_draw_category_statistics (draw_canvas, GCanvasGC[1], bg->aggr, 
					  x_shift, origin_aggr.y);
    bg = bg->next;
  }
}
      
	    

/****************************************************************************
 *  Query_cb_execute: Gathers the user input in a global buffer, and calls yyparse *
 ****************************************************************************/

void Query_cb_execute (Widget w, caddr_t client_data, caddr_t call_data)
{
  char tmp[80];
  char *quit = "\n quit;";
  GInputBufCnt = 0;
  sprintf (GInputBuf,"%s", (char *) XmTextGetString (query_command));
  strcat (GInputBuf, quit);
  printf("%s\n", GInputBuf);
  XmTextSetString(query_warn, (char *) NULL);
  AveScanInit();
  yyparse();
  XmTextSetString(query_command, (char *) NULL);

  /* The parser may put a file pointer in the GScriptFile. In that 
     case, we have to loop before we do things */
  if (GScriptFile != NULL) {
    GInputBuf[0] = '\0';
    while (fscanf(GScriptFile, "%s\n", tmp) != -1) 
      strcat (GInputBuf, tmp);
    strcat (GInputBuf, "\n quit;");
    AveScanInit();
    yyparse();
    fclose(GScriptFile);
    GScriptFile = NULL;
  }
    
  if (get_show_attribute_flag() == 1) {
    set_show_attribute_flag(0);
    refresh_query_canvas(w, client_data, call_data);
    refresh_drawing_canvas(w, client_data, call_data);
  }
}


void WarnCB (Widget w, caddr_t client_data, caddr_t call_data)
{
}

void Query_cb_view_partitions (Widget w, caddr_t client_data, caddr_t call_data)
{
}
  
void Query_cb_file (Widget w, caddr_t client_data, caddr_t call_data)
{
  create_file_window(w);
}

/********************************
 * QuitCB: Quit the application.*
 ********************************/

void QuitCB (Widget w, caddr_t client_data, caddr_t call_data)
{
  exit(0);
}



/**********************************************************************
 * create_gc: Creates the GC for the canvas: the only GC we will need *
 **********************************************************************/

GC create_gc (Widget w)
{
  XGCValues values;
  GC gc;
  values.foreground = 1;
  values.background = 0;
  gc = XtGetGC (w, GCForeground | GCBackground, &values);
  return gc;
}

void AveInitialize()
{
  GCommandFile = fopen ("AveFile", "rw");
  GInputBuf = StrAlloc (1000);
  GInputBuf[0] = '\0';
  GMatchToShow = NULL;
  GMatchVarName = NULL;
  scaninit();
  parseinit();
}


/*
 ***************************************************************************
 * FUNCTION:     ave_create_cascade_button()                               *
 * DESCRIPTION:  Creates a simple cascade button in a menubar              *
 * NOTES:        Used instead of two calls to macros at every place a      *
 *               cascade button is required. It may make the code a bit    *
 *               slow. But it is only invoked at the initialization phase, *
 *               and speed trade-off with readability is worth it.         *
 ***************************************************************************
*/
  
void ave_create_cascade_button (Widget menu_bar, char *name, Arg *args, int n,
				XtCallbackProc func, XtPointer client_data)
{
  Widget   cascade = XmCreateCascadeButton (menu_bar, name, args, n);
  XtManageChild (cascade);
  XtAddCallback (cascade, XmNactivateCallback, func, client_data);
}


void ave_load_color (Widget w)
{
  Display *dpy  = XtDisplay(w);
  int scr       = DefaultScreen(dpy);
  Colormap cmap = DefaultColormap (dpy, scr);
  XColor ignore;
  FILE *fcolor = fopen ("ave.color.map", "r");
  int i;
  char *default_color[8] = {"SkyBlue", "Red", "Green", "Cyan", 
			       "Magenta", "aquamarine", "Black", "Gold"};
  
  if (fcolor == NULL) {
    printf("Ave expects colors to be specified in ave.color.map file\n");
    printf("Using default colors ...\n");
    for (i=0; i<g_max_color; i++)
      XAllocNamedColor (dpy, cmap, default_color[i], &ave_color[i], &ignore);
    return;
  }

  for (i=0; i<g_max_color; i++) {
    char str[64];
    int r = fscanf(fcolor, "%s", &str);
    if (r == -1)
      XAllocNamedColor (dpy, cmap, default_color[i], &ave_color[i], &ignore);
    else
      XAllocNamedColor (dpy, cmap, str, &ave_color[i], &ignore);
  }
}




void refresh_drawing_canvas(Widget w, caddr_t client_data, caddr_t call_data)
{
  bar_shift = 0;
  bar_shift_aggr = 0;
  XClearWindow(XtDisplay(draw_canvas), XtWindow(draw_canvas));
  render_drawing_canvas();
}

void refresh_query_canvas (Widget w, caddr_t client_data, caddr_t call_data)
{
  XClearWindow (XtDisplay(query_canvas), XtWindow(query_canvas));
  render_query_canvas();
  ave_draw_line(query_canvas, GCanvasGC[0], GLines2Draw);
}

void Draw_cb_scatter (Widget w, caddr_t client_data, caddr_t call_data)
{
  GLineOrScatter = 0;
  refresh_drawing_canvas(w, client_data, call_data);
}

void Draw_cb_line (Widget w, caddr_t client_data, caddr_t call_data)
{
  GLineOrScatter = 1;
  refresh_drawing_canvas(w, client_data, call_data);
}


  
void create_drawing_canvas (Widget parent)
{
  int n;
  int stat;
  Arg args[10];
  Widget menu_bar, cascade;
  
  n = 0;
  draw_app_shell = 
    XtAppCreateShell ("Ave: Drawing Canvas", "DrawWindow", 
		      applicationShellWidgetClass, XtDisplay(parent),
		      NULL, 0);


  /* Create main window */
  n = 0;
  draw_main_window = XmCreateMainWindow (draw_app_shell, "Ave", args, n);
  XtManageChild (draw_main_window);

  /* Create menubar in mainwindow */
  n = 0;
  menu_bar = XmCreateMenuBar (draw_main_window, "menu_bar", args, n);
  XtManageChild (menu_bar);

  
  n = 0;
  ave_create_cascade_button 
    (menu_bar,"Scatter",args, n, (XtCallbackProc) Draw_cb_scatter, NULL);

  ave_create_cascade_button 
    (menu_bar, "Line", args, n, (XtCallbackProc) Draw_cb_line, NULL);

  n = 0;
  XtSetArg (args[n], XmNlistVisibleItemCount, 4); n++;
  draw_form = XmCreateForm (draw_main_window, "form", args, n);
  XtManageChild (draw_form);


  /* Create Frame MainWindow and ScrolledWindow in Frame */
  n = 0;
  XtSetArg (args[n], XmNmarginWidth, 2); n++;
  XtSetArg (args[n], XmNmarginHeight, 2); n++;
  XtSetArg (args[n], XmNshadowThickness, 1); n++;
  XtSetArg (args[n], XmNshadowType, XmSHADOW_OUT); n++;
  draw_frame = XmCreateFrame (draw_form, "frame", args, n);
  XtManageChild (draw_frame);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
  XtSetValues (draw_frame, args, n);

  n = 0;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
  XtSetArg (args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
  draw_swindow = XmCreateScrolledWindow (draw_frame, "swindow", args, n);
  XtManageChild (draw_swindow);

  n = 0;
  XtSetArg (args[n], XmNwidth, 10000); n++;
  XtSetArg (args[n], XmNheight, 10000); n++;
  XtSetArg (args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  draw_canvas = XmCreateDrawingArea (draw_swindow, "canvas", args, n);
  XtManageChild(draw_canvas);
  XtAddCallback
    (draw_canvas, XmNexposeCallback, (XtCallbackProc) refresh_drawing_canvas, NULL);

  GCanvasGC[1] = create_gc (draw_canvas);
  ave_load_color (draw_canvas);

  n = 0;
  XtSetArg (args[n], XmNworkWindow, draw_canvas); n++;
  XtSetValues (draw_swindow, args, n);

  XmMainWindowSetAreas (draw_main_window, menu_bar, NULL, NULL, NULL, draw_form);
  XtRealizeWidget (draw_app_shell);

}


void create_gui_for_ariadne (Display *display)
{
  Widget  menu_bar, menu_pane, cascade; 
  Widget  button;
  Arg args[MAX_ARGS];
  int n;


  /* Create application shell */
  query_app_shell = 
    XtAppCreateShell ("Ave", "XMdemos", applicationShellWidgetClass,
		      display, NULL, 0);


  /* Create main window */
  n = 0;
  query_main_window = XmCreateMainWindow (query_app_shell, "Ave", args, n);
  XtManageChild (query_main_window);
  

  /* Create menubar in mainwindow */
  n = 0;
  menu_bar = XmCreateMenuBar (query_main_window, "menu_bar", args, n);
  XtManageChild (menu_bar);


  /* Create "Exit" PulldownMenu */
  n = 0;
  menu_pane = XmCreatePulldownMenu (menu_bar, "menu_pane", args, n);

  /* Function calls,not macro -- No need to reinitialize "n" every time */
  n = 0;
  ave_create_cascade_button 
    (menu_bar, "Show", args, n, (XtCallbackProc) Query_cb_show_match_variable, NULL);

  ave_create_cascade_button 
    (menu_bar, "Execute", args, n, (XtCallbackProc) Query_cb_execute, NULL);

  ave_create_cascade_button 
    (menu_bar, "Cancel", args, n, (XtCallbackProc) Query_cb_cancel, NULL);

  ave_create_cascade_button 
    (menu_bar, "View", args, n, (XtCallbackProc) Query_cb_view_partitions, NULL);

  ave_create_cascade_button 
    (menu_bar, "File", args, n, (XtCallbackProc) Query_cb_file, NULL);

  ave_create_cascade_button 
    (menu_bar, "Msgs", args, n, (XtCallbackProc) WarnCB, NULL);

  ave_create_cascade_button 
    (menu_bar, "IPD", args, n, (XtCallbackProc) WarnCB, NULL);

  n = 0;
  button = XmCreatePushButton (menu_pane, "Quit", args, n);
  XtManageChild (button);
  XtAddCallback (button, XmNactivateCallback, (XtCallbackProc) QuitCB, NULL);
  
  n = 0;
  XtSetArg (args[n], XmNsubMenuId, menu_pane); n++;
  cascade = XmCreateCascadeButton (menu_bar, "Exit", args, n);
  XtManageChild (cascade);

  n = 0;
  XtSetArg (args[n], XmNlistVisibleItemCount, 4); n++;
  query_form = XmCreateForm (query_main_window, "form", args, n);
  XtManageChild (query_form);


  n = 0;
  XtSetArg (args[n], XmNeditable, TRUE); n++;
  XtSetArg (args[n], XmNeditMode, XmMULTI_LINE_EDIT); n++;
  XtSetArg (args[n], XmNrows, 6); n++;
  XtSetArg (args[n], XmNcolumns, 40); n++;
  query_command = XmCreateText (query_form, "command", args, n);
  XtManageChild (query_command); 

  n = 0;
  XtSetArg (args[n], XmNeditable, FALSE); n++;
  query_warn = XmCreateText (query_form, "warn", args, n);
  XtManageChild (query_warn); 


  /* Create Frame MainWindow and ScrolledWindow in Frame */
  n = 0;
  XtSetArg (args[n], XmNmarginWidth, 2); n++;
  XtSetArg (args[n], XmNmarginHeight, 2); n++;
  XtSetArg (args[n], XmNshadowThickness, 1); n++;
  XtSetArg (args[n], XmNshadowType, XmSHADOW_OUT); n++;
  query_frame = XmCreateFrame (query_form, "frame", args, n);
  XtManageChild (query_frame);

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetValues (query_command, args, n);
  

  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNtopWidget, query_command); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetValues (query_warn, args, n);
  
  n = 0;
  XtSetArg (args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg (args[n], XmNtopWidget, query_warn); n++;
  XtSetArg (args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNrightAttachment, XmATTACH_FORM); n++;
  XtSetArg (args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
  XtSetValues (query_frame, args, n);

  n = 0;
  XtSetArg (args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++;
  XtSetArg (args[n], XmNscrollingPolicy, XmAUTOMATIC); n++;
  query_swindow = XmCreateScrolledWindow (query_frame, "swindow", args, n);
  XtManageChild (query_swindow);


  n = 0;
  XtSetArg (args[n], XmNwidth, 10000); n++;
  XtSetArg (args[n], XmNheight, 10000); n++;
  XtSetArg (args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  query_canvas = XmCreateDrawingArea (query_swindow, "canvas", args, n);
  XtManageChild(query_canvas);
  XtAddCallback
    (query_canvas, XmNexposeCallback, (XtCallbackProc) refresh_query_canvas, NULL);

  GCanvasGC[0] = create_gc (query_canvas);
  ave_load_color (query_canvas);

  n = 0;
  XtSetArg (args[n], XmNworkWindow, query_canvas); n++;
  XtSetValues (query_swindow, args, n);


  XmMainWindowSetAreas (query_main_window, menu_bar, NULL, NULL, NULL, query_form);
  /* Create the drawing canvas */
  create_drawing_canvas(query_app_shell);  
  XtRealizeWidget (query_app_shell);

}


/*
 ***********************************************************
 * FUNCTION:      main()                                   *
 * DESCRIPTION:   Sets up parameters. Creates windows      *
 ***********************************************************
*/

main(int argc, char **argv)
{
  Display *display;

  /* Initialization of parseinit(), scaninit(), setting up of 
     assorted global variables */
  AveInitialize();

  /* Initialize toolkit and open the display */
  XtToolkitInitialize();
  app_context = XtCreateApplicationContext();
  display = XtOpenDisplay (app_context, NULL, "Ave", "XMdemos", NULL, 0,
			   &argc, argv);
  
  if (!display) {
    XtWarning ("Ave: Can't open display, exiting ...");
    exit(0);
  }

  create_gui_for_ariadne(display);

  XFlush (display);
  XtAppMainLoop (app_context);
}
	

/*
 **************************************************************************
 * FUNCTION:      view_abstract_event_through_named_partition()           *
 * DESCRIPTION:   
 **************************************************************************
*/

