%{
# include "std.h"
# include <stdio.h>
# include "xstdio.h"
# include <varargs.h>

# include "structs.h"
#include "expr.h"
# include "symtab.h"
# include "scan.h"
# include "event.h"
# include "mtree.h"
# include "assert.h"
# include "break.h"
# include "init.h"
# include "view.h"
#include "decl.h"
#include "stmts.h"
#include "abstract.h"

extern FILE *GScriptFile;
extern void* get_value_of_attribute_with_name (void *, void *);
extern boolean equal_value (void *, void *);
/* forward routines */
void yyerror();
void parseinit();
void AddToName(IDENT*, char*);


static int CntPushPop = 0;
int me = 0;
int InForLoop = 0;
FILE *fp;
%}


/* The following %union translates into the union type YYSTYPE in parse.tab.h
   and defines the possible types for elements on bison's semantic stack.
*/
%union {
  int 	     ival;	      /* INTEGER literals */
  IDENT	     *id;	      /* identifiers (points to identifier 
				 table entry) */
  IDE        *ide;
  IDLIST     *id_list;
  SET_OBJ    *sobj;
  TREE_OBJ   *tobj;
  char       *str;
  SYN_TREE   *st;
  DEF_LIST   *dfl;
  SYN_TREE_LIST *stl;
  ABSTRACT_EVENT_LIST *ael;
  ON_KIND    onkind;
  OPER       op;
  WRT_LIST   *wrt;
  ExprOper   eop;
}

/* tokens that come with values */
%token <ival> INT_
%token <id>   ID_  
%token <id>   _ID 

/* keywords */
%token  ABS_, ADD_, AFTER_, AND_, BEFORE_, BEGIN_, BREAK_, BY_
%token  CHAIN_, CHILD_, COMPUTE_, CONST_, COUNT_, CREATE_, CROSS_
%token  DISJOINT_, DISTINCT_, DUMP_, END_, EQUAL_, EVENT_, EXEC_, EXPLAIN_
%token  FIF_ FIRST_ FLOAT_, FOR_, FOREACH_
%token  GEC_, GTC_, IMAP_, IN_, INFO_, INTEGER_
%token  LAST_, LEAF_, LEC_, LEFT_, LOAD_, LTC_, MARKER_, MATCH_, MLC_, MRC_
%token  ONALL_, ONSOME_, OPTION_, OR_, OWNERS_
%token  PARENT_, PART_, PCHAIN_, PRINT_,  PRODUCT_, QUIT_
%token  RELATION_, REMAINDER_, REMOVE_, REPORT_, RESET_, RESTORE_, RIGHT_
%token  SECTION_, SET_, SHOW_, SIBLING_, SIF_, SIZE_, SOURCE_, SUM_
%token  SUBSET_, SUPERSET_ 
%token  TCHAIN_, TO_, UNION_, UTAG_, VIEW_, WITH_, WRT_

/* multi-character operators not declared below */
%token DOTDOT_  COLON_COLON_ COLON_EQ_  PUSH_ POP_  
%token AT_

/* precedence and associativity among operators */
%left  ':'
%left  OR_
%left  AND_
%right NOT_
%left  '<' '>' '|' LE_ GE_ '=' '#'
%left  '+' '-' UNION_
%left  '*' '/' SECTION_ CROSS_


/* non-terminal semantic record types */
%type <ival> 	Relop BreakWhen
%type <str>     OptVarDecl TypeField WhatToAdd Anchor
%type <ide>     LeftId OptImap
%type <tobj>    ChainPrimary ChainItem ChainList Chain
%type <tobj>    PtPrimary PtItem PtList PtChain
%type <stl>     Expr Primary OptExpr
%type <sobj>    Set
%type <stl>     Range SetRightSide
%type <dfl>     Dfntns DfntnsBlock ChainDef ForDef UserInputs Session SetDef
%type <dfl>     PchainDef PtChainDef MatchDef LoadStmt ErrStmt OwnerStmt CountStmt
%type <dfl>     RestoreStmt AttachStmt PartStmt ShowAttrStmt BreakStmt IndexDef
%type <dfl>     AssignStmt SkipStmt FailStmt FilterBlock OptFilterBlock
%type <dfl>     FilterAssignStmt FilterSuperBlock  ExecStmt
%type <dfl>     AddStmt, ConstStmt PrintStmt ViewStmt
%type <dfl>     AddMarkerStmt 
%type <str>     MyFileName  
%type <onkind>  Onword
%type <op>      SpChSymbols
%type <wrt>     IDList, OptWrtList OptWith
%type <id>      AnyIdent
%type <eop>     FuncCall Addop Mulop
%type <ael>     RootNodes
%start          Session
%%


/* Compilation Unit Productions */

Session	
  : UserInputs {fprintf(stdout, "? "); }
  | Session UserInputs {fprintf(stdout, "? "); }
  ;

UserInputs
  : Dfntns {interpret_definition ($1);}
  ;

DfntnsBlock
  : Dfntns {$$ = $1;}
  | DfntnsBlock Dfntns  
    {
      DEF_LIST *dl = $1;
      while (dl->next)
	dl = dl->next;
      dl->next = $2;
      $$ = $1;
    }
  | '{' DfntnsBlock '}' {$$ = $2;}
  ;


Dfntns	
  : ChainDef { $$ = $1;}
  | ForDef {$$ = $1;}
  | PchainDef {$$ = $1;}
  | PtChainDef { $$ = $1; }
  | SetDef {$$ = $1;}
  | MatchDef {$$ = $1;}
  | LoadStmt { $$ = $1; }
  | RestoreStmt { $$ = $1; }
  | OwnerStmt {$$ = $1;}
  | ErrStmt {$$ = $1;}
  | CountStmt {$$ = $1;}
  | AttachStmt {$$ = $1;}
  | AddMarkerStmt { $$ = $1; }
  | ShowAttrStmt  { $$ = $1; }
  | PartStmt { $$ = $1; }
  | AssignStmt {$$ = $1;}
  | BreakStmt {$$ = $1;}
  | ExecStmt {$$ = $1;}
  | IndexDef {$$ = $1;}
  | QuitStmt {$$ = NULL;}
  | AddStmt {$$ = NULL;}
  | ConstStmt {$$ = NULL;}
  | PrintStmt {$$ = NULL;}
  | ViewStmt  {$$ = NULL;}
  ;

ViewStmt
  : VIEW_ ID_ WRT_ ID_ ';'
  {
    view_node_through_named_partition ($2, $4->id_chars);
    $$ = (DEF_LIST *) NULL;
  }
;

PrintStmt
  : PRINT_ ID_ ';'
  { 
    print_identifier ($2);
    $$ = (DEF_LIST *) NULL;
  }
AddStmt
  : ADD_  WhatToAdd IDList ';'
  {
    add_reserved_words($3, $2);
    $$ = (DEF_LIST *) NULL;
  }
  | ADD_ TypeField IDList TO_ IDList ';'
  {
    add_additional_fields ($2, $3, $5);
    $$ = (DEF_LIST *) NULL;
  }
;

WhatToAdd
  : UTAG_ {$$ = "utag";}
  | EVENT_ {$$ = "event";}
  
TypeField
  : FLOAT_ {$$ = "float";}
  | INTEGER_ {$$ = "int";}
;

ConstStmt
  : CONST_ ID_ '=' INT_ ';'
  {
    set_global_constants ($2, $4);
    $$ = (DEF_LIST *) NULL;
  }
;
    
AssignStmt
  : LeftId '=' Expr ';' {$$ = (DEF_LIST *) NULL;}
  ;

ExecStmt
  : EXEC_ MyFileName ';'
  {
    GScriptFile = fopen($2, "r");
    $$ = (DEF_LIST *) NULL;
  }
;

BreakWhen
  : AFTER_ {$$  = BREAK_AFTER_EVENT;}
  | BEFORE_ {$$ = BREAK_BEFORE_EVENT;}
;

BreakStmt
  : BREAK_ BreakWhen ID_ OptWrtList ';'
    {
      ABSTRACT_EVENT_LIST *ael = NULL;
      ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) ($3->id_decl->decl_object);

      if ($4 == NULL)
	add_node_to_abstract_list (&ael, ae);
      else {
	WRT_LIST *wrt = $4;
	while (wrt) {
	  ABSTRACT_EVENT_LIST *tael = 
	    get_named_nodes (wrt->marker, ae, 0);
	  abstract_list_concat (&ael, tael);
	  wrt = wrt->next;
	}
      }
      break_with_select_abstract_event(ael, $2);
      $$ = (DEF_LIST *) NULL;
    }
   | BREAK_ BreakWhen '(' ID_ OptWrtList ')' ';'
    {
      ABSTRACT_EVENT_LIST *ael = NULL;
      ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) ($4->id_decl->decl_object);

      if ($5 == NULL)
	add_node_to_abstract_list (&ael, ae);
      else {
	WRT_LIST *wrt = $5;
	while (wrt) {
	  ABSTRACT_EVENT_LIST *tael = 
	    get_named_nodes (wrt->marker, ae, 0);
	  abstract_list_concat (&ael, tael);
	  wrt = wrt->next;
	}
      }
      break_with_select_abstract_event(ael, $2);
      $$ = (DEF_LIST *) NULL;
    }

;

  
  
ShowAttrStmt
  : WITH_ RootNodes SHOW_ ID_ '(' ID_ ')' WRT_ ID_ ';'
    {
      int root_cnt = get_size_of_abstract_list ($2);
      if (root_cnt == 0) {
	printf("Show Command is ignored\n");
      }
      else if (root_cnt > 1) {
	printf("with <ident>::<ident> NOT allowed in show command\n");
	printf("Show command ignored\n");
      }
      else {
	ABSTRACT_EVENT *ae = get_element_from_abstract_list ($2, 0);
	show_select_attribute (ae, $4->id_chars, $6->id_chars, $9->id_chars);
      }
      $$ = (DEF_LIST *) NULL;
    }
;
PartStmt
  : WITH_ RootNodes FOREACH_ ID_ CREATE_ ID_ PART_ WRT_ ID_ ';'
    {
      int root_cnt = get_size_of_abstract_list ($2);
      int i;
      for (i=0; i<root_cnt; i++) {
	ABSTRACT_EVENT *ae = get_element_from_abstract_list($2, i);
	partition_abstract_event_wrt_attribute
	  (ae, $4->id_chars, $6->id_chars, $9->id_chars); 
      }
      $$ = (DEF_LIST *) NULL;
    }
;


AttachStmt
   : WITH_ RootNodes FOREACH_ Anchor OptExpr COMPUTE_ ID_ '=' Expr ';'
     {
       $7->id_reserved = 2;
       new_compute_attribute_for_named_nodes ($2, $4, $5, $7->id_chars, $9);
       $$ = (DEF_LIST *) NULL;
     }
;

OptExpr
    : {$$ = NULL;}
    | '{' Expr ';' '}'  {$$ = $2;}
;



RootNodes
   : ID_
     {
       ABSTRACT_EVENT_LIST *ael = NULL;
       DECL *d = $1->id_decl;
       if (d == NULL) {
	 printf("WARNING: %s has not yet been defined.\n", $1->id_chars);
	 printf("Command will be ignored.\n");
	 $$ = NULL;
       }
       else if (d->decl_kind != DeclMatch) {
	 printf("WARNING: %s is not an abstract event.\n", 
		$1->id_chars);
	 $$ = NULL;
       }
       else {
	 ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) d->decl_object;
	 add_node_to_abstract_list (&ael, ae);
	 $$ = ael;
       }
     }
   | ID_ COLON_COLON_ ID_
     {
       ABSTRACT_EVENT_LIST *ael = NULL;
       DECL *d = $1->id_decl;
       if (d == NULL) {
	 printf("WARNING: %s has not yet been defined.\n", $1->id_chars);
	 printf("Command will be ignored.\n");
	 $$ = NULL;
       }
       else if (d->decl_kind != DeclMatch) {
	 printf("WARNING: %s is not an abstract event.\n", 
		$1->id_chars);
	 $$ = NULL;
       }
       else {
	 ABSTRACT_EVENT *ae = (ABSTRACT_EVENT *) d->decl_object;
	 ael = get_named_nodes ($3->id_chars, ae, 0); 
	 $$ = ael;
       }
     }
;

AddMarkerStmt
   : WITH_ RootNodes FOREACH_ Anchor OptExpr ADD_ MARKER_ ID_ ';'
     {
       add_marker_for_named_nodes ($2, $4, $5, $8->id_chars);
       $$ = (DEF_LIST *) NULL;
     }
;

Anchor
  : ID_ {$$ = $1->id_chars;}
  | EVENT_ {$$ = "event";}
  ;
      
RestoreStmt
  : RESTORE_ '(' ID_ ')' ';'
    {
      restore_match_variable ($3);
      $$ = (DEF_LIST *) NULL;
    }
;

CountStmt
  : COUNT_ '(' ID_ OptWrtList ')' ';'
    {
      count_listed_events ($3, $4);
      $$ = (DEF_LIST *) NULL;
    }
;

OwnerStmt
  : OWNERS_ '(' ID_ OptWrtList ')' ';' 
    { 
      get_owner_process_set ($3, $4);
      $$ = (DEF_LIST *) NULL;
    }
  ;
ErrStmt
  : error ';' { printf("parse error\n"); $$ = (DEF_LIST *) NULL;}
;

QuitStmt
  : QUIT_ ';' {printf("Exiting Ariadne\n"); return;}
  ;

LoadStmt
  : LOAD_ MyFileName ';'
    { 
      int success = LoadTraceFile($2, "START", 1, "END", 1);
      printf("File %s loaded successfully.\n", $2);
      $$ = (DEF_LIST *) NULL;
    }
;

MyFileName
  : ID_
    { $$ = $1->id_chars; }
  | MyFileName '.' ID_
    { 
      char *str = (char *) malloc (strlen($1) + strlen ($3->id_chars)+2);
      strcat (str, $1);
      strcat (str, ".");
      strcat (str, $3->id_chars);
      free ($1);
      $$ = str;
    }
  ;
MatchDef
  : MATCH_ LeftId '=' PtChain OptWith';'
    {
      DEF_LIST *dfl = create_reg_def ($2, (OBJECT *) $4, $5);
      dfl->def_kind = DefMatch;
      $2->decl_kind = DeclMatch;
      $$ = dfl;
    }
;

OptWith
   : {$$ = NULL;}
   | WITH_ '[' IDList ']' {$$ = $3;}
   ;

IndexDef
  : IMAP_ LeftId '=' Expr ';'
  {
    IMAP_OBJ *imo = allocate_imap_object ($4);
    DEF_LIST *dfl = create_reg_def ($2, (OBJECT *) imo, NULL);
    dfl->def_kind = DefImap;
    $2->decl_kind = DeclImap;
    $$ = dfl;
  }
  ;

PtChainDef
  : TCHAIN_ LeftId '=' PtChain ';'
    { 
      DEF_LIST *dfl = create_reg_def ($2, (OBJECT *) $4, NULL);
      dfl->def_kind = DefTchain;
      $2->decl_kind = DeclTchain;
      $$ = dfl;
    }
  ;

PtChain
  : PtList {$$=$1;}
  ;

PtList
  : PtItem {$$ = $1; }
  | PtList PtItem {concat_tree_objects ($1,$2); $$=$1;}
  ;

PtItem
  : PtItem '*' {$$ = build_tree_item ($1,0, (WRT_LIST *) NULL);}
  | PtItem '^' INT_ {$$ = build_tree_item ($1, $3, (WRT_LIST *) NULL);}
  | OptVarDecl PtPrimary 
    {$2->marker = $1; $$ = $2;}
  ;

OptWrtList
  : {$$ = (WRT_LIST *) NULL;}
  | WRT_ '[' IDList ']' { $$ = $3;}
  ;

AnyIdent
  : ID_ {$$ = $1;}
  | _ID {$$ = $1;}
  ;

IDList 
  : AnyIdent
    { 
      WRT_LIST *tmp = (WRT_LIST *) malloc (sizeof (WRT_LIST));
      tmp->next = NULL;
      tmp->marker = (char *) malloc (strlen ($1->id_chars) + 1);
      strcpy (tmp->marker, $1->id_chars);
      $$ = tmp;
    }
      
  | IDList ID_
    {
      WRT_LIST *tmp = $1;
      WRT_LIST *new = (WRT_LIST *) malloc (sizeof (WRT_LIST));
      new->next = NULL;
      new->marker = (char *) malloc (strlen ($2->id_chars)+1);
      strcpy (new->marker, $2->id_chars);
      while (tmp->next)
	tmp = tmp->next;
      tmp->next = new;
      $$ = $1;
    }
  ;

PtPrimary
  : LeftId {$$ = build_unresolved_tree_item ($1, DeclTchain);}
  | '(' '+' PtChain '+' ')' 
    {
      TREE_OBJ *tch = build_tree_item ($3, 1 , (WRT_LIST *) NULL);
      tch->oper = OperAny;
      $$ = tch;
    }
  | '{' PtChain OptWrtList '}' 
    { 
      TREE_OBJ *tch = build_tree_item($2, 1, $3);
      tch->oper = OperParen;
      $$ = tch;
    }
  | '|' '['  PtChain OptWrtList  ']' '|'
    {
      TREE_OBJ *tch = build_tree_item($3,1,$4);
      tch->oper = OperParallel;
      $$ = tch;
    }
  | '<' PtChain  OptWrtList '>'
    { 
      TREE_OBJ *tch = build_tree_item($2, 1, $3);
      tch->oper = OperPrecede;
      $$ = tch;
    }
  | '(' PtChain ')'  
    {
      TREE_OBJ *to = build_tree_item ($2,1, (WRT_LIST *) NULL);
      to->oper = OperParen;
      $$ = to;
    }
  ;



PchainDef
  : PCHAIN_ LeftId '=' Chain Onword  Set ';'
    {
      DEF_LIST *dfl;
      PCH_OBJ *po = allocate_pchain();
      po->chain = $4;
      po->set = $6;
      po->on_kind = $5;
      dfl = create_reg_def($2, (OBJECT *) po, NULL);
      dfl->def_kind = DefPchain;
      $2->decl_kind = DeclPchain;
      $$ = dfl;
    }
  ;

Onword
  : ONALL_ {$$ = OnAll;}
  | ONSOME_ {$$ = OnSome;}
  ;
/*
OptInteger
  : 
  | '(' Expr ')'
  ;
*/
Set
  : LeftId 
    { 
      SET_OBJ *so = allocate_set();
      so->ide = $1;
      $$ = so;
    }
	     
  | '{' Range '}' 
    {
      SET_OBJ *so = allocate_set();
      so->condsyn = $2;
      $$ = so;
    }
  ;

ChainDef 
  : CHAIN_  LeftId '=' Chain OptWrtList';'
    { DEF_LIST *dfl = create_reg_def ($2, (OBJECT *) $4, NULL);
      dfl->def_kind = DefChain;
      $2->decl_kind = DeclChain;
      $4->wrt = $5;
      $$ = dfl;
    }
  ;
	
Chain
  : ChainList {$$ = $1;}
  ;

ChainList
  : ChainItem {$$ = $1;}
  | ChainList ChainItem
   {
     concat_tree_objects ($1, $2);
     $$ = $1;
   }
  ;



ChainItem
  : ChainItem '*' 
    { $$ = build_tree_item ($1, 0, (WRT_LIST *) NULL); }
  | ChainItem '^' INT_ { $$ = build_tree_item ($1, $3, (WRT_LIST *) NULL); }
  | OptVarDecl ChainPrimary { $2->marker = $1; $$ = $2 }
  ;

OptFilterBlock
  : {$$ = NULL;}
  | '{' FilterSuperBlock '}' {$$ = $2}
  ;

FilterSuperBlock
  : FilterBlock {$$ = $1;}
  | FilterSuperBlock FilterBlock
    {
      DEF_LIST *dfl = $1;
      while (dfl->next)
	dfl = dfl->next;
      dfl->next = $2;
      $$ = $1;
    }
   ;

FilterBlock
  : FilterAssignStmt {$$ = $1;}
  | SkipStmt {$$ = $1;}
  | FailStmt {$$ = $1;}
  ;

FilterAssignStmt
  :  LeftId '=' Expr ';'
  {
    IMAP_OBJ *imo = allocate_imap_object ($3);
    DEF_LIST *dfl = create_reg_def ($1, (OBJECT *)imo, NULL);
    dfl->def_kind = DefAssign;
    $1->decl_kind = DeclImap;
    $$ = dfl;
  }
  ;

SkipStmt
  : SIF_ Expr ';'
  {
    IMAP_OBJ *imo = allocate_imap_object ($2);
    DEF_LIST *dfl = create_reg_def (NULL, (OBJECT *)imo, NULL);
    dfl->def_kind = DefSkip;
    $$ = dfl;
  }
  ;

FailStmt
  : FIF_ Expr ';' 
  {
    IMAP_OBJ *imo = allocate_imap_object ($2);
    DEF_LIST *dfl = create_reg_def (NULL, (OBJECT *)imo, NULL);
    dfl->def_kind = DefFail;
    $$ = dfl;
  }
  ;


ChainPrimary
  : LeftId OptFilterBlock 
   {
     TREE_OBJ *to = build_unresolved_tree_item($1, DeclChain);
     to->dfl = $2;
     $$ = to;
   }
  | '(' '+' Chain  '+' ')'
    {
      TREE_OBJ *co = build_tree_item ($3, 1, (WRT_LIST *) NULL);
      co->oper = OperAny;
      $$ = co;
    }
  | '(' Chain ')'  
    { TREE_OBJ *co = build_tree_item ($2,1, (WRT_LIST *) NULL);
      co->oper = OperParen;
      $$ = co;
    }
  | SpChSymbols {$$ = build_special_chain_symbol($1);}
  ;


SpChSymbols
  : '@' {$$ = OperAt;}
  | PUSH_ {$$ = OperPush; ++CntPushPop;}
  | POP_  {$$ = OperPop; --CntPushPop;}
  ;

ForDef
  : FOR_ '(' ID_ '=' INT_ TO_ INT_ BY_ INT_ ')' BEGIN_ DfntnsBlock END_
    {
      DEF_LIST *dl = create_for_def($3, $5, $7, $9);
      dl->next = $12;
      $$ = dl;
    }
  ;
LeftId
  : ID_  
    { 
      IDE *ide = allocate_ide();
      ide->id_chars = (char *) malloc (strlen($1->id_chars) + 1);
      strcpy(ide->id_chars, $1->id_chars);
      $$ = ide;
    }
  | LeftId '[' Expr ']' 
    {
      add_syntree_list_to_ide ($1, $3);
      $$ = $1;
    }
;

OptVarDecl
  : {$$ = (char *) NULL;}
  | _ID ':'  {$$ = $1->id_chars;}
  ;



Expr	
  : Expr Relop Expr  %prec '='
  | Expr Addop Expr  %prec '+'
    { 
      SYN_TREE_LIST *stl = (SYN_TREE_LIST *) malloc (sizeof(SYN_TREE_LIST));
      stl->next = NULL;
      stl->st = build_syntax_tree ($1, $3, $2, NULL);
      $$ = stl;
    }
  | Expr Mulop Expr  %prec '*'
    { 
      SYN_TREE_LIST *stl = (SYN_TREE_LIST *) malloc (sizeof(SYN_TREE_LIST));
      stl->next = NULL;
      stl->st = build_syntax_tree ($1, $3, $2, NULL);
      $$ = stl;
    }
  | Primary {$$ = $1; }
  ;

Relop
  : '<' { $$ = '+' }
  | '>' { $$ = '>' }
  | LE_ { $$ = LE_ }
  | GE_ { $$ = GE_ }
  | '#' { $$ = '#' }
  ;

Addop
  : '+'  {$$ = ApplyAdd;}
  | '-'  {$$ = ApplySub;}
  | UNION_ {$$ = ApplyUnion;}
  ;

Mulop
  : '*'  {$$ = ApplyMult;}
  | '/'  {$$ = ApplyDiv;}
  | CROSS_ {$$ = ApplyCross;}
  | SECTION_ {$$ = ApplySection;}
  ;

Primary
  :'(' Expr ')' {$$ = $2;}
  | INT_ 
  { 
    SYN_TREE_LIST *stl = (SYN_TREE_LIST *) malloc (sizeof (SYN_TREE_LIST));
    stl->next = NULL;
    stl->st = build_int_node($1);
    $$ = stl;
  }
  | ID_ 
  { 
    SYN_TREE_LIST *stl = (SYN_TREE_LIST *) malloc (sizeof (SYN_TREE_LIST));
    stl->next = NULL;
    stl->st = build_id_node($1);
    $$ = stl;
  }
  | FuncCall '(' Range ')'
  {
    SYN_TREE_LIST *stl = (SYN_TREE_LIST *) malloc (sizeof (SYN_TREE_LIST));
    stl->next = NULL;
    stl->st = build_syntax_tree ($3, NULL, $1, NULL); 
    $$ = stl;
  }
  | ID_ '(' Range ')'
    { 
      SYN_TREE_LIST *stl;
      if ($1->id_reserved == 0) {
	yyerror("Field %s is not defined", $1->id_chars);
	$$ = NULL;
      }
      stl = (SYN_TREE_LIST *) malloc (sizeof (SYN_TREE_LIST));
      stl->next = NULL;
      stl->st = build_syntax_tree ($3, NULL, ApplyNone, $1->id_chars);
      $$ = stl;
    }
  ;


FuncCall
  : ABS_      {$$ = ApplyAbs; }
  | CHILD_    {$$ = ApplyChild; }
  | DISJOINT_ {$$ = ApplyDisjoint;}
  | DISTINCT_ {$$ = ApplyDistinct;}
  | EQUAL_    {$$ = ApplyEqual; }
  | FIRST_    {$$ = ApplyFirst;}
  | GEC_      {$$ = ApplyGec; }
  | GTC_      {$$ = ApplyGtc; }
  | LAST_     {$$ = ApplyLast; }
  | LEAF_     {$$ = ApplyLeaf;}
  | LEC_      {$$ = ApplyLec; }
  | LEFT_     {$$ = ApplyLeft; }
  | LTC_      {$$ = ApplyLtc; }
  | PARENT_   {$$ = ApplyParent; }
  | PART_     {$$ = ApplyPart; }
  | PRODUCT_  {$$ = ApplyProd; }
  | RIGHT_    {$$ = ApplyRight; }
  | SIBLING_  {$$ = ApplySibling; }
  | SIZE_     {$$ = ApplySize;}
  | SUM_      {$$ = ApplySum; }
  | SUBSET_   {$$ = ApplySubSet;}
  | SUPERSET_   {$$ = ApplySuperSet;}
  ;

SetDef
  : SET_ LeftId '=' SetRightSide OptCond OptImap';'
    {
      DEF_LIST *dfl;
      SET_OBJ *so = allocate_set();
      so->condsyn = $4;
      so->map  = $6;
      $2->decl_kind = DeclSet;
      dfl = create_reg_def ($2, (OBJECT *) so, NULL);
      dfl->def_kind = DefSet;
      $$ = dfl;
    }
  ;

SetRightSide
  : '{' Range '}' {$$=$2;}
  | Range {$$=$1;}
  ;

OptCond
  :
  | '|' Expr
  ;

OptImap
  : {$$ = NULL;}
  | WITH_ IMAP_ LeftId {$$ = $3;}   /* should change it to LftId later */
  ;


Range	
  : Expr  { $$ = $1; }
  | Range ',' Expr 
    {
      SYN_TREE_LIST *stl = $1;
      while (stl->next)
	stl = stl->next;
      stl->next = $3;
      $$ = $1;
    }
  | Expr DOTDOT_ Expr 
    { 
      SYN_TREE_LIST *stl = 
	(SYN_TREE_LIST *) malloc (sizeof (SYN_TREE_LIST));
      stl->next = NULL;
      if ($1->st->oper == ApplyPart)
	stl->st= build_syntax_tree ($1, $3, ApplyExpandPart, NULL); 
      else if ($1->st->oper == ApplyChild)
	stl->st= build_syntax_tree ($1, $3, ApplyExpandChild, NULL); 	
      else
	stl->st = build_syntax_tree ($1, $3, ApplySet, NULL);
      stl->st->value = 1;
      $$ = stl;
    }
  | Expr DOTDOT_ Expr BY_ INT_
    { 
      SYN_TREE_LIST *stl = (SYN_TREE_LIST *) malloc (sizeof (SYN_TREE_LIST));
      stl->next = NULL;
      if ($1->st->oper == ApplyPart)
	stl->st= build_syntax_tree ($1, $3, ApplyExpandPart, NULL);
      else if ($1->st->oper == ApplyChild)
	stl->st= build_syntax_tree ($1, $3, ApplyExpandChild, NULL); 	
      else
	stl->st = build_syntax_tree ($1, $3, ApplySet, NULL); 

      stl->st->value = $5;
      $$ = stl;
    }
;


%%

void yyerror (va_alist)
/* Use of <varargs> allows a call to yyerror to specify a format string
   followed by multiple arguments */
va_dcl
{
  va_list args;
  char *fmt;
  /* Print filename and line number at beginning of every error message.
     Use same format as gcc produces, so can compile and check errors
     within emacs. */
/*  fprintf (stderr, "%s:%d ", inname, yylineno); */
  /* Print specific error message */
  va_start (args);
  fmt = va_arg (args, char *);
  vfprintf (stderr, fmt, args);
  va_end (args);
  fprintf (stderr, "\n");
}


void parseinit ()
{
  
}




 
/************************************************************************
* Function changes int to string, starting with '[' and ending with ']' *
* e.g., 10 is changed to "[10]", 253 to "[253]" etc.                    *
* As you may have guessed already, this is our way of handling arrays:  *
* where we just make a number of new identifiers, after computing the   *
* indices. In our domain of interpreter design, there is no distinction *
* between run-time and compile-time.                                    *
************************************************************************/

void AddToName (IDENT *id, char *temp)
{
  int size = strlen(id->id_chars) + strlen(temp) + 1;
  char *name = (char *) malloc(size);
  (*name) = '\0';
  strcat(name,id->id_chars);
  strcat(name,temp);
  free(id->id_chars);
  id->id_chars = name;
  return;
}
