#include <stdio.h>
#include <X11/Intrinsic.h>
#include <Xm/Xm.h>
#include "structs.h"
#include "event.h"
#include "strs.h"
#include "match.h"
#include "symtab.h"
#include "abstract.h"
#include "DrawingCanvas.h"

float xmax = 1;

void ave_calc_coord (ABSTRACT_EVENT *ae, int level);
void ave_draw_tree  (ABSTRACT_EVENT *ae, int propagate);
void ave_mark_nodes (ABSTRACT_EVENT *ae);
extern ave_draw_error_arrow (Widget w, GC lgc, int origin_x, 
			     int origin_y, int scale);
extern Widget query_canvas;
extern GC GCanvasGC[2];

char *get_name (ABSTRACT_EVENT *ae)
{
  int size = 0;
  char *name;
  ABSTRACT_EVENT *t_ae = ae;

  while (t_ae) {
    size += StrLen(t_ae->name);
    t_ae = t_ae->next;
  }

  name = (char *) malloc (size+10);
  
  t_ae = ae;
  while (t_ae) {
    StrCat (name, t_ae->name);
    t_ae = t_ae->next;
  }

  return name;
}


void add_abstract_list_to_collapse_child 
  (COLLAPSE_NODE *cc, ABSTRACT_EVENT_LIST *ael)
{
  ABSTRACT_EVENT_LIST *temp = cc->ael;
  linked_list_concat (&temp, (LINKED_LIST *) ael);
  return;
}
SLIDE_COLLAPSE_NODE *allocate_slide_collapse_node(int total, int lo, int hi)
{
  SLIDE_COLLAPSE_NODE *sln = 
    (SLIDE_COLLAPSE_NODE *) malloc (sizeof (SLIDE_COLLAPSE_NODE));
  sln->total = total;
  sln->lo = lo;
  sln->hi = hi;
  sln->prev = sln->next = NULL;
  sln->ave = NULL;
  sln->winfo = NULL;
  return sln;
}

  
void add_partition_for_view (COLLAPSE_NODE *cn, ABSTRACT_EVENT *ae, char *name)
{
  COLLAPSE_NODE *cc = allocate_collapse_node();
  COLLAPSE_NODE *tcc = cn->down;
  ABSTRACT_EVENT_LIST *ael = NULL;
/*  char *name = get_name (ae->down); */
  cc->ave = allocate_ave_str (&NodeId, cn->ave->my_id);
  cc->pattern = (char *) malloc (StrLen (name)+1);
  cc->parent = cn;
  strcpy(cc->pattern, name);

  add_node_to_abstract_list (&ael, ae);
  cc->ael = ael;
  
  if (tcc == NULL)
    cn->down = cc;
  else
    {
      while (tcc->next)
	tcc = tcc->next;
      tcc->next = cc;
    }
  return;
}

boolean InCollapseChild (ABSTRACT_EVENT *ae, COLLAPSE_NODE *cc)
{
  COLLAPSE_NODE *tcc = cc;
  char *name = get_name (ae->down);
  for (; tcc != NULL; tcc = tcc->next) {
    if (! StrCmp(tcc->pattern, name))  /* Strings match */
      {
	int size = get_size_of_list ((LINKED_LIST *) tcc->ael);
	ABSTRACT_EVENT_LIST *ael = NULL;
	add_node_to_abstract_list (&ael, ae);
	add_abstract_list_to_collapse_child(tcc, ael);
	if (tcc->scn == NULL && (size+1) > 3) {
	  tcc->scn = allocate_slide_collapse_node(size+1, 0, 3);
	  tcc->scn->prev = 
	    (ABSTRACT_EVENT *) get_element_from_list((LINKED_LIST *) tcc->ael,
						       0);
	  tcc->scn->next = ae;
	}
	else if (tcc->scn)
	  tcc->scn->next = ae;
	
	return TRUE;
      }
  }
  return FALSE;
}
  

void DoDefaultCollapse (ABSTRACT_EVENT *ae)
{
  int total = 0;
  int part = 0;
  ABSTRACT_EVENT *tae;
  char str[64];

  for (tae = ae->down; tae != NULL; tae = tae->next) {
    COLLAPSE_NODE *cc = ae->clps->down;
    if (tae->error)
      break;
    if ((cc == NULL) || (InCollapseChild (tae, cc) == FALSE)) {
      ++part;
      add_partition_for_view (ae->clps, tae, get_name(tae->down));
    }
    ++total;
  }
  sprintf(str, "default");
  ae->clps->total = total;
  ae->clps->part = part;
  ae->clps->pattern = (char *) malloc (strlen(str)+1);
  sprintf(ae->clps->pattern, "%s",  str);
  return;
}
      
  
void AveDefaultCollapse (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;

  if (ae->clps) {
    DoDefaultCollapse (ae);
    AveDefaultCollapse (ae->next);
    return;
  }

  AveDefaultCollapse (ae->next);
  AveDefaultCollapse (ae->down);
  return;
}

int get_child_count (ABSTRACT_EVENT *ae)
{
  int cnt = 1;
  if (ae == NULL)
    return cnt;

  while (ae->next) {
    ++cnt;
    ae = ae->next;
  }
  return cnt;
}

int get_group_size (TREE_OBJ *to)
{
  int cnt = 1;
  if (to == NULL)
    return cnt;

  while (to->next) {
    ++cnt;
    to = to->next;
  }
  return cnt;
}



/****************************************************************************
 * ave_horiz_collapse_mark():  Follows ave_vert_collapse(). It looks at an  *
 *        abstract event, and if the cnt > 3, it builds a slide_clps node.  *
 *        It also marks the nodes to be displayed, for ave_calc_coord() to  *
 *        process. Once you have hit a pchain node (evident from ae->clps)  *
 *        we should not mark its descendents as candidates for display.     *
 ****************************************************************************/

void ave_horiz_collapse (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  

  /* If I am a pchain node, then my siblings get benefit of doubt, but my
     children are not for horizontal collapsing. */
  if (ae->clps) {
    ave_horiz_collapse(ae->next);
    ave_horiz_collapse(ae->down);
    return;
  }

  /* Otherwise, check my "cnt" field, to see if my children are ready to be
     grouped. in slide_clps node. */
  if (ae->cnt > 2) {
    ABSTRACT_EVENT *tae = ae->down;
    int i;
    while (tae->next)
      tae = tae->next;

    ae->down->slide_clps = allocate_slide_collapse_node (ae->cnt, 1, ae->cnt);
    ae->down->slide_clps->next = tae;
    ae->down->slide_clps->prev = ae->down;
    ae->down->slide_clps->ave = allocate_ave_str (&NodeId, ae->ave->my_id);
    ae->down->slide_clps->ave->draw_flag = 1;
  }
  
  ave_horiz_collapse(ae->next);
  ave_horiz_collapse (ae->down);
  return;
}

void mark_slide_node (SLIDE_COLLAPSE_NODE *scn)
{
  scn->ave->draw_flag = 1;
  ave_mark_nodes (scn->next);
  return;
}

void ave_mark_nodes (ABSTRACT_EVENT *ae)
{
  if (ae == NULL)
    return;
  ae->ave->draw_flag = 1;

  if (ae->slide_clps) 
    mark_slide_node (ae->slide_clps);
  else
    ave_mark_nodes (ae->next);

  if (ae->clps && !StrCmp(ae->down->name, "proc")) 
    ae->clps->ave->draw_flag = 1;
  else
    ave_mark_nodes (ae->down);
  return;
}

  
void InitCoordinate ()
{
  xmax = 1;
}




void ave_draw_slide_clps (SLIDE_COLLAPSE_NODE *sc)
{
  if (sc == NULL) { /* insanity check: this should never execute */
    printf("PANIC: ave_draw_slide_clps(): passed NULL parameter!!!\n");
    return;
  }
      
  if (sc->ave->done_flag)
    XtUnmanageChild (sc->winfo->w);
  XtManageChild (sc->winfo->w);
  sc->ave->done_flag = 1;
  ave_draw_tree (sc->next, 1);
  return;
}
  
void ave_draw_clps (COLLAPSE_NODE *cn, int propagate)
{
  if (cn == NULL)
    return; 

  if (cn->ave->done_flag)
    XtUnmanageChild (cn->winfo->w);
  
  if (cn->ave->draw_flag) {
    XtManageChild (cn->winfo->w);
    cn->ave->done_flag = 1;
    if (cn->ael && propagate) {
      if (cn->scn) {
	ave_draw_tree(cn->scn->prev, 0);
	ave_draw_slide_clps (cn->scn);
      }
      else {
	int i;
	int size = get_size_of_list ((LINKED_LIST *)cn->ael);
	for (i=0; i<size; i++) {
	  ABSTRACT_EVENT *ae = get_element_from_abstract_list (cn->ael, i);
	  ave_draw_tree(ae, 0);
	}
      }
      ave_draw_clps (cn->next, 1);
    }
  }
  if (propagate)
    ave_draw_clps (cn->down, 1);
  return;
}


void ave_draw_tree (ABSTRACT_EVENT *ae, int propagate)
{
  if (ae == NULL)
    return;

  if (ae->ave->draw_flag && ae->winfo) {
    if (ae->ave->done_flag)
      XtUnmanageChild (ae->winfo->w);
    XtManageChild (ae->winfo->w);
    ae->ave->done_flag = 1;
  }

  if (propagate) {
    if (ae->slide_clps)
      ave_draw_slide_clps (ae->slide_clps);
    else
      ave_draw_tree (ae->next, 1);
  }

  if (ae->clps) {
    if (!StrCmp(ae->down->name, "proc") && (strchr(ae->clps->pattern, ',') != NULL))
      ave_draw_clps (ae->clps, 0);
    else
      ave_draw_clps (ae->clps, 1);
  }
  else if (ae->ave->draw_flag)
    ave_draw_tree (ae->down, 1);

  return;
}
  

      

void horiz_clps_calc_coord (SLIDE_COLLAPSE_NODE *sc, int level)
{
  float x;
  if (sc->prev->ave->x <= xmax)
    sc->ave->x = sc->prev->ave->x + 3;

  if (sc->ave->x > xmax)
    xmax = sc->ave->x;

  sc->ave->y = level;
  ave_calc_coord (sc->next, level);
  return;
}

extern void vert_clps_calc_coord (COLLAPSE_NODE *cn, int level);

void ave_calc_coord_view_part (ABSTRACT_EVENT *ae, int level)
{
  float lx, rx;
  ABSTRACT_EVENT *child = ae->down;
  if (ae->ave->draw_flag == 0) 
    return;

  if (ae->clps && ae->clps->ave->draw_flag) {
    vert_clps_calc_coord (ae->clps, level+2);
    ae->ave->x = ae->clps->ave->x;
    if (ae->ave->x > xmax)
      xmax = ae->ave->x;
    ae->ave->y = level;
  }
  else if (child && child->ave->draw_flag) {
    if (ae->clps)
      ave_calc_coord (child, level+3);
    else
      ave_calc_coord (child, level+2);
  
    /* Get the leftmost child's coordinate */
    lx = child->ave->x;

    /* Get the rightmost child's coordinate */

    while (child->next)
      child = child->next;
    
    rx = child->ave->x;


    ae->ave->x = (lx+rx)/2;
    if (ae->ave->x > xmax)
      xmax = ae->ave->x;
    ae->ave->y = level;
  }
  else {
    xmax += 2;
    ae->ave->x = xmax;
    ae->ave->y = level;
  }

  if (ae->slide_clps)
    horiz_clps_calc_coord (ae->slide_clps, level);
  return;
}

void vert_clps_calc_coord (COLLAPSE_NODE *cn, int level)
{
  if (cn == NULL)
    return;
  if (cn->ael) {
    int size = get_size_of_abstract_list (cn->ael);
    ABSTRACT_EVENT *ae_first = get_element_from_abstract_list (cn->ael, 0);
    ABSTRACT_EVENT *ae_last = get_element_from_abstract_list (cn->ael, size-1);
    ave_calc_coord_view_part (ae_first, level+2);
    if (cn->scn)
      horiz_clps_calc_coord (cn->scn, level+2);
    else {
      int size = get_size_of_list ((LINKED_LIST *)cn->ael);
      int i;
      for (i=1; i<size; i++) {
	ABSTRACT_EVENT *ae  = 
	  (ABSTRACT_EVENT *) get_element_from_list ((LINKED_LIST *)cn->ael, i);
	ave_calc_coord_view_part (ae, level+2);
      }
    }
    if (ae_first->ave->x < 0) {
      xmax += 2;
      cn->ave->x = xmax;
    }
    else {
      cn->ave->x = (ae_first->ave->x + ae_last->ave->x)/2;
      if (cn->ave->x > xmax)
	xmax = cn->ave->x;
    }
    vert_clps_calc_coord (cn->next, level);
  }
  else if (cn->down){ /* cn->down != NULL */
    COLLAPSE_NODE *child_first, *child_last;
    vert_clps_calc_coord (cn->down, level+2);
    child_first = child_last = cn->down;
    while (child_last->next)
      child_last = child_last->next;

    cn->ave->x = (child_first->ave->x + child_last->ave->x)/2;
  }

  cn->ave->y = level + 2;  /* 3 originally */
  return;
}

void ave_calc_coord (ABSTRACT_EVENT *ae, int level)
{
  float lx, rx;
  ABSTRACT_EVENT *child = ae->down;
  if (ae->ave->draw_flag == 0) {
    if (ae->next)
      ave_calc_coord (ae->next, level);
    return;
  }

  if (ae->clps && ae->clps->ave->draw_flag) {
    vert_clps_calc_coord (ae->clps, level+2);
    ae->ave->x = ae->clps->ave->x;
    if (ae->ave->x > xmax)
      xmax = ae->ave->x;
    ae->ave->y = level;
  }
  else if (child && child->ave->draw_flag) {
    if (ae->clps)
      ave_calc_coord (child, level+3);
    else
      ave_calc_coord (child, level+2);
  
    /* Get the leftmost child's coordinate */
    lx = child->ave->x;

    /* Get the rightmost child's coordinate */

    while (child->next)
      child = child->next;
    
    rx = child->ave->x;


    ae->ave->x = (lx+rx)/2;
    if (ae->ave->x > xmax)
      xmax = ae->ave->x;
    ae->ave->y = level;
  }
  else {
    xmax += 2;
    ae->ave->x = xmax;
    ae->ave->y = level;
  }

  if (ae->slide_clps)
    horiz_clps_calc_coord (ae->slide_clps, level);
  else if (ae->next)
    ave_calc_coord (ae->next, level);



  return;
}
