#include <X11/Intrinsic.h>
#include "defs.h"

#define  LOCAL_  0
#define  DESCENDANTS_ 1

typedef struct link_list {
  struct link_list *prev;
  struct link_list *next;
  void *contents;
} LINKED_LIST;

typedef LINKED_LIST INT_LINKED_LIST;
typedef LINKED_LIST ABSTRACT_EVENT_LIST;
typedef LINKED_LIST EVENT_DESC_LIST;
typedef enum oper { OperParen, OperPrecede, OperParallel, 
		      OperPush, OperPop, OperAt, OperAny, OperNone} OPER;


typedef struct ident {
  struct ident   *id_next;     /* chain of identifiers in same hash bucket */
  struct decl    *id_decl;
  char           *id_chars;    /* null terminated name of identifier */
  int            id_reserved;  /* 1 => reserved, 0 otherwise */
} IDENT;

typedef struct idlist {
  IDENT *id;
  struct idlist *next;
} IDLIST;

typedef enum decl_kind {
  DeclSet, DeclChain, DeclPchain, DeclTchain, DeclInt, DeclUnknown, DeclMatch,
DeclImap} DECL_KIND;

typedef struct decl {
  DECL_KIND decl_kind;
  struct decl *decl_next;
  union object *decl_object;
} DECL;

typedef struct ide {
  DECL_KIND decl_kind; 
  char *id_chars;
  struct syn_tree_list *stl;
} IDE;  

typedef struct NameList {
  IDENT *id;
  struct NameList *next;
} NAME_LIST;

typedef struct LineData {
  float x1, y1, x2, y2;  /* coordinates */
  float w1, h1, w2, h2;  /* correction factors */
  int color;
  struct LineData *next;
} LINE_DATA;

typedef struct RectData {
  int    bar_id, row_id, col_id;
  int    width, height;
  int    value;
  int    color;
  struct RectData *next;
} RECT_DATA;

typedef struct StrData {
  float x1, y1;
  char *value;
  struct StrData *next;
} STR_DATA;


typedef struct draw_canvas_data {
  LINE_DATA *line;
  RECT_DATA *rect;
  STR_DATA  *str;
} DRAW_CANVAS_DATA;


typedef struct core_set_obj {
  int dim;
  int size;
  int tail;
  int *elem;
} CORE_SET_OBJ;



typedef struct set_obj {
  IDE *ide;
  IDE *map;
  struct syn_tree_list *condsyn;
  char *name;
  CORE_SET_OBJ *core;
} SET_OBJ;

typedef struct wrt_list {
  char *marker;
  struct wrt_list *next;
} WRT_LIST;

typedef struct imap_obj {
  struct syn_tree_list *stl;
} IMAP_OBJ;

typedef struct tree_obj {
  DECL_KIND  decl_kind;
  IDE  *ide;
  OPER oper;
  int  cnt;
  char *name;
  char *marker;
  int  flag;
  WRT_LIST *wrt;     
  struct def_list *dfl;
  struct tree_obj *down;
  struct tree_obj *next;
} TREE_OBJ;

typedef enum on_kind {OnSome, OnAll} ON_KIND;

typedef struct pch_obj {
  TREE_OBJ *chain;
  SET_OBJ *set;
  ON_KIND on_kind;
} PCH_OBJ;




/* typedef enum {FALSE, TRUE} boolean; */
typedef enum {FALSE_, TRUE_} boolean;

/**********************************************************************
 * SO_Set contains the receive events corresponding to a "W" or an    *
 * "M" events, or the "W" or "M" event corresponding to a "R" event,  *
 * SO => Significant Other . They are computed by the                 *
 * "compute_significant_others() function.                            *
 * call in a lazy fashion. All the receives that have the same MsgID  *
 * as the "W" or the "M" event will be included in the WriteSet.      *
 *********************************************************************/

typedef struct SoSet {
  int size;
  struct EventPtrList *epl;  /* An array of events */
} SO_SET;


/************************************************************************
 * Event_Set is the struct that contains the relevant information about *
 * each event.                                                          *
 ***********************************************************************/


typedef enum {Phase, Read, Write, MCast, UserDefined, Unknown} EVENT_TYPE;

typedef struct Event {
  int          myid;       
  int          otherid;
  char         event_type[16];
  char         event_tag[MSG_TYPE_SIZE];
  int          event_size;

  char         event_name[MSG_TYPE_SIZE+10];
  int          ltime;    /* logical time */
  int          msg_id;   /* we compute this */
  int          instr_flag; /* user can set this for dynamic instrumentation */
  int          bp_flag; 
  int          bp_cut; 
  int          bpcount;
  int          error;
  int          active;       /* Has it already been matched or not */
  boolean            filtered;     /* True => Filtered out. Extension */
  SO_SET             *so_set;
  struct Event       *prev;
  struct Event       *next;
  struct Attribute   *attr;
} EVENT;


typedef struct EventPtrList {
  EVENT *ep;
  struct EventPtrList *next;
} EVENT_PTR_LIST;

typedef enum {OFF, ON} EV_SWTCH;

typedef struct ave_str {
  float   x, y;
  float   w, h;  /* correction factor */
  int    draw_flag;  /* 1 => should be drawn */
  int    done_flag;  /* 1 => already drawn */
} AVE_STR;

typedef struct WidgetInfo {
  GC     gc;          /* Can now have different GCs to bring back and forth */
  Widget w;
  int    height;
  int    width;
} WIDGET_INFO;

typedef struct SlideCollapseNode {
  AVE_STR *ave;
  int total;
  int lo;
  int hi;
  struct AbstractEvent *prev;
  struct AbstractEvent *next;
  struct WidgetInfo *winfo;
} SLIDE_COLLAPSE_NODE;

typedef struct CollapseNode {
  AVE_STR *ave;
  int     my_id, par_id;
  int  total, part;
  char *pattern;
  struct AbstractEvent *pae;
  struct CollapseNode  *parent;
  struct CollapseNode  *down;
  struct CollapseNode  *next;
  struct WidgetInfo    *winfo;
  SLIDE_COLLAPSE_NODE  *scn;
  ABSTRACT_EVENT_LIST  *ael;
} COLLAPSE_NODE;


typedef struct Attribute {
  char  *name;
  char  *atype;
  LINKED_LIST *val_list;
  int color;
  struct Attribute *next;
} ATTRIBUTE;

typedef struct CollapseChild {
  AVE_STR *ave;
  int  my_id;
  int  par_id;
  char *pattern;
  struct CollapseChild *down;
  struct CollapseChild *next;
  ABSTRACT_EVENT_LIST *ael;
  struct WidgetInfo *winfo;
} COLLAPSE_CHILD;


typedef struct AbstractEvent {
  AVE_STR *ave;
  int proc;   /* Normally -1, holds the proc_id for pchains */
  int error;  /* 1 => Match Error, 2 => Parallel Failure, 3 => Precede, 4=>dummy */
  int cnt;    /* How many times did it match? Default = 1 , -1 => dummy?*/
  int dummy;
  EV_SWTCH status;
  char *orig;
  char *name;
  char *marker;
  struct AbstractEvent *parent;
  struct AbstractEvent *prev;          /* sibling */
  struct AbstractEvent *down;          /* child */
  struct AbstractEvent *next;          /* sibling */
  struct CollapseNode *clps;
  struct SlideCollapseNode *slide_clps;
  EVENT *ep;
  struct Attribute *attr;
  struct WidgetInfo *winfo;
} ABSTRACT_EVENT;

typedef struct val_obj {
  int value;
} VAL_OBJ;

typedef union object {
  IMAP_OBJ imap_obj;
  SET_OBJ  set_obj;
  PCH_OBJ  pch_obj;
  TREE_OBJ tree_obj;
  VAL_OBJ  val_obj;
  ABSTRACT_EVENT abs_evnt_obj;
} OBJECT;

typedef enum {ApplyAbs, ApplyAdd, ApplySub, ApplyMult, ApplyDiv, 
		ApplyDistinct, ApplyDisjoint,
		ApplyEqual, ApplyGec, ApplyGtc, ApplyLec,
		ApplyLtc, ApplySize, ApplyGetPid, 
		ApplyProd, ApplySum, ApplyNone, ApplyChild, 
		ApplyLeft, ApplyRight,
		ApplyLeaf, ApplyParent, ApplyFirst, ApplyLast,
		ApplySibling, ApplyPart, ApplyExpandPart, ApplyExpandChild,
	        ApplyUnion, ApplySection, ApplySubSet, ApplySuperSet,
                ApplyCross, ApplySet} ExprOper;

typedef struct syn_tree {
  struct syn_tree_list *left;
  struct syn_tree_list *right;
  char *name;
  int value;
  ExprOper oper;
} SYN_TREE;

typedef struct int_list {
  int elem;
  struct int_list *next;
} IntList;

typedef struct syn_tree_list {
  SYN_TREE *st;
  struct syn_tree_list *next;
} SYN_TREE_LIST;


typedef enum def_kind {
  DefMatch, DefChain, DefPchain, DefTchain, DefSet, DefSetPlus, DefFor, 
  DefImap, DefAssign, DefSkip, DefFail} DEF_KIND;

typedef struct for_def {
  IDENT *index;
  int   lower;
  int   upper;
  int   by;
} FOR_DEF;

typedef struct reg_def {
  IDE *left;
  WRT_LIST *wrt;
  union object *right;
} REG_DEF;
  
typedef union def {
  FOR_DEF for_def;
  REG_DEF reg_def;
} DEF;

typedef struct def_list {
  DEF_KIND def_kind;
  DEF *def;
  struct def_list *next;
} DEF_LIST;




typedef struct hash_table {
  int size;
  EVENT_PTR_LIST **buckets;  /* array of pointers to EVENT_PTR_LIST */
} HASH_TABLE;


typedef struct bar_graph_node {
  ABSTRACT_EVENT *xae;
  ABSTRACT_EVENT_LIST *first, *last;
  RECT_DATA *discrete;
  RECT_DATA *aggr;
  int cnt;
  struct bar_graph_node *next;
} BAR_GRAPH_NODES;


typedef struct event_attribute_object {
  char *name;
  char *atype;
} EVENT_ATTR_OBJ;

typedef LINKED_LIST EVENT_ATTR_LIST;

typedef struct event_desc_object {
  char *name;
  char *utag;
  EVENT_ATTR_LIST *eal;
} EVENT_DESC_OBJ;



