#include <stdio.h>
#include "math.h"

#include "structs.h"
#include "array.h"
#include "decl.h"
#include "expr.h"
#include "global.h"
#include "scan.h"


/*
 ************************************************************************
 * Include file dependencies:                                           *
 *   structs.h :       Data Structures                                  *
 *   array.h:          find_name()                                      *
 *   decl.h:           allocate_decl()                                  *
 *                     replicate_set_object()                           *
 *                     replicate_tree_object()                          *
 *                     replicate_pchain_object()                        *
 *                     interpret_set_object()                           *
 *                     interpret_tree_object()                          *
 *                     interpret_pchain_object()                        *
 *   global.h:         get_match_variable_name()                        *
 *                     set_match_variable_name()                        *
 *                     get_trace_file_ptr()                             *
 *   expr.h:           allocate_value_object()                          *
 *   scan.h:           id_find_insert()                                 *
 ************************************************************************
*/             


#define Value(id) (id->id_decl->decl_object->val_obj.value)


int NodeId;  /* needed for node numbering */


/******************************************
 * var_init(): Initialize everything to 0 *
 ******************************************/

/*
 ************************************************************
 * FUNCTION:   allocate_name_object()                       *
 ************************************************************
*/

NAME_LIST *allocate_name_list(IDENT *id)
{
  NAME_LIST *nl = (NAME_LIST *) malloc (sizeof(NAME_LIST));
  nl->id = id;
  nl->next = NULL;
  return nl;
}

void add_node_to_name_list (IDENT *id)
{
  NAME_LIST *nl = allocate_name_list(id);
  NAME_LIST *tnl = get_match_variable_name();
  nl->next = tnl;
  set_match_variable_name (nl);
  return;
}



  
/*
 ******************************************************************
 * FUNCTION(S):  create different kinds of definitions (DEF_LIST) *
 *               + create_regular_definition()                    *
 *               + create_for_definition()                        *
 * DESCRIPTION:  There are two types of definition in Ariadne:    *
 *                 Regular definition                             *
 *                 For definition                                 *
 ******************************************************************
*/


DEF_LIST *create_reg_def (IDE *ide, OBJECT *obj, WRT_LIST *wrt)
{
  DEF_LIST *dfl = (DEF_LIST *) malloc (sizeof (DEF_LIST));
  dfl->def = (DEF *) malloc(sizeof (DEF));
  dfl->def->reg_def.left = ide;
  dfl->def->reg_def.wrt = wrt;
  dfl->def->reg_def.right = obj;
  dfl->next = NULL;
  return dfl;
}


DEF_LIST *create_for_def (IDENT *id, int lower, int upper, int by, 
			  DEF_LIST *tail)
{
  DEF_LIST *dfl = (DEF_LIST *) malloc (sizeof (DEF_LIST));
  dfl->next = tail;
  dfl->def_kind = DefFor;
  dfl->def = (DEF *) malloc (sizeof (DEF));
  dfl->def->for_def.index =  id;
  dfl->def->for_def.lower = lower;
  dfl->def->for_def.upper = upper;
  dfl->def->for_def.by = by;
  return dfl;
}


/*
 ******************************************************************
 * FUNCTION:    interpret_definition()                            *
 * DESCRIPTION: main interpreter for all definitions              *
 ******************************************************************
*/


void interpret_definition (DEF_LIST *df)
{
  IDENT *id;
  IDENT *LftId;
  PCH_OBJ *po;
  TREE_OBJ *to;
  ABSTRACT_EVENT *ae;
  ABSTRACT_EVENT *t_ae;
  int pid;


  if (df == NULL)
    return;

  switch (df->def_kind)
    {
    case DefFor:
      id = df->def->for_def.index;
      create_decl_int (id);
      Value(id) = df->def->for_def.lower;
      while (Value(id) < df->def->for_def.upper) {
	interpret_definition (df->next);
	Value(id) += df->def->for_def.by;
      }
      break;
    case DefImap:
      {
	IMAP_OBJ *imo = (IMAP_OBJ *) df->def->reg_def.right;
	LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	LftId->id_decl->decl_object = (OBJECT *) imo;
	interpret_definition (df->next);
	break;
      }
    case DefSet:
      {
	int i;
	SET_OBJ *so = 
	  replicate_set_object ((SET_OBJ *) df->def->reg_def.right);
	LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	if (interpret_set_object(so) == FALSE)
	  break;
	LftId->id_decl->decl_object = (OBJECT *) so;
	interpret_definition(df->next);
	break;
      }
    case DefSetPlus:
      {
	int i;
	SET_OBJ *so = ((SET_OBJ *) df->def->reg_def.right);
	LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	interpret_set_object(so);
		 
	LftId->id_decl->decl_object = (OBJECT *) so;
	interpret_definition(df->next);
	break;
      }
    case DefPchain:
      po = replicate_pchain_object((PCH_OBJ *) df->def->reg_def.right);
      if (interpret_pchain(po) == TRUE)
	{
	  LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	  LftId->id_decl->decl_object = (OBJECT *) po;
	}
      interpret_definition(df->next);
      break;
    case DefChain:
      to = replicate_tree_object((TREE_OBJ *) df->def->reg_def.right, DeclChain );
      if (interpret_tree_object(to) == TRUE)
	{
	  LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	  LftId->id_decl->decl_object = (OBJECT *) to;
	}
      interpret_definition(df->next);
      break;
    case DefTchain:
      to = replicate_tree_object((TREE_OBJ *) df->def->reg_def.right, DeclTchain );
      if (interpret_tree_object(to) == TRUE)
	{
	  LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	  LftId->id_decl->decl_object = (OBJECT *) to;
	}
      interpret_definition(df->next);
      break;
    case DefMatch:
      if (get_trace_file_ptr() == NULL) {
	printf("No tracefile has been opened for this session.");
	break;
      }
/*      var_init(); */
      to = replicate_tree_object ((TREE_OBJ *) df->def->reg_def.right, DeclTchain);
      if (interpret_tree_object (to) == TRUE) {
	LftId = interpret_identifier_tobe_defined (df->def->reg_def.left, 0);
	process_match_statement (to, LftId, df->def->reg_def.wrt);
      }
      free_tree_object(to);
      to = NULL;
      interpret_definition(df->next);
      break;
    default:
      break;
    }
  return;

}


/*
 ************************************************************************
 * FUNCTION:      filter_interp_definition()                            *
 * DESCRIPTION:   Three kinds of user defined filter directives are     *
 *                allowed in Ariadne, all at the chain level.           *
 *                (1) DefAssign:  ch x = W {a = pid(W);} @ R;           *
 *                (2) DefSkip:                                          *
 *                ch x = W {a = pid(W);} @ R W @ R {sof eq(pid(R),a);}; *
 *                (2) DefFail:                                          *
 *                ch x = W {a = pid(W);} @ R W @ R {aof eq(pid(R),a);}; *
 ************************************************************************
*/



int filter_interp_def (DEF_LIST *df, IDLIST **idl, ABSTRACT_EVENT *context)
{
  IDENT *id;
  IMAP_OBJ *imo;
  int result;
  VAL_OBJ *vo;
  int x;
  if (df == NULL)
    return 1;

  switch (df->def_kind) {
  case DefAssign:
    id = interpret_identifier_tobe_defined (df->def->reg_def.left,1);
    add_to_id_list (idl, id);
    imo = (IMAP_OBJ *) df->def->reg_def.right;
    result = get_int (evaluate_int_list(imo->stl, context));
    vo = allocate_value_object (result);
    id->id_decl->decl_object = (OBJECT *) vo;
    x = filter_interp_def (df->next, idl, context);
    return x;
  case DefSkip:
    imo = (IMAP_OBJ *) df->def->reg_def.right;
    result = get_int (evaluate_int_list(imo->stl, context));
    if (result == 0)
      return result;
    x = filter_interp_def (df->next, idl, context);
    return x;
  case DefFail:
    imo = (IMAP_OBJ *) df->def->reg_def.right;
    result = get_int (evaluate_int_list(imo->stl, context));
    if (result == 0)
      return -1;
    x = filter_interp_def (df->next, idl, context);
    return x;
  }
}
    





