#include <stdio.h>
#include "structs.h"

#include "strs.h"
#include "event.h"
#include "poeg.h"


FILE *fp_trace, *fp_map, *fp_out;
int ltime[MAX_PROCS];



int get_msg_id (int pid, int count)
{
  int msg_id = (pid << 24)+count;
  return msg_id;
}

void set_msg_id_of_matching_read (EVENT *ep)
{
  int msgid = MsgID(ep);
  int otherid = OID(ep);
  int myid = PID(ep);
  int total_procs = get_total_proc();
  int pid;

  if (otherid == -1) { /* multicast */
    for(pid=0; pid<total_procs; pid++) {
      EVENT *tep = get_first_event_on_proc(pid);
      if (pid == myid)
	continue;
      while (tep) {
	if ((EventType(tep) == 'R') && (OID(tep) == myid) &&
	    (MsgID(tep) == AA_NONE)) {
	  MsgID(tep) = msgid;
	  break;
	}
	tep = tep->next;
      }
    }
  }
  else {
    EVENT *tep = get_first_event_on_proc(otherid);
    while (tep) {
      if ((EventType(tep) == 'R') && (OID(tep) == myid) &&
	  (MsgID(tep) == AA_NONE)) {
	MsgID(tep) = msgid;
	break;
      }
      tep = tep->next;
    }
  }
  
}


void set_logical_time(int pid) 
{
  EVENT *ep = get_current_event_on_proc(pid);
  
  while (ep) {
    if (LTIME(ep) == AA_NONE) {
      if (EnumType(ep) == Read) {
	int cand = get_logical_time_from_other_proc (OID(ep), MsgID(ep));
	if (cand > ltime[pid])
	  ltime[pid] = cand;
      }
      LTIME(ep) = ++ltime[pid];
    }
    ep = ep->next;
  }
}

int get_logical_time_from_other_proc (int pid, int msgid)
{
  EVENT *ep = get_current_event_on_proc(pid);
  
  while(ep && (MsgID(ep) != msgid)) {
    if (LTIME(ep) == AA_NONE) {
      if (EnumType(ep) == Read) {
	int cand = get_logical_time_from_other_proc (OID(ep), MsgID(ep));
	if (cand > ltime[pid])
	  ltime[pid] = cand;
      }
      LTIME(ep) = ++ltime[pid];
    }
    ep = ep->next;
  }
  
  if (LTIME(ep) == AA_NONE)
    LTIME(ep) = ++ltime[pid];

  
  return (LTIME(ep));
}
    
      
      
void add_logical_time_and_msg_id ()
{
  int pid;
  int total_procs = get_total_proc();
  
  /* Number all writes, Multicasts and Reads */
  for (pid=0; pid<total_procs; pid++) {
    int cnt = 1;
    EVENT *ep = get_first_event_on_proc(pid);
    while (ep) {
      if ((EventType(ep) == 'W') || (EventType(ep) == 'M')) {
	++cnt;
	MsgID(ep) = get_msg_id (pid, cnt);
	set_msg_id_of_matching_read (ep);
      }
      ep = get_event_next(ep);
    }
  }

  /* Now set the logical times: the hard part */
  for (pid=0; pid<total_procs; pid++)
    set_logical_time (pid);
    
  
}


main(int argc, char **argv)
{
  int i;
  int msg_type;
  char tag[80];
  int cnt = 0;
  int pid;
  if (argc < 3) {
    printf("usage: ave2ave <trace_file> <outfile> \n");
    exit(0);
  }

  fp_trace = fopen(argv[1], "r");
  fp_out = fopen(argv[2], "w");
  
  for (pid=0; pid<MAX_PROCS; pid++)
    ltime[pid] = 0;

  LoadTraceFile(argv[1], "START", 1, "END", 1);

  add_logical_time_and_msg_id ();

  for (pid=0; pid<MAX_PROCS; pid++) {
    EVENT *ep = get_first_event_on_proc(pid);
    while(ep) {
      fprintf(fp_out,"%d %d %c %d %d\n", 
	      ep->myid, ep->otherid, ep->event_type,
	      ep->msg_id, ep->ltime);
      ep = ep->next;
    }
  }
  
}
