#include <stdio.h>
#include "strs.h"
#include "defs.h"

/*
 ********************************************************************
 * Program expects the whole trace file to be in argv[1]            *
 * and it generates individual traces in                            *
 *        "pgon_aa_trace0", "pgon_aa_trace1" etc.                   *
 * Reads the file argv[2]  which specifies how the strings          *
 * specifying user tags relate to the message_type s in Paragon     * 
 * program                                                          *
 ********************************************************************
*/

#define MAX_TAGS   2000
FILE *fp_trace, *fp_map, *fp_out;
typedef struct tag_map {
  int  msg_type;      /* paragon's msg_type */
  char event_tag[MSG_TYPE_SIZE];
} TAG_MAP;


TAG_MAP user_tags[MAX_TAGS];
static int current_node = -1;


initialize_tag_database()
{
  int cnt = 0;
  int i;
  int msg_type;
  char tag[MSG_TYPE_SIZE];

  /* init of tags */
  for (i=0; i<MAX_TAGS; i++) {
    user_tags[i].msg_type = -1;
    user_tags[i].event_tag[0] = '\0';
  }


  while (fscanf(fp_map, "%d %s\n", &msg_type, &tag) != -1) {
    user_tags[cnt].msg_type = msg_type;
    sprintf(user_tags[cnt].event_tag,"%s", tag);
    cnt++;
  }
}

char *get_user_tag (int msg_type)
{
  int i;
  for (i=0; i<MAX_TAGS; i++) {
    if (user_tags[i].event_tag == NULL)
      return NULL;
    if (user_tags[i].msg_type == msg_type)
      return user_tags[i].event_tag;
  }
  return NULL;
}
  

int get_msg_type (char *event_tag)
{
  int i;
  for (i=0; i<MAX_TAGS; i++) {
    if (user_tags[i].event_tag == NULL)
      return AA_UNDEFINED;
    if ( !StrCmp(user_tags[i].event_tag, event_tag))
      return user_tags[i].msg_type;
  }
  return AA_UNDEFINED;
}


void convert_ave_paragon()
{
  int myid, otherid, size, bpoint;
  char msg_rw;
  char event_tag[MSG_TYPE_SIZE];
  int pgon_msg;

  while (fscanf(fp_trace,"%d %d %c %s %d %d\n", &myid, &otherid, &msg_rw,
		&event_tag, &size, &bpoint) != -1) {
    int msg_type = get_msg_type (event_tag);
    if (current_node != myid) {
      char fname[MSG_TYPE_SIZE];
      sprintf(fname,"pgon_aa_trace%d",myid);
      fp_out = fopen(fname,"w");
      current_node = myid;
    }
    if (msg_rw == 'W')
      pgon_msg = AA_SEND;
    else if (msg_rw == 'M')
      pgon_msg = AA_MCAST;
    else if (msg_rw == 'R')
      pgon_msg = AA_RECEIVE;
    fprintf (fp_out, "%d %d %d %d %d %d\n", myid, otherid, pgon_msg,
	     msg_type, size, bpoint);
  }
}


main(int argc, char **argv)
{
  int i;
  int msg_type;
  char tag[80];
  int cnt = 0;

  if (argc < 3) {
    printf("usage: ave2pgon <map_file> <trace_file>\n");
    exit(0);
  }

  fp_map = fopen(argv[1], "r");
  fp_trace = fopen(argv[2], "r");

  initialize_tag_database();
  convert_ave_paragon();
  
}
