/* --------------------------------------------------
 *
 *  Darryl Brown
 *  
 *  October, 1994
 *
 *  astrlib.c
 *  
 *  This file is the source file for the implementation of 
 *  several useful string functions.
 *
 * --------------------------------------------------
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
#include "astrlib.h"
/*
 *
 *  char *strsave(char *str)- accepts a string, and mallocs space 
 *       for it, or returns an error if malloc fails...;
 *
 *  int strsame(char *str1, char *str2) - returns true (1) is strings
 *       are the same, false (0) if not.
 *
 *  int isAllDigits(char *str1) - returns true (1) if string
 *       is made of all digits.
 *
 *  char strappend(char *str1, char *str2) - returns concatenation
 *       of str2 and str2, allocating memory for the result.  
 *       It does not free any space used by str1 and str2.
 *
 *  strtokStruct *initstrtok(char *buf) -
 *       initializes a structure for parsing functions that
 *       work like strtok, except that there is a unique
 *       buffer associated with each call to initastrtok.
 *       In this way, you can use the functions on more
 *       than one string at a time.
 *
 *  char *astrtok(strtokStruct *st, char *delim) - works
 *       exactly like strtok (except you supply the 
 *       string tok struct 'st'.
 *  
 *  void freestrtok(strtokStruct *st) - frees all space 
 *       associated with a string tok struct.
 *
 */

/*  strsave  -- accepts a string, and mallocs space for it, or;
 *    returns an error if malloc fails...;
 */
char * strsave(char * str)
{
  char * tmp;
  if ((tmp = (char *) malloc ( strlen(str)+1 )) == NULL) 
    {
      printf("malloc failed\n");
      exit(1);
    }
  strcpy (tmp, str);
  return tmp;
}

/*
 *  int strsame(char *str1, char *str2) - returns true (1) is strings
 *       are the same, false (0) if not.  Note that if either (or
 *       both) are NULL, this func returns false.
 *
 */
int strsame(char *str1, char *str2) {
  if ((str1 == NULL) || (str2 == NULL)) {
    return 0;
  }
  return (strcmp(str1, str2) == 0);
}

/*
 *
 *  char *strappend(char *str1, char *str2) - returns concatenation
 *       of str2 and str2, allocating memory for the result.  
 *       It does not free any space used by str1 and str2.
 */
char *strappend(char *str1, char *str2) {
  char * tmp;
  if ((tmp = (char *) 
       malloc(sizeof ( strlen(str1)+ strlen(str2)+1 ))) == NULL) 
    {
      printf("malloc failed\n");
      exit(1);
    }
  strcpy (tmp, str1);
  strcat (tmp, str2);
  return tmp;
}
  

/*
 *
 *  int isAllDigits(char *str1) - returns true (1) if string
 *       is made of all digits.
 *
 */
int isAllDigits(char *str1) {
  char * tmp = str1;

  while (tmp) {
    if (!isdigit(*tmp)) {
      return 0;
    }
  }
  return 1;
}
  

/*
 *  strtokStruct *initstrtok(char *buf) -
 *       initializes a structure for parsing functions that
 *       work like strtok, except that there is a unique
 *       buffer associated with each call to initastrtok.
 *       In this way, you can use the functions on more
 *       than one string at a time.
 *
 */
strtokStruct *initstrtok(char *buf) {
  strtokStruct *st;

  if ((st = (strtokStruct *) 
       malloc(sizeof(strtokStruct))) == NULL) {
    printf("malloc failed");
    exit(1);
  }
  
  st->curPtr = st->buf = buf;
  st->delimStr = NULL;

  return st;
}

/* 
 * char *sstrtok(strtokStruct *st, char *delim) - works
 *       exactly like strtok (except you supply the 
 *       string tok struct 'st').  Note that this 
 *       only returns a non-zero length delimited 
 *       string, or else NULL (as opposed to astrtok,
 *       which will return zero length strings);
 *  
 */
char *sstrtok(strtokStruct *st, char *delim) {
  char *tmp, *cp;
  int dlen;

  cp = st->curPtr;

  if ((cp == NULL) || (*cp == '\0')) return NULL;
  tmp = strpbrk(cp, delim);

  if (tmp == NULL) {
    /* nothing left, put at end of string */
    tmp = (cp + strlen(cp));
    /* determine length of token */
    dlen = tmp-(cp);
    st->curPtr += dlen;
  } else {
    /* there is something left in the string */
    if (tmp == cp) {
      /* they are at same place, so keep looking for
       * a non-zero length delimeted string 
       */
      st->curPtr += 1;
      return astrtok(st, delim);
    }
    /* determine length of token */
    dlen = tmp-(cp);
    st->curPtr += (dlen+1);
  }

  /* allocate memory (if necessary) */
  if (!(st->delimStr) || (strlen(st->delimStr) < dlen)) {
    if (st->delimStr) free(st->delimStr);
    if ((st->delimStr = (char *)malloc(dlen+1)) == NULL) {
      fprintf(stderr, "malloc failed");
      exit(1);
    }
  }

  strncpy(st->delimStr, cp, dlen); 
  /* put terminator on end of string */
  *(st->delimStr+dlen) = '\0';
    

  return st->delimStr;
}

/* 
 * char *astrtok(strtokStruct *st, char *delim) - works
 *       exactly like strtok (except you supply the 
 *       string tok struct 'st'.
 *  
 */
char *astrtok(strtokStruct *st, char *delim) {
  char *tmp, *cp;
  int dlen;

  cp = st->curPtr;

  if (*cp == '\0') return NULL;
  tmp = strpbrk(cp, delim);

  if (tmp == NULL) {
    /* nothing left, put at end of string */
    tmp = (cp + strlen(cp));
    /* determine length of token */
    dlen = tmp-(cp);
    st->curPtr += dlen;
  } else {
    /* determine length of token */
    dlen = tmp-(cp);
    st->curPtr += (dlen+1);
  }

  /* allocate memory (if necessary) */
  if (!(st->delimStr) || (strlen(st->delimStr) < dlen)) {
    if (st->delimStr) free(st->delimStr);
    if ((st->delimStr = (char *)malloc(dlen+1)) == NULL) {
      fprintf(stderr, "malloc failed");
      exit(1);
    }
  }

  strncpy(st->delimStr, cp, dlen); 
  /* put terminator on end of string */
  *(st->delimStr+dlen) = '\0';
    

  return st->delimStr;
}

/*
 *  char freestrtok(strtokStruct *st) - frees all space 
 *       associated with a string tok struct.
 *
 */
void freestrtok(strtokStruct *st) {
  if (st->delimStr) free(st->delimStr);
  free(st);
}

char * trimLeadingSpaces(char *st) {
  char *tmp = st;
  while (*tmp == ' ') {
    tmp++;
  }
  return tmp;
}

int wordSame(char *str1, char *str2) {
  /* tests to see if first non-blank characters
   * in each string are the same.
   */
  char  *idx1 = NULL, *idx2 = NULL;
  
  if (strsame(str1, str2)) {
    /* if they are the same to start with, fine... 
     */
    return 1;
  }
  if (((idx1 = strchr(str1, ' ')) == NULL) &&
      ((idx2 = strchr(str2, ' ')) == NULL)) {
    /* if niether have spaces in them, they are
     * not the same
     */
    return 0;
  }
  if (idx1 != 0) {
    /* it means we did find a space in str1, but
     * we never tried to find a space in str2.
     * But we need to try in case str2 also has a space
     * but is longer than str1 
     */
    if (((idx2 = strchr(str2, ' ')) != NULL) &&
	((idx1-str1) != (idx2-str2))) {
      /* we found a space in str2 also, so if they are 
       * not the same, then the first word in each string
       * must be different.  If they are the same, test that
       * number of characters.
       */
      return 0;
    }
    return (strncmp(str1, str2, (idx1-str1))==0);
  } else if (idx2 != 0) {
    /* there were no spaces in str1, check length of str1.
     * if it is longer than the space index into str2, then
     * they are different strings.  If it is not longer,
     * we can compare (idx2-str2) characters.
     */
    if (strlen(str1) > (idx2-str2)) {
      return 0;
    }
    return (strncmp(str1, str2, (idx2-str2))== 0);
  }
  return 0;
}
	

