/* --------------------------------------------------
 *
 *  Darryl Brown
 *  
 *  October, 1994
 *
 *  distApi.c
 *  
 *  This file implements the protocol for the API
 *  between an external application the a parallel
 *  pC++ program.
 *
 * --------------------------------------------------
 */

/* see if this file has already been included; */
#ifndef BRK_EXEC_DEF

#include "GlobalStruct.h"
#include "stdio.h"
#include "stdlib.h"
#include "astrlib.h"
#include "clServ.h"
#include "typeApi.h"
#include "cargo.h"
#include "collInst.h"
#include "distApi.h"

#include "compatible.h"
#if SYS5
#include <string.h>
#else
#include <strings.h>
#endif
/* 
 * --------------------------------------------------------------
 *  GLOBAL DECLARATIONS
 * --------------------------------------------------------------
 */


/* toggle variable for whether or not to breakpoint */
int brk_BREAKPOINT_TOGGLE_VAR=1;

/*
 * The variable that specifies whether or not access funcs
 * have been initialized.
 */
int DA_AF_Init = 0;

/*
 * The variable that holds the connection number
 */
int brk_ConnectionNum;

/*
 * continue var, for setting #of breakpoints to skip before stopping;
 */ 
int DA_Continue = 0;

/*
 * these variables will be set to identify the location
 * before call to generated barrier
 */
int brk_funcId = -1;
char *brk_fileName = "";
int brk_lineNumber = -1;

// a flag for invoking parallel functions. 
static int BRK_DO_PARALLEL;
static int Brk_Do_Exit = 0;

extern void brk_AccessFunctionTableInit();

extern "C" {
/*
 * global structure for including pcxx_param,
 *  allowing access to environment variables.
 */
extern pcxx_PARAM_t pcxx_param;
}


/*
 *  ------------------------------------------------------------
 *  Extern Declarations
 *  ------------------------------------------------------------
 */
extern connRoleEnum ConnectionRole;

/* 
 * --------------------------------------------------------------
 *  ACCESS FUNCTIONS
 * --------------------------------------------------------------
 */


/*
 *  char *DA_getTypes() -
 *       retrieve all the types for 
 *       the parallel data structures.
 *
 */
aPackage *DA_getTypes() {
  aPackage *m;
  int i;
  char *pbuf, store[10000];

  m = CRG_InitPackage(CRG_shared);
  /* print the list of composites */
  for (i = 0; i < TA_CompListIdx; i++) {
    pbuf = TA_writeComposite(TA_CompositeList[i]);
    strcat(store, pbuf);
    free (pbuf);
    strcat(store, "\n");
  }
  
  /* print the list of type descriptions */
  for (i = 0; i < TA_TypeListIdx; i++) {
    pbuf = TA_writeTypeDesc(TA_typeDescList[i]);
    strcat(store, pbuf);
    free(pbuf);
    strcat(store, "\n");
  }
  m->data = strsave(store);
  m->length = strlen(store)+1;
  return m;

}


/*
 *  aPackage *DA_getVarType(strtokStruck *tokStr) {
 *       retrieve the type of a particular
 *       variable.
 *
 */
aPackage *DA_getVarType(strtokStruct *tokStr) {
  aPackage *m;
  char *tmp;

  m = CRG_InitPackage(CRG_shared);
  /* print the list of composites */
  tmp = astrtok(tokStr, " ");

  if (tmp) {
    m->data = TA_getVarType(tmp);
    m->length = strlen((char *)m->data)+1;
    return m;
  }
  m->data = "no type";
  m->length = strlen((char *)m->data)+1;
  return m;
}


/*
 *  aPackage *DA_getInstances() {
 *       retrieve all the Instances of 
 *       the parallel data objects. 
 *
 */
aPackage *DA_getInstances() {
  char pbuf[5000];
  aPackage *m;

  strcpy(pbuf,"");
  m = CRG_InitPackage(CRG_shared);
  /* print the list of collections */
  DA_collInstanceTable->printToBuf(pbuf);
  
  m->data = strsave(pbuf);
  m->length = strlen(pbuf)+1;
#if DEBUG      
  printf("instances: \"%s\"\n\t\tbytes %d\n\n",
	 (char *)m->data, m->length);
#endif
  return m;
}


/*
 *  aaPackage *DA_getFuncInfo() {
 *       retrieve the function ID
 *       of the current breakpoint.
 *
 */
aPackage *DA_getFuncInfo() {
  aPackage *m;
  char tbuf[100];

  m = CRG_InitPackage(CRG_shared);
  /* report funcId */
  sprintf(tbuf, "id=%d#fn=%s#ln=%d#", brk_funcId, brk_fileName,
	  brk_lineNumber);
  
  m->data = strsave(tbuf);
  m->length = strlen(tbuf)+1;
  return m;
}


static char *BRK_PARALLEL_FUNC;
static DA_collInstance *BRK_COLLINST;

/*
 *  void DA_callAccessFuncFromInst(DA_collInstance collInst)
 *      calls the access function for the given collection
 *      instance. 
 *
 */
void DA_callAccessFuncFromInst(DA_collInstance *collInst, 
			       int elemNum,
			       char *fname, int accessType) {
  char buf[50];
  if (accessType == BRK_GET_INSTANCE){ 
    sprintf(buf, "Collection %s <%s>", collInst->collName, 
	    collInst->elemName);
  } else if (BRK_CALL_USER_MEM_FUNC == accessType) {
    sprintf(buf, "%s <%s>::%s", collInst->collName, 
	    collInst->elemName, fname);
  } else if (BRK_CALL_USER_MEM_FUNC_IN_PARALLEL == accessType) {
    //only set up the call, dont actually do it, have to
    // wait until other threads exit barriers before we can
    // make this call.
    sprintf(buf, "%s <%s>::%s", collInst->collName, 
	    collInst->elemName, fname);
    BRK_COLLINST = collInst;
    BRK_PARALLEL_FUNC = strdup(fname);
    return;
  } 
  TA_callAccessFunc(buf, 
		    (void *) collInst->collectionPtr, 
		    elemNum, 
		    fname);
}


/*
 *  void DA_callAccessFuncFromInst(DA_collInstance collInst)
 *      calls the access function for the given collection
 *      instance. 
 *
 */
void DA_callAccessFuncFromInstOverAll(DA_collInstance *collInst, 
			       char *fname, int accessType) {
  char buf[50];
  int elemNumi, elemNumj, maxi, maxj;

  if (accessType == BRK_GET_ALL_ELEMS) {
    sprintf(buf, "Collection %s <%s>", collInst->collName, 
	    collInst->elemName);
  } else if (accessType == BRK_CALL_USER_MEM_FUNC_OVER_ALL) {
      sprintf(buf, "%s <%s>::%s", collInst->collName, 
	      collInst->elemName, fname);
  }
  
  if (collInst->numDims == 1) {
    maxi = collInst->dim1size;
    for (elemNumi = 0; elemNumi < maxi; elemNumi++) {
      TA_callAccessFunc(buf, 
			(void *) collInst->collectionPtr, 
			elemNumi, 
			fname);
    }
  } else {   
    /* must be 2 dimensional */
    maxi = collInst->dim1size;
    maxj = collInst->dim2size;
    for (elemNumi = 0; elemNumi < maxi; elemNumi++) {
      for (elemNumj = 0; elemNumj < maxj; elemNumj++) {
        TA_callAccessFunc(buf, 
			  collInst->collectionPtr, 
			  collInst->elemIndex(elemNumi, elemNumj),
			  fname);
      }
    }
  }
}
   
/*
 *  void DA_callUserFunc(strtokStruct *tokStr);
 *      calls the access function for the given collection
 *      instance. 
 *
 */
void DA_callUserFunc(strtokStruct *tokStr) {
  char *tmp;
  tmp = astrtok(tokStr, " ");
#if DEBUG      
  printf("calling user func %s", tmp);
  fflush(stdout);
#endif
  TA_callUserFunc(tmp);
}

/*
 *  void DA_getUserFuncs() {
 *      send a list of the user defined functions
 *
 */
void DA_getUserFuncs() {
  char *tmp;
  aPackage *m;
  m = CRG_InitPackage(CRG_shared);
  tmp = TA_writeUserDefFuncs();
  CRG_Pack(m, tmp, strlen(tmp)+1);
  CRG_SendPackage(brk_ConnectionNum, m);
}

/*
 *  void DA_getUserMemFuncs() {
 *      send a list of the user defined functions
 *
 */
void DA_getUserMemFuncs() {
  char *tmp;
  aPackage *m;
  m = CRG_InitPackage(CRG_shared);
  tmp = TA_writeUserDefMemFuncs();
  CRG_Pack(m, tmp, strlen(tmp)+1);
  CRG_SendPackage(brk_ConnectionNum, m);
}




/*
 *  void DA_invalidCollSpec(char *argstr) {
 *       returns an invalid collection spec message.
 *       
 */
void DA_invalidCollSpec(char *argstr, aPackage *m) {
  char buf[100];
  sprintf(buf, "invalid collection specification: %d\n", argstr);
  m->data = buf;
  m->length = strlen(buf)+1;
  CRG_SendPackage(brk_ConnectionNum, m);
  return;
}

/*
 *  int DA_isValidCollNum(int collnum)
 *       checks to see if collnum is valid 
 *       collection number.
 *
 */
int DA_isValidCollNum(int collnum) {
  return ((collnum >= 0)  && (collnum < DA_collInstanceTable->length));
}

/*
 *  aPackage *DA_getCollElem(strtokStruct *tokStr)
 *       retrieve a specific collection
 *       or field of that collection.
 *
 */
void DA_AccessCollElem(strtokStruct *tokStr, int accessType) {
  DA_collInstance * collInst;
  int collnum = 0, elemnum = 0, elemnum2 = 0;
  char fname[40], *arg, argstr[40];
  int numericArgc = 0, dum;
  aPackage *m;

  m = CRG_InitPackage(CRG_shared);
  *fname = 0;
  
  strcpy(argstr, "");
  while ((arg = astrtok(tokStr, " ")) != NULL) {
    strcat(argstr, strsave(arg));
    strcat(argstr, " ");
    if (sscanf(arg, "%d", &dum) == 1) {
      numericArgc++;
    }
  }

#if DEBUG
     printf("args: \"argstr \"%s\"\n", argstr); 
#endif

  /* Do different things depending on the number of arguments;
   */
  switch (numericArgc) {
  case 0: {
    DA_invalidCollSpec(argstr, m);
    return;
  }
  case 1: {
    if (accessType == BRK_GET_ALL_ELEMS) {
      /* check for collnum out of range;
       */
      *fname = 0;
      if (sscanf(argstr, " %d  %[^.]s",
		 &collnum, fname) >= 1) {
	if (!DA_isValidCollNum(collnum))  {
	  DA_invalidCollSpec(argstr, m);
	  return;
	}
	collInst = (DA_collInstance *) DA_collInstanceTable->getIth(collnum);
	DA_callAccessFuncFromInstOverAll(collInst, fname, accessType);
	return;
      }
    } else if (accessType == BRK_CALL_USER_MEM_FUNC_IN_PARALLEL) {
      *fname = 0;
      if (sscanf(argstr, " %d  %[^.]s",
		 &collnum, fname) >= 1) {
	if (!DA_isValidCollNum(collnum))  {
	  DA_invalidCollSpec(argstr, m);
	  return;
	}
	collInst = (DA_collInstance *) DA_collInstanceTable->getIth(collnum);
	DA_callAccessFuncFromInst(collInst, -1, fname, accessType);
	return;
      }
    } else {
      /* Not enough arguments to do anything.
       */
      DA_invalidCollSpec(argstr, m);
      return;
    }
  }
  case 2: {
    *fname = 0;
    if (sscanf(argstr, " %d %d %[^.]s",
	       &collnum, &elemnum, fname) >= 2) {
      if (!DA_isValidCollNum(collnum))  {
	DA_invalidCollSpec(argstr, m);
	return;
      }
      collInst = (DA_collInstance *) DA_collInstanceTable->getIth(collnum);
      if (collInst->numDims == 1) {
	/* must be a request to retrieve the whole element of a 
	 * single dimension collection.;
	 */
	DA_callAccessFuncFromInst(collInst, elemnum, fname, accessType);
	return;
      }
    } else {
      DA_invalidCollSpec(argstr, m);
      return;
    }
  }
  default: {
    /* 
     *   (1) a request to retrieve entire element of a two-dimensional 
     *       collection distribution.
     */
    *fname = 0;
    if (sscanf(argstr, " %d %d %d %[^.]s",
	       &collnum, &elemnum, &elemnum2,
	       fname) < 3) {
      DA_invalidCollSpec(argstr, m);
      return;
    }
    if (!DA_isValidCollNum(collnum))  {
      DA_invalidCollSpec(argstr, m);
      return;
    }
    collInst = (DA_collInstance *) DA_collInstanceTable->getIth(collnum);
    if (collInst->numDims < 2) {
      // Then, the last argument must be an element field;
      DA_invalidCollSpec(argstr, m);
      return;
    }
    DA_callAccessFuncFromInst(collInst, 
			      collInst->elemIndex(elemnum, elemnum2),
			      fname, accessType);
    return;
  }
  }
}


/*
 *  aPackage *DA_getRequest()
 *       retrieves the next request from the connection.
 *
 */
aPackage *DA_getRequest() {
  return CRG_ReceivePackage(brk_ConnectionNum);
}


/*
 *  int DA_requestToInt(char *req)
 *       converts a character request to an integer constant.
 *
 */
int DA_requestToInt(char *req) {

  if (strsame("getCollElem", req)) {
    return BRK_GET_INSTANCE;
  }
  if (strsame("getAllElems", req)) {
    return BRK_GET_ALL_ELEMS;
  }
  if (strsame("getTypes", req)) {
    return BRK_LIST_TYPES;
  }
  if (strsame("getVars", req)) {
    return BRK_LIST_INSTANCES;
  }
  if (strsame("getVarType", req)) {
    return BRK_GET_INSTANCE_TYPE;
  }
  if (strsame("getFuncInfo", req)) {
    return BRK_GET_FUNC_ID;
  }
  if (strsame("callUserFunc", req)) {
    return BRK_CALL_USER_FUNC;
  }
  if (strsame("getUserFuncs", req)) {
    return BRK_GET_USER_FUNCS;
  }
  if (strsame("callUserMemFunc", req)) {
    return BRK_CALL_USER_MEM_FUNC;
  }
  if (strsame("callUserMemFuncOverAll", req)) {
    return BRK_CALL_USER_MEM_FUNC_OVER_ALL;
  }
  if (strsame("callUserMemFuncInParallel", req)) {
    return BRK_CALL_USER_MEM_FUNC_IN_PARALLEL;
  }
  if (strsame("getUserMemFuncs", req)) {
    return BRK_GET_USER_MEM_FUNCS;
  }
  if (strsame("exit", req)) {
    return BRK_EXIT;
  }
  if (strsame("setContinue", req)) {
    return BRK_SET_CONTINUE;
  }

  return 0;
}


/*
 *  int DA_getContinue(strtokStruct *tokStr)
 *       gets the integer value of the continue parameter
 *       passed in. 
 *
 */
int DA_getContinue(strtokStruct *tokStr) {
  int i;
  char *cnum;
  cnum = astrtok(tokStr, " ");
  if (sscanf(cnum, " %d ", &i) != 1) {
    return 0;
  } else {
    return i;
  }
}



/*
 *  int DA_beginConnection()
 *       Initiate connection with external application.
 *
 */
int DA_beginConnection() {

  if (strsame(CONNECT_ROLE, "server")) {
    // I am a server;
    brk_ConnectionNum = TRNS_OpenServer
      (pcxx_param.SocketNum);
  } else {
    // I am the client;
#if DEBUG      
    printf("opening client connection on port num: %d on host: %s\n", 
	   pcxx_param.SocketNum, pcxx_param.HostAddr);
#endif
    brk_ConnectionNum = TRNS_OpenClient
      (pcxx_param.SocketNum, pcxx_param.HostAddr);
  }
  return brk_ConnectionNum;
}




/*
 *  void DA_cleanup()
 *       Close connection.
 *
 */
void DA_cleanup() {
  /*  
    int close(int); 
    close(brk_ConnectionNum);
    */
}


/*
 *  void DA_ServeRequest(void) {
 *       Serves the next request in the connection.
 *
 */
int DA_ServeRequest(void) {
  int i;
  char *tmp;
  aPackage *m, *request;
  strtokStruct *tokStr;

  request = DA_getRequest();

#if DEBUG      
  printf("serving request:  %s \n", (char *)request->data);
  fflush(stdout);
#endif

  if (((char *) request->data) == NULL) {
    // no request, just return
    return 0;
  }

  tokStr = initstrtok((char *)request->data);
  // get the first word, i.e. the command
  tmp = astrtok(tokStr, " ");

  switch (DA_requestToInt(tmp)) {
  case BRK_LIST_INSTANCES: {
    m = DA_getInstances();
    CRG_SendPackage(brk_ConnectionNum, m);
    return 1;
  } 
  case BRK_GET_FUNC_ID: {
    m = DA_getFuncInfo();
    CRG_SendPackage(brk_ConnectionNum, m);
    return 1;
  }
  case BRK_CALL_USER_FUNC: {
    DA_callUserFunc(tokStr);
    return 1;
  }
  case BRK_CALL_USER_MEM_FUNC: {
    DA_AccessCollElem(tokStr, BRK_CALL_USER_MEM_FUNC);
    return 1;
  }
  case BRK_CALL_USER_MEM_FUNC_OVER_ALL: {
    DA_AccessCollElem(tokStr, BRK_CALL_USER_MEM_FUNC_OVER_ALL);
#if 1    
    m = CRG_InitPackage(CRG_shared);
    CRG_Pack(m, "0", 1);
    CRG_SendPackage(brk_ConnectionNum, m);
#endif
    return 1;
  }
  case BRK_CALL_USER_MEM_FUNC_IN_PARALLEL: {
    BRK_DO_PARALLEL = 1;
    DA_AccessCollElem(tokStr, BRK_CALL_USER_MEM_FUNC_IN_PARALLEL);
#if 1    
    m = CRG_InitPackage(CRG_shared);
    CRG_Pack(m, "0", 1);
    CRG_SendPackage(brk_ConnectionNum, m);
#endif
    return 0;
  }
  case BRK_GET_USER_FUNCS: {
    DA_getUserFuncs();
    return 1;
  }
  case BRK_GET_USER_MEM_FUNCS: {
    DA_getUserMemFuncs();
    return 1;
  }
  case BRK_LIST_TYPES: {
    m = DA_getTypes();
    CRG_SendPackage(brk_ConnectionNum, m);
    return 1;
  }
  case BRK_GET_ALL_ELEMS: {
    /* lookup correct access function for each collection, and  */
    /* access the collections */
    DA_AccessCollElem(tokStr, BRK_GET_ALL_ELEMS);
    return 1;
  }
  case BRK_GET_INSTANCE: {
    /* lookup correct access function for each collection, and  */
    /* access the collections */
    DA_AccessCollElem(tokStr, BRK_GET_INSTANCE);
    return 1;
  }
  case BRK_GET_INSTANCE_TYPE: {
    /* lookup correct access function for each collection, and  */
    /* access the collections */
    DA_getVarType(tokStr);
    return 1;
  }
  case BRK_SET_CONTINUE: {
    i = DA_getContinue(tokStr);
    if (i > 0) {
#if 1
      printf("setting continue %d\n", i);
#endif
      DA_Continue = i;
    }
#if 1    
    m = CRG_InitPackage(CRG_shared);
    CRG_Pack(m, "0", 1);
    CRG_SendPackage(brk_ConnectionNum, m);
#endif
    brk_funcId = -1;
    brk_fileName = "";
    brk_lineNumber = -1;
    return 0;
  }
  case BRK_EXIT: {
    // stop executing program;
    DA_cleanup();
    BRK_DO_PARALLEL = 1;
    Brk_Do_Exit = 1;
  }
  default: {
#if 1    
    m = CRG_InitPackage(CRG_shared);
    CRG_Pack(m, "0", 1);
    CRG_SendPackage(brk_ConnectionNum, m);
#endif
    return 0;
  }
  }
#if DEBUG      
  printf("done servicing requests for this breakpoint.\n");
  fflush(stdout);
#endif
}



/*
 * perform clean up functions, tell client program
 * is exiting.
 */
void DA_finishUp() {
  aPackage *request;
  char *retstring = "";
  printf("Program Exiting, Breezy performing cleanup\n");

  // we need to get the client's request, but ignore it and
  // tell the client instead that the program is exiting.
  
  request = DA_getRequest();
  while (! strsame(retstring, "done OK")) {
    printf("request: %s\n", (char *) request->data);
    CRG_RecyclePackage(request);
    
    request = CRG_InitPackage(CRG_shared);

    CRG_Pack(request, "Program Exiting", strlen( "Program Exiting") + 1);
    CRG_SendPackage(brk_ConnectionNum, request);
    CRG_RecyclePackage(request);
    printf("done sending package\n");
    fflush(stdout);
    request = DA_getRequest();
    retstring = (char *) request->data;
  }
  DA_cleanup();
}
  


//----------------------------------------------------------;
void brk_Init () {

  // init sockets, 
  if (DA_beginConnection() == -1 ) {
    DA_cleanup(); exit(1);
  }

  Brk_Do_Exit = 0;

  // read in type info;
#if DEBUG      
  printf("reading tif file");
  fflush(stdout);
#endif
  TA_readTIFFile();

#if DEBUG      
  printf("I am about done ...");
  fflush(stdout);
#endif

  // initialize collection access functions;
  brk_AccessFunctionTableInit();
  DA_AF_Init = 1;

  // set up what to do at the exit of the program...

}

#ifdef __cplusplus
extern "C" {
#endif

void pcxx_BroadcastBytes(int flag, int length, void *buffer);
int pcxx_NodeNumber();

extern void pcxx_Barrier();




/*
 * void brk_EXITFUNC();
 * wrapper for DA_finishUp()
 * is exiting.
 */
void brk_EXITFUNC() {
  DA_finishUp();
}


/*
 *  void brk_CHECK_PARALLEL();
 *       Function called by parallel program to
 *       answer requests of external application.
 *
 */
void brk_CHECK_PARALLEL() {
  /*
   *  broadcast the value of BRK_DO_PARALLEL.  This is 
   *  originally only set in processor 0, so we need to
   *  propogate it to all elements.  Note that we can
   *  use pcxx_MyNodeNumber() for the flag to the
   *  broadcast call because the node with the correct value
   *  is always 0 (and 0 in this call means, I am supplying
   *  the data, non-zero means I am receiving the data).
   */
  pcxx_BroadcastBytes(pcxx_NodeNumber(), sizeof(BRK_DO_PARALLEL),
		      &BRK_DO_PARALLEL);

  if (BRK_DO_PARALLEL) {
    /*
     *  broadcast the value of Brk_Do_Exit.  This is 
     *  originally only set in processor 0, so we need to
     *  propogate it to all elements
     */
    pcxx_BroadcastBytes(pcxx_NodeNumber(), sizeof(Brk_Do_Exit),
			&Brk_Do_Exit);
    if (Brk_Do_Exit) {
      exit(1);
    }
    printf("parallel function %s\n",
	   BRK_PARALLEL_FUNC);
  } 
    
  //pcxx_Barrier();
  BRK_DO_PARALLEL = 0;
  return;
}
    
    
  


#ifdef __cplusplus
}
#endif


/*
 *  void brk_BREAKPOINT_EXECUTIVE() {
 *       Function called by parallel program to
 *       answer requests of external application.
 *
 */
#ifdef __cplusplus
extern "C" {
#endif
void brk_BREAKPOINT_EXECUTIVE() {

  if (DA_Continue > 0) {
#if DEBUG
    printf("decrementing continue %d\n", DA_Continue);
#endif
    DA_Continue--;
    return;
  }

  if (!brk_BREAKPOINT_TOGGLE_VAR) {
    return;
  }

  if (DA_AF_Init == 0) {
#if DEBUG      
    printf("initializing again\n");
#endif
    brk_Init();
  }
  
  BRK_DO_PARALLEL = 0;

  while(DA_ServeRequest()) {};
#if DEBUG      
  printf("done serving requests\n");
#endif
}
#ifdef __cplusplus 
}
#endif

#endif
