// *******************************************
//
// Darryl Brown 
// March, 94
// Sage utilities
//
// sageUtils.C
//
// *******************************************


#include "sage++user.h"
#include <stdlib.h>
#include <ctype.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

#define DEBUG 0

// an integer maximum func;
int max(int a, int b) {
  if (a>b) return a;
  return b;
}

//;
SgType *extractBaseType(SgType *ty) {
  if (ty->hasBaseType()) {
    return (extractBaseType(ty->baseType()));
  } else {
    return ty;
  }
}

SgExpression *extractFuncExpFromHedr(SgStatement *fhs) {
  
  SgExpression *funcPtr;

  funcPtr = fhs->expr(0);
  funcPtr = funcPtr->lhs();
  if (funcPtr->variant() == POINTST_OP) {
    // member function call..
    // member name is left hand side, func call is right->a
    funcPtr = funcPtr->rhs();
  } else if (funcPtr->variant() == DEREF_OP) {
    funcPtr = funcPtr->lhs();
  }
  if (!(funcPtr->variant() == FUNCTION_REF) &&
      !(funcPtr->variant() == FUNC_CALL)) {
    printf("error in addParame to func headr \n");
    return NULL;
  }
  return funcPtr;
}



//  add a parameter expression to a function expression,
//  both for func references and func calls.;
SgExpression *addParamExpToFuncExp(SgExpression *fs, SgExpression *vexp) {

  SgExprListExp *funcList, *tmpExpList;
  SgExpression *funcPtr;
  SgSymbol *vs;


  // try to find actual function expression (skipping past 
  //     pointer expression, for example).


  if (fs->variant() == FUNCTION_REF) {
    // It is a function reference.

    // Get the func symbol of parameter expression
    // to see if it is a pointer type.
    vs = vexp->symbol();
    if (vs && 
	(vs->type()->variant() == T_POINTER)) {
      // Create a deref op for the pointer type...;
      funcPtr = new SgExpression(DEREF_OP);
      funcPtr->setLhs(*vexp);
    } else {
      // else just point to the new param expression..;
      funcPtr = vexp;
    }
  } else if (fs->variant() == FUNC_CALL) {
    // It is a func call, point to the param expression.
    funcPtr = vexp;
  } else {
    // What is it?   Point to the param expression anyway, but give warning.
    funcPtr = vexp;
    printf("\nAdding parameter to expression node with wierd variant %d\n", 
	   fs->variant());
  }
  

  // put the func ptr exp in a list;
  funcList = new SgExprListExp();
  funcList->setValue(*funcPtr);

  // add this to the function ref expression;
  if (!fs->lhs()) {
    // if left is empty, put it in the left, it is the first argument;
    fs->setLhs(*funcList);
  } else {
    // there is something already in the left side, so append this new 
    //    expression to the expression list on the left side;
    tmpExpList = (SgExprListExp *) fs->lhs();
    
    // go to end of expression list...;
    while (tmpExpList->rhs()) {
      // there is something in the right hand side, append to the end;
      tmpExpList = (SgExprListExp *) tmpExpList->rhs();
    }

    // append it onto the right side;
    tmpExpList->setRhs(*funcList);
  }
  return funcList;
}

//  This function adds a parameter symbol to a function ref expression (fs).;
SgExpression *addParamToFuncExp(SgExpression *fs, SgSymbol *vs) {
  
  SgVarRefExp *vexp;

  // create a function pointer expression;
  vexp = new SgVarRefExp(*vs);
  return addParamExpToFuncExp(fs, vexp);
}

// add a parameter to a function header statement;
SgExpression *addParamToFuncStmt(char *paramName, SgType *varType, 
				 SgStatement *fhs) {
  SgVariableSymb *vs;

  vs = new SgVariableSymb(paramName, *varType);
  return addParamToFuncExp(extractFuncExpFromHedr(fhs), vs);
}



// add a parameter to a function header statement;
SgExpression *addParamToFuncCall(char *paramName, SgType *varType, 
				 SgStatement *fhs) {
  SgVariableSymb *vs;

  vs = new SgVariableSymb(paramName, *varType);
  return addParamToFuncExp(extractFuncExpFromHedr(fhs), vs);
}

//  This function adds a member function to a class.  
//  It requires the class declaration statement, the function
//  type, the function name, and the status of the function 
//  (public, private, protected);
//
//  (Note that it returns a pointer to the function expression it 
//  creates.  This is so that this expression can be used to add
//  parameters to the function via the addParamToFuncExp function 
//  above.)
//;
SgExpression *addFunctionToClass(SgClassStmt *cs, 
				 SgType *funcType, 
				 char *funcName, 
				 int status) {
  SgVarDeclStmt *vds;
  SgMemberFuncSymb *fs;
  SgExprListExp *funcList;
  SgExpression *fexp;

#define COLL_STATUS_NUM (256)

   // create function symbol;
  fs = new SgMemberFuncSymb(funcName, *funcType, 
			    *cs, status);

  // create a function pointer expression;
  fexp = new SgExpression(FUNCTION_REF);
  fexp->setSymbol(*fs);

  // put the func ptr exp in a list;
  funcList = new SgExprListExp();
  funcList->setValue(*fexp);
  funcList->setType(*funcType);

  // create var declaration statement;
  
  vds = new SgVarDeclStmt(*funcList, *funcType);
  cs->insertStmtAfter(*vds);

#if DEBUG 
  printf("adding %s function to class %s\n",
	 funcName, cs->symbol()->identifier());
#endif

  return fexp;

}

void addFuncDeclToSymb(SgStatement *fhdr, SgSymbol *fsymb) {
  fsymb->thesymb->entry.member_func.func_hedr = fhdr->thebif;

}


//  create Function definition.
//
//  (Note that it returns a pointer to the function expression it 
//  creates.  This is so that this expression can be used to add
//  parameters to the function via the addParam function 
//  above.)


SgFuncHedrStmt  *buildFuncDefStmt(char *funcName, SgType *funcType) {

  SgFuncHedrStmt *fhs;

  // create func header statement and give it the correct return type;
  fhs = new SgFuncHedrStmt(funcName);
  fhs->setReturnedType(*funcType);

  return fhs;
}



/*
 *  int funcDefinedInClass(SgStatement *clhdr, char *fname) {
 *      Checks to see if fname has already been defined
 *      in the class header clhdr
 */   
int funcDefinedInClass(SgStatement *clhdr, char *fname) {
  
  SgStatement *lst, *cur;
  SgVarDeclStmt *vds;
  SgMemberFuncSymb *mfs;

  lst = clhdr->lastNodeOfStmt();
  cur = clhdr;
  while (cur != lst) {
    cur = cur->lexNext();
    if (vds = isSgVarDeclStmt(cur)) {
      for (int i = 0; i < vds->numberOfSymbols(); i++) {
	if ((mfs = isSgMemberFuncSymb(vds->symbol(i)) ) && 
	    (strcmp(mfs->identifier(), fname)==0)) {
	  return 1;
	}
      }
    }
  }
  return 0;
}


SgStatement *lastExecutableOfBlock(SgStatement *st) {

  SgStatement *tmpst;
#if 1   
  tmpst = st->lastNodeOfStmt();
  if (tmpst->variant() == CONTROL_END) {
    return tmpst->lexPrev();
  } else {
    return tmpst;
  }
#else   
  SgStatement *last;
  last = st;
  tmpst = st->lexNext();
  while ((tmpst->controlParent() == st->controlParent()) &&
	 (tmpst->variant() != CONTROL_END)) {
#if DEBUG 
    printf ("looking for last executable, var (%d), cp var (%d)\n",
	    tmpst->variant(), tmpst->controlParent()->variant());
#endif
    last = tmpst ;
    tmpst = tmpst->lexNext();
  }
  return last;
#endif
}
    
//  This function creates a variable declaration statement inside a 
//  function.

SgSymbol *declareVarInFunc(SgStatement *fhs, SgType *vt, char *vname) {
  SgVariableSymb *vs;

  vs = new SgVariableSymb(vname, *vt);
  vs->declareTheSymbol(*fhs);
  return vs;
}


// Build array type;
SgType *buildArrayType(int numDims, int *dimSizes, 
		       SgType *baseType) {
  
  SgValueExp *ve;
  SgArrayType *at;

  // start with dummy type;
  at = new SgArrayType(*baseType);

  // now, set correct type;
  at->setBaseType(*(baseType));
  
  for (int i = 0; i < numDims; i++) {
    ve = new SgValueExp(dimSizes[i]);
    at->addRange(*ve);
  }
  return at;
}
			
// BuildDerivedCollectionType;
SgType *buildDerivedCollectionType(SgType *elementType, char *collName, 
				   SgType *collType) {

  SgSymbol *collSymb;
  SgDerivedCollectionType *ctype;

  collSymb = new SgSymbol(COLLECTION_NAME, collName);
  ctype = new SgDerivedCollectionType(*collSymb, *collType);
  ctype->setElementClass(*elementType);
  
  return ctype;
}


// build the assign op expression given right and left hand sides
SgExpression *buildAssignExp(SgExpression *lhs, SgExpression *rhs) {
  SgExpression *ae;
  
  // create assign_op expression;
  ae = new SgExpression(ASSGN_OP);
  ae->setLhs(*lhs);
  ae->setRhs(*rhs);
  ae->setType(*lhs->type());

	  return ae;
}

// build new assignment statement given basic  
// expressions for rhs and lhs;
SgStatement *buildAssignStmt(SgExpression *lhs, SgExpression *rhs) {
  SgExprListExp *elist;
  SgExpression *ae;
  SgStatement *cexp;

  ae = buildAssignExp(lhs, rhs);

  // create the expression list
  elist = new SgExprListExp();
  elist->setLhs(*ae);
  if (ae->type()) {
    elist->setType(*ae->type());
  }

  // return a CExpStmt;
  cexp = new SgCExpStmt(*elist);
  return cexp;
}     


// build cast expression given symbol and type to cast to ;
SgExpression *buildCastExp(SgSymbol *s, SgType *ty) {
  SgVarRefExp *vre;
  
  // create variable_ref expression;
  vre = new SgVarRefExp(*s);

  // create cast_op expression;
  return new SgCastExp(*ty, *vre);
}

// compose a cast statement;
SgStatement *composeCastAssignStatement(SgSymbol *lhs, SgSymbol *rhs) {
       
  SgVarRefExp *vre;
  SgExpression *cexp;

  // create variable_ref for lhs of assign op;
  vre = new SgVarRefExp(*lhs);

  // create cast expression
  cexp = buildCastExp(rhs, lhs->type());

  return buildAssignStmt(vre, cexp);
}

// findCoCompDeclStmt
SgStatement *findCompDeclStmt(char *cname, SgFile *f) {
  SgStatement *stmp;
  SgClassStmt *clSt;
  SgStructStmt *stSt;
  SgExpression *exp;
  SgVarRefExp *vexp;

  stmp = f->firstStatement();
  while (stmp) {

    if ((clSt = isSgClassStmt(stmp)) &&
	((clSt->variant() == COLLECTION_DECL) ||
	 (clSt->variant() == CLASS_DECL)) &&
	// and compare the name to the one were looking for
	(strcmp(clSt->name()->identifier(), cname) == 0)) {
      return stmp;
    }
    // check to see if it is a struct.
    if ((stSt = isSgStructStmt(stmp)) &&
	// grab the first expression of the struct statement,
	// this will be an expression list
	(exp = stSt->expr(0)) &&
	// grab the first expression of the expression list,
	// this should be a var_ref expression
	(vexp = isSgVarRefExp(exp->lhs())) &&
	// check if the name is the same
	(strcmp(vexp->symbol()->identifier(), cname) == 0)) {
      return stmp;
    }
    stmp = stmp->lexNext();
  }
  return NULL;
}

// findCollectionSymbol;
SgSymbol *findCollectionSymbol(char *cname, SgFile *f) {
  SgCollectionStmt *stmp;

  stmp = (SgCollectionStmt *) findCompDeclStmt(cname, f);
  if (stmp) {
    return stmp->name();
  } else {
    return NULL;
  }
}



//  builds a dereference expression from an expression;
SgExpression *buildDerefExpFromExp(SgExpression *sy) {
  
  SgExpression *dexp;

  dexp = new SgExpression(DEREF_OP);
  dexp->setLhs(*sy);
  if (sy->type()) {
    dexp->setType(*(new SgPointerType(*sy->type())));
  }
  return dexp;
}


//  builds a dereference expression;
SgExpression *buildDerefExpFromSym(SgSymbol *sy) {
  
  SgVarRefExp *vre;

  // create the variable ref expression
  vre = new SgVarRefExp(*sy);
  return buildDerefExpFromExp(vre);
}


//  buildCollGetElemExp
SgExpression *buildCollGetElemExp(SgExpression *collExp, SgExpression *elemNum) {
  
  SgExprListExp *elist;
  SgExpression *func;

  elist = new SgExprListExp();
  elist->setLhs(*elemNum);

  func = new SgExpression(FUNCTION_OP);
  func->setLhs(*collExp);
  func->setRhs(*elist);

  return func;
}

// determines if a statement is a func declaration &;
// returns the function symbol if it is, else NULL;
SgSymbol *isFuncDeclStmt(SgStatement *st) {
  SgExpression *tmpExp;
  if (st->variant() == VAR_DECL &&
      (tmpExp = st->expr(0)) &&
      (tmpExp = tmpExp->lhs()) &&
      (tmpExp->variant() == FUNCTION_REF) &&
      (tmpExp->symbol()->variant() == FUNCTION_NAME)) {
    return tmpExp->symbol();
  } else {
    return NULL;
  }
}
    

//   findFunctionSymb;
SgSymbol *findFunctionSymb(char *fname, SgFile *f) {
  SgStatement *stmp;
  SgSymbol *ts;
  
  stmp = f->firstStatement();
  while (stmp) {
    if ((stmp->variant() == FUNC_HEDR) &&
	 (strcmp(stmp->symbol()->identifier(), fname) == 0)) {
      return stmp->symbol();
    } else if ((ts = isFuncDeclStmt(stmp)) && 
	       (strcmp(ts->identifier(), fname) == 0)) {
      return ts;
    } else {
      stmp = stmp->lexNext();
    }
  }
  return NULL;
}

SgSymbol *extractSymbIfSame(SgExpression *e, char *fname) {

  char *tmp;
  SgSymbol *symb;

  if ((!e) || (!fname))  return NULL;

  if (e->variant() == DEREF_OP) {
    // retrieve child for object pointed to.;
    e = e->lhs();
    return extractSymbIfSame(e, fname);
  }
  
  // might be a function pointer;
  if (e->variant() == FUNCTION_OP) {
    // retrieve child for name of function
    e = e->lhs();
    return extractSymbIfSame(e, fname);
  }

  symb = e->symbol();
 
  if ((!symb) || !(tmp = symb->identifier())) {
    return NULL;
  }
  if (strcmp(tmp, fname) == 0) {
    return e->symbol();
  }

  return NULL;
}

SgSymbol *findVarRefInExpTree(char *fname, SgExpression *exp) {
  SgSymbol *s;
  
  if (!exp || (isSgValueExp(exp))) {
    return NULL;
  }
  
  if (s = extractSymbIfSame(exp, fname)) {
    return s;
  } else if (s = findVarRefInExpTree(fname, exp->lhs())) {
    return s;
  } else if (s = findVarRefInExpTree(fname, exp->rhs())) {
    return s;
  }

  return NULL;
}


//   find variable declaration of a given name 
//   beginning from given statement;
SgStatement *findVarDeclStmt(char *fname, 
			     SgStatement *stmp) {
  SgVarDeclStmt *var;
  SgSymbol *varsymb;

  while ( stmp ) {
    if ( var = isSgVarDeclStmt (stmp) ) {
      for (int i=0; i<var->numberOfSymbols(); i++) {
	varsymb = var->symbol(i);
	if (strcmp(fname, varsymb->identifier())==0) {
	  return stmp;
	}
      }
    }
    stmp = stmp->lexNext();
  }

#if DEBUG       
  printf("\nCouldnt find var decl %s.\n\n", fname);
  fflush(stdout);
#endif	
  return NULL;
}


//   find variable declaration symbol of a given name 
//   beginning from given statement;
SgSymbol *findVarDeclSymbFromStmt(char *fname, 
				  SgStatement *stmp) {
  SgVarDeclStmt *var;
  SgSymbol *varsymb;

  var = (SgVarDeclStmt *) findVarDeclStmt(fname, stmp);
  if (var != NULL) {
    for (int i=0; i < var->numberOfSymbols(); i++) {
      varsymb = var->symbol(i);
#if DEBUG
      printf("looking at symbol %s", varsymb->identifier());
      fflush(stdout);
#endif
      if (strcmp(fname, varsymb->identifier())==0) {
	return varsymb;
      }
    }
  }
#if DEBUG
  printf("did not find var decl symb %s", fname);
  fflush(stdout);
#endif
  return NULL;
}




// find a the first variable declaration of a given name in;
// a file;
SgSymbol *findVarDeclSymb(char *fname, SgFile *f) {
  return findVarDeclSymbFromStmt(fname, f->firstStatement());
}


//   find variable declaration of a given name beginning from given
//   statement;
SgStatement *findDefStmt(char *fname, 
			     SgStatement *stmp) {
  SgStatement *var;
  SgSymbol *varsymb;

  while ( stmp ) {
    if (( var = (SgStatement *) isSgClassStmt (stmp)) || 
	(var = (SgStatement *) isSgStructStmt (stmp)) || 
	(var = (SgStatement *) isSgCollectionStmt (stmp))  
	) {
      varsymb = var->symbol();
      if (strcmp(fname, varsymb->identifier())==0) {
	return stmp;
      }
    }
    stmp = stmp->lexNext();
  }
  
#if DEBUG       
  printf("\nCouldnt find var decl %s.\n\n", fname);
  fflush(stdout);
#endif	
  return NULL;
}


//   find variable definition of a given name beginning from given
//   statement;
SgSymbol *findDefSymbFromStmt(char *fname, SgStatement *stmp) {
  SgClassStmt *cs;
  SgSymbol *varsymb;

  cs = (SgClassStmt *) findDefStmt(fname, stmp);
  if (cs != NULL) {
    varsymb = cs->symbol();
    if (strcmp(fname, varsymb->identifier())==0) {
      return varsymb;
    }
  }
#if DEBUG
  printf("did not find var decl symb %s", fname);
  fflush(stdout);
#endif
  return NULL;
}


// find a the first variable declaration of a given name in;
// a file;
SgSymbol *findDefSymb (char *fname, SgFile *f) {
  return findDefSymbFromStmt(fname, f->firstStatement());
}


SgExpression *buildFuncCallExp(char *funcName, SgFile *f) {
  SgFunctionCallExp *fce;
  SgSymbol *fs;

  fs = findFunctionSymb(funcName, f);
  if (!fs) {
#if DEBUG
    printf("couldnt find function %s, creating ref to it\n", funcName);
#endif
    fs = new SgFunctionSymb(FUNCTION_NAME, funcName,
			    *(new SgType(T_VOID)), *(f->firstStatement()));
  }
  fce = new SgFunctionCallExp(*fs);
  return fce;
}


// build a function call statement;
SgStatement *buildFuncCallStmt(char *funcName, SgFile *f) {
  SgCExpStmt *ces;
  SgExprListExp *elist;
  SgExpression *fce;

  fce = buildFuncCallExp(funcName, f);
  elist = new SgExprListExp();
  elist->setValue(*fce);
  ces = new SgCExpStmt(*elist);
  return ces;
}


// grab the paramNum_th parameter symbol from func statement;
SgSymbol *extractParamIFromFuncStmt(SgStatement *fs, int paramNum) {
  
  int i;
  SgExpression *fe;
  
  fe = extractFuncExpFromHedr(fs);
  for (i = 0; (i < paramNum) && (fe->rhs()); i++) {
    fe = fe->rhs();
  }
  if (i < paramNum) {
    // means we ran out of parameters before we got to paramNum;
    return NULL;
  } else {
    fe = fe->lhs();
    if (fe->variant() == DEREF_OP) {
      fe = fe->lhs();
    }
    return fe->symbol();
  }
}



// search for the parameter of name pname;
SgSymbol *searchForParamInFuncStmt(SgStatement *fs, char *pname) {
  
  SgExpression *fe, *ft;
  char *symname;

  // get expression list of parameters;
  fe = extractFuncExpFromHedr(fs);
  fe = fe->lhs();

  while (fe) {
    ft = fe->lhs();
    if (ft->variant() == DEREF_OP) {
      ft = ft->lhs();
    }
    symname = ft->symbol()->identifier();
    fflush(stdout);
    if (strcmp(symname, pname) == 0) {
      return ft->symbol();
    }
    fe = fe->rhs();
  }
  return NULL;
}

// build a function definition from a symbol (works for member functions, too);
SgStatement *buildFuncDefFromSymb(SgSymbol *symb) {

  SgFuncHedrStmt *fhs;

  fhs = new SgFuncHedrStmt(*symb);
  addFuncDeclToSymb(fhs, symb);
  return fhs;
}

SgSymbol *findMemb(char *declName, char *membName, SgFile *f) {
  SgStatement *decl;

  decl = findCompDeclStmt(declName, f);
  return findVarDeclSymbFromStmt(membName, decl);
#if 0
  int i;
  SgClassSymb *clSym;
  decl = findCompDeclStmt(declName, f);
  if (clSym = isSgClassSymb(decl->symbol())) {
    i = 0;
    while (i <= clSym->numberOfFields()) {
      if (strcmp((clSym->field(i))->identifier(), membName) == 0) {      
	return clSym;
      }
      i++;
    }
  }
  return NULL;
#endif
}

// build a member function call expression.;
SgExpression *buildMembFuncCallExpOfExp(SgSymbol *cvar, 
					SgExpression *fce) {

  SgExpression *fexp;

  fexp = &(SgPointStOp(*(new SgVarRefExp(*cvar)),
		       *fce));
  return fexp;
}

// build a member function call expression.;
SgExpression *buildMembFuncCallExp(char *className, char *membName, 
				   SgFile *f, SgSymbol *cvar) {

  SgFunctionCallExp *fce;
  SgSymbol *fs;
  SgExpression *fexp;

  fs = findMemb(className, membName, f);
  
  fce = new SgFunctionCallExp(*fs);
  fexp = buildMembFuncCallExpOfExp(cvar, fce);
  return fexp;
}


// build a function call statement;
SgStatement *buildMembFuncCallStmt(char *className, 
				   char *membName, 
				   SgFile *f, 
				   SgSymbol *cvar) {
  SgCExpStmt *ces;
  SgExprListExp *elist;
  SgExpression *fce;

  fce = buildMembFuncCallExp(className, membName, f, cvar);
  elist = new SgExprListExp();
  elist->setValue(*fce);
  ces = new SgCExpStmt(*elist);
  return ces;
}


// build a member function call expression.;
SgExpression *buildThisMembCallExp(char *className, char *membName, 
				   SgFile *f) {

  SgFunctionCallExp *fce;
  SgStatement *tmps;
  SgSymbol *fs;
  SgExpression *fexp;
  SgType *ty;

  // find the correct collection symbol;
  tmps = findCompDeclStmt(className, f);
  ty = tmps->symbol()->type();

  fs = findVarDeclSymbFromStmt(membName, tmps);
  
  fce = new SgFunctionCallExp(*fs);
  fexp = &(SgPointStOp(*(new SgThisExp(*ty)),
			 *fce));
  return fexp;
}

SgStatement *buildThisMembCallStmt(char *className, char *membName, 
				   SgFile *f) {
  SgCExpStmt *ces;
  SgExprListExp *elist;
  SgExpression *fce;

  fce = buildThisMembCallExp(className, membName, f);
  elist = new SgExprListExp();
  elist->setValue(*fce);
  ces = new SgCExpStmt(*elist);
  return ces;
}


// build a derived type, for struct, classes, etc.;
SgType *buildDerivedType(char *sName, SgFile *f) {
  SgStatement *st;
  SgSymbol *stSym;
  SgDerivedType *dty;

  st = findCompDeclStmt(sName, f);
  if (!st) {
    return NULL;
  }
  stSym = st->symbol();
  dty = new SgDerivedType(*stSym);
  return dty;
}
  
/* returns the last lexical statement with given control parent cp */
SgStatement *lastStatementWithControlParent(SgStatement *cp) {
  SgStatement *st, *lst, *fst;
  lst = fst = st = cp;
  st = st->lexNext();

  while (st) {
    if (st->controlParent() == fst) {
      lst = st;
    }
    st = st->lexNext();
  }
  return lst;
}

// retrieve last statement of a dep file
SgStatement *lastGlobalStatement(SgFile *f) {
  return lastStatementWithControlParent(f->firstStatement());
}

  
// retrieve last statement of a dep file
SgStatement *lastStatement(SgFile *f) {
  
  SgStatement *st, *lst;

  lst = st = f->firstStatement();
  st = st->lexNext();

  while (st) {
    lst = st;
    st = st->lexNext();
  }
  return lst;
}


SgStatement *findFuncHedrStmt(char *fname, SgFile *f) {

  SgStatement *st;
  SgSymbol *fs;

  fs = findFunctionSymb(fname, f);

  if (fs == NULL) {
    return NULL;
  }

  st = f->firstStatement();
  while (st) {
    if (st->symbol() == fs) {
      return st;
    }
    st = st->lexNext();
  }
  return NULL;
}
  
SgExpression *buildCompareExp(SgExpression *lhs, SgExpression *rhs, 
			     int op) {
  SgExpression *compExp;

  compExp = new SgExpression(op);
  compExp->setLhs(*lhs);
  compExp->setRhs(*rhs);
  return compExp;
}
 

SgStatement *buildForStmt(int start, int end, SgSymbol *loopvar,
			  SgStatement *body) {
  SgExpression *fexp, *fexp2, *fexp3, *loopvarExp;
  SgStatement *nbody;

  loopvarExp = new SgVarRefExp(*loopvar);

  fexp = buildAssignExp(loopvarExp,
			new SgValueExp(start));

  fexp2 = buildCompareExp(loopvarExp,
			 new SgValueExp(end), LT_OP);

  fexp3 = new SgUnaryExp(PLUSPLUS_OP, *(new SgVarRefExp(*loopvar)));

  if (!body) {
    nbody = new SgStatement(CONTROL_END);
  } else {
    nbody = body;
  }
  nbody =  new SgForStmt(*fexp, 
			 *fexp2,
			 *fexp3,
			 *nbody);

  /* because of the you cant build a for statement with
   * a NULL body, you must build a fake one, then delete
   * it after creating the for statement.
   */
  if (!body) {
    body = nbody;
    nbody = nbody->lexNext();
    nbody->deleteStmt();
  } else {
    body = nbody;
  }
  return body;
}


// ----------------------------------;
char *unparseExp(SgExpression *e) {
  char *buf, buf2[20], *tbuf;
  SgValueExp *ve;

  buf = new char[100];
  *buf = '\0';
 
 switch (e->variant()) {
 case ASSGN_OP:{
#if DEBUG
    printf("assign op, ");
#endif    
   tbuf = unparseExp(e->rhs());
   strcat(buf, tbuf);
   delete [] tbuf;
   return buf;
   /*break;*/
 }
 case EXPR_LIST: {
#if DEBUG
    printf("paren? op, ");
#endif    
   strcat(buf, "(");
   tbuf = unparseExp(e->lhs());
   strcat(buf, tbuf);
   strcat(buf, ")");
   delete [] tbuf;
   return buf;
   /*break;*/
 } 
 }

  if (e->lhs()) {
#if DEBUG
    printf("  unparseing left hand expression: { \n", tbuf);
#endif
    tbuf = unparseExp(e->lhs());
    strcpy(buf, tbuf);
#if DEBUG
    printf("        }  unparsed left hand expression: %s\n", tbuf);
#endif
    delete [] tbuf;
    
  }

  switch (e->variant()) {
  case ADD_OP: {
#if DEBUG
    printf(" add op ");
#endif
    strcat(buf, "+");
    break;
  }
  case SUBT_OP: {
#if DEBUG
    printf(" sub op ");
#endif
    strcat(buf, "-");
    break;
  }
  case MULT_OP: {
#if DEBUG
    printf(" mult op ");
#endif
    strcat(buf, "*");
    break;
  }
  case DIV_OP: {
#if DEBUG
    printf(" div op ");
#endif
    strcat(buf, "/");
    break;
  }
  case MOD_OP: {
#if DEBUG
    printf(" mod op ");
#endif
    strcat(buf, "%");
    break;
  }
  case LSHIFT_OP: {
#if DEBUG
    printf(" left shift op ");
#endif
    strcat(buf, "<<");
    break;
  }
  case RSHIFT_OP: {
#if DEBUG
    printf(" right shift op ");
#endif 
    strcat(buf, ">>");
    break;
  }
  case INT_VAL: {
    ve = (SgValueExp *) e;
    sprintf(buf2, "%d", ve->intValue());
    strcat(buf, buf2);
#if DEBUG
    printf(" ival (%s) ", buf2);
#endif 
    break;
  }
  case VAR_REF: {
    
    //    # check for a const declaration;
#if DEBUG
    printf("\n\nFOUND VARIABLE REFERENCE: %s\n", e->symbol()->identifier());
#endif    
    if (e->symbol()->type()) {
      SgStatement *vds, *cp;
      cp = e->symbol()->scope();
      while (cp->variant() != GLOBAL) {
	cp = cp->controlParent();
      }
      vds = findVarDeclStmt(e->symbol()->identifier(), cp);
      if (!vds) {
	break;
      }
      // get expr_list;

#if DEBUG
      fflush(stdout);
#endif      
      tbuf =  unparseExp(vds->expr(0));
      strcat(buf, tbuf);
      delete [] tbuf;

#if DEBUG
      printf("unparsed so far: %s \n", buf);
      fflush(stdout);
#endif
    }
    break;
  }
  default: {
    printf("I cant unparse: default variant %d, lhs %d rhs %d \n", 
	   e->variant(),
	   e->lhs(), e->rhs());
    break;
  }
  }
  if (e->rhs()) {
#if DEBUG
    printf("getting rhs\n");
#endif
    strcat(buf, unparseExp(e->rhs()));
  }
  return buf;
}
#if 0
      if (e->symbol()->type()->variant() == T_DESCRIPT) {
	SgDescriptType *sdt;
	sdt = (SgDescriptType *) e->symbol()->type();
	if (sdt->modifierFlag() && BIT_CONST) {
	  // get past assign op;
	  e = e->lhs();
	  // get right hand side, the value side;
	  e = e->rhs();
	  // now, parse the operations;
	}
#endif

char *skipSpaces(char *ptr) {
  while (*ptr == ' ') {
    ptr++;
  }
  return ptr;
}


char *skipNumber(char *ptr) {
  while (isdigit(*ptr)) {
    ptr++;
  }
  return ptr;
}

char *strsave(char *s) {
  char *str;

  str = new char[strlen(s)+1];
  strcpy(str, s);
  return str;
}
  
char *copyParenExp (char *ptr) {
  char tbuf[100], *t, *p;
  int paren = 1;

  t = tbuf;
  p = ptr;
  while (paren != 0) {
    if (*p == '(') {
      paren++;
    } else if (*p == ')') {
      paren--;
      if (paren == 0) {
	*t = '\0';
	break;
      }
    } 
    *t = *p;
    t++ ;
    p++;
  }
  return strsave(tbuf);
}
  

char *skipParenExp (char *ptr) {
  int paren = 1;
  char *p;

  p = ptr;
  while (paren != 0) {
    if (*p == '(') {
      paren++;
    } else if (*p == ')') {
      paren--;
      if (paren == 0) {
	return (++p);
      }
    } else if (*p == '\0') {
      printf("malformed expression\n");
      exit(1);
      return p;
    }
    p++;
  }
  printf("malformed expression\n");
  exit(1);
  return p;
}
  
int evalStr(char *expStr) {
  
  char *ptr, *p1, *p2, *warn = "Warning: ";
  int tmp1;
  
#if DEBUG
  printf("evaluating string \"%s\"\n", expStr);
  fflush(stdout);
#endif
  ptr = skipSpaces(expStr);

  if (*ptr != '\0') {
    if (isdigit(*ptr) || (*ptr == '(') ) {
      if (*ptr == '(') {
	ptr++;
	p2 = ptr;
	p1 = copyParenExp(ptr);
	tmp1 = evalStr(p1);
	ptr = skipParenExp(p2);
      } else if (sscanf(ptr, "%d", &tmp1) != 1) {
        printf("%s non integer in expressoin string:%s\n, ", warn,
               ptr);
        return -1;
      } 

      ptr = skipNumber(ptr);
      ptr = skipSpaces(ptr);
      
      switch (*ptr) {
      case '*': {
	ptr++;
	return tmp1 * evalStr(ptr);
	/*break;*/
      }
      case '+': {
	ptr++;
	return tmp1 + evalStr(ptr);
	/*break;*/
      }
      case '-': {
	ptr++;
	return tmp1 - evalStr(ptr);
	/*break;*/
      }
      case '/': {
	ptr++;
	return tmp1 - evalStr(ptr);
	/*break;*/
      }
      case '<': {
	// test for left shift operator;
	ptr++;
	if (*ptr == '<') {
	  ptr++;
	  return tmp1 << evalStr(ptr);
	} else {
	  return (tmp1 < evalStr(ptr));
	}
	/*break;*/
      }
      case '>': {
	// test for left shift operator;
	ptr++;
	if (*ptr == '>') {
	  ptr++;
	  return tmp1 >> evalStr(ptr);
	} else {
	  return (tmp1 > evalStr(ptr));
	}
	/*break;*/
      }
	// done with string...;
      case '\0': {
	return tmp1;
	/*break;*/
      }
	// should never happen
	default: {
	  printf("%s something wrong in expression %s\n", warn, ptr);
          return -1;
	}
      }
    } else {
      // could be unary operator???;
      printf("%s something wrong in expression doesnt start with digit %s\n",
             warn, ptr);
      return -1;
    }
  } else {
    // could be unary operator???;
    printf("%s something wrong in expression (nothing in string?) %s\n",
           warn, ptr);
    return -1;
  }
}


int evalExp(SgExpression *e) {
  return evalStr(unparseExp(e));
}

int sameType(char *tname, SgType *ty) {
  SgSymbol *sym;

  if (isSgDescriptType(ty)) {
    return sameType(tname, ty->baseType());
  } else {
    if (sym = ty->symbol()) {
      return (strcmp(sym->identifier(), tname) == 0);
    } 
  }
  return 0;
}
    
SgType *findUserType(char *tname, SgFile *f) {
  SgType *tmp;

  tmp = f->firstType();
  while(tmp) {
    if (sameType(tname, tmp)) {
      return tmp;
    } else {
      tmp->next();
    }
  }
  return NULL;
}
      

SgType *dupType(SgType *ty) {
#if 0
  SgType *retType, *tmpType;
#endif
 
  return &(ty->copy());

#if 0
  switch (ty->variant()) {
    
  case T_ARRAY: {
    SgArrayType *arrType;
    tmpType = dupType(ty->baseType());
    arrType = new SgArrayType(*tmpType);
    for (int i = 0; i < ty->dimension(); i++) {
      arrType->addRange(*(ty->sizeInDim[i]));
    }
    return arrType;
    break;
  }
  case T_DERIVED_TYPE: {
    SgDerivedType *sdt, *t;
    t = (SgDerivedType *) ty;
    tmpType = dupType(ty->baseType());
    sdt = new SgDerivedType(ty->typeName());
    sdt->setBaseType(*tmpType);
    return sdt;
    break;
  }
  case T_DESCRIPT: {
    SgDescriptType *dt;
    tmpType = dupType(ty->baseType());
    sdt = new SgDescriptType(tmpType);
    return dt;
    break;
  }
  case T_POINTER: {
    SgPointerType *pt;
    tmpType = dupType(ty->baseType());
    sdt = new SgPointerType(tmpType);
    return pt;
    break;
  }
  }
#endif
}
