/* --------------------------------------------------
 *
 *  Darryl Brown
 *  
 *  October, 1994
 *
 *  typeApi.c
 *  
 *  This file is the source file for the implementation of 
 *  the API for data/data structure transfer above the 
 *  transportation layer (cargo.c and transport.c).
 *
 * --------------------------------------------------
 */

#include "astrlib.h"
#include "typeApi.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

/* 
 * --------------------------------------------------------------
 *  MACROS
 * --------------------------------------------------------------
 */

#ifndef NULL
#define NULL 0
#endif
#define DEBUG 0

/* 
 * --------------------------------------------------------------
 *  GLOBAL DECLARATIONS
 * --------------------------------------------------------------
 */

/*  array of MAXCOMPS TA_composite pointers */
TA_composite *TA_CompositeList[MAXCOMPS];

/*  array of MAXTYPES type pointers */
TA_typeDesc *TA_typeDescList[MAXTYPES];

/*  array of MAXINST of var name & type pairs */
TA_nameTypePair *TA_VarTypeList[MAXINST];

/*  array of MAXINST of func name & type pairs */
TA_nameTypePair *TA_FuncTypeList[MAXINST];

/* table of function names and function pointers */
TA_AccessFuncEntry *TA_AccessFuncTable[MAXFUNCS];

/* table of function names and function pointers */
TA_UserFuncEntry *TA_UserFuncTable[MAXFUNCS];

/* indices for the above lists */
int TA_CompListIdx, TA_TypeListIdx, TA_VarTypeIdx,
  TA_FuncTypeIdx, TA_AccessFuncTableIdx,
  TA_UserFuncTableIdx;

/* 
 * --------------------------------------------------------------
 *  ACCESS FUNCTIONS
 * --------------------------------------------------------------
 */


/*
 *  TA_typeEnum TA_strToType(char *s) - changes string to
 *       corresponding TA_typeEnum representation. 
 *  
 *  char * TA_typeToStr (int ty) - changes TA_typeEnum into
 *        string representation  of that type.
 *  
 *  char * TA_typeToFormat(TA_typeEnum ty) - changes TA_typeEnum
 *        representation into a string format used in
 *        printf statements (to print that type).
 *
 *  char *TA_getType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 *  void TA_putType(char *buf) - adds a type string.
 *
 *  char *TA_callFunc(char *name, int argc, char **argv) - 
 *       returns the string representation of the type
 *       that corresponds to name.
 *
 *  int TA_putFunc(char *buf, void *funcPtr) - adds a 
 *       function name and pointer to that func.
 *
 *  int TA_putVarType(char *name, char *type) - stores
 *       the name and type pair for later retrieval.
 *
 *  int TA_putFuncType(char *name, char *type) - stores
 *       the func name and type pair for later retrieval.
 *
 *  char *TA_getVarType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 *  char *TA_getFuncType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 *  char *TA_writeComposite(TA_composite *comp) - returns string 
 *       representation of TA_composite comp.
 *
 *  char *TA_writeTypeDesc(TA_typeDesc *td) - returns string
 *       representation of type description td.
 *
 *  char *TA_writeField(TA_field *fd) - returns string representation
 *       of field fd.
 *
 *  TA_typeDesc *TA_writeTypeDesc(char *buf) - returns TA_typeDesc
 *       representation of type description in buf.
 *
 *  TA_field *TA_writeField(char *buf) - returns field representation
 *       of field in buf.
 *
 *  TA_nameList *TA_writeNameList(char *buf) - returns TA_nameList
 *       representation of TA_nameList in buf.
 * 
 *  TA_typeDesc *TA_findtypeDesc(char *typeName) returns a pointer to 
 *       TA_composite structure of type compName (or NULL if none
 *       exists).
 *
 *  char *TA_findType(char *typeName) returns a pointer to 
 *       either TA_composite structure or TA_typeDesc structure 
 *       of type typeName or (or NULL if none exists).
 *
 *  void TA_addDim(TA_dimType *dt, int dim) - adds a dimension
 *       to TA_dimType dt.
 *
 *  TA_dimType *TA_createDimType() - creates the TA_dimType
 *       representation of dimensions dt, allocating memory.
 *
 *  TA_typeDesc *TA_createTypeDesc() - creates a TA_typeDesc
 *       structure, allocating memory.
 *
 *  int TA_isComposite(TA_typeDesc te) 
 *       returns true if te is TA_composite type.
 *
 *  TA_field *TA_createField() - creates a field
 *       structure, allocating memory.
 *
 *  TA_composite TA_createComposite() - creates a TA_composite
 *       structure, allocating memory.
 *
 *  void TA_putComposite(TA_composite *comp) - adds comp to 
 *       global TA_composite list.
 *
 *  comp *TA_findComposite(char *compName) returns a pointer to 
 *       TA_composite structure of type compName (or NULL if none
 *       exists).
 *
 */

/* 
 * --------------------------------------------------------------
 *  INTERNAL FUNCTIONS
 * --------------------------------------------------------------
 */

/*
 *
 *  TA_typeEnum TA_readTypeEnum(char *buf) - returns a enumerated
 *       type representation of the type in buf.
 *
 *  TA_dimType *TA_readDims(char *buf) - returns a TA_dimType
 *       representation of dimensions if they are in buf.
 *
 *  TA_descriptEnum TA_readDescript(char *buf) -
 *       returns enumerated representation if there
 *       is a descriptor in buf.
 *
 *  char *TA_writeType(TA_typeEnum *te) - returns string
 *       representation of type enum te.
 *
 *  char *TA_writeUserDefFuncs() - returns string
 *       listing of all userDefFuncs, seperated by blanks
 *
 *  char *TA_writeUserDefMemFuncs(TA_typeEnum *te) - returns string
 *       representation of all user defined member functions of
 *       element classes, each line separated by a newline.
 *
 *  char *TA_writeDims(TA_dimType *dt) - writes the string
 *       representation of dimensions dt.
 *
 *  char *TA_writeDescriptor(TA_descriptEnum de) - returns string 
 *       representation of de.
 *
 *  void TA_putTypeDesc(TA_typeDesc *ty) - adds ty to global 
 *       TA_typeDesc list.
 *
 *  TA_nameList *TA_createNameList(char *name) - creates a TA_nameList
 *       structure, allocating memory.
 *
 *  TA_nameEntry *TA_createNameEntry(char *name) - creates a TA_nameList
 *       structure, allocating memory.
 *
 *  varTypePair *TA_createVarTypePair() - creates a 
 *       varTypePair structure, allocating memory.
 *
 *  TA_funcEntry *TA_createFuncEntry() - creates a 
 *       TA_funcEntry structure, allocating memory.
 * 
 *  TA_funcEntry *TA_findFuncEntry(char *name) returns a pointer 
 *       to TA_funcEntry that corresponds to name from the
 *       TA_AccessFuncTable.
 *
 */


/* 
 * --------------------------------------------------------------
 *  INTERNAL FUNCTION IMPLEMENTATION
 * --------------------------------------------------------------
 */

/*
 *  int TA_isComposite(TA_typeDesc te) 
 *       returns true if te is TA_composite type.
 */
int TA_isComposite(TA_typeEnum te)  {
  return ((te == t_class) ||
	  (te == t_struct) ||
	  (te == t_collection)) ;
}



/*
 *  TA_descriptEnum TA_readDescript(char *buf) -
 *       returns enumerated representation if there
 *       is a descriptor in buf.
 */
TA_descriptEnum TA_readDescript(char *buf) {
  switch (*buf) {
  case 'v':
    return d_volatile;
  case 'r':
    return d_register;
  default:
    return d_nodescriptor;
  }
}

/*
 *
 *  char *TA_writeDescriptor(TA_descriptEnum de) - returns string 
 *       representation of de.
 *
 */
char *TA_writeDescriptor(TA_descriptEnum de) {
  switch (de) {
  case d_volatile: {
    return "volatile";
  }
  case d_register: {
    return "register";
  }
  case d_nodescriptor: {
    return "";
  }
  }
  return "";
}
    

/*
 *  TA_typeEnum TA_readTypeEnum(char *buf) - returns a enumerated
 *       type representation of the type in buf.
 *
 */
TA_typeEnum TA_readTypeEnum(char *buf) {
  switch (*buf) {
  case 'i' :
    return t_int;
  case 'f' :
    return t_float;
  case 'd' :
    return t_double;
  case 'c': {
    /* check second letter */
    switch (*(buf+1)) {
    case 'h':
      /* must be char */
      return t_char;
    case 'o':
      /* must be collection */
      return t_collection;
    case 'l':
      /* must be class */
      return t_class;
    }
  }
  case 'e':
    return t_enum;
  case 's':
    return t_struct;
  default:
    return t_int;
  }
}

/*
 *  char *TA_writeType(TA_typeEnum te) - returns string
 *       representation of type enum te.
 *
 */
char *TA_writeType(TA_typeEnum te)  {
  switch (te) {
  case t_int :
    return "int";
  case t_float :
    return "float";
  case t_double:
    return "double";
  case t_char :
    return "char";
  case t_enum:
    return "enum";
  case t_class :
    return "class";
  case t_collection :
    return "collection";
  case t_struct :
    return "struct";
  }
  return "";
}


/*
 *  comp *TA_findComposite(char *compName) returns a pointer to 
 *       TA_composite structure of type compName (or NULL if none
 *       exists).
 *
 */
TA_composite *TA_findComposite(char *compName) {
  TA_composite *tmpc;
  int i;

  i = 0;

  tmpc = TA_CompositeList[0];
  while (tmpc != NULL) {
    if (strsame(tmpc->name, compName))  
      return tmpc;
    i++;
    tmpc = TA_CompositeList[i];
  }
  return NULL;
}
    

/*
 *  TA_typeDesc *TA_findTypeDesc(char *typeName) returns a pointer to 
 *       TA_composite structure of type compName (or NULL if none
 *       exists).
 *
 */
TA_typeDesc *TA_findTypeDesc(char *typeName) {
  TA_typeDesc *tmpt;
  int i;

  i = 0;
  tmpt = TA_typeDescList[0];
  while (tmpt != NULL) {
#if DEBUG
    printf("looking for TA_typeDesc %s l at %s\n", 
	   typeName, tmpt->name);
#endif    
    
    if (strsame(tmpt->name, typeName))  {
#if DEBUG
      printf("found TA_typeDesc %s \n");
#endif
      return tmpt;
    }
    i++;
    tmpt = TA_typeDescList[i];
  }
  return NULL;
}
    

/*       
 *  int TA_putComposite(TA_composite *comp) - adds comp to global 
 *       TA_composite list.
 *
 */
int TA_putComposite(TA_composite *comp) {
  TA_CompositeList[TA_CompListIdx] = comp;
  TA_CompListIdx++;
  return   TA_CompListIdx-1;
}


/*
 *  void TA_putTypeDesc(TA_typeDesc *ty) - adds ty to global TA_typeDesc list.
 *
 */
int TA_putTypeDesc(TA_typeDesc *ty) {
  TA_typeDescList[TA_TypeListIdx] = ty;
  TA_TypeListIdx++;
  return   TA_TypeListIdx-1;
}


/*
 *
 *  TA_dimType *TA_createDimType() - creates the TA_dimType
 *       representation of dimensions dt, allocating memory.
 *
 */
TA_dimType *TA_createDimType() {
  TA_dimType *dt;

  if ((dt = (TA_dimType *)malloc(sizeof(TA_dimType)))
      == NULL) {
    perror("malloc failed");
  }
  dt->numDims = 0;
  dt->sizes = NULL;

  return dt;
}

/*
 *
 *  TA_field *TA_createField() - creates a field
 *       structure, allocating memory.
 *
 */
TA_field *TA_createField() {
  TA_field *f;

  if ((f = (TA_field *)malloc(sizeof(TA_field)))
      == NULL) {
    perror("malloc failed");
  }
  f->next = NULL;
  f->name = NULL;
  f->type = NULL;

  return f;
}


/*
 *
 *  TA_typeDesc *TA_createTypeDesc() - creates a TA_typeDesc
 *       structure, allocating memory.
 *
 */
TA_typeDesc *TA_createTypeDesc() {
  TA_typeDesc *t;

  if ((t = (TA_typeDesc *) malloc(sizeof(TA_typeDesc)))
      == NULL) {
    printf("malloc failed");
    exit(1);
  }
  t->desc = d_nodescriptor;
  t->thisType = t_notype;
  t->dims = NULL;
  t->isPointer = 0;
  t->name = NULL;
  return t;
}


/*
 *
 *  TA_nameEntry *TA_createNameEntry(char *name) - creates a TA_nameList
 *       structure, allocating memory.
 *
 */
TA_nameEntry *TA_createNameEntry(char *name) {
  TA_nameEntry *ne;

  if ((ne = (TA_nameEntry *) malloc(sizeof(TA_nameEntry)))
      == NULL) {
    printf("malloc failed");
    exit(1);
  }
  ne->name = strsave(name);
  ne->next = NULL;
  return ne;
}
 
/*
 *
 *  TA_nameList *TA_createNameList() - creates a TA_nameList
 *       structure, allocating memory.
 *
 */
TA_nameList *TA_createNameList() {
  TA_nameList *nl;

  if ((nl = (TA_nameList *) malloc(sizeof(TA_nameList)))
      == NULL) {
    printf("malloc failed");
    exit(1);
  }
  nl->first = nl->last = NULL;
  nl->length = 0;
  return nl;
}


/*
 *
 *  TA_composite TA_createComposite() - creates a TA_composite
 *       structure, allocating memory.
 *
 */
TA_composite *TA_createComposite() {
  TA_composite *c;

  if ((c = (TA_composite *)malloc(sizeof(TA_composite)))
      == NULL) {
    perror("malloc failed");
  }
  c->parents = NULL;
  c->name = NULL;
  c->type = t_notype;
  c->fields = NULL;

  return c;
}


/*
 *
 *  TA_nameTypePair *TA_createNameTypePair() - creates a 
 *       varTypePair structure, allocating memory.
 *
 */
TA_nameTypePair *TA_createNameTypePair() {
  TA_nameTypePair *c;

  if ((c = (TA_nameTypePair *)malloc(sizeof(TA_nameTypePair)))
      == NULL) {
    perror("malloc failed");
  }
  c->name = NULL;
  c->thisType = t_notype;
  c->i = -1;

  return c;
}


/*
 *
 *  TA_funcEntry *TA_createFuncEntry() - creates a 
 *       TA_funcEntry structure, allocating memory.
 *
 */
TA_UserFuncEntry *TA_createUserFuncEntry() {
  TA_UserFuncEntry *fe;

  if ((fe = (TA_UserFuncEntry *)malloc(sizeof(TA_UserFuncEntry)))
      == NULL) {
    perror("malloc failed");
  }
  fe->typeName = NULL;
  fe->funcPtr = NULL;
  fe->name = NULL;

  return fe;
}

/*
 *
 *  TA_funcEntry *TA_createFuncEntry() - creates a 
 *       TA_funcEntry structure, allocating memory.
 *
 */
TA_AccessFuncEntry *TA_createAccessFuncEntry() {
  TA_AccessFuncEntry *fe;

  if ((fe = (TA_AccessFuncEntry *)malloc(sizeof(TA_AccessFuncEntry)))
      == NULL) {
    perror("malloc failed");
  }
  fe->typeName = NULL;
  fe->funcPtr = NULL;
  fe->name = NULL;

  return fe;

}

/*
 *
 *  void TA_addDim(TA_dimType *dt, int dim) - adds a dimension
 *       to TA_dimType dt.
 *
 */
void TA_addDim(TA_dimType *dt, int dim) {
  int *s;

  if ((s = (int *) malloc(sizeof(int)*(dt->numDims+1)))
      == NULL) {
    perror("malloc failed");
  }
  if (dt->numDims > 0) {
    memcpy(s, dt->sizes, (dt->numDims*sizeof(int)));
  }
  s[dt->numDims] = dim;
  dt->numDims += 1;
  free(dt->sizes);
  dt->sizes = s;
}


/*
 *  TA_dimType *TA_readDims(char *buf) - returns a TA_dimType
 *       representation of dimensions if they are in buf.
 *
 */
TA_dimType *TA_readDims(char *buf) {
  TA_dimType *dt;
  int d, numdim = 0;
  char *p;
  strtokStruct *st;

  st = initstrtok(buf);
  dt = TA_createDimType();
  p = sstrtok(st, "[]");
  while (p) {
    if (sscanf(p, "%d", &d) == 1) {
      numdim++;
      TA_addDim(dt, d);
    }
    p = sstrtok(st, "[]");
  }
  return dt;
}

/*
 *
 *  char *TA_writeDims(TA_dimType *dt) - writes the string
 *       representation of dimensions dt.
 *
 */
char *TA_writeDims(TA_dimType *dt) {
  int i;
  char buf[100], tmp[50], *t;

  strcpy(buf, "");
  for (i = 0; i < dt->numDims; i++) {
    sprintf(tmp, "[%d]", dt->sizes[i]);
    strcat(buf, tmp);
  }
  t = strsave(buf);
  return t;
}

 

/*
 *  TA_nameList *TA_readnameList(char *buf) - returns TA_nameList
 *       representation of TA_nameList in buf.
 * 
 */
TA_nameList *TA_readnameList(char *buf) {
  char *t;
  strtokStruct *st;
  TA_nameList *nl;
  TA_nameEntry *ne;

  nl = TA_createNameList();
  st = initstrtok(buf);
  while (t = sstrtok(st, " ")) {
    ne = TA_createNameEntry(t);
    if (nl->last) {
      nl->last->next = ne;
    } else {
      /* must be the first one, */
      nl->last = nl->first = ne;
    }
    nl->last = ne;
  }
  return nl;
}



/*
 *  char *TA_writeNameList(TA_nameList *nl) - returns string
 *       representation of TA_nameList nl.
 * 
 */
char *TA_writeNameList(TA_nameList *nl) {
  char buf[5000];
  TA_nameEntry *ne;

  if (nl == NULL) {
    return NULL;
  }
  ne = nl->first;

  strcpy(buf, ne->name);
  ne = ne->next;
  while (ne) {
    /* add a space */
    strcat(buf, " ");
    strcat(buf, ne->name);
    ne = ne->next;
  }
  return strsave(buf);
}
    


/*
 *  TA_field *TA_readField(char *buf) - returns TA_field representation
 *       of field in buf.
 *
 */
TA_field *TA_readField(char *buf) {
  TA_field *f;
  char nbuf[50], tbuf[200];

  f = TA_createField();
 
  /* scan name and TA_typeDescription */
  sscanf(buf, " %[^ ] %[^)] ", nbuf, tbuf);
  f->name = strsave(nbuf);
  f->type = TA_readTypeDesc(tbuf);

  return f;
}

/*
 *  TA_field *TA_readFields(char *buf) - returns all fields
 *       in buf.
 *
 */
TA_field *TA_readFields(char *buf) {
  TA_field *f, *prev, *first;
  char *t;
  strtokStruct *st;

  first = NULL;
  st = initstrtok(buf);
  prev = NULL;
  /* parse a TA_field */
  astrtok(st, "(");
  while (t = sstrtok(st, ")")) {
    f = TA_readField(t);
    if (prev) {
      /* not the first */
      prev->next = f;
    } else {
      /* must be the first */
      first = f;
    }
    prev = f;
    astrtok(st, "(");
  }

  return first;
}

/*
 *  char *TA_writeField(TA_field *fd) - returns string representation
 *       of TA_field fd.
 *
 */
char *TA_writeField(TA_field *fd) {
  char *tmp, buf[50];

  strcpy(buf, fd->name);
  strcat(buf, " ");
  tmp = TA_writeTypeDesc(fd->type);
  strcat(buf, tmp);
  free(tmp);
  
  return strsave(buf);
}
  
/* 
 *  TA_UserFuncEntry *TA_findUserFuncEntry(char *name) returns a pointer to 
 *       TA_funcEntry that corresponds to name from the
 *       functable paramemeter.
 *
 */
TA_UserFuncEntry *TA_findUserFuncEntry(char *name) {
  TA_UserFuncEntry *fe;
  int i;

  for (i = 0; i < TA_UserFuncTableIdx; i++) {
    fe = TA_UserFuncTable[i];
    if (strsame(fe->name, name))  
      return fe;
  }
  
  return NULL;
}


/* 
 * TA_AccessFuncEntry *TA_findAccessFuncEntry(char *name) returns a pointer to 
 *       TA_funcEntry that corresponds to name from the
 *       functable paramemeter.
 *
 */
TA_AccessFuncEntry *TA_findAccessFuncEntry(char *name) {
  TA_AccessFuncEntry *fe;
  int i;


  for (i = 0; i < TA_AccessFuncTableIdx; i++) {
    fe = TA_AccessFuncTable[i];
    if (strsame(fe->name, name))  
      return fe;
  }
  return NULL;
}


/* 
 * --------------------------------------------------------------
 *  ACCESS FUNCTION IMPLEMENATION
 * --------------------------------------------------------------
 */





/*
 *  TA_typeEnum TA_strToType(char *s) - changes string to
 *       corresponding TA_typeEnum representation. 
 *  
 */
TA_typeEnum TA_strToType(char *s) {
  
  if (strsame(s, "class")) {
    return t_class;
  }  else if (strsame(s, "Collection")) {
    return t_collection;
  } else if (strsame(s, "struct")) {
    return t_struct;
  } else if (strsame(s, "derived type")) {
    return t_derived_type;
  } else if (strsame(s, "union")) {
    return t_union;
  } else if (strsame(s, "int")) {
    return t_int;
  } else if (strsame(s, "long")) {
    return t_int;
  } else if (strsame(s, "short")) {
    return t_int;
  } else if (strsame(s, "char")) {
    return t_char;
  } else if (strsame(s, "float")) {
    return t_float;
  } else if (strsame(s, "double")) {
    return t_double;
  } else if (strsame(s, "void")) {
    return t_void;
  } else if (strsame(s, "unknown")) {
    return t_notype;
  } else if (strsame(s, "function")) {
    return t_function;
  } else {
    return t_notype;
  }
}


/*
 *  char * TA_typeToStr (int ty) - changes TA_typeEnum into
 *        string representation  of that type.
 *  
 */

char * TA_typeToStr (int ty) {
  
  switch (ty) {
  case t_enum: {
    return "enum";
    /*break;*/
  }
  case t_class: {
    return "class";
    /*break;*/
  }
  case t_collection: {
    return "Collection";
    /*break;*/
  }
  case t_struct: {
    return "struct";
    /*break;*/
  }
  case t_int: {
    return "int";
    /*break;*/
  }
  case t_float: {
    return "float";
    /*break;*/
  }
  case t_double: {
    return "double";
    /*break;*/
  }
  case t_char: {
    return "char";
    /*break;*/
  }
  case t_void: {
    return "void";
    /*break;*/
  }
  case t_union: {
    return "union";
    /*break;*/
  }
  case t_function: {
    return "function";
    /*break;*/
  }
  case t_derived_type: {
    return "derived type";
    /*break;*/
  }
  case t_notype: {
    return "notype";
    /*break;*/
  }
  }
  return NULL;
}

/*
 *  char * TA_typeToFormat(TA_typeEnum ty) - changes TA_typeEnum
 *        representation into a string format used in
 *        printf statements (to print that type).
 *
 */

char * TA_typeToFormat(TA_typeEnum ty) {

  switch (ty) {
  case t_class: {
    return "";
    /*break;*/
  }
  case t_collection: {
    return "follow";
    /*break;*/
  }
  case t_enum:
  case t_int: 
  case t_function: {
    return "%10d";
    /*break;*/
  }
  case t_struct: {
    return "follow";
    /*break;*/
  }
  case t_double:
  case t_float: {
    return "%9.3g";
    /*break;*/
  }
  case t_char: {
    return "%c";
    /*break;*/
  }
  case t_union:
  case t_void: {
    return "";
    /*break;*/
  }
  case t_derived_type: {
    printf("derived type to format?\n");
    return "";
    /*break;*/
  }
  case t_notype: {
    printf("unknown type type to format?\n");
    return "";
    /*break;*/
  }
  default: {
    return "";
    /*break;*/
  }
  }
}



/*
 * char * TA_typeDescToFormat(TA_typeDesc *tyde) builds the
 *        TA_typeDesc representation into a string format 
 *        used in printf statements (to print that type).
 *
 */

char * TA_typeDescToFormat(TA_typeDesc *tyde) {
  TA_typeDesc *tytmp;

  if (tyde->thisType == t_derived_type) {
    tytmp = TA_findTypeDesc(tyde->name);
    if (tytmp == NULL) {
      printf("problem with type desc(s): %s\n", TA_writeTypeDesc(tyde));
      return "";
    }
    return TA_typeDescToFormat(tytmp);
  } else if (tyde->isPointer) {
    return "%d";
  } else {
    return TA_typeToFormat(tyde->thisType);
  }
}

/*
 *  TA_nameEntry *TA_pushName(char *name, TA_nameList *nl) - stores
 *       the name in a list.
 *
 */
TA_nameEntry *TA_pushName(char *name, TA_nameList *nl) {
  TA_nameEntry *ne;

  if (nl == NULL) {
    return NULL;
  }

  ne = TA_createNameEntry(name);
  if (nl->last != NULL) {
    /* it is not the first name... */
    nl->last->next = ne;
  } else {
    /* it IS the first name... */
    nl->first = ne;
  }
  nl->last = ne;
  nl->length++; 
  ne->next = NULL;
  return ne;
}



/*
 *  char *popName(TA_nameList *nl) - pops 
 *       the top name in the list.
 *
 */
char *popName(TA_nameList *nl) {
  TA_nameEntry *ne, *te;
  char *t;

  if (nl == NULL) {
    return NULL;
  }
  ne = nl->last;
  te = nl->first;

  if (ne == te) {
    /* it the only entry */
    nl->first = nl->last = NULL;
  } else {
    /* find the next to last entry,
        and make it the last */
    while (te->next != ne ) {
      te = te->next;
    }
    nl->last = te;
    te->next = NULL;
  }
  
  nl->length--; 

  t = ne->name;
#if 1
  t = strsave(ne->name);
  free(ne->name);
#endif
  free(ne);
  return t;
}


/*
 *  TA_typeDesc *TA_readTypeDesc(char *buf) - returns TA_typeDesc
 *       representation of type description in buf.
 *
 */
TA_typeDesc *TA_readTypeDesc(char *buf) {
  char b[50], *t;
  strtokStruct *st;
  TA_typeDesc *td;

  td = TA_createTypeDesc();
  st = initstrtok(buf);

  t = sstrtok(st, " ");
  td->desc = TA_readDescript(t);

  if (td->desc != d_nodescriptor) {
    t = sstrtok(st, " ");
  } 
  /* initialize the type */
  td->thisType = TA_readTypeEnum(t);
  
  if (td->thisType == t_notype) {
    td->name = strsave(t);
  }

  /* see whats next */
  t = sstrtok(st, " ");
  if (t == NULL) {
    td->dims = TA_createDimType();
    td->isPointer = 0;
    return td;
  }
    
  if (*t == '*') {
    /* it is a pointer */
    td->isPointer = 1;
    /* so grab whats next */
    t = sstrtok(st, " ");
  } else {
    /* it is not a pointer */
    td->isPointer = 0;
  }
  
  /* now, 't' has name and dimensions (if they exist) */
  td->dims = TA_readDims(t);

  if (t == NULL) {
    td->name = NULL;
  } else {
    if (sscanf(t, " %[^ [\\!#$%] ", b) == 1)  {
      td->name = strsave(b);
    } else {
      td->name = NULL;
    }
  }

  return td;
}

/*  
 *  char *TA_writeTypeDesc(TA_typeDesc *td) - returns string
 *       representation of type description td.
 *
 */
char *TA_writeTypeDesc(TA_typeDesc *td) {
  char *tmp, buf[1000];
  int i;

  if (td->desc != d_nodescriptor) {
    strcpy(buf, TA_writeDescriptor(td->desc));
    strcat(buf, " ");
  } else {
    strcpy(buf, "");
  }

  if (((td->thisType == t_derived_type) ||
       (td->thisType == t_class))
       && td->name) {
    strcat(buf, "class ");
    strcat(buf, td->name);
  } else {
    strcat(buf, TA_writeType(td->thisType));
  }

  strcat(buf, " ");
  for (i = 0; i < td->isPointer; i++ ) {
    strcat(buf, "*");
  }
  if (td->isPointer) strcat(buf, " ");
  
  if (td->dims) {
    /* it is an array */
    if (td->dims->numDims < 1) {
      /* no meaning.. */
    } else {
      /* add dimensions */
      tmp = TA_writeDims(td->dims);
      strcat(buf, tmp);
      free(tmp);
    }
  }

  return strsave(buf);
}

/* 
 *  comp *TA_readComposite(char *buf) - returns TA_composite 
 *       representation of TA_composite in buf.
 *
 */
TA_composite *TA_readComposite(char *buf) {
  TA_composite *comp;
  char *t;
  strtokStruct *st;

  comp = TA_createComposite();
  st = initstrtok(buf);
  t = sstrtok(st, " ");
  comp->type = TA_readTypeEnum(t);
  comp->name = strsave(sstrtok(st, " "));

  if (comp->type == t_class ||
      comp->type == t_collection) {
    t = astrtok(st, "(");
    t = astrtok(st, ")");
    comp->parents = TA_readnameList(t);
  } 
  astrtok(st, "{");
  t = sstrtok(st, "}");

  comp->fields = TA_readFields(t);
  freestrtok(st);

  return comp;
}
  
  
/*
 *
 *  char *TA_writeComposite(TA_composite *comp) - returns string 
 *       representation of TA_composite comp.
 *
 */
char *TA_writeComposite(TA_composite *comp) {
  char buf[5000], *tmp;
  TA_field *f;

  strcpy(buf, TA_writeType(comp->type));
  strcat(buf, " ");
  strcat(buf, comp->name);
  strcat(buf, " ");
  if (comp->type != t_struct) {
    /* must be class or collection */
    strcat(buf, "( ");
    tmp = TA_writeNameList(comp->parents);
    if (tmp) {
      strcat(buf, tmp);
      free(tmp);
    }
    strcat(buf, " ) ");
  }
  strcat(buf, "{ ");
  f = comp->fields;
  while (f) {
    tmp = TA_writeField(f);
    strcat(buf, " ( ");
    strcat(buf, tmp);
    strcat(buf, " ) ");
    free(tmp);
    f = f->next;
  }
  strcat(buf, " }");
  tmp = strsave(buf);

  return tmp;
}
 

/*
 *  void TA_putType(char *buf) - adds a type string.
 *
 */
int TA_putType(char *buf) {
  strtokStruct *st;
  char *t;
  TA_typeEnum te;

  st = initstrtok(buf);
  t = sstrtok(st, " ");
  te = TA_readTypeEnum(t);
  if (TA_isComposite(te)) {
    return TA_putComposite(TA_readComposite(buf));
  } else {
    return TA_putTypeDesc(TA_readTypeDesc(buf));
  }
}

/*
 *  char *TA_getType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 */
char *TA_getType(char *typeName) {
  TA_typeDesc *tmpt;
  TA_composite *comp;
  int i;

  /* must check user defined TA_composites first */
  i = 0;
  comp = TA_CompositeList[0];
  while (comp != NULL) {
    if (strsame(comp->name, typeName))  
      return TA_writeComposite(comp);
    i++;
    comp = TA_CompositeList[i];
  }

  i = 0;
  tmpt = TA_typeDescList[0];
  while (tmpt != NULL) {
    if ((tmpt->name) && (strsame(tmpt->name, typeName)))
      return TA_writeTypeDesc(tmpt);
    i++;
    tmpt = TA_typeDescList[i];
  }
  return "";
}



/*
 *  char *TA_getVarType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 */
char *TA_getVarType(char *name) {
  TA_nameTypePair *nt;
  int i;

  
  i = 0;
  nt = TA_VarTypeList[0];
  while (nt != NULL) {
    if (strsame(nt->name, name)) {
      if (TA_isComposite(nt->thisType)) {
	return TA_writeComposite(TA_CompositeList[nt->i]);
      } else {
	return TA_writeTypeDesc(TA_typeDescList[nt->i]);
      }
    }
    i++;
    nt = TA_VarTypeList[i];
  }
  return "";
}  
  

/*
 *  int TA_putVarType(char *name, char *type) - stores
 *       the name and type pair for later retrieval.
 *
 */
int TA_putVarType(char *name, char *type) {
  strtokStruct *st;
  char *t;
  TA_typeEnum te;
  TA_nameTypePair *nt;

  nt = TA_createNameTypePair();

  /* save name */
  nt->name = strsave(name);

  /* save type and index into storage array */
  st = initstrtok(type);
  t = sstrtok(st, " ");
  te = TA_readTypeEnum(t);
  nt->thisType = te;
  if (TA_isComposite(te)) {
    nt->i = TA_putComposite(TA_readComposite(type));
  } else {
    nt->i = TA_putTypeDesc(TA_readTypeDesc(type));
  }
  TA_VarTypeList[TA_VarTypeIdx] = nt;
  TA_VarTypeIdx++;
  return TA_VarTypeIdx-1;
}  


/*
 *  char *TA_getFuncType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 */
char *TA_getFuncType(char *name) {
  TA_nameTypePair *nt;
  int i;

  
  i = 0;
  nt = TA_FuncTypeList[0];
  while (nt != NULL) {
    if (strsame(nt->name, name)) {
      if (TA_isComposite(nt->thisType)) {
	return TA_writeComposite(TA_CompositeList[nt->i]);
      } else {
	return TA_writeTypeDesc(TA_typeDescList[nt->i]);
      }
    }
    i++;
    nt = TA_FuncTypeList[i];
  }
  return "";
}  
  
/*
 *  int TA_putFuncType(char *name, char *type) - stores
 *       the func name and type pair for later retrieval.
 *
 */
int TA_putFuncType(char *name, char *type) {
  strtokStruct *st;
  char *t;
  TA_typeEnum te;
  TA_nameTypePair *nt;

  nt = TA_createNameTypePair();

  /* save name */
  nt->name = strsave(name);

  /* save type and index into storage array */
  st = initstrtok(type);
  t = sstrtok(st, " ");
  te = TA_readTypeEnum(t);
  nt->thisType = te;
  if (TA_isComposite(te)) {
    nt->i = TA_putComposite(TA_readComposite(type));
  } else {
    nt->i = TA_putTypeDesc(TA_readTypeDesc(type));
  }
  TA_FuncTypeList[TA_FuncTypeIdx] = nt;
  TA_FuncTypeIdx++;
  return TA_FuncTypeIdx-1;
}  

  
/*
 *  int TA_putAccessFunc(char *name, void (*funcPtr)(void *,
 *                       int, char *)) - stores
 *       the func name and type pair for later retrieval.
 *
 */
int TA_putAccessFunc(char *name, void (*funcPtr)(void *,
						 int, char *)) {
  TA_AccessFuncEntry *fe;

  fe = TA_createAccessFuncEntry();

  /* save name */
  TA_AccessFuncTable[TA_AccessFuncTableIdx] = fe;
  fe->name = strsave(name);
  fe->funcPtr = funcPtr;
  
  TA_AccessFuncTableIdx++;
  return TA_AccessFuncTableIdx-1;
}  


/*
 *  int TA_putUserFunc(char *name, void *funcPtr) - stores
 *       the func name and type pair for later retrieval.
 *
 */
int TA_putUserFunc(char *name, void (*funcPtr)()) {
  TA_UserFuncEntry *fe;

  fe = TA_createUserFuncEntry();

  /* save name */
  TA_UserFuncTable[TA_UserFuncTableIdx] = fe;
  fe->name = strsave(name);
  fe->funcPtr = funcPtr;
  
  TA_UserFuncTableIdx++;
  return TA_UserFuncTableIdx-1;
}  

/*
 *  char *TA_writeUserDefFuncs() - returns string
 *       listing of all userDefFuncs, seperated by blanks
 *
 */
char *TA_writeUserDefFuncs()  {
  TA_UserFuncEntry *fe;
  int i;
  char buf[500];
  char newline[3];

  sprintf(newline, "\n");
  strcpy(buf, "");
  i = 0;
  fe = TA_UserFuncTable[0];
  while (fe != NULL) {
    strcat(buf, fe->name);
    strcat(buf, newline);
    i++;
    fe = TA_UserFuncTable[i];
  }
  
  return strdup(buf);
}

/*
 *  char *TA_writeUserDefMemFuncs() - returns string
 *       representation of all user defined member functions of
 *       element classes, each line separated by a newline.
 *
 */
char *TA_writeUserDefMemFuncs() {
  TA_AccessFuncEntry *fe;
  int i;
  char buf[1000];
  char newline[3];

  sprintf(newline, "\n");
  strcpy(buf, "");
  i = 0;
  fe = TA_AccessFuncTable[0];
  while (fe != NULL) {
    if (strstr(fe->name, "::")) {
      /* then the function must be a user defined method 
       * because I use 2 semicolons in the user defined function
       *  string representation 
       */
      strcat(buf, fe->name);
      strcat(buf, newline);
    }
    i++;
    fe = TA_AccessFuncTable[i];
  }

  return strdup(buf);
}

/*
 *  void TA_callAccessFunc(char *name, void *collPtr, int elemNum,
 *                         char *fname) 
 *       returns the string representation of the type
 *       that corresponds to name.
 *
 */
void TA_callAccessFunc(char *name, void *collPtr, int elemNum,
		       char *fname) {
  TA_AccessFuncEntry *fe;

  fe = TA_findAccessFuncEntry(name);
  if (fe) {
    fe->funcPtr(collPtr, elemNum, fname);
  } else {
#if DEBUG
    printf("no entry of %s found.\n", name);fflush(stdout);
#endif    
  }
}

/*
 *  void TA_callUserFunc(char *name)
 *       returns the string representation of the type
 *       that corresponds to name.
 *
 */
void TA_callUserFunc(char *name) {
  TA_UserFuncEntry *fe;

  fe = TA_findUserFuncEntry(name);
  if (fe) {
    fe->funcPtr();
  }
}





/*  
 *  void TA_readTIFFile()
 *
 *    Reads in a Type Information File that will
 *    be read in by the parallel program during startup
 *    phase.  This type information will then be given
 *    to any applications that will be connecting to the
 *    parallel program for runtime access to the parallel
 *    data.
 *
 */
void TA_readTIFFile() {
  char line[TA_MAX_LN];
  FILE *fptr;

  /*  --------------------------------------------- 
   *
   * Open the file "TypeInfoFile.TIF"
   */ 
  strcpy(line, "TypeInfoFile.TIF");
  fptr = fopen(line, "r");
  if (!fptr) {
    perror("file open of \"TypeInfoFile.TIF\" failed\n");
    exit(1);
  }

 /*  ---------------------------------------------
  *
  *   read in all the types needed for accessing 
  *   parallel data at runtime.
  * 
  *  First, the collections and element classes they use
  */
  while (fgets(line, TA_MAX_LN, fptr)) {
    TA_putType(line);
  }
}


