/*
//  ----------------------------------
//  Darryl Brown
//  University of Oregon pC++/Sage++
//
//  brkAccess.h
//
//  This program is a library that will interface with the parallel
//  program to gain data access.
//  
//  ----------------------------------
//  ----------------------------------
*/

/*
// ---------------------------------------------
// Necessary Includes
// ---------------------------------------------
 */

#ifdef __cplusplus
extern "C" {
#endif
#include "transport.h"
#include "cargo.h"
#include "typeApi.h"

/*
// ---------------------------------------------
// EXTERNAL DECLARATIONS
// ---------------------------------------------
 */

/*
// ---------------------------------------------
// Functions
// ---------------------------------------------
 */


/*
// ----------------------
// int initServerConnection(int socketNum);
//     initializes server connection given a socket number.
// ----------------------
 */
int initServerConnection(int socketNum);
/*
  // socketNum - number of socket to open server connection
  //             from.
 */


/*
// ----------------------
// int initClientConnection(int socketNum);
//     initializes server connection given a socket number.
// ----------------------
 */
int initClientConnection(int socketNum, char *host);
/*
  // socketNum - number of socket to open server connection
  //             from.
 */


/*
// ----------------------
// int sendData(void *ms, int numbytest);
//     send a package of data over the previously
//     opened connection.
// ----------------------
 */
int sendData(void *ms, int numbytes);
/*
  // ms - data to send.
  // numbytes - number of bytes in ms to send.
 */




/*
// ----------------------
// int sendText(char *ms);
//     sends textual data over the previously
//     opened connection.
// ----------------------
 */
int sendText(char *ms);
/*
  // ms - textual message to send.
 */




/*
// ----------------------
// CRG_package *receiveData();
//     receives data package over previously opened
//     connection.
// ----------------------
 */
aPackage * receiveData();




/*
// ----------------------
// void updateCurrentCollections()
//     request and receive current collection variable
//     instantiations from the parallel program.
// ----------------------
 */
void updateCurrentCollections();



/*
// ----------------------
// void printCurrentCollections();
//     print current collection variable instantiations
//     stored.
// ----------------------
 */
void printCurrentCollections();



/*
// ----------------------
// int findIthCollection(char *varname);
//     given a variable name, finds the FIRST occurence 
//     of a collection with that name in the variable 
//     list.  Note that since collections can be in different
//     scopes, they could share the same name, and therefore 
//     they would occur multiple times in the instantiation
//     list.
// ----------------------
 */
int findIthCollection(char *varname);
/*
  // varname - variable name to find in variable list.
 */




/*
// ----------------------
// void getTypes()
//     request and receive type information from
//     the executing parallel program.
// ----------------------
 */
void getTypes();




/*
// ----------------------
// char * printTypes();
//     print the types currently stored (could be none).
// ----------------------
 */
char * printTypes();


/*
 * ----------------------
 * char *getSourceInfo()
 *     request and receive current source code line number
 *     information.
 * ----------------------
 */
char *getSourceInfo();



/*
 * ----------------------
 * char *getVars()
 *     request and receive current variables in scope
 * ----------------------
 */
char *getVars();

/*
 * ----------------------
 * char * getUserDefFuncs()
 *     this func retrieves a list of all the user defined functions
 *     in the program.  "User defined" functions are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     function name.  Note that this does not include the methods, 
 *     which can be list using the getUserMemFuncs routine following
 *     this routine.
 * ----------------------
 */
char * getUserDefFuncs();



/*
 * ----------------------
 * char * getUserDefMethods()
 *     this func retrieves a list of all the user defined methods
 *     in the program.  "User defined" methods are those which the user
 *     specifies by prepending the string "UserDefined_" onto the 
 *     method name.  Note that this does not include the user defined
 *     functions which can be list using the getUserFuncs routine.
 * ----------------------
 */
char * getUserDefMethods();



/*
// ----------------------
// void skipNBrkpnts(int numskip);
//     execute program, skipping 'numskip' breakpoints.
// ----------------------
 */
void skipNBrkpnts(int numskip);
/*
  // numskip - number of breakpoints to skip.
 */



/*
// ----------------------
// void nextBrkpoint();
//     continue to the next breakpoint;
// ----------------------
 */
void nextBrkpoint();


/*
 * ----------------------
 * void terminateProgram();
 *     terminate parallel program execution.
 * ----------------------
 */
void terminateProgram();



/*
// ----------------------
// void callAccessFuncByName(char *collName, int elemNum, char *fname);
//     this func calls one of the collection element access functions
//     automatically generated by the BREEZY pre-compilation process.
//     It calls the FIRST occurence of a collection with that name in 
//     the variable list.  Note that since collections can be in different
//     scopes, they could share the same name, and therefore 
//     they would occur multiple times in the instantiation
//     list.  See 'callAccessFuncByNum' for calling a specific 
//     collection instance by number instead of name.
// ----------------------
 */
void callAccessFuncByName(char *collName, int elemNumi, 
			  int elemNumj, char *fname);
/*
  // collName - variable name of collection to access.
  // elemNumi - index of the first dimension of the element to
  //            access.
  // elemNumj - index of the second dimension of the element to
  //            access.  IF THERE IS NO SECOND DIMENSION, THIS 
  //            PARAMETER MUST BE A NEGATIVE NUMBER.
  // fname    - the function name to access (i.e. the name of 
  //            the element class variable).
 */




/*
// ----------------------
// void callAccessFuncByNum(int collnum, int elemNum, char *fname);
//     this func calls one of the collection element access functions
//     automatically generated by the BREEZY pre-compilation process.
//     It calls the collection by unique number in the list of 
//     instance variables.
// ----------------------
 */
void callAccessFuncByNum(int collnum, int elemNumi, 
			 int elemNumj, char *fname);
/*
  // collNum - number in the collection variable list 'VarList'
  //            of collection to access.
  // elemNumi - index of the first dimension of the element to
  //            access.
  // elemNumj - index of the second dimension of the element to
  //            access.  IF THERE IS NO SECOND DIMENSION, THIS 
  //            PARAMETER MUST BE A NEGATIVE NUMBER.
  // fname    - the function name to access (i.e. the name of 
  //            the element class variable).
 */






/*
// ----------------------
//  void callUserMemFunc(int collnum, 
//		     int elemnumi, int elemnumj, char *methodname);
//     this func calls a user defined element method in the parallel
//     program.  "User defined" methods are those which the user
//     specifies by prepending the string "UserDefined_" onto the 
//     method name.  The BREEZY precompilation process identifies
//     these methods and prepares runtime access to them.
// ----------------------
 */
void callUserMemFunc(int collnum, 
		     int elemnumi, int elemnumj, char *methodname);
/*
  // collNum - number in the collection variable list 'VarList'
  //            of collection to access.
  // elemNumi - index of the first dimension of the element to
  //            access.
  // elemNumj - index of the second dimension of the element to
  //            access.  IF THERE IS NO SECOND DIMENSION, THIS 
  //            PARAMETER MUST BE A NEGATIVE NUMBER.
  // methodname - the method name to access, _WITHOUT_ the prepended
  //            "UserDefined_" string.
 */



/*
// ----------------------
// void callUserFunc(char *funcname);
//     this func calls a user defined c function in the parallel
//     program.  "User defined" functions are those which the user
//     specifies by prepending the string "UserDefined_" onto the 
//     function name.  The BREEZY precompilation process identifies
//     these functions and prepares runtime access to them.
// ----------------------
 */
void callUserFunc(char *funcname);
/*
  // funcname - the function name to access, _WITHOUT_ the prepended
  //            "UserDefined_" string.
 */


/*
// ----------------------
// void callUserMemFuncOverAll(int collnum, char *memfunc);
//     this func calls the user defined member function for all elements
//     of the specified collection in the parallel program.  
//     "User defined" methods are those which the user
//     specifies by prepending the string "UserDefined_" onto the 
//     method name.  Note that this does not include the user defined
//     -functions- which can be list using the getUserFuncs routine.
// ----------------------
 */
void callUserMemFuncOverAll(int collnum, char *memfunc);



/*
 * ----------------------
 * char * printCollections()
 *     prints a list of the collections
 * ----------------------
 */
char *printCollections();


/*
// ----------------------
// void callAccessFuncOverAll(int collnum);
//     this func calls the generated member function for accessing the
//     fields of the element of the specified collection.  It makes
//     this call for all elements in the collection.  So, there will
//     be as many packages sent back by the parallel program as there
//     are elements in the collection.
// ----------------------
 */
void callAccessFuncOverAll(int collnum);


#ifdef __cplusplus
}
#endif
