/*
  Darryl Brown
  April 94
  An implementation of sockets for tk/tcl.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/file.h>
#include "clServ.h"
#include "cargo.h"
#include "brkAccess.h"

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
#include "tk.h"
#include "tcl.h"

#if 0
#include "tclInt.h" 
#endif

#define DEBUG 1
#define FRMT_SEPARATOR " #"


int Tk_server_connection(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {

  int socketfd, socketNum;
  char buf[100];
  Tcl_DString resultBuf;

  Tcl_DStringInit(&resultBuf);

  if (argc != 2) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  /* print socket number into an int; */
  sscanf(argv[1], "%d", &socketNum);
  socketfd = initServerConnection(socketNum);
  sprintf(buf, "%d", socketfd);
  Tcl_DStringAppend(&resultBuf, buf, -1);
  Tcl_DStringResult(interp, &resultBuf);
  return TCL_OK;
}


int Tk_brk_client_connection(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {

  int socketfd, socketNum;
  char buf[100];

  if (argc != 3) {
    interp->result = "wrong # args";
    return TCL_ERROR;
  }
  
  sscanf(argv[1], "%d", &socketNum);
  printf("Trying to connect to port %d with host %s\n", socketNum, argv[2]);
  fflush(stdout);
  /* call client connection with host addr (argv[2]) and 
     socket num (argv[1]);*/
  socketfd = initClientConnection(socketNum, argv[2]);
  sprintf(buf, "%d", socketfd);
  strcpy(interp->result, buf);
  return TCL_OK;
}


int Tk_close_connection(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {

  int socketfd;
  char buf[100];

  if (argc != 2) {
    strcpy(interp->result, "wrong # args for proc");
    return TCL_ERROR;
  }

  if (sscanf(argv[1], "%d", &socketfd) != 1) {
    strcpy(interp->result, "not a socket number");
    Tcl_AppendResult(interp, "\nUsage:\n\tbrk_CloseConnection>");
    return TCL_ERROR;
  }

  sprintf(buf, "%d", close(socketfd));
  strcpy(interp->result, buf);
  return TCL_OK;
}


int Tk_SendMessage(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  char buf[100];
  Tcl_DString resultBuf;

  Tcl_DStringInit(&resultBuf);

  if (argc != 2) {
    Tcl_DStringAppend(&resultBuf, 
	"\nUsage:\n\tbrk_SendText \"<str>\"", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* brkAccess function sendText */
  sendText(argv[1]);
  
  sprintf(buf, "%d", strlen(argv[1]+1));
  Tcl_DStringAppend(&resultBuf, buf, -1);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  return TCL_OK;
}

int Tk_ReceiveMessage(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  Tcl_DString resultBuf;
  aPackage *pkg;

  Tcl_DStringInit(&resultBuf);

  if (argc != 1) {
    Tcl_DStringAppend(&resultBuf, 
         "\nUsage:\n\tbrk_ReceiveMessage", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }
  
  pkg = receiveData();
  Tcl_DStringAppend(&resultBuf, pkg->data, pkg->length);
  CRG_RecyclePackage(pkg);
  
  Tcl_DStringResult(interp, &resultBuf);
  return TCL_OK;
}

int Tk_SendDataToViz(ClientData clientData, Tcl_Interp *interp,
		     int argc, char **argv) {
  FILE *f;
  int cmdArgc;
  char *cmdArgv[6], *fileId, *curPtr, line[30], fname[40], *strchr();
  int i;
  float tmpf;
  int numcount = 0;
  Tcl_DString resultBuf;
  int hasField = 0;

#if DEBUG    
  int tsti;
#endif

  /* just for initializing the result to nothing at the end of the 
     procedure  */

  Tcl_DStringInit(&resultBuf);
  
  if ((argc < 3) || (argc > 4)) {
    sprintf(line, " argc(%d) ", argc);
    for (i = 1; i < argc; i++) {
      strcat(line, argv[i]);
      strcat(line, " ");
    }
    Tcl_AppendResult(interp, "wrong # args: should be \"", 
     argv[0],
     " pipeNum dataRequest <?extract?>\"", line,
	            (char *) NULL);
    return TCL_ERROR;
  }

  for (curPtr = strtok(argv[2], " "), i = 0;
       (curPtr); 
       curPtr = strtok(NULL, " "), i++) {
    cmdArgv[i]= strdup(curPtr);
  }

  cmdArgc = i;

  strcpy(fname, cmdArgv[cmdArgc-1]);
  /* check to see if last item in data call is
   * a non-numeric (i.e. fieldname)
   */
  if (isAllDigits(fname)) {
    strcpy(fname, "");
    hasField = 0;
  } else {
    hasField = 1;
  }

#if DEBUG
  printf("fname: %s\n", fname);
  for (tsti = 0; tsti < argc; tsti++) {
    printf("argv[%d]: %s\n", tsti, argv[tsti]);
  }
  fflush(stdout);
#endif



#if DEBUG    
  for (tsti = 0; tsti < cmdArgc; tsti++) {
    printf("cmdArgv[%d]: %s\n", tsti, cmdArgv[tsti]);
  }
#endif

  Tk_CallAccessFunc((ClientData) NULL, interp, 
		    cmdArgc, cmdArgv);

#if DEBUG    
  printf("done writeing data request \"%s\"\n",   
	 interp->result); 
  fflush(stdout);
#endif

  /* store fileId argument */
  fileId = argv[1];
  
  /* get the open file */
  if (Tcl_GetOpenFile(interp, fileId, 1, 1, &f) != TCL_OK) {
    return TCL_ERROR;
  }

#if DEBUG    
  printf("retrieved open file descriptor\n"); fflush(stdout);
  fflush (stdout);
#endif
  
#if DEBUG    
  printf("done reading\n"); fflush(stdout);
fflush (stdout);
#endif

  /* now pipe interp result to viz process */
  
  clearerr(f);
  
  /* write retrieved data to pipe */
  curPtr = interp->result;
#if DEBUG    
  if (!curPtr) {
    printf("interpreter returned null pointer\n");
  } else {
    printf("interpreter returned \"%s\"\n", curPtr);
  }
  fflush(stdout);
#endif
  
  if (!curPtr)  {
    Tcl_AppendResult(interp, "error in return from data call: \"", 
		     interp->result,
		     "\": ", Tcl_PosixError(interp), (char *) NULL);
    return TCL_ERROR;
  }
  if (0) {
  curPtr = strchr(curPtr, ' ');
  curPtr++;
  strchr(curPtr, ')');
  if (!curPtr)  {
    Tcl_AppendResult(interp, "error in return from data call: \"", 
		     interp->result,
		     "\": ", Tcl_PosixError(interp), (char *) NULL);
    return TCL_ERROR;
  }
  curPtr = strtok(curPtr, " #@");
  if (!curPtr)  {
    Tcl_AppendResult(interp, "error in return from data call: \"", 
		     interp->result,
		     "\": ", Tcl_PosixError(interp), (char *) NULL);
    return TCL_ERROR;
  }
  }

  if (!hasField) {
    /* there is no field name */
    while (curPtr) {
      if (sscanf(curPtr, "%f", &tmpf) == 1) {
	numcount++;
	fprintf(f, " %s", curPtr);
      } else {
	printf( "%s was not a number, substituting 0", curPtr);
	fprintf(f, " 0");
      }
      
      if (ferror(f)) {
	Tcl_AppendResult(interp, "error writing \"", fileId,
			 "\": ", Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
      }
      curPtr = strtok(NULL, FRMT_SEPARATOR);
    }
  } else {

    /* we must pick out all non numbers... 
     * there is no field name, so we are extricating the numbers from
     * the fieldnames
     */
    if (curPtr) {
      fprintf(f, "%s", curPtr);
    } else {
      while (curPtr) {
	/* find field name */
	while (0 && strcmp(curPtr, fname) != 0) {
#if DEBUG    
	  printf("looking at string %s\n", curPtr);
	  fflush(stdout);
#endif
	  curPtr = strtok(NULL, FRMT_SEPARATOR);
	  if (curPtr == NULL) {
	    break;
	  }
	}
	if (curPtr) {

	  /* now get next item, should be the data */
	  curPtr = strtok(NULL, FRMT_SEPARATOR);

	  
	  if (sscanf(curPtr, "%f", &tmpf) == 1) {
#if 0
	    if (0 && (tmpf > 1000000000)) {
	      printf("big number: %s\n", curPtr);
	      fprintf(f, " 0.00000");
	    } else {
#endif
	      numcount++;
	      fprintf(f, " %s", curPtr);

#if DEBUG    
	      printf("(n:%d) ", curPtr);
	      fflush(stdout);
#endif
#if 0
	    }
#endif
	  } else {
	    printf( "%s was not a number, substituting 0", curPtr);
	    fprintf(f, " 0");
	  }
	
	  if (ferror(f)) {
	    Tcl_AppendResult(interp, "error writing \"", fileId,
			     "\": ", Tcl_PosixError(interp), (char *) NULL);
	    return TCL_ERROR;
	  }
	  curPtr = strtok(NULL, ")");
	  curPtr = strtok(NULL, FRMT_SEPARATOR);
#if DEBUG    
	  if (curPtr) {
	    printf("\n\tnext token: \"%s\"\n", curPtr);
	    fflush(stdout);
	  }
#endif
	}
      }
    }
  }
    
  fprintf(f, "\n");

#if DEBUG    
  printf("done parsing interp result (returning from vizDataFromSocket)\n");
  printf("the result : %s\n", interp->result);
  fflush(stdout);
#endif
  
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  return TCL_OK;
}

int Tk_GetTypes(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  Tcl_DString resultBuf;

  Tcl_DStringInit(&resultBuf);

  if (argc != 1) {
    Tcl_DStringAppend(&resultBuf, 
		      "\nUsage:\n\tbrk_getTypes", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* brkAccess function sendText */
  getTypes();
  /* printf ("types are \"%s\"\n", printTypes()); */
  Tcl_DStringAppend(&resultBuf, printTypes(), -1);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  return TCL_OK;
}


int Tk_getSourceInfo(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  Tcl_DString resultBuf;
  char *info;

  Tcl_DStringInit(&resultBuf);

  if (argc != 1) {
    Tcl_DStringAppend(&resultBuf, 
		      "\nUsage:\n\tbrk_getFuncInfo", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* brkAccess function sendText */
  info = getSourceInfo();
  Tcl_DStringAppend(&resultBuf, info, -1);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  free(info);
  return TCL_OK;
}



int Tk_getVars(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  Tcl_DString resultBuf;
  char *info;

  Tcl_DStringInit(&resultBuf);

  if (argc != 1) {
    Tcl_DStringAppend(&resultBuf, 
		      "\nUsage:\n\tbrk_getVars", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* brkAccess function sendText */
  info = getVars();
  Tcl_DStringAppend(&resultBuf, info, -1);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  free(info);
  return TCL_OK;
}


int Tk_skipNBreakpoints(ClientData clientData, Tcl_Interp *interp,
			int argc, char **argv) {
  Tcl_DString resultBuf;
  int tmpi;

  Tcl_DStringInit(&resultBuf);

  if (argc != 2) {
    Tcl_DStringAppend(&resultBuf, 
	"\nUsage:\n\tbrk_skipBreakpoints <num>", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  if (sscanf(argv[1], "%d", &tmpi) != 1) {
    Tcl_DStringAppend(&resultBuf, 
	"\nUsage:\n\tbrk_skipBreakpoints <num>", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* call function to skip breakpoints */
  
  skipNBrkpnts(tmpi);
  Tcl_DStringAppend(&resultBuf, argv[1], -1);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  return TCL_OK;
}

int Tk_terminateProgram(ClientData clientData, Tcl_Interp *interp,
			int argc, char **argv) {
  Tcl_DString resultBuf;
  Tcl_DStringInit(&resultBuf);

  if (argc != 1) {
    Tcl_DStringAppend(&resultBuf, 
	"\nUsage:\n\tbrk_terminateProgram", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* call function to skip breakpoints */
  
  terminateProgram();
  Tcl_DStringAppend(&resultBuf, "1", -1);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  return TCL_OK;
}


int Tk_CallAccessFunc(ClientData clientData, Tcl_Interp *interp,
		      int argc, char **argv) {
  aPackage *pkg;
  Tcl_DString resultBuf;
  int i, dum, numericArgc;
  char fieldSpec[100];

  Tcl_DStringInit(&resultBuf);
  numericArgc = 0;
  *fieldSpec = '\0';

  for (i = 1; i < argc; i++) {
#if DEBUG    
    printf("argv[%d]: %s\n", i, argv[i]);
#endif
    if (sscanf(argv[i], "%d", &dum) == 1) {
      numericArgc++;
    }
    if (sscanf(argv[i], "%d", &dum) != 1) {
      if (*fieldSpec) {
	/*  already something in the spec... */
	strcat(fieldSpec, " ");
	strcat(fieldSpec, argv[i]);
      } else {
	strcpy(fieldSpec, argv[i]);
      }
    }  
  }
  
#if DEBUG    
  printf("numericArgc = %d\n", numericArgc);
  printf("fieldSpec = \"%s\"\n", fieldSpec);
#endif
  if ((numericArgc < 2) || (numericArgc > 3)) {
    Tcl_DStringAppend(&resultBuf, 
		      "\nUsage:\n\tbrk_callAccessFunc <collNum> <index1> (<index2>) <fieldName>", -1);
    Tcl_DStringResult(interp, &resultBuf);
    return TCL_ERROR;
  }

  /* brkAccess function sendText */
  switch (numericArgc) {
  case 2:
    if (argc < 4) {
      callAccessFuncByNum(atoi(argv[1]),
			  atoi(argv[2]),
			  -1,
			  "");
    } else {
      callAccessFuncByNum(atoi(argv[1]),
			  atoi(argv[2]),
			  -1,
			  fieldSpec);
    }
    break;
  case 3:
    if (argc < 5) {
      callAccessFuncByNum(atoi(argv[1]),
			  atoi(argv[2]),
			  atoi(argv[3]),
			  "");
    } else {
      callAccessFuncByNum(atoi(argv[1]),
			  atoi(argv[2]),
			  atoi(argv[3]),
			  fieldSpec);
    }
    break;
  }
  
  pkg = receiveData();
  
  Tcl_DStringAppend(&resultBuf, (char *) pkg->data, pkg->length);
  Tcl_DStringResult(interp, &resultBuf);
  Tcl_DStringFree(&resultBuf);
  return TCL_OK;
}


int brk_Init(Tcl_Interp *interp) {
  Tcl_CreateCommand(interp, "brk_OpenServer", 
		    Tk_server_connection,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_OpenClient", 
		    Tk_brk_client_connection,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_CloseConnection", 
		    Tk_close_connection,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_SendMessage", 
		    Tk_SendMessage,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_ReceiveMessage",
		    Tk_ReceiveMessage,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_VizData",
		    Tk_SendDataToViz,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_GetTypes", 
		    Tk_GetTypes,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_skipBreakpoints", 
		    Tk_skipNBreakpoints,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_terminateProgram", 
		    Tk_terminateProgram,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_getSourceInfo",
		    Tk_getSourceInfo,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_getVars",
		    Tk_getVars,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand(interp, "brk_callAccessFunc",
		    Tk_CallAccessFunc,
		    (ClientData) NULL,
		    (Tcl_CmdDeleteProc *) NULL);
  
  return TCL_OK;
}

