# -----------------------------------------
#
#  Darryl Brown
#  April 1994
#
#  Socket communication for breakpoint executive;
#  
#
# -----------------------------------------
# comm.tcl
# -----------------------------------------


# -----------------------------------------
# Set up globals
set SocketNum 6871;
set ControlSocketNum 6872;
set SocketID 0;
set ControlSocketID 0;

set HostAddr 128.223.8.35

# Set up menu requests that can be made from 
# breakpoint server.
set serverMenu(get_func_info) "getFuncInfo"
set serverMenu(list_all) "getVars"
set serverMenu(get_types) "getTypes"
set serverMenu(set_continue) "setContinue"


# -----------------------------------------
# initialize connection, assign socket 
# number.

set ConnectRole "server";

proc setUpConnection {} {
  global SocketID SocketNum HostAddr ConnectRole env;
  
  puts "Setting up connection"
  if {[info exist env(pcxx_BRKHOST)]} {
    set HostAddr $env(pcxx_BRKHOST);
  } 

  if {[info exist env(pcxx_BRKCTRLSOCKET)]} {
    set ControlSocketNum $env(pcxx_BRKCTRLSOCKET);
  } 

  if {[info exist env(pcxx_BRKSOCKET)]} {
    set SocketNum $env(pcxx_BRKSOCKET);
  } 

  if {[string compare $ConnectRole "client"] == 0} {
    puts "setting client up connection.. $SocketNum $HostAddr";
    flush stdout;
    set SocketID [brk_OpenClient $SocketNum $HostAddr];
    puts "connection set.";
    flush stdout;
  } else {
    #must be server
    puts "setting server up connection.. $SocketNum $HostAddr";
    flush stdout;
    set SocketID [brk_OpenServer $SocketNum];
    puts "connection set.";
    flush stdout;
  }
}


# -----------------------------------------
proc sendMessage {mess} {
  
    # puts "sending message $mess"
    return [brk_SendMessage $mess];
}  


# -----------------------------------------
proc continueExec {} {
  sendMessage "";
  set retstr [brk_ReceiveMessage];
}


# -----------------------------------------
proc getAnswer {} {
 
  set retstr "";
    # puts "TRYING TO GET ANSWER";
  set retstr [brk_ReceiveMessage];
    # puts [format "\nFINAL RETURN STRING  :$retstr:\n"];
  return $retstr;
}

# -----------------------------------------
proc requestData {reqStr} {
  puts " in request data"
  sendMessage $reqStr;
  return [getAnswer];
}

# -----------------------------------------
proc vizData {vizChannel collNum elemNumi elemNumj fname extract} {
  
  if {$extract==1} {
    set addendum " 1 ";
  } else {
    set addendum " ";
  }

  #puts stdout \
  # "vizDT collNum $collNum elemNumi $elemNumi elemNumj $elemNumj field $fname";

  if {[llength $elemNumj]} {
    set vcom [concat brk_VizData $vizChannel \
	    [list \
	    "brk_callAccessFunc $collNum $elemNumi $elemNumj $fname"] \
	    $addendum];
    return [eval $vcom];
  } else {
    set vcom [concat brk_VizData $vizChannel \
	    [list \
	    "brk_callAccessFunc $collNum $elemNumi $fname"] \
	    $addendum];
    return [eval $vcom];
  }
}

# -----------------------------------------
proc getTypes {} {
  return [brk_GetTypes];
}


# -----------------------------------------
proc quitExecution {} {
    return [brk_terminateProgram];
}


# -----------------------------------------
proc getCollElem {args} {
  return [eval brk_callAccessFunc $args];
}
  


# -----------------------------------------
proc getCollElemField {args} {
    #coll elemi elemj field
    return [eval brk_callAccessFunc $args];
}
  

# -----------------------------------------
proc skipBreakpoints {skipNum} {
  append rexp {[^} "0-9" {]};
  # if I cant find any non-digits, continue
  if {![regexp $rexp $skipNum]} {
    brk_skipBreakpoints $skipNum;
    return $skipNum;
  } else {
    return -1;
  }
}


# -----------------------------------------
proc contBreakpoints {contNum} {
  append rexp {[^} "0-9" {]};
  # if I cant find any non-digits, continue
  if {![regexp $rexp $contNum]} {
    for {set i 0} {$i < $contNum} {incr i} {
      continueExec;
      flush stdout;
      getInstances;
      flush stdout;
      update;
      update idletasks;
      flush stdout;
    }
    flush stdout;
    return $contNum;
  } else {
    flush stdout;
    return -1;
  }
}

proc getFuncInfo {} {

  set strPointer [brk_getSourceInfo];
    
  if [regexp "^id=(.*)#fn=(.*)#ln=(.*)#(.*)$" $strPointer dummy \
      funcId fileName lineNumber strPointer] {
    if { $funcId != -1 } {
      #puts "stopped in $funcId called from $fileName:$lineNumber";
      #flush stdout;
      global_selectLine $lineNumber $fileName $funcId
    }
  } else {
    puts "communication problem: no funcId";
  }
}

# -----------------------------------------
proc getInstances {} {
  global InstanceList;

  set InstanceList "";

  brk_SendMessage "getVars";
  set strPointer [getAnswer];
  puts "getInstances returned $strPointer";
  flush stdout;

  if {[strsame $strPointer "Program Exiting"]} {
      tk_dialog .done "" "Parallel program has exited" "" 0 OK;
  }

  set funcInfo [getFuncInfo];
  puts "funcInfo $funcInfo";

  set rexp [format "\[^\n\]*\n"];
  
  while {[regexp -indices $rexp $strPointer ind]} {
    set tline [string range $strPointer [lindex $ind 0] [lindex $ind 1]];
    set strPointer [string range $strPointer [expr [lindex $ind 1]+1] end] ;
    if {[scan $tline { %s < %s > %s %s} collName elemName numElems varName] \
	  == 4} {
      set line "$collName < $elemName > $numElems $varName";
      lappend InstanceList [list $line];
    } else {
      puts "communication problem .$tline.";
    }
  }
  
  autoUpdate;

  return $InstanceList;
}

  
