#include <stdio.h>                                                              
#include <stdlib.h>                                                              
#include <string.h>                                                            
#include <ctype.h>                                                            
#include "functab.h"

static Func_List *Functions = 0;   /* list of functions */
static int num_funcs = 0;          /* number of functions */
static Func_List *mainfunc = 0;    /* pointer to "main" function */

/* ----------------------------------------------------- get_function --- */
Func_List *get_function               /* get function list entry */
   ( char *funcname,                  /* function name */
     int tag,                         /* unique id or UNKNOWN_TAG */
     Func_List *scope                 /* stmt or dummy func: pointer to parent func */
                                      /*               else: 0 */
   )
{
   Func_List *curr_func, *last_func;

   curr_func = last_func = Functions;
   while (curr_func) {
      if ((strcmp(curr_func->name, funcname) == 0) &&
          (curr_func->scope == scope)) {
         if (tag != UNKNOWN_TAG)
            curr_func->tag = tag;
         return curr_func;
      }
      last_func = curr_func;
      curr_func = curr_func->next;
   }

/*
   NOT FOUND -> add to function list
*/
   if ((curr_func = malloc(sizeof(Func_List))) == 0) {
      fprintf(stderr,"Not enough memory available.\n");
      exit(1);
   }
   curr_func->name = strdup(funcname);
   if (strcmp(funcname, "main") == 0) mainfunc = curr_func;
   if (tag == UNKNOWN_TAG)
      curr_func->tag = num_funcs++;
   else
      curr_func->tag = tag;
   curr_func->lines = 0;
   curr_func->pos = 0;
   curr_func->file = 0;
   curr_func->calls = 0;
   curr_func->numcalled = 0;
   curr_func->scope = scope;
   curr_func->next = 0;

   if (last_func)
     last_func->next = curr_func;
   else
     Functions = curr_func;
   return curr_func;
}

/* ----------------------------------------------------- output_func --- */
static void output_func                    /* print function information */
   ( FILE *pdb,                            /* output file */
     Func_List *curr_func                  /* pointer to function */
   )
{
   Child_List *curr_child;
   int num_childs;

   fprintf(pdb,"  ftag: %d %s\n",curr_func->tag,curr_func->name);
   if (curr_func->lines) {
     fprintf(pdb," ffile: %d %s @ %d.%d %d.%d %d.%d %d.%d\n",
        curr_func->lines[PU_HEAD_START], curr_func->file,
        curr_func->lines[PU_HEAD_START], curr_func->pos[PU_HEAD_START],
        curr_func->lines[PU_HEAD_END  ], curr_func->pos[PU_HEAD_END  ],
        curr_func->lines[PU_BODY_START], curr_func->pos[PU_BODY_START],
        curr_func->lines[PU_BODY_END  ], curr_func->pos[PU_BODY_END  ]
     );
   }
   else
     fprintf(pdb," ffile: 0 -\n");
   num_childs = 0;
   fprintf(pdb,"fcalls: %d -",curr_func->numcalled);
   curr_child = curr_func->calls;
   while (curr_child) {
      fprintf(pdb," %d",curr_child->func->tag);
      num_childs++;
      curr_child = curr_child->next;
   }
   fprintf(pdb,"\n");
   if (num_childs) {
      fprintf(pdb," fline:");
      curr_child = curr_func->calls;
      while (curr_child) {
         fprintf(pdb," %d",curr_child->line);
         curr_child = curr_child->next;
      }
      fprintf(pdb,"\n");
   }
   if (num_childs) {
      fprintf(pdb,"  fpos:");
      curr_child = curr_func->calls;
      while (curr_child) {
         fprintf(pdb," { %d.%d %d.%d }",curr_child->line,curr_child->start,
                                        curr_child->line,curr_child->end);
         curr_child = curr_child->next;
      }
      fprintf(pdb,"\n");
   }
   fprintf(pdb,"fclass: - - -2\n");
   fprintf(pdb," ftype: seq %s %s\n",
      (strcmp(curr_func->name,"main")==0?"Main":"Ord"),
      (curr_func->numcalled?"used":"not"));
   fprintf(pdb,"\n");
   curr_func = curr_func->next;
}

/* --------------------------------------------------- write_pdb_file --- */
void write_pdb_file                 /* generate TAU program database file */
   ( char *filename,                /* name of PDB file */
     char *arg1                     /* name of 1. argument file */
   )
{
   FILE *pdb;
   Func_List *curr_func;
   char *ptr;
   int i, roots;

   if (filename[0]) {
     if ((ptr = malloc(strlen(filename)+8)) == 0) {
        fprintf(stderr,"Not enough memory available.\n");
        exit(1);
     }
     /* Changed to be consistent with sage depfile naming convention */
     /* kurtw - 8/6/96 */
     sprintf(ptr,"%s.dep",filename);
     /* sprintf(ptr,"%s.c.dep",filename); */ 
     filename = ptr;
   } else if (mainfunc) {
     if ((ptr = malloc(strlen(mainfunc->file)+8)) == 0) {
        fprintf(stderr,"Not enough memory available.\n");
        exit(1);
     }
     strcpy(ptr,mainfunc->file);
     filename = ptr;
     /* Changed to be consistent with sage depfile naming convention */
     /* kurtw - 8/6/96 */
     if ((ptr = strrchr(filename, '.')) == 0) {
       sprintf(filename,"%s.dep",mainfunc->file);
       /* sprintf(filename,"%s.c.dep",mainfunc->file); */
     } else {
        strcpy(ptr,".dep");
       /* strcpy(ptr,".c.dep"); */
     }
   } else {
     if ((ptr = malloc(strlen(arg1)+8)) == 0) {
        fprintf(stderr,"Not enough memory available.\n");
        exit(1);
     }
     strcpy(ptr,arg1);
     filename = ptr;
     /* Changed to be consistent with sage depfile naming convention */
     /* kurtw - 8/6/96 */
     if ((ptr = strrchr(filename, '.')) == 0) {
       sprintf(filename,"%s.dep",arg1);
       /* sprintf(filename,"%s.c.dep",arg1); */
     } else {
       strcpy(ptr,".dep");
       /* strcpy(ptr,".c.dep"); */
     }
   }

   pdb = fopen(filename,"w");
   if (pdb == NULL) {
     perror(filename);
     exit(1);
   }

   /* -- add 1 if we have no main function defined -- */
   fprintf(pdb,"%d\n\n",num_funcs + (mainfunc == 0));

   curr_func = Functions;
   while (curr_func) {
      output_func(pdb,curr_func);
      curr_func = curr_func->next;
   }

   if (!mainfunc) {
     fprintf(pdb,"  ftag: %d -\n",num_funcs+1);
     fprintf(pdb," ffile: 0 -\n");
     fprintf(pdb,"fcalls: 0 -");
     roots = 0;
     curr_func = Functions;
     while (curr_func) {
        if ((curr_func->numcalled == 0) && (curr_func->lines != 0)) {
          fprintf(pdb," %d", curr_func->tag);
          roots++;
        }
        curr_func = curr_func->next;
     }
     fprintf(pdb,"\n");
     fprintf(pdb," fline:");
     for (i=0; i<roots; i++) fprintf(pdb," 0");
     fprintf(pdb,"\n");
     fprintf(pdb,"  fpos:");
     for (i=0; i<roots; i++) fprintf(pdb," {- -}");
     fprintf(pdb,"\n");
     fprintf(pdb,"fclass: - - -2\n");
     fprintf(pdb," ftype: seq Main used\n\n");
   }
   fclose(pdb);
}
