%{
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "functab.h"

extern char yytext[];
extern int yyleng;
extern int yylex(void);
extern void yyerror(char *s);

static Func_List *scope = 0;
static Child_List *centry, *clast;
static Name_List *curr;

static int *make_array(int hs, int he, int bs, int be)
{
   int *array;

   if ((array = malloc(PU_MAX*sizeof(int))) == 0 ) {
     exit(1);
   }
   array[PU_HEAD_START] = hs;
   array[PU_HEAD_END  ] = he;
   array[PU_BODY_START] = bs;
   array[PU_BODY_END  ] = be;

   return array;
}

%}
%union {
  struct {
    char *name;
    char *file;
    int line;
    int pos;
  } term;

  struct {
    Name_List *namelist;
    char *name;
    char *file;
    int type;
    int sline;
    int spos;
    int eline;
    int epos;
    int pointer;
  } nonterm;
}

%token <term> IDENTIFIER CONSTANT STRING_LITERAL SIZEOF
%token <term> PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP EQ_OP NE_OP
%token <term> AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN ADD_ASSIGN
%token <term> SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token <term> XOR_ASSIGN OR_ASSIGN TYPE_NAME
%token <term> TYPEDEF EXTERN STATIC AUTO REGISTER
%token <term> CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE CONST VOLATILE VOID
%token <term> STRUCT UNION ENUM ELLIPSIS
%token <term> CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE BREAK RETURN
%token <term> ';' '{' '}' ',' ':' '=' '(' ')' '[' ']' '.' '&' '!' '~'
%token <term> '-' '+' '*' '/' '%' '<' '>' '^' '|' '?'

%type  <term> identifier unary_operator2 typedef_name assignment_operator2 tag

%type <nonterm> declarator pointer direct_declarator type_qualifier_list
%type <nonterm> type_qualifier function_body declaration_specifiers
%type <nonterm> storage_class_specifier type_specifier translation_unit
%type <nonterm> struct_or_union_specifier enum_specifier cs_decl_list
%type <nonterm> struct_or_union struct_body enum_head enumerator_list
%type <nonterm> enumerator constant_expr declaration conditional_expr
%type <nonterm> logical_or_expr expr logical_and_expr inclusive_or_expr
%type <nonterm> exclusive_or_expr and_expr equality_expr relational_expr
%type <nonterm> shift_expr additive_expr multiplicative_expr cast_expr
%type <nonterm> unary_expr postfix_expr unary_operator primary_expr
%type <nonterm> assignment_expr initializer compound_statement 
%type <nonterm> declaration_list argument_expr_list assignment_operator
%type <nonterm> init_declarator init_declarator_list specifier_qualifier_list
%type <nonterm> function_definition jump_statement struct_declaration
%type <nonterm> iteration_statement statement selection_statement
%type <nonterm> expression_statement statement_list labeled_statement
%type <nonterm> initializer_list direct_abstract_declarator abstract_declarator
%type <nonterm> type_name parameter_declaration identifier_list struct_declarator
%type <nonterm> struct_declarator_list cs_stmt_list parameter_list empty
%type <nonterm> struct_declaration_list external_declaration parameter_type_list

%start translation_unit

%%

primary_expr
	: identifier
            {
              $$.name = $1.name;
              $$.pointer = 0;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	| CONSTANT
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	| STRING_LITERAL
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	| '(' expr ')'
            {
              $$.name = $2.name;
              $$.pointer = $2.pointer;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	;

postfix_expr
	: primary_expr
	| postfix_expr '[' expr ']'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	| postfix_expr '(' ')'
            {
              if ((centry = malloc(sizeof(Child_List))) == 0) {
                fprintf(stderr,"Not enough memory available.\n");
                exit(1);
              }
              if ($1.pointer)
                centry->func = get_function($1.name,UNKNOWN_TAG,scope);
              else
                centry->func = get_function($1.name,UNKNOWN_TAG,0);
              centry->func->numcalled++;
              centry->line  = $1.sline;
              centry->start = $1.spos;
              centry->end   = $1.epos;
              if (clast)
                clast = clast->next = centry;
              else
                clast = scope->calls = centry;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| postfix_expr '(' argument_expr_list ')'
            {
              if ((centry = malloc(sizeof(Child_List))) == 0) {
                fprintf(stderr,"Not enough memory available.\n");
                exit(1);
              }
              if ($1.pointer)
                centry->func = get_function($1.name,UNKNOWN_TAG,scope);
              else
                centry->func = get_function($1.name,UNKNOWN_TAG,0);
              centry->func->numcalled++;
              centry->line  = $1.sline;
              centry->start = $1.spos;
              centry->end   = $1.epos;
              if (clast)
                clast = clast->next = centry;
              else
                clast = scope->calls = centry;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	| postfix_expr '.' identifier
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + strlen($3.name);
            }
	| postfix_expr PTR_OP identifier
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + strlen($3.name);
            }
	| postfix_expr INC_OP
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 2;
            }
	| postfix_expr DEC_OP
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.line;
              $$.epos = $2.pos = 2;
            }
	;

argument_expr_list
	: assignment_expr
	| argument_expr_list ',' assignment_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

unary_expr
	: postfix_expr
	| INC_OP unary_expr
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| DEC_OP unary_expr
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| unary_operator cast_expr
            {
              $$.name = $2.name;
              $$.pointer = ($1.name[0] == '*');
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| SIZEOF unary_expr
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| SIZEOF '(' type_name ')'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	;

unary_operator
	: unary_operator2
            {
              $$.name = $1.name;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 1;
            }
	;

unary_operator2
	: '&'
	| '*'
	| '+'
	| '-'
	| '~'
	| '!'
	;

cast_expr
	: unary_expr
	| '(' type_name ')' cast_expr
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $4.eline;
              $$.epos = $4.epos;
            }
	;

multiplicative_expr
	: cast_expr
	| multiplicative_expr '*' cast_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| multiplicative_expr '/' cast_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| multiplicative_expr '%' cast_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

additive_expr
	: multiplicative_expr
	| additive_expr '+' multiplicative_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| additive_expr '-' multiplicative_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

shift_expr
	: additive_expr
	| shift_expr LEFT_OP additive_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| shift_expr RIGHT_OP additive_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

relational_expr
	: shift_expr
	| relational_expr '<' shift_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| relational_expr '>' shift_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| relational_expr LE_OP shift_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| relational_expr GE_OP shift_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

equality_expr
	: relational_expr
	| equality_expr EQ_OP relational_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| equality_expr NE_OP relational_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

and_expr
	: equality_expr
	| and_expr '&' equality_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

exclusive_or_expr
	: and_expr
	| exclusive_or_expr '^' and_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

inclusive_or_expr
	: exclusive_or_expr
	| inclusive_or_expr '|' exclusive_or_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

logical_and_expr
	: inclusive_or_expr
	| logical_and_expr AND_OP inclusive_or_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

logical_or_expr
	: logical_and_expr
	| logical_or_expr OR_OP logical_and_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

conditional_expr
	: logical_or_expr
	| logical_or_expr '?' expr ':' conditional_expr
	;

assignment_expr
	: conditional_expr
	| unary_expr assignment_operator assignment_expr
	;

assignment_operator
	: assignment_operator2
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 1;
            }
	;

assignment_operator2
	: '='
	| MUL_ASSIGN
	| DIV_ASSIGN
	| MOD_ASSIGN
	| ADD_ASSIGN
	| SUB_ASSIGN
	| LEFT_ASSIGN
	| RIGHT_ASSIGN
	| AND_ASSIGN
	| XOR_ASSIGN
	| OR_ASSIGN
	;

expr
	: assignment_expr
	| expr ',' assignment_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

constant_expr
	: conditional_expr
	;

declaration
	: declaration_specifiers ';'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	| declaration_specifiers init_declarator_list ';'
            {
              curr = $2.namelist;
              while (curr) {
                if ($1.type) add_typedef(curr->name);
                curr = curr->next;
              }
              free_name_list($2.namelist);
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	;

declaration_specifiers
	: storage_class_specifier
	| storage_class_specifier declaration_specifiers
            {
              $$.type = $1.type;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| type_specifier
            {
              $$.type = 0;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $1.eline;
              $$.epos = $1.epos;
            }
	| type_specifier declaration_specifiers
            {
              $$.type = $2.type;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| type_qualifier
            {
              $$.type = 0;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $1.eline;
              $$.epos = $1.epos;
            }
	| type_qualifier declaration_specifiers
            {
              $$.type = $2.type;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

init_declarator_list
	: init_declarator
            {
              $$.namelist = alloc_name();
              $$.namelist->name = $1.name;
              $$.namelist->next = 0;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $1.eline;
              $$.epos = $1.epos;
            }
	| init_declarator_list ',' init_declarator
            {
              $$.namelist = alloc_name();
              $$.namelist->name = $3.name;
              $$.namelist->next = $1.namelist;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

init_declarator
	: declarator
	| declarator '=' initializer
            {
              $$.name = $1.name;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

storage_class_specifier
	: TYPEDEF
            {
              $$.type = 1;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 7;
            }
	| EXTERN
            {
              $$.type = 0;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 6;
            }
	| STATIC
            {
              $$.type = 0;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 6;
            }
	| AUTO
            {
              $$.type = 0;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 4;
            }
	| REGISTER
            {
              $$.type = 0;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 8;
            }
	;

type_specifier
	: VOID
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 4;
            }
	| CHAR
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 4;
            }
	| SHORT
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 5;
            }
	| INT
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 3;
            }
	| LONG
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 4;
            }
	| FLOAT
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 5;
            }
	| DOUBLE
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 6;
            }
	| SIGNED
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 6;
            }
	| UNSIGNED
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 8;
            }
	| struct_or_union_specifier
	| enum_specifier
	| typedef_name
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	;

struct_or_union_specifier
	: struct_or_union tag struct_body
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.sline;
              $$.epos = $3.spos;
            }
	| struct_or_union struct_body
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.sline;
              $$.epos = $2.spos;
            }
	| struct_or_union tag
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.line;
              $$.epos = $2.pos + strlen($2.name);
            }
	;

tag
	: IDENTIFIER
	| TYPE_NAME /*HACK!!!*/
	;

struct_body
	: '{' struct_declaration_list '}'
           {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
           }
	;

struct_or_union
	: STRUCT
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 6;
            }
	| UNION
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 5;
            }
	;

struct_declaration_list
	: struct_declaration
	| struct_declaration_list struct_declaration
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

struct_declaration
	: specifier_qualifier_list struct_declarator_list ';'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	;

specifier_qualifier_list
	: type_specifier
	| type_specifier specifier_qualifier_list
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| type_qualifier
	| type_qualifier specifier_qualifier_list
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

struct_declarator_list
	: struct_declarator
	| struct_declarator_list ',' struct_declarator
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

struct_declarator
	: declarator
	| ':' constant_expr
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| declarator ':' constant_expr
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

enum_specifier
	: enum_head '{' enumerator_list '}'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	| enum_head identifier '{' enumerator_list '}'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $5.line;
              $$.epos = $5.pos + 1;
            }
	| enum_head identifier
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.line;
              $$.epos = $2.pos + strlen($2.name);
            }
	;

enum_head
	: ENUM
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 4;
            }
	;

enumerator_list
	: enumerator
	| enumerator_list ',' enumerator
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

enumerator
	: identifier
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	| identifier '=' constant_expr
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

type_qualifier
	: CONST
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 5;
            }
	| VOLATILE
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 8;
            }
	;

declarator
	: direct_declarator
	| pointer direct_declarator
            {
              $$.name = $2.name;
              $$.file = $2.file;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

direct_declarator
	: identifier
            {
              $$.name = $1.name;
              $$.file = $1.file;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	| '(' declarator ')'
            {
              $$.name = $2.name;
              $$.pointer = $2.pointer;
              $$.file = $2.file;
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_declarator '[' ']'
            {
              $$.name = $1.name;
              $$.file = $1.file;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_declarator '[' constant_expr ']'
            {
              $$.name = $1.name;
              $$.file = $1.file;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	| direct_declarator '(' ')'
            {
              $$.name = $1.name;
              $$.pointer = $1.pointer;
              $$.file = $1.file;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_declarator '(' parameter_type_list ')'
            {
              $$.name = $1.name;
              $$.pointer = $1.pointer;
              $$.file = $1.file;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	| direct_declarator '(' identifier_list ')'
            {
              $$.name = $1.name;
              $$.pointer = $1.pointer;
              $$.file = $1.file;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	;

pointer
	: '*'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 1;
            }
	| '*' type_qualifier_list
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| '*' pointer
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| '*' type_qualifier_list pointer
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

type_qualifier_list
	: type_qualifier
	| type_qualifier_list type_qualifier
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

identifier_list
	: identifier
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + strlen($1.name);
            }
	| identifier_list ',' identifier
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + strlen($3.name);
            }
	;

parameter_type_list
	: parameter_list
	| parameter_list ',' ELLIPSIS
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + strlen($3.name);
            }
	;

parameter_list
	: parameter_declaration
	| parameter_list ',' parameter_declaration
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

parameter_declaration
	: declaration_specifiers declarator
            {
              $$.type = $1.type;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	| declaration_specifiers
	| declaration_specifiers abstract_declarator
            {
              $$.type = $1.type;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

type_name
	: specifier_qualifier_list
	| specifier_qualifier_list abstract_declarator
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

abstract_declarator
	: pointer
	| direct_abstract_declarator
	| pointer direct_abstract_declarator
            {
              $$.name = $2.name;
              $$.pointer = 1;
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

direct_abstract_declarator
	: '(' abstract_declarator ')'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| '[' ']'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	| '[' constant_expr ']'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_abstract_declarator '[' ']'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_abstract_declarator '[' constant_expr ']'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	| '(' ')'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	| '(' parameter_type_list ')'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_abstract_declarator '(' ')'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| direct_abstract_declarator '(' parameter_type_list ')'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $4.line;
              $$.epos = $4.pos + 1;
            }
	;

typedef_name
	: TYPE_NAME
	;

initializer
	: assignment_expr
	| '{' initializer_list '}'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| '{' initializer_list ',' '}'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	;

initializer_list
	: initializer
	| initializer_list ',' initializer
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

statement
	: labeled_statement
	| compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	;

labeled_statement
	: identifier ':' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	| CASE constant_expr ':' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $4.eline;
              $$.epos = $4.epos;
            }
	| DEFAULT ':' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.eline;
              $$.epos = $3.epos;
            }
	;

compound_statement
	: '{'
            {
              enter_block();
            }
          cs_decl_list cs_stmt_list
            {
              leave_block();
            }
          '}'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $6.line;
              $$.epos = $6.pos + 1;
            }
	;

cs_decl_list
	: declaration_list
	| empty
	;

cs_stmt_list
	: statement_list
	| empty
	;

empty
	: /*empty*/
            {
              $$.name = 0;
              $$.sline= 0;
              $$.spos = 0;
              $$.eline= 0;
              $$.epos = 0;
            }
	;

declaration_list
	: declaration
	| declaration_list declaration
	;

statement_list
	: statement
	| statement_list statement
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

expression_statement
	: ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $1.line;
              $$.epos = $1.pos + 1;
            }
	| expr ';'
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	;

selection_statement
	: IF '(' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $5.eline;
              $$.epos = $5.epos;
            }
	| IF '(' expr ')' statement ELSE statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $7.eline;
              $$.epos = $7.epos;
            }
	| SWITCH '(' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $5.eline;
              $$.epos = $5.epos;
            }
	;

iteration_statement
	: WHILE '(' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $5.eline;
              $$.epos = $5.epos;
            }
	| DO statement WHILE '(' expr ')' ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $7.line;
              $$.epos = $7.pos + 1;
            }
	| FOR '(' ';' ';' ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $6.eline;
              $$.epos = $6.epos;
            }
	| FOR '(' ';' ';' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $7.eline;
              $$.epos = $7.epos;
            }
	| FOR '(' ';' expr ';' ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $7.eline;
              $$.epos = $7.epos;
            }
	| FOR '(' ';' expr ';' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $8.eline;
              $$.epos = $8.epos;
            }
	| FOR '(' expr ';' ';' ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $7.eline;
              $$.epos = $7.epos;
            }
	| FOR '(' expr ';' ';' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $8.eline;
              $$.epos = $8.epos;
            }
	| FOR '(' expr ';' expr ';' ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $8.eline;
              $$.epos = $8.epos;
            }
	| FOR '(' expr ';' expr ';' expr ')' statement
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $9.eline;
              $$.epos = $9.epos;
            }
	;

jump_statement
	: GOTO identifier ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	| CONTINUE ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	| BREAK ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	| RETURN ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $2.line;
              $$.epos = $2.pos + 1;
            }
	| RETURN expr ';'
            {
              $$.sline= $1.line;
              $$.spos = $1.pos;
              $$.eline= $3.line;
              $$.epos = $3.pos + 1;
            }
	;

translation_unit
	: external_declaration
	| translation_unit external_declaration
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.eline;
              $$.epos = $2.epos;
            }
	;

external_declaration
	: function_definition
	| declaration
	;

function_definition
	: declarator
            {
              scope = get_function($1.name, UNKNOWN_TAG, 0);
              scope->file = strdup($1.file);
              clast = 0;
            }
	  function_body
            {
              scope->lines = make_array($1.sline,$1.eline,$3.sline,$3.eline);
              scope->pos   = make_array($1.spos ,$1.epos ,$3.spos ,$3.epos );
            }
	| declaration_specifiers declarator
            {
              scope = get_function($2.name, UNKNOWN_TAG, 0);
              scope->file = strdup($2.file);
              clast = 0;
            }
	  function_body
            {
              scope->lines = make_array($1.sline,$2.eline,$4.sline,$4.eline);
              scope->pos   = make_array($1.spos ,$2.epos ,$4.spos ,$4.epos );
            }
	;

function_body
	: compound_statement
	| declaration_list  compound_statement
            {
              $$.sline= $1.sline;
              $$.spos = $1.spos;
              $$.eline= $2.sline;
              $$.epos = $2.spos;
            }
	;

identifier
	: IDENTIFIER
	;
%%

#include <stdio.h>

extern int yylineno;
extern char yyfilename[];

void yyerror(char *s)
{
	fflush(stdout);
	printf ("\"%s\", line %d: %s\n", yyfilename, yylineno, s);
}
