#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "util.h"
#include "functab.h"

extern int yyparse(void);
extern char yyfilename[];
extern FILE *yyin;

extern int yydebug;

typedef struct mystr {
  char str[8192];
  char *last;
  char *mark;
} String;

#ifndef CPP
#  define CPP "cc -E"
#endif

static String cppline;
static char opt[] = " -I";

static void InitString (String *s) {
  s->str[0] = '\0';
  s->last = s->str;
  s->mark = 0;
}

static char *MarkString (String *s) {
  s->mark = s->last;
  return s->mark;
}

static void CatString (String *s, const char *str, int atmark) {
  if (atmark) s->last = s->mark;
  while ((*(s->last)++ = *str++));
  (s->last)--;
}

static void usage(char *prog) {
  fprintf (stderr, "Usage: %s\n", prog);
  fprintf (stderr, "    [-h] [-o <depfile basename>]\n");
  fprintf (stderr, "    [cpp switches -I/-D/-U]\n");
  fprintf (stderr, "    filenames...\n");
}

int main(int argc, char **argv)
{
  int ch;
  extern char *optarg;
  extern int optind;
  char *filename = "";
  char *arg1;
   
  InitString(&cppline);
  CatString (&cppline, CPP, 0);

  while ((ch = getopt(argc, argv, "hI:D:U:o:")) != EOF) {
    switch (ch) {
      case 'h':
        usage(argv[0]);
        exit(1);
      case 'I':
      case 'D':
      case 'U':
        opt[2] = ch;
        CatString (&cppline, opt, 0);
        CatString (&cppline, optarg, 0);
        break;
      case 'o':
        filename = optarg;
    }
  }
  MarkString(&cppline);
  arg1 = argv[optind];

  if (optind == argc) {
    usage(argv[0]);
    exit(1);
  }

  for (; optind < argc; optind++) {
    CatString (&cppline, " ", 1);
    CatString (&cppline, argv[optind], 0);
    if ((yyin = popen(cppline.str, "r")) == NULL) {
      perror(argv[optind]);
      exit(1);
    }
#   ifdef EBUG
    printf("%s:\n", argv[optind]);
#   endif
    init_scanner();
    if (yyparse() == 1) exit(1);
    pclose(yyin);
  }
  if (arg1) write_pdb_file(filename, arg1);
  exit(0);
}
