D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][+-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*

%{
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include "util.h"
#include "y.tab.h"

#undef input
#undef unput
extern int input(void);
extern void unput(int c);
extern int check_type(void);
extern void line_number(void);
extern void comment(void);

extern char *strdup(const char *);

char yyfilename[PATH_MAX] = "stdin";
int yycolumn = 0;

#define yyreturn(t) {\
  yylval.term.name = strdup(yytext); \
  yylval.term.file = yyfilename; \
  yylval.term.line = yylineno; \
  yylval.term.pos  = yycolumn - strlen(yytext); \
  return(t); \
}
%}

%%

"#"			{ line_number(); }
"/*"			{ comment(); }

"auto"			{ yyreturn(AUTO); }
"break"			{ yyreturn(BREAK); }
"case"			{ yyreturn(CASE); }
"char"			{ yyreturn(CHAR); }
"const"			{ yyreturn(CONST); }
"continue"		{ yyreturn(CONTINUE); }
"default"		{ yyreturn(DEFAULT); }
"do"			{ yyreturn(DO); }
"double"		{ yyreturn(DOUBLE); }
"else"			{ yyreturn(ELSE); }
"enum"			{ yyreturn(ENUM); }
"extern"		{ yyreturn(EXTERN); }
"float"			{ yyreturn(FLOAT); }
"for"			{ yyreturn(FOR); }
"goto"			{ yyreturn(GOTO); }
"if"			{ yyreturn(IF); }
"int"			{ yyreturn(INT); }
"long"			{ yyreturn(LONG); }
"register"		{ yyreturn(REGISTER); }
"return"		{ return(RETURN); }
"short"			{ yyreturn(SHORT); }
"signed"		{ yyreturn(SIGNED); }
"sizeof"		{ yyreturn(SIZEOF); }
"static"		{ yyreturn(STATIC); }
"struct"		{ yyreturn(STRUCT); }
"switch"		{ yyreturn(SWITCH); }
"typedef"		{ yyreturn(TYPEDEF); }
"union"			{ yyreturn(UNION); }
"unsigned"		{ yyreturn(UNSIGNED); }
"void"			{ yyreturn(VOID); }
"volatile"		{ yyreturn(VOLATILE); }
"while"			{ yyreturn(WHILE); }

{L}({L}|{D})*		{ yyreturn(check_type()); }

0[xX]{H}+{IS}?		{ yyreturn(CONSTANT); }
0[xX]{H}+{IS}?		{ yyreturn(CONSTANT); }
0{D}+{IS}?		{ yyreturn(CONSTANT); }
0{D}+{IS}?		{ yyreturn(CONSTANT); }
{D}+{IS}?		{ yyreturn(CONSTANT); }
{D}+{IS}?		{ yyreturn(CONSTANT); }
'(\\.|[^\\'])+'		{ yyreturn(CONSTANT); }

{D}+{E}{FS}?		{ yyreturn(CONSTANT); }
{D}*"."{D}+({E})?{FS}?	{ yyreturn(CONSTANT); }
{D}+"."{D}*({E})?{FS}?	{ yyreturn(CONSTANT); }

\"(\\.|[^\\"])*\"	{ yyreturn(STRING_LITERAL); }

"..."                   { yyreturn(ELLIPSIS); }
">>="			{ yyreturn(RIGHT_ASSIGN); }
"<<="			{ yyreturn(LEFT_ASSIGN); }
"+="			{ yyreturn(ADD_ASSIGN); }
"-="			{ yyreturn(SUB_ASSIGN); }
"*="			{ yyreturn(MUL_ASSIGN); }
"/="			{ yyreturn(DIV_ASSIGN); }
"%="			{ yyreturn(MOD_ASSIGN); }
"&="			{ yyreturn(AND_ASSIGN); }
"^="			{ yyreturn(XOR_ASSIGN); }
"|="			{ yyreturn(OR_ASSIGN); }
">>"			{ yyreturn(RIGHT_OP); }
"<<"			{ yyreturn(LEFT_OP); }
"++"			{ yyreturn(INC_OP); }
"--"			{ yyreturn(DEC_OP); }
"->"			{ yyreturn(PTR_OP); }
"&&"			{ yyreturn(AND_OP); }
"||"			{ yyreturn(OR_OP); }
"<="			{ yyreturn(LE_OP); }
">="			{ yyreturn(GE_OP); }
"=="			{ yyreturn(EQ_OP); }
"!="			{ yyreturn(NE_OP); }
";"			{ yyreturn(';'); }
"{"			{ yyreturn('{'); }
"}"			{ yyreturn('}'); }
","			{ yyreturn(','); }
":"			{ yyreturn(':'); }
"="			{ yyreturn('='); }
"("			{ yyreturn('('); }
")"			{ yyreturn(')'); }
"["			{ yyreturn('['); }
"]"			{ yyreturn(']'); }
"."			{ yyreturn('.'); }
"&"			{ yyreturn('&'); }
"!"			{ yyreturn('!'); }
"~"			{ yyreturn('~'); }
"-"			{ yyreturn('-'); }
"+"			{ yyreturn('+'); }
"*"			{ yyreturn('*'); }
"/"			{ yyreturn('/'); }
"%"			{ yyreturn('%'); }
"<"			{ yyreturn('<'); }
">"			{ yyreturn('>'); }
"^"			{ yyreturn('^'); }
"|"			{ yyreturn('|'); }
"?"			{ yyreturn('?'); }

[ \t\v\n\f]		{ }
.			{ /* ignore bad characters */ }

%%

int yywrap(void)
{
	return(1);
}

extern int input(void)
{
	if (yysptr > yysbuf) {
		/* retrieve pushed-back character */
		yytchar = *--yysptr;
	} else {
		yytchar = getc(yyin);
		if (yytchar == EOF) {
			return 0;
		}
	}

	/* count yycolumn and yylineno */
	if (yytchar == '\n') {
		yylineno++;
		yycolumn = 0;
	} else {
		yycolumn++;
	}

	return yytchar;
}

extern void unput(int c)
{
  yytchar = c;
  if (yytchar=='\n') {
    yylineno--;
  }
  *yysptr++=yytchar;
  yycolumn--;
}

void comment(void)
{
	char c, c1;

loop:
	/* we have already seen a / and a * */
	while ((c = input()) != '*' && c != 0) /* NOTHING */;

	if (c != 0 && (c1 = input()) != '/' )
	{
		unput(c1);
		goto loop;
	}
}

#define READWHILE(cond)	while(cond) c = input();

void line_number(void)
{
	char c;
	/* skip spaces */
	c = input();
	READWHILE ((c == ' ' || c == '\t'));

	if (c >= '0' && c <= '9') {
		/* line number specification */
		int line_num = 0;
		while (c >= '0' && c <= '9') {
			line_num = line_num * 10 + c - '0';
			c = input();
		}
		if (line_num > 0) {
		    yylineno = line_num - (c != '\n');
                }
		READWHILE ((c == ' ' || c == '\t'));
		if (c == '"') {
			char *yf = yyfilename;
			do {
				c = input();
				*yf++ = c;
			} while (c != '"');
			*(yf-1) = '\0';
		}
	}

	/* flush rest of line */
	READWHILE ((c != '\n'));
}

int check_type(void)
{

    if (is_typedef(yytext)) {
	return (TYPE_NAME);
    } else {
	return (IDENTIFIER);
    }
}
