#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"

#define MAX_NESTING 64

int in_typedef = 0;

static Name_List *typedef_stack[MAX_NESTING];
static int level = 0;
static Name_List *free_list = 0;

void free_name_list(Name_List *list)
{
  Name_List *h;

  while (list) {
    h = list->next;
    list->next = free_list;
    free_list = list;
#   ifdef EBUG
    printf ("free %x\n", list);
#    endif
    list = h;
  }
}

void init_scanner (void)
{
  int i;

  for (i=0; i<MAX_NESTING; i++) {
     free_name_list(typedef_stack[i]);
     typedef_stack[i] = 0;
   }
  level = 0;
  in_typedef = 0;
}

void enter_block (void)
{
# ifdef EBUG
  printf("enter_block\n");
# endif
  level++;
  if (level == MAX_NESTING) {
    fprintf (stderr, "too deeply nested\n");
    exit (1);
  }
}

void leave_block (void)
{
# ifdef EBUG
  printf("leave_block\n");
# endif
  free_name_list(typedef_stack[level]);
  typedef_stack[level] = 0;
  level--;
  if (level < 0) {
    fprintf (stderr, "internal nesting error\n");
    exit (1);
  }
}

Name_List *alloc_name (void)
{
  Name_List *h;

  if (free_list) {
    h = free_list;
    free_list = free_list->next;
#   ifdef EBUG
    printf ("reusing %x\n", h);
#   endif
    return h;
  } else
    return malloc(sizeof(Name_List));
}

void add_typedef (char *name)
{
  Name_List *new = alloc_name();
  new->name = name;
  new->next = typedef_stack[level];
  typedef_stack[level] = new;
# ifdef EBUG
  printf ("  add typedef %s\n", name);
# endif
}

int is_typedef (char *name)
{
  Name_List *curr;
  int block = level;

  while (block >= 0) {
    curr = typedef_stack[block];
    while (curr) {
      if (strcmp(curr->name,name) == 0) {
#       ifdef EBUG
        printf ("  type %s\n", name);
#       endif
        return 1;
      }
      curr = curr->next;
    }
    block--;
  }
# ifdef EBUG
  printf ("  identifier %s\n", name);
# endif
  return 0;
}
