// -*- C++ -*-
//
// Control-flow-graph analyzer data structures.

struct CFG;
struct CFGE;
struct CFGTEXT;

///////////////////////////////////////////////////////////////////////////
// One CFG Edge node
//
// Edges can be linked together to form sets of edges.

struct CFGE {
  CFGE( CFG *, const char *, CFGE * );
  ~CFGE();
  CFG     *sink;
  CFGTEXT *label;
  CFGE    *next;
};

///////////////////////////////////////////////////////////////////////////
// Text label on CFG and CFGE nodes.
struct CFGTEXT {
  CFGTEXT( const char *, CFGTEXT * );
  ~CFGTEXT();
  int operator==( const char * );
  char *label;
  CFGTEXT *next;
};

///////////////////////////////////////////////////////////////////////////
// One CFG Vertex node.

// Tags for the 'xtag' (eXit TAG) field.

enum CFG_xtag { 
  L_END, L_IF, L_FOR, L_WHILE, L_DO_WHILE, L_SWITCH, L_BREAK, L_CONTINUE,
  L_RETURN, L_GOTO, L_CASE
};

struct CFG {
  CFG();
  CFG( const char *, const int xtag );
  ~CFG();
  void init();
  void label( const char * ); // add label
  void label( const char *, const int xtag ); // add label
  int  unlabelled();          // test: is node unlabelled?

  int     tag;                // node tag
  int     xtag;               // node exit tag (type of exit structure)
  int     level;              // max level of node in graph
  int     width;              // width of subgraph
  CFGTEXT *thelabel;          // textual representation of label
  CFGE    *children;          // list of out-edges

  int     dumped;             // 1 if has been dumped (used by cgm)

// internal
  int     childcount;         // number of children (out edges)
  int     parentcount;        // number of parents (in edges)
  int     links;              // number of links made during flooding
  int     mustbranch;         // 1 if branch target
  CFG     *gclink;            // link to another allocated node.
  CFG     *gclist;            // 0 or link to all allocated nodes.
};

///////////////////////////////////////////////////////////////////////////
// Interface.

#ifndef CFG_NOPROTO
CFG *compute_cfg( SgFuncHedrStmt * );
#endif

// eof
