/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include <stdio.h>
# include <stdlib.h>
# include <signal.h>

# include "sage++user.h"
# include "sage++callgraph.h"

# ifndef __GNUC__
#   include <unistd.h>
#else
extern "C" int unlink(const char *path);
extern "C" int access(char *path, int amode);
# endif

# include "compatible.h"
# ifdef SYS5
#   include <string.h>
# else
#   include <strings.h>
# endif

# ifndef R_OK
#   define R_OK 04
# endif

/* -- external functions -------------------------------------------------- */
/* -- from: scanfile.c ---------------------------------------------------- */
extern "C" void openFile (char *path);
extern "C" void closeFile (void);
extern "C" char *getLine (int l);

/* -- global variables ---------------------------------------------------- */
# ifndef TRUE
#   define FALSE 0
#   define TRUE  1
# endif

static int search = TRUE;
static int warnings = FALSE;
static int foundErrors = FALSE;
int isFortran = FALSE;
int allLowercase = TRUE;

static char *dumped;

static char *filepath = "";

static char *findStr (register char *text, register char *pattern)
{
  do {
    register char *tt = text;
    register char *pp = pattern;
    do {
      if (*pp == '\0') return ((char *)text);
    } while (*tt++ == *pp++);
  } while (*text++ != '\0');
  return (NULL);
}

static int DumpFunc (SgSymbol *funcsymb)
{
  SgMemberFuncSymb *memsymb;
  SgStatement *body;
  char *classname = NULL, *filename = "-", *funcname = funcsymb->identifier();
  int funcline = 0;
  char *l1, *l2, *ptr;

  if ( memsymb = isSgMemberFuncSymb(funcsymb) ) {
    body      = memsymb->body();
    classname = memsymb->className()->identifier();
  }
  else {
    body      = funcsymb->body();
    classname = NULL;
  }

  if ( body ) {
    funcline = body->lineNumber();
    filename = body->fileName();
  }

  if ( search ) {
    if ( strcmp (filename, filepath) != 0 ) {
      search = FALSE;
      filepath = filename;
      openFile (filepath);
    }
    else
      return FALSE;
  }

  if ( strcmp (filename, filepath) == 0 ) {
    l1 = getLine (funcline);
    l2 = getLine (funcline+1);
    ptr = NULL;
    if ( !findStr (l1, funcname) ) {
      if ( warnings ) ptr = "???";
      if ( !findStr (l2, funcname) ) {
        ptr = "!!!";
        foundErrors = TRUE;
      }
    }

    if ( ptr ) {
      if ( *filename == '.' )
        printf ("%s", filename);
      else
        printf ("./%s", filename);
      printf (":%d:", funcline);
      if ( classname )
        printf (" %s::%s\n", classname, funcname);
      else
        printf (" %s\n", funcname);
      printf ("%s %4d: %s\n", ptr, funcline, l1);
      printf ("%s %4d: %s\n\n", ptr, funcline+1, l2);
    }
    return TRUE;
  }
  else
    return FALSE;
}

void checklines( SgFile *f, SgFunctionSymb *main )
{
  int prev, ln, lasterr;
  SgStatement *s;
  char *fn = main->body()->fileName(); 
  
  openFile( fn );
  prev = -1;
  lasterr = -2;
  for ( s = f->firstStatement() ; s != 0 ; s = s->lexNext() ) {
    ln = s->lineNumber();
    if (getLine( ln ) == 0) {
      // do some basic filtering of runs to avoid too many errors.
      if ( lasterr != prev )
	printf( "%s: bogus line# %d (prev = %d)\n", fn, ln, prev );
      lasterr = ln;
    }
    prev = ln;
  }
}

/* -- M A I N  function --------------------------------------------------- */

main (int argc, char **argv)
{
  SgProject *project = 0;
  char *projFile = NULL;
  char *depFile = NULL;
  char *p;
  FILE *pf = 0;
  int i, l;
  int scanned;
  SgSymbol *mainSymb = 0, *f;
  SgFile *file;

  if (argc < 2 || strlen( argv[1] ) < 4) {
    fprintf( stderr, "Usage: %s depfile\n", argv[0] );
    exit( 1 );
  }

  depFile = argv[1];
  l = strlen(argv[1]);
  p = argv[1] + l - 4;
  if ( strcmp (p, ".dep") != 0 ) {
    fprintf (stderr,"ERROR: `%s': no dep file\n", argv[1]);
    exit (1);
  }
  else {
    *p = '\0';
    projFile = new char[l+2];
    sprintf (projFile, "%s.proj", argv[1]);
    *p = '.';
  }

  if ( access (projFile, R_OK) == -1 ) {
    pf = fopen (projFile, "w");
    fprintf (pf, "%s\n", depFile);
    fclose (pf);
  }
  project = new SgProject (projFile);
  file = &(project->file(0));

  int num = file->numberOfFunctions();

  dumped = new char[num];
  for (i=0; i<file->numberOfFunctions(); i++) {
    dumped[i] = FALSE;
    f = file->functions(i)->symbol();
    p = f->identifier();
    if ( strcmp (p, "Processor_Main") == 0 ) mainSymb = f;
  }
  if (!mainSymb) {
    fprintf (stderr, "%s: no main function\n", depFile);
    exit (1);
  }

  scanned = 0;
  while ( scanned != num ) {
    search = TRUE;
    for (i=0; i<file->numberOfFunctions(); i++) {
      f = file->functions(i)->symbol();

      if ( ! dumped[i] ) {
        if ( DumpFunc (f) ) {
          dumped[i] = TRUE;
          scanned++;
        }
      }
    }
  }

  checklines( file, (SgFunctionSymb*) mainSymb );

  closeFile ();

  delete [] dumped;

  if ( pf ) unlink(projFile);

  exit (foundErrors);
  return foundErrors; // to make the compiler happy
}
