/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include <stdio.h>
# include "regexp.h"
# include "instr.h"
# include "tdltemplate.h"
# include "edftemplate.h"

# define LISTCHUNK 100

extern int profileMode;

static int eventId = 0;
static int maxList = 0;

static struct nameelem
{
  char *str;
  int   cnt;
  int   tag;
}
*nameList = 0;

int newEventId (char *classname, char *funcname, int tag)
{
  char *name;
  int i;

  /* -- build full functionname including class -------------------- */
  if ( classname )
  {
    name = xalloc (strlen(classname)+strlen(funcname)+3, char);
    sprintf (name, "%s::%s", classname, funcname);
  }
  else
    name = funcname;

  /* -- function already used ? (C++ allows overloading names!)----- */
  /* -- search from the back, as same name was most probably ------- */
  /* -- specified last --------------------------------------------- */
  for (i=eventId-1; i>=0; i--)
    if (strcmp (name, nameList[i].str) == 0)
    {
      if ( nameList[i].cnt == 0 ) nameList[i].cnt++;
      break;
    }

  /* -- insert new funtion name ------------------------------------ */
  if ( eventId >= maxList )
  {
    maxList += LISTCHUNK;
    if ( nameList )
      nameList = xrealloc (nameList, maxList, struct nameelem);
    else
      nameList = xalloc (maxList, struct nameelem);
  }
  nameList[eventId].str = name;
  nameList[eventId].cnt = ( i<0 ? 0 : nameList[i].cnt+1 );
  nameList[eventId].tag = tag;

  eventId++;
  if ( profileMode )
    return (eventId + 10);
  else
    return ((eventId-1)*2);
}

void genTdl (char *tdlFile)
{
  int i;
  int little;
  char **ptr;
  FILE *out;
  union { long l; char c[4]; } convert;

  if ( (out = fopen (tdlFile, "w")) == NULL )
  {
    perror (tdlFile);
    exit (1);
  }

  /* -- test for little endian -- */
  convert.l = 0x00010203;
  for (i=0; i<4; i++) if ( convert.c[i] != 3-i ) break;
  little = ( i == 4 );

  for (ptr=tdlHeader[i=0]; ptr[0]; ptr=tdlHeader[++i])
    if ( ptr[little] == SAME )
      fprintf (out, "%s\n", ptr[0]);
    else if ( ptr[little] != EMPTY )
      fprintf (out, "%s\n", ptr[little]);

  if ( eventId > 0 )
    fprintf (out, "    VALUES ARE [ 0..%d, 60000..60006, ", eventId*2-1);
  else
    fprintf (out, "    VALUES ARE [ 60000..60006, ");
  fprintf (out, "60100..60109, 60199, 60200..60209,");
  fprintf (out, "60300..60309, 60399, 60400..60409,");
  fprintf (out, "61100..61122, 61150..61157, 61170, 61200..61213 ];\n");
  fprintf (out, "    INTERPRETATION\n");

  for (i=0; i<eventId; i++)
  {
    if ( nameList[i].cnt )
    {
      fprintf (out, "       %5d = '%s-%d-enter',\n", i*2, nameList[i].str,
              nameList[i].cnt);
      fprintf (out, "       %5d = '%s-%d-exit',\n", i*2+1, nameList[i].str,
              nameList[i].cnt);
    }
    else
    {
      fprintf (out, "       %5d = '%s-enter',\n", i*2, nameList[i].str);
      fprintf (out, "       %5d = '%s-exit',\n", i*2+1, nameList[i].str);
    }
  }

  for (ptr=tdlFooter1[i=0]; ptr[0]; ptr=tdlFooter1[++i])
    if ( ptr[little] == SAME )
      fprintf (out, "%s\n", ptr[0]);
    else if ( ptr[little] != EMPTY )
      fprintf (out, "%s\n", ptr[little]);

  if ( eventId > 0 )
  {
    for (i=0; i<eventId; i++)
    {
      if ( nameList[i].cnt )
        fprintf (out, "  '%s-%d-exit'", nameList[i].str, nameList[i].cnt);
      else
        fprintf (out, "  '%s-exit'", nameList[i].str);
      if ( i == (eventId-1) )
        fprintf (out, ":\n");
      else
        fprintf (out, ",\n");
    }
    fprintf (out, "    DATA:\n");
    fprintf (out, "      NAME IS SAGEID;\n");
    fprintf (out, "      FORMAT IS INTEGER*4;\n\n");
  }

  for (ptr=tdlFooter2[i=0]; ptr[0]; ptr=tdlFooter2[++i])
    if ( ptr[little] == SAME )
      fprintf (out, "%s\n", ptr[0]);
    else if ( ptr[little] != EMPTY )
      fprintf (out, "%s\n", ptr[little]);

  fclose (out);
}

void genEdf (char *edfFile)
{
  int i;
  char **ptr;
  FILE *out;

  if ( (out = fopen (edfFile, "w")) == NULL )
  {
    perror (edfFile);
    exit (1);
  }

  fprintf (out, "%d event types\n\n", eventId*2 + pCC_NUMEVENTS);
  fprintf (out,
           "#type state evtag eventname                  parametername\n\n");

  if ( eventId )
    fprintf (out,
             "# user code events ---------------------------------------\n");

  for (i=0; i<eventId; i++)
  {
    if ( nameList[i].cnt )
    {
      fprintf (out, "%5d USER  %5d %s-%d-enter\n", i*2, nameList[i].tag,
              nameList[i].str, nameList[i].cnt);
      fprintf (out, "%5d USER     -9 %s-%d-exit %*s\n", i*2+1, nameList[i].str,
              nameList[i].cnt, 32-strlen(nameList[i].str), "tag");
    }
    else
    {
      fprintf (out, "%5d USER  %5d %s-enter\n", i*2, nameList[i].tag,
              nameList[i].str);
      fprintf (out, "%5d USER     -9 %s-exit %*s\n", i*2+1, nameList[i].str,
              34-strlen(nameList[i].str), "tag");
    }
  }

  ptr = edfFooter; while ( *ptr ) fprintf (out, "%s\n", *ptr++);

  fclose (out);
}

void genFuncTable (char *tabFile, char *depfile, int *pcxx_tag)
{
  int i;
  FILE *out;

  if ( (out = fopen (tabFile, "w")) == NULL)
  {
    perror (tabFile);
    exit (1);
  }
  fprintf (out, "%s\n", depfile);
  fprintf (out, "%d func\n", eventId+11);

  fprintf (out, "%d\tProcessor_Main\n", pcxx_tag[0]);
  fprintf (out, "%d\tpcxx_Barrier\n",   pcxx_tag[1]);
  fprintf (out, "%d\tpcxx_Poll\n",      pcxx_tag[2]);
  fprintf (out, "%d\tWhereIsData\n",    pcxx_tag[3]);
  fprintf (out, "-1\t\n");
  fprintf (out, "-1\t\n");
  fprintf (out, "-1\t\n");
  fprintf (out, "-1\t\n");
  fprintf (out, "-1\t\n");
  fprintf (out, "-1\t\n");
  fprintf (out, "-1\t\n");

  for (i=0; i<eventId; i++)
  {
    if ( nameList[i].cnt )
      fprintf (out, "%d\t%s-%d\n", nameList[i].tag,
               nameList[i].str, nameList[i].cnt);
    else
      fprintf (out, "%d\t%s\n", nameList[i].tag, nameList[i].str);
  }
  fclose (out);
}
