/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# include "regexp.h"
# include "instr.h"

typedef struct descr
{
  REXP *file;
  REXP *class;
  REXP *func;
  struct descr *next;
}
DESCR;

LIMIT  excl_limit = { 0, 0, OP_OR };
LIMIT count_limit = { 0, 0, OP_OR };
int level_limit = 9999999;
int ignoreInlines = FALSE;

static char *REXP2Str (REXP *rexp)
{
  if ( ! rexp )
    return ("(null)");
  else
    return ( rexp->regstr );
}

static DESCR *base[COM_MAX];

void initCommands (void)
{
  int c;

  for (c=0; c<COM_MAX; c++) base[c] = (DESCR *) 0;
}

void addCommand (int command, REXP *file, REXP *class, REXP *func)
{
  DESCR *ptr;

  ptr = xalloc (1, DESCR);
  ptr->file  = file;
  ptr->class = class;
  ptr->func  = func;
  ptr->next = base[command];
  base[command] = ptr;
  if ( debugMode )
  {
    printf ("\tADD com=%s, file=%s, class=%s, func=%s\n", command ? "incl" : "excl",
            REXP2Str(file), REXP2Str(class), REXP2Str(func));
  }
}

static int isSingleMatch (REXP *rexp, char *name)
{
  if ( debugMode )
  {
    printf ("\tisMATCH rexp=%s, name=%s\n", REXP2Str(rexp), name);
  }

  if ( ! rexp )
    /* -- empty regExpr matches -- */
    return (TRUE);
  else if ( ! name )
    /* -- empty regExpr mame doesn't match -- */
    return (FALSE);
  else
  {
    if ( rexp->regexpr )
    {
      return ( regexec (rexp->regexpr, name) );
    }
    else
    {
      return ( strcmp (rexp->regstr, name) == 0 );
    }
  }
}

static int isMatch (DESCR *ptr, char *filename, char *classname, char *funcname)
{
  return ( isSingleMatch (ptr->file,  filename) &&
           isSingleMatch (ptr->class, classname) &&
           isSingleMatch (ptr->func,  funcname) );
}

int isValidFunc (int minlevel, char *filename, char *classname, char *funcname)
{
  int isNotInc;
  DESCR *ptr;

  /* -- check level first ----------------------------------------------- */
  if ( debugMode )
  {
    printf ("\tlevel %d >? %d\n", minlevel, level_limit);
  }
  if ( minlevel > level_limit ) return (FALSE);

  /* -- check includes next --------------------------------------------- */
  if ( (ptr = base[COM_INCLUDE]) )
  {
    isNotInc = TRUE;
    while ( ptr )
    {
      if ( isMatch (ptr, filename, classname, funcname) ) isNotInc = FALSE;
      ptr = ptr->next;
    }
    if ( isNotInc ) return (FALSE);
  }

  /* -- check excludes next --------------------------------------------- */
  ptr = base[COM_EXCLUDE];
  while ( ptr )
  {
    if ( isMatch (ptr, filename, classname, funcname) ) return (FALSE);
    ptr = ptr->next;
  }

  return (TRUE);
}

static int logic_table[OP_MAX][2][2] = {
  { /* or  */ { FALSE,  TRUE }, {  TRUE,  TRUE } },
  { /* and */ { FALSE, FALSE }, { FALSE,  TRUE } }
};

int isNotTooSmall (int numstmts, int numchild, int isInline)
{
  int es, ec, er, cs, cc, cr;

  if ( isInline && ignoreInlines )
  {
    if ( debugMode )
    {
      printf ("\tisNotTooSmall inline ignored\n");
    }
    return ( IS_TOO_SMALL );
  }

  es = (numstmts < excl_limit.stmts);
  ec = (numchild < excl_limit.child);
  er = ( logic_table[excl_limit.op][es][ec] );

  cs = (numstmts < count_limit.stmts);
  cc = (numchild < count_limit.child);
  cr = ( logic_table[count_limit.op][cs][cc] );

  if ( debugMode )
  {
    printf ("\tisNotTooSmall  excl s=%d, c=%d ==> %d\n", es, ec, !er);
    printf ("\tisNotTooSmall count s=%d, c=%d ==> %d\n", cs, cc, !cr);
  }

  if ( er )
    return ( IS_TOO_SMALL );
  else if ( cr )
    return ( COUNT_ONLY );
  else
    return ( IS_FINE );
}
