/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
   This file is a horrible hack!!
   It is needed because the depfiles of libSage++ don't store
   symbol position information.

   The functions here try to find pieces of code in Fortran, C, and
   C++ source files. However, as they are not implementing a full
   language parser, they can only handle 95% of the cases, but that
   is better than 0% :-)

   Bernd Mohr
*/

# include <stdio.h>
# include <stdlib.h>

#ifndef NeXT
# include <unistd.h>
#endif

# include <errno.h>
# include <fcntl.h>
# ifdef sequent
#   include <sys/types.h>
# endif
# include <sys/stat.h>

# ifndef TRUE
#   define TRUE 1
#   define FALSE 0
# endif

extern int isFortran;
extern int allLowercase;

static char *filebuf = 0;
static char *lastptr = 0;
static char *ptr = 0;
static int nolines = 0;
static char **lines = 0;
static char pathname[ 1024 ];

void openFile (char *path)
{
  int i;
  int infd;
  char *ptr;
  struct stat statbuf;

  if ( filebuf ) { free (filebuf); filebuf = 0; }
  if ( lines   ) { free (lines  ); lines   = 0; }

  if ( (infd = open (path, O_RDONLY)) == -1 ) {
    perror (path);
    exit (1);
  }
  if ( fstat (infd, &statbuf) == -1 ) {
    fprintf (stderr, "%s: cannot determine size\n", path);
    exit (1);
  }

  if ( (filebuf = malloc (statbuf.st_size)) == NULL ) {
    fprintf (stderr, "%s: cannot get enough memory\n", path);
    exit (1);
  }

  if ( read (infd, (void *) filebuf, statbuf.st_size) != statbuf.st_size ) {
    fprintf (stderr, "%s: cannot read\n", path);
    exit (1);
  }
  close (infd);
  strcpy( pathname, path );

  nolines = 0;
  for (i=0; i<statbuf.st_size; i++) {
    if ( filebuf[i] == '\n' ) {
      filebuf[i] = '\0';
      nolines++;
    }
  }

  if ( (lines = malloc (nolines * sizeof(char *))) == NULL ) {
    fprintf (stderr, "%s: cannot setup line buffer\n", path);
    exit (1);
  }

  nolines = 0;
  ptr     = filebuf;
  for (i=0; i<statbuf.st_size; i++) {
    if ( filebuf[i] == '\0' ) {
      lines[nolines++] = ptr;
      ptr = filebuf+i+1;
    }
  }

  ptr     = filebuf;
  lastptr = filebuf + statbuf.st_size;
}

void closeFile ()
{
  free (lines);
  free (filebuf);
}

static int line = 1;
static int pos = -1;
static int lastchar = -1;
static int lastfoo = -1;

static int readch ()
{
  int ch;

  lastchar = lastfoo;
  ch = *ptr++;
  if ( ptr == lastptr ) return (lastfoo=EOF);

  if (ch == '\0') {
    ch = '\n';
    line++;
    pos = -1;
  }
  else
    pos++;

  return (lastfoo=ch);
}

static int nextch ()
{
  int ch;

  ch = *ptr;
  if (ch == '\0') ch = '\n';
  return (ch);
}

static void gotopos (int l, int p)
{
  int i;

  /* Not all instrumenters insert the right line numbers... */
  if (l >= nolines) {
    fprintf( stderr,
      "Error: gotopos( %d, %d ). Current file is %s. No. of lines is %d.\n",
      l, p, pathname, nolines );
    exit( 1 );
  }

  line = l;
  pos  = -1;
  ptr  = lines[l-1];
  for (i=0; i<p; i++) readch();
}

/* used by checkpos.C */
char *getLine (int l)
{
  if (l > 0 && l < nolines+1)
    return lines[l-1];
  else
    return 0;
}

static char resbuf[80];

static int body_end_line = -1;
static int look_next_line = -1;
static int look_next_pos  = -1;

void setNextPos (int line, int pos, int endline)
{
  look_next_line = line;
  look_next_pos  = pos;
  body_end_line  = endline;
}

#define X 0x00
#define L 0x01   /* Letter */
#define D 0x02   /* Digit */
#define U 0x04   /* Underscore */
#define W 0x08   /* WhiteSpace */
#define F 0x10   /* Fortran Comment */
#define S 0x20   /* Special: char after C++/C func call */

#define I 0x07   /* Ident: L+D+U */
#define C 0x11   /* Comment letter: L+F */

static char validch[] = {           /* valid identifier characters   */
/* ^@  ^A  ^B  ^C  ^D  ^E  ^F  ^G  ^H  \t  \n  ^K  \f  \r  ^N  ^O    */
    X,  X,  X,  X,  X,  X,  X,  X,  X,  W,  W,  X,  W,  W,  X,  X,

/* ^P  ^Q  ^R  ^S  ^T  ^U  ^V  ^W  ^X  ^Y  ^Z  ^[  ^\  ^]  ^^  ^_    */
    X,  X,  X,  X,  X,  X,  X,  X,  X,  X,  X,  X,  X,  X,  X,  X,

/* sp   !   "   #   $   %   &   '   (   )   *   +   ,   -   .   /    */
    W,  F,  X,  X,  F,  X,  X,  X,  S,  S,  F,  X,  X,  X,  X,  X,

/*  0   1   2   3   4   5   6   7   8   9   :   ;   <   =   >   ?    */
    D,  D,  D,  D,  D,  D,  D,  D,  D,  D,  X,  S,  S,  X,  X,  X,

/*  @   A   B   C   D   E   F   G   H   I   J   K   L   M   N   O    */
    X,  L,  L,  C,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,

/*  P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]   ^   _    */
    L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  S,  X,  X,  X,  U,

/*  `   a   b   c   d   e   f   g   h   i   j   k   l   m   n   o    */
    X,  L,  L,  C,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,

/*  p   q   r   s   t   u   v   w   x   y   z   {   |   }   ~        */
    L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  L,  S,  X,  X,  X
};

static int IsIdent  (int ch) { return ((ch <= '~') && (validch[ch] & I)); }
static int IsWhite  (int ch) { return ((ch <= '~') && (validch[ch] & W)); }
static int IsLetter (int ch) { return ((ch <= '~') && (validch[ch] & L)); }
static int IsFComm  (int ch) { return ((ch <= '~') && (validch[ch] & F)); }
static int IsSpecial(int ch) { return ((ch <= '~') && (validch[ch] & S)); }

#undef X
#undef L
#undef D
#undef U
#undef W
#undef F
#undef I
#undef C

void findFuncOrClass (int start, char **result)
{
  int ch;
  int braces = 1;
  int last = 0;
  int header_start_line = -1;
  int header_start_pos  = -1;
  int header_end_line = -1;
  int header_end_pos  = -1;
  int hline = -1;
  int hpos  = -1;
  int body_start_line = -1;
  int body_start_pos  = -1;

  gotopos (start, 0);

  while ( (ch = readch()) != EOF ) {
    if ( header_start_line == -1 ) {

      /* -- look for start of function or class header ------------------- */
      if ( IsLetter(ch) ) {
        header_start_line = line;
        header_start_pos  = pos;
      }
    }
    else if ( header_end_line == -1 ) {

      /* -- look for end of function or class header --------------------- */
      if ( (ch != '{') && ! IsWhite(ch) ) {
        hline = line;
        hpos  = pos;
      }
      else if ( ch == '{' ) {
        header_end_line = hline;
        header_end_pos  = hpos;
        body_start_line = line;
        body_start_pos  = pos;

        /* -- for findNextChild -- */
        look_next_line  = line;
        look_next_pos   = pos;
      }
    }
    else {

      /* -- look for end of function or class body ----------------------- */
      if ( ch == '/' ) {
        if ( (ch = readch()) == '/' ) {
          while ( (ch = readch()) != '\n' ) /*skip*/ ;
        }
        else if ( ch == '*' ) {
          last = 0;
          do {
            if ( ((ch = readch()) == '/') && (last == '*') )
              break;
            last = ch;
          }
          while ( ch != EOF );
        }
      }

      if ( ch == '\'' ) {
        do {
          if ( ((ch = readch()) == '\\') && ((ch = readch()) == '\'') )
            ch = readch();
        }
        while ( ch != '\'' );
      }
      else if ( ch == '\"' ) {
        do {
          if ( (ch = readch()) == '\\' ) {
            if ( (ch = readch()) == '\"' )
              ch = readch();
          }
        }
        while ( ch != '\"' );
      }
      else if ( ch == '{' ) {
        braces++;
      }
      else if ( ch == '}' ) {
        braces--;
        if ( braces == 0 ) {
          sprintf (resbuf, "@ %d.%d %d.%d %d.%d %d.%d",
                  header_start_line, header_start_pos,
                  header_end_line,   header_end_pos + 1,
                  body_start_line,   body_start_pos,
                  line, pos + 1);
          *result = resbuf;

          /* -- for findNextChild -- */
          body_end_line = line;
          return;
        }
      }
    }
  }
  sprintf (resbuf, "@ - - - -");
  *result = resbuf;
}

static void findNext (char *child, char **result, int marknewpos,
                      int start_line, int start_pos)
{
  int ch;
  int last = 0;
  int child_start_line = -1;
  int child_start_pos  = -1;
  char *pattern;

  if ( IsLetter(*child) ) {
    gotopos (start_line, start_pos);
    pattern = child;
    while ( ((ch = readch()) != EOF) && (line <= body_end_line) ) {
      if ( isFortran ) {
        /* -- skip fortran comments and compiler directives -- */
        if ( IsFComm(ch) && (pos == 0) ) {
            while ( (ch = readch()) != '\n' ) /*skip*/;
        }

        /* -- fortran is case-insensitive! -- */
        if ( allLowercase && (ch >= 'A') && (ch <= 'Z') ) {
            ch = ch - 'A' + 'a';
        }
        if ( !allLowercase && (ch >= 'a') && (ch <= 'z') ) {
            ch = ch - 'a' + 'A';
        }

        /* -- spaces are ignored in fortran -- */
        if ( ch == ' ' ) continue;

        /* -- F90 comments -- */
        if ( ch == '!' ) {
            while ( (ch = readch()) != '\n' ) /*skip*/;
        }
      }
      else {
        /* -- skip C and C++ comments -- */
        if ( ch == '/' ) {
          if ( (ch = readch()) == '/' ) {
            while ( (ch = readch()) != '\n' ) /*skip*/ ;
          }
          else if ( ch == '*' ) {
            last = 0;
            do {
              if ( ((ch = readch()) == '/') && (last == '*') )
              break;
              last = ch;
            }
            while ( ch != EOF );
          }
        }
      }

      /* -- skip strings -- */
      if ( ch == '\'' ) {
        do {
          if (((ch = readch()) == '\\') && ((ch = readch()) == '\''))
            ch = readch();
        }
        while ( ch != '\'' );
      }
      else if ( ch == '\"' ) {
        do {
          if ( (ch = readch()) == '\\' ) {
            if ( (ch = readch()) == '\"' )
              ch = readch();
          }
        }
        while ( ch != '\"' );
      }
      /* -- otherwise try matching the child name -- */
      else if ( ch == *pattern ) {
        if ( child_start_line == -1) {
          /* -- check if this is really the start of a new ident -- */
          if ( IsLetter(lastchar) ) {
            /* -- nice try -- */
            continue;
          }
          child_start_line = line;
          child_start_pos  = pos;
        }
        pattern++;
        if ( *pattern == '\0' ) {
          /* -- check if end of identifier -- */
          ch = nextch();
          if ( IsIdent(ch) ) {
            /* -- almost but not quite -- */
            pattern = child;
            child_start_line = -1;
          }
          else {
            sprintf (resbuf, "%d.%d %d.%d",
                     child_start_line, child_start_pos,
                     line, pos + 1);
            *result = resbuf;

            if ( !isFortran ) {
              /* heuristic: check for valid next character */
              while ( IsWhite(ch) ) {
                readch(); ch = nextch();
              }
              if ( ! IsSpecial(ch) && ! IsLetter(ch) ) {
                /* -- almost but not quite -- */
                pattern = child;
                child_start_line = -1;
                continue;
              }
            }
            if ( marknewpos ) {
              look_next_line = line;
              look_next_pos  = pos + 1;
            }
            return;
          }
        }
      }
      else {
        pattern = child;
        child_start_line = -1;
      }
    }
  }

  /* -- not found or non-alphabetic name -- */
  sprintf (resbuf, "- -");
  *result = resbuf;
}

void findNextChild (char *child, char **result, int marknewpos)
{
  findNext (child, result, marknewpos, look_next_line, look_next_pos);
}

void findNextMember (char *name, int line, char **result)
{
  static char nbuf[12];

  if ( IsLetter(*name) )
    findNext (name, result, FALSE, line, 0);
  else {
    sprintf (nbuf, "operator%s", name);
    findNext (nbuf, result, FALSE, line, 0);
  }
}

# ifdef EBUG

int main (int argc, char **argv)
{
  int i;
  char *r;

  openFile (argv[1]);
  findFuncOrClass (atoi(argv[2]), &r);
  printf ("%s: %s\n", argv[1], r);
  for (i=3; i<argc; i++) {
    findNextChild (argv[i], &r, TRUE);
    printf ("%s: %s\n", argv[i], r);
  }
  closeFile ();
  exit (0);
}
# endif
