; c m d - d i s p l a y . s c h
; $Id: cmd-display.sch,v 1.1 1996/05/01 20:07:50 lth Exp $
; 
; Command window for Sneaky
; kurtw@cs.uoregon.edu / Feb 28, 1996


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Globals

(define cmdwin-mutex   #f)           ;mutex for the below variables
(define cmd-windows #f)              ;list of command windows
(define num-cmd-wins #f)             ;list of command windows


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Command Windows Bookkeeping

(define (init-cmdwins) 
  (set! cmdwin-mutex (make-mutex))
  (set! cmd-windows '())
  (set! num-cmd-wins 0))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Command Window Class

(define-class <CommandWin> ()
  ((mutex :initform (make-mutex) 
	  :getter get-mutex)
   (toplevel :getter get-toplevel)
   (text :getter get-text)
   (tthreads :initform (make-tthread-set) 
	     :getter get-tthreads)
   (tthread-str :initform "" 
		:getter get-tthread-str 
		:setter set-tthread-str!)
   (tthread-error :initform #f
		  :getter get-tthread-error 
		  :setter set-tthread-error!)))


; Could be overloaded for other windows...
(define-generic win-init)
(define-generic cmd.return)
(define-generic cmd.erase)
(define-generic cmd.backspace)
(define-generic cmd.msg)
(define-generic cmd.get-command-line)
(define-generic cmd.display-prompt)
(define-generic cmd.win-title)


(define (foo w)
  (cmd.return w))

;cmdwin-mutex:notheld
;FIX - bind all the cursor control keys to procs that do the right thing
;      (ie, command line editing, history, etc)
(define-method win-init((w <CommandWin>))
  (call-with-mutex cmdwin-mutex
      (lambda ()
	(slot-set! w 'toplevel (make <Toplevel> :title "Thread(s): all"))
	(slot-set! w 'text (make <scroll-text> :parent (get-toplevel w)))
	(pack (get-text w))
	(focus (slot-ref (get-text w) 'text))

	(let ((textw (slot-ref (get-text w) 'text)))
	  (bind textw "<Any-Button-1>" `(focus ,(slot-ref (get-text w) 'text)))
	  (bind textw "<Any-Button1-ButtonPress>" '())
	  (bind textw "<Any-Button1-ButtonRelease>" '())
	  (bind textw "<Any-Button1-Motion>" '())
	  (bind textw "<Key-Up>" '())
	  (bind textw "<Key-Down>" '())
	  (bind textw "<Key-Left>" '())
	  (bind textw "<Key-Right>" '())
	  (bind textw "<Control-a>" '())
	  (bind textw "<Control-e>" '())
	  (bind textw "<Control-b>" '())
	  (bind textw "<Control-n>" '())
	  (bind textw "<Control-f>" '())
	  (bind textw "<Control-p>" '())
	  (bind textw "<Control-u>" `(cmd.erase ,(address-of w)))
	  (bind textw "<Key-BackSpace>" `(cmd.backspace ,(address-of w)))
	  (bind textw "<Key-Delete>" `(cmd.backspace ,(address-of w)))
	  (bind textw "<Return>" `(cmd.return ,(address-of w))))

	(set! num-cmd-wins (+ num-cmd-wins 1))
	(set! cmd-windows (append cmd-windows (list w)))
	(cmd.msg w "Sneaky command line debugger~%")
	(cmd.msg w "Enter 'help' for command summary.~%"))))

;w.mutex:notheld
(define-method cmd.msg (w . rest)
  (call-with-mutex (get-mutex w)
    (lambda ()
	(text-insert (get-text w) 'end (apply format #f rest))
	(let ((line (car (string-split (text-index (get-text w) 'end) #\.)))
	      (tkwidget (slot-ref (slot-ref (get-text w) 'text) 'id)))
	  (tkwidget 'yview `-pickplace line)))))
		   
;w.mutex:notheld
(define-method cmd.erase (w)
  (call-with-mutex (get-mutex w)
    (lambda ()
      (text-delete (get-text w) "insert linestart + 9 c" "insert lineend"))))

;w.mutex:notheld
(define-method cmd.backspace (w)
  (call-with-mutex (get-mutex w)
    (lambda ()
      (let ((ch (string->number 
		 (cadr (string-split 
			(text-index (get-text w) "insert") #\.)))))
	(if (> ch 9)
	    (text-delete (get-text w) "insert - 1 c" "insert"))))))

;w.mutex:notheld
(define-method cmd.return (w)
  (parse-command w (cmd.get-command-line w))
  (cmd.display-prompt w))

;w.mutex:notheld
(define-method cmd.get-command-line (w)
  (call-with-mutex (get-mutex w)
    (lambda ()
      (text-get (get-text w) "insert linestart" "insert lineend"))))
    
;w.mutex:notheld
(define-method cmd.win-title (w str)
  (call-with-mutex (get-mutex w)
    (lambda ()
      (wm 'title (get-toplevel w) str))))
  
;w.mutex:notheld
(define-method cmd.display-prompt (w)
  (cmd.msg w "(sneaky) "))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Command display

(define (create-command-display)
  (let ((win (make <CommandWin>)))
    (win-init win)
    (cmd.display-prompt win)
    ((slot-ref (slot-ref (get-text win) 'text) 'id) 'yview 0)
    win))



(provide "command-display.sch")

; eof
