; c m d - p r o c s . s c h
; $Id: cmd-procs.sch,v 1.2 1996/06/28 21:34:09 lth Exp $
; 
; Command implementations for Sneaky
; kurtw@cs.uoregon.edu / Feb 29, 1996
; FIX - check that threads are responsive before issuing certain commands
;       (continue, handle)

(require "ct-info.sch")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Command Implementations


;FIX - This should terminate the debuggee...
(define (cmd:exit w args num-args)
  (exit))


(define (cmd:help w args num-args)
  (cmd.msg w "The following commands are available:~%")
  (cmd.msg w "  continue                      - continue each selected thread~%")
  (cmd.msg w "  exit                          - exit sneaky~%")
  (cmd.msg w "  handle                        - list installed event handlers for~%")
  (cmd.msg w "                                  selected threads~%")
  (cmd.msg w "  handle [event-handler] ...    - install event handlers for selected~%")
  (cmd.msg w "                                  threads~%")
  (cmd.msg w "  help - display this text~%")
  (cmd.msg w "  info collections              - list registered collections for~%")
  (cmd.msg w "                                  selected threads~%")
  (cmd.msg w "  info events                   - list available event handlers~%")
  (cmd.msg w "  info threads                  - list running status of selected threads~%")
  (cmd.msg w "  inspect coll index ...        - inspect a collection at the given indices.~%")
  (cmd.msg w "  quit                          - exit sneaky~%")
  (cmd.msg w "  source context:thread         - display source in fancy~%")
  (cmd.msg w "  target [[host] port]          - connect to a debuggee~%")
  (cmd.msg w "  thread                        - list selected threads~%")
  (cmd.msg w "  thread [context]:[thread] ... - select a list of threads~%")
  (cmd.msg w "  unhandle [event-handler] ...  - uninstall event handler for selected~%")
  (cmd.msg w "  where context:thread          - display stack backtrace~%")
  (cmd.msg w "                                  threads.~%")
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: info threads
(define (cmd:info w args num-args)

  (define (info.threads)
    (cmd.msg w "Status of selected threads:~%")
    (for-each-selected-tthread
     (lambda (cid tid)
       (cmd.msg w "  ~a:~a - ~a~%" cid tid (symbol->string 
					  (ct-table-ref ct-state cid tid))))
     (get-tthreads w)))

  (define (info.collections)
    (cmd.msg w "Available collections:~%")
    (for-each (lambda (c)
		(cmd.msg w "  ~a ~a (number ~a), defined in ~a~%"
			 (typename c)
			 (tag c)
			 (collection-number c)
			 (scope c)))
	      (all-collections)))

  (define (info.events)
    (cmd.msg w "Available event handlers:~%")
    (map (lambda (id) 
	   (cmd.msg w "  ~a~%" (sneezy.event-name id)))
       (list sneezy.etype:begin-barrier
             sneezy.etype:end-barrier
             sneezy.etype:begin-create-collection
             sneezy.etype:end-create-collection
             sneezy.etype:begin-delete-collection
             sneezy.etype:end-delete-collection
             sneezy.etype:begin-parallel
             sneezy.etype:end-parallel
             sneezy.etype:enter-element
             sneezy.etype:exit-element
             sneezy.etype:user-event)))

  (define sub-commands
    (list 
     (cons "c" info.collections)
     (cons "e" info.events)
     (cons "t" info.threads)
     ))

  (if (= num-args 1)
      (let ((sub-closure (lookup-keyword (car args) sub-commands)))
	(if sub-closure 
	    (sub-closure)
	    (cmd.msg w "Invalid info command: ~a~%" (car args))))
      (cmd.msg w "Usage: info threads~%")))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: inspect collection index0 ...

(define (cmd:inspect w args num-args)

  (define (check-collection name coll indices)
    (if (not coll)
	(begin (cmd.msg w "Not a known collection: ~a~%" name)
	       #f)
	(let ((dims (dimension-list coll)))
	  (if (and (= (length indices) (length dims))
		   (every? (lambda (x y)
			     (and (number? x)
				  (exact? x)
				  (integer? x)
				  (>= x 0)
				  (< x y)))
			   indices
			   dims))
	      coll
	      (begin (cmd.msg w "Invalid indices for collection ~a: ~a~%" 
			      name indices)
		     (cmd.msg w "Collection indices are: ~a~%" dims)
		     #f)))))

  (define (find-collection cname indices)
    (cond ((string? cname)
	   (check-collection cname (find-collection-by-tag cname) indices))
	  ((number? cname)
	   (check-collection cname (find-collection-by-number cname) indices))
	  (else
	   (cmd.msg w "Not a valid collection expression: ~a~%" cname)
	   #f)))

  (cond ((< num-args 2)
	 (cmd.msg w "Collection or index not specified.~%"))
	((> num-args 3)
	 (cmd.msg w "Too many indices (max is 2).~%"))
	(else
	 (let* ((indices (map string->number (cdr args)))
		(coll    (find-collection (car args) indices)))
	   (if (not coll)
	       #f
	       (let ((elt  (apply inspect-collection-element coll indices))
		     (type (element-type coll)))
		 (cmd.msg w "~a {~%" (car type))
		 ; FIXME: It should really do substructures correctly,
		 ; too.
		 (for-each (lambda (field)
			     (cmd.msg w "~a ~a~%" (car field) (cadr field)))
			   elt)
		 (cmd.msg w "}~%")))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: target [[host] port]
(define (cmd:target w args num-args)
  (let ((port (sneezy.debugger-port))
        (host "localhost"))
    (if (not connected?)
	(cond ((and (= num-args 2)
		    (integer? (string->number (cadr args)))
		    (exact? (string->number (cadr args)))
		    (> (string->number (cadr args)) 0))
	       (set! host (car args))
	       (set! port (string->number (cadr args)))
	       (run-sneaky-cmd host port #t))
	      ((and (= num-args 1)
		    (integer? (string->number(car args)))
		    (exact? (string->number (car args)))
		    (> (string->number (car args)) 0))
	       (set! port (string->number (car args)))
	       (run-sneaky-cmd host port #t))
	      ((= num-args 0) 
	       (run-sneaky-cmd host port #t))
	      (else
	       (cmd.msg w "Usage: target [[host] port]~%")))
	(cmd.msg w "Already connected to a debuggee!~%"))))

    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: continue [n]
; FIX so [n] works
; FIX - check state of thread (running or stopped)
(define (cmd:continue w args num-args)

  (define (cont.cont-tthread cid tid)
    (if (eq? (ct-table-ref ct-state cid tid) 'ready)
	(begin 
	  (thread-is-continuing cid tid)
	  (stepping.continue cid tid))
	(cmd.msg w "Couldn't continue ~a:~a - Thread not responsive.~%"
		 cid tid)))


  ;Body of cmd:continue
  (cond ((= num-args 0) 
	 (eventmsg "CONTINUING THREADS: ~a~%" (get-tthread-str w))
	 (for-each-selected-tthread cont.cont-tthread 
				    (get-tthreads w)))
	(else 
           (cmd.msg w "Usage: continue [n]~%"))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: thread                     - lists selected threads
;         thread [cid:tid ...]       - replace selected threads with listed
(define (cmd:thread w args num-args)

  (define (thread.add-tthread tt-str)
    (let ((ct-pair (thread.split-tthread-id tt-str)))
      (if ct-pair
	  (if (and  
	       (memv (car ct-pair) (client.data 'all-contexts '()))
	       (< (cdr ct-pair) (client.data 'threads-in-context 
					     (car ct-pair))))
	      (select-tthread! (get-tthreads w) (car ct-pair) (cdr ct-pair))
	      (begin 
		(cmd.msg w 
			 "Thread ~a does not exist in the debuggee.~%" tt-str)
		(set-tthread-error! w #t)))
	  (begin 
	    (cmd.msg w "Invalid thread identifier ~a - must be [context-id]:[thread-id]~%"
		     tt-str)
	    (set-tthread-error! w #t)))))
  
  ;FIX parse '*' as context or thread
  (define (thread.split-tthread-id tt-str)
    (let ((len (string-length tt-str)))
      (define (find-colon)
	(let ((index 
	       (do ((i 0 (+ i 1)))
		   ((or (char=? (string-ref tt-str i) #\:) (= i len)) i))))
	  (if (= index len) #f index)))
      (let* ((colon-index (find-colon))
	     (cid (string->number (substring tt-str 0 colon-index)))
	     (tid (string->number (substring tt-str (+ colon-index 1) len))))
	(if (and (number? cid) (number? tid)
		 (integer? cid) (integer? tid)
		 (not (negative? cid)) (not (negative? tid)))
	    (cons cid tid)
	    #f))))
  

  ;Body of cmd:thread
  (if connected?
      (let ((old-tthreads (tthread-get (get-tthreads w))))
	(cond ((= num-args 0) 
	       (cmd.msg w "Selected theads:~%")
	       (cmd.msg w "  ~a~%" (get-tthread-str w)))
	      ((>= num-args 1) 
	       (remove-all-tthreads! (get-tthreads w))
	       (for-each thread.add-tthread args)
	       (if (get-tthread-error w) 
		   (tthread-set! (get-tthreads w) old-tthreads))
	       (set-tthread-error! w #f)
	       (set-tthread-str! w (format-selected-tthreads (get-tthreads w)))
	       (cmd.win-title w
			      (format #f "Thread(s) ~a" 
				      (get-tthread-str w))))))
      (cmd.msg w "Not connected - use 'target' to connect.~%")))
  

			 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: handle [event ...]
; FIX so that 0 or multiple events can be installed (0 lists installed evts)
; FIX: it's now undefined what happens in the GUI if you install on one
;      thead but not another.
(define (cmd:handle w args num-args)

  (define (handle.display-events)
    (for-each-selected-tthread 
     (lambda (cid tid)
       (cmd.msg w "Installed events for thread ~a:~a~%" cid tid)
       (for-each 
	(lambda (etype)
	  (cmd.msg w "  ~a~%" (sneezy.event-name etype)))
	(ct-table-ref ct-handlers cid tid)))
     (get-tthreads w)))
     
  (define (handle.handle-events estring)
    (let ((event (lookup-keyword estring handler-events)))
      (if event 
	  (for-each-selected-tthread
	   (lambda (cid tid)
	     (if (memv event (ct-table-ref ct-handlers cid tid))
		 (cmd.msg w "Event (~a) already installed for thread ~a:~a~%"
			 (sneezy.event-name event) cid tid)
		 (begin
		   (if (not (= (sneezy.event-type 'user-event) event))
		       (client.handle event cid tid general-event-handler))
		   (ct-table-set! ct-handlers cid tid
				  (append (ct-table-ref ct-handlers cid tid)
					  (list event)))
		   (slot-set! (list-ref event-cb (lookup-cb event))
			      'value (number->string event))
		   (cmd.msg w "Installed (~a) on thread ~a:~a~%" 
			   (sneezy.event-name event) cid tid))))
	   (get-tthreads w))
	  (cmd.msg w "Invalid event: ~a~%" estring))))
  
  (define (general-event-handler etype cid tid edata)
    (thread-has-stopped cid tid)
    (cmd.msg w "Event Reached: ~a ~a ~a ~a~%" (sneezy.event-name etype)
	    cid tid edata)
    (eventmsg "Event Reached: ~a ~a ~a ~a~%" (sneezy.event-name etype)
	    cid tid edata)
    (ct-table-set! ct-state cid tid 'ready)
    (state.set-state! cid tid (sneezy.event-symbol etype))
    (cmd.display-prompt w))
  

  ;Body of cmd:handle
  (cond ((= num-args 0)  (handle.display-events))
	((>= num-args 1) (for-each handle.handle-events args))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Syntax: unhandle [event ...]
; FIX so that 0 or multiple events can be installed (0 lists installed evts)
(define (cmd:unhandle w args num-args)

  (define (unhandle.unhandle-events estring)
    (let ((event (lookup-keyword estring handler-events)))
      (if event 
	  (for-each-selected-tthread
	   (lambda (cid tid)
	     (if (memv event (ct-table-ref ct-handlers cid tid))
		 (begin
		   (if (not (= (sneezy.event-type 'user-event) event))
		       (client.unhandle event cid tid))
		   (ct-table-set! ct-handlers cid tid
				  (remove event 
					  (ct-table-ref ct-handlers cid tid)))
		   (slot-set! (list-ref event-cb (lookup-cb event))
			      'value "")
		   (cmd.msg w "Uninstalled (~a) from thread ~a:~a~%" 
			    (sneezy.event-name event) cid tid))
		 (cmd.msg w "Event (~a) not installed for thread ~a:~a~%"
			 (sneezy.event-name event) cid tid)))
	   (get-tthreads w))
	  (cmd.msg w "Invalid event: ~a~%" estring))))
  
  ;Body of cmd:unhandle
  (cond ((= num-args 0)  (cmd:handle w args num-args))
	((>= num-args 1) (for-each unhandle.unhandle-events args))))

; source ctx:thread 
(define (cmd:source w args num-args)
  (ctx/thr-arg w args num-args 
	       (lambda (ctx thr)
		 (let ((stack-trace (get-thread-stack ctx thr)))
		   (if (null? stack-trace)
		       (cmd.msg w "ERROR: Stack trace is empty!~%")
		       (TAU:global-select-function (car stack-trace)))))))

; where ctx:thread
(define (cmd:where w args num-args)
  (ctx/thr-arg w args num-args 
	       (lambda (ctx thr)
		 (cmd.msg w "Stack backtrace:~%")
		 (for-each (lambda (tag)
			     (cmd.msg w "~a~%" (tau:get-tag-name tag)))
			   (get-thread-stack ctx thr)))))

(define (ctx/thr-arg w args num-args k)
  (if (not (= num-args 1))
      (cmd.msg w "Wrong number of arguments: ~a~%" args)
      (let ((s (string-split (car args) #\:)))
	(if (or (not (= (length s) 2))
		(not (every? (lambda (p)
			       (let ((n (string->number p)))
				 (and (number? n)
				      (exact? n)
				      (integer? n)
				      (>= n 0))))
			     s)))
	    (cmd.msg w "Bad argument: ~a~%" (car args))
	    (let ((ctx (string->number (car s)))
		  (thr (string->number (cadr s))))
	      ; HACK HACK HACK FIXME!
	      (format #t "ctx=~a thr=~a~%" ctx thr)
	      (if (or (>= ctx (client.data 'server-context-count))
		      (> thr 0))
		  (cmd.msg w "Argument out of range: ~a~%" (car args))
		  (k ctx thr)))))))

(provide "cmd-procs.sch")

;eof
