; c o m m a n d s . s c h
; $Id: commands.sch,v 1.1 1996/05/01 20:07:52 lth Exp $
; 
; Command parsing for Sneaky
; kurtw@cs.uoregon.edu / Feb 29, 1996

(require "cmd-procs.sch")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Commands

;alist of command string-lambda pairs
(define commands
       (list 
	(cons "c" cmd:continue)
	(cons "e" cmd:exit)
	(cons "ha" cmd:handle)
	(cons "he" cmd:help)
	(cons "inf" cmd:info)
	(cons "ins" cmd:inspect)
	(cons "q" cmd:exit)
	(cons "s" cmd:source)
	(cons "ta" cmd:target)
	(cons "th" cmd:thread)
	(cons "u" cmd:unhandle)
	(cons "w" cmd:where)
	))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Events

;alist of event string-number pairs
(define handler-events
       (list (cons "begin-b" sneezy.etype:begin-barrier)
             (cons "end-b" sneezy.etype:end-barrier)
             (cons "begin-c" sneezy.etype:begin-create-collection)
             (cons "end-c" sneezy.etype:end-create-collection)
             (cons "begin-d" sneezy.etype:begin-delete-collection)
             (cons "end-d" sneezy.etype:end-delete-collection)
             (cons "begin-p" sneezy.etype:begin-parallel)
             (cons "end-p" sneezy.etype:end-parallel)
             (cons "ent" sneezy.etype:enter-element)
             (cons "ex" sneezy.etype:exit-element)
             (cons "u" sneezy.etype:user-event)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Command Parsing

(define (parse-command w cmd-str)
  (cmd.msg w "~%")
  (let ((cmd-list (string-split cmd-str #\space)))
    (if (not (null? (cdr cmd-list)))
	(let* ((cmd-name (cadr cmd-list))
	       (arg-list (cddr cmd-list))
	       (cmd-closure (lookup-keyword cmd-name commands)))
	  (if cmd-closure 
	      (cmd-closure w arg-list (length arg-list))
	      (cmd.msg w "Invalid Command: ~a~%" cmd-name))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Utiliites

(define (lookup-keyword str table) 
  (define (iter l) 
    (if (null? l) #f
	(if (string-start? (caar l) str) (cdar l) (iter (cdr l)))))
  (iter table))

;Splits a string into a list of space separated words
(define (string-split str sep)
  (let ((len (string-length str)))

    (define (string-split-iter index)
      ;skip leading separators when computing first
      (let ((first (do ((next index (+ next 1)))
		       ((or (= next len)
			    (not (char=? (string-ref str next) sep))) 
			next))))
	(if (= first len) 
	    '()

	    ; first is the first char of a word
	    (do ((last first (+ last 1)))
		((or (= last len)
		     (char=? (string-ref str last) sep))
		 (append (list (substring str first last)) 
			 (string-split-iter last)))))))
    
    (append '() (string-split-iter 0))))


;Bool: is abbrev an initial substring of str
(define (string-start? abbrev str)
  (let ((abbrev-len (string-length abbrev))
	(str-len (string-length str)))
    (if (<= abbrev-len str-len)
	(string=? abbrev (substring str 0 abbrev-len))
	#f)))
	 

; works for numbers, strings, or symbols
; The result if returned.
; NOTE that the list passed may not be side effected 
; (if element removed is first)
(define (list-remove l item) 
  (let ((eq-test (cond ((number? item) =)
		       ((string? item) string=?)
		       ((symbol? item) eq?))))
    (cond ((null? l) #f)
	  ((apply eq-test (list item (car l))) l)
	  (else
	   (let ((ptr (do ((p l (cdr p))
			   (q (cdr l) (cdr q)))
                          ((or (null? q) 
			       (apply eq-test (list item (car q)))) p))))
	     
	     (if (null? (cdr ptr))
		 #f
		 (begin (set-cdr! ptr (cddr ptr))
			l)))))))
