; s n e a k y - b a s i c . s c h
; $Id: sneaky-basic.sch,v 1.2 1996/06/28 21:34:10 lth Exp $
;
; Command line debugger interface for Sneaky - commands based on gdb.
; To use, load all sneaky files except sneaky-basic.
;
; May 7, 1996
;
; kurtw@cs.uoregon.edu / Feb 28, 1996
; lth@cs.uoregon.edu / March 18, 1996

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Globals

(define connected? #f)

(define cmd-mutex          #f)         ; controls access to these variables
(define bvec               #f)         ; for count stepping
(define interesting-events '())        ; for event stepping
(define collections        '())        ; assoc list (id . name) of live colls.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Startup code

;cmd-mutex:irrelevent
(define (sneaky . rest)
;  (wm 'withdraw *root*)
;  (init-gui)
  (if (not (tasking-enabled?))
      (begin (display "ENABLING TASKING.")
	     (newline)
	     (begin-tasking)))
  (let ((l    (length rest))
	(port (sneezy.debugger-port))
	(host "localhost"))
    (cond ((and (= l 2)
		(string? (car rest))
		(integer? (cadr rest))
		(exact? (cadr rest))
		(> (cadr rest) 0))
	   (set! host (car rest))
	   (set! port (cadr rest))
	   (run-sneaky-cmd host port #f))
	  ((and (= l 1)
		(integer? (car rest))
		(exact? (car rest))
		(> (car rest) 0))
	   (set! port (car rest))
	   (run-sneaky-cmd host port #f))
	  ((and (= l 1)
		(string? (car rest)))
	   (set! host (car rest))
	   (run-sneaky-cmd host port #f))
	  ((not (= l 0))
	   (error "Usage: (sneaky-cmd [host] [port])")))))



;cmd-mutex:notheld
(define (init-gui cidcnt)

  ;state window
  (create-state-display cidcnt)

  ;command window
  (init-cmdwins)
  (let ((win (create-command-display)))
    ((slot-ref (slot-ref (get-text win) 'text) 'id) 'yview 0))

)

;cmd-mutex:notheld
;ew-mutex:notheld
(define (run-sneaky-cmd host port call-from-cmdwin?)
  (set! cmd-mutex (make-mutex))
; This defaults to "off"; turn on manually if you need it.
;  (client.control 'monitor)

  ; Grab the mutex to prevent the ready handlers from running until
  ; after initialization.

  (call-with-mutex cmd-mutex
    (lambda ()
      (let ((rdy (let ((gui-setup #f)
		       (rdy (make-ready-handler))
		       (m   (make-mutex)))
		   (lambda (etype cid tid edata)
		     (call-with-mutex m
		       (lambda ()
			 (if (not gui-setup)
			     (begin
			       (init-gui (client.data 'server-context-count))
			       (set! gui-setup #t)))
			 (rdy etype cid tid edata)))))))
	(display "Connecting...")
	(newline)
	(client.initialize host port rdy))))
  (display "Waiting for READY messages...") (newline)

  (for-each-thread (lambda (cid tid)
		     (ct-table-set! ct-state cid tid 'starting)))
  (for-each-thread (lambda (cid tid)
		     (ct-table-set! ct-handlers cid tid '())))

  (wait-for-ready)

  (set! connected? #t)
  (for-each (lambda (w) 
	      (cmd.msg w "~%Connection established~%")
	      (cmd.win-title w "Thread(s) all")
	      (select-all-tthreads! (get-tthreads w))
	      (set-tthread-str! w "all")
	      (if (not call-from-cmdwin?) (cmd.display-prompt w)))
	    cmd-windows)
  (for-each-thread (lambda (cid tid)
		     (client.handle 'terminate cid tid terminate-handler)))
  (for-each-thread (lambda (cid tid)
		     (client.handle 'user-event cid tid user-evt-handler)))
  (setup-function-entry/exit)
  (for-each-thread (lambda (cid tid)
		     (client.handle 'enter-function cid tid enterfunc)
		     (client.handle 'exit-function cid tid exitfunc)))
  #t)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; READY management.

(define *ready-cnt* 0)
(define *ready-sem* #f)

;cmd-mutex:held
(define (make-ready-handler)
  (set! *ready-cnt* 0)
  (set! *ready-sem* (make-semaphore 0))
  *ready-handler*)

;cmd-mutex:notheld
;ew-mutex:notheld
(define (*ready-handler* etype cid tid edata)
  (call-with-mutex cmd-mutex
    (lambda ()
      (eventmsg "READY from ~a~%" (list cid tid))
      (add-context-display cid)
      (ct-table-set! ct-state cid tid 'ready)
      (state.set-state! cid tid 'ready)
      (set! *ready-cnt* (+ *ready-cnt* 1))
      (if (= *ready-cnt* (client.data 'server-context-count))
	  (semaphore-up *ready-sem*)))))

;cmd-mutex:notheld
(define (wait-for-ready)
  (semaphore-down *ready-sem*))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; TERMINATE management

;ew-mutex:notheld
(define (terminate-handler etype cid tid edata)
  (eventmsg "TERMINATE from ~a ~a~%" cid tid))

;cmd-mutex:notheld
(define (terminate-all)
  (set! connected? #f)
  (call-with-mutex cmd-mutex
    (lambda ()
      (for-each-thread client.terminate))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; User Event Handling
;   user events must be handled specially since they must always be 
;   trapped for collection data.

(define (user-evt-handler etype cid tid edata)
  (let ((code (car edata)))
    (if (and (>= code 600) (< code 700))
	(begin (user-event-6xx cid tid edata
			       (lambda ()
				 (stepping.continue cid tid))))
	(if (memq (sneezy.event-type 'user-event) 
		  (ct-table-ref ct-handlers cid tid))
	    (general-event-handler etype cid tid edata)
	    (stepping.continue cid tid)))))
	       
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Function entry/exit event, hacked.
; Works for one thread per context.
;
; We maintain a set of stacks, one per thread, which shows the position
; of the thread in instrumented (user) code by retaining the function IDs.

(define function-stack #f)

; This could/should be a ct-table.  FIXME.

(define (setup-function-entry/exit)
  (set! function-stack (make-vector (client.data 'server-context-count) '())))

; data is ( function-tag )

(define (enterfunc etype cid tid data)
  (assert (= tid 0) "enterfunc")
  (vector-set! function-stack cid (cons (car data)
					(vector-ref function-stack cid)))
  (client.continue cid tid))

(define (exitfunc etype cid tid data)
  (assert (= tid 0) "exitfunc")
  (vector-set! function-stack cid (cdr (vector-ref function-stack cid)))
  (client.continue cid tid))

(define (get-thread-stack cid tid)
  (assert (= tid 0) "get-thread-stack")
  (vector-ref function-stack cid))

(provide "sneaky-cmd.sch")

; eof
