; s t a t e - d i s p l a y . s c h
; $Id: state-display.sch,v 1.1 1996/05/01 20:08:05 lth Exp $
;
; Sneaky state display
; This is a natural for a monitor.

(define cw-mutex #f)    ; controls access to these variables
(define step-cnt #f)    ; the entry for the step count
(define labels #f)      ; a vector of state labels
(define buttons #f)     ; a vector of continue buttons
(define event-cb #f)    ; a list of event checkboxes
(define etrace #f)      ; checkbox for tracing/no tracing
(define control-top #f) ; toplevel control window 

(define ew-mutex #f)     ; mutex for the event window
                         ; should never wait for cw-mutex within ew-mutex
(define event-window #f) ; text widget for events)


(define cb-events
  (map (lambda (id)
	 (cons id (sneezy.event-name id)))
       (list sneezy.etype:begin-barrier
	     sneezy.etype:end-barrier
	     sneezy.etype:begin-create-collection
	     sneezy.etype:end-create-collection
	     sneezy.etype:begin-delete-collection
	     sneezy.etype:end-delete-collection
	     sneezy.etype:begin-parallel
	     sneezy.etype:end-parallel
	     sneezy.etype:enter-element
	     sneezy.etype:exit-element
	     sneezy.etype:user-event)))

(define (lookup-cb id) 
  (define (iter-cb cbe i) 
    (cond ((null? cbe) '())
	  ((eq? id (caar cbe)) i)
	  (else (iter-cb (cdr cbe) (+ i 1)))))
  (iter-cb cb-events 0))



; Notes
; - The use of strings as check button values is an STk botch; it should
;   be possible to use arbitrary values without having them be folded to
;   strings or silently rejected (if they don't valid have Tcl syntax).
; - Should probably have a "clear all" button, when the events become many.
;
;cw-mutex:irrelevant
;ew-mutex:irrelevant
(define (create-state-display cidcnt)
  (set! cw-mutex (make-mutex))
  (set! ew-mutex (make-mutex))
  (let* ((control
	  (make <Toplevel> :title "Sneaky Control Panel"))
	 (beframe
	  (make <Frame> :parent control))
	 (bframe
	  (make <Frame> :parent beframe))
	 (eframe
	  (make <Frame> :parent beframe))
	 (contall
	  (make <Button> :text "Continue all" :width 20 :command '(**contall)
		:parent bframe))
	 (quit
	  (make <Button> :text "Quit/Stop" :width 20 :command '(**quit/stop)
		:parent bframe))
	 (step/cnt
	  (make <Button> :text "Step count" :width 20 :command '(**step/cnt)
		:parent bframe))
	 (count-entry
	  (make <Labeled-entry> :title "Step count:" :value "1" :width 5
		:parent bframe))
	 (ebuttons
	  (map (lambda (x)
		 (let ((id (number->string (car x)))
		       (str (cdr x))
		       (cmd (format #f "(**toggle/evt ~a)" (car x))))
		   (make <Check-button>
			 :on-value id
			 :off-value ""
			 :value ""
			 :parent eframe
			 :anchor "w"
			 :text str
			 :command cmd)))
	       cb-events))
	 (traceoff
	  (make <Check-button>
		:parent control
		:value "1"
		:anchor "w"
		:text "Event trace while stepping"))
	 (eventw (make <scroll-text> :parent control
		       :width 40 :height 10)))
    (apply pack (append ebuttons (list :fill 'x)))
    (pack quit contall step/cnt count-entry
	  :side 'top :fill 'x)
    (pack (make <Frame> :parent bframe) :side 'bottom :fill 'y)
    (pack bframe eframe :side 'left)
    (pack beframe traceoff :side 'top :fill 'x)
    (pack eventw :side 'top :fill 'x :expand 1)
    (wm 'minsize control 307 400)
    (wm 'maxsize control 585 595)

    (state.init)
    (stepping.init)
    (state.install-state-change-proc (lambda (cid tid)
				       (displaystate cid tid)))
    (set! control-top control)
    (set! labels (make-vector cidcnt #f))
    (set! event-cb ebuttons)
    (set! buttons (make-vector cidcnt #f))
    (set! step-cnt count-entry)
    (set! etrace traceoff)
    (set! event-window eventw)))

(define (**step/cnt)
  (let* ((v (get-step-count))
	 (n (string->number v)))
    (if n (begin
	    (thread-is-continuing cid tid)
	    (stepping.step-all-count n)))))

; FIX - Assumes only one thread (0)
(define (**cont cid)
  (thread-is-continuing cid tid)
  (stepping.step-thread cid 0))


(define (**quit/stop)
  (if (stepping.some-is-stepping?)
      (stepping.stop-stepping-all)
      (begin (terminate-all) (exit))))

(define (**contall)
  (stepping.stop-stepping-all)
  (for-each-thread (lambda (cid tid)
		     (if (eq? (ct-table-ref ct-state cid tid) 'ready)
			 (thread-is-continuing cid tid))))
  (stepping.continue-all))

;cw-mutex:notheld
(define (**toggle/evt evt-id)
  (call-with-mutex cw-mutex
    (lambda ()
      (let* ((index (lookup-cb evt-id))
	     (button-state (value (list-ref event-cb index))))
	(if (string=? "" button-state)
	    (for-each-thread 
	     (lambda (cid tid)
	       (if (not (= (sneezy.event-type 'user-event) evt-id))
		   (client.unhandle evt-id cid tid))
	       (format #t "~a ~a ~a~%" evt-id cid tid)
	       (ct-table-set! ct-handlers cid tid
			      (remove evt-id 
				      (ct-table-ref ct-handlers cid tid)))))
	    (for-each-thread 
	     (lambda (cid tid)
	       (if (not (= (sneezy.event-type 'user-event) evt-id))
		   (client.handle evt-id cid tid general-event-handler))
	       (ct-table-set! ct-handlers cid tid
			      (append (ct-table-ref ct-handlers cid tid)
				      (list evt-id))))))))))


;ew-mutex:notheld
; FIX - Assumes only one thread (0)
(define (general-event-handler etype cid tid edata)
  (thread-has-stopped cid tid)
  (eventmsg "Event Reached: ~a ~a ~a ~a~%" (sneezy.event-name etype)
	    cid tid edata)

  (cond ((eq? (stepping-type cid tid) 'count-event)
	 (vector-set! bvec cid (- (vector-ref bvec cid) 1))
	 (if (<= (vector-ref bvec cid) 0)
	     (begin (set-stepping-type! cid tid 'none)
		    (ct-table-set! ct-state cid tid 'ready))))
	((eq? (stepping-type cid tid) 'none)
	 (ct-table-set! ct-state cid tid 'ready)))

  (state.set-state! cid tid (sneezy.event-symbol etype))

  (if (eq? (stepping-type cid tid) 'count-event)
      (stepping.continue cid tid)))


;cw-mutex:notheld
(define (add-context-display cid)
  (call-with-mutex cw-mutex
    (lambda ()
      (let* ((frame  (make <Frame>))
	     (text   (make <Label>
			   :parent frame
			   :text (format #f "Context ~a:" cid)))
	     (button (make <Button>
			   :parent frame
			   :text "Continue"
			   :command (list '**cont cid)))
	     (label (make <Label>
			  :parent frame
			  :text (symbol->string (state.get-state cid 0)))))
	(vector-set! labels cid label)
	(vector-set! buttons cid button)
	(pack text label button :side 'left)
	(pack frame :expand #t)))))

;cw-mutex:notheld
(define (tracing-enabled?)
  (call-with-mutex cw-mutex
    (lambda ()
      (string=? (value etrace) "1"))))

;cw-mutex:notheld
(define (displaystate cid tid)
  (let* ((s  (state.get-state cid tid))
	 (d? (eq? (ct-table-ref ct-state cid tid) 'running)))
    (call-with-mutex cw-mutex
      (lambda ()
	(let ((l (vector-ref labels cid))
	      (b (vector-ref buttons cid)))
	  (slot-set! l 'text (symbol->string s))
	  (slot-set! l 'background (if d? 'red 'green))
	  (slot-set! b 'state (if d? 'disabled 'normal)))))))

;cw-mutex:notheld
(define (get-step-count)
  (call-with-mutex cw-mutex
    (lambda ()
      (value step-cnt))))

;cw-mutex:notheld
(define (get-checked-events)
  (call-with-mutex cw-mutex
    (lambda ()
      (map string->number
	   (select-if (lambda (x) (not (string=? x "")))
		      (map value event-cb))))))


;ew-mutex:notheld
(define (eventmsg . rest)
  (call-with-mutex ew-mutex
   (lambda ()
     (text-insert event-window 'end (apply format #f rest))
     (let ((line (car (string-split (text-index event-window 'end) #\.)))
	   (tkwidget (slot-ref (slot-ref event-window 'text) 'id)))
       (tkwidget 'yview `-pickplace line)))))
     


(provide "state-display.sch")

; eof
