; s t e p p i n g . s c h
;
; Thread stepping and state control.
; lth@cs.uoregon.edu / July 27, 1995

(require "monitor.sch")
(require "ct-tbl.sch")

(define stepping-info 'uninitialized-stepping-info)

(define (init-stepping-info)
  (set! stepping-info (make-ct-table)))

(define (set-stepping-type! cid tid type)
  (ct-table-set! stepping-info cid tid type))

(define (stepping-type cid tid)
  (ct-table-ref stepping-info cid tid 'none))

(monitor stepping (init
		   some-is-stepping?
		   stop-stepping-all
		   step-thread
		   step-all-count
		   step-all-until-events
		   continue
		   continue-all)

  (define (init)
    (init-stepping-info))

  (define (some-is-stepping?)
    ; We could compute the value of is-stepping as we go along, but
    ; this procedure is not called very often.
    (let ((is-stepping? #f))
      (for-each-thread (lambda (cid tid)
			 (if (not (eq? (stepping-type cid tid) 'none))
			     (set! is-stepping? #t))))
      is-stepping?))

  (define (stop-stepping-all)
    (for-each-thread (lambda (cid tid)
		       (if (not (eq? (stepping-type cid tid) 'none))
			   (set-stepping-type! cid tid 'none)))))

  (define (step-thread cid tid)
    (set-stepping-type! cid tid 'none) ; for now
    (continue cid tid))

  (define (step-all-count ntimes)
    (eventmsg "STEPPING ~a~%" ntimes) 
    (set! bvec (make-vector (client.data 'server-context-count) ntimes))
    (for-each-thread (lambda (cid tid)
		       (set-stepping-type! cid tid 'count-event)))
    (continue-all))

  (define (step-all-until-events events)
    (if (not (null? events))
	(begin (for-each-thread (lambda (cid tid)
				  (set-stepping-type! cid tid 'until-event)))
	       (set! interesting-events events)
	       (continue-all))))

  (define (continue cid tid)
    (client.continue cid tid)
    (ct-table-set! ct-state cid tid 'running)
    (state.run-procs cid tid))

  ; FIX - Assumes only one thread (0)
  (define (continue-all)
    (eventmsg "CONTINUING THREADS: all~%") 
    (for-each (lambda (n)
		(if (not (eq? (ct-table-ref ct-state n 0) 'ready))
		    #t
		    (continue n 0)))
	      (client.data 'all-contexts)))

)


(monitor state (init
		run-procs
		set-state! 
		get-state
		state-change!
		install-state-change-proc)

  (define tbl 'uninitialized-state.tbl)
  (define procs '())
  (define running-procs #f)

  (define (init)
    (set! running-procs #f)
    (set! procs '())
    (set! tbl (make-ct-table)))

  ; FIX - procs are run even if state is unchanged - needed hack (for now)
  (define (set-state! cid tid state)
      (ct-table-set! tbl cid tid state)
      (run-procs cid tid))
    
  (define (state-change! cid tid)
    (set-state! cid tid (sneezy.new-state (get-state cid tid))))

  (define (get-state cid tid)
    (ct-table-ref tbl cid tid 'unknown))

  (define (install-state-change-proc p)
    (set! procs (cons p procs)))

  ; This is not very sleek. The problem is that we want the procs to
  ; be run with monitor locking off (so that procs which access the monitor
  ; won't deadlock); at the same time we want others to stay out while the
  ; procs are being run. We have not accomplished the latter. We can probably
  ; do something with a thread-ID sensitive mutex, if we had one.
  ;
  ; Perhaps a procedure "call-without-monitor-lock" would be better.

  (define (run-procs cid tid)
    (if (not running-procs)
	(begin (set! running-procs #t)
	       (MONITOR-UNLOCK)
	       (for-each (lambda (x) (x cid tid)) procs)
	       (MONITOR-LOCK)
	       (set! running-procs #f))))
    
)

; FIXME: names obsolete; works for now.
; Perhaps belong inside monitor?

; Given a state, is it a disabled state (i.e., program is stopped but beyond
; control of the debugger)?

(define (sneezy.disabled-state? state)
  (memq state '(barrier indeterminate)))

; Given a state, is it a running state (i.e., program is actually executing)?

(define (sneezy.running-state? state)
  (memq state '(sequential parallel)))

; State transition based on event: the event determines the new state,
; without knowing what the old state is. This is based on symbolic
; events. Should accept numbers too. FIXME.

(define (sneezy.new-state x)
  (let ((probe (assq x '((ready                        sequential)
			 (begin-barrier                barrier)
			 (end-barrier                  sequential)
			 (begin-parallel               parallel)
			 (end-parallel                 sequential)
			 (begin-create-collection      sequential)
			 (end-create-collection        sequential)
			 (begin-delete-collection sequential)))))
    (if probe
	(cadr probe)
	'indeterminate)))

; Give a state or an event a color. This is based on symbolically
; represented states and events; should accept numbers too. FIXME.

(define (sneezy.state-color s)
  (let ((probe (assq s '((ready                   gray)
			 (begin-barrier           red)
			 (end-barrier             red)
			 (begin-parallel          green)
			 (enter-element           green)
			 (exit-element            green)
			 (end-parallel            green)
			 (create-collection       gray)
			 (delete-collection       gray)
			 (begin-create-collection gray)
			 (end-create-collection   gray)
			 (begin-delete-collection gray)
			 (end-delete-collection   gray)

			 (sequential              gray)
			 (barrier                 red)
			 (parallel                green)))))
    (if probe
	(cadr probe)
	'blue)))


(provide "stepping.sch")

; eof
