; Interface to TAU via a Tk proxy; allows Tk3.6 based Sneaky to communicate
; with TK4 based TAU.
;
; written by kurtw@cs.uoregon.edu / March 19, 1996
; hacked by lth@cs.uoregon.edu / March 21, 1996 (mainly generalization)

; Functions for communication with TAU via a socket-connected tk proxy.
;   tau:global-select-function  - Execute tau's global select on a 
;                                 specified tag.  This automatically
;                                 starts the proxy and fancy/tau if
;                                 they are not already running.
;   tau:get-tag-name            - Talks to tau via the proxy to get
;                                 a given tag's name.
;   send-to-proxy-sync          - Sends a tcl/tk command to a running
;                                 tcl/tk program via the proxy.  Blocks 
;                                 until a reply is received, which is
;                                 then returned.  The reply is a double-quoted
;                                 string.
;   send-to-proxy-async         - Sends a tcl/tk command to a running
;                                 tcl/tk program via the proxy. Returns
;                                 immediately with an undefined value.
;   launch-tk                   - Starts both the proxy and a specified
;                                 tcl/tk program if they are not already
;                                 running.  The specifed path should have a
;                                 trailing '/'.
;
; BUGS
; - The proxy port should be automatically selected, rather than relying
;   on command line parameters/defaults/environment variables.  The proxy
;   can and should be invisible, and no fixed or known port is needed.

(require "Tk-classes")
(require "connect.sch")
(require "threads.sch")
(require "synch.sch")
(require "sleep.sch")

; For testing
;(begin-tasking)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 
; Global Variables

(define tkproxy-connection #f)
(define tkproxy-name "tk4proxy.tcl")
(define tkproxy-host "localhost")
(define tkproxy-port
  (let ((e       (getenv "TAU_TKPROXYPORT"))
	(default 5668))
    (if e
	(or (string->number e) default)
	default)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 
; Proxy interface functions


; Message format is:
;   A(<tclscript>) - for asynchronous
;   S(<tclscript>) - for synchronous

(define (send-to-proxy-sync c msg)
  (let ((cmd (format #f "S(~a)" msg)))
    (display cmd (connection.output c))
    (flush (connection.output c)))
  (read (connection.input c)))
    
(define (send-to-proxy-async c msg)
  (let ((cmd (format #f "A(~a)" msg)))
    (display cmd (connection.output c))
    (flush (connection.output c))))


(define (tau:global-select-function tag)
  (launch-tk "./" "fancy")
  (send-to-proxy-async tkproxy-connection 
		       (format #f "global_selectFuncTag ~a" tag)))

(define (tau:get-tag-name tag)
  (launch-tk "./" "fancy")
  (send-to-proxy-sync tkproxy-connection 
		       (format #f "set funcs(~a,name)" tag)))

; Launch a tk program if not already running.

(define (launch-tk tkpath tkprog)

  (define (make-shell-command path prog params)
    (format #f "cd ~a ; ~a ~a &" path prog params))
			   
  (if (not (member tkproxy-name (winfo 'interps)))
      (begin
	(display "Starting TkProxy") (newline)
	(system (make-shell-command *tkproxy-path* tkproxy-name
				    (format #f "~a ~a ~a" 
					    tkprog tkproxy-host tkproxy-port)))
	(set! tkproxy-connection (open-proxy-connection tkproxy-port))))

  (if (or (zero? (string-length tkpath))
	  (not (char=? (string-ref tkpath (- (string-length tkpath) 1)) #\/)))
      (set! tkpath (string-append tkpath "/")))

  (if (not (member tkprog (winfo 'interps)))
      (system (make-shell-command tkpath tkprog ""))))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; 
; Utilities

(define open-proxy-connection accept-connection)

;CHANGED when replacing proxy server w/ proxy client
;(define (open-proxy-connection port)
;  (or (connect-to "localhost" port)
;      (begin (display "[tkproxy-connect: connect failed; sleeping]") (newline)
;	     (sleep 5 'seconds)
;	     (open-proxy-connection port))))

(provide "tau-interface.sch")

; eof
