#!/local/apps/tcltk/bin/dpwish -f
#
# Tk4Proxy - proxy for forwarding commands from a socket to a tk4 interpreter
#
# Usage: tk4proxy dest-interp [host port] 
#
# This proxy is a client and must connect to a server socket.
# 

# Should _never_ hardcode the port here!
# Port should be command line argument.

# Server defaults
set Proxy_ServerHost invalid_host
set Proxy_ServerPort invalid_port
set Proxy_DestInterp invalid_interpreter

# Globals
set sockid 0

# CHANGE - This function replaces AcceptConnection
# 
# ConnectToServer - attempt to connect to the server, repeating until
#                   connection is established.  Return socket id.
#
proc ConnectToServer {host port} {
    # Open the socket
    puts -nonewline "Connect to $host $port; waiting..."
    while {[catch "dp_connect $host $port" retval]} {
	puts "Failed; sleeping..."
	after 1000
	update idletasks
	update
    }

    # Connected
    puts "PROXY CONNECTED!"
    return [lindex $retval 0]
}

#
# receiveMsg - reads from the socket until a message has been read.
#              The full message will be enclosed in quotes.
#              Returns the message
proc receiveMsg {sockid} {

    # Wait for socket to be readable
    while {![lindex [dp_isready $sockid] 0]} {
	update idletasks
    }
	
    if {[catch "dp_receive $sockid" msg]} {
	showError "The server has disconnected unexpectedly." 1
    }

    while {!([string index $msg 1] == "(" && \
	    [string index $msg [expr [string length $msg] - 1]] == ")")} {

        # Keep reading
	update idletasks
        if {[catch "dp_receive $sockid" frag]} {
            showError "The client has disconnected unexpectedly." 1
	    tkwait window .dataerror
        }
	set msg "$msg$frag"
    }

    return $msg
}


proc sendReply {sockid msg} {
    # Wait for socket to be writable
    while {![lindex [dp_isready $sockid] 1]} {
	update idletasks
    }

    set send_cmd [format "dp_send %s \"\\\"%s\\\"\"" $sockid $msg]
    if {[catch $send_cmd]} {
	showError "The client has disconnected unexpectedly." 1
    }
}


proc ForwardMsg {mode sockid} {
    global Proxy_DestInterp

    set msg [receiveMsg $sockid]
    if {$msg == ""} {
	showError "The server has disconnected unexpectedly." 1
    }

    set cmd_part [string range $msg 2 [expr [string length $msg] - 2]]
    if {[string index $msg 0] == "S"} {
	set synchronous 1
    } else {
	set synchronous 0
    }

    set reply [send $Proxy_DestInterp $cmd_part]
    
    if {$synchronous} {
	sendReply $sockid $reply
    }
}

    


proc showError {message {fatal 0}} {
    
    if {![winfo exists .dataerror]} {
	toplevel .dataerror
	wm title .dataerror "Error"
	
	label .dataerror.l1 -bitmap error -foreground red
	message .dataerror.m1 -text $message -width 250 -foreground black
	if {$fatal} {
	    button .dataerror.b -text "OK" -command "exit"
	} else {
	    button .dataerror.b -text "OK" -command "grab release .dataerror; destroy .dataerror"
	}
	pack .dataerror.b  -side bottom -fill x
	pack .dataerror.l1 -side left
	pack .dataerror.m1 -side left -fill both
	
	tkwait visibility .dataerror
	grab .dataerror
    }
}




# MAIN CODE

if {($argc != 3)} {
    puts stderr "usage: tk4proxy interpreter-name host port"
    exit
}
set Proxy_DestInterp [lindex $argv 0]
set Proxy_ServerHost [lindex $argv 1]
set Proxy_ServerPort [lindex $argv 2]

puts stderr "DEBUG: $Proxy_DestInterp $Proxy_ServerHost $Proxy_ServerPort"
#exit

# CHANGE - Call to AcceptConnection changed to ConnectToServer w/ extra arg
wm withdraw .
set sockid [ConnectToServer $Proxy_ServerHost $Proxy_ServerPort]
if {[string length $sockid] > 0} {
    dp_filehandler $sockid re ForwardMsg
}
