/* c l i e n t - l o w . h
 * $Id: client-low.h,v 1.1 1996/05/01 20:09:40 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: October 9-25, 1995
 * Modified:
 *
 * Low-level types and #defines for Sneezy client interface.
 */

#if defined(SERVER_SIDE) || !defined(CLIENT_SIDE)
#error "client-low.h valid only on the client side."
#endif

#ifndef INCLUDED_CLIENT_LOW
#define INCLUDED_CLIENT_LOW

#include "cidtid.h"

#ifdef __cplusplus
extern "C" {
#endif

#define EID_OFFSET 100

/* A connection has a tag field (which tells us the type of connection),
 * a table of functions implementing the lowest level of the operations, 
 * and some tag-specific information.
 *
 * A connection structure is pretty big and contains pointer data, and 
 * should always be treated as an object with pointer identity (although
 * that's not required by the API; perhaps it should be).
 */
struct snz_conn_t {
  short tag;
  void (*close)( snz_conn_t * );
  void (*install)( snz_conn_t *, int, int, int, int );
  void (*uninstall)( snz_conn_t *, int, int, int );
  int  (*wait)( snz_conn_t *, int );
  void (*cont)( snz_conn_t *, int, int );
  void (*terminate)( snz_conn_t *, int, int );
  void (*read)( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len );
  void (*write)( snz_conn_t *c, int cid, int tid, snz_addr_t p, int len, 
	         void *data );
  void (*call_func)( snz_conn_t *c, int cid, int tid, snz_addr_t addr, 
		     snz_param_t *p );
  int  (*read_reply)( snz_conn_t *c, int cid, int tid, void *data,
		      int (*k)( snz_conn_t *, int, int, void * ) );
  /* call_reply */
  snz_proginfo_t proginfo;
  snz_hstruct_t  handlers;
  snz_cidtid_t   *sysmap;
  union {
    struct {
      char *addr;   /* IP address */
      int  port;    /* IP port number */
      FILE *fp_in;  /* Input stream */
      FILE *fp_out; /* Output stream */
    } socket;
  } u;
  int sizeof_int;
  int sizeof_long;
  int sizeof_addr;
};

#define SNZ_CONN_INVALID    0
#define SNZ_CONN_SOCK_LISP  1   /* socket; lisp syntax */
#define SNZ_CONN_SOCK_C     2   /* socket; C binary data */

#define SNZ_MAX_PARAMS 10

struct snz_param_t {
  int params;
  struct {
    union {
      struct {
	snz_str_t *data;
	short copied;
      } s;
      snz_int_t i;
      snz_long_t l;
      snz_addr_t p;
    } u;
    short tag;
  } param[ SNZ_MAX_PARAMS ];
};


#define SNZ_PARAM_INT       0
#define SNZ_PARAM_LONG      1
#define SNZ_PARAM_STRING    2
#define SNZ_PARAM_ADDR      3

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_CLIENT_LOW */

/* eof */
