/* c l i e n t - r e p l y . c
 * $Id: client-reply.c,v 1.1 1996/05/01 20:09:41 lth Exp $
 * 
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 20, 1995
 * Modified:
 *
 * Reply management.
 */

#define CLIENT_SIDE

#include "client-api.h"
#include "client-reply.h"

#define REQ_NONE 0
#define REQ_READ 1

typedef struct read_request read_request;
struct read_request {
  snz_addr_t p;
  int len;
  int (*k)( snz_conn_t*, int, int, void * );
  read_request *next;
};

typedef struct {
  int request;
  struct {
    int (*read)( snz_conn_t*, int, int, void * );
  } handler;
  struct {
    read_request *read;
  } queues;
} snz_request_t;

static snz_request_t *get_request_record( snz_conn_t *c, int cid, int tid );
static int read_reply( snz_conn_t *c, int cid, int tid, void *d );

void 
dispatch_read_request( snz_conn_t *c, int cid, int tid, snz_addr_t p,
		       int len, 
		       int (*k)( snz_conn_t *c, int cid, int tid, void *b ) )
{
  snz_request_t *pr;
  read_request *rr;

  pr = get_request_record( c, cid, tid );
  if (pr->request != REQ_NONE) {
    snz_message( "[dispatch_read_request: deferring.]" );
    rr = (read_request*)malloc( sizeof( read_request ) );
    if (rr == 0) snz_perror_panic( "dispatch_read_request: malloc" );
    rr->p = p;
    rr->len = len;
    rr->k = k;
    rr->next = pr->queues.read;
    pr->queues.read = rr;
  }
  else {
    pr->request = REQ_READ;
    pr->handler.read = k;
    c->read( c, cid, tid, p, len );
  }
}

int snz_reply( snz_conn_t *c, int cid, int tid, void *d )
{
  snz_request_t *pr;

  pr = get_request_record( c, cid, tid );
  switch (pr->request) {
  case REQ_READ :
    return read_reply( c, cid, tid, d );
  case REQ_NONE :
    snz_panic( "snz_reply: no pending request for %d %d!", cid, tid );
    break;
  default :
    snz_panic( "snz_reply: invalid code in request field: %d!", pr->request );
  }
  /*NOTREACHED*/
  return 0;
}

static int read_reply( snz_conn_t *c, int cid, int tid, void *d )
{
  snz_request_t *pr;
  read_request *rr;
  int (*k)( snz_conn_t *, int, int, void * );

  pr = get_request_record( c, cid, tid );
  pr->request = REQ_NONE;
  k = pr->handler.read;

  /* FIXME: When there are multiple queues this must check them all; 
   * create an abstraction of this procedure.
   */
  if (pr->queues.read != 0) {
    rr = pr->queues.read;
    pr->queues.read = pr->queues.read->next;
    dispatch_read_request( c, cid, tid, rr->p, rr->len, rr->k );
    free( rr );
  }
  return c->read_reply( c, cid, tid, d, k );
}

static snz_request_t *get_request_record( snz_conn_t *c, int cid, int tid )
{
  snz_request_t *pr;

  pr = (snz_request_t*)snz_cidtid_lookup( c->sysmap, cid, tid );
  if (pr == 0) {
    pr = (snz_request_t*)malloc( sizeof( snz_request_t ) );
    if (pr == 0) snz_perror_panic( "malloc" );
    pr->request = REQ_NONE;
    pr->handler.read = 0;
    pr->queues.read = 0;
    snz_cidtid_insert( c->sysmap, cid, tid, (void*)pr );
  }
  return pr;
}

/* eof */
