/* c l i e n t - r e p r . h
 * $Id: client-repr.h,v 1.1 1996/05/01 20:09:42 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 29, 1995
 * Modified:
 *
 * Client-side representations for server-side quantities.
 *
 * Numeric quantities are represented as 64-bit sign-and-magnitude
 * quantities.
 */

#ifndef INCLUDED_CLIENT_REPR
#define INCLUDED_CLIENT_REPR

#ifdef __cplusplus
extern "C" {
#endif

/* The type signed16 must hold _at least_ 16 bits; more is OK. 
 * Ditto unsigned. Ditto 32.
 */
typedef int _signed16;
typedef unsigned int _unsigned16;
typedef long _signed32;
typedef unsigned long _unsigned32;

typedef unsigned char _octet;

/* Type definitions for machine-independent data. */
struct snz_int_t {
  _unsigned16 dig[4];
  signed char size;     /* server-dependent: 4 or 8 */
  signed char sign;
};

struct snz_long_t {
  _unsigned16 dig[4];
  signed char size;     /* server-dependent: 4 or 8 */
  signed char sign;
};

struct snz_str_t {
  _signed32 length;
  _octet *data;   /* assume 8-bit chars */
};

struct snz_addr_t {
  _unsigned16 dig[4];
  signed char size;     /* server-dependent: 4 or 8 */
};

/* snz_time_t is a data type which has enough granularity to record times 
 * to an interesting level of precision (which is system-dependent, by 
 * the way).
 */
struct snz_time_t {   /* FIXME */
  double d;
};

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_CLIENT_REPR */

/* eof */
