/* c l i e n t - t y p e s . c
 * $Id: client-types.c,v 1.1 1996/05/01 20:09:44 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 2, 1995
 * Modified:
 *
 * Operations on abstract data types representing server-side quantities.
 */

#define CLIENT_SIDE

#include <string.h>
#include "client-api.h"

static void depint( int size, int sign, _unsigned16* dig, char *s );
static void extract( int size, signed char *sign, _unsigned16 *dig, char *s );

static void dumpnum( int sign, _unsigned16 *dig )
{
  int k;

  printf( "<%d", sign );
  for ( k = 3 ; k >= 0 ; k-- ) printf( " %04x", dig[k] );
  printf( ">\n" );
}

snz_int_t snz_mkint( snz_conn_t *c, int i )
{
  snz_int_t d;
  unsigned u;
  int k;

  d.size = c->sizeof_int;
  d.sign = (i < 0 ? 1 : 0);
  u = (i < 0 ? -i : i);
  for ( k = 0 ; k < 4 ; k++ ) {
    d.dig[k] = u & 0xFFFF;
    u >>= 16;
  }
  return d;
}

snz_long_t snz_mklong( snz_conn_t *c, long l )
{
  snz_long_t d;
  unsigned long u;
  int k;

  d.size = c->sizeof_int;
  d.sign = (l < 0 ? 1 : 0);
  u = (l < 0 ? -l : l);
  for ( k = 0 ; k < 4 ; k++ ) {
    d.dig[k] = u & 0xFFFF;
    u >>= 16;
  }
  return d;
}

snz_str_t snz_mkstr( snz_conn_t *c, char *s )
{
  snz_panic( "snz_mkstr" );    /* FIXME */
  /*NOTREACHED*/
  /* Ignore warning here */
}

snz_addr_t snz_mkaddr( snz_conn_t *c, unsigned long s )
{
  snz_addr_t p;
  int k;

  p.size = c->sizeof_addr;
  for ( k = 0 ; k < 4 ; k++ ) {
    p.dig[k] = s & 0xFFFF;
    s >>= 16;
  }
  return p;
}

int val( snz_long_t l )
{
  int i;

  i = (int)lval( l );
#if 0
  dumpnum( l.sign, l.dig );
  snz_message( "<val returns %d>", i );
#endif
  return i;
}

int ival( snz_int_t i )
{
  int k;

  if (sizeof(int) == 4)
    k = (int)(((unsigned)i.dig[1] << 16) | (unsigned)i.dig[0]);
  else
    k = (int)(((unsigned)i.dig[3] << 16) 
	     |((unsigned)i.dig[2] << 16)
	     |((unsigned)i.dig[1] << 16) 
	     |((unsigned)i.dig[0]));
  return (i.sign ? -k : k);
}

long lval( snz_long_t l )
{
  long k;

  if (sizeof(long) == 4)
    k = (long)(((unsigned long)l.dig[1] << 16) | (unsigned long)l.dig[0]);
  else
    k = (long)(((unsigned long)l.dig[3] << 16) 
	      |((unsigned long)l.dig[2] << 16) 
	      |((unsigned long)l.dig[1] << 16)
	      |((unsigned long)l.dig[0]));
  return (l.sign ? -k : k);
}

snz_addr_t addr( snz_conn_t *c, snz_long_t l )
{
  snz_addr_t a;
  int k;

#if 0
  dumpnum( l.sign, l.dig );
#endif
  a.size = c->sizeof_addr;
  for ( k = 0 ; k < 4 ; k++ )
    a.dig[k] = l.dig[k];
  return a;
}

void snz_int2str( snz_int_t i, char *buf )
{
  depint( i.size, i.sign, i.dig, buf );
}

void snz_long2str( snz_long_t l, char *buf )
{
  depint( l.size, l.sign, l.dig, buf );
}

void snz_addr2str( snz_addr_t p, char *buf )
{
  depint( p.size, 0, p.dig, buf );
}

void snz_time2str( snz_time_t t, char *buf )
{
  strcpy( buf, "0" );  /*FIXME*/
}

snz_int_t snz_str2int( snz_conn_t *c, char *s )
{
  snz_int_t d;

  d.size = c->sizeof_int;
  extract( c->sizeof_int, &d.sign, d.dig, s );
  return d;
}

snz_long_t snz_str2long( snz_conn_t *c, char *s )
{
  snz_long_t d;

  d.size = c->sizeof_long;
  extract( c->sizeof_long, &d.sign, d.dig, s );
#if 0
  printf( "<%s> ", s );
  dumpnum( d.sign, d.dig );
#endif
  return d;
}

snz_addr_t snz_str2addr( snz_conn_t *c, char *s )
{
  snz_addr_t d;
  signed char sign;

  d.size = c->sizeof_addr;
  extract( c->sizeof_addr, &sign, d.dig, s );
  if (sign != 0)
    snz_panic( "snz_str2addr: negative!" );
  return d;
}

snz_str_t *snzstr_dup( snz_str_t *s )
{
  snz_panic( "snzstr_dup" );  /* FIXME */
  return 0;
}

void snz_str_free( snz_str_t *s )
{
  /* FIXME */
}

/***************************************************************************/

static void add_i( _unsigned16 *dig, _unsigned16 i )
{
  _unsigned32 carry, tmp;
  int k;

  carry = i;
  for ( k = 0 ; k < 4 ; k++ ) {
    tmp = (_unsigned32)dig[k]+carry;
    dig[k] = (_unsigned16)(tmp&0xFFFF);
    carry = tmp >> 16;
  }
}

static void add( _unsigned16 *dig1, _unsigned16 *dig2 )
{
  _unsigned32 tmp, carry;
  int k;

  carry = 0;
  for ( k = 0 ; k < 4 ; k++ ) {
    tmp = (_unsigned32)dig1[k]+(_unsigned32)dig2[k]+carry;
    dig1[k] = (_unsigned16)(tmp & 0xFFFF);
    carry = tmp >> 16;
  }
}

/* a*10 == (a*2)+(a*2*2*2); a*2=a+a */
static void mul10( _unsigned16 *dig )
{
  _unsigned16 tmp[4];

  add( dig, dig );
  memcpy( tmp, dig, sizeof(tmp) );
  add( dig, dig );
  add( dig, dig );
  add( dig, tmp );
}

#define xval( x ) ((x)<='9' ? (x)-'0' : ((x) <= 'Z' ? (x)-'A'+10 : (x)-'a'+10))
#define dval( x ) ((x)-'0')
#define hexlo( n )  ("0123456789abcdef"[(n)&15])
#define hexhi( n )  ("0123456789abcdef"[((n)>>4)&15])

static void extract( int size, signed char *sign, _unsigned16 *dig, char *s )
{
  int j, k;
  char *p;

  *sign = 0;
  if (*s == '#' && *(s+1) == 'x') {
    s+=2;
    if (*s == '-') {
      *sign = 1;
      s++;
    }
    p = strchr( s, 0 )-1;
    for ( k = 0 ; k < 4 ; k++ ) dig[k] = 0;
    for ( k = 0 ; k < 4 && p >= s ; k++ ) {
      j = 0;
      while (j < 16 && p >= s) {
	dig[k] |= (xval(*p) << j);
	p--;
	j+=4;
      }
    }
  }
  else {
    if (*s == '-') { 
      *sign = 1; 
      s++; 
    }
    for ( k = 0 ; k < 4 ; k++ ) dig[k] = 0;
    while (*s) {
      mul10( dig );
      add_i( dig, dval(*s) );
      s++;
    }
  }
}

static void deposit( char *s, _unsigned16 k )
{
  *(s+3) = hexlo( k & 255 );
  *(s+2) = hexhi( k & 255 );
  k >>= 8;
  *(s+1) = hexlo( k & 255 );
  *(s+0) = hexhi( k & 255 );
}

static void depint( int size, int sign, _unsigned16* dig, char *s )
{
  int k;

  *s++ = '#';
  *s++ = 'x';
  if (sign) *s++ = '-';
  for ( k = 3 ; k > 0 && dig[k] == 0 ; k-- )
    ;
  for ( ; k >= 0 ; k-- ) {
    deposit( s, dig[k] );
    s+=4;
  }
  *s = 0;
}

/* eof */
