/* This program exhibits a bug in the IRIX C compiler.
 * Lars Thomas Hansen [lth@cs.uoregon.edu] / November 30, 1995
 *
 * Description.
 * The returned value from the function str() is wrong when the complex
 * conditional in the _check() macro is used, but correct when the simple
 * definition is used. The incorrect return value is fetched from the union
 * at an offset several bytes too high.
 *
 * The correct output from the program should be 
 *    12345678 37
 * but is instead
 *    2500000000 0
 * One notes that 0x25==37.
 *
 * To exhibit the bug, compile with -DEXHIBIT_BUG
 * The bug does not show up when compiling with -32.
 *
 * The machine:
 *

earth(5) % uname -a
IRIX64 geos 6.1 07121823 IP21 mips

 *
 * The command line with -show:
 *

earth(12) % cc -show -DEXHIBIT_BUG -o irix-bug irix-bug.c
/usr/lib64/cmplrs/cc -show -DEXHIBIT_BUG -o irix-bug irix-bug.c -64
/usr/lib64/cmplrs/fec -G8 -DMIPSEB -D_MIPSEB -TENV:PIC -D_PIC -D__DSO__ -O0 -Ac=xansi -D__EXTENSIONS__ -D__INLINE_INTRINSICS -D__MATH_HAS_NO_SIDE_EFFECTS -DLANGUAGE_C -Dmips -Dunix -Dsgi -Dhost_mips -D_SGI_SOURCE -D_LONGLONG -D_SVR4_SOURCE -D_LANGUAGE_C -D_MODERN_C -D__sgi -D__host_mips -Amachine(mips) -D_SYSTYPE_SVR4 -D__unix -Asystem(unix) -D_COMPILER_VERSION=601 -m1 -show -DEXHIBIT_BUG -D__mips=4 -D_MIPS_ISA=4 -D_ABI64=3 -D_MIPS_SIM=_ABI64 -D_MIPS_FPSET=32 -D_MIPS_SZINT=32 -D_MIPS_SZLONG=64 -D_MIPS_SZPTR=64 -TARG:abi=64:isa=mips4 -I/usr/include -fB,/tmp/ctmB.AAAa000as irix-bug.c 
Compiling _sexpr_panic 
Compiling str 
Compiling main 
/usr/lib64/cmplrs/be -G8 -TENV:PIC -O0 -m1 -show -TARG:abi=64:isa=mips4 -fB,/tmp/ctmB.AAAa000as -fo,irix-bug.o irix-bug.c 
Compiling irix-bug.c (/tmp/ctmB.AAAa000as) -- Back End
_sexpr_panic str main 
/usr/lib64/cmplrs/ld64 -no_unresolved -call_shared -transitive_link -elf -_SYSTYPE_SVR4 -o irix-bug -mips4 -64 -L/usr/lib64/mips4 -L/usr/lib64 -nocount /usr/lib64/mips4/crt1.o -count irix-bug.o -nocount -dont_warn_unused -lc /usr/lib64/mips4/crtn.o -warn_unused 

 *
 */

typedef struct snz_long_t snz_long_t;
typedef struct sexpr_obj_t sexpr_obj_t;

struct snz_long_t {
  unsigned int dig[4];
  signed char size;
  signed char sign;
};

struct _sexpr_str {
  char *data;
  int length;
};

struct sexpr_obj_t {
  union {
    snz_long_t num;
    struct _sexpr_str str;
    struct {
      sexpr_obj_t *car;
      sexpr_obj_t *cdr;
    } pair;
  } u;
  short tag;
};

#define TAG_STRING 2

sexpr_obj_t *_sexpr_panic(const char*p,int t, sexpr_obj_t*q) {
  /* Normally there's a panic routine here */
  return 0; 
}

#ifdef EXHIBIT_BUG
#define _check(m,t,op,name) \
    ((m&&m->tag==t)?(m->op):(_sexpr_panic(name,t,m)->op))
#else
#define _check( m,t,op,name ) (m&&m->tag==t?m->op:m->op)
#endif

#define _str( m ) _check((m),TAG_STRING,u.str,"str")

static struct _sexpr_str str( sexpr_obj_t *p ) { return _str(p); }

main()
{
  sexpr_obj_t o, *p;
  struct _sexpr_str s;

  p = &o;
  p->tag = TAG_STRING;
  p->u.str.data = (char*)0x12345678;
  p->u.str.length = 37;
  s = str(p);
  printf( "%lx %d\n", (unsigned long)s.data, s.length );
  return 0;
}
