/* p r o x y - s e r v e r . c
 * $Id: proxy-server.c,v 1.1 1996/05/01 20:09:50 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu], 
 *           based loosely on a design by Darryl Ivan Brown.
 *  Created: November 21, 1995
 * Modified: January 24, 1996 / lth -- implemented USE_STDIO
 *
 * Procedures which are called from the parallel program to initiate 
 * the connection to the proxy.
 */

#define SERVER_SIDE

#include "sneezy.h"
#include "server-api.h"
#include "proxy-sock.h"
#include "server-low.h"
#include "socket.h"

int _proxy_portno;

struct cidinfo {
  int cid;
  int nthreads;
};

/* Called from each _context_ just before system goes into Processor_Main().
 * This procedure is never run in the proxy.
 */
snz_sconn_t *snz_server_init_context( int cid, int nthreads )
{
  snz_sconn_t *c;
  struct cidinfo x;
  int fd, one;

  fd = connect_to_server( myaddr(), _proxy_portno );

  x.cid = cid;
  x.nthreads = nthreads;
  if (write( fd, (void*)&x, sizeof( x ) ) == -1)
    snz_perror_panic( "snz_server_init_context: write" );

  c = _snz_create_remote_connection();
  c->tag = TAG_SOCKET;
  c->send_event = _snz_sock_send_event;
  c->recv_cmd = _snz_sock_recv_cmd;
  c->send_reply = _snz_sock_send_reply;

  /* Synchronize */
  snz_message( "Synchronizing read." );
  if (read( fd, (void*)&one, sizeof( one ) ) == -1) 
    snz_perror_panic( "snz_server_init_context: read" );
  snz_message( "Synchronized read." );

#ifdef USE_STDIO
  c->u.socket.fp_out = fdopen( fd, "w" );
  c->u.socket.fp_in = fdopen( fd, "r" );
  if (c->u.socket.fp_in == 0 || c->u.socket.fp_out == 0) 
    snz_panic( "snz_server_init_context: fdopen" );
#else
  c->u.socket.fd = fd;
#endif
  return c;
}


/* Called from each _context_ just before the context terminates. 
 * This procedure is never run in the proxy.
 */
void snz_server_shutdown_context( int cid, snz_sconn_t *c )
{
#ifdef USE_STDIO
  fclose( c->u.socket.fp_in );
  fclose( c->u.socket.fp_out );
#else
  close( c->u.socket.fd );
#endif
}


/* eof */
