/* p r o x y - s o c k - s i n g l e . c
 * $Id: proxy-sock-single.c,v 1.1 1996/05/01 20:09:53 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 6, 1995
 * Modified:
 *
 * Single client support: very fast and simple. 
 */

#define SERVER_SIDE

#include <unistd.h>
#include <stdio.h>
#include "sneezy.h"
#include "server-low.h"
#include "proxy-sock.h"

static int hasclient = 0;
static snz_clnt_t *client;

int _proxy_accept_new_client_single_client( int fd, snz_proginfo_t *info )
{
  snz_clnt_t *c;
  snz_event_t e;

  if (hasclient)
    snz_panic( "Single-client proxy cannot accept multiple clients." );
  hasclient = 1;
  c = _snz_new_client( fd );

  /* Negotiate protocol; for now assume S-expressions over sockets.
   * Negotiation is pretty independent of single/multi and should be 
   * supported directly in proxy-sock.
   */
  /* _proxy_sock_negotiate( c ); */

  c->protocol = PROTO_LISP;
  c->read = _proxy_read_lisp;
  c->write = _proxy_write_lisp;
  c->tag = TAG_SOCKET;
  c->u.socket.fp_in = fdopen( fd, "r" );
  c->u.socket.fp_out = fdopen( fd, "w" );
  if (c->u.socket.fp_in == 0 || c->u.socket.fp_out == 0)
    snz_panic( "_proxy_accept_new_client_single_client: fdopen." );

  e.etype = _SNZE_STARTUP;
  e.u.startup.info = info;
  c->write( c, &e );
  client = c;
  return fd;
}

int 
_proxy_route_to_program_single_client( int *program_fds, snz_clnt_t *src )
/*NOTUSED:src*/
{
  int i;
  snz_cmd_t cmd;

  if (!client->read( client, &cmd )) {
    snz_message( "_proxy_route_to_program_single_client: EOF on channel." );
    return 0;
  }
  if (!write_bytes( program_fds[ cmd.cid ], (void*)&cmd, sizeof( cmd ) ))
    snz_perror_panic( "write" );

  if (cmd.cmd == SNZC_WRITE && cmd.u.write.data != 0) {
    if (!write_bytes( program_fds[ cmd.cid ], (void*)cmd.u.write.data,
		    cmd.u.write.len ))
      snz_perror_panic( "write" );
    free( cmd.u.write.data );
  }

  if (cmd.cmd == SNZC_CALL_USER_FUNCTION) {
    if (!write_bytes( program_fds[ cmd.cid ], (void*)cmd.u.call.argv,
		    cmd.u.call.argc*sizeof( snz_arg_t ) ))
      snz_perror_panic( "write" );

    /* String arguments carry data: each string is an snz_str. */
    /* Strings are written after the argv structure, in order, with a
     * native length field followed by the bytes.
     * String data is _not_ freed since it is shared (see proxy-sock-lisp.c).
     */
    for ( i = 0 ; i < cmd.u.call.argc ; i++ )
      if (cmd.u.call.argv[i].tag == SNZAT_CHARPTR) {
	if (!write_bytes( program_fds[ cmd.cid ], 
			  (void*)&cmd.u.call.argv[i].u.str.len,
			  sizeof( int ) ))
	  snz_perror_panic( "write" );
	if (!write_bytes( program_fds[ cmd.cid ],
			  (void*)cmd.u.call.argv[i].u.str.cp,
			 cmd.u.call.argv[i].u.str.len ))
	  snz_perror_panic( "write" );
      }
    free( cmd.u.call.argv );
  }

  return 1;
}

int
_proxy_route_to_clients_single_client( int context_fd )
{
  snz_event_t e;
  char *buf = 0;

  if (client == 0)
    snz_panic( "_proxy_route_to_clients_single_client: no client" );

  if (!read_bytes( context_fd, (void*)&e, sizeof( e ))) return 0;
  if (e.etype == _SNZE_REPLY) {
    buf = e.u.reply.data = _snz_allocbuf( e.u.reply.len );
    if (!read_bytes( context_fd, (void*)buf, e.u.reply.len )) return 0;
  }
  if (e.etype == _SNZE_PUSH) {
    buf = e.u.push.value = _snz_allocbuf( e.u.push.len );
    if (!read_bytes( context_fd, (void*)buf, e.u.push.len )) return 0;
  }

  client->write( client, &e );
  if (buf) _snz_freebuf( buf );
  return 1;
}

/* eof */
