/* p r o x y - s o c k . h
 * $Id: proxy-sock.h,v 1.1 1996/05/01 20:09:55 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 9, 1995
 * Modified:
 *
 * Definitions exported by proxy comm layer.
 * None of this is really socket specific; parts of the implementation are.
 */

#if !defined( SERVER_SIDE ) || defined( CLIENT_SIDE )
#error "proxy-sock.h valid on server side only."
#endif

#ifndef INCLUDED_PROXY_SOCK
#define INCLUDED_PROXY_SOCK

#include "server-low.h"

#ifdef __cplusplus
extern "C" {
#endif

#define TAG_NONE       0
#define TAG_SOCKET     1

#define FD_PROGRAM     0
#define FD_CLIENT      1
#define FD_CONNECT     2

/* Reply tags */
#define _REPL_READ  0
#define _REPL_CALL  1

/* A client structure encodes the connection between the proxy and a
 * client.  It has read and write methods and connection-specific data.
 */
typedef struct snz_clnt_t snz_clnt_t;

struct snz_clnt_t {
  int tag;
  int protocol;
  int fd;            /* more for identification purposes than anything */
  union {
    struct {
      FILE *fp_in;
      FILE *fp_out;
    } socket;
  } u;
  int (*read)( snz_clnt_t *clnt, snz_cmd_t *cmd );
  void (*write)( snz_clnt_t *clnt, snz_event_t *e );
};

#define PROTO_LISP  0
#define PROTO_BIN   1
#define PROTO_XDR   2

/* In proxy-sock.c */

void snz_server_init_agent( int port, int multi );
snz_sconn_t *snz_server_init_context( int cid, int nthreads );
void snz_server_shutdown_context( int cid, snz_sconn_t *c );
snz_clnt_t *_snz_new_client( int fd );


/* In proxy-sock-single.c */

int _proxy_accept_new_client_single_client( int fd, snz_proginfo_t *info );
int _proxy_route_to_program_single_client( int *fds, snz_clnt_t *c );
int _proxy_route_to_clients_single_client( int context_fd );


/* In proxy-sock-multi.c */

int _proxy_accept_new_client_multi_client( int fd, snz_proginfo_t *info );
int _proxy_route_to_program_multi_client( int *fds, snz_clnt_t *c );
int _proxy_route_to_clients_multi_client( int context_fd );


/* In proxy-sock-lisp.c */

int _proxy_read_lisp( snz_clnt_t *, snz_cmd_t *cmd );
void _proxy_write_lisp( snz_clnt_t *, snz_event_t *e );


/* In proxy-sock-bin.c */

int _proxy_read_bin( snz_clnt_t *, snz_cmd_t *cmd );
void _proxy_write_bin( snz_clnt_t *, snz_event_t *e );

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_PROXY_SOCK */

/* eof */
