/* s e r v e r - a p i . h
 * $Id: server-api.h,v 1.1 1996/05/01 20:09:56 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: October 11, 1995
 * Modified:
 *
 * Sneezy server API defintion (C). PROTOTYPE.
 */

#include <stddef.h>

#if defined(CLIENT_SIDE) || !defined(SERVER_SIDE)
#error "server-api.h may be used on server side only."
#endif

#ifndef INCLUDED_SERVER_API
#define INCLUDED_SERVER_API

#ifdef __cplusplus
extern "C" {
#endif

/* Startup and shutdown */

void snz_server_init_agent( int agent_port, int multi ); /* called once */

void snz_register_context( void );  /* once from each context */
void snz_shutdown_context( void );  /* once from each context */

/* Introspection */

void snz_whoami( int *context, int *thread );
int snz_get_event_id( int event_type );

/* Event signalling */

void snz_ready_event( void );
void snz_error_event( int code );
void snz_terminate_event( void );
void snz_begin_barrier_event( int barrier_id );
void snz_end_barrier_event( int barrier_id );
void snz_begin_fetch_event( int src_cid, void *src_addr, void *dest_addr, 
			    int len );
void snz_end_fetch_event( int src_cid, void *src_addr, void *dest_addr, 
			  int len );
void snz_begin_put_event( int dest_cid, void *dest_addr, void *src_addr,
			  int len );
void snz_end_put_event( int dest_cid, void *dest_addr, void *src_addr,
		        int len );
/* dest_addr is 'this' ptr, and function is member pointer. */
/* FIXME: that's too C++ specific. */
void snz_begin_rpc_event( int dest_cid, void *dest_addr, int function, 
			  void *params, int plen );
void snz_end_rpc_event( int dest_cid, void *dest_addr, int function,
		        void *params, int plen );
void snz_begin_service_rpc_event( int src_cid, void *buf, int len ); /*FIXME*/
void snz_end_service_rpc_event( int src_cid, void *buf, int len );   /*FIXME*/
void snz_begin_parallel_event( void );
void snz_end_parallel_event( void );
void snz_enter_element_event( void *eltaddr );
void snz_exit_element_event( void *eltaddr );
void snz_user_event( int event_code, void *p, int len );
void snz_begin_create_collection_event( int cnum );
void snz_end_create_collection_event( int cnum );
void snz_begin_delete_collection_event( int cnum );
void snz_end_delete_collection_event( int cnum );
void snz_enter_function_event( int function_code );
void snz_exit_function_event( int function_code );

/* Value transmission */

void snz_push_value( int code, void *p, size_t len );

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_SERVER_API */

/* eof */
