/* s e r v e r - l o w . h
 * $Id: server-low.h,v 1.1 1996/05/01 20:09:58 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 9, 1995
 * Modified: January 24, 1996 / lth -- implemented USE_STDIO.
 *
 * Definitions private to the server.
 */

#if !defined( SERVER_SIDE ) || defined( CLIENT_SIDE )
#error "server-low.h valid only on client side."
#endif

#ifndef INCLUDED_SERVER_LOW
#define INCLUDED_SERVER_LOW

#include <stdio.h>
#include "server-types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* This is the structure for parameters to call-user-function,
 * on the server side.
 */
struct _snz_arg_t {
  union {
    int i;
    unsigned int ui;
    long l;
    unsigned long ul;
    short s;
    unsigned short us;
    char c;
    unsigned char uc;
    signed char sc;
    struct {
      char *cp;
      int  len;
    } str;
  } u;
  short tag;  /* SNZAT_ values below. */
};

/* Argument type tags for call_user_function */

#define SNZAT_INT    1
#define SNZAT_UINT   2
#define SNZAT_SHORT  3
#define SNZAT_USHORT 4
#define SNZAT_CHAR   5
#define SNZAT_UCHAR  6
#define SNZAT_SCHAR  7
#define SNZAT_LONG   8
#define SNZAT_ULONG  9
#define SNZAT_CHARPTR 10

/* A server connection encodes the agent's interface to the proxy or client;
 * there's one connection for each thread (probably) in the context.  The 
 * server connection holds function pointers for communication procedures 
 * and event callouts.
 */

typedef struct snz_sconn_t snz_sconn_t;

struct snz_sconn_t {
  /* Connection state */
  int tag;
  union {
    struct {
#ifdef USE_STDIO
      FILE *fp_in;
      FILE *fp_out;
#else
      int fd;
#endif
    } socket;
  } u;

  /* Command handlers */
  void (*terminate)( snz_sconn_t *c, int cid, int tid );
  void (*install)( snz_sconn_t *c, int cid, int tid, int etype, int eid );
  void (*uninstall)( snz_sconn_t *c, int cid, int tid, int etype );
  void (*read_data)( snz_sconn_t *, snz_gptr_t, int, void * );
  void (*write_data)( snz_sconn_t *, snz_gptr_t, int, void * );
  void (*call_func)( snz_sconn_t *, void (*fptr)(void), int, snz_arg_t * );

  /* Sending and receiving (for low-level use) */
  void (*send_event)( snz_sconn_t *c, snz_event_t *e );
  void (*recv_cmd)( snz_sconn_t *c, snz_cmd_t *cmd );
  void (*send_reply)( snz_sconn_t *c, int cid, int tid, int code, 
		      void *buf, int len );

  /* Procedures which signal events */
  void (*ready)( snz_sconn_t *, int, int, int );
  void (*err)( snz_sconn_t *, int, int, int, int );
  void (*term)( snz_sconn_t *, int, int, int );
  void (*bbarr)( snz_sconn_t *, int, int, int, int );
  void (*ebarr)( snz_sconn_t *, int, int, int, int );
  void (*bfetch)( snz_sconn_t *, int, int, int, snz_gptr_t, snz_gptr_t,int );
  void (*efetch)( snz_sconn_t *, int, int, int, snz_gptr_t, snz_gptr_t,int );
  void (*bput)( snz_sconn_t *, int, int, int, snz_gptr_t, snz_gptr_t, int );
  void (*eput)( snz_sconn_t *, int, int, int, snz_gptr_t, snz_gptr_t, int );
  void (*brpc)( snz_sconn_t *, int, int, int, snz_gptr_t, int,snz_gptr_t,int);
  void (*erpc)( snz_sconn_t *, int, int, int, snz_gptr_t, int,snz_gptr_t, int);
/* FIXME: Service?? */
  void (*bpar)( snz_sconn_t *, int, int, int );
  void (*epar)( snz_sconn_t *, int, int, int );
  void (*eelt)( snz_sconn_t *, int, int, int, snz_gptr_t );
  void (*xelt)( snz_sconn_t *, int, int, int, snz_gptr_t );
  void (*user)( snz_sconn_t *, int, int, int, int, snz_gptr_t, int );
  void (*bcc)( snz_sconn_t *, int, int, int, int );
  void (*ecc)( snz_sconn_t *, int, int, int, int );
  void (*bdc)( snz_sconn_t *, int, int, int, int );
  void (*edc)( snz_sconn_t *, int, int, int, int );
  void (*efun)( snz_sconn_t *, int, int, int, int, snz_time_t );
  void (*xfun)( snz_sconn_t *, int, int, int, int, snz_time_t );

  void (*push)( snz_sconn_t *c, int eid, int cid, int tid, int code, 
	        void *buf, int len );

};


/* server-remote.c */

struct snz_sconn_t *_snz_create_remote_connection( void );


/* server-sock.c */

void _snz_sock_send_event( snz_sconn_t *, snz_event_t * );
void _snz_sock_recv_cmd( snz_sconn_t *, snz_cmd_t * );
void _snz_sock_send_reply( snz_sconn_t *, int, int, int, void*, int );


/* server-spmd.c et al */

void        _snz_setup( void );
void        _snz_shutdown( void );
void        _snz_whoami( int *cid, int *tid );
int         _snz_get_event_id( int cid, int tid, int _etype );
snz_sconn_t *_snz_lookup_connection( int cid, int tid );
void        _snz_spmd_terminate_thread( void );
int         _snz_spmd_nodenumber( void );
snz_time_t  _snz_get_time( void );
int         _snz_total_nodes( void );
void        *_snz_allocbuf( int len );
void        _snz_freebuf( void *p );

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_SERVER_LOW */

/* eof */
