/* s e r v e r - r e m o t e . c
 * $Id: server-remote.c,v 1.1 1996/05/01 20:09:59 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 3, 1995
 * Modified:
 *
 * Generic server code for non-direct communication.
 */

#define SERVER_SIDE

#include "sneezy.h"
#include "server-low.h"
#include "proxy-sock.h"

/* Command loop reads commands, processes them until continue or terminate */
static void cmdloop( snz_sconn_t *c )
{
  int i;
  snz_cmd_t cmd;
  char *buf;

  while (1) {
    c->recv_cmd( c, &cmd );
    switch ( cmd.cmd ) {
    case SNZC_CONTINUE : 
      return;
    case SNZC_TERMINATE :
      c->terminate( c, cmd.cid, cmd.tid );
      return;
    case SNZC_INSTALL_HANDLER :
      c->install( c, cmd.cid, cmd.tid, cmd.u.install.etype, cmd.u.install.eid);
      break;
    case SNZC_UNINSTALL_HANDLER :
      c->uninstall( c, cmd.cid, cmd.tid, cmd.u.uninstall.etype );
      break;
    case SNZC_READ :
      snz_message( "(%d,%d): READ", cmd.cid, cmd.tid );
      buf = _snz_allocbuf( cmd.u.read.len );
      c->read_data( c, cmd.u.read.g, cmd.u.read.len, buf );
      c->send_reply( c, cmd.cid, cmd.tid, _REPL_READ, buf, cmd.u.read.len );
      _snz_freebuf( buf );
      break;
    case SNZC_WRITE :
      /* Buffer has been allocated for us by comm layer */
      snz_message( "(%d,%d): WRITE", cmd.cid, cmd.tid );
      c->write_data( c, cmd.u.write.g, cmd.u.write.len, cmd.u.write.data );
      _snz_freebuf( cmd.u.write.data );
      break;
    case SNZC_CALL_USER_FUNCTION :
      /* Argument list buffer has been allocated by comm layer */
      snz_message( "(%d,%d): CALL-USER-FUNC %d", cmd.cid, cmd.tid, 
		   cmd.u.call.argc );
      c->call_func( c, cmd.u.call.fptr, cmd.u.call.argc, cmd.u.call.argv );

      for ( i = 0 ; i < cmd.u.call.argc ; i++ )
	if (cmd.u.call.argv[i].tag == SNZAT_CHARPTR)
	  _snz_freebuf( cmd.u.call.argv[i].u.str.cp );
      _snz_freebuf( cmd.u.call.argv );
      c->send_reply( c, cmd.cid, cmd.tid, _REPL_CALL, 0, 0 );
      break;
#if 0
    case SNZC_INSTALL_BREAKPOINT :
      /* Arguments: cid, tid, code address */
      break;
    case SNZC_UNINSTALL_BREAKPOINT :
      /* Arguments: cid, tid, code address */
      break;
#endif
    default :
      break;
    }
  }
}

#define PROLOGUE( code ) \
  snz_event_t e; \
  e.etype = code; \
  e.eid = eid; \
  e.cid = cid; \
  e.tid = tid

#define EPILOGUE() \
  c->send_event( c, &e ); \
  cmdloop( c ) 

static void
remote_ready( snz_sconn_t *c, int eid, int cid, int tid )
{
  PROLOGUE( _SNZE_READY );
  EPILOGUE();
}

static void 
remote_error( snz_sconn_t *c, int eid, int cid, int tid, int code )
{
  PROLOGUE( SNZE_ERROR );
  e.u.err.code = code;
  EPILOGUE();
}

static void 
remote_terminate( snz_sconn_t *c, int eid, int cid, int tid )
{
  PROLOGUE( SNZE_TERMINATE );
  EPILOGUE();
}

static void 
remote_begin_barrier( snz_sconn_t *c, int eid, int cid, int tid, int bid )
{
  PROLOGUE( SNZE_BEGIN_BARRIER );
  e.u.bbarr.bid = bid;
  EPILOGUE();
}

static void 
remote_end_barrier( snz_sconn_t *c, int eid, int cid, int tid, int bid ) 
{
  PROLOGUE( SNZE_END_BARRIER );
  e.u.ebarr.bid = bid;
  EPILOGUE();
}

static void 
remote_begin_fetch( snz_sconn_t *c, int eid, int cid, int tid, snz_gptr_t gs, 
		    snz_gptr_t gd, int len )
{
  PROLOGUE( SNZE_BEGIN_FETCH );
  e.u.bfetch.src = gs;
  e.u.bfetch.dest = gd;
  e.u.bfetch.len = len;
  EPILOGUE();
}

static void 
remote_end_fetch( snz_sconn_t *c, int eid, int cid, int tid, snz_gptr_t gs, 
		  snz_gptr_t gd, int len )
{
  PROLOGUE( SNZE_END_FETCH );
  e.u.efetch.src = gs;
  e.u.efetch.dest = gd;
  e.u.efetch.len = len;
  EPILOGUE();
}

static void
remote_begin_put( snz_sconn_t *c, int eid, int cid, int tid, snz_gptr_t gs, 
		  snz_gptr_t gd, int len )
{
  PROLOGUE( SNZE_BEGIN_PUT );
  e.u.bput.src = gs;
  e.u.bput.dest = gd;
  e.u.bput.len = len;
  EPILOGUE();
}

static void
remote_end_put( snz_sconn_t *c, int eid, int cid, int tid, snz_gptr_t gs,
	        snz_gptr_t gd, int len )
{
  PROLOGUE( SNZE_END_PUT );
  e.u.eput.src = gs;
  e.u.eput.dest = gd;
  e.u.eput.len = len;
  EPILOGUE();
}

static void 
remote_begin_rpc( snz_sconn_t *c, int eid, int cid, int tid, snz_gptr_t gd,
		  int fun, snz_gptr_t gs, int plen )
{
  PROLOGUE( SNZE_BEGIN_RPC );
  e.u.brpc.elt = gd;
  e.u.brpc.param = gs;
  e.u.brpc.paramlen = plen;
  e.u.brpc.code = fun;
  EPILOGUE();
}

static void
remote_end_rpc( snz_sconn_t *c, int eid, int cid, int tid, snz_gptr_t gd,
	        int fun, snz_gptr_t gs, int plen )
{
  PROLOGUE( SNZE_END_RPC );
  e.u.erpc.elt = gd;
  e.u.erpc.param = gs;
  e.u.erpc.paramlen = plen;
  e.u.erpc.code = fun;
  EPILOGUE();
}

#if 0
static void
remote_begin_service_rpc( snz_sconn_t *c, int eid, int cid, int tid, 
			 snz_gptr_t g )
{
  PROLOGUE( SNZE_BEGIN_SERVICE_RPC );
  e.u.bsrpc.dest = g;
  e.u.bsrpc.code = code;
  EPILOGUE();
}

static void
remote_end_service_rpc( snz_sconn_t *c, int eid, int cid, int tid,
		        snz_gptr_t g, int code )
{
  PROLOGUE( SNZE_END_SERVICE_RPC );
  e.u.esrpc.dest = g;
  e.u.esrpc.code = code;
  EPILOGUE();
}
#endif

static void
remote_begin_parallel( snz_sconn_t *c, int eid, int cid, int tid )
{
  PROLOGUE( SNZE_BEGIN_PARALLEL );
  EPILOGUE();
}

static void
remote_end_parallel( snz_sconn_t *c, int eid, int cid, int tid )
{
  PROLOGUE( SNZE_END_PARALLEL );
  EPILOGUE();
}

static void
remote_enter_element( snz_sconn_t *c, int eid, int cid, int tid, 
		     snz_gptr_t elt )
{
  PROLOGUE( SNZE_ENTER_ELEMENT );
  e.u.eelt.elt = elt;
  EPILOGUE();
}

static void
remote_exit_element( snz_sconn_t *c, int eid, int cid, int tid, 
		     snz_gptr_t elt )
{
  PROLOGUE( SNZE_EXIT_ELEMENT );
  e.u.xelt.elt = elt;
  EPILOGUE();
}

static void
remote_user( snz_sconn_t *c, int eid, int cid, int tid, int code, 
	     snz_gptr_t p, int len )
{
  PROLOGUE( SNZE_USER_EVENT );
  e.u.user.code = code;
  e.u.user.p = p;
  e.u.user.len = len;
  EPILOGUE();
}

static void
remote_begin_create_collection( snz_sconn_t *c, int eid, int cid, int tid,
			        int coll_id )
{
  PROLOGUE( SNZE_BEGIN_CREATE_COLLECTION );
  e.u.bcc.cid = coll_id;
  EPILOGUE();
}

static void
remote_end_create_collection( snz_sconn_t *c, int eid, int cid, int tid, 
			     int coll_id )
{
  PROLOGUE( SNZE_END_CREATE_COLLECTION );
  e.u.ecc.cid = coll_id;
  EPILOGUE();
}

static void
remote_begin_delete_collection( snz_sconn_t *c, int eid, int cid, int tid,
			        int coll_id )
{
  PROLOGUE( SNZE_BEGIN_DELETE_COLLECTION );
  e.u.bdc.cid = coll_id;
  EPILOGUE();
}

static void
remote_end_delete_collection( snz_sconn_t *c, int eid, int cid, int tid, 
			     int coll_id )
{
  PROLOGUE( SNZE_END_DELETE_COLLECTION );
  e.u.edc.cid = coll_id;
  EPILOGUE();
}

static void
remote_enter_function( snz_sconn_t *c, int eid, int cid, int tid,
		      int fun_id, snz_time_t time )
{
  PROLOGUE( SNZE_ENTER_FUNCTION );
  e.u.efun.id = fun_id;
  e.u.efun.time = time;
  EPILOGUE();
}

static void
remote_exit_function( snz_sconn_t *c, int eid, int cid, int tid,
		     int fun_id, snz_time_t time )
{
  PROLOGUE( SNZE_EXIT_FUNCTION );
  e.u.xfun.id = fun_id;
  e.u.xfun.time = time;
  EPILOGUE();
}

static void
remote_push_value( snz_sconn_t *c, int eid, int cid, int tid, 
		   int code, void *value, int len )
{
  snz_event_t e;
  e.etype = _SNZE_PUSH;
  e.eid = eid;
  e.cid = cid;
  e.tid = tid;
  e.u.push.code = code;
  e.u.push.value = value;
  e.u.push.len = len;
  c->send_event( c, &e );
}

/* Create a remote connection object */
snz_sconn_t *_snz_create_remote_connection( void )
{
  snz_sconn_t *c;

  c = (snz_sconn_t*)malloc( sizeof( snz_sconn_t ) );
  if (c == 0) 
    snz_perror_panic( "_snz_create_remote_connection: malloc" );
  c->ready = remote_ready;
  c->err = remote_error;
  c->term = remote_terminate;
  c->bbarr = remote_begin_barrier;
  c->ebarr = remote_end_barrier;
  c->bfetch = remote_begin_fetch;
  c->efetch = remote_end_fetch;
  c->bput = remote_begin_put;
  c->eput = remote_end_put;
  c->brpc = remote_begin_rpc;
  c->erpc = remote_end_rpc;
/*
  c->bsrpc = remote_begin_service_rpc;
  c->esrpc = remote_end_service_rpc;
*/
  c->bpar = remote_begin_parallel;
  c->epar = remote_end_parallel;
  c->eelt = remote_enter_element;
  c->xelt = remote_exit_element;
  c->user = remote_user;
  c->bcc = remote_begin_create_collection;
  c->ecc = remote_end_create_collection;
  c->bdc = remote_begin_delete_collection;
  c->edc = remote_end_delete_collection;
  c->efun = remote_enter_function;
  c->xfun = remote_exit_function;
  c->push = remote_push_value;
  c->send_event = 0;
  c->recv_cmd = 0;
  c->send_reply = 0;
  c->terminate = 0;
  c->install = 0;
  c->uninstall = 0;
  c->read_data = 0;
  c->write_data = 0;
  c->tag = TAG_NONE;
  return c;
}

/* eof */
