/* s e r v e r - s o c k . c
 * $Id: server-sock.c,v 1.1 1996/05/01 20:09:59 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 6, 1995
 * Modified:
 *
 * Low-level comm layer for communicating raw events over sockets.
 */

#define SERVER_SIDE

#include <stdio.h>
#include "sneezy.h"
#include "server-low.h"
#include "socket.h"

#ifdef USE_STDIO
#error "STDIO no longer supported.  You're not on Un*x?  Too bad, so sad."
#endif

void _snz_sock_send_event( snz_sconn_t *c, snz_event_t *e )
{
  if (c == 0)
    snz_panic( "_snz_sock_send_event: null client!" );

  if (!write_bytes( c->u.socket.fd, (void*)e, sizeof( snz_event_t ) ))
    snz_panic( "write_bytes" );

  if (e->etype == _SNZE_PUSH) {
    if (!write_bytes( c->u.socket.fd, e->u.push.value, e->u.push.len ))
      snz_panic( "write_bytes" );
  }
}


void _snz_sock_recv_cmd( snz_sconn_t *c, snz_cmd_t *cmd )
{
  int l, i;
  void *p;

  if (c == 0)
    snz_panic( "_snz_sock_recv_cmd: null client!" );

  if (!read_bytes( c->u.socket.fd, (void*)cmd, sizeof(snz_cmd_t) ))
    snz_panic( "_snz_sock_recv_cmd: read_bytes" );

  if (cmd->cmd == SNZC_WRITE) {
    l = cmd->u.write.len;
    p = cmd->u.write.data = _snz_allocbuf( l );
    if (!read_bytes( c->u.socket.fd, (void*)p, l ))
      snz_perror_panic( "read" );
  }

  if (cmd->cmd == SNZC_CALL_USER_FUNCTION) {
    l = cmd->u.call.argc;
    p = cmd->u.call.argv = _snz_allocbuf( l*sizeof( snz_arg_t ) );
    if (!read_bytes( c->u.socket.fd, (void*)p, l*sizeof(snz_arg_t) ))
      snz_perror_panic( "read" );

    /* Strings follow the argv structure in order of the string arguments in
     * the argv structure: length field in native size/order followed by bytes.
     * Alloc buffers for all; these will be freed by same agent which frees
     * argv buffer.
     */

    for ( i = 0 ; i < cmd->u.call.argc ; i++ )
      if (cmd->u.call.argv[i].tag == SNZAT_CHARPTR) {
	if (!read_bytes( c->u.socket.fd, 
			 (void*)&cmd->u.call.argv[i].u.str.len,
			 sizeof( int ) ))
	  snz_perror_panic( "read" );
	cmd->u.call.argv[i].u.str.cp = _snz_allocbuf( cmd->u.call.argv[i].u.str.len);
	if (!read_bytes( c->u.socket.fd,
			 (void*)cmd->u.call.argv[i].u.str.cp,
			 cmd->u.call.argv[i].u.str.len ))
	  snz_perror_panic( "read" );
      }
  }
}

void _snz_sock_send_reply( snz_sconn_t *c, int cid, int tid, int tag,
			   void *buf, int len )
{
  snz_event_t e;

  if (c == 0)
    snz_panic( "_snz_sock_send_reply: null client!" );
  e.etype = _SNZE_REPLY;
  e.eid = 2;
  e.cid = cid;
  e.tid = tid;
  e.u.reply.tag = tag;
  e.u.reply.len = len;
  e.u.reply.data = 0;

  if (!write_bytes( c->u.socket.fd, (void*)&e, sizeof(e) ))
    snz_panic( "_snz_sock_send_reply: write_bytes" );
  if (buf != 0)
    if (!write_bytes( c->u.socket.fd, (void*)buf, len ))
      snz_panic( "_snz_sock_send_reply: write_bytes" );
}

/* eof */
