/* s e x p r . h
 * $Id: sexpr.h,v 1.1 1996/05/01 20:10:03 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November, 1995
 * Modified:
 *
 * S-expression parser: public definitions.
 */

#if !defined( SERVER_SIDE ) && !defined( CLIENT_SIDE )
#error "Must predefine SERVER_SIDE or CLIENT_SIDE"
#endif

#ifndef INCLUDED_SEXPR_H
#define INCLUDED_SEXPR_H

#ifdef __cplusplus
extern "C" {
#endif

#ifdef SERVER_SIDE
#include "server-types.h"
#endif

typedef struct sexpr_obj_t sexpr_obj_t;
typedef struct sexpr_seg_t sexpr_seg_t;

struct _sexpr_str {
  char *data;
  int length;
};

/* NIL is not an object, but a 0 pointer. */

struct sexpr_obj_t {
  union {
    snz_long_t num;
    struct _sexpr_str str;          /* Should this be a snz_str_t? */
    struct {
      sexpr_obj_t *car;
      sexpr_obj_t *cdr;
    } pair;
  } u;
  short tag;
};

#define TAG_PAIR   0
#define TAG_NUMBER 1
#define TAG_STRING 2
#define TAG_LP     100
#define TAG_RP     200
#define TAG_ERROR  300

extern sexpr_obj_t *_sexpr_panic(const char*,int,sexpr_obj_t*);

#ifdef FULL_TYPECHECKS

/* Workaround for SGI IRIX6.01 (at least) 64-bit compiler bug.
 * See the file irix-bug.c for details.
 */
#if defined(sgi) && defined(mips) && defined(_MIPS_SZPTR) && _MIPS_SZPTR==64
static sexpr_obj_t *_check2(sexpr_obj_t*m, int t, char *name )
{
  if (m && m->tag == t)
    return m;
  else
    return _sexpr_panic( name, t, m );
}
#define _check(m,t,op,name) (_check2(m,t,name)->op)
#else
#define _check(m,t,op,name) \
   ((m&&m->tag==t)?(m->op):(_sexpr_panic(name,t,m)->op))
#endif

/* Using these definitions cause a very rapid blowup in the size of the
 * definitions of caar, etc.  Pretty amazing to watch the compiler struggle...
 * So for the checked version, compile primitives into functions.
 */
#define _car( m ) _check((m),TAG_PAIR,u.pair.car,"car")
#define _cdr( m ) _check((m),TAG_PAIR,u.pair.cdr,"cdr")
#define _num( m ) _check((m),TAG_NUMBER,u.num,"num")
#define _str( m ) _check((m),TAG_STRING,u.str,"str")

static sexpr_obj_t *car( sexpr_obj_t *p ) { return _car(p); }
static sexpr_obj_t *cdr( sexpr_obj_t *p ) { return _cdr(p); }
static snz_long_t num( sexpr_obj_t *p ) { return _num(p); }
static struct _sexpr_str str( sexpr_obj_t *p ) { return _str(p); }
#else
#define car( m ) ((m)->u.pair.car)
#define cdr( m ) ((m)->u.pair.cdr)
#define num( m ) ((m)->u.num)
#define str( m ) ((m)->u.str)
#endif

#define caar( m ) car(car(m))
#define cadr( m ) car(cdr(m))
#define cddr( m ) cdr(cdr(m))
#define cadar( m ) car(cdr(car(m)))
#define caddr( m ) car(cdr(cdr(m)))
#define cdddr( m ) cdr(cdr(cdr(m)))
#define cadddr( m ) car(cdr(cdr(cdr(m))))
#define caddddr( m ) car(cdr(cdr(cdr(cdr(m)))))

typedef struct {
  sexpr_obj_t *freelist;
  sexpr_seg_t *segments;
} sexpr_mem_t;


sexpr_mem_t *sexpr_alloc_mem( void );
void sexpr_clear_mem( sexpr_mem_t *p );
void sexpr_release_mem( sexpr_mem_t *mem );
#ifdef CLIENT_SIDE
sexpr_obj_t *sexpr_parse( snz_conn_t *c, sexpr_mem_t *mem );
#else
sexpr_obj_t *sexpr_parse( FILE *fp, sexpr_mem_t *mem );
#endif

int length( sexpr_obj_t * );

char *sexpr_encode_data( unsigned char *d, unsigned int len );
unsigned char *sexpr_decode_data( char *d, unsigned int *len );

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_SEXPR */

/* eof */
