/* s n e e z y . h
 * $Id: sneezy.h,v 1.1 1996/05/01 20:10:06 lth Exp $
 *
 * Copyright (C) 1995 The University of Oregon.
 *
 *   Author: Lars Thomas Hansen [lth@cs.uoregon.edu]
 *  Created: November 9, 1995
 * Modified:
 *
 * Basic functionality and types common to client and server.
 */

#if !defined( SERVER_SIDE ) && !defined( CLIENT_SIDE )
#error "Must predefine SERVER_SIDE or CLIENT_SIDE"
#endif

#ifndef INCLUDED_SNEEZY
#define INCLUDED_SNEEZY

#ifdef __cplusplus
extern "C" {
#endif

/* A global pointer is represented as context ID and local address */

#ifdef SERVER_SIDE
#include "server-types.h"
#endif

#ifdef CLIENT_SIDE
#include "client-types.h"
#endif

snz_gptr_t snz_mkgptr( int, snz_addr_t );

#include "proginfo.h"

#ifdef SERVER_SIDE
/* The command structure snz_cmd_t encodes a command. */

typedef struct {
  short cmd;                /* The command identifier */
  short cid;                /* The context ID */
  short tid;                /* The thread ID */
  union {
    struct {
      int etype;            /* Event type */
      int eid;              /* Event ID being installed */
    } install;
    struct {
      int etype;            /* Event type */
    } uninstall;
    struct {
      snz_gptr_t g;         /* global pointer */
      int len;              /* length */
    } read;
    struct {
      snz_gptr_t g;         /* global pointer */
      int len;              /* length */
      unsigned char *data;  /* data pointer (must free!) */
    } write;
    struct {
      void (*fptr)(void);   /* Function address */
      int  argc;            /* Argument length */
      snz_arg_t *argv;      /* Arguments */
    } call;
  } u;
} snz_cmd_t;
#endif

/* These are the command identifiers */

#define SNZC_CONTINUE 0
#define SNZC_TERMINATE 1
#define SNZC_INSTALL_HANDLER 2
#define SNZC_UNINSTALL_HANDLER 3
#define SNZC_READ 4
#define SNZC_WRITE 5
#define SNZC_GET_PARSECT_INFO 6
#define SNZC_GET_ELEM_TYPE 7
#define SNZC_GET_COLL_INFO 8
#define SNZC_CALL_USER_FUNCTION 9
#define SNZC_INSTALL_BREAKPOINT 10
#define SNZC_UNINSTALL_BREAKPOINT 11

/* The event structure snz_event_t encodes an event. */

typedef struct {
  short etype;
  short eid;
  short cid;
  short tid;
  union {
    /* READY has no data */ 
    struct {
      snz_proginfo_t *info;  /* Startup info */
    } startup;
    struct {
      short tag;
      int len;
      void *data;
    } reply;
    struct {
      int code;
      void *value;
      int len;
    } push;
    struct {
      short code;      /* error code */
    } err;
    /* No data for terminate */
    struct {
      short bid;       /* Barrier ID */
    } bbarr;
    struct {
      short bid;       /* Barrier ID */
    } ebarr;
    struct {
      snz_gptr_t src;  /* Fetch from */
      snz_gptr_t dest; /* Put to */
      int        len;  /* # bytes */
    } bfetch;
    struct {
      snz_gptr_t src;  /* Fetch from */
      snz_gptr_t dest; /* Put to */
      int        len;  /* # bytes */
    } efetch;
    struct {
      snz_gptr_t src;   /* Get from */
      snz_gptr_t dest;  /* Put to */
      int        len;   /* # bytes */
    } bput;
    struct {
      snz_gptr_t src;   /* Get from */
      snz_gptr_t dest;  /* Put to */
      int        len;   /* # bytes */
    } eput;
    struct {
      snz_gptr_t elt;      /* Element on which RPC is invoked */
      snz_gptr_t param;    /* parameter buffer */
      int        paramlen; /* # bytes */
      int        code;     /* Function code */
    } brpc;
    struct {
      snz_gptr_t elt;   /* Element on which RPC was invoked */
      snz_gptr_t param;    /* parameter buffer */
      int        paramlen; /* # bytes */
      int        code;  /* Function code */
    } erpc;
    struct {
      snz_gptr_t dest;  /* Element which is target of request */
      int        code;  /* Function code */
    } bsrpc;
    struct {
      snz_gptr_t dest;  /* Element which was target of request */
      int        code;  /* Function code */
    } esrpc;
    /* No data for begin-parallel */
    /* No data for end-parallel */
    struct {
      snz_gptr_t elt;   /* Element entered */
    } eelt;
    struct {
      snz_gptr_t elt;   /* Element exited */
    } xelt;
    struct {
      int code;         /* User event code */
      snz_gptr_t p;     /* Pointer to data */
      int len;          /* Length of data */
    } user;
    struct {
      int cid;          /* Collection ID */
    } bcc;
    struct {
      int cid;          /* Collection ID */
    } ecc;
    struct {
      int cid;          /* Collection ID */
    } bdc;
    struct {
      int cid;          /* Collection ID */
    } edc;
    struct {
      int id;           /* Function ID */
      snz_time_t time;  /* Timestamp */
    } efun;
    struct {
      int id;           /* Function ID */
      snz_time_t time;  /* Time stamp */
    } xfun;
  } u;
} snz_event_t;


/* The event types */

#define _SNZE_PUSH                   -4
#define _SNZE_READY                  -3
#define _SNZE_STARTUP                -2
#define _SNZE_REPLY                  -1
#define SNZE_ERROR                   0
#define SNZE_TERMINATE               1
/* #define SNZE_CREATE_COLLECTION 2 */ /* Obsolete */
/* #define SNZE_DELETE_COLLECTION 3 */ /* Obsolete */
#define SNZE_BEGIN_BARRIER           4
#define SNZE_END_BARRIER             5
#define SNZE_BEGIN_FETCH             6
#define SNZE_END_FETCH               7
#define SNZE_BEGIN_PUT               8
#define SNZE_END_PUT                 9
#define SNZE_BEGIN_RPC               10
#define SNZE_END_RPC                 11
#define SNZE_BEGIN_PARALLEL          12
#define SNZE_ENTER_ELEMENT           13
#define SNZE_EXIT_ELEMENT            14
#define SNZE_BEGIN_SERVICE_RPC       15
#define SNZE_END_SERVICE_RPC         16
#define SNZE_USER_EVENT              17
#define SNZE_BEGIN_CREATE_COLLECTION 18
#define SNZE_END_CREATE_COLLECTION   19
#define SNZE_BEGIN_DELETE_COLLECTION 20
#define SNZE_END_DELETE_COLLECTION   21
#define SNZE_END_PARALLEL            22
#define SNZE_ENTER_FUNCTION          23
#define SNZE_EXIT_FUNCTION           24

#define SNZ_MAX_EVENTS 25

/* Predefined event IDs */

#define SNZEID_READY 0
#define SNZEID_TERM  1
#define SNZEID_REPLY 2
#define SNZEID_PUSH  3

#include "message.h"

#ifdef __cplusplus
}
#endif

#endif /* ifndef INCLUDED_SNEEZY */

/* eof */
