/* t e s t - c l i e n t . c
 * $Id: test-client.c,v 1.1 1996/05/01 20:10:10 lth Exp $
 *
 * Client test program for Sneezy C API implementation.
 */

#define CLIENT_SIDE

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "client-api.h"

#include "socket.h"  /* for myaddr() */

#define DLEN 4

int numcontexts;   /* Number of contexts in program*/
int ready_cnt;     /* Number of contexts READY */
int term_cnt;      /* Number of contexts TERMINATED */
int *live_cnta;    /* Number of live threads, one slot per context. */
int b4cnt;         /* Number of contexts in barrier 4 */
int b5cnt;
snz_gptr_t *u;
int *l;
int kcnt;

msg( char *s, ... )
{
  va_list args;

  va_start( args, s );
  vfprintf( stderr, s, args );
  va_end( args );
  fputc( '\n', stderr );
  fflush( stderr );
  return 0;
}

/* Intelligent handlers */

/* Counts ready contexts until all have signalled ready. */
int h_ready( snz_conn_t *c, int cid, int tid )
{
  msg( "READY %d %d", cid, tid );
  return (++ready_cnt < numcontexts);
}

/* This handler will just count termination events until all threads in
 * all contexts have terminated. There are several ways to do this; this
 * is the hard way (it can be put to good use).
 */
int h_term( snz_conn_t *c, int cid, int tid )
{
  msg( "TERMINATE %d %d", cid, tid );
  if (--live_cnta[cid] == 0) {
    /* Context terminated: we could synthesize event here. */
    return (++term_cnt != numcontexts);
  }
  else
    return 1;
}

/* Barrier #4 is treated specially */

int h_bbarr( snz_conn_t *c, int cid, int tid, int bid ) 
{ 
  msg( "BEGIN-BARRIER %d %d %d", cid, tid, bid ); 
  if (bid != 4 && bid != 5) {
    snz_continue( c, cid, tid );
    return 1;
  }
  if (bid == 4)
    return (++b4cnt < numcontexts);
  else
    return (++b5cnt < numcontexts);
}


/* Stupid handlers */

int h_ebarr( snz_conn_t *c, int cid, int tid, int bid ) { 
  msg( "END-BARRIER %d %d %d", cid, tid, bid ); 
  snz_continue( c, cid, tid );
  return 1;
}

int h_bfetch( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, snz_gptr_t gd, 
	      int len ) { 
  char sp[20],gp[20];

  snz_addr2str( gs.addr, sp );
  snz_addr2str( gd.addr, gp );
  msg( "BEGIN-FETCH %d %d (%d %s) (%d %s) %d", cid, tid, gs.context, sp,
       gd.context, gp, len );
  snz_continue( c, cid, tid );
  return 1;
}

int h_efetch( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, snz_gptr_t gd, 
	      int len ) { 
  char sp[20],gp[20];

  snz_addr2str( gs.addr, sp );
  snz_addr2str( gd.addr, gp );
  msg( "END-FETCH %d %d (%d %s) (%d %s) %d", cid, tid, gs.context, sp,
       gd.context, gp, len );
  snz_continue( c, cid, tid );
  return 1;
}

int h_bput( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, snz_gptr_t gd, 
	      int len ) { 
  char sp[20],gp[20];

  snz_addr2str( gs.addr, sp );
  snz_addr2str( gd.addr, gp );
  msg( "BEGIN-PUT %d %d (%d %s) (%d %s) %d", cid, tid, gs.context, sp,
       gd.context, gp, len );
  snz_continue( c, cid, tid );
  return 1;
}

int h_eput( snz_conn_t *c, int cid, int tid, snz_gptr_t gs, snz_gptr_t gd, 
	      int len ) { 
  char sp[20],gp[20];

  snz_addr2str( gs.addr, sp );
  snz_addr2str( gd.addr, gp );
  msg( "END-PUT %d %d (%d %s) (%d %s) %d", cid, tid, gs.context, sp,
       gd.context, gp, len );
  snz_continue( c, cid, tid );
  return 1;
}

int h_brpc( snz_conn_t *c, int cid, int tid, snz_gptr_t param, int paramlen,
	   snz_gptr_t elt, int code ) { 
  char sp[20],gp[20];

  snz_addr2str( param.addr, sp );
  snz_addr2str( elt.addr, gp );
  msg( "BEGIN-RPC %d %d (%d %s) %d (%d %s) %d", cid, tid, param.context, sp,
       paramlen, elt.context, gp, code );
  snz_continue( c, cid, tid );
  return 1;
}

int h_erpc( snz_conn_t *c, int cid, int tid, snz_gptr_t param, int paramlen,
	   snz_gptr_t elt, int code ) { 
  char sp[20],gp[20];

  snz_addr2str( param.addr, sp );
  snz_addr2str( elt.addr, gp );
  msg( "END-RPC %d %d (%d %s) %d (%d %s) %d", cid, tid, param.context, sp,
       paramlen, elt.context, gp, code );
  snz_continue( c, cid, tid );
  return 1;
}

int h_bpar( snz_conn_t *c, int cid, int tid ) { 
  msg( "BEGIN-PARALLEL %d %d", cid, tid );
  snz_continue( c, cid, tid );
  return 1;
}

int h_epar( snz_conn_t *c, int cid, int tid ) { 
  msg( "END-PARALLEL %d %d", cid, tid );
  snz_continue( c, cid, tid );
  return 1;
}

int h_eelt( snz_conn_t *c, int cid, int tid, snz_gptr_t elt )
{
  char p[20];

  snz_addr2str( elt.addr, p );
  msg( "ENTER-ELEMENT %d %d (%d %s)", cid, tid, elt.context, p );
  snz_continue( c, cid, tid );
  return 1;
}

int h_xelt( snz_conn_t *c, int cid, int tid, snz_gptr_t elt )
{
  char p[20];

  snz_addr2str( elt.addr, p );
  msg( "EXIT-ELEMENT %d %d (%d %s)", cid, tid, elt.context, p );
  snz_continue( c, cid, tid );
  return 1;
}

int h_user( snz_conn_t *c, int cid, int tid, int code, snz_gptr_t p, int len ) 
{
  char s[20];

  snz_addr2str( p.addr, s );
  msg( "USER-EVENT %d %d %d (%d %s) %d", cid, tid, code, p.context, s, len );
  u[cid] = p;
  l[cid] = len;
  snz_continue( c, cid, tid );
  return 1;
}

int h_bcc( snz_conn_t *c, int cid, int tid, int coll_id ) {
  msg( "BEGIN-CREATE-COLLECTION %d %d %d", cid, tid, coll_id );
  snz_continue( c, cid, tid );
  return 1;
}

int h_ecc( snz_conn_t *c, int cid, int tid, int coll_id ) {
  msg( "END-CREATE-COLLECTION %d %d %d", cid, tid, coll_id );
  snz_continue( c, cid, tid );
  return 1;
}

int h_bdc( snz_conn_t *c, int cid, int tid, int coll_id ) {
  msg( "BEGIN-DELETE-COLLECTION %d %d %d", cid, tid, coll_id );
  snz_continue( c, cid, tid );
  return 1;
}

int h_edc( snz_conn_t *c, int cid, int tid, int coll_id ) {
  msg( "END-DELETE-COLLECTION %d %d %d", cid, tid, coll_id );
  snz_continue( c, cid, tid );
  return 1;
}

int h_efun( snz_conn_t *c, int cid, int tid, int fun_id, snz_time_t time )
{
  char s[30];
  snz_time2str( time, s );
  msg( "ENTER-FUNCTION %d %d %d %s", cid, tid, fun_id, s );
  snz_continue( c, cid, tid );
  return 1;
}

int h_xfun( snz_conn_t *c, int cid, int tid, int fun_id, snz_time_t time )
{
  char s[30];
  snz_time2str( time, s );
  msg( "EXIT-FUNCTION %d %d %d %s", cid, tid, fun_id, s );
  snz_continue( c, cid, tid );
  return 1;
}

/* Utilities */

/* Continue distringuished thread in each context */
continue_all_contexts( snz_conn_t *c )
{
  int i;

  for ( i = 0 ; i < numcontexts ; i++ )
    snz_continue( c, i, 0 );
  return 0;
}

/* Enable the given handler in all known threads */
enable_all_handlers( snz_conn_t *c, int etype )
{
  int i, j;

  /* This is not quite right: we need to keep track of _which_ threads
   * are alive in each context.
   */
  for ( i=0 ; i< numcontexts ; i++ )
    for ( j=0 ; j < live_cnta[i] ; j++ )
      snz_install_handler( c, i, j, etype );
  return 0;
}

step_all( snz_conn_t *c )
{
  char buf[10];

  msg( "press key to continue all" );
  gets( buf );
  continue_all_contexts( c );

  msg( "enter wait." );
  snz_wait( c, 0 );
  msg( "exit wait." );
  return 0;
}

int read_k( snz_conn_t *c, int cid, int tid, void *data )
{
  msg( "read_k: %d %d %s", cid, tid, sexpr_encode_data( (char*)data, DLEN ) );
  return 0;
}

int read_k_cinfo( snz_conn_t *c, int cid, int tid, void *data )
{
  msg( "read_k_cinfo: %d %d %s", cid, tid, (char *)data );
  return (++kcnt < numcontexts);
}

int main( int argc, char **argv )
{
  snz_hstruct_t hstruct;     /* Handler structure */
  snz_conn_t *c;             /* The connection */
  snz_proginfo_t *pi;        /* The program info */
  int i;
  unsigned int one = 0xDEADBEEF;
  char *s;
  int port;
  char buf[10];

  if (argc == 2) 
    s = argv[1];
  else 
    s = getenv( "SNZPORT" );

  if (s != 0) port = atoi( s ); else port = 5666;
  snz_null_hstruct( &hstruct );

  c = snz_make_connection();
  hstruct.ready = h_ready;
  snz_setup_handlers( c, &hstruct );

  msg( "connecting" );
  snz_connect_socket( c, myaddr(), port );

  msg( "getting program info" );
  pi = snz_program_info( c );
  printf( "got program info: %d;", pi->numcontexts );
  for ( i = 0 ; i < pi->numcontexts ; i++ )
    printf( " %d", pi->numthreads[i] );
  printf( "\n" );
  fflush( stdout );
  
  numcontexts = pi->numcontexts;
  ready_cnt = 0;
  term_cnt = 0;
  live_cnta = (int*)malloc( sizeof(int)*numcontexts );
  for (i = 0 ; i < numcontexts ; i++ )
    live_cnta[i] = pi->numthreads[i];
  u = (snz_gptr_t*)malloc( sizeof(snz_gptr_t)*numcontexts );
  l = (int*)malloc( sizeof(int)*numcontexts );

  msg( "wait" );
  snz_wait( c, 0 );  /* wait for all to be ready */
  msg( "end wait" );

  /* All contexts are now waiting */

  hstruct.term = h_term;
  hstruct.bbarr = h_bbarr;
  hstruct.ebarr = h_ebarr;
  hstruct.bfetch = h_bfetch;
  hstruct.efetch = h_efetch;
  hstruct.bput = h_bput;
  hstruct.eput = h_eput;
  hstruct.brpc = h_brpc;
  hstruct.erpc = h_erpc;
  hstruct.bpar = h_bpar;
  hstruct.epar = h_epar;
  hstruct.eelt = h_eelt;
  hstruct.xelt = h_xelt;
  hstruct.user = h_user;
  hstruct.bcc = h_bcc;
  hstruct.ecc = h_ecc;
  hstruct.bdc = h_bdc;
  hstruct.edc = h_edc;
  hstruct.efun = h_efun;
  hstruct.xfun = h_xfun;
  snz_setup_handlers( c, &hstruct );

  enable_all_handlers( c, SNZE_TERMINATE );
  enable_all_handlers( c, SNZE_BEGIN_BARRIER );
  enable_all_handlers( c, SNZE_END_BARRIER );
  enable_all_handlers( c, SNZE_BEGIN_FETCH );
  enable_all_handlers( c, SNZE_END_FETCH );
  enable_all_handlers( c, SNZE_BEGIN_PUT );
  enable_all_handlers( c, SNZE_END_PUT );
  enable_all_handlers( c, SNZE_BEGIN_RPC );
  enable_all_handlers( c, SNZE_END_RPC );
  enable_all_handlers( c, SNZE_BEGIN_PARALLEL );
  enable_all_handlers( c, SNZE_END_PARALLEL );
  enable_all_handlers( c, SNZE_ENTER_ELEMENT );
  enable_all_handlers( c, SNZE_EXIT_ELEMENT );
  enable_all_handlers( c, SNZE_USER_EVENT );
  enable_all_handlers( c, SNZE_BEGIN_CREATE_COLLECTION );
  enable_all_handlers( c, SNZE_END_CREATE_COLLECTION );
  enable_all_handlers( c, SNZE_BEGIN_DELETE_COLLECTION );
  enable_all_handlers( c, SNZE_END_DELETE_COLLECTION );
  enable_all_handlers( c, SNZE_ENTER_FUNCTION );
  enable_all_handlers( c, SNZE_EXIT_FUNCTION );

  step_all( c );
  msg( "all should now be at barrier 4." );

  for ( i = 0 ; i < numcontexts ; i++ ) {
    msg( "read" );
    snz_read_data( c, i, 0, u[i].addr, DLEN, read_k );
    msg( "begin wait" );
    snz_wait( c, 0 );
    msg( "end wait" );
    msg( "begin write" );
    snz_write_data( c, i, 0, u[i].addr, DLEN, (void*)&one );
    msg( "end write" );
    msg( "read" );
    snz_read_data( c, i, 0, u[i].addr, DLEN, read_k );
    msg( "begin wait" );
    snz_wait( c, 0 );
    msg( "end wait" );
  }

  step_all( c );
  msg( "all should now be at barrier 5" );

  /* At this point, u and l contain access info for collection info, so
     we need to get it. */
  kcnt = 0;
  for ( i = 0 ;i < numcontexts ; i++ )
    snz_read_data( c, i, 0, u[i].addr, l[i], read_k_cinfo );
  snz_wait( c, 0 );

  /* Disable handlers for the RPC events */
  for ( i = 0 ; i < numcontexts ; i++ ) {
    if (i & 1)
      snz_uninstall_handler( c, i, 0, SNZE_BEGIN_RPC );
    else
      snz_uninstall_handler( c, i, 0, SNZE_END_RPC );
  }

  step_all( c );
  msg( "all should now have terminated." );

  /* All contexts are now terminated but still responding */

  msg( "press ENTER to continue all." );
  gets( buf );
  continue_all_contexts( c );

  /* All should now be dead */

  /* snz_client_terminate( c ); */

  snz_close( c );
  return 0;
}
