/* t e s t - p r o g r a m . c
 * $Id: test-program.c,v 1.1 1996/05/01 20:10:13 lth Exp $
 *
 * Simulates a parallel SPMD-mode program; tests the sneezy implementation.
 */

#define SERVER_SIDE

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include "server-api.h"

#define min(a,b) ((a) < (b) ? (a) : (b))

int MyTotalNodes = 2;  /* total number of compute contexts */
int MyNodeNumber;      /* local context number */

void usage( char *s )
{
  fprintf( stderr, "usage: %s [-n numprocs] [portno]\n", s );
  exit( 1 );
}

int main( int argc, char **argv )
{
  char *s = 0;
  int port;
  int i;

  for ( i = 1 ; i < argc ; ) {
    if (strcmp( argv[i], "-n" ) == 0) {
      if (i == argc-1) 
	usage( *argv );
      else {
	MyTotalNodes = atoi( argv[i+1] );
	i+=2;
      }
    }
    else if (s == 0)
      s = argv[i++];
    else
      usage( *argv );
  }

  if (s == 0) 
    s = getenv( "SNZPORT" );
  if (s != 0) 
    port = atoi(s); 
  else
    port=5666;

  snz_server_init_agent( port, 0 );

  /* Go parallel */
  
  fork_thyself( pcxx_TotalNodes() );

  /* Now in n copies */

  msg( "Registering." );
  snz_register_context();
  msg( "Entering Processor_Main." );
  Processor_Main();
  msg( "Leaving Processor_Main." );
  snz_shutdown_context();
  msg( "Exiting." );
  exit(0);
}

msg( char *m )
{
  fprintf( stderr, "[%d: %s]\n", pcxx_NodeNumber(), m );
  fflush( stderr );
  return 0;
}

/* IRIX defines barrier() */
the_barrier( int n )
{
  msg( "Begin barrier." );    snz_begin_barrier_event( n );
  pcxx_barrier();
  msg( "End barrier." );      snz_end_barrier_event( n );
  return 0;
}

fetch( int node, void *addr, void *local, int len )
{
  msg( "Begin fetch." ); snz_begin_fetch_event( node, addr, local, len );
  sleep( 1 );
  msg( "End fetch." );   snz_end_fetch_event( node, addr, local, len );
  return 0;
}

put( int node, void *addr, void *local, int len )
{
  msg( "Begin put." ); snz_begin_fetch_event( node, addr, local, len );
  sleep( 1 );
  msg( "End put." );   snz_end_fetch_event( node, addr, local, len );
  return 0;
}

rpc( int node, void *this, int memp, void *params, int plen )
{
  msg( "Begin rpc." ); snz_begin_rpc_event( node, this, memp, params, plen );
  sleep( 1 );
  msg( "End rpc." );   snz_end_rpc_event( node, this, memp, params, plen );
  return 0;
}

struct collection {
  int  nelements;
  int  id;
  int  eltsize;
  int  firstlocal;
  int  lastlocal;
  void *data;
};

struct collection *create_collection1( int id, int size, int eltsize )
{
  struct collection *c;
  int nodes, me, pernode, first, last, i;

  nodes   = pcxx_TotalNodes();
  me      = pcxx_NodeNumber();
  pernode = (size+nodes-1) / nodes;
  first   = me*pernode;
  last    = min( first+pernode-1, size-1 );

  msg( "Begin-create-collection." ); snz_begin_create_collection_event( id );
  pcxx_barrier();
  c = (struct collection *)malloc( sizeof( struct collection ) );
  c->nelements = size;
  c->id = id;
  c->eltsize = eltsize;
  c->firstlocal = first;
  c->lastlocal = last;
  c->data = malloc( (last-first+1)*eltsize );
  snz_message( "%d: First: %d last: %d", me, first, last );
  pcxx_barrier();
  msg( "End-create-collection." );   snz_end_create_collection_event( id );
  return c;
}

delete_collection( struct collection *c )
{
  msg( "Begin-delete-collection." );
  snz_begin_delete_collection_event( c->id );
  pcxx_barrier();
  free( c->data );
  c->data = 0;
  pcxx_barrier();
  msg( "End-delete-collection." );
  snz_end_delete_collection_event( c->id );
  return 0;
}

void *coll_elt_addr( struct collection *c, int i )
{
  if (i >= c->firstlocal && i <= c->lastlocal)
    return (void*)((char*)(c->data)+(i-c->firstlocal)*c->eltsize);
  else {
    snz_panic( "COLL_ELT_ADDR: %d %d %d", i, c->firstlocal, c->lastlocal );
    /*NOTREACHED*/
    return 0;
  }
}

/* Get some data about the local part of the given collection in a string */
char *get_collection_info( struct collection *c )
{
  static char buf[ 256 ];

  /* ID FIRSTLOCAL LASTLOCAL ELTSIZE DATAPTR */
  sprintf( buf, "%d %d %d %d #x%lx", 
	   c->id, c->firstlocal, c->lastlocal, c->eltsize,
	   (unsigned long)(c->data) );
  return buf;
}

parsect( struct collection *c )
{
  int i;
  int *p;

  snz_begin_parallel_event();
  for ( i = c->firstlocal ; i <= c->lastlocal ; i++ ) {
    p = (int*)(coll_elt_addr(c,i));
    snz_enter_element_event( (void*)p );
    *p = pcxx_NodeNumber()*100+i;
    sleep( 1 );
    snz_exit_element_event( (void*)p );
  }
  the_barrier( 3 );
  snz_end_parallel_event();
  return 0;
}


int factfun( int n )
{
  int _r;
  msg( "Enter-fun" ); snz_enter_function_event( 33 );
  if ( n < 2 ) {
    _r = 1; goto end;
  }
  else {
    _r = n*factfun( n-1 );
    goto end;
  }
 end:
  msg( "Exit-fun" ); snz_exit_function_event( 33 );
  return _r;
}

void myfun( int i1, int i2, int i3, int i4 )
{
  static char buf[] = "(this is the dawning of the age of aquarius)";

  printf( "In myfun: %d %d %d %d\n", i1, i2, i3, i4 );
  snz_push_value( 0, &buf, sizeof(buf) );
}

Processor_Main()
{
  int f;
  char buf[30];
  struct collection *the_collection;
  char *p;

  msg( "Getting ready." ); 
  snz_ready_event(); /* FIXME: should probably be implicit. */

  /* Assuming a function pointer fits in a void * */
  snz_user_event( 666, (void*)&myfun, sizeof( void* ) );

  the_barrier( 1 );
  fetch( (pcxx_NodeNumber()+1)%pcxx_TotalNodes(), (void*)0x37, (void*)0x42, 7 ); 
  put( 0, (void*)37, (void*)42, 7 );
  rpc( (pcxx_NodeNumber()+1)%pcxx_TotalNodes(), (void*)0x38, 8, (void*)0x43, 10 );
  the_barrier( 2 );
  the_collection = create_collection1( 11, 10, sizeof(int) );
  parsect(the_collection);
  /* The user event signals the willingness of the program to be modified */
  msg( "User" ); snz_user_event( 1, (void*)&f, sizeof(f) );
  the_barrier( 4 );
  sprintf( buf, "f=%x", f );
  msg( buf );
  factfun( 3 );

  p = get_collection_info( the_collection );
  snz_user_event( 2, (void*)p, strlen(p)+1 );
  /* Signal stop, again */
  the_barrier( 5 );

  /* Do some RPC to test handler disabling */
  rpc( (pcxx_NodeNumber()+1)%pcxx_TotalNodes(), (void*)0x38, 8, (void*)0x43, 10 );
  delete_collection(the_collection);

  msg( "Terminating." );
  snz_terminate_event();
  return 0;
}

/***************************************************************************/
/* Minimalist RTS */

/* Channels are used for doing barrier (and could be used to simulate IPC) */

struct channel_t {
  int in;
  int out;
} *channel = 0;

/* Fork n-1 times to create a total of n worker processes. */
fork_thyself( int n )
{
  int r, i;
  int p1[2], p2[2];

  channel = (struct channel_t *)malloc( n*sizeof( struct channel_t ));
  for ( i = 1 ; i < n ; i++ ) {
    MyNodeNumber = i;
    if (pipe(p1)) perror( "pipe" );
    if (pipe(p2)) perror( "pipe" );
    channel[0].in = p1[0];
    channel[0].out = p2[1];
    channel[i].out = p1[1];
    channel[i].in = p2[0];
    r = fork();
    if (r < 0) abort();
    if (r == 0) {
      close( p1[1] ); 
      close( p2[0] );
      return 0;
    }
    close( p1[0] );
    close( p2[1] );
  }
  MyNodeNumber = 0;
  return 0;
}

pcxx_barrier()
{
  int i;
  char one = '*';

  if (pcxx_NodeNumber() == 0) {
    /* I am the master */
    for ( i = 1 ; i < pcxx_TotalNodes() ; i++ ) {
      if (read( channel[i].in, (void*)&one, sizeof(one) ) < sizeof(one))
	perror( "read" );
    }
    for ( i = 1 ; i < pcxx_TotalNodes() ; i++ ) {
      if (write( channel[i].out, (void*)&one, sizeof(one) ) < sizeof(one))
	perror( "write" );
    }
  }
  else {
    /* Fan-in, fan-out */
    if (write( channel[0].out, (void*)&one, sizeof( one ) ) < sizeof(one))
      perror( "write" );
    if (read( channel[0].in, (void*)&one, sizeof( one ) ) < sizeof(one)) 
      perror( "read" );
  }
  return 0;
}

int pcxx_TotalNodes()
{
  return MyTotalNodes;
}

int pcxx_NodeNumber()
{
  return MyNodeNumber;
}

double pcxx_get_time()
{
  return 0.0;
}

/* eof */
