; c o n n e c t . s c h
; $Id: connect.sch,v 1.1 1996/05/01 20:11:12 lth Exp $
;
; Connections are abstract communication channels.
; lth@cs.uoregon.edu / June 19, 1995
;
; For STk 2.1.7 sockets.
;
; FIXME:
; If read and write are not atomic (they are in STk, so not too much to
; worry about) we need a per-connection mutex to avoid interleaving messages
; on the same connection.

(define *connection-monitor* #f)

(define (accept-connection portno)
  (let ((s (make-server-socket portno)))
    (socket-accept-connection s)
    (make-connection s)))

(define (connect-to hostname portno)
  (let ((s (make-client-socket hostname portno #f)))
    (if (not s)
	s
	(make-connection s))))

(define (close-connection conn)
  (if (connection? conn)
      (begin (socket-shutdown (connection.socket conn))
	     #t)
      (sneezy.error "close-connection: Not a connection: " conn)))

(define (connection-control which)
  (cond ((eq? which 'monitor)
	 (set! *connection-monitor* #t))
	((eq? which 'nomonitor)
	 (set! *connection-monitor* #f))
	(else
	 (format #t "connection-control: use one of monitor, nomonitor.~%"))))

(define (connection-send c msg)
  (if *connection-monitor*
      (format #t "SEND: ~a~%" msg))
  (write msg (connection.output c))
  (flush (connection.output c)))

(define (connection-receive c)
  (let ((r (read (connection.input c))))
    (if *connection-monitor*
	(format #t "RECV: ~a~%" r))
    r))

(define (make-connection s)
  (list 'connection
	(socket-input s)
	(socket-output s)
	s))

(define (connection? x)
  (and (pair? x) (eq? (car x) 'connection)))

(define connection.input cadr)
(define connection.output caddr)
(define connection.socket cadddr)

(provide "connect.sch")

; eof
