; s n e e z y . s c h
; $Id: sneezy.sch,v 1.1 1996/05/01 20:11:13 lth Exp $
;
; Sneezy base module
; lth@cs.uoregon.edu / June 20, 1995
;
; Definitions of the 'sneezy' module, including all protocol values.

(require "utils.sch")

(define sneezy.eid:ready 0)
(define sneezy.eid:terminate 1)
(define sneezy.eid:reply 2)
(define sneezy.eid:push 3)

(define sneezy.eid-base 100)

(define sneezy.etype:error 0)
(define sneezy.etype:terminate 1)
(define sneezy.etype:create-collection 2)  ; OBSOLETE
(define sneezy.etype:delete-collection 3)  ; OBSOLETE
(define sneezy.etype:begin-barrier 4)
(define sneezy.etype:end-barrier 5)
(define sneezy.etype:begin-fetch 6)
(define sneezy.etype:end-fetch 7)
(define sneezy.etype:begin-put 8)
(define sneezy.etype:end-put 9)
(define sneezy.etype:begin-rpc 10)
(define sneezy.etype:end-rpc 11)
(define sneezy.etype:begin-parallel 12)
(define sneezy.etype:enter-element 13)
(define sneezy.etype:exit-element 14)
(define sneezy.etype:begin-handle-rpc 15)  ; UNSUPPORTED
(define sneezy.etype:end-handle-rpc 16)    ; UNSUPPORTED
(define sneezy.etype:user-event 17)
(define sneezy.etype:begin-create-collection 18)
(define sneezy.etype:end-create-collection 19)
(define sneezy.etype:begin-delete-collection 20)
(define sneezy.etype:end-delete-collection 21)
(define sneezy.etype:end-parallel 22)
(define sneezy.etype:enter-function 23)
(define sneezy.etype:exit-function 24)

(define sneezy.event-count 25)

(define sneezy.cmd:continue 0)
(define sneezy.cmd:terminate 1)
(define sneezy.cmd:install-handler 2)
(define sneezy.cmd:uninstall-handler 3)
(define sneezy.cmd:read 4)
(define sneezy.cmd:write 5)
;(define sneezy.cmd:get-parsection-info 6)
;(define sneezy.cmd:get-element-type 7)
;(define sneezy.cmd:get-collection-info 8)
(define sneezy.cmd:call-user-function 9)

(define sneezy.cmd-count 10)

(define sneezy.argtype:int 1)
(define sneezy.argtype:uint 2)
(define sneezy.argtype:short 3)
(define sneezy.argtype:ushort 4)
(define sneezy.argtype:char 5)
(define sneezy.argtype:uchar 6)
(define sneezy.argtype:schar 7)
(define sneezy.argtype:long 8)
(define sneezy.argtype:ulong 9)
(define sneezy.argtype:charptr 10)

(define sneezy.event-symbol
  (let ((l '#(error
	      terminate
	      create-collection
	      delete-collection
	      begin-barrier
	      end-barrier
	      begin-fetch
	      end-fetch
	      begin-put
	      end-put
	      begin-rpc
	      end-rpc
	      begin-parallel
	      enter-element
	      exit-element
	      begin-handle-rpc
	      end-handle-rpc
	      user-event
	      begin-create-collection
	      end-create-collection
	      begin-delete-collection
	      end-delete-collection
	      end-parallel
	      enter-function
	      exit-function
	      )))
    (lambda (etype)
      (if (or (< etype 0) (>= etype (vector-length l)))
	  (begin (format #t "sneezy.event-symbol: unknown type: ~a~%" etype)
		 'unknown)
	  (vector-ref l etype)))))

(define sneezy.event-name
  (let ((l '#("Error"
	      "Terminate"
	      "Create collection"
	      "Delete collection"
	      "Begin barrier"
	      "End barrier"
	      "Begin fetch"
	      "End fetch"
	      "Begin put"
	      "End put"
	      "Begin RPC"
	      "End RPC"
	      "Begin parallel"
	      "Enter element"
	      "Exit element"
	      "Begin handle RPC"
	      "End handle RPC"
	      "User event"
	      "Begin create collection"
	      "End create collection"
	      "Begin delete collection"
	      "End delete collection"
	      "End parallel"
	      "Enter function"
	      "Exit function"
	      )))
    (lambda (etype)
      (if (or (< etype 0) (>= etype (vector-length l)))
	  (begin (format #t "sneezy.event-name: unknown type: ~a~%" etype)
		 "Unknown")
	  (vector-ref l etype)))))


; Given a symbol representing an event type, return the event type's code
; or #f if the symbol is not valid.

(define sneezy.event-type 
  (let ((tbl (alist->hash-table 
	      `((error . ,sneezy.etype:error)
		(terminate . ,sneezy.etype:terminate)
		(create-collection  . ,sneezy.etype:create-collection)
		(delete-collection . ,sneezy.etype:delete-collection)
		(begin-barrier . ,sneezy.etype:begin-barrier)
		(end-barrier . ,sneezy.etype:end-barrier)
		(begin-fetch . ,sneezy.etype:begin-fetch)
		(end-fetch . ,sneezy.etype:end-fetch)
		(begin-put . ,sneezy.etype:begin-put)
		(end-put . ,sneezy.etype:end-put)
		(begin-rpc . ,sneezy.etype:begin-rpc)
		(end-rpc . ,sneezy.etype:end-rpc)
		(begin-parallel . ,sneezy.etype:begin-parallel)
		(enter-element . ,sneezy.etype:enter-element)
		(exit-element . ,sneezy.etype:exit-element)
		(begin-handle-rpc . ,sneezy.etype:begin-handle-rpc)
		(end-handle-rpc . ,sneezy.etype:end-handle-rpc)
		(user-event . ,sneezy.etype:user-event)
		(begin-create-collection
		     . ,sneezy.etype:begin-create-collection)
		(end-create-collection . ,sneezy.etype:end-create-collection)
		(begin-delete-collection
		     . ,sneezy.etype:begin-delete-collection)
		(end-delete-collection . ,sneezy.etype:end-delete-collection)
		(end-parallel . ,sneezy.etype:end-parallel)
		(enter-function . ,sneezy.etype:enter-function)
		(exit-function . ,sneezy.etype:exit-function)
		))))
    (lambda (id)
      (hash-table-get tbl id #f))))

; Return list of all events. The list is to be considered immutable!

(define sneezy.all-events
  (let ((k #f))
    (lambda ()
      (if k 
	  k
	  (let loop ((i 0) (l '()))
	    (if (= i sneezy.event-count)
		(begin (set! k l)
		       l)
		(loop (+ i 1) (cons i l))))))))

; Print a message on some suitable display.

(define (sneezy.msg f . rest)
  (apply format #t f rest)
  (newline))

; Signal an error.

(define sneezy.error error)

; Return the debugger port to use on the parallel host.

(define sneezy.default-port-number 5666)

(define (sneezy.debugger-port)
  (let ((p (getenv "SNZPORT")))
    (or (and p (string->number p))
	sneezy.default-port-number)))

(provide "sneezy.sch")

; eof
