#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#
displayMessage hourglass "Loading classy..."

source "$TAUDIR/inc/depfile.tcl"
source "$TAUDIR/inc/printcan.tcl"
source "$TAUDIR/inc/bdbm_utils.tcl"; #To get the maintag information.
# Need fileio.tcl for remote operations
source "$TAUDIR/inc/fileio.tcl"

set ch_maxlevel 0;     # depth of class hierarchy (maximum x coordinate)
set ch_maxpos 0;       # maximum y coordinate
set ch_showtype 1;     # mark collections with "||" ?
set ch_sel_ctag NONE;  # id of currently selected class
set classy_classes(1) "";#setting up the local storage array SAS

# A debugging framework, for installing scaffolding.
set DEBUG_SET 0


proc DEBUG {} {
    global DEBUG_SET

    return $DEBUG_SET
}

#
# traverseHierarchy: determine graph layout by recursively traversing it
#                    position determined by "level" (x) and "(max)pos" (y)
#
#               sub: Sage++ id of root of subgraph
#             level: current level in the graph (passed down)
#

proc traverseHierarchy {file sub level} {
    global myself classy_classes
    global ch_maxlevel ch_maxpos
    
    
    set classy_classes($sub,level) $level
    if { $level > $ch_maxlevel } {set ch_maxlevel $level}
    if {![info exists classy_classes($sub,fold)]} {
	set classy_classes($sub,fold) 0
    }
    
    if { $classy_classes($sub,fold) } {
	# -- current node is "folded"
	incr ch_maxpos
	set mypos $ch_maxpos
    } else {
	incr level
	set temp_subInfo [Cgm_ClassInfo $file $sub "subs"]
	switch -exact [llength $temp_subInfo] {
	    0 {
		# -- no subclasses
		# -- take next free position
		#set classy_classes($sub,level) $level
		set classy_classes($sub,fold) 0
		set classy_classes($sub,subs) [list]
		incr ch_maxpos
		set mypos $ch_maxpos
		set classy_classes($sub,pos) $mypos
	    }
	    1 {
		# -- exactly one subclass
		set c [lindex $temp_subInfo 0]
		if [ info exists classy_classes($c,pos) ] {
		    # -- subclass has already y position elsewhere
		    # -- handle like no subclass case
		    set classy_classes($c,level) $level
		    incr ch_maxpos
		    set mypos $ch_maxpos
		    #set classy_classes($c,pos) $mypos
		} else {
		    # -- subclass has NOT yet y position
		    # -- use the y coordinate of subclass for myself too
		    set classy_classes($c,level)  $level
		    set mypos [traverseHierarchy $file $c $level]
		    set classy_classes($sub,pos) $mypos
		    set classy_classes($sub,fold) 0
		    #incr ch_maxpos
		    
		}
		set classy_classes($sub,subs) $temp_subInfo
	    }
	    default {
		# -- more than one subclass
		# -- position myself in the middle between first and last subclass
		set startpos $ch_maxpos
		foreach c $temp_subInfo {
		    if { ! [info exists classy_classes($c,pos)] } {
			set endpos [traverseHierarchy $file $c $level]
			set classy_classes($c,level) $level
			set classy_classes($c,fold) 0
			set mypos [expr ($startpos+$endpos-($startpos+1))/2]
			
			#set classy_classes($c,pos) $mypos
		    }
		}
		set classy_classes($sub,subs) $temp_subInfo
	    }
	}
    }
    if {![info exists classy_classes($sub,pos)]} {
	set classy_classes($sub,pos) $mypos
    } else {
	set mypos $classy_classes($sub,pos)
    }
    
    if { $level == 0 } {
	.$myself.graph.can configure -scrollregion \
	    [list 0 0 [expr 260+$ch_maxlevel*200] [expr 60+$ch_maxpos*30]]
    }
    return $mypos
}

#
# selectFuncTag: implementation of global feature "selectFunction"
#                this tool doesn't provide one
#

proc selectFuncTag {progfile tag} {}

#
# functions to invoke in class member table
#

set selectclass(show) global_showFuncTag
set selectclass(select) global_selectFuncTag

#
# proc selectClassTag: implementation of global feature "selectClass"
#                      classy displays the selected class in red
#                      and pop up a class member table
#
#                   c: Sage++ id of class to select
#

proc selectClassTag {file c} {
    global classy_classes members selectclass classes
    global ch_sel_ctag myself
    
    # -- de-select previously selected class if necessary
    if { $ch_sel_ctag != "NONE" } {
	if [info exists classy_classes($ch_sel_ctag,obj)] {
	    .$myself.graph.can itemconfigure $classy_classes($ch_sel_ctag,obj) -fill black
	}
    }
    
    # -- display selected class in red, if it is displayed
    if [info exists classy_classes($c,obj)] {
	.$myself.graph.can itemconfigure $classy_classes($c,obj) -fill red
    }
    set ch_sel_ctag $c
    if [DEBUG] {
	puts "The tag has been colored. Going to build member window."
	puts "file == $file, c == $c, mem is still mem."
    }
    
    # -- pop up class member table, if class has members

    # Abstracting, using the Cgm interface. This will make the module
    # more independent of existing features of TAU.
    # SAS 10/18
    set classy_classes($file,$c,mem) [Cgm_ClassInfo $file $c "mem"]
    if {$classy_classes($file,$c,mem) != "CGM_DOESNT_EXIST"} {
	if [DEBUG] {puts "Info exists; into the membering function"}
	# -- build sorted list of member functions
	foreach m $classy_classes($file,$c,mem) {
	    lappend mems $members($file,$m)
	}
	set mems [lsort $mems]
	set selectclass(mem) $mems
	set selectclass(file) $file
	
	if [winfo exists .member] {
	    # -- there is already a class member window; reset it
	    .member.l1 delete 0 end
	} else {
	    # -- create new class member window
	    toplevel .member
	    wm minsize .member 50 50
	    listbox .member.l1 -relief sunken -width 60 -height 15 \
		-exportselection false -selectmode single
	    scrollbar .member.s1 -orient vert -relief sunken \
		-command ".member.l1 yview"
	    .member.l1 configure -yscrollcommand ".member.s1 set"
	    
	    frame .member.bottom
	    frame .member.f1 -relief sunken -bd 1
	    button .member.b1 -text "show" -command {
		set s [.member.l1 curselection]
		if { $s != "" } {
		    set type [lindex [lindex $selectclass(mem) $s] 2]
		    if { [string first f $type] == 1 } {
			set tag [lindex [lindex $selectclass(mem) $s] 1]
			$selectclass(show) $selectclass(file) $tag
		    }
		}
	    }
	    pack .member.b1 -in .member.f1 -side left -padx 5 -pady 5
	    pack .member.f1 -in .member.bottom -side left -padx 10 -pady 10
	    button .member.b2 -text "select" -command {
		set s [.member.l1 curselection]
		if { $s != "" } {
		    set type [lindex [lindex $selectclass(mem) $s] 2]
		    if { [string first f $type] == 1 } {
			set tag [lindex [lindex $selectclass(mem) $s] 1]
			$selectclass(select) $selectclass(file) $tag
		    }
		}
	    }
	    pack .member.b2 -in .member.bottom -side left -padx 15 -pady 15
	    button .member.b3 -text "cancel" -command "destroy .member"
	    pack .member.b3 -in .member.bottom -side right -padx 15 -pady 15
	    
	    pack .member.bottom -side bottom -fill x
	    pack .member.s1 -side right -fill y
	    pack .member.l1 -side left  -expand yes -fill both
	    
	    bind .member.l1 <Double-1> ".member.b1 invoke"
	}
	
	# -- display title
	set classy_classes($file,$c,coll) [Cgm_ClassInfo $file $c "coll"]
	if {$classy_classes($file,$c,coll) == "CGM_DOESNT_EXIST"} {
	    showError "Error checking for collection information"
	    return
	}
	set classy_classes($file,$c,name) [Cgm_ClassInfo $file $c "name"]
	if {$classy_classes($file,$c,name) == "CGM_DOESNT_EXIST"} {
	    showError "Error checking for class name information"
	    return
	}
	if { $classy_classes($file,$c,coll) == "-" } {
	    wm title .member "class  $classy_classes($file,$c,name)"
	} else {
	    wm title .member "collection  $classy_classes($file,$c,name)"
	}
	
	# -- display formatted class member list
	foreach m $mems {
	    .member.l1 insert end \
		"[format "%-25s" [lindex $m 0]] [typeDescr [lindex $m 2]]"
	}
    }
}

#
# typeDescr: transform short typedescription into readable string
#
#      type: short form
#

proc typeDescr {type} {
  set t [split $type ""]

  switch -exact [lindex $t 0] {
    o  {set r "protected"}
    u  {set r "public   "}
    i  {set r "private  "}
    e  {set r "element  "}
  }

  switch -exact [lindex $t 4] {
    " "  { switch -exact [lindex $t 1] {
             v  {append r " variable   "}
             f  {append r " function   "}
             c  {append r " constructor"}
           }
         }
    c    {append r " constructor"}
    d    {append r " destructor "}
    o    {append r " operator   "}
  }

  if { [lindex $t 2] == "v" } {append r " virtual"}
  if { [lindex $t 3] == "i" } {append r " inline"}

  return $r
}

#
# displayHierarchy: display class hierarchy by recursively traversing graph
#                   traverseHierarchy has to be called first to compute
#                   positions!
#
#              sub: Sage++ id of root of graph to display
#

proc displayHierarchy {sub} {
    global myself classy_classes
    #global classes
    global ch_sel_ctag ch_showtype
    
    # -- compute screen coordinates from level/pos computed by traverseHierarchy
    set x [expr 30+$classy_classes($sub,level)*200]
    set y [expr 30+$classy_classes($sub,pos)*30]
    set file [lindex $classy_classes(maintag) 0]
    set name [Cgm_ClassInfo $file $sub "name"]
    if {[string match $name "NOT_OK"]} {
	puts "Not name info for tag $sub in $file."
	return
    }
    
    
    # -- mark collections with "||" if necessary
    set temp_collInfo [Cgm_ClassInfo $file $sub "coll"] 
    if { $ch_showtype && $temp_collInfo != "-" } {
	set name "|| $name"
    }
    
    # -- if class selected, display name in red
    if { $sub == $ch_sel_ctag } {
	set obj [.$myself.graph.can create text $x $y \
		     -text $name -fill red -anchor w \
		     -font -Adobe-Helvetica-Bold-R-Normal--*-120-*]
    } else {
	set obj [.$myself.graph.can create text $x $y \
		     -text $name -fill black -anchor w \
		     -font -Adobe-Helvetica-Bold-R-Normal--*-120-*]
    }
    .$myself.graph.can bind $obj <Button-1> "toggleClass $sub"
    .$myself.graph.can bind $obj <Button-2> "selectClassTag $file $sub"
    .$myself.graph.can bind $obj <Button-3> "global_selectClassTag $file $sub"
    set classy_classes($sub,obj) $obj
    
    # -- compute center of node
    set bbox [.$myself.graph.can bbox $obj]
    set ex [expr [lindex $bbox 0] + ([lindex $bbox 2] - [lindex $bbox 0]) / 2]
    set ey [expr [lindex $bbox 1] + ([lindex $bbox 3] - [lindex $bbox 1]) / 2]
    
    # This was going in the space below.
	set temp_subInfo [Cgm_ClassInfo $file $sub "subs"]
	if {[string match $temp_subInfo "NOT_OK"]} {
	    puts "No sub information for tag $sub in $file."
	    return
	}


    # -- if class has subclasses, draw them and connect with lines
    if {  $classy_classes($sub,fold) } {
	#Right here: I'm the space below what's above.
	foreach c $temp_subInfo {
	    set pos [displayHierarchy $c]
	    .$myself.graph.can create line $ex $ey [lindex $pos 0] [lindex $pos 1] \
		-arrow last
	}
    }
    
    # -- draw surrounding rectangle
    .$myself.graph.can create rectangle \
	[expr [lindex $bbox 0]-3] [expr [lindex $bbox 1]-3] \
	[expr [lindex $bbox 2]+3] [expr [lindex $bbox 3]+3] \
	-fill white
    
    # -- if node folded, draw double-border rectangle
    if { !$classy_classes($sub,fold) && [llength $temp_subInfo] } {
	.$myself.graph.can create rectangle \
	    [expr [lindex $bbox 0]-1] [expr [lindex $bbox 1]-1] \
	    [expr [lindex $bbox 2]+1] [expr [lindex $bbox 3]+1] \
	    -fill white
    }
    
    # -- raise text above last drawn rectangles
    # -- cannot done the other way round, as the size of the rectangle
    # -- depends on the text
    .$myself.graph.can raise $obj
    
    return [list [expr $x-10] $y]
}

#
# toggleClass: toggle class node between "expanded" and "folded"
#
#           c: id of class to toggle
#

proc toggleClass {c} {
    global classy_classes myself
    
    set f $classy_classes($c,fold)
    set classy_classes($c,fold) [expr 1-$f]
    set hasSubs [llength $classy_classes($c,subs)]
    
    if { $hasSubs } {
	set oldobj [expr 30+$classy_classes($c,pos)*30]
	set oldwin [.$myself.graph.can canvasy 0]
	
	redrawHierarchy
	
	set maxyscroll [lindex [.$myself.graph.can cget -scrollregion] 3]
	set newobj [expr 30+$classy_classes($c,pos)*30]
	.$myself.graph.can yview moveto \
	    [expr (($newobj-$oldobj+$oldwin)/$maxyscroll)+1]
    }
}

#
# redrawHierarchy: redraw class hierarchy by deleteing old one,
#                  then redrawing it
#

proc redrawHierarchy {} {
    global myself classy_classes
    global classes ch_maxlevel ch_maxpos
    global depfile
    
    # -- reset variables
    .$myself.graph.can delete all
    #set ch_maxlevel 0
    set ch_maxpos 0
    
    set file [lindex $classy_classes(maintag) 0]
    
    set temp_allctags [Cgm_DepInfo $file "allctags"]
    if {[string match $temp_allctags "NOT_OK"] \
	    || [string match $temp_allctags "CGM_FAILED"]} {
	exit
    }
    #foreach t $temp_allctags {
    #if [info exists classy_classes($t,pos)] { unset classy_classes($t,pos) }
    #}
    
    # -- recompute positions and display new graph
    # -- start with "virtual" root which has id -1 and level -1
    set classy_classes(-1,fold) 0
    traverseHierarchy $file -1 -1
    set starting_subsInfo [Cgm_ClassInfo $file -1 "subs"]
    if {[string match $starting_subsInfo "NOT_OK"]} {
	puts "Initial subs information invalid."
	return
    }
    foreach c $starting_subsInfo {displayHierarchy $c}
}

#
# loadDep: switch to another application
#          this is invoked from the TAU master control window
#
#    path: UNIX pathname of depfile to read
#    root: patchname of SAGEROOT
#    arch: architecture of machine where depfile resides
#

proc loadDep {path root arch} {
  global myself classy_classes
  #global classes members
  global ch_sel_ctag
  global depfile

  # -- delete old class database
    #if {[info exists  classy_classes]} {
	#unset classy_classes
    #}
    if {[info exists members]} {
	unset members
    }
    #  unset classes members

  # -- read new data, and display new graph
    if { [Cgm_LoadDep [lindex $classy_classes(maintag) 0] "-dumpch"] == "NOT_OK" } {
	puts "Could not load Depfile. Exiting."
	exit
    }
    
    redrawHierarchy
    set ch_sel_ctag NONE
    
    # -- adjust "expand" menu list
    .$myself.bar.b2.m1.1 delete 0 last
    expandMenu
}

#
# createWindow: create and display main class hierarchy window
#

proc createWindow {} {
    global myself TAUDIR classy_classes
    
    toplevel .$myself
    wm title .$myself "CLASSY"
    wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm
    wm minsize .$myself 50 50
    
    frame .$myself.graph
    
    # -- configure graph area
    canvas .$myself.graph.can -width 600 -height 350 -background white
    bind .$myself.graph.can <2> ".$myself.graph.can scan mark %x %y"
    bind .$myself.graph.can <B2-Motion> ".$myself.graph.can scan dragto %x %y"
    
    frame .$myself.graph.f1
    
    scrollbar .$myself.graph.hscroll -orient horiz -relief sunken \
	-command ".$myself.graph.can xview"
    .$myself.graph.can configure -xscrollcommand ".$myself.graph.hscroll set"
    
    frame .$myself.graph.f2 -width 20 -height 20
    
    pack .$myself.graph.hscroll -in .$myself.graph.f1 -side left \
	-expand yes -fill x
    pack .$myself.graph.f2      -in .$myself.graph.f1 -side right
    
    scrollbar .$myself.graph.vscroll -orient vert -relief sunken \
	-command ".$myself.graph.can yview"
    .$myself.graph.can configure -yscrollcommand ".$myself.graph.vscroll set"
    
    pack .$myself.graph.f1      -side bottom -fill x
    pack .$myself.graph.vscroll -side right  -fill y
    pack .$myself.graph.can     -side top -padx 15 -pady 15 -fill both -expand yes
    
    # -- configure menu bar
    frame .$myself.bar -relief raised -borderwidth 2
    
    menubutton .$myself.bar.b1 -text File -menu .$myself.bar.b1.m1 -underline 0
    menu .$myself.bar.b1.m1
    .$myself.bar.b1.m1 add command -label "Print graph" -underline 0 \
	-command "printCanvas .$myself.graph.can $myself"
    .$myself.bar.b1.m1 add separator
    .$myself.bar.b1.m1 add command -label "Exit"  -underline 0 -command "exit"
    
    menubutton .$myself.bar.b2 -text View -menu .$myself.bar.b2.m1 -underline 0
    menu .$myself.bar.b2.m1
    .$myself.bar.b2.m1 add checkbutton -label "Show type of class" \
	-underline 0 -variable ch_showtype -onvalue 1 -offvalue 0 \
	-command "redrawHierarchy"
    .$myself.bar.b2.m1 add separator
    
    .$myself.bar.b2.m1 add cascade -label "Expand" -underline 0 \
	-menu .$myself.bar.b2.m1.1
    menu .$myself.bar.b2.m1.1
    
    menubutton .$myself.bar.b3 -text Help -menu .$myself.bar.b3.m1 -underline 0
    menu .$myself.bar.b3.m1
    .$myself.bar.b3.m1 add command -label "on $myself" -underline 3 \
	-command "xsend tau \[list showHelp $myself 1-$myself 1\]"
    .$myself.bar.b3.m1 add separator
    .$myself.bar.b3.m1 add command -label "on menubar" -underline 3 \
	-command "xsend tau \[list showHelp $myself 1.1-menu 1\]"
    .$myself.bar.b3.m1 add command -label "on display area" -underline 3 \
	-command "xsend tau \[list showHelp $myself 1.2-display 1\]"
    .$myself.bar.b3.m1 add separator
    .$myself.bar.b3.m1 add command -label "on using help" -underline 3 \
	-command "xsend tau {showHelp general 1-help 1}"
    
    createToolMenu .$myself.bar.b4
    
    pack .$myself.bar.b1 .$myself.bar.b2 -side left -padx 5
    pack .$myself.bar.b3 .$myself.bar.b4 -side right -padx 5
    
    pack .$myself.bar   -side top -fill x
    pack .$myself.graph -side left -padx 15 -pady 15 -fill both -expand yes
}

#
# expandMenu: update "expand" menu, so that it has an entry for each
#             level of the callgraph
#

proc expandMenu {} {
    global myself ch_maxlevel ch_exp_level
    
    set ch_exp_level $ch_maxlevel
    for {set i 0} {$i<=$ch_maxlevel} {incr i} {
	.$myself.bar.b2.m1.1 add radiobutton -label "upto level $i" -underline 11 \
	    -variable ch_exp_level -value $i \
	    -command "expandHierarchy $i"
    }
}

#
# expandHierarchy: expand class hierarchy after user selected level through
#                  the "expand" menu
#
#           level: level up to which graph has to be expanded
#

proc expandHierarchy {level} {
    global depfile classy_classes
    global classes

    set temp_allctags [Cgm_DepInfo [lindex $classy_classes(maintag) 0] "allctags"]
    if {[string match $temp_allctags "NOT_OK"] \
	    || [string match $temp_allctags "CGM_FAILED"]} {
	puts "Reading allctags failed: $temp_allctags"
	exit
    }
  foreach t $temp_allctags {
    if { $classy_classes($t,level) < $level } {
      set classy_classes($t,fold) 1
    } elseif { $classy_classes($t,level) >= $level } {
      set classy_classes($t,fold) 0
    }
  }
  redrawHierarchy
}

proc Tool_AcceptChanges {progfile flag} {

    global depfile classy_classes

    switch $flag {

	d {
	    # Delete a file
	    # Since the file that has the main procedure in it may have
	    # been the file deleted, we need to reinitialize cagey with the 
	    # updated file list. . .
	    set classy_classes(maintag) [Bdb_GetMaintag]
	    if {[llength $classy_classes(maintag)] == 1} {
		if {[string match $classy_classes(maintag) "BDBM_FAILED"] \
			|| !$classy_classes(maintag)} {
		    showError "No project loaded, or incomplete compilation"
		    return
		}
	    }
	    
	    loadDep [lindex $classy_classes(maintag) 0] \
		$depfile(root) $depfile(arch)
	    redrawHeirarchy
	}


	a {
	    # Add a file
	    # If cagey was already running, then we already have a maintag function. However,
	    # we need to check for incomplete compilation of the new file. . .
	    loadDep [lindex $cagey_funcs(maintag) 0] $depfile(root) $depfile(arch)
	    redrawHeirarchy
	}


	u {
	    # Update a file
	    loadDep [lindex $cagey_funcs(maintag) 0] $depfile(root) $depfile(arch)
	    redrawHeirarchy
	}
	
	p {
	    #Modify project information. 
	    set temp [PM_Status]
	    if {![string match $temp "UNDEFINED"]} {
		set depfile(project) [lindex $temp 0]
		set depfile(host) [lindex $temp 1]
		set depfile(arch) [lindex $temp 2]
		set depfile(root) [lindex $temp 3]
       		set depfile(dir) [lindex $temp 4]
	    } else {
		showError "There is no project to modify."
	    }
	}

	e {
	    #This is a flag for updating during execution. No action is needed here.
	}
    }
    puts "in classy: returning from Tool_AcceptChanges"
}


#
# exit - TAU exit function that communicates the event to other tau tools.
#
rename exit exit.old
proc exit {{status 0}} {
    # Taken and modified from Spiffy SAS 7/31/96
    global myself

    PM_RemTool $myself
    exit.old $status
}


# ------------
# -- main code
# ------------
if {$argc == 0} {
    set parg [pwd]
} elseif {$argc == 1} {
    set parg [lindex $argv 0]
    if {[file extension $parg] != ".pmf"} {
	set parg "$parg.pmf"
    }
} else {
    puts stderr "usage: $myself \[\[host:\]projFile \| \[host:\]directory\]"
    exit
}

# Init the project manager (taud)
launchTauDaemon -waitfor
PM_AddTool $myself
PM_AddGlobalSelect $myself { global_selectLine \
				 global_selectFuncTag \
				 global_selectClassTag}


# So. . . where once dfile was set as the depfile, now a series of calls to the BDBM is needed
# for whatever files are in the project.
# SAS 7/22/96

# Initialize the project
# Coordinate w/ PM
set pm_status [PM_Status]
if {[lindex $pm_status 0] == "NO_PROJECT"} {
    # Open or create a project
    set colon [string first ":" $parg]
    if {$colon > 0} {
	set hostarg [string range $parg 0 [expr $colon - 1]]
	set patharg [string range $parg [expr $colon + 1] end]
    } else {
	set hostarg localhost
	set patharg $parg
    }
    
    set projfile [PM_OpenProject $patharg $hostarg]
    if {$projfile == "NO_PROJECT"} {
	showError "No project opened!"
	exit
    }
    set pm_status [PM_Status]
}

set depfile(project) [lindex $pm_status 0]
set depfile(host)    [lindex $pm_status 1]
set depfile(arch)    [lindex $pm_status 2]
set depfile(root)    [lindex $pm_status 3]
set depfile(dir)     [lindex $pm_status 4]
set depfile(hostarch) "$depfile(host) ($depfile(arch))"

createWindow
set classy_classes(maintag) [Bdb_GetMaintag]
# Since we just initialized a project in some fashion, we should get *some* kind 
# of maintag. However, the project may need compiling.

if {[llength $classy_classes(maintag)] == 1} {
    if {[string match $classy_classes(maintag) "BDBM_FAILED"] \
	    || !$classy_classes(maintag)} {
	showError "No project loaded, or incomplete compilation\n$classy_classes(maintag)"
	return
    }
}

loadDep [lindex $classy_classes(maintag) 0] $depfile(root) $depfile(arch)
# -- compute positions and draw class hierarchy
# -- start with "virtual" root which has id -1 and level -1
traverseHierarchy [lindex $classy_classes(maintag) 0] -1 -1
set init_subsInfo [Cgm_ClassInfo [lindex $classy_classes(maintag) 0] -1 "subs"]
if {[string match $init_subsInfo "NOT_OK"]} {
    puts "No subs info for -1 in $file"
    exit
}
foreach c $init_subsInfo {displayHierarchy $c}
#expandMenu

removeMessage
