#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#
displayMessage hourglass "Loading racy..."

source "$TAUDIR/inc/printcan.tcl"
source "$TAUDIR/inc/fileio.tcl"

set racy_progfile "";   # the single pc++ program file

set tcl_precision 17

set Colors [list \
    magenta green red blue yellow purple orange gold brown \
    tan grey coral forestgreen skyblue seagreen beige steelblue1 royalblue \
    pink cyan violet snow steelblue white ]

set TextColors [list \
    black black black white black white black black white \
    black black black black black black black black black \
    black black black black black black ]

set Stipples [list \
    hline1.bm cross.bm xline1.bm gray1.bm hline22.bm \
    vline1.bm gray3.bm cross0.bm xline2.bm \
    cross1.bm cross3.bm xline3.bm cross2.bm \
    cross5.bm vline2.bm xline4.bm cross4.bm hline2.bm \
    vline22.bm cross6.bm ]


#
# exit - TAU exit function that communicates the event to other tau tools.
#
rename exit exit.old
proc exit {{status 0}} {
    global myself

    PM_RemGlobalSelect $myself \
	    { global_selectFuncTag }
    PM_RemTool $myself

    exit.old
}


#
# wraplindex: like lindex, but index "wraps around" if bigger
#             than length of list
#

proc wraplindex {list index} {
  lindex $list [expr $index%[llength $list]]
}

#
# min: returns minimum of two values
#

proc min {x1 x2} {
  if { $x1 < $x2 } { return $x1 } else { return $x2 }
}

#
# max: returns maximum of two values
#

proc max {x1 x2} {
  if { $x1 > $x2 } { return $x1 } else { return $x2 }
}

#
# vectorsum: returns sum of all elements in a vector
#

proc vectorsum {vector} {
  upvar $vector v
  set r 0.0
  foreach elem $v {
    set r [expr $r+$elem]
  }
  return $r
}

#
# vectormax: returns maximum of all elements in a vector
#

proc vectormax {vector} {
  upvar $vector v
  set r [lindex $v 0]
  foreach elem $v {
    if { $elem > $r } { set r $elem }
  }
  return $r
}

#
# vectorop: apply scalar operation to vector
#

proc vectorop {vector op val} {
  upvar $vector v
  set r ""
  foreach elem $v {
    lappend r [format "%.1f" [expr ${elem}${op}${val}]]
  }
  return $r
}

#
# vectorpercent: scale elements of vector according to max and scale
#

proc vectorpercent {vector max scale} {
  upvar $vector v

  if { $max == 0 } { return $v }
  set r ""
  foreach elem $v {
    lappend r [format "%.2f" [expr $elem/$max*$scale]]
  }
  return $r
}

#
# numberInput: widget for input of 1 integer number
#
#         win: widget name
#       label: optional label to display
#     varname: name of global variable which holds input value
#

proc numberInput {win label varname} {
  global $varname

  frame $win
  if { $label != "" } {
    label $win.t -text $label
    pack $win.t -side top -fill x
  }
  button $win.b1 -text "<<" -width 3 -command "incr $varname -5"
  button $win.b2 -text "<"  -width 3 -command "incr $varname -1"
  entry $win.e -textvariable $varname -width 5 -relief sunken
  button $win.b3 -text ">"  -width 3 -command "incr $varname +1"
  button $win.b4 -text ">>" -width 3 -command "incr $varname +5"
  pack $win.b1 $win.b2 $win.e $win.b3 $win.b4 -side left -padx 10 -pady 10
}

set pr_sel_tag -99;  # currently selected function

#
# bargraph: horizontal bargraph widget
#
#        win: pathname for bargraph
# leftlabels: list of labels
#    percent: list of percent values
#     values: list of values
#       tags: list of corresponding Sage++ id of functions
#      nodes: list of corresponding node identifiers
#             (if nonempty used instead of tags)
#       mode: what to show as lables on the right side: per / val / none
#

proc bargraph {win leftlabels percent values tags {nodes {}} {mode per}} {
  global tagcol tagstip
  global pr_sel_tag
  global racy_progfile

  set num [llength $leftlabels];  # number of bars
  set h [expr ($num+1)*20];       # height of window

  # -- create or reset canvas
  if { [winfo exists $win] } {
    $win delete all
    $win configure -height [min 420 $h]
  } else {
    canvas $win -width 420 -height [min 420 $h] -background white
    $win configure -scrollregion [list 0 0 420 $h]
  }

  # -- scale percentage vector, so that it uses most of the display
  set pmax [vectormax percent]
  if { $pmax < 0.1 } {
    set pers [vectorop percent * 750]
  } elseif { $pmax < 1 } {
    set pers [vectorop percent * 75]
  } elseif { $pmax < 10 } {
    set pers [vectorop percent * 7.5]
  } elseif { $pmax < 25 } {
    set pers [vectorop percent * 3]
  } elseif { $pmax < 50 } {
    set pers [vectorop percent * 1.5]
  } else {
    set pers $percent
  }
  
  for {set i 0} {$i<$num} {incr i} {
    # -- display left label; highlight in red if it is selected function 
    set t [lindex $tags $i]
    if { $t == $pr_sel_tag } {
      set obj [$win create text 150 [expr 20+$i*20] \
               -text [lindex $leftlabels $i] -anchor e -fill red \
               -font -Adobe-Helvetica-Bold-R-Normal--*-120-*]
    } else {
      set obj [$win create text 150 [expr 20+$i*20] \
               -text [lindex $leftlabels $i] -anchor e \
               -font -Adobe-Helvetica-Bold-R-Normal--*-120-*]
    }

    # -- setup bindings for clicking on left labels
    if { $nodes == {} } {
      $win bind $obj <Button-1> "showFuncgraph $t"
      $win bind $obj <Button-2> "PM_GlobalSelect $racy_progfile global_showFuncTag $t"
      $win bind $obj <Button-3> "PM_GlobalSelect $racy_progfile global_selectFuncTag $t"
    } else {
      $win bind $obj <Button-1> \
                "showBargraph [lindex $nodes $i] {[lindex $leftlabels $i]}"
      $win bind $obj <Button-2> \
                "showText [lindex $nodes $i] {[lindex $leftlabels $i]}"
    }

    # -- draw bars and right labels according to mode
    set p [lindex $pers $i]
    if { $p > 0 } {
      $win create rectangle 160 [expr 29+$i*20] [expr 160+$p*2] [expr 11+$i*20]\
                  -fill $tagcol($t) -stipple $tagstip($t)
      switch -exact $mode {
        per   { $win create text [expr 170+$p*2] [expr 20+$i*20] \
                            -text "[lindex $percent $i]%" -anchor w \
                            -font -Adobe-Helvetica-Bold-R-Normal--*-120-*}

        val   { $win create text [expr 170+$p*2] [expr 20+$i*20] \
                            -text "[wraplindex $values $i]" -anchor w \
                            -font -Adobe-Helvetica-Bold-R-Normal--*-120-*}

        none  {}
      }
    }
  }
}

#
# selectFuncTag: implementation of global feature "selectFunction" for racy
#                display all instances of selected function in red and
#                show function profile, if possible
#
#           progfile: had better be == racy_progfile
#           tag: id of function to select
#

proc selectFuncTag {progfile tag} {
  global pr_sel_tag \
	  tagname racy_progfile

  if {$progfile != $racy_progfile} {
      return
  }

  set pr_sel_tag $tag
  redraw {} {}
  if [ info exists tagname($tag) ] { showFuncgraph $tag }
}

#
# showText: display text node profile window
#
#     node: id of node
#     name: full nameof node
#

proc showText {node name} {
  global myself
  global data
  global textorder1 textorder2
  global proforder profvalue profmode profunit

  if { ! [winfo exists .text$node] } {
    set textorder1($node) $profvalue(all)
    if { $proforder(all) == "all" } {
      set textorder2($node) $node
    } else {
      set textorder2($node) $proforder(all)
    }
    toplevel .text$node
    wm title .text$node "$name profile"

    frame .text$node.mbar -relief raised -borderwidth 2

    menubutton .text$node.mbar.b1 -text File -menu .text$node.mbar.b1.m1 \
                                  -underline 0
    menu .text$node.mbar.b1.m1
    .text$node.mbar.b1.m1 add command -label "Close"  -underline 0 \
                    -command "destroy .text$node"

    menubutton .text$node.mbar.b2 -text Order -menu .text$node.mbar.b2.m1 \
                                  -underline 0
    menu .text$node.mbar.b2.m1
    submenu2 .text$node.mbar.b2.m1 textorder1 redrawText $node {$name}
    .text$node.mbar.b2.m1 add separator
    submenu3 .text$node.mbar.b2.m1 textorder2 redrawText $node {$name}

    menubutton .text$node.mbar.b3 -text Help -menu .text$node.mbar.b3.m1 \
                                  -underline 0
    menu .text$node.mbar.b3.m1
    .text$node.mbar.b3.m1 add command -label "on text node profile" \
                    -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.2.3-text 1\]"

    pack .text$node.mbar.b1 .text$node.mbar.b2 -side left -padx 5
    pack .text$node.mbar.b3 -side right -padx 5

    redrawText $node {$name}
    button .text$node.but -text "close" -command "destroy .text$node"
    pack .text$node.mbar -side top -fill x
    pack .text$node.text -side top -ipadx 10 -ipady 10 -anchor center
    pack .text$node.but -side top -fill x
  } else {
    raise .text$node
  }
}

#
# redrawText: create or update contents of text node profile window
#
#       node: id of node
#       name: full nameof node
#

proc redrawText {node name} {
  global data
  global tagname tagcol tagfcol
  global textorder1 textorder2
  global pr_sel_tag
  global racy_progfile

  set win .text$node.text
  set v $textorder1($node)
  set tags $data($textorder2($node),${v}tags)

  # -- create or reset text widget
  if { [winfo exists $win] } {
    $win configure -state normal
    $win delete 1.0 end
    $win configure -height [expr 3+[llength $tags]]
  } else {
    text $win -height [expr 3+[llength $tags]] \
              -background white -foreground black -padx 5 -pady 5
  }

  # -- redraw text area
  $win insert end \
       "-------------------------------------------------------------------\n"
  $win insert end \
       "%time         msec   total msec    #call  usec/call name\n"
  $win insert end \
       "-------------------------------------------------------------------\n"
  set n 3
  set txt [order $textorder2($node) $node $v text {}]
  foreach line $txt {
    incr n
    set t [lindex $tags [expr $n-4]]
    if { $line == "" } {
      $win insert end \
           "                                                    $tagname($t)\n"
    } else {
      $win insert end "$line\n"
      if { $t == $pr_sel_tag } {
        $win tag add seltag $n.0 $n.51
        $win tag configure seltag -foreground red
      }
    }
    $win tag add t$n $n.52 "$n.52 lineend"
    if { [winfo depth .] == 2 } {
      $win tag configure t$n -background white -foreground black
    } else {
      $win tag configure t$n -background $tagcol($t) -foreground $tagfcol($t)
    }
    $win tag bind t$n <Button-1> "showFuncgraph $t"
    $win tag bind t$n <Button-2> "PM_GlobalSelect $racy_progfile global_showFuncTag $t"
    $win tag bind t$n <Button-3> "PM_GlobalSelect $racy_progfile global_selectFuncTag $t"
  }
  $win configure -state disabled
}

#
# order: return field of a data item in the same order it appear in
#        other data item. order is determined by "tags" field
#
# source: id of data which values determine sorting
# target: id of data to order
#  value: excl or incl
#  field: field of data to be sorted
#     NA: NULL value for this field
#

proc order {source target value field NA} {
  global data

  if { $source == $target } {
    return $data($target,$value$field)
  } else {
    set result ""
    foreach t $data($source,${value}tags) {
      set i [lsearch -exact $data($target,${value}tags) $t]
      if { $i == -1 } {
        lappend result $NA
      } else {
        lappend result [lindex $data($target,$value$field) $i]
      }
    }
    return $result
  }
}

#
# specialOrder: basically like order above, but only for numeric fields
#               corrects "other" entry in field
#

proc specialOrder {source target value field} {
  global data

  if { $source == $target } {
    return $data($target,$value$field)
  } else {
    set result ""
    set na [vectorsum data($target,$value$field)]; #sum of all values
    foreach t $data($source,${value}tags) {
      set i [lsearch -exact $data($target,${value}tags) $t]
      if { $i == -1 } {
        lappend result 0.0
      } else {
        set x [lindex $data($target,$value$field) $i]
        if { $t == -1 } {
          # -- tag -1 is the "others" element; always last one
          # -- use "notused" values instead of precalculated ones
          lappend result [format "%4.2f" $na]
        } else {
          set na [expr $na-$x]; #value used; subtract it from "notused" value
          lappend result $x
        }
      }
    }
    return $result
  }
}

#
# submenu2: create "Value" menu
#
#      menu: pathname of menu
#       var: basename of global variable to store selection
#      func: function to call for redraw
# node,name: arguments to func
#

proc submenu2 {menu var func node name} {
  $menu add radiobutton -label "microseconds" \
                 -underline 0 -variable ${var}($node) -value excl \
                 -command "$func $node {$name}"
  $menu add radiobutton -label "total microseconds" \
                 -underline 0 -variable ${var}($node) -value incl \
                 -command "$func $node {$name}"
}

#
# submenu3: create "Order" menu
#
#      menu: pathname of menu
#       var: basename of global variable to store selection
#      func: function to call for redraw
# node,name: arguments to func
#

proc submenu3 {menu var func node name} {
  global funodes fulabels

  $menu add radiobutton -label "global" \
            -underline 0 -variable ${var}($node) -value m \
            -command "$func $node {$name}"
  $menu add radiobutton -label "decreasing" \
            -underline 0 -variable ${var}($node) -value $node \
            -command "$func $node {$name}"
  $menu add cascade -label "node" \
            -underline 0 -menu $menu.1

  # -- create "node" submenu
  menu $menu.1
  set i 0
  foreach n $funodes {
    if { $n != {} } {
      $menu.1 add radiobutton -label [lindex $fulabels $i] \
                  -variable ${var}($node) -value $n \
                  -command "$func $node {$name}"
    } else {
      $menu.1 add separator
    }
    incr i
  }
}

#
# submenu4: create function profile "Mode" menu
#
#      menu: pathname of menu
#       var: basename of global variable to store selection
#      func: function to call for redraw
# node,name: arguments to func
#

proc submenu4 {menu var func node name} {
  $menu add radiobutton -label "percent" \
            -underline 0 -variable ${var}($node) -value per \
            -command "$func $node {$name}"
  $menu add radiobutton -label "value" \
            -underline 0 -variable ${var}($node) -value val \
            -command "$func $node {$name}"
}

#
# submenu5: create "Unit" menu
#
#      menu: pathname of menu
#       var: basename of global variable to store selection
#      func: function to call for redraw
# node,name: arguments to func
#

proc submenu5 {menu var func node name} {
  $menu add radiobutton -label "microseconds" \
            -underline 0 -variable ${var}($node) -value 1.0 \
            -command "$func $node {$name}"
  $menu add radiobutton -label "milliseconds" \
            -underline 0 -variable ${var}($node) -value 1.0e3 \
            -command "$func $node {$name}"
  $menu add radiobutton -label "seconds" \
            -underline 0 -variable ${var}($node) -value 1.0e6 \
            -command "$func $node {$name}"
}

#
# submenu6: create collection profile "Mode" menu
#
#      menu: pathname of menu
#       var: basename of global variable to store selection
#      func: function to call for redraw
# node,name: arguments to func
#

proc submenu6 {menu var func node name} {
  $menu add radiobutton -label "local percent" \
            -underline 0 -variable ${var}($node) -value lper \
            -command "$func $node {$name}"
  $menu add radiobutton -label "global percent" \
            -underline 0 -variable ${var}($node) -value gper \
            -command "$func $node {$name}"
  $menu add radiobutton -label "value" \
            -underline 0 -variable ${var}($node) -value val \
            -command "$func $node {$name}"
}

#
# bargraphmenufunction: create menubar for node profile windows
#
#               parent: pathname for parent of menubar
#               prefix: prefix for global variables to store selections
#                 func: function to call for redraw
#            node,name: arguments to func
#

proc bargraphmenu {parent prefix func node name} {
  global myself

  set menubar $parent.mbar
  frame $menubar -relief raised -borderwidth 2
  menubutton $menubar.b1 -text File -menu $menubar.b1.m1 -underline 0
  menu $menubar.b1.m1
  $menubar.b1.m1 add command -label "Print" -underline 0 \
                    -command "printCanvas .bar${node}.bar node.$node"
  $menubar.b1.m1 add separator
  $menubar.b1.m1 add command -label "Close" -underline 0 \
                    -command "destroy .bar$node"

  menubutton $menubar.b2 -text Value -menu $menubar.b2.m1 -underline 0
  menu $menubar.b2.m1
  submenu2 $menubar.b2.m1 ${prefix}value $func $node {$name}

  menubutton $menubar.b3 -text Order -menu $menubar.b3.m1 -underline 0
  menu $menubar.b3.m1
  submenu3 $menubar.b3.m1 ${prefix}order $func $node {$name}

  menubutton $menubar.b4 -text Mode -menu $menubar.b4.m1 -underline 0
  menu $menubar.b4.m1
  submenu4 $menubar.b4.m1 ${prefix}mode $func $node {$name}

  menubutton $menubar.b5 -text Units -menu $menubar.b5.m1 -underline 0
  menu $menubar.b5.m1
  submenu5 $menubar.b5.m1 ${prefix}unit $func $node {$name}

  menubutton $menubar.b6 -text Help -menu $menubar.b6.m1 -underline 0
  menu $menubar.b6.m1
  $menubar.b6.m1 add command -label "on node profile" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.2.2-node 1\]"

  pack $menubar.b1 $menubar.b2 $menubar.b3 $menubar.b4 $menubar.b5\
       -side left -fill x -padx 5
  pack $menubar.b6 -side right -fill x -padx 5
}

#
# showBargraph: display node profile window
#
#         node: id of node
#         name: full nameof node
#

proc showBargraph {node name} {
  global barorder barmode barvalue barunit
  global proforder profvalue profmode profunit

  if { ! [winfo exists .bar$node] } {
    if { $proforder(all) == "all" } {
      set barorder($node) $node
    } else {
      set barorder($node) $proforder(all)
    }
    set barmode($node) $profmode(all)
    set barvalue($node) $profvalue(all)
    set barunit($node) $profunit(all)
    set win [toplevel .bar$node]
    wm title $win "$name profile"

    bargraphmenu $win bar redrawBargraph $node {$name}

    redrawBargraph $node {$name}
    button $win.but -text "close" -command "destroy $win"
    pack $win.mbar $win.bar $win.but -side top -fill x
  } else {
    raise .bar$node
  }
}

#
# redrawBargraph: create or update contents of node profile window
#
#           node: id of node
#           name: full nameof node
#

proc redrawBargraph {node name} {
  global data
  global barorder barmode barvalue barunit
  global tagname

  set v $barvalue($node)
  set u [specialOrder $barorder($node) $node $v usecs]
  foreach t $data($barorder($node),${v}tags) {
    lappend n $tagname($t)
  }
  bargraph .bar$node.bar $n \
                         [specialOrder $barorder($node) $node $v percent] \
                         [vectorop u / $barunit($node)] \
                         [order $barorder($node) $node $v tags -2] \
                         {} $barmode($node)
  if { $barmode($node) == "val" } {
    .bar$node.mbar.b5 configure -state normal
  } else {
    .bar$node.mbar.b5 configure -state disabled
  }
}

#
# multiFuncgraph: create summary function bargraphs
#
#            win: pathname for canvas
#     leftlabels: list of labels for left side of bargraphs
#          nodes: list of node ids
#       percents: list of list of percentage values
#           tags: list of tags 
#

proc multiFuncgraph {win leftlabels nodes percents tags} {
  global tagcol tagstip
  global minheight
  global racy_progfile

  set num [llength $leftlabels];  #number of bars
  set h [expr ($num+1)*20];       #height of the display

  # -- create or reset canvas
  if { [winfo exists $win] } {
    $win delete all
    $win configure -height [min 420 $minheight]
  } else {
    canvas $win -width 300 -height [min 420 $minheight] -background white
  }
  bind $win <2> "$win scan mark %x %y"
  bind $win <B2-Motion> "$win scan dragto %x %y"
  $win configure -scrollregion [list 0 0 300 $h]

  for {set i 0} {$i<$num} {incr i} {
    # -- display left label, and create bindings for them
    set obj [$win create text 60 [expr 20+$i*20] \
             -text [lindex $leftlabels $i] -anchor e \
             -font -Adobe-Helvetica-Bold-R-Normal--*-120-*]
    $win bind $obj <Button-1> \
              "showBargraph [lindex $nodes $i] {[lindex $leftlabels $i]}"
    $win bind $obj <Button-2> \
              "showText [lindex $nodes $i] {[lindex $leftlabels $i]}"

    # -- display multibar, and create bindings for each subbar
    set p [lindex $percents $i]
    set sum 0.0
    for {set j 0} {$j<[llength $p]} {incr j} {
      if { [lindex $p $j] > 0 } {
        set obj [$win create rectangle [expr 70+$sum*2] [expr 29+$i*20] \
                             [expr 70+($sum+[lindex $p $j])*2] [expr 11+$i*20]\
                             -fill $tagcol([lindex $tags $j]) \
                             -stipple $tagstip([lindex $tags $j])]
        set t [lindex $tags $j]
        $win bind $obj <Button-1> "showFuncgraph $t"
        $win bind $obj <Button-2> "PM_GlobalSelect $racy_progfile global_showFuncTag $t"
        $win bind $obj <Button-3> "PM_GlobalSelect $racy_progfile global_selectFuncTag $t"
        set sum [expr $sum+[lindex $p $j]]
      }
    }
  }
}

#
# multiCollgraph: create collection profile bargraphs
#
#            win: pathname for canvas
#     leftlabels: list of labels for left side of bargraphs
#         idents: list of collection ids
#       percents: list of list of percentage values
#         values: list of values
#           mode: display mode: none or lper or gper
#          scale: scale factor for bars
#

proc multiCollgraph {win leftlabels idents percents \
                     {values {}} {mode none} {scale 1.0}} {
  global minheight
  global TAUDIR

  if { [winfo depth .] == 2 } {
    set cols {black black}
    set stipps [list @$TAUDIR/xbm/gray3.bm {}]
  } else {
    set cols {green red}
    set stipps [list {} {}]
  }
  set num [llength $leftlabels];           #number of bars (lines)
  set h [expr ($num+1)*20];                #screen height
  set nbar [llength [lindex $percents 0]]; #number of subbars per bar

  # -- screen width different with and without right side labels
  if { $mode == "none" } {
    set w 300
  } else {
    set w [expr 300+$nbar*60]
  }

  # -- create or update drawing canvas
  if { [winfo exists $win] } {
    $win delete all
    $win configure -height [min 420 $minheight]
  } else {
    canvas $win -width $w -height [min 420 $minheight] -background white
    bind $win <2> "$win scan mark %x %y"
    bind $win <B2-Motion> "$win scan dragto %x %y"
  }
  $win configure -scrollregion [list 0 0 $w $h]

  for {set i 0} {$i<$num} {incr i} {
    # -- draw left labels
    set obj [$win create text 60 [expr 20+$i*20] \
             -text "          [lindex $leftlabels $i]" -anchor e \
             -font -Adobe-Helvetica-Bold-R-Normal--*-120-*]

    # -- setup bindings for left labels, if necessary
    if { $idents != "" } {
      $win bind $obj <Button-1> "showCollgraph [lindex $idents $i]"
    }

    # -- draw bars and right labels
    set p [lindex $percents $i]
    set sum 0.0
    for {set j 0} {$j<$nbar} {incr j} {
      # -- draw subbars, if value > 0
      set sc [expr [lindex $p $j]*$scale]
      if { $sc > 0 } {
        # -- don't draw first bar (total) if mode is global percent, as this
        # -- bar would be *very* big
        if { $i != 0 || $mode == "lper" || $mode == "none" } {
          set obj [$win create rectangle [expr 70+$sum*2] [expr 29+$i*20] \
                        [expr 70+($sum+$sc)*2] [expr 11+$i*20] \
                        -fill [lindex $cols $j] -stipple [lindex $stipps $j]]
          set sum [expr $sum+$sc]
        }

        # -- draw right labels according to mode
        switch -glob $mode {
          none {}
          [lg]per { $win create text [expr 340+$j*60] [expr 20+$i*20] \
                 -text [format "%.2f%%" [lindex $p $j]] -anchor e \
                 -font -Adobe-Helvetica-Bold-R-Normal--*-120-*}
          val  { $win create text [expr 340+$j*60] [expr 20+$i*20] \
                 -text [lindex [lindex $values $i] $j] -anchor e \
                 -font -Adobe-Helvetica-Bold-R-Normal--*-120-*}
        }
      }
    }
  }
}

#
# readProfile: read in profile data by using the "pprof" tool
#
#         dir: directory in which profile data should be looked for
#

set wantnum 10

proc readProfile {} {
    global depfile \
	    data \
	    nodes \
	    numfunc wantnum \
	    tagname tagcol tagfcol tagstip Colors TextColors Stipples \
	    coll colls \
	    BINDIR REMDIRBIN REMSH TAUDIR

    if {$depfile(host) == "localhost"} {
	set dir "."
	set file "$dir/profile"
	
	# -- open connection to pprof to read profile data
	if { ! [file readable $file.ftab] } {
	    return NOT_OK
	}
	set in [open "|$BINDIR/pprof -d -m -n $wantnum -f $file" r]
    } else {
	set REMBINDIR "$depfile(root)/bin/$depfile(arch)"
	set in [open "|$REMSH $depfile(host) \
		-n \"cd $dir; $REMBINDIR/pprof -d -m -n 10\"" r]
    }
    set n NA

    # -- read header: 1) name of corresponding depfile
    # --              2) number of profiled functions
    gets $in line
    scan $line "%s" depfile(file)
    gets $in line
    scan $line "%d" numfunc

    # -- init
    set depfile(path) $dir/$depfile(file)
    set tagcol(-1) black
    set tagfcol(-1) white
    set tagstip(-1) {}
    set tagname(-1) "-others-"
    set tagcol(-2) white
    set tagfcol(-2) black
    set tagstip(-2) {}
    set tagname(-2) ""
    set i 0
    set colls ""
    
    # -- read line by line
    while {[gets $in line] >= 0} {
	scan $line "%s" node
	if { $node == "coll" } {
	    scan $line "%s %s %d %d %f %d %f" dummy node no local lper remote rper
	    set coll($no,$node,val) [list $local $remote]
	    set coll($no,$node,per) [list $lper $rper]
	} elseif { $node == "cinfo" } {
	    scan $line "%s %s %s %s %d %d %d %d %f %d %f" \
		    dummy no type name elem size dim local lper remote rper
	    lappend colls $no
	    set coll($no,name) $name
	    set coll($no,type) $type
	    set coll($no,info) [list $elem $size $dim]
	    set coll($no,t,val) [list $local $remote]
	    set coll($no,t,per) [list $lper $rper]
	} else {
	    # -- function profile data
	    scan $line "%s %d %s %s %g %f" node tag name mode usec per
	    if { $n != $node && $mode == "incl"} {
		set n $node
		lappend nodes $node
	    }
	    
	    lappend data($node,${mode}tags) $tag
	    lappend data($node,${mode}usecs) $usec
	    lappend data($node,${mode}percent) [format "%4.2f" $per]
	    
	    if { ! [info exists tagcol($tag)] } {
		if { [winfo depth .] == 2 } {
		    set tagcol($tag) black
		    set tagfcol($tag) white
		    set tagstip($tag) @$TAUDIR/xbm/[wraplindex $Stipples $i]
		} else {
		    set tagcol($tag) [wraplindex $Colors $i]
		    set tagfcol($tag) [wraplindex $TextColors $i]
		    set tagstip($tag) {}
		}
		set tagname($tag) $name
		incr i
	    }

	    gets $in line
	    lappend data($node,${mode}text) $line
	}
    }
    if [catch {close $in} errmsg] {
	showError "$readcom: `$errmsg'."
	return NOT_OK
    }
    
    if { ! [info exists nodes] } {
	return NOT_OK
    }
    return OK
}

#
# funcgraphmenu: create menubar for function profile windows
#
#        parent: pathname for parent of menubar
#        prefix: prefix for global variables to store selections
#          func: function to call for redraw
#     node,name: arguments to func
#

proc funcgraphmenu {parent prefix func node name} {
  global myself

  set menubar $parent.mbar
  frame $menubar -relief raised -borderwidth 2
  menubutton $menubar.b1 -text File -menu $menubar.b1.m1 -underline 0
  menu $menubar.b1.m1
  $menubar.b1.m1 add command -label "Print" -underline 0 \
                    -command "printCanvas .func${node}.bar func.$node"
  $menubar.b1.m1 add separator
  $menubar.b1.m1 add command -label "Close" -underline 0 \
                    -command "destroy .func$node"

  menubutton $menubar.b2 -text Value -menu $menubar.b2.m1 -underline 0
  menu $menubar.b2.m1
  submenu2 $menubar.b2.m1 ${prefix}value $func $node {$name}

  menubutton $menubar.b4 -text Mode -menu $menubar.b4.m1 -underline 0
  menu $menubar.b4.m1
  submenu4 $menubar.b4.m1 ${prefix}mode $func $node {$name}

  menubutton $menubar.b5 -text Units -menu $menubar.b5.m1 -underline 0
  menu $menubar.b5.m1
  submenu5 $menubar.b5.m1 ${prefix}unit $func $node {$name}

  menubutton $menubar.b6 -text Help -menu $menubar.b6.m1 -underline 0
  menu $menubar.b6.m1
  $menubar.b6.m1 add command -label "on function profile" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.2.4-func 1\]"

  pack $menubar.b1 $menubar.b2 $menubar.b4 $menubar.b5\
       -side left -fill x -padx 5
  pack $menubar.b6 -side right -fill x -padx 5
}

#
# showFuncgraph: display function profile window
#
#           tag: Sage++ id of function
#

proc showFuncgraph {tag} {
  global data
  global funcvalue funcmode funcunit
  global proforder profvalue profmode profunit
  global tagname

  if { ! [winfo exists .func$tag] } {
    set funcvalue($tag) $profvalue(all)
    set funcmode($tag) $profmode(all)
    set funcunit($tag) $profunit(all)
    set win [toplevel .func$tag]

    set name $tagname($tag)
    wm title $win "$name profile"
    wm minsize $win 300 50

    funcgraphmenu .func$tag func redrawFuncgraph $tag {}
    pack .func$tag.mbar -side top -fill x

    redrawFuncgraph $tag {}
    bind .func$tag.bar <2> ".func$tag.bar scan mark %x %y"
    bind .func$tag.bar <B2-Motion> ".func$tag.bar scan dragto %x %y"

    scrollbar .func$tag.s1 -orient vert -relief sunken \
                           -command ".func$tag.bar yview"
    .func$tag.bar configure -yscrollcommand ".func$tag.s1 set"

    button .func$tag.but -text "close" -command "destroy .func$tag"
    pack .func$tag.but -side bottom -fill x
    pack .func$tag.s1 -side right -fill y -expand yes
    pack .func$tag.bar -side top -fill both -expand yes
  } else {
    raise .func$tag
  }
}

#
# redrawFuncgraph: create or update contents of function profile window
#
#             tag: Sage++ id of function
#           dummy: unused parameter so it can be used with the generic
#                  menu construction functions
#

proc redrawFuncgraph {tag dummy} {
  global data
  global funodes fulabels
  global funcvalue funcmode funcunit

  # -- compute arguments for bargraph widget
  set value $funcvalue($tag)
  foreach n $funodes {
    if { $n == "" } {
      lappend pers 0.0
      lappend vals 0.0
      lappend tags -2
    } else {
      set i [lsearch -exact $data($n,${value}tags) $tag]
      if { $i == -1 } {
        lappend pers 0.0
        lappend vals 0.0
      } else {
        lappend pers [lindex $data($n,${value}percent) $i]
        lappend vals [lindex $data($n,${value}usecs) $i]
      }
      lappend tags $tag
    }
  }
  if { $funcmode($tag) == "val" } {
    set vmax [vectormax vals]
    set pers [vectorpercent vals $vmax 80.0]
  }

  # -- create display using bargraph widget
  bargraph .func$tag.bar $fulabels $pers [vectorop vals / $funcunit($tag)] \
                         $tags $funodes $funcmode($tag)
  if { $funcmode($tag) == "val" } {
    .func$tag.mbar.b5 configure -state normal
  } else {
    .func$tag.mbar.b5 configure -state disabled
  }
}

#
# collgraphmenu: create menubar for collection profile windows
#
#        parent: pathname for parent of menubar
#        prefix: prefix for global variables to store selections
#          func: function to call for redraw
#     node,name: arguments to func
#

proc collgraphmenu {parent prefix func node name} {
  global myself

  set menubar $parent.mbar
  frame $menubar -relief raised -borderwidth 2
  menubutton $menubar.b1 -text File -menu $menubar.b1.m1 -underline 0
  menu $menubar.b1.m1
  $menubar.b1.m1 add command -label "Print" -underline 0 \
                    -command "printCanvas .coll${node}.bar coll.$node"
  $menubar.b1.m1 add separator
  $menubar.b1.m1 add command -label "Close" -underline 0 \
                    -command "destroy .coll$node"

  menubutton $menubar.b6 -text Mode -menu $menubar.b6.m1 -underline 0
  menu $menubar.b6.m1
  submenu6 $menubar.b6.m1 ${prefix}mode $func $node {$name}

  menubutton $menubar.b7 -text Help -menu $menubar.b7.m1 -underline 0
  menu $menubar.b7.m1
  $menubar.b7.m1 add command -label "on collection profile" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.3.1-coll 1\]"

  pack $menubar.b1 $menubar.b6 -side left -fill x -padx 5
  pack $menubar.b7 -side right -fill x -padx 5
}

#
# showCollgraph: display collection profile window
#
#         ident: collection id
#

proc showCollgraph {ident} {
  global coll
  global collmode profmode

  if { ! [winfo exists .coll$ident] } {
    set win [toplevel .coll$ident]

    set name $coll($ident,name)
    wm title $win "$name profile"
    wm minsize $win 300 50

    if { $profmode(all) == "per" } {
      set collmode($ident) lper
    } else {
      set collmode($ident) $profmode(all)
    }

    collgraphmenu .coll$ident coll redrawCollgraph $ident {}
    pack .coll$ident.mbar -side top -fill x

    set e [lindex $coll($ident,info) 0]
    set s [lindex $coll($ident,info) 1]
    set d [lindex $coll($ident,info) 2]
    set t [expr $e*$s]
    if { $t > 1000000 } {
      set t "[format {%.1f MB} [expr $t/1e6]]"
    } elseif { $t > 1000 } {
      set t "[format {%.1f kB} [expr $t/1e3]]"
    } else {
      set t "$t B"
    }
    frame .coll$ident.info -background white -relief raised -borderwidth 2
    label .coll$ident.l1 -text "$coll($ident,type) $name" \
                         -background white -foreground black
    label .coll$ident.l2 -text "$e elements of size $s \[$t\]" \
                         -background white -foreground black
    label .coll$ident.l3 -text "$d-dimensional shape" \
                         -background white -foreground black
    pack .coll$ident.l1 .coll$ident.l2 .coll$ident.l3 \
                        -side top -anchor w -in .coll$ident.info
    pack .coll$ident.info -side top -fill x

    redrawCollgraph $ident {}

    bind .coll$ident.bar <2> ".coll$ident.bar scan mark %x %y"
    bind .coll$ident.bar <B2-Motion> ".coll$ident.bar scan dragto %x %y"

    scrollbar .coll$ident.s1 -orient vert -relief sunken \
                           -command ".coll$ident.bar yview"
    .coll$ident.bar configure -yscrollcommand ".coll$ident.s1 set"

    button .coll$ident.but -text "close" -command "destroy .coll$ident"
    pack .coll$ident.but -side bottom -fill x
    pack .coll$ident.s1 -side right -fill y -expand yes
    pack .coll$ident.bar -side top -fill both -expand yes
  } else {
    raise .coll$ident
  }
}

#
# redrawCollgraph: create or update contents of collection profile window
#
#           ident: collection id
#           dummy: unused parameter so it can be used with the generic
#                  menu construction functions
#

proc redrawCollgraph {ident dummy} {
  global conodes colabels
  global coll
  global collmode

  set mode $collmode($ident)
  if { $mode == "gper" || $mode == "val" } {
    # -- global percent / value display mode
    # -- compute data from local percent available from pprof
    set max   0.0
    foreach n $conodes {
      if [ info exists coll($ident,$n,val) ] {
        set v $coll($ident,$n,val)
        lappend value $v
        if { $n != "t" } {
          set s [expr [lindex $v 0]+[lindex $v 1]]
          if { $s > $max } { set max $s }
        }
      } else {
        lappend value [list 0.0 0.0]
      }
    }
    set total [expr \
          double([lindex $coll($ident,t,val) 0]+[lindex $coll($ident,t,val) 1])]
    set scale [expr $total/$max]

    foreach n $conodes {
      if [ info exists coll($ident,$n,val) ] {
        set v $coll($ident,$n,val)
        lappend percent [list [expr [lindex $v 0]/$total*100.0] \
                              [expr [lindex $v 1]/$total*100.0]]
      } else {
        lappend percent [list 0.0 0.0]
      }
    }
  } elseif { $mode == "lper" } {
    # -- local percent display mode
    foreach n $conodes {
      if [ info exists coll($ident,$n,per) ] {
        lappend percent $coll($ident,$n,per)
      } else {
        lappend percent [list 0.0 0.0]
      }
    }
    set value {}
    set scale 1.0
  }
  multiCollgraph .coll$ident.bar $colabels {} $percent $value $mode $scale
}

#
# computeMultiBars: compute data for summary function profile
#                   as well as often used lists of labels
#

proc computeMultiBars {} {
  global data
  global nodes funodes conodes
  global fulabels colabels clabels
  global exclpercent collpercent
  global coll colls
  global minheight

  foreach n $nodes {
    set p ""
    set na [vectorsum data($n,exclpercent)]
    foreach t $data(m,excltags) {
      set i [lsearch -exact $data($n,excltags) $t]
      if { $i == -1 } {
        lappend p 0.0
      } else {
        set x [lindex $data($n,exclpercent) $i]
        if { $t == -1 } {
          lappend p $na
        } else {
          set na [expr $na-$x]
          lappend p $x
        }
      }
    }
    set data($n,meanper) $p
  }

  set fulabels [list "mean" "min" "max" {}]
  set colabels [list "total" {}]
  set exclpercent \
      [list $data(m,exclpercent) $data(<,meanper) $data(>,meanper) {}]
  set funodes "m < > {} [lrange $nodes 0 [expr [llength $nodes]-5]]"
  set conodes "t {} [lrange $nodes 0 [expr [llength $nodes]-5]]"

  foreach n [lrange $nodes 0 [expr [llength $nodes]-5]] {
    lappend fulabels [format "node %2d" $n]
    lappend colabels [format "node %2d" $n]
    lappend exclpercent $data($n,meanper)
  }

  set clabels ""
  set collpercent ""
  foreach c $colls {
    lappend clabels $coll($c,name)
    lappend collpercent $coll($c,t,per)
  }

  set minheight [expr ([max [llength $fulabels] [llength $clabels]]+1)*20]
}

#
# redraw: redraw all windows currently displayed, after user selected
#         global order from Configure main menu
#
# dummy1:
# dummy2: unused parameters so it can be used with the generic
#         menu construction functions
#

proc redraw {dummy1 dummy2} {
  global myself
  global data
  global funodes fulabels 
  global barorder barmode barvalue barunit
  global textorder1 textorder2
  global funcmode funcvalue funcunit
  global proforder profvalue profmode profunit
  global collmode colls

  # -- redraw node and text node profile windows
  set i 0
  foreach n $funodes {
    if [winfo exists .bar$n] {
      set barvalue($n) $profvalue(all)
      if { $proforder(all) == "all" } {
        set barorder($n) $n
      } else {
        set barorder($n) $proforder(all)
      }
      set barmode($n) $profmode(all)
      set barunit($n) $profunit(all)
      redrawBargraph $n [lindex $fulabels $i]
    }
    if [winfo exists .text$n] {
      set textorder1($n) $profvalue(all)
      if { $proforder(all) == "all" } {
        set textorder2($n) $n
      } else {
        set textorder2($n) $proforder(all)
      }
      redrawText $n [lindex $fulabels $i]
    }
    incr i
  }

  # -- redraw function profile windows
  foreach t $data(m,excltags) {
    if [winfo exists .func$t] {
      set funcvalue($t) $profvalue(all)
      set funcmode($t) $profmode(all)
      set funcunit($t) $profunit(all)
      redrawFuncgraph $t {}
    }
  }

  # -- redraw collection profile windows
  foreach c $colls {
    if [winfo exists .coll$c] {
      if { $profmode(all) == "per" } {
        set collmode($c) lper
      } else {
        set collmode($c) $profmode(all)
      }
      redrawCollgraph $c {}
    }
  }

  # -- redraw function legend
  if [winfo exists .$myself.led.can] {
    redrawLegend .$myself.led.can
  }

  if { $profmode(all) == "val" } {
    .$myself.mbar.b2.m1 entryconfigure 3 -state normal
  } else {
    .$myself.mbar.b2.m1 entryconfigure 3 -state disabled
  }
}

#
# showLegend: display function legend
#

proc showLegend {} {
  global myself

  if { ! [winfo exists .$myself.led] } {
    toplevel .$myself.led
    wm title .$myself.led "Function Legend"

    # -- create canvas and "close" button
    canvas .$myself.led.can -width 250 -background white
    pack .$myself.led.can

    button .$myself.led.b -text close -command "destroy .$myself.led"
    pack .$myself.led.b -fill x

    # -- draw color legend
    redrawLegend .$myself.led.can
  } else {
    raise .$myself.led
  }
}

#
# redrawLegend: draw or update function legend canvas
#
#          can: canvas id
#

proc redrawLegend {can} {
  global tagcol tagstip tagname
  global pr_sel_tag
  global racy_progfile

  # -- generate alphabetical list of functions
  foreach t [array names tagcol] {
    if { $t != -2 && $t != -1 } {
      lappend map [list $tagname($t) $t $tagcol($t) $tagstip($t)]
    }
  }
  set map [lsort $map]
  lappend map [list $tagname(-1) $t $tagcol(-1) {}]

  # -- set height of map
  set num [llength $map];    # number of funcs
  set h [expr ($num+1)*20];  # height of canvas
  $can configure -height $h

  # -- draw color legend
  $can delete all
  set i 11
  foreach m $map {
    set t [lindex $m 1]
    $can create rectangle 10 [expr $i] 28 [expr 18+$i] \
         -fill [lindex $m 2] -tag fl$t -stipple [lindex $m 3]
    if { $t == $pr_sel_tag } {
      $can create text 40 [expr 9+$i] -text [lindex $m 0] \
           -anchor w -tag fl$t -fill red \
           -font -Adobe-Helvetica-Bold-R-Normal--*-120-*
    } else {
      $can create text 40 [expr 9+$i] -text [lindex $m 0] -anchor w -tag fl$t \
           -font -Adobe-Helvetica-Bold-R-Normal--*-120-*
    }
    $can bind fl$t <Button-1> "showFuncgraph $t"
    $can bind fl$t <Button-2> "PM_GlobalSelect $racy_progfile global_showFuncTag $t"
    $can bind fl$t <Button-3> "PM_GlobalSelect $racy_progfile global_selectFuncTag $t"
    incr i 20
  }
}

#
# setNumFunc: set maximal number of functions included in profile data displays
#

proc setNumFunc {} {
  global myself
  global depfile
  global wantnum

  set copy $wantnum

  if { ! [winfo exists .$myself.num] } {
    toplevel .$myself.num
    wm title .$myself.num "Set #Functions"

    numberInput .$myself.num.i "Number of functions to display:" wantnum
    pack .$myself.num.i -side top

    frame .$myself.num.b -relief sunken -bd 1
    pack .$myself.num.b -side left -padx 15 -pady 10
    button .$myself.num.b.b1 -text "set" -command {
      destroy .$myself.num
      loadProfile
    }
    pack .$myself.num.b.b1 -side top -padx 5 -pady 5

    button .$myself.num.b2 -text "cancel" -command "
      set wantnum $copy
      destroy .$myself.num
    "
    pack .$myself.num.b2 -side right -padx 15

    bind .$myself.num.i.e <Return> {.$myself.num.b.b1 invoke}
    bind .$myself.num <Return> {.$myself.num.b.b1 invoke}
  } else {
    raise .$myself.num
  }
}

#
#  mainmenu: create menubar for main window
#
#    parent: pathname for parent of menubar
#    prefix: prefix for global variables to store selections
#      func: function to call for redraw
# node,name: arguments to func
#

proc mainmenu {parent prefix func node name} {
  global myself

  set menubar $parent.mbar
  frame $menubar -relief raised -borderwidth 2

  menubutton $menubar.b1 -text File -menu $menubar.b1.m1 -underline 0
  menu $menubar.b1.m1
  $menubar.b1.m1 add command -label "set #Functions" -underline 0 \
                    -command setNumFunc
  $menubar.b1.m1 add command -label "show Legend" -underline 5 \
                    -command showLegend
  $menubar.b1.m1 add command -label "print Functions" -underline 6 \
                    -command "printCanvas .$myself.fu.bar profile.fu"
  $menubar.b1.m1 add command -label "print Collections" -underline 6 \
                    -command "printCanvas .$myself.co.bar profile.co"
  $menubar.b1.m1 add separator
  $menubar.b1.m1 add command -label "Exit" -underline 0 -command "exit"

  menubutton $menubar.b2 -text Configure -menu $menubar.b2.m1 -underline 0
  menu $menubar.b2.m1

  $menubar.b2.m1 add cascade -label "Value" -menu $menubar.b2.m1.1 -underline 0
  menu $menubar.b2.m1.1
  submenu2 $menubar.b2.m1.1 ${prefix}value $func $node {$name}

  $menubar.b2.m1 add cascade -label "Order" -menu $menubar.b2.m1.2 -underline 0
  menu $menubar.b2.m1.2
  submenu3 $menubar.b2.m1.2 ${prefix}order $func $node {$name}

  $menubar.b2.m1 add cascade -label "Mode" -menu $menubar.b2.m1.3 -underline 0
  menu $menubar.b2.m1.3
  submenu4 $menubar.b2.m1.3 ${prefix}mode $func $node {$name}

  $menubar.b2.m1 add cascade -label "Unit" -menu $menubar.b2.m1.4 -underline 0
  menu $menubar.b2.m1.4
  submenu5 $menubar.b2.m1.4 ${prefix}unit $func $node {$name}
  $menubar.b2.m1 entryconfigure 3 -state disabled

  menubutton $menubar.b3 -text Help -menu $menubar.b3.m1 -underline 0
  menu $menubar.b3.m1
  $menubar.b3.m1 add command -label "on $myself" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1-$myself 1\]"
  $menubar.b3.m1 add separator
  $menubar.b3.m1 add command -label "on menubar" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.1-menu 1\]"
  $menubar.b3.m1 add command -label "on function summary" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.2-funcsum 1\]"
  $menubar.b3.m1 add command -label "on collection summary" -underline 3 \
                   -command "xsend tau \[list showHelp $myself 1.3-collsum 1\]"
  $menubar.b3.m1 add separator
  $menubar.b3.m1 add command -label "on using help" -underline 3 \
                   -command "xsend tau {showHelp general 1-help 1}"

  createToolMenu $menubar.b4

  pack $menubar.b1 $menubar.b2 -side left -padx 5
  pack $menubar.b3 $menubar.b4 -side right -padx 5
}

#
# loadProfile: switch to another application
#              this is invoked from the TAU master control window
#

proc loadProfile {} {
  global myself
  global data
  global numfunc
  global nodes funodes fulabels conodes colabels exclpercent
  global barorder barmode barvalue barunit
  global textorder1 textorder2
  global funcmode funcvalue funcunit
  global collmode
  global proforder profvalue profmode profunit
  global tagname tagcol tagfcol tagstip
  global colls coll clabels collpercent
  global depfile

  set dir  $depfile(dir)
  set root $depfile(root)
  set arch $depfile(arch)

  foreach n $funodes {
    if [winfo exists .bar$n]  { destroy .bar$n  }
    if [winfo exists .text$n] { destroy .text$n }
  }
  foreach t $data(m,excltags) {
    if [winfo exists .func$t] { destroy .func$t }
  }
  foreach c $colls {
    if [winfo exists .coll$c] { destroy .coll$c }
  }
  if [winfo exists .$myself.num] { destroy .$myself.num }
  if [winfo exists .$myself.led] { destroy .$myself.led }

  # -- destroy old profile data database
  unset data
  unset nodes funodes fulabels conodes colabels exclpercent
  unset tagname tagcol tagfcol tagstip
  unset clabels collpercent
  if { [info exists barorder]   } { unset barorder barmode barvalue barunit }
  if { [info exists textorder1] } { unset textorder1 textorder2 }
  if { [info exists funcmode]   } { unset funcmode funcvalue funcunit }
  if { [info exists collmode]   } { unset collmode }
  if { [info exists colls]      } { unset colls }
  if { [info exists coll]       } { unset coll }

  # -- reset global configuration parameters
  set profvalue(all) excl
  set proforder(all) m
  set profmode(all) per
  set profunit(all) 1.0

  # -- read new profile data
  if { [readProfile] == "NOT_OK" } {
    showError "No profile data available in directory $dir."
    after 1 exit
    return
  }
  xsend tau "set depfile(numprof) $numfunc"

  # -- redraw summary bargraphs
  computeMultiBars
  multiFuncgraph .$myself.fu.bar $fulabels $funodes \
                 $exclpercent $data(m,excltags)
  multiCollgraph .$myself.co.bar $clabels $colls $collpercent

  # rebuilt configure-order-node submenu
  .$myself.mbar.b2.m1.2.1 delete 0 last
  set i 0
  foreach n $funodes {
    if { $n != {} } {
      .$myself.mbar.b2.m1.2.1 add radiobutton -label [lindex $fulabels $i] \
                  -variable proforder(all) -value $n \
                  -command "redraw all {}"
    } else {
      .$myself.mbar.b2.m1.2.1 add separator
    }
    incr i
  }
}

#
# createWindow: create and display main window of racy
#

proc createWindow {} {
  global myself
  global TAUDIR
  global data
  global fulabels funodes clabels colls collpercent exclpercent

  toplevel .$myself
  wm title .$myself "RACY"
  wm minsize .$myself 300 50
  wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm

  mainmenu .$myself prof redraw all {}
  pack .$myself.mbar -side top -fill x

  frame .$myself.fu
  label .$myself.fu.l1 -text Functions -relief raised
  multiFuncgraph .$myself.fu.bar $fulabels $funodes \
                 $exclpercent $data(m,excltags)
  scrollbar .$myself.fu.s1 -orient vert -relief sunken \
                        -command ".$myself.fu.bar yview"
  .$myself.fu.bar configure -yscrollcommand ".$myself.fu.s1 set"
  pack .$myself.fu.l1 -side top -fill x -expand yes
  pack .$myself.fu.s1 -side right -fill y -expand yes
  pack .$myself.fu.bar -side top -fill both -expand yes

  frame .$myself.co
  label .$myself.co.l1 -text Collections -relief raised
  multiCollgraph .$myself.co.bar $clabels $colls $collpercent
  scrollbar .$myself.co.s1 -orient vert -relief sunken \
                        -command ".$myself.co.bar yview"
  .$myself.co.bar configure -yscrollcommand ".$myself.co.s1 set"
  pack .$myself.co.l1 -side top -fill x -expand yes
  pack .$myself.co.s1 -side right -fill y -expand yes
  pack .$myself.co.bar -side top -fill both -expand yes

  pack .$myself.fu .$myself.co -side left -padx 15 -pady 15 \
                               -fill both -expand yes
}

#
# getColorCode: return list of actual function name <-> color mapping
#

proc getColorCode {} {
  global tagcol

  foreach t [array names tagcol] {
    if { $t != -2 } { lappend result [list $t $tagcol($t)] }
  }
  return $result
}


proc Tool_AcceptChanges {progfile flag} {
    global myself depfile \
	    showFile selectBox

    switch $flag {

        d {
        }


        a {
        }


        u {
        }
	
	e {
	    loadProfile
	}

	p {
	    set status [PM_Status]
	    if {$status != "NO_PROJECT"} {
		set depfile(project) [lindex $pm_status 0]
		set depfile(host)    [lindex $pm_status 1]
		set depfile(arch)    [lindex $pm_status 2]
		set depfile(root)    [lindex $pm_status 3]
		set depfile(dir)     [lindex $pm_status 4]
	    }
	}
    }
}



#  initRacy - Racy only works with simple pc++ projects, ie, pc++ projects
#             with only one program file.
proc initRacy {} {
    global racy_progfile

    set files [PM_GetFiles]
    if {[llength $files] > 1 || [llength $files] == 0} {
	showError "Racy is only compatible with single-file pC++ projects."
	exit
    }
    if {[file extension [lindex $files 0]] != ".pc"} {
	showError "Racy is only compatible with single-file pC++ projects."
	exit
    }

    set racy_progfile [lindex $files 0]
}


# ------------
# -- main code
# ------------

set profvalue(all) excl
set proforder(all) m
set profmode(all) per
set profunit(all) 1.0

if {$argc == 0} {
  set parg [pwd]
} elseif {$argc == 1} {
    set parg [lindex $argv 0]
    if {[file extension $parg] != ".pmf"} {
	set parg "$parg.pmf"
    }
} else {
  puts stderr "usage: $myself \[\[host:\]projFile \| \[host:\]directory\]"
  exit
}

# Init the project manager (taud)
launchTauDaemon -waitfor
PM_AddTool $myself
PM_AddGlobalSelect $myself {global_selectFuncTag}

# Initialize the project
# Coordinate w/ PM
set pm_status [PM_Status]
if {[lindex $pm_status 0] == "NO_PROJECT"} {
    # Open or create a project
    set colon [string first ":" $parg]
    if {$colon > 0} {
        set hostarg [string range $parg 0 [expr $colon - 1]]
        set patharg [string range $parg [expr $colon + 1] end]
    } else {
        set hostarg localhost
        set patharg $parg
    }

    set projfile [PM_OpenProject $patharg $hostarg]
    if {$projfile == "NO_PROJECT"} {
        showError "No project opened!"
        exit
    }
    set pm_status [PM_Status]
}

set depfile(project) [lindex $pm_status 0]
set depfile(host)    [lindex $pm_status 1]
set depfile(arch)    [lindex $pm_status 2]
set depfile(root)    [lindex $pm_status 3]
set depfile(dir)     [lindex $pm_status 4]

# Tool Init
initRacy
if { [readProfile] == "NOT_OK" } {
    showError "No profile data available in current directory."
    exit
}

# -- create new toplevel window
computeMultiBars
createWindow
launchTAU

# -- notify tau about number of functions profiled
xsend tau "set depfile(numprof) $numfunc"

removeMessage
