#*********************************************************************#
#*                TAU/pC++/Sage++  Copyright (C) 1994                *#
#*                        Jerry Manic Saftware                       *#
#*  Indiana University  University of Oregon  University of Rennes   *#
#*********************************************************************#
displayMessage hourglass "Loading $myself..."

source "$TAUDIR/inc/selectfile.tcl"
source "$TAUDIR/inc/depfile.tcl"
source "$TAUDIR/inc/selectfunc.tcl"
source "$TAUDIR/inc/stack.tcl"
source "$TAUDIR/inc/help.tcl"
source "$TAUDIR/inc/pm_interface.tcl"


# A debugging ramework, for conditional scaffolding.
set DEBUG_SET 0


proc DEBUG {} {
    global DEBUG_SET

    return $DEBUG_SET
}



set selectfunc(show) showFuncTag
set selectfunc(select) global_selectFuncTag

set ${myself}winnr 0
# Global for setting SAGROOT in project options
set srVar ""


#
# loadCgmIfNeeded - check if the depfile needs to be loaded and load it if so.
#        progfile : the depfile to load
#
proc loadCgmIfNeeded {progfile} {
    if [Cgm_IsDepLoaded $progfile] {
	return 1
    } else {
	Cgm_LoadDep $progfile -dumpall
    }
}


#
# showFuncTag: implementation of global feature "showFunction" for this tool
#              show small function information window in form of a table
#    progfile: depfile containing the tag
#         tag: id of function to show
#

proc showFuncTag {progfile tag} {
    global myself
    global ${myself}winnr

    if {[loadCgmIfNeeded $progfile] == "NOT_OK" } {
	showError "No valid depfile for $progfile."
	return
    }

    if {[Cgm_FuncInfo $progfile $tag name] == "NOT_OK" } {
	showError "No info about tag $tag."
	return
    } elseif { ! [winfo exists .$myself.t$tag] } {
	set name  [Cgm_FuncInfo $progfile $tag name]
	set class [Cgm_FuncInfo $progfile $tag class]
	set file  [Cgm_FuncInfo $progfile $tag file]
	set ptype [lindex [Cgm_FuncInfo $progfile $tag type] 0]
	set type  [lindex [Cgm_FuncInfo $progfile $tag type] 1]
	set num   [Cgm_FuncInfo $progfile $tag num]
	set calls [llength [Cgm_FuncInfo $progfile $tag calls]]
	
	if { [lindex $class 0] == "-" } {
	    set class ""
	} elseif { [lindex $class 1] == "-" } {
	    set class [lindex $class 0]
	} else {
	    set class "Collection [lindex $class 0]"
	}

	if { [lindex $file 0] == "0" } {
	    set file ""
	} else {
	    set file "[lindex $file 1]:[lindex $file 0]"
	}

	switch $ptype {
	    seq { set ptype "sequential" }
	    par { set ptype "parallel" }
	}

	switch $type {
	    Mem { set type "Member Function" }
	    MoE { set type "Method of Element" }
	    MoC { set type "Member of Collection" }
	    Ord { set type "Ordinary Function" }
	    Main { set type "Main Function" }
	    Prcs { set type "Fortran-M Process" }
	}

	set w [toplevel .$myself.t$tag]
	wm title $w "Function Info"

	button $w.b -text "close" -command "destroy $w"
	frame $w.l
	frame $w.r
	
	pack $w.b -side bottom -fill x
	pack $w.l $w.r -side left -padx 5 -pady 5
	
	label $w.l1 -text "name:"
	label $w.l2 -text "class:"
	label $w.l3 -text "file:"
	label $w.l4 -text "type:"
	label $w.l5 -text "call sites:"
	label $w.l6 -text "calls:"
	pack $w.l1 $w.l2 $w.l3 $w.l4 $w.l5 $w.l6 -in $w.l -side top -anchor e
	
	label $w.r1 -text $name -relief sunken -anchor w
	label $w.r2 -text $class -relief sunken -anchor w
	label $w.r3 -text $file -relief sunken -anchor w
	label $w.r4 -text "$ptype $type" -relief sunken -anchor w
	label $w.r5 -text $num -relief sunken -anchor w
	label $w.r6 -text $calls -relief sunken -anchor w
	pack $w.r1 $w.r2 $w.r3 $w.r4 $w.r5 $w.r6 -in $w.r -side top -fill x
    } else {
	raise .$myself.t$tag
    }
}

#
# selectFuncTag: implementation of global feature "selectFunction"
#                this tool doesn't provide one
#

proc selectFuncTag {progfile tag} {}

#
# bitmapbutton: button widget which has a text AND a bitmap
#
#          win: pathname for new button
#       bitmap: bitmap to display
#         text: text label to display
#      command: command to execute if button is pressed
#

proc bitmapbutton {win bitmap text command} {
    frame $win
    button $win.b -bitmap "$bitmap" -command "$command" -back white -fore black \
	-highlightthickness 0
    button $win.t -text "$text" -command "$command" -back white -fore black \
	-takefocus 0 -highlightthickness 0
    bind $win.t <Enter> "$win.b configure -state active"
    bind $win.t <Leave> "$win.b configure -state normal"
    bind $win.b <Enter> "$win.t configure -state active"
    bind $win.b <Leave> "$win.t configure -state normal"
    pack $win.b $win.t -side top -fill x
    return $win
}

#
# createWindow: create and display main window for TAU master control
#

proc createWindow {} {
    global myself
    global TAUDIR
    global depfile

    toplevel .$myself
    wm title .$myself "TAU"
    wm iconbitmap .$myself @$TAUDIR/xbm/$myself.xbm
    wm minsize .$myself 655 445

    # -- menubar
    frame .$myself.bar -relief raised -borderwidth 2
    menubutton .$myself.bar.b1 -text File -menu .$myself.bar.b1.m1 -underline 0 
    menu .$myself.bar.b1.m1
    .$myself.bar.b1.m1 add command -label "Load New Project" -underline 5 \
	-command "tau_OpenProject"
    .$myself.bar.b1.m1 add command -label "Copy Project (Backup)" -underline 0 \
	-command {PM_ChangeProjectName; set depfile(project) [lindex [PM_Status] 0]}
    .$myself.bar.b1.m1 add command -label "Add Source File" -underline 0 \
	-command "tau_AddFile"
    .$myself.bar.b1.m1 add command -label "Delete Selected Source File" -underline 0 \
	-command "tau_DeleteFile"
    .$myself.bar.b1.m1 add command -label "View/Change Project Options" \
	-underline 0 \
	-command "tau_ProjectOptions"
    .$myself.bar.b1.m1 add command -label "Select function" -underline 0 \
	-command "selectFunction"
    .$myself.bar.b1.m1 add separator
    .$myself.bar.b1.m1 add command -label "Exit"  -underline 0 -command "atExit"
    
    menubutton .$myself.bar.b2 -text Appl -menu .$myself.bar.b2.m1 -underline 0 
    menu .$myself.bar.b2.m1

    createToolMenu .$myself.bar.b4
    menubutton .$myself.bar.b3 -text Help -menu .$myself.bar.b3.m1 -underline 0 
    menu .$myself.bar.b3.m1
    .$myself.bar.b3.m1 add command -label "on $myself" -underline 3 \
	-command "showHelp $myself 1-$myself 1"
    .$myself.bar.b3.m1 add separator
    .$myself.bar.b3.m1 add command -label "on menubar" -underline 3 \
	-command "showHelp $myself 1.1-menu 1"
    .$myself.bar.b3.m1 add command -label "on display area" -underline 3 \
	-command "showHelp $myself 1.2-display 1"
    .$myself.bar.b3.m1 add command -label "on command buttons" -underline 3 \
	-command "showHelp $myself 1.3-buttons 1"
    .$myself.bar.b3.m1 add separator
    .$myself.bar.b3.m1 add command -label "on using help" -underline 3 \
	-command "showHelp general 1-help 1"

    pack .$myself.bar.b1 .$myself.bar.b2 -side left -padx 5
    pack .$myself.bar.b3 .$myself.bar.b4 -side right -padx 5
    pack .$myself.bar -side top -fill x

    # -- display area
    frame .$myself.info1
    frame .$myself.info1.left
    label .$myself.info1.l1 -text "host:"
    label .$myself.info1.l2 -text "dir:"
    label .$myself.info1.l3 -text "project:"
    pack .$myself.info1.l1 .$myself.info1.l2 .$myself.info1.l3\
	-side top -anchor e -in .$myself.info1.left
    pack .$myself.info1.left -side left -padx 5

    frame .$myself.info1.right
    label .$myself.info1.r1 \
	-relief sunken \
	-anchor w \
	-textvariable depfile(host)
    label .$myself.info1.r2 \
	-relief sunken \
	-anchor w \
	-textvariable depfile(dir)
    label .$myself.info1.r3 \
	-relief sunken \
	-anchor w \
	-textvariable depfile(project)
    pack .$myself.info1.r1 .$myself.info1.r2 .$myself.info1.r3 \
	-side top -anchor w -in .$myself.info1.right -fill x -expand 1 -ipadx 10
    pack .$myself.info1.right -side right -padx 5 -fill x -expand 1

    frame .$myself.info2
    #Modified 7/9/96, to accomodate new file information. SAS

    set temp [frame .$myself.info2.ext]
    frame $temp.int
    frame $temp.int.top
    label $temp.int.top.filelabel \
	-text File
    label $temp.int.top.langlabel \
	-text Language
    pack $temp.int.top.filelabel \
	-side left \
	-padx 10 \
	-anchor nw 
    pack $temp.int.top.langlabel \
	-side right \
	-padx 20 \
	-anchor ne
    pack $temp.int.top \
	-side top \
	-fill x
    frame $temp.int.bot
   listbox $temp.int.bot.list \
	-width 75 \
	-font -*-courier-bold-r-*-*-12-*-*-*-*-*-*-* \
	-yscrollcommand "$temp.int.sb set"
    pack $temp.int.bot.list \
	-side left \
	-anchor nw \
	-fill y \
	-expand 1
    pack $temp.int.bot \
	-side left \
	-expand 1 \
	-fill y \
	-anchor n
    scrollbar $temp.int.sb \
	-command "$temp.int.bot.list yview" 
    pack $temp.int.sb \
	-side left \
	-fill y \
	-expand 1
    pack $temp.int \
	-side top \
	-anchor n \
	-expand 1 \
	-fill y
    pack $temp \
	-side top \
	-expand 1 \
	-anchor n \
	-fill y

    pack .$myself.info1 -side top -padx 10 -pady 10 -fill x
    pack .$myself.info2 -side top -padx 10 -pady 10 -fill both -expand 1

    # -- command button bar
    frame .$myself.coms
    bitmapbutton .$myself.b0 @$TAUDIR/xbm/cosy.xbm cosy "launch cosy .cosy"
    bitmapbutton .$myself.b1 @$TAUDIR/xbm/mighty.xbm mighty "launch mighty .mighty"
    bitmapbutton .$myself.b2 @$TAUDIR/xbm/fancy.xbm fancy "launch fancy .fancy"
    bitmapbutton .$myself.b3 @$TAUDIR/xbm/cagey.xbm cagey "launch cagey .cagey"
    bitmapbutton .$myself.b4 @$TAUDIR/xbm/classy.xbm classy \
	"launch classy .classy"
    bitmapbutton .$myself.b5 @$TAUDIR/xbm/racy.xbm racy "launch racy .racy"
    bitmapbutton .$myself.b6 @$TAUDIR/xbm/speedy.xbm speedy \
	"launch speedy .speedy"
    bitmapbutton .$myself.b7 @$TAUDIR/xbm/spiffy.xbm spiffy \
	"launch spiffy .spiffy"
    pack .$myself.b0 .$myself.b1 .$myself.b2 .$myself.b3 .$myself.b4\
	.$myself.b5 .$myself.b6 .$myself.b7\
	-side left -padx 10 -pady 5 -in .$myself.coms
    pack .$myself.coms -side bottom -pady 10

    wm protocol .$myself WM_DELETE_WINDOW atExit
    bind all Q atExit
}

#
# toggleButtons: map/unmap command buttons depending on depfile type
#

proc toggleButtons {} {
    global myself depfile


    #  Stuff to display proper tool set for mighty-based languages
    #  ----------------------------------------------------------------------
    #  if { $depfile(type) == "c" } {
    #    pack forget .$myself.b0 .$myself.b4 .$myself.b5 .$myself.b6
    #    pack .$myself.b1 .$myself.b2 .$myself.b3 .$myself.b7 \
	#                     -side left -padx 10 -pady 5 -in .$myself.coms
    #  } else {
    pack forget .$myself.b1 
    pack .$myself.b0 .$myself.b2 .$myself.b3 .$myself.b4 \
	.$myself.b5 .$myself.b6 .$myself.b7 \
	-side left -padx 10 -pady 5 -in .$myself.coms
    #  }
}

#
# atExit: cleanup procedure; makes sure all launched tools are exited
#

proc atExit {} {
    global myself

    async_send racy exit
    async_send cagey exit
    async_send classy exit
    async_send fancy exit
    async_send cosy exit
    async_send speedy exit
    async_send spiffy exit
    async_send mighty exit
    PM_RemTool $myself
    exit
}

#
# getLocalHost: get hostname of computer
#
proc getLocalHost {} {
    global depfile

    if { $depfile(host) == "localhost" } {
	if [catch {exec uname -n} h] {
	    if { ![catch {exec hostname} h] } {
		set depfile(hostarch) "$h ($depfile(arch))"
	    } else {
		set depfile(hostarch) "localhost"
	    }
	} else {
	    set depfile(hostarch) "$h ($depfile(arch))"
	}
    } else {
	set depfile(hostarch) "$depfile(host) ($depfile(arch))"
    }
}




proc PM_AcceptChanges {progfile} {

    # Scaffolding, retained for debugging.
    puts "Project Manager say: $progfile has changed."
    
}


proc tau_ProjectOptions {} {
    #This procedure takes no arguments. It access the current set of project options, and
    #allows the user to make changes in the variables that tau is using to manipulate
    #the project. Currently the SAGEROOT variable is accessible. . .

    global depfile srVar
    proc tau_ProjOptsInternal {} {
	
	global depfile 
	
    }
    proc maintainPMF {throw these away} {
	global depfile

	if {![string match ".pmf" [file ext $depfile(projnameTemp)]]} {
	    set depfile(projnameTemp) "$depfile(projnameTemp).pmf"
	    .tauProjOpts.fr3.newNameEnt icursor \
		[expr [string length $depfile(projnameTemp)] - 4]
	}
    }

    if [winfo exists .tauProjOpts] {
	raise .tauProjOpts
    } else {
	toplevel .tauProjOpts
	wm title .tauProjOpts "Project Options"
	set rootTemp [frame .tauProjOpts.fr1]
	pack $rootTemp \
	    -expand 1 \
	    -fill both \
	    -padx 10 \
	    -pady 10
	label $rootTemp.sageRootlabel -text "Sageroot:"
	entry $rootTemp.sageRootentry -width 50 -textvariable srVar
	pack $rootTemp.sageRootlabel $rootTemp.sageRootentry \
	    -side left \
	    -anchor nw \
	    -fill x
	$rootTemp.sageRootentry delete 0 end
	$rootTemp.sageRootentry insert end $depfile(root)
	bind $rootTemp.sageRootentry <Return> {.tauProjOpts.fr2.enter flash;\
						   .tauProjOpts.fr2.enter invoke}
	button $rootTemp.sageRootSearch \
	    -text Search \
	    -command {set srVar \
			  [getFile "Set SAGEROOT Directory" \
			       "" 0 $depfile(host) ""]}
	pack $rootTemp.sageRootSearch -side left -anchor nw
	set nameTemp [frame .tauProjOpts.fr3]
	pack $nameTemp \
	    -expand 1 \
	    -fill both \
	    -padx 10 \
	    -pady 10
	label $nameTemp.newNameLabel \
	    -text "Project Name:"
	entry $nameTemp.newNameEnt -width 30 -textvariable depfile(projnameTemp)
	pack $nameTemp.newNameLabel $nameTemp.newNameEnt \
	    -side left \
	    -anchor nw \
	    -fill x
	$nameTemp.newNameEnt delete 0 end
	set depfile(projnameTemp) $depfile(project)
	trace variable depfile(projnameTemp) w maintainPMF
	set depfile(projnameTemp) $depfile(project)
	set this [frame .tauProjOpts.fr2]
	pack $this \
	    -expand 1 \
	    -fill both \
	    -padx 10 \
	    -pady 10
	button $this.enter -text Enter \
	    -command {set depfile(root) $srVar;\
			  puts "new name == $depfile(projnameTemp)"; \
			  set depfile(project) $depfile(projnameTemp); \
			  Bdb_ChangeName $depfile(projnameTemp); \
			  puts "Returned from Bdb_ChangeName" ; \
			  PM_ChangeProjectName $depfile(projnameTemp); \
			  PM_BroadcastChanges - p; \
			  destroy .tauProjOpts}
	button $this.cancel -text Cancel -command "destroy .tauProjOpts"
	pack $this.enter \
	    -side left 

	pack $this.cancel \
	    -side right 

    }
}



proc tau_OpenProject {} {
    #For use by tau after the initial project is opened, to
    # switch projects and associated information.
    global depfile pm_status

    PM_OpenProject
    set pm_status [PM_Status]
    set depfile(project) [lindex $pm_status 0]
    set depfile(host)    [lindex $pm_status 1]
    set depfile(arch)    [lindex $pm_status 2]
    set depfile(root)    [lindex $pm_status 3]
    set depfile(dir)     [lindex $pm_status 4]
    if {[string match $depfile(dir) "."]} {
	set depfile(dir) [pwd]
    }
    SetFileDisplay
}

proc tau_AddFile {} {
    # Interface for adding a file to the project, and updating the display.

    PM_AddFile
    SetFileDisplay
}

proc tau_DeleteFile {{filename ""}} {

    global myself depfile

    set lb .$myself.info2.ext.int.bot.list

    if [string match $filename ""] {
	if [string match [$lb curselection] ""] {
	    showError "No File Selected"
	    return
	} else {
	    set temp [$lb get [$lb curselection]]
	    scan $temp "%s" filename
	}
    }
#    puts "tau_DF: Expanded filename == $depfile(dir)/$filename"
#    PM_RemoveFile $depfile(dir)/$filename
    PM_RemoveFile $filename
    SetFileDisplay
}

proc SetFileDisplay {} {
    # This procedure gets the current list of files and associated language
    # information, and displays it in the GUI.
    # Returns nothing.
    
    global myself

    set filelist [PM_GetFiles]
    .$myself.info2.ext.int.bot.list delete 0 end
    
    foreach elem $filelist {
	set langtemp [PM_SetLangOption $elem]
	set holder $elem
	for {set i 1} {$i < [expr 68 - [string length $elem]]} {incr i} {
	    set holder "$holder "
	}
	set holder "$holder$langtemp"
	.$myself.info2.ext.int.bot.list insert end $holder
    }
    
}

proc Tool_AcceptChanges {progfile flag} {
    if [DEBUG] { puts "In tau: Project Manager says: $progfile has changed ($flag)."}

    switch $flag {

	d {
	    # Delete a file
	    # Remove from GUI
	    SetFileDisplay
	}


	a {
	    # Add a file
	    # Add to GUI
	    SetFileDisplay
	}


	u {
	    # Update a file
	    # Reset GUI
	}
	
	p {
	    #Modify project information. Since tau is the interpretter effecting
	    #the changes, do nothing here.
	}

	e {
	    #This flag is for udating during execution of the project binary. No action needed.
	}
    }
    if [DEBUG] {puts "in tau: returning from Tool_AcceptChanges"}
}

# ------------
# -- main code
# ------------

if {$argc == 0} {
    set parg [pwd]
} elseif {$argc == 1} {
    set parg [lindex $argv 0]
} else {
    puts stderr "usage: $myself \[\[host:\]projFile \| \[host:\]directory\]"
    exit
}

# Init the project manager (taud)
if [DEBUG] {puts "TAU - launching daemon"}
launchTauDaemon -waitfor
if [DEBUG] {puts "TAU - daemon running - initing PM for tool."}
PM_AddTool $myself
PM_AddGlobalSelect $myself { \
				 global_showFuncTag }

# Initialize the project
# Coordinate w/ PM
if [DEBUG] {puts "TAU - opening a project"}
set pm_status [PM_Status]
if [DEBUG] {puts "PM_Status - $pm_status"}
if {[lindex $pm_status 0] == "NO_PROJECT"} {
    # Open or create a project
    set colon [string first ":" $parg]
    if {$colon > 0} {
	set hostarg [string range $parg 0 [expr $colon - 1]]
	set patharg [string range $parg [expr $colon + 1] end]
    } else {
	set hostarg localhost
	set patharg $parg
    }

    if [DEBUG] {puts "hostarg - $hostarg"}
    if [DEBUG] { puts "patharg - $patharg"}
    set projfile [PM_OpenProject $patharg $hostarg]
    if {$projfile == "NO_PROJECT"} {
	showError "No project opened!  This should work but it doesn't!"
	atExit
    }
    set pm_status [PM_Status]
}

set depfile(project) [lindex $pm_status 0]
set depfile(host)    [lindex $pm_status 1]
set depfile(arch)    [lindex $pm_status 2]
set depfile(root)    [lindex $pm_status 3]
set depfile(dir)     [lindex $pm_status 4]
if {[string match $depfile(dir) "."]} {
    set depfile(dir) [pwd]
}

# Tool Init
if [DEBUG] {puts "TAU - initializing tool GUI"}
getLocalHost
createWindow
set dispHost [PM_GetHost]
set dispDir [PM_GetDir]
SetFileDisplay
toggleButtons
removeMessage
if [DEBUG] {puts "TAU - done"}
