/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/



class pcxx_vector{
  REAL data[SIZEVECT+1];
 public:
  int size;
  pcxx_vector();
  pcxx_vector(int s);
  pcxx_vector(pcxx_vector *);
  pcxx_vector(double d);
  pcxx_vector operator +(const pcxx_vector &rhs);
  pcxx_vector operator *(const pcxx_vector &rhs);
  pcxx_vector operator -(const pcxx_vector &rhs);
  pcxx_vector operator -();
  pcxx_vector operator /(const pcxx_vector &rhs);
  pcxx_vector &operator =(const pcxx_vector &rhs);
  pcxx_vector &operator +=(const pcxx_vector &rhs);
  void allocData(int s);
  double Dot(pcxx_vector &arg1);
  void setValue(int i, double x);
  double value(int i);
  void load(double x); // sets whole vect to value.
  void print();
};

pcxx_vector::pcxx_vector(){
  size = SIZEVECT;
};

pcxx_vector::pcxx_vector(double x){
  int i;
  size = SIZEVECT;
  // printf("(pcxx_vector)%d\n", size);
  for (i = 0; i< SIZEVECT; i++)
    data[i] = x;
}

pcxx_vector::pcxx_vector(int){
   size = SIZEVECT;
}

pcxx_vector::pcxx_vector(pcxx_vector *v){
  int i;
  size = v->size;
  for (i = 0; i< SIZEVECT; i++)
    data[i] = v->data[i];
}

void pcxx_vector ::allocData(int s){
  int i;
  size = s;
  for (i = 0; i< SIZEVECT; i++)
    data[i] = 0;
}

REAL pcxx_vector::Dot(pcxx_vector &arg1)
{
  int i;
  REAL x;
  REAL *p1, *p2;
  x = 0.0;
  p1 = &(this->data[0]);
  p2 = &(arg1.data[0]);
// modified to avoid averflow
  for(i = 0; i < size; i++) x += (*p1++)*(*p2++); 
  return x;
}


void pcxx_vector ::print(){
  int i;
  
  for (i = 0; i< SIZEVECT; i++)
    printf("%f ",data[i]);

}

void pcxx_vector ::load(double x){
  int i;
  
  for (i = 0; i< SIZEVECT; i++)
    data[i] = x;

}

void pcxx_vector::setValue(int i, double x)
{
  data[i] = x;
}
double pcxx_vector::value(int i)
{
  return data[i];
}

pcxx_vector pcxx_vector::operator +(const pcxx_vector &rhs){
  pcxx_vector T;
  int i;
  T.size = this->size;
  // printf("(+)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = this->data[i] + rhs.data[i];
  return T;
}

pcxx_vector pcxx_vector::operator -(const pcxx_vector &rhs){
  pcxx_vector T;
  int i;
  T.size = this->size;
  // printf("(-)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = this->data[i] - rhs.data[i];
  return T;
}

pcxx_vector pcxx_vector::operator -(){
  pcxx_vector T;
  int i;
  T.size = this->size;
  // printf("(-u)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = -this->data[i];
  return T;
}

pcxx_vector pcxx_vector::operator /(const pcxx_vector &rhs){
  pcxx_vector T;
  int i;
  T.size = this->size;
  // printf("(/)%d\n", size);
  for(i = 0; i < size; i++)
    if(rhs.data[i] != 0.0)
       T.data[i] = this->data[i]/rhs.data[i];
     else printf("divide error %d\n");
  return T;
}


pcxx_vector pcxx_vector::operator *(const pcxx_vector &rhs){
  pcxx_vector T;
  int i;
  T.size = this->size;
  // printf("(*)%d\n", size);
  for(i = 0; i < size; i++)
    T.data[i] = this->data[i] * rhs.data[i];
  return T;
}

pcxx_vector &pcxx_vector::operator =(const pcxx_vector &rhs){
	int i;
        // printf("(=)%d\n", size);
	for(i = 0; i < rhs.size; i++)
		this->data[i] = rhs.data[i];
	return *this;
}

pcxx_vector &pcxx_vector::operator +=(const pcxx_vector &rhs)
{
  int i;
  // printf("(+=)%d\n", size);
  for(i = 0; i < rhs.size; i++)
    {
      this->data[i] += rhs.data[i];
    }
  return *this;
}

