/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


#include <stdio.h>
#define MAXNU 3
#define MAXVECTOR 15

extern "C" double fabs(double);


/* dtrsv library function      **********************************************/
/*      solves the system of equations:                                     */
/*      Ax = b              (triangular solver)                             */
/*                                                                          */
/* void dtrsv(uplo,trans,diag,n,A,ldA,x,incx)                               */
/*      char uplo        :- 'u' or 'l' for upper, lower triangular matrix   */
/*      char trans       :- variations of solver (use 'n')                  */
/*      char diag        :- 'u' for unit triangular, 'n' for not            */
/*      int n            :- order of matrix A                               */
/*      double A[n][n]   :- matrix                                          */
/*      int ldA          :- length  dimension of A (MAX_L of rows)          */
/*      double x[l]      :- vector x gets overwritten with solution         */
/*      int incx         :- stride to be used for vectors (usually 1)       */
/****************************************************************************/

extern "C" void dtrsv(char uplo, char trans, char diag,
                      int n,double A[MAXNU][MAXNU],
		      int ldA, double x[MAXNU] , int incx);

/****************************************************************************/
/* void dgemm(transa,transb,m,n,k,a,A,ldA,B,ldB,b,C,ldC)                    */
/*      char transa,transb :- variation to be used (both 'N')               */
/*      int m,n            :- rows and columns of matrix C                  */
/*      int k              :- # of columns in A and # of rows in B          */
/*      double a,b         :- alpha, beta scalars                           */
/*      double A[m][n],B[m][n],C[m][n] :- matrixes                          */
/*      int ldA,ldB,ldC  :- length dimension of A,B and C (length of rows)  */
/****************************************************************************/

extern "C" void dgemm(char ta, char tb , int m, int n,
                      int k, double a,
                      double A[][MAXNU], int ldA,
                      double B[MAXNU][MAXNU], int ldB, double b,
                      double C[][MAXNU], int ldC);

/* solves T*X = B for unit lower triangular T n by n */
/*   and n by n X overwriting n by n B */
void triag_solve(double T[][MAXNU], double B[][MAXNU], int n){
    int i, j, k;
    for( i = 1; i < n; i++)
       for(j = 0; j < i; j++)
	   for(k = 0; k < n; k++) 
            B[i][k] -= T[i][j]*B[j][k];
}

/* compute C -= A*B with C n by nu, A n by nu, B nu by nu */

void rank_nu(double A[][MAXNU], double B[][MAXNU], double C[][MAXNU], 
             int n, int nu){
     
     double a, b;
     a = -1.0; b = 1.0;
     dgemm('N', 'N',  n, nu, nu, a, A, MAXNU, B, MAXNU, b, C, MAXNU);
      
     /* same as 
     int i, j, k;
      for(i = 0; i < n; i++)
           for(j = 0; j < nu; j++)
               for(k = 0; k < nu; k++)
		   C[i][j] -= A[i][k]*B[k][j];
     */
     
}

typedef double Block[MAXNU][MAXNU];

class ColumnBlock{
     int pos; /* position */
     int nu;  /* block size */
     int n;   /* column size */
     double a[MAXVECTOR][MAXNU];
  public:
     int ipv[MAXVECTOR]; /* pivot ColumnBlock */
     ColumnBlock(int N, int NU, int K);
     void initBlock(int N, int NU, int k);
     double &operator ()(int i, int j){ return a[i][j]; }
     int permute(int i){ return ipv[i]; }
     double *row(int i){ return &(a[i][0]); }
     Block * block(int i){ return (Block *) &(a[i][0]); }
     void setWidth(int k){ nu = k; };
     void setPosition(int p){ pos = p; };
     int position(){ return pos; };
     void swap(int i, int j);
     void factor();
     void update( ColumnBlock * fromB);
     void print();

};
     
ColumnBlock::ColumnBlock(int N, int NU, int K){
      int i,j;
      n = N; nu = NU; pos = K;
      for(i = 0; i < N;  i++) ipv[i] = i;
      for(i = 0; i < N; i++)
	for(j = 0; j < nu; j++) 
		a[i][j] = 0.0;
};
void ColumnBlock::initBlock(int N, int NU, int K){
      int i,j;
      n = N; nu = NU; pos = K;
      for(i = 0; i < N;  i++) ipv[i] = i;
      for(i = 0; i < N; i++)
	for(j = 0; j < nu; j++) 
		a[i][j] = 0.0;
};

void ColumnBlock::print(){
      int i, j;
      printf("-------------- block %d----------------\n", pos);
      for(i = 0; i < n; i++){
	printf("%3d |", ipv[i]);
	for(j = 0; j < nu; j++) printf(" %6.3f ", a[i][j]);
	printf("\n");
	}
}
      
void ColumnBlock::swap(int i, int j){
   int k;
   double *ip, *jp;
   double t;
   k = ipv[i];  ipv[i] = ipv[j]; ipv[j] = k;
   for(k = 0; k < nu; k++){
	  t = a[i][k]; a[i][k] = a[j][k]; a[j][k] = t;
	  }
 }


void ColumnBlock::factor(){
    int i, j, k;
    double max, t;
    int maxpo;
    double *ip, *jp;
    for(i = 0; i < nu; i++){
	if (i + pos >= n) break;
        max = fabs( a[i+pos][i] );
	maxpo = i+pos;
        for (j = i+pos+1; j < n; j++){
		t = fabs( a[j][i]);
		if(t > max){ max = t; maxpo = j; }
		}
        if(maxpo != i+pos){
		 swap(i+pos,maxpo); 
		}
        for(j = i+pos+1; j < n; j++){
                t = a[j][i]/a[i+pos][i];
                a[j][i] = t;
                for (k = i+1; k < nu; k++)
                     a[j][k] -= t*a[i+pos][k];
		}
	}
}

void ColumnBlock::update(ColumnBlock * fromB){

    int piv_row, i, j, k;
    int *want, *have;
    int whereis[MAXVECTOR];
    int t;
    /* first apply the permutation. do this by comparing */
    /* the old permutation againt the new one */
    if((piv_row = fromB->position()) == pos) return;
    /* compute the inverse of the permutation */
    want = &(fromB->ipv[0]);
    have = &(ipv[0]);
    for(i = 0; i < n; i++) whereis[have[i]] = i;
    for( i = 0; i < n; i++ )
	if( want[i] != have[i]){
		j = whereis[want[i]];
		swap(i,j);
    		whereis[have[j]] = j;
    		whereis[have[i]] = i;
		}

    if ( pos > piv_row){
    	triag_solve(*(fromB->block(piv_row)), *(block(piv_row)), nu);
	rank_nu(*(fromB->block(piv_row+nu)), *block(piv_row),
		 *block(piv_row+nu), n-(piv_row+nu), nu);
	}
}

void check(ColumnBlock *A[], int n){
	double M[MAXVECTOR][MAXVECTOR], L[MAXVECTOR][MAXVECTOR], U[MAXVECTOR][MAXVECTOR];
	int *p;
        int i,j,k;
	for(i = 0; i < n; i++){
		for(j = 0; j < n; j++){
			M[i][j] = L[i][j] = U[i][j] = 0.0;
 			k = j/3;
			/* M[i][j] = (*A[k])(i, j-3*k); */
		        if(i > j)
			   L[i][j] = (*A[k])(i, j-3*k);
			else
			   U[i][j] = (*A[k])(i, j-3*k);
			}
		L[i][i] = 1.0;
		}
	p = A[0]->ipv;
        	
	for(i = 0; i < n; i++)
		for(j = 0; j < n; j++)
			for(k = 0; k < n; k++)
				M[p[i]][j] += L[i][k]*U[k][j];
	
        printf("L is \n");
	for(i = 0; i < n; i++){
		printf(" %2d |", p[i]);
		for(j = 0; j < n; j++) printf(" %5.2f ", L[i][j]);
		printf("\n");
		}
        printf("U is \n");
	for(i = 0; i < n; i++){
		printf(" %2d |", p[i]);
		for(j = 0; j < n; j++) printf(" %5.2f ", U[i][j]);
		printf("\n");
		}
        printf("M is \n");
	for(i = 0; i < n; i++){
		printf(" %2d |", p[i]);
		for(j = 0; j < n; j++) printf(" %5.2f ", M[i][j]);
		printf("\n");
		}
		
}
			

void main(){
	ColumnBlock *A[5];
	int numblocks = 5;
        int bs = 3;  /* the size of each block */
        int n;       /* the size of the system */
        int i,j;
        n = bs*numblocks;
        for(i = 0; i < numblocks; i++) {
		A[i] = new ColumnBlock(n,bs,bs*i);
		for(j = 0; j < 3; j++){
			(*A[i])(0, j) = 1.0*(bs*i+1);
			(*A[i])(bs*i+j, j) = 2.0;
			}
		}
	 for(i = 1; i < n; i++) (*A[0])(i,0) = 0.1*i;

         /* now factor */
	 for(i = 0; i < numblocks; i++){
		A[i]->factor();
		for(j = 0; j < 5; j++) A[j]->update(A[i]); 
		}
         
	 check(A, n);
}


