/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/



extern "C" double fabs(double);
extern void triag_solve(double T[][MAXNU], double B[][MAXNU], int n);
extern void triag_solve2(double T[][MAXNU], double B[][MAXNU], int n);
extern void rank_nu(double A[][MAXNU], double B[][MAXNU], double C[][MAXNU], 
             int n, int nu);
extern void rank_nu2(double A[][MAXNU], double B[][MAXNU], double C[][MAXNU], 
             int n, int nu);


typedef double Block[MAXNU][MAXNU];

class ColumnBlock{
     int pos; /* position */
     int nu;  /* block size */
     int n;   /* column size */
     double a[MAXVECTOR][MAXNU];
  public:
     int ipv[MAXVECTOR]; /* pivot ColumnBlock */
     ColumnBlock() {};
     ColumnBlock(int N, int NU, int K);
     void initBlock(int N, int NU, int k);
     double &operator ()(int i, int j){ return a[i][j]; }
     int permute(int i){ return ipv[i]; }
     double *row(int i){ return &(a[i][0]); }
     Block * block(int i){ return (Block *) &(a[i][0]); }
     void setWidth(int k){ nu = k; };
     void setPosition(int p){ pos = p; };
     int position(){ return pos; };
     void swap(int i, int j);
     void factor();
     void update( ColumnBlock * fromB);
     void print();
     /* the rest for dist array */
     ColumnBlock(int N){};
     void setValue(int i, double v){ a[i][0] = v; };
     double  value(int i);
  ColumnBlock operator +(ColumnBlock &rhs){return *this;};
  ColumnBlock operator *(ColumnBlock &rhs){return *this;};
  ColumnBlock operator -(ColumnBlock &rhs){return *this;};
  ColumnBlock operator -(){return *this;};
  ColumnBlock operator /(ColumnBlock &rhs){return *this;};
  ColumnBlock &operator =(ColumnBlock &rhs){return *this;};
  ColumnBlock &operator +=(ColumnBlock &rhs){return *this;};
  void allocData(int s){};
  double dot(ColumnBlock &arg1){return 0.0;};

};
     
ColumnBlock::ColumnBlock(int N, int NU, int K){
      int i,j;
      n = N; nu = NU; pos = K;
      for(i = 0; i < N;  i++) ipv[i] = i;
      for(i = 0; i < N; i++)
	for(j = 0; j < nu; j++) 
		a[i][j] = 0.0;
};
void ColumnBlock::initBlock(int N, int NU, int K){
      int i,j;
      n = N; nu = NU; pos = K;
      for(i = 0; i < N;  i++) ipv[i] = i;
      for(i = 0; i < N; i++)
	for(j = 0; j < nu; j++) 
		a[i][j] = 0.0;
};

void ColumnBlock::print(){
      int i, j;
      printf("-------------- block %d----------------\n", pos);
      for(i = 0; i < n; i++){
	printf("%3d |", ipv[i]);
	for(j = 0; j < nu; j++) printf(" %6.3f ", a[i][j]);
	printf("\n");
	}
}
      
void ColumnBlock::swap(int i, int j){
   int k;
   double *ip, *jp;
   double t;
   k = ipv[i];  ipv[i] = ipv[j]; ipv[j] = k;
   for(k = 0; k < nu; k++){
	  t = a[i][k]; a[i][k] = a[j][k]; a[j][k] = t;
	  }
 }


void ColumnBlock::factor(){
    int i, j, k;
    double max, t;
    int maxpo;
    double *ip, *jp;
    for(i = 0; i < nu; i++){
	if (i + pos >= n) break;
        max = fabs( a[i+pos][i] );
	maxpo = i+pos;
        for (j = i+pos+1; j < n; j++){
		t = fabs( a[j][i]);
		if(t > max){ max = t; maxpo = j; }
		}
        if(maxpo != i+pos){
		 swap(i+pos,maxpo); 
		}
        for(j = i+pos+1; j < n; j++){
                t = a[j][i]/a[i+pos][i];
                a[j][i] = t;
                for (k = i+1; k < nu; k++)
                     a[j][k] -= t*a[i+pos][k];
		}
	}
}

void ColumnBlock::update(ColumnBlock * fromB){

    int piv_row, i, j, k;
    int *want, *have;
    int whereis[MAXVECTOR];
    int t;
    /* first apply the permutation. do this by comparing */
    /* the old permutation againt the new one */
    if((piv_row = fromB->position()) == pos) return;
    /* compute the inverse of the permutation */
    want = &(fromB->ipv[0]);
    have = &(ipv[0]);
    for(i = 0; i < n; i++) whereis[have[i]] = i;
    for( i = 0; i < n; i++ )
	if( want[i] != have[i]){
		j = whereis[want[i]];
		swap(i,j);
    		whereis[have[j]] = j;
    		whereis[have[i]] = i;
		}

    if ( pos > piv_row){
    	triag_solve2(*(fromB->block(piv_row)), *(block(piv_row)), nu);
	rank_nu2(*(fromB->block(piv_row+nu)), *block(piv_row),
		 *block(piv_row+nu), n-(piv_row+nu), nu);
	}
}
