/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/



// *********************************************************************
//  Sparse Matrix Class Element for Distributed Sparse Matrix.
// *********************************************************************
class SparseMatrix 
{
       int rows, columns, nnz, size;
       REAL *value;
       int *row_ptr, *col_indx;
    public:
       SparseMatrix(int row_size, int col_size, int mtrx_size); 
       void initialize(int row_size, int col_size, int mtrx_size);//temporarily
       // needed because pC++ cannot call a constructor automatically.
       REAL operator ()(int row, int col); // returns value of element
       void insert_element(REAL val, int row, int col); 
       int number_of_rows() { return rows; };
       int number_of_columns() { return columns; };
       int number_of_nonzeros() { return nnz; };
       int allocated_size(){ return size; }; 
       SparseMatrix amub(SparseMatrix& a, SparseMatrix& b); // A * B
       SparseMatrix amub1(SparseMatrix& a, SparseMatrix& b, int &count);
       void amux(Vector& in_vector, Vector& out_vector); // A * x 
       void atmux(Vector& in_vector, Vector& out_vector); // tranpose(A) * x
//       SparseMatrix operator *(SparseMatrix& B); // A * B
       SparseMatrix &operator =(SparseMatrix &rhs);
       SparseMatrix &operator +=(SparseMatrix &rhs);
       void allocData(int s, int initval);
       void allocData(int s);
};

SparseMatrix & SparseMatrix::operator =(SparseMatrix &rhs){ return rhs; };
SparseMatrix & SparseMatrix::operator +=(SparseMatrix &rhs){ return rhs; };
void SparseMatrix::allocData(int s, int initval) { return;};
void SparseMatrix::allocData(int s) { return; };

void SparseMatrix::initialize(int row_size, int col_size, int mtrx_size) 
{ 
  int i;
       rows = row_size;
       columns = col_size;
       size = mtrx_size;
       nnz = 0;

       row_ptr = new int [rows+1];
       value = new REAL [size];
       col_indx = new int [size];

       for(i = 0; i <= rows; i++) row_ptr[i] = 0;
       for(i = 0; i < size; i++) col_indx[i] = -1;
};

SparseMatrix::SparseMatrix(int row_size, int col_size, int mtrx_size) 
{ 
  int i;
       rows = row_size;
       columns = col_size;
       size = mtrx_size;
       nnz = 0;

       row_ptr = new int [rows+1];
       value = new REAL [size];
       col_indx = new int [size];

       for(i = 0; i <= rows; i++) row_ptr[i] = 0;
       for(i = 0; i < size; i++) col_indx[i] = -1;
};

//SparseMatrix SparseMatrix::operator *(SparseMatrix& b)
//{
//       int mtrx_size = rows*b.columns;
// pb with pC++       SparseMatrix c(rows, b.columns, mtrx_size); 
//       c.amub(*this, b);
//       return c;
//}

REAL SparseMatrix::operator ()(int row, int col)
{
  int num;
     for(num = row_ptr[row]; num < row_ptr[row+1]; num++) {
       if(col_indx[num] == col) 
	 return value[num];
     }
     return (REAL) 0;
}


void SparseMatrix::amux(Vector& in_vector, Vector& out_vector)
{
     /*
     -----------------------------------------------------------------------
     A times a vector
     
     y = A * x
     
     Notes:
     -------
     Based on SPARSKIT amux.
     
     Shelby Yang, CICA, Indiana University, 10/31/92.
     
     ----------------------------------------------------------------------- 
     multiplies a matrix by a vector using the dot product form
     Matrix A is stored in compressed sparse row storage.
     
     on entry:
     ----------
     in_vector = real array of length equal to the column dimension of
                 the A matrix.
		 
     on return:
     -----------
     out_vector = real array of length equal to the row dimension of the A 
                  matrix, containing the product y = A x

     -----------------------------------------------------------------------
     */

     REAL sum;
     int num, row;

     /*  --- main loop  */
     for(row = 0; row < rows; row++) {
       sum = (REAL) 0;
       for(num = row_ptr[row]; num < row_ptr[row+1]; num++) {
	 sum += value[num]*in_vector[col_indx[num]];
       }
       out_vector[row] = sum;
     }
     return;
}


void SparseMatrix::atmux(Vector& in_vector, Vector& out_vector)
{
    /*
    -----------------------------------------------------------------------
    transp( A ) times a vector
    ----------------------------------------------------------------------- 
    multiplies the transpose of a matrix by a vector when the original
    matrix is stored in compressed sparse row storage. Can also be
    viewed as the product of a matrix by a vector when the original
    matrix is stored in the compressed sparse column format.
    -----------------------------------------------------------------------
    Notes:
    -------
    Based on SPARSKIT atmux.

    Shelby Yang, CICA, Indiana University, 11/2/92.
    
    ----------------------------------------------------------------------- 
    
    on entry:
    ----------
    rows       = row dimension of A
    in_vector  = real array of length equal to the row dimension of
                 the A matrix.
    this       = input matrix in compressed sparse row format.
    
    on return:
    -----------
    out_vector = real array of length equal to the column dimension of A, 
                 containing the product transp( A ) * in_vector

    -----------------------------------------------------------------------
    */

      int row, num;

      /* initialize out_vector */
      for(row = 0; row < rows; row++) {
	out_vector[row] = (REAL) 0;
      }
      /* --- main loop */
      for(row = 0; row < rows; row++) {
	for(num = row_ptr[row]; num < row_ptr[row+1]; num++) {
            out_vector[col_indx[num]] += in_vector[row]*value[num];
	  }
      }
      return;
}

 
void SparseMatrix::insert_element(REAL val, int row, int col)
{
     int num, i;
     REAL *tmp_val;
     int *tmp_col;

     if(row >= rows || row < 0) {
       printf("SparseMatrix::insert_element() : ");
       printf("element row index is out of range!\n");
       return;
     }
     if(col >= columns || col < 0) {
       printf("SparseMatrix::insert_element() : ");
       printf("element column index is out of range!\n");
       return;
     }
     if(nnz > size) {
       printf("SparseMatrix::insert_element() : ");
       printf("number of elements > size allocated!\n");
       return;
     }

     /* remember nnz = row_ptr[rows]                                */
     /*	             = total number of nonzero elements             */
     /*              = 1 + the number assigned to the last element  */


     
     num = row_ptr[row];
     if(num != row_ptr[row+1]) {  /* if statment takes care of a special   */
				  /* situation when there is no element    */
				  /* in a row.                             */
        while(num < row_ptr[row+1] && col_indx[num] < col)
	 num++;        /* search for element */
	if(col_indx[num] == col) { 
	  value[num] = val; /* if found, replace old value with new value */
	  return;
	}
      }

     /* if not found, need to insert an element.          */
     /* first decide whether need to expand array or not. */

     if(nnz < size) {   /* no need to expand */

       for(i = nnz; i > num; i--) { /*copy element backward */
	 value[i] = value[i-1];
	 col_indx[i] = col_indx[i-1];
       }
       value[num] = val;      /* insert new element */
       col_indx[num] = col;
       for(i = row+1; i <= rows ; i++) row_ptr[i] = row_ptr[i]+1;
       nnz = nnz + 1;

       return;

     } else {  /* expand array */

       tmp_val = new REAL [size+1];
       tmp_col = new int [size+1];  /* increase size by 1 */
       size = size + 1;   

       for(i = 0; i < num; i++) {
	 tmp_val[i] = value[i]; /* copy part in front of element to new array*/
	 tmp_col[i] = col_indx[i];
       }

       tmp_val[num] = val;        /* insert element */
       tmp_col[num] = col;

       for(i = num+1; i <= nnz; i++) {
	 tmp_val[i] = value[i-1];  /* copy part behind element to new array */
	 tmp_col[i] = col_indx[i-1];
       }

       for(i = row+1; i <= rows ; i++) row_ptr[i] = row_ptr[i]+1;
       nnz = nnz + 1;

       delete [] value;    /* delete old copy */
       value = tmp_val;
       
       delete [] col_indx;
       col_indx = tmp_col;
       
       return;
     }

}

SparseMatrix SparseMatrix::amub(SparseMatrix& a, SparseMatrix& b)
{      
    /*
      -----------------------------------------------------------------------
      performs the matrix by matrix product C = A B 

      Notes:
      -------
      Based on SPARSKIT amub.

      Shelby Yang, CICA, Indiana University, 10/27/92.
      -----------------------------------------------------------------------
      on entry:
      ---------
      a = Matrix A in compressed sparse row format.
 
      b = Matrix B in compressed sparse row format.

      on return:
      ----------
      this = resulting matrix C in compressed sparse row sparse format.
      
      work array:
      ------------
      iw = integer work array of length equal to the number of
           columns in B.

     -----------------------------------------------------------------
     */

      REAL scal;
      int a_row, a_col, b_col, a_num, b_num, c_num, a_row_size, b_col_size;
      int *iw, len,j;
      
      len = 0;
      row_ptr[0] = 0; 
      a_row_size = a.rows;
      b_col_size = b.columns;

      iw = new int [b_col_size];

      /*     initialize array iw. */
      for(j = 0; j < b_col_size; j++) iw[j] = -1;
      
      for(a_row = 0; a_row < a_row_size; a_row++){
	                                        /* starting with a row of A */
	for(a_num = a.row_ptr[a_row]; a_num < a.row_ptr[a_row+1]; a_num++) { 
	                            /* for each nonzero element in this row */
	  scal = a.value[a_num];    /* first record its value, then         */
	  a_col = a.col_indx[a_num]; /* record its column number, a_col.    */
	  for(b_num = b.row_ptr[a_col]; b_num < b.row_ptr[a_col+1]; b_num++) {
	                         /* scan over the a_col row of B,           */
	                         /* since all elements in the a_col         */
	                         /* row will be multiplied with this nonzero*/
	                         /* element of A.                           */
	    b_col = b.col_indx[b_num]; /* get the column number of the B    */
				       /* element.                          */
	    c_num = iw[b_col];   /* the number for a C element in b_col     */
	                         /* column is stored earlier in previous    */
	                         /* iteration in iw[b_col].                 */
	    if (c_num < 0 ) {    /* if the number is -1, element has not    */
	                         /* been created.  counter increases by 1.  */
	      value[len]  = scal*b.value[b_num];   /* create element.       */
	      col_indx[len] = b_col; /* column index of this new element is */
	      iw[b_col]= len;        /* equal to B element's column index.  */
	      len = len +1;          /* record the number for new element.  */
	    }else{                   
	      value[c_num] += scal*b.value[b_num];      /* if not -1,       */
	                                /* element has already been created,*/
		                        /* add to element.                  */
	    }
	  }
	}
	for(c_num = row_ptr[a_row]; c_num < len; c_num++) {
	  iw[col_indx[c_num]] = -1 ;  /* reset elements in iw that have been */
				      /* used. */
	}
	row_ptr[a_row+1] = len;
      }
      
      delete [] iw;

      rows = a_row_size;
      columns = b_col_size;
      nnz = row_ptr[rows];

      return *this;
}


/*
Comment out to use the default memberwise initialization instead of  
overloading the assigment operator, becasue the default assigment is faster.

SparseMatrix& SparseMatrix::operator =(const SparseMatrix& A)
{
  int i;
      rows = A.rows;
      columns = A.columns;
      nnz = A.nnz;
      
      for(i = 0; i < nnz; i++) {
	value[i] = A.value[i];
	col_indx[i] = A.col_indx[i];
      }

      for(i = 0; i <= rows; i++) {
	row_ptr[i] = A.row_ptr[i];
      }
      return *this;
}
*/


SparseMatrix SparseMatrix::amub1(SparseMatrix& a, SparseMatrix& b, int &count)
{      
/*
-----------------------------------------------------------------------
 performs the matrix by matrix product C = A B and count the number of
 floating point operations.

 Notes:
-------
         Based on SPARSKIT amub.

         Shelby Yang, CICA, Indiana University, 10/27/92.
-----------------------------------------------------------------------
 on entry:
 ---------
 a = Matrix A in compressed sparse row format.
 
 b = Matrix B in compressed sparse row format.

 on return:
----------
 this = resulting matrix C in compressed sparse row sparse format.
	    
 work array:
------------
 iw	= integer work array of length equal to the number of
         columns in B.

-----------------------------------------------------------------
*/
      REAL scal;
      int a_row, a_col, b_col, a_num, b_num, c_num, a_row_size, b_col_size;
      int *iw, len,j;

      len = 0;
      row_ptr[0] = 0; 
      a_row_size = a.rows;
      b_col_size = b.columns;

      count = 0;

      iw = new int [b_col_size];

      /*     initialize array iw. */
      for(j = 0; j < b_col_size; j++) iw[j] = -1;
      
      for(a_row = 0; a_row < a_row_size; a_row++){
	                                        /* starting with a row of A */
	for(a_num = a.row_ptr[a_row]; a_num < a.row_ptr[a_row+1]; a_num++) { 
	                            /* for each nonzero element in this row */
	  scal = a.value[a_num];    /* first record its value, then         */
	  a_col = a.col_indx[a_num]; /* record its column number, a_col.    */
	  for(b_num = b.row_ptr[a_col]; b_num < b.row_ptr[a_col+1]; b_num++) {
	                         /* scan over the a_col row of B,           */
	                         /* since all elements in the a_col         */
	                         /* row will be multiplied with this nonzero*/
	                         /* element of A.                           */
	    b_col = b.col_indx[b_num]; /* get the column number of the B    */
				       /* element.                          */
	    c_num = iw[b_col];   /* the number for a C element in b_col     */
	                         /* column is stored earlier in previous    */
	                         /* iteration in iw[b_col].                 */
	    if (c_num < 0 ) {    /* if the number is -1, element has not    */
	                         /* been created.  counter increases by 1.  */
	      value[len]  = scal*b.value[b_num];   /* create element.       */
	      count++;
	      col_indx[len] = b_col; /* column index of this new element is */
	      iw[b_col]= len;        /* equal to B element's column index.  */
	      len = len +1;          /* record the number for new element.  */
	    }else{                   
	      value[c_num] += scal*b.value[b_num];      /* if not -1,       */
	      count +=2;
	                                /* element has already been created,*/
		                        /* add to element.                  */
	    }
	  }
	}
	for(c_num = row_ptr[a_row]; c_num < len; c_num++) {
	  iw[col_indx[c_num]] = -1 ;  /* reset elements in iw that have been */
				      /* used. */
	}
	row_ptr[a_row+1] = len;
      }
      
      delete [] iw;

      rows = a_row_size;
      columns = b_col_size;
      nnz = row_ptr[rows];

      return *this;
}
/*
-------------end-of-amub-----------------------------------------------
-----------------------------------------------------------------------
*/


