#ifndef EVENTS_H
#define EVENTS_H

#include <stdio.h>
#include "XtraPRoot.h"

#define TRACE_OKAY 0
#define TRACE_NOT_OKAY 1
#define TRACE_TYPE_TRC 0
#define TRACE_TYPE_CRT 1
#define TRACE_TYPE_DES 2
#define TRACE_NO_SUBSCRIPT -1
#define TRACE_READ 1
#define TRACE_WRITE 2
#define TRACE_UPDATE 3

#define DEF_TRACEFILE "ms"
#define DEF_DESFILE "Trace"
#define DEF_PARFILE "ms"
#define DEF_INTRACEFILE "ms"
#define DEF_OUTTRACEFILE "ms_"
#define DEF_STATFILE "ms"
#define UNKNOWN_EVENT "unknown_event"

#define MAX_PAR 5

#define PCXX_BUFSIZE 65536  /* -- 64 K -- */

/* Relic Events */
#define PCXX_LOCAL_ACCESS 	60500
#define PCXX_REMOTE_ACCESS 	60501
#define PCXX_REMOTE_BEGIN 	60502
#define PCXX_REMOTE_END		60503
#define PCXX_NO_EVENT	        60504
#define PCXX_SUBREMOTE_ACCESS   60505

#define PCXX_ALLOC_ELEM_ENTER     60109     /* R */    /* process    */
#define PCXX_ALLOC_ELEM_EXIT      60110     /* R */    /* MACH_Elem  */
#define PCXX_GET_ELEM_PTR         60111     /* R */    /* process    */
#define PCXX_COLL_ACCESS          60165     /* K */    /* coll_num   */
#define PCXX_TIMER1_START         60301     /* Ti*/
#define PCXX_TIMER1_STOP          60401     /* Ti*/

/* Relic Events end */

/* -- pcxx tracer events ------------------- */
#       define PCXX_EV_INIT         60000
#       define PCXX_EV_FLUSH_ENTER  60001
#       define PCXX_EV_FLUSH_EXIT   60002
#       define PCXX_EV_CLOSE        60003
#       define PCXX_EV_INITM        60004
#       define PCXX_EV_WALL_CLOCK   60005
#       define PCXX_EV_CONT_EVENT   60006

/* -- the following two events are only the ----- */
/* -- base numbers, actually both represent ----- */
/* -- 64 events (60[1234]00 to 60[1234]64)  ----- */
#       define PCXX_WTIMER_CLEAR    60199
#       define PCXX_WTIMER_START    60100
#       define PCXX_WTIMER_STOP     60200
#       define PCXX_UTIMER_CLEAR    60399
#       define PCXX_UTIMER_START    60300
#       define PCXX_UTIMER_STOP     60400

/* -- used in the benchmarks -------------------- */
# define PCXX_USERCODE_ENTER       61122     /* B */

/* -- general events for all SMRTS machines ----- */
/* -- used in the runtime system ---------------- */
# define PCXX_SETUP_ENTER          61100     /* B */    /* num_proc   */
# define PCXX_SETUP_INIT           61101     /* R */
# define PCXX_SETUP_EXIT           61102     /* B */    /* MACH_Num_P */
# define PCXX_FORK_ENTER           61103     /* B */
# define PCXX_FORK_EXIT            61104     /* B */

# define PCXX_NEWPROC_ENTER        61105     /* R */
# define PCXX_NEWPROC_EXIT         61106     /* R */    /* MACH_My_ID */
# define PCXX_BARRIER_ENTER        61107     /* R */
# define PCXX_BARRIER_EXIT         61108     /* R */
# define PCXX_DEALLOC_ENTER        61109     /* R */    /* ptr        */
# define PCXX_DEALLOC_EXIT         61110     /* R */

# define PCXX_MAIN_ENTER           61120     /* B */
# define PCXX_MAIN_EXIT            61121     /* B */

# define PCXX_INSTL_COLL           61150     /* K */    /* coll_num   */
# define PCXX_COLL_INFO_1          61151     /* K */    /* num_dim    */
# define PCXX_COLL_INFO_2          61152     /* K */    /* sizeelem   */
# define PCXX_COLL_INFO_3          61153     /* K */    /* nb_elem    */
# define PCXX_COLL_INFO_4          61154     /* K */    /* dim1size   */
# define PCXX_COLL_INFO_5          61155     /* K */    /* dim2size   */
# define PCXX_ELEM_INFO            61156     /* K */    /* elem_num   */
# define PCXX_ELEM_ACCESS          61157     /* K */    /* elem_ident */
# define PCXX_ELEM_SIZE            61158     /* K */    /* size       */
# define PCXX_ACS_ON               61159     /* E */    /* func. id   */
# define PCXX_ACS_OFF              61160     /* E */    /* func. id   */

# define PCXX_SYNC_MARK            61170     /* K */

typedef struct
{
  long int           ev;    /* -- event id        -- */
  short unsigned int nid;   /* -- node id         -- */
  short unsigned int tid;   /* -- thread id       -- */
  long int           par;   /* -- event parameter -- */
  unsigned long int  ti;    /* -- time [us]?      -- */
} Trc_Event_t;


typedef struct
{
  long int           ev;     /* -- event id              -- */
  short unsigned int nid;    /* -- node id               -- */
  short unsigned int tid;    /* -- thread id             -- */
  double	     ti;     /* -- time [us]?            -- */
  int 		     num_par;/* -- number of parameters  -- */
  long int           *par;   /* -- pointer to parameters -- */
} Crt_Event_t;

/* event description : event id, event name */
typedef struct {
  long int id;
  char name[256];
} EventName_t;

typedef struct {
  int numNames;
  int init;
  EventName_t *evNames; 
} EventNamesDB_t;

#ifndef TRCLIB_C
extern char *GetBaseName(int argc, char **argv);
extern char *GetBaseName1(int argc, char **argv);
extern char *GetBaseName2(int argc, char **argv);
extern FILE *OpenTraceFile(char *basename, int Type, int Sub, int RorW);
extern int ReadTrcEvent(FILE *fp, Trc_Event_t *event);
extern int WriteTrcEvent(FILE *fp, Trc_Event_t *event);
extern int SkipUntilEvent(FILE *fp, Trc_Event_t *event, long int EvType);
extern int ReadCrtEvent(FILE *fp, Crt_Event_t *event);
extern int WriteCrtEvent(FILE *fp, Crt_Event_t *event);
extern int RewindCrtFile(FILE *fp, Crt_Event_t *event);
extern int GetNumThreads(FILE *file, int Type);
extern int GetNumCollns(FILE *file, int Type);
extern void InitEventNamesDB();
extern char *GetEventName(long int id);
extern void PrintTrcEvent(Trc_Event_t *event);
extern void PrintCrtEvent(Crt_Event_t *event);
#endif

#endif
