#define GENERATOR_C
#include <stdio.h>
#include <stdlib.h>
#include "TaskLibIncludes.h"
#include "Trace.h"
#include "Main.h"
#include "Generator.h"

double BarrierExitTime = 0.0;
Semaphore mutex(1);
double ForkEnterTime = 0.0;
 
void BarrierEnter(double ti)
{
  GlobalBarr.wait();
  mutex.wait();
  if (ti > BarrierExitTime) BarrierExitTime = ti;
  mutex.signal();
  GlobalBarr.wait();
}

Generator_c::Generator_c(int pid_) 
{
  pid=pid_;
  fp=OpenTraceFile(OutFileName, TRACE_TYPE_CRT, pid, TRACE_WRITE);
  event.par = new long int[MAX_PAR];

  erand interval(RequestInterval);  
  urand remote(0, NumThreads - 1); 

  interval.seed(pid);
  remote.seed(pid);

  event.nid = pid;

// Generate the part until fork_enter for thread 0
  if (pid == 0) {

    event.ti = 0.0;

    event.ev = PCXX_SETUP_ENTER;
    event.ti += interval.draw();
    event.num_par = 1;
    event.par[0] = NumThreads;
    WriteCrtEvent(fp, &event);

    event.ev = PCXX_SETUP_EXIT;
    event.ti += interval.draw();
    WriteCrtEvent(fp, &event);

    event.ev = PCXX_FORK_ENTER;
    event.ti += interval.draw();
    event.num_par = 0;
    WriteCrtEvent(fp, &event);
 
    ForkEnterTime = event.ti + 1.0 ;    
  }

// Sync. so that everybody knows where to start
  GlobalBarr.wait();

// Start from fork enter event
  event.ti = ForkEnterTime;

  event.ev = PCXX_BARRIER_ENTER;
  event.ti += interval.draw();
  event.num_par = 0;
  WriteCrtEvent(fp, &event);

  BarrierEnter(event.ti); // enter the barrier recording the last time 

// Repeat the cycle - Barrier exit, remote accesses, barrier enter

  for(int i=0;i<NumBarriers - 1;i++) {

    event.ev = PCXX_BARRIER_EXIT;
    event.ti = BarrierExitTime + 1.0;
    WriteCrtEvent(fp, &event);

    event.ev = PCXX_REMOTE_ACCESS;
    event.num_par = 4;
    event.par[0] = 0;
    event.par[1] = 0;
    event.par[2] = ElemSize;
    for(int j=0;j<NumRequests;j++) {
     event.ti += interval.draw();
     event.par[3] = remote.draw();
     if (event.par[3] == pid) continue;
     WriteCrtEvent(fp, &event);
    }

    event.ev = PCXX_BARRIER_ENTER;
    event.ti += interval.draw();
    event.num_par = 0;
    WriteCrtEvent(fp, &event);

    BarrierEnter(event.ti); // enter the barrier recording the last time 
  }

  event.ev = PCXX_BARRIER_EXIT;
  event.ti = BarrierExitTime + 1.0;
  WriteCrtEvent(fp, &event);

// Sync so that thread 0 knows where to start the trailer

  GlobalBarr.wait();

// Thread 0 generates the trailier
  if (pid == 0) {
    event.ev = PCXX_FORK_EXIT;
    event.ti += interval.draw();
    WriteCrtEvent(fp, &event);
  }

  thistask->resultis(0);
}

#undef GENERATOR_C
