#define REMAP_C
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Trace.h"

/* 
  Format of the map file 

  <thread id> <thread id> ...
  <thread id> <thread id> ...
  ...
*/

typedef struct {
  int id;
  int numthreads;
  int *threads;
} Cluster_t;
  
main(int argc,char **argv)
{
  char *trace_base;
  char *new_base;
  FILE *trace_file, *Fin, *Fout;
  int *ClusterMap;
  int NumClusters, tid, i, threads_in_cluster;
  char Line[256];
  char *word;
  int NumThreads, remote_tid, cluster_id, myCluster;
  Crt_Event_t event;
  Cluster_t *Clusters;

  trace_base=GetBaseName1(argc,argv);
  new_base=GetBaseName2(argc,argv);

  /* open file 0 and get number of threads */
  trace_file = OpenTraceFile(trace_base, TRACE_TYPE_CRT, 0, TRACE_READ);
  NumThreads = GetNumThreads(trace_file, TRACE_TYPE_CRT);

  /* Create cluster map */
  Clusters = new Cluster_t[NumThreads];
  for(i=0;i<NumThreads;i++) {
    Clusters[i].id = i;
    Clusters[i].numthreads = 0;
    Clusters[i].threads = new int [NumThreads];
  }

  /* create reverse map. Each thread should know which cluster it belongs to */
  ClusterMap = new int [NumThreads];

  /* read the mapping line by line. The numbers in each line correspond to */
  /* threads that have been mapped together into the same cluster */

  NumClusters = 0;
  while (!feof(stdin)) {
    if (fgets(Line, 256, stdin) == NULL) break;
    word = strtok(Line, " .,?\"\n\t\r\f");
    if (word == NULL) continue;

    /* found a line containing threads in a cluster */
    while (word != NULL) {
      tid = atoi(word);

      /* note the current cluster with the thread */
      ClusterMap[tid] = NumClusters;

      /* add this thread to the current cluster */
      threads_in_cluster = Clusters[NumClusters].numthreads;
      Clusters[NumClusters].threads[threads_in_cluster] = tid;
      Clusters[NumClusters].numthreads++;
      word = strtok(NULL, " .,?\"\n\t\r\f");
    }
    NumClusters++;
  }

  event.par = new long int [MAX_PAR];

  /* For each cluster, create a dummy trace file */

  for(i=0;i<NumClusters;i++) {
     Fin = OpenTraceFile(trace_base, TRACE_TYPE_CRT, 
                         Clusters[i].threads[0], TRACE_READ);
     Fout = OpenTraceFile(new_base, TRACE_TYPE_CRT, 
                          NumThreads + i, TRACE_WRITE);
     while (!ReadCrtEvent(Fin, &event)) {
       if (event.ev == PCXX_BARRIER_ENTER ||
           event.ev == PCXX_BARRIER_EXIT ||
           event.ev == PCXX_MAIN_EXIT)
          WriteCrtEvent(Fout, &event);
     }
   }

  /* In each trace file, reclassify remote accesses */

  for(i=0;i<NumThreads;i++) {
    Fin = OpenTraceFile(trace_base, TRACE_TYPE_CRT, i, TRACE_READ);
    Fout = OpenTraceFile(new_base, TRACE_TYPE_CRT, i, TRACE_WRITE);
    myCluster = ClusterMap[i];
    while(!ReadCrtEvent(Fin, &event)) {
      if (event.ev == PCXX_REMOTE_ACCESS) {
        /* ASSUMPTION: # of threads < MAXINT */
        remote_tid = (int)(event.par[3]);
        cluster_id = ClusterMap[remote_tid];
        if (cluster_id == myCluster) 
          event.ev = PCXX_SUBREMOTE_ACCESS;
        else
          event.par[3] = NumThreads + cluster_id;
      }
      if (event.ev == PCXX_SETUP_ENTER || PCXX_SETUP_EXIT) {
        event.par[0] = NumThreads + NumClusters;
      }
      WriteCrtEvent(Fout, &event);
    }
  }
      
      
}

#undef REMAP_C
