#define TRCLIB_C
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include "Trace.h"

/* Convention: TRACE means either trc or crt */

char *GetBaseName(int argc,char **argv)
{
  if (argc==1)
    return DEF_TRACEFILE;
  else if (argc==2)
    return argv[1];
  else {
    fprintf(stderr,"Usage: %s [basename]\n",argv[0]);
    exit(1);
  }
  return NULL; //for the picky compilers
}

char *GetBaseName1(int argc,char **argv)
{
  if (argc==1)
    return DEF_TRACEFILE;
  else if (argc==2 || argc==3)
    return argv[1];
  else {
    fprintf(stderr,"Usage: %s [basename1 [basename2]]\n",argv[0]);
    exit(1);
  }
  return NULL; //for the picky compilers
}

char *GetBaseName2(int argc,char **argv)
{
  if (argc==1)
    return DEF_TRACEFILE;
  else if (argc==2)
    return argv[1];
  else if (argc==3)
    return argv[2];
  else {
    fprintf(stderr,"Usage: %s [basename1 [basename2]]\n",argv[0]);
    exit(1);
  }
  return NULL; //for the picky compilers
}

/* 
   Type = TRACE_TYPE_TRC | TRACE_TYPE_CRT 
   Subscript = TRACE_NO_SUBSCRIPT | integer >=0
   RorW = TRACE_READ | TRACE_WRITE
*/
FILE *OpenTraceFile(char *base_name, int Type, int Subscript, int RorW)
{
  char f_name[100];
  FILE *fp;
  char *suffix, *mode;

  if (Type == TRACE_TYPE_TRC) 
    suffix = "trc";
  else if (Type == TRACE_TYPE_CRT) 
    suffix = "crt";
  else {
   fprintf(stderr, "Unknown trace file type: %d\n", Type);
   exit(1);
  }
  
  if (Subscript == TRACE_NO_SUBSCRIPT)
    sprintf(f_name,"%s.%s", base_name, suffix);
  else
    sprintf(f_name,"%s%d.%s", base_name, Subscript, suffix);

  if (RorW == TRACE_READ)
    mode = "r";
  else if (RorW == TRACE_WRITE)
    mode = "w";
  else if (RorW == TRACE_UPDATE)
    mode = "r+";
  else {
    fprintf(stderr, "Unknown trace file read/write mode: %d\n", RorW);
    exit(1);
  }

  if ((fp=fopen(f_name,mode))==NULL) {
    fprintf(stderr, "Unable to open %s\n", f_name);
    exit(1);
  }
 
  return fp;
}
 
int ReadTrcEvent(FILE *fp, Trc_Event_t *event)
{
  if ((fread((char *)event, sizeof(Trc_Event_t), 1, fp)) == 0) 
     return TRACE_NOT_OKAY;
  else
     return TRACE_OKAY;
}

int WriteTrcEvent(FILE *fp, Trc_Event_t *event)
{
  if ((fwrite((char *)event, sizeof(Trc_Event_t), 1, fp)) == 0)
     return TRACE_NOT_OKAY;
  else
     return TRACE_OKAY;
}

int SkipUntilEvent(FILE *fp, Trc_Event_t *event, long int evType)
{
  while (1) {
     if (fread((char *)event, sizeof(Trc_Event_t), 1, fp) == 0) {
       fprintf(stderr,"Reached end of file while looking for event %ld\n",
               evType);
       exit(1);
     }
    if (event->ev == evType)
      return TRACE_OKAY;
   }
}

int ReadCrtEvent(FILE *fp, Crt_Event_t *event)
{  
  if ((fread((char *)event, sizeof(Crt_Event_t) - sizeof(long int *), 1, fp)) == 0)
    return TRACE_NOT_OKAY;
  if (event->num_par) {
    if ((fread((char *)(event->par), sizeof(long int), event->num_par, fp)) 
        == 0)
      return TRACE_NOT_OKAY;
  }
  return TRACE_OKAY;
}

int WriteCrtEvent(FILE *fp, Crt_Event_t *event)
{  
  if ((fwrite((char *)event, sizeof(Crt_Event_t) - sizeof(long int *), 1, fp)) == 0)
    return TRACE_NOT_OKAY;
  if (event->num_par) {
    if ((fwrite((char *)(event->par), sizeof(long int), event->num_par, fp)) 
        == 0)
      return TRACE_NOT_OKAY;
  }
  return TRACE_OKAY;
}

int RewindCrtFile(FILE *fp, Crt_Event_t *event)
{
  fseek (fp, - (event->num_par * sizeof(long int)), 1);
  fseek (fp, - (sizeof(Crt_Event_t) - sizeof(long int *)), 1);
  return TRACE_OKAY;
}

int GetNumCollns(FILE *file, int Type)
{
  int num_collns=0;

  if (Type == TRACE_TYPE_TRC) {
    Trc_Event_t t_event;
    while (!ReadTrcEvent(file, &t_event)) 
      if (t_event.ev == PCXX_INSTL_COLL ) 
        num_collns++;
  }
  else if (Type == TRACE_TYPE_CRT) {
    Crt_Event_t c_event;
    c_event.par = new long int[MAX_PAR];
    while(!ReadCrtEvent(file, &c_event))
      if (c_event.ev == PCXX_INSTL_COLL)
        num_collns++;
  }
  else {
    fprintf(stderr,"Unknown trace file type: %d\n", Type);
    exit(1);
  }

  fseek(file,0L,0);
  return num_collns;
}

int GetNumThreads(FILE *file, int Type)
{
  int found=0;
  int num_threads=0;

  if (Type == TRACE_TYPE_CRT) {
    Crt_Event_t c_event;
    c_event.par = new long int[MAX_PAR];
    while (1) {
      if (ReadCrtEvent(file, &c_event)) {
        fprintf(stderr,"setup-enter event not found\n");
        exit(1);
      }
      if (c_event.ev == PCXX_SETUP_ENTER) {
        /* ASSUMPTION: # of threads in a pC++ program < MAX_INT */
        num_threads = (int)c_event.par[0];  
        found = 1;
        break;
      }
    }
  }

  else if (Type == TRACE_TYPE_TRC) {
    Trc_Event_t t_event;
    while (1) {
      if (ReadTrcEvent(file, &t_event)) {
        fprintf(stderr,"setup-enter event not found\n");
        exit(1);
      }
      if (t_event.ev == PCXX_SETUP_ENTER) {
        /* ASSUMPTION: # of threads in a pC++ program < MAX_INT */
        num_threads = (int)t_event.par; 
        found = 1;
        break;
      }
    }
  }
  
  else {
    fprintf(stderr, "Unknown trace file type: %d\n", Type);
    exit(1);
  }

  if (found == 0) {
    fprintf(stderr,"setup-enter event not found\n");
    exit(1);
  }
  if (num_threads <= 0) {
    fprintf(stderr,"Invalid number of threads - %d\n",num_threads);
    exit(1);
  }

  fseek(file,0L,0);
  return num_threads;
}

EventNamesDB_t EventNamesDB;

void InitEventNamesDB()
{
  int i;
  FILE *file;
  FILE *in;
  char DBFile[1024];
  char *p;

  /* Construct DBFile name */
  DBFile[0]=0;
  if (fopen(XTRAPROOT, "r") != NULL) {
    strcat(DBFile,XTRAPROOT);
  }
  else if ((in = fopen("./SAGEROOT", "r")) != NULL) {
    fgets(DBFile, 1024, in);
    fclose(in);
    DBFile[strlen(DBFile)-1] = '\0'; /*chop off newline*/
    strcat(DBFile,"/XtraP");
  }
  else if ((p = getenv("SAGEROOT")) != NULL) {
    strcat(DBFile, p);
    strcat(DBFile,"/XtraP");
  }
  else {
    fprintf(stderr,"ERROR: cannot find XTRAPROOT\n");
    exit(1);
  }

  strcat(DBFile,XTRAPROOT);
  strcat(DBFile,"/include/");
  strcat(DBFile,DEF_DESFILE);
  strcat(DBFile,".des");

  /* Open DBFile */
  file = fopen(DBFile,"r");
  if (file == NULL) {
    fprintf(stderr, "Cannot open %s\n", DBFile);
    exit(1);
  }

  /* Read number of event names */
  fscanf(file,"%d",&(EventNamesDB.numNames));
  EventNamesDB.evNames = (EventName_t *)malloc(sizeof(EventName_t) * 
                                                EventNamesDB.numNames);

  /* Read event name pairs */
  for(i=0;i<EventNamesDB.numNames;i++) 
    fscanf(file,"%ld %s",&(EventNamesDB.evNames[i].id), 
                         EventNamesDB.evNames[i].name);

  /* DB has been initialized */
  EventNamesDB.init = 1;
}

int BinSearch(long int id, int start, int end)
{
  int middle;

  if (EventNamesDB.init != 1) {
   fprintf(stderr, "Event names database not initialized\n");
   exit(1);
  }

  if (id < EventNamesDB.evNames[start].id) return -1;
  if (id > EventNamesDB.evNames[end].id) return -1;
  if (start==end) {
    if (id == EventNamesDB.evNames[start].id) return start;
    else return -1;
  }
  middle = (end-start)/2 + start;
  if (id == EventNamesDB.evNames[middle].id) return middle;
  else if (id < EventNamesDB.evNames[middle].id)
    return BinSearch(id, start,middle-1);
  else
    return BinSearch(id, middle+1,end);
}

char *GetEventName(long int ev_id)
{
  int found;

  if (EventNamesDB.init != 1) {
   fprintf(stderr, "Event names database not initialized\n");
   exit(1);
  }
  
  found=BinSearch(ev_id, 0, EventNamesDB.numNames-1);
  if (found == -1)
    return UNKNOWN_EVENT;
  else
    return EventNamesDB.evNames[found].name;
}

void PrintCrtEvent(Crt_Event_t *event)
{
  char *event_name;
  int i;

  event_name=GetEventName(event->ev);
  fprintf(stdout, "%-10.2f %-8ld %-20s %-5hu %-d   ", 
	   event->ti, event->ev, event_name, event->nid, event->num_par);
  for(i=0;i<event->num_par;i++) 
    fprintf(stdout, " %-ld",event->par[i]);
  fprintf(stdout, "\n");
}

void PrintTrcEvent(Trc_Event_t *event)
{
  char *event_name;

  event_name=GetEventName(event->ev);
  fprintf(stdout, "%11u %8ld %25s %5u %15ld\n", 
	   event->ti, event->ev, event_name, event->nid, event->par);
  fprintf(stdout, "\n");
}

#undef TRCLIB_C
