#define MAIN_C
#include <stdio.h>
#include <stdlib.h>
#include "TaskLibIncludes.h"
#include "GetOpt.h"
#include "Trace.h"
#include "Collection.h"
#include "Main.h"
#include "Reader.h"
#include "Translator.h"

CollInfo_t *CollInfo;
int NumThreads,NumCollns, NumBarriers;
int Overhead = 130;
double ForkTime = 0.0, Timer1Time = 0.0, SelectedTime = 0.0;
Barrier Reader_Translators_Barr(0);
Barrier Translators_Barr(0);
Reader_c *Reader;
Translator_c **Translator;
char *InFileName = DEF_INTRACEFILE;
char *OutFileName = DEF_OUTTRACEFILE;
char *StatFileName = DEF_STATFILE;
int NoOutFlag = 0;
int UseElemSizeFlag = 0;

void AllocateCollInfo()
{
  int i,j,p,q;

  CollInfo=(CollInfo_t *)malloc(sizeof(CollInfo_t) * NumCollns);
  for(i=0;i<NumCollns;i++) {
    CollInfo[i].AccessMatrix = (int **)malloc(sizeof(int *) * NumThreads);
    for(j=0;j<NumThreads;j++)
      CollInfo[i].AccessMatrix[j] = (int *)malloc(sizeof(int) * NumThreads);
    for(p=0;p<NumThreads;p++)
      for(q=0;q<NumThreads;q++)
	CollInfo[i].AccessMatrix[p][q] = 0;
    CollInfo[i].done = 0;
  }
}

void PrintUsage(char *pgm)
{
  fprintf(stderr, "Usage: %s [ -i <infile> ] [ -o <outfile> | -n ] [ -s <statfile> | -N] [-t <overhead>] [-e]\n", pgm);
}

void PrintStats()
{
  int i,j,k,p,q;
  int globalindex;
  char filename[256];
  FILE *file;

  sprintf(filename, "%s.info", StatFileName);
  file = fopen(filename, "w");
  if (file == NULL) {
    fprintf(stderr, "Unable to open %s\n", filename);
    exit(1);
  }

  fprintf(file, "Number of threads       : %d\n", NumThreads);
  fprintf(file, "Number of collections   : %d\n", NumCollns);
  fprintf(file, "Number of barriers      : %d\n", NumBarriers);
  fprintf(file, "Fork time               : %f\n", (ForkTime)/1000000);
  fprintf(file, "Timer1 time             : %f\n", (Timer1Time)/1000000);
  fprintf(file, "Selected time           : %f\n", (SelectedTime)/1000000);

  fprintf(file, "---------------------------------------------\n");
  fprintf(file, "Collection Profile\n");
  fprintf(file, "---------------------------------------------\n");

  for(i=0;i<NumCollns;i++) {
    fprintf(file, "---------------------------------------------\n");
    fprintf(file, "Collection number     : %d\n",i);
    fprintf(file, "---------------------------------------------\n");
    fprintf(file, "  Number of dimensions: %d\n", CollInfo[i].num_dim);
    fprintf(file, "  Number of elements  : %d\n", CollInfo[i].nb_elem);
    fprintf(file, "  Size of element     : %d\n", CollInfo[i].sizeelem);
    fprintf(file, "  Size of dimension 1 : %d\n", CollInfo[i].dim1size);
    fprintf(file, "  Size of dimension 2 : %d\n", CollInfo[i].dim2size);

    fprintf(file, "  Data distribution   : \n");
    fprintf(file, "      Element #     Thread # \n");

    if (CollInfo[i].num_dim == 1) {
      for(j=0;j<CollInfo[i].dim1size;j++) {
        globalindex = j;
        fprintf(file, "%10d %10d\n", j, CollInfo[i].DistVector[globalindex]);
      }
    }
    else if (CollInfo[i].num_dim == 2) {
      for(j=0;j<CollInfo[i].dim1size;j++)
        for(k=0;k<CollInfo[i].dim2size;k++) {
          globalindex = j + k * CollInfo[i].dim1size;
          fprintf(file, "(%5d, %5d) %10d\n",j, k, CollInfo[i].DistVector[globalindex]);
        }
    }

    fprintf(file, "  Access matrix       : \n");
    for(p=0;p<NumThreads;p++){
      for(q=0;q<NumThreads;q++)
        fprintf(file, "%7d", CollInfo[i].AccessMatrix[p][q]);
      fprintf(file, "\n");
    }
    fprintf(file, "-----------------------------------------\n");
  }

}

main(int argc,char **argv)
{
  FILE *file;
  int i, ch;
  int InFlag, OutFlag, StatFlag, NoStatFlag, OverheadFlag;

  GetOpt getopt(argc, argv, "i:o:s:t:nNe");

  InFlag = OutFlag = StatFlag = NoOutFlag = NoStatFlag = OverheadFlag = 0;
  UseElemSizeFlag = 0;
  while((ch = getopt()) != -1) {
    switch(ch) {
     case 'i':if (InFlag == 1) {
               fprintf(stderr, "-i flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              InFileName = getopt.optarg;
              InFlag = 1;
              break;
     case 'o':if (OutFlag == 1) {
               fprintf(stderr, "-o flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              if (NoOutFlag == 1) {
               fprintf(stderr, "conflicting flags -o and -n\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              OutFileName = getopt.optarg;
              OutFlag = 1;
              break;
     case 's':if (StatFlag == 1) {
               fprintf(stderr, "-s flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              if (NoStatFlag == 1) {
               fprintf(stderr, "conflicting flags -s and -N\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              StatFileName = getopt.optarg;
              StatFlag = 1;
              break;
     case 't':if (OverheadFlag == 1) {
               fprintf(stderr, "-t flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              Overhead = atoi(getopt.optarg);
              OverheadFlag = 1;
              break;
     case 'n':if (NoOutFlag == 1) {
               fprintf(stderr, "-n flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              if (OutFlag == 1) {
               fprintf(stderr, "conflicting flags -o and -n\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              NoOutFlag = 1;
              break;
     case 'N':if (NoStatFlag == 1) {
               fprintf(stderr, "-N flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              if (StatFlag == 1) {
               fprintf(stderr, "conflicting flags -s and -N\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              NoStatFlag = 1;
              break;
     case 'e':if (UseElemSizeFlag == 1) {
               fprintf(stderr, "-e flag encountered twice\n");
               PrintUsage(argv[0]);
               exit(1);
              }
              UseElemSizeFlag = 1;
              break;
     default: fprintf(stderr, "Unknown flag : -%c\n", ch);
              PrintUsage(argv[0]);
              exit(1);
    }
  }


  file=OpenTraceFile(InFileName,TRACE_TYPE_TRC, TRACE_NO_SUBSCRIPT, TRACE_READ);
  NumThreads=GetNumThreads(file, TRACE_TYPE_TRC);
  Reader_Translators_Barr.height(NumThreads + 1);
  Translators_Barr.height(NumThreads);
  NumCollns=GetNumCollns(file, TRACE_TYPE_TRC);
  AllocateCollInfo();

  NumBarriers = 0;

  Reader = new Reader_c();
  Translator = new Translator_c*[NumThreads];
  for(i=0;i<NumThreads;i++) 
    Translator[i] = new Translator_c(i);

  Reader->result();
  for(i=0;i<NumThreads;i++)
    Translator[i]->result();

  if (!NoStatFlag)
    PrintStats();

  thistask->resultis(0);

  /* Programs using CC task library should not return. They should
     use resultis instead */
  /* return 0; */
}

#undef MAIN_C
