#define TRANSLATOR_C
#include <stdio.h>
#include <stdlib.h>
#include "TaskLibIncludes.h"
#include "Trace.h"
#include "Collection.h"
#include "Main.h"
#include "Reader.h"
#include "Translator.h"
    

/* To be called by the reader for delivering an event to this translator */

void Translator_c::DeliverEvent(Trc_Event_t *event)
{
  /* wait for this translator to give permission to deliver */
  barr.wait(); 

  /* have permission to deliver. Deliver either an event or end of stream */
  if (event != NULL) {
   t_event.ev = event->ev;
   t_event.ti = event->ti;
   t_event.nid = event->nid;
   t_event.tid = event->tid;
   t_event.par = event->par;
  }
  else {
   done = 1;
  }

  /* signal this translator that the reader has delivered an event */
  barr.wait();
}

/* To be called by this translator to wait for an event to be delivered by */
/* the reader */

int Translator_c::WaitForEvent()
{
  /* signal reader to deliver */
  barr.wait();
  /* wait for event to be delivered */
  barr.wait();
  /* return 1 if there is an event */
  return !done;
}
 
Translator_c::Translator_c(int pid_) 
{
  int collnum, elemnum, owner, index, installing;
  double timer1time, selectedtime, forktime;
  int accessingSize = -1;

  pid=pid_;
  if (!NoOutFlag)
   fp=OpenTraceFile(OutFileName, TRACE_TYPE_CRT, pid, TRACE_WRITE);
  c_event.par = new long int[MAX_PAR];
  barr.height(2);
  done = 0;

  /* Make sure every body starts properly */
  Reader_Translators_Barr.wait();

  while (WaitForEvent()) { 
    switch( t_event.ev ) {
    case PCXX_SETUP_ENTER:
    case PCXX_SETUP_EXIT:
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 1;
      c_event.par[0] = t_event.par;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      break;

    case (PCXX_UTIMER_START+1):
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 1;
      c_event.par[0] = t_event.par;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      timer1time = c_event.ti;
      break;

    case (PCXX_UTIMER_STOP+1):
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 1;
      c_event.par[0] = t_event.par;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      Timer1Time += c_event.ti - timer1time;
      break;

    case PCXX_ACS_ON:
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 1;
      c_event.par[0] = t_event.par;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      selectedtime = c_event.ti;
      break;

    case PCXX_ACS_OFF:
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 1;
      c_event.par[0] = t_event.par;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      SelectedTime += c_event.ti - selectedtime;
      break;

    case PCXX_FORK_ENTER:
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 0;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      forktime = c_event.ti;
      break;

    case PCXX_FORK_EXIT:
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 0;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      ForkTime = c_event.ti - forktime;
      break;

    case PCXX_BARRIER_EXIT:
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 0;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      break;
      
    case PCXX_BARRIER_ENTER:
      Translators_Barr.wait();
      c_event.ti = (double) t_event.ti;
      c_event.ev = t_event.ev;
      c_event.nid = t_event.nid;
      c_event.num_par = 0;
      if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      if (pid == 0) NumBarriers++;
      break;
      
    case PCXX_INSTL_COLL:
      /* ASSUMPTION: A pC++ program doesn't use more than MAX_INT collections */
      installing = (int)t_event.par; 
      break;
      
    case PCXX_COLL_INFO_1:
      if (!CollInfo[installing].done)
        /* ASSUMPTION: the number of dimensions of a collections < MAX_INT */
	CollInfo[installing].num_dim = (int)t_event.par;
      break;
      
    case PCXX_COLL_INFO_2:
      if (!CollInfo[installing].done)
        /* ASSUMPTION: the size of an element in a collection is < MAX_INT */
	CollInfo[installing].sizeelem = (int)t_event.par;
      break;
      
    case PCXX_COLL_INFO_3:
      if (!CollInfo[installing].done) {
        /* ASSUMPTION: the number of elements in a collection is < MAX_INT */
	CollInfo[installing].nb_elem = (int)t_event.par;
	CollInfo[installing].DistVector = 
	  (int *)malloc(sizeof(int) * (int)t_event.par);
      }
      break;

    case PCXX_COLL_INFO_4:
      if (!CollInfo[installing].done)      
        /* ASSUMPTION: size of first dimension in a collection is < MAX_INT */
	CollInfo[installing].dim1size = (int)t_event.par;
      break;

    case PCXX_COLL_INFO_5:
      if (!CollInfo[installing].done) {
        /* ASSUMPTION: size of second dimension in a collection is < MAX_INT */
	CollInfo[installing].dim2size = (int)t_event.par;
	CollInfo[installing].done = 1; 
      }
      break;

    case PCXX_ELEM_INFO:
      /* ASSUMPTION: # of threads (owners) in a pC++ program is < MAX_INT */
      owner = (int) ((t_event.par & (~0 << 24)) >> 24);
      /* ASSUMPTION: # of elmts & index of an elmnt in a colln is < MAX_INT */
      index = (int) (t_event.par - (owner << 24));
      CollInfo[installing].DistVector[index] = owner;
      break;

    case PCXX_ELEM_SIZE:
      /* ASSUMPTION: size of an elmnt or its part being accessed < MAX_INT */
      accessingSize = (int)t_event.par;
      break;

    case PCXX_ELEM_ACCESS:
      /* ASSUMPTION: # of collns & index of a colln is < MAX_INT */
      collnum = (int)((t_event.par & (~0 << 24)) >> 24);
      /* ASSUMPTION: # of elmts & index of an elmnt is < MAX_INT */
      elemnum = (int)(t_event.par - (collnum << 24));
      c_event.ti = (double) t_event.ti;
      c_event.nid = t_event.nid;
      c_event.num_par = 4;
      c_event.par[0] = collnum;
      c_event.par[1] = elemnum;
      if ((accessingSize != -1) && (!UseElemSizeFlag)) {
        c_event.par[2] = accessingSize;
        accessingSize = -1;  
      }
      else
        c_event.par[2] = CollInfo[collnum].sizeelem;
      c_event.par[3] = CollInfo[collnum].DistVector[elemnum];
      CollInfo[collnum].AccessMatrix[pid][c_event.par[3]] ++;
      if (pid == c_event.par[3])
	c_event.ev = PCXX_LOCAL_ACCESS;
        /* if (!NoOutFlag) WriteCrtEvent(fp, &c_event); */
      else {
	c_event.ev = PCXX_REMOTE_ACCESS;
        if (!NoOutFlag) WriteCrtEvent(fp, &c_event);
      }
      break;
      
    default: 
      break;
    }
  }

  Reader_Translators_Barr.wait();

  thistask->resultis(0);
}

#undef TRANSLATOR_C
