#define BARRIERMODEL_C
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

int NumNodes;
int NumNodesInBarrier;

double BarrierEntryTime()
{
  return _BarrierEntryTime;
}

double BarrierCheckTime()
{
  return _BarrierCheckTime;
}

double BarrierModelTime()
{
  return NumNodes * _BarrierModelTime;
}

double BarrierExitCheckTime()
{
  return _BarrierExitCheckTime;
}

double BarrierExitTime()
{
  return _BarrierExitTime;
}

void BarrierModelInit(int numnodes)
{
  NumNodes = numnodes;
  NumNodesInBarrier = 0;
}

void BarrierModel(Node_c *Node)
{
  Network_c *Network;
  FILE *Fpout;
  Crt_Event_t Event,*CurrentEvent, *PrevEvent;
  double ComputationTime;
  Msg_c *Msg, *BarrierExitMsg, *BarrierEntryMsg;
  int i;

  Network = Node->GetSys()->GetNetwork();
  Fpout = Node->GetFpout();
  CurrentEvent = Node->GetCurrentEvent();
  PrevEvent = Node->GetPrevEvent();
  Event.par = new long int[MAX_PAR];

  ComputationTime = CurrentEvent->ti - PrevEvent->ti;
  Computation(Node, ComputationTime);

  if (!NoOutTrace) {
    Event.ev = PCXX_BARRIER_ENTER;
    Event.ti = GetCurrentTime();
    Event.nid = CurrentEvent->nid;
    Event.num_par = CurrentEvent->num_par;
    for(i=0;i<Event.num_par;i++)
      Event.par[i] = CurrentEvent->par[i];
    WriteCrtEvent(Fpout, &Event);
  }

  Suspend(BarrierEntryTime());

  if (Node->GetNodeId() == 0) {
    NumNodesInBarrier++;
    Suspend(BarrierCheckTime());
    while (NumNodesInBarrier != NumNodes) {
      Msg = Node->GetMsg();
      ProcessMsg(Node, Msg);
      Suspend(BarrierCheckTime());
    }
    NumNodesInBarrier = 0;
    Suspend(BarrierModelTime());
    for(i=1;i<NumNodes;i++) {
      if (_BarrierByMsgs == 0)
        BarrierExitMsg = new Msg_c(IMAG, BARRIEREXITMSG, 0, 0, i, 0);
      else
        BarrierExitMsg = new Msg_c(REAL, BARRIEREXITMSG, _BarrierMsgSize, 
                                   0, i, 0);
      Network->SendMsg(BarrierExitMsg);
    }
  }
  else {
    if (_BarrierByMsgs == 0) 
      BarrierEntryMsg = new Msg_c(IMAG, BARRIERENTRYMSG, 0,
  				  Node->GetNodeId(), 0, 0);
    else
      BarrierEntryMsg = new Msg_c(REAL, BARRIERENTRYMSG, _BarrierMsgSize,
  				  Node->GetNodeId(), 0, 0);
      
    Network->SendMsg(BarrierEntryMsg);
    Msg = Node->GetMsg();
    Suspend(BarrierExitCheckTime());
    while (Msg->GetMsgType() != BARRIEREXITMSG) {
      ProcessMsg(Node,Msg);
      Msg = Node->GetMsg();
      Suspend(BarrierExitCheckTime());
    }
    ProcessMsg(Node,Msg);
  }

  Suspend(BarrierExitTime());
  
  if (!NoOutTrace) {
    Event.ev = PCXX_BARRIER_EXIT;
    Event.ti = GetCurrentTime();
    Event.nid = CurrentEvent->nid;
    Event.num_par = CurrentEvent->num_par;
    for(i=0;i<Event.num_par;i++)
      Event.par[i] = CurrentEvent->par[i];
    WriteCrtEvent(Fpout, &Event);
  }

  delete Event.par;
}

#undef BARRIERMODEL_C
