#define COMPUTATION_C
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

void UninterruptedComputation( Node_c *, double Time)
{
  double NewTime;
  
  NewTime = Time * _MipsRatio;
  Suspend(NewTime);
}

void HandleMsgAtEndComputation( Node_c *Node, double Time)
{
  double NewTime;
  Msg_c *Msg;
 
  NewTime = Time * _MipsRatio;
  Suspend(NewTime);
  while ( (Msg = Node->GetMsgNoBlock()) != (Msg_c *)NULL) 
    ProcessMsg(Node, Msg);
}
   
void PolledComputation( Node_c *Node, double Time)
{
  int Num,i;
  double NewTime,Remainder;
  Msg_c *Msg;
  
  NewTime = Time * _MipsRatio;
  Num = (int)(NewTime / _Delta);
  Remainder = NewTime - (Num * _Delta);
  for(i=0;i < Num; i++) {
    Suspend(_Delta);
    while ( (Msg = Node->GetMsgNoBlock()) != (Msg_c *)NULL) {
      ProcessMsg(Node, Msg);
    }
  }
  Suspend(Remainder);
}

void PolledAndMsgAtEndComputation( Node_c *Node, double Time)
{
  int Num,i;
  double NewTime,Remainder;
  Msg_c *Msg;
  
  NewTime = Time * _MipsRatio;
  Num = (int)(NewTime / _Delta);
  Remainder = NewTime - (Num * _Delta);
  for(i=0;i < Num; i++) {
    Suspend(_Delta);
    while ( (Msg = Node->GetMsgNoBlock()) != (Msg_c *)NULL) {
      ProcessMsg(Node, Msg);
    }
  }
  Suspend(Remainder);
  while ( (Msg = Node->GetMsgNoBlock()) != (Msg_c *)NULL) {
    ProcessMsg(Node, Msg);
    }
}

void InterruptedComputation( Node_c *Node, double Time)
{
  double NewTime, expectedOutTime, rem;
  Msg_c *Msg;
 
  NewTime = Time * _MipsRatio;
  expectedOutTime = GetCurrentTime() + NewTime;
  rem = NewTime;
  while (Node->iSuspend(rem)) {
     rem = expectedOutTime - GetCurrentTime();
     while ( (Msg = Node->GetMsgNoBlock()) != (Msg_c *)NULL) {
      ProcessMsg(Node, Msg);
     }
  } 
}

void Computation( Node_c *Node, double Time)
{
  switch(_ProcessMsgType) {
   case 0: UninterruptedComputation(Node, Time);
           break;
   case 1: HandleMsgAtEndComputation(Node, Time);
           break;
   case 2: PolledComputation(Node, Time);
           break;
   case 3: PolledAndMsgAtEndComputation(Node, Time);
           break;
   case 4: InterruptedComputation(Node, Time);
           break;
   default:UninterruptedComputation(Node, Time);
           break;
  }
}
#undef COMPUTATION_C
