#define MSGQUEUE_C
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

MsgQueue_c::MsgQueue_c(int MaxMsg):Mutex(1), Empty(MaxMsg), Full(MaxMsg)
{
  int i;
  Front = End = (Msg_c *)NULL;
  QCount = 0;
  for(i=0;i<MaxMsg;i++)
    Full.wait();
}

double AddMsgTime(Msg_c *Msg)
{
  return (_AddMsgTime * Msg->GetMsgSize() + _AddMsgStartupTime);
}

double AddMsgTryTime()
{
  return _AddMsgTryTime;
}

double GetMsgTryTime()
{
  return _GetMsgTryTime;
}

double GetMsgTime(Msg_c *Msg)
{
  return (_GetMsgTime * Msg->GetMsgSize() + _GetMsgStartupTime);
}

int MsgQueue_c::AddMsg(Msg_c *Msg)
{
  if (Msg->GetMsgCat() == REAL)
    Suspend(AddMsgTryTime());
  Empty.wait();
  Mutex.wait();
  if (QCount == 0)
    Front = Msg;
  else 
    End->SetNextMsg(Msg);
  End = Msg;
  End->SetNextMsg((Msg_c *)NULL);
  QCount++;
  if (Msg->GetMsgCat() == REAL)
    Suspend(AddMsgTime(Msg));
  Mutex.signal();
  Full.signal();
  return 1;
}

int MsgQueue_c::AddMsgNoBlock(Msg_c *Msg)
{
  if (Msg->GetMsgCat() == REAL)
    Suspend(AddMsgTryTime());
  if (! Empty.waitNoBlock()) 
    return -1;
  Mutex.wait();
  if (QCount == 0)
    Front = Msg;
  else 
    End->SetNextMsg(Msg);
  End = Msg;
  End->SetNextMsg((Msg_c *)NULL);
  QCount++;
  if (Msg->GetMsgCat() == REAL)
    Suspend(AddMsgTime(Msg));
  Mutex.signal();
  Full.signal();
  return 1;
}

Msg_c * MsgQueue_c::GetMsg()
{
  Msg_c *Msg;
 
  Suspend(GetMsgTryTime());
  Full.wait();
  Mutex.wait();
  Msg = Front;
  Front = Front->GetNextMsg();
  QCount--;
  if (Msg->GetMsgCat() == REAL)
    Suspend(GetMsgTime(Msg));
  Mutex.signal();
  Empty.signal();
  return Msg;
}

Msg_c * MsgQueue_c::GetMsgNoBlock()
{
  Msg_c *Msg;

  Suspend(GetMsgTryTime());
  if (! Full.waitNoBlock())
    return (Msg_c *)NULL;
  Mutex.wait();
  Msg = Front;
  Front = Front->GetNextMsg();
  QCount--;
  if (Msg->GetMsgCat() == REAL)
    Suspend(GetMsgTime(Msg));
  Mutex.signal();
  Empty.signal();
  return Msg;
}

#undef MSGQUEUE_C
