#define MSGTHREAD_C
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

Semaphore Mutex(1);
int NumActiveMsgs = 0;
int SizeActiveMsgs = 0;

double MsgTransferTime(Msg_c *Msg)
{
  double time;
  int Dist,Size;
  int i,j,n;

  Mutex.wait();
  if (Msg->GetMsgCat() == REAL) 
    switch (_Topology) {

    case 0:   /* Fully connected, no contention for links */
      Size = Msg->GetMsgSize();
      time = _CommStartupTime + _ByteTransferTime * Size;
      break;

    case 1:   /* Ring */
      if (Msg->GetFromNodeId() < Msg->GetToNodeId()) {
	i = Msg->GetFromNodeId();
	j = Msg->GetToNodeId();
      }
      else {
	j = Msg->GetFromNodeId();
	i = Msg->GetToNodeId();
      }
      n = GetNumNodes();
      if (n-j+i < j-i)
	Dist = n-j+i;
      else
	Dist = j-i;
      Size = Msg->GetMsgSize();
      time = _CommStartupTime + _ByteTransferTime * Dist * Size;
      break;

   case 3: /* Ring with contention */
      if (Msg->GetFromNodeId() < Msg->GetToNodeId()) {
	i = Msg->GetFromNodeId();
	j = Msg->GetToNodeId();
      }
      else {
	j = Msg->GetFromNodeId();
	i = Msg->GetToNodeId();
      }
      n = GetNumNodes();
      if (n-j+i < j-i)
	Dist = n-j+i;
      else
	Dist = j-i;
      Size = Msg->GetMsgSize();
      time = _CommStartupTime + _ByteTransferTime * Dist * Size;
      time += ((NumActiveMsgs-1) / 2) * _CommStartupTime +
                  ((SizeActiveMsgs-Size) / 2) * (GetNumNodes() / 2) *
                  _ByteTransferTime;
      break;

    case 4: /* Pipelined Ring */
      if (Msg->GetFromNodeId() < Msg->GetToNodeId()) {
	i = Msg->GetFromNodeId();
	j = Msg->GetToNodeId();
      }
      else {
	j = Msg->GetFromNodeId();
	i = Msg->GetToNodeId();
      }
      n = GetNumNodes();
      if (n-j+i < j-i)
	Dist = n-j+i;
      else
	Dist = j-i;
      Size = Msg->GetMsgSize();
      time = _CommStartupTime + (Dist + Size - 1) * _ByteTransferTime;
      break;

    case 2: /* shared network */
      Size = Msg->GetMsgSize();
      time = _CommStartupTime + _ByteTransferTime * Size;
      time += ((NumActiveMsgs-1) / 2) * _CommStartupTime +
              ((SizeActiveMsgs-Size) / 2) * _ByteTransferTime;
      break;

    default:
      Size = Msg->GetMsgSize();
      time = _CommStartupTime + _ByteTransferTime * Size;
      break;
    }
  else
    time = 0;
  Mutex.signal();
  return time;
}

int
MsgThread_c::Interrupt(Node_c *t)
{
   if (t->GetState() == 1) {  /* interruptible */
      t->SetIntrTask(thistask);
      long rem = t->preempt();
      t->SetRemain(rem);
      // t->delay(0);
      t->sleep(thistask);
      alert();
      return 1;
   }
   return 0; /* cannot interrupt */
}

MsgThread_c::MsgThread_c(System_c *sys,Msg_c *msg)
{
// real constructor
  Sys = sys;
  Msg = msg;
  Mutex.wait();
  NumActiveMsgs++;
  SizeActiveMsgs += Msg->GetMsgSize();
  Mutex.signal();

// body
  int tonodeid;
  Node_c *tonode;
  double res;

  tonodeid = Msg->GetToNodeId();
  tonode = Sys->GetNode(tonodeid);
  if (Msg->GetMsgCat() == REAL)
    Suspend(MsgTransferTime(Msg));
  tonode->DeliverMsg(Msg);
  if (_ProcessMsgType == 4 && tonode->GetState() == 1) {
    res = Interrupt(tonode); 
    if (!res) {
      fprintf(stderr, "Unsuccesfull interrupt\n");
      exit(1);
    }
  }      

// real destructor
  Mutex.wait();
  NumActiveMsgs--;
  SizeActiveMsgs -= Msg->GetMsgSize();
  Mutex.signal();

  thistask->resultis(0);
}

Msg_c *MsgThread_c::GetMsg()
{
  return Msg;
}

#undef MSGTHREAD_C
