#define NODE_C
#include <sys/time.h>
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

#if 0
/* sys/time.h should define this! */
extern "C" int gettimeofday(struct timeval *t, struct timezone *z);
#endif

static struct timeval tv1, tv2;
void TimerBegin()
{
  gettimeofday(&tv1, NULL);
}

double TimerEnd()
{
  gettimeofday(&tv2, NULL);
  return ((tv2.tv_sec - tv1.tv_sec) + (tv2.tv_usec - tv1.tv_usec) / 1000000.0);
}

Node_c::Node_c(System_c *sys,int node_id)
{

// prolog
  NodeId = node_id;
  Sys = sys;
  MsgQueue = new MsgQueue_c(100);
  State = 0;
  CurrentEvent = new Crt_Event_t;
  CurrentEvent->par = new long int [MAX_PAR];
  PrevEvent = new Crt_Event_t;
  PrevEvent->par = new long int [MAX_PAR];
  PrevEvent->ev = PCXX_NO_EVENT;
  PrevEvent->nid = NodeId;
  PrevEvent->ti = 0;
  PrevEvent->num_par = 0;
  Fpin = OpenTraceFile(InFileName, TRACE_TYPE_CRT, NodeId, TRACE_READ);
  if (!NoOutTrace)
    Fpout = OpenTraceFile(OutFileName, TRACE_TYPE_CRT, NodeId, TRACE_WRITE);

// body
  Crt_Event_t *event_ptr;
  double timer1time1, timer1time2, forktime1, forktime2;
  double selectedtime1, selectedtime2, selectedtime;
  double timer1time, simtime, forktime;

  timer1time1 = timer1time2 = 0;
  selectedtime1 = selectedtime2 = 0;
  forktime1 = forktime2 = 0;
  timer1time = 0.0;
  selectedtime = 0.0;
  TimerBegin();
  while (!ReadCrtEvent(Fpin, CurrentEvent)) {
    switch(CurrentEvent->ev) {
    case PCXX_REMOTE_ACCESS:
      RemoteAccess(this);
      break;
    case PCXX_BARRIER_ENTER:
      BarrierModel(this);
      break;
    case PCXX_BARRIER_EXIT:
      break;
    case PCXX_TIMER1_START:
      DefaultSimulation(this);
      timer1time1 = GetCurrentTime();
      break;
    case PCXX_TIMER1_STOP:
      DefaultSimulation(this);
      timer1time2 = GetCurrentTime();
      timer1time += timer1time2 - timer1time1;
      break;
    case PCXX_ACS_ON:
      DefaultSimulation(this);
      selectedtime1 = GetCurrentTime();
      break;
    case PCXX_ACS_OFF:
      DefaultSimulation(this);
      selectedtime2 = GetCurrentTime();
      selectedtime += selectedtime2 - selectedtime1;
      break;
    case PCXX_FORK_ENTER:
      DefaultSimulation(this);
      forktime1 = GetCurrentTime();
      break;
    case PCXX_FORK_EXIT:
      DefaultSimulation(this);
      forktime2 = GetCurrentTime();
    case PCXX_SETUP_ENTER:
    case PCXX_SETUP_EXIT:
    case PCXX_LOCAL_ACCESS:
    case PCXX_SUBREMOTE_ACCESS:
    default:
      DefaultSimulation(this);
      break;
    }

    event_ptr = PrevEvent;
    PrevEvent = CurrentEvent;
    CurrentEvent = event_ptr;
    
  }

  simtime = TimerEnd(); 
  timer1time = timer1time / 1000000;
  selectedtime = selectedtime / 1000000;
  forktime = (forktime2 - forktime1) / 1000000;

  if (NodeId == 0)  {
    Sys->GetStatistics()->SetSimTime(simtime);
    Sys->GetStatistics()->SetTimer1Time(timer1time);
    Sys->GetStatistics()->SetSelectedTime(selectedtime);
    Sys->GetStatistics()->SetForkTime(forktime);
  } 

// conclusion
  delete MsgQueue;
 
  thistask->resultis(0);
}

int
Node_c::iSuspend(double d)
{
  if (d <= 0) return 0;
  Remain = 0;
  State = 1;
  delay((long)d);
  State = 0;
  if (Remain) return 1;
  else return 0;
}
  
#undef NODE_C
