#define REMOTEACCESS_C
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

void RemoteAccess(Node_c *Node)
{
  FILE *Fpout;
  Crt_Event_t Event,*PrevEvent, *CurrentEvent;
  double ComputationTime;
  Network_c *Network;
  Msg_c *Msg, *Reply;
  int i;

  Fpout = Node->GetFpout();
  CurrentEvent = Node->GetCurrentEvent();
  PrevEvent = Node->GetPrevEvent();
  Network = Node->GetSys()->GetNetwork();
  Event.par = new long int[MAX_PAR];

  ComputationTime = CurrentEvent->ti - PrevEvent->ti;
  Computation(Node, ComputationTime);

  if (!NoOutTrace) {
    Event.ev = PCXX_REMOTE_BEGIN;
    Event.ti = GetCurrentTime();
    Event.nid = CurrentEvent->nid;
    Event.num_par = CurrentEvent->num_par;
    for(i=0;i<Event.num_par;i++)
      Event.par[i] = CurrentEvent->par[i];
    WriteCrtEvent(Fpout, &Event);
  }

  /* ASSUMPTION: # of threads < MAXINT, CurrentEvent->par[3] = to_node_id */
  /* ASSUMPTION: size of elmt access=CurrentEvent->par[2] < MAXINT */ 
  Msg = new Msg_c(REAL, REQUESTMSG, _RequestMsgSize, Node->GetNodeId(), 
                  (int)CurrentEvent->par[3], (int)CurrentEvent->par[2]);
  Network->SendMsg(Msg);
  
  Reply = Node->GetMsg();
  while (Reply->GetMsgType() != REPLYMSG) {
    ProcessMsg(Node,Reply);
    Reply = Node->GetMsg();
  }
  ProcessMsg(Node,Reply);

  if (!NoOutTrace) {
    Event.ev = PCXX_REMOTE_END;
    Event.ti = GetCurrentTime();
    Event.nid = CurrentEvent->nid;
    Event.num_par = CurrentEvent->num_par;
    for(i=0;i<Event.num_par;i++)
      Event.par[i] = CurrentEvent->par[i];
    WriteCrtEvent(Fpout, &Event);
  }

  delete Event.par;
}

#undef REMOTEACCESS_C
