#define SYSTEM_C
#include "Trace.h"
#include "TaskLibIncludes.h"
#include "SimulatorIncludes.h"

static int NumNodes;

double GetCurrentTime()
{
  return (double)(thistask->get_clock());
}

void Suspend(double time)
{
  if (time <= 0) return;
  thistask->delay((long)time);
}

void NumNodesInit(int numnodes)
{
  NumNodes = numnodes;
}

int GetNumNodes()
{
  return NumNodes;
}

FILE *ParFileOpen(char *parfilename)
{
  char filename[80];
  FILE *file;

  sprintf(filename,"%s.par",parfilename);
  if ((file = fopen(filename, "r")) == NULL) {
    printf("Unable to open %s\n",filename);
    exit(1);
  }
  return file;
}

void WriteParFile(FILE *parfile)
{
  fprintf(parfile,"BarrierEntryTime %lf\n",_BarrierEntryTime);
  fprintf(parfile,"BarrierExitTime %lf\n" ,_BarrierExitTime);
  fprintf(parfile,"BarrierCheckTime %lf\n" ,_BarrierCheckTime);
  fprintf(parfile,"BarrierModelTime %lf\n" ,_BarrierModelTime);
  fprintf(parfile,"BarierExitCheckTime %lf\n" ,_BarrierExitCheckTime);
  fprintf(parfile,"BarrierByMsgs %d\n",_BarrierByMsgs);
  fprintf(parfile,"BarrierMsgSize %d\n",_BarrierMsgSize);

  fprintf(parfile,"BytePackTime %lf\n" ,_BytePackTime);
  fprintf(parfile,"PackStartupTime %lf\n" ,_PackStartupTime);
  fprintf(parfile,"ByteToNWTransferTime %lf\n" ,_ByteToNWTransferTime);
  fprintf(parfile,"ToNWStartupTime %lf\n" ,_ToNWStartupTime);
  fprintf(parfile,"AddMsgTime %lf\n" ,_AddMsgTime);
  fprintf(parfile,"GetMsgTime %lf\n" ,_GetMsgTime);
  fprintf(parfile,"AddMsgTryTime %lf\n" ,_AddMsgTryTime);
  fprintf(parfile,"AddMsgStartupTime %lf\n" ,_AddMsgStartupTime);
  fprintf(parfile,"GetMsgTryTime %lf\n" ,_GetMsgTryTime);
  fprintf(parfile,"GetMsgStartupTime %lf\n" ,_GetMsgStartupTime);
  fprintf(parfile,"ByteTransferTime %lf\n" ,_ByteTransferTime);
  fprintf(parfile,"MsgProcessingTime %lf\n" ,_MsgProcessingTime);
  fprintf(parfile,"RequestMsgSize %d\n",_RequestMsgSize);

  fprintf(parfile,"MipsRatio %lf\n" ,_MipsRatio);
  fprintf(parfile,"ProcessMsgType %d\n", _ProcessMsgType);
  fprintf(parfile,"Delta %lf\n" ,_Delta);

  fprintf(parfile,"Topology %d\n",_Topology);
  fprintf(parfile,"CommStartupTime %lf\n" ,_CommStartupTime);

}

void ReadParFile(FILE *parfile)
{
  char parname[80],dummypar[80];
  char Line[256];

  while(fgets(Line, 256, parfile)) {
    if (Line[0] == '#') /* comments, skip */
      continue;
    if (sscanf(Line, "%s", parname) <= 0) /* empty line */
      continue;
    if (!strcmp(parname,"BarrierEntryTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_BarrierEntryTime);
    else if (!strcmp(parname,"BarrierExitTime"))
      sscanf(Line,"%s %lf", dummypar ,&_BarrierExitTime);
    else if (!strcmp(parname,"BarrierCheckTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_BarrierCheckTime);
    else if (!strcmp(parname,"BarrierModelTime"))
      sscanf(Line,"%s %lf", dummypar ,&_BarrierModelTime);
    else if (!strcmp(parname,"BarrierExitCheckTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_BarrierExitCheckTime);
    else if (!strcmp(parname,"BarrierByMsgs"))
      sscanf(Line,"%s %d", dummypar,&_BarrierByMsgs);
    else if (!strcmp(parname,"BarrierMsgSize"))
      sscanf(Line,"%s %d", dummypar,&_BarrierMsgSize);

    else if (!strcmp(parname,"BytePackTime"))
      sscanf(Line,"%s %lf", dummypar ,&_BytePackTime);
    else if (!strcmp(parname,"PackStartupTime"))
      sscanf(Line,"%s %lf", dummypar ,&_PackStartupTime);
    else if (!strcmp(parname,"ByteToNWTransferTime"))
      sscanf(Line,"%s %lf", dummypar ,&_ByteToNWTransferTime);
    else if (!strcmp(parname,"ToNWStartupTime"))
      sscanf(Line,"%s %lf", dummypar ,&_ToNWStartupTime);
    else if (!strcmp(parname,"AddMsgTime"))
      sscanf(Line,"%s %lf", dummypar ,&_AddMsgTime);
    else if (!strcmp(parname,"GetMsgTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_GetMsgTime);
    else if (!strcmp(parname,"AddMsgTryTime"))
      sscanf(Line,"%s %lf", dummypar ,&_AddMsgTryTime);
    else if (!strcmp(parname,"AddMsgStartupTime"))
      sscanf(Line,"%s %lf", dummypar ,&_AddMsgStartupTime);
    else if (!strcmp(parname,"GetMsgTryTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_GetMsgTryTime);
    else if (!strcmp(parname,"GetMsgStartupTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_GetMsgStartupTime);
    else if (!strcmp(parname,"ByteTransferTime"))
      sscanf(Line,"%s %lf", dummypar ,&_ByteTransferTime);
    else if (!strcmp(parname,"MsgProcessingTime")) 
      sscanf(Line,"%s %lf", dummypar ,&_MsgProcessingTime);
    else if (!strcmp(parname,"RequestMsgSize"))
      sscanf(Line,"%s %d", dummypar,&_RequestMsgSize);

    else if (!strcmp(parname,"MipsRatio"))
      sscanf(Line,"%s %lf", dummypar ,&_MipsRatio);
    else if (!strcmp(parname, "ProcessMsgType"))
      sscanf(Line,"%s %d", dummypar, &_ProcessMsgType);
    else if (!strcmp(parname,"Delta")) 
      sscanf(Line,"%s %lf", dummypar ,&_Delta);

    else if (!strcmp(parname,"Topology"))
      sscanf(Line,"%s %d", dummypar,&_Topology);
    else if (!strcmp(parname,"CommStartupTime"))
      sscanf(Line,"%s %lf", dummypar ,&_CommStartupTime);

    else {
      printf("Ignoring parameter %s\n",parname);
    }
  }
}
  
System_c::System_c(int numnodes) 
{
  FILE *parfile;
  NumNodes = numnodes;
  NumNodesInit(numnodes);
  BarrierModelInit(numnodes);
  parfile = ParFileOpen(ParFileName);
  ReadParFile(parfile);
  Nodes = new Node_c*[numnodes];
  Network = new Network_c(this);
  Statistics = new Statistics_c(this, numnodes);
}

System_c::~System_c()
{ 
  delete Network;
  delete Statistics;
}

void System_c::Run()
{ 
  for(int i=0;i<NumNodes;i++) 
    Nodes[i] = new Node_c(this, i);

  for(i=0;i<NumNodes;i++)
    Nodes[i]->result();

  Statistics->PrintAll();
}
#undef SYSTEM_C
