/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdio.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
#include "macro.h"
#include "ext_low.h"
#include "ext_ann.h"

#define MAXSTRING 256
#define ALIGNSTRING "ALIGN"
#define ALIGN_FILE "align.h"
#define ALIGNCLASSNAME "Align"

static int Nb_align_Function_in_Program =0;
extern int Verbose;
extern char *Tool_Unparse2_LLnode();

static int linecount;
               /* Linecount keeps track of the number of lines printed */

/*********************************************************************/
/*********************** Data Alignment Routines *********************/
/*********************************************************************/
/* Uses the parser for annotation to compute the alignement */
/* Take a string and generate an alignement function */
/* assumed shape 
[i,j] -> [i,j] 
[i] -> [0,j] 0 is constante 
[j] -> [*,k]
[i,j] -> [j] is this legal ?
Presently limited to one and two dimensions */

/*********************************************************************/
static int GenerateAlignFunction(str,bif,file)
char *str;
PTR_BFND bif;
FILE *file;
{
  char *name;
  PTR_LLND align;
  PTR_LLND func,first_param, second_param;
  PTR_LLND firstvar, secondvar;
  int nb_dim_first = 0, nb_dim_second = 0;
  int num = -1;
  align =  Parse_Annotation(str,bif);

  if (!align)
    {
      fprintf(stderr,"%s  ",str);
      Message("Incorrect syntax of align function",0);
      return -1;
    }
  
  if (Verbose)
    fprintf(stderr,"\nAlignement found :: %s\n",Unparse_Annotation(align)); 
  
  /* now generate the alignement function */  
  name = Get_Annotation_Field_Label(align);
  if (strcmp(name,ALIGNSTRING)==0)
    {
      func = Get_Annotation_Field(align);
      first_param = Get_Th_Parameter_For_Call(func,0);
      second_param = Get_Th_Parameter_For_Call (func,1);
      if (first_param && (NODE_CODE(first_param) == ARRAY_REF))
        {
          if (second_param && (NODE_CODE(second_param) == ARRAY_REF))
            { /* we are safe here */
              nb_dim_first = exprListLength(NODE_OPERAND0(first_param));
              nb_dim_second = exprListLength(NODE_OPERAND0(second_param));
              if (Verbose)
                fprintf(stderr,"Dimension %d ---> %d\n",nb_dim_first, nb_dim_second );
              if ( nb_dim_first > nb_dim_second)
                {
                  Message("Alignement: No replication handled",0);
                  return -1;
                }
              if ((nb_dim_first == 0) || (nb_dim_second == 0))
                {
                  Message("Alignement: Null dimension ",0);
                  return -1;
                }                       
              /* check the index ....... of the first param */
              if (nb_dim_first == 1)
                {
                  firstvar = NODE_OPERAND0(NODE_OPERAND0(first_param));
                  if (firstvar && (NODE_CODE(firstvar) == VAR_REF))
                    {
                      char res1[MAXSTRING];
                      char res2[MAXSTRING];
                      Reset_Unparser();
                      strcpy(res1,Tool_Unparse2_LLnode(NODE_OPERAND0(NODE_OPERAND0(second_param))));
                      /* there may be another dimension */
                      if (nb_dim_second == 2)
                        {
                          Reset_Unparser();
                          strcpy(res2,Tool_Unparse2_LLnode(NODE_OPERAND0(NODE_OPERAND1(NODE_OPERAND0(second_param)))));
			  linecount += 7;  /* NOTE! # of newlines in format string. */
                          fprintf(file,"case %d:\n{int %s;\n%s = tmp_i;\n *ri1 = %s;\n *ri2 = %s;\n}\nbreak;\n",
                              Nb_align_Function_in_Program,
                              SYMB_IDENT(NODE_SYMB(firstvar)),
                              SYMB_IDENT(NODE_SYMB(firstvar)),
                              res1,res2);
                        } else {
			  linecount += 7;  /* NOTE! # of newlines in format string. */
                          fprintf(file,"case %d:\n{int %s;\n%s = tmp_i;\n *ri1 = %s;\n *ri2 = 0;\n}\nbreak;\n",
                                  Nb_align_Function_in_Program,
                                  SYMB_IDENT(NODE_SYMB(firstvar)),
                                  SYMB_IDENT(NODE_SYMB(firstvar)),
                                  res1);
			}
                      num = Nb_align_Function_in_Program;
                      Nb_align_Function_in_Program++;
                    } else
                     {
                       Message("Alignement: Variables ....... ",0);
                       return -1;
                     }  
                } else
                  {
                    firstvar = NODE_OPERAND0(NODE_OPERAND0(first_param));
                    secondvar = NODE_OPERAND0(NODE_OPERAND1(NODE_OPERAND0(first_param)));
                    if ((firstvar && (NODE_CODE(firstvar) == VAR_REF)) &&
                        (secondvar && (NODE_CODE(secondvar) == VAR_REF)))
                      {
                        char res1[MAXSTRING];
                        char res2[MAXSTRING];
                        Reset_Unparser();
                        strcpy(res1,Tool_Unparse2_LLnode(NODE_OPERAND0(NODE_OPERAND0(second_param))));
                        Reset_Unparser();
                        strcpy(res2,Tool_Unparse2_LLnode(NODE_OPERAND0(NODE_OPERAND1(NODE_OPERAND0(second_param)))));
			linecount += 8; /* NOTE! # of newlines in format string. */
                        fprintf(file,"case %d:\n{int %s,%s;\n%s = tmp_i;\n%s = tmp_j;\n *ri1 = %s;\n *ri2 = %s;\n}\nbreak;\n",
                               Nb_align_Function_in_Program,
                               SYMB_IDENT(NODE_SYMB(firstvar)),
                               SYMB_IDENT(NODE_SYMB(secondvar)),
                                SYMB_IDENT(NODE_SYMB(firstvar)),
                               SYMB_IDENT(NODE_SYMB(secondvar)),
                               res1,res2);
                        num = Nb_align_Function_in_Program;
                        Nb_align_Function_in_Program++;
                      } else
                        {
                          Message("Alignement: Variables ....... ",0);
                          return -1;
                        }  
                  }
            }
          else
            {
              Message("Alignement:Second parameter of Alignement wrong",0);
              return -1;
            }
        } else
          {
            Message("Alignement:First parameter of Alignement wrong",0);
            return -1;
          }

    } else
      {
        Message("Alignement:Align not recognized",0);
        return -1;
      }
  return num;
}

/*********************************************************************/
static char *LookInNodeForAlignDecl(thedec,found)
     PTR_LLND thedec, *found;
{
  PTR_SYMB symb;
  PTR_LLND param;
  PTR_TYPE type;
  char *str;
  str = NULL;

  if (thedec && (NODE_CODE(thedec) == CLASSINIT_OP))
    {
      symb = NULL;
      if (NODE_OPERAND0(thedec) && NODE_SYMB(NODE_OPERAND0(thedec)))
        symb = NODE_SYMB(NODE_OPERAND0(thedec));
      if (symb)
        {
          type = SYMB_TYPE(symb);
          if (type && (TYPE_CODE(type) == T_DERIVED_TYPE))
            symb = TYPE_SYMB_DERIVE(type);
          if (symb && (strcmp(SYMB_IDENT(symb),ALIGNCLASSNAME) == 0))
            {
              /* got one; Look for string in the parameter */
              param = NODE_OPERAND1(thedec);
              while (param)
                {
                              
                  if (NODE_OPERAND0(param) && (NODE_CODE(NODE_OPERAND0(param)) == STRING_VAL))
                    {
                      str = NODE_STR(NODE_OPERAND0(param));
                      *found = NODE_OPERAND0(param);
                      /* printf("found alignement %s \n",str);*/
                    }
                  param = NODE_OPERAND1(param);
                }
            }
        }                      
    }

  return str;
}

/*********************************************************************/
int Look_For_Align_Declaration(file)
  FILE *file;
{
  PTR_BFND bif;
  PTR_LLND pt,pt2,found,assop;
  char *str;
  int num;

  linecount = 0;

  Init_Unparser();
  Nb_align_Function_in_Program =0;

  /* create the file */
  /* (PHB) No need to create it */
  /* (PHB) the output will be put directly into temp.C */
  /*  file = fopen(ALIGN_FILE,"w"); */

  /* Write prefunction preamble */
  linecount += 2;  /* NOTE! # of newlines in the following format strings. */
  fprintf(file,
   "void GenAlignFunc(int x,int tmp_i,int tmp_j,int *ri1,int *ri2) {\n");
  fprintf(file,"   switch(x) {\n");

  for (bif = PROJ_FIRST_BIF (); bif; bif = BIF_NEXT (bif))
    {
      if (BIF_CODE(bif) == VAR_DECL)
	{
          /* now look for Align declaration */
          pt =  BIF_LL1(bif);
          while (pt)
            {
              str = LookInNodeForAlignDecl(NODE_OPERAND0(pt),&found); 
              if (str)
                {
                  num = GenerateAlignFunction(str,bif,file);
                  if (num >= 0)
                    {
                      NODE_STR(found) = NULL;
                      NODE_CODE(found) = INT_VAL;
                      NODE_INT_CST_LOW(found) = num;
                    }
                }
              pt = NODE_OPERAND1(pt);
            }
        }
      if (BIF_CODE(bif) == EXPR_STMT_NODE)
	{ /* new operator may have an alignement */
          pt =  BIF_LL1(bif);
          while (pt)
            {
              if (NODE_OPERAND0(pt) && (NODE_CODE(NODE_OPERAND0(pt)) == ASSGN_OP))
                {
                  assop = NODE_OPERAND1(NODE_OPERAND0(pt));
                  if (assop && (NODE_CODE(assop) == NEW_OP))
                    {
                      pt2 = NODE_OPERAND1(assop);
                      while (pt2)
                        {
                          str = NULL;
                          if (NODE_OPERAND0(pt2) && (NODE_CODE(NODE_OPERAND0(pt2)) == STRING_VAL))
                            {
                              str = NODE_STR(NODE_OPERAND0(pt2));
                              found = NODE_OPERAND0(pt2);
                              /* printf("found alignement %s \n",str);*/
                            }
                          if (str)
                            {
                              num = GenerateAlignFunction(str,bif,file);
                              if (num >= 0)
                                {
                                  NODE_STR(found) = NULL;
                                  NODE_CODE(found) = INT_VAL;
                                  NODE_INT_CST_LOW(found) = num;
                                }
                            }
                          pt2 = NODE_OPERAND1(pt2);
                        }               
                    }
                }
              pt = NODE_OPERAND1(pt);
            }
        }
    } 

  /* Post-amble (end of in-lined alignment function */
  linecount += 7;  /* NOTE! # of newlines in following code */
  fprintf(file,"      default :\n");
  fprintf(file,"        /* identity */ \n");
  fprintf(file,"        *ri1 = tmp_i;\n");
  fprintf(file,"      *ri2 = tmp_j;\n");
  fprintf(file,"      break;\n");
  fprintf(file,"    }\n");
  fprintf(file,"}\n");

  return linecount;
}
