/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// include "macro.h"
// include "vpc.h"
// temporary solution pb with the linker should be h
#include "sage.h"
#include "sage++user.h"
#include "sage++callgraph.h"

static SgStatement       * ResetLocal=NULL;
static SgStatement	 * EnterPar = NULL;
static SgStatement	 * ExitPar = NULL;
static SgSymbol          * SymbResetLocal=NULL;
static SgSymbol          * CheckLocal=NULL;
static SgFunctionCallExp * Local=NULL;
static SgExpression      * FirstLocal=NULL;
static int                 IndexVarInUse = 0;
int  Verbose = 0;

#include "paraSect.h"

extern "C" void initAnnotation();
extern "C" int  Look_For_Align_Declaration(...);
extern "C" int CommentOut; // no comment are output
FILE *finput;
FILE *debugoutfile;
extern int Warning_count;
extern int DEBUG;
extern int CommentOut;

// 960108 / lth@cs.uoregon.edu
// These support the creation of a mapping file during the unparse.

extern "C" void unparser_begin_mapping( char *, char *, int );
extern "C" void unparser_end_mapping( void );
static void 
do_unparse( SgFile *file, FILE *outfile, char *source, int lineno, int map );

char Version[256];
int  VeryVerbose = 0;
int  Brk_special = 0;
int  SetStop = -1;
static int  SimpleCodeGeneration = 0;
int arch_flag = 0;
int spmd_flag = 1; // SPMD = 1; MASTER/SLAVE = 0;

//
// code for adding __private attribute for KSR  (BW)
//
void my_setBaseType(SgType * tp, SgType * base) {

// there exist a method setBaseType but for SgPointerType 
// this one is for SgDEscriptType

      TYPE_BASE(((isSgDescriptType(tp))->thetype)) = base->thetype;
}

int addAttribute(SgSymbol * sym, int attr, SgType * dtype) {
    int varnt, aflag = 0;
    SgType * tp, *oldtp, *newtp;
    SgPointerType *pptr;
    SgReferenceType *rptr;
 
                 varnt=sym->variant();
                 if(varnt == VARIABLE_NAME) {
                    tp = sym->type();

//
//   ******pointers, references, and arrays *******
                    if(tp->variant() == T_POINTER ||
                       tp->variant() == T_REFERENCE ||
                       tp->variant() == T_ARRAY){
                           while(tp->baseType()->variant() == T_POINTER ||
                                 tp->baseType()->variant() == T_REFERENCE ||
                                 tp->baseType()->variant() == T_ARRAY){
                                      tp = tp->baseType();}
                        oldtp = tp;
                        tp = tp->baseType();
                        varnt = tp ->variant();
                        
                        if(varnt == T_DESCRIPT) {
                           isSgDescriptType(tp)->setModifierFlag( 
                                attr | (isSgDescriptType(tp)->modifierFlag()));
                           if(dtype != NULL) 
                                {my_setBaseType(tp, dtype);}
                           }
                        else {
                          newtp = new SgDescriptType(* tp, attr);
                          if(pptr = isSgPointerType(oldtp)) pptr->setBaseType(*newtp);
                          else if(rptr=isSgReferenceType(oldtp)) rptr->setBaseType(*newtp);
                          else isSgArrayType(oldtp)->setBaseType(*newtp);   
                          }                      

                      }

// ******descript types **********
                   else if(tp->variant() == T_DESCRIPT) {
                           isSgDescriptType(tp)->setModifierFlag( 
                                attr | (isSgDescriptType(tp)->modifierFlag()));
                           if(dtype != NULL) 
                                {my_setBaseType(tp, dtype);}
                           }
// *********** others *************
                  else {
                         newtp = new SgDescriptType(* tp, attr);
                         sym->setType(*newtp);   }                      
                aflag = 1;
               }
  
  return aflag;
}

void replace_globals (SgFile * f) {
    int i, is, ii, aflag;
    SgStatement *s = f->firstStatement();
    SgStatement *st, *prev, *stt;
    SgVarDeclStmt *decl;
    SgClassStmt *cls;
    SgSymbol *sym;
    SgVarRefExp *varref;
    SgExpression *texp, *myexp, *q_list, *ap, *bp;
    SgType *dtype = NULL;                        
    int n_gl_st = s->numberOfChildrenList1();

    for (i=0; i < n_gl_st; i++) {
      st = s->childList1(i);

// *********** VarDeclStmt ********
// ********************************

      if(decl = isSgVarDeclStmt(st)) {
             ii = decl->numberOfSymbols();  
             ap = decl->expr(0);
             prev = decl;
             for (is = 0; is < ii; is++){
                 sym = decl->symbol(is);           
                 (void)addAttribute(sym, BIT_KSRPRIVATE, 0);
                 bp = ap->lhs();
                 q_list = new SgExprListExp(*bp);
                 stt = new SgVarDeclStmt(*q_list, *(sym->type()));
                 if(decl->isExtern()) stt->addDeclSpec(BIT_EXTERN);
                 if(decl->isStatic()) stt->addDeclSpec(BIT_STATIC);
                 if(decl->isInline()) stt->addDeclSpec(BIT_INLINE);
                 if(decl->isFriend()) stt->addDeclSpec(BIT_FRIEND);
                 prev->insertStmtAfter(*stt);
                 prev = stt;
                 i++;  n_gl_st++;
                 ap = ap->rhs();
             }
                   decl->deleteStmt();
                   i--;  n_gl_st--;
         }

// *********** ClassStmt **********
// ********************************

     else if(cls = isSgClassStmt(st)) {
            if(texp = cls->expr(0)) {  // have expr list
            prev = cls;
// get rid of the T_DESCRIPT type for class (if any)
       
            if(cls->name()->type() != NULL && 
               cls->name()->type()->variant() == T_DESCRIPT) {
                     cls->name()->setType(*(cls->name()->type()->baseType()));
// adjust variable type, pass new type as arg to addAttribute
                     dtype = new SgDerivedType(*(cls->name())); 
                     }

            while (texp) {
// not all expressions are variables, can have typedefs
// so have to look for variable name
// addAttribute finds variable names, returns a flag
// but still have to find symbols
//
                 bp = texp->lhs();
                 myexp = bp;

// find symbol
                 if(myexp->variant() == ASSGN_OP) 
                   {
                   myexp = myexp->lhs(); 
                   }
               while(myexp->variant() == ADDRESS_OP ||
                     myexp->variant() == DEREF_OP ||
                     myexp->variant() == FUNCTION_OP){ 
                              myexp =  (myexp->operand(0)); 
                     }
               if(varref = isSgVarRefExp(myexp)) {
                  sym = varref->symbol();
                 aflag = addAttribute(sym, BIT_KSRPRIVATE, dtype);
                 if (aflag == 1){
                   q_list = new SgExprListExp(*bp);
                   stt = new SgVarDeclStmt(*q_list, *(sym->type()));
                   prev->insertStmtAfter(*stt, *(cls->controlParent()));
                 if(prev->isExtern()) stt->addDeclSpec(BIT_EXTERN);
                 if(prev->isStatic()) stt->addDeclSpec(BIT_STATIC);
                 if(prev->isInline()) stt->addDeclSpec(BIT_INLINE);
                 if(prev->isFriend()) stt->addDeclSpec(BIT_FRIEND);
                   i++;  n_gl_st++;
                   prev = stt;
                   texp = texp->rhs();
                   }
                   else texp = texp->rhs();
                 }
                 else texp = texp->rhs();
              }
              if(cls->numberOfChildrenList1() == 0) {
                   cls->deleteStmt();
                   i--;  n_gl_st--;
                   }
              else if(aflag == 1) {
                  cls->thebif->entry.Template.ll_ptr1= NULL;
              }
          }
     }


}
}
// end of __private stuff (BW)

//
// control functions for the arguments of the program


// read the parameters value or check if here and return the value

int
isArgumentThere(char * s, int argc, char *argv[])
{
  int i;
  for (i = 1 ; i<argc; i++)
    {
      if (strcmp(argv[i],s) == 0)
        { // we found the parameter 
          return(1);
        }
    }
  return(0);
}

int 
getValueForArgument(char * s, int argc, char *argv[])
{
  int i;

  for (i = 1 ; i<argc; i++)
    {
      if (strcmp(argv[i],s) == 0)
        { // we found the parameter 

          if (i+1<argc)
            return(atoi(argv[i+1]));
          else
            fprintf(stderr,"Argument value missing for %s\n",s);
          return(-1);
        }
    }
  return(-1);
}

char *
getStringForArgument(char * s, int argc, char *argv[])
{
  int i;

  for (i = 1 ; i<argc; i++)
    {
      if (strcmp(argv[i],s) == 0)
        { // we found the parameter 

          if (i+1<argc)
            return argv[i+1];
          else
            fprintf(stderr,"Argument value missing for %s\n",s);
          return(NULL);
        }
    }
  return(NULL);
}

void 
writePreambule()
{
  strcpy(Version,".dep parse tree to C++ prototype version - 12/01/92");
  fprintf(stderr,"%s\n", Version);
  fprintf(stderr,"Copyright Indiana University\n");
}

void 
writeUsage()
{
 fprintf(stderr,"usage:\n");
 fprintf(stderr,"dep2C++ -i file.proj -o file.c\n");
 fprintf(stderr,"dep2C++ -i file.proj -o file.c -version -verbose -debug  -help -simple -unparse -comment -arch -ms -map\n");
}

void 
printHelp()
{
 fprintf(stderr,"help:\n");
 fprintf(stderr,"dep2C++ is a converter from a .dep parse tree to a C++ program.\n");
 fprintf(stderr,"dep2C++ takes as an input a project file containing 1 file.dep.\n");
 fprintf(stderr,"This file is obtained by parsing a pC++ source file with pC++2dep.\n");
 fprintf(stderr,"-version : print the version of the compiler\n");
 fprintf(stderr,"-debug   : output the file debug.dep that contains \n");
 fprintf(stderr,"           the dep file corresponding to the output of the compiler\n");
 fprintf(stderr,"-help    : this output\n");
 fprintf(stderr,"-simple  : simple code generation, does not check for overloading\n");
 fprintf(stderr,"-unparse : only read the dep file and unparse it\n");
 fprintf(stderr,"-unparsecodegen : unparse after code generation and  before instance creation (debug only)\n");
 fprintf(stderr,"-comment : unparse comments in output file\n");
 fprintf(stderr,"-arch    : architecture flag (optional), if -arch ksr generate code for KSR\n");
 fprintf(stderr,"-ms      : generate master/slave (rather than SPMD) code \n");
 fprintf(stderr,"-map     : generate unparse.map file\n" );
 fprintf(stderr,SAGE_INFO);
}

//
// end for utilitary, we start here the pC++ to C++ translator 
//

#define MAXSTRING 256
#define MAX_COLLECTION 1000
#define MAX_METHOD 1000
#define MAXINDEXVAR  10

// table for the compiler 

static SgType          * Tab_Collection_Type[MAX_COLLECTION];
static int               Nb_Elem_in_Tab_Collection = 0;

static SgType          * Tab_Collection_Created[MAX_COLLECTION];
static SgStatement     * Tab_Collection_Created_Stmt[MAX_COLLECTION];
static int               Nb_Elem_in_Tab_Collection_Created = 0;

static SgStatement     * Tab_Remove[MAX_COLLECTION];
static int               Nb_Elem_in_Tab_Remove = 0;
static SgStatement       * barrier=NULL;
static SgStatement       * poll=NULL;
SgSymbol                 * pcxx_inpar_symb;
static SgStatement       * NextInSubset = NULL;
static SgStatement       * FirstInSubset = NULL;
static SgDerivedCollectionType  * Tab_Collection_Derived_Type[MAX_COLLECTION];
static int               Nb_Elem_in_Tab_Collection_Derived_Type = 0;
static SgFunctionSymb 	*Sympcxx_NodeNumber = NULL;
static SgFunctionSymb 	*SymBcast = NULL;
static SgSymbol         *brk_funcId, *brk_fileName, *brk_lineNumber;

//
// record collection type derived, dedicated to 
// take the collection and the basic type
// return the type, and create it if not 
// already done
//

SgDerivedCollectionType *
typeCollectionDerived(SgSymbol * symb, SgType * type)
{
  int i;
  SgDerivedCollectionType *cur;
  
  if (!symb || !type)
    {
      Message("typeCollectionDerived should not have null parameter(s)",0);
      return NULL;
    }

  for (i =0 ; i <Nb_Elem_in_Tab_Collection_Derived_Type ; i++)
    {
      cur = Tab_Collection_Derived_Type[i];
      if ( (cur->collectionName() == symb) && (type->equivalentToType(*cur->elementClass())))
        return cur;
    }
  // I have to create the type;
  cur = new SgDerivedCollectionType(*symb,*type);
  if (Nb_Elem_in_Tab_Collection_Derived_Type >= MAX_COLLECTION)
    {
      Message("Table Overflow",0);
      exit(1);
    }
  Tab_Collection_Derived_Type[Nb_Elem_in_Tab_Collection_Derived_Type] = cur;
  Nb_Elem_in_Tab_Collection_Derived_Type++;
  return cur;
}



 //
 // record what has been done
 //

 void
 setTabCollectionCreated(SgType *t, SgStatement *s)
 {
   if (!t || !s)
     return ;
   if (Nb_Elem_in_Tab_Collection_Created >= MAX_COLLECTION)
     {
       Message("Table Overflow",0);
       exit(1);
     }
   Tab_Collection_Created[Nb_Elem_in_Tab_Collection_Created] = t;
   Tab_Collection_Created_Stmt[Nb_Elem_in_Tab_Collection_Created] = s;
   Nb_Elem_in_Tab_Collection_Created++;
 }

 SgStatement *
 checkTabCollectionCreated(SgType *type)
 {
   int i;
   if (!type)
     return NULL;
   for (i= 0; i< Nb_Elem_in_Tab_Collection_Created; i++)
     {
       if (type->equivalentToType(*Tab_Collection_Created[i]))
         {
           return Tab_Collection_Created_Stmt[i];
         }
     }
   return NULL;
 }


 //
 // Initiliaze the compiler, records collection and create basic statement for code generation
 // also record the basic type of derived collections Collection<Atype>
 // also record class declaration 
 //

 void 
 initializeCompiler(SgFile *file)
 {
   SgStatement      *stmt, *first;
   SgSymbol         *symb;
   SgExpression     *expr, *expr1, *expr0;
   SgMemberFuncSymb *mf;
   char strin[256];

   if (!file)
     return;

   // init the global variable
   Nb_Elem_in_Tab_Collection = 0;
   Nb_Elem_in_Tab_Collection_Created = 0;
   Nb_Elem_in_Tab_Remove = 0;
   brk_funcId=brk_fileName=brk_lineNumber=NULL;
   barrier=NULL;
   poll=NULL;
   EnterPar = NULL;
   ExitPar = NULL;
   ResetLocal=NULL;
   CheckLocal=NULL;
   SymbResetLocal=NULL;
   Local=NULL;
   FirstLocal=NULL;


   initAnnotation();
   
   for (stmt = file->firstStatement(); stmt ; stmt = stmt->lexNext())
     {
       if (isSgCollectionStmt(stmt))
         {
 // the function in Tab_Remove will be removed later
 // record them when it is still easy

           Tab_Remove[Nb_Elem_in_Tab_Remove] = stmt;
           Nb_Elem_in_Tab_Remove++;
         }
 //      if (isSgClassStmt(stmt))
 //         {
 //           Tab_Class[Nb_Elem_in_Tab_Class] = stmt;
 //           if (Nb_Elem_in_Tab_Remove >= MAX_COLLECTION)
 //             {
 //               Message("Table Overflow",0);
 //               exit(1);
 //             }
 //           Nb_Elem_in_Tab_Class++;
 //         }
     }
 // now look for symbol that are a member function of a collection

   for (symb = file->firstSymbol(); symb; symb = symb->next())
     {
       if (mf = isSgMemberFuncSymb(symb))
         {
           if (mf->className() && ((mf->className())->variant() == COLLECTION_NAME))
             {
               if (mf->body())
                 {
                   Tab_Remove[Nb_Elem_in_Tab_Remove] = mf->body();
                   if (Nb_Elem_in_Tab_Remove >= MAX_COLLECTION)
                     {
                       Message("Table Overflow",0);
                       exit(1);
                     }
                   Nb_Elem_in_Tab_Remove++;
                 }
             }
         }
     }
 // now create the basic statement needed for code generation 

   first = file->firstStatement();

   symb = new SgFunctionSymb(FUNCTION_NAME,"pcxx_Barrier",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*symb);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   barrier= new SgCExpStmt(*expr1);

   symb = new SgFunctionSymb(FUNCTION_NAME,"pcxx_Poll",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*symb);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   poll = new SgCExpStmt(*expr1);

   symb = new SgFunctionSymb(FUNCTION_NAME,"NextInSubset",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*symb);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   NextInSubset = new SgCExpStmt(*expr1);

   symb = new SgFunctionSymb(FUNCTION_NAME,"FirstInSubset",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*symb);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   FirstInSubset = new SgCExpStmt(*expr1);

   CheckLocal = new SgFunctionSymb(FUNCTION_NAME,"CheckLocal",*SgTypeInt(),*first);

   SymbResetLocal = new SgFunctionSymb(FUNCTION_NAME,"ResetLocal",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*SymbResetLocal);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   ResetLocal = new SgCExpStmt(*expr1);

  Sympcxx_NodeNumber = new SgFunctionSymb(FUNCTION_NAME,"pcxx_NodeNumber",
			     *SgTypeInt(),*first);
  SymBcast = new SgFunctionSymb(FUNCTION_NAME,"pcxx_ScanBcast",
			     *SgTypeInt(),*first);

/* delete this to make local induction vars.
   for (i=0; i< MAXINDEXVAR; i++)
     {
       sprintf(strin,"VarpC_%d",i);
       IndexVar[i] = new SgVariableSymb(strin);
       IndexVar[i]->declareTheSymbol(*first);
       IndexVarExpr[i] = new SgVarRefExp(*IndexVar[i]);
     }
*/ 

   // the pcxx_inpar variable need not be declared because it is done
   // by a direct text insert.  however, we need a reference to a symbol  
   sprintf(strin,"pcxx_inpar");
   pcxx_inpar_symb = new SgVariableSymb(strin);
   // symb->declareTheSymbol(*first);
   expr = new SgVarRefExp(*pcxx_inpar_symb);
   expr0 = new SgValueExp(0);
   // SgStatement *newdecl = first->lexNext();
   // if(!(mydecl = isSgVarDeclStmt(newdecl))) Message("compiler error pcxx_inpar", 0);
   // else if(mydecl->symbol(0) != symb) Message("compiler error: wrong pcxx_inpar symbol",0);
   // else mydecl->setInitialValue(0, expr0->copy());
   
   expr1 = &(SgPostPlusPlusOp(expr->copy()));
   expr0  = new SgExpression(EXPR_LIST);
   expr0 ->setLhs(*expr1);
   EnterPar = new SgCExpStmt(*expr0);
   expr1 = &(SgPostMinusMinusOp(expr->copy()));
   expr0  = new SgExpression(EXPR_LIST);
   expr0 ->setLhs(*expr1);
   ExitPar = new SgCExpStmt(*expr0);

   symb = new SgFunctionSymb(FUNCTION_NAME,"Local",*SgTypeInt(),*first);
   Local = new SgFunctionCallExp(*symb);

   symb = new SgFunctionSymb(FUNCTION_NAME,"FirstLocal",*SgTypeInt(),*first);
   FirstLocal = new SgFunctionCallExp(*symb);

   if(!spmd_flag)
     initParallelSections(file);

   if(Brk_special){
     brk_funcId     = new SgVariableSymb("brk_funcId");
     brk_fileName   = new SgVariableSymb("brk_fileName");
     brk_lineNumber = new SgVariableSymb("brk_lineNumber");
   }
}

//
// given a pointer to a function declaration and a name as a string
// this will find the symbol if it is declared in the scope of the
// function.
//
SgSymbol * findNameInScope(SgStatement * function, char * name){
     SgSymbol *s;
     SgFuncHedrStmt *f;
     if(!function) return NULL;
     if(!name) return NULL;
     if(f = isSgFuncHedrStmt(function)){
		s = &(f->name());
		for(; s; s = s->next()){
			if((strcmp(s->identifier(), name) == 0) &&
			    (s->scope() == function))
			        return s;
			}
		return NULL;
	}
      else{  printf("function arg in findNameInScope not a function");
             fflush(stdout);
	     return NULL;
	}
}

//
// give a statement, this finds the enclosing function if there is one.
// note it does not find inclosing structure or class or template definitions.
//

SgStatement *findEnclosingFunction(SgStatement *s){
     while(s && (s->variant() != GLOBAL) && (s->variant() != FUNC_HEDR) &&
	   (s->variant() != PROC_HEDR) && (s->variant() != PROG_HEDR))
		s = s->controlParent();
     if(s && s->variant() != GLOBAL) return s;
     else return NULL;
}

//
// this function will declare an integer induction variable initialized to 0
// with the name pcxx_i  in the same function as statement where
//
SgExpression *makeLocalInductionVarRef(int i, SgStatement *where){
       char strin[32];
       SgSymbol *s;
       SgStatement *fun;
       SgExpression *e, *expr0;
       SgVarDeclStmt *mydecl;
       
       if(!where){
		Message("no reference statement in makeLocalInductionVarRef\n",0);
		return NULL;
		}
       if((fun = findEnclosingFunction(where)) == NULL){
		Message("could not find enclosing function in makeLocalInductionVarRef\n", 0);
		return NULL;
		}
       sprintf(strin,"pcxx_%d",i);
       if((s = findNameInScope(fun, strin)) == NULL){
       		s  = new SgVariableSymb(strin);
       		s->declareTheSymbol(*fun);
   		expr0 = new SgValueExp(0);
   		SgStatement *newdecl = fun->lexNext();
   		if(!(mydecl = isSgVarDeclStmt(newdecl))) 
			Message("error in initializer for makeLocalInductionVarRef\n", 0);
   		else mydecl->setInitialValue(0, *expr0);
		}
       e = new SgVarRefExp(*s);
       return e;
}



void
addToTabCollectionType(SgDerivedCollectionType *ct)
{
  int i, created;

  if (ct)
    {
      if (Nb_Elem_in_Tab_Collection >= MAX_COLLECTION)
        {
          Message("Table Overflow",0);
          exit(1);
        }
      if (ct->elementClass() && !((ct->elementClass())->isTheElementType()))
        {
          // look at if already created
            created = 0;
          for (i= 0; i< Nb_Elem_in_Tab_Collection; i++)
            {
              if (ct->equivalentToType(*Tab_Collection_Type[i]))
                {
                  created = 1;
                  break;
                }
            }
          if (!created)
            {
              Tab_Collection_Type[Nb_Elem_in_Tab_Collection] = ct;
              Nb_Elem_in_Tab_Collection++;
            }
        }
    }
}

//
// Look for Collection derived Type we have to treat
//
void
  recordCollectionDerivedType(SgFile *file)
{
   SgType *type;
   SgDerivedCollectionType *ct;
   for (type = file->firstType(); type ; type = type->next())
     {       
       if (ct = isSgDerivedCollectionType(type))
         {
           addToTabCollectionType(ct);
         }
     }
 }

//
// Look for the call to the constructor name
// change the call to be on the new superclass
//

void
updateCallToSuperClassConstructor(SgStatement *s,SgSymbol *oldinhs, SgSymbol *newinhs)
{
  SgStatement *last, *pt,  *body;
  SgSymbol *symb;
  SgVarDeclStmt *stmt;
  SgMemberFuncSymb *func;
  int i, nbsymb;
  
  if (!s || !oldinhs || !newinhs)
    return;

  s->replaceSymbBySymb(*oldinhs, *newinhs);
  last =  s->lastNodeOfStmt();
  for (pt = s; pt; pt = pt->lexNext())
    {
      if (stmt = isSgVarDeclStmt(pt))
        {
          nbsymb = stmt->numberOfSymbols();
          for (i = 0 ; i < nbsymb; i++)
            {
              symb = stmt->symbol(i);
              if (symb)
                {
                  if (func = isSgMemberFuncSymb(symb))
                    {                      
                      if (body = func->body())
                        body->replaceSymbBySymbSameName(*oldinhs, *newinhs);
                      // if body not there this is probably a virtual function 
                    }
                } else
                  Message("internal error in updateCallToSuperClassConstructor-2",0);
            }
        } 
      if (pt == last)
        break;
    }
}


 // 
 // create instance of the collection we need according to the element type declare
 //

 SgStatement *
 makeACollectionInstance(SgDerivedCollectionType *dt)
 {
   SgStatement *s, *inhs;
   SgClassStmt *cl;
   SgSymbol *symb;
   SgDerivedCollectionType *inht;
   int nb;
   if (!dt)
     return NULL;

   if (s=checkTabCollectionCreated(dt))
     return s;
   s = dt->createCollectionWithElemType();  /* <--- DEBUG THIS */
   setTabCollectionCreated(dt,s);
 // get the name and look at if inherit 
   if (cl = isSgClassStmt(s))
     {
       if (nb = cl->numberOfSuperClasses())
         {
           if (nb != 1)
             {
               Message("Sorry no mulitple inheritance allowed",0);
               exit(1);
             }
           symb = cl->superClass(0);
           if (!symb || !dt->elementClass())
             {
               Message("Internale error in makeACollectionInstance-1",0);
               return NULL;
             }
           // we have to do that only if of type collection
             if (symb->variant() == COLLECTION_NAME)
               {
                 // we have to create the instance for the inherited class
                 // first we create the type we want 
                 // that type is what we need
                 inht = new SgDerivedCollectionType(*symb, *(dt->elementClass()));
                 inhs = makeACollectionInstance(inht);
                 // now we have to update the inheritage of the current collection 
                 if (!inhs || !s->symbol())
                   {
                     Message("Internale error in makeACollectionInstance-2",0);
                     return NULL;
                   }
                 cl->setSuperClass(0, *inhs->symbol(), BIT_PUBLIC);
                 // we also have to update the constructor call to the superclass;
                 updateCallToSuperClassConstructor(s, symb, inhs->symbol());
               }
         }
     } else
       Message("Internale error in makeACollectionInstance",0);
   return s;
 }

 void
 createCollectionInstance()
 {
   int i;
   SgDerivedCollectionType *dt;

   for (i= 0; i< Nb_Elem_in_Tab_Collection; i++)
     {
       if (dt = isSgDerivedCollectionType(Tab_Collection_Type[i]))
         {
           makeACollectionInstance(dt);
         }
     }
 }

 void
 removeCollections()
 {
   int i;
   for (i = 0; i < Nb_Elem_in_Tab_Remove; i++)
     {
       Tab_Remove[i]->extractStmt();
     }
 }

 //
 // For all type we have been creating replace by the right one
 //
 updateTypeInTheProgram(SgFile *file)
 {
   int i;
   SgType *newtype;

   if (!file)
     return NULL;
   for (i= 0; i< Nb_Elem_in_Tab_Collection_Created; i++)
     {
       if (Tab_Collection_Created_Stmt[i] && 
           Tab_Collection_Created_Stmt[i]->symbol())
         {
           newtype = new SgDerivedType(*(Tab_Collection_Created_Stmt[i]->symbol()));
           (file->firstStatement())->replaceTypeInStmt(*Tab_Collection_Created[i],*newtype);
         } else
           Message("Internal Error in updateTypeInTheProgram",0);
     }
   return NULL;
 }

 //
 // function to generate the call to the runtime system
 // WARNING That function has to enhance to get the element class declared
 //

 SgSymbol *
 isReferenceToMethodOfElement(SgExpression *expr)
 {
   SgVarRefExp *varef;
   SgFunctionCallExp *fcall;
   SgSymbol *symb=NULL;
   SgMemberFuncSymb *symbm;
   SgFieldSymb *symbf;
   if (!expr)
     return NULL;
   if (varef = isSgVarRefExp(expr))
     {
       symb = varef->symbol();
     }
   if (fcall = isSgFunctionCallExp(expr))
     {
       symb = fcall->symbol();
     }
   if (symbm = isSgMemberFuncSymb(symb))
     {
       if (symbm->isMethodOfElement())
         return symbm;
       else
         return NULL;
     }
   if (symbf = isSgFieldSymb(symb))
     {
       if (symbf->isMethodOfElement())
         return symbf;
       else
         return NULL;
     }
   return 0;
 }

SgSymbol *
isReferenceToClassOfElement(SgDerivedCollectionType *type,SgExpression *expr)
{
  SgVarRefExp *varef;
  SgFunctionCallExp *fcall;
  SgSymbol *symb,*elsymb;
  SgMemberFuncSymb *symbm;
  SgFieldSymb *symbf;
  SgType *elmetype;
  SgDerivedType *eltype;

   if (!expr || !type)
     return NULL;
  elmetype = type->elementClass();
  if (!elmetype)
    return NULL;
  if (elmetype->isTheElementType())
    return NULL;
  
  if (eltype = isSgDerivedType(elmetype))
    {
      if (!(elsymb = eltype->typeName()))
        {
          return NULL;
        }
    } else
      return NULL;

   if (varef = isSgVarRefExp(expr))
     {
       symb = varef->symbol();
     }
   if (fcall = isSgFunctionCallExp(expr))
     {
       symb = fcall->symbol();
     }

   if (symbm = isSgMemberFuncSymb(symb))
     {
       if (symbm->className() == elsymb)
         return symbm;
       else
         return NULL;
     }
   if (symbf = isSgFieldSymb(symb))
     {
       if (symbf->structureName() == elsymb)
         return symbf;
       else
         return NULL;
     }
   return NULL;
}


int
inMethodOfTheElement(SgStatement *where)
{
  SgFuncHedrStmt *fh;
  SgMemberFuncSymb *symb;

  if (!where)
    return 0;

  if (where->variant() == GLOBAL)
    return 0;

  if (fh = isSgFuncHedrStmt(where))
    {
      if (symb = isSgMemberFuncSymb(fh->symbol()))
        {
          return symb->isMethodOfElement();
        }
    }
  return inMethodOfTheElement(where->controlParent());
}


SgFuncHedrStmt *
whichFunctionAmI(SgStatement *where)
{
  SgFuncHedrStmt *fh;

  if (!where)
    return NULL;

  if (where->variant() == GLOBAL)
    return NULL;

  if (fh = isSgFuncHedrStmt(where))
    {
      return fh;
    }
  return whichFunctionAmI(where->controlParent());
}


//
// where is used to disambiguate this
//
SgDerivedCollectionType *
isReferenceToCollection(SgExpression *expr, SgStatement *where)
 {
   SgType *type = NULL;
   SgExpression *lhs;
   SgVarRefExp *varef;
   SgPointerType *tpt;
   SgReferenceType *tref;
   SgDerivedCollectionType *dstype;
   SgArrayRefExp *array;
   SgSymbol *symb;
   SgFunctionCallExp *func;

   if (!expr)
     return NULL;
   switch (expr->variant())
     {
     case RECORD_REF:
     case POINTST_OP:
     case DEREF_OP:
       lhs = expr->lhs();
       break;
     case VAR_REF:
     case ARRAY_REF:
     case FUNC_CALL:
       lhs = expr;
       break;
       default :
// default should be to try to generate a copy for the expression
// and see if expression becomes call the function again.
         return NULL;
     }
   if (!lhs)
     return NULL;
   if (varef = isSgVarRefExp(lhs))
     {
       if (varef->symbol())
         type = varef->symbol()->type();
     } else
       if (isSgThisExp(lhs))
         {
           if (inMethodOfTheElement(where))
             return 0; // this refer to the element 
           type = lhs->type();
         } else
           if (array = isSgArrayRefExp(lhs))
             {
               symb = array->symbol();
               if (symb)
                 type = symb->type();
               if (type)
                 type = type->internalBaseType();
             } else  
               if (func = isSgFunctionCallExp(lhs))
                 {
                   symb = func->symbol();
                   if (symb)
                     type = symb->type();
                 } else
                   if  (isSgPointerDerefExp(lhs))
                     {
                       type = isReferenceToCollection(lhs,where);
                     } else
                       if  (isSgExprListExp(lhs))
                         {
                           type = isReferenceToCollection(lhs->lhs(),where);
                         } else
                           return NULL;
   if (!type)
     return NULL;
   switch (expr->variant())
     {
     case FUNC_CALL:
     case ARRAY_REF:
     case VAR_REF:
     case RECORD_REF:
       if (dstype = isSgDerivedCollectionType(type))
         return dstype;
       if (tref = isSgReferenceType(type))
         {
           type = tref->baseType();
           if (dstype = isSgDerivedCollectionType(type))
             return dstype;
         }
       break;
     case DEREF_OP:
     case POINTST_OP: 
       if (tpt = isSgPointerType(type))
         {
           type = tpt->baseType();
           if (dstype = isSgDerivedCollectionType(type))
             return dstype;
           if (tref = isSgReferenceType(type))
             {
               type = tref->baseType();
               if (dstype = isSgDerivedCollectionType(type))
                 return dstype;
             }
         }
       break;
     }
   return NULL;
 }


SgDerivedCollectionType *
isReferenceToSubset(SgExpression *expr, SgStatement *where, SgSubscriptExp ** resddot,  SgExpression ** ar)
 {
   SgType *type = NULL;
   SgExpression *lhs;
   SgVarRefExp *varef;
   SgPointerType *tpt;
   SgPntrArrRefExp *array;
   SgArrayRefExp   *arrayref;
   SgSubscriptExp *range;
   SgExprListExp  *expl;
   SgDerivedCollectionType *res1;

   if (!expr)
     return 0;
   if (expr->variant() == RECORD_REF)
     lhs = expr->lhs();
   else
     return 0;

   if (!lhs)
     return 0;

   if (array = isSgPntrArrRefExp(lhs))
     {
       lhs = array->lhs();
       if (expl = isSgExprListExp(array->rhs()))
         {
           if (range = isSgSubscriptExp(expl->elem(0)))
             {
               *resddot = range;
               *ar = lhs;
             } else
               return 0;
           if (expl->length() != 1)
             {
               Message("Only one dimension is accepted for subset operation",where->lineNumber());
               return 0;
             }
           } else
             return 0;
       if (varef = isSgVarRefExp(lhs))
         {
           if (varef->symbol())
             type = varef->symbol()->type();
         } else
           if (isSgThisExp(lhs))
             {
               if (inMethodOfTheElement(where))
                 return 0; // this refer to the element 
                   type = lhs->type();
             } else
               return 0;
     } else
       if (arrayref = isSgArrayRefExp(lhs))
         {
           lhs = arrayref->lhs();
           if (expl = isSgExprListExp(lhs))
             {
               if (expl->length() != 1)
                 {
                   Message("Only one dimension is accepted for subset operation",where->lineNumber());
                   return 0;
                 }
               if (range = isSgSubscriptExp(expl->elem(0)))
                 {
                   *resddot = range;
                 } else
                   return 0;
               if (expl->length() != 1)
                 {
                   Message("Only one dimension is accepted for subset operation",where->lineNumber());
                   return 0;
                 }
             } else
               return 0;
           if (arrayref->symbol())
             {
               type = arrayref->symbol()->type();
               *ar = new SgVarRefExp(*arrayref->symbol());
             }
         } else
           return 0;
   if (!type)
     return 0;
   switch (expr->variant())
     { // in that case the type must be a pointer type (may change with the new parser);
     case RECORD_REF:
       if (tpt = isSgPointerType(type))
         {
           type = tpt->baseType();
           if (res1 = isSgDerivedCollectionType(type))
             return res1;
         }
       break;
     }
   return 0;
 }

//
// Basic functions for check hazard
//
  int
isInIfStmt(SgStatement *where)
  {
    if (!where)
      return FALSE;
    if (where->variant() == GLOBAL)
      return FALSE;
    if (isSgIfStmt(where))
      return TRUE;    
    return isInIfStmt(where->controlParent());
  }

// return in which function I am
#if 0
SgStatement *
whereAmI(SgStatement *where)
  {
    // later

  }


//
// The next function check for hazard in a parallel statement
// for instance if a parallel activation may be in an hazardous place
// i.e. where it is not set the same way by all the processor
//
//
//

 checkHazard(SgStatement *where)
   {
     // later;
     // check if call in a if statement;
     // buffer problem;
   }
#endif

//
// Look at Collection derived type and check against the element type
//

int
checkForDeridedCollectionType(SgType *type1, SgType *type2)
  {
    SgDerivedCollectionType *td1, *td2;
    SgType                  *et1, *et2;

    if (!type1 || !type2 )
      return 0;    
    if (!(td1 = isSgDerivedCollectionType(type1)))
      return 0;
    if (!(td2 = isSgDerivedCollectionType(type2)))
      return 0;
    
    if (td2->collectionName() != td1->collectionName())
      return 0;
    
    if (!(et1 = td1->elementClass()))
      return 0;
    if (!(et2 = td2->elementClass()))
      return 0;

    if (!et1->isTheElementType() && !et2->isTheElementType())
       return 0;
    
    // for now element type match any type;
    return TRUE;
  }

//
// used to compute if overloading applies
//

SgType *
compareForOverloading_1(SgMemberFuncSymb *func,SgType *trhs,SgType *tlhs)
{
  SgSymbol *param;
  SgType   *type,*type1;
  SgReferenceType *reftype;

  if (!func|| !trhs || !tlhs)
    return NULL;

  if ( func->numberOfParameters() != 1)
    return NULL;

  if (param = func->parameter(0))
    {
      if (type = param->type())
        {// a special rule is needed her for dealing with element type...;
         if (!type->equivalentToType(*tlhs) && !checkForDeridedCollectionType(tlhs,type))
           {
             if (reftype = isSgReferenceType(type))
               {
                 type = reftype->baseType();
                 if (!tlhs->equivalentToType(*type))
                   {
                     if (!checkForDeridedCollectionType(tlhs,type))                     
                       return NULL;
                   }
               } else
                 if (reftype = isSgReferenceType(tlhs))
                   {
                     type1 = reftype->baseType();
                     if (!type->equivalentToType(*type1))
                       {
                         if (!checkForDeridedCollectionType(type,type1))                     
                           return NULL;
                       }
                   } else
                     return NULL;
           } 
       }
    }

  if (reftype = isSgReferenceType(func->type()))
    return reftype->baseType();
  else
    return func->type();
}

//
// look in the body of a class to find an overloading function
//

SgType *
lookForOverloading_1(int code, SgStatement *tolook,SgType * trhs,SgType * tlhs)
{
  SgStatement *last, *pt;
  SgSymbol *symb;
  SgVarDeclStmt *stmt;
  SgMemberFuncSymb *func;
  SgFuncHedrStmt *funh;
  SgClassStmt    *cl;
  int i, nbsymb;
  SgType *valret;
  
  if ( !tolook || !trhs || !tlhs)
    return NULL;

  last =  tolook->lastNodeOfStmt();
  for (pt = tolook; pt; pt = pt->lexNext())
    {
      valret = NULL;
      if (stmt = isSgVarDeclStmt(pt))
        {
          nbsymb = stmt->numberOfSymbols();
          for (i = 0 ; (i < nbsymb) && !valret; i++)
            {
              symb = stmt->symbol(i);
              if (symb)
                {
                  if (func = isSgMemberFuncSymb(symb))
                    {                      
                      valret =  compareForOverloading_1(func,trhs,tlhs);
                    }
                }
            }
        } else
          if (funh = isSgFuncHedrStmt(pt))
            {
              symb = funh->symbol();
              if (symb)
                {
                  if (func = isSgMemberFuncSymb(symb))
                    {                      
                      valret =  compareForOverloading_1(func,trhs,tlhs);
                    }
                } 
            } 
      if (valret)
        { // check for the name;
          switch (code)
            {
            case MOD_OP:
              if (!strcmp(symb->identifier(),"%"))
                return valret;
              break;
            case OR_OP:
              if (!strcmp(symb->identifier(),"|"))
                return valret;
              break;
            case AND_OP:
              if (!strcmp(symb->identifier(),"&"))
                return valret;
              break;
            case SUBT_OP:
              if (!strcmp(symb->identifier(),"-"))
                return valret;
              break;
            case NOT_OP :
              if (!strcmp(symb->identifier(),"!"))
                return valret;
              break;
            case PLUSPLUS_OP:
              if (!strcmp(symb->identifier(),"++"))
                return valret;
              break;
            case MINUSMINUS_OP:
              if (!strcmp(symb->identifier(),"--"))
                return valret;
              break;
            case PLUS_ASSGN_OP:
              if (!strcmp(symb->identifier(),"+="))
                return valret;
              break;
            case MINUS_ASSGN_OP:	
              if (!strcmp(symb->identifier(),"-="))
                return valret;
              break;
            case AND_ASSGN_OP	:
              if (!strcmp(symb->identifier(),"&="))
                return valret;
              break;
            case IOR_ASSGN_OP	:	
              if (!strcmp(symb->identifier(),"??"))
                return valret;
              break;
            case MULT_ASSGN_OP:		
              if (!strcmp(symb->identifier(),"*="))
                return valret;
              break;
            case DIV_ASSGN_OP	:	
              if (!strcmp(symb->identifier(),"/="))
                return valret;
              break;
            case MOD_ASSGN_OP	:	
              if (!strcmp(symb->identifier(),"%="))
                return valret;
              break;
            case XOR_ASSGN_OP	:	
              if (!strcmp(symb->identifier(),"??"))
                return valret;
              break;
            case LSHIFT_ASSGN_OP:
              if (!strcmp(symb->identifier(),"<<="))
                return valret;
              break;
            case RSHIFT_ASSGN_OP:
              if (!strcmp(symb->identifier(),">>="))
                return valret;
              break;
            case LSHIFT_OP:
              if (!strcmp(symb->identifier(),"<<"))
                return valret;
              break;
            case RSHIFT_OP:
              if (!strcmp(symb->identifier(),">>"))
                return valret;
              break;
            case INTEGER_DIV_OP:
              if (!strcmp(symb->identifier(),"??"))
                return valret;
              break;
            case ADD_OP:
              if (!strcmp(symb->identifier(),"+"))
                return valret;
              break;
            case MULT_OP:
              if (!strcmp(symb->identifier(),"*"))
                return valret;
              break;
            case ASSGN_OP:
              if (!strcmp(symb->identifier(),"="))
                return valret;
              break;
            }
        }
      if (pt == last)
        break;
    }
  // if the class inherit, then look at the inherited class;
  if (cl = isSgClassStmt(tolook))
    {
      for (i=0; i < cl->numberOfSuperClasses(); i++)
        {
          valret = lookForOverloading_1(code, cl->superClass(i)->body(),trhs,tlhs);
          if (valret)
            return valret;
        }
    }
  return NULL;
}

//
// Compute the type of the expression for collection 
// also using the overloading the type of an expression
// 
//

SgType *
computeTypeOfExpression(SgExpression *expr, SgStatement *where)
  {
    SgDerivedCollectionType *ret1;
    SgType *type1,*type2, *type3, *type;
    SgSymbol *symb;
    SgFieldSymb *field;
    SgMemberFuncSymb *mf;

    if (!expr || !where)
      return NULL;
    switch(expr->variant())
      {
      case EXPR_LIST:
        return computeTypeOfExpression(expr->lhs(), where);
      case THIS_NODE:
        type = expr->type();
        if (!type)
          return NULL;
        if (inMethodOfTheElement(where))
          {
            if (ret1 = isSgDerivedCollectionType(type->baseType()))
              return ret1->elementClass();
          } 
        return type;
//        break;        
        // list of operators we considered for overloading ;
        // more to come;
      case MOD_OP:
      case OR_OP:
      case AND_OP:
      case SUBT_OP:
      case NOT_OP :
      case PLUSPLUS_OP:
      case MINUSMINUS_OP:
      case PLUS_ASSGN_OP:
      case MINUS_ASSGN_OP:		
      case AND_ASSGN_OP	:	
      case IOR_ASSGN_OP	:	
      case MULT_ASSGN_OP:		
      case DIV_ASSGN_OP	:	
      case MOD_ASSGN_OP	:	
      case XOR_ASSGN_OP	:	
      case LSHIFT_ASSGN_OP:
      case RSHIFT_ASSGN_OP:
      case LSHIFT_OP:
      case RSHIFT_OP:
      case INTEGER_DIV_OP:
      case MULT_OP:
      case ADD_OP:
      case ASSGN_OP: // that part is incorrect for unary operator; to be modified
        type2 = computeTypeOfExpression(expr->rhs(), where);
        type1 = computeTypeOfExpression(expr->lhs(), where);
        if (type1 && type2)
          {
            if (ret1 = isSgDerivedCollectionType(type1))
              { // overloading found
                type3 = lookForOverloading_1(expr->variant(),ret1->collectionName()->body(),type1,type2);
                if (type3)
                  return type3;
                else
                  return type1;
              }
            else
              return type1;
          } else
            return type1;
        //break;
      case DEREF_OP:
      case FUNC_CALL:
      case ARRAY_REF:
      case VAR_REF:
        if (ret1 = isReferenceToCollection(expr,where))
          {
            return ret1;
          } else
            if (symb = isReferenceToMethodOfElement(expr))
              { // symb must be from a collection; 
               if (!inMethodOfTheElement(where))
                 {
                   // we have to find to which collection it is coming from so we;
                   // so we can return the write type;
                   if (symb->type())
                     {
                       if (field = isSgFieldSymb(symb))
                         {
                           return typeCollectionDerived(field->structureName(), symb->type());
                         } else
                           if (mf = isSgMemberFuncSymb(symb))
                             {
                               return typeCollectionDerived(mf->className(), symb->type());
                             } 
                     } 
                 }
              }
        return expr->type();
//        break;
     case RECORD_REF: // this will have to change later to generalize thing (recursively compute the type of the expression)
        if (ret1 = isReferenceToCollection(expr,where)) 
          {
            if (symb = isReferenceToMethodOfElement(expr->rhs()))
              {
                // need to create the appropriate type if not already created; 
                return typeCollectionDerived(ret1->collectionName(), symb->type());
              } else
                { // look at if the symb is a field of the element class;
                  if (symb = isReferenceToClassOfElement(ret1,expr->rhs()))
                    {
                      // need to create the appropriate type if not already created; 
                      return typeCollectionDerived(ret1->collectionName(), symb->type());
                    }
                }
          }
        return expr->type();
//        break;
      case POINTST_OP:
        if (ret1 = isReferenceToCollection(expr,where))
          {
            if (symb = isReferenceToMethodOfElement(expr->rhs()))
              {
                // need to create the appropriate type if not created;
                return typeCollectionDerived(ret1->collectionName(), symb->type());
              } else
                {
                  if (symb = isReferenceToClassOfElement(ret1,expr->rhs()))
                    {
                      // need to create the appropriate type if not already created; 
                      return typeCollectionDerived(ret1->collectionName(), symb->type());
                    }
                }
          }
        return expr->type();
//        break;
        default :
          return expr->type();
//        break;
      }
  }

//
// look if addr needed
//

int
LookIfAddrNeeded(SgExpression *coll, SgStatement *where)
{
  SgType                  *type;
  if (!coll)
    return 0;    

  type = computeTypeOfExpression(coll,where);
  if (!type)
    {
      Message("internal error  in LookIfAddrNeeded",0);
      return 0;
    }
  
  if (isSgDerivedCollectionType(type))
    return TRUE;
  else
    return 0;
}

int
findMoECallinExpr(SgExpression *e)
{
  SgFunctionCallExp *f;
  int r;

  if(!e) return -1;
  if(f = isSgFunctionCallExp(e)){
    SgSymbol *s = f->funName();
    SgMemberFuncSymb* m;

    if ( (m = isSgMemberFuncSymb(s)) && m->isMethodOfElement() )
      return s->id();
    return -1;
  }

  if ( (r = findMoECallinExpr(e->lhs())) != -1 ) return r;
  if ( (r = findMoECallinExpr(e->rhs())) != -1 ) return r;
  return -1;
}

int
findMoECallinStmt(SgStatement *stmt)
{
  int r;
  if ( (r = findMoECallinExpr (stmt->expr(0))) != -1 ) return r;
  if ( (r = findMoECallinExpr (stmt->expr(1))) != -1 ) return r;
  if ( (r = findMoECallinExpr (stmt->expr(2))) != -1 ) return r;
  return -1;
}

//
// generate the loop around a statement
//
SgStatement *
generateCodeForStmt(SgStatement *stmt1, SgStatement *savenext, SgStatement  *previous1, char found, SgExpression *coll)
{
  SgStatement       *stmt, *cp, *copy, *tokeep, *previous;
  SgStatement       *brk_stmt1, *brk_stmt2, *brk_stmt3;
  SgStatement       *ite, *where;
  SgForStmt         *forloop;
  SgWhileStmt       *whileloop;
  SgExpression      *IndexVarExpr;
  SgExprListExp     *nlist;
  SgFunctionCallExp *fcall;
  SgValueExp        *con0;
  char              addrneeded;
  int               funcId;

  if (!stmt1 || !coll || !previous1) 
    return stmt1;

  if(!spmd_flag)
    return MSgenerateCodeForStmt(stmt1, savenext, previous1, found, coll);

  addrneeded = LookIfAddrNeeded(coll,stmt1);
  previous = previous1;
  stmt = stmt1;
  brk_stmt1 = NULL;
  if(Brk_special && (funcId=findMoECallinStmt(stmt1))){
    SgExpression *expr, *expr0, *expr1;

    expr  = new SgVarRefExp(*brk_funcId);
    expr0 = new SgValueExp(funcId);
    expr1 = new SgExpression(ASSGN_OP);
    expr1->setLhs(*expr);
    expr1->setRhs(*expr0);
    expr  = new SgExprListExp();
    expr->setLhs(*expr1);
    brk_stmt1 = new SgCExpStmt(*expr);

    expr  = new SgVarRefExp(*brk_fileName);
    expr0 = new SgValueExp(stmt1->fileName());
    expr1 = new SgExpression(ASSGN_OP);
    expr1->setLhs(*expr);
    expr1->setRhs(*expr0);
    expr  = new SgExprListExp();
    expr->setLhs(*expr1);
    brk_stmt2 = new SgCExpStmt(*expr);

    expr  = new SgVarRefExp(*brk_lineNumber);
    expr0 = new SgValueExp(stmt1->lineNumber());
    expr1 = new SgExpression(ASSGN_OP);
    expr1->setLhs(*expr);
    expr1->setRhs(*expr0);
    expr  = new SgExprListExp();
    expr->setLhs(*expr1);
    brk_stmt3 = new SgCExpStmt(*expr);
  }
  if (found == 1)
    { // we have to generate the loop;
      if ( Verbose )
        {
          fprintf(stderr,"parallelism activation line %d\n",stmt->lineNumber());
        }
      cp = stmt->controlParent();
      copy = &(ResetLocal->copy());
      fcall = new SgFunctionCallExp(*SymbResetLocal);
      if (addrneeded)
        {
          if (isSgPointerDerefExp(coll))
            {
              if (coll->lhs())
                fcall->addArg(coll->lhs()->copy());
            } else
              fcall->addArg(SgAddrOp(coll->copy()));
        }
      else
        fcall->addArg(coll->copy());
      nlist = new SgExprListExp(*fcall);
      copy->setExpression(0,*nlist);
      previous->insertStmtAfter(*copy,*cp);
      previous = copy;
      copy = &(EnterPar->copy());
      previous->insertStmtAfter(*copy,*cp);
      previous = copy;                   
      // at that point we have create a stmt ResetLocal(collection);
      // now we have to create the loop;
      tokeep = stmt->extractStmt();
      con0 =  new SgValueExp(0);
      /*  this version will put the induction variables declared
          directly into the for header as for(int pcxx_i = 0; ... )
      IndexVarExpr = makeLocalInductionVarRef(0, previous);
      SgExprListExp *fori = new SgExprListExp(SgAssignOp(*IndexVarExpr,*FirstLocal));
      SgExpression *fordecl = new SgExpression(FORDECL_OP);
      fordecl->setLhs(*fori);
      forloop = new SgForStmt(*fordecl,
                              *IndexVarExpr >= *con0,
                              SgAssignOp(*IndexVarExpr,*Local), 
                              *tokeep);
      */ 
      IndexVarExpr = makeLocalInductionVarRef(0, previous);
      SgFunctionCallExp *newLocal = isSgFunctionCallExp(&(Local->copy()));
      newLocal->addArg(IndexVarExpr->copy());
      forloop = new SgForStmt(SgAssignOp(*IndexVarExpr,*FirstLocal),
                              *IndexVarExpr >= *con0,
                              SgAssignOp(*IndexVarExpr, *newLocal), 
                              *tokeep);
      previous->insertStmtAfter(*forloop,*cp);
      forloop->insertStmtAfter(poll->copy(),*forloop);
      stmt = forloop->lastNodeOfStmt();
      if (brk_stmt1){
        stmt->insertStmtAfter(*brk_stmt1,*cp);
        brk_stmt1->insertStmtAfter(*brk_stmt2,*cp);
        brk_stmt2->insertStmtAfter(*brk_stmt3,*cp);
        stmt = brk_stmt3;
      }
      copy = &(barrier->copy());
      stmt->insertStmtAfter(*copy,*cp);
      stmt = copy;
      copy = &(ExitPar->copy());
      stmt->insertStmtAfter(*copy,*cp);
      stmt = copy;
    } else
      if (found == 2)
        { // we have to generate the loop;
          // side effect have inserted statment after the one we are treating; 
          // those statements must be in the loop
            if ( Verbose )
              {
                fprintf(stderr,"Subset operation found line %d\n",stmt->lineNumber());
              }
          // reset the previous it may have change;
          previous = stmt->nodeBefore();
          IndexVarExpr = makeLocalInductionVarRef(IndexVarInUse, stmt);
          cp = stmt->controlParent();
          fcall = new SgFunctionCallExp(*CheckLocal);
          fcall->addArg(*coll);
          fcall->addArg(*IndexVarExpr);
          // now we have to create the loop;
          tokeep = stmt->extractStmt();
          tokeep = new SgIfStmt(*fcall,*tokeep);
          con0 =  new SgValueExp(0);
          whileloop = new SgWhileStmt(*IndexVarExpr >= *con0,*tokeep);
          copy = &(EnterPar->copy());
          previous->insertStmtAfter(*copy,*cp);
          previous = copy;                   
          previous->insertStmtAfter(*whileloop,*cp);
          where = tokeep->lastNodeOfStmt();
          tokeep = whileloop->lastNodeOfStmt(); // take the last node of while
          tokeep = tokeep->lexNext();// take the next so it points to the first to include in the loop
          for (ite = tokeep; ite && (ite != savenext); )
            {
              tokeep = ite->lexNext();
              where->insertStmtAfter(*ite->extractStmt(),*whileloop);
             ite = tokeep;
            }
          whileloop->insertStmtAfter(poll->copy(),*whileloop);
          stmt = whileloop->lastNodeOfStmt();
          if (brk_stmt1){
            stmt->insertStmtAfter(*brk_stmt1,*cp);
            brk_stmt1->insertStmtAfter(*brk_stmt2,*cp);
            brk_stmt2->insertStmtAfter(*brk_stmt3,*cp);
            stmt = brk_stmt3;
          }
          copy = &(barrier->copy());
          stmt->insertStmtAfter(*copy,*cp);
          stmt = copy;                   
          copy = &(ExitPar->copy());
          stmt->insertStmtAfter(*copy,*cp);
          stmt = copy;                   
        }
  return stmt;
}

//
// give if it is a terminal expression for which code must be generated
//

  int
 isTerminalExpr(SgExpression *expr, SgStatement *where)
 {
   SgSubscriptExp  *resddot; 
   SgExpression    *array;
   SgSymbol        *symb;
   SgType          *type;
   SgDerivedCollectionType *ret1;
   if (!expr || !where)
     return FALSE;
    
   switch(expr->variant())
     {
       // set of comparaison operation; ignore, overloading gives the meaning to that;
       // we also ignore operation on integer; the operator have to be overloaded;
       // also other op that have nothing to do with collection (avoid it);
     case EQ_OP:
     case LT_OP:
     case GT_OP:
     case NOTEQL_OP:	
     case LTEQL_OP:
     case GTEQL_OP:
     case NOT_OP:
     case BITAND_OP:
     case BITOR_OP:
     case DDOT:
     case RANGE_OP:
     case UPPER_OP:
     case LOWER_OP:
     case INT_VAL:
     case FLOAT_VAL:
     case DOUBLE_VAL:
     case BOOL_VAL:
     case CHAR_VAL:
     case STRING_VAL:
     case KEYWORD_VAL:
     case LONG_DOUBLE_VAL:
     case LONG_INT_VAL:
     case UNSIGNED_INT_VAL:
     case UNSIGNED_LONG_INT_VAL:
     case COMPLEX_VAL:
     case CONST_REF:
     case ENUM_REF:
     case LABEL_REF:
     case VAR_LIST:
     case RANGE_LIST:
     case CASE_CHOICE:
     case DEF_CHOICE:
     case VARIANT_CHOICE:
     case CONCAT_OP:
     case SIZE_OP:
     case UNARY_ADD_OP:
     case CLASSINIT_OP:
     case CAST_OP:
     case COPY_NODE:
     case INIT_LIST:
     case VECTOR_CONST:
     case NEW_OP:
     case ARITH_ASSGN_OP:
     case BIT_NUMBER:
     case DELETE_OP:
     case NAMELIST_LIST:
     case SCOPE_OP:
     case FUNCTION_OP:
     case ARRAY_OP:
       return FALSE;
//       break;
     case FUNC_CALL:
       return 6;
     case THIS_NODE:
       type = expr->type();
       if (!type)
         return FALSE;
       type = type->baseType();
       if (!type)
         return FALSE;
       if (inMethodOfTheElement(where))
         return FALSE;
       if (isSgDerivedCollectionType(type))
         return 1;
       break;
     case DEREF_OP:
     case ARRAY_REF:
     case VAR_REF:
       if (isReferenceToCollection(expr,where))
         {
           return 1;
         } else
           if (symb = isReferenceToMethodOfElement(expr))
             { // symb must be from a collection; 
               if (!inMethodOfTheElement(where))
                 {
                   if (symb->type())
                     {
                       return 2;
                     }
                 }
             }else
               if (expr->variant() == DEREF_OP)
                 {
                   if (isTerminalExpr(expr->lhs(), where) > 2)
                     return 5;
                 }
       break;
     case RECORD_REF:
         if (ret1 = isReferenceToSubset(expr,where,&resddot,&array))
           {
             if (isReferenceToMethodOfElement(expr->rhs()))
               return 3;
             if (isReferenceToClassOfElement(ret1,expr->rhs()))
               return 3;
            } else
              {
                ret1 = isReferenceToCollection(expr,where);
                symb =  isReferenceToMethodOfElement(expr->rhs());
                if (ret1 && !symb)
                  symb = isReferenceToClassOfElement(ret1,expr->rhs());
                if (ret1 && symb)
                  {
                    return 3;
                  } else
                    { 
                      if (isTerminalExpr(expr->lhs(), where) > 2)
                        return 5;
                    } 
              } 
       break;
    case POINTST_OP:
       ret1 = isReferenceToCollection(expr,where);
       symb =  isReferenceToMethodOfElement(expr->rhs());
       if (ret1 && !symb)
         symb = isReferenceToClassOfElement(ret1,expr->rhs());
       if (ret1 && symb)
         {
           return 3;
         } else
           { 
             if (isTerminalExpr(expr->lhs(), where) > 2)
               return 5;
           } 
      break;
       default :
         return 4;
    }
   return FALSE;
}

//
// check if a collection derived type is a good one 
//

int
checkElementCollectionType(SgType *type1)
{
  SgDerivedCollectionType *td1;
  SgType                  *et1;
  SgDerivedType           *eltype;
  SgSymbol                *elsymb;
  SgReferenceType         *reftype;
  SgDescriptType          *dsctype;
  int change;
  if (!type1)
    return 0;    
  if (!(td1 = isSgDerivedCollectionType(type1)))
    return 0;
    
  if (!td1->collectionName())
    return 0;
    
  if (!(et1 = td1->elementClass()))
    return 0;
  
  change = 1;
  while (change)
    {
      change = 0;
      if (reftype = isSgReferenceType(et1))
        {
          et1 = reftype->baseType();
          change = 1;
        }
      if (dsctype = isSgDescriptType(et1))
        {
          et1 = dsctype ->baseType();
          change = 1;
        }
    }

  if (et1->isTheElementType())
    return 1;

  // et1 should a class type;
  if (!(eltype = isSgDerivedType(et1)))
    return 0;
  if (!(elsymb = eltype->typeName()))
     return 0;

  if (elsymb->variant() != CLASS_NAME)
    return 0;

    return 1;
}


 //
 // modify a reference so it has the right call to the collection element
 // A.method or A->method of element must be converted in 
 // Forall element i that I own A(i)->method1
 // also treat subset operation A[2:3].method
 // return NULL if nothing has been done 
 //if (copieneeded) the results is ask to be put in a copy

 SgExpression *
 modifyCollectionExpression(SgExpression *expr, char &found, SgExpression **coll, SgStatement *where, int recursif)
 {
   SgExpression    * resul = expr;
   SgExpression    *IndexVarExpr;
   SgFuncPntrExp   * fop;
   SgSubscriptExp  *resddot; 
   SgExpression    *array;
   SgStatement     *stmt;
   SgFunctionCallExp *fcal;
   SgThisExp       *implthis;
   SgSymbol        *symb;
   SgDerivedCollectionType *ret1;
   SgValueExp      *constante;
   int valret;
   if (!expr || !where)
     return expr;
    
   switch(expr->variant())
     {
     case DDOT:
     case PROC_CALL:
     case SIZE_OP:
     case EXPR_IF:
     case CLASSINIT_OP:
     case CAST_OP:
     case VECTOR_CONST:
     case NEW_OP:
     case DELETE_OP:
     case INIT_LIST:
     case BIT_NUMBER:
     case NAMELIST_LIST:
     case SCOPE_OP:
     case FUNCTION_REF:
     case FUNCTION_OP:
     case ARRAY_OP:
       return expr;
//       break;
     case DEREF_OP:
     case ARRAY_REF:
     case VAR_REF:
       if (isReferenceToCollection(expr,where))
         {
           found = 1;
           *coll = expr;
	   IndexVarExpr = makeLocalInductionVarRef(0, where);
           fop = new SgFuncPntrExp(*expr);
           fop->addArg(*IndexVarExpr);
           resul = &(SgDerefOp(*fop));
         } else
           if (symb = isReferenceToMethodOfElement(expr))
             { // symb must be from a collection; 
               if (!inMethodOfTheElement(where))
                 {
                   // in that case we have to create an impicit this;
                   if (symb->type())
                     {
                       found = 1;
                       implthis = new SgThisExp(*symb->type());
                       *coll = implthis;
	   	       IndexVarExpr = makeLocalInductionVarRef(0, where);
                       fop = new SgFuncPntrExp(*(implthis));
                       fop->addArg(*IndexVarExpr);
                       resul = &(SgPointStOp(*fop, *expr));
                     } else
                       Message("modifyCollectionExpression",0);
                 }
             } else
               if (expr->variant() == DEREF_OP)
                 {
                   valret = isTerminalExpr(expr->lhs(), where);
                   if (valret > 2)
                     expr->setLhs(*modifyCollectionExpression(expr->lhs(),found,coll,where,recursif));
                 }
       break;
     case RECORD_REF:
           if ((ret1 = isReferenceToSubset(expr,where,&resddot,&array)) &&
               (isReferenceToMethodOfElement(expr->rhs()) || 
                isReferenceToClassOfElement(ret1,expr->rhs())))
             {// we have a reference to a do subset to treat ;
              
              found = 2;
              // we have to generate the code for setting properly access to the;
              IndexVarInUse++;
              *coll = array;
	      IndexVarExpr = makeLocalInductionVarRef(IndexVarInUse, where);
              fop = new SgFuncPntrExp(SgDerefOp(*array));
              fop->addArg(*IndexVarExpr);
              resul = &(SgPointStOp(*fop, *expr->rhs()));
              // now we have to generate initialisation of the subset operation;
              stmt = &(FirstInSubset->copy());
              // get the parameters;
              if (fcal = isSgFunctionCallExp(stmt->expr(0)->lhs()))
                {// no copy needed here
                   fcal->addArg(*array);
                 if (resddot->lbound())
                   fcal->addArg(*resddot->lbound());
                 else
                   {
                     constante = new SgValueExp(-1);
                     fcal->addArg(*constante);
                   }
                 if (resddot->lbound())
                   fcal->addArg(*resddot->ubound());
                 else
                   {
                     constante = new SgValueExp(-1);
                     fcal->addArg(*constante);
                   }
                 fcal->addArg(*resddot->step());
                 where->insertStmtBefore(*stmt, *where->controlParent());
	      	 IndexVarExpr = makeLocalInductionVarRef(IndexVarInUse, where);
                 // make it an assignement to the index var; 
                 stmt->expr(0)->setLhs(SgAssignOp(*IndexVarExpr, *fcal));
               } else
                  Message("internal error in modifyCollectionExpression",0);
              // same here;
              stmt = &(NextInSubset->copy());
              if (fcal = isSgFunctionCallExp(stmt->expr(0)->lhs()))
                {// no copy needed here;
                 fcal->addArg(*array);
                 if (resddot->lbound())
                   fcal->addArg(*resddot->lbound());
                 else
                   {
                     constante = new SgValueExp(-1);
                     fcal->addArg(*constante);
                   }
                 if (resddot->lbound())
                   fcal->addArg(*resddot->ubound());
                 else
                   {
                     constante = new SgValueExp(-1);
                     fcal->addArg(*constante);
                   }
                 fcal->addArg(*resddot->step());
	      	 IndexVarExpr = makeLocalInductionVarRef(IndexVarInUse, where);
                 fcal->addArg(*IndexVarExpr);
                 where->insertStmtAfter(*stmt, *where->controlParent());
                 stmt->expr(0)->setLhs(SgAssignOp(IndexVarExpr->copy(), *fcal));
               } else
                 Message("internal error in modifyCollectionExpression",0); 
            }
           else
             {
               ret1 = isReferenceToCollection(expr,where);
               symb =  isReferenceToMethodOfElement(expr->rhs());
               if (ret1 && !symb)
                 symb = isReferenceToClassOfElement(ret1,expr->rhs());
             if (ret1 && symb)
               {// we have a reference to treat
                  found = 1;
                *coll = expr->lhs();
	      	 IndexVarExpr = makeLocalInductionVarRef(0, where);
                fop = new SgFuncPntrExp(*(expr->lhs()));
                fop->addArg(*IndexVarExpr);
                resul = &(SgPointStOp(*fop, *expr->rhs()));
              }  else
                { // may have to look recursively
                  valret = isTerminalExpr(expr->lhs(), where);
                  if (valret > 2)
                    expr->setLhs(*modifyCollectionExpression(expr->lhs(),found,coll,where,recursif));
                }
           }
       break;
     case POINTST_OP:
       ret1 = isReferenceToCollection(expr,where);
       symb =  isReferenceToMethodOfElement(expr->rhs());
       if (ret1 && !symb)
         symb = isReferenceToClassOfElement(ret1,expr->rhs());
       if (ret1 && symb)
        {// we have a reference to treat
         found = 1;
         *coll = expr->lhs();
	 IndexVarExpr = makeLocalInductionVarRef(0, where);
         fop = new SgFuncPntrExp(SgDerefOp(*(expr->lhs())));
         fop->addArg(*IndexVarExpr);
         resul = &(SgPointStOp(*fop, *expr->rhs()));
        } else
          { // may have to look recursively
              valret = isTerminalExpr(expr->lhs(), where);
            if (valret > 2)
              expr->setLhs(*modifyCollectionExpression(expr->lhs(),found,coll,where,recursif));
          }
      break;
       default :
         if (recursif)
           {
             if (expr->rhs())
               expr->setRhs(*modifyCollectionExpression(expr->rhs(),found,coll,where,recursif));
             if (expr->lhs())
               expr->setLhs(*modifyCollectionExpression(expr->lhs(),found,coll,where,recursif));
           }
       break;
    }
   return resul;
}



int 
  isThereSubsetOp(SgExpression *expr, SgStatement *where)
{
  int resul = 0;
  SgSubscriptExp  *resddot; 
  SgExpression    *array;
  SgDerivedCollectionType *ret1;

  if (!expr || !where)
    return 0;
    
  switch(expr->variant())
    {
     case RECORD_REF:
      if ((ret1 = isReferenceToSubset(expr,where,&resddot,&array)) &&
          (isReferenceToClassOfElement(ret1,expr->rhs()) ||
            isReferenceToMethodOfElement(expr->rhs())))
        return 1;
       break;
       default :
         if (expr->rhs())
           resul = isThereSubsetOp(expr->rhs(),where);
       if (expr->lhs())
         resul = resul | isThereSubsetOp(expr->lhs(),where);
       break;
     }
   return resul;
}

// forward definition 
SgExpression *generateCopyStmt(SgExpression *expr, SgStatement *where, SgExpression *coll, SgDerivedCollectionType *res, SgStatement  **stcopy);

 SgExpression *
 modifyReferenceToCollection(SgExpression *expr, SgStatement *where, char &found,  SgExpression **coll, int copieneeded)
 {
   SgExpression    * resul = expr;
   SgExpression    * copres ,*texp;
   SgType          *type, *type1, *type2;
   SgDerivedCollectionType *ret1, *ret2;
   SgStatement  *stcopy, *savenext, *previous;
   int valret;
   SgReferenceType *reftype;
   char savefound;
//SimpleCodeGeneration to be used later

   if (!expr || !where)
     return expr;

   if (SetStop == where->lineNumber())
     Message("You reached the point you would like to stop, Master",where->lineNumber());

   // first check for subset;
   if (isThereSubsetOp(expr, where))
       {
         return modifyCollectionExpression(expr, found, coll, where,1);
       }

   // which kind of expression do we have;
   valret = isTerminalExpr(expr, where);
   // in not valret, dont touch to it;
   if (!valret)
     return expr; 
   if (valret == 1)
     {
       // do not modify the expression return it, it will be parameter;
       if (copieneeded && !SimpleCodeGeneration)
         return expr; 
       else
         {
           return modifyCollectionExpression(expr, found, coll, where,0);
         }
     } else
       if ((valret == 2) ||(valret == 3) || (valret == 5))
           {
             if (!copieneeded || SimpleCodeGeneration)
               {
                 return modifyCollectionExpression(expr, found, coll, where,1);
               }
             type = computeTypeOfExpression(expr, where);
             if (!checkElementCollectionType(type))
               {
                 Message("Collection with element not a class cannot be created",where->lineNumber());
                 return modifyCollectionExpression(expr, found, coll, where,1);
               }
             // we have to generate a copy;
             // find out what is the collection here
             if (valret == 2)
               { // this is an implicit if
                   
               } else
                 {
                   *coll = expr->lhs();
                 }
             if (reftype = isSgReferenceType(type))
               type = reftype->baseType();
             if (!(ret1 = isSgDerivedCollectionType(type)))
               return expr;
  
             stcopy = NULL;
             resul = generateCopyStmt(expr, where, *coll, ret1, &stcopy);
             if (stcopy)
               {
                 previous = stcopy->nodeBefore();
                 savenext = stcopy->lexNext();
               } else
                 Message("internal error",0);   
             found = 1;
             Message("Generating copies (may affect performances)",where->lineNumber());
             // now modify the expressions;
             if (stcopy)
               {                 
                 modifyCollectionExpression(stcopy->expr(0), found, coll, where,1);
                 generateCodeForStmt(stcopy, savenext,previous,found,resul);
                 found = 0;
                 if (resul)
                   resul = &(SgDerefOp(*resul));
               }
           } else
             if (valret == 4)
               {
                 // the expression may be subject to overloading;
                 type1 = computeTypeOfExpression(expr->lhs(), where);
                 if (reftype = isSgReferenceType(type1))
                   type1 = reftype->baseType();
                 if (!(ret1 = isSgDerivedCollectionType(type1)))
                   return expr;
                 type2 = computeTypeOfExpression(expr->rhs(), where);

                 // look at unary operator like ADDRESS_OP;
                 switch (expr->variant())
                   {
                   case ADDRESS_OP:  
                     if (!SimpleCodeGeneration)
                       {
                         // here we have an expression list; 
                         savefound = found;
                         if (expr->lhs() && isSgExprListExp(expr->lhs()) && expr->lhs()->lhs())
                           expr->lhs()->setLhs(*modifyReferenceToCollection(expr->lhs()->lhs(), where,found,coll,1));
                         else
                           if (expr->lhs())
                             expr->setLhs(*modifyReferenceToCollection(expr->lhs(), where,found,coll,1));
                         found = savefound;
                       }
                     return expr;
//                     break;
                   }                   
                 // check if there is an overload operation on Collection concerning;
                 // the current type;
                 if (lookForOverloading_1(expr->variant(),ret1->collectionName()->body(),type1,type2)  && !SimpleCodeGeneration)
                   {
                     if (VeryVerbose)
                       {
                         fprintf(stderr,"Overloading found line %d\n",where->lineNumber());
                         fprintf(stderr,"SURGEON GENERAL'S WARNING: Overloading causes memory consumption\n");
                         fprintf(stderr,"and performance degradation, it also makes program difficult to read.\n");
                         fprintf(stderr,"Do you really need it?\n");
                       }
                     savefound = found;
                     if (expr->lhs())
                       expr->setLhs(*modifyReferenceToCollection(expr->lhs(), where,found,coll,1));
                     if (expr->rhs())
                       expr->setRhs(*modifyReferenceToCollection(expr->rhs(), where,found,coll,1));
                     // reset found, that statement does not need the loop around;
                     found = savefound;
                     return expr;
                   } else
                     { 
                       // no overloading, generate the code for the expression in an elementwise;
                      if (copieneeded && !SimpleCodeGeneration )
                        { // compute the type first;
                          type = computeTypeOfExpression(expr, where);
                          if (!checkElementCollectionType(type))
                            {
                              Message("Collection with element not a class cannot be created\n",where->lineNumber());
                              Message("Code generation will not be correct\n",where->lineNumber());
                              if (VeryVerbose)
                                {
                                  fprintf(stderr,"Only a class can be the element of a collection\n");
                                  fprintf(stderr,"collection of int, float, ... are not yet implemented\n");
                                  fprintf(stderr,"We can implement it but we wait for HPF, so you can\n");
                                  fprintf(stderr,"also wait a long time\n");
                                }
                              copieneeded = 0;
                            }
                          if (reftype = isSgReferenceType(type))
                            type = reftype->baseType();
                          if (!(ret1 = isSgDerivedCollectionType(type)))
                            return expr;
                        }
                       *coll = NULL;
                       if (expr->rhs())
                         expr->setRhs(*modifyReferenceToCollection(expr->rhs(), where,found,coll,0));
                       if (expr->lhs())
                         expr->setLhs(*modifyReferenceToCollection(expr->lhs(), where,found,coll,0));

                       if (copieneeded && *coll  && !SimpleCodeGeneration)
                         { // something has been done, we need to make that a copy;
                           stcopy = NULL;
                           copres = NULL;                           
                           resul = generateCopyStmt(expr, where, *coll, ret1, &stcopy);                           
                           if (resul)
                             copres = &(resul->copy());  
                           if (texp = stcopy->expr(0)) // do not forget to generate the copy
                             {
                               if (texp->lhs() && (texp->lhs()->lhs()))
                                 texp->lhs()->setLhs(*modifyCollectionExpression(texp->lhs()->lhs(), 
                                                                          found,  coll, where,1));
                             }
                           resul = copres;
                           if (stcopy)
                             {
                               previous = stcopy->nodeBefore();
                               savenext = stcopy->lexNext();
                             } else
                               Message("internal error",0);   
                           found = 1;
                           Message("Generating copies (may affect performances)",where->lineNumber());
                           // now modify the expressions;
                           if (stcopy)
                             {
                               generateCodeForStmt(stcopy, savenext,previous,found,resul);
                             }
                           if (resul)
                             return &(SgDerefOp(*resul));
                           else 
                             return NULL;
                         }
                       return expr;
                     }
               } else
                 if (valret == 6)
                   { 
                     SgExpression   *param, *onepar;
                     SgExprListExp  *paramlist;
                     SgFunctionSymb *thefunc;
                     SgSymbol       *decparam;
                     int   j, flag;
                     // here we treat a function call;
                     // look at parameters to check if they are of type element or of type collection;
                     param = expr->lhs();
                     if (thefunc = isSgFunctionSymb(expr->symbol()))
                       {
                         // check first what is the type of the return value;
                         type1 = thefunc->type();
                         if (reftype = isSgReferenceType(type1))
                           type1 = reftype->baseType();
                         ret1 = isSgDerivedCollectionType(type1); 
                         if (ret1)
                           {
                             // that function return a collection, so this may be incompatible;
                             flag = 1;
                           } else
                             flag = 0;
                         j = 0;
                         for (paramlist = isSgExprListExp(param); paramlist ; paramlist = paramlist->next()) 
                           {
                             onepar = paramlist->value();
                             decparam = thefunc->parameter(j);
                             if (onepar && decparam)
                               {
                                 type1 = computeTypeOfExpression(onepar, where);
                                 type2 = decparam->type();
                                 if (reftype = isSgReferenceType(type1))
                                   type1 = reftype->baseType();
                                 ret1 = isSgDerivedCollectionType(type1);                                      
                                 if (reftype = isSgReferenceType(type2))
                                   type2 = reftype->baseType();
                                 ret1 = isSgDerivedCollectionType(type1);
                                 ret2 = isSgDerivedCollectionType(type2);
                                 // if !ret1 && !ret2 nothing to do;
                                 // 
                                   // if ret1 && ret2 the expression may need copie generation
                                     
                                     if (ret1 && ret2)
                                       {
                                         savefound = found;
                                         paramlist->setValue(*modifyReferenceToCollection(onepar,where,found,coll,1));
                                         found = savefound;
                                       } else
                                         if (ret1 && !ret2)
                                           {
                                             // here we have a collection as a parameter and we expect probably;
                                             // an element type; we have to generate the code;
                                             if (flag)
                                               {
                                               Message("You may have a parallel operation conflict\n (I have not made my mind on what means what you wrote... to see later)",where->lineNumber());
                                               break;
                                             }
                                             else
                                               paramlist->setValue(*modifyCollectionExpression(onepar, found, coll, where,1));
                                           } else
                                             if (!ret1 && ret2)
                                               {
                                                 Message("Collection type error in parameter list",where->lineNumber());
                                               }
                               }
                             j++;
                           }
                       }
                   }
   return resul;
}


//
// goes over the program and generate the appropriate code for the call to
// the element of the collection
// basically put a loop around the reference if found 
// we inspect mainly EXPR_STMT_NODE
//

void
treatCalltoMethodOfTheElement(SgFile *file)
{
  SgStatement       *stmt, *previous;
  SgStatement       *savenext;
  SgExpression      *expr, *coll;
  SgExprListExp     *list;
  char              found;
  if (!file)
    return;

   for (stmt = file->firstStatement(); stmt ; stmt = stmt->lexNext())
     {
       IndexVarInUse = 0;
       savenext = stmt->lexNext(); // use because we want to keep track of next that may be modified;
       if (isSgCExpStmt(stmt) || isSgReturnStmt(stmt))
         {// must be that to be candidate 
          found = 0;
          if (list = isSgExprListExp(stmt->expr(0)))
            {
              expr = list->value();
              if (expr)
                {// no copy needed for that call => ,0;
                 if (isSgReturnStmt(stmt))
                   {
                     if (!SimpleCodeGeneration)                       
                       list->setValue(*modifyReferenceToCollection(expr,stmt,found,&coll,1));
                     found = 0;
                   }
                 else
                   list->setValue(*modifyReferenceToCollection(expr,stmt,found,&coll,0));
                 if (found)
                   {
                     if (found != 2)
                       {
                         previous = stmt->nodeBefore();
                         savenext = stmt->lexNext();
                       }
                     stmt = generateCodeForStmt(stmt, savenext,previous,found,coll);  
                     if (inMethodOfTheElement(stmt) && found)
                       {
                         Message("Nested parallelism not yet implemented",stmt->lineNumber());
                         if (VeryVerbose)
                           {
                             fprintf(stderr,"The SPMD exection model used does not allow a general form\n");
                             fprintf(stderr,"of nested parallelism; Complains are accepted, and if you\n");
                             fprintf(stderr,"justify very seriously you asking (for instance a pay raise)\n");
                             fprintf(stderr,"we may do something about it\n");
                           }
                       }
                   }
               }
              if (list->next())
                {
                  Message("Statement does not allow code generation",stmt->lineNumber());
                  break;
                }
            }
        }
       previous  = stmt;
     }     
}


//
// we have to move method of the collection into the object
// we use the instance created to do that
//
// normally for one element class we can only have one instance of 
// each collection to move
void
moveMOfE(SgStatement *coll,SgStatement *elstmt)
{
  SgStatement *firstMethOfE, *pt, *extracted;
  SgStatement *previous;
  SgCollectionStmt *thecoll;
  if (!coll || !elstmt)
    return;
  
  if (thecoll = isSgCollectionStmt(coll))
    {
      firstMethOfE = thecoll->firstElementMethod();
      if (firstMethOfE)
        previous = firstMethOfE->nodeBefore();
      if (firstMethOfE == thecoll->childList1(0))
        {
          pt = thecoll->childList1(0);
          while (pt)
            { // everything has to be moved
              extracted = pt->extractStmt();
              elstmt->insertStmtAfter(*extracted,*elstmt);
              pt = thecoll->childList1(0);
            }
        } else
          {
            for (pt = firstMethOfE; pt;)
              {
                extracted = pt->extractStmt();
                elstmt->insertStmtAfter(*extracted,*elstmt);
                pt = previous->nextInChildList();
              }
          }
    }
}

int
isVirtualDecl(SgStatement *pt)
{
  SgVarDeclStmt *stmt;
  SgSymbol *symb;
  SgType *type;
  SgDescriptType *drtype;
  if (!pt)
    return 0;
  
  if (stmt = isSgVarDeclStmt(pt))
    {
      if (stmt->numberOfSymbols()>0)
        {
          symb = stmt->symbol(0);
          if (symb)
            {
              type = symb->type();
              if (type)
                {
                  while (type)
                    {
                      if (drtype = isSgDescriptType(type))
                        {
                          if (drtype->modifierFlag() & BIT_VIRTUAL)
                            return 1;
                        }
                      type = type->baseType();
                    }
                } else {
                  Message("internal error in isVirtualDecl-1",0);
		  pt->unparsestdout();
		}
            } else
              Message("internal error in isVirtualDecl-2",0);
        }
    } 
  return 0;
}


void
removeVirtual(SgStatement *elstmt)
{
  SgStatement *pt, *last, *previous, *lastofstmt;
  int i;
  if (!elstmt)
    return;
  last =  elstmt->lastNodeOfStmt();
  previous = elstmt;
  i = 0;
  for (pt = elstmt; pt && (pt != last->lexNext())  ; pt =  pt->lexNext())
    {
      if (isVirtualDecl(pt))
        {
          lastofstmt = pt->lastNodeOfStmt();
          pt->extractStmt();
          if (last == lastofstmt)
            break;
          pt = previous;
        }      
      if (pt == last)
        break;
      previous = pt;
      i++;
      if ( i> 10000)
        { // dirty trick for debugging
          Message("Infinite loop in removeVirtual",0);
          return;
        }
    }

}

void
resetOwnerOfMemberFunction(SgStatement *elstmt)
{
  SgStatement *pt, *last;
  SgSymbol *symb;
  SgMemberFuncSymb *func;
  SgVarDeclStmt *stmt;
  SgFuncHedrStmt *fun1;
  int i, nbsymb;

  if (!elstmt)
    return;
  last =  elstmt->lastNodeOfStmt();
  for (pt = elstmt; pt; pt =  pt->lexNext())
    {
      if (stmt = isSgVarDeclStmt(pt))
        {
          nbsymb = stmt->numberOfSymbols();
          for (i = 0 ; i < nbsymb; i++)
            {
              symb = stmt->symbol(i);
              if (symb)
                {
                  if (func = isSgMemberFuncSymb(symb))
                    {
                      // this is the declaration, we have to change the owner 
                      // of the symbol;
                      if (elstmt->symbol())
                        func->setClassName(*elstmt->symbol());
                      else
                        Message("internal error in resetOwnerOfMemberFunction-1",0);
                    }
                } else
                  Message("internal error in resetOwnerOfMemberFunction-2",0);
            }
        } else
          {
            if (fun1 = isSgFuncHedrStmt(pt))
              {
                symb = fun1->symbol();
                if (symb)
                  {
                    if (func = isSgMemberFuncSymb(symb))
                      {
                        // this is the declaration, we have to change the owner 
                          // of the symbol;
                        if (elstmt->symbol())
                          func->setClassName(*elstmt->symbol());
                        else
                          Message("internal error in resetOwnerOfMemberFunction-2",0);
                      }
                  } 
              }
          }
      if (pt == last)
        break;
    }
}

void
changeCollectionDeclToClassDecl(SgFile *file)
{
  SgStatement       *stmt;

  if (!file)
    return;

   for (stmt = file->firstStatement(); stmt ; stmt = stmt->lexNext())
     { 
       if(isSgCollectionStmt(stmt))
         stmt->setVariant(CLASS_DECL);
     }
}


void
moveMethodOfTheElementToElement()
 {
   int i;
   SgType *type;
   SgStatement *coll, *elstmt;
   SgDerivedCollectionType *colltype;
   SgDerivedType *eltype;
   SgSymbol *elsymb;
   // this is stored in this table, they should not be redundancy in that 
   // table, so do it for each collection and each element 
   for (i= 0; i< Nb_Elem_in_Tab_Collection_Created; i++)
     {
       coll = Tab_Collection_Created_Stmt[i];
       if (coll)
         {
           if (colltype = isSgDerivedCollectionType(Tab_Collection_Created[i]))
             {
               type = colltype->elementClass();
               if (eltype = isSgDerivedType(type))
                 {
                   if (elsymb = eltype->typeName())
                     {
                       if (elstmt = elsymb->body())
                         {
                           // here we do the job; 
                           moveMOfE(coll,elstmt);
                           removeVirtual(elstmt);
                           resetOwnerOfMemberFunction(elstmt);
                         } else 
                           Message("internal error in moveMethodOfTheElementToElement-1",0);                          
                     } else 
                       Message("internal error in moveMethodOfTheElementToElement-2",0);
                 } else
                   Message("Element type for a collection must be a class",0);
             }
           else 
             Message("internal error in moveMethodOfTheElementToElement-3",0);
         }
       else 
         Message("internal error in moveMethodOfTheElementToElement-4",0);
     }
}

//
// Add forward definition 
//

void
addForwardDefinitions(SgFile *file)
 {
   int i;
   SgStatement *coll, *first;
   SgClassStmt *fordef;

   first = file->firstStatement();
   if (!first)
     {
        Message("internal error in addForwardDefinitions-1",0);
        return ;
     }
   for (i= 0; i< Nb_Elem_in_Tab_Collection_Created; i++)
     {
       coll = Tab_Collection_Created_Stmt[i];
       if (!coll)
         {
           Message("internal error in addForwardDefinitions-2",0);
           return ;
         }
       fordef = new SgClassStmt(*coll->symbol());
       first->insertStmtAfter(*fordef);
     }
}

//
//
//
// functions to find overloading operators in a class or in a collection
//   take an expression as input and return the function
//   return the symbol corresponding to the function.
//   will probably return more than that......
//


//
// Look at an expression and return a basic type if it is an expression
// on a collection and return the type of the operation
// that function ends recursivity 
//
SgType *
lookForOpOnCollection(SgExpression *expr, SgStatement *where, SgSymbol **col)
{
  SgType *base = NULL;
  SgSymbol *symb;
  SgDerivedCollectionType *type;

  if (!expr || ! where)
    return NULL;
  if ((type = isReferenceToCollection(expr,where)) && (symb = isReferenceToMethodOfElement(expr->rhs())))
    {
      base = symb->type();
      if (base->isTheElementType())
        base = type->elementClass();
      *col = type->collectionName();
    } else
      if (type = isReferenceToCollection(expr,where))
        {
          base = type->elementClass();
          *col = type->collectionName();
        }
  return base;
}

//
// look a type and get the class corresponding, this is a statement
//

SgStatement *
getTheClass(SgType *type)
{
  SgDerivedType *eltype;
  SgSymbol      *symb;

  if (!type)
    return NULL;
  
  if (eltype = isSgDerivedType(type))
    {
      if (symb = eltype->typeName())
        {
          if (symb->variant() == CLASS_NAME)
            {
              return symb->body();
            }
        }
    }
  if (type->hasBaseType())
    {
      return getTheClass(type->baseType());
    }
  return NULL;
}

//
//
// return the expression ; new of variable with a T_DERIVED_COLLECTION type
//

SgCExpStmt *
genNewCollection(SgDerivedCollectionType *type,SgSymbol *symb, SgExpression *coll)
{
  SgCExpStmt *stmt;
  SgNewExp *theexpr;
  SgVarRefExp *varef;
  SgExprListExp *list;

  if (!type || !symb || !coll)
    return NULL;
  list = new SgExprListExp(*coll);
  theexpr = new SgNewExp(*type, *list);
  varef = new  SgVarRefExp(*symb);
  stmt = new SgCExpStmt(*varef, *theexpr);
  return stmt;
}


//
SgCExpStmt *
genDeleteCollection(SgSymbol *ptcoll)
{
  SgCExpStmt *stmt;
  SgDeleteExp *theexpr;
  SgVarRefExp *varef;
  SgExprListExp *list;

  if (!ptcoll)
    return NULL;
  varef = new  SgVarRefExp(*ptcoll);
  list = new SgExprListExp(*varef);
  theexpr = new SgDeleteExp(*list);
  stmt = new SgCExpStmt(*theexpr);
  return stmt;
}

// Look an expression and transform the expression to generate copies
// of collection, coll gives the reference template and alignement
//

static int TemporaryNum = 0;

SgSymbol *
generateCopyOfCollection(SgExpression *expr, SgStatement *where, SgDerivedCollectionType *type,SgExpression *coll)
{
  SgSymbol      *tempsymb;
  SgStatement   *funs;
  SgPointerType *pttype;
  SgCExpStmt    *stmt;
  char          nn[512];

  if (!expr || !where || !type)
    return NULL;

  // we first need to declare a temporary variable;
  // create the symbol;
  // make it so

  funs =  whichFunctionAmI(where);
  if (!funs)
    {
      Message("internal error",0);
      return NULL;
    }
  // the new constructor here is assume to do a copy of the;
  // collection that is used as reference;
  sprintf(nn,"pCTemp__%d__%d",TemporaryNum,where->lineNumber());
  TemporaryNum++;
  if (!isSgPointerType(type))
    {
      pttype = new SgPointerType(*type);
      tempsymb = new SgSymbol(VARIABLE_NAME,nn, *pttype, *funs);
    } else
      tempsymb = new SgSymbol(VARIABLE_NAME,nn, *type, *funs);
  tempsymb->declareTheSymbol(*funs); 
  // now we have a variable ;
  // we have to create the new expression to build the temporary collection;
  stmt = genNewCollection(type, tempsymb, coll);
  where->insertStmtBefore(*stmt,*where->controlParent());
  stmt->setlineNumber(where->lineNumber());
  if (!isSgReturnStmt(where))
    {
      stmt = genDeleteCollection(tempsymb);
      where->insertStmtAfter(*stmt,*where->controlParent());
      stmt->setlineNumber(where->lineNumber());
    }
  return tempsymb;
}

//
// fix type for copy statement
//

SgDerivedCollectionType *
fixTypeForCopy(SgExpression *coll, SgDerivedCollectionType *td1in, SgStatement *where)
{
  SgType                  *et1, *type2;
  SgReferenceType         *reftype;
  SgDescriptType          *dsctype;
  SgDerivedCollectionType *td2;
  SgDerivedCollectionType *td1;
  int change;

  if (!td1in)
    return NULL;    
    
  if (!(td1 = isSgDerivedCollectionType(&(td1in->copy()))))
    {
      Message("internal error  in fixTypeForCopy-1",0);
      return td1in;
    }

  if (!td1->collectionName())
    return td1;
    
  if (!(et1 = td1->elementClass()))
    return td1;
  
  change = 1;
  while (change)
    {
      change = 0;
      if (reftype = isSgReferenceType(et1))
        {
          et1 = reftype->baseType();
          change = 1;
        }
      if (dsctype = isSgDescriptType(et1))
        {
          et1 = dsctype ->baseType();
          change = 1;
        }
    }
  
  if (et1)
    td1->setElementClass(*et1);

  if (et1->isTheElementType())
    {
      // we have to find the collection and we take the symbol element class;
      type2 = computeTypeOfExpression(coll, where);
      if (td2 = isSgDerivedCollectionType(type2))
        {
          if (td2->elementClass())
            td1->setElementClass(*td2->elementClass());
        } else
          {
            Message("internal error  in fixTypeForCopy-2",0);
            return td1;
          }
    }
  return td1;
}


// coll is the collection uses as a basis for the copy

SgExpression *
generateCopyStmt(SgExpression *expr, SgStatement *where, SgExpression *coll, SgDerivedCollectionType *res, SgStatement  **stcopy)
{
  SgSymbol     *ncol;
  SgCExpStmt   *stmt;
  SgVarRefExp  *varef;

  if (!expr || !where || !coll || !res)
    return expr;
  res = fixTypeForCopy(coll,res, where); 
  // the type must also be added in the list for instance generation;
  addToTabCollectionType(res);
    
  ncol = generateCopyOfCollection(expr,where,res,coll);
  // now we have to transform the expression;
  // to be assigned;
  varef= new SgVarRefExp(*ncol);                      
  stmt = new SgCExpStmt(SgDerefOp(*varef), *expr);
  *stcopy =  stmt;
  where->insertStmtBefore(*stmt,*where->controlParent());
  stmt->setlineNumber(where->lineNumber());
  return varef;
}

//
// for the header and the tail of the output file
//
int writeHeader(FILE *fout)
{
  if((arch_flag == 1) && spmd_flag){
    fprintf(fout,"__private int pcxx_inpar = 0;\n"); //test insert for KSR pcxx_inpar BW
  }
  else fprintf(fout,"int pcxx_inpar = 0;\n"); 
  fprintf(fout,"class pCInParClass{\n");
  fprintf(fout,"  public:\n");
  fprintf(fout,"    pCInParClass(){ pcxx_inpar++;}\n");
  fprintf(fout,"   ~pCInParClass(){ pcxx_inpar--;}\n");
  fprintf(fout,"};\n");
  fprintf(fout,"#include \"runtime.h\"\n");
  return 7;  /* NOTE! # of newlines in preceding code */
}

void writeTail(FILE *fout)
{
  fprintf(fout,"\n#include \"startup.h\"\n");

#if 0
  fprintf(fout,"int main(int argc, char *argv[]) { \n");
  fprintf(fout,"  pcxx_MainBegin(&argc, argv);\n");
  fprintf(fout,"     Processor_Main();  /* soon: argc & argv PHB */\n");
  fprintf(fout,"  pcxx_MainEnd();\n");
  fprintf(fout,"  return 0;\n");
  fprintf(fout,"}\n");
#endif
}

SgCallGraph CallGraph;
//
// this function is used to create a new printf and fprintf calls.
// it builds the new function and inserts it at the top of the file.
// there is a unique pcxx_printf function for each occurance of a printf
// or fprintf.  
//
#define SGMAX_ARGS 100
int ScanCounter = 1;
SgSymbol * makeNewPrintf(SgSymbol *scan, SgFile *file, SgExpression *args){
  // first create a new function 
  char name[50];
  SgStatement *s;
  int i = 0;
  int isFscan = 0;
  int no_of_args;
  SgExprListExp *list;
  SgExpression *passedarg[SGMAX_ARGS];  // the args that are passed
  SgType *argtype[SGMAX_ARGS];          // the types of the args
  SgSymbol *newname[SGMAX_ARGS];        // the new names for parameters
  SgExpression *nameref[SGMAX_ARGS];    // references to the new names.
  //
  // create the function header and insert it at top of file
  //
  sprintf(name, "pcxx_printf%d", ScanCounter++);
  SgFuncHedrStmt *ptfunc = new SgFuncHedrStmt(name);
  SgSymbol *new_name = &(ptfunc->name());
  s = file->firstStatement();
  s->insertStmtAfter(*ptfunc);
  //
  // extract the list of arguments and put them in the function header
  //
  if((list = isSgExprListExp(args)) == NULL){
		printf("bad arg list to makeNewPrintf\n");
                fflush(stdout);
		return NULL;
		}
  for(i = 0; (i < SGMAX_ARGS) && list; i++, list=list->next()){
	passedarg[i] = list->value();
        if(i == 0) argtype[i] = passedarg[i]->type();
        else if((argtype[i] = makeReducedType(passedarg[i]->type())) == NULL){
		printf("unknown type for expression:");
		passedarg[i]->unparsestdout();
                printf("\n");
                fflush(stdout);
                return NULL;
		}
	sprintf(name, "pCar_%d", i);
        newname[i] = new SgVariableSymb(name, *(argtype[i]), *ptfunc);
	nameref[i] = new SgVarRefExp(*(newname[i]));
	if(isSgPointerType(argtype[i])) // || isSgArrayType(argtype[i]))
        	ptfunc->AddArg(SgDerefOp(*(nameref[i])));
	else    ptfunc->AddArg(*(nameref[i]));
	}
  if(i == SGMAX_ARGS) printf("too many args for printf\n");
  no_of_args = i;
  //
  // create the real call to scanf
  //
  SgFunctionCallExp *call = new SgFunctionCallExp(*scan);
  for(i = 0; i < no_of_args; i++)
	call->addArg(nameref[i]->copy());
  SgExprListExp *e = new SgExprListExp(*call);
  SgCExpStmt *cexp = new SgCExpStmt(*e);
  //
  // create the fake return value  
  //
  call = new SgFunctionCallExp(*SymBcast);
  if(strcmp(scan->identifier(), "pcxx_fprintf") == 0) isFscan = 1;
  // add the format string arg. (2nd if a fscanf, 1st if scanf)
  call->addArg(nameref[isFscan]->copy());
  // pass the number of args of pointer type.
  SgValueExp *val = new SgValueExp(no_of_args -1 - isFscan);
  call->addArg(*val);
  SgReturnStmt *cexp2 = new SgReturnStmt(*(new SgValueExp(1)));
  ptfunc->insertStmtAfter(*cexp2, *ptfunc );
  //
  //  make the if statement
  //
  SgFunctionCallExp *pcxx_nodenumber = new SgFunctionCallExp(*Sympcxx_NodeNumber);
  SgValueExp *zero = new SgValueExp(0);
  SgExprListExp *cond = new SgExprListExp(SgEqOp(*pcxx_nodenumber, *zero));
  SgVarRefExp *inpar = new SgVarRefExp(*pcxx_inpar_symb);
  cond = new SgExprListExp(*cond || *inpar);
  SgIfStmt *myif = new SgIfStmt(*cond, *cexp);
  ptfunc->insertStmtAfter(*myif, *ptfunc );
  // ptfunc->unparsestdout();
  return new_name;
}	

//
// this function is used to create a new scanf and fscanf call.
// it builds the new function and inserts it at the top of the file.
// there is a unique scanf function for each occurance of a scanf
// or fscanf.  
//

SgSymbol * makeNewScanf(SgSymbol *scan, SgFile *file, SgExpression *args){
  // first create a new function 
  char name[50];
  SgStatement *s;
  SgPointerType *pt;
  SgDescriptType *disc;
  SgValueExp *v1;
  SgArrayType *ar;
  int disc_flag; 
  SgType *t;
  int i = 0;
  int isFscan = 0;
  int no_of_args;
  SgExprListExp *list;
  SgExpression *passedarg[SGMAX_ARGS];  // the args that are passed
  SgType *argtype[SGMAX_ARGS];          // the types of the args
  SgSymbol *newname[SGMAX_ARGS];        // the new names for parameters
  SgVarRefExp *nameref[SGMAX_ARGS];     // references to these names
  //
  // create the function header and insert it at top of file
  //
  sprintf(name, "pcxx_scanf%d", ScanCounter++);
  SgFuncHedrStmt *ptfunc = new SgFuncHedrStmt(name);
  SgSymbol *new_name = &(ptfunc->name());
  s = file->firstStatement();
  s->insertStmtAfter(*ptfunc);
  //
  // extract the list of arguments and put them in the function header
  //
  if((list = isSgExprListExp(args)) == NULL){
		printf("bad arg list to makeNewScanf\n");
                fflush(stdout);
		return NULL;
		}
  for(i = 0; (i < SGMAX_ARGS) && list; i++, list=list->next()){
	passedarg[i] = list->value();
        if((argtype[i] = passedarg[i]->type()) == NULL){
		printf("unknown type for expression:"); 
		passedarg[i]->unparsestdout();
                printf("\n");
                fflush(stdout);
                return NULL;
		}
	sprintf(name, "pCar_%d", i);
        newname[i] = new SgVariableSymb(name, *(argtype[i]), *ptfunc);
        nameref[i] = new SgVarRefExp(*(newname[i]));
        ptfunc->AddArg(SgDerefOp(*(nameref[i])));
	}
  if(i == SGMAX_ARGS) printf("too many args for scanf\n");
  no_of_args = i;
  //
  // create the real call to scanf
  //
  SgFunctionCallExp *call = new SgFunctionCallExp(*scan);
  for(i = 0; i < no_of_args; i++)
	call->addArg(nameref[i]->copy());
  SgExprListExp *e = new SgExprListExp(*call);
  SgCExpStmt *cexp = new SgCExpStmt(*e);
  //
  // create the call to the broadcast scan.  
  //
  call = new SgFunctionCallExp(*SymBcast);
  if(strcmp(scan->identifier(), "fscanf") == 0) isFscan = 1;
  // add the format string arg. (2nd if a fscanf, 1st if scanf)
  call->addArg(nameref[isFscan]->copy());
  // pass the number of args of pointer type.
  SgValueExp *val = new SgValueExp(no_of_args -1 - isFscan);
  call->addArg(*val);
  // pass three values for each paramenter: 
  // the long-short flag, the type (1 for int, etc.), the pointer
  for(i = 1+isFscan; i < no_of_args; i++){
        int base_found;
	t = argtype[i];
        disc_flag = 0; base_found = 0;
        while(!base_found){
          if(disc = isSgDescriptType(t)){
		disc_flag |= disc->modifierFlag();
		t = disc->baseType();
		}
          else if(pt = isSgPointerType(t)) {
	    if (pt->baseType() == NULL){
		printf("error in scanf parser\n");
                fflush(stdout);
		return NULL;
		}
	     t = pt->baseType();
	     }
           else if(ar = isSgArrayType(t)) t = ar->baseType();
           else base_found = 1;
	  }
	     
	switch(t->variant()){
		case T_INT:   v1 = new SgValueExp(1);
			      break;
		case T_FLOAT: v1 = new SgValueExp(2);
			      break;
		case T_DOUBLE:v1 = new SgValueExp(3);
			      break;
		case T_CHAR:  v1 = new SgValueExp(4);
			      break;
		default:
			printf("base type error in scanf parser");
			v1 = new SgValueExp(0);
			break;
		}
	SgValueExp *discv = new SgValueExp(disc_flag);
	call->addArg(*discv); 
	call->addArg(*v1);		
	call->addArg(nameref[i]->copy());
	}
  e = new SgExprListExp(*call);
  SgReturnStmt *cexp2 = new SgReturnStmt(*e);
  ptfunc->insertStmtAfter(*cexp2, *ptfunc );
  //
  //  make the if statement
  //
  SgFunctionCallExp *pcxx_nodenumber = new SgFunctionCallExp(*Sympcxx_NodeNumber);
  SgValueExp *zero = new SgValueExp(0);
  SgExprListExp *cond = new SgExprListExp(SgEqOp(*pcxx_nodenumber, *zero));
  SgIfStmt *myif = new SgIfStmt(*cond, *cexp);
  ptfunc->insertStmtAfter(*myif, *ptfunc );
  // ptfunc->unparsestdout();
  return new_name;
}	

//
// this function looks for all instances of fscanf and scanf,
// whith calls to a special function that is generated for each call, 
//
int findAndReplaceScanf(SgFile *f){
  SgStatement *s;
  SgExpression *e;
  SgSymbol *scan = NULL;
  SgFunctionCallExp *call;
  SgExpression *args;
  SgSymbol *new_name;

  // do scanf first.
  for(scan = CallGraph.firstFunction(); scan; scan = CallGraph.nextFunction()){
	if(strcmp(scan->identifier(), "scanf") == 0) break;
	}
  if(scan){
        for(s = CallGraph.firstCallSiteStmt(scan),
            e = CallGraph.firstCallSiteExpr(scan);
            s;
            s = CallGraph.nextCallSiteStmt(scan),
            e = CallGraph.nextCallSiteExpr(scan)){
		// printf("call site:"); s->unparsestdout();
		if(e && e->variant() == EXPR_LIST) e = e->lhs();
		if((call = isSgFunctionCallExp(e)) == NULL){
			printf("cant find the function call!\n");
                        fflush(stdout);
			return 0;
			}
		args = call->args();
		new_name = makeNewScanf(call->funName(), f, args);
                if(new_name == NULL) {
                  printf("possible error on line %i in file %s\n",
                         s->lineNumber(), s->fileName());
                  exit(1);
                  }
		call->setSymbol(*new_name);
		}
	}

  // do fscanf next.
  for(scan = CallGraph.firstFunction(); scan; scan = CallGraph.nextFunction()){
	if(strcmp(scan->identifier(), "fscanf") == 0) break;
	}
  if(scan){
        for(s = CallGraph.firstCallSiteStmt(scan),
            e = CallGraph.firstCallSiteExpr(scan);
            s;
            s = CallGraph.nextCallSiteStmt(scan),
            e = CallGraph.nextCallSiteExpr(scan)){
		// printf("call site:"); s->unparsestdout();
		if(e && e->variant() == EXPR_LIST) e = e->lhs();
		if((call = isSgFunctionCallExp(e)) == NULL){
			printf("cant find the function call!\n");
                        fflush(stdout);
			return 0;
			}
		args = call->args();
		new_name = makeNewScanf(call->funName(), f, args);
                if(new_name == NULL) {
                  printf("possible error on line %i in file %s\n",
                         s->lineNumber(), s->fileName());
                  exit(1);
                  }
		call->setSymbol(*new_name);
		}
	}
  //
  // ig++ has a varars problem so we will need to make a printf as well.
  // do printf first.
  //
  for(scan = CallGraph.firstFunction(); scan; scan = CallGraph.nextFunction()){
	if(strcmp(scan->identifier(), "printf") == 0) break;
	}
  if(scan){
        for(s = CallGraph.firstCallSiteStmt(scan),
            e = CallGraph.firstCallSiteExpr(scan);
            s;
            s = CallGraph.nextCallSiteStmt(scan),
            e = CallGraph.nextCallSiteExpr(scan)){
		// printf("call site:"); s->unparsestdout();
		if(e && e->variant() == EXPR_LIST) e = e->lhs();
		if((call = isSgFunctionCallExp(e)) == NULL){
			printf("cant find the function call!\n");
                        fflush(stdout);
			return 0;
			}
		args = call->args();
		new_name = makeNewPrintf(call->funName(), f, args);
                if(new_name == NULL) {
                  printf("possible error on line %i in file %s\n",
                         s->lineNumber(), s->fileName());
                  exit(1);
                  }

		call->setSymbol(*new_name);
		}
	}

  // do fprintf next.
  for(scan = CallGraph.firstFunction(); scan; scan = CallGraph.nextFunction()){
	if(strcmp(scan->identifier(), "fprintf") == 0) break;
	}
  if(scan){
        for(s = CallGraph.firstCallSiteStmt(scan),
            e = CallGraph.firstCallSiteExpr(scan);
            s;
            s = CallGraph.nextCallSiteStmt(scan),
            e = CallGraph.nextCallSiteExpr(scan)){
		// printf("call site:"); s->unparsestdout();
		if(e && e->variant() == EXPR_LIST) e = e->lhs();
		if((call = isSgFunctionCallExp(e)) == NULL){
			printf("cant find the function call!\n");
                        fflush(stdout);
			return 0;
			}
		args = call->args();
		new_name = makeNewPrintf(call->funName(), f, args);
                if(new_name == NULL) {
                  printf("possible error on line %i in file %s\n",
                         s->lineNumber(), s->fileName());
                  exit(1);
                  }
		call->setSymbol(*new_name);
		}
	}

   return 0;
}
	
//
// This function will search for and delete declarations of varaibles and
// types that lie between the declarations 
//  int pcxx_cut_start;
// and
//  int pcxx_cut_end;
// the symbols are left in the symbol table for future use.
//
// first a function to identify the cut points.
//

static char include_name[500]; // the name of the include file (no .h)
static char end_name[500]; // pcxx_cut_end_ + include_name
int isCutStart(SgStatement *st){
   SgVarDeclStmt *d;
   SgSymbol *s;
   char *name;
   if(d = isSgVarDeclStmt(st)){
	s = d->symbol(0);
	if((name = s->identifier()) && 
	   (strncmp(name, "pcxx_cut_start", 14) == 0)){
                 strcpy(include_name, &(name[15]));
                 strcpy(end_name, "pcxx_cut_end_");
                 strcpy(&(end_name[13]),&(name[15])); 
		return 1;
		}
	}
    return 0;
}

int isCutEnd(SgStatement *st){
   SgVarDeclStmt *d;
   SgSymbol *s;
   char *name;
   if(d = isSgVarDeclStmt(st)){
	s = d->symbol(0);
	if((name = s->identifier()) && 
	   (strcmp(name, end_name) == 0))
		return 1;
	}
    return 0;
}
	
int removeDummyStatements(SgFile *f, FILE *of){
   SgStatement *s = f->firstStatement();
   SgStatement *rem;
   int done;
   int linecount = 0;

   while(s){
      while(s && !isCutStart(s)) s = s->lexNext();
      if(s){
        if(!of)
		printf("#include <%s.h>\n", include_name);
	else    fprintf(of,"#include <%s.h>\n", include_name);
	linecount++;
	}
      done = 0;
      while(s && !done){
	rem = s;
        done = isCutEnd(rem);
        // s->unparsestdout();
	s = s->lastNodeOfStmt();
	s = s->lexNext();
	rem->extractStmt();
	}
     }
   return linecount;
}
   
//
// this function will declar a pCInParClass object in each collection
// method function, but not in MOE functions.
//

void createInParClassObjects(SgFile *f){
   SgStatement *s = f->firstStatement();
   SgSymbol *aFun;
   SgMemberFuncSymb *memfname;
   SgStatement *body;
//   CallGraph.GenCallTree(f); // commented out lth@oregon 941230; moved into main()
   // create a statement of the form
   // class NewClass;
   // and insert it in the top of the program.
   SgSymbol *classname = new SgSymbol(CLASS_NAME, "pCInParClass");
   // SgClassStmt *class_decl = new SgClassStmt(*classname);
   // s->insertStmtAfter(*class_decl);

   // now declare a variable of that type
   SgDerivedType *t = new SgDerivedType(*classname);
   for(aFun = CallGraph.firstFunction(); aFun; aFun = CallGraph.nextFunction()){
        if(CallGraph.isACollectionFunc(aFun)){   
                if((memfname = isSgMemberFuncSymb(aFun)) && (body = memfname->body())){
   			SgSymbol *x = new SgVariableSymb("pCInPar", *t);
   			x->declareTheSymbol(*body);
			}
		else{ /*
		     if(!memfname) printf("A member function that isn't: %s\n", aFun->identifier());
		     else printf("Warning: No body found for: %s\n", aFun->identifier());
		     */
		     }
                }
	}

   // now seach for a collection variable of the
   // form C<A> z;

   // extract the element type.
   // extract the collection type.

}

/**********************************************************************/
/*  Convert TEClass instances to regular classes                      */
/**********************************************************************/


//
// find the TEClass definitions in the file and add a new constructor
// function.
//
// the first function adds the construtor  classname(Processors &P): TEKernel(P){}
// to the TEClass referenced by the argument.
//
SgSymbol *TEKernel;
SgSymbol *Processors;
SgSymbol *TEKconstructor;
//
void FixTEClass(SgClassStmt *cl, SgCallGraph &CG){
  SgMemberFuncSymb *con;
  SgSymbol *scan;
  SgExpression *el, *fr, *ci, *bo;
  SgVarDeclStmt *decl;
  char *className;
  int n, i, count;
  if(cl == NULL) return;
  if((n = cl->numberOfChildrenList1()) == 0){
	// a forward def
        cl->setVariant(CLASS_DECL);
        return;
	}
  // printf("doing a fix\n");
  cl->setSuperClass(cl->numberOfSuperClasses(), *TEKernel, BIT_PUBLIC);
  className = cl->symbol()->identifier();
  //
  // build a reference to a "Processors" parameter P
  SgDerivedType *processorType = new SgDerivedType(*Processors);
  //
  // next, for each non-inlined constructor, add the parameter P.
  //
  for(count = 0, i = 0; i < n; i++){
      SgExpression *funr;
      SgExprListExp *args;
      SgVariableSymb *arg;
      SgVarRefExp *e;
      if(decl = isSgVarDeclStmt(cl->childList1(i))){
           el = decl->expr(0);
           if(((funr = el->lhs())->variant() == FUNCTION_REF) && 
              (strcmp(funr->symbol()->identifier(), className) == 0)){
			// found a constructor!
			count++;
  			arg = new SgVariableSymb("P", *processorType, *decl);
  			e = new SgVarRefExp(*arg);
			// constructor found, append P to end of arg list.
			if(args = isSgExprListExp(funr->lhs()))
				args->append(*e);
			else{ 
                           funr->setLhs(*(new SgExprListExp(*e)));
			   }
			}
	   }
    }
  //
  // Now go through the call graph looking for the bodies of constructors
  // we need to add the parameter and the call to the TEKernel constructor
  // to each of these.
  //
  SgFuncHedrStmt *myfun;
  for(scan = CG.firstFunction(); scan; scan = CG.nextFunction()){
        SgVariableSymb *arg;
        SgVarRefExp *e;
	SgExprListExp *lis;
	if( (strcmp(className, scan->identifier()) == 0) &&
	     (myfun = isSgFuncHedrStmt(CG.functionBody(scan)))){
		// printf(" found a live one\n");
  		arg = new SgVariableSymb("P", *processorType, *myfun);
  		e = new SgVarRefExp(*arg);
                myfun->AddArg(*e);
  		el = myfun->expr(0); // an expression list
  		fr = el->lhs();  // the FUNCTION_REF or BIT_NUMBER
		if(fr->variant() == FUNCTION_REF){
  			ci = new SgExpression(CLASSINIT_OP); 
  			ci->setLhs(*( new SgVarRefExp(*TEKconstructor)));
  			ci->setRhs(*(new SgExprListExp(e->copy())));
  			bo = new SgExpression(BIT_NUMBER); // used to generate :
  			bo->setLhs(*fr);
  			bo->setRhs(*ci);
  			el->setLhs(*bo);
			}
                 else if(fr->variant() == BIT_NUMBER){
  			ci = new SgExpression(CLASSINIT_OP); 
  			ci->setLhs(*( new SgVarRefExp(*TEKconstructor)));
  			ci->setRhs(*(new SgExprListExp(e->copy())));
			if(lis = isSgExprListExp(fr->rhs()))
				lis->append(*ci);
			else {printf("wrong!\n"); fflush(stdout);}
			}
     		}
  }
           
  if(count == 0){	
  //
  // now create a constructor of the form
  // teclassname(Processors P):TEKernel(P){}
  //
  	con = new SgMemberFuncSymb(className, *SgTypeDefault(), *cl, 0);
  	SgFuncHedrStmt *ptfunc = new SgFuncHedrStmt(*con);
  	SgVariableSymb *arg = new SgVariableSymb("P", *processorType, *ptfunc);
  	SgVarRefExp *e = new SgVarRefExp(*arg);
  	cl->insertStmtAfter(*ptfunc, *cl);
  	ptfunc->AddArg(e->copy());
  	// now add the call to the super constructor
  	el = ptfunc->expr(0); // an expression list
  	fr = el->lhs();  // the FUNCTION_REF or BIT_NUMBER
	if(fr->variant() == FUNCTION_REF){
  		ci = new SgExpression(CLASSINIT_OP); 
  		ci->setLhs(*( new SgVarRefExp(*TEKconstructor)));
  		ci->setRhs(*(new SgExprListExp(e->copy())));
  		bo = new SgExpression(BIT_NUMBER); // used to generate :
  		bo->setLhs(*fr);
  		bo->setRhs(*ci);
  		el->setLhs(*bo);
		}
   	}  
}
  
//
//
void findTEClass(SgFile *f, SgCallGraph &CG){
  SgStatement *st = f->firstStatement();
  SgClassStmt *cl;
  TEKernel = new SgSymbol(CLASS_NAME, "TEKernel");
  TEKconstructor = new SgMemberFuncSymb("TEKernel", *SgTypeDefault(), *st, 0);
  Processors = new SgSymbol(CLASS_NAME, "Processors");
  while(st){
	if((cl = isSgClassStmt(st)) && (cl->variant() == TECLASS_DECL)){
		 FixTEClass(cl, CG);
		 cl->setVariant(CLASS_DECL);
		}
        st = st->lexNext();
	}
}

//
// main program
//
main(int argc,char **argv)
{
  SgProject *project;
  SgFile    *file;
  char *ifi, *arch_string;
  FILE      *outfile;
  int       lineno = 1;
               /* The variable lineno keeps track of the line number
		* we're on by counting the number of of boilerplate header 
		* lines are printed out; we need to know this to do 
		* mapping correctly during unparsing.
		*/
  int       generate_map_file = 0;
               /* The variable generate_map_file is 1 if the user wants
		* the unparse.map mapping file to be generated.
		*/
  char      basename[ 1024 ];
               /* The string basename holds the name of the project file
		* sans extension.
		*/

//
// read the parameters for the program
//
  if (isArgumentThere("-version",argc,argv) ||
      isArgumentThere("-v",argc,argv) ||
      isArgumentThere("-V",argc,argv))
    {
      writePreambule();
    }
  if (isArgumentThere("-help",argc,argv) ||
      isArgumentThere("-?",argc,argv) ||
      isArgumentThere("-H",argc,argv) ||
      isArgumentThere("-h",argc,argv))
    {
      printHelp();
    }
  if (isArgumentThere("-verbose",argc,argv))
    Verbose = 1;

  if (isArgumentThere("-veryverbose",argc,argv))
    {
      VeryVerbose = 1;
      Verbose = 1;
    }

  if (isArgumentThere("-breakpoint",argc,argv) ||
      isArgumentThere("-b",argc,argv))
    {
      Brk_special = 1;
    }
  
  if (isArgumentThere("-arch",argc,argv))
    {
       arch_string = getStringForArgument("-arch",argc,argv);
       if(strcmp("ksr", arch_string) == 0) arch_flag = 1;
    }

  if (isArgumentThere("-ms",argc,argv))
    {
      spmd_flag = 0;
    }

  generate_map_file = isArgumentThere("-map",argc,argv);

  SetStop = getValueForArgument("-setdebug",argc,argv);


  // input parameter, MUST be there
  ifi = getStringForArgument("-i",argc,argv);
  if (!ifi)
    {
    // (PHB) if there is no "-i" option, try using the LAST file in argv

    if (argc < 1) { writeUsage(); exit(1); }

    char LastParm[40], *lpp;
    int lng;
    FILE *newProj;

    // Use *.dep name to create *.proj file on the fly...
    // see if it ends in .dep //
    strcpy(LastParm,argv[argc-1]);
    lng = strlen(LastParm);
    if (lng < 4) 
      { Message("Invalid .dep filename",0); exit(1); }

    lpp = &LastParm[lng-4];
    if (strcmp(lpp,".dep"))
      { Message("Must use a *.dep filename",0); exit(1); }

    // I've got a .dep file!!!
    lpp[0] = 0;
    strcat(LastParm,".proj");  // truncate .dep, add .proj

    // See if *.proj file already exists!
    newProj = fopen(LastParm,"r");  
    if (!newProj) {
      if (Verbose)
        fprintf(stderr,"Automatically creating project file: %s\n",LastParm);
      newProj = fopen(LastParm,"w");
      if (!newProj)
        { Message("Error, could not automatically generate .dep file",0);
          exit(1); }

      // write out filename into new *.proj and close
      fprintf(newProj,"%s\n",argv[argc-1]);
      }

    else
	if (Verbose)
	      fprintf(stderr,
	           "Automatically using existing project file: %s\n",LastParm);
 
    fclose(newProj);    
    ifi = LastParm;
    }

  { int l;
    char *p;
    strcpy( basename, ifi );
    l = strlen( basename );
    p = basename+l-1;
    while (p >= basename && *p != '/' && *p != '.') p--;
    if (p >= basename && *p == '.') *p = 0;
  }
  project =  new SgProject(ifi);
  file = &(project->file(0));
  if (!file)  
    {
      Message("file not found",0);
      exit(1);
    }

  // check this is c file
  if (file->languageType() != CSrc)
    {
      Message("This is not a C file; exiting",0);
      exit(1);
    }

  // Changed the structure of this 950125 / lth@cs.uoregon.edu
  // File checking control now in only one place (here); outfile
  // is always made valid.
  // Assumption: "unparsestdout()" is same as "unparse( stdout )".

  { char *ifo;
    ifo = getStringForArgument("-o",argc,argv);
    if(ifo) {
      outfile = fopen(ifo,"w");
      if (outfile == 0) {
	Message( "Cannot open output file", 0 );
	outfile = stdout;
      }
    }
    else
      outfile = stdout;
  }
  // At this point outfile is valid, no matter what.

  // unparse data base only;
  if (isArgumentThere("-unparse",argc,argv))
    {
      file->unparse( outfile );
      exit(0);
    }

  // set the code generator to it simple version (no copy generated and no overloading check);
  if (isArgumentThere("-simple",argc,argv))
      SimpleCodeGeneration = 1;

  if (isArgumentThere("-comment",argc,argv))                                 
   CommentOut = 0;
  else
   CommentOut = 1;                                

// initialize the compiler

  initializeCompiler(file);

  CallGraph.GenCallTree( file ); // moved here from below; lth@oregon 941230

  /* VERY, VERY, VERY, first thing in the file (PHB) */
  lineno += 1;
  fprintf(outfile,"// This program was generated by 'dep2C++' \n");

// delete the excess stdio dummy declarations
  lineno += removeDummyStatements(file, outfile);

// Now look for all TEClass instances and convert them to regular
// class, but make sure that they can still be seen as collections.
//

   findTEClass(file, CallGraph);
//
// look at all collection (non-element) functions and
// add the declaration for the sGInPar objects.
//
  createInParClassObjects(file); 
  findAndReplaceScanf(file); 

//
// Look at first the alignement function use old function
//
  lineno += Look_For_Align_Declaration(outfile);

//
// now treat the collections
//
  recordCollectionDerivedType(file);

//
// generate the code for calls to collections
// in short put a loop around the call to collection so
// that all processors iterate on its local element
//  * for master/slave code, just mark the calls for *
//  * now and generate the loop later                *
//
  treatCalltoMethodOfTheElement(file);


// unparse data base only;
  if (isArgumentThere("-unparsecodegen",argc,argv))
    {
      strcat( basename, ".pc" );
      do_unparse( file, outfile, basename, lineno, 
		  generate_map_file );
      exit(0);
    }

//
// create the collection instances
//

  createCollectionInstance();

//
// for all instance created we need to replace the types 
// for instance DistributedArray<vector> becomes DistributedArray__pC__Vector
//

  updateTypeInTheProgram(file);

//
// move method of the elementr in the element class
//
  moveMethodOfTheElementToElement();

//
// change collection decl by class declaration
//

  changeCollectionDeclToClassDecl(file);

/* moves to the end
//
// add forward definitions
//

  addForwardDefinitions(file);
*/

//
// remove the collection
//

  removeCollections();

//
//  * for m/s code, generate the loops for parallel calls now *
//

  if(!spmd_flag){
    treatParallelSections(file);
    makeWorkerSwitch(file);
  }

//
// add forward definitions
//

  addForwardDefinitions(file);

// unparse the file

  
  lineno += writeHeader(outfile);
  if((arch_flag == 1) && spmd_flag) replace_globals(file); //BW
  strcat( basename, ".pc" );
  do_unparse( file, outfile, basename, lineno, generate_map_file );
  basename[ strlen( basename )-3 ] = 0;
  writeTail(outfile);

  if (isArgumentThere("-debug",argc,argv))
    file->saveDepFile("debug.dep");

// print the warning count
  if (Warning_count > 1)
    fprintf(stderr,"There were %d warnings\n",Warning_count);
  else
    if (Warning_count)
      fprintf(stderr,"There was %d warning\n",Warning_count);

  exit(0);
  return 0; // to keep some compilers happy;
}

static void 
do_unparse( SgFile *file, FILE *outfile, char *source, int lineno, int map )
{
  if (map) unparser_begin_mapping( "unparse.map", source, lineno );
  file->unparse(outfile);
  if (map) unparser_end_mapping();
}

// eof
