/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * more globals
 */


static SgSymbol         *SymbEnterPar;
static SgSymbol         *SymbSubsetEnterPar;
static SgSymbol         *SymbExitPar;
static SgVarRefExp      *CurrentCollectionVar;
static int               workId=0;
static SgSwitchStmt     *parallelSections;
static SgSymbol		*SymbFirstInSubset;
static SgSymbol		*SymbNextInSubset;
static SgVarRefExp	*pcaw_First;
static SgVarRefExp	*pcaw_Last;
static SgVarRefExp	*pcaw_Step;

/*
 * The M-S compiler takes all the arguments to a MoE invocation and
 * creates global variables that can hold the arguments. The global
 * variables are declared at the beginning of the file. So if the
 * arguments are of class/structure/derived type, there must be a
 * forward declaration of the class etc.. To avoid multiple forward
 * declarations, the compiler keeps the following table of classes
 * that have been forward declared already. (K7 11/9/94)
 */
#define MAXTABSIZE	1024
static SgClassSymb      *forwardDeclsTab[MAXTABSIZE];
static int		numForwardDecls=0;

/*
 * For a subset operation in a MoE invocation, the M-S compiler sets
 * the following global variables, which are used in WorkerSwitch
 * directly. (K7 11/9/94)
 */
static SgSymbol		*Symbpcaw_First;
static SgSymbol		*Symbpcaw_Last;
static SgSymbol		*Symbpcaw_Step;

/*
 * forward decl
 */

extern SgExpression *makeLocalInductionVarRef(int, SgStatement *);
extern int LookIfAddrNeeded(SgExpression *, SgStatement *);


/*
 * initialize
 */

void initParallelSections(SgFile *file)
{
   SgStatement      *first;
   SgSymbol         *symb;
   SgExpression     *expr, *expr1;
   char strin[256];

   if (!file)
     return;

   first = file->firstStatement();

   SymbEnterPar = new SgFunctionSymb(FUNCTION_NAME,"pcaw_EnterPar",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*SymbEnterPar);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   EnterPar = new SgCExpStmt(*expr1);

   SymbSubsetEnterPar = new SgFunctionSymb(FUNCTION_NAME,"pcaw_SubsetEnterPar",*SgTypeInt(),*first);
   SymbFirstInSubset = new SgFunctionSymb(FUNCTION_NAME,"FirstInSubset",*SgTypeInt(),*first);
   SymbNextInSubset = new SgFunctionSymb(FUNCTION_NAME,"NextInSubset",*SgTypeInt(),*first);

   SymbExitPar = new SgFunctionSymb(FUNCTION_NAME,"pcaw_ExitPar",*SgTypeInt(),*first);
   expr = new SgFunctionCallExp(*SymbExitPar);
   expr1 = new SgExpression(EXPR_LIST);
   expr1->setLhs(*expr);
   ExitPar = new SgCExpStmt(*expr1);

   CurrentCollectionVar = new SgVarRefExp(*(new SgVariableSymb("pcaw_cc")));

// setup switch stmt for worker threads

   sprintf(strin,"pcaw_cf");
   symb = new SgVariableSymb(strin);

   parallelSections = new SgSwitchStmt(*(new SgVarRefExp(*symb)),
		      NULL, *(new SgConstructSymb("ws", *first)));

 // Global variables for handling subset operations in MoE invocation
  Symbpcaw_First = new SgVariableSymb("pcaw_First", *SgTypeInt());
  Symbpcaw_Last = new SgVariableSymb("pcaw_Last", *SgTypeInt());
  Symbpcaw_Step = new SgVariableSymb("pcaw_Step", *SgTypeInt());

  pcaw_First = new SgVarRefExp(*Symbpcaw_First);
  pcaw_Last = new SgVarRefExp(*Symbpcaw_Last);
  pcaw_Step = new SgVarRefExp(*Symbpcaw_Step);
}


/*
 * mark
 */
//
// (was: generate the loop around a statement)
//
//  ** now only marks statements **
//
SgStatement *
MSgenerateCodeForStmt(SgStatement *stmt1, SgStatement *savenext, SgStatement  *previous1, char found, SgExpression *coll)
{
  SgStatement       *stmt, *cp, *copy, *previous;
  SgStatement       *ite, *glassop;
  //SgExpression      *IndexVarExpr;
  SgExpression      *vref, *expr;
  SgExprListExp     *nlist;
  SgFunctionCallExp *fcall;
  SgExpression      *anassign, *collp;
  char              addrneeded;

  if (!stmt1 || !coll || !previous1) 
    return stmt1;

  addrneeded = LookIfAddrNeeded(coll,stmt1);
  previous = previous1;
  stmt = stmt1;
  if (found == 1)
    { // we have to generate the loop;
      if ( Verbose )
        {
          fprintf(stderr,"parallelism activation line %d\n",stmt->lineNumber());
        }
      cp = stmt->controlParent();
      copy = &(EnterPar->copy());
      fcall = new SgFunctionCallExp(*SymbEnterPar);
      if (addrneeded)
        {
          if (isSgPointerDerefExp(coll))
            {
              if (coll->lhs())
                fcall->addArg(coll->lhs()->copy());
            } else
              fcall->addArg(SgAddrOp(coll->copy()));
        }
      else
        fcall->addArg(coll->copy());
      nlist = new SgExprListExp(*fcall);
      copy->setExpression(0,*nlist);
      previous->insertStmtAfter(*copy,*cp);
      previous = copy;

      // create the loop after collection types are resolved later

      copy = &(ExitPar->copy());
      stmt->insertStmtAfter(*copy,*cp);
      stmt = copy;
    } else
      if (found == 2)
        { 
            if ( Verbose )
              {
                fprintf(stderr,"Subset operation found line %d\n",stmt->lineNumber());
              }
          // set previous to  pcxx_1 = FirstInSubset(...)  stmt
          previous = stmt->nodeBefore();

	  // this is what we nedd to take apart:
	  // 2170: stmt->expr(0)->setLhs(SgAssignOp(*IndexVarExpr, *fcal));
	  
	  if(!((anassign = previous->expr(0)) 
	       && (anassign = anassign->lhs()))){
	    Message("FirstInSubset stmt expected in mark-subset", 0); 
	    exit(0); 
	  }
	  if(!(fcall = isSgFunctionCallExp(anassign->rhs()))){
	    Message("FunctionCallExp expected in mark-subset", 0); 
	    exit(0); 
	  }
 

          // pcxx_1 = FirstInSubset(this, lb, ub, step);           
          // arg0 - collection pointer, arg1 - first, arg2 - last, arg3 - step

          vref = new SgVarRefExp(*Symbpcaw_First);
          expr = &(fcall->arg(1)->copy());
          glassop = new SgCExpStmt(SgAssignOp(*vref, *expr));
          previous->insertStmtBefore(*glassop, *previous->controlParent());
 
          vref = new SgVarRefExp(*Symbpcaw_Last);
          expr = &(fcall->arg(2)->copy());
          glassop = new SgCExpStmt(SgAssignOp(*vref, *expr));
          previous->insertStmtBefore(*glassop, *previous->controlParent());

          vref = new SgVarRefExp(*Symbpcaw_Step);
          expr = &(fcall->arg(3)->copy());
          glassop = new SgCExpStmt(SgAssignOp(*vref, *expr));
          previous->insertStmtBefore(*glassop, *previous->controlParent());

          collp = &(fcall->arg(0)->copy());
          fcall = new SgFunctionCallExp(*SymbSubsetEnterPar);
          fcall->addArg(*collp);
          nlist = new SgExprListExp(*fcall);
	  copy = new SgCExpStmt(*nlist);
	  previous->replaceWithStmt(*copy);
	  previous = copy;	

          // create the loop after collection types are resolved later

          copy = &(ExitPar->copy());

	  // find stmt before savenext (savenext is possibly null)
          cp = stmt->controlParent();
	  for(ite = stmt; ite->lexNext() != savenext; ite = ite->lexNext());
          ite->insertStmtAfter(*copy,*cp);
	  // remove NextInSubset stmt (different stmt needed in workerthread)
          ite->deleteStmt();
          stmt = copy;                   
        }
  return stmt;
}

/*
 * generate 
 */

//
// type bug work around
//
SgType *myTrueType(SgExpression *e)
{
  SgPointerDerefExp *pe;
  SgReferenceType *t;

  if (pe = isSgPointerDerefExp(e)) {
    SgPointerType *t = isSgPointerType(myTrueType(pe->pointerExp()));
    return t->baseType();
  }
  else if (e->variant() == ADDRESS_OP) {
    SgType *t = myTrueType(e->lhs());
    return new SgPointerType(*t);
  }
  else if (t = isSgReferenceType(e->type())) {
    return t->baseType();  
  }
  else {
    return e->type();
  }
}

//
//  'declareAVar not completed yet' workaround
//
SgType *substituteType(SgType *t)
{
  SgSymbol *ts;

  switch(t->variant()){
    case T_POINTER:
        return new SgPointerType(*(substituteType(t->baseType())));
    case T_STRUCT:
    case T_CLASS:
    case T_COLLECTION:
    case T_DERIVED_CLASS:
    case T_DERIVED_COLLECTION:
        ts = new SgSymbol(TYPE_NAME, t->symbol()->identifier());
	return new SgDerivedType(*ts);
    default:
        return t;
  }
}

//
// make t a pointer type if it is a collection (or other structure)
// 
int adjustTypeForStructures(SgType **t)
{
  SgPointerType *pt;

  switch((*t)->variant()){
    case T_STRUCT:
    case T_CLASS:
    case T_COLLECTION:
    case T_DERIVED_CLASS:
    case T_DERIVED_TYPE:
    case T_DERIVED_COLLECTION:
    case T_DERIVED_TEMPLATE:
        pt = new SgPointerType(**t);
	*t = pt;
	return 1;
    default:
        return 0;
  }
}

/* 
   Written by Lars (11/9/94)
   Given a type, this procedure returns a Class symbol (i.e., a class
   or collection type name) iff, at the very bottom of its soul, the
   type is a class or collection type. To do this, it attempts to strip
   off all other attributes.
   It's probably not complete.
*/
SgClassSymb *lookForClassOrCollectionSymbol( SgType *type )
{
  SgDescriptType *des;
  SgDerivedType *dt;
  SgArrayType *at;
  SgClassSymb *cs;
  SgDerivedCollectionType *ct;
  SgPointerType *pt;
  SgReferenceType *rt;

  if (type == 0) 
    return 0;
  else if (dt = isSgDerivedType( type )) {
    // class/struct/union X
    if (cs = isSgClassSymb( dt->typeName() ))
      return cs;
    else
      return lookForClassOrCollectionSymbol( dt->typeName()->type() );
  }
  else if (ct = isSgDerivedCollectionType( type ))
    // collection<X>
    return isSgClassSymb( ct->collectionName() );
  else if (at = isSgArrayType( type ))
    // array of X
    return lookForClassOrCollectionSymbol( at->baseType() );
  else if ((des = isSgDescriptType( type )) && (des->modifierFlag() & BIT_TYPEDEF)) {
    // typedef X Y;
    if(des->hasBaseType() && des->baseType()) 
      return lookForClassOrCollectionSymbol( des->baseType() );
    else
      return 0;
  }
  else if (pt = isSgPointerType( type ))
    // pointer to X
    return lookForClassOrCollectionSymbol( pt->baseType() );
  else if (rt = isSgReferenceType( type ))
    // reference to X
    return lookForClassOrCollectionSymbol( rt->baseType() );
  else
    return 0;
}

//
// this function takes a SgClassSymb and checks if it is already
// there in the forwardDeclsTab. If it is not already there,
// it inserts it in the table.
//
int addToForwardDeclsTab(SgClassSymb *cs)
{
  int i;

  for(i=0;i<numForwardDecls;i++)
    if (!strcmp(cs->identifier(), forwardDeclsTab[i]->identifier()))
      return 0;
  if (numForwardDecls == MAXTABSIZE) 
    Message("addToForwardDeclsTab: table full\n", 0);
  forwardDeclsTab[numForwardDecls++] = cs;
  return 1;
}

//
// this function will declare and return a global variable 
// with name pcaw_globali and set it to expr0, 
//
SgExpression *makeGlobalVarRef(SgStatement *previous, SgExpression *expr0)
{
  static int        i=0;
  char              strin[32];
  SgSymbol          *s;
  SgStatement       *scope = previous;
  SgExpression      *e, *expr;
  SgExprListExp     vlist;
  SgType            *t;
  SgStatement       *glassop = NULL;
  int               byReference=0;
  SgClassSymb       *cs;

  sprintf(strin,"pcaw_global%d",++i);
  while(scope && (scope->variant() != GLOBAL))
    scope = scope->controlParent();

  t = myTrueType(expr0);  
  cs = lookForClassOrCollectionSymbol(t); // look for possible forward decls
  if (cs) addToForwardDeclsTab(cs);
  byReference = adjustTypeForStructures(&t);
  t = substituteType(t);  // =*=
  s  = new SgVariableSymb(strin, *t);

  s->declareTheSymbol(*scope);
  SgStatement *newdecl = scope->lexNext();
  if(! isSgVarDeclStmt(newdecl)) 
    Message("error in initializer for makeGlobalVarRef\n", 0);
 

  e = new SgVarRefExp(*s);
  
  // insert assignment
  expr = byReference?&(SgAddrOp(expr0->copy())):&expr0->copy();
  glassop = new SgCExpStmt(SgAssignOp(*e, *expr));
  previous->insertStmtBefore(*glassop, *previous->controlParent());
  
  return byReference?&(SgDerefOp(*e)):e;
}

//
// generate the loop for subset operation
//
//  ** we do this AFTER collections are turned into classes etc. **
//
SgStatement *
generateCodeForSubsetStmt(SgStatement *stmt1, SgStatement *previous, SgExpression *coll, SgFunctionCallExp *enterpar)
{
  SgStatement       *stmt, *tokeep, *nextStmt;
  SgForStmt         *forloop;
  SgExpression      *currentCollection, *psop, *index;
  SgFuncPntrExp     *funop;
  SgFunctionCallExp *funcall, *newfun;
  SgCaseOptionStmt  *acase;
  SgValueExp        *con0, *wid;
  int               i;


  if (!stmt1 || !coll || !previous) 
    return stmt1;

  // get work id and broadcast it 

  wid = new SgValueExp(++workId);
  enterpar->addArg(wid->copy());

  stmt = stmt1;
  nextStmt = stmt->lexNext();
  currentCollection = new SgCastExp(*myTrueType(coll), 
				    CurrentCollectionVar->copy());
  tokeep = stmt->extractStmt();

  // need to massage tokeep:
  //  - replace collection reference by global with correct typecast

  psop = tokeep->expr(0)->lhs();
  if(psop->variant() != POINTST_OP)
    { Message("parallel expression not a POINTST_OP", 0); exit(1); }

  if(!(funop = isSgFuncPntrExp(psop->lhs())))
    { Message("FUNCTION_OP expected as lhs of parallel expression", 0); exit(1); }

  funop->setLhs(*(new SgPointerDerefExp(*currentCollection)));

  //  - replace all args by globals and insert corresponding
  //    assignment statements before previous

  if(!(funcall = isSgFunctionCallExp(psop->rhs())))
    { Message("FUNC_CALL expected as rhs of parallel expression", 0); exit(1); }

  newfun = new SgFunctionCallExp(*funcall->funName());
  for(i=0; i<funcall->numberOfArgs(); i++)
    newfun->addArg(*(makeGlobalVarRef(previous, funcall->arg(i))));

  psop->setRhs(*newfun);
    
  // now generate case stmt for workerthread
/*
  con0 =  new SgValueExp(0);
  index = makeLocalInductionVarRef(1, previous);
  SgFunctionCallExp *newLocal = isSgFunctionCallExp(&(Local->copy()));
  newLocal->addArg(index->copy());
  forloop = new SgForStmt(SgAssignOp(*index,*FirstLocal),
			  *index >= *con0,
			  SgAssignOp(*index, *newLocal), 
			  *tokeep);
*/
////

  con0 =  new SgValueExp(0);
  index = makeLocalInductionVarRef(1, previous);

  SgFunctionCallExp *newFirstInSubset = new SgFunctionCallExp(*SymbFirstInSubset);
  newFirstInSubset->addArg(currentCollection->copy());
  newFirstInSubset->addArg(pcaw_First->copy());
  newFirstInSubset->addArg(pcaw_Last->copy());
  newFirstInSubset->addArg(pcaw_Step->copy());
 
  SgFunctionCallExp *newNextInSubset = new SgFunctionCallExp(*SymbNextInSubset);
  newNextInSubset->addArg(currentCollection->copy());
  newNextInSubset->addArg(pcaw_First->copy());
  newNextInSubset->addArg(pcaw_Last->copy());
  newNextInSubset->addArg(pcaw_Step->copy());
  newNextInSubset->addArg(index->copy());

  SgFunctionCallExp *newCheckLocal = new SgFunctionCallExp(*CheckLocal);
  newCheckLocal->addArg(currentCollection->copy());
  newCheckLocal->addArg(index->copy());

  tokeep = new SgIfStmt(*newCheckLocal, *tokeep);

  forloop = new SgForStmt(SgAssignOp(*index, *newFirstInSubset),
			  *index >= *con0,
			  SgAssignOp(*index, *newNextInSubset),
			  *tokeep);
////
  acase = new SgCaseOptionStmt(
	   *wid, NULL, *(new SgConstructSymb("wc", *parallelSections)));
  parallelSections->addCaseOption(*acase);
  acase->insertStmtAfter(*forloop, *parallelSections);
  forloop->insertStmtAfter(*(new SgBreakStmt), *parallelSections);


  // return last stmt
  return nextStmt;
}


//
// generate the loop around a statement
//
//  ** we do this AFTER collections are turned into classes etc. **
//
SgStatement *
generateCodeForParStmt(SgStatement *stmt1, SgStatement *previous, SgExpression *coll, SgFunctionCallExp *enterpar)
{
  SgStatement       *stmt, *tokeep, *nextStmt;
  SgForStmt         *forloop;
  SgExpression      *currentCollection, *psop, *index;
  SgFuncPntrExp     *funop;
  SgFunctionCallExp *funcall, *newfun;
  SgCaseOptionStmt  *acase;
  SgValueExp        *con0, *wid;
  int               i;


  if (!stmt1 || !coll || !previous) 
    return stmt1;

  // get work id and broadcast it 

  wid = new SgValueExp(++workId);
  enterpar->addArg(wid->copy());

  stmt = stmt1;
  nextStmt = stmt->lexNext();
  currentCollection = new SgCastExp(*myTrueType(coll), 
				    CurrentCollectionVar->copy());
  tokeep = stmt->extractStmt();

  // need to massage tokeep:
  //  - replace collection reference by global with correct typecast

  psop = tokeep->expr(0)->lhs();
  if(psop->variant() != POINTST_OP)
    { Message("parallel expression not a POINTST_OP", 0); exit(1); }

  if(!(funop = isSgFuncPntrExp(psop->lhs())))
    { Message("FUNCTION_OP expected as lhs of parallel expression", 0); exit(1); }

  funop->setLhs(*(new SgPointerDerefExp(*currentCollection)));

  //  - replace all args by globals and insert corresponding
  //    assignment statements before previous

  if(!(funcall = isSgFunctionCallExp(psop->rhs())))
    { Message("FUNC_CALL expected as rhs of parallel expression", 0); exit(1); }

  newfun = new SgFunctionCallExp(*funcall->funName());
  for(i=0; i<funcall->numberOfArgs(); i++)
    newfun->addArg(*(makeGlobalVarRef(previous, funcall->arg(i))));

  psop->setRhs(*newfun);
    
  // now generate case stmt for workerthread

  con0 =  new SgValueExp(0);
  index = makeLocalInductionVarRef(0, previous);
  SgFunctionCallExp *newLocal = isSgFunctionCallExp(&(Local->copy()));
  newLocal->addArg(index->copy());
  forloop = new SgForStmt(SgAssignOp(*index,*FirstLocal),
			  *index >= *con0,
			  SgAssignOp(*index, *newLocal), 
			  *tokeep);
  acase = new SgCaseOptionStmt(
	   *wid, NULL, *(new SgConstructSymb("wc", *parallelSections)));
  parallelSections->addCaseOption(*acase);
  acase->insertStmtAfter(*forloop, *parallelSections);
  forloop->insertStmtAfter(*(new SgBreakStmt), *parallelSections);


  // return last stmt
  return nextStmt;
}


//
// find stmts enclosed by EnterPar/ExitPar stmts and 
// move them to the workerthreads main() method
//
void
treatParallelSections(SgFile *file)
{
  SgFunctionCallExp *fcall;
  SgExprListExp *l;
  SgStatement *stmt, *first;
  SgExpression *coll;
  int i;
 
  if (!file) return;

  for(stmt = file->firstStatement(); stmt; stmt = stmt->lexNext()) {
    if(isSgCExpStmt(stmt)){
      for(l = isSgExprListExp(stmt->expr(0)); l; l = l->next()) {
	if(fcall = isSgFunctionCallExp(l->value())){
	  if(!strcmp(fcall->funName()->identifier(), 
		     SymbEnterPar->identifier())){  
  	    // next stmt is a parallel MoE invocation
	    if(!stmt->lexNext())
	      { Message("lonely EnterPar() encountered", 0); exit(1); }
	    if(!(coll = fcall->arg(0)))
	      { Message("argument to EnterPar() missing", 0); exit(1); }
	    stmt = generateCodeForParStmt(stmt->lexNext(), stmt, coll, fcall);
	  }
	  else if(!strcmp(fcall->funName()->identifier(), 
			  SymbSubsetEnterPar->identifier())){
	    // next stmts are subset operations
	    if(!stmt->lexNext())
	      { Message("lonely EnterSubset() encountered", 0); exit(1); }
	    if(!(coll = fcall->arg(0)))
	      { Message("argument to EnterSubset() missing", 0); exit(1); }
	    stmt = generateCodeForSubsetStmt(stmt->lexNext(), stmt, coll, fcall);
	  }
	}
      }
    }
  }	

  first = file->firstStatement();
  for(i=0;i<numForwardDecls;i++){
     SgStatement * stst = new SgClassStmt(*(forwardDeclsTab[i]));
     first->lexNext()->insertStmtBefore(* stst);
  }
  Symbpcaw_First->declareTheSymbol(*first);
  Symbpcaw_Last->declareTheSymbol(*first);
  Symbpcaw_Step->declareTheSymbol(*first);
}

/*
 * make worker switch
 */

//
//  generates the function around the workers' switch stmt
//
void makeWorkerSwitch(SgFile *f)
{
  SgStatement    *first, *last;
  SgSymbol       *pcxx_0, *pcxx_1;
  SgFunctionSymb *funsymb;
  SgStatement    *thefun;

  if (!f) return;

  first = f->firstStatement();
  for(last = first; last->lexNext(); last = last->lexNext());
  pcxx_0 = new SgVariableSymb("pcxx_0");
  pcxx_1 = new SgVariableSymb("pcxx_1");

// add/insertParameter 'not implemented yet' workaround,  what a hack !! 
  funsymb = new SgFunctionSymb(FUNCTION_NAME,
			 "WorkerSwitch(int pcaw_cf, char *pcaw_cc){  // ",
			 *SgTypeVoid(),*first);

  thefun = new SgFuncHedrStmt(*funsymb, *parallelSections);
  pcxx_0->declareTheSymbol(*thefun);
  pcxx_1->declareTheSymbol(*thefun);

  last->insertStmtAfter(*thefun, *first);
}
