/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * main.c --- Calls parser, dependence graph analyzer, subscript analyzer.
 * Opens and closes the files that are needed. 
 */

#include <stdio.h>
#include <stdlib.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

#include <fcntl.h>
#include "db.h"
#include "defines.h"
#include "version.h"

typedef FILE	*FILEP;

extern int	yylineno, yyleng, yylisting, yylonglines, yydebug;
extern int      parstate;               
extern int	errcnt,	prflag, errline, num_files;
extern PTR_BFND global_bfnd;
extern PTR_BLOB head_blob;
extern PTR_FNAME head_file,cur_thread_file;
extern int     num_bfnds;		/* total # of bif nodes */
extern int     num_llnds;		/* total # of low level nodes */
extern int     num_symbs;		/* total # of symbol nodes */
extern int     num_types;		/* total # of types nodes */
extern int     num_blobs;		/* total # of blob nodes */
extern int     num_sets;		/* total # of set nodes */
extern int     num_cmnt;
extern int     num_def;		/* total # of dependncy nodes */
extern int     num_dep;
extern int     num_label;		/* total # of label nodes */
extern PTR_BFND head_bfnd,	/* start of bfnd chain */
	        cur_bfnd;	/* poextern int to current bfnd */

extern PTR_LLND head_llnd, cur_llnd;
extern PTR_SYMB head_symb, cur_symb;
extern PTR_TYPE head_type, cur_type;
extern PTR_LABEL head_label, cur_label;
extern PTR_FNAME head_file,cur_thread_file;
extern PTR_BLOB head_blob, cur_blob;
extern PTR_SETS head_sets, cur_sets;
extern PTR_DEF head_def, cur_def;
extern PTR_DEFLST head_deflst, cur_deflst;
extern PTR_DEP head_dep, cur_dep, pre_dep;
PTR_CMNT head_cmnt, cur_cmnt;
extern PTR_CMNT cur_comment;
PTR_TYPE global_unknown = NULL;

int _filbuf();
PTR_SETS build_sets();
int relink();
int write_nodes();
int fclose();
int inilex();
int yyparse();

PTR_BFND get_bfnd();
void fatalstr(), initialize(), errstr(), release_nodes(), close_files();
void make_file_list();
char *copys();
char *getenv();
int system();
void free();
void exit();


int	language = ForSrc;
FILE	*outf;
static	char *defeditor = "/usr/ucb/vi +%d %s";
int show_deps = 0; /* set to 1 by -sd flag */
int garb_collect = 1; /* disabled by -c flag */
static int dep = 0;
char *outname;
PTR_FILE fi;
extern PTR_FILE cur_file;


void
gen_out_name (char *filename)
{
   register int    i;
   char *p = filename,
	*q, *r;

   if (!filename)
     return;

   
   outname = (char *) malloc((unsigned)(strlen(filename)+4));

   strcpy (outname,filename);
   
   for (i = strlen(filename)-1 ; i >= 0 ; i --)
   {
        if ( filename[i] == '.' )
        {
             q = &(outname[i+1]);
             break;
        }
   }

#if 0 /* OBSOLETE */
   for(i = strlen(p), q = outname; i; i--)
      if ((*q++ = *p++) == '.') break;
#endif /* OBSOLETE */

   r = q;
   if (yylisting) {
      *q++ = 'l';
      *q++ = 's';
      *q++ = 't';
      *q = '\0';
      if ((outf = fopen(outname, "w")) == NULL)
	 fatalstr("Cant Open File %s for Write\n", outname);
   }
   *r++ = 'd';
   *r++ = 'e';
   *r++ = 'p';
}

void
main(int argc, char *argv[])
{
   int k;
   int fromfile = (argv != 0);	/* Flag to see if read from a file */
   void procinit();
   char *input_file;
   int gendeps();
   int collect_garbage();

   argv++;
   while ((argc > 1) && (*argv)[0] == '-'){
      if (!strcmp(argv[0],"-d"))
	 yydebug = 1;
      else if (!strcmp(argv[0],"-sd")){
	 (void)printf(" show dependence mode \n");
	 show_deps = 1;
      } else if (!strcmp(argv[0],"-dep"))
	 dep = 1;
      else if (!strcmp(argv[0],"-l"))
	 yylisting = 1; 
      else if (!strcmp(argv[0],"-c"))
	 garb_collect = 0;
      else
	 (void)fprintf (stderr, "Unkown switch %s ignored\n", argv[0]);
      argc--;
      argv++;
   }
   
 retry:	if (*argv)
    gen_out_name(input_file = *argv);
 else
    input_file = "stdin";
   
   yylonglines = 1;
   prflag = 1;
   cur_file = fi = (PTR_FILE) calloc(1, (unsigned)sizeof(struct file_obj));
   fi->lang = language;
   initialize();
   make_file_list(input_file);
   
   if (inilex(copys(*argv ? *argv : ""))) {
      (void)printf("Error in inilex\n");
      goto finish;
   }
   procinit();
   printf("|...............................................|\n");
   printf("      Sage/Indiana F2DEP version %s\n",VERSION_NUMBER);
   printf("|...............................................|\n");
   /* parsing */
   if (k = yyparse()) {
      (void)printf("Bad parse, return code %d\n", k);
      goto finish;
   }
   if(parstate != OUTSIDE){
     (void) errstr("cftn.c:main:missing final end statement");
     goto finish;
   }

   global_bfnd->filename = cur_thread_file;
   if (errcnt) {
      (void)fprintf(stderr, "%d error(s)\n", errcnt);
      if (fromfile) {
	 int ans;
	 
	 (void)printf("Do you want to invoke editor? [y/n] ");
	 while ((ans = getchar()) == '\n');
	 if (ans == 'y' || ans == 'Y') {
	    char *edtcmd;
	    char cmd[50];
	    
	    edtcmd = getenv("FOREDIT");
	    if (!edtcmd) edtcmd = defeditor;
	    (void)sprintf(cmd, edtcmd, errline, *argv);
	    (void)system(cmd);
	    release_nodes();
	    close_files();
	    goto retry;
	 }
	 exit(1);
      }
   }
   
   if (dep) {
      (void)build_sets(0, global_bfnd,NULL,NULL,1); 
      (void)build_sets(0, global_bfnd,NULL,NULL,2);
      (void)gendeps(global_bfnd); 
      (void)relink(fi);
      printf("garbage collecting\n");
      
      if (garb_collect) (void)collect_garbage(fi);
   }

   if (yylisting)
      (void)fprintf(outf,"$f %s\n",input_file); 
   write_nodes(cur_file, outname); 
 finish:
   exit(0);
}


void 
clf(FILEP *p)
{
   void fatal();
   
   if (p && *p && *p != stdout) {
      if (ferror(*p))
	 fatal("writing error");
      (void)fclose(*p);
   }
   *p = (FILEP)NULL;
}


void
make_file_list(char *filename)
{
   char  *namep = (char *) malloc((unsigned)(strlen(filename)+1));
   PTR_FNAME thisF = (PTR_FNAME) malloc((unsigned)sizeof(struct file_name));
   PTR_FNAME p;
   
   (void)strcpy(namep, filename);
   thisF->id = fi->num_files = ++num_files;
   thisF->name = namep;
   thisF->next = (PTR_FNAME)NULL;
   if (head_file == (PTR_FNAME)NULL)
      fi->head_file = head_file = thisF;
   else {
      for (p = head_file; p->next != NULL; p = p->next)
	 ;
      p->next = thisF;
   }
   cur_thread_file = thisF;
}
