/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * errors.c
 *
 * Miscellaneous error routines
 */

#include <stdio.h>
#include "defs.h"
#include "symb.h"
#include "extern.h"


static char buff[100];

/*
 * fatal -- print the fatal error message then exit
 *
 * input:
 *	  s - the message to be printed out
 */
void 
fatal(char *s)
{
   (void)fprintf(stderr, "Compiler error line %d of %s: %s\n", yylineno, infname, s);
	exit(3);
}


/*
 * fatalstr -- sets up the error message according to the given format
 *		then call "fatal" to print it out
 *
 * input:
 *	  t - a string that specifies the output format
 *	  s - a string that contents the error messaged to be formatted
 */
void 
fatalstr(char *t, char *s)
{
	(void)sprintf(buff, t, s);
	fatal(buff);
}


/*
 * fatali -- formats a fatal error message which contains a number
 *	     then call "fatal" to print it
 *
 * input:
 *	  t - a string that specifies the output format
 *	  d - an integer to be converted according to t
 */
void 
fatali(char *t, int d)
{
	(void)sprintf(buff, t, d);
	fatal(buff);
}


/*
 * warn1 -- formats a warning message then call "warn" to print it out
 *
 * input:
 *	  s - string that specifies the conversion format
 *	  t - string that to be converted according to s
 */
void 
warn1(char *s, char *t)
{
   void warn();

   (void)sprintf(buff, s, t);
   warn(buff);
}


/*
 * warn -- print the warning message if specified
 *
 * input:
 *	  s - string to be printed
 */
void
warn(char *s)
{
   if (!nowarnflag) {
      ++nwarn;
      (void)fprintf(stderr, "Warning on line %d of %s: %s\n", yylineno, infname, s);
   }
}


/*
 * errstr -- formats the non-fatal error message then call "err" to print it
 *
 * input:
 *	  s - string that specifies the conversion format
 *	  t - string that to be formated according to s
 */
void 
errstr(char *s, char *t)
{
   void err();

   (void)sprintf(buff, s, t);
   err(buff);
}


/*
 * erri -- formats an error number then prints it out
 *
 * input:
 *	  s - string that specifies the output format
 *	  t - error number to be formatted
 */
void
erri(char *s, int t)
{
   void err();

   (void)sprintf(buff, s, t);
   err(buff);
}


/*
 * err -- prints the error message and does the bookkeeping
 *
 * input:
 *	  s - string to be printed out
 */
void
err(char *s)
{
	if (! errcnt++)
		errline = yylineno;
	(void)fprintf(stderr, "Error on line %d of %s: %s\n", yylineno, infname, s);
/*	exit(3);*/
}


/*
 * yyerror -- the error handling routine called by yacc
 *
 * input:
 *	  s - the error message to be printed
 */
void 
yyerror(char *s)
{
	err(s);
}


/*
 * dclerr -- prints the error message when find error in declaration part
 *
 * input:
 *	  s - error message string
 *	  v - pointer to the symble table entry
 */
void 
dclerr(char *s, PTR_SYMB v)
{
	char    buf[100];

	if (v) {
		(void)sprintf(buf,"Declaration error for %s: %s",v->ident, s);
		err(buf);
	} else
		errstr("Declaration error: %s", s);
}


/*
 * execerr -- prints error message for executable part
 *
 * input:
 *	  s - the error message string1
 *	  n - the error message string2
 */
void 
execerr(char *s, char *n)
{
	char    buf1[100], buf2[100];

	(void)sprintf(buf1, "Execution error: %s", s);
	(void)sprintf(buf2, buf1, n);
	err(buf2);
}

