/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/*
 * exec.c
 *
 * Routines for handling the semantics of control structures.
 * 
 */

#include "defs.h"


/*
 *  pushctl  --  push a control node of type "code" to control stack
 */
void
pushctl(int code)
{
	if (++ctlsp >= lastctl)
		many("loops or if-then-elses", 'c');
	ctlsp->ctltype = code;
	++blklevel;
}


/*
 *  popctl -- pop up a control node from control stack
 */
void
popctl()
{
	if (ctlsp-- < ctls)
		fatal("control stack empty");
	--blklevel;
}


/*
 *  poplab either marks all labels in inner blocks unreachable
 *    or moves all labels referred to in inner blocks out a level
 */
void
poplab()
{
	register PTR_LABEL lp;

	for (lp = head_label; lp; lp = lp->next)
		if (lp->labdefined) {
		/* mark all labels in inner blocks unreachable */
			if (lp->blklevel > blklevel)
				lp->labinacc = YES;
		} else if (lp->blklevel > blklevel) {
		/* move all labels referred to in inner blocks out a level */
			lp->blklevel = blklevel;
		}
}

/*
 *   Logical IF codes
 */
void
exif(expptr p)
{
	register int k;

	(void) pushctl(CTLIF);

	if ((k = (p = fixtype(p))->headblock.vtype) != TYLOGICAL) {
		if (k != TYERROR)
			err("non-logical expression in IF statement");
		frexpr(p);
	}
}


void
exelif(expptr p)
{
	int     k, oldelse;

	if ((k = p->headblock.vtype) != TYLOGICAL) {
		if (k != TYERROR)
			err("non-logical expression in IF statement");
	} else {
		if (ctlsp->ctltype == CTLIF) {
		} else
			execerr("elseif out of place", CNULL);
	}
}


void
exelse()
{
	if (ctlsp->ctltype == CTLIF)
		ctlsp->ctltype = CTLELSE;
	else
		execerr("else out of place", CNULL);
}


void
exendif()
{
	if (ctlsp->ctltype == CTLIF || ctlsp->ctltype == CTLELSE)
		(void) popctl();
	else
		execerr("endif out of place", CNULL);
}


void
exequals(register struct Primblock *lp, register expptr rp)
{
	register Namep np;

	if (lp->tag != TPRIM) {
		err("assignment to a non-variable");
		frexpr(lp);
		frexpr(rp);
	} else if (lp->namep->vclass != CLVAR && lp->argsp) {
		if (parstate >= INEXEC)
			err("undimensioned array or statement function out of order");
		else
			mkstfunct(lp, rp);
	} else {
		np = (Namep) lp->namep;
		if (np->vclass == CLPROC && np->vprocclass == PTHISPROC
		    && proctype == TYSUBR) {
			err("assignment to a subroutine name");
			return;
		}
		if (parstate < INDATA)
			enddcl();
		parstate = INEXEC;
	}
}


void
mkstfunct(struct Primblock *lp, expptr rp)
{
	register struct Primblock *p;
	register Namep np;
	chainp  args;

	if (parstate < INDATA) {
		enddcl();
		parstate = INDATA;
	}
	np = lp->namep;
	if (np->vclass == CLUNKNOWN)
		np->vclass = CLPROC;
	else {
		dclerr("redeclaration of statement function", np);
		return;
	}
	np->vprocclass = PSTFUNCT;
	np->vstg = STGSTFUNCT;
	impldcl(np);
	args = (lp->argsp ? lp->argsp->listp : CHNULL);
	np->varxptr.vstfdesc = mkchain(args, rp);

	for (; args; args = args->nextp)
		if (args->datap->tag != TPRIM ||
		    (p = (struct Primblock *) (args->datap))->argsp ||
		    p->fcharp || p->lcharp)
			err("non-variable argument in statement function definition");
		else {
			args->datap = (tagptr) (p->namep);
			vardcl(p->namep);
			free(p);
		}
}


void
excall(Namep name, struct Listblock *args, int nstars, struct Labelblock *labels)
{
	register expptr p;

	if (name->vdcldone)
		if (name->vclass != CLPROC && name->vclass != CLENTRY) {
			dclerr("call to non-subroutine", name);
			return;
		} else if (name->vtype != TYSUBR) {
			dclerr("subroutine invocation of function", name);
			return;
		}
	settype(name, TYSUBR, ENULL);
	p = mkfunct(mkprim(name, args, CHNULL));
	p->exprblock.vtype = p->exprblock.leftp->headblock.vtype = TYINT;
	if (nstars > 0)
		if (optimflag)
			optbuff(SKCMGOTO, p, nstars, labels);
		else
			putcmgo(p, nstars, labels);
	else if (optimflag)
		optbuff(SKCALL, p, 0, 0);
	else
		putexpr(p);
}


void
exstop(int stop, register expptr p)
{
	char   *q;
	int     n;
	expptr  mkstrcon();

	if (p) {
		if (!ISCONST(p)) {
			execerr("pause/stop argument must be constant", CNULL);
			frexpr(p);
			p = mkstrcon(0, CNULL);
		} else if (ISINT(p->constblock.vtype)) {
			q = convic(p->constblock.const.ci);
			n = strlen(q);
			if (n > 0) {
				p->constblock.const.ccp = copyn(n, q);
				p->constblock.vtype = TYCHAR;
				p->constblock.vleng = (expptr) ICON(n);
			} else
				p = (expptr) mkstrcon(0, CNULL);
		} else if (p->constblock.vtype != TYCHAR) {
			execerr("pause/stop argument must be integer or string", CNULL);
			p = (expptr) mkstrcon(0, CNULL);
		}
	} else
		p = (expptr) mkstrcon(0, CNULL);

	if (optimflag)
		optbuff((stop ? SKSTOP : SKPAUSE), p, 0, 0);
	else
		putexpr(call1(TYSUBR, (stop ? "s_stop" : "s_paus"), p));
}

/* DO LOOP CODE */

#define DOINIT	par[0]
#define DOLIMIT	par[1]
#define DOINCR	par[2]

#define CONSTINIT  const[0]
#define CONSTLIMIT const[1]
#define CONSTINCR  const[2]

#define VARSTEP	0
#define POSSTEP	1
#define NEGSTEP	2


void
exdo(int range, chainp spec)
{
	register expptr p, q;
	expptr  q1;
	register Namep np;
	chainp  cp;
	register int i;
	int     dotype, incsign;
	Addrp   dovarp, dostgp;
	expptr  par[3];
	expptr  const[3];
	Slotp   doslot;

	(void) pushctl(CTLDO);
	dorange = ctlsp->dolabel = range;
	np = (Namep) (spec->datap);
	ctlsp->donamep = NULL;
	if (np->vdovar) {
		errstr("nested loops with variable %s", varstr(VL, np->varname));
		return;
	}
	dovarp = mkplace(np);
	dotype = dovarp->vtype;

	if (!ONEOF(dotype, MSKINT | MSKREAL)) {
		err("bad type on DO variable");
		return;
	}
	for (i = 0, cp = spec->nextp; cp != NULL && i < 3; cp = cp->nextp) {
		p = fixtype((expptr) cpexpr((tagptr) q = cp->datap));
		if (!ONEOF(p->headblock.vtype, MSKINT | MSKREAL)) {
			err("bad type on DO parameter");
			return;
		}
		if (ISCONST(q))
			const[i] = mkconv(dotype, q);
		else {
			frexpr(q);
			const[i] = NULL;
		}

		par[i++] = mkconv(dotype, p);
	}

	frchain(&spec);
	switch (i) {
	    case 0:
	    case 1:
		err("too few DO parameters");
		return;

	    case 2:
		DOINCR = (expptr) ICON(1);
		CONSTINCR = ICON(1);

	    case 3:
		break;

	    default:
		err("too many DO parameters");
		return;
	}

	ctlsp->donamep = np;

	np->vdovar = YES;
	if (!optimflag && enregister(np)) {
	/* stgp points to a storage version, varp to a register version */
		dostgp = dovarp;
		dovarp = mkplace(np);
	} else
		dostgp = NULL;

	for (i = 0; i < 4; i++)
		ctlsp->ctlabels[i] = newlabel();

	if (CONSTLIMIT)
		ctlsp->domax = DOLIMIT;
	else
		ctlsp->domax = (expptr) mktemp(dotype, PNULL);

	if (CONSTINCR) {
		ctlsp->dostep = DOINCR;
		if ((incsign = conssgn(CONSTINCR)) == 0)
			err("zero DO increment");
		ctlsp->dostepsign = (incsign > 0 ? POSSTEP : NEGSTEP);
	} else {
		ctlsp->dostep = (expptr) mktemp(dotype, PNULL);
		ctlsp->dostepsign = VARSTEP;
	}
}


void
enddo(int here)
{
	register struct Ctlframe *q;
	Namep   np;
	Addrp   ap, rv;
	expptr  t;
	register int i;
	Slotp   doslot;

	while (here == dorange) {
		while (ctlsp->ctltype == CTLIF || ctlsp->ctltype == CTLELSE) {
			execerr("missing endif", CNULL);
			exendif();
		}

		(void) popctl();
		(void) poplab();

		dorange = 0;
		for (q = ctlsp; q >= ctls; --q)
			if (q->ctltype == CTLDO) {
				dorange = q->dolabel;
				break;
			}
	}
}
