/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/****************************************************************************

DESCRIPTION:	This is the Cedar Fortran lexical scanner.  It reads an input
		Cedar Fortran program and returns lexemes to the Cedar 
		Fortran Preprocessor.  It is based on the Unix 4.2 BSD f77
		lexical scanner.
----------------------------------------------------------------------------
CALLING FORMAT: yylex() 
----------------------------------------------------------------------------
ARGUMENTS:	<none>
----------------------------------------------------------------------------
OUTPUTS:	<return param>: integer lexeme value 
----------------------------------------------------------------------------
SIDE EFFECTS:	1].  A listing will be written to standard output based on
		     the value of global variable "yylisting".
		2].  The variable "yylineno" will be changed to reflect the
		     current line number.
		3].  The variables "yytext" and "yyleng" will be changed to
		     hold the lexeme text and its length. 
----------------------------------------------------------------------------
EXTERNAL REFS:	1].  file: <strings.h>
		2].  file: <sys/file.h>
		3].  file: <stdio.h>
		4].  file: <ctype.h>
		5].  file: defs.h
		6].  file: extern.h
		7].  file: tokdefs.h
		8].  extern int	 debug; ! when debug=0: D in col 1 is comment
----------------------------------------------------------------------------
INVOKED BY:	1].  yyparse()
----------------------------------------------------------------------------
GLOBALS:	1].  char yytext[];	! lexeme text
		2].  int  yyleng;	! length of lexeme text (in yytext)
		3].  int  yylineno;	! current line number
		4].  int  yyi4;		! =4 default/ =2 if OPTIONS/NOI4 used
		5].  int  yylisting;	! when yylisting=1: listing on stdout
		6].  int  yylonglines;	! when yylonglines=0: cols >73 comment
---------------------------------------------------------------------------- */

#include <string.h>
#include <sys/file.h>
#include <stdio.h>
#include <ctype.h>
#include <memory.h>
#include "defs.h"
#include "extern.h"
#include "tokdefs.h"

void fatal();
void fatali();
void fatalstr();
int chkalloc();
int getcd();
int _filbuf();
int randomio();
int eqn();
int getkwd();
FILE *outf;
static void flush_comments();
static void store_comment();
static void store_comment_end_permbuff();
static void move_comments_to_permbuffer();

/* Scanner constants */

#define XEOF		0

/* Character definitions */

#define BLANKC		' '
#define MYQUOTE		(2)
#define MYHOLLERITH	(3)
#define BADHOLLERITH	(4)
#define NEWLINE		10

/* input Fortran card types */

#define STEOF		1
#define STINITIAL	2
#define STCONTINUE	3
#define CMT             4

/* lexical scanner states */

#define NEWSTMT		1
#define FIRSTTOKEN	2
#define OTHERTOKEN	3
#define RETEOS		4
#define COMMENTS	5
#define COMMENTSEOS     6
#define OVER            7

/* Declared limits */

#define MAX_NAME_LEN		29	/* max length of Fortran name */
#define MAXINCLUDES		8	/* max number of include levels */
#define MAXPARLEV		100	/* max parenthesis nesting level */
#define MAX_CONTIN_CARDS	99	/* max number of continuation lines */
#define NCHARS_PER_LINE		66	/* max chars/line without longlines */
#define BUFFERSIZE		1024	/* number bytes read in one read */
#define COMMENT_LENGTH          1000    /* Size of comment buffer */

/* is_equal added for FORTRAN M */
#define is_equal(x) (((x)=='='))


char   *infname;

/* Variables shared with calling routine */

int     yylisting = 1;		/* flag: =1: make listing on stdout */
int     yylonglines = 0;	/* flag: =0: col >72 for comments */
int     yyleng;			/* length of text in yytext */
int     yylineno;		/* source line num : current lexeme */
long    yystno = 0;		/* statement number */

int	freeformat = 0;		/* scanner accepts F90 free format */
                                /* when format=1. Added by cychen */

/* Globals */

static FILE *infile = {stdin};	/* file descriptor num: current file */

static char *nextch;		/* next to be considered in stmtbuf */
static char *lastch;		/* last char in this stmt in stmtbuf */
static char *nextcd = (char *)NULL;	/* place for next card in stmtbuf */
static char *endcd;		/* current end-of-card in stmtbuf */
static char *newname = (char *)NULL;	/* pointer to string */

static int lineno;
static int prevlin;
static int thislin;
static int stkey;		/* lexeme for overall type of stmt */
static int lastend = 1;
static int parlev;		/* current parenthesis level */
static int lexstate = NEWSTMT;	/* state within yylex */
/*
 * Comments need special care when * the last processed statement is
 * a INCLUDE statement --- Ming Q. Xu
 */
int last_stmt_is_include = 0;

/* Added by cychen by supporting F90 free format */
static int last_c = ' ';

static int nincl = 0;		/* current number of include levels */

/* Globals for: getcds/getcd/list_line/is_include/xgetc/unxgetc */

static int code = 0;		/* result of getcd (STEOF,STINITIAL,STCONTINUE) */
static char prefix[6];		/* 6 chars of Fortran stmt prefix */
static char *prefixend = prefix + 6;	/* end of prefix */
static int lines_returned = 0;	/* num lines read in getcd */


/* Globals for: crunch/analyz */

static int expcom;		/* flag: =1: exposed , found */
static int expeql;		/* flag: =1: exposed = found */
static int expdcolon;           /* flag: =1; exposed :: found */
static int expscolon;           /* flag: =1; exposed : found */
static int expeqlt;             /* flag: =1; exposed => found */

/* Statement numbers and the statement buffer */

static long int stno;		/* stmt number for this Fortran stmt */
static long int nxtstno;	/* stmt number for next Fortran stmt */
char stmtbuf[NCHARS_PER_LINE * (MAX_CONTIN_CARDS + 1)];
				/* this Fortran stmt + 1 more card */
char *commentbuf;
static char tempbuf[COMMENT_LENGTH];
static char *stmtend = stmtbuf + (MAX_CONTIN_CARDS * NCHARS_PER_LINE);
				/* max addr for this Fortran stmt */
static char crunchbuf[NCHARS_PER_LINE * (MAX_CONTIN_CARDS + 1)];
/* "Crunch"ed statement is placed in crunchbuf so that unmodified stmtbuf
    is avialable to the parser                                             */
#define COMMENT_BUF_STORE 161

typedef struct comment_buf {
	struct comment_buf *next;
	char *last;
	char buf[COMMENT_BUF_STORE];
	} comment_buf;
static struct comment_buf *cbfirst, *cblast, *pcbfirst, *pcblast;

/* Comment buffering data

	Comments are kept in a list until the statement before them has
   been parsed.  This list is implemented with the above comment_buf
   structure and the pointers cbnext and cblast.

	The comments are stored with terminating NULL, and no other
   intervening space.  The last few bytes of each block are likely to
   remain unused.
*/


/* INCLUDE file information block */

struct Inclfile {
	struct Inclfile *inclnext;	/* pointer to next higher block */
	FILE   *inclfp;		/* file descriptor number for file */
	char   *inclname;	/* filename of this file */
	int     incllno;
	char   *incllinp;	/* point to saved nextcd when doing include */
	int     incllen;	/* length of the saved card */
	int     inclcode;	/* saved status when read nextcd */
	int     inclstno;	/* saved statement number */
	char   *prefix;		/* save columns 1-6 for later listing */
};

struct Inclfile *inclp = (struct Inclfile *)NULL;	/* pointer to current INCLUDE block */


/* Valid punctuation */

struct Punctlist {
	char    punchar;
	int     punval;
} puncts[] = {
	{'(', LEFTPAR},
	{'<', LEFTAB},
	{'>', RIGHTAB},
	{')', RIGHTPAR},
	{'=', EQUAL},
	{',', COMMA},
	{'+', PLUS},
	{'-', MINUS},
	{'*', ASTER},
	{'&', AMPERSAND},
	{'/', SLASH},
	{':', COLON},
	{'\'', QUOTE},
	{'"', DQUOTE},
        {'%', PERCENT},
	{0, 0}
};


/* Valid relational operators */

struct Dotlist {
	char   *dotname;
	int     dotval;
} dots[] = {
	{"and.", AND},
	{"or.", OR},
	{"xor.", XOR},
	{"not.", NOT},
	{"true.", TTRUE},
	{"false.", FFALSE},
	{"eq.", EQ},
	{"ne.", NE},
	{"lt.", LT},
	{"le.", LE},
	{"gt.", GT},
	{"ge.", GE},
	{"neqv.", NEQV},
	{"eqv.", EQV},
	{0, 0}
};


/* Lists of keywords:  Keylist for statement keywords */

struct Keylist {
	char   *keyname;
	int     keyval;
};
struct Keylist *keystart[26], *keyend[26];	/* ptrs: start/end places for
						 * each letter of alphabet */

struct Keylist keys[] = {
	{"accept", ACCEPT},
	{"align", ALIGN},
        {"allocatable", ALLOCATABLE},
        {"allocate", ALLOCATE},
	{"assignment", ASSIGNMENT},
	{"assign", ASSIGN},
/* Automatic support --- Ming Q. Xu */
        {"automatic", AUTOMATIC},
	{"backspace", BACKSPACE},
        {"block_cyclic", BLOCK_CYCLIC},
	{"blockdata", BLOCKDATA},
	{"byte", BYTE},
	{"call", CALL},
        {"casedefault", DEFAULT_CASE},	
	{"case", CASE},
        {"channel", CHANNEL},
	{"character", CHARACTER},
/*	{"cdoacross", CDOACROSS}, */
/*	{"cdoall", CDOALL}, */
        {"clear", CLEAR},
        {"close", CLOSE},
	{"cluster", CLUSTER},
	{"common", COMMON},
	{"complex", COMPLEX},
	{"concurrent", CONCURRENT},
	{"contains", CONTAINS},
	{"continue", CONTINUE},
        {"criticalsection", CRITICALSECTION},
	{"cycle", CYCLE},
	{"data", DATA},
        {"deallocate", DEALLOCATE},
/*	{ "decode", DECODE }, */
	{"decomposition", DECOMPOSITION},
	{"default", DEFAULT_CASE},
	{"define", DEFINE},
	{"dimension", DIMENSION},
	{"distribute", DISTRIBUTE},
/*	{"doacross", DOACROSS}, */
	{"doubleprecision", DOUBLEPRECISION},
	{"doublecomplex", DOUBLECOMPLEX},
	{"do", DOWHILE},
	{"elseif", ELSEIF},
	{"elsewhere", ELSEWHERE},
	{"else", ELSE},
        {"empty", EMPTY},
/*	{ "encode",  ENCODE }, */
	{"endcdoacross", ENDCDOACROSS},
	{"endcdoall", ENDCDOALL},
        {"endchannel", ENDCHANNEL},
	{"endcloop", ENDCLOOP},
	{"endconcurrent", ENDCONCURRENT},
        {"endcriticalsection", ENDCRITICALSECTION},
	{"enddoacross", ENDDOACROSS},
	{"enddo", ENDDO},
        {"endextended", ENDEXTENDED},
        {"endextend", ENDEXTEND},
	{"endfile", ENDFILE},
	{"endfunction", ENDFUNCTION},
	{"endif", ENDIF},
        {"endinterface", ENDINTERFACE},
        {"endmodule", ENDMODULE},
        {"endparalleldo", ENDPARALLELDO},
        {"endparallelsections", ENDPARALLELSECTIONS},
        {"endparallel", ENDPARALLEL}, 
        {"endpdo", ENDPDO},
        {"endprocessdo", ENDPROCESSDO},
        {"endprocesses", ENDPROCESSES},
        {"endprogram", ENDPROGRAM},
        {"endpsections", ENDPSECTIONS},
	{"endsdoall", ENDSDOALL},
        {"endsections", ENDSECTIONS},
        {"endselect", ENDSELECT},
        {"endsingleprocess", ENDSINGLEPROCESS},
        {"endsubroutine", ENDSUBROUTINE},
        {"endtype", ENDTYPE},
	{"endwhere", ENDWHERE},
	{"end", ENDUNIT},
	{"entry", ENTRY},
	{"equivalence", EQUIVALENCE},
        {"err", ERR},
        {"event", EVENT},
	{"exit", EXIT},
        {"extended", EXTENDED},
        {"extend", EXTEND},
	{"external", EXTERNAL},
	{"find", FIND},
        {"foralldo", FORALLDO},
	{"format", FORMAT},
        {"from", FROM},
	{"function", FUNCTION},
        {"gate", GATE},
	{"global", GLOBAL_A},
	{"goto", PLAINGOTO},
        {"guards", GUARDS},
	{"implicitnone", IMPLICITNONE},
	{"implicit", IMPLICIT},
	{"include", INCLUDE},
	{"inout", INOUT},
        {"inport", INPORT},
	{"inquire", INQUIRE},
	{"integer", INTEGER},
        {"intent", INTENT},
	{"interfaceassignment", INTERFACEASSIGNMENT},
	{"interfaceoperator", INTERFACEOPERATOR},
	{"interface", INTERFACE},
	{"intrinsic", INTRINSIC},
	{"in", IN},
        {"iostat", IOSTAT},
        {"location", LOCATION},
        {"lock", LOCK},
	{"logical", LOGICAL},
	{"loop", LOOP},
        {"maxparallel", MAXPARALLEL},
        {"merger", MERGER},
	{"module", MODULE},
        {"moveport", MOVEPORT},
/* Used for MPCALL --- Ming Q. Xu */
        {"mpcall", MPCALL},
	{"namelist", NAMELIST},
        {"nullify", NULLIFY},
	{"only", ONLY},
	{"on", ON},
	{"open", OPEN},
	{"operator", OPERATOR},
        {"optional", OPTIONAL},
        {"ordered", ORDERED},
	{"otherwise", OTHERWISE},
        {"outport", OUTPORT},
	{"out", OUT},
	{"overflow", OVERFLOW},
        {"paralleldo", PARALLELDO},
        {"parallelsections", PARALLELSECTIONS},
        {"parallel", PARALLEL},
	{"parameter", PARAMETER},
	{"pause", PAUSE},
        {"pdo", PDO},
        {"pointer", POINTER},
        {"port", PORT},
        {"post", POST},
	{"print", PRINT},
        {"private", PRIVATE},
        {"probe", PROBE},
        {"procedure", PROCEDURE},
	{"program", PROGRAM},
	{"processcluster", PROCESS_CLUSTER},
	{"processcommon", PROCESS_COMMON},
        {"processdo", PROCESSDO},
        {"processes", PROCESSES},
	{"processglobal", PROCESS_GLOBAL},
        {"processors", PROCESSORS},
        {"process", PROCESS},
        {"psections", PSECTIONS},
        {"public", PUBLIC},
	{"range", RANGE},
	{"read", READ},
	{"real", REAL},
        {"receive", RECEIVE},
	{"recursive", RECURSIVE},
	{"reduce", REDUCE}, 
        {"release", RELEASE},
        {"result", RESULT},
	{"return", RETURN},
	{"rewind", REWIND},
	{"save", SAVE},
        {"scommon", SCOMMON},
/*	{"sdoall", SDOALL}, */
        {"sectionwait", SECTION},
	{"section", SECTION},
	{"select", SELECT},
        {"send", SEND},
        {"sequence", SEQUENCE},
        {"set", SET},
        {"singleprocess", SINGLEPROCESS},
        {"skippasteof", SKIPPASTEOF},
/*        {"stat", STAT}, */
	{"stop", STOP},
        {"submachine", SUBMACHINE},
	{"subroutine", SUBROUTINE},
	{"sync", SYNC},
	{"taskcluster", TASK_CLUSTER},
	{"taskglobal", TASK_GLOBAL},
        {"target", TARGET},
	{"then", THEN},
	{"to", TO},
	{"type", TYPE},
        {"unlock", UNLOCK},
	{"use", USE},
	{"virtual", VIRTUAL},
        {"wait", WAIT},
	{"where", WHERE},
	{"while", WHILE},
        {"with", WITH},
	{"wrap", WRAP},
	{"write", WRITE},
	{0, 0}
};


/*
 * This routine closes all files opened in this session
 */
void
close_files()
{
	register struct Inclfile *p = inclp;
	while (p) {
		(void)fclose(p->inclfp);
		p = p->inclnext;
	}
}


/*
 * This routine initializes the search tables for the keyword search
 * and the parameter keyword search.
 */
void
initkey()
{
	register struct Keylist *p;
	register int i, j;

	for (i = 0; i < 26; ++i)
		keystart[i] = (struct Keylist *)NULL;

	for (p = keys; p->keyname; ++p) {
		j = p->keyname[0] - 'a';
		if (keystart[j] == (struct Keylist *)NULL)
			keystart[j] = p;
		keyend[j] = p;
	}
}


int 
inilex(char *name)	/* filename to be scanned  */
{
	void    doinclude();

	initkey();
	cbfirst = NULL;
	cblast = NULL;
	pcbfirst = NULL;
	pcblast = NULL;
	nincl = 0;
	inclp = NULL;
	doinclude(name);
	lexstate = NEWSTMT;
	return (NO);
}


/* throw away the rest of the current line */
void 
flline()
{
	lexstate = RETEOS;
}


void 
doinclude(char *name)	/* file name string of the INCLUDE file to be
				 	* set up */
{
	FILE   *fp;
	struct Inclfile *t;
/*	char    temp[100]; */
/*	register char *lastslash, *s; */
	register char *lastslash;
/*	register char *p, *q; */

	if (inclp) {
		--lines_returned;
		inclp->incllno = thislin;
		inclp->inclcode = code;
		if(nextcd)
			inclp->incllinp = copyn(inclp->incllen = endcd-nextcd , nextcd);
		else
			inclp->incllinp = 0;
	}
	nextcd = (char *)NULL;

	if(++nincl >= MAXINCLUDES)
		fatal("includes nested too deep\n");
	if (name[0] == '\0')
		fp = stdin;
	else 
		fp = fopen(name, "r");
	if (fp)
        {
		t = inclp;
		inclp = (struct Inclfile *) chkalloc(sizeof(struct Inclfile));
		inclp->inclnext = t;
		prevlin = thislin = 0;
		inclp->inclname = name;
		infname = copys(name);
		infile = inclp->inclfp = fp;
	} else
	   fatalstr("Cannot open file %s\n", name);
}


/*
 * This routine eliminates the current INCLUDE file block and sets up the
 * immediately previous INCLUDE file block as the current block.
 *
 *	return value = YES (1) if we are at INCLUDE level >= 2
 *		       NO  (0) if we are at the top level (we hit end-of-file)
 */
int 
popinclude()
{
	struct Inclfile *t;
	register char *p;
	register int k;
	void clf();

/* Free the space for this include file's buffer. */

	if (infile != stdin)
		clf(&infile);
	free(infname);

	--nincl;
	t = inclp->inclnext;
	free((char *)inclp->inclname);
	free((char *) inclp);
	inclp = t;
	if (!inclp)
		return (NO);

	infile = inclp->inclfp;
	infname = copys(inclp->inclname);
	prevlin = thislin = inclp->incllno;
	code = inclp->inclcode;
	stno = nxtstno = inclp->inclstno;
	if (inclp->incllinp) {
		endcd = nextcd = stmtbuf;
		k = inclp->incllen;
		p = inclp->incllinp;
		while (--k >= 0)
			*endcd++ = *p++;
		free((char *) (inclp->incllinp));
	} else
		nextcd = (char *)NULL;
	return (YES);
}


/*
 * This routine does one line of a listing on stdout.
 *
 *   ptr    (INPUT) -	a pointer to the body of the line to be listed (col 7)
 *   line_number (INPUT) - the line number for this line
 */
void
list_line(char *ptr, int line_number)
{
	if (ptr)
		(void)fprintf(outf,"%6d %c %c%c%c%c%c%c%s\n",
		       line_number, ((nincl > 1) ? 'I' : ' '),
		       prefix[0], prefix[1], prefix[2], prefix[3], prefix[4], prefix[5], ptr);
}


/*
 * This routine fetches all the cards necessary to get a complete Fortran
 * statement.  One additional card after the full Fortran statement must 
 * be read to determine that it is not a continuation card.
 *
 *	return value - STINITIAL (a Fortran statement is ready)
 *		       STEOF	 (end of entire program)
 */
int
getcds()
{
	register char *p, *q;
	void err();

/* First, record how many bytes and lines have been read from the file */

first:	yylineno = lines_returned;
	if (yylisting && (code != STEOF) )	/* LISTING */
		list_line(nextcd, yylineno);

	if (newname) {
		free(infname);
		infname = newname;
		newname = (char *)NULL;
	}
	move_comments_to_permbuffer();
top:
	if (!nextcd) {
		code = getcd(nextcd = stmtbuf);
		yylineno = lines_returned;

		if (yylisting && (code != STEOF)) /* LISTING */
			(void) list_line(nextcd, yylineno);

		stno = nxtstno;
		if (newname) {
			free(infname);
			infname = newname;
			newname = (char *)NULL;
		}
		prevlin = thislin;
	}
	if (code == STEOF) {
    	        move_comments_to_permbuffer();
		if (popinclude()) {
			++lines_returned;
			goto first;
		} else
			return (STEOF);
	}
	if (code == STCONTINUE) {
		if (newname) {
			free(infname);
			infname = newname;
			newname = (char *)NULL;
		}
		lineno = thislin;
		err("Illegal continuation card ignored");
		nextcd = (char *)NULL;
		goto top;
	}
	if (code == CMT) {
		nextcd = (char *)NULL;
		flline();
		return (CMT);
	}
	if (nextcd > stmtbuf) {
		for (q = nextcd, p = stmtbuf; q < endcd; *p++ = *q++)
			;
		endcd = p;
	}
	move_comments_to_permbuffer();
	for (nextcd = endcd;
	     (nextcd+NCHARS_PER_LINE <= stmtend) && ((code=getcd(nextcd)) == STCONTINUE);
	     nextcd = endcd-1) 
	{
	     memcpy(nextcd-1,nextcd,endcd-nextcd);
	     move_comments_to_permbuffer();
		if (yylisting && (code != STEOF) && (code != CMT)) /* LISTING */
			list_line(nextcd-1, lines_returned);
	}
	nextch = stmtbuf;
	lastch = nextcd - 1;
	if (nextcd >= stmtend)
		nextcd = (char *)NULL;
	lineno = prevlin;
	prevlin = thislin;
	return (STINITIAL);
}


/*
 * This routine retrieves one card from the Fortran source file.
 *
 *	b (INPUT) -	The address at which to put the new card
 *
 *	return value - STINITIAL (the first card of a Fortran statement)
 *		       STCONTINUE (a continuation card)
 *		       STEOF (end-of-file on the current file)
 */
int 
getcd(register char *b)
{
	register int c;
	register char *p, *bend, *cmt;
        char *last_amp;
	extern int debug;
	int     cnext, only_amp;
	void err(), store_comment();

        /* Ming Q. Xu */
        if (strlen(b))
          memset(b,'\0',strlen(b));

top:
	endcd = b;
        bend = yylonglines? stmtend : b + NCHARS_PER_LINE;
	/* bend = b + NCHARS_PER_LINE; */

	/* Get first character on a line */

	if ((c = getc(infile)) == '&') { /* whole-card continuation */
		prefix[0] = BLANKC;
		prefix[5] = 'x';
		bend = stmtend;
	}

	else if (((c == 'c' || c == 'C' || c == '*'|| c == '#') &&
                  !(freeformat))
                 || ( c == '!' || (debug ? 0 : (c == 'D') || (c == 'd')))) {
                /* comment card */
		++lines_returned;

		/* For the listing:  fill the prefix array from the comment */

		prefix[0] = c;
		for (p = prefix + 1; p < prefixend; p++)
			if ((c = getc(infile)) == '\n')
				break;
			else if (c == EOF)
				break;
			else
				*p = c;

		/*
		 * If we saw EOF or '\n' and quit early above,
		 * fill the rest of the prefix with blanks. 
		 */

		for (; p < prefixend; p++)
			*p = ' ';

		/* Now, begin consuming the rest of the comment line- */

		if ((c != EOF) && (c != '\n'))
			while (((c = getc(infile)) != '\n') &&
			       (c != EOF)) 
				*endcd++ = c;
		if (c == '\n') *endcd++ = c;
		*endcd = '\0';

		if (yylisting)	/* LISTING */ 
			(void) list_line(b, lines_returned);
		for (cmt = tempbuf, p = prefix; p < prefixend; )
		  *cmt++ = *p++;
		for (p = b; *p;)
		  *cmt++ = *p++;
		*cmt = '\0';
		store_comment();
		goto top; 
	}  else if (c != EOF) {	/* Supporting F90 free format by cychen */
            if (freeformat) {
                (void)ungetc(c, infile);

                p = prefix;
                while (p < prefixend)
                    *p++ = BLANKC;

                if (last_c == '&') {	/* see if the last char of last */
                    prefix[5] = 'x';	/* line is a continuation symbol */
                    last_c = ' ';
                };

                only_amp = 0;	/* only_amp = 1 means that there are only */
                                /* '&' symbols beside blanks & comments */

                while (((c = getc(infile)) != '\n') && (c != EOF))
                    if ((c != '\t') && (c != ' ')) { /* ignoring starting */
                                                     /* blanks */

                        if (c == '&') {		/* line starts with '&' is */
                                                /* a continuation line */
                            prefix[5] = 'x';
                            only_amp = 1;
                        } else if (isdigit(c)) { /* The beginning five digits */
                                                 /* is the label of the line */
                            p = prefix;
                            while (p < prefixend-1) {
                                *p++ = c;
                                if (! isdigit(c = getc(infile))) {
                                    (void)ungetc(c, infile);
                                    break;
                                } 
                            }
                        } else
                            (void)ungetc(c, infile);
                        break;
                    }
                bend = stmtend;
            } else {
                /* a tab in columns 1-6 skips to column 7 */
                (void)ungetc(c, infile);
 
                for (p = prefix; p < prefixend &&
                     ((c = getc(infile)) != '\n') && (c != EOF);)
                        if (c == '\t') {
                                while (p < prefixend)
                                        *p++ = BLANKC;
 
                        /*
                         * check out next char: if digit, it goes in col 6 of
                         * the card
                         */
 
                                if (isdigit(cnext = getc(infile)))
                                        prefix[5] = cnext;
                                else
                                        (void)ungetc(cnext, infile);
                                bend = stmtend;
                        } else
                                *p++ = c;
            }
	}


	if (c == EOF)
		return (STEOF);
	if (c == '\n') {
		while (p < prefixend)
			*p++ = BLANKC;
		*endcd = '\0';	/* EOLN:  make statement empty */
	} else {		/* Read body of line */
		while ((endcd < bend) && ((c = getc(infile)) != '\n') &&
                       (c != EOF)) {
                    /* Ming Q. Xu getting rid of dangling comments */
                    if (c == '!')
                    {
                        int i;
                        int no_alnum = 1;

                        /* check to see whether b has
                           not get any alphanum chars */
                        for (i = 0; i < strlen(b); i++) {
                            if (isalnum(b[i])) {
                                no_alnum = 0;
                                break;
                            }
                        }
                        if (no_alnum) {
                            while (1) {
                                c = getc(infile);
                                if (c == '\n' ||
                                    c == EOF)
                                {
                                    break;
                                }
                            }
                            break;
                        } else {
                            *endcd++ = c;
                        }
                    } else {
                        if (c != ' ') {
                            last_c = c;
                            if (c == '&')
                                last_amp = endcd; /* records the position */
                                                  /* of the last '&' char */
                            else
                                only_amp = 0;
                        }
                        *endcd++ = c;
                    }
                }

                if (freeformat && only_amp)
                    err("Line with only chars '&' not allowed");

                if (freeformat && (last_c == '&')) /* clear the '&' char at */
                    *last_amp = ' ';		   /* the end of the line */

		if (c == EOF)
			return (STEOF);
		if (c != '\n') {
			while ((c = getc(infile)) != '\n')
				if (c == EOF)
					return (STEOF);
		}
		*endcd++ = '\0';  /* put NULL char in buffer as end marker */
	}

	++lines_returned;

	/* Check for a comment char in the prefix */

	for (p = prefix; p < prefixend; ++p)
		if (*p == '!') {
			if (yylisting)
				(void) list_line(b, lines_returned);
			for (cmt = tempbuf; *p != '\n';)
			  *cmt++ = *p++;
			*p = '\n';
			store_comment();
			goto top; 
		}
	if ((prefix[5] != ' ') && (prefix[5] != '0'))
		return (STCONTINUE);
	for (p = prefix; p < prefixend; ++p)
		if (isspace(*p) == 0)
			goto initline;
	/* Check for blank line */
	for (p = b; p < endcd - 1; ++p)	/* char at endcd is always \0 */
		if (isspace(*p) == 0) {
			nxtstno = 0;
			return (STINITIAL);
		}
	/* We have a completely blank line */

	if (yylisting)	/* LISTING */
		(void) list_line(b, lines_returned);
	goto top;

initline:			/* there is a label */
	nxtstno = 0;

 /*
  * Start out conversion loop at second char in prefix if we have debug set and
  * a D/d in col 1.  Otherwise, start at first. 
  */

	if (debug)
		if (prefix[0] == 'd' || prefix[0] == 'D')
			p = prefix + 1;
		else
			p = prefix;

	while (p < prefix + 5) {
		if (isspace(*p) == 0)
			if (isdigit(*p))
				nxtstno = 10 * nxtstno + (*p - '0');
			else {
				if (newname) {
					free(infname);
					infname = newname;
					newname = (char *)NULL;
				}
				lineno = thislin;
				err("nondigit in statement number field");
				nxtstno = 0;
				break;
			}
		p++;
	}
	return (STINITIAL);
}


/*
 * This routine removes all blanks, reduces everything to lower case
 *   (except char strings), replaces outer quotes with MYQUOTE, encloses
 *   a Hollerith string with MYHOLLERITH, and determines whether this
 *   statement contains either an exposed comma, or an exposed equal.
 */
void
crunch()
{
	register char *i, *j, *j0, *j1, *prvstr, *cmt;
	int     bracklev;
	int     ten, nh, quote;
	void    erri(), err();

/* i is the next input character to be looked at.
   j is the next output character */

	--lastch;		/* at this point, lastch is pointing at the
				 * ending '\0' decrement it to eliminate that
				 * '\0' from consideration */

	bracklev = 0;		/* level of <,> pairs */
	parlev = 0;		/* level of (,) pairs */
	expcom = 0;		/* exposed ','s */
	expeql = 0;		/* exposed '='s */
	expdcolon = 0;          /* exposed '::'s */
	expscolon = 0;          /* exposed ':' eg. construct_name : .... */
	expeqlt = 0;
	prvstr = j = crunchbuf;

	for (i = stmtbuf; i <= lastch; ++i) {
		if (isspace(*i) || (*i == '\0'))
			continue;
	/* eliminate ! comments	 */
		if (*i == '!') {
/*			*j++ = '!'; why? */
			cmt = tempbuf;
			while (*i) {
                          *cmt++ = *i;
                          *i = ' ';
                          i++;
                        }
/*                              *cmt++ = *i++; */
                        *cmt = '\n';
                        *cmt++;
			*cmt = '\0';
			store_comment_end_permbuff();
			continue;
		}
	/* Is this a QUOTE and is this not a random I/O operator? */

		if ((*i == '\'') && (!randomio(j))) {
			quote = *i;
			*j = MYQUOTE;	/* special marker */
			for (;;) {
				if (++i > lastch) {
					err("Unbalanced quotes; closing quote supplied");
					break;
				}
				if (*i == quote)
					if (i < lastch && i[1] == quote) {
                                                *++j = *i; /* fixed by cychen */
						++i;
					} else
						break;
				*++j = *i;
			}
			if (i > lastch)
				++j;	/* quotes were unbalanced */
			else {
				j[1] = MYQUOTE;
				j += 2;
			}
			prvstr = j;
                } else if ((*i == '\"') && (!randomio(j))) {
                        quote = *i;
                        *j = MYQUOTE;   /* special marker */
                        for (;;) {
                                if (++i > lastch) {
                                        err("Unbalanced quotes; closing quote supplied");
                                        break;
                                }
                                if (*i == quote)
                                        if (i < lastch && i[1] == quote) {
                                                *++j = *i; /* fixed by cychen */
                                                ++i;
                                        } else
                                                break;
                                *++j = *i;
                        }
                        if (i > lastch)
                                ++j;    /* quotes were unbalanced */
                        else {
                                j[1] = MYQUOTE;
                                j += 2;
                        }
                        prvstr = j;
		} else if ((*i == 'h' || *i == 'H') && j > prvstr) {	/* test for Hollerith
									 * strings */
			if (isdigit(j[-1])==0)
				goto copychar;
			nh = j[-1] - '0';
			ten = 10;
			j1 = prvstr - 1;
			if (j1 < j - 5)
				j1 = j - 5;
			for (j0 = j - 2; j0 > j1; --j0) {
				if (isdigit(*j0)==0)
					break;
				nh += ten * (*j0 - '0');
				ten *= 10;
			}
			if (j0 <= j1)
				goto copychar;
/*
 * A hollerith must be preceded by a punctuation mark.
 * '*' is possible only as repetition factor in a data statement
 * not, in particular, in character*2h
 */

			if (!(*j0 == '*' && stmtbuf[0] == 'd' && stmtbuf[1] == 'a') &&
			    *j0 != '/' && *j0 != '(' &&
			    *j0 != ',' && *j0 != '=' && *j0 != '.')
				goto copychar;
			if (i + nh > lastch) {
				erri("%dH too big", nh);
				nh = lastch - i;
			}
		/* Form the string for the Hollerith constant. */
			j0[1] = MYHOLLERITH;	/* special marker */
			j = j0 + 1;
			while (nh-- > 0)
				*++j = *++i;
			j[1] = MYHOLLERITH;
			j += 2;
			prvstr = j;
		} else {
			if (*i == '(') {
				if (++parlev > MAXPARLEV)
					fatal("Too many levels of parenthesis nesting");
			} else if (*i == ')')
				--parlev;
/*			else if (*i == '<')
				++bracklev;
			else if (*i == '>')
				--bracklev;
			else if ((*i == ':') && (i[1] == ':') && (bracklev == 0))
  			        expdcolon = 1;
*/			else if ((parlev == 0) && (bracklev == 0))
                                if ((*i == ':') && (i[1] == ':')) 
				{
				     *j++ = *i++;
				     expdcolon = 1;
				}
			        else if (*i == ':')
				     expscolon = 1;
				else if ((*i == '=') && (i[1] == '>') && (expdcolon == 0))
				{
				     *j++ = *i++;
				     expeqlt = 1;
				}
				else if ((*i == '=') && (i[1] == '='))
				     *j++ = *i++;
				else if ((*i == '=') && (expdcolon == 0))
				     expeql = 1;
				else if (*i == ',')
				     expcom = 1;
	copychar:		/* not a string or space -- copy, shifting case
				 * if necessary */
			*j++ = (shiftcase && isupper(*i)) ? tolower(*i) : *i;
		}
	}
	{ register char *k;
	for (i = k = stmtbuf; i <= lastch; ++i)
            if (*i != '\0') { *k = *i; ++k; }
        }
	lastch = j - 1;
 	nextch = crunchbuf;
}


/*
 * This routine is only called when the current character in crunch is 
 * a single quote (').  It determines whether the beginning of the current
 * statement looks like the beginning of a random I/O statement.  The 
 * parameter j is a pointer to the next place where a 'crunch'ed character
 * will be put within crunchbuf.  The characters under scrutiny are:
 * crunchbuf[0] through j[-1].
 *
 *      j (INPUT) - the next place within crunchbuf where a 'crunch'ed
 *                  character will be placed.
 *
 *	return value - YES if the chars in cruncbuf look like a random I/O
 *                     stmt NO  otherwise
 */
int 
randomio(register char *j)
{
	int     nchars;
	char   *start;
	register char *p;

	nchars = j - crunchbuf;	/* compute number of characters in buffer */

	if ((nchars > 6) && eqn(6, crunchbuf, "write("))
		start = crunchbuf + 6;
	else if ((nchars > 5) && eqn(5, crunchbuf, "read("))
		start = crunchbuf + 5;
	else if ((nchars > 5) && eqn(5, crunchbuf, "find("))
		start = crunchbuf + 5;
	else
		return (NO);

	if (isalpha(*start))
	/* We have an identifier for a unit number.  Rest must be alphanum-- */
	{
		for (p = start + 1; p < j; ++p)
			if ((isalpha(*p)==0) && (isdigit(*p)==0) && !(*p == '_'))
				return (NO);
		return (YES);
	} else if (isdigit(*start))
	/* We have an integer constant for a unit number. */
	{
		for (p = start + 1; p < j; ++p)
			if (isdigit(*p) == 0)
				return (NO);
		return (YES);
	} else
		return (NO);
}


int 
named_stmt()
{
     char *i, *p;
     
     i = nextch;
     p = yytext;
     *p++ = *i++;
     while (i <= lastch)
	  if (isalpha(*i) || isdigit(*i) || (*i == '_'))
	       *p++ = *i++;
	  else
	       break;
     yyleng = p - yytext;
     *p = '\0';
     if ((*i == ':') && (i[1] != ':') && (i != nextch)) 
     {
	  nextch = i;
	  return (1);
     }
     else return (0);
}

	  
/*
 * This routine determines the type of statement we have.  It either
 * finds the first lexeme, (like FORMAT, LOGICALIF, etc), or determines
 * that this is an assignment statement (LET).  It assigns that value
 * to the variable "stkey" which stays around as a reminder of what
 * type statement we have as the rest of the statement is scanned.
 */
void
analyz()
{
	register char *i, *ii;
	int forallstmt, level;
	void err();

	if (parlev) {
		err("Unbalanced parentheses, statement skipped");
		stkey = UNKNOWN;
		return;
	}

	if ((expscolon) && (named_stmt()))
	{
	     stkey = CONSTRUCT_ID;
	     return;
	}
	
	if (nextch + 2 <= lastch && eqn(3, nextch, "if(")) {
               /* assignment or if statement -- look at character after balancing paren */
		parlev = 1;
		for (i = nextch + 3; i <= lastch; ++i)
			if (*i == (MYQUOTE))
				while (*++i != MYQUOTE);
			else if (*i == (MYHOLLERITH))
				while (*++i != MYHOLLERITH);
			else if (*i == '(')
				++parlev;
			else if (*i == ')') {
				if (--parlev == 0)
					break;
			}
		if (i >= lastch)
			stkey = LOGICALIF;
		else if (i[1] == '=')
			stkey = LET;
		else if (isdigit(i[1]))
			stkey = ARITHIF;
		else
			stkey = LOGICALIF;
		if (stkey != LET)
			nextch += 2;
	} else if (nextch + 5 <= lastch && eqn(6, nextch, "where(")) {
               /* assignment or where statement -- look at character after balancing paren */
		parlev = 1;
		for (i = nextch + 6; i <= lastch; ++i)
			if (*i == (MYQUOTE))
				while (*++i != MYQUOTE);
			else if (*i == (MYHOLLERITH))
				while (*++i != MYHOLLERITH);
			else if (*i == '(')
				++parlev;
			else if (*i == ')') {
				if (--parlev == 0)
					break;
			}
		if (i >= lastch)
			stkey = WHERE;
		else if (i[1] == '=')
			stkey = LET;
		else
			stkey = WHERE_ASSIGN;
		if (stkey != LET)
			nextch += 5;
	} else if (expeql) {	/* may be an assignment */
		if (expcom && nextch < lastch) {
			if ((nextch[0] == 'd') && (nextch[1] == 'o')) {
				stkey = PLAINDO;
				nextch += 2;
                        } else if (eqn(9, nextch, "processdo")) {
                                stkey = PROCESSDO;
                                nextch += 9;
			} else if (eqn(10, nextch, "paralleldo")) {
				stkey = PARALLELDO;
				nextch += 10;
			} else if (eqn(3, nextch, "pdo")) {
				stkey = PDO;
				nextch += 3;
			} else if (eqn(6, nextch, "forall")) {
				stkey = FORALLDO;
				nextch += 6;
			} else if (eqn(5, nextch, "alldo")) {
				stkey = CDOALL;
				nextch += 5;
			} else if (eqn(8, nextch, "acrossdo")) {
				stkey = CDOACROSS;
				nextch += 8;
			} else if (eqn(6, nextch, "cdoall")) {
				stkey = CDOALL;
				nextch += 6;
			} else if (eqn(6, nextch, "sdoall")) {
				stkey = SDOALL;
				nextch += 6;
			} else if (eqn(8, nextch, "doacross")) {
				stkey = DOACROSS;
				nextch += 8;
			} else if (eqn(9, nextch, "cdoacross")) {
				stkey = CDOACROSS;
				nextch += 6;
			} else
				stkey = LET;
		}
		else if (eqn(7, nextch, "forall(")) {
			level = 0;
			forallstmt = -1;
			for (ii = nextch + 7; ii <= lastch; ii++) {
				switch (*ii) {
				    case (')'):
					if (level <= 0)
						if (*(ii + 1) == '=') {
							forallstmt = NO;
							break;
						} else {
							forallstmt = YES;
							break;
						}
					else
						level--;
					break;
				    case ('('):
					level++;
				    default:
					break;
				}
				if (forallstmt >= 0) {
					if (forallstmt == 1) {
						stkey = FORALL;
						nextch += 6;
					} else
						stkey = LET;
					break;
				     }
			     }
			if (ii > lastch)
				stkey = LET;
		} else
			stkey = LET;
	}
	else if (expeqlt && !expcom)
	     stkey = POINTERLET;
/* otherwise search for keyword */
	else {
		stkey = getkwd();
		if ((stkey == TYPE) && (nextch[0] != '('))
		     stkey = TYPE_DECL;
		if (stkey == PLAINGOTO && lastch >= nextch)
		     if (nextch[0] == '(')
			  stkey = COMPGOTO;
		     else if (isalpha(nextch[0]))
			  stkey = ASSIGNGOTO;
	}
	parlev = 0;
}


/*
 * This routine checks the characters starting at "nextch" against the
 * table of Fortran keywords.
 */
int 
getkwd()
{
	register char *i, *j;
	register struct Keylist *pk, *pend;
	int     k;

        /* the following line is useful for testing parser */
        /* printf("getkwd:  %d %s\n", yylineno, stmtbuf); */
	if (isalpha(nextch[0])==0)
		return (LET);
	k = nextch[0] - 'a';
	if (pk = keystart[k])
		for (pend = keyend[k]; pk <= pend; ++pk) {
			i = pk->keyname;
			j = nextch;
			while (*++i == *++j && *i != '\0');
			if (*i == '\0' && j <= lastch + 1) {
				nextch = j;
				return (pk->keyval);
			}
		}
	return (UNKNOWN);
}


/*
 * This routine gets the next token in a statement.  It starts at "nextch"
 * within "stmtbuf".  As a side effect, "yytext" is filled with the
 * consumed input and "yyleng" is set to the length of that text.
 *
 *	return value - the lexeme value for the token
 */
int 
gettok()
{
	register struct Dotlist *pd;
	register struct Punctlist *pp;
/*	register struct oplist *op; */
	register char *p, *i, *j, *saveptr;

	int     havdot, havexp, havdbl;
	int     val, ijk;
	extern struct Punctlist puncts[];
	extern struct Dotlist dots[];
	extern struct oplist operands[];
	extern int yyleng;
	void err();

	char   *n1;

        /* the following line is useful for testing parser */
        /* printf("gettok:  %d %s\n", yylineno, nextch); */

/* Check for ordinary character string: 'xxxxxxx' */
	if (*nextch == (MYQUOTE)) {
		++nextch;
		p = yytext;
		while ((nextch <= lastch) && (*nextch != MYQUOTE))
			*p++ = *nextch++;
		if (nextch > lastch) {
			yyleng = p - yytext;
			*p = '\0';
			return (BAD_CCONST);
		}
		++nextch;
		yyleng = p - yytext;
		*p = '\0';

	/* Are we sure this is a char constant? (could be hex or octal) */
/*		if ((*nextch != 'x') &&
		    (*nextch != 'o'))
*/			return (CHAR_CONSTANT);

	/* It's a hex or octal constant! */

/*		else {
			if (*nextch == 'x')
				radix = 16;
			else
				radix = 8;

*/		/* Check for bad chars in constant. */

/*			for (p = yytext + 1; p < (yytext + (yyleng - 1));)
				if (*p == BLANKC || *p == '\t')
					p++;
				else {
					if (isupper(*p))
						*p = tolower(*p);
					if (hextoi(*p++) >= radix) {
						radix = 0;
						break;
					}
				}
			++nextch;
			return (radix == 0 ? BAD_SYMBOL :
				(radix == 16 ? HEX_CONSTANT : OCTAL_CONSTANT));
	        }
*/       }
/* Check for a Hollerith constant */

	if (*nextch == (MYHOLLERITH)) {
		++nextch;
		p = yytext;
		*p++ = '\'';	/* opening quote */
		while (*nextch != MYHOLLERITH) {
			if (*nextch == '\'')
				*p++ = '\'';	/*-explode quotes-*/
			*p++ = *nextch++;
		}
		++nextch;
		*p++ = '\'';	/* closing quote */
		yyleng = p - yytext;
		*p = '\0';
		return (CHAR_CONSTANT);
	}
/* Check for a bad Hollerith constant (as found by crunch) */

	if (*nextch == (BADHOLLERITH)) {
		++nextch;
		yytext[0] = '1';
		yytext[1] = 'h';
		yytext[yyleng = 2] = '\0';
		return (BAD_CCONST);
	}
	if (needkwd) {
		needkwd = 0;
		return (getkwd());
	}
        if (optkwd) {
                optkwd = 0;
                saveptr = nextch;
                ijk = getkwd();
                if ((ijk != UNKNOWN) && (is_equal(*nextch)))
                    return(ijk);
                else
                    nextch = saveptr;
        }
        if (optcall) {
                optcall = 0;
                saveptr = nextch;
                ijk = getkwd();
                if ((ijk != UNKNOWN) && (ijk == 54))
                    return(ijk);
                else
                    nextch = saveptr;
        }
        if (opt_l_s) {
                opt_l_s = 0;
                saveptr = nextch;
                ijk = getkwd();
                if ((ijk != UNKNOWN) && ((ijk == 242) || (ijk == 243)))
                    return(ijk);
                else
                    nextch = saveptr;
        }


/* Check for punctuation */
	for (pp = puncts; pp->punchar; ++pp)
		if (*nextch == pp->punchar) {
			if ((*nextch == '*' || *nextch == '/') &&
			    nextch < lastch && nextch[1] == nextch[0]) {
				if (*nextch == '*')
					val = DASTER;
				else
					val = DSLASH;
				nextch += 2;
			} 
			else if ((*nextch == '/') && ( nextch < lastch))
			{
			     if (nextch[1] == ')')
			     {
				  val = RIGHTAB;
				  nextch += 2;
			     }
                             else if (nextch[1] == '=')
                             {
                                  val = NE;
                                  nextch += 2;
                             }
			     else 
			     {
				  val = SLASH;
				  ++nextch;
			     }
			}
			else if ((*nextch == '(') && ( nextch < lastch))
			{
			     if (nextch[1] == '/')
			     {
				  val = LEFTAB;
				  nextch += 2;
			     }
			     else 
			     {
				  val = LEFTPAR;
				  ++parlev;
				  ++nextch;
			     }
			}
			else if ((*nextch == '=') && (nextch < lastch))
			{
			     if (nextch[1] == '>')
			     {
				  val = POINT_TO;
				  nextch += 2;
			     }
			     else if (nextch[1] == '=')
			     {
				  val = EQ;
			     	  nextch += 2;
			     }
			     else 
			     {
				  val = EQUAL;
				  ++nextch;
			     }
			}
			else if ((*nextch == '<') && (nextch < lastch))
			{
			     if (nextch[1] == '>')
			     {
				  val = NE;
				  nextch += 2;
			     }
			     else if (nextch[1] == '=')
			     {
				  val = LE;
			     	  nextch += 2;
			     }
			     else
			     {
				  val = LT;
				  ++nextch;
			     }
			}
			else if ((*nextch == '>') && (nextch < lastch))
			{
			     if (nextch[1] == '=')
			     {
				  val = GE;
			     	  nextch += 2;
			     }
			     else 
			     {
				  val = GT;
				  ++nextch;
			     }
			}
			else {
				val = pp->punval;
				if (val == LEFTPAR)
					++parlev;
				else if (val == RIGHTPAR)
					--parlev;
				++nextch;
			}
			return (val);
		}
/* Check for dotted lexeme (.AND. .OR. .XOR.  ...) */

	if (*nextch == '.')
		if (nextch >= lastch)
			goto badchar;
		else if (isdigit(nextch[1]))
			goto numconst;
		else {
			for (pd = dots; (j = pd->dotname); ++pd) {
				for (i = nextch + 1; i <= lastch; ++i)
					if (*i != *j)
						break;
					else if (*i != '.')
						++j;
					else {
						nextch = i + 1;
						return (pd->dotval);
					}
			}
			p = yytext;
			*p++ = *nextch++;
			while ((nextch <= lastch) && (*nextch != '.'))
			     if (isalpha(*nextch))
				  *p++ = *nextch++;
			     else
				  goto badchar;
			if (*nextch != '.')
			     goto badchar;
			*p++ = *nextch++;
			yyleng = p - yytext;
			return (DEFINED_OPERATOR);
		}

/* Check for an alphabetic lexeme */
	if (isalpha(*nextch)) {
	/* Now, match an ordinary identifier. */

		p = yytext;
		*p++ = *nextch++;
		while (nextch <= lastch)
		  if (isalpha(*nextch) || isdigit(*nextch) || (*nextch == '_') || (*nextch == '$'))
		    *p++ = *nextch++;
		  else
		    break;
		yyleng = p - yytext;
		*p = '\0';
                if (colon_flag && nextch <= lastch && (*nextch == ':') &&
                    !strcmp(yytext, "only"))
		{
		     ++nextch;
		     colon_flag = 0;
		     return (ONLY);
		}
		if (inioctl && nextch <= lastch && *nextch == '=') { 
			++nextch;
			return (NAMEEQ);
		}
	/* Check to make sure that it was not a FUNCTION def */

		if (yyleng > 17 && eqn(17, yytext, "recursivefunction") &&
		    isalpha(yytext[8]) && (nextch < lastch) && (nextch[0] == '(') &&
		    ((nextch[1] == ')') || isalpha(nextch[1]))) {
			nextch -= (yyleng - 9);
			return (RECURSIVE);
		}
		if (yyleng > 8 && eqn(8, yytext, "function") &&
		    isalpha(yytext[8]) && (nextch < lastch) && (nextch[0] == '(') &&
		    ((nextch[1] == ')') || isalpha(nextch[1]))) {
			nextch -= (yyleng - 8);
			return (FUNCTION);
		}
		if (yyleng > MAX_NAME_LEN) {
			char    buff[100];
			(void)sprintf(buff, "name %s too long, truncated to %d",
				yytext, MAX_NAME_LEN);
			err(buff);
			yyleng = MAX_NAME_LEN;
			yytext[MAX_NAME_LEN] = '\0';
		}
		return (IDENTIFIER);
	}
	if (*nextch == '!') {
		++nextch;
	        return(COMMENT);
	}

/* Check for a numeric lexeme */

	if (isdigit(*nextch)==0)
		goto badchar;
numconst:
	havdot = NO;
	havexp = NO;
	havdbl = NO;
	for (n1 = nextch; nextch <= lastch; ++nextch) {
		if (*nextch == '.')
			if (havdot)
				break;
			else if (nextch + 2 <= lastch && isalpha(nextch[1])
				 && isalpha(nextch[2]))
				break;
			else
				havdot = YES;
		else if (!intonly && (*nextch == 'd' || *nextch == 'e')) {
			p = nextch;
			havexp = YES;
			if (*nextch == 'd')
				havdbl = YES;
			if (nextch < lastch)
				if (nextch[1] == '+' || nextch[1] == '-')
					++nextch;
			if ((nextch >= lastch) || (isdigit(*++nextch)==0)) {
				nextch = p;
				havdbl = havexp = NO;
				break;
			}
			for (++nextch;
			     nextch <= lastch && isdigit(*nextch);
			     ++nextch);
			break;
		} else if (isdigit(*nextch)==0)
			break;
	}
	p = yytext;
	i = n1;
	while (i < nextch)
		*p++ = *i++;
	yyleng = p - yytext;
	*p = '\0';
	if (havdbl)
		return (DP_CONSTANT);
	if (havdot || havexp)
		return (REAL_CONSTANT);

/* Have an INTEGER constant.  Is it a label? */
	return (INT_CONSTANT);
badchar:
	stmtbuf[0] = *nextch++;
	return (UNKNOWN);
}


/*
 * This routine is the top level of the Scanner, called to get a lexeme.
 *
 *	return value - lexeme value
 */
int 
yylex()
{
	static int tokno;
	static int nt;

/* Initialize the token value and length. */

	/* yytext[0] = '\0';*/
	yyleng = 0;

	switch (lexstate) {
new_stmt:
 	    case OVER:
	        return XEOF;
	    case NEWSTMT:	/* need a new statement */
		nt = getcds();
		if (nt == STEOF) {
		     nt = XEOF;
                     /* Solving misplacement of comments
                        before Include Statement --- Ming Q. Xu */
                     if (last_stmt_is_include) {
                         last_stmt_is_include = 0;
                         return (XEOF);
                     }
		     if (pcbfirst != NULL)
			  goto commentscase;
		     else return (XEOF);
		} /* if getcds() == SEOF */

		lastend = stkey == ENDUNIT;

	/*
	 * Now, crunch the new statement.  If nothing's left after crunching,
	 * then go get another statement. 
	 */

/* cychen
                printf("prefix %d: %c%c%c%c%c%c\n", yylineno, prefix[0],
                        prefix[1],  prefix[2],  prefix[3],  prefix[4],
                        prefix[5]);
                printf("crunch %d: %s\n", yylineno, nextch);
*/
		crunch();
		if (nextch > lastch)
			goto new_stmt;
		tokno = 0;
		lexstate = FIRSTTOKEN;
		yystno = stno;
		stno = nxtstno;
		yyleng = 0;
		return (LABEL);

first:
	    case FIRSTTOKEN:	/* first step on a statement */

		(void) analyz();
                lexstate = OTHERTOKEN;
		tokno = 1;
		return (stkey);

	    case OTHERTOKEN:	/* return next token */
		if (stkey == FORMAT)
		        goto reteos;
/*		if (stkey == READ)
		        goto reteos;
		if (stkey == WRITE)
		        goto reteos;
*/		if (nextch > lastch) {
			nextch = lastch + 1;
			goto reteos;
		}
 		++tokno;
		if ((stkey == LOGICALIF || stkey == ELSEIF) && parlev == 0 && tokno > 3)
			goto first;
		if (stkey == ASSIGN && tokno == 3 && nextch < lastch &&
		    nextch[0] == 't' && nextch[1] == 'o') {
			nextch += 2;
			return (TO);
		}
		if ((stkey == CONSTRUCT_ID) && (tokno == 2) && (nextch[0] == ':')) 
		     lexstate = FIRSTTOKEN;
		return (gettok());

reteos:
	    case RETEOS:
                /* Solving the misplacement of comments before
                   include statement --- Ming Q. Xu */
                if (last_stmt_is_include) {
                    last_stmt_is_include = 0;
                    lexstate = NEWSTMT;
                    return (EOLN);
                }
		if (pcbfirst != NULL)
		     lexstate = COMMENTS;
		else lexstate = NEWSTMT;
		return (EOLN);
commentscase:
	   case COMMENTS:
		lexstate = COMMENTSEOS;
		flush_comments();
		return (COMMENT);
	   case COMMENTSEOS:
		if (nt == XEOF)
		     lexstate = OVER;
		else lexstate = NEWSTMT;
		return (EOLN);
	}

	if (nt == EOF)
	     if (pcbfirst != NULL) {
		  flush_comments();
		  return (COMMENT);
	     }
	
        return nt;
	
}

/* Comment buffering code */

static void
store_comment()

{
     comment_buf *ncb;
     
     ncb = (comment_buf *) chkalloc(sizeof(comment_buf));
     if ((int)strlen(tempbuf) > COMMENT_BUF_STORE) 
     {
	  warn("comment too long. truncated to 160 characters.");
	  tempbuf[COMMENT_BUF_STORE] = '\0';
     }

     ncb->buf[0]='\0';
     strcpy(ncb->buf,tempbuf);
     ncb->next = NULL;
     if (!cbfirst) {
	  cbfirst = ncb;
	  cblast = cbfirst;
     }
     else 
     {
	  cblast->next = ncb;
	  cblast = ncb;
     }
}

static void
store_comment_end_permbuff()

{
     comment_buf *ncb;
     
     ncb = (comment_buf *) chkalloc(sizeof(comment_buf));
     ncb->buf[0]='\0';
     strcpy(ncb->buf,tempbuf);
     ncb->next = NULL;
     if (!pcbfirst) {
	  pcbfirst = ncb;
	  pcblast = pcbfirst;
     }
     else 
     {
	  pcblast->next = ncb;
	  pcblast = ncb;
     }
}

static void
move_comments_to_permbuffer()
{
     if (pcbfirst == NULL)
     {
	  pcbfirst = cbfirst;
	  pcblast = cblast;
     }
     else 
     {
	  if (cbfirst != NULL) 
	  {
	       pcblast->next = cbfirst;
	       pcblast = cblast;
	  }
	  
     }
     
     cbfirst = cblast = NULL;
}

     
static void
flush_comments()
{
     comment_buf *ncb;
     int len = 0;

     for (ncb = pcbfirst; ncb; ncb = ncb->next)
	  len = len + strlen(ncb->buf) + 1;
     
     commentbuf = (char *) chkalloc(len * sizeof(char));
     
     for (ncb = pcbfirst; ncb; ncb = ncb->next) 
	  strcat(commentbuf, ncb->buf);
     

     pcbfirst = pcblast = NULL;
/*     printf("%s\n", commentbuf);*/
     
}     

