/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#define EXTEND_NODE 2  /* move this definition to h/ files. */
                       /* should agree with cftn.gram definition. */
/*
 * stat.c
 *
 * Routines for handling Fortran statements
 *
 */

#include "inc.h"
#include "defines.h"
#include "extern.h"

extern int parstate;
extern PTR_SYMB head_symb, global_list, star_symb;
extern PTR_BFND cur_bfnd, global_bfnd, pred_bfnd, cur_scope();
extern PTR_TYPE global_float, global_int, global_bool, global_default, vartype;
extern PTR_LABEL head_label;
extern PTR_TYPE impltype[];
extern PTR_BLOB head_blob, cur_blob;
extern PTR_CMNT comments;
extern int yylineno;
extern int mod_offset;
extern int nioctl;
extern int yydebug;

void fatalstr();
void execerr();
int chk_params();
void err();

PTR_BFND get_bfnd();
PTR_LLND make_llnd();
PTR_SYMB make_symb();
PTR_BLOB make_blob();
PTR_LABEL make_label();
PTR_SYMB install_entry();
PTR_SYMB get_proc_symbol();
PTR_HASH correct_symtab();
int end_group = 0;


/* 
   The following two routines are used for reading in input/output
   control lists. 
*/

void
startioctl()
{

	inioctl = YES;
        nioctl = 0; 

}

void
endioctl()
{
	inioctl = NO;

}


/*
 * Follow a chain of blob nodes to get the last
 *
 * input:
 *	 blob - the list to be searched
 *
 * output:
 *	 pointer to the last node in the list
 */
PTR_BLOB
follow_blob(PTR_BLOB blob)
{
	register PTR_BLOB next, last;

	for (next = last = blob; next; next = next->next)
		last = next;
	return (last);
}


/* 
 * make_if takes an expression to make an IF_NODE
 * Also allocates a collection point and points the false branch
 * to the collection point
 */
PTR_BFND
make_if(PTR_LLND expr)
{
	PTR_BFND p;
	void set_blobs(), make_prog_header();
	void err();

	if (pred_bfnd->variant == GLOBAL)
	  make_prog_header();

	if (expr->type->variant != T_BOOL) {
		err("non-logical expression in IF statement");
		expr = LLNULL;
	}
	p = get_bfnd(fi,IF_NODE, SMNULL, expr, LLNULL, LLNULL);
/*	set_blobs(p, pred_bfnd, NEW_GROUP1); */
	return (p);
}


/*
 * make_elseif fixes the control frame to reflect the current paring state
 */

void
make_elseif(PTR_LLND expr)
{
	register PTR_BFND p;
	void execerr(), make_endblock();

	if (expr->type->variant != T_BOOL) {
		err("non-logical expression in IF statement ");
		expr = LLNULL;
	}
	if (pred_bfnd->variant == IF_NODE || pred_bfnd->variant == ELSEIF_NODE)
		p = get_bfnd(fi,ELSEIF_NODE, SMNULL, expr, LLNULL, LLNULL);
	else
		execerr("elseif out of place", (char *)NULL);
	p->control_parent = pred_bfnd;
	pred_bfnd->entry.Template.bl_ptr2 = make_blob(fi,p, BLNULL);
	pred_bfnd = p;
	cur_blob = p->entry.Template.bl_ptr1 = make_blob(fi,BFNULL, BLNULL);
}


/*
 * make_else fixes the control stack to reflect the current state of parsing
 * and put all BIF nodes after the saved control frame as the true branch of
 * the If statement
 */
void
make_else()
{
	void execerr(), set_blobs();
	PTR_BFND p;

	if (pred_bfnd->variant != IF_NODE && pred_bfnd->variant != ELSEIF_NODE)
	  execerr("else out of place", (char *)NULL);
	p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	set_blobs(p, pred_bfnd, SAME_GROUP);
        cur_blob = pred_bfnd->entry.Template.bl_ptr2
	  = make_blob(fi,BFNULL, BLNULL);
}

void
make_elsewhere()
{
	void execerr(), set_blobs();
	PTR_BFND p;

	if (pred_bfnd->variant != WHERE_BLOCK_STMT)
	  execerr("elsewhere out of place", (char *)NULL);
	p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	set_blobs(p, pred_bfnd, SAME_GROUP);
        cur_blob = pred_bfnd->entry.Template.bl_ptr2
	  = make_blob(fi,BFNULL, BLNULL);
}


/*
 * make_endblock sets up the statement list to the proper branch according
 * to the control stack's status and then pop the control stack.
 */
void
make_endblock(PTR_BFND p) 
{
	PTR_BFND past_pred;
	PTR_BLOB q;
	void set_blobs();

	if ((pred_bfnd==BFNULL) || (pred_bfnd->control_parent==BFNULL))
	   fatalstr("Illegal End of Block\n",(char *)NULL);
	if (cur_blob->ref==BFNULL) { /* empty block body */
/*	  pred_bfnd->entry.Template.bl_ptr1 = BLNULL; */
	  cur_blob->ref = p;
        }
	else
	   if (p) set_blobs(p, pred_bfnd, SAME_GROUP);
	past_pred = pred_bfnd;
	pred_bfnd = pred_bfnd->control_parent;
	if (pred_bfnd->variant == GLOBAL) parstate = OUTSIDE;
	q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	cur_blob = (q && (q->ref == past_pred)) ?
	             q
		    :follow_blob(pred_bfnd->entry.Template.bl_ptr2);
}


void
make_endif() 
{
	PTR_BFND p, past_pred;
	PTR_BLOB q;
	void set_blobs();

	if ((pred_bfnd==BFNULL) || (pred_bfnd->control_parent==BFNULL)
	     || (pred_bfnd->variant != IF_NODE) &&
	        (pred_bfnd->variant != ELSEIF_NODE)) {
		fatalstr("Illegal endif\n", (char *)NULL);
	}
	if ((!cur_blob->ref) || (cur_blob->ref->variant != CONTROL_END)) {
	   p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	   set_blobs(p, pred_bfnd, SAME_GROUP);
	}
	while (pred_bfnd->variant == ELSEIF_NODE)
	  pred_bfnd = pred_bfnd->control_parent;
	past_pred = pred_bfnd;
	pred_bfnd = pred_bfnd->control_parent;
	q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	cur_blob = (q && (q->ref == past_pred)) ?
	             q
		    :follow_blob(pred_bfnd->entry.Template.bl_ptr2);
/*	cur_blob = follow_blob(pred_bfnd->entry.Template.bl_ptr1);*/
}

void
make_endwhere() 
{
	PTR_BFND p;
	void set_blobs();

	if ((pred_bfnd==BFNULL) || (pred_bfnd->control_parent==BFNULL)
	     || (pred_bfnd->variant != WHERE_BLOCK_STMT)) {
		fatalstr("Illegal endwhere\n", (char *)NULL);
	}
	if ((!cur_blob->ref) || (cur_blob->ref->variant != CONTROL_END)) {
	   p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	   set_blobs(p, pred_bfnd, SAME_GROUP);
	}
	pred_bfnd = pred_bfnd->control_parent;
	cur_blob = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
}

void
make_endextend() 
{
	PTR_BFND p;
	void set_blobs();

	if ((pred_bfnd==BFNULL) || (pred_bfnd->control_parent==BFNULL)
	     || (pred_bfnd->variant != PDO_NODE) &&
	        (pred_bfnd->variant != PSECTIONS_NODE)) {
/*		fatalstr("Illegal end extend[ed] statement\n", (char *)NULL);*/
	}
	if ((!cur_blob->ref) || (cur_blob->ref->variant != CONTROL_END)) {
	   p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	   set_blobs(p, pred_bfnd, SAME_GROUP);
	}
	pred_bfnd = pred_bfnd->control_parent;
	cur_blob = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	if (cur_blob->ref->variant == CONTROL_END)
	cur_blob = follow_blob(pred_bfnd->entry.Template.bl_ptr2);
}

/* 18th, Dec. 90 */
/* modified to take care of PCF FORTRAN's variety of END statements */

PTR_BFND
make_enddoall()
{
        PTR_BFND p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	PTR_BLOB q;

	if ((!pred_bfnd) ||
	    ((pred_bfnd->variant != CDOALL_NODE) &&
	     (pred_bfnd->variant != SDOALL_NODE) &&
	     (pred_bfnd->variant != DOACROSS_NODE) &&
	     (pred_bfnd->variant != CDOACROSS_NODE) &&
             (pred_bfnd->variant != PARDO_NODE) &&
             (pred_bfnd->variant != PROCESS_DO_STAT) &&
             (pred_bfnd->variant != PDO_NODE)))
	   execerr("enddoall statement out of place", (char *)NULL);
	q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	if (q != cur_blob) {
	   pred_bfnd->entry.Template.bl_ptr1 =
                 pred_bfnd->entry.Template.bl_ptr2;
	   pred_bfnd->entry.Template.bl_ptr2 = BLNULL;
	}
	return p;
}

PTR_BFND
make_endprocesses()
{
        PTR_BFND p = get_bfnd(fi,PROCESSES_END, SMNULL, LLNULL, LLNULL, LLNULL);
	PTR_BLOB q;

	if ((!pred_bfnd) || (pred_bfnd->variant != PROCESSES_STAT))
	   execerr("endprocesses statement out of place", (char *)NULL);
	q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	if (q != cur_blob) {
	   pred_bfnd->entry.Template.bl_ptr1 =
                 pred_bfnd->entry.Template.bl_ptr2;
	   pred_bfnd->entry.Template.bl_ptr2 = BLNULL;
	}
	return p;
}


void
make_loop()
{
	PTR_BFND p;
	void set_blobs();

	if ((pred_bfnd->variant != CDOALL_NODE) &&
	    (pred_bfnd->variant != SDOALL_NODE) &&
	    (pred_bfnd->variant != DOACROSS_NODE) &&
	    (pred_bfnd->variant != CDOACROSS_NODE))
	  execerr("loop statement out of place", (char *)NULL);
	p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	set_blobs(p, pred_bfnd, SAME_GROUP);
	cur_blob = pred_bfnd->entry.Template.bl_ptr1
	  = make_blob(fi,(PTR_BFND) NULL, (PTR_BLOB) NULL);
}


/*
 * Setup of logical IF statement
 */
PTR_BFND
make_logif(PTR_BFND if_stmt, PTR_BFND body)
{
/*	PTR_BFND temp; */
        PTR_BLOB new_blob;
	void set_blobs(), make_prog_header();

/*
	if (expr->type->variant != T_BOOL) {
		err("non-logical expression in IF statement");
		expr = LLNULL;
	} 
	if (pred_bfnd->variant == GLOBAL)
	  make_prog_header();
	temp = get_bfnd(fi,LOGIF_NODE, SMNULL, expr, LLNULL, LLNULL);
	set_blobs(temp, pred_bfnd, SAME_GROUP);
*/
	new_blob = make_blob(fi,body, BLNULL);
	if_stmt->entry.Template.bl_ptr1 = new_blob;
	body->control_parent = if_stmt;
	return (if_stmt);
}

void
make_extend()
{
	void execerr(), set_blobs();
	PTR_BFND p, past_pred;
	PTR_BLOB q;

	p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	set_blobs(p, pred_bfnd, SAME_GROUP);
	if ((pred_bfnd->variant == PDO_NODE) &&
            (pred_bfnd->index == EXTEND_NODE))
	{
	     cur_blob = pred_bfnd->entry.Template.bl_ptr2
		                          = make_blob(fi,BFNULL, BLNULL);
        }
	else 
	{
	     past_pred = pred_bfnd;
	     pred_bfnd = pred_bfnd->control_parent;
	     q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	     cur_blob = (q && (q->ref == past_pred)) ?
		         q : follow_blob(pred_bfnd->entry.Template.bl_ptr2);
	}
}

void
make_section_extend()
{
	void execerr(), set_blobs();
	PTR_BFND p, past_pred, set_stat_list();
	PTR_BLOB q;
	
   {
	/* mark end of section */
	p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	set_stat_list(pred_bfnd, p);
	/* mark end of psections's sections */
	p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	set_blobs(p, pred_bfnd, SAME_GROUP);
	/* prepare for extend statements of psections in case */
	if ((pred_bfnd->variant == PSECTIONS_NODE) && 
            (pred_bfnd->index == EXTEND_NODE))
	{
	     cur_blob = pred_bfnd->entry.Template.bl_ptr2
		                          = make_blob(fi,BFNULL, BLNULL);
        }
	else 
	{
	     past_pred = pred_bfnd;
	     pred_bfnd = pred_bfnd->control_parent;
	     q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	     cur_blob = (q && (q->ref == past_pred)) ?
		         q : follow_blob(pred_bfnd->entry.Template.bl_ptr2);
	}
   }
}

void
make_section(PTR_LLND section_name, PTR_LLND wait_list)
{
	void execerr(), set_blobs();
	PTR_BFND p, set_stat_list();
	
   {
	if (pred_bfnd->variant == SECTION_NODE)
	{
	     p = get_bfnd(fi,CONTROL_END, SMNULL, LLNULL, LLNULL, LLNULL);
	     set_stat_list(pred_bfnd, p);
	}
	if ((pred_bfnd->variant == PSECTIONS_NODE) ||
            (pred_bfnd->variant == PARSECTIONS_NODE))
	{
        p = get_bfnd(fi,SECTION_NODE, SMNULL, section_name, wait_list, LLNULL);
	set_stat_list(pred_bfnd, p);
/*	pred_bfnd = p;*/
	}
	else printf("stat.c:make_section: SECTION NODE's attachment point is not a PSECTION NODE\n"); 
	
   }
}

/* 
 * procedure_call gets an id and parameter list.
 * It handles procedures not seen yet
 */
PTR_SYMB
procedure_call(PTR_HASH entry)
{
	register PTR_SYMB symb_ptr;

	entry = correct_symtab(entry, PROCEDURE_NAME);
	for (symb_ptr = entry->id_attr; symb_ptr; symb_ptr = symb_ptr->outer)
		if (symb_ptr->variant == PROCEDURE_NAME)
			return (symb_ptr);

	symb_ptr = get_proc_symbol(entry);
	symb_ptr->variant = PROCEDURE_NAME;
	symb_ptr->type = global_default;
	symb_ptr->entry.proc_decl.seen = NO;
	return (symb_ptr);
}


/*
 * proceduer match_parameters gets a proc_id and param_list and performs type
 * checking
 */
void
match_parameters(PTR_SYMB proc_id, PTR_LLND param_list)
{
	PTR_LLND newll;
	void err();

	newll = make_llnd(fi,FUNC_CALL, param_list, LLNULL, proc_id);
	if (proc_id->entry.proc_decl.seen == YES)
		if (!chk_params(proc_id->entry.proc_decl.in_list, param_list))
			err(" Parameter mismatch ");

 /*
  * if procedure declaration not seen yet and otherwise too add this call to
  * the call list 
  */
	newll->entry.proc.next_call = proc_id->entry.proc_decl.call_list;
	proc_id->entry.proc_decl.call_list = newll;
}


/*
 * chk_params checks formals against actuals -- to be added later
 */
int
chk_params(PTR_SYMB formal, PTR_LLND actual)
{
	return (1);
}


/*
 * set_stat_list  -- links together an old BIF node list with a new one
 *
 * input:
 *	  old_list - old BIF node list
 *	  stat	   - new BIF node
 *
 * output:
 *	  a BIF node list that links these two together
 */
PTR_BFND
set_stat_list(PTR_BFND old_list, PTR_BFND stat)
{
	PTR_BFND ret;
	BOOL    start_new_group = NO;
	void fatal(), set_blobs(), make_prog_header(), close_groups();

	if (!stat) return (old_list);
        /* The proper place for this piece of code is in cur_scope(). */
	/* if (pred_bfnd->variant == GLOBAL)
	    make_prog_header(); */

	switch (stat->variant) {
 	    case (IF_NODE):
	    case (LOGIF_NODE):
	          close_groups();
	          return (old_list);
  	    /* start of group */
            case (INTERFACE_STMT):
            case (STRUCT_DECL):
	    case (CDOALL_NODE): 
	    case (SDOALL_NODE):
	    case (DOACROSS_NODE):
	    case (CDOACROSS_NODE):
	    case (FOR_NODE):
	    case (PROCESS_DO_STAT):
	    case (WHILE_NODE):
	    case (FORALL_NODE):
            case (PARDO_NODE):
            case (PROCESSES_STAT):
	    case (PDO_NODE):
	    case (PARREGION_NODE):
	    case (PSECTIONS_NODE):
            case (PARSECTIONS_NODE):
	    case (SECTION_NODE):
            case (CRITSECTION_NODE):
            case (SINGLEPROCESS_NODE):
	    case (CASE_NODE):
	    case (WHERE_BLOCK_STMT):
	        start_new_group = NEW_GROUP1;
	    /* NO_OP nodes */
	    case (VAR_DECL):
	    case (PARAM_DECL):
	    case (COMM_STAT):
	    case (NAMELIST_STAT):
	    case (PUBLIC_STMT):
	    case (PRIVATE_STMT):
	    case (NULLIFY_STMT):
	    case (ALLOCATE_STMT):
	    case (DEALLOCATE_STMT):
	    case (PROS_COMM):
	    case (DIM_STAT):
            case (PROCESSORS_STAT):
	    case (DATA_DECL):
	    case (EXTERN_STAT):
	    case (INTRIN_STAT):
	    case (EQUI_STAT):
	    case (IMPL_DECL):
	    case (SAVE_DECL):
	    case (INCLUDE_STAT):
	    case (ATTR_DECL):
	    case (INPORT_DECL):
	    case (OUTPORT_DECL):
	    /* not an assignment stat */
	    case (GOTO_NODE):
	    case (ASSGOTO_NODE):
	    case (COMGOTO_NODE):
	    case (ARITHIF_NODE):
	    case (LOOP_NODE):
	    case (EXIT_NODE):
            case (CONT_STAT):
	    case (RETURN_STAT):
	    case (STOP_STAT):
	    case (PAUSE_NODE):
            /* other */
	    case (ASSIGN_STAT):
	    case (STMTFN_STAT):
	    case (SUM_ACC):
	    case (MULT_ACC):
	    case (MAX_ACC):
	    case (MIN_ACC):
	    case (CAT_ACC):
	    case (OR_ACC):
	    case (AND_ACC):
	    case (READ_STAT):
	    case (WRITE_STAT):
	    case (PRINT_STAT):
	    case (BACKSPACE_STAT):
	    case (REWIND_STAT):
	    case (ENDFILE_STAT):
	    case (INQUIRE_STAT):
	    case (OPEN_STAT):
	    case (CLOSE_STAT):
	    case (OTHERIO_STAT):
	    case (FORMAT_STAT):
	    case (PROC_STAT):
            case (PROS_STAT):
            case (PROS_STAT_LCTN):
            case (PROS_STAT_SUBM):
            case (MP_STAT):
            case (MP_STAT_LCTN):
            case (MP_STAT_SUBM):
	    case (ASSLAB_STAT):
            case (LOCK_NODE):
	    case (UNLOCK_NODE):
            case (POST_NODE):
	    case (WAIT_NODE):
	    case (CLEAR_NODE):
            case (POSTSEQ_NODE):
	    case (WAITSEQ_NODE):
	    case (SETSEQ_NODE):
	    case (PRIVATE_NODE):
            case (GUARDS_NODE):
	    case (CYCLE_STMT):
	    case (EXIT_STMT):
	    case (CONTAINS_STMT):
	    case (SWITCH_NODE):
	    case (WHERE_NODE):
	    case (USE_STMT):
	    case (MODULE_PROC_STMT):
	    case (OVERLOADED_ASSIGN_STAT): 
	    case (POINTER_ASSIGN_STAT):
	    case (OVERLOADED_PROC_STAT):
            case (INTENT_STMT):
	    case (CHANNEL_STAT):
	    case (MERGER_STAT):
	    case (MOVE_PORT):
	    case (SEND_STAT):
	    case (RECEIVE_STAT):
	    case (ENDCHANNEL_STAT):
	    case (PROBE_STAT):
	      if (start_new_group) {
		if (stat->variant == CDOALL_NODE
		    || stat->variant == SDOALL_NODE
		    || stat->variant == DOACROSS_NODE
		    || stat->variant == CDOACROSS_NODE)
		  set_blobs(stat, pred_bfnd, NEW_GROUP2);
		else
		  set_blobs(stat, pred_bfnd, NEW_GROUP1);
	      } else
		  set_blobs(stat, pred_bfnd, SAME_GROUP);
		break;
	    case (CONTROL_END):
	    case (PROCESSES_END):
	        make_endblock(stat);
		break;
	    default:
                printf("%d\n", stat->variant);
		err("stat.c:1:Compiler bug -- undesirable state!");
		ret = old_list;
	}
	close_groups();
	return (ret);
}


void
close_groups()
{
     PTR_BFND past_pred;
     PTR_BLOB q;
     
     while (end_group > 0) {
	  past_pred = pred_bfnd;
	  pred_bfnd = pred_bfnd->control_parent;
	  q = follow_blob(pred_bfnd->entry.Template.bl_ptr1);
	  cur_blob = (q && (q->ref == past_pred)) ?
	       q : follow_blob(pred_bfnd->entry.Template.bl_ptr2); 
/*	  cur_blob = follow_blob(pred_bfnd->entry.Template.bl_ptr1); */
	  --end_group;
     }
}


/*
 *  Makes a PROG_HEDR statement for programs which does not start with
 *  a program name.
 */ 
void
make_prog_header()
{
     PTR_BFND first_bfnd;
/*   PTR_BLOB b; */
     void set_blobs();

     parstate = INSIDE;
     set_blobs(first_bfnd, global_bfnd, NEW_GROUP1);
}



PTR_SYMB
proc_decl_init(PTR_HASH entry, int type)
{
	PTR_SYMB symb_ptr;

	undeftype = NO;
	entry = correct_symtab(entry, type);
	mod_offset = yylineno - 1;
	for (symb_ptr = head_symb; symb_ptr; symb_ptr = symb_ptr->thread)
		if ((strcmp(symb_ptr->ident, entry->ident) == 0) &&
		    (symb_ptr->variant == PROCEDURE_NAME))
			break;
	if (!symb_ptr)
		symb_ptr = install_entry(entry, SOFT);
	if (type == FUNCTION_NAME)
	   symb_ptr->type = (vartype == global_default) ?
	                        (undeftype ? global_unknown
			               :impltype[*entry->ident - 'a'])
			          : vartype;

	symb_ptr->variant = type;
	symb_ptr->scope = global_bfnd;
	symb_ptr->entry.proc_decl.seen = YES;
	return (symb_ptr);
}



/*
 * End of declaration section of procedure.  Allocate storage.
 */
void
enddcl()
{
	parstate = INEXEC;
/*
	docommon();
	doequiv();
	docomleng();
*/
}

/*
 * Start a new procedure
 */
void
newprog()
{
	void execerr();

	mod_offset = yylineno - 1;
	/* if (parstate != OUTSIDE)
		execerr("missing end statement", (char *)NULL); */
	parstate = INSIDE;
	setimpl(global_float, 'a', 'z');
	setimpl(global_int, 'i', 'n');
}


/*
 * Main program or Block data
 */
void
startproc(PTR_SYMB prgname, int class)
{
}


/*
 *  pop_lab either marks all labels in inner blocks unreachable
 *    or moves all labels referred to in inner blocks out a level
 */ /*
void
pop_lab()
{
	register PTR_LABEL lp;

	for (lp = head_label; lp; lp = lp->next)
		if (lp->labdefined) {
*/		/* mark all labels in inner blocks unreachable */
/*			if (lp->blklevel > blklevel)
				lp->labinacc = YES;
		} else if (lp->blklevel > blklevel) {
*/		/* move all labels referred to in inner blocks out a level */
/*			lp->blklevel = blklevel;
		}
}
*/

/*
 * make_do handles the DO statement.
 *
 * input:
 *	  label - the label of the last statement of the DO loop
 *	  spec  - list of low level nodes specifying the starting,
 *		  value, end value, and increment of the loop
 *
 * output:
 *	  returns a BIF node which contains the necessary information
 *	  of a loop header
 */
PTR_BFND
make_do(int type, PTR_LABEL label, PTR_SYMB symbol, PTR_LLND start, PTR_LLND end, PTR_LLND stride)
{
       
	PTR_BFND bfnd;
	PTR_LLND ddot;

        if (symbol != SMNULL)
	  symbol->dovar = 1;

	ddot = make_llnd(fi, DDOT, start, end, SMNULL);
	bfnd = get_bfnd(fi, type, symbol, ddot, stride, LLNULL);
	bfnd->entry.for_node.doend = label;
	return (bfnd);
}

/* Assuming where_cond is unused in FORTRAN */
PTR_BFND
make_pardo(int type, PTR_LABEL label, PTR_SYMB symbol, PTR_LLND start, PTR_LLND end, PTR_LLND stride, PTR_LLND qualities)
{
       
	PTR_BFND bfnd;
	PTR_LLND ddot;

	ddot = make_llnd(fi, DDOT, start, end, SMNULL);
	bfnd = get_bfnd(fi, type, symbol, ddot, stride, LLNULL);
	bfnd->entry.for_node.doend = label;
	bfnd->entry.for_node.where_cond = qualities;
	return (bfnd);
}

PTR_BFND
make_processdo(int type, PTR_LABEL label, PTR_SYMB symbol, PTR_LLND start, PTR_LLND end, PTR_LLND stride)
{
       
	PTR_BFND bfnd;
	PTR_LLND ddot;

        symbol->dovar = 1;
	ddot = make_llnd(fi, DDOT, start, end, SMNULL);
	bfnd = get_bfnd(fi, type, symbol, ddot, stride, LLNULL);
	bfnd->entry.for_node.doend = label;
	return (bfnd);
}

PTR_BFND
make_processes()
{
	PTR_BFND p;
	void make_prog_header();

	if (pred_bfnd->variant == GLOBAL)
	  make_prog_header();

	p = get_bfnd(fi, PROCESSES_STAT, SMNULL, LLNULL, LLNULL, LLNULL);
	return (p);
}

void
set_blobs(PTR_BFND stat, PTR_BFND parent, int new_group)
{
	if (((new_group == NEW_GROUP1) && stat->entry.Template.bl_ptr1) ||
	    ((new_group == NEW_GROUP2) && stat->entry.Template.bl_ptr2))
	  return;
	stat->control_parent = parent;
	if (!cur_blob) {
	   (void)fprintf(stderr, "cur_blob is null !\n");
	   return;
	}
	if (cur_blob->ref) {
		cur_blob->next = make_blob(fi,stat, BLNULL);
		cur_blob = cur_blob->next;
	} else 
	  cur_blob->ref = stat;
	if (new_group) {
		pred_bfnd = stat;
		cur_blob = make_blob(fi,BFNULL, BLNULL);
		if (new_group == NEW_GROUP1)
		   stat->entry.Template.bl_ptr1 = cur_blob;
		else
		   stat->entry.Template.bl_ptr2 = cur_blob;
	}
}

