/*
 * Header file for operator.C
 *
 * Stuff to include operators in the call graph.
 * lth@cs.uoregon.edu / november, 1994
 */

/* Linked list of SgTypes */

class SgTypeList {
public:
  SgTypeList( SgType *t, SgTypeList *n ) 
    { type = t; next = n; }
  ~SgTypeList() 
    { if (next != 0) delete next; }
  SgType *type;
  SgTypeList *next;
};

/* According to h/tag.doc, operators are in range 325 - 359.
   However, there are a lot more operators floating around than
   what is found in tag.doc, and in addition, some other idents
   (FUNC_CALL, ARRAY_REF, EXPR_LIST) are used as operators
   in calls to members of optbl. Since memory is plentiful, we
   just define the range 1 ... 999 as valid, even if this
   compromises error checking somewhat.
 */

#define OPTBL_FIRST_OP 1
#define OPTBL_LAST_OP  999

class optbl {
  struct entry {
    /* Structure holds signature of a single operator */
    SgType *lhs;                   /* type of lhs for binops, type of only arg for unops */
    SgTypeList *rhs;               /* type of rhs for binops, 0 for unops, more for multiops */
    SgCallGraphFunRec *funrec;     /* pointer to function record (may be 0) */
    entry *next;                   /* next record in chain */
  };

  struct entry *tbl[ OPTBL_LAST_OP - OPTBL_FIRST_OP + 1 ];  /* table of chains */

  void warning( char *fmt, ... );
  void panic( char *fmt, ... );
  int checkop( int op );
  int normalize( int op );
  int unnormalize( int op );
  int exact_match( SgType *lhs, SgTypeList *rhs, entry *e );
  int approx_match( SgType *lhs, SgTypeList *rhs, entry *e, int op );
  SgType *find_operator_in_base( SgClassType *q, const char *opname );
  const char *opname( int );

public:
  optbl();
  ~optbl();
  void add_operator( int op, SgType *lhs, SgTypeList *rhs, SgCallGraphFunRec *funrec );
  SgCallGraphFunRec *match( int op, SgType *lhs, SgTypeList *rhs );
};

int isOperatorName( SgSymbol *s, int unary=0 );
SgTypeList *buildtype( SgExpression *e );

/* eof */
