#!/usr/local/bin/perl
# *-*-perl-*-*

# Code Search, v0.3
#
#  author:  Juan Villacis
# project:  pC++/Sage++
# started:  05-16-95
# updated:  05-23-95
#
# This program searches through a program's database (as generated by dep2html)
# for various patterns that have been specified by the user.

$DEBUG = '';				  # if non-null, will show the FORM data


# ************************ system dependent stuff *****************************

# SYSDEPENDENTSTUFFGOESHERE


# **************************** some constants *********************************

# NOTE:  "$*TAG" are just names used to index %FORM

$PRJCTAG = "project_name";
$DBSETAG = "database";
$WEBDTAG = "webdir";
$ICONTAG = "icon";
$URLTAG  = "link_home";
$MAILTAG = "mailto";

$TITLE{ $SRCETAG = "source"     } = "Source" ;
$TITLE{ $INCLTAG = "include"    } = "Include";
$TITLE{ $SYSLTAG = "system"     } = "SysLib" ;
$TITLE{ $EXTNTAG = "external"   } = "External" ;

$TITLE{ $FUNCTAG = "function"   } = "function" ;
$TITLE{ $SUBRTAG = "subroutine" } = "subroutine" ;
$TITLE{ $MAINTAG = "mainprog" }   = "main program" ;
$TITLE{ $CLASTAG = "class"      } = "class" ;
$TITLE{ $COLLTAG = "collection" } = "collection" ;

$TITLE{ $CASETAG = "case"       } = "case sensitive";
$TITLE{ $EXACTAG = "exact"      } = "exact";
$TITLE{ $PERLTAG = "perl"       } = "perl regexp";

$PATTTAG = "pattern";
$TITLE{ $HELPTAG = "help"       } = "special symbols";
$TITLE{ $EXAMTAG = "example"    } = "examples";

# *********************** form-related subroutines ****************************

sub readFORM {

    read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
    print "<pre>" if ($DEBUG);
    print "<B>RAW=</B>$buffer\n" if ($DEBUG);
    @pairs = split(/&/, $buffer);
    foreach $pair (@pairs) {
	   print "($pair) \n" if ($DEBUG);
	   $pair =~ s/^\s+//;	  # remove leading spaces
	   $pair =~ s/\s+$//;	  # remove trailing spaces
	   $pair =~ tr/+/ /;	  # replace token divider with space
	   ($name, $value) = split(/=/, $pair);
	   $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
	   $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
	   print "  ($name,$value)\n" if ($DEBUG);
	   if ($FORM{$name}) {
		  $FORM{$name} .= " $value";
	   } else {
		  $FORM{$name} = $value;
	   }
	   print "  FORM{<b>$name</b>} = <b>$FORM{$name}</b> \n" if ($DEBUG);
    }
    print "</pre>" if ($DEBUG);
}


# *********************** print-related subroutines ***************************

sub printHidden {
    local($tag) = shift(@_);

    print "<INPUT TYPE=hidden NAME=\"$tag\" VALUE=\"$FORM{$tag}\">";
}

# *************************** sort routines ***********************************

sub byArgName {	  # we assume all these name+args are to be DISTINCT!
    local($ia)=sprintf("%s%s",$ID_NAME{$a},$ID_ARGLIST{$a});
    local($ib)=sprintf("%s%s",$ID_NAME{$b},$ID_ARGLIST{$b});

    local($la,$lb)=($ia,$ib);
    $la=~y/A-Z/a-z/; $lb=~y/A-Z/a-z/;
    $la cmp $lb;
}

# ***************************** main program **********************************

if ($#ARGV>=0) {	# only use this when making the initial form!

    ($#ARGV<=1) && 
	   die "Usage:  $0 projectname database webdir [-icon webImage] [-url URL] [-mailto user\@address] \n" ;

    $FORM{$PRJCTAG} = shift(@ARGV);
    $FORM{$DBSETAG} = &fixDir(shift(@ARGV));
    $FORM{$WEBDTAG} = &fixDir(shift(@ARGV));

    while ($a=shift(@ARGV)) {
	   ($a eq "-icon")   && ($FORM{$ICONTAG}=shift(@ARGV)) && next;
	   ($a eq "-url")    && ($FORM{$URLTAG}=shift(@ARGV))  && next;
	   ($a eq "-mailto") && ($FORM{$MAILTAG}=shift(@ARGV)) && next;
    }

    $FORM{ICONTAG} = $PCXX_ICON unless ($FORM{$ICONTAG});
    $FORM{URLTAG}  = $LINK_HOME unless ($FORM{$URLTAG});
    $FORM{MAILTAG} = $MAIL_TO   unless ($FORM{$MAILTAG});

    $dbase = $FORM{$DBSETAG};	  # shorthand for awhile ...

    (-r $dbase) || die "$0: Cannot read $dbase ! \n";
    (-s $dbase) || die "$0: Zero length file, $dbase \n";
    (-T $dbase) || die "$0: Not a text file, $dbase \n";

    $FIRST_TIME_THRU = 'true';

} else {			# no arguments, so assume it's an http request

# ---------------- Prelimminary HTML Stuff

    print "Content-type: text/html \n\n";

    &readFORM;
    print "<TITLE>csearch $FORM{$PRJCTAG}</TITLE>\n";
    do {
	   print "<H1><A HREF=\"$FORM{$URLTAG}\"><IMG SRC=\"$FORM{$ICONTAG}\"></A> Code Browser</H1>";
	   print "<HR>\n";
	   print "ERROR:  Must specify <B><I>project</I></B> name and <B><I>database</I></B> file!";
	   print " cannot read $FORM{$DBSETAG}" if ($FORM{$DBSETAG}) ;
	   exit 0;
    } unless ( ($FORM{$PRJCTAG}) && (-r $FORM{$DBSETAG}) ) ;
}

# ---------------- Initialize Defaults for FORM

require "$FORM{$DBSETAG}" ;	  # yet another security hole ...

$FORTRAN = ($LANGUAGE eq "Fortran");

do {
    $FORM{$SRCETAG} = $FORM{$INCLTAG} = $FORM{$SYSLTAG} = ' ';
    if ($FORTRAN) {
	   $FORM{$SUBRTAG} = $FORM{$FUNCTAG} = $FORM{$MAINTAG} = ' ';
    } else {
	   $FORM{$CLASTAG} = $FORM{$COLLTAG} = $FORM{$FUNCTAG} = ' ';
    }
} if ($FIRST_TIME_THRU);

# ---------------- Begin Form

print "<HTML>\n";

# put POST as the method, NOT "post" because Mosaic won't recognize it
# otherwise!  also, only need to define action in the INITIAL document

print "<FORM METHOD=POST" ;
print " ACTION=\"$CSEARCH" ;
print "\">\n" ;

# ---------------- Form State (Level 0)

&printHidden($PRJCTAG);		  # yields project name
&printHidden($DBSETAG);		  # yields local file path to project database
&printHidden($WEBDTAG);		  # yields web path to project database
&printHidden($ICONTAG);		  # yields icon to display at top of the page
&printHidden($URLTAG);		  # yields url to user home page (or link back)
&printHidden($MAILTAG);		  # yields mailto ref at the bottom of the page

# ---------------- Form Overview (Level 1)

print "<H1><A HREF=\"$FORM{$URLTAG}\"><IMG SRC=\"$FORM{$ICONTAG}\"></A> Search <I>$FORM{$PRJCTAG}</I></H1>";
print "<HR> \n";
print "<I>Check the items below to narrow down and/or customize the search</I>\n";
sub option {
    local($tag)=shift(@_);

    local($a,$b,$c);
    $a="<INPUT TYPE=checkbox NAME=\"$tag\" VALUE=\"$tag\" ";
    $b="CHECKED" if ($FORM{$tag});
    $c="><B> $TITLE{$tag} </B>";
    "$a$b$c";
}
print "<PRE>";
print "<FONT SIZE=4><B>   Group</B></FONT> : ";
print &option($SRCETAG), &option($INCLTAG), &option($SYSLTAG), &option($EXTNTAG), "\n";
print "<FONT SIZE=4><B>    Type</B></FONT> : ";
if ($FORTRAN) {
    print &option($SUBRTAG), &option($FUNCTAG), &option($MAINTAG), "\n";
} else {
    print &option($CLASTAG), &option($COLLTAG), &option($FUNCTAG), "\n";
}
print "<FONT SIZE=4><B>     How</B></FONT> : ";
print &option($CASETAG), &option($EXACTAG), &option($PERLTAG), "\n";
print "<FONT SIZE=4><B> Pattern</B></FONT> : ";
print "<INPUT NAME=\"$PATTTAG\" VALUE=\"$FORM{$PATTTAG}\" SIZE=30> \n";
print "<FONT SIZE=4><B>    Help</B></FONT> : ";
print &option($HELPTAG), &option($EXAMTAG), "\n";
do {
    print "<P><HR><P>\n";
    print "<DL><DL>";
    print "<H4>Special Symbols:</H4>\n";
    print "   <B>*</B>    matches <B>0</B> or more characters\n";
    print "   <B>+</B>    matches <B>1</B> or more characters\n";
    print "   <B>?</B>    matches <B>exactly 1</B> character\n";
    print "  <B>[  ]</B>  matches any character <B>in</B> the set [ ]\n";
    print "  <B>[! ]</B>  matches any character <B>not in</B> the set [ ]\n";
    print "   <B>\\c</B>   matches character \`<B>c</B>\' literally\n";
    print "   <B>^</B>    matches from the <B>beginning</B> of the line\n";
    print "   <B>\$</B>    matches to the <B>end</B> of the line\n";
    print "</DL></DL>";
    print "<P>\n";
} if ($FORM{$HELPTAG});
do {
    print "<P><HR><P>\n";
    print "<DL><DL>";
    print "<H4>Examples:</H4>\n";
    print "   match everything that <I>begins</I> with an \`<B>a</B>\'\n\n";
    print "       ^a*\n\n";
    print "   match everything that <I>does NOT</I> begin with an \`<B>a</B>\' or \`<B>B</B>\'\n\n";
    print "       ^[!aB]+\n\n";
    print "   match all multiple argument functions whose <I>first</I> argument\n   is an integer\n\n";
    print "       (int *,*)\n\n";
    print "   match any single letter pointer variable that appears in a\n";
    print "   function's argument list\n\n";
    print "       (+ \\*?)\n";
    print "</DL></DL>\n";
} if ($FORM{$EXAMTAG});
print "</PRE>";

# ---------------- Form Results

sub getIDs {
    local(*GROUP,*list)=@_;

    if ($FORM{$FUNCTAG}) {
	   foreach $cID (keys %GROUP) {
		  foreach $id (split(/\s+/,$GROUP{$cID})) {
			 $list{$id} = ' ';
		  }
	   }
    } else {
	   foreach $cID (keys %GROUP) {
		  $list{$cID} = ' ';
	   }
    }
}

if ($FORTRAN) {
} else {
    ($FORM{$CLASTAG}) && ( %classList=(), &getIDs(*GROUP_CLASS,*classList) );
    ($FORM{$COLLTAG}) && ( %collList=(),  &getIDs(*GROUP_COLLECTION,*collList));
}

sub union {
    local(*GROUP)=@_;

    local(@a)=();
    foreach $t (keys %GROUP) {
	   push(@a,split(/\s+/,$GROUP{$t}));
    }
    @a;
}

sub match {
    print "<P>Looking at: " if ($DEBUG);
    foreach $ID (@_) {
	   print " $ID_NAME{$ID} " if ($DEBUG);
	   if ($FORTRAN) {
		  $MATCH1= ( $FORM{$FUNCTAG} && $GROUP_FUNCTION{$ID}) || 
				 ( $FORM{$SUBRTAG} && $GROUP_SUBROUTINE{$ID}) ||
				 ( $FORM{$MAINTAG} && $GROUP_MAINPROG{$ID} );
	   } else {
		  $MATCH1=( ( $classList{$ID}) || ( $collList{$ID}) ||
				 ( ( ($FORM{$FUNCTAG}) && 
				   (! $FORM{$CLASTAG}) && (! $FORM{$COLLTAG}) ) &&
				   $GROUP_FUNCTION{$ID} ) );
	   }
	   do {
		  $nmatches++;
		  $mList{$ID} = "$prefix$ID.html\">$`<B>$1$2</B>$3$'</A>";
	   } if ( $MATCH1 
		    &&
		    ( (($na="$ID_NAME{$ID}$ID_ARGLIST{$ID}") =~ y/ / /s) || 1 )
		    &&
		    ( ( (! $FORM{$CASETAG}) && ($na=~m~$pattern~i) ) ||
		      ( $na=~m~$pattern~) ) );
    }
    print "done.\n" if ($DEBUG);
    1;
}

$regexp=$pattern=$FORM{$PATTTAG};
do {
    $pattern =~ s/\./\\./g;	  # `.' is now just a `.'
    $pattern =~ s/\?/./g;	  # `?' is now a perl `.'
    $pattern =~ s/\[\^/[\\^/g;  # `^' is now just a `^'
    $pattern =~ s/\[\!/[^/g;    # `!' is now a perl `^'
    $pattern =~ s/([^\\\]])(\*|\+)/$1.$2/g;  # `*|+' now a perl `*|+'
    $pattern = ".$pattern" if ( ($pattern eq "*") || ($pattern eq "+") );
    print "<P><TT>pattern1=<B>$pattern</B></TT><P>\n" if ($DEBUG);
} unless $FORM{$PERLTAG};
$pattern =~ s/([\(\)\{\}])/\\\1/g;   # escape out parens
if ($pattern) {
    $pattern = '(' . $pattern . ')';     # surround with parens for match
    print "<P><TT>pattern2=<B>$pattern</B></TT><P>\n" if ($DEBUG);
    if ($FORM{$EXACTAG}) {
	   $pattern = ('(^|\s+)' . $pattern . '(\s+|$)');
    } else {
	   $pattern =~ y/ / /s;	  # compress blank spaces into one
    }
} else {
    $pattern = '^()$';
}
print "<P><TT>pattern=<B>$pattern</B></TT><P>\n" if ($DEBUG);
($FORM{$SRCETAG}) && ($prefix="<A HREF=\"$FORM{$WEBDTAG}/SOURCE/")
    && &match(&union(*GROUP_SOURCE));
($FORM{$INCLTAG}) && ($prefix="<A HREF=\"$FORM{$WEBDTAG}/INCLUDE/")
    && &match(&union(*GROUP_INCLUDE));
($FORM{$SYSLTAG}) && ($prefix="<A HREF=\"$FORM{$WEBDTAG}/SYSLIB/")
    && &match(&union(*GROUP_SYSLIB));
($FORM{$EXTNTAG}) && ($prefix="<A HREF=\"$FORM{$WEBDTAG}/EXTERNAL/index.html#")
    && &match((keys %GROUP_EXTERNAL));

print "</FORM>\n";

do {
    print "<PRE><HR><P>\n";
    print "Results for \`<B>$regexp</B>\' : ";

    if ($nmatches>0) {
	   print "<FONT SIZE=3><B>$nmatches</B> match";
	   print "es" if ($nmatches>1);
	   print "</FONT>\n";
	   print "<DL><DL>\n";
	   foreach $ID (sort byArgName (keys %mList)) {
		  print "<LI> ",$mList{$ID},"\n";
	   }
	   print "</DL></DL>";
    } else {
	   print "<B><I>None found</I></B>";
    }
    print "</PRE>\n";
} unless ($FIRST_TIME_THRU);

# ---------------- End of Form

print "<HR>\n";
print "<ADDRESS>\n";
print "<A HREF=\"mailto:$FORM{$MAILTAG}\">$FORM{$MAILTAG}</a>\n";
print "</ADDRESS>\n";
print "<A NAME=\"foofoo\"></A>\n";

print "</HTML>\n";

# The End.

1;
