#!/usr/local/bin/perl
# *-*-perl-*-*

# HyperText, v1.1
#
#  author:  Juan Villacis
# project:  pC++/Sage++
# started:  04-04-95
# updated:  09-27-95
#
# This program takes as input C or fortran source file(s) and/or directories
# and runs them through a make program.  Web pages are produced that contain
# links to the hypertexted programs in "alphabetical matrix" order (see the
# variable $COLUMNS_PER_ROW below).  The main web page is called "index.html"
# and the others are named "index-<dir>.html" if directory <dir> was found or
# specified on the command line.  These web pages are created within the
# current working directory.
#
# The following hypertext files are generated for each source program `foo.x'
# in its respective subdirectory:
#
#     file in ./HTML-foo       produced by ...      brief description ...
#     ------------------       ---------------      ---------------------
#         index.html              dep2html           index by name & args
#         form.html               mccp               code browser form
#         search.html             csearch            search code form
# 
# If Hypertext is run with the argument "-combine", then source `foo.x' 
# will be combined with other sources of the same language type into a file
# called `dummy_main.x'.  If Hypertext is run with the argument "-index", 
# then only the HTML page will be generated (no make's will be done).
# 

$DEBUG = '';				  # not used ... yet


# ************************ system dependent stuff *****************************

# SYSDEPENDENTSTUFFGOESHERE

# **************************** some constants *********************************

$PATTERN= '\.(' . $VALID_EXTENSIONS . ')$'; # valid source file extensions
$IGNORE_DIRs = '(\.|\.\.|CVS|HTML-.+)';	# directories to ignore

$COLUMNS_PER_ROW = 3;		  # change this to something else if you want

$FORMSYMBOL = ":";
$SEARCHSYMBOL = "?";
$PARSESYMBOL = "*";


# *************************** sort routines ***********************************

sub byname {
    local($la,$lb)=($a,$b);
    $la=~y/A-Z/a-z/; $lb=~y/A-Z/a-z/;
    $la cmp $lb;
}


# ************************** htmlize sources **********************************

sub htmlize {
    local(*sources,$BASEDIR)=@_;

    local($SUBDIR);
    print "\nFound source file(s) : ",join(", ",@sources),"\n";
    foreach $source (@sources) {
	   $source=~s/$PATTERN$//i;
	   $ext = $1;			  # since we put () around pattern
	   $lang = ($ext=~/f|fcm/i) ? "fortran" : "c"; # what to pass to Makefile
	   $flags = ($lang eq "c") ? $CFLAGS : $FFLAGS;
	   if ($source=~/^(.+)\/([^\/]+)$/) {
		  $source = $2;
		  $SUBDIR = $1;
	   } else {
		  $SUBDIR = $BASEDIR;
	   }
	   chop($pwd=`cd $SUBDIR && pwd`);	# all this so that the freakin' 
	   $ENV{"PWD"}=$pwd;               # Makefile does the right thing !!!
	   $pwd=&fixDir($pwd);	# remove extraneous cruft
	   print "\n   ----------------- doing $source.$ext ------------------\n";
	   if ($INDEXONLY) {
		  $SRC{"$SUBDIR"} .= "$source.$ext ";
	   } else {
		  ($CLEAN) && 
			 (system "(cd $SUBDIR && $MAKE -f $MAKEFILE s=$source e=$ext clean)");
		  $SRC{"$SUBDIR"} .= "$source.$ext "
			 if ( (-r "$SUBDIR/$source.dep") || 
				 ((system "(cd $SUBDIR && $MAKE -f $MAKEFILE s=$source e=$ext $INCLUDE TMPDIR=$TMPDIR $lang flags=$flags)")==0) ) && ($DEPONLY || ((system "(cd $SUBDIR && $MAKE -f $MAKEFILE s=$source e=$ext $how i=\'-icon $ICON\' u=\'-url $URL\' m=\'-mailto $MAILTO\')")==0));
	   }
    }
}

sub readDir {
    local($dir)=shift(@_);

    local(@foo)=();

    if (opendir(FOO,"$dir")) {
	   @foo=readdir(FOO);
	   closedir(FOO);
    }

    @foo;
}


sub getSourceFiles {
    local(*FILES,$SDIR)=@_;
    local(@files)=();
    foreach $s (grep(/$PATTERN/i,@FILES)) {
	   push(@files,$s) if (-T "$SDIR/$s");
    }
    @files;
}

sub searchDir {
    local(@ROOTDIR)=@_;

    return unless (@ROOTDIR);

    while ($ROOTDIR=shift(@ROOTDIR)) {
	   print "\nExamining contents of ROOTDIR=$ROOTDIR \n";

# scan through all the subdirectories, and grab as many source files as
# you can, sticking each one found into @sources stack; also, make the
# source files at the top level ...

	   @FILES = &readDir("$ROOTDIR");

	   next unless (@FILES);

	   &htmlize(*sources,$ROOTDIR,"ROOTDIR") 
		  if (@sources=&getSourceFiles(*FILES,$ROOTDIR)) ;

	   next unless ($RECURSIVE);

#	   @DIRs=grep(-d,@FILES);  # doesn't always work; don't know why ...
	   @DIRs=@FILES;

	   foreach $DIR (@DIRs) {

		  next if ($DIR=~/$IGNORE_DIRs/);

		  $SUBDIR="$ROOTDIR/$DIR";

		  $SUBDIR =~ s/\/\//\//g;

		  next unless (-d $SUBDIR);

		  @SUBFILES = &readDir("$SUBDIR");

		  next unless (@SUBFILES);

		  print "\nMaking $SUBDIR ...\n";

		  do {
			 foreach $d (@SUBFILES) {
				next unless (-d ($sd="$SUBDIR/$d"));
				push(@ROOTDIR,"$sd") unless ($d=~/$IGNORE_DIRs/);
			 }
		  } if ($RECURSIVE);

		  next unless (@sources=&getSourceFiles(*SUBFILES,$SUBDIR));

		  &htmlize(*sources,$SUBDIR);
	   }
    }
}

sub createIndex {
    local($subdirName)=shift(@_);
    local($dirName)=shift(@_);
    local($linkBack)=shift(@_);
    local(*HTML)=shift(@_);

    $indexName = "index" . ( ($subdirName) ? "-$subdirName" : "" ) . ".html";

    do {
	   print "\nNo sources compiled correctly ... web page \`$indexName\' NOT made.\n";
	   return;
    } unless (@sources = sort byname (keys %HTML));

    $nums = $#sources; $cols = $COLUMNS_PER_ROW; $rows = int($nums/$cols);

    $dirName=~s/\/$//;  $dirName=~s/^\.\///;
    ($dirName=~/\/?([^\/]+)$/) ? ($relDirName=$1) : ($relDirName=$dirName);

# now create the html index file ...
    $indexName =~s/\//_/g; # replace '/' with '_' when creating index

    open(HTML,"> $dirName/$indexName");

    print HTML "<TITLE>Table of Hypertexted Projects in $dirName/$subdirName</TITLE>\n";
    print HTML "<H1><A HREF=\"$linkBack\"><IMG SRC=\"$ICON\"></A> <FONT SIZE=6>Hypertexted Projects in <STRONG>$relDirName/$subdirName</STRONG></FONT></H1>\n<HR>\n";

    print HTML "Below is a list of Sage++ programs in <B>$dirName/$subdirName</B> which have been automatically converted to HTML via the <A HREF=\"\">dep2html</A> program.  Clicking on the \' <B>$FORMSYMBOL</B> \' will take you to a code browser for the program.  Clicking on the \` <B>$SEARCHSYMBOL</B> \' will put you into search mode for the program.  Items marked with a \` <B>$PARSESYMBOL</B> \' have generated parse errors.  For more information, please refer to the <A HREF=\"http://www.extreme.indiana.edu:80/metatext/sc95/paper.html\">WIPPSCA</A> paper.<HR>";

    printf("\nMaking web page \`$indexName\' with rows=%s, cols=%s, numElts=%s ... ",$rows,$cols,$nums+1);

    print HTML "<CENTER><TABLE NOBORDER>\n";

    sub doit {
	   $SOURCE=shift(@sources) || return;
	   local($n,$d)=split(/\s+/,$SOURCE);
	   print HTML "<TD><STRONG>";
	   if ($d) {
		  $rdir = "$d/HTML-$n";
		  print HTML "<FONT SIZE=4><A HREF=\"$rdir/form.html\">$FORMSYMBOL</A></FONT> <FONT SIZE=5><A HREF=\"$rdir/index.html\"><I>$n</I></A></FONT> <FONT SIZE=4><A HREF=\"$rdir/search.html\">$SEARCHSYMBOL</A></FONT>";
		  print HTML (-e "$rdir/ERROR")? "<A HREF=\"$rdir/ERROR/index.html\">*</A>" : "";
	   } else {
		  $rrr = $n;
		  $rrr =~s/\//_/g;	  # replace '/' with '_' when creating dir index
		  print HTML "<FONT SIZE=5><A HREF=\"./index-$rrr.html\">$n/</A></FONT>";
	   }
	   print HTML "</STRONG></TD>";
	   print HTML "<TD><PRE>    </PRE></TD>\n";
    }

# get the ones that fit nicely within a matrix ...
    for ($r=1; $r<=$rows; $r++) {
	   print HTML "<TR>\n";
	   for ($c=1; $c<=$cols; $c++) {
		  &doit();
	   }
	   print HTML "</TR>\n";
    }

# get the dregs ...
    for ($r=0; $r<($nums-$rows*$cols+1); $r++) {
	   &doit();
    }

    print HTML "</TABLE></CENTER>\n";

    print HTML "<HR>\n";
    print HTML "<ADDRESS>\n";
    print HTML "<A HREF=\"mailto:$MAILTO\">$MAILTO</a>\n";
    print HTML "</ADDRESS>\n";

    close(HTML);

    print "done.\n";
}


# ***************************** main program **********************************

$|=1;					  # print out stuff IMMEDIATELY
$usage=
"Usage: $0 [ -r ] [ dir1 .. dirN ] [ file1 .. fileN ] [ opts ] \n\
 where\
       -r       recursively descends all specified directories\n\
       `dir'    makes source files in specified directory\n\
       `file'   makes the specified source file\n\
       `opts'   is any of the following options:\n\
                   [ -clean ]        removes .dep files and HTML-*\
                                     subdirectories\
                   [ -combine ]      combine source files of same type\
                                     into one large file via \#INCLUDEs\
                   [ -deponly ]      only generate dep files\
                   [ -index ]        only creates the root index menu\
                                     file (does not pass thru Make)\
                   [ -C flag ]       send \`flag\' to C preprocesor\
                   [ -F flag ]       send \`flag\' to Fortran preprocessor\
                   [ -I(incDir) ]    use include directory \`incDir\'\
                   [ -outdir d ]     write to output directory \`d\'\
                   [ -prefix x ]     x = browser access method to be\
                                     prefixed to the dep2html, mccp,\
                                     and csearch icon links; e.g.,\
                                     use \"http://webServer/\" if you\
                                     want to browse through the server\
                   [ -icon webImg ]  web link to an image file\
                   [ -url URL ]      web link to another URL\
                   [ -mailto addr ]  address to send email to\n\n\
NOTE:  You MUST specify at least one file or directory!\n
";

# read in command line arguments ...
@sources=();
while ($arg=shift(@ARGV)) {
    ($arg=~/^-h(elp)?/)      && ($HELPONLY=' ')          && last;
    ($arg=~/^-clean/)        && ($CLEAN  = ' ')          && next;
    ($arg=~/^-combine/)      && ($COMBINE  = '-cgmtest') && next;
    ($arg=~/^-deponly/)      && ($DEPONLY  = ' ') && next;
    ($arg=~/^-F/)            && ($FFLAGS .= shift(@ARGV) . ' ') && next;
    ($arg=~/^-C/)            && ($CFLAGS .= shift(@ARGV) . ' ') && next;
    ($arg=~/^-index/)        && ($INDEXONLY = ' ')       && next;
    ($arg=~/^-I(\S+)/)       && ($INCLUDE .= "-I$1 ")    && next;
    ($arg=~/^-tmpdir/)       && ($TMPDIR = shift(@ARGV)) && next;
    ($arg=~/^-prefix/)       && ($PREFIX = shift(@ARGV)) && next;
    ($arg=~/^-icon/)         && ($ICON = shift(@ARGV))   && next;
    ($arg=~/^-url/)          && ($URL = shift(@ARGV))    && next;
    ($arg=~/^-mailto/)       && ($MAILTO = shift(@ARGV)) && next;
    ($arg=~/^-r/)            && ($RECURSIVE = ' ')       && next;
    if ( ($arg=~/$PATTERN/) && (-r $arg) && (-T $arg) ) {
	   push(@sources,$arg);
    } elsif ( (-d $arg) && (-w $arg) ) {
	   push(@ROOTDIR,$arg);
    } else {
	   print "Ignoring unknown flag \'$arg\' \n";
    }
}

die $usage if ($HELPONLY);
die "$usage\nUnable to locate VALID source file(s) or directory!\n\n"
    if (! ( (@sources) || (@ROOTDIR) ) );

# quote include variables if user provides them ...
$INCLUDE = "INC=\"$INCLUDE\"" if ($INCLUDE);

# use system defaults if user doesn't provide these ...
$ICON   = $PCXX_ICON unless ($ICON);
$URL    = $LINK_HOME unless ($URL);
$MAILTO = $MAIL_TO   unless ($MAILTO);

# determine the current working directory
chop($dirName = `cd . && pwd`);
$dirName= &fixDir($dirName);

# determine the temp directory
if ($TMPDIR) {
    chop($TMPDIR = `cd $TMPDIR && pwd`);
    $TMPDIR = &fixDir($TMPDIR);   
} else {
    $TMPDIR = $dirName;
}

$url = "$PREFIX/index.html"; # link which "sub" URLs point back to

$how = ($COMBINE) ? "cgmtest" : "html";	# how to initially process each 
                                        # file we encounter
if ($COMBINE) {
    $DUMMY_MAIN = "_main";
    $mainF = "$DUMMY_MAIN.f";
    $mainC = "$DUMMY_MAIN.c";
    if ($CLEAN) {
	   system "$RM -f $dirName/$DUMMY_MAIN.*";
    } else {
	   $SRC2{$mainF} = ' ' if (-r "$dirName/$mainF");
	   $SRC2{$mainC} = ' ' if (-r "$dirName/$mainC");
    }
}

# begin htmlizing command-line source files ...
if (@sources) {
    &htmlize(*sources,$dirName);
}

# begin htmlizing command-line directories ...
if ( (@ROOTDIR) && ( (! $COMBINE) || (! %SRC2)) ) {
    &searchDir(@ROOTDIR);
}

# create/compile combined source files if asked to ...
if ($COMBINE) {
    if (%SRC2) {
	   print "\nRe-using previous combined source file(s) : ",%SRC2;
    } else {
	   print "\nCreating combined source file(s) : ";
	   foreach $sdir (keys %SRC) {
		  foreach $s (split(/\s+/,$SRC{$sdir})) {
			 $s =~s/$PATTERN$//i;
			 $e = $1;
			 $fortran = ($e=~/$FORTRAN_EXTENSIONS/i) ? "fortran" : "";
			 $mainFile = ($fortran) ? $mainF : $mainC;
			 $openHow = ($SRC2{$mainFile}) ? ">>" : ">";
			 print "$mainFile " unless ($SRC2{$mainFile});
			 $SRC2{$mainFile} = ' ';
			 open(GOO,"$openHow $mainFile");
			 print GOO ( ($fortran) ? "\t " : "#" );
			 print GOO "include \"$sdir/$s.$e\"\n";
			 close(GOO);
		  }
	   }
    }
    print "\n";
    $ENV{"PWD"}=$dirName;		  # needed so that make works ...
    undef %SRC;
    foreach $s (keys %SRC2) {
	   $s =~s/$PATTERN$//i;
	   $e = $1;
	   $l = ($e=~/$FORTRAN_EXTENSIONS/i) ? "fortran" : "c";
	   $flags = ($l eq "c") ? $CFLAGS : $FFLAGS;
	   print "\n";
	   $SRC{"$dirName"} .= "$s.$e "
		  if ( (system "(cd $dirName; $MAKE -f $MAKEFILE s=$s e=$e $INCLUDE clean $l flags=$flags TMPDIR=$TMPDIR html i=\'-icon $ICON\' u=\'-url $URL\' m=\'-mailto $MAILTO\')") == 0);
    }
    undef %SRC2;
}

if (! $DEPONLY) {

    $pdirName = $dirName;           # set pattern matching dirName
    $pdirName =~s/(\W)/\\\1/g;      # escape out metacharacters

    foreach $sdir (keys %SRC) {
	   chop($sdN = `cd $sdir && pwd`); # fully-qualified directory name
	   $sdN=&fixDir($sdN);
	   if ($sdN eq $dirName) {
		  foreach $s (split(/\s+/,$SRC{$sdir})) {
			 $HTML{sprintf("%s %s",$s,$sdir)}=' ';
		  }
	   } else {
		  $n = $sdir;
		  $n =~s/^$pdirName//;	  # try to get unique wrt  working dir!
		  $n =~s/^\.?\///;		  # remove beginning '/'
		  $n =~s/\/$//;		  # remove ending '/'
		  $HTML{sprintf("%s",$n)}=' ';
		  foreach $s (split(/\s+/,$SRC{$sdir})) {
			 $HTMLD{sprintf("%s %s",$s,$sdir)}=' ';
		  }
		  &createIndex($n,$dirName,"./index.html",*HTMLD);
		  undef %HTMLD;
	   }
    }

    &createIndex("",$dirName,$URL,*HTML);

    if ($COMBINE) {
	   system "$RM -f $dirName/$DUMMY_MAIN.dep";
    }
}

print "\n";

# The End.

1;
