#!/usr/local/bin/perl
# *-*-perl-*-*

# Master Control/Configuration Panel, v0.92
#
#  author:  Juan Villacis
# project:  pC++/Sage++
# started:  01-04-95
# updated:  05-16-95
#
# This program creates a forms-based html file that can be used to browse
# various pC++ source files for a particular pet-peeve project.

$DEBUG = '';				  # if non-null, will show the FORM data


# ************************ system dependent stuff *****************************

# SYSDEPENDENTSTUFFGOESHERE


# **************************** some constants *********************************

# NOTE:  "$*TAG" are just names used to index %FORM

$PRJCTAG = "project_name";
$DBSETAG = "database";
$WEBDTAG = "webdir";
$ICONTAG = "icon";
$URLTAG  = "link_home";
$MAILTAG = "mailto";

$TITLE{ push(@TAG_ORDER,$t=$ANNOTAG = "Annotations") && $t } = "Annotations" ;
$TITLE{ push(@TAG_ORDER,$t=$SYSLTAG = "System"     ) && $t } = "System Libs" ;
$TITLE{ push(@TAG_ORDER,$t=$CLASTAG = "Class"      ) && $t } = "Classes" ;
$TITLE{ push(@TAG_ORDER,$t=$SRCETAG = "Source"     ) && $t } = "Source Files" ;
$TITLE{ push(@TAG_ORDER,$t=$INCLTAG = "Include"    ) && $t } = "Include Files";
$TITLE{ push(@TAG_ORDER,$t=$COLLTAG = "Collection" ) && $t } = "Collections" ;

$TAG_ROWS = 2;  $TAG_COLS = 3;

$TITLE{ $FUNCTAG = "Function"   } = "Functions of %s" ;
$TITLE{ $METHTAG = "Method"     } = "Methods of %s" ;

$TITLE{ $UPDATE  = "update_form" } = "update";
$TITLE{ $VIEWHOW = "how_to_view" } = "View selected items: ";
$TITLE{ $BACKTAG = "go_back"     } = "<- go back";

$TITLE{ $INLINED_FORMAT = "inlined" } = "inlined";
$TITLE{ $NEWPAGE_FORMAT = "newpage" } = "in a new page";
$TITLE{ $REPORT_FORMAT  = "report"  } = "in report form";

$SHOWTAG = "what_to_show";

$VIEWTAG = "view";

$KEEPTAG = "keep";

sub btag {  sprintf("%s-%s",$VIEWTAG,shift(@_));  }

sub ktag {  sprintf("%s-%s",$KEEPTAG,shift(@_));  }

sub otag {  sprintf("showlist_%s",shift(@_));  }

sub ttag {  sprintf("toggle_%s",shift(@_));  }

sub stag {  sprintf("%s_of_%s",shift(@_),shift(@_));  }

sub utag {  sprintf("updatelist_%s",shift(@_));  }


# *********************** form-related subroutines ****************************

sub readFORM {

    read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'});
    print "<pre>" if ($DEBUG);
    print "<B>RAW=</B>$buffer\n" if ($DEBUG);
    @pairs = split(/&/, $buffer);
    foreach $pair (@pairs) {
	   print "($pair) \n" if ($DEBUG);
	   $pair =~ s/^\s+//;	  # remove leading spaces
	   $pair =~ s/\s+$//;	  # remove trailing spaces
	   $pair =~ tr/+/ /;	  # replace token divider with space
	   ($name, $value) = split(/=/, $pair);
	   $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
	   $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
	   print "  ($name,$value)\n" if ($DEBUG);
	   if ($FORM{$name}) {
		  $FORM{$name} .= " $value";
	   } else {
		  $FORM{$name} = $value;
	   }
	   print "  FORM{<b>$name</b>} = <b>$FORM{$name}</b> \n" if ($DEBUG);
    }
    print "</pre>" if ($DEBUG);
}

sub member {				  # true if $value is a member of $FORM{$tag}
    local($value,$tag) = @_;

    $value =~ s/(\W)/\\\1/g;	# escape out metacharacters!

    ($value) && ($FORM{$tag}=~/$value/) ;
}

sub showIt {

    &member(@_,$SHOWTAG) ;
}

sub viewIt {
    local($tag) = shift(@_);

    (($FORM{&btag($tag)}) || ($FORM{&ktag($tag)})) && (! $FORM{$BACKTAG}) && ((! $FORM{$UPDATE}) || ($FORM{$VIEWHOW} eq $INLINED_FORMAT));
}


# ***************************** fly with files ********************************

sub basename { $_=shift(@_); (/^.+\/([^\/]+)$/) ? $1  : $_; }

sub getSYSfile { sprintf("$BASE_DIR/SYSLIB/%s.html",&basename(@_)); }

sub getINCfile { sprintf("$BASE_DIR/INCLUDE/%s.html",&basename(@_)); }

sub getSRCfile { sprintf("$BASE_DIR/SOURCE/%s.html",&basename(@_)); }

sub getEXTfile { sprintf("$BASE_DIR/EXTERNAL/%s.html",&basename(@_)); }

sub getCCfile {
    local($thing) = shift(@_);

    sub checkGroup {
	   local(*GROUP)=@_;

	   local($found)="";
	   foreach $file (keys %GROUP) {
		  last if ( ($GROUP{$file}=~/\s$id\s/) && ($found='FOUND') );
	   }
	   $found;
    }

    local($id)=($thing=~/.+\((\d+)\)$/) ? $1 : $thing;
    ( &checkGroup(*GROUP_SYSLIB)  &&  &getSYSfile($id) ) ||
    ( &checkGroup(*GROUP_INCLUDE) &&  &getINCfile($id) ) ||
    ( &checkGroup(*GROUP_SOURCE)  &&  &getSRCfile($id) ) || 
    &getEXTfile($id);
}

sub getFile {
	local($foo) = shift(@_);

	($foo=~/\(([^\)]*)\)\s*$/) ? &getCCfile($1) : $foo ;
}


# ************************** how to create lists ******************************

sub fList { 
    local($fist)=shift(@_);

    (sort (split(/\s+/,$fist))) ; 
}

sub ffList {				  # list of files ...
    local(*FGROUP)=@_;

    (keys %FGROUP);
}

sub cList {
    local($cist)=shift(@_);

    local(@sublist) = ();
    while ($cist=~/\s+\((\d+)\)/) {
	   $cist=$';
	   push(@sublist,$1);
    }
    @sublist;
}

sub ccList {
    local(*CGROUP)=@_;

    local(@sublist) = ();
    foreach $elt (keys %CGROUP) {
	   push(@sublist,sprintf("%s (%s)",$ID_NAME{$elt},$elt)) if ($elt);
    }
    @sublist ;
}

sub groupList {			  # list of ID-IDs ...
    local(*IDsOf,$key) = @_;

    local(@sublist) = ();
    foreach $elt (split(/\s+/,$IDsOf{$key})) {
	   push(@sublist,sprintf("%s (%s)",$ID_NAME{$elt},$elt)) if ($elt);
    }
    @sublist ;
}


# *********************** print-related subroutines ***************************

sub printHidden {
    local($tag) = shift(@_);

    print "<INPUT TYPE=hidden NAME=\"$tag\" VALUE=\"$FORM{$tag}\">";
}

sub printFiles {

    local($ttab,$tab,$tag,*listOfThings,*getFileThing) = @_;

    return if ($#listOfThings < 0 );

    do {				# flatten left-indentation if on new page
	   $ttab = "";
	   $tab = " ";
    } if ($viewOnNewPage);

    local($bb,$be) = ($viewOnNewPage) ? ("<H3>","</H3>") : ("<H5> ","</H5>");
    local($i,$min_width,$max_width,$filething) = (0,2,4,"");

    sub fqnhref {				  # fully-qualified href ...
	   local($x)=shift(@_);

	   $x=~s/(\<A HREF=\")\.\/\.\.(\/)/$1$FORM{$WEBDTAG}$2/gi;
	   $x;
    }

    foreach $thing (sort @listOfThings) {
	   if (&member($thing,$tag)) {
		  print "\n$bb $ttab$tab<I>$tag : $thing</I> $be\n";
		  $filething = &getFileThing($thing);
		  if ( ($filething) && (-r $filething) && (! -d $filething) ) {
			 open(FOO,"$filething");
			 # skip past the header (the first <HR>) ...
			 while (<FOO>) {
				last if ((/\<HR\>/i) && ($dregs="$'")) ;
			 }
			 $dregs=~s/\<(\/)?PRE\>//ig; # remove PRE's
			 print $tab,&fqnhref($dregs);
			 # now just print the remainder ...
			 while (<FOO>) {
				s/\<(\/)?PRE\>//ig;	# remove PRE's
				print $tab,&fqnhref($_);
			 }
			 close(FOO);
		  } else {
			 print $tab, "<I>Unable to access file for <B>$thing</B>! ";
			 print "($filething)" if (($DEBUG) && (! -d $filething)) ;
			 print "</I>\n";
		  }
		  print $tab,"\n\n";
	   }
    }
    print $tab, "\n";
}

sub printOptionList {

    local($tab,$title,$tag,*listOfThings,$how) = @_;

    do {
	   &printHidden($tag);
	   return;
    } if ($viewOnNewPage);
    
    print "<A NAME=\"$tag\"></A>";
    print $tab, $title,"\n";
    
    local($size,$min,$max) = ($#listOfThings + 1, 2, 5);
    
    if ($size < 1 ) {
	   print $tab, "<I>None to choose from.</I>\n";
	   return ;
    }
    
    $size = ($size < $min) ? $min : ( ($size > $max) ? $max : $size ) ;
    
    printf("%s<SELECT NAME=\"%s\" SIZE=%d ",$tab,$tag,$size);
    print "MULTIPLE" if ($how eq "view");
    print ">\n";
    
    foreach $thing (sort @listOfThings) {
	   print "<OPTION";
	   print " SELECTED" if (&member($thing,$tag)) ;
	   print "> $thing \n";
    }
    
    print "</SELECT>";
    print "<A NAME=\"",&btag($tag),"\"></A>";
    print " <INPUT TYPE=submit NAME=\"",&btag($tag), "\" VALUE=\"$how\"> ";
}

sub printDualList {
    local($tag,*gList,*GROUP,*sList,*getFileThing,$othertag) = @_;

    local(@listOfThings) = &gList(*GROUP);

    &printOptionList(" ","<H3>$TITLE{$tag}</H3>",$tag,*listOfThings,"view");

    do {
	   local($onoff) = &otag($tag);   # button state (name) variable
	   local($tnoff) = &ttag($onoff); # button hidden state variable
	   local($unoff) = &utag($onoff); # button update-only variable
	   local($checked) = ($FORM{$tnoff} eq "show");
	   local($conoff) = ($FORM{$tag}) ? ( ($FORM{$onoff}) ? ( ($checked) ? "hide" : "show" ) : ($FORM{$tnoff}) ? $FORM{$tnoff} : "show" ) : "show";

	   if ($viewOnNewPage) {

		  &printHidden($unoff);
		  &printHidden($onoff);

	   } else {
		  
		  $checked = ( ($conoff eq "hide") || ($FORM{$unoff}) );
		  
		  print "<I><B>or</B> ";
		  
		  print "<INPUT TYPE=submit NAME=\"$unoff\" VALUE=\"show\">/" unless ($conoff eq "show");
		  
		  print "<INPUT TYPE=submit NAME=\"$onoff\" VALUE=\"$conoff\"> <B>$othertag</B>s</I><A NAME=\"$onoff\"></A>\n";
	   }
	   
	   print "<INPUT TYPE=hidden NAME=\"$tnoff\" VALUE=\"$conoff\">";
	   
	   &printFiles("","  ",$tag,*listOfThings,*getFileThing) if (&viewIt($tag));
	   do {
		  local(@sublist) = (); local($subtag) = "";
		  foreach $subThing (&sList($FORM{$tag})) {
			 @sublist = &groupList(*GROUP,$subThing);
			 $subtag = &stag($othertag,$ID_NAME{$subThing} || $subThing);
			 &printOptionList("\t",sprintf("<H4>\t$TITLE{$othertag}</H4>",$ID_NAME{$subThing} || $subThing),$subtag,*sublist,"view");
			 print $tab, "\n" unless ($viewOnNewPage);
			 &printFiles("\t","\t",$subtag,*sublist,*getFile) if (&viewIt($subtag));
		  } ;
	   } if ( ($checked) || ($viewOnNewPage) );
    } if ($#listOfThings >= 0) ;
    print "<HR>" unless ($viewOnNewPage);
}


# ***************************** main program **********************************

if ($#ARGV>=0) {	# only use this when making the initial form!

    ($#ARGV<=1) && 
	   die "Usage:  $0 projectname database webdir [-icon webImage] [-url URL] [-mailto user\@address] \n" ;

    $FORM{$PRJCTAG} = shift(@ARGV);
    $FORM{$DBSETAG} = &fixDir(shift(@ARGV));
    $FORM{$WEBDTAG} = &fixDir(shift(@ARGV));

    while ($a=shift(@ARGV)) {
	   ($a eq "-icon")   && ($FORM{$ICONTAG}=shift(@ARGV)) && next;
	   ($a eq "-url")    && ($FORM{$URLTAG}=shift(@ARGV))  && next;
	   ($a eq "-mailto") && ($FORM{$MAILTAG}=shift(@ARGV)) && next;
    }

    $FORM{ICONTAG} = $PCXX_ICON unless ($FORM{$ICONTAG});
    $FORM{URLTAG}  = $LINK_HOME unless ($FORM{$URLTAG});
    $FORM{MAILTAG} = $MAIL_TO   unless ($FORM{$MAILTAG});

    $dbase = $FORM{$DBSETAG};	  # shorthand for awhile ...

    (-r $dbase) || die "$0: Cannot read $dbase ! \n";
    (-s $dbase) || die "$0: Zero length file, $dbase \n";
    (-T $dbase) || die "$0: Not a text file, $dbase \n";
    
    $FIRST_TIME_THRU = 'true';

} else {			# no arguments, so assume it's an http request

# ---------------- Prelimminary HTML Stuff

    print "Content-type: text/html \n\n";

    &readFORM;

    print "<TITLE>mccp $FORM{$PRJCTAG}</TITLE>\n";

    do {
	   print "<H1><A HREF=\"$FORM{$URLTAG}\"><IMG SRC=\"$FORM{$ICONTAG}\"></A> Code Browser</H1>";
	   print "<HR>\n";
	   print "ERROR:  Must specify <B><I>project</I></B> name and <B><I>database</I></B> file!";
	   print " cannot read $FORM{$DBSETAG}" if ($FORM{$DBSETAG}) ;
	   exit 0;
    } unless ( ($FORM{$PRJCTAG}) && (-r $FORM{$DBSETAG}) ) ;
}

# ---------------- Initialize Defaults for FORM

require "$FORM{$DBSETAG}" ;	  # yet another security hole ...

do {
    $FORM{$SHOWTAG} = "$SRCETAG";
    $FORM{$VIEWHOW} = "$INLINED_FORMAT";
} if ($FIRST_TIME_THRU);

# ---------------- Check to see if a NON-FORM page is requested

do {
    print "<H2>Sorry, not implemented yet.\n</H2>";
#    exec "$EDOC foo.dbs foo.anno.html";
} if ( ($FORM{&btag($ANNOTAG)}) && ($GROUP_ANNO{$FORM{$ANNOTAG}}) ) ;

$foundTag = (1==2);
foreach $key (keys %FORM) {
    $foundTag = ($key=~/^$VIEWTAG\-(.*)/);
    do {
	   $FORM{&ktag($1)}=$KEEPTAG;
	   $jumpToTag = sprintf("#%s",&btag($1));
	   last;
    } if ($foundTag);
}

$viewOnNewPage = ($FORM{$VIEWHOW} ne $INLINED_FORMAT) && ($foundTag);

# ---------------- Begin Form

# put POST as the method, NOT "post" because Mosaic won't recognize it
# otherwise!  also, only need to define action in the INITIAL document

print "<FORM METHOD=POST" ;
print " ACTION=\"$MCCP" ;
# intent: jump to the place where you last last pressed a submit button
# print "$jumpToTag" ;
print "\">\n" ;

# ---------------- Form State (Level 0)

&printHidden($PRJCTAG);		  # yields project name
&printHidden($DBSETAG);		  # yields local file path to project database
&printHidden($WEBDTAG);		  # yields web path to project database
&printHidden($ICONTAG);		  # yields icon to display at top of the page
&printHidden($URLTAG);		  # yields url to user home page (or link back)
&printHidden($MAILTAG);		  # yields mailto ref at the bottom of the page

foreach $key (keys %FORM) {	# remember which previously viewed items to keep
    &printHidden(&ktag($1)) if ( ($key=~/^$KEEPTAG\-(.*)/) && ($FORM{$1}) ) ;
}

print "<INPUT TYPE=submit NAME=\"$BACKTAG\" VALUE=\"$TITLE{$BACKTAG}\">" if ($viewOnNewPage);

# ---------------- Form Overview (Level 1)


if ($viewOnNewPage) {

    &printHidden($SHOWTAG);
    &printHidden($VIEWHOW);

} else {

    print "<H1><A HREF=\"$FORM{$URLTAG}\"><IMG SRC=\"$FORM{$ICONTAG}\"></A> Code Browser for <I>$FORM{$PRJCTAG}</I></H1>";
    print "<HR> \n";
    printf("<I>Check the items you wish to see, and then press a button to <INPUT TYPE=submit NAME=\"%s\" VALUE=\"%s\"> this page.</I><P><A NAME=\"%s\"></A>\n",$UPDATE,$TITLE{$UPDATE},$UPDATE);
    print "<PRE>";
    for ($x=1; $x<=$TAG_ROWS; $x++) {
	   for ($y=1; $y<=$TAG_COLS; $y++) {
		  last unless ($tag=shift(@TAG_ORDER));
		  $checked = &showIt($tag);
		  print "\t";
		  print "<INPUT TYPE=checkbox NAME=\"$SHOWTAG\" VALUE=\"$tag\" ";
		  print "CHECKED" if ($checked);
		  print "> ";
		  print "<B>" if ($checked);
		  print "<A HREF=\"$MCCP#$tag\">$TITLE{$tag}</A>" ;
		  print "</B>" if ($checked);
		  print " ";
	   }
	   print "\n";
    }
    print "</PRE>\n";

    sub printShowHowTag {
	   local($howVal) = shift(@_);

	   local($checked) = ($FORM{$VIEWHOW} eq $howVal);
	   print "<INPUT TYPE=radio NAME=\"$VIEWHOW\" VALUE=\"$howVal\"";
	   print " CHECKED" if ($checked) ;
	   print "> ";
	   print "<B>" if ($checked) ;
	   print "$TITLE{$howVal} ";
	   print "</B> " if ($checked) ;
    }

    print "<I>$TITLE{$VIEWHOW}</I><TT> ";
    &printShowHowTag($INLINED_FORMAT);
    &printShowHowTag($NEWPAGE_FORMAT);
    print "</TT><P>\n";
}

print "<PRE>"; print "<HR>" unless ($viewOnNewPage) ;

# ---------------- Form Lists and Links (Level 2)

do {
    @list = (keys %GROUP_ANNO);
    &printOptionList(" ","<H3>$TITLE{$ANNOTAG}</H3>",$ANNOTAG,*list,"goto");
    print "<HR>";
} if ( (&showIt($ANNOTAG)) && (! $viewOnNewPage) );

&printDualList($SYSLTAG,*ffList,*GROUP_SYSLIB,*fList,*getSYSfile,$FUNCTAG)
    if (&showIt($SYSLTAG));

&printDualList($INCLTAG,*ffList,*GROUP_INCLUDE,*fList,*getINCfile,$FUNCTAG)
    if (&showIt($INCLTAG));

&printDualList($SRCETAG,*ffList,*GROUP_SOURCE,*fList,*getSRCfile,$FUNCTAG)
    if (&showIt($SRCETAG));

&printDualList($CLASTAG,*ccList,*GROUP_CLASS,*cList,*getCCfile,$METHTAG)
    if (&showIt($CLASTAG));

&printDualList($COLLTAG,*ccList,*GROUP_COLLECTION,*cList,*getCCfile,$METHTAG)
    if (&showIt($COLLTAG));

print "</PRE>\n";

# ---------------- End of Form

print "</FORM>\n";

do {
    print "<ADDRESS>\n";
    print "<A HREF=\"mailto:$FORM{$MAILTAG}\">$FORM{$MAILTAG}</a>\n";
    print "</ADDRESS>\n";
    print "<A NAME=\"foofoo\"></A>\n";
} unless ($viewOnNewPage);

# The End.

1;
