/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/* Modified and overhauled completely by Pete Beckman */


#include <stdio.h>
#include <stdlib.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
/**********************************************************************/
/* main.c
   
   This file calls the rest of the program to parse a pC++ input file and
   create a .dep file.  The .dep file is passed through the Faust system,
   but is not strictly needed to create the .dep file */
/**********************************************************************/

/* Include files */
/********************/
#include "macro.h"        /* All the macros for node manipulation */
#include "vpc.h"   

#define DECLARE
/* declares some of the variables (phb) */
#include "vextern.h"

#include "sage.h"
#include "ext_low.h"
#include "ext_parser_lib.h"
#include "ext_vhash.h"
#include "ext_vinit.h"
#include "ext_utility.h"

/* Define macros */
/********************/
#define  DEP               /* Compile in call to the dependance analyzer */
/* VERSION_NUMBER is defined in ../h/sage.h" */
   
/* External Procedures */ 
/********************/
extern int  yyparse();     /* from yy++.tab.c, created by BISON from y++.y */
extern void init_lex();    /* from vlex.c */
extern void initialize();  /* from vinit.c */
extern void check_usage_of_elementtype();      /* from convert.c */
extern void clean_void_argument();             /* from convert.c */
extern void normalize_overloading_expr();      /* from convert.c */
extern void convert_reference_to_value();      /* from convert.c */
extern void expand_expr_for_struct_fun_ref();  /* from convert.c */
extern void transform_collection();            /* from transform.c */


/* External /lib Procedures */ 
/********************/
extern int write_nodes();      /* from ../lib/writenodes.c */
extern PTR_SETS build_sets();  /* from ../lib/sets.c */
extern void relink();          /* from ../lib/sets.c */

/* External References */
/********************/
extern int yydebug;            /* from yy++.tab.c, from BISON from y++.y */
extern PTR_FILE cur_file;      /* from db.c */
extern FILE *finput;           /* from vlex.c */
extern int TRACEON;            /* from vinit.c */
extern char *first_line_name;  /* from vinit.c */
extern char *new_cur_fname;    /* from vinit.c */
extern int language;           /* External, declared in libdb.a */

/* Global Module Variables */
/********************/
static int unp_flag  = 0 ;
int    show_deps = 0 ; /* will be set to be 1 if '-sd' is specified */
int    store_overloaded_flag = 0;
int    verboseMode = 0; /* Will there be lots of messages to console? */
int    NoWarnings = 0; /* Print warnings by default */
int    field_flag = 0;
int    is_old_DOALL = 0;
int    growing_flag = 0;
int    convert_reference = 0;
int    errors_cmnt = 1;
char *outname;
PTR_FILE fi;

/**********************************************************************/
/* gen_out_name - creates *.dep filename, set!s outname
   
   Input: 
   argv - pointer to char containing pc++ input filename
   Output: 
   outname - modifies global variable, and mallocs space */
/**********************************************************************/
void gen_out_name (argv)
     char *argv;
{
  register int	i;
  register char  *p = argv,*q,*old;
  
  outname = (char *) malloc(strlen(argv)+5);
  for(i = strlen(p), q = outname; i; i--)
    if ((*q++ = *p++) == '.') {
      old= q;
    }
  *old++ = 'd';
  *old++ = 'e';
  *old++ = 'p';
  *old++ = '\0';
}

/**********************************************************************/
/* printUsage - prints command line switches and options
   
   Input:
   execName - pointer to char containing name of executable
   Output: 
   To stderr console */
/**********************************************************************/
void printUsage(execName)
     char* execName;
{
  fprintf(stderr,"Usage: %s\n",execName);
  fprintf(stderr,"    [-help | -h] [-usage | -u]\n");
  fprintf(stderr,"    [-debug] [-verbose | -v] [-version]\n");
  fprintf(stderr,"    [-sd] [-yydebug] [-w]\n");
  fprintf(stderr,"    [-unparse] [-soln] [-OLDDOALL]\n");
  fprintf(stderr,"    [-NEWDOALL] [-cref] [-synset]\n");
  fprintf(stderr,"    [-field] [-trans] [-ne]\n");
  fprintf(stderr,"    [filename]\n");
  fprintf(stderr,
     "\n'%s -help' prints description of switches, parameters\n",execName);
}

/**********************************************************************/
/* printHelp - prints description of switches and options
   
   Input:
   execName - pointer to char containing name of executable
   Output:
   to stdout console */
/**********************************************************************/
void printHelp(execName)
     char* execName;
{
  printf("Usage: %s\n",execName);
  printf("    [-help | -h]  Print this help text\n");
  printf("    [-usage | -u] Print usage info (command line switches)\n");
  printf("    [-debug] Print useful debugging information\n");
  printf("    [-verbose | -v] Turn on Verbose Mode \n");
  printf("    [-version]  Print version number of compiled program\n\n");
  printf("    [-yydebug]  Turn on BISON debugging output\n");
  printf("    [-w]  Turn off warnings\n");
  printf("    [-sd] Show dependance mode\n");
  printf("    [-unparse] Parse input (make .dep) then unparse to stdout\n");
  printf("    [-soln] Set store_overload_flag=1\n");
  printf("    [-OLDDOALL] Set is_old_DOALL=1\n");
  printf("    [-NEWDOALL] Set is_old_DOALL=0 (DEFAULT) \n");
  printf("    [-synset] Set growing_flag=1\n");
  printf("    [-field] Set field_flag=1\n");
  printf("    [-ne] Set errors_cmnt=0\n");
  printf("    [filename]\n\n");
  
  printf(SAGE_INFO);
}

/**********************************************************************/
/* main
   
   Input :
   command line switches and a filename, or stdin */
/**********************************************************************/
int main(argc,argv) 
     int    argc ;
     char   *argv[];
{
  FILE *fopen();
  int i ;
  char *input_filename;
  char *executableName = argv[0]; /* Save a pointer to the exec name */
  TRACEON = 0;  /* should yyparse be traced ? */

  language = CSrc;  /* Set language for .dep file */

  argv++;
  while ((argc > 1) && (*argv)[0] == '-') {
    if ((!strcmp(argv[0],"-help")) || (!strcmp(argv[0],"-h"))) {
      printHelp(executableName); 
      return 1;  /* error: non-zero exit condition */
    }
    else if ((!strcmp(argv[0],"-usage")) || (!strcmp(argv[0],"-u"))) {
      printUsage(executableName); 
      return 1;  /* error: non-zero exit condition */
    }
    else if (!strcmp(argv[0],"-debug")) {
      printf("Suggestion: Use -yydebug switch for BISON's debugging output\n");
      TRACEON = 1;
    }
    else if ((!strcmp(argv[0],"-verbose")) || (!strcmp(argv[0],"-v")))
      verboseMode = 1;
    else if (!strcmp(argv[0],"-version")) {
      printf("This is version %s of %s\n",VERSION_NUMBER,executableName); 
      return 1;  /* error: non-zero exit condition */
    }
    else if (!strcmp(argv[0],"-sd")) {
      printf(" show dependence mode \n");
      show_deps = 1;
    }
    else if (!strcmp(argv[0],"-w"))
      NoWarnings = 1;
    else if (!strcmp(argv[0],"-yydebug"))
      yydebug = 1;
    else if (!strcmp(argv[0],"-unparse"))
      unp_flag = 1;
    else if (!strcmp(argv[0],"-soln"))
      store_overloaded_flag=1;
    else if (!strcmp(argv[0],"-OLDDOALL"))
      is_old_DOALL=1;
    else if (!strcmp(argv[0],"-NEWDOALL"))
      is_old_DOALL=0;
    else if (!strcmp(argv[0],"-synset"))
      growing_flag=1;
    else if (!strcmp(argv[0],"-field"))
      field_flag=1;
    else if (!strcmp(argv[0],"-ne"))
      errors_cmnt = 0;
    else 
      {
	fprintf (stderr, "%s: unrecognized option `%s'\n",
		 executableName, argv[0]);
	printUsage(executableName);
	return 1;   /* error: non-zero exit condition */
      }    
    argc--;
    argv++;
  }
  
  if (*argv) { 
    finput = fopen(new_cur_fname = input_filename = *argv,"r");
    if (finput == NULL) 
      { 
	fprintf(stderr, "Can't open the file %s for read\n", *argv);
	return 1;  /* error: non-zero exit condition */
      }
    gen_out_name(*argv) ;
  }
  else { 
    finput = stdin ;
    input_filename = "stdin";
    new_cur_fname  = "";
    outname = "dep";
  }
  
  cur_file = fi = (PTR_FILE) calloc(1, sizeof(struct file_obj));
  fi->lang = language;
  
  initialize(); 			/* routine in vinit.c */
  init_lex();

  if ( (i = yyparse()) ) 
    {
      fprintf(stderr,"bad parse, return code %d\n", i);
      return 1;  /* error: non-zero exit condition */
    }
  /* this is not set */
  
  Reset_Bif_Next();

  if (first_line_name == 0) 
    {  /* PHB */
      fprintf(stderr,"%s: NULL program, nothing generated\n",executableName);
      return 1;  /* error: non-zero exit condition */
    }
  
  if (!strcmp(first_line_name,""))  
    first_line_name = input_filename;
  clean_void_argument(fi);
  if (unp_flag)
    UnparseProgram(stdout);
  if (verboseMode)
    printf("pC++2dep: Writing Parse Tree (.dep) File: %s\n",outname);
  if (write_nodes(cur_file, outname) < 0)
    fprintf(stderr,"Error, write_nodes() failed\n");
  return 0; /* exit normally */
}
