/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* This file is used to automatically generate a "#include" header */
/* 
mkCextern $SAGEROOT/pC++2dep/parser_lib.c > ! $SAGEROOT/pC++2dep/ext_parser_lib.h
*/

/**************************************************************************
*                   Parser Functions for pC++2dep                         *
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

#include "macro.h"

/* Extern declarations */
#include "ext_lib.h"
#include "ext_low.h"
#include "ext_vinit.h"
#include "ext_vhash.h"
#include "ext_utility.h"
#include "ext_comments.h"

/* basic external variable decls */
#include "vextern.h"

/* Basic defines for pC++ */
#include "vpc.h"

static int StrongChecking = 1;

extern void err(); /* from verrors.c */
extern void err_symb(); /* from verrors.c */
extern void warning(); /* from verrors.c */
char warning_buff[1000];
extern int GLOBAL_DEFAULT_TYPE;
extern PTR_SYMB getFirstFieldOfStruct();
extern int pete;
extern int bif_level_decl_specs;  /* BW */
extern PTR_LLND member_templ_parms;

/**************************************************************************/
/************************* FORWARD DECLARATIONS ***************************/
void eliminate_lastlink();
int map2();
char *ckalloc();

static PTR_SYMB checkExternSecondPart();
static PTR_SYMB find_algebraic_base();
static PTR_HASH find_entry_from_type();
static PTR_TYPE fetch_type_for_constructor();
static PTR_HASH look_up_expanded_string();
static int getSimpleIndirectionLevel();
static int isSameTypeVariant();
/* static int map_status();  BW */
static int isSameType();
static int is_same_entry();

/**************************************************************************/
/********************* New function for Parser ******************************
*                                                                           *
*       Very low level function for the parser construction                 *
*                                                                           *
****************************************************************************/

/* Change this to modify the default variable and function types */
int GLOBAL_DEFAULT_TYPE=T_INT;  

/* 
 *
 * Warning: function here are of two kinds : 
      library function and parser functions. 
 * -------
 * Parser functions are going to be moved somewhere else later. 
 *
 */

/**************************************************************************/
/*****************************************************
  ResetSymbolTable is used to change the scope of parameters
  for a template class or function.  the parser puts the
  symbols at the global scope.  we need to change them to
  be of the current scope.
*******************************************************/
void ResetSymbolTable(b, ll)
PTR_BFND b;
PTR_LLND ll;
{
    PTR_LLND p;
    PTR_SYMB s;
    PTR_TYPE t;
    while(ll){
    	p = NODE_OPERAND0(ll);
	while(p && NODE_SYMB(p) == NULL) p = NODE_OPERAND0(p);
	if(p == NULL) err("error in ResetSymbolTable. contact dbg");
	s = NODE_SYMB(p);
	if(strlen(SYMB_IDENT(s)) == 0){ /* a type parameter */
		SYMB_SCOPE(s) = b;
                t = SYMB_TYPE(s);
		if(t == NULL || TYPE_CODE(t) != T_DERIVED_TYPE)
		  err("wrong type for class argument in template");
	        s = TYPE_SYMB_DERIVE(t);
		if(s == NULL) err("bad derived type: internal error");
		}
	/* printf("ident:%s\n", SYMB_IDENT(s)); */
	SYMB_SCOPE(s) = b;
	ll = NODE_OPERAND1(ll);
	}
}
		
    
/**************************************************************************/
static int getSimpleIndirectionLevel(type)
     PTR_TYPE type;
{
  PTR_TYPE tmp;
  int ind = 0;
  if (!type)
    return 0;
  for (tmp = type; tmp; )
    {
      if (isAtomicType(TYPE_CODE(tmp))) break;
      if (TYPE_CODE(tmp) == T_ARRAY)
	 ind  =  ind + getTypeNumDimension(tmp);
      else
	if (TYPE_CODE(tmp) == T_POINTER)
	  ind = ind + TYPE_TEMPLATE_DUMMY1(tmp);
      if (hasTypeBaseType(TYPE_CODE(tmp))) 
	{
	  if (TYPE_BASE(tmp))
	    tmp  = TYPE_BASE(tmp);
	  else
	    return ind;
	} else
	  return ind;
    }
  return ind;
}

/**************************************************************************/
static int getIndirectionLevel(type)
     PTR_TYPE type;
{
  PTR_TYPE tmp;
  PTR_SYMB symb;
  int ind = 0;
  if (!type)
    return 0;
  for (tmp = type; tmp; )
    {
      if (isAtomicType(TYPE_CODE(tmp))) break;
      if (TYPE_CODE(tmp) == T_ARRAY)
	 ind  =  ind + getTypeNumDimension(tmp);
      else
	if (TYPE_CODE(tmp) == T_POINTER)
	  ind = ind + TYPE_TEMPLATE_DUMMY1(tmp);
      if (hasTypeBaseType(TYPE_CODE(tmp))) 
	{
	  tmp  = TYPE_BASE(tmp);
	} else
	  if (hasTypeSymbol(TYPE_CODE(tmp)))
	    {
	      symb = TYPE_SYMB_DERIVE(tmp);
	      if (symb && SYMB_TYPE(symb))
		tmp  = SYMB_TYPE(symb);
	      else
		return ind;
	    } else
	      return ind;
    }
  return ind;
}

/**************************************************************************/
/* check if type1 a subclass of type two for atomic type */
static int isAtomicTypeAsubClass(type1,type2)
PTR_TYPE type1,type2;
{

  if (!type1 || !type2) return(0);
  if (!isAtomicType(TYPE_CODE(type1)))  return(0);
  if (!isAtomicType(TYPE_CODE(type2)))  return(0);

  if (isSameTypeVariant(type1,type2)) return 1;   /* Beckman */

  if (((TYPE_CODE(type1) == T_CHAR) || (TYPE_CODE(type1) == T_INT) 
       || (TYPE_CODE(type1) == DEFAULT))
      &&
      ((TYPE_CODE(type2) == T_CHAR) || (TYPE_CODE(type2) == T_INT) 
       || (TYPE_CODE(type2) == DEFAULT)))
    return(1);

  if (((TYPE_CODE(type1) == T_DOUBLE) || (TYPE_CODE(type1) == T_FLOAT))
      &&
      ((TYPE_CODE(type2) == T_DOUBLE) || (TYPE_CODE(type2) == T_FLOAT)))
    return(1);

  return 0;
}

/**************************************************************************/
static int isClassOfSuperclass(subclass,superclass)
PTR_TYPE subclass,superclass;
{
  PTR_SYMB subclass_symb,superclass_symb,object;
  PTR_BFND subclass_bfnd,superclass_bfnd;
  PTR_LLND expr_list,list,list1;
  
  if (!subclass || !superclass)
    return(0);

  if (isStructType(TYPE_CODE(subclass)) && isStructType(TYPE_CODE(superclass))) 
    {
      subclass_bfnd= TYPE_COLL_ORI_CLASS(subclass);
      if (!subclass_bfnd)
	return(0);
      subclass_symb = BIF_SYMB(subclass_bfnd);
      superclass_bfnd =TYPE_COLL_ORI_CLASS( superclass);
      if (!superclass_bfnd)
	return(0);
      superclass_symb =  BIF_SYMB(subclass_bfnd);
      if ((!subclass_symb)||(!superclass_symb))
	return(0);
      
      if ( (expr_list =(BIF_LL2(subclass_bfnd))) )
	{
	  for ( list = expr_list; list; list =NODE_OPERAND1(list))
	    { 
	      list1  = Follow_Llnd0(list);
	      if (!list1)
		return(0);
	      object = NODE_SYMB(list1);
	      if (object == superclass_symb) 
		return(1);
	      else 
		if (isClassOfSuperclass(NODE_TYPE(object),superclass)) 
		  return(1);
	    }
	}
    }
  return(0);
 }


/**************************************************************************/
/********************************************************************/
/* type1 : a type                                                   */
/* type2 : a type                                                   */
/* type1==type2 if type type1 is equivalent to type2                */
/* This procedure was added by Beckman to fix the DEFAULT != INT    */
/********************************************************************/
static int isSameTypeVariant(type1,type2)
PTR_TYPE type1,type2;
{
  if ((!type1)&&(!type2)) return(1);
  if ((!type1)||(!type2)) return(0);

  if (TYPE_CODE(type1) == TYPE_CODE(type2))
      return 1;

  if ( ((TYPE_CODE(type1) == GLOBAL_DEFAULT_TYPE) ||
       (TYPE_CODE(type1) == DEFAULT))
      &&
       ((TYPE_CODE(type2) == GLOBAL_DEFAULT_TYPE) ||
       (TYPE_CODE(type2) == DEFAULT)) )
    return 1;

  return 0;
}

/**************************************************************************/
/********************************************************************/
/* type1 : new type                                                 */
/* type2 : old-declared type                                        */
/* type1==type2 if new type type1 is a subclass of type2            */
/********************************************************************/
/* have to be enhance to take into account LONG_SHORT_FLAG .... */
static int isStrictlySameType(type1,type2)
PTR_TYPE type1,type2;
{
   PTR_SYMB c1,c2,symb1,symb2;
   PTR_TYPE b1,b2;

   if ((!type1)&&(!type2)) return(1);
   if ((!type1)||(!type2)) return(0);

   if (type1 == type2)
     return (1);
   
   if (!isSameTypeVariant(type1,type2)) return 0;   /* Beckman */

   if (isAtomicType(TYPE_CODE(type1)))
     return(1);
   if (hasTypeBaseType(TYPE_CODE(type1))) 
     {
       if (TYPE_BASE(type1)){
	 return isStrictlySameType(TYPE_BASE(type1),TYPE_BASE(type2));
	 }
     }   
   if ((TYPE_CODE(type1) ==T_DERIVED_COLLECTION) &&
       (TYPE_CODE(type2) ==T_DERIVED_COLLECTION))
     { /* check for class<nnn> */
       c1 = TYPE_COLL_NAME(type1);
       c2 = TYPE_COLL_NAME(type2);
       b1 = TYPE_COLL_BASE(type1);
       b2 = TYPE_COLL_BASE(type2);
       if (!c1 || !c2 || !b1 || !b2)
	 return(0);
       /* is same entry should be replace by same type */
       if ((c1==c2) && is_same_entry(b1,b2)) return(1);
       if ((c1==c2)&&(TYPE_CODE(b1) ==T_DERIVED_TYPE)&&
	   (TYPE_CODE(b2) ==T_DERIVED_TYPE))
	 {
	   symb1 =TYPE_SYMB_DERIVE(b1);
	   symb2 =TYPE_SYMB_DERIVE(b2);
	   if ((symb1==elementtype_symb)||(symb2==elementtype_symb))
	     return(1);
	 }
     } else 
       if (hasTypeSymbol(TYPE_CODE(type1)))
	 {
	   symb1 = TYPE_SYMB_DERIVE(type1);
	   symb2 = TYPE_SYMB_DERIVE(type2);
	   if (symb1 && symb2)
	     {
	       if (symb1 == symb2)
		 return 1; /* should be replace by same symbol later */
	       else
		 return 0;
	     }
	 }
   return(0);
 }

/**************************************************************************/
/* replace is_same_type_conversion() */ 
/* is the same type should be written later */
static int isSameType(type1,type2)
PTR_TYPE type1,type2;
{  
   PTR_SYMB c1,c2,symb1,symb2;
   PTR_TYPE b1,b2;

   if ((!type1)&&(!type2)) return(1);
   if ((!type1)||(!type2)) return(0);

   if (type1 == type2)
     return (1);
   
   if (TYPE_CODE(type1) != TYPE_CODE(type2))
     {
       if (isAtomicTypeAsubClass(type1,type2))
	 return (1);
       else
	 return(0);
     }

   if (isAtomicType(TYPE_CODE(type1)))
     return(1);

   if (hasTypeBaseType(TYPE_CODE(type1))) 
     {
       if (TYPE_BASE(type1))
	 return isSameType(TYPE_BASE(type1),TYPE_BASE(type2));
     }   

   if (isClassOfSuperclass(type1,type2)) 
     return(1);
   if ((TYPE_CODE(type1) ==T_DERIVED_COLLECTION) &&
       (TYPE_CODE(type2) ==T_DERIVED_COLLECTION))
     { /* check for class<nnn> */
       c1 = TYPE_COLL_NAME(type1);
       c2 = TYPE_COLL_NAME(type2);
       b1 = TYPE_COLL_BASE(type1);
       b2 = TYPE_COLL_BASE(type2);
       if (!c1 || !c2 || !b1 || !b2)
	 return(0);
       /* is same entry should be replace by same type */
       if ((c1==c2) && is_same_entry(b1,b2)) return(1);
       if ((c1==c2)&&(TYPE_CODE(b1) ==T_DERIVED_TYPE)&&
	   (TYPE_CODE(b2) ==T_DERIVED_TYPE))
	 {
	   symb1 =TYPE_SYMB_DERIVE(b1);
	   symb2 =TYPE_SYMB_DERIVE(b2);
	   if ((symb1==elementtype_symb)||(symb2==elementtype_symb))
	     return(1);
	 }
     } else 
       if (hasTypeSymbol(TYPE_CODE(type1)))
	 {
	   symb1 = TYPE_SYMB_DERIVE(type1);
	   symb2 = TYPE_SYMB_DERIVE(type2);
	   if (symb1 && symb2)
	     {
	       if (symb1 == symb2)
		 return 1; /* should be replace by same symbol later */
	       else
		 return 0;
	     }
	 }
   return(0);
 }

/**************************************************************************/
/* replace is_same_type_conversion2() when combine with 
   is_same_type_conversion ( || )*/
static int typePointerForClassOrCollMatch(type1,type2)
PTR_TYPE type1,type2;
{  
   int level1, level2 ;
   PTR_SYMB c1,c2,symb1,symb2;
   PTR_TYPE b1,b2;

   if ((!type1)&&(!type2)) return(1);
   if ((!type1)||(!type2)) return(0);
  /* type_a = lookForInternalBasetype(type1); */
  /* type_b = lookForInternalBasetype(type2); */
   level1 = getSimpleIndirectionLevel(type1);
   level2 = getSimpleIndirectionLevel(type2);
   if ((level1==0)&&(level2==1)) 
     {
       if (isClassOfSuperclass(type1,type2)) 
	 return(1);
       if ((TYPE_CODE(type1) ==T_DERIVED_COLLECTION) &&
	   (TYPE_CODE(type2) ==T_DERIVED_COLLECTION))
	 { /* check for class<nnn> */
	   c1 = TYPE_COLL_NAME(type1);
	   c2 = TYPE_COLL_NAME(type2);
	   b1 = TYPE_COLL_BASE(type1);
	   b2 = TYPE_COLL_BASE(type2);
	   if (!c1 || !c2 || !b1 || !b2)
	     return(0);
	   /* is same entry should be replace by same type */
	   if ((c1==c2) && is_same_entry(b1,b2)) return(1);
	   if ((c1==c2)&&(TYPE_CODE(b1) ==T_DERIVED_TYPE)&&
	       (TYPE_CODE(b2) ==T_DERIVED_TYPE))
	     {
	       symb1 =TYPE_SYMB_DERIVE(b1);
	       symb2 =TYPE_SYMB_DERIVE(b2);
	       if ((symb1==elementtype_symb)||(symb2==elementtype_symb))
		 return(1);
	     }
	 } 
     }
   return(0);
 }

/**************************************************************************/
static int isAboutSameType(type1,type2)
PTR_TYPE type1,type2;
{ 
  PTR_TYPE type_a;
  PTR_TYPE type_b;
  PTR_TYPE basetype,finaltype;
  int extra_level,finallevel;
  int level1,level2;
  if (isSameType(type1,type2))
    return TRUE;
  
  if  (!type1 || !type2)
    return FALSE;
  type_a= FollowTypeBaseAndDerived(type1);
  level1 = getIndirectionLevel(type1);
  type_b= FollowTypeBaseAndDerived(type2);
  level2 = getIndirectionLevel(type2);

  if (!type_a || (TYPE_CODE(type_a) !=T_DERIVED_COLLECTION) )
      return(0);
  
  basetype = TYPE_COLL_BASE(type_a);
  finaltype = FollowTypeBaseAndDerived(basetype);
  extra_level = getIndirectionLevel(finaltype);

  if (!isSameType(finaltype,type_b)) 
      return(0);
  finallevel = level1 + extra_level;
  if (level2== finallevel) 
    return(1);
  return(0);
}

/**************************************************************************/
/* indicate if same function according to the symbols and the argument lists */
static int functionMatch(symb1,symb2,arg_list1,arg_list2)
     PTR_SYMB symb1,symb2,arg_list1,arg_list2;
{
  PTR_SYMB arg1,arg2;

  if (!sameName(symb1,symb2))
    return FALSE;

  /* check for the type first the return type then for the arguments type */
  if (!isSameType(SYMB_TYPE(symb1), SYMB_TYPE(symb2)))
    return FALSE;

  /* check for the argument list */
  for (arg1 =arg_list1,arg2 =arg_list2 ; arg1&& arg2;)
    {
      if ((!arg1 )&&(!arg2)) return(1);
      if (arg1 && SYMB_TYPE(arg1) && TYPE_CODE(SYMB_TYPE(arg1)) == T_VOID)
	return(1);
      if (arg2 && SYMB_TYPE(arg2) && TYPE_CODE(SYMB_TYPE(arg2)) == T_VOID)
	return(1);
      if ((!arg1 )||(!arg2)) return(0);
      if (!isSameType(SYMB_TYPE(arg1),SYMB_TYPE(arg2))) 
	return(0);
      arg1 = SYMB_NEXT_DECL(arg1) ;
      arg2 = SYMB_NEXT_DECL(arg2); 
    }
  if (!arg1 && !arg2)
    return 1;
  else
    return(0);
}

/**************************************************************************/
/* this is almost the same as functionMatch but without any
   type conversions in the arg list                                       */

static int functionExactMatch(symb1,symb2,arg_list1,arg_list2)
     PTR_SYMB symb1,symb2,arg_list1,arg_list2;
{
  PTR_SYMB arg1,arg2;

  if (!sameName(symb1,symb2))
    return FALSE;

  /* check for the type first the return type then for the arguments type */
  if (!isSameType(SYMB_TYPE(symb1), SYMB_TYPE(symb2)))
    return FALSE;

  /* check for the argument list */
  for (arg1 =arg_list1,arg2 =arg_list2 ; arg1&& arg2;)
    {
      if ((!arg1 )&&(!arg2)) return(1);
      if (arg1 && SYMB_TYPE(arg1) && TYPE_CODE(SYMB_TYPE(arg1)) == T_VOID)
	return(1);
      if (arg2 && SYMB_TYPE(arg2) && TYPE_CODE(SYMB_TYPE(arg2)) == T_VOID)
	return(1);
      if ((!arg1 )||(!arg2)) return(0);
      if (!isStrictlySameType(SYMB_TYPE(arg1),SYMB_TYPE(arg2))) 
	return(0);
      arg1 = SYMB_NEXT_DECL(arg1) ;
      arg2 = SYMB_NEXT_DECL(arg2); 
    }
  if (!arg1 && !arg2)
    return 1;
  else
    return(0);
}

/**************************************************************************/
static int exprMatchAnyType(expr)
     PTR_LLND expr;
{
  /* if expression is null then match any type */
  int *res1;

  res1 = evaluateExpression(expr);
  if ((res1[0] != -1) &&
      (res1[1] == 0))
    {
      free(res1);
      return TRUE;
    }
  free(res1);
  return FALSE;
}

/* replace is_function_type_matched() 
   if flag is turned on than Pointer match is also checked */

/**************************************************************************/
static int functionMatchExpr(symb1,arg_list1,expr2, flag)
     PTR_SYMB symb1,arg_list1;
     PTR_LLND expr2;
     int flag;
{
  PTR_SYMB arg1;
  PTR_LLND  expr, ll;

  if (!symb1)
    return FALSE;

  /* check for the argument list */
  if (expr2  && (NODE_CODE(expr2) != EXPR_LIST))
    {
      ll = expr2;
      arg1 =arg_list1;
      if ((!arg1 )&&(!ll)) return(1);
      if (arg1 && SYMB_TYPE(arg1) && TYPE_CODE(SYMB_TYPE(arg1)) == T_VOID)
	return(1);
      if (ll && NODE_TYPE(ll) && TYPE_CODE(NODE_TYPE(ll)) == T_VOID)
	return(1);
      if ((!arg1 )||(!ll)) return(0);
      if (!flag)
	{
	  if (!isSameType(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
	      !exprMatchAnyType(ll))
	    return(0);
	} else
	  {
	    if (!isSameType(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
		!typePointerForClassOrCollMatch(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
		!exprMatchAnyType(ll))
	      return(0);
	  }
    } else
      {
	for (arg1 =arg_list1,expr=expr2 ; arg1&& expr;)
	  {
	    ll = NODE_OPERAND0(expr);
	    if ((!arg1 )&&(!ll)) return(1);
	    if (arg1 && SYMB_TYPE(arg1) && TYPE_CODE(SYMB_TYPE(arg1)) == T_VOID)
	      return(1);
	    if (ll && NODE_TYPE(ll) && TYPE_CODE(NODE_TYPE(ll)) == T_VOID)
	      return(1);
	    if ((!arg1 )||(!ll)) return(0);
	    if (!flag)
	      {
		if (!isSameType(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
		    !exprMatchAnyType(ll))
		  return(0);
		} else
		  {
		    if (!isSameType(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
			!typePointerForClassOrCollMatch(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
			!exprMatchAnyType(ll))
		      return(0);
		  }
	    arg1 = SYMB_NEXT_DECL(arg1);
	    expr = NODE_OPERAND1(expr);
	  }
	if ( !arg1 && !expr)
	   return 1;
	else
	  return(0);
      }
  return 1;
}

/**************************************************************************/
/* does not use the same type checking */
static int functionAboutMatchExpr(symb1,arg_list1,expr2)
     PTR_SYMB symb1,arg_list1;
     PTR_LLND expr2;
{
  PTR_SYMB arg1;
  PTR_LLND  expr, ll;
     
  if (!symb1)
    return FALSE;

  /* check for the argument list */
  if (expr2  && (NODE_CODE(expr2) != EXPR_LIST))
    {
      arg1 =arg_list1;
      ll = expr2;
      if ((!arg1 )&&(!ll)) return(1);
      if (arg1 && SYMB_TYPE(arg1) && TYPE_CODE(SYMB_TYPE(arg1)) == T_VOID)
	return(1);
      if (ll && NODE_TYPE(ll) && TYPE_CODE(NODE_TYPE(ll)) == T_VOID)
	return(1);
      if ((!arg1 )||(!ll)) return(0);
      if (!isAboutSameType(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
	  !exprMatchAnyType(ll))
	return(0);
    } else
      {
	for (arg1 =arg_list1,expr=expr2 ; arg1&& expr;)
	  {
	    ll = NODE_OPERAND0(expr);
	    if ((!arg1 )&&(!ll)) return(1);
	    if (arg1 && SYMB_TYPE(arg1) && TYPE_CODE(SYMB_TYPE(arg1)) == T_VOID)
	      return(1);
	    if (ll && NODE_TYPE(ll) && TYPE_CODE(NODE_TYPE(ll)) == T_VOID)
	      return(1);
	    if ((!arg1 )||(!ll)) return(0);
	    if (!isAboutSameType(SYMB_TYPE(arg1),NODE_TYPE(ll)) &&
		!exprMatchAnyType(ll))
	      return(0);
	    arg1 = SYMB_NEXT_DECL(arg1);
	    expr = NODE_OPERAND1(expr);
	  }
	if ( !arg1 && !expr)
	   return 1;
	else
	  return(0);
      }
  return 1;
}


/**************************************************************************/
/**************************************************************************/
/******************** Begin Parser Functions *************************/
/* functions are going to be moved somewhere else parser dedicated function */
/* Warning We keep vhash.c for the moment */
/**************************************************************************/
/**************************************************************************/

/**************************************************************************/
/* routine we kept from vdecl.c */
static int map_bif_to_type(var_decl)
     int var_decl;
{
  switch (var_decl) {
  case STRUCT_DECL:
    return(T_STRUCT);
  case CLASS_DECL:
    return(T_CLASS);
  case TECLASS_DECL:
    return(T_TECLASS);
  case UNION_DECL:
    return(T_UNION);
  case ENUM_DECL:
    return(T_ENUM);
  case COLLECTION_DECL:
    return(T_COLLECTION);
  }
err("Internal Error: map_bif_to_type, illegal case"); /* (phb) */
abort();
return 0;
}


/**************************************************************************/
/* was get_bfnd() */
PTR_BFND installBifNode(node_type, symb_ptr, ll1, ll2, ll3)
     int     node_type;
     PTR_SYMB symb_ptr;
     PTR_LLND ll1, ll2, ll3;
{
  register PTR_BFND new_bfnd;
  int  line_position;

  if(!isABifNode(node_type))
    {
      Message("installBifNode() called for creating something that is not a bif Node",0);
      abort();
    }
  new_bfnd = (PTR_BFND) newNode(node_type);
  BIF_SYMB(new_bfnd) = symb_ptr;
  BIF_LL1(new_bfnd) = ll1;
  BIF_LL2(new_bfnd) = ll2;
  BIF_LL3(new_bfnd) = ll3;

  if ((node_type != GLOBAL)&&(wait_first_include_name == ON))
    { 
      wait_first_include_name = OFF;
      first_line_name         = new_cur_fname;
      BIF_FILE_NAME(global_bfnd) = find_file_entry(new_cur_fname);
      BIF_LINE(global_bfnd)   = 0;
    }

  BIF_CMNT(new_bfnd) = fetch_comment_queue();
  BIF_FILE_NAME(new_bfnd) = fetch_line_queue( &line_position);
  BIF_LINE(new_bfnd) = line_position;
  BIF_LOCAL_LINE(new_bfnd) = 0; /* relative line number  */
  BIF_DECL_SPECS(new_bfnd) = bif_level_decl_specs;   /* BW */
  bif_level_decl_specs = 0;    
  last_bfnd = new_bfnd;
  return (new_bfnd);
}

/**************************************************************************/
/* Was make_llnd ()*/
PTR_LLND installLowLevelNode(code,ll1,ll2,symb)
     int code;
     PTR_LLND ll1,ll2;
     PTR_SYMB symb;
{
  PTR_LLND tmp;

  if (!isALoNode(code))
    {
      Message("installLowLevelNode() called for creating something that is not a bif Node",0);
      abort();
    }
  tmp = (PTR_LLND) newNode(code);
  NODE_OPERAND0(tmp) = ll1;
  NODE_OPERAND1(tmp) = ll2;

  /* Added for 64bit pointer/int conversion (phb) */
  switch (code) {
    case INT_VAL:
      NODE_INT_CST_LOW(tmp) = (int) symb;
      break;
    case BOOL_VAL:
      NODE_BOOL_CST(tmp) = (int) symb;
      break;
    default:
      NODE_SYMB(tmp) = symb;
      break;
    }
  
  return tmp;
}

/**************************************************************************/
/* was is_complete_chain() check if a type chain is only 
   a right hand side of a type */
static int isTypeCompleteChain(type)
PTR_TYPE type;
{
  PTR_TYPE tmp;

  if (!type)
    return FALSE;
  tmp = lookForInternalBasetype(type);
  if (!tmp)
    return FALSE;

  if (hasTypeBaseType(TYPE_CODE(tmp)))
    return FALSE;
  else
    return TRUE;
}

/**************************************************************************/
/* was find_element() */
static PTR_SYMB getObjectSymbol(etype)
PTR_TYPE etype;
{ PTR_TYPE type;
  PTR_BFND pbf;

  if (getIndirectionLevel(etype))
    return NULL;
  type = FollowTypeBaseAndDerived(etype);
  if (type)
    {
      if (isStructType(TYPE_CODE(type)) ||
	  isEnumType(TYPE_CODE(type)) ||
	  isUnionType(TYPE_CODE(type))
	  )
	{    
	  pbf=TYPE_COLL_ORI_CLASS(type);
	  if (pbf) 
	    return BIF_SYMB(pbf);
	  else
	    return NULL;
	}
    }
  return NULL;
 }
 
/**************************************************************************/
/* replace find_current_superobject */
static PTR_SYMB getEnglobingSuperObject(pred_bfnd)
     PTR_BFND pred_bfnd;
{
  PTR_BFND p;
  PTR_BFND oldp;
  PTR_SYMB functor;

  p =pred_bfnd;
  for (oldp=p ; p; p= BIF_CP(p))
    {
      if ((BIF_CODE(p)==GLOBAL)||(BIF_CODE(p)==COLLECTION_DECL)) 
	break;
      oldp =p;
    }

  if (BIF_CODE(p)==GLOBAL) 
    {
      if (BIF_CODE(oldp)==FUNC_HEDR)
	{
	  functor = BIF_SYMB(oldp);
	  if (SYMB_CODE(functor) !=MEMBER_FUNC) 
	    return(NULL);
	  else
	    return(SYMB_MEMBER_BASENAME(functor));
	}
      else 
	if (BIF_CODE(oldp)==CLASS_DECL)
	  {
	    return(BIF_SYMB(oldp));
	  }
	else 
	  return(NULL);
    }
  if (BIF_CODE(p)==COLLECTION_DECL)
    return(BIF_SYMB(p));
  return(NULL);
}


/**************************************************************************/
static void updateParamScope(functor,bif_ptr)
PTR_SYMB functor;
PTR_BFND bif_ptr;
{
   PTR_SYMB list1 ;
   if (!functor)
     return;   
   for (list1 = SYMB_MEMBER_PARAM(functor); list1 ; list1 = SYMB_NEXT_DECL(list1))
     {
       SYMB_SCOPE(list1) = bif_ptr ;
     }
 }


/* intend to replace find_overloading_function
   Look if parameters matching with a declaration 
   That is between the declaration in the structure and 
   in the code of the function ???
 */

/**************************************************************************/
static PTR_SYMB lookForOverloadingFunction(functor)
     PTR_SYMB  functor;
{
  PTR_SYMB  new_arg_list,arg_list,symb=NULL, stmp=NULL;
  PTR_TYPE new_return_type;

  if (!functor)
    return(NULL);
  if (SYMB_CODE(functor) != MEMBER_FUNC) 
    return(NULL);
 if (!SYMB_MEMBER_LIST(functor))  
   return(NULL);
  
  /* Put everything back to normal setup */
  new_arg_list = SYMB_MEMBER_PARAM(functor);
  arg_list = SYMB_MEMBER_OUTLIST(functor);
  SYMB_MEMBER_PARAM(functor)   = SYMB_MEMBER_OUTLIST(functor);
  SYMB_MEMBER_LIST(functor)    = (PTR_SYMB) NULL;
  SYMB_MEMBER_OUTLIST(functor) = (PTR_SYMB) NULL;
  new_return_type = SYMB_TYPE(functor);
  
  if (SYMB_UD_CHAIN(functor)){ 
    if(!SYMB_TYPE(functor)){
         SYMB_TYPE(functor) = (PTR_TYPE) SYMB_UD_CHAIN(functor);
	 symb = functor;
	 }
    else{
       /* both types are there.  */
       if(isSameType(new_return_type, SYMB_UD_CHAIN(functor)))
	    symb = functor;
       else 
	  symb = SYMB_MEMBER_NEXT(functor);
       }
    }
/* try to find exact match first, if this fails do type
   conversion */
  stmp = symb;
  while(stmp)
    {
      if (SYMB_CODE(stmp) == FIELD_NAME)
	stmp = SYMB_NEXT_FIELD(stmp);
      else  
	{
	  arg_list = SYMB_MEMBER_PARAM(stmp);
	  if (functionExactMatch(functor,stmp,new_arg_list,arg_list))
            break;
	  else   
	    stmp= SYMB_MEMBER_NEXT(stmp);
	}
    }
  if(stmp) {
    symb = stmp;
    }
  else{
  while( symb) 
    {
      if (SYMB_CODE(symb) == FIELD_NAME)
	symb = SYMB_NEXT_FIELD(symb);
      else  
	{
	  arg_list = SYMB_MEMBER_PARAM(symb);
	  if (functionMatch(functor,symb,new_arg_list,arg_list))
            break;
	  else   
	    symb= SYMB_MEMBER_NEXT(symb);
	}
    }
  }
  SYMB_TYPE(functor) = (PTR_TYPE) SYMB_UD_CHAIN(functor);
  if (!symb) {
    /* Message("function type mismatched",0); */
    }
  else {
    SYMB_MEMBER_PARAM(symb) = new_arg_list;
    }
  return(symb);
  
}

/**************************************************************************/
/* this function is intended to replace "chain_type_declarator()" */
PTR_LLND addToDeclaration(ptr1, access_flag, ll_ptr1,update_functor, body_appear)
     PTR_TYPE ptr1;
     PTR_LLND ll_ptr1;
     int access_flag ;
     PTR_SYMB *update_functor;
     int body_appear;
{
  PTR_TYPE oldptr,pp1 ;
  PTR_LLND lptr1,lptr2 ;
  PTR_SYMB symbptr,symb2,symb3;
  int type_flag;

  if (!ptr1 || !ll_ptr1){
    return NULL;
    }

  *update_functor= (PTR_SYMB) NULL;
  pp1 = ptr1;
  type_flag = OFF ;
  if (TYPE_CODE(ptr1) == T_DESCRIPT)
    {  
      if (TYPE_LONG_SHORT(ptr1)&BIT_TYPEDEF)
	type_flag = ON ; 
      else  
	type_flag = OFF ;
    }
  else   
    {  
      if (access_flag) 
	{  
	  pp1 = (PTR_TYPE) newNode (T_DESCRIPT);
	  TYPE_BASE(pp1) = ptr1 ;
	  type_flag = OFF ;
	  TYPE_LONG_SHORT(pp1) |= map_status(access_flag) ;
	}
    }
  /* set type of declaration and update symbol 
   if the type was T_POINTER then you have to add the basetype
   somthing like char *toto, type of toto is T_POINTER but no basetype 
   which is now set by this function */
  for ( lptr1=ll_ptr1 ; lptr1  ; lptr1 = NODE_OPERAND1(lptr1))
    { 
      lptr2 = giveLlSymbInDeclList(lptr1);
      if (!lptr2) {Message("internal error",0); abort();}
      symbptr = NODE_SYMB(lptr2);
      /*
      printf("in addToDeclaration, symbol is %s\n", SYMB_IDENT(symbptr));
      if(!SYMB_TYPE(symbptr)) printf("********* no type yet **********\n");
      else printf("type = %d\n", TYPE_CODE(SYMB_TYPE(symbptr)) );
      */
      if (type_flag == ON) SYMB_CODE(symbptr) = TYPE_NAME ;
      oldptr = lookForInternalBasetype(SYMB_TYPE(symbptr));
      if(oldptr && (TYPE_CODE(oldptr) == T_MEMBER_POINTER)){
             TYPE_COLL_BASE(oldptr) = pp1;
             }
      else if (oldptr && hasTypeBaseType(TYPE_CODE(oldptr)))	
	TYPE_BASE(oldptr) = pp1 ;
      else
	if (oldptr == NULL)
          {
          /*
          if(!SYMB_TYPE(symbptr)) printf("no type ---\n"); 
          printf("setting type of %s to %d\n", SYMB_IDENT(symbptr), pp1->variant);
	  */
          if(!SYMB_TYPE(symbptr)) SYMB_TYPE(symbptr) = pp1; 
          }
/*	else
	  Message("Something wrong in addToDeclaration, type was already defined",0); */

      symb2= checkExternSecondPart(symbptr);
      if(symb2) NODE_SYMB(lptr2) = symb2;
    }

  /* treat function declaration */
  for ( lptr1=ll_ptr1 ; lptr1  ; lptr1 = NODE_OPERAND1(lptr1))
    { 
      lptr2 = giveLlSymbInDeclList(lptr1);      
      if (!lptr2) {Message("internal error",0); abort();}
      symbptr = NODE_SYMB(lptr2);
      if (!symbptr) {Message("internal error",0); abort();}
      
      if ((body_appear==NO)&&((SYMB_CODE(symbptr) == FUNCTION_NAME)||
			      (SYMB_CODE(symbptr) ==MEMBER_FUNC)))
	{
	  updateParamScope(symbptr,lastBifInBlobList(pred_bfnd));
	  /* merge the two declarations of a member function
	     the one in the class and the one outside that class.
	     deal with situation like
	     class xxx {
	     int func (param);
	     }
	     
	     int xxx:: func (param);
	     */
	  if ( (symb3 = lookForOverloadingFunction(symbptr)) )
	    *update_functor = NODE_SYMB(lptr2) = symb3;
	  SYMB_UD_CHAIN(symbptr) = (PTR_BLOB) NULL;
	  SYMB_DECL(symbptr)     = 0;
	}
      if (body_appear != YES)  
	{
	  SYMB_UD_CHAIN(symbptr) = (PTR_BLOB) NULL;
	  SYMB_DECL(symbptr)     = 0;
	}
    }
  NODE_TYPE(ll_ptr1) = ptr1;
  return(ll_ptr1);
}
  
/**************************************************************************/
/* these functions, setPure and setConst are added by dbg to
   use the unused field local_size in the symbol table to indicate
   when a function is a pure virtual function (128) or a
   const function.  later we will need to change the name
   of the field if it is not used in fortran.
*/
int setPure(p)
PTR_LLND p;
{
      PTR_LLND p1;
      PTR_SYMB symb;
      if(!p) {Message("bad pure function",0); abort();}

      p1 = giveLlSymbInDeclList(p);
      if (!p1) {Message("bad pure function name",0); abort();}
      symb = NODE_SYMB(p1);
      SYMB_TEMPLATE_DUMMY8(symb) |=  128;
      return 0;
}

/**************************************************************************/
/* not used? BW Dec 20 1994 */
#if 0
static int setConst(p)
PTR_LLND p;
{
      PTR_LLND p1;
      PTR_SYMB symb;
      if(!p) {Message("bad const function",0); abort();}

      p1 = giveLlSymbInDeclList(p);
      if (!p1) {Message("bad const function name",0); abort();}
      symb = NODE_SYMB(p1);
      SYMB_TEMPLATE_DUMMY8(symb) |=  64;
      return 0;
}
#endif

/*
  was check_field_decl_4 is exactly the same as check_field_decl_3 except
  for that check_field_decl_4 will check the overloading field
  (strong type-checking) and find the exact one.
  check_field_decl_3 only find the  one with the same field name it meets.
  
  argument  (1) object. ident
            (2) range is the same as check_field_decl_3
            (3) expr2 is either an argument or a list of arguments

  non_type_flag: if flag ==0 then the overloaded operator need exeact type
                 if flag ==1 then the the arguments of the 
	     overloaded operator can be a collection of the original arguments
*/

/* use to set the level of checking of checkTypeForField 
 default is StrongChecking = 1*/

/**************************************************************************/
static void setWeakChecking()
{ /* just check ident */
 StrongChecking = 0;
}

/**************************************************************************/
static void setVeryWeakChecking()
{ /* just check ident */
 StrongChecking = -1;
}

/**************************************************************************/
static void setStrongChecking()
{
 StrongChecking = 1;
}

/**************************************************************************/
static PTR_SYMB checkTypeForField(object,operator_ident,expr2,range,non_type_flag)
     PTR_SYMB object;
     char     *operator_ident;
     PTR_LLND expr2;
     int      range;
     int      non_type_flag;
{
  
   PTR_SYMB first,object2,base_name;
   PTR_BLOB blob;
   PTR_LLND ll;
   PTR_BFND pbf;

   pbf  = getObjectStmt(object);
   if (!pbf)
     return NULL;

   /* find first symbol */
   first = NULL;
   if(BIF_SYMB(pbf)){
	first = BIF_SYMB(pbf);
	while(first && (SYMB_SCOPE(first) != pbf)) first = first->thread;
	}
   if(!first)
   for(blob =BIF_BLOB1(pbf);blob; blob=BLOB_NEXT(blob))
     {
       if (BLOB_VALUE(blob)) {
	 if (first = BIF_SYMB(BLOB_VALUE(blob)))  /* needed for enums -jkg */
	   break;
	 if (ll = giveLlSymbInDeclList(BIF_LL1(BLOB_VALUE(blob)))) {
	   first = NODE_SYMB(ll);
	   /* if(SYMB_SCOPE(first) == pbf) break; */
           while(first && (SYMB_SCOPE(first) != pbf)) first = first->thread;
	 }
       }
     }
   if (!first) return(NULL);

   for (; first  ;  )
     { 
       if (StrongChecking == 1)
	 {
	   if ((!strcmp(SYMB_IDENT(first),operator_ident)) &&
	       ((range != ELEMENT_ONLY)||
		((range==ELEMENT_ONLY)&&(SYMB_ATTR(first) & ELEMENT_FIELD))) &&
	       (((!non_type_flag)&& 
		 functionMatchExpr(first,SYMB_MEMBER_PARAM(first),expr2,FALSE))||
		((non_type_flag)&&
		 functionAboutMatchExpr(first,SYMB_MEMBER_PARAM(first),expr2))))
	     return(first);
	 } else
 	if (StrongChecking == 0)
		{
		  if ((!strcmp(SYMB_IDENT(first),operator_ident)) &&
		      ((range != ELEMENT_ONLY)||
		       ((range==ELEMENT_ONLY)&&(SYMB_ATTR(first) & ELEMENT_FIELD))))
		    return(first);
		} 
	else if (!strcmp(SYMB_IDENT(first),operator_ident))
		    break;

       /* first = getClassNextFieldOrMember(first);
	  Trying to go to the next symbol in the class here.
	  The problem is that this doesn't work for ENUM_NAMEs.
	  Why not just use SYMB_NEXT(first)? -jkg */
       /* first = SYMB_NEXT(first); */
       first = first->thread;
       while(first && (SYMB_SCOPE(first) != pbf)) first = first->thread;
       /* this seems right but it fails iostreambuf - dbg
       if(first && (SYMB_SCOPE(first) != pbf)) first = SYMB_NEXT(first);
       */
     }

   if (StrongChecking == -1)
     {
       while (first) 
	 {
	 if (first->variant == MEMBER_FUNC)
	   if (range == FUNCTION_ONLY) 
	     if (!strcmp(SYMB_IDENT(first),operator_ident))
                 return(first);
             else {
		/* first = SYMB_NEXT(first); */
		first = first->thread;
		while(first && (SYMB_SCOPE(first) != pbf)) first = first->thread;
		}
	   else                     
	     err("error: member function not found1");
	 else
	   if (range == FUNCTION_ONLY) {
	     /* first = SYMB_NEXT(first); */
	     first = first->thread;
	     while(first && (SYMB_SCOPE(first) != pbf)) first = first->thread;
	     }
	   else                     
	     if (!strcmp(SYMB_IDENT(first),operator_ident))
                 return(first);
             else {
               /* first = SYMB_NEXT(first); */
	       first = first->thread;
	       while(first && (SYMB_SCOPE(first) != pbf)) first = first->thread;
	       }
       }
       err("error: member function not found3");
       return(NULL);
     }

   if ( (object2=doesClassInherit(pbf)) )
     {
       if ((first =checkTypeForField(object2,operator_ident,
                 expr2,range,non_type_flag)))
	 return(first);
     }

   if (range==ALL_FIELDS) 
     {
       base_name = find_algebraic_base(object);
       if (base_name)
	 return(checkTypeForField(base_name,operator_ident,expr2,ELEMENT_ONLY,non_type_flag));
     }
   return(NULL);
 }

/**************************************************************************/
/* was check_field_decl_3() only check for id not for parameter */
PTR_SYMB weakCheckTypeForField(object,entry,sw,range)
PTR_SYMB object;
PTR_HASH entry ;
int sw ;
int range;
{
  PTR_SYMB symb;
  setWeakChecking();
  symb = checkTypeForField(object,HASH_IDENT(entry),NULL,range,0);
  setStrongChecking();
  return symb;
}


/**************************************************************************/
/* was check_field_decl_2() only check for id not for parameter */
PTR_SYMB veryWeakCheckTypeForField(object,entry,sw)
PTR_SYMB object;
PTR_HASH entry ;
int sw ;
{
  PTR_SYMB symb;
  setVeryWeakChecking();
  symb = checkTypeForField(object,HASH_IDENT(entry),NULL,sw,0);
  setStrongChecking();
  return symb;
}

/**************************************************************************/
/* find_overloading_function2() */
static PTR_SYMB getOverloadingFunction(functor,expr_list)
     PTR_SYMB  functor;
     PTR_LLND expr_list;
{
  PTR_SYMB symb;

  if (!functor)
    return NULL;
  
  if (SYMB_CODE(functor) != MEMBER_FUNC) 
    return(NULL);

  for ( symb = functor; symb;   )
    {
      if (SYMB_CODE(symb) == MEMBER_FUNC)
	{
	  if ((!strcmp(SYMB_IDENT(functor),SYMB_IDENT(symb))) && 
	      functionMatchExpr(symb,SYMB_MEMBER_PARAM(symb),expr_list,TRUE))
	    break;
	}
      symb = getClassNextFieldOrMember(symb);
    }
  
  if (!symb) 
    {
      if ((!strcmp(SYMB_IDENT(functor),"DoSubSet"))||
	  (!strcmp(SYMB_IDENT(functor),"DOSUBSET")))
	{
	  warning("DoSubSet operator is used");
     }
      else 
	if ((!strcmp(SYMB_IDENT(functor),"Doall"))||
	    (!strcmp(SYMB_IDENT(functor),"DOALL")))
	  {
	    warning("Doall operator is used");
	  }
	else 
	  { sprintf(warning_buff, "%s :possible member function arg type missmatch",  functor->ident);
	    warning(warning_buff,functor->ident);
	  }
      return(functor);
    }
  return(symb);
}

/**************************************************************************/
/**************************************************************************/
/*  Was Find_overloaded_operator()                                        */
/*  expr1: the object expression                                          */
/*  expr2: the argument expression                                        */
/*  operator: the variant of the C operator                               */
/*  flag : Unary                                                          */
/*         Binary                                                         */
/*         Multiple (in this case, expr2 is expr_list)                    */
/**************************************************************************/
static PTR_SYMB lookForOverloadOperator(expr1,expr2, operator_ident)
PTR_LLND expr1,expr2;
char *operator_ident;
{
  PTR_SYMB object,symb,element,object_base;
  PTR_TYPE type,elementtype;
  PTR_BFND pbf;   
  
  element = symb =  object = NULL;
  elementtype =NULL;
  type = FollowTypeBaseAndDerived(NODE_TYPE(expr1));

  if (!type) return(NULL);
  
  if (getIndirectionLevel(NODE_TYPE(expr1))==0)
    {
      if (TYPE_CODE(type) ==  T_DERIVED_COLLECTION)
	{
	  object = TYPE_COLL_NAME(type);
	  elementtype = TYPE_COLL_BASE(type);
	} else
	  if (isStructType(TYPE_CODE(type)))
	    {    
	      pbf=TYPE_COLL_ORI_CLASS(type);
	      if (pbf) object= BIF_SYMB(pbf);
	    } else
	      {
		if (isElementType(type))
		  { /* check thing otherwise not */
		    object_base=  getEnglobingSuperObject(pred_bfnd);
		      if (object_base)
		      symb=checkTypeForField(object_base,operator_ident,expr2,ELEMENT_ONLY,0);
		    else   
		      symb =NULL;
		    return(symb);
		  }
	      }
      if (object) 
	{
	  symb = checkTypeForField(object,operator_ident,expr2,ALL_FIELDS,0);
	}
      if (!symb) 
	{
	  if (elementtype) 
	    element = getObjectSymbol(elementtype);
	  /* was element = find_element(elementtype); */	  
	  if (element) 
	    {
	      symb=checkTypeForField(element,operator_ident,expr2,CLASS_ONLY,0);
	      /* we are looking at a collection and elementwise parameter */
	      if (!symb) 
		symb=checkTypeForField(element,operator_ident,expr2,CLASS_ONLY,1);
	    }
	}
    }
return(symb);
}
  
/**************************************************************************/
/* was type_conversion need to be enhance... */
PTR_TYPE typeConversion(type1,type2)
PTR_TYPE type1,type2;
{  
  PTR_TYPE type_a,type_b ;
  int level1, level2;
  PTR_SYMB collection_name;
  PTR_TYPE base_type,type3;
  
  if (isStrictlySameType(type1,type2))
    return type1;

  if(!type1 || !type2)
    return NULL;

  /* don't go thru T_DERIVED TYPE in the old version that 
   routine may have been confused with symbol return by 
   follow_type_1 */
  type_a = lookForInternalBasetype(type1);
  type_b = lookForInternalBasetype(type2);

  level1 = getSimpleIndirectionLevel(type1);
  level2 = getSimpleIndirectionLevel(type2);

  if ((type_a == type_b) && (level1 == level2)) 
    return(type1);
  
  if(!type_a || !type_b)
    return(type1);

  /* a pointer can be cast to int or char level indicate if pointer */
   if ((TYPE_CODE(type_a)== T_CHAR) || (TYPE_CODE(type_a) ==T_INT)
       || (TYPE_CODE(type_a) == DEFAULT))
     { 
       if (level1 <= level2) 
	 return(type2);      
       else                  
	 return(type1);      
     }

  /* a pointer can be cast to int or char */
   if ((TYPE_CODE(type_b)== T_CHAR) || (TYPE_CODE(type_b) ==T_INT)
       || (TYPE_CODE(type_b) ==DEFAULT))
    { 
      if (level2 <= level1) 
	return(type1);
      else                  
	return(type2);
    }

  if ((((TYPE_CODE(type_a) == T_DOUBLE)||(TYPE_CODE(type_a)==T_FLOAT))&&
       ((TYPE_CODE(type_b) == T_DOUBLE)||(TYPE_CODE(type_b)==T_FLOAT)))
      && (level1==level2)) 
    { 
      if (TYPE_CODE(type_a)==T_DOUBLE) 
	return(type1);
      else                           
	return(type2);
    }

   if ((TYPE_CODE(type_a)==T_DERIVED_COLLECTION)&&
       (TYPE_CODE(type_b)==T_COLLECTION))
     {
       collection_name=TYPE_COLL_NAME(type_a);
       base_type= TYPE_COLL_BASE(type_a);
       type3=FollowTypeBaseAndDerived(SYMB_TYPE(collection_name));
       if (isElementType(base_type) && (type3==type_b)&&
	   (level1==level2))
	 return(type1);
     }
   if ((TYPE_CODE(type_b)==T_DERIVED_COLLECTION)&&
       (TYPE_CODE(type_a)==T_COLLECTION))
     {
       collection_name=TYPE_COLL_NAME(type_b);
       base_type=TYPE_COLL_BASE(type_b);
       type3=FollowTypeBaseAndDerived(SYMB_TYPE(collection_name));
       if (isElementType(base_type) && (type3 == type_a)&&
	   (level1==level2))
	 return(type2);
     }

   if ((TYPE_CODE(type_b)==T_DERIVED_COLLECTION)&&
       (TYPE_CODE(type_a)==T_DERIVED_COLLECTION))
     {
       if (TYPE_COLL_NAME(type_b)== TYPE_COLL_NAME(type_a))
	 {
	   return(type1);
	 }
     }
   if((TYPE_CODE(type1) != T_DERIVED_TYPE) &&
      (TYPE_CODE(type2) != T_DERIVED_TYPE))
          warning("illegal type combination",0);
   return(type1);

 }

/**************************************************************************/
/* was adjust_element_type */
PTR_TYPE adjustElementType(pllnd,field_type)
     PTR_LLND pllnd;
     PTR_TYPE field_type;
{
  PTR_TYPE type,element_type,new_type;

  if (!pllnd || !field_type)
    return NULL;

  if (NODE_CODE(pllnd)==THIS_NODE) 
    return(field_type);
  
  type = field_type;
  type = FollowTypeBaseAndDerived(type);
  if (!isElementType(type))
    return(field_type);

  /* now check if the LHS's type  is T_DERIVED_COLLECTION */
  type = NODE_TYPE(pllnd);
  type = lookForInternalBasetype(type);
  if (type && TYPE_CODE(type) == T_DERIVED_COLLECTION) 
    {
      element_type= TYPE_COLL_BASE(type);
      /* keep that one, to check if useful, will need more work 
       should use my routines to replace types instead */
      new_type = substitute_element_type(field_type, element_type);
      return(new_type);
    }
  return(field_type);
}


/**************************************************************************/
/* was adjust_array_type()
 *
 * <1> Given the array symbol in the primary_expression
 * <2> Since the A and A[i] and A[i][j] means different type, this routine
 *     adjust these difference.
 * <3> If the array range_list appear later, the next routine adjust_deref_type
 *     will work correspondingly.
 */
PTR_TYPE adjustArrayType(symb_ptr)
     PTR_SYMB symb_ptr;
{  
  int counter ;
  PTR_TYPE n_type,new_base_type ;
  PTR_LLND ll_ptr2 ;
   
  counter = 0 ;
  if (!symb_ptr) 
    err("member instances/methods not found");
  n_type = SYMB_TYPE(symb_ptr);
  if (n_type &&(TYPE_CODE(n_type) == T_ARRAY))
    {   
      counter = exprListLength(TYPE_DECL_RANGES(n_type));
      ll_ptr2 = TYPE_DECL_RANGES(n_type);
      if (counter)
	{  
	  n_type = (PTR_TYPE) newNode (T_POINTER);
	  new_base_type = (PTR_TYPE) newNode (T_ARRAY);
	  TYPE_DECL_RANGES(new_base_type) = NODE_OPERAND1(ll_ptr2);
	  TYPE_DIM(new_base_type) =  counter -1;
	  TYPE_DECL_BASE(new_base_type) = TYPE_DECL_BASE(SYMB_TYPE(symb_ptr));
	  TYPE_BASE(n_type) =new_base_type;
	}
      return(n_type);
    }
  return(n_type);
} 

/**************************************************************************/
/* was adjust_deref_type() */
PTR_TYPE adjustDerefType(type, index)
     PTR_TYPE type ;
     int index ;
{ 
  PTR_TYPE n_type ;
  switch (index) 
    {
    case DEREF_OP:
      { 
        /* i think that this is wrong.  if it is
           a defref op then just go to base type. dbg
        */
	n_type = (PTR_TYPE) newNode (T_POINTER) ;
        TYPE_TEMPLATE_DUMMY1(n_type) = -1 ;
        TYPE_BASE(n_type) = type ;
      }
      break;
    case SCOPE_OP:
      {   
	n_type = (PTR_TYPE) newNode (T_MEMBER_POINTER) ;
	TYPE_COLL_NAME(n_type) = SMNULL;
        TYPE_BASE(n_type) = type ;
      }
      break;
    case ADDRESS_OP:
      {   
	n_type = (PTR_TYPE) newNode (T_POINTER) ;
        TYPE_BASE(n_type) = type ;
      }
      break;
    }
  return(n_type); 
}

/**************************************************************************/
/*
  Was typeconversion_or_oo_handling
  Do the conventional type_conversion 
  or 
  processing the overloaded operator
  Limitation: Only used for binary or unary overloaded operator.
              [], SIZEOF,CAST_OP, (), NEW, and Delete, are excluded

  flag: if (flag) then type_conversion is needed
*/
int typeConversionOrObjectHandling(expr1,expr2, op,expr3,flag)
     PTR_LLND expr1,expr2;
     PTR_LLND expr3;
     int flag;
char *op;
{
  PTR_SYMB symb;

  if (!expr3)
    return 0;

  symb=lookForOverloadOperator(expr1,expr2, op);
  if (!symb) 
    {
      if (flag) 
	NODE_TYPE(expr3) = typeConversion(NODE_TYPE(expr1),NODE_TYPE(expr2));
      return(0);
    }
  else 
    {
      NODE_TYPE(expr3) = adjustElementType(expr1,SYMB_TYPE(symb));
      if (store_overloaded_flag) {
	if ((NODE_CODE(expr3) == FUNC_CALL)||(NODE_CODE(expr3) == ARRAY_REF))
	  NODE_OPERAND1(expr3) = newExpr(VAR_REF,SYMB_TYPE(symb),symb);
	else
	  NODE_SYMB(expr3) = symb;
      }
      return(1);
    }
}


/**************************************************************************/
/* Was lazy_install(entry , variant,err_sw) */
PTR_LLND lazyInstall(pred_bfnd, entry , variant,err_sw)
PTR_BFND pred_bfnd;
PTR_HASH entry;
int variant ;
int err_sw;
{  int sw ;
   int decl_temp;
   PTR_LLND ll_node,expr_llnd;
   PTR_BFND bf1 ;
   PTR_SYMB  name, symb;
   PTR_BFND adj_scope;

   if (!pred_bfnd || !entry)
     return NULL;

   if ((BIF_CODE(pred_bfnd) == FUNC_HEDR)&& 
       (BIF_SYMB(pred_bfnd)) &&
       (SYMB_DOVAR(BIF_SYMB(pred_bfnd)) >= XDECL ))
     {
       new_cur_symb = lookForNameInParamList(BIF_SYMB(pred_bfnd),HASH_IDENT(entry));
       if (!new_cur_symb) 
	 { 
	   err("Unmatched Parameter symbol ");
	   ll_node = LLNULL ;
	 }
       else 
	 {
	   ll_node = (PTR_LLND) newNode(map2(variant));
	   NODE_SYMB(ll_node) = new_cur_symb;
	   new_cur_type = (PTR_TYPE) NULL ;
	   cur_flag = RULE_ID ;
	 }
     }
   else
     { /* consider structure here */   
       /* sw means
	ONE : this is not a struct or union or enum 
	TWO : no first symbol
	THREE : first symbol exist, the new symbol is going to be
	        chained to that one 
		*/
     sw = ONE;
     symb = getFirstFieldOfStruct(pred_bfnd);
     if (symb)
       {
	 sw = THREE;
	 first_symbol = symb;
       }
     else
       {
	 if (isAStructDeclBif(BIF_CODE(pred_bfnd)) || 
	     isAUnionDeclBif(BIF_CODE(pred_bfnd)) ||
	     isAEnumDeclBif(BIF_CODE(pred_bfnd)))
	   sw = (!first_symbol) ? TWO : THREE ;
       }

     if (sw >= 2)
       {  /* in structure so set the variant properly */
	 if (variant == FUNCTION_NAME) 
	   decl_temp = MEMBER_FUNC ;
	 else                          
	   decl_temp = FIELD_NAME ;
       }
     else
       decl_temp = variant ; /* this is not a struct field */

     /* from vhash.c, keep it for the moment */
     new_cur_symb = find_type_symbol(entry,decl_temp,area());

/* if pred_bfnd is EXTERN_C_STAT, then we need to compare
   with global scope not pred_bfnd */

     adj_scope = pred_bfnd;
     while( BIF_CODE(adj_scope) == EXTERN_C_STAT )
        adj_scope = BIF_CP(adj_scope);

     if ( (! new_cur_symb) || ( SYMB_SCOPE(new_cur_symb) != adj_scope)) /* BW */
       {  
	 /*
	   from vhash.c, keep it for the moment 
	   but to be changed soon 
	   */
         if(isAEnumDeclBif(BIF_CODE(pred_bfnd)))
	     new_cur_symb = 
                install_entry_2(entry,decl_temp,UP_ONE_LEVEL_SKIP_EXTERN);
/*
 above was not quite right
 I've added a new case: UP_ONE_LEVEL_SKIP_EXTERN
 to skip extern statements that are not really scopes
*/
         else
	     new_cur_symb = install_entry_2(entry,decl_temp,YES); 
       }
     else if((BIF_CODE(pred_bfnd) == GLOBAL) && 
             (SYMB_CODE(new_cur_symb) == FUNCTION_NAME))
       {
             new_cur_symb = install_duplicate_entry(new_cur_symb,decl_temp); 
       }
     else
       {  
	 if (err_sw ==  YES) 
	   err_symb("(00) redeclaration symbol",SYMB_IDENT(new_cur_symb));
	 else 
	   {
             if (isAStructDeclBif(BIF_CODE(pred_bfnd)))
	       { 
	         /* check to see if this is a use befor a def. */
                 if((SYMB_CODE(new_cur_symb) == FIELD_NAME) && 
		    !(SYMB_TYPE(new_cur_symb)))
		    /* if we use T_UNKNOWN then this is needed
                    (SYMB_TYPE(new_cur_symb))) && 
		    (SYMB_TYPE(new_cur_symb)->variant == T_UNKNOWN))
		    */
		    {
			/* this is a previously used but undeclared field */
                        /* In this case we cuse the new_cur_symb because */
                        /* it was not declared befor it was used. -- dbg */
			}
		else{
		 if ((SYMB_CODE(new_cur_symb) != MEMBER_FUNC)
                      && strcmp("", SYMB_IDENT(new_cur_symb)))
		   err_symb("(01) redeclaration of symbol",SYMB_IDENT(new_cur_symb));
		 /* from vhash.c, keep it for the moment */
		 new_cur_symb = install_duplicate_entry(new_cur_symb,decl_temp);
                 }
	       }
             else {
	       /* ud_chain used for type */
	       if (!SYMB_DECL(new_cur_symb))
		 {
		   SYMB_UD_CHAIN(new_cur_symb) = (PTR_BLOB) SYMB_TYPE(new_cur_symb);
		   SYMB_DECL(new_cur_symb)  = SYMB_CODE(new_cur_symb);
		 }
	       else 
		 {
		   if (isTypeCompleteChain(SYMB_UD_CHAIN(new_cur_symb)))
		     {  
		       if (SYMB_SCOPE(new_cur_symb) != global_bfnd)
			 err_symb("(02) redeclaration of symbol",SYMB_IDENT(new_cur_symb));
		     }
		 }
	       /*
               if(SYMB_TYPE(new_cur_symb)) printf("%s has a type! %d\n",
                      new_cur_symb->ident, new_cur_symb->type->variant);
	       */
	       SYMB_TYPE(new_cur_symb) = (PTR_TYPE) NULL;
	       SYMB_CODE(new_cur_symb) = decl_temp;
	     }
	   }
	 }

       if ( new_cur_symb) 
	 {
	   if ((sw >= 2) && (name = BIF_SYMB(pred_bfnd)))
	      {  /* set the name of the structure the field depends on */
		if (decl_temp == MEMBER_FUNC)
		  SYMB_MEMBER_BASENAME(new_cur_symb) = name ;
		else 
		  SYMB_FIELD_BASENAME(new_cur_symb)   = name ;
	       }
	   if (sw == 3) addSymbToFieldList(first_symbol,new_cur_symb);
            if (sw == 2) first_symbol = new_cur_symb ;
	   ll_node = (PTR_LLND) newNode(map2(decl_temp));
	   NODE_SYMB(ll_node) = new_cur_symb;
	   new_cur_type  = (PTR_TYPE) NULL ;
	   cur_flag = RULE_ID ;
	 }
   }

   if ((new_cur_symb)&&
       isAStructDeclBif(BIF_CODE(pred_bfnd)))
     {
       SYMB_ATTR(new_cur_symb) = SYMB_ATTR(new_cur_symb)|
	 ((int) BIF_PROPLIST(pred_bfnd));
     }
   switch (main_type_flag) 
     {
     case TYPE_ONE :
     case TYPE_TWO : 
     case TYPE_FOUR : 
       /* keep it to be change later */
       bf1 = installBifNode(VAR_DECL,SMNULL,LLNULL,LLNULL,LLNULL);
       appendBfndToList(pred_bfnd,bf1);        
       expr_llnd = newExpr(EXPR_LIST,NULL,ll_node,NULL);
       BIF_LL1(bf1) = expr_llnd ;
       main_type_flag = TYPE_CLEAN ;       
       break ;
     case TYPE_THREE: 
       bf1 = lastBifInBlobList(pred_bfnd);
       expr_llnd = newExpr(EXPR_LIST,NULL,ll_node,NULL);
       BIF_LL1(bf1) = expr_llnd ;
       main_type_flag = TYPE_CLEAN ;       
       break;
     case TYPE_CLEAN: 
       bf1 = lastBifInBlobList(pred_bfnd);
       if (bf1)
	 {
	    BIF_LL1(bf1) = addToExprList(BIF_LL1(bf1),ll_node);
	 }
       break;
     }
     return(ll_node);
 }


/**************************************************************************/
/* was derived_lazy_install() */
PTR_LLND derivedLazyInstall(symb)
PTR_SYMB symb;
{ 
  PTR_BFND bf1;
  PTR_LLND ll_node,expr_llnd;
  
  if (!symb)
    return NULL;
  new_cur_symb =first_symbol = symb;
  ll_node = installLowLevelNode(FUNCTION_REF,LLNULL,LLNULL,first_symbol);
  switch (main_type_flag) 
    {
     case TYPE_ONE :
     case TYPE_TWO : 
     case TYPE_FOUR : 
       /* keep it to be change later */
       bf1 = installBifNode(VAR_DECL,SMNULL,LLNULL,LLNULL,LLNULL);
       appendBfndToList(pred_bfnd,bf1);        
       expr_llnd = newExpr(EXPR_LIST,NULL,ll_node,NULL);
       BIF_LL1(bf1) = expr_llnd ;
       main_type_flag = TYPE_CLEAN ;       
       break ;
     case TYPE_THREE: 
       bf1 = lastBifInBlobList(pred_bfnd);
       expr_llnd = newExpr(EXPR_LIST,NULL,ll_node,NULL);
       BIF_LL1(bf1) = expr_llnd ;
       main_type_flag = TYPE_CLEAN ;       
       break;
     case TYPE_CLEAN: 
       bf1 = lastBifInBlobList(pred_bfnd);
       if (bf1)
	 {
	    BIF_LL1(bf1) = addToExprList(BIF_LL1(bf1),ll_node);
	 }
       break;
     }
  return(ll_node);
}

/**************************************************************************/
/* was install_parameter(entry,variant) */
PTR_LLND installParameter(new_cur_symb,entry,variant)
PTR_HASH entry;
int variant ;
PTR_SYMB *new_cur_symb;
{
   PTR_LLND ll;
   /* NO is the right one to use, but the func_hedr bif node might not
      be created yet in typename_functor case. Therefore we use OTHER
      meaning global node as a temporary one. The real scope of the
      parameter will be installed in the routine change_parameter_scope */
   if (!entry)
     return NULL;
    *new_cur_symb = install_entry_2(entry,variant,OTHER);
   ll = newExpr(VAR_REF,SYMB_TYPE( *new_cur_symb), *new_cur_symb);
   return(ll);
 }

/*was fill_type_for_param() */

/**************************************************************************/
void setParamTypeToDefault(functor)
PTR_SYMB functor;
{ 
   PTR_SYMB list1;
   if (!functor)
     return;
   for (list1 = SYMB_MEMBER_PARAM(functor) ; list1 ; 
	list1 = SYMB_NEXT_DECL(list1))
     if (!SYMB_TYPE(list1)) SYMB_TYPE(list1) = GetAtomicType(DEFAULT);
 }


/**************************************************************************/
/* WAS find_constant_type() */
PTR_TYPE getTypeForConstantNode(pllnd)
PTR_LLND pllnd;
{

 PTR_TYPE dnode;
  if (!pllnd)
    return NULL;
  switch(NODE_CODE(pllnd)) 
    {
    case INT_VAL:
      return(GetAtomicType(T_INT));
    case FLOAT_VAL:
      return(GetAtomicType(T_FLOAT));
    case CHAR_VAL:
      return(GetAtomicType(T_CHAR));
    case DOUBLE_VAL:
      return(GetAtomicType(T_DOUBLE));
    case LONG_DOUBLE_VAL:
      dnode = (PTR_TYPE) newNode(T_DESCRIPT);
      TYPE_LONG_SHORT(dnode) |= BIT_LONG;
      TYPE_BASE(dnode) = GetAtomicType(T_DOUBLE);
      return dnode;
    case LONG_INT_VAL:
      dnode = (PTR_TYPE) newNode(T_DESCRIPT);
      TYPE_LONG_SHORT(dnode) |= BIT_LONG;
      TYPE_BASE(dnode) = GetAtomicType(T_INT);
      return dnode;
    case UNSIGNED_INT_VAL:
      dnode = (PTR_TYPE) newNode(T_DESCRIPT);
      TYPE_LONG_SHORT(dnode) |= BIT_UNSIGNED;
      TYPE_BASE(dnode) = GetAtomicType(T_INT);
      return dnode;
    case UNSIGNED_LONG_INT_VAL:
      dnode = (PTR_TYPE) newNode(T_DESCRIPT);
      TYPE_LONG_SHORT(dnode) |= ( BIT_LONG | BIT_UNSIGNED);
      TYPE_BASE(dnode) = GetAtomicType(T_INT);
      return dnode;

    default:
      return(GetAtomicType(DEFAULT));
    }
}


/**************************************************************************/
/* Was confirm_overloaded_function() */
void confirmOverloadedFunction(f_llnd,exprlist)
     PTR_LLND f_llnd,exprlist;
{
  PTR_SYMB functor,symb;
  
  if (!f_llnd)
    return;

  functor = NODE_SYMB(f_llnd);
  if (functor && (NODE_CODE(functor)== MEMBER_FUNC))
    {
      symb=getOverloadingFunction(functor,exprlist);
      if (symb) 
	NODE_SYMB(f_llnd) = symb;
    }
}


/**************************************************************************/
/* was check_array_id_format() */
PTR_LLND checkArrayIdFormat(ll_ptr,state)
     int *state;
     PTR_LLND ll_ptr ;
{  
  PTR_LLND temp,temp1;

  if (!ll_ptr)
    {
      *state = NO;
      return NULL;
    }
  temp = ll_ptr;
  switch (NODE_CODE(ll_ptr)) 
    {
    case VAR_REF : 
      *state = ID_ONLY ;
      return(ll_ptr);
    case POINTST_OP:
    case RECORD_REF :
    case SCOPE_OP:
      temp = NODE_OPERAND1(ll_ptr);
      if ((NODE_CODE(temp) == VAR_REF)||
	  (NODE_CODE(temp) == FUNCTION_REF))
	{ 
	  *state = ID_ONLY ;
	  return(temp);
	}
      if  (NODE_CODE(temp) == ARRAY_REF) 
	{
	  temp1 = Follow_Llnd(NODE_OPERAND0(temp),2);
	  *state = RANGE_APPEAR;
	  return(temp1);
	}
      else 
	err("unexpected state in array format");
      return(temp);
    case ARRAY_REF :
      temp1 = Follow_Llnd(NODE_OPERAND0(temp),2);
      *state = RANGE_APPEAR;
      return(temp1);
    case  ARRAY_OP:
      temp1 = Follow_Llnd(NODE_OPERAND0(temp),2);
      *state =RANGE_APPEAR ;
      return(temp1);
      default :    
	*state = ARRAY_OP_NEED ;
      return(temp);
    }
}

/**************************************************************************/
/* was is_object_legal() */
static int isObjectLegal(object,pbf)
     PTR_SYMB object;
     PTR_BFND pbf;
{
  /* temporarily */
  return(1);
}

/**************************************************************************/
/* Was install_scope_operator()
  The routine to actually install object::field
  (1) check if the object is legally derived with the current object
  (2) if yes, then use the object to find the proper field
  (3) do some preparing work for the coming [] or () just like
      "primary . identifier" generated
  (4) generate proper low level nodes just like  
      "primary . identifier" generated
*/
PTR_LLND installScopeOperator(object_type_entry,field_entry,current_pred_bfnd)
     PTR_HASH object_type_entry,field_entry;
     PTR_BFND current_pred_bfnd;
{
  PTR_SYMB object,symbol;
  PTR_LLND ll,l_ptr1,l_ptr2;
  int tag;
  
  /* coming from vhash.c keep it for the moment */
  object = find_object_symbol(object_type_entry);                 
  if (isObjectLegal(object,current_pred_bfnd)) 
    {
      symbol=weakCheckTypeForField(object,field_entry,0,CLASS_ONLY);
      if (!symbol)
	{
	  err("error member"); 
	  return(NULL);
	}
      else 
	{
	  tag  = map2(SYMB_CODE(symbol))  ;
	  l_ptr2 = (PTR_LLND) newNode(tag);
	  NODE_SYMB(l_ptr2) = symbol;
	  l_ptr1 = newExpr(VAR_REF,NULL,object);
	  ll = newExpr(SCOPE_OP,NULL,l_ptr1,l_ptr2);
	  NODE_TYPE(ll) = adjustArrayType(symbol);
	  return(ll);
	}
    }
  else 
    err("improper derived class");
  return(NULL);
}


/**************************************************************************/
/* Was find_field_for_this(entry) */
/* new for pC++2dep, to find the class object corresponds with this      */
/* and return the type of the object class                               */
/*                                                                       */
/* There is a bug with the class which has no tag for the class name     */
PTR_SYMB getFieldForThis(pred_bfnd, entry)
     PTR_BFND pred_bfnd;
     PTR_HASH entry;
{
  PTR_BFND p;
  PTR_SYMB functor,base_name;
  PTR_SYMB ret_symb=NULL;
  
  p = getFuncScope(pred_bfnd);

  if (BIF_CODE(p) == FUNC_HEDR) 
    {
      functor=BIF_SYMB(p);
      if (functor && (SYMB_CODE(functor) == MEMBER_FUNC) )
	{
	  base_name = SYMB_MEMBER_BASENAME(functor);
	  ret_symb=weakCheckTypeForField(base_name,entry,0,area());
          if(!ret_symb){
             ret_symb=install_entry_2(entry,FIELD_NAME,UP_TO_CLASS);
            }
          return ret_symb;      
	}	 
    }
  err("not a proper member of this");
  return(NULL);
}


/**************************************************************************/
/* Was install_declmod()
 * <1> original_type is given
 * <2> Initial and duplicated error checking need more
 * <3> type_mod may be type_mod or sc_class
 * <4> return the original type_ptr
 */
PTR_TYPE installDeclmod(original_type,type_mod)
PTR_TYPE original_type ;
int type_mod ;
{
   if(!original_type) 
      original_type = (PTR_TYPE) newNode(T_DESCRIPT);

   TYPE_LONG_SHORT(original_type) |= map_status(type_mod);
   return(original_type);

 }


/**************************************************************************/
/* Was combine_2
  combine two descriptive type and clean second one */
PTR_TYPE combineDeclmod(original, second_one)
     PTR_TYPE original;
     PTR_TYPE second_one;
{
  if(!original && !second_one)
    return(original);
  if(!original) original = (PTR_TYPE) newNode(T_DESCRIPT);

  TYPE_LONG_SHORT(original) |= TYPE_LONG_SHORT(second_one);
  TYPE_LONG_SHORT(second_one) = OFF ;
  return(original);
  
}



/**************************************************************************/
/* Was find_proc_name()
 * <1> Using pred_bfnd to find the bif_node of the current funtion_header
 *     or the symbol_declarator which is recently parsed.
 * <2> called by install_parameter
 * <3> return symbol of the function_name
 */
PTR_SYMB getCurrentProcName(pred_bfnd)
     PTR_BFND pred_bfnd;
{
  PTR_BFND bf_node;
  PTR_SYMB symb ;
  PTR_LLND ll_node,llptr;
  bf_node = lastBifInBlobList(pred_bfnd);
  if (!bf_node)
      {   
	symb = NULL ;
	err("UNEXPECTED (1) error in getCurrentProcName");
      }
    else
      {  
	llptr = BIF_LL1(bf_node);
	ll_node = Follow_Llnd(llptr,1);
	if (ll_node)
	  symb = NODE_SYMB(ll_node);
	else
	  err("UNEXPECTED (2) error in getCurrentProcName");
      }
  return(symb);
}

/**************************************************************************/
/* to change y++.y from here */
/* was transform_symb_node() */
void setSymbNodeToFunction(functor)
     PTR_SYMB functor ;
{
 PTR_SYMB base_name,next_one ;
 int tag, offset;

 if (!functor)
   return;
 switch(SYMB_CODE(functor)) 
   {
   case FIELD_NAME :
     SYMB_CODE(functor) = MEMBER_FUNC ;
     base_name = SYMB_FIELD_BASENAME(functor);
     next_one = SYMB_NEXT_FIELD(functor);
     tag      = SYMB_FIELD_TAG(functor);
     offset   = SYMB_FIELD_OFFSET(functor);
     SYMB_FIELD_BASENAME(functor) = NULL ;
     SYMB_NEXT_FIELD(functor)     = NULL ;
     SYMB_FIELD_TAG(functor)      = 0;
     SYMB_FIELD_OFFSET(functor)   = 0;
     SYMB_FIELD_DECLARED_NAME(functor) = NULL ;
     SYMB_MEMBER_TAG(functor) = tag;
     SYMB_MEMBER_OFFSET(functor) = offset;
     SYMB_MEMBER_BASENAME(functor) = base_name;
     SYMB_MEMBER_NEXT(functor) = next_one ;
     SYMB_MEMBER_NB_OUTPUT(functor) = 0 ;
     SYMB_MEMBER_NB_IO(functor) = 0 ;
     SYMB_MEMBER_OUTLIST(functor)=  NULL;
     break;

   case VARIABLE_NAME :
     SYMB_CODE(functor) = FUNCTION_NAME ;
     break;
   }
}

/**************************************************************************/
/* Was func_declarator_appear() */
PTR_LLND setFuncAfterDeclaratorAppear(pred_bfnd,cur_flag,l_ptr1,l_ptr2,symb1)
     int *cur_flag;
     PTR_BFND pred_bfnd;
     PTR_LLND l_ptr1,l_ptr2;
     PTR_SYMB symb1 ;   	
{   
  PTR_TYPE type_1 ;
  PTR_LLND ll_p1 ;
  PTR_SYMB functor;
  PTR_LLND old_ptr, tmp;
  PTR_TYPE type ;
  PTR_SYMB symb2;

  if (!l_ptr1)
    return NULL;
  switch(NODE_CODE(l_ptr1))  
    { 
    case SCOPE_OP: /* this line is probably not right: dbg */
    case DEREF_OP :  
      ll_p1 = newExpr(FUNCTION_OP,NULL,l_ptr1,l_ptr2);
      /* needs check */
      type_1 = (PTR_TYPE) newNode(T_FUNCTION);
      if (symb1 && !SYMB_TYPE(symb1))
	{ 
	  SYMB_TYPE(symb1) = type_1 ;
	}
      else 
	{ 
	  if (symb1)
	    type = lookForInternalBasetype(SYMB_TYPE(symb1));
	  else
	    type = NULL;
	  if (type){
          if(NODE_CODE(type) != T_MEMBER_POINTER)
	    TYPE_BASE(type) = type_1 ;
	     }
	  else
	    Message("Internal error",0);
	}                        
      break ;
    case FUNCTION_REF :
    case RECORD_REF :
    case VAR_REF :
      old_ptr = l_ptr1;
      functor = NODE_SYMB(old_ptr);
      tmp = giveLlSymbInDeclList(l_ptr2);
      if (tmp)
	symb2 = NODE_SYMB(tmp);
      else
	symb2 = NULL;
      if (((BIF_CODE(pred_bfnd) ==GLOBAL) || 
           (BIF_CODE(pred_bfnd) == TEMPLATE_FUNDECL)) &&
	  (NODE_CODE(functor) == MEMBER_FUNC)) 
	{
	  SYMB_MEMBER_OUTLIST(functor) = SYMB_MEMBER_PARAM(functor);
	  SYMB_MEMBER_PARAM(functor) = symb2;
	  SYMB_MEMBER_LIST(functor)  = (PTR_SYMB) ON;
	}
      else 
	if (symb2)   
	  SYMB_MEMBER_PARAM(functor) = symb2;
      NODE_OPERAND0(old_ptr) = l_ptr2 ; 
      NODE_CODE(old_ptr)  = FUNCTION_REF ;
      *cur_flag = *cur_flag | RULE_PARAM ;
      ll_p1 =l_ptr1 ;
      break ;
    default:        
      err("unexpected format ");
    }
  return(ll_p1);
   
}
   

   	
/**************************************************************************/
/* Was chain_array_op_range() 
 * <1> ll_node1 : LHS low level node
 *
 * <2> ll_node2 : range_list in the RHS
 */
PTR_LLND addRangeToArray(new_cur_type,ll_node1,ll_node2)
     PTR_TYPE *new_cur_type;
     PTR_LLND ll_node1,ll_node2 ;
   	
{
  PTR_LLND llpr_1;
  PTR_LLND p,pllnd;
  PTR_SYMB symb;

  if (!ll_node1)
    return NULL;
  switch(ll_node1->variant)
    {
    case DEREF_OP:
      llpr_1= newExpr(EXPR_LIST,NULL,ll_node2,NULL);
      p = newExpr(ARRAY_OP,NULL,ll_node1,llpr_1);
      pllnd = giveLlSymbInDeclList(ll_node1);
      symb = NODE_SYMB(pllnd);
      *new_cur_type = (PTR_TYPE) newNode(T_ARRAY);
      TYPE_DECL_RANGES(*new_cur_type) =  llpr_1;
      SYMB_TYPE(symb) = addToBaseTypeList(*new_cur_type,SYMB_TYPE(symb));
      return(p);
    case ARRAY_OP:
      addToExprList(NODE_OPERAND1(ll_node1),ll_node2);
      return(ll_node1);
    case ARRAY_REF:
    case VAR_REF:
    default:
      NODE_CODE(ll_node1) = ARRAY_REF;
      if (! NODE_OPERAND0(ll_node1))
	{  /* find one */
	  llpr_1= NODE_OPERAND0(ll_node1) = newExpr(EXPR_LIST,NULL,ll_node2,NULL);
	  pllnd = giveLlSymbInDeclList(ll_node1);
	  symb = NODE_SYMB(pllnd);
	  *new_cur_type = (PTR_TYPE) newNode(T_ARRAY);
	  TYPE_DECL_RANGES(*new_cur_type) =  llpr_1 ;
	  SYMB_TYPE(symb) = addToBaseTypeList(*new_cur_type,symb->type);
	}
      else 
	{
	  addToExprList(NODE_OPERAND0(ll_node1),ll_node2);
	} 
      return(ll_node1);
    }
}
   	
   	

/**************************************************************************/
/* Was set_struct_typename()
 * <1> type_decl : CLASS_DECL, ENUM_DECL STRUCT_DECL, UNION_DECL
 * <2> symbol    : SMNULL if not appear
 */
PTR_TYPE setBfndForStruct(pred_bfnd,cur_flag,var_decl,symbol)
     PTR_BFND pred_bfnd;
     int      var_decl ;
     PTR_SYMB symbol;
     int *cur_flag;
{   
  PTR_BFND bf1 ;

  bf1 = installBifNode(var_decl,symbol,LLNULL,LLNULL,LLNULL);
  appendBfndToList(pred_bfnd,bf1);
  *cur_flag = var_decl;
  return(SYMB_TYPE(symbol));
}
   	
   	

    	
/**************************************************************************/
/* was install_struct()
 * <1> type_decl : T_CLASS, T_ENUM , T_STRUCT, T_UNION, T_CLASS
 * <2> original_class : SMNULL 
 */
PTR_TYPE installStruct(pred_bfnd, new_cur_type, type_decl, original_class)
     PTR_BFND *pred_bfnd;
     PTR_TYPE *new_cur_type;
     int      type_decl ;
     PTR_LLND original_class ;
{  
  PTR_SYMB symb_ptr;
  PTR_SYMB count_first();
    	
  symb_ptr = BIF_SYMB(*pred_bfnd);
  if (symb_ptr) 
    *new_cur_type = SYMB_TYPE(symb_ptr);
  else          
    *new_cur_type = (PTR_TYPE) newNode(type_decl);
  TYPE_COLL_ORI_CLASS(*new_cur_type) = *pred_bfnd;
  /* probably wrong, to check later count_fields_num() */
  /*  TYPE_COLL_NUM_FIELDS(*new_cur_type) = count_fields_num( *pred_bfnd);*/
  TYPE_COLL_NUM_FIELDS(*new_cur_type) = 0;
  TYPE_COLL_RECORD_SIZE(*new_cur_type) = 0;
  TYPE_COLL_FIRST_FIELD(*new_cur_type) = getFirstFieldOfStruct(*pred_bfnd);
  *pred_bfnd = BIF_CP(*pred_bfnd);
  return(*new_cur_type);
}
    	


    	

/* Was enter_inner_level()
 * <1> type_decl : CLASS_DECL, ENUM_DECL STRUCT_DECL, UNION_DECL
 * <2> symbol    : SMNULL if not appear
 * <3> typename_list: LLNULL if it is not a derived_class
 */
void innerLevelStruct(pred_bfnd,cur_flag, var_decl,symbol,typename_list)
     PTR_BFND *pred_bfnd;
     int      var_decl ;
     PTR_SYMB symbol ;
     PTR_LLND typename_list;
     int *cur_flag;
{   
  PTR_BFND bf1 ;

  bf1 = installBifNode(var_decl,symbol,NULL,NULL,NULL);
  appendBfndToList(*pred_bfnd,bf1);
  *pred_bfnd = bf1 ;
  *cur_flag = var_decl;
  first_symbol = NULL;
  BIF_LL2(bf1) = typename_list;
  if (symbol) 
    {
      SYMB_TYPE(symbol) = (PTR_TYPE) newNode(map_bif_to_type(var_decl));
      TYPE_COLL_ORI_CLASS(SYMB_TYPE(symbol)) = bf1;
    }
  main_type_flag = TYPE_FOUR;
  /* prop_list is used for public, methodofelement,private... */
  switch (BIF_CODE(bf1)) 
    {
    case COLLECTION_DECL:
      BIF_PROPLIST(bf1) =(PTR_PLNK) (PUBLIC_FIELD|COLLECTION_FIELD);
      break;
    case CLASS_DECL:
      BIF_PROPLIST(bf1) =(PTR_PLNK) PRIVATE_FIELD;
      break;
    case STRUCT_DECL:
    default:
      BIF_PROPLIST(bf1) =(PTR_PLNK) PUBLIC_FIELD;
    }
}


/**************************************************************************/
/* NEW (phb and dbg) */
void makeExternC(pred_bfnd,cur_flag, var_decl)
     PTR_BFND *pred_bfnd;
     int *cur_flag;
     int var_decl ;

  {   
  PTR_BFND bf1 ;

  bf1 = installBifNode(var_decl,SMNULL,NULL,NULL,NULL);
  appendBfndToList(*pred_bfnd,bf1);
  *pred_bfnd = bf1 ;
  *cur_flag = var_decl;
  first_symbol = NULL;
  BIF_LL2(bf1) = NULL;

  main_type_flag = TYPE_FOUR;
  /* prop_list is used for public, methodofelement,private... */
  BIF_PROPLIST(bf1) = (PTR_PLNK) NULL;
  }

void dontmakeExtrn(){
   int i;
   PTR_FNAME fetch_line_queue();
   fetch_line_queue(&i);
}

/**************************************************************************/
/********************************************************************/
/*  make name from type will take a pointer to a type and construct */
/*  a char. string that represents the type                         */
/********************************************************************/
char * makeStringForType(type)
PTR_TYPE type;
{
    char *name, *buffer;
    int index;
    PTR_TYPE base;
    if(!type){
        return NULL;
        }
    name = (char *) malloc(256);
    buffer = (char *) malloc(256);
    strcpy(name,"");
    base = lookForInternalBasetype(type);
    switch(TYPE_CODE(base)){
       case T_DERIVED_TYPE:
	   if(TYPE_SYMB_DERIVE(base)) 
                   strcpy(name, SYMB_IDENT(TYPE_SYMB_DERIVE(base)));
	   break;
       case T_INT:
           strcpy(name, "int");
           break;
       case  T_CHAR:
           strcpy(name, "char");
           break;
       case T_FLOAT:
           strcpy(name, "float");
	   break;
       case T_DOUBLE:
           strcpy(name, "double");
           break;
       case T_VOID:
           strcpy(name, "void");
           break; 
        default: break;
        }	
    while(type){
	switch(TYPE_CODE(type)){
	case T_POINTER :
	  name = strcat(name, "*"); 
	  type = TYPE_BASE(type);
	  break;
	case T_REFERENCE :
	  name = strcat(name, "&"); 
	  type = TYPE_BASE(type);
	  break;
	case T_DESCRIPT :
	  index = TYPE_LONG_SHORT(type);
	  if(index & BIT_UNSIGNED){
		  strcpy(buffer,"unsigned ");
		  name = strcat(buffer, name);
		  } 
	  if(index & BIT_SIGNED){
		  strcpy(buffer,"signed ");
		  strcpy(name,strcat(buffer, name));
		  } 
	  if(index & BIT_LONG){
                  strcpy(buffer,"long ");
		  strcpy(name,strcat(buffer, name)); 
		  }
	  if(index & BIT_SHORT){
		  strcpy(buffer,"short ");
		  strcpy(name,strcat(buffer, name));
		  } 
	  if(index & BIT_VOLATILE){
		  strcpy(buffer,"volatile ");
		  strcpy(name,strcat(buffer, name)); 
		  }
	  if(index & BIT_CONST){
		 strcpy(buffer,"const ");
		 strcpy(name,strcat(buffer, name)); 
                 }
	  if(index & BIT_SYNC){
		 strcpy(buffer,"sync ");
		 strcpy(name,strcat(buffer, name)); 
                 }
	  if(index & BIT_GLOBL){
		 strcpy(buffer,"global ");
		 strcpy(name,strcat(buffer, name)); 
                 }
	  type = TYPE_DESCRIP_BASE_TYPE(type);
	  break;
       default:
          type = NULL;
          break;
        }
     }
     /*
     strcpy(buffer, "_ovop_");
     strcpy(name,strcat(buffer, name));
     */
     return name;
}

/**************************************************************************/
char* makeNameString(p, s, t)
char *s;
PTR_TYPE p, t;
{
    char *prefix, *postfix;
    if(p){
         prefix = makeStringForType(p);
         }
    else {
       prefix = (char *) malloc(256);
       prefix[0] = 0;
       }
    strcat(prefix, s);
    if(t){
        postfix = makeStringForType(t);
        prefix = strcat(prefix, postfix);
        }
    /* printf("makeNameString:%s:\n", prefix); */
    return prefix;
}

/**************************************************************************/
/* Was component_constructor_action() */
/*****************************************************************/
/*                                                               */
/*  Grammar :typed_declspecs "(" parmlist ')'                    */
/*                                                               */
/* flag ==0: for CONSTRUCTOR                                     */
/* flag ==1: for DESTRUCTOR                                      */
/* returns the llnode for the expression list for the declaration */
/*****************************************************************/
PTR_LLND componentConstructorAction(pred_bfnd,cur_flag,type,parmlist,flag,body_appear, entry)
     PTR_BFND *pred_bfnd;
     int *cur_flag,flag,body_appear;
     PTR_TYPE type;
     PTR_LLND parmlist;
     PTR_HASH entry;
{
   PTR_LLND ll;
   PTR_SYMB symb;
   PTR_SYMB dummy_symb;
   PTR_TYPE new_type;
   PTR_BFND bif;

   if (!type)
     {
       Message("Internal Error",0);
       return NULL;
     }

  if(flag != OVOPERATOR) 
    if ((TYPE_CODE(type) != T_DESCRIPT)&&(TYPE_CODE(type) !=T_DERIVED_TYPE)
        && (TYPE_CODE(type) != T_DERIVED_TEMPLATE))
      err("illegal syntax for (de)(con)structor or overloaded op");

   if (entry == NULL) entry = find_entry_from_type(type);
   
   if (flag==DESTRUCTOR || flag==VIRTUAL_DESTRUCTOR)
     entry = look_up_expanded_string(entry);
     

   if (flag != OVOPERATOR && TYPE_CODE(type) ==  T_DESCRIPT)
     {
       TYPE_DESCRIP_BASE_TYPE(type) = GetAtomicType(DEFAULT);
       new_type = type;       
     } 
  else if (flag == OVOPERATOR) new_type = type;
  else if (flag == VIRTUAL_DESTRUCTOR) {
        new_type = (PTR_TYPE) newNode(T_DESCRIPT);
       TYPE_DESCRIP_BASE_TYPE(new_type) = GetAtomicType(DEFAULT);
       (void)installDeclmod(new_type,RID_VIRTUAL);
       }
     else
       new_type = GetAtomicType(DEFAULT);
   main_type_flag = TYPE_TWO;
   ll = lazyInstall(*pred_bfnd,entry,VARIABLE_NAME,NO);
   symb = getCurrentProcName(*pred_bfnd);
   if(flag == OVOPERATOR) SYMB_ATTR(symb) |= OVOPERATOR;
   if ((NODE_CODE(ll)== VAR_REF)||(NODE_CODE(ll)==RECORD_REF) ||
       (NODE_CODE(ll)== FUNCTION_REF))                 
     setSymbNodeToFunction(symb);
   
   ll = setFuncAfterDeclaratorAppear(*pred_bfnd,cur_flag,ll,parmlist,symb) ;

   addToDeclaration(new_type,OFF,ll,&dummy_symb,body_appear);
   if (body_appear)
     {
       setParamTypeToDefault(symb);
       bif = lastBifInBlobList(*pred_bfnd);
       if (bif)
	 BIF_SYMB(bif) = symb ;
     }
   return ll;
 }


/**************************************************************************/
/* was late_adjust_functor() */
void lateAdjustFunctor(pbf)
PTR_BFND pbf;
{

  PTR_SYMB symbptr,symb3;
  PTR_LLND lptr2;

  if (!pbf)
    return;

  if (BIF_CODE(pbf) != FUNC_HEDR) 
    return;
  lptr2=  giveLlSymbInDeclList(BIF_LL1(pbf));
  symbptr= BIF_SYMB(pbf);
  /* printf("syb id is %d\n", symbptr->id); */
  if ( (symb3=lookForOverloadingFunction(symbptr)) )
    {
     BIF_SYMB(pbf) = NODE_SYMB(lptr2) = symb3; 
    }
  SYMB_MEMBER_HEADER(BIF_SYMB(pbf)) = pbf;
  SYMB_UD_CHAIN(symbptr) = (PTR_BLOB) NULL;
  SYMB_DECL(symbptr)   = 0;
}

/**************************************************************************/
void lateAdjustFunctorWithInit(pbf, initl)
PTR_BFND pbf;
PTR_LLND initl;
{

  PTR_SYMB symbptr,symb3;
  PTR_LLND lptr2;
  PTR_LLND lp, lp3;
 
  if (!pbf)
    return;
  if (BIF_CODE(pbf) != FUNC_HEDR) 
    return;
  lp = BIF_LL1(pbf);
  lptr2=  giveLlSymbInDeclList(lp);
  symbptr= BIF_SYMB(pbf);
  if ( (symb3=lookForOverloadingFunction(symbptr)) )
    {
       BIF_SYMB(pbf) = NODE_SYMB(lptr2) = symb3; 
    }
  SYMB_MEMBER_HEADER(BIF_SYMB(pbf)) = pbf;
  SYMB_UD_CHAIN(symbptr) = (PTR_BLOB) NULL;
  SYMB_DECL(symbptr)   = 0;
  lp = BIF_LL1(pbf);
  lp3 = NODE_OPERAND0(lp);
  NODE_OPERAND0(lp) = installLowLevelNode(BIT_NUMBER,lp3,initl,SMNULL);
}

/**************************************************************************/
/* 
  was enter_func_body()
  arg1: type
  arg2: llnd
*/
void enterFuncBody(pred_bfnd,arg1,arg2)
     PTR_BFND pred_bfnd;
     PTR_TYPE arg1;
     PTR_LLND arg2;
{
  PTR_BFND bf1;
  PTR_LLND expr_llnd;
  PTR_SYMB dummy_symb;

  if (!pred_bfnd)
    {
      Message("Internal Error",0);
      return;
    }
  bf1 = pred_bfnd;
  expr_llnd = BIF_LL1(bf1);
  NODE_OPERAND0(expr_llnd) = arg2 ;
  addToDeclaration(arg1,OFF,expr_llnd,&dummy_symb,YES);
  if (dummy_symb) 
    BIF_SYMB(bf1) = dummy_symb;
  SYMB_DOVAR(BIF_SYMB(bf1)) += XDECL ;
  /* To be done later in late_adjust_functor()   */
  main_type_flag = TYPE_FOUR;
}

/**************************************************************************/
/* was change_param_scope() */
static void setParamScope(functor, scope)
     PTR_SYMB functor;
     PTR_BFND scope;
{ 
   PTR_SYMB list1;
   if (!functor)
     return;
   for (list1 = SYMB_MEMBER_PARAM(functor) ; list1 ; 
	list1 = SYMB_NEXT_DECL(list1))
      SYMB_SCOPE(list1) = scope;
 }


/**************************************************************************/
/* was set_function_hedr() */
void installFunctionHedr(pred_bfnd)
     PTR_BFND *pred_bfnd;
{
  PTR_BLOB blob ;
  PTR_LLND expr_llnd ;
  PTR_SYMB symb1 ;
  PTR_BFND bf1 ;

  bf1 = lastBifInBlobList(*pred_bfnd);
  blob = lastBlobInBlobList(*pred_bfnd);
  if (!bf1)
    return;
  
  if (BIF_CODE(bf1) == VAR_DECL)
    {
      BIF_CODE(bf1) = FUNC_HEDR ;
      symb1 = BIF_SYMB(bf1) = getCurrentProcName(*pred_bfnd);
      setParamScope(symb1,bf1);	
      *pred_bfnd = bf1 ;
    }
  else
    {
      symb1 = getCurrentProcName(*pred_bfnd);
      expr_llnd = BIF_LL1(bf1);
      bf1 = installBifNode(FUNC_HEDR,symb1,expr_llnd,NULL,NULL);
      BLOB_VALUE(blob) =bf1 ;
      setParamScope(symb1,bf1);
      BIF_CP(bf1) = *pred_bfnd ;
      BIF_SYMB(bf1) = symb1;
      *pred_bfnd = bf1 ;
    }

  /*if this is a member function of a template class, it may */
  /*have its own parameter list.  we will store this in llnd3 */
  /* of the function header. */
  if((SYMB_CODE(symb1) == MEMBER_FUNC) && member_templ_parms){
	BIF_LL3(bf1) = member_templ_parms;
	member_templ_parms = NULL;
	}

}



/**************************************************************************/
/* Was chain_parameter_list()
   The followings are the routine to chain parameters together
   (1) chain parameter symbol list
   (2) chain exprlist of paramter
*/
PTR_LLND chainParameterList(p1,p2)
PTR_LLND p1,p2;
{
  PTR_SYMB old_symb,name_list, first_arg,new_arg;
  PTR_LLND tmp;
  char *s;
  
  tmp = giveLlSymbInDeclList(p1);
  if (tmp)
    first_arg = NODE_SYMB(tmp);
  else
    first_arg = NULL;
  tmp = giveLlSymbInDeclList(p2);
   if (tmp)
     new_arg   = NODE_SYMB(tmp);
  else
    new_arg   = NULL;
  if (first_arg&&new_arg)  
    {
      for (old_symb=name_list=first_arg;name_list;
	   name_list= SYMB_NEXT_DECL(name_list))
	{
	  old_symb = name_list;
	  s = SYMB_IDENT(new_arg);
	  if ((!strcmp(SYMB_IDENT(name_list),SYMB_IDENT(new_arg)))&&
	      ( (*s !='$') && (*s != 0)) )
	    err_symb("parameter redeclaration (chainParameterList)"
                            ,SYMB_IDENT(new_arg));
	} 	     
      if (old_symb) SYMB_NEXT_DECL(old_symb) = new_arg;
    }
  addToExprList(p1,p2);
  return(p1);  
}



/**************************************************************************/
/* Was primary_identifier_action()
   Primary: IDENTIFIER:                                          */
PTR_LLND primaryIdentifierAction( cur_id_entry,exception_flag, entry)
     int *exception_flag;
     PTR_HASH *cur_id_entry;
     PTR_HASH entry;
{
  PTR_SYMB symbptr;
  PTR_LLND return_llnd;
  /* printf("entry is %s\n", entry->ident); */
  symbptr = find_type_symbol(entry,VARIABLE_NAME,area());
  /*
  if(!symbptr) printf("no symbol yet\n");
  else {
     printf("symb %d found.", symbptr->id);
     if(!(symbptr->type)) printf("no type yet.\n");
     else printf(" with type %d\n", symbptr->type->variant);
     }
   */
  if (! symbptr)
    symbptr=fts_by_inherited_class(entry,VARIABLE_NAME,area());
  /* if(!symbptr) printf("still null\n"); */
  if( !symbptr){
	/* it is possible that this is a use befor a declaration */
        /* inside a class.  install_entry_2 will add a field name */
        /* entry in the inclosing class definition and return the */
        /* value in symbptr. it returns null if there is no symbptr */
        /* note: this will not always be the correct thing to do! */
        /* in some cases, such as nested classes this will add the */
        /* entry at the wrong level, but it will take a second pass */
        /* type analysis to correct this problem. - dbg */ 
         /* printf("symb: %s\n", entry->ident); */
         symbptr = install_entry_2(entry,FIELD_NAME,UP_TO_CLASS);
	 /* one way to do this would be T_UNKNOWN but it does not work.
	 if(symbptr){
	    SYMB_TYPE(symbptr) = GetAtomicType(T_UNKNOWN);
	    if(SYMB_TYPE(symbptr) == NULL) 
		printf("internal error!!!\n");
	    }
	 */
         /* printf("variant = %d\n", symbptr->variant); */
	}
  if (! symbptr)
    {  /* default function expected */
      *exception_flag = ON ;
      *cur_id_entry = entry ;
      return_llnd= NULL;
    }
  else 
    {
      switch (SYMB_CODE(symbptr)) 
	{
	case MEMBER_FUNC :
	case FUNCTION_NAME:
	  return_llnd = installLowLevelNode(FUNCTION_REF,NULL,NULL,symbptr);
	  break;
	case FIELD_NAME :
	case VARIABLE_NAME:
	  return_llnd = installLowLevelNode(VAR_REF,NULL,NULL,symbptr);
	  break;
	}
      return_llnd->type = adjustArrayType(symbptr);
    }
  return(return_llnd);
}



/**************************************************************************/
/* was check_field_decl() (symbrequired==0) and 
    find_field_symb_by_type() (symbrequired==1)
 * Used Place :
 *      Primary :- primary  . identifier
 *
 * <1> given two low level node
 *     ptr1 : stands for the current LHS format for record structure
 *     entry : stands for the new field hash_entry
 *
 * <2> sw : FUNCTION_ONLY represent Needs function
 *      
 * <3> side effect : right_symbol = symbol_ptr for h_ptr2;
 */
PTR_SYMB checkFieldDecl(pred_bfnd, right_symbol, ptr1,entry,sw,symbrequired)
     PTR_BFND pred_bfnd;
     PTR_SYMB * right_symbol;
     PTR_LLND ptr1 ;
     PTR_HASH entry ;
     int sw ;
     int symbrequired;
{  
  PTR_BFND pbf;
  PTR_SYMB first,object;
  PTR_TYPE typeptr_1,otype;
  int      flag;
  int      dummy_level,empty_var;
  PTR_TYPE element_type;
  PTR_SYMB element,object_base;   

  element_type =NULL;
  element = NULL;
  first = NULL;
  if (ptr1) 
    {
      typeptr_1 =  lookForInternalBasetype(NODE_TYPE(ptr1));
      /* printf(" type is %d\n",typeptr_1->variant); */
      dummy_level = getSimpleIndirectionLevel(NODE_TYPE(ptr1));
      if (typeptr_1 && (TYPE_CODE(typeptr_1) == T_DERIVED_TYPE) &&
	  TYPE_SYMB_DERIVE(typeptr_1) && !SYMB_TYPE(TYPE_SYMB_DERIVE(typeptr_1)))
	empty_var = YES;
      else
	{
	  empty_var = NO;
	  /* we want for instance the T_STRUCT attach to it */
	  dummy_level = getIndirectionLevel(NODE_TYPE(ptr1));
	  typeptr_1 = FollowTypeBaseAndDerived(NODE_TYPE(ptr1));
	}
    }
  else 
    {
      err(" not a proper member or union/struct requierd");
      return 0;
    }
  /*
  printf("empty var = %d dummy_level = %d type_ptr_1 = %d\n", empty_var, dummy_level,
            typeptr_1->variant);
  */
  if (!symbrequired)
    {
      if (dummy_level != 0) 
	{
	  err(" not a proper member or union/struct requierd");
	  return 0;
	}
    }
  else
    {
      if (dummy_level != 1) 
	{
	  err(" not a proper member or union/struct requierd");
	  return 0;
	}  
    }
  if (empty_var == YES) 
    {
      typeptr_1 = FollowTypeBaseAndDerived(NODE_TYPE(ptr1));
      if (isElementType(typeptr_1))
	{
	  object_base= getEnglobingSuperObject(pred_bfnd);
	  if (object_base)
	    {
	      if (!symbrequired)
		{
		  first = weakCheckTypeForField(object_base,entry,0,ELEMENT_ONLY);
		  if (!first) 
		    err("struct/union type definition have not appeared");
		} else
		  {
		    if (SYMB_CODE(object_base) ==COLLECTION_NAME)
		      first = weakCheckTypeForField(object_base,entry,0,ELEMENT_ONLY);
		    else
		      first = weakCheckTypeForField(object_base,entry,0,ALL_FIELDS);
		  }
	    }
	  if (!first) err("struct/union type definition have not appeared");
	}
      else  
	err("struct/union type definition have not appeared");
    }

  otype = typeptr_1;
  if (typeptr_1 && (TYPE_CODE(typeptr_1) == T_DERIVED_COLLECTION))
    {
      element_type= TYPE_COLL_BASE(typeptr_1);
      if (TYPE_COLL_NAME(typeptr_1))
	typeptr_1= SYMB_TYPE(TYPE_COLL_NAME(typeptr_1));
      flag = TRUE;
    }
  else
    flag = FALSE;

  if (otype && 
   (isStructType(TYPE_CODE(otype)) ||
      isEnumType(TYPE_CODE(otype)) ||
      isUnionType(TYPE_CODE(otype)) || flag))
    {
      pbf =TYPE_COLL_ORI_CLASS(typeptr_1);
      if (pbf )
	{
	  object = BIF_SYMB(pbf);
	  first = weakCheckTypeForField(object,entry,0,ALL_FIELDS);
	} else
	  Message("internal error",0);
    }
     
  if (!first) 
    {
     if(NODE_TYPE(ptr1) && (TYPE_CODE(NODE_TYPE(ptr1)) == T_DERIVED_TEMPLATE)){
            PTR_LLND args, arg1;
            PTR_TYPE argt;
	    PTR_SYMB argclass;
            args = TYPE_TEMPL_ARGS(NODE_TYPE(ptr1));
            /* printf("is a derived template\n"); */
            while(args){
		if(NODE_CODE(args) == EXPR_LIST){
			arg1 = NODE_OPERAND0(args);
			args = NODE_OPERAND1(args);
			/*
			printf("expr1:%d\n", NODE_CODE(arg1));
			if(NODE_TYPE(arg1)){
			         printf("type=%d\n",TYPE_CODE(NODE_TYPE(arg1)));
				}
			*/
			if((NODE_CODE(arg1) == TYPE_REF) && (NODE_TYPE(arg1)) &&
                          (TYPE_CODE(NODE_TYPE(arg1)) == T_DERIVED_TYPE)){
				argt = NODE_TYPE(arg1);
                                argclass = argt->entry.derived_type.symbol;
				/* printf("symb:%s \n", argclass->ident); */
	  			first = weakCheckTypeForField(argclass,entry,
							0,ALL_FIELDS);
				if(first) args = NULL;
				}
			}
		else args = NULL;
		}
     }
     else{ 
      if (element_type) 
	element =  getObjectSymbol(element_type);
      if (element)  
	first= weakCheckTypeForField(element,entry,0,CLASS_ONLY);
     }
    }	
  if (!symbrequired) 
    *right_symbol = first;
  if (!first)
    { 
      if (!symbrequired)
	err("unknown member function or field"); 
      else{ /* this is to allow a pointer to a member field in a
               class definition that has not yet been defined - dbg
            */
        first = install_entry_2(entry,MEMBER_FUNC,UP_TO_CLASS);
	/* err("not a proper member"); */
        return first;
        }
      return(NO); 
    }
  else 
    {  
      if (symbrequired) 
	return first;

      if (SYMB_CODE(first) == MEMBER_FUNC)
	if (sw == FUNCTION_ONLY) 
	  return(first);
	else                    
	  return(first);
      else
	if (sw == FUNCTION_ONLY) 
	  return(NO);
	else                     
	  return(first);
    }
}
/**************************************************************************
 resolveFinal is used to look at the second type pointer and see if
 it is a type modifier that should go befor the first.  For example,
 int const * p;
 should be 
 const int *p;
*/
PTR_TYPE resolveFinal(first, second)
    PTR_TYPE first, second;
{
   PTR_TYPE ndes;
   if(!second) return first;
   /* if((second == second_part) || (NODE_CODE(second) == T_DESCRIPT) ){ */
   if(TYPE_LONG_SHORT(second)){
		if(NODE_CODE(first) == T_DESCRIPT){
		  TYPE_LONG_SHORT(first) |= TYPE_LONG_SHORT(second);
		  TYPE_LONG_SHORT(second) = OFF;
		  return first;
		  }
		else{
		  ndes = (PTR_TYPE) newNode(T_DESCRIPT);
		  TYPE_LONG_SHORT(ndes) = TYPE_LONG_SHORT(second);
		  TYPE_BASE(ndes) = first;
		  TYPE_LONG_SHORT(second) = OFF;
		  return ndes;
		  }
	}
    return first;
}
		   	
			
			
			

/**************************************************************************/
PTR_TYPE typed_declspecs_without_declmods(main_type1,type2,flag, Scope_Name)
     PTR_TYPE main_type1,type2;
     int flag;
     PTR_HASH Scope_Name;
{ PTR_TYPE type_ptr1,type_ptr2 ;
  PTR_BFND bf2;
  PTR_SYMB gsym;
  PTR_TYPE return_type;
  PTR_SYMB scsymb;
  PTR_TYPE scnt;
   /* BW 
  printf("in typed_declspecs_without_declmods\n"); */

  switch (main_type_flag) {
  case TYPE_ONE :
    /* printf("case TYPE_ONE\n");     BW */
    type_ptr1=main_type1 ;
    return_type = type_ptr1 ;
    break ;
  case TYPE_TWO :
    /* printf("case TYPE_TWO\n");    BW */
    type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
    /*
    if (Scope_Name)printf("scope_name = %s\n", Scope_Name->ident);
     printf("main_type1 = %s\n", ((PTR_SYMB)main_type1)->ident);
    else printf("no identifier\n");
    */
    TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) main_type1;
    type_ptr1 = type_ptr2;
    return_type = type_ptr1 ;
    break ;
   case TYPE_FIVE :
    /* printf("case TYPE_FIVE\n");    BW */
    type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
    TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) main_type1;
    if(!main_type1 || !type_ptr2){
          warning("possible type error",0);
	  }
    else {
  /* TYPE_SCOPE_SYMB_DERIVE(type_ptr2) */
    scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) main_type1), Scope_Name);
    scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
    TYPE_SYMB_DERIVE(scnt) = scsymb;
    TYPE_SCOPE_SYMB_DERIVE(type_ptr2) = scnt;
    }
    main_type_flag = TYPE_TWO;
    type_ptr1 = type_ptr2;
    return_type = type_ptr1 ;
    break ;
 case TYPE_THREE :
    /* printf("case TYPE_THREE\n");  BW */
    bf2 =lastBifInBlobList(pred_bfnd);   
    if ((gsym= BIF_SYMB(bf2)))
      { type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
        TYPE_SYMB_DERIVE(type_ptr2) = gsym ;
        type_ptr1 = type_ptr2;
        return_type = type_ptr1;
      }
    else {
      return_type= main_type1 ;
    }
    passed_bfnd=new_cur_bfnd = bf2 ;
    if (flag) {
      if (bf2 && (isAStructDeclBif(BIF_CODE(bf2)) ||
		  isAUnionDeclBif(BIF_CODE(bf2)) ||
		  isAEnumDeclBif(BIF_CODE(bf2))))
	{   
	  eliminate_lastlink();
	}
    }
    break;
    default :      err("System Crash ");
   }
  return(return_type);
}

/**************************************************************************/
PTR_TYPE typed_declspecs_with_declmods(main_type1,type2,flag,second, Scope_Name)
     PTR_TYPE main_type1,type2,second;
     int flag;
     PTR_HASH Scope_Name;
{ PTR_TYPE type_ptr1,type_ptr2 ;
  PTR_SYMB gsym;
  PTR_BFND bf2;
  PTR_TYPE return_type;
  PTR_SYMB scsymb;
  PTR_TYPE scnt;

  switch (main_type_flag) {
  case TYPE_ONE :
    if(type2 || (second && TYPE_LONG_SHORT(second))){
    	type_ptr1 = combineDeclmod(type2, second);
    	TYPE_DESCRIP_BASE_TYPE(type_ptr1) = main_type1;
    	return_type = type_ptr1 ;
	}
    else return_type = main_type1;
    break ;
  case TYPE_TWO : 
    type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
    TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) main_type1 ;
    if(type2 || (second && TYPE_LONG_SHORT(second))){
    	type_ptr1 = combineDeclmod(type2, second);
    	TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2 ;
    	return_type = type_ptr1 ;
	}
    else return_type = type_ptr2;
    break ;
  case TYPE_FIVE : 
    type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
    TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) main_type1 ;
/*    TYPE_SCOPE_SYMB_DERIVE(type_ptr2) */
     scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) main_type1), Scope_Name);
     scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
     TYPE_SYMB_DERIVE(scnt) = scsymb;
     TYPE_SCOPE_SYMB_DERIVE(type_ptr2) = scnt;

     main_type_flag = TYPE_TWO;
    if(type2 || (second && TYPE_LONG_SHORT(second))){
    	type_ptr1 = combineDeclmod(type2, second);
    	TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2 ;
    	return_type = type_ptr1 ;
	}
    else return_type = type_ptr2;
    break ;
  case TYPE_THREE :
    type_ptr1 = type2;
    bf2 =lastBifInBlobList(pred_bfnd);
    if ( (gsym = BIF_SYMB(bf2)) )
      { type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
        TYPE_SYMB_DERIVE(type_ptr2) = gsym ;
	if(type_ptr1){
        	TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
        	return_type = type_ptr1 ;
		}
	else return_type = type_ptr2;
        if(isAStructDeclBif(BIF_CODE(bf2)) && type_ptr1){
		if(main_type1 && TYPE_CODE(main_type1) == T_DESCRIPT){
			/* we don't need two T_DESC enties. */
                        TYPE_LONG_SHORT(main_type1) |= TYPE_LONG_SHORT(type_ptr1);
                        TYPE_LONG_SHORT(type_ptr1) = OFF;
			SYMB_TYPE(gsym) = main_type1;
			return_type = main_type1;
			type_ptr1 = main_type1;
			}
		else if (main_type1){
                	TYPE_DESCRIP_BASE_TYPE(type_ptr1) = main_type1;
			SYMB_TYPE(gsym) = type_ptr1;
			}
		}
      }
    else {
      type_ptr2 = main_type1;
      /* new_cur_type must be set inside structsp */
      if(type_ptr1){
      		TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
      		return_type=type_ptr1;
		}
      else return_type =main_type1;
    }
    passed_bfnd=new_cur_bfnd = bf2 ;
    if (flag) {
      if (bf2 && (isAStructDeclBif(BIF_CODE(bf2)) ||
		  isAUnionDeclBif(BIF_CODE(bf2)) ||
		  isAEnumDeclBif(BIF_CODE(bf2))))
	{   
	  eliminate_lastlink();
	}
    }
    break;
    default :      err("System Crash ");
   }
  return(return_type);
}

/**************************************************************************/
/* was typed_declspecs_without_declmods() and 
   typed_declspecs_with_declmods() */
/*****************************************************************/
/*                                                               */
/*  Grammar :typed_declspecs                                     */
/*                                                               */
/*  flag == 0 : normal case                                      */
/*          1 : need to eliminate unnecessary links              */
/*****************************************************************/

/* not used? BW Dec 20 1994 */
/*
static PTR_TYPE typedDeclspecsWithoutDeclmods(pred_bfnd, passed_bfnd,
   new_cur_bfnd, main_type1,type2,flag,second)
     PTR_BFND pred_bfnd, *passed_bfnd, *new_cur_bfnd;
     PTR_TYPE main_type1,type2;
     PTR_TYPE *second;
     int flag;
{ 
  PTR_TYPE type_ptr1,type_ptr2 ;
  PTR_BFND bf2;
  PTR_SYMB gsym;
  PTR_TYPE return_type;

  switch (main_type_flag) 
    {
    case TYPE_ONE :
      if (second)
	{
	  type_ptr1 = combineDeclmod(type2, second);
	  TYPE_DESCRIP_BASE_TYPE(type_ptr1) = main_type1;
	}
      else
	type_ptr1=main_type1 ;
      return_type = type_ptr1 ;
      break ;
    case TYPE_TWO :
      type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
      TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) main_type1;
      type_ptr1 = type_ptr2;

      if (second)
	{
	  type_ptr1 = combineDeclmod(type2, second);
	  TYPE_DESCRIP_BASE_TYPE(type_ptr1) = main_type1;
	}
      return_type = type_ptr1 ;
    break ;
  case TYPE_THREE :
    bf2 =lastBifInBlobList(pred_bfnd);
    if (bf2 && (gsym= BIF_SYMB(bf2)))
      { 
	if (second)
	  {
	    type_ptr1 = type2;
	    type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
	    TYPE_SYMB_DERIVE(type_ptr2) = gsym ;
	    TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
	    return_type = type_ptr1 ;
	  }
	else
	  {
	    type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
	    TYPE_SYMB_DERIVE(type_ptr2) = gsym ;
	    type_ptr1 = type_ptr2;
	    return_type = type_ptr1;
	  }
      }
    else 
      {
	if (second)
	  {
	    type_ptr2 = main_type1;
	    TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
	  }
	return_type= main_type1 ;
      }
      *passed_bfnd= *new_cur_bfnd = bf2 ;                
      if (flag) 
	{
	  if (bf2 && (isAStructDeclBif(BIF_CODE(bf2)) ||
		      isAUnionDeclBif(BIF_CODE(bf2)) ||
		      isAEnumDeclBif(BIF_CODE(bf2))))
	    {   
	      eliminate_lastlink();
	    }
	}
      break;
      default :      
	err("System Crash ");
    }
  return(return_type);
}
*/

/**************************************************************************/
/* was check_extern_first_part*/
/* This procedure returns true or false.  False is an error condition */
int checkExternFirstPart(symb)
PTR_SYMB symb;
{ 
  PTR_TYPE type2;
  if (!symb) {
    err("NULL symbol in checkExternFirstPart"); /* Beckman */
    return 0; }



  /* was is_same_type so make some change 
   look at initial function for changes */
  type2 = (PTR_TYPE) SYMB_UD_CHAIN(symb);
  if (!SYMB_TYPE(symb) && type2) 
    {
      switch (TYPE_CODE(type2)) 
	{
	case T_POINTER  :
	case T_REFERENCE:
	case T_FUNCTION :    
	case T_ARRAY    :
	 err_symb("(03) redeclaration symbol",SYMB_IDENT(symb)); 
	 return 0;
	default:
	  return(1);
	}
    }
  if (!isStrictlySameType(SYMB_TYPE(symb),type2)) 
    { /* function to check */
      if (isPointerType(TYPE_CODE(type2)) && isPointerType(TYPE_CODE(SYMB_TYPE(symb))))
	return 1;
      else if((TYPE_CODE(type2) == T_ARRAY) && 
              (TYPE_CODE(SYMB_TYPE(symb)) == T_ARRAY)){
		  PTR_TYPE types;
		  types = SYMB_TYPE(symb);
		  if(!TYPE_BASE(types) || !TYPE_BASE(type2) ||
		      (isStrictlySameType(TYPE_BASE(types), TYPE_BASE(type2)))){
		          return 1;
			  }
		  else {
		    /* err_symb("(04.1) redeclaration of symbol", SYMB_IDENT(symb)); */
                    Message("checkExternFirstPart,  possible redeclaration of a symbol", yylineno);
		    return 0;
		    } 
	      }
      else
	{
	  /* err_symb("(04) redeclaration symbol",SYMB_IDENT(symb));  */    
	  return 0;
	}
    }
  return 1;
}


/**************************************************************************/
/* was handle_complicated_paramters()*/
PTR_LLND handleComplicatedParameters(p1,p2)
PTR_LLND p1,p2;
{
  PTR_LLND p, tmp;
  PTR_SYMB  symbptr1;
  PTR_TYPE type;

  p = installLowLevelNode(FUNCTION_OP,p1,p2,NULL);
  tmp = giveLlSymbInDeclList(p1);  
  symbptr1 = NODE_SYMB(tmp);
  if (! SYMB_TYPE(symbptr1)) 
    type = SYMB_TYPE(symbptr1) = (PTR_TYPE) newNode(T_FUNCTION);
  else  /* no type checking here */
    { 
      type =  lookForInternalBasetype(SYMB_TYPE(symbptr1));
      switch (TYPE_CODE(type)) {
      case T_FUNCTION:
      case T_POINTER :
      case T_REFERENCE:
	TYPE_BASE(type) = (PTR_TYPE) newNode(T_FUNCTION);
	break;
	default : err("too complicated parameter expressions");
      }
    }
  return(p);
}

/**************************************************************************/
/* Was query_extern(passing_type) */

/* not used? BW Dec 20 1994 */

/*
static int queryExtern(passing_type)
     PTR_TYPE passing_type;
{ 
 
  PTR_TYPE type;
  int flag;

  flag =(int) RID_MAX;
  
  for (type=passing_type ;type ; )
   {   
     if (TYPE_CODE(type) == T_DESCRIPT)
       {
	 if (TYPE_LONG_SHORT(type) & BIT_EXTERN)
	   {
	     flag = (int) RID_EXTERN;
	     return flag;
	   }
	 type = TYPE_BASE(type);
       }
     type = lookForTypeDescript(type);
   }
  return(flag);
}
*/

/**************************************************************************/
/* Was is_complete_chaintype)
  check if a type chain is only a right hand side of a type */
static int pendingBaseType(type)
PTR_TYPE type;
{
  PTR_TYPE temptr;

  for (temptr = type;temptr;temptr= TYPE_BASE(temptr))
    { 
      if (!hasTypeBaseType(TYPE_CODE(temptr)))
	return(1);
    }
  
  return(0);
}


/**************************************************************************/
/* was is_skip_scope(symb) */
/*****************************************************************************/
/* Given a symbol not defined in global but going to changed to be global    */
/* and check if there is any same name symbol lies between pred_bfnd and     */
/* global_bfnd                                                               */
/*         Used for FIELD_NAME, VARIABLE_NAME, FUNCTION_NAME,MEMBER_FUNC     */
/*****************************************************************************/
static int isSkipScope(symb)
PTR_SYMB symb;
{  PTR_SYMB symb_ptr;
   extern PTR_BFND real_scope(),pre_scope();
   int flag;

   flag = SYMB_CODE(symb);
   for (symb_ptr =SYMB_PARENT(symb)->id_attr ;symb_ptr ; symb_ptr = symb_ptr->outer)
     { 
       if (symb_ptr==symb) continue;
       if (SYMB_SCOPE(symb_ptr)==global_bfnd) 
	 continue;
       if (is_same_variant_group(SYMB_CODE(symb_ptr),flag))
	{  
	  PTR_BFND flow ;
	  PTR_BFND type_scope ;                    
	   
	  type_scope = real_scope(SYMB_SCOPE(symb_ptr)) ;
	  for ( flow = SYMB_SCOPE(symb); flow ; flow = pre_scope(flow) )
	    if (flow == type_scope) 
	      return(1);
	}   
     }
   return(0);
 }

  
/**************************************************************************/
/* find_type_symbol_global(symb) */
/*****************************************************************************/
/* Given a symbol not defined in global but going to changed to be global    */
/* return: SMNULL if there is no same entry in global already                */
/*         symb_g if exist                                                   */
/*         Used for FIELD_NAME, VARIABLE_NAME, FUNCTION_NAME,MEMBER_FUNC     */
/*****************************************************************************/
static PTR_SYMB findTypeSymbolGlobal(symb)
     PTR_SYMB symb;
{ 
  PTR_SYMB symb_ptr;
  int flag;

  if (!symb)
    return NULL;
  flag = SYMB_CODE(symb);
  /* missing macro */
  for (symb_ptr = SYMB_PARENT(symb)->id_attr ;symb_ptr; symb_ptr = symb_ptr->outer)
    {  
      if ((is_same_variant_group(SYMB_CODE(symb_ptr),flag))
	  &&(SYMB_SCOPE(symb_ptr) == global_bfnd))
	return(symb_ptr);
    }
  return(NULL);
 }
   

/**************************************************************************/
/* was check_extern_second_part() */
static PTR_SYMB checkExternSecondPart(symb)
PTR_SYMB symb;
{
  int extern1_flag;
  PTR_SYMB symb2;
  PTR_BFND temp;

  if (!symb)
    return NULL;

  if ((SYMB_CODE(symb) == MEMBER_FUNC)
     || (SYMB_CODE(symb) == FUNCTION_NAME))
    return(NULL);
  if (SYMB_DECL(symb)) 
    {
      if(temp = lastBifInBlobList(pred_bfnd))
        extern1_flag =  (temp->decl_specs & BIT_EXTERN) 
                                      ? (int) RID_EXTERN : (int) RID_MAX;
      else extern1_flag = (pred_bfnd->decl_specs & BIT_EXTERN) 
                                      ? (int) RID_EXTERN : (int) RID_MAX;
      if (!pendingBaseType(SYMB_UD_CHAIN(symb)))
	{
	  if ((extern1_flag !=((int) RID_EXTERN))&&(SYMB_CODE(symb)!=FUNCTION_NAME))
	    err_symb("(05) redeclaration symbol",SYMB_IDENT(symb));
	}
      else
	{ 
	  if (!isStrictlySameType(SYMB_TYPE(symb),SYMB_UD_CHAIN(symb)))
	    err_symb("(06) redeclaration symbol",SYMB_IDENT(symb));
	  else 
	    if (SYMB_CODE(symb) !=((int) SYMB_DECL(symb)))
	      err_symb("redelaration symbol",SYMB_IDENT(symb));
	    else 
	      if (SYMB_CODE(symb) == FIELD_NAME)  
		err_symb("(07) redeclaration symbol",SYMB_IDENT(symb));  
	      else
                {
/*                extern2_flag = queryExtern(SYMB_UD_CHAIN(symb));   */
		if ((extern1_flag !=((int) RID_EXTERN))
		/*    &&(extern2_flag !=((int) RID_EXTERN))  */
		    &&(SYMB_CODE(symb)!=FUNCTION_NAME)
		    &&(SYMB_CODE(symb)!=MEMBER_FUNC))
		/*  err_symb("(08) redeclaration symbol",SYMB_IDENT(symb)); */
                    Message("possible redeclaration of a symbol,", yylineno);
                }
	}
    }
  else 
    {
      if (SYMB_SCOPE(symb) == global_bfnd) 
	return(NULL);
      if(temp = lastBifInBlobList(pred_bfnd))
        extern1_flag =  (temp->decl_specs & BIT_EXTERN) 
                                      ? (int) RID_EXTERN : (int) RID_MAX;
      else extern1_flag = (pred_bfnd->decl_specs & BIT_EXTERN) 
                                      ? (int) RID_EXTERN : (int) RID_MAX;
      if ((extern1_flag != ((int) RID_EXTERN))
	  &&(SYMB_CODE(symb)!=FUNCTION_NAME)) 
	return(NULL);
      if (isSkipScope(symb))
	err_symb("invalid scope",SYMB_IDENT(symb));
      if (!(symb2=findTypeSymbolGlobal(symb)))
	{
	  /* vhash */
	  change_scope(symb,global_bfnd);
	}
      else 
	{  
	  if (!isStrictlySameType(SYMB_TYPE(symb),SYMB_TYPE(symb2)))
	    err_symb("(09) redeclaration symbol",SYMB_IDENT(symb));
	  else
	    remove_entry(symb); /* vhash.c */
	  return(symb2);
	}
    }
  return(SMNULL);
}

/**************************************************************************/
/* was 	chain_type_declarator_2 Same as add to base type
   To be change later */
/* Handle decl statement like  <long int foo()> ;   */
/* ptr1    stand for <long int>  i.e. type_pt   r   */
/* ll_ptr1 stand for <foo>  i.e. FUNCTION_REF       */
/* return  stand for FUNCTION    ll_node            */
PTR_TYPE addToDeclaration2(ptr1, symbptr)
     PTR_TYPE ptr1;
     PTR_SYMB symbptr ;
{
  PTR_TYPE oldptr,temptr;
  int double_appear = OFF;

  if (!symbptr)
    return NULL;

  for (oldptr = temptr = SYMB_TYPE(symbptr); temptr ; )
    { 
      if ((TYPE_CODE(temptr) != T_MEMBER_POINTER) &&
          !hasTypeBaseType(TYPE_CODE(temptr)))
	{  
	  double_appear = ON;
	  break;
	}
      else   
	oldptr = temptr;
      if(TYPE_CODE(temptr) == T_MEMBER_POINTER) 
		temptr = TYPE_COLL_BASE(temptr);
      else temptr = TYPE_BASE(temptr);
    }
   if (double_appear == OFF) 
     {
        if (oldptr == NULL){
           SYMB_TYPE(symbptr) = ptr1;
         }
       else{
         if(TYPE_CODE(oldptr) == T_MEMBER_POINTER)
              TYPE_COLL_BASE(oldptr) = ptr1;
	 else TYPE_BASE(oldptr) = ptr1;
	}
   }
  return(SYMB_TYPE(symbptr));
}
    	

/**************************************************************************/
/* Was get_label*/
/**********************************************************************/
/*
 * Make a new label node
 * return <1> new_lab if this is a label never appear for definition and ref
 *        <2> LBNULL  otherwise
 */
/**********************************************************************/
PTR_LABEL InstallLabel(entry)
     PTR_HASH entry;
{
  register PTR_LABEL new_lab;
  PTR_SYMB label_symb;
  
  label_symb = find_type_symbol(entry,LABEL_NAME,area());
  if (!label_symb) 
    {
      new_lab = (PTR_LABEL) newNode(LABEL_KIND);/* one is pseudo for fortran */
      label_symb = install_entry_2(entry,LABEL_NAME,UP_TO_FUNC_HEDR);
      LABEL_SYMB(new_lab) = label_symb;
      LABEL_STMTNO(new_lab)= 0;
      LABEL_SCOPE(new_lab) =   LABEL_SCOPE(label_symb);
      LABEL_USED(new_lab) = NO;
      LABEL_DEFINED(new_lab) = NO ;
      LABEL_ILLEGAL(new_lab) = NO;
/*      LABEL_TYPE(new_lab) = LABUNKNOWN; To find out later */
      LABEL_BODY(new_lab) = BFNULL;
      return (new_lab);
    }
  else
    return((PTR_LABEL) NULL);
}


/**************************************************************************/
/* Was chain_type_with_array_range()*/
/*****************************************************************************/
/* type: a type pointer                                                      */
/* pllnd: a range_list llnd                                                  */
/*****************************************************************************/
PTR_TYPE chainTypeWithArrayRange(type,pllnd)
     PTR_TYPE type;
     PTR_LLND pllnd;
{ 
  PTR_TYPE oldptr;
  PTR_TYPE return_type;
  PTR_LLND p2;
  
  if (!type)
    return NULL;
  return_type = type;
  
  oldptr = lookForInternalBasetype(type);


  if (!oldptr)
    return_type = oldptr = (PTR_TYPE) newNode(T_ARRAY);

  if (TYPE_CODE(oldptr) != T_ARRAY) 
    oldptr = TYPE_BASE(oldptr) = (PTR_TYPE) newNode(T_ARRAY);

  if (!TYPE_DECL_RANGES(oldptr)) 
    TYPE_DECL_RANGES(oldptr) = pllnd;
  else
    {  
      p2 = Follow_Llnd(TYPE_DECL_RANGES(oldptr),2);
      NODE_OPERAND1(p2) = pllnd;
    }
  return(return_type);
}

/**************************************************************************/
PTR_SYMB set_func_cast_with_typename(type,pllnd,flag,entry)
     PTR_TYPE type;
     PTR_LLND pllnd;
     int flag;
     PTR_HASH entry;
{
  PTR_LLND ll,lptr2;
  PTR_BLOB blob;
  PTR_BFND bf1;
  PTR_SYMB symb;
  PTR_SYMB dummy_symb;
  PTR_TYPE new_type;

  if(entry==NULL){
    if ((type->variant!= T_DESCRIPT)&&(type->variant!=T_DERIVED_TYPE))
         err("illegal syntax");
    entry = find_entry_from_type(type);
    new_type = fetch_type_for_constructor(type);
  }
  else{
    new_type = type;
  }
  main_type_flag = TYPE_TWO;
  ll = lazyInstall(pred_bfnd,entry,VARIABLE_NAME,NO);
  symb = getCurrentProcName(pred_bfnd);
  if ((ll->variant == VAR_REF)||(ll->variant ==RECORD_REF) ||
      (ll->variant == FUNCTION_REF))                 
    setSymbNodeToFunction(symb);
  ll = setFuncAfterDeclaratorAppear(pred_bfnd,&cur_flag,ll,pllnd,symb) ;
  lptr2 = addToDeclaration(new_type,OFF,ll,&dummy_symb,YES);
  blob = lastBlobInBlobList(pred_bfnd);
  bf1 = blob->ref ;
  bf1->variant = FUNC_HEDR ;
  bf1->entry.Template.symbol = symb;

  symb->entry.member_func.seen = FUNCTION_BODY_APPEAR;
  SYMB_ATTR(symb) |= OVOPERATOR;
  setParamScope(symb,bf1);
  pred_bfnd = bf1 ;
  main_type_flag = TYPE_FOUR ;
  setParamTypeToDefault(pred_bfnd->entry.Template.symbol);
  return symb;
}
 

/**************************************************************************/
/* to rewrite */
void set_func_hedr2_with_typename(type,pllnd,flag)
PTR_TYPE type;
PTR_LLND pllnd;
int flag;
{
  PTR_LLND ll;
  PTR_BLOB blob;
  PTR_BFND bf1;
  PTR_SYMB symb;
  PTR_SYMB dummy_symb;
  PTR_HASH entry;
  PTR_TYPE new_type;

  if ((type->variant!= T_DESCRIPT)&&(type->variant!=T_DERIVED_TYPE) &&
      (type->variant!= T_DERIVED_TEMPLATE))
    err("illegal syntax");
  entry = find_entry_from_type(type);
  if (flag)   entry = look_up_expanded_string(entry);
  new_type = fetch_type_for_constructor(type);
  main_type_flag = TYPE_TWO;
  ll = lazyInstall(pred_bfnd,entry,VARIABLE_NAME,NO);
  symb = getCurrentProcName(pred_bfnd);
  if ((ll->variant == VAR_REF)||(ll->variant ==RECORD_REF) ||
      (ll->variant == FUNCTION_REF))                 
    setSymbNodeToFunction(symb);
  ll = setFuncAfterDeclaratorAppear(pred_bfnd,&cur_flag,ll,pllnd,symb) ;
  addToDeclaration(new_type,OFF,ll,&dummy_symb,YES);
  blob = lastBlobInBlobList(pred_bfnd);
  bf1 = blob->ref ;
  bf1->variant = FUNC_HEDR ;
  bf1->entry.Template.symbol = symb;

  symb->entry.member_func.seen = FUNCTION_BODY_APPEAR;
  setParamScope(symb,bf1);
  pred_bfnd = bf1 ;
  main_type_flag = TYPE_FOUR ;
  setParamTypeToDefault(pred_bfnd->entry.Template.symbol);
  
}

/**************************************************************************/
void set_func_hedr3_with_typename()
{
  main_type_flag = TYPE_FOUR ;
}


/**************************************************************************/
/*
   routine used by check_field_decl
   return  the  algebraic super-collection of the object.
*/
static PTR_SYMB find_algebraic_base(object)
     PTR_SYMB object;
{ 
  PTR_LLND pllnd;
  PTR_BFND bf1;
  PTR_TYPE type;

   type = lookForInternalBasetype(object->type);
   if (isStructType(TYPE_CODE(type)) ||
       isEnumType(TYPE_CODE(type)) ||
       isUnionType(TYPE_CODE(type))
       )
     {
       bf1=type->entry.derived_class.original_class;
       if ( (pllnd =bf1->entry.Template.ll_ptr3) )
	 return(pllnd->entry.Template.symbol);
     }
  return(NULL);
}

/**************************************************************************/
/* Need More for record_ref point_stat */
int is_variable(pllnd)
PTR_LLND pllnd;
{
  PTR_LLND p1;

  p1= giveLlSymbInDeclList(pllnd);
  if (!p1)
    return(0);
  switch (p1->variant) {
  case ARRAY_REF:
    return(0);
  case FUNCTION_REF:
  default:
    return(1);
  }
}

/**************************************************************************/
/* to handle e.g., &x[1]  BW */
int is_variable_1(pllnd)
PTR_LLND pllnd;
{
  PTR_LLND p1;

  p1= giveLlSymbInDeclList(pllnd);
  if (!p1)
    return(0);
  return (1);
}

/**************************************************************************/
int map_status(rid_value)    /* BW */
int rid_value;
{
  switch (rid_value) 
     {
     case (int) RID_PRIVATE: return(BIT_PRIVATE);
     case (int) RID_FUTURE:  return(BIT_FUTURE);
     case (int) RID_VIRTUAL: return(BIT_VIRTUAL);
     case (int) RID_ATOMIC:  return(BIT_ATOMIC);
     case (int) RID_INLINE:  return(BIT_INLINE);
     case (int) RID_UNSIGNED:return(BIT_UNSIGNED);
     case (int) RID_SIGNED : return(BIT_SIGNED);
     case (int) RID_SHORT :  return(BIT_SHORT);
     case (int) RID_LONG :   return(BIT_LONG);
     case (int) RID_VOLATILE:return(BIT_VOLATILE);
     case (int) RID_CONST   :return(BIT_CONST);
     case (int) RID_GLOBL   :return(BIT_GLOBL);
     case (int) RID_SYNC    :return(BIT_SYNC);
     case (int) RID_TYPEDEF :return(BIT_TYPEDEF); 
     case (int) RID_EXTERN  :return(BIT_EXTERN);
     case (int) RID_AUTO :   return(BIT_AUTO);
     case (int) RID_STATIC : return(BIT_STATIC);
     case (int) RID_REGISTER:return(BIT_REGISTER);
     case (int) RID_FRIEND:  return(BIT_FRIEND);
     case (int) RID_RESTRICT:return(BIT_RESTRICT);
     case (int) RID_KSRPRIVATE:return(BIT_KSRPRIVATE);
     default: return(0);
     }
 }


/**************************************************************************/
int is_basic_block()
{
  PTR_BLOB blob;
  PTR_BFND bf1 ;
  PTR_SYMB symb, functor,list1 ;
  PTR_LLND lptr1,ll_kptr ;

 switch (pred_bfnd->index) {
 case  BRANCH_ON :
        blob = pred_bfnd->entry.Template.bl_ptr2;
        break;
 case BRANCH_OFF:
 default:
        blob = pred_bfnd->entry.Template.bl_ptr1;
	break;
      }
 if ((blob)&&( pred_bfnd->variant!= FUNC_HEDR))
   return(1);
if (pred_bfnd->variant == BASIC_BLOCK) return(1); 
 if (!blob) return(0);
 blob = lastBlobInBlobList(pred_bfnd);
 bf1 =  blob->ref ;
 lptr1 = bf1->entry.Template.ll_ptr1 ;
 if (( lptr1) && ((bf1->variant == VAR_DECL) ||
                  (bf1->variant == CLASS_DECL) ||
                  (bf1->variant == COLLECTION_DECL) ||
		  (bf1->variant == ENUM_DECL) ||
		  (bf1->variant == UNION_DECL) ||
		  (bf1->variant == STRUCT_DECL) ))
   {   ll_kptr = giveLlSymbInDeclList(lptr1);
       if (ll_kptr) symb = ll_kptr->entry.Template.symbol ;
       functor = pred_bfnd->entry.Template.symbol ;
       for ( list1 = functor->entry.member_func.in_list ; list1 ; list1 = list1->entry.var_decl.next_in)
       {
          if ( symb == list1)  return(0);
        }

       if (!list1 )   return(1);

     }

   return(1);
}

/**************************************************************************/
static int is_same_entry(base,new_base)
PTR_TYPE  base,new_base;
{

    if (base==new_base) return(1);
    if ((!base)||(!new_base)) return(0);
    if ((base->variant==T_DERIVED_TYPE)&&(new_base->variant==T_DERIVED_TYPE))
      {
	if (base->entry.derived_type.symbol==new_base->entry.derived_type.symbol) 
           return(1);
      }
    return(0);
  }

  
/**************************************************************************/
/********************************************************************/
/* routine find_entry_from_type:                                    */
/* Used in parser for the constructor function header               */
/*                                                                  */
/********************************************************************/
static PTR_HASH find_entry_from_type(type)
PTR_TYPE type;
{
  /* PTR_TYPE type1; */
  PTR_SYMB s;

 /*
  if (type->variant== T_DESCRIPT)
    type1=type->entry.descriptive.base_type;
  else type1= type;
 */

  if (type->variant==T_DERIVED_TYPE) {
    s=type->entry.derived_type.symbol;
    return(s->parent);
  }
  else if(type->variant == T_DERIVED_TEMPLATE){
	s = TYPE_TEMPL_NAME(type);
	return s->parent;
	}
  else err("illegal syntax");
  return((PTR_HASH) NULL);
}

/**************************************************************************/
static PTR_TYPE fetch_type_for_constructor(type)
PTR_TYPE type;
{

  if (type->variant== T_DESCRIPT){
    type->entry.descriptive.base_type =GetAtomicType(DEFAULT);
    return(type);
  }
  return(GetAtomicType(DEFAULT));
}


/**************************************************************************/
int map2( variant)
int variant ;
{
   switch (variant)  
     {
     case  VARIABLE_NAME : return(VAR_REF);
     case  FUNCTION_NAME :
     case  MEMBER_FUNC   : return(FUNCTION_REF);
     case  FIELD_NAME    : return(VAR_REF);
       default :           return(VAR_REF);
     }
 }

char * standardTypeNames[] = { "int", "char", "float", "double", "void", "long", "short", "unsigned"};

/**************************************************************************/
char* map_rid_to_name(i)
int i;
{
  switch (i) {
  case (int) RID_CHAR : 
    return(standardTypeNames[1]);
  case (int) RID_INT  :
    return(standardTypeNames[0]);
  case (int) RID_FLOAT :
    return(standardTypeNames[2]);
  case (int) RID_DOUBLE : 
    return(standardTypeNames[3]);
  case (int) RID_VOID :
    return(standardTypeNames[4]);
  case (int) RID_LONG :
    return(standardTypeNames[5]);
  case (int) RID_SHORT :
    return(standardTypeNames[6]);
  case (int) RID_UNSIGNED :
    return(standardTypeNames[7]);
  }
  return(standardTypeNames[0]);
}

/**************************************************************************/
PTR_TYPE map_rid_to_type(i)
int i;
{

  switch (i) {
  case (int) RID_CHAR : 
    return(GetAtomicType(T_CHAR));
  case (int) RID_INT  :
    return(GetAtomicType(T_INT));
  case (int) RID_FLOAT :
    return(GetAtomicType(T_FLOAT));
  case (int) RID_DOUBLE : 
    return(GetAtomicType(T_DOUBLE));
  case (int) RID_VOID :
    return(GetAtomicType(T_VOID));
  }
  return(GetAtomicType(T_INT));
}
  

/**************************************************************************/
void eliminate_lastlink()
{
  PTR_BLOB blob, old_blob ;
  
  switch ( pred_bfnd->index ) {
  case BRANCH_OFF:   blob = pred_bfnd->entry.Template.bl_ptr1 ;
                     if (!blob) return;
                     old_blob = (PTR_BLOB) NULL ;
                     for ( blob = pred_bfnd->entry.Template.bl_ptr1; 
                          (blob) && (blob->next) ;
                           blob = blob->next)
                                    old_blob = blob ;
                     if (!old_blob)
	           	        pred_bfnd->entry.Template.bl_ptr1 = (PTR_BLOB) NULL;
                     else
	          	        old_blob->next = (PTR_BLOB) NULL ;
                     break;
   case BRANCH_ON:
                     blob = pred_bfnd->entry.Template.bl_ptr2 ;
                     if (!blob) return;
                     old_blob = (PTR_BLOB) NULL ;
                     for ( blob = pred_bfnd->entry.Template.bl_ptr2; 
                          (blob) && (blob->next) ;
                           blob = blob->next)
                                    old_blob = blob ;
                     if (!old_blob)
	           	        pred_bfnd->entry.Template.bl_ptr2 = (PTR_BLOB)  NULL;
                     else
	          	        old_blob->next = (PTR_BLOB) NULL ;
                     break;
  }
}


/**************************************************************************/
int map_assgn_op(value)
int value;
{
  char * map_op();
  switch (value) {
  case ((int) PLUS_EXPR) : 
      return(PLUS_ASSGN_OP);
  case ((int) MINUS_EXPR):
      return(MINUS_ASSGN_OP);
  case ((int) BIT_AND_EXPR):
      return(AND_ASSGN_OP);
  case ((int) BIT_IOR_EXPR):
      return(IOR_ASSGN_OP);
  case ((int) MULT_EXPR): 
      return(MULT_ASSGN_OP);
  case ((int) TRUNC_DIV_EXPR): 
      return(DIV_ASSGN_OP);
  case ((int) TRUNC_MOD_EXPR): 
      return(MOD_ASSGN_OP);
  case ((int) BIT_XOR_EXPR): 
      return(XOR_ASSGN_OP);
  case ((int) LSHIFT_EXPR):
      return(LSHIFT_ASSGN_OP);
  case ((int) RSHIFT_EXPR):
      return(RSHIFT_ASSGN_OP);
  }
err("Internal Error: map_assgn_op, illegal case");  /* (phb) */
abort();
return 0;
}


/**************************************************************************/
void fill_comment()
{ PTR_BLOB blob1 ;
  PTR_BFND bf1 ;
 
 blob1= lastBlobInBlobList(pred_bfnd);
 if(!blob1) return;
 bf1 =blob1->ref ;
 if (!(bf1->entry.Template.cmnt_ptr))
   {   bf1->entry.Template.cmnt_ptr = fetch_comment_queue(); }

}

/**************************************************************************/
PTR_LABEL find_label(symb)
     PTR_SYMB   symb ;
{  
  PTR_LABEL temp_label;
  if(!fi) return LBNULL;
  temp_label=fi->head_lab;
  for ( ; temp_label ; temp_label = temp_label->next)
    {    if (!strcmp(temp_label->label_name->ident, symb->ident))
	   return(temp_label);
       }
  return(LBNULL);
  
}

/**************************************************************************/
static PTR_FNAME alloc_file()
{
  register PTR_FNAME new;
  
  new = ALLOC(file_name);
  new->id = fi->num_files = ++num_files;
  new->next = (PTR_FNAME) NULL;
  if (num_files == 1)
    head_file = fi->head_file = new;
  else
    cur_thread_file->next = new;
  cur_thread_file = new;
  return (new);
}

/**************************************************************************/
PTR_FNAME make_file(name)
     char *name ;
{
  register PTR_FNAME new_file;
  
  new_file = alloc_file();
  
  new_file->name = name ;
  return (new_file);
}


/**************************************************************************/
char *ckalloc(n)
	register int n;
{
	register char *p;
	void fatal();

	if ( (p = (char *) calloc(1, (unsigned) n)) )
		return (p);

	fatal("out of memory");
	abort();
        return NULL;
}


/**************************************************************************/
static PTR_HASH look_up_expanded_string(entry)
PTR_HASH  entry;
{
  char tmp[256],*s1;
  int  status;

  s1 = tmp;
  *s1++ = '~';
  strcpy(s1,entry->ident);
  s1 = copys(tmp);
  return(look_up_type(s1,&status));
}



/**************************************************************************/
/*
   clean_void_argument(pbf)

    C++ class allows users to specify arguments as void to
    mean there is no argument, this routine eliminate
    the  argument
*/
void clean_void_argument(fi)
PTR_FILE fi;
{
  PTR_SYMB symb1, s1,s2;
  PTR_TYPE type;
  int level;

  for ( symb1=fi->head_symb; symb1; symb1=symb1->thread)
    {
      if ((symb1->variant==FUNCTION_NAME)|| 
	  (symb1->variant==MEMBER_FUNC))
	{
	    s1 = symb1->entry.member_func.in_list;	  
	    if (s1)  s2=s1->entry.var_decl.next_in;
	    if (s1 &&(!s2) ){
	      type =FollowTypeBaseAndDerived(s1->type);
	      level = getIndirectionLevel(s1->type);
	      if (type && (type->variant==T_VOID)&&(level==0))
		symb1->entry.member_func.in_list = (PTR_SYMB) NULL;
	    }
	  }
    }
}

PTR_LLND makeDeclarExp(sym, t)
  PTR_SYMB sym;
  PTR_TYPE t;
{
  PTR_LLND s = NULL;
  int first = 1, done = 0;
  PTR_TYPE tsave = t;
  if( (sym != NULL) && (t != NULL))
	SYMB_TYPE(sym) = t;
  while((! done) && (t != NULL)){ 
	      /* printf("loop var = %d\n", t->variant); */
     switch(TYPE_CODE(t)){
   	case T_POINTER:
	      if(first){
		s = installLowLevelNode(VAR_REF,LLNULL,LLNULL,sym);
		SYMB_TYPE(s) = tsave;
		}
              s = installLowLevelNode(DEREF_OP,s,LLNULL,SMNULL);
	      SYMB_TYPE(s) = t; /* this is wrong but it is consistant with parser. */
	      t = TYPE_BASE(t);
	      /* SYMB_TYPE(s) = t; this should be correct, but because of paser.. */
	      first = 0;
	      break;
	case T_REFERENCE:
              if( first){
		s = installLowLevelNode(VAR_REF,LLNULL,LLNULL,sym);
		SYMB_TYPE(s) = tsave;
		}
              s = installLowLevelNode(ADDRESS_OP,s,LLNULL,SMNULL);
	      SYMB_TYPE(s) = t; /* this is wrong but it is consistant with parser. */
	      t = TYPE_BASE(t);
	      /* SYMB_TYPE(s) = t; this should be correct, but because of paser.. */
	      first = 0;
	      break;
        case T_ARRAY: {
              PTR_TYPE art = t;
	      if(first){
		 s = installLowLevelNode(ARRAY_REF,TYPE_RANGES(art),LLNULL,sym);
		 }
	      else{
		 s = installLowLevelNode(ARRAY_OP,s,TYPE_RANGES(art),SMNULL);
		 }
	      t = TYPE_BASE(t);
	      SYMB_TYPE(s) = tsave;
	      first = 0;
              }
	      break;
        case T_FUNCTION: {
	      PTR_TYPE f = t;
	      if(s == NULL){ warning("error in AddArg", 0); return NULL; }
	      s = installLowLevelNode(FUNCTION_OP,s,LLNULL,SMNULL);
	      t = TYPE_BASE(f);
	      SYMB_TYPE(s) = t;
	      first = 0;
              }
	      break;
	case T_DESCRIPT:
	      t = TYPE_BASE(t);
	      break;
	default:
	      done = 1;
	      if(first){
		s = installLowLevelNode(VAR_REF,LLNULL,LLNULL,sym);
	        SYMB_TYPE(s) = tsave;
		}
	      first = 0;
	      break;
	      }
	}
  return s;
}

void fixTemplateArgs(ll, flag)
PTR_LLND ll; 
int flag; /* flag is true when we are converting to a arg list. */
          /* flag is zero when we are converting to a parameter list. */
{
   PTR_LLND p, tref;
   PTR_TYPE nt;
   PTR_SYMB scsymb;
   PTR_TYPE scnt;

   /* printf("being called\n"); */
 if(flag == 1){
   while(ll){
     /* printf("ll variant is %d\n",  ll->variant); */
      p = NODE_OPERAND0(ll);
      if(NODE_CODE(p) == VAR_REF){
        /*
        if(!NODE_SYMB(p)) printf("a var ref with no symb\n");
	else if(!SYMB_IDENT(NODE_SYMB(p))) 
	   printf(" a var ref with a symbol but no ident\n");
        else printf("got a var ref = %s|\n",  SYMB_IDENT(NODE_SYMB(p)));
	*/
        if( !NODE_SYMB(p) ||
         (NODE_SYMB(p) && ( !(SYMB_IDENT(NODE_SYMB(p))) ||
	             (strlen(SYMB_IDENT(NODE_SYMB(p))) == 0)))){
           /* printf("got inside\n"); */
	   if(NODE_SYMB(p)) nt = SYMB_TYPE(NODE_SYMB(p));
	   else nt = NODE_TYPE(p);
 	   tref = installLowLevelNode(TYPE_REF, LLNULL, LLNULL, SMNULL);
	   NODE_TYPE(tref) = nt;
	   NODE_OPERAND0(ll) = tref;
	   /*
           if (main_type_flag==TYPE_FIVE) {    
                scsymb = find_scope_of_scoped_type_symbol(NODE_SYMB(p), Scope_Name);
                scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                TYPE_SYMB_DERIVE(scnt) = scsymb;
                TYPE_SCOPE_SYMB_DERIVE(nt) = scnt;

                main_type_flag = TYPE_TWO;
                }
	   */
	   }
	}
      ll = NODE_OPERAND1(ll);
      }
   }
}

