/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* Re-written by Pete Beckman 10/13/92 */
/* extended   by Bernd Mohr   07/06/93 */
/* further extented by Darryl Brown 01/27/94 */

/**********************************************************************/
/* pc++.c
   This module is a simple compiled "wrapper" for pC++2dep, that makes it
   look like a standard c++ compiler. */
/**********************************************************************/

# define R_OK 04

/* Include files */
/********************/

#include <stdio.h>
#include <stdlib.h>

#include "sage.h"

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
/* Define macros */
/********************/

/* External Procedures */
/********************/
extern int system();                 /* from ?? */

/* External /lib Procedures */ 
/********************/

/* External References */
/********************/

/* Global Module Variables */
/********************/
int    profileMode = 0; /* Generate profiling code ? 1 = profile, */
                        /*  2 = tracing, 3 = breakpoint */
int    verboseMode = 0; /* Will there be lots of messages to console? */
int    debugMode = 0;   /* Save aborted files? */
int    spmdMode = 1;    /* generate SPMD or master/slave code ? */
int    DepOnly = 0;
char   DefaultArch[20] = "uniproc";

/**********************************************************************/
/* printUsage - prints command line switches and options
   
   Input:
   execName - pointer to char containing name of executable
   Output: 
   To stderr console */
/**********************************************************************/
void printUsage(execName)
     char* execName;
{
  fprintf(stderr,"Usage: %s\n",execName);
  fprintf(stderr,"    [-help | -h] [-usage | -u]\n");
  fprintf(stderr,"    [-debug] [-verbose | -v] [-version]\n");
  fprintf(stderr,"    [-yydebug] [-nodep] [-unparse] \n");
  fprintf(stderr,"    [-deponly] [-CC] [-w] [-ms] \n");
  fprintf(stderr,"    [-profile | -p]\n");
  fprintf(stderr,"    [-trace | -t]\n");
  fprintf(stderr,"    [-breakpoint | -b]\n");
  fprintf(stderr,"    [all standard cpp switches, e.g. -I]\n");
  fprintf(stderr,"    filename\n");
  fprintf(stderr,
     "\n'%s -help' prints description of switches, parameters\n",execName);
}

/**********************************************************************/
/* printHelp - prints description of switches and options
   
   Input:
   execName - pointer to char containing name of executable
   Output:
   to stdout console */
/**********************************************************************/
void printHelp(execName)
     char* execName;
{
  printf("Usage: %s\n",execName);
  printf("  [-h | -help]  Print this help text\n");
  printf("  [-u | -usage]  Print usage information (command line switches)\n");
  printf("  [-v | -verbose]  Turn on Verbose Mode \n");
  printf("  [-debug]  Print useful debugging information\n");
  printf("  [-version]  Print version number of compiled program\n\n");
  printf("  [-deponly]  Compile to .dep only, do not call dep2C++\n");
  printf("  [-CC]  look for include files in /usr/include/CC first\n");
  printf("  [-w]  do not print warnings\n");
  printf("  [-yydebug]  Turn on BISON debugging output\n");
  printf("  [-nodep]  Do not run Faust \n");
  printf("  [-unparse]  parse input (make .dep), then unparse to stdout\n");
  printf("  [-profile | -p]  Generate function profiling code\n");
  printf("  [-trace | -t]  Generate function tracing code\n");
  printf("  [-ms]  Generate master/slave (rather than SPMD) code\n");
  printf("  [-breakpoint | -b]  Allow breakpoint debugging\n");
  printf("  [all standard cpp switches, e.g. -I] \n");
  printf("  filename\n\n");
  
  printf(SAGE_INFO);
}

/**********************************************************************/
void ExitNice()
  {
  if (debugMode)
    {
    fprintf(stderr,"Debug on: Temp File TEMP_pC++.pc.C NOT deleted\n");
    fprintf(stderr,"Debug on: Temp File TEMP_pC++.C NOT deleted\n");
    fprintf(stderr,"Debug on: Temp File TEMP_pC++.dep NOT deleted\n");
    }
  else
    {
    system("/bin/rm -f TEMP_pC++.pc.C");
    system("/bin/rm -f TEMP_pC++.C");
    system("/bin/rm -f TEMP_pC++.dep");
    }
  exit(1);
  }

/**********************************************************************/
void EXECUTE(cbuff)
  char * cbuff;
  {
  int errno;

  if (verboseMode) printf("EXECUTING: %s\n",cbuff);
  errno = system(cbuff);
  if(errno) { 
    fprintf(stderr,"Error in pc++ while executing:\n"); 
    fprintf(stderr,"------->  %s\n",cbuff);
    
    ExitNice();
    }
  }

/**********************************************************************/
/* main
   
   Input :
   command line switches and a filename, or stdin */
/**********************************************************************/
int main(argc, argv)
     int argc;
     char *argv[];
{
  int i;
  char command_buff[1024], cppSwitches[512], tmpSwitches[512], CCinclude[20];
  char SageRoot[1024];
  char *p, *files[30];
  int num_files = 0;
  char prf_name[100];
  char tdl_name[100];
  char edf_name[100];
  char brk_name[100];
  char dep_name[100];
  char out_name[100];
  char pcxx2dep[100];
  char instr[100];
  char dep2cxx[100];
  char fwdSwitches[100];
  char BaseFilename[100];

  char *executableName = argv[0]; /* Save a pointer to the exec name */
  char *IncludePath,*DelimiterPos,*EndString;
  char IncludeString[255];
  char sysinstrcomfile[255];
  char *EnvPathDelimiters=";: ";  /* semicolon, colon, and space valid */
  FILE *tmppcsrc;
  FILE *in;

  if (argc == 1) { printUsage(executableName); return 1; }

  else {
    /* Initialize variables and arrays */
    for(i = 0; i < 30; i++) files[i] = NULL;
    tmpSwitches[0] = 0;
    cppSwitches[0] = 0;
    fwdSwitches[0] = 0;
    CCinclude[0] = 0;
    BaseFilename[0] = 0;
    out_name[0] = 0;

    for(i = 1; i < argc; i++)
      switch(argv[i][0]) {
      case '-': p = argv[i];

	if ((!strcmp(p,"-help")) || (!strcmp(p,"-h"))) {
	  printHelp(executableName); 
	  return 1;  /* error: non-zero exit condition */
	}
	else if ((!strcmp(p,"-usage")) || (!strcmp(p,"-u"))) {
	  printUsage(executableName); 
	  return 1;  /* error: non-zero exit condition */
	}
	else if (!strcmp(p,"-debug")) {
	  debugMode = 1;
	}
	else if (!strcmp(p,"-ms")) {
	  spmdMode = 0;
	}
	else if (!strcmp(p,"-deponly")) {
	  DepOnly = 1;
        }
	else if ((!strcmp(p,"-verbose")) || (!strcmp(p,"-v"))) {
	  verboseMode = 1;
	  strcat(fwdSwitches,"-verbose ");
        }
	else if ((!strcmp(p,"-profile")) || (!strcmp(p,"-p"))) {
	  profileMode = 1;
        }
	else if ((!strcmp(p,"-trace")) || (!strcmp(p,"-t"))) {
	  profileMode = 2;
        }
	else if ((!strcmp(p,"-breakpoint")) || (!strcmp(p,"-b"))) {
 	  profileMode = 3;
	}
	else if ((!strcmp(p,"-sneezy")) || (!strcmp(p,"-z"))) {
	  profileMode = 4;
	}
	else if (!strcmp(p,"-version")) {
	  printf("This is version %s of %s\n",VERSION_NUMBER,executableName); 
	  return 1;  /* error: non-zero exit condition */
	}
	else if (!strcmp(p,"-o")) {
          if ((i == (argc-1)) || (argv[i+1][0] == '-')) {
            fprintf(stderr,"ERROR: not enough arguments for -o switch\n");
            return 1;
          }
          strcpy(out_name,argv[i+1]);
          i++;
          break;
	}

        /* Parse architecture options */
	else if (!strcmp(p,"-arch")) {
          if ((i == (argc-1)) || (argv[i+1][0] == '-')) {
            fprintf(stderr,"ERROR: not enough arguments for -arch\n");
            return 1;
          }
          strcpy(DefaultArch,argv[i+1]);
          i++;
          break;
	}

	else if (!strcmp(p,"-nodep"))
	  strcat(fwdSwitches,"-nodep ");
	else if (!strcmp(p,"-unparse"))
	  strcat(fwdSwitches,"-unparse ");
	else if (!strcmp(p,"-lldebug"))
	  strcat(fwdSwitches,"-debug ");
	else if (!strcmp(p,"-yydebug"))
	  strcat(fwdSwitches,"-yydebug ");
	else if (!strcmp(p,"-CC")) 
	  strcat(CCinclude,"-I/usr/include/CC ");
	else if (!strcmp(p,"-w"))
	  strcat(fwdSwitches,"-w ");
	else {
          strcat(tmpSwitches,p);
          strcat(tmpSwitches," ");
	}
	break;
      default:  files[num_files] = 
	(char *) malloc(strlen(argv[i])+2);
	strcpy(files[num_files++],argv[i]);
	break;
      }	

    pcxx2dep[0] = 0;
     dep2cxx[0] = 0;
       instr[0] = 0;
    DelimiterPos = (char *) strrchr(executableName,47); /* find last "/" */
    if (DelimiterPos) {  /* There is a path, use it for finding pC++2dep */
      *DelimiterPos = 0;
      strcpy(pcxx2dep,executableName); strcat(pcxx2dep,"/");
      strcpy(dep2cxx,executableName);  strcat(dep2cxx,"/");
      strcpy(instr,executableName);    strcat(instr,"/");
      *DelimiterPos = 47; /* reset it back to previous value */
      }
    strcat(pcxx2dep,"pC++2dep");
    strcat(dep2cxx,"dep2C++");
    strcat(instr,"instr");

    /* determine SAGEROOT */
    if (fopen(SAGEROOT, "r") != NULL) {
      strcat(SageRoot, SAGEROOT);
      }
    else if ((in = fopen("./SAGEROOT", "r")) != NULL) {
      fgets(SageRoot, 1024, in);
      fclose(in);
      SageRoot[strlen(SageRoot)-1] = '\0'; /*chop off newline*/
      }
    else if ((p = getenv("SAGEROOT")) != NULL) {
      strcat(SageRoot, p);
      }
    else {
      fprintf(stderr,"ERROR: cannot find SAGEROOT\n");
      return 1;
      }

    /********************************************************/
    /***** Extract and include paths stored in the      *****/
    /***** enviroment variable SAGE_INCLUDE_PATH        *****/
    /********************************************************/
    IncludeString[0] = 0;   /* Set IncludeString to be empty */
    IncludePath = (char *) getenv("SAGE_INCLUDE_PATH");

    if (IncludePath) /* There was a environment var declared */
      {
      /* parse string and pull out pieces */
      EndString = (char *) 
        ((long) strlen(IncludePath) + (long) IncludePath + 1);
      while ((long) IncludePath < (long) EndString)
        {
        strcat(IncludeString," -I");
        /* Find first string delimiter */
        DelimiterPos = (char *) strpbrk(IncludePath,EnvPathDelimiters);
        if (!DelimiterPos) /* There are no more words after this one */
          DelimiterPos = (char *) 
            ((long) strlen(IncludePath) + (long) IncludePath);
        /* Copy word */
        strncat(IncludeString,IncludePath,
             ( (long) DelimiterPos - (long) IncludePath ));
        IncludePath = (char *) ( (long) DelimiterPos + 1);
        }
      if (verboseMode)
        { 
        printf("Include String formed from enviroment variable:\n");
        printf("%s\n",IncludeString);
        }
      }
    else
    if (verboseMode)
       printf("No additional include paths found in enviroment...\n"); 

    if (profileMode) {
      /* generate name for system instrumentation command file */
      sysinstrcomfile[0] = 0;
      strcat(sysinstrcomfile,SageRoot);
      strcat(sysinstrcomfile,"/target/include/sysinstr.prf");
    }

    /********************************************************/
    /***** Begin Parsing C++ Files, each in turn ****/
    /********************************************************/

    strcat(cppSwitches,CCinclude);  /* add CC include string first, if !null */
    strcat(cppSwitches,tmpSwitches);  /* Add any other switches */
    strcat(cppSwitches,IncludeString); /* Add -I from environment var */
    strcat(cppSwitches," -D__PCXX__");  

    for(i = 0; i < num_files; i++){
      /****** HACK FIXME... REMOVE COMMENTS ********/

      strcpy(BaseFilename,files[i]);  /* Make copy of source filename */

      if (profileMode) {
        /* Unfortunately, Cfront only compiles *.C files, so we had
           to name the temp file TEMP_pC++.pc.C  */

        tmppcsrc = fopen ("TEMP_pC++.pc.C", "w");
        if (profileMode == 1) {
          fprintf (tmppcsrc, "#define PROFILING_ON\n");
          fprintf (tmppcsrc, "#include \"profiler.h\"\n");
        }
        else if (profileMode == 2) {
          fprintf (tmppcsrc, "#define TRACING_ON\n");
          fprintf (tmppcsrc, "#include \"pcxx_events_def.h\"\n");
          fprintf (tmppcsrc, "#include \"tracer.h\"\n");
        }
        if (profileMode == 3) {
          fprintf (tmppcsrc, "#define BRKPNT_ON\n");
          fprintf (tmppcsrc, "#include \"pcxx.h\"\n");
	  /* fprintf (tmppcsrc, "#include \"cargo.h\"\n"); */
	  fprintf (tmppcsrc, "#include \"%s\"\n", files[i]);
        } else {
	  fprintf (tmppcsrc, "#include \"%s\"\n", files[i]);
	}
        fclose (tmppcsrc);
#ifdef PCXX_gnu

/* A REAL hack because of wierd __attribute__ special stuff in include files */
#ifdef __bsdi__
        sprintf(command_buff, "g++ -x c++ -E %s TEMP_pC++.pc.C | sed -e s/__attribute__\\(.\\*\\)\\;/\\;/g > TEMP_pC++.C ",
          cppSwitches);
#else
        sprintf(command_buff, "g++ -x c++ -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#endif


#else
#ifdef PCXX_dec
        sprintf(command_buff, "cxx -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#else
#ifdef PCXX_ibm
        sprintf(command_buff, "xlC -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#else
#ifdef PCXX_sgi
        sprintf(command_buff, "NCC -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#else
#ifdef PCXX_cfront
#ifdef __hpux
        sprintf(command_buff, "CC -E -Ac %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#else
        sprintf(command_buff, "CC -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#endif
#else
#ifdef PCXX_pgonxdev
	if (!strcmp(DefaultArch,"paragon")) /* Cross compiling... */
	sprintf(command_buff,"ig++ -x c++ -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
	else
        sprintf(command_buff, "g++ -x c++ -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#else

ERROR, PCXX_* must be defined

#endif
#endif
#endif
#endif
#endif
#endif

      }
      else {
#ifdef PCXX_gnu
/* A REAL hack because of wierd __attribute__ special stuff in include files */
#ifdef __bsdi__
        sprintf(command_buff, "g++ -x c++ -E %s %s | sed -e s/__attribute__\\(.\\*\\)\\;/\\;/g > TEMP_pC++.C ",
          cppSwitches,  files[i]);
#else
        sprintf(command_buff, "g++ -x c++ -E %s %s > TEMP_pC++.C ",
          cppSwitches,  files[i]);
#endif


#else
#ifdef PCXX_dec
        /* Create a dummy file since CC can only do .C files */
        tmppcsrc = fopen ("TEMP_pC++.pc.C", "w");
        fprintf (tmppcsrc, "#include \"%s\"\n", files[i]);
        fclose (tmppcsrc);
	
        sprintf(command_buff, "cxx -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);

#else
#ifdef PCXX_ibm
        /* Create a dummy file since CC can only do .C files */
        tmppcsrc = fopen ("TEMP_pC++.pc.C", "w");
        fprintf (tmppcsrc, "#include \"%s\"\n", files[i]);
        fclose (tmppcsrc);
	
        sprintf(command_buff, "xlC -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);

#else
#ifdef PCXX_sgi
        /* Create a dummy file since CC can only do .C files */
        tmppcsrc = fopen ("TEMP_pC++.pc.C", "w");
        fprintf (tmppcsrc, "#include \"%s\"\n", files[i]);
        fclose (tmppcsrc);
	
        sprintf(command_buff, "NCC -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);

#else
#ifdef PCXX_cfront
        /* Create a dummy file since CC can only do .C files */
        tmppcsrc = fopen ("TEMP_pC++.pc.C", "w");
        fprintf (tmppcsrc, "#include \"%s\"\n", files[i]);
        fclose (tmppcsrc);
	
#ifdef __hpux
        sprintf(command_buff, "CC -E -Ac %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#else
        sprintf(command_buff, "CC -E %s TEMP_pC++.pc.C > TEMP_pC++.C ",
          cppSwitches);
#endif
#else
#ifdef PCXX_pgonxdev
	if (!strcmp(DefaultArch,"paragon")) /* Cross compiling... */
        sprintf(command_buff,"ig++ -x c++ -E %s %s > TEMP_pC++.C ",
          cppSwitches,  files[i]);
	else
        sprintf(command_buff, "g++ -x c++ -E %s %s > TEMP_pC++.C ",
          cppSwitches,  files[i]);
#else


ERROR, PCXX_* must be defined

#endif
#endif
#endif
#endif
#endif
#endif
      }

      DelimiterPos = (char *) strrchr(BaseFilename,'.');
      if (!DelimiterPos) {
        fprintf(stderr,"%s not a valid input filename\n",BaseFilename);
        exit(1);
        }
      *DelimiterPos = 0;   /* Terminate string at . (delete extension) */

      strcpy(dep_name,BaseFilename);  /* copy base name */
      strcat(dep_name,".dep");        /* add extension */

      if (profileMode == 2) {
        /* generate TDL description file name */
        strcpy(tdl_name,BaseFilename);  /* copy base name */
        strcat(tdl_name,".tdl");        /* add extension */

        /* generate edf description file name */
        strcpy(edf_name,BaseFilename);  /* copy base name */
        strcat(edf_name,".edf");        /* add extension */
      }

      if (profileMode == 3) {
	/*  generate typeDescription file name for breakpoint debugging */
        strcpy(brk_name,BaseFilename);  /* copy base name */
        strcat(brk_name,"");        /* add extension "" */
	                                /* for type information file */
      } else if (profileMode) {
        /* test for user instrumenation command file */
        strcpy(prf_name,BaseFilename);  /* copy base name */
        strcat(prf_name,".prf");        /* add extension */
        if ( access (prf_name, R_OK) == 0 )
          sprintf(prf_name, " -c %s.prf", BaseFilename);
        else
          prf_name[0] = 0;
      }

      /* create default output filename if none was specified */
      if (out_name[0] == 0) { /* no output filename, create one... */
        strcpy(out_name,BaseFilename);  /* copy base name */
        strcat(out_name,".C");         /* add extension */
      }

      if (!strcmp(out_name,files[i])) {
        fprintf(stderr,
           "Error: attempting to overwrite source file %s\n",out_name);
        ExitNice();
      }

      /* preprocess the input */
      EXECUTE(command_buff);


#if (!defined(PCXX_gnu) && !defined(PCXX_pgonxdev))
      if (!profileMode) {
        /* Remove dump file */

        sprintf(command_buff, "/bin/rm -f TEMP_pC++.pc.C");
	EXECUTE(command_buff);
      }
#endif      
      /* parse the input */
      sprintf(command_buff, "%s %s TEMP_pC++.C", pcxx2dep, fwdSwitches);
      EXECUTE(command_buff);

      if (profileMode) {
        /* instrument dep file for profiling */
        if (profileMode == 1) 
          sprintf(command_buff,"%s TEMP_pC++.dep -o %s -p -c %s%s",
		  instr, dep_name, sysinstrcomfile, prf_name);
        else if (profileMode == 2) {
          sprintf(command_buff,"%s TEMP_pC++.dep -o %s -t %s -e %s -c %s%s",
		  instr, dep_name, tdl_name, edf_name, sysinstrcomfile, prf_name);
	} else if (profileMode == 3) {
	  /* breakpoint command build access functions */
          sprintf(command_buff,"%s TEMP_pC++.dep -o %s -b %s",
		  instr, dep_name, brk_name);
	}
	else if (profileMode == 4) {
	  /* Sneezy instrumentation. */
	  sprintf( command_buff, "%s TEMP_pC++.dep -o %s -z",
		  instr, dep_name );
	}
        if (debugMode) strcat(command_buff, " -debug");
        if (verboseMode) strcat(command_buff, " -v");
        EXECUTE(command_buff);

        /* remove no longer necessary tmp files */
	if (!debugMode) {
	  sprintf(command_buff,"rm -f TEMP_pC++.pc.C TEMP_pC++.dep");
	  EXECUTE(command_buff);
	}
      }
      else
      {
        /* move the temporary file */
        sprintf(command_buff,"mv TEMP_pC++.dep %s",dep_name);
        EXECUTE(command_buff);
      }

      /* unparse (compile) the .dep file */

      if (profileMode == 3)
      sprintf(command_buff,"%s %s -arch %s -b -o %s %s", 
	      dep2cxx, spmdMode?" ":" -ms ", DefaultArch, out_name, dep_name);
      else if (profileMode == 4)
      sprintf(command_buff,"%s %s -arch %s -map -o %s %s", 
	      dep2cxx, spmdMode?" ":" -ms ", DefaultArch, out_name, dep_name);
      else
      sprintf(command_buff,"%s %s -arch %s -o %s %s", 
	      dep2cxx, spmdMode?" ":" -ms ", DefaultArch, out_name, dep_name);

      if (!DepOnly) EXECUTE(command_buff);

      if (debugMode)
        fprintf(stderr,"Debug on: Temp File TEMP_pC++.C NOT deleted\n");
      else
	system("/bin/rm -f TEMP_pC++.C");
    }
    return 0;
  }
}

