/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* A front-end for ANSI C programs (without C++ keywords) */
/* A near-clone (simplified) of pc++.c */

/* Re-written by Pete Beckman 10/13/92 */
/* extended   by Bernd Mohr   07/06/93 */
/* further extented by Darryl Brown 01/27/94 */

/**********************************************************************/
/* pc++.c
   This module is a simple compiled "wrapper" for pC++2dep, that makes it
   look like a standard c compiler. */
/**********************************************************************/

# define R_OK 04

/* Include files */
/********************/

#include <stdio.h>
#include <stdlib.h>

#include "sage.h"

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
 
/* Define macros */
/********************/

/* External Procedures */
/********************/
extern int system();                 /* from ?? */

/* External /lib Procedures */ 
/********************/

/* External References */
/********************/

/* Global Module Variables */
/********************/
int    verboseMode = 0; /* Will there be lots of messages to console? */
int    debugMode = 0;   /* Save aborted files? */

/**********************************************************************/
/* printUsage - prints command line switches and options
   
   Input:
   execName - pointer to char containing name of executable
   Output: 
   To stderr console */
/**********************************************************************/
void printUsage(execName)
     char* execName;
{
  fprintf(stderr,"Usage: %s\n",execName);
  fprintf(stderr,"    [-help | -h] [-usage | -u]\n");
  fprintf(stderr,"    [-debug] [-verbose | -v] [-version]\n");
  fprintf(stderr,"    [-yydebug] [-nodep] [-unparse] [-w] \n");
  fprintf(stderr,"    [-trace | -t]\n");
  fprintf(stderr,"    [all standard cpp switches, e.g. -I]\n");
  fprintf(stderr,"    filename\n");
  fprintf(stderr,
     "\n'%s -help' prints description of switches, parameters\n",execName);
}

/**********************************************************************/
/* printHelp - prints description of switches and options
   
   Input:
   execName - pointer to char containing name of executable
   Output:
   to stdout console */
/**********************************************************************/
void printHelp(execName)
     char* execName;
{
  printf("Usage: %s\n",execName);
  printf("  [-h | -help]  Print this help text\n");
  printf("  [-u | -usage]  Print usage information (command line switches)\n");
  printf("  [-v | -verbose]  Turn on Verbose Mode \n");
  printf("  [-debug]  Print useful debugging information\n");
  printf("  [-version]  Print version number of compiled program\n\n");
  printf("  [-w]  do not print warnings\n");
  printf("  [-yydebug]  Turn on BISON debugging output\n");
  printf("  [-nodep]  Do not run Faust \n");
  printf("  [-unparse]  parse input (make .dep), then unparse to stdout\n");
  printf("  [-trace | -t]  Generate function tracing code\n");
  printf("  [all standard cpp switches, e.g. -I] \n");
  printf("  filename\n\n");
  
  printf(SAGE_INFO);
}

/**********************************************************************/
void ExitNice()
  {
  if (debugMode)
    {
    fprintf(stderr,"Debug on: Temp File TEMP_Sgcc.c NOT deleted\n");
    fprintf(stderr,"Debug on: Temp File TEMP_Sgcc.dep NOT deleted\n");
    }
  else
    {
    system("/bin/rm -f TEMP_Sgcc.c");
    system("/bin/rm -f TEMP_Sgcc.dep");
    }
  exit(1);
  }

/**********************************************************************/
void EXECUTE(cbuff)
  char * cbuff;
  {
  int errno;

  if (verboseMode) printf("EXECUTING: %s\n",cbuff);
  errno = system(cbuff);
  if(errno) { 
    fprintf(stderr,"Error in sagecc while executing:\n"); 
    fprintf(stderr,"------->  %s\n",cbuff);
    
    ExitNice();
    }
  }

/**********************************************************************/
/* main
   
   Input :
   command line switches and a filename, or stdin */
/**********************************************************************/
int main(argc, argv)
     int argc;
     char *argv[];
{
  int i;
  char command_buff[1024], cppSwitches[512], tmpSwitches[512];
  char *p, *files[30];
  int num_files = 0;
  char dep_name[100];
  char pcxx2dep[100];
  char instr[100];
  char fwdSwitches[100];
  char BaseFilename[100];

  char *executableName = argv[0]; /* Save a pointer to the exec name */
  char *IncludePath,*DelimiterPos,*EndString;
  char IncludeString[255];
  char *EnvPathDelimiters=";: ";  /* semicolon, colon, and space valid */

  if (argc == 1) { printUsage(executableName); return 1; }

  else {
    /* Initialize variables and arrays */
    for(i = 0; i < 30; i++) files[i] = NULL;
    tmpSwitches[0] = 0;
    cppSwitches[0] = 0;
    fwdSwitches[0] = 0;
    BaseFilename[0] = 0;

    for(i = 1; i < argc; i++)
      switch(argv[i][0]) {
      case '-': p = argv[i];

	if ((!strcmp(p,"-help")) || (!strcmp(p,"-h"))) {
	  printHelp(executableName); 
	  return 1;  /* error: non-zero exit condition */
	}
	else if ((!strcmp(p,"-usage")) || (!strcmp(p,"-u"))) {
	  printUsage(executableName); 
	  return 1;  /* error: non-zero exit condition */
	}
	else if (!strcmp(p,"-debug")) {
	  debugMode = 1;
	}
	else if ((!strcmp(p,"-verbose")) || (!strcmp(p,"-v"))) {
	  verboseMode = 1;
	  strcat(fwdSwitches,"-verbose ");
        }
	else if (!strcmp(p,"-version")) {
	  printf("This is version %s of %s\n",VERSION_NUMBER,executableName); 
	  return 1;  /* error: non-zero exit condition */
	}
	else if (!strcmp(p,"-nodep"))
	  strcat(fwdSwitches,"-nodep ");
	else if (!strcmp(p,"-unparse"))
	  strcat(fwdSwitches,"-unparse ");
	else if (!strcmp(p,"-yydebug"))
	  strcat(fwdSwitches,"-yydebug ");
	else if (!strcmp(p,"-w"))
	  strcat(fwdSwitches,"-w ");
	else {
          strcat(tmpSwitches,p);
          strcat(tmpSwitches," ");
	}
	break;
      default:  files[num_files] = 
	(char *) malloc(strlen(argv[i])+2);
	strcpy(files[num_files++],argv[i]);
	break;
      }	

    pcxx2dep[0] = 0;
       instr[0] = 0;
    DelimiterPos = (char *) strrchr(executableName,47); /* find last "/" */
    if (DelimiterPos) {  /* There is a path, use it for finding pC++2dep */
      *DelimiterPos = 0;
      strcpy(pcxx2dep,executableName); strcat(pcxx2dep,"/");
      strcpy(instr,executableName);    strcat(instr,"/");
      *DelimiterPos = 47; /* reset it back to previous value */
      }
    strcat(pcxx2dep,"pC++2dep");
    strcat(instr,"instr");

    /********************************************************/
    /***** Extract and include paths stored in the      *****/
    /***** enviroment variable SAGE_INCLUDE_PATH        *****/
    /********************************************************/
    IncludeString[0] = 0;   /* Set IncludeString to be empty */
    IncludePath = (char *) getenv("SAGE_INCLUDE_PATH");

    if (IncludePath) /* There was a environment var declared */
      {
      /* parse string and pull out pieces */
      EndString = (char *) 
        ((long) strlen(IncludePath) + (long) IncludePath + 1);
      while ((long) IncludePath < (long) EndString)
        {
        strcat(IncludeString," -I");
        /* Find first string delimiter */
        DelimiterPos = (char *) strpbrk(IncludePath,EnvPathDelimiters);
        if (!DelimiterPos) /* There are no more words after this one */
          DelimiterPos = (char *) 
            ((long) strlen(IncludePath) + (long) IncludePath);
        /* Copy word */
        strncat(IncludeString,IncludePath,
             ( (long) DelimiterPos - (long) IncludePath ));
        IncludePath = (char *) ( (long) DelimiterPos + 1);
        }
      if (verboseMode)
        { 
        printf("Include String formed from enviroment variable:\n");
        printf("%s\n",IncludeString);
        }
      }
    else
    if (verboseMode)
       printf("No additional include paths found in enviroment...\n"); 

    /********************************************************/
    /***** Begin Parsing C Files, each in turn ****/
    /********************************************************/

    strcat(cppSwitches,tmpSwitches);  /* Add any other switches */
    strcat(cppSwitches,IncludeString); /* Add -I from environment var */
    strcat(cppSwitches," -D_SAGE_");  

    for(i = 0; i < num_files; i++){
      /****** HACK FIXME... REMOVE COMMENTS ********/

      strcpy(BaseFilename,files[i]);  /* Make copy of source filename */

#ifdef SAGE_gcc
        sprintf(command_buff, "gcc -E %s %s > TEMP_Sgcc.c ",
          cppSwitches,  files[i]);
#else
        sprintf(command_buff, "cc -E %s %s > TEMP_Sgcc.c ",
          cppSwitches, files[i]);
#endif

      DelimiterPos = (char *) strrchr(BaseFilename,'.');
      if (!DelimiterPos) {
        fprintf(stderr,"%s not a valid input filename\n",BaseFilename);
        exit(1);
        }
      *DelimiterPos = 0;   /* Terminate string at . (delete extension) */

      strcpy(dep_name,BaseFilename);  /* copy base name */
      strcat(dep_name,".dep");        /* add extension */

      /* preprocess the input */
      EXECUTE(command_buff);

      /* parse the input */
      sprintf(command_buff, "%s %s TEMP_Sgcc.c", pcxx2dep, fwdSwitches);
      EXECUTE(command_buff);

      /* move the temporary file */
      sprintf(command_buff,"mv TEMP_Sgcc.dep %s",dep_name);
      EXECUTE(command_buff);

      if (debugMode)
        fprintf(stderr,"Debug on: Temp File TEMP_Sgcc.c NOT deleted\n");
      else
	system("/bin/rm -f TEMP_Sgcc.c");
    }
    return 0;
  }
}




