/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************
  A special collection class for Fortran interface.  Used in 
  conjunction with FArray class.  Include this file after kernel.h
  is included.  

  This version interfaces with the new runtime system.  User can use
  lower level functions, FetchFArray and FetchFArrayElem, or higher
  level functions, GetFArray and GetFArrayElem to fetch FArrays and 
  FArray elements.

  Written by Shelby Yang, Indiana University.  (1/6/95)
**********************************************************************/

Collection Fortran : public SuperKernel {
 public:
  Fortran(Distribution *T, Align *A);

  void FetchFArray_1D(int index1,int offset,char *buff,int size);
  void FetchFArray_2D(int index1,int index2,int offset,char *buff,
		      int size);
  void FetchFArrayElem_1D(int index1,int offset,
			  int array_elem_offset,char *buff,int size); 
  void FetchFArrayElem_2D(int index1,int index2,int offset,
			  int array_elem_offset,char *buff,int size); 

  // FArrayDouble related

  void GetFArray(int index1,int offset,FArrayDouble &buff);
  void GetFArray(int index1,int index2,int offset,FArrayDouble &buff);
  double GetFArrayElem(int index,int offset,FArrayDouble &which_farray,
                       int i);
  double GetFArrayElem(int index1,int offset,FArrayDouble &which_farray,
		       int i,int j);
  double GetFArrayElem(int index1,int offset,FArrayDouble &which_farray,
		       int i,int j,int k);
  double GetFArrayElem(int index1,int index2,int offset,
		       FArrayDouble &which_farray,int i);
  double GetFArrayElem(int index1,int index2,int offset,
		       FArrayDouble &which_farray,int i,int j);
  double GetFArrayElem(int index1,int index2,int offset,
		       FArrayDouble &which_farray,int i,int j,int k);

  // FArrayFloat related

  void GetFArray(int index1,int offset,FArrayFloat &buff);
  void GetFArray(int index1,int index2,int offset,FArrayFloat &buff);
  float GetFArrayElem(int index,int offset,FArrayFloat &which_farray,
                       int i);
  float GetFArrayElem(int index1,int offset,FArrayFloat &which_farray,
		       int i,int j);
  float GetFArrayElem(int index1,int offset,FArrayFloat &which_farray,
		       int i,int j,int k);
  float GetFArrayElem(int index1,int index2,int offset,
		       FArrayFloat &which_farray,int i);
  float GetFArrayElem(int index1,int index2,int offset,
		       FArrayFloat &which_farray,int i,int j);
  float GetFArrayElem(int index1,int index2,int offset,
		       FArrayFloat &which_farray,int i,int j,int k);

  // FArrayComplexFloat related

  void GetFArray(int index1,int offset,FArrayComplexFloat &buff);
  void GetFArray(int index1,int index2,int offset,FArrayComplexFloat &buff);
  ComplexFloat GetFArrayElem(int index,int offset,
			     FArrayComplexFloat &which_farray,int i);
  ComplexFloat GetFArrayElem(int index1,int offset, 
			     FArrayComplexFloat &which_farray,int i,int j);
  ComplexFloat GetFArrayElem(int index1,int offset, 
			     FArrayComplexFloat &which_farray,int i,int j, 
			     int k);
  ComplexFloat GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplexFloat &which_farray,int i);
  ComplexFloat GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplexFloat &which_farray,int i,int j);
  ComplexFloat GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplexFloat &which_farray,int i,int j, 
			     int k);

  // FArrayComplex (double) related

  void GetFArray(int index1,int offset,FArrayComplex &buff);
  void GetFArray(int index1,int index2,int offset,FArrayComplex &buff);
  Complex GetFArrayElem(int index,int offset,
			     FArrayComplex &which_farray,int i);
  Complex GetFArrayElem(int index1,int offset, 
			     FArrayComplex &which_farray,int i,int j);
  Complex GetFArrayElem(int index1,int offset, 
			     FArrayComplex &which_farray,int i,int j, 
			     int k);
  Complex GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplex &which_farray,int i);
  Complex GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplex &which_farray,int i,int j);
  Complex GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplex &which_farray,int i,int j, 
			     int k);
};  
// be careful with semicolons; makes strange errors (not always syntax errors)

Fortran::Fortran(Distribution *T, Align *A) : SuperKernel(T, A) 
{}

void Fortran::FetchFArray_1D(int index1,int offset,char *buff,
			     int size)
{
  // for 1D collections
  // offset and size are in bytes. 
  // offset measured from beginning of element class to FArray pointer d.

  int proc_id, elem_index;
  char *source;
  tulip_GlobalPointer_t mygp;

  elem_index = ElemIndex(index1);

  if(Is_Local(index1)) {
    source = (char *) 
      *( (long *) ((char*)tulip_ToLocalPointer(&(ElementList[elem_index])) +offset) );
    memcpy((char*)buff, (char*)source, size);
    return ;
  }

  /* COPY existing global pointer to remote element */
  mygp = ElementReference(elem_index);

  /* Offset the address of remote element by "offset" bytes */
  tulip_SetLocalPointer(&mygp,((char*)tulip_ToLocalPointer(&mygp) + offset));

  /* copy remote FArray pointer to source */ 
  CopyBytes((char *)&source, mygp, sizeof(char*));

  tulip_SetLocalPointer(&mygp, source);

  /* copy remote FArray to the local buff */ 
  CopyBytes((char *)buff, mygp, size);
}

void Fortran::FetchFArray_2D(int index1,int index2,int offset,char *buff,
			     int size)
{
  // for 2D collections
  // offset and size are in bytes. 
  // offset measured from beginning of element class to FArray pointer d.

  int proc_id, elem_index;
  char *source;
  tulip_GlobalPointer_t mygp;

  elem_index = ElemIndex(index1,index2);

  if (Is_Local(index1,index2)) {
    source = (char *)
      *( (long *) tulip_ToLocalPointer(&(ElementList[elem_index])) +offset);
    memcpy((char*)buff, (char*)source, size);
    return ;
  }

  /* Get a global pointer to the remote element */
  mygp = ElementReference(elem_index);

  /* Offset the address of remote element by "offset" bytes */
  tulip_SetLocalPointer(&mygp, ((char*)tulip_ToLocalPointer(&mygp) + offset));

  /* copy remote FArray pointer to source */ 
  CopyBytes((char *)&source, mygp, sizeof(char*));

  tulip_SetLocalPointer(&mygp, source);

  /* copy remote FArray to the local buff */ 
  CopyBytes((char *)buff, mygp, size);
  //tulip_GlobalPointerDestroy(&mygp);
}

void Fortran::FetchFArrayElem_1D(int index1,int offset,
				 int array_elem_offset,char *buff,int size) 
{
  // offset, size and array_elem_offset are in bytes. 
  // offset measured from beginning of element class to FArray pointer d.
  // array_elem_offset measured from beginning of array to specific element  

  int proc_id, elem_index;
  char *source;
  tulip_GlobalPointer_t mygp;

  elem_index = ElemIndex(index1);

  if (Is_Local(index1)) {
    source = (char *)
      *( (long *) tulip_ToLocalPointer(&(ElementList[elem_index]) +offset) );
    // compute address of array element
    source += array_elem_offset;
    memcpy(buff, source, size);
    //tulip_GlobalPointerDestroy(&mygp);
    return ;
  }

  /* Get a global pointer to the remote element */
  mygp = ElementReference(elem_index);

  /* Offset the address of remote element by "offset" bytes */
  tulip_SetLocalPointer(&mygp, ((char*)tulip_ToLocalPointer(&mygp) + offset));

  /* copy remote FArray pointer to source */ 
  CopyBytes((char *)&source, mygp, sizeof(char*));

  // compute address of array element
  source += array_elem_offset;

  tulip_SetLocalPointer(&mygp, source);

  /* copy remote FArray element to buffer */ 
  CopyBytes((char *) buff, mygp, size);
  
  //tulip_GlobalPointerDestroy(&mygp);
}

void Fortran::FetchFArrayElem_2D(int index1,int index2,int offset,
				 int array_elem_offset,char *buff,int size) 
{
  // offset, size  and array_elem_offset are in bytes. 
  // offset measured from beginning of element class to FArray pointer d.
  // array_elem_offset measured from beginning of array to specific element  

  int proc_id, elem_index;
  char *source;
  tulip_GlobalPointer_t mygp;

  elem_index = ElemIndex(index1,index2);

  if (Is_Local(index1,index2)) {
    source = (char *)
      *( (long *) tulip_ToLocalPointer(&(ElementList[elem_index]) +offset) );
    // compute address of array element
    source += array_elem_offset;
    memcpy(buff, source, size);
    //tulip_GlobalPointerDestroy(&mygp);
    return ;
  }

  /* Get a global pointer to the remote element */
  mygp = ElementReference(elem_index);

  /* Offset the address of remote element by "offset" bytes */
  tulip_SetLocalPointer(&mygp, ((char*)tulip_ToLocalPointer(&mygp) + offset));

  /* copy remote FArray pointer to source */ 
  CopyBytes((char *)&source, mygp, sizeof(char*));

  // compute address of array element
  source += array_elem_offset;

  tulip_SetLocalPointer(&mygp, source);

  /* copy remote FArray element to buffer */ 
  CopyBytes((char *) buff, mygp, size);
  
  //tulip_GlobalPointerDestroy(&mygp);
}

//--------- FArrayDouble -------------------
void Fortran::GetFArray(int index1,int offset,FArrayDouble &buff)
{
  FetchFArray_1D(index1, offset, (char*)buff.d, buff.size);
}

void Fortran::GetFArray(int index1,int index2,int offset,FArrayDouble &buff)
{
  FetchFArray_2D(index1, index2, offset, (char*)buff.d, buff.size);
}

double Fortran::GetFArrayElem(int index1,int offset,FArrayDouble &which,
			      int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len) * 
    sizeof(double);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(double));
  return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int offset,FArrayDouble &which,int i,
			      int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(double);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(double));
  return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int offset,FArrayDouble &which,int i,
			      int j, int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(double);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(double));
  return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayDouble &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len)*sizeof(double);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(double));
  return *(double*)buffer;
}


double Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayDouble &which,int i,int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(double);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(double));
  return *(double*)buffer;
}


double Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayDouble &which,int i,int j,int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(double);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(double));
  return *(double*)buffer;
}

//--------- FArrayFloat -------------------
void Fortran::GetFArray(int index1,int offset,FArrayFloat &buff)
{
  FetchFArray_1D(index1, offset, (char*)buff.d, buff.size);
}

void Fortran::GetFArray(int index1,int index2,int offset,FArrayFloat &buff)
{
  FetchFArray_2D(index1, index2, offset, (char*)buff.d, buff.size);
}

float Fortran::GetFArrayElem(int index1,int offset,FArrayFloat &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len)*sizeof(float);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(float));
  return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int offset,FArrayFloat &which,int i,
			      int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(float);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(float));
  return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int offset,FArrayFloat &which,int i,
			      int j, int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(float);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(float));
  return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayFloat &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len)*sizeof(float);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(float));
  return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayFloat &which,int i,int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(float);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(float));
  return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayFloat &which,int i,int j,int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(float);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(float));
  return *(float*)buffer;
}

//------- FArarrayComplexFloat ----------
void Fortran::GetFArray(int index1,int offset,FArrayComplexFloat &buff)
{
  FetchFArray_1D(index1, offset, (char*)buff.d, buff.size);
}

void Fortran::GetFArray(int index1,int index2,int offset,
			FArrayComplexFloat &buff)
{
  FetchFArray_2D(index1, index2, offset, (char*)buff.d, buff.size);
}

ComplexFloat Fortran::GetFArrayElem(int index1,int offset,
				    FArrayComplexFloat &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len) *
    sizeof(ComplexFloat);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,
		     sizeof(ComplexFloat));
  return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int offset,
				    FArrayComplexFloat &which,int i,int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len) *
    sizeof(ComplexFloat);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,
		     sizeof(ComplexFloat));
  return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int offset,
				    FArrayComplexFloat &which,
				    int i,int j, int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(ComplexFloat);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(ComplexFloat));
  return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplexFloat &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len)*sizeof(ComplexFloat);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(ComplexFloat));
  return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplexFloat &which,int i,int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(ComplexFloat);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(ComplexFloat));
  return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplexFloat &which,int i,int j,int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(ComplexFloat);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(ComplexFloat));
  return *(ComplexFloat*)buffer;
}


//------- FArarrayComplex ----------
void Fortran::GetFArray(int index1,int offset,FArrayComplex &buff)
{
  FetchFArray_1D(index1, offset, (char*)buff.d, buff.size);
}

void Fortran::GetFArray(int index1,int index2,int offset,FArrayComplex &buff)
{
  FetchFArray_2D(index1, index2, offset, (char*)buff.d, buff.size);
}

Complex Fortran::GetFArrayElem(int index1,int offset,FArrayComplex &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len)*sizeof(Complex);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(Complex));
  return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int offset,FArrayComplex &which,int i,
			      int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(Complex);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(Complex));
  return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int offset,FArrayComplex &which,int i,
			      int j, int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(Complex);
  FetchFArrayElem_1D(index1,offset,array_elem_offset,buffer,sizeof(Complex));
  return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplex &which,int i)
{
  int array_elem_offset;
  array_elem_offset=(i+ 0*which.i_len+ 0*which.i_len*which.j_len)*sizeof(Complex);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(Complex));
  return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplex &which,int i,int j)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ 0*which.i_len*which.j_len)*sizeof(Complex);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(Complex));
  return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplex &which,int i,int j,int k)
{
  int array_elem_offset;
  array_elem_offset=(i+ j*which.i_len+ k*which.i_len*which.j_len)*sizeof(Complex);
  FetchFArrayElem_2D(index1,index2,offset,array_elem_offset,buffer,sizeof(Complex));
  return *(Complex*)buffer;
}
