/*------------------------------------------------------------
 *  Darryl Brown
 *  
 *  pC++ tools research group, University of Oregon
 *  October, 1990
 *  
 *  cargo.h
 *
 *  An implementation of a packaging system for sending/receiving
 *  data over connections using sockets.
 *
 *------------------------------------------------------------
 */

#ifndef CRG_HDR
#define CRG_HDR
#include "transport.h"

#ifdef __cplusplus 
extern "C" {
#endif

/*
 *------------------------------------------------------------
 *  Macros
 *
 */

#define MAX_CONNECTIONS 100
#define BASE_PORT_NUM 7300
#define CRG_HDR_LEN 7

/*------------------------------------------------------------
 *  Types
 */

/*  package types */

typedef enum packageEnum {
  CRG_shared, CRG_unique
} packageTypes;

#define CRG_shared 0
#define CRG_unique 1

typedef struct aPackageStruct {
  int length;
  int bufLength;
  void *data;
  int packageType;
} aPackage;

/*------------------------------------------------------------
 *  int CRG_SendPackage(int connNum, aPackage *pkg)
 *    Sends a message over connection, preceded by control
 *    information.
 *
 *------------------------------------------------------------
 */
int CRG_SendPackage(int connNum, aPackage *pkg); 

/*------------------------------------------------------------
 *  aMessage *CRG_ReceivePackage(int connNum)
 *    Receieves a package over connection.
 *
 *------------------------------------------------------------
 */
aPackage * CRG_ReceivePackage(int connNum); 
     
/* ------------------------------------------------------------
 *  void CRG_RecyclePackage(aPackage *pkg)
 *    Frees up the space allocated for a package
 *
 */
void CRG_RecyclePackage(aPackage *pkg);

/*
 *------------------------------------------------------------
 *  void CRG_InitPackage(int packType)
 *    Initializes a package so that it can be packed with 
 *    one or more calls to CRG_Pack.
 *
 *------------------------------------------------------------
 */
aPackage  *CRG_InitPackage(int packType);


/*
 *------------------------------------------------------------
 *  int CRG_Pack(aPackage *pkg, void *data, int numBytes)
 *    Packs new information onto the package, pkg.
 *
 *------------------------------------------------------------
 */
int CRG_Pack(aPackage *pkg, void *data, int numBytes);

#ifdef __cplusplus 
}
#endif


#endif


