/* --------------------------------------------------
 *
 *  Darryl Brown
 *  
 *  March, 1995
 *
 *  collInst.C
 *  
 *  It includes the 
 *  neccessary items for the keeping track of the collection
 *  instances during runtime. 
 *
 * --------------------------------------------------
 */

/* see if this file has already been included; */
#ifdef BRKPNT_ON
#ifndef COLL_INST_DEF
#define COLL_INST_DEF


/* 
 * --------------------------------------------------------------
 *  GLOBALS
 * --------------------------------------------------------------
 */

// The variable that specifies whether or not global queues
// have been initialized.
int DA_Init = 0;

/* This points to the queue which names of collection variables are */
/*  placed just before they are created. */
DA_collQueue *DA_collDeclQueue = new DA_collQueue;

/* This points to the tables (list) of collection instances and a record */
/* of information about them. */
DA_collQueue *DA_collInstanceTable = new DA_collQueue;

/* 
 * --------------------------------------------------------------
 *  CLASS METHOD IMPLEMENTATION
 * --------------------------------------------------------------
 */


/*
 *   DA_collQueue() - the class constructor.
 *
 */
DA_collQueue::DA_collQueue(void ) {
  last = first = NULL;
  length = 0;
}

/*
 *  void push(DA_collInstance *collInst) - stores
 *       the collInstance in a list.
 *
 */
void DA_collQueue::push(DA_collInstance *collInst) {
  if (last != NULL) {
    /* it is not the first name... */
    last->next = collInst;
  } else {
    /* it IS the first name... */
    first = collInst;
  }
  last = collInst;
  length++; 
  collInst->next = NULL;
}


/*
 *  DA_collInstance *DA_collQueue::pop(void) - pops 
 *       the top Instance in the list.
 *
 */
DA_collInstance *DA_collQueue::pop(void) {
  DA_collInstance *collInst, *tci;
  
  collInst = last;
  

  if (last == first) {
    /* it was the only entry */
    first = last = NULL;
  } else {
    /* find the next to last entry,
     *  and make it the last .
     */
    tci = first;
    while (tci->next != collInst ) {
      tci = tci->next;
    }
    last = tci;
    tci->next = NULL;
  }
  
  length--; 

  return collInst;
}


/*
 *  int *elemIndex(int idx1, int idx2) - pops 
 *       the top Instance in the list.
 *
 */
int *DA_collInstance::elemIndex(int idx1, int idx2) {
  Kernel *kptr;

  kptr = (Kernel *) collectionPtr;
  return kptr->ElemIndex(idx1, idx2);
}



virtual void DA_collInstance::printToBuf(char *buf) {
  char tbuf[1000];
  Kernel *kptr;

#if DEBUG
  printf("printing collInstance varName \"%s\"\n",
	 varName);
  fflush(stdout);
#endif

  kptr = (Kernel *) collectionPtr;
  if (numDims == 1) {
    sprintf(tbuf, "%s < %s > [%d] %s", 
	    collName, elemName, kptr->CollectionInfo.dim1size, varName);
  } else if (numDims == 2) {
    // numDims must equal 2;
    sprintf(tbuf, "%s < %s > [%d][%d] %s", 
	    collName, elemName, kptr->CollectionInfo.dim1size, 
	    kptr->CollectionInfo.dim2size, varName);
  } else {
    // error, cannot handle more than 2 dimensions;
    sprintf(tbuf, "More than 2 dimensional distribution, no work\n");
  }
      
  strcpy(buf, tbuf);
}

/*
 *  DA_collQueue *printToBuf() - 
 *       prints all elements into buffer.
 *
 */
void DA_collQueue::printToBuf(char *pbuf) {
  DA_collInstance *collInst;
  char buf[500], retStr[4];

  strcpy(pbuf, "");
  sprintf(retStr, "\n");
  collInst = first;
  while (collInst) {
    collInst->printToBuf(buf);
#if DEBUG      
    printf("collInst %s <%s> %s: %s\n", 
	    collInst->collName,
	    collInst->elemName,
	    collInst->varName,
	    buf);
#endif
    strcat(pbuf, buf);
    strcat(pbuf, retStr);
    collInst = collInst->next;
  }
}

/*
 *  DA_collQueue *getIth() - pops 
 *       the top Instance in the list.
 *
 */
DA_collQueue *DA_collQueue::getIth(int i) {
  DA_collInstance *collInst;
  int tmp;

  collInst = first;
  tmp = 1;

  while (collInst && (tmp < i)) {
    collInst = collInst->next;
    tmp++;
  }
  return collInst;
}



// ---------------------------------------------------------------------
// this function adds a colleciton name and type to a queue, so that when;
// it is created (immediatly after this), we know the variable name and type;
// of the collection.;

void brk_QUE_COLLECTION_INSTANCE (char *collName, char *elemType, 
                              char *varName) {
  DA_collInstance *newColl;

  // only execute if you are the zeroth process 
  if (!(pcxx_NodeNumber() == 0)) {
    return;
  }
  if (! (DA_Init)) {
    DA_collDeclQueue = new DA_collQueue;
    DA_collInstanceTable = new DA_collQueue;
    DA_Init = 1;
  }

#if DEBUG
  printf("queing collection instance \n");
  fflush(stdout);
#endif

  newColl = new DA_collInstance;
  newColl->varName = strsave(varName);
  newColl->collName = strsave(collName);
  newColl->elemName = strsave(elemType);
  DA_collDeclQueue->push(newColl);
}


// ---------------------------------------------------------------------
// this function deletes a colleciton entry from the table, it is called
// from the SuperKernel destructor.

void brk_DELETE_COLLECTION_ENTRY (void *collPtr) {
  DA_collInstance *coll, *last;

  // only execute if you are the zeroth process 
  if (!(pcxx_NodeNumber() == 0)) {
    return;
  }
  
  coll = DA_collInstanceTable->first;
  last = NULL;
  while (coll) {
    if (collPtr == coll->collectionPtr) {
      if (last) {
	/* delete the coll node (it isnt the first */
	last->next = coll->next;
      } else {
	/* delete the coll node (it IS the first */
	DA_collInstanceTable->first = coll->next;
      }
      DA_collInstanceTable->length --;
      break;
    }
  }
}


// ---------------------------------------------------------------------
// this function is called during the construction of a collection.  It
// retrieves some important information regarding the collection and
// stores it int the collection instance table.;
void brk_ADD_COLLECTION_ENTRY(int numElems, int sizeElem, 
			      Kernel *newC) {
  
  DA_collInstance *collInst;

  // only execute if you are the zeroth process 
  if (!(pcxx_NodeNumber() == 0)) {
    return;
  }

#if DEBUG
  printf("in adding collection entry\n");
  fflush(stdout);
#endif

  collInst = DA_collDeclQueue->pop();

  if (!collInst) {
    perror("created a collection that was not instrumented.\n");
    exit(1);
  }
  
  collInst->numElems = numElems;
  collInst->sizeElem = sizeElem;
  collInst->numDims = newC->CollectionInfo.number_dimension;
  collInst->dim1size = newC->CollectionInfo.dim1size;
  if (collInst->numDims > 1) {
    collInst->dim2size = newC->CollectionInfo.dim2size;
  }
  collInst->sizeElem = sizeElem;
  collInst->collectionPtr = (void *) newC;

  // ;
  DA_collInstanceTable->push(collInst);
}

#endif
#endif
