/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/* FIND an element.  Returns a reference (global pointer) to the
element.  This function uses any mapping information, caching stuff,
remapping information, etc.  Put all the "locator" smarts here */

/* For this distributed memory version, we assume each processor can,
if it does not have a global pointer to an element, find it by using
AllocatedProc() to find the owner, and then looking in the owner's own
ElementList, and caching the result.  A version which supports
remapping or movement of elements, must be able to invalidate these
caches. */

/*********************************************************************/
tulip_GlobalPointer_t Kernel::ElementReference(int elemIndex)
{ 
  /* elemIndex = 1 dimensional ID for element */

  tulip_GlobalPointer_t mygp1, mygp2;
  C_Record record;
  int proc_id;
  tulip_ACK_t *ackhandle;

  ackhandle = tulip_ACKCreate(1);

  /* Get global pointer to element from CollDbase */
  mygp2 = (CollDbase[CollectionNumber].TableOfElements)[elemIndex];

  /* Is the whereabouts of this element cached? */
  if (tulip_ToContext(&mygp2) < 0) {

    /* No! I don't know where it is */

    /* Is there one or two dimensions? */
    if (CollectionInfo.number_dimension == 1) {
      /* Calculate the processor that owns the element */
      proc_id = AllocatedProc(elemIndex);
    }
    else {
      /* 2 dimensions... */
      proc_id = 
	AllocatedProc( elemIndex % CollectionInfo.dim1size,
		       elemIndex / CollectionInfo.dim1size );
    }

    /* Global pointer which points to CollDBase on remote processor */
    mygp1 = GlobalPtrTable[proc_id];

    /* Compute collection record address on remote processor */
    tulip_SetLocalPointer(&mygp1, ((char *) 
      ( ((C_Record *) tulip_ToLocalPointer(&mygp1)) + CollectionNumber )));

    pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_begin_fetch, aa_poll_count);

    pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_begin_fetch); 
  /* sets Ariadne break point if needed */
    /* Fetch collection record pointed by LocalPointer */ 
#ifdef SNEEZY
    snz_begin_fetch_event(pcxx_NodeNumber(), &record, &mygp1, sizeof(C_Record));
#endif

    tulip_Get((char *) &record, mygp1, sizeof(C_Record), ackhandle);
    tulip_WaitACK(ackhandle);
#ifdef SNEEZY
    snz_end_fetch_event(pcxx_NodeNumber(), &record, &mygp1, sizeof(C_Record));
    /* check - the srcid, srcadd, destaddr, size --> dest address is in the 
       LocalAddress field of the global pointer. */
#endif

    /* Extract TableOfElement pointer from fetched collection record */
    tulip_SetLocalPointer(&mygp1, ((char *) record.TableOfElements));

    /* Compute address of global pointer that points to element */
    tulip_SetLocalPointer(&mygp1, ((char *) 
        ( ((tulip_GlobalPointer_t *) 
	   tulip_ToLocalPointer(&mygp1)) + elemIndex )));

#ifdef SNEEZY
    snz_begin_fetch_event(pcxx_NodeNumber(), &mygp2, &mygp1, sizeof(tulip_GlobalPointer_t));
/* sneezy event signalling */
#endif
    /* Fetch global pointer that points to element */
    tulip_Get((char *) &mygp2, mygp1, 
	      sizeof(tulip_GlobalPointer_t), ackhandle);
    tulip_WaitACK(ackhandle);
#ifdef SNEEZY
    snz_end_fetch_event(pcxx_NodeNumber(), &mygp2, &mygp1, sizeof(tulip_GlobalPointer_t));
#endif
    pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_end_fetch, aa_poll_count);
    /* There are two tulip_Gets here - but they can be considered as one for 
	Ariadne */

    pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_end_fetch); 
  /* sets Ariadne break point if needed */

    /* Cache global pointer to element */
    (CollDbase[CollectionNumber].TableOfElements)[elemIndex] = mygp2;
  } 

  tulip_ACKDestroy(ackhandle);

  return mygp2;
}

/*********************************************************************/
/* Very Similar to the run-time primitive, tulip_Get, but gives
the function to the "User", without having to understand the RTS */
/*********************************************************************/
void Kernel::CopyBytes(char *localbuffer, tulip_GlobalPointer_t GPsrc, int len)
{ 
  /* localbuffer = the destination */
  /* GPsrc       = global pointer to (remote) source */
  /* len         = length of bytes to grab (in bytes) */

  tulip_ACK_t *ackhandle;
  tulip_GlobalPointer_t mygp1;

  ackhandle = tulip_ACKCreate(1);

  /* Call the RTS primitives */
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_begin_fetch, aa_poll_count);

  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_begin_fetch); 
  /* sets Ariadne break point if needed */
#ifdef SNEEZY
  mygp1.NodeID = pcxx_NodeNumber(); 
  mygp1.LocalAddress = localbuffer;
  snz_begin_fetch_event(pcxx_NodeNumber(), &mygp1, &GPsrc, len);
#endif
  tulip_Get((char *) localbuffer, GPsrc, len, ackhandle);
  tulip_WaitACK(ackhandle);
#ifdef SNEEZY
  snz_end_fetch_event(pcxx_NodeNumber(), &mygp1, &GPsrc, len);
#endif
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_end_fetch, aa_poll_count);

  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_end_fetch); 
  /* sets Ariadne break point if needed */

  tulip_ACKDestroy(ackhandle);

  return;
}




