/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/* FIND an element.  Returns a reference (global pointer) to the
element.  This function uses any mapping information, caching stuff,
remapping information, etc.  Put all the "locator" smarts here */

/* For this shared memory version, we assume each processor ALREADY
has a complete and full list global pointers to elements.  Other
versions, which let the elements move, or be remapped, will need
rewrite this simplistic approach. */

/*********************************************************************/
tulip_GlobalPointer_t Kernel::ElementReference(int elemIndex)
{ 
  /* elemIndex = 1 dimensional ID for element */

  tulip_GlobalPointer_t mygp1,mygp2;
  C_Record record;
  int proc_id;

  /* Get global pointer to element from CollDbase */
  mygp2 = (CollDbase[CollectionNumber].TableOfElements)[elemIndex];

  /* Is the whereabouts of this element cached? */
  if (tulip_ToContext(&mygp2) < 0) {

    /* No! I don't know where it is */

    /* Is there one or two dimensions? */
    if (CollectionInfo.number_dimension == 1) {
      /* Calculate the processor that owns the element */
      proc_id = AllocatedProc(elemIndex);
    }
    else {
      /* 2 dimensions... */
      proc_id = 
	AllocatedProc( elemIndex % CollectionInfo.dim1size,
		       elemIndex / CollectionInfo.dim1size );
    }

    /* Global pointer which points to CollDBase on remote processor */
    mygp1 = GlobalPtrTable[proc_id];

    /* Compute collection record address on remote processor */
    tulip_SetLocalPointer(&mygp1, ((char *) 
      ( ((C_Record *) tulip_ToLocalPointer(&mygp1)) + CollectionNumber )));
    
    /* Fetch pointer to TableOfElements */
    tulip_SetLocalPointer(&mygp1,
      ((char *) ((C_Record *) tulip_ToLocalPointer(&mygp1))->TableOfElements));

    /* Compute address of global pointer that points to element */
    tulip_SetLocalPointer(&mygp1, 
      ((char *) ( ((tulip_GlobalPointer_t *) 
		   tulip_ToLocalPointer(&mygp1)) + elemIndex)));

    /* Fetch global pointer that points to element */
    mygp2 = *(tulip_GlobalPointer_t *) tulip_ToLocalPointer(&mygp1);

    /* Cache global pointer to element */
    (CollDbase[CollectionNumber].TableOfElements)[elemIndex] = mygp2;
  } 

  return mygp2;
}

/*********************************************************************/
/* Very Similar to the run-time primitive, tulip_Get, but gives
the function to the "User", without having to understand the RTS */
/*********************************************************************/
void Kernel::CopyBytes(char *localbuffer, tulip_GlobalPointer_t GPsrc, int len)
{
  /* localbuffer = the destination */
  /* GPsrc       = global pointer to (remote) source */
  /* len         = length of bytes to grab (in bytes) */

  /* No reason to do fancy stuff, on this shared memory, we can memcpy */
  memcpy((void *) localbuffer, tulip_ToLocalPointer(&GPsrc), len);
  return;
}

