/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/****************************************************************
 *                                                              *
 *         pC++ kernel classes and SuperKernel collection       *
 *         This file should be included only after are          *
 *         element classes defined.                             *
 *                                                              *
 *         Originally written by Francois Bodin                 *
 *         Modified often by:                                   *
 *         Pete Beckman, Bernd Mohr, Shelby Yang                *
 *         Computer Science, Indiana University                 *
 *                                                              *
 ****************************************************************/
/* ASSUME pcxx.h already included, and therefore some RTS stuff */

#include "kernel_defs.h"
#include "profiler.h"
#include "pcxx_events_def.h"
// This defines brkpointing files...
#include "distApi.h"
// The breezyInc.h file is included AFTER the kernel class definition
#include "pcxx_ariadne.h" 
//Ariadne specific events defined here 

#ifdef TRACING_ON
#  define TRC_DECL_PARAM long trcparam
#  define TRC_SET_PARAM(s) trcparam = (mygp.NodeID<<24)+(CollectionNumber<<16)+s
#else
#  define TRC_DECL_PARAM
#  define TRC_SET_PARAM(s)
#endif

#ifdef _AA_REPLAY
extern "C" int pcxx_CheckForBreakPoint(int rwtype, int msgtag_no);
#else 
#define pcxx_CheckForBreakPoint(rwtype, msgtag_no)
/* define it as null in case we don't have replay */
#endif
 /* pcxx_GetBkFlag(pcxx_NodeNumber(), rwtype, mtag_string_array[msgtag], 1) */
/* here we've defined a dummy function. It calls the breakpoint function with 
True as the stop_at_break_point argument. */

#ifdef SNEEZY
#include "pcxx_snz.h"
#endif
/*******************************************************************/
/* Global Variable to ALL Run-time environments */

/*******************************************************************/

#define WHOLE_ELEMENT -1
#define MAXCOLLECTION 100
#define PCXX_BOOTSTRAP 23
#define PCXX_RPC 29
#define UNDEFINED_NODE -99

static int NumberOfCollection = 0;
static char *CurrentCollectionTraversal[MAXPARAL];
static int TraversalLevel = -1;



/* Forward Declaration:  In "user" code */
void Processor_Main(int argc, char **argv);

#ifdef _MASTER_SLAVE_
/* Forward declaration */
int MShack_IamtheOwner(int elem);
int MShack_InParallelSection();
#endif

/*******************************************************************/
/*******************************************************************/
/* Globally overload "new" and "delete */

void * operator new(size_t size) {
    return (void *)tulip_SharedMalloc(size);
}

void operator delete(void *p) {
  tulip_SharedFree(p);
}

/*******************************************************************/
/*******************************************************************/

C_Record *CollDbase;         /* The Database of all known Collections */
tulip_GlobalPointer_t *GlobalPtrTable;   /* Memory Communication Area */

void pcxx_Barrier() {
#ifdef SNEEZY
  static int barrier_id = 0;
  barrier_id++;
#endif
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_begin_barrier, aa_poll_count);
  /* Ariadne Trace with parameter 0 */
  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_begin_barrier); /* sets Ariadne break point if needed */
#ifdef SNEEZY
  snz_begin_barrier_event(barrier_id);
#endif
  
  if (pcxx_NodeNumber() == 0) {
    brk_BREAKPOINT_EXECUTIVE(); 
  }
  tulip_Barrier();
  brk_CHECK_PARALLEL();
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_end_barrier, aa_poll_count);
  /* Ariadne Trace with parameter 0 */
  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_end_barrier); /* sets Ariadne break point if needed */
#ifdef SNEEZY
  snz_end_barrier_event(barrier_id);
#endif
}


/*******************************************************************/
/*******************************************************************/

class Processors{
  public :
    int dimension;
};

/*********************************************************************/
class Distribution {
 public:
   int nbdim;
   int nbelement;
   int dim1size, dim2size, dim3size;
   int dist1, dist2, dist3;
   Processors P;
   Distribution();
   Distribution(int d1size, Processors *P, int dt1);
   Distribution(int d1size, int d2size, Processors *P, int dt1, int dt2);
   Distribution(int d1size, int d2size, int d3size, Processors *P, int dt1, 
	    int dt2, int dt3);
   int GiveElementProc(int i);
   int GiveElementProc(int i, int j);
   int GiveElementProc(int i, int j, int k);
};

/*********************************************************************/
Distribution::Distribution() {
  nbdim =1;
  nbelement = pcxx_TotalNodes();
  dim1size = pcxx_TotalNodes();
  dist1 = CYCLIC;
}

/*********************************************************************/
Distribution::Distribution(int d1size, Processors *, int dt1) {
  nbdim =1;
  nbelement = d1size;
  dim1size = d1size;
  dist1 = dt1;
}

/*********************************************************************/
Distribution::Distribution(int d1size, int d2size, Processors *, int dt1, int dt2) {
  nbdim =2;
  nbelement = d1size*d2size;
  dim1size = d1size;
  dim2size = d2size;
  dist1 = dt1;
  dist2 = dt2;
}

/*********************************************************************/
Distribution::Distribution(int d1size, int d2size, int d3size, Processors *, 
		   int dt1, int dt2, int dt3) {
  nbdim =2;
  nbelement = d1size*d2size*d3size;
  dim1size = d1size;
  dim2size = d2size;
  dim3size = d3size;
  dist1 = dt1;
  dist2 = dt2;
  dist3 = dt3;
  pcxx_Message("Distribution(): Sorry, 3D template not implemented.",PCXX_FATAL);
}
     
/*********************************************************************/
int Distribution::GiveElementProc(int i) {
  int proc = 0;
  int p, rem, q;
  switch (dist1) {
    case BLOCK:

/**************************************************************************
Notes from Suresh Srinivas (6/23/93):

New version of Block distribution, tries to keep the allocation
of elements to processors balanced. It tries to divide the data-structure
into contiguous blocks whose sizes are identical, except for the last
block which may be of a smaller size.  

Let `dimsize' be the size of a distributee in a certain dimension
and let `Numproc' be the number of processors in that dimension.
[Note: Integer division (i.e x/y where x and y are integers) gives the
same result as ceil((x*1.0)/(y*1.0))]

Let p   = dimsize/Numproc
    rem = dimsize - Numproc * (dimsize/Numproc)
    q = if (p <= rem) p + 1  else  p

The distribution function `delta' on the index space in that dimension is
delta(i) = { j } for all i, 0 <= i <= (dimsize - 1)
            where  j = i/q
            (note if (j >= Numproc) then j = (Numproc - 1) - j % Numproc
[Note: When j exceeds Numproc; it is better to keep j close to the computed
value by setting j to (Numproc - 1) - j % Numproc instead of trying
j = j % Numproc. The reason being that the former promotes better locality]

Example:
a) dimsize = 32, Numproc = 4.
   p = 8, rem = 0, q = 8.
   Each processor gets 8 elements.
b) dimsize = 33, Numproc = 4.
   p = 8, rem = 1, q = 8.
   Each processor gets 8 except one processor which gets 9.
c) dimsize = 99, Numproc = 50
   p = 1, rem = 49, q = 2.
   Each processor gets 2 except one processor which gets 1.

However, for example, when dimsize = 8 and Numproc = 6, each of four 
processors will get 2 and two processors will get nothing.   So don't assume 
that if n > p every processor will get something.   --- Kate.

****************************************************************************/
      p = dim1size/pcxx_TotalNodes();
      rem = dim1size - pcxx_TotalNodes() * p;
      q = (p <= rem) ? p+1 : p;
      proc = i/q;
      if (proc >= pcxx_TotalNodes()) proc = (pcxx_TotalNodes() - 1) - proc % pcxx_TotalNodes();
      break;
    case CYCLIC:
      proc = (i % pcxx_TotalNodes());
      break;
    case WHOLE:
      pcxx_Message("GiveElementProc()_1D: WHOLE has no meaning in 1D Distribution,",
	      TULIP_WARNING);
      pcxx_Message("GiveElementProc()_1D: use BLOCK distribution instead.",
	      TULIP_WARNING);
      p = dim1size/pcxx_TotalNodes();
      rem = dim1size - pcxx_TotalNodes() * p;
      q = (p <= rem) ? p+1 : p;
      proc = i/q;
      if (proc >= pcxx_TotalNodes()) proc = (pcxx_TotalNodes() - 1) - proc % pcxx_TotalNodes();
      break;
    default:
      pcxx_Message("GiveElementProc()_1D: Unknow distribution for 1D Distribution.", 
	      PCXX_FATAL);
    }

  if (proc >= pcxx_TotalNodes()) {
    pcxx_Message("GiveElementProc()_1D: Error in 1D Distribution object.",TULIP_WARNING);
    proc = proc %pcxx_TotalNodes();
  }

  return proc;
}

/*********************************************************************/
int Distribution::GiveElementProc(int i, int j) {
  int proc = 0;
  int index1,index2, sq, p, rem, q;
  float x;

  if ((dist1 == QBLOCK) && (dist2 == QBLOCK)) {
    q = (int) QGridToProc(i,j,dim1size);
    x = dim1size*dim2size/ (float) pcxx_TotalNodes();
    proc = (int) (q/x);
    if (proc >= pcxx_TotalNodes()) {
      pcxx_Message("GiveElementProc()_2D: Processor # Out of Bounds)", 
	      TULIP_WARNING);
      proc = proc %pcxx_TotalNodes();
    }
    return proc;
  }
  
  sq = (int) sqrt((double)  pcxx_TotalNodes());
  if((dist1 == WHOLE) && (dist2 != WHOLE)) sq = pcxx_TotalNodes(); 
  if((dist1 != WHOLE) && (dist2 == WHOLE)) sq = pcxx_TotalNodes(); 
  if((dist1 == WHOLE) && (dist2 == WHOLE)) sq = 1;

  switch (dist1) {
    case BLOCK:
      p = dim1size/sq;
      rem = dim1size - sq * p;
      q = (p <= rem) ? p+1 : p;
      index1 = i/q;
      if (index1 >= sq) index1 = (sq - 1) - index1 % sq;
      break;
    case CYCLIC:
      index1 = (i % sq);
      break;
    case WHOLE:
      index1 = 0;
      break;
    default:
      pcxx_Message("GiveElementProc()_2D: Unknow distribution for 2D Distribution.", 
	      PCXX_FATAL);
    }
  switch (dist2) {
    case BLOCK:
      p = dim2size/sq;
      rem = dim2size - sq * p;
      q = (p <= rem) ? p+1 : p;
      index2 = j/q;
      if (index2 >= sq) index2 = (sq - 1) - index2 % sq;
      break;
    case CYCLIC:
      index2 = (j % sq);
      break;
    case WHOLE:
      index2 = 0;
      break;
    default:
      pcxx_Message("GiveElementProc()_2D: Unknow distribution for 2D Distribution.",
	      PCXX_FATAL);
    }
  
  if (dist1==WHOLE)
    proc = index2;
  else proc = index1 + index2*(sq);

  if (proc >= pcxx_TotalNodes()) {
    pcxx_Message("GiveElementProc()_2D: Error in 2D Distribution object.", TULIP_WARNING);
    proc = proc %pcxx_TotalNodes();
  }
  
  return proc;
}

/*********************************************************************/
int Distribution::GiveElementProc(int, int, int) {
  pcxx_Message("GiveElementProc()_3D: Sorry, 3D template not implemented.",PCXX_FATAL);
  return 0;
}

class Align {
  public:
    int nbelement;
    int nbdim;
    int dim1size,dim2size,dim3size;
    int ret_index1;
    int ret_index2;
    int ret_index3;
    int chosen;
    Align(int, int, char *) {
      /* fool the compiler, the real one is the next */
    };
    /* the real one */
    Align(int n1, int n2, int choix) {
      nbdim = 2;
      dim1size = n1;
      dim2size = n2;
      nbelement = n1*n2;
      chosen = choix;
    };
    Align(int n1, int choix) {
      nbdim = 1;
      dim1size = n1;
      dim2size = 0;
      nbelement = n1;
      chosen = choix;
    };
    void align_func(int var_i, int var_j); /* avoid shadow */
};

/*********************************************************************/
void Align::align_func(int varZZ_i, int varZZ_j)
{
  /* Set ret_index1 and ret_index2 depending on dep2C+
     generated align function (PHB) */
  GenAlignFunc(chosen,varZZ_i,varZZ_j,&ret_index1,&ret_index2);
}

/*********************************************************************/
/*********************************************************************/
/*********************** The Almighty Kernel! ************************/
/*********************************************************************/
/*********************************************************************/
class Kernel {
  public:
    Distribution *distribution;
    Align * align;
    struct INFO_HEAD CollectionInfo;
    tulip_GlobalPointer_t *ElementList;
#ifdef AWEDUDE
    void *IterBarPtr;
    tulip_GlobalPointer_t **IterList;
#endif
    int mycontextnum;   // WARNING! SPMD!! Different value on each context
    int CollectionNumber;
    int CollectionNumber_Type; 
    char *buffer;



    Kernel(Distribution *T, Align *A, int sizelem);
    Kernel(Kernel *col);
    int Install_Collection(Distribution *T, Align *A, int sizeelem);
    ~Kernel();

    int Is_Local(int index); 
    int Is_Local(int index1, int index2); 
    int ElemIndex(int i);   
    int ElemIndex(int i, int j);   
    int Get_2DHighIndex(int index); 
    int Get_2DLowIndex(int index); 
    int AllocatedProc(int index);
    int AllocatedProc(int index1, int index2);
    int *Get_Element(int index);
    int *Get_Element(int index1, int index2);
    int *Get_ElementPart(int index, int offset, int size);
    int *Get_ElementPart(int index1, int index2, int offset, int size);
    int *Get_CopyElem(int index);
    int *Get_CopyElem(int index1, int index2);
    void ElementCopy(int elemIndex, char *localbuffer);

    /* Architecture Dependant (kernel-SMRTS.h & kernel-DMRTS.h) */
    tulip_GlobalPointer_t ElementReference(int index);
    void CopyBytes(char *localbuffer, tulip_GlobalPointer_t GPsrc, int len);
};

/*********************************************************************/
// Darryl's weird include file for breezy breakpoint debugging
#include "breezyInc.h"

/*********************************************************************/
Kernel::Kernel(Distribution *T, Align *A, int elem_size) 
{
  if (!T || !A)
    pcxx_Message("Kernel(): No alignment or template provided.",PCXX_FATAL);
  distribution = T;
  align = A;
  if(A->nbelement > T->nbelement) {
    pcxx_Message("Kernel(): Elements in Align are more than elements in Distribution.",
	    TULIP_WARNING);
  }
#ifdef AWEDUDE
  IterBarPtr = NULL;
#endif  
  CollectionInfo.type = -1; /* undefined */
  CollectionInfo.number_dimension = A->nbdim; /* the default */
  CollectionInfo.element_size = elem_size;
  CollectionInfo.nb_element = A->nbelement;
  CollectionInfo.dim1size  = A->dim1size;
  CollectionInfo.dim2size  = A->dim2size;
  CollectionInfo.dim3size  = A->dim3size;
  if (!Install_Collection(T, A, elem_size))
    pcxx_Message("Kernel(): Error in creating collection.",PCXX_FATAL);
  pcxx_Barrier();
}

/*********************************************************************/
Kernel::Kernel(Kernel *col)
{
  Distribution *T;
  Align *A;
  int sizelem;
  int i;

  if (!col)
    pcxx_Message("Kernel(): Collection not provided.",PCXX_FATAL);
  T = col->distribution;
  A = col->align;
  sizelem = col->CollectionInfo.element_size;

  if (!T || !A)
    pcxx_Message("Kernel(): No alignment or template provided.",PCXX_FATAL);
  distribution = T;
  align= A;

  CollectionInfo.type             = -1; /* undefined */
  CollectionInfo.number_dimension = A->nbdim; /* the default */
  CollectionInfo.element_size     = sizelem;
  CollectionInfo.nb_element       = A->nbelement;
  CollectionInfo.dim1size         = A->dim1size;
  CollectionInfo.dim2size         = A->dim2size;
  CollectionInfo.dim3size         = A->dim3size;
  if (!Install_Collection(T,A,sizelem))
    pcxx_Message("Kernel(): Error in creating collection.",PCXX_FATAL);

  for (i=0; i < CollectionInfo.nb_element; i++) {

  /* I'm going to check first both are local... (PHB) */
  if ( (tulip_ToContext(&(     ElementList[i])) ==      mycontextnum) &&
       (tulip_ToContext(&(col->ElementList[i])) == col->mycontextnum))
    /* Check for valid addresses */
    if ( (tulip_ToLocalPointer(&(     ElementList[i])) == 0) ||
         (tulip_ToLocalPointer(&(col->ElementList[i])) == 0) )
      pcxx_Message("Kernel(): Error in copying data.",PCXX_FATAL);
    else
      memcpy((char *) tulip_ToLocalPointer(&(     ElementList[i])),
             (char *) tulip_ToLocalPointer(&(col->ElementList[i])), 
	     sizelem);
  }
  pcxx_Barrier();
}

/*********************************************************************/
int Kernel::Install_Collection(Distribution *, Align *A, int elem_size)
{
  int i, j;

  /* initialize the number of collection */
  CollectionNumber = GetNumberOfCollection();
  IncNumberOfCollection();

  /* Record a whole bunch of events for the creation of this collection */
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_INSTL_COLL, CollectionNumber);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_1, CollectionInfo.number_dimension);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_2, CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_3, CollectionInfo.nb_element);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_4, CollectionInfo.dim1size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_5, CollectionInfo.dim2size);

  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_create_collection,aa_poll_count);
  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_create_collection); 
  /* sets Ariadne break point if needed */

  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_begin_create_collection, aa_poll_count);
  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_begin_create_collection); 
  /* sets Ariadne break point if needed */
#ifdef SNEEZY
  snz_begin_create_collection_event(pcxx_NodeNumber());
#endif

  PROFILE_COLL_INIT(A->nbelement, CollectionInfo.number_dimension, elem_size);

  // This fills in the ContextNumber for this collection.
  // For MASTER_SLAVE on a shared memory, this is always 0
  mycontextnum = pcxx_NodeNumber(); /* Really, Context number */

  /* Allocate vector of elements */
  ElementList = (tulip_GlobalPointer_t *) 
    tulip_SharedMalloc(sizeof(tulip_GlobalPointer_t)* (A->nbelement));
#ifdef AWEDUDE
  IterList = (tulip_GlobalPointer_t **) 
    tulip_SharedMalloc(sizeof(tulip_GlobalPointer_t *)* (A->nbelement));
  for (i=0; i<A->nbelement; i++)
    IterList[i] = NULL;
#endif
  CollDbase[CollectionNumber].cinfo = &CollectionInfo;
  CollDbase[CollectionNumber].TableOfElements = 
    (tulip_GlobalPointer_t*) ElementList;

  /* Allocate space for a copy from a remote node */
  buffer = (char *) new char[elem_size+16];

  if (A->nbelement == 0)
    pcxx_Message("Install_Collection(): No element in collection.",PCXX_FATAL);
  
  // Fill Global Info Table...
  // Mark each element pointer, so we know if we should allocate for it
  // The actual allocation happens in parallel, in SuperKernel.

  if (CollectionInfo.number_dimension == 1) {
    // Cycle through each element
    for (i=0; i < CollectionInfo.dim1size; i++) {
      // Is this element in this context?
      if (AllocatedProc(i) == mycontextnum) {
        //SuperKernel will allocate the memory
	tulip_SetContext(&(ElementList[ElemIndex(i)]),mycontextnum);
	PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, ElemIndex(i));
      } else {
	tulip_SetContext(&(ElementList[ElemIndex(i)]),UNDEFINED_NODE);
	tulip_SetLocalPointer(&(ElementList[ElemIndex(i)]),0);
      }
    }
  } 
  
  if (CollectionInfo.number_dimension == 2) {
    // Cycle through each element
    for (i=0; i < CollectionInfo.dim1size; i++) {
      for (j=0; j < CollectionInfo.dim2size; j++) {
	// Is this element in this context?
	if (AllocatedProc(i, j) == mycontextnum) {
          //SuperKernel will allocate the memory
          tulip_SetContext(&(ElementList[ElemIndex(i,j)]),mycontextnum);
	  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, ElemIndex(i,j));
	} else {
          tulip_SetContext(&(ElementList[ElemIndex(i,j)]),UNDEFINED_NODE);
	  tulip_SetLocalPointer(&(ElementList[ElemIndex(i,j)]),0);
	}
      }
    }
  }
 
  if (CollectionInfo.number_dimension > 2) {
    pcxx_Message("Install_Collection(): 3D implementation not available.",
		 PCXX_FATAL);
  }
  
  /* add collection entry for breakpointing */
  brk_ADD_COLLECTION_ENTRY(A->nbelement, elem_size, this);
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_end_create_collection, aa_poll_count);

  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_end_create_collection); 
  /* sets Ariadne break point if needed */ 
#ifdef SNEEZY
  snz_end_create_collection_event(pcxx_NodeNumber());
#endif
  return 1; 
}

/*********************************************************************/
void Kernel::~Kernel()
{
  int i;

/****** begin of Ariadne tracing */
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_delete_collection, aa_poll_count);

  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_delete_collection); 
  /* sets Ariadne break point if needed */
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_begin_delete_collection, aa_poll_count);

  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_begin_delete_collection); 
  /* sets Ariadne break point if needed */
#ifdef SNEEZY
  snz_begin_delete_collection_event(pcxx_NodeNumber());
#endif
/***** end of Ariadne tracing */
  delete [] buffer;


  // cycle through all elements
  for (i=0; i < CollectionInfo.nb_element; i++) {
    // Do I own this element?  Then delete it
    if (tulip_ToContext(&ElementList[i]) == mycontextnum) {

      delete [] (tulip_ToLocalPointer(&ElementList[i]));
    }
  }

  /* Delete vector of element pointers */
  tulip_SharedFree(ElementList);

  /* Mark them as empty (why is this needed) */
  CollDbase[CollectionNumber].cinfo           = (struct INFO_HEAD *) 0;
  CollDbase[CollectionNumber].TableOfElements = (tulip_GlobalPointer_t *) 0;

/******** Ariadne tracing */
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_end_delete_collection, aa_poll_count);

  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_end_delete_collection); 
  /* sets Ariadne break point if needed */
#ifdef SNEEZY
  snz_end_delete_collection_event(pcxx_NodeNumber());
#endif
/********* end of Ariadne tracing */
}

/*********************************************************************/
// The name of this function is misleading.  It is not about locality,
// the memory heirarchy sense, but about whether the calling compute
// resource is responsible for computing on this resource.
#ifdef AWEDUDE
int Kernel::Is_Local(int index)
{
  return 1;
}
int Kernel::Is_Local(int index1, int index2)
{
  return 1;
}
#else //#ifdef AWEDUDE
int Kernel::Is_Local(int index)
{
#ifdef _MASTER_SLAVE_
  if (!MShack_InParallelSection()) 
    return 1;
  else
    return MShack_IamtheOwner(index);
#else
  if (tulip_ToContext(&ElementList[ElemIndex(index)]) == mycontextnum) 
    return 1;
  else
    return 0; 
#endif
}

/*********************************************************************/
int Kernel::Is_Local(int index1, int index2)
{
#ifdef _MASTER_SLAVE_
  if (!MShack_InParallelSection())
    return 1;
  else
    return MShack_IamtheOwner(ElemIndex(index1,index2));
#else
  if (tulip_ToContext(&ElementList[ElemIndex(index1,index2)]) == mycontextnum)
    return 1;
  else
    return 0;
#endif
}
#endif //#ifdef AWEDUDE
/*********************************************************************/
int Kernel::ElemIndex(int index)
{
  return index; 
}

/*********************************************************************/
int Kernel::ElemIndex(int index1, int index2)
{
  return (index1+index2*CollectionInfo.dim1size); 
}

/*********************************************************************/
int Kernel::Get_2DHighIndex(int index)
{
  return (index/CollectionInfo.dim1size);
}

/*********************************************************************/
int Kernel::Get_2DLowIndex(int index)
{
  return (index%CollectionInfo.dim1size);
}

/*********************************************************************/
int  Kernel::AllocatedProc(int index)
{
  // Assign the indexth element to a processor according to the
  // given template and alignment 
  int proc = 0;
  align->align_func(index, 0);
  if(distribution->nbdim == 1)
    proc = distribution->GiveElementProc(align->ret_index1);
  else
    proc = distribution->GiveElementProc(align->ret_index1,align->ret_index2);
  return proc;
}

/*********************************************************************/
int Kernel::AllocatedProc(int index1, int index2)
{
  int proc = 0;
  align->align_func(index1, index2);
  proc = distribution->GiveElementProc(align->ret_index1, align->ret_index2);
  return proc;
}

/*********************************************************************/
/* The Get_Element functions are slightly odd.  
   They use the following logic:
    1) If the element is local, return a pointer to it
    2) Else, FIND the element, by calling ElementReference
       2a) If shared memory, return a pointer to the found element
       2b) Else, [distributed memory], make a shallow copy of the element
           and return a pointer to the copy */
/*********************************************************************/
int *Kernel::Get_Element(int index)
{
  tulip_GlobalPointer_t mygp;
  TRC_DECL_PARAM;

  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_Element(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS(CollectionInfo.element_size);

  if (Is_Local(index)) {
    PROFILE_LOCAL_ACS();
    return (int *) tulip_ToLocalPointer(&(ElementList[ElemIndex(index)]));
  }

  PROFILE_REMOTE_ACS();

  /* Get a global pointer to the remote element */
  mygp = ElementReference(ElemIndex(index));
  TRC_SET_PARAM(CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_START, trcparam);

#ifdef _SMRTS_
  /* Access is simple, just return the pointer */
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) tulip_ToLocalPointer(&mygp);
#else
  /* Access must be through a COPY */
  /* I could call ElementCopy, but that would be redundant at this point */
  CopyBytes((char *) this->buffer, mygp, CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) this->buffer;
#endif
}

/*********************************************************************/
int *Kernel::Get_Element(int index1, int index2)
{
  tulip_GlobalPointer_t mygp;
  TRC_DECL_PARAM;

  TRACE_2D_ACS(CollectionInfo.element_size);

  if (Is_Local(index1, index2)) {
    PROFILE_LOCAL_ACS();
    return (int *) 
      tulip_ToLocalPointer(&(ElementList[ElemIndex(index1,index2)]));
  }

  PROFILE_REMOTE_ACS();

  /* Get a global pointer to the remote element */
  mygp = ElementReference(ElemIndex(index1, index2));
  TRC_SET_PARAM(CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_START, trcparam);

#ifdef _SMRTS_
  /* Access is simple, just return the pointer */
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) tulip_ToLocalPointer(&mygp);
#else
  /* Access must be through a COPY */
  /* I could call ElementCopy, but that would be redundant at this point */
  CopyBytes((char *) this->buffer, mygp, CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) this->buffer;
#endif
}

/*********************************************************************/
int *Kernel::Get_ElementPart(int index, int offset, int size)
{
  tulip_GlobalPointer_t mygp;
  TRC_DECL_PARAM;

  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_ElementPart(Get_2DLowIndex(index), Get_2DHighIndex(index),
			   offset, size);
  }

  /* For backward compatibility, the input parameters are in sizeof INTS */
  size = size*sizeof(int);
  offset =offset*sizeof(int);

  TRACE_1D_ACS(size);

  if (Is_Local(index)) {
    PROFILE_LOCAL_ACS();
    return (int *) 
      ((char *) tulip_ToLocalPointer(&(ElementList[ElemIndex(index)])) 
       + offset);
  }

  PROFILE_REMOTE_ACS();

  /* Get a global pointer to the remote element */
  mygp = ElementReference(ElemIndex(index));
  TRC_SET_PARAM(size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_START, trcparam);

  /* Offset the local pointer by "offset" bytes */
  tulip_SetLocalPointer(&mygp,((char *) tulip_ToLocalPointer(&mygp) + offset));

#ifdef _SMRTS_
  /* Access is simple, just return the pointer */
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) tulip_ToLocalPointer(&mygp);
#else
  /* Access must be through a COPY */
  /* I could call ElementCopy, but that would be redundant at this point */
  CopyBytes((char *) this->buffer, mygp, size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) this->buffer;
#endif
}

/*********************************************************************/
int *Kernel::Get_ElementPart(int index1, int index2, int offset, int size)
{
  tulip_GlobalPointer_t mygp;
  TRC_DECL_PARAM;

  /* For backward compatibility, the input parameters are in sizeof INTS */
  size = size*sizeof(int);
  offset =offset*sizeof(int);

  TRACE_2D_ACS(size);

  if (Is_Local(index1, index2)) {
    PROFILE_LOCAL_ACS();
    return (int *) 
      ((char *) tulip_ToLocalPointer(&(ElementList[ElemIndex(index1, index2)]))
       + offset);
  }

  PROFILE_REMOTE_ACS();

  /* Get a global pointer to the remote element */
  mygp = ElementReference(ElemIndex(index1, index2));
  TRC_SET_PARAM(size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_START, trcparam);

  /* Offset the local pointer by "offset" bytes */
  tulip_SetLocalPointer(&mygp,((char *) tulip_ToLocalPointer(&mygp) + offset));

#ifdef _SMRTS_
  /* Access is simple, just return the pointer */
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) tulip_ToLocalPointer(&mygp);
#else
  /* Access must be through a COPY */
  /* I could call ElementCopy, but that would be redundant at this point */
  CopyBytes((char *) this->buffer, mygp, size);
  PCXX_EVENT(PCXX_EC_COLLACS, PCXX_COLLACS_END, trcparam);
  return (int *) this->buffer;
#endif
}

/*********************************************************************/
/* Another Wierd set of functions, included ONLY for backward
compatibility, but these functions really don't make much sense.
Function logic: If local, return reference, otherwise, "new" a new
element */
/*********************************************************************/
int *Kernel::Get_CopyElem(int index)
{
  int *NBuf;

  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_CopyElem(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }

  TRACE_1D_ACS(CollectionInfo.element_size);
  if (Is_Local(index)) {
    PROFILE_LOCAL_ACS();
    return (int *) 
      (tulip_ToLocalPointer(&(ElementList[ElemIndex(index)])));
  }
  
  PROFILE_REMOTE_ACS();

  /* Allocate space for element */
  NBuf = (int *) new char[CollectionInfo.element_size+16];
  ElementCopy(ElemIndex(index), (char *) NBuf);
  return NBuf;
}

/*********************************************************************/
int *Kernel::Get_CopyElem(int index1, int index2)
{
  int *NBuf;

  TRACE_2D_ACS(CollectionInfo.element_size);
  if (Is_Local(index1, index2)) {
    PROFILE_LOCAL_ACS();
    return (int *) 
      (tulip_ToLocalPointer(&(ElementList[ElemIndex(index1,index2)])));
  }

  PROFILE_REMOTE_ACS();

  /* Allocate space for element */
  NBuf = (int *) new char[CollectionInfo.element_size+16];
  ElementCopy(ElemIndex(index1,index2), (char *) NBuf);
  return NBuf;
}

/*********************************************************************/
/* Strict element copy.  Put (shallow) copy of element into "buffer",
without regard for locality.  Routine does NOT call constructors, but
simply copies bytes.  This routine uses ElementReference to FIND the
element in question, then CopyBytes to fetch the data. */
/*********************************************************************/
void Kernel::ElementCopy(int elemIndex, char *localbuffer)
{ 
  /* elemIndex   = 1 dimensional ID for element */
  /* localbuffer = global pointer to source (remote) element */

  tulip_GlobalPointer_t mygp;

  /* Get a global pointer to the remote element */
  mygp = ElementReference(elemIndex);
  
  CopyBytes((char *) localbuffer, mygp, CollectionInfo.element_size);
  return;
}

/*********************************************************************/
/********************** non-class functions **************************/
/*********************************************************************/

int GetNumberOfCollection()
{
  return NumberOfCollection;
}

void IncNumberOfCollection()
{
  if (NumberOfCollection >= MAXCOLLECTION)
    pcxx_Message("Too many collections.",PCXX_FATAL);
  NumberOfCollection++;
}

void SetNextCollectionTraversalLevel(char * p)
{
  TraversalLevel++;
  if (TraversalLevel >= MAXPARAL)
    pcxx_Message("ResetLocal(): Too much parallelism.",PCXX_FATAL);
  CurrentCollectionTraversal[TraversalLevel] = p;
}

char *GetCollectionTraversalLevel ()
{
  return CurrentCollectionTraversal[TraversalLevel];
}

void DecCollectionTraversalLevel ()
{
  TraversalLevel--;
}

void ResetCollectionTraversalLevel ()
{
  TraversalLevel = -1;
}

int GetFirstLocalElementIndex(Kernel *p)
{
  Kernel *pt;
  int i;
  pt = (Kernel *) p;

#ifdef _MASTER_SLAVE_
// To the main thread, the first local element is always 0
  if (!MShack_InParallelSection()) return 0;

  for (i = 0; i < pt->CollectionInfo.nb_element; i++) {
    if (MShack_IamtheOwner(i)) return i;
  }
  return -1;
#else
  for (i = 0; i < pt->CollectionInfo.nb_element; i++) {
    if (tulip_ToContext(&(pt->ElementList[i])) == pt->mycontextnum) return i;
  }   
  return -1;
#endif
}

// Definition of Local,FirstLocal and ResetLocal is in kernel-AWE.h
void ResetLocal(Kernel *p)
{
#ifndef AWEDUDE  
  p->CollectionInfo.LocalTraversalCount = 0;
  SetNextCollectionTraversalLevel((char *) p);
#endif
}

#ifndef AWEDUDE  
int Local(int elem)
{

  Kernel *pt;
  int i;

  pt = (Kernel *) GetCollectionTraversalLevel();

  if (pt->CollectionInfo.LocalTraversalCount > pt->CollectionInfo.nb_element) {
#ifdef _MASTER_SLAVE_
  if(!MShack_InParallelSection())
#endif
    DecCollectionTraversalLevel();
    return -1;
  }
  
#ifdef _MASTER_SLAVE_
  for (i = elem+1; i < pt->CollectionInfo.nb_element; i++) {
    if (MShack_IamtheOwner(i)) {
#else
  for (i = pt->CollectionInfo.LocalTraversalCount; 
       i < pt->CollectionInfo.nb_element;
       i++) {
    if (tulip_ToContext(&(pt->ElementList[i])) == pt->mycontextnum) {
      pt->CollectionInfo.LocalTraversalCount = i+1;
#endif
      return i;
    }
  }   
#ifdef _MASTER_SLAVE_
  if(!MShack_InParallelSection())
#endif
  DecCollectionTraversalLevel();
  return -1;
}

// I think this is only used by MASTER_SLAVE anyway
int  FirstLocal ()
{
  return Local(-1);
}
#endif // of ifdef AWEDUDE


/* functions to deal with subsets for a collection with one dimension */

int FirstInSubset(Kernel *coll ,int lb,int ub,int)
{
  if (!coll)
    return -1;
  if (lb == -1)
    return 0;
  if (lb <= ub)
    return lb;
  return -1;
}

int NextInSubset(Kernel *coll ,int,int ub,int step, int var)
{
  int bsup;

  if (!coll)
    return -1;

  if (ub == -1)
    bsup = coll->CollectionInfo.dim1size;
  else 
    bsup = ub;

  if (var + step <= bsup)
    return var + step;
  else
    return -1;
}

CheckLocal(Kernel *coll, int index)
{
  if (!coll)
    return 0;
  
  return coll->Is_Local(index);
}

/********************BEGIN MACHINE DEPENDANT SECTION*********************** 
   Shared memory runtime system is used by:
   BBN TC2000, Sequent Symmetry, KSR1, SGIMP.
   Distributed memory runtime system is used by:
   cm5, paragon, sp1, pvm3.3, uniprocessor (regarded as 1-node distributed 
   systems).
 ***************************************************************************/

#if defined(_SMRTS_) || defined(_MASTER_SLAVE_)
#  include "kernel-SMRTS.h"
#else
#  include "kernel-DMRTS.h"
#endif

#ifdef _MASTER_SLAVE_
#include "kernel-MS.h"
#endif

/***************** END MACHINE DEPENDANT FUNCTIONS**********************/

/* *************************************************************************
 * Functions to initialize GlobalPtrTable.  There is one copy of the
 * table on every processor.  GlobalPtrTable[i] Local Pointer points
 * to CollDBase on processor i.  GlobalPtrTable[i] Context is equal to
 * i.
 * ************************************************************************/

/* HANDLER!  This handler is involked when a RemoteAction request is
 * sent to this processor.  It initializes GlobalPtrTable[from_whom]
 * */

extern "C" {
  void RemoteFunctionList(void *element_address,int function_number,
			  void *arg_buffer);
/* hacked code. if user does not define RemoteFunctionList() define it. */
#ifndef pcxx_RemoteFunctionList
  void RemoteFunctionList(void *,int, void *) {};
#endif

/*********************************************************************/
  void pcxx_RemoteActionHandler(int from_whom, char *buffer, int) {
    tulip_GlobalPointer_t *t;
    char* arg_buffer;
    long CollectionNumber, elem_index, function_number;

    switch (((long*)buffer)[0]) {
    case PCXX_BOOTSTRAP:
      t = (tulip_GlobalPointer_t *) (buffer + sizeof(long));
      GlobalPtrTable[from_whom]= *t;
      break;
    case PCXX_RPC:
      CollectionNumber = ((long*)buffer)[1];
      elem_index = ((long*)buffer)[2];
      function_number = ((long*)buffer)[3];
      arg_buffer = buffer+4*sizeof(long);
      RemoteFunctionList( (void *)
        (tulip_ToLocalPointer(
            &(CollDbase[CollectionNumber].TableOfElements)[elem_index])),
       	(int) function_number, (void *) arg_buffer);
      break;
    default:
      pcxx_Message("pcxx_RemoteActionHandler: Wrong RPC switch.",PCXX_FATAL);
    }
  }
}  /* End of extern "C" */


/**************************************************************************
   A kernel collection, SuperKernel, from which all new collections should 
   be derived.  It can also be used directly since basic element access and
   basic element management are provided.  Based on Francios Bodin's code
   in DistributedArray with modifications.

   Shelby Yang, Computer Science, Indiana University, 6/22/93.
***************************************************************************/
Collection SuperKernel : public Kernel {
 public:
  int dim1size;
  int dim2size;
  int dimensions;

  SuperKernel(Distribution *T, Align *A);
  ElementType  *operator()(int i1);           //pointer to a collection element
  ElementType  *operator()(int i1, int i2);   //pointer to a collection element

  void BroadcastBufferToField(int proc_id, char *buffer,int offset,int size);
  void BroadcastField(int index,int offset,int size);
  void BroadcastField(int index1,int index2,int offset,int size);

  void ReduceDoubleFieldAdd(int offset,int elem_count);
  void ReduceFloatFieldAdd(int offset,int elem_count);
  void ReduceIntFieldAdd(int offset,int elem_count);

  void ReduceDoubleFieldMax(int offset,int elem_count);
  void ReduceFloatFieldMax(int offset,int elem_count);
  void ReduceIntFieldMax(int offset,int elem_count);

  void ReduceDoubleFieldMin(int offset,int elem_count);
  void ReduceFloatFieldMin(int offset,int elem_count);
  void ReduceIntFieldMin(int offset,int elem_count);

  void RemoteExecute(int index,int function_number,char *arg_buffer,int size,
		     int *tag_location);
  void RemoteExecute(int index,int function_number,char *arg_buffer,int size);
  void RemoteExecute(int index1,int index2,int function_number,
		     char *arg_buffer,int size,int *tag_location);
  void RemoteExecute(int index1,int index2,int function_number,
		     char *arg_buffer,int size);
 MethodOfElement:
  int Ident;
  int index1;
  int index2;
  SuperKernel<ElementType> *ThisCollection;
  ElementType *Self(int i);        //copy an element with an offset i 
  ElementType *Self(int i, int j); //copy an element with an offset i,j

#ifdef _SMRTS_
  void * operator new(size_t size) {
    return (void *)tulip_SharedMalloc(size);
  }
  void operator delete(void *p, size_t ) {
    tulip_SharedFree(p);
  }
#endif
};

/*********************************************************************/
SuperKernel::SuperKernel(Distribution *T, Align *A)
                        :Kernel(T, A, (int)sizeof(ElementType))
{
  int i, j;

  if (!T || !A)
    pcxx_Message("No alignment or template for collection initialization.",
		 PCXX_FATAL);
  dim1size = A->dim1size;
  dim2size = A->dim2size;
  dimensions = A->nbdim;


  /* The **REAL** Allocating of elements */
  if (A->nbdim == 1) {
    for (i = 0;  i < dim1size; i++) {
      if (Is_Local(i)) {
        tulip_SetLocalPointer(&(ElementList[ElemIndex(i)]),
			      (void *) new ElementType);
        tulip_SetContext(&(ElementList[ElemIndex(i)]),mycontextnum);
        (*this)(i)->index1 = i;
        (*this)(i)->index2 = 0;
        (*this)(i)->ThisCollection = (SuperKernel<ElementType> *) this;
        (*this)(i)->Ident = this->CollectionNumber*16777216+i;
      } 
    }
  }

  if (A->nbdim == 2) {
    for (i = 0;  i < dim1size; i++) {
      for (j = 0;  j < dim2size; j++) {
        if (Is_Local(i,j)) {
	  tulip_SetLocalPointer(&(ElementList[ElemIndex(i,j)]), 
				(void *) new ElementType);
	  tulip_SetContext(&(ElementList[ElemIndex(i,j)]),mycontextnum);
          (*this)(i,j)->index1 = i;
          (*this)(i,j)->index2 = j;
          (*this)(i,j)->ThisCollection = (SuperKernel<ElementType> *) this;
          (*this)(i,j)->Ident = this->CollectionNumber*16777216+i+j*this->dim1size;
        } 
      }
    }
  }

  /* OK, everyone has allocated all the elements! */
  pcxx_Barrier();

  if (A->nbdim >= 3)
    pcxx_Message("3D and higher-dimensional collection not implemented.",
		 PCXX_FATAL);

  /* Now, we can allow users to access data at breakpoints as data has
     been initialized. */
  brk_TOGGLE_BREAKPOINT(1);
}

/*********************************************************************/
ElementType * SuperKernel::operator()(int i1)
{
  ElementType  *pt;
  pt = (ElementType  *) Get_Element(i1);
  return pt;
}

/*********************************************************************/
ElementType * SuperKernel::operator()(int i1, int i2)
{
  ElementType  *pt;
  pt = (ElementType  *) Get_Element(i1,i2);
  return pt;
}


;
/*********************************************************************/
ElementType * SuperKernel::Self(int i)
{
  return (ElementType  *) ThisCollection->Get_Element(index1 + i);
}

/*********************************************************************/
ElementType * SuperKernel::Self(int i, int j)
{
  return (ElementType  *) ThisCollection->Get_Element(index1 + i, index2 + j);
};


/*********************************************************************
  The arguments for the remote function should be copied into the
  arg_buffer. index1 and index2 are the indices of the element. 
  function_number is the number assigned to the function that will be 
  executed. size is the size of arg_buffer in bytes.
***********************************************************************/
void SuperKernel::RemoteExecute(int index,int function_number,
                                char *arg_buffer,int size)
{
  int *p, proc_id;
  int elem_index = ElemIndex(index);
  tulip_ACK_t *ackhandle;

  if(Is_Local(index)) {
    RemoteFunctionList(
          (void*)(tulip_ToContext(&(ElementList[elem_index]))), 
          function_number, (void*)arg_buffer);
    return;
  }

  if ( tulip_ToContext(
	&(CollDbase[CollectionNumber].TableOfElements)[elem_index]) < 0) {
    proc_id = AllocatedProc(index);
  } else {
    proc_id = tulip_ToContext(
      &((CollDbase[CollectionNumber].TableOfElements)[elem_index]));
  }

  p = (int*)buffer;
  p[0] = PCXX_RPC;  // code for remote function call
  p[1] = CollectionNumber;
  p[2] = elem_index;
  p[3] = function_number;
  memcpy(buffer+4*sizeof(int), arg_buffer, size);

  ackhandle = tulip_ACKCreate(1);
  pcxx_RemoteAction(proc_id,buffer,size+4*sizeof(int),ackhandle);

  tulip_WaitACK(ackhandle);
  tulip_ACKDestroy(ackhandle);

}

/*********************************************************************/
void SuperKernel::RemoteExecute(int index,int function_number,
                                char *arg_buffer,int size,int *tag_location)
{
  int *pt, *p, proc_id;
  int elem_index = ElemIndex(index);
  tulip_ACK_t *ackhandle;

  if (Is_Local(index)) {
    RemoteFunctionList(
          (void*)(tulip_ToContext(&(ElementList[elem_index]))), 
          function_number, (void*)arg_buffer);
    if(tag_location != 0) {
      pt = (int*)tag_location;
      (*pt)++;
    }
    return;
  }

  if (tulip_ToContext(
     &(CollDbase[CollectionNumber].TableOfElements)[elem_index]) < 0) {
    proc_id = AllocatedProc(index);
  } else {
    proc_id = tulip_ToContext(
      &((CollDbase[CollectionNumber].TableOfElements)[elem_index]));
  }

  p = (int*)buffer;
  p[0] = PCXX_RPC;  // code for remote function call
  p[1] = CollectionNumber;
  p[2] = elem_index;
  p[3] = function_number;
  memcpy(buffer+4*sizeof(int), arg_buffer, size);

  ackhandle = tulip_ACKCreate(1);
  pcxx_RemoteAction(proc_id,buffer,size+4*sizeof(int),ackhandle);

  tulip_WaitACK(ackhandle);
  tulip_ACKDestroy(ackhandle);

  if(tag_location != 0) {
    pt = (int*)tag_location;
    (*pt)++;
  }
}

/*********************************************************************/
void SuperKernel::RemoteExecute(int index1,int index2,int function_number,
                                char *arg_buffer,int size)
{
  int *p, proc_id;
  int elem_index = ElemIndex(index1,index2);
  tulip_ACK_t *ackhandle;

  if(Is_Local(index1,index2)) {
    RemoteFunctionList(
          (void*)(tulip_ToContext(&(ElementList[elem_index]))), 
          function_number, (void*)arg_buffer);
    return;
  }

  if ( tulip_ToContext(
     &(CollDbase[CollectionNumber].TableOfElements)[elem_index]) < 0) {
    proc_id = AllocatedProc(index1,index2);
  } else {
    proc_id = tulip_ToContext(
      &((CollDbase[CollectionNumber].TableOfElements)[elem_index]));
  }

  p = (int*)buffer;
  p[0] = PCXX_RPC;  // code for remote function call
  p[1] = CollectionNumber;
  p[2] = elem_index;
  p[3] = function_number;
  memcpy(buffer+4*sizeof(int), arg_buffer, size);

  ackhandle = tulip_ACKCreate(1);
  pcxx_RemoteAction(proc_id,buffer,size+4*sizeof(int),ackhandle);

  tulip_WaitACK(ackhandle);
  tulip_ACKDestroy(ackhandle);
}

/*********************************************************************/
void SuperKernel::RemoteExecute(int index1,int index2,int function_number,
                                char *arg_buffer,int size,int *tag_location)
{
  int *pt, *p, proc_id;
  int elem_index = ElemIndex(index1,index2);
  tulip_ACK_t *ackhandle;

  if(Is_Local(index1,index2)) {
    RemoteFunctionList(
	  (void*)(tulip_ToContext(&(ElementList[elem_index]))), 
          function_number, (void*)arg_buffer);
    if(tag_location != 0) {
      pt = (int*)tag_location;
      (*pt)++;
    }
    return;
  }

  if ( tulip_ToContext(
     &(CollDbase[CollectionNumber].TableOfElements)[elem_index]) < 0) {
    proc_id = AllocatedProc(index1,index2);
  } else {
    proc_id = tulip_ToContext(
      &((CollDbase[CollectionNumber].TableOfElements)[elem_index]));
  }

  p = (int*)buffer;
  p[0] = PCXX_RPC;  // code for remote function call
  p[1] = CollectionNumber;
  p[2] = elem_index;
  p[3] = function_number;
  memcpy(buffer+4*sizeof(int), arg_buffer, size);

  ackhandle = tulip_ACKCreate(1);
  pcxx_RemoteAction(proc_id,buffer,size+4*sizeof(int),ackhandle);

  tulip_WaitACK(ackhandle);
  tulip_ACKDestroy(ackhandle);

  if(tag_location != 0) {
    pt = (int*)tag_location;
    (*pt)++;
  }
}

/*********************************************************************/
void SuperKernel::BroadcastBufferToField(int proc_id, char *buffer,int offset,
int size)
{
  /************************************************************************ 
    This function broadcasts a buffer located on processor "proc_id" to the
    element class field which is offset by "offset" bytes from the beginning 
    of the element class.  "size" is the size of the buffer given in bytes.
   *************************************************************************/
  int i, j, k, flag;
  char *p;

  flag = 1;
  if(pcxx_NodeNumber() == proc_id) flag = 0;

  pcxx_BroadcastBytes(flag, size, buffer);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = ((char *)(*this)(i)) + offset;
	for(j = 0; j < size; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = ((char *)(*this)(i,j)) + offset;
	  for(k = 0; k < size; k++) p[k] = buffer[k];
	}
      }
    }
  }

}

/*********************************************************************/
void SuperKernel::BroadcastField(int index,int offset,int size)
{
  /************************************************************************ 
    This function broadcasts a field of an element to the same field of all 
    other elements.  "index" denotes the element that broadcasts.
    The field is offset from the beginning of the element by "offset" bytes.
    "size" is the size of the field in bytes.
   *************************************************************************/
  int i, j;
  char *p, *buffer;
  int flag = 1;

  buffer = new char[size];

  if(Is_Local(index)) {
    /* this processor will broadcast.  set flag to zero. */
    flag = 0;
    p = ((char *)(*this)(index)) + offset;
    for(j = 0; j < size; j++) buffer[j] = p[j]; 
  }
  pcxx_Barrier();

  pcxx_BroadcastBytes(flag, size, buffer);

  for(i = 0; i < dim1size; i++) {
    if(Is_Local(i) && i != index) {
      p = ((char *)(*this)(i)) + offset;
      for(j = 0; j < size; j++) p[j] = buffer[j];
    }
  } 
  
  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::BroadcastField(int index1,int index2,int offset,int size)
{
  /************************************************************************ 
    This function broadcasts a field of an element to the same field of all 
    other elements.  "index1" and "index2" denotes the element that broadcasts.
    The field is offset from the beginning of the element by "offset" bytes.
    "size" is the size of the field in bytes.
   *************************************************************************/
  int i, j, k;
  char *p, *buffer;
  int flag = 1;

  buffer = new char[size];

  if(Is_Local(index1,index2)) {
    /* this processor will broadcast.  set flag to zero. */
    flag = 0;
    p = ((char *)(*this)(index1,index2)) + offset;
    for(j = 0; j < size; j++) buffer[j] = p[j]; 
  }

  pcxx_BroadcastBytes(flag, size, buffer);

  for(i = 0; i < dim1size; i++) {
    for(j = 0; j < dim2size; j++) {
      if(Is_Local(i,j) && i!=index1 && j!=index2) {
	p = ((char *)(*this)(i,j)) + offset;
	for(k = 0; k < size; k++) p[k] = buffer[k];
      }
    }
  } 
  
  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceDoubleFieldAdd(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function computes the sum of the field
   from all elements in the collection.  The sum is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   summed together.  Note "offset" is in bytes and "elem_count" is the number 
   of array elements.
   ***************************************************************************/
  int i, j, k;
  double *p, *buffer;

  buffer = new double[elem_count];

  for(i = 0; i < elem_count; i++) buffer[i] = 0.0;
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] += p[j];
      }
    }
  } 

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] += p[k];
	}
      }
    }
  }


  /* reduce from and broadcast to processors */ 
  pcxx_ReduceDoubleAdd((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceFloatFieldAdd(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function computes the sum of the field
   from all elements in the collection.  The sum is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   summed together.  Note "offset" is in bytes and "elem_count" is the number 
   of array elements.
   ***************************************************************************/
  int i, j, k;
  float *p, *buffer;

  buffer = new float[elem_count];

  for(i = 0; i < elem_count; i++) buffer[i] = 0.0;
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] += p[j];
      }
    }
  } 

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] += p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceFloatAdd((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceIntFieldAdd(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function computes the sum of the field
   from all elements in the collection.  The sum is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   summed together.  Note "offset" is in bytes and "elem_count" is the number 
   of array elements.
   ***************************************************************************/
  int i, j, k;
  int *p, *buffer;

  buffer = new int[elem_count];

  for(i = 0; i < elem_count; i++) buffer[i] = 0;
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] += p[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] += p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceIntAdd((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  } 

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceIntFieldMin(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the minimum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  int *p, *buffer;

  buffer = new int[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] < buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] < buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceIntMin((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceFloatFieldMin(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the minimum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  float *p, *buffer;

  buffer = new float[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] < buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] < buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceFloatMin((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceDoubleFieldMin(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the minimum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  double *p, *buffer;

  buffer = new double[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] < buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] < buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceDoubleMin((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceIntFieldMax(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the maximum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  int *p, *buffer;

  buffer = new int[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] > buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
    if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] > buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceIntMax((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (int*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (int*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceFloatFieldMax(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the maximum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  float *p, *buffer;

  buffer = new float[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] > buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] > buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceFloatMax((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (float*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (float*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}

/*********************************************************************/
void SuperKernel::ReduceDoubleFieldMax(int offset,int elem_count)
{
  /**************************************************************************
   Suppose a field of an element class is offset from the beginning of the
   class by "offset" bytes.  This function finds the maximum of the field
   of all elements in the collection.  The result is broadcast back to each 
   element and stored in the same field.  The old content of the field is 
   overwritten.  If the field is an array, corresponding array elements are 
   compared with each other.  Note "offset" is in bytes and "elem_count" is 
   the number of array elements.
   ***************************************************************************/
  int i, j, k, first_i, first_j, found;
  double *p, *buffer;

  buffer = new double[elem_count];
 
  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	first_i = i;
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) buffer[j] = p[j];
	break;
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) 
	  if(p[j] > buffer[j]) buffer[j] = p[j];
      }
    }
  }

  found = 0;
  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      if(found == 1) break;
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  first_i = i;
	  first_j = j;
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) buffer[k] = p[k];
	  found = 1;
	  break;
	}
      }
    }
    for(i = first_i+1; i < dim1size; i++) {
      for(j = first_j+1; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) 
	    if(p[k] > buffer[k]) buffer[k] = p[k];
	}
      }
    }
  }

  /* reduce from and broadcast to processors */ 
  pcxx_ReduceDoubleMax((void*)buffer,elem_count);

  if(dimensions == 1) {
    for(i = 0; i < dim1size; i++) {
      if(Is_Local(i)) {
	p = (double*)(((char *)(*this)(i)) + offset);
	for(j = 0; j < elem_count; j++) p[j] = buffer[j];
      }
    }
  }

  if(dimensions == 2) {
    for(i = 0; i < dim1size; i++) {
      for(j = 0; j < dim2size; j++) {
	if(Is_Local(i,j)) {
	  p = (double*)(((char *)(*this)(i,j)) + offset);
	  for(k = 0; k < elem_count; k++) p[k] = buffer[k];
	}
      }
    }
  }

  delete [] buffer;
}


/*********************************************************************/
/*************************  Other Functions **************************/
/*********************************************************************/

/*********************************************************************/
/* This procedure initializes GlobalPtrTable on every processor.
GlobalPtrTable[i] contains a global pointer that provides the address
to CollDbase on processor i. GlobalPtrTable[i] is initialized through
RemoteAction call by processor i, if i is not equal this node's
Context */
/*********************************************************************/
void KernelBootStrap()
{
  tulip_GlobalPointer_t mygp;
  int i;
  char buffer[100];
  char *t;
  tulip_ACK_t *ackhandle_array;
  
  brk_TOGGLE_BREAKPOINT(0);

  ackhandle_array = tulip_ACKCreate(MAXCOLLECTION);

  GlobalPtrTable = (tulip_GlobalPointer_t *) 
    tulip_SharedMalloc(MAXCOLLECTION*sizeof(tulip_GlobalPointer_t));

  CollDbase = (C_Record  *) 
    tulip_SharedMalloc(MAXCOLLECTION*sizeof(C_Record));

  ResetCollectionTraversalLevel();
  
  pcxx_Barrier();

  /* Package up the Global Pointer for MY CollDbase */
  tulip_SetLocalPointer(&mygp,(void *) CollDbase);
  tulip_SetContext(&mygp,pcxx_NodeNumber());

  /* Create a RemoteAction */
  t = buffer;
  *((long *) t) = PCXX_BOOTSTRAP;     /* Code for boot */
  t = t + sizeof(long);
  *((tulip_GlobalPointer_t *) t) = mygp;
  
  /* invoke a RemoteAction on every remote node, handing it buffer */
  for (i=0; i<pcxx_TotalNodes(); i++)
    if (i != pcxx_NodeNumber())
      pcxx_RemoteAction(i,buffer,sizeof(long)+sizeof(tulip_GlobalPointer_t),
			&ackhandle_array[i]);
    else
      GlobalPtrTable[i] = mygp;
  
  for (i=0; i<pcxx_TotalNodes(); i++)
    if (i != pcxx_NodeNumber())
      tulip_WaitACK(&ackhandle_array[i]);

  tulip_ACKDestroy(ackhandle_array);

  pcxx_Barrier();
}

/*********************************************************************/
/* The "entry point" for the RTS.  */
/*********************************************************************/

extern "C" {
  int pcxx_ParallelMain(int argc, char **argv) {
    KernelBootStrap();
#ifdef _MASTER_SLAVE_
    thread_Init(&argc,argv);
#endif
 pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_begin_parallel, aa_poll_count);

 pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_begin_parallel); 
  /* sets Ariadne break point if needed */

#ifndef AWEDUDE
    Processor_Main(argc,argv);
#endif
    if (pcxx_NodeNumber() == 0) {
      brk_EXITFUNC();
    }
 pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
		    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_end_parallel, aa_poll_count);

 pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_end_parallel); 
  /* sets Ariadne break point if needed */

    return 1;
  }
}

#ifdef AWEDUDE
#include "kernel-AWE.h"
#endif

int pcxx_AriadneUserBreakPoint(int parameter1,int parameter2)
{ 
/* here we would log the parameter by putting it with the aa_poll_count  */
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
                    pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_user_event, aa_poll_count);
  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_user_event);
/* */
  return 1;
}

unsigned int pack_var(unsigned int var, int length, int starting_lsb_position);

int pcxx_AriadneRecordUserEvent(int  event_no, int no_of_variables, int var1, int var2, int var3)
{
/* This sets puts the value of the parameter in the trace for Ariadne */
/* This is used for logging user event in pc++ */
/* out of 32 bits for the integer parameter we use
    xx yyyy vv var1 var2 var3
    2   4    2  8    8    8
    xx - 2 bits unused - signed etc.
    yyyy - event no. goes from 0 to 15 - different types  of events- merge
                and record - values 0 1 ?
    vv - how many variables do you want to store.
    var1 - one variable,
    var2 - second variable - 8 bits.
    var3 - third variable - 8 bits.
*/
unsigned int parameter = 0;

#ifdef TRACING_ON
  parameter = pack_var(event_no,4, 26) | pack_var(no_of_variables,2, 24) | pack_var (var3, 8, 16) | pack_var(var2, 8, 8) | pack_var(var1, 8, 0);
#endif

 /* trace it and check while replaying the program */ 
  pcxx_AriadneTrace(PCXX_EC_ARIADNE, PCXX_ARIADNE_PCXX_EV, pcxx_NodeNumber(),
         pcxx_NodeNumber(), PCXX_AA_MTAG_USER, PCXX_user_event, parameter);
  pcxx_CheckForBreakPoint(PCXX_AA_MTAG_USER, PCXX_user_event);
  return 1;
}


unsigned int pack_var(unsigned int var, int length, int starting_lsb_position)
{
    /* make sure its of the right length */

    var = ((var << (32 - length)) >> (32 - length)); 
	/* its of length - won't interfere with other bits */
    /* now shift it starting at point position */
    /* for eg. if we want to put x (8 bits length at position lsb 16) */
    /* then 8........16  x goes in the dots */
    return (var << starting_lsb_position);
} 

