/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# ifndef PROFILE_ENTER
#   ifdef PROFILING_ON

#     ifndef MaxCollection
#       define MaxCollection 50
#     endif

typedef struct p_func_descr {
    int     numcalls;
    double  usec;
    double  cumusec;
} P_FUNC_DESCR;

typedef struct p_coll_descr {
    int numelem;
    int dim;
    int size;
    int localacs;
    int remoteacs;
    char *collname;
    char *elemname;
    char *varname;
} P_COLL_DESCR;

typedef struct p_cqueue {
    char *name;
    struct p_cqueue *next;
} P_CQUEUE;

#     ifndef PROF_INITIALIZED
#       if defined(__ksr__) && ! defined(UNIPROC)
          extern __private P_COLL_DESCR *p_coll_list[MaxCollection];
          extern __private P_FUNC_DESCR *p_func_list;
          extern __private int p_curr_func;
          extern __private char *p_collname, *p_elemname;
          extern __private P_CQUEUE *p_varname;
#       else
          extern P_COLL_DESCR *p_coll_list[MaxCollection];
          extern P_FUNC_DESCR *p_func_list;
          extern int p_curr_func;
          extern char *p_collname, *p_elemname;
          extern P_CQUEUE *p_varname;
#       endif

#       include <stdio.h>
#       include <stdlib.h>
#       include <fcntl.h>

        void PROFILE_DUMP(int numfunc) {
          int out;
          char buf[256];
          char profname[16];
          int numcoll, i;

          sprintf (profname, "profile.%d", pcxx_NodeNumber());
          if ( (out = creat (profname, 0644)) == -1 ) {
              perror (profname);
              exit (1);
          }
          numcoll = GetNumberOfCollection();
          sprintf (buf, "%d func\n", numfunc);
          write (out, buf, strlen(buf));
          for (i = 0; i<numfunc; i++) {
            if (p_func_list[i].usec > 0.0 )
              sprintf (buf, "%d %G %G\n", p_func_list[i].numcalls,
                            p_func_list[i].usec, p_func_list[i].cumusec);
            else
              sprintf (buf, "%d 0 0\n", p_func_list[i].numcalls);
            write (out, buf, strlen(buf));
          }
          sprintf (buf, "%d coll\n", numcoll);
          write (out, buf, strlen(buf));
          for (i = 0; i<numcoll; i++) {
            sprintf (buf, "%d %d %d %d %d %s %s %s\n",
                     p_coll_list[i]->numelem,
                     p_coll_list[i]->dim,
                     p_coll_list[i]->size,
                     p_coll_list[i]->localacs,
                     p_coll_list[i]->remoteacs,
                     p_coll_list[i]->collname,
                     p_coll_list[i]->elemname,
                     p_coll_list[i]->varname);
            write (out, buf, strlen(buf));
          }
          close (out);
        }
#     else
        extern "C" void PROFILE_DUMP(int numfunc);
#     endif

#     define PROFILE_ENTER(id)           \
        int prf_parent = p_curr_func;    \
        double prf_start = pcxx_GetUSecD(); \
        p_curr_func = id;

#     define PROFILE_EXIT(id)                           \
        do {                                            \
          if (p_func_list) {                            \
            double duration = pcxx_GetUSecD() - prf_start; \
            p_func_list[id].numcalls++;                 \
            p_func_list[id].usec    += duration;        \
            p_func_list[id].cumusec += duration;        \
            p_curr_func = prf_parent;                   \
            p_func_list[prf_parent].usec -= duration;   \
          }                                             \
        } while (0)

#     define PROFILE_COLL_INIT(n,d,s)                                        \
        do {                                                                 \
          p_coll_list[CollectionNumber] = (P_COLL_DESCR *) new P_COLL_DESCR; \
          p_coll_list[CollectionNumber]->numelem = n;                        \
          p_coll_list[CollectionNumber]->dim     = d;                        \
          p_coll_list[CollectionNumber]->size    = s;                        \
          p_coll_list[CollectionNumber]->localacs  = 0;                      \
          p_coll_list[CollectionNumber]->remoteacs = 0;                      \
          p_coll_list[CollectionNumber]->collname  = p_collname;             \
          p_coll_list[CollectionNumber]->elemname  = p_elemname;             \
          if ( p_varname->name ) {                                           \
            p_coll_list[CollectionNumber]->varname = p_varname->name;        \
            p_varname = p_varname->next;                                     \
          } else                                                             \
            p_coll_list[CollectionNumber]->varname = "???";                  \
        } while (0)

#     define PROFILE_LOCAL_ACS() \
        p_coll_list[CollectionNumber]->localacs++

#     define PROFILE_REMOTE_ACS() \
        p_coll_list[CollectionNumber]->remoteacs++

#     define PROFILE_Barrier 1
#     define PROFILE_Poll    2
#     define PROFILE_Where   3

#   else
#     define PROFILE_COLL_INIT(n,d,s)
#     define PROFILE_LOCAL_ACS()
#     define PROFILE_REMOTE_ACS()
#     define PROFILE_ENTER(id)
#     define PROFILE_EXIT(id)
#     define PROFILE_COLL_NAMES(c,e,v)
#   endif
# endif
