/*------------------------------------------------------------
 *  Darryl Brown
 *  
 *  pC++ tools research group, University of Oregon
 *  October, 1990
 *  
 *  transport.h
 *
 *  An implementation of a low level data transfer library using
 *  sockets, the include file.
 *
 *------------------------------------------------------------
 */

#ifndef TRNS_HDR
#define TRNS_HDR
#ifdef __cplusplus 
extern "C" {
#endif

/*------------------------------------------------------------
 *  Definitions
 */
/* largest message size */
#define MESSAGE_BUF_SIZE 2028
/* number of seconds before retrying connection */
#define WAIT_TIME 5
#define TRNS_BUFBLOCK 1024

/*------------------------------------------------------------
 *  Types
 */
typedef struct aMessageStruct {
  int length;
  void *data;
} aMessage;

/*
 *------------------------------------------------------------
 *  FUNCTIONS
 *------------------------------------------------------------
 */

/*------------------------------------------------------------
 *
 *  Int TRNS_OpenClient(int portNum, char *hostName)
 *    Opens a new data client connection using portNum to hostName, 
 *    returns an ID that can be used in other data transer functions.
 *
 */
int TRNS_OpenClient(int portNum, char *hostName);

/*------------------------------------------------------------
 *
 *  int TRNS_OpenServer(int portNum)
 *    Opens a new data server connection using portNum, returns 
 *    an ID that can be used in other data transer functions.
 *
 */
int TRNS_OpenServer(int portNum); 

/*------------------------------------------------------------
 *  void TRNS_CloseConnection(int ConnNum)
 *    Closes the connection designated by ConnNum.
 *
 */
void TRNS_CloseConnection(int connNum); 

/*------------------------------------------------------------
 *  msg *TRNS_Receive(int connNum, int dataLen) 
 *    Reads all data from connection specified by ConnNum. 
 *    
 */
aMessage *TRNS_Receive(int connNum, int dataLen); 

     
/* ------------------------------------------------------------
 *  int TRNS_Send(int connNum, aMessage *msg)
 *    Writes msg into connection specified by ConnNum
 *
 */
int TRNS_Send(register int connNum, aMessage *msg);

     
/* ------------------------------------------------------------
 *  void TRNS_RecycleMessage(aMessage *msg)
 *    Writes msg into connection specified by ConnNum
 *
 */
void TRNS_RecycleMessage(aMessage *msg);

/* ------------------------------------------------------------
 *  void *TRNS_addBufferBlock(void *buf, int bufLen, int copyLen)
 *    Increases buffer size, and copies copylen bytes from old buf
 *    to new buf.  Returns the new buffer pointer.
 *
 */
void *TRNS_addBufferBlock(void *buf, int bufLen, int copyLen);

/* end of file */

#ifdef __cplusplus 
}
#endif

#endif
