/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/***************************************************************************
 *                                                                         *
 *                                                                         *
 *                HPC++ Routine for managing access to data                *
 *                Sequential Set Of routine                                *
 *                                                                         *
 *                F. Bodin, Indiana University                             *
 *                                                                         *
 ***************************************************************************/
#include <stdarg.h>

#define MaxCollection 100

/* these numbers have to be the same as they are in sgilib.c, ksrlib.c, ... */
#define MAXBUFF 4096
#define MAX_PROC 128
extern char *ReduceBuff[MAX_PROC];

#include <stdio.h>
#include "GlobalStruct.h"

#ifndef __ksr__
#  define __private
#endif

#ifdef _MASTER_SLAVE_
#  define __private

extern int IamtheMainThread();
#endif

__private static int NumberOfCollection = 0;
__private static PTR_INFO TableOfCollection[MaxCollection];
__private static char *CurrentCollectionTraversal[MAXPARAL];
__private static int TraversalLevel = -1;
__private static int WarningCount =0;

/* Forward definition of activated function */
void GiveWhereDataIs();
void SendData();
void UpdateMyData();

/* Forward definition of local management */
int WhereIsData();
void Message();

/**************************************************************************
 ********************** General Routine ***********************************
 **************************************************************************/
#include "common.h" /* Common support routines for EVERY KernelCom.c (phb) */

/*************************************************************/
void Message(st, flag)
char *st;
int flag;
{
  if (flag == WARNING)
    {
      WarningCount++;
      fprintf(stderr, "Warning: %s\n",st);
      return;
    }
  if (flag == FATAL)
    {
      fprintf(stderr, "Fatal error: %s\n",st);
      return;
    }
}

int GetNumberOfCollection()
{
  return NumberOfCollection;
}

void IncNumberOfCollection()
{
  if (NumberOfCollection >= MaxCollection)
    Message("Too many collections.",FATAL);
  NumberOfCollection++;
}

void SetTableOfCollection (num, addr)
int num;
char *addr;
{
  TableOfCollection[num] = (PTR_INFO) addr;

}

void SetNextCollectionTraversalLevel (addr)
char *addr;
{
  TraversalLevel++;
  if (TraversalLevel >= MAXPARAL)
    Message("ResetLocal(): Too much parallelism.",FATAL);
  CurrentCollectionTraversal[TraversalLevel] = addr;
}

char *GetCollectionTraversalLevel ()
{
  return CurrentCollectionTraversal[TraversalLevel];
}

void DecCollectionTraversalLevel ()
{
#ifdef _MASTER_SLAVE_
  if(IamtheMainThread())
#endif
    TraversalLevel--;
}

void ResetCollectionTraversalLevel ()
{
  TraversalLevel = -1;
}

/**************************************************************************
 ********************** Kernel Request ************************************
 **************************************************************************/

int IamtheManager(id)
int id;
{
  Message("IamtheManager called in shared memory mode",FATAL);
  return 1;
}

int WhereIsData(id,info)
int id;
int info;
{
  static int manager = 0;
  return 0;
}

int AskForDataBlock(int where, int index, int offset,
                       int Coll_Id, int size, int *buffer)
{
  char *p;
  p = (char*) (TableOfCollection[Coll_Id]->Table[index]+offset);
  memcpy(buffer,p,size);
  return 0;
}

#ifndef TASK_MS

main(argc, argv)
int argc;
char **argv;
{
  pcxx_InitParam (&argc, argv);
  SMRTSSetup ();
  SMRTSFork (argc, argv);
}

#endif

void pcxx_MainBegin(int *argc, char *argv[]) { }
void pcxx_MainEnd() {}

extern int  MyTotalNodes;  /* NO __private!! */
int pcxx_TotalNodes() { return(MyTotalNodes); }

#ifndef _MASTER_SLAVE_
__private extern int  MyNodeNumber;

int pcxx_NodeNumber() { return(MyNodeNumber); }
#endif

extern void RemoteFunctionList(void *element_address,int function_number,
                               void *arg_buffer);

void AskForRemoteExecute(int CollectionNumber,int element_index,
                         int function_number,char *arg_buffer,int size,
                         long tag_location)
{
  long *p, where;
  where = TableOfCollection[CollectionNumber]->Where[element_index];
  RemoteFunctionList((void*)TableOfCollection[CollectionNumber]->Table[element_index],
                     function_number, (void*) arg_buffer);
  if(tag_location != 0) {
    p = (long*)tag_location;
    (*p)++;
  }
}

int AskForFArray(int where, unsigned index, unsigned Coll_Id, int offset,
                       char* buffer, int size)
{
  char *pt;
  pt = (char*)*((long*)(TableOfCollection[Coll_Id]->Table[index]+offset));
  memcpy(buffer,pt,size);
  return 0;
}

void CopyFArray(char *dest,char *source,int rank,int size)
{
  memcpy(dest,source,size);
}

void CopyFArrayElem(char* source, char *buffer,
                  int elem_size, int rank,
                  int i_len, int i_inc, int i_off, 
                  int j_len, int j_inc, int j_off,
                  int k_len, int k_inc, int k_off, 
                  int i, int j, int k)
{
  int m;
  char *pt;
  pt = source + (i + j*i_len + k*i_len*j_len)*elem_size;
  for(m = 0; m < elem_size; m++) buffer[m] = pt[m];
}

int AskForFArrayElem(int where,int index,int Coll_Id,int offset,char *buffer,
                   int elem_size, int rank,
                   int i_len, int i_inc, int i_off, 
                   int j_len, int j_inc, int j_off,
                   int k_len, int k_inc, int k_off, 
                   int i, int j, int k)
{
  char *pt;
  pt = (char*)*((long*)(TableOfCollection[Coll_Id]->Table[index]+offset));
  pt = pt + (i + j*i_len + k*i_len*j_len)*elem_size;
  memcpy(buffer,pt,elem_size);
  return 0;
}
    
void pcxx_ReduceIntAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	((int*)ReduceBuff[self])[elem] += 
	  ((int*)ReduceBuff[self+span])[elem];
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceFloatAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	((float*)ReduceBuff[self])[elem] += 
	  ((float*)ReduceBuff[self+span])[elem];
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceDoubleAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	((double*)ReduceBuff[self])[elem] += 
	  ((double*)ReduceBuff[self+span])[elem];
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceDoubleMax(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((double*)ReduceBuff[self+span])[elem] >
	   ((double*)ReduceBuff[self])[elem]) 
	  {
	    ((double*)ReduceBuff[self])[elem] = 
	      ((double*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceFloatMax(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((float*)ReduceBuff[self+span])[elem] >
	   ((float*)ReduceBuff[self])[elem]) 
	  {
	    ((float*)ReduceBuff[self])[elem] = 
	      ((float*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceIntMax(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((int*)ReduceBuff[self+span])[elem] >
	   ((int*)ReduceBuff[self])[elem]) 
	  {
	    ((int*)ReduceBuff[self])[elem] = 
	      ((int*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceDoubleMin(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((double*)ReduceBuff[self+span])[elem] <
	   ((double*)ReduceBuff[self])[elem]) 
	  {
	    ((double*)ReduceBuff[self])[elem] = 
	      ((double*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceFloatMin(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((float*)ReduceBuff[self+span])[elem] <
	   ((float*)ReduceBuff[self])[elem]) 
	  {
	    ((float*)ReduceBuff[self])[elem] = 
	      ((float*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

void pcxx_ReduceIntMin(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    Message("Reduction buffer too small!  Increase buffer", FATAL);
  self = pcxx_NodeNumber();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  pcxx_Barrier();
  /* tree reduction */
  for(span = 1; span < pcxx_TotalNodes(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < pcxx_TotalNodes()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((int*)ReduceBuff[self+span])[elem] <
	   ((int*)ReduceBuff[self])[elem]) 
	  {
	    ((int*)ReduceBuff[self])[elem] = 
	      ((int*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    pcxx_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  pcxx_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive pcxx_ReduceIntAdd are called               */
  pcxx_Barrier();
}

