/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include "Thread.h"       /* Awesime thread class */
#include "Barrier.h"       /* Awesime thread class */
#include "CpuMux.h"       /* Awesime CpuMux class */
#include "MultiFifoMux.h" /* Awesime Mux class */

extern "C" void  MSNewProc(int);
extern "C" void  MSFinishProc(void);
extern "C" void  pcaw_Barrier(void);
extern "C" void  pcaw_Barrier_Uninstrumented(void);
extern "C" int   pcaw_inseq;

char *pcaw_cc = NULL;
int pcaw_cf = 0;

MultiFifoMux pcaw_Cpu;
int pcaw_shmemsize;
int pcaw_numcpus;
Barrier AweBarrier;

class pcxx_Thread : public Thread {
 public:
  int tid;
 public:
  pcxx_Thread(int i, int stacksize) : Thread("pcxx_Thread", stacksize)
    { tid = i; };
  virtual void main();
};

void pcxx_Thread::main()
{

  if (tid == 0) {			// Main thread
    pcaw_inseq--;
    pcaw_Barrier_Uninstrumented();
    MSNewProc(tid);
    pcaw_Barrier();
    pcaw_inseq++;

    ResetCollectionTraversalLevel ();

    Processor_Main();

    /* terminate threads */
    pcaw_inseq--;
    pcaw_cc = (char *)0;
    pcaw_Barrier();
    MSFinishProc();
    pcaw_Barrier_Uninstrumented();
    pcaw_inseq++;
  }
  else {				// Worker threads
    int done = 0;

    pcaw_Barrier_Uninstrumented();
    MSNewProc(tid);
    pcaw_Barrier();

    while(!done){
      pcaw_Barrier();
      if(pcaw_cc) {
        WorkerSwitch(pcaw_cf, pcaw_cc);
        pcaw_Barrier();
      }
      else{
        done = 1;
        MSFinishProc();
        pcaw_Barrier_Uninstrumented();
      }
    }
  }
}

extern "C" void rAwedezvous()
{
  AweBarrier.rendezvous();
}

extern "C" void AweSetup(int nodes, int cpus, int shmemsize, int stacksize)
{
  int i;

  pcaw_shmemsize = shmemsize;
  pcaw_numcpus = cpus;

  AweBarrier.height(nodes);

  for(i=1; i<nodes; i++){
    CpuMux::add(new pcxx_Thread(i, stacksize));
  }
  CpuMux::add(new pcxx_Thread(0, stacksize));
}

extern "C" int AweD()
{
  pcxx_Thread *thisThread = (pcxx_Thread *)(CpuMux::CurrentThread());

  return thisThread->tid;
}  

extern "C" void New_Main()
{
  /* startup the threads */
  pcaw_Cpu.fireItUp(pcaw_numcpus , pcaw_shmemsize);
}


void pcaw_EnterPar(Kernel *coll, int selector)
{
  if(coll){
    ResetLocal(coll);
  }
  pcxx_inpar++;
  pcaw_inseq--;
  pcaw_cf = selector;
  pcaw_cc = (char *)coll;
  pcaw_Barrier();
  WorkerSwitch(selector, (char *)coll);
}

void pcaw_SubsetEnterPar(Kernel *coll, int selector)
{
  pcaw_EnterPar(coll, selector);
}

void pcaw_ExitPar()
{
  pcaw_Barrier();  
  pcaw_inseq++;
  pcxx_inpar--;
  ResetCollectionTraversalLevel();
}

