/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/************************************************************************
 FArray class used in conjuction with Fortran collection to provide 
 Fortran interface. Include this file at the top of your pC++ program.

 Written by Shelby Yang, Computer Science, Indiana University (11/11/93).
 *************************************************************************/

// On some shared memory machines, memory to be accessed by other processors
// has to be allocated on the "shared-memory."  Furthermore, for the SGI, 
// the shared-memory has to be initialized before it can be used.  But since 
// the initialization is usually done in the beginning of the main program,
// allocating a global variable on shared-memory outside main would be
// a problem.  

# ifdef _SMRTS_
#include <stddef.h>
  void * operator new(size_t size) {
    return (void *)tulip_SharedMalloc(size);
  }
  void operator delete(void *p) {
    tulip_SharedFree(p);
  }
# endif

#include "Complex.h"

#define NEWS 'n'
#define SERIAL 's'

class FArrayDouble {
  double *p0, *p1, *p2, *p3;
 public:
  int i_len,i_inc,i_off;
  int j_len,j_inc,j_off;
  int k_len,k_inc,k_off;
  double *d;
  int rank, size;
  FArrayDouble() {d = 0;};
  FArrayDouble(int i);
  FArrayDouble(int i, int j);
  FArrayDouble(int i, int j, int k);
  FArrayDouble(int i, char layout1);
  FArrayDouble(int i, int j, char layout1, char layout2); 
  FArrayDouble(int i, int j, int k, char layout1, char layout2, char layout3);
  ~FArrayDouble();
  double& operator()(int i);
  double& operator()(int i, int j);
  double& operator()(int i, int j, int k);
};

class FArrayFloat {
  float *p0, *p1, *p2, *p3;
 public:
  int i_len,i_inc,i_off;
  int j_len,j_inc,j_off;
  int k_len,k_inc,k_off;
  float *d;
  int rank, size;
  FArrayFloat() {d = 0;};
  FArrayFloat(int i);
  FArrayFloat(int i, int j);
  FArrayFloat(int i, int j, int k);
  FArrayFloat(int i, char layout1);
  FArrayFloat(int i, int j, char layout1, char layout2); 
  FArrayFloat(int i, int j, int k, char layout1, char layout2, char layout3);
  ~FArrayFloat();
  float& operator()(int i);
  float& operator()(int i, int j);
  float& operator()(int i, int j, int k);
};

// double complex
class FArrayComplex {
  Complex *p0, *p1, *p2, *p3;
 public:
  int i_len,i_inc,i_off;
  int j_len,j_inc,j_off;
  int k_len,k_inc,k_off;
  Complex *d;
  int rank, size;
  FArrayComplex() {d = 0;};
  FArrayComplex(int i);
  FArrayComplex(int i, int j);
  FArrayComplex(int i, int j, int k);
  FArrayComplex(int i, char layout1);
  FArrayComplex(int i, int j, char layout1, char layout2); 
  FArrayComplex(int i, int j, int k, char layout1, char layout2, char layout3);
  ~FArrayComplex();
  Complex& operator()(int i);
  Complex& operator()(int i, int j);
  Complex& operator()(int i, int j, int k);
};

// single complex
class FArrayComplexFloat {
  ComplexFloat *p0, *p1, *p2, *p3;
 public:
  int i_len,i_inc,i_off;
  int j_len,j_inc,j_off;
  int k_len,k_inc,k_off;
  ComplexFloat *d;
  int rank, size;
  FArrayComplexFloat() {d = 0;};
  FArrayComplexFloat(int i);
  FArrayComplexFloat(int i, int j);
  FArrayComplexFloat(int i, int j, int k);
  FArrayComplexFloat(int i, char layout1);
  FArrayComplexFloat(int i, int j, char layout1, char layout2); 
  FArrayComplexFloat(int i, int j, int k, char layout1, char layout2, char layout3);
  ~FArrayComplexFloat();
  ComplexFloat& operator()(int i);
  ComplexFloat& operator()(int i, int j);
  ComplexFloat& operator()(int i, int j, int k);
};

#ifdef CM5
/****************************************************************************
  CM5 implementation note:

 The class member functions allocate arrays on the vector units (VU) and
  store information that is necessary for accessing each individual array
  element.

  The algorithm used to compute the address of an element is described in
  CMRTS Architectural Specification Section 4.2, Array Geometries -- A
  Technical Description.  Given indexes of an array element , we first compute
  physical grid coordinates and subgrid coordinates for the element.  The
  physical coordinates and "offchip_position" are then used to find out on
  which vector unit the element resides.  The subgrid coordinates and
  "subgrid_axis_increment" are used to compute the memory offset into the
  vector unit.  "offchip_position" takes care of the physical axis sequence
  and "subgrid_axis_increment" takes care of the subgrid axis sequence.
  "subgrid_axis_increment" is the number of elements that one needs to 
  skip to reach the next element along a given axis.  This algorithm should 
  work for any array layout.

  Shelby Yang, Computer Science, Indiana University.  (10/1/93)

  Note: 1) The values of the two integers, news_order and serial_order 
           may be changed by TMC without notice.  Check with <cm/cmrt_types> 
	   for their correctness.
	   
	2) The pointer returned by allocate_vu_array() is a pointer
	   to CMRT array descriptor, CMRT_desc, which is a structure.  
	   Don't treat it as a normal double or float pointer. 

        3) For argument, type, in allocate_vu_array()   
           type = 0 for double
           type = 1 for float
           type = 2 for double complex
           type = 3 for float complex 

 ***************************************************************************/

extern "C" {
  char *allocate_vu_array(int* extents,int *weights,
			  int *axis_order,int rank,int num_of_elem,int type);
  void deallocate_vu_array(char *d);
  char *get_address_on_vu_0(unsigned); 
  char *get_address_on_vu_1(unsigned);
  char *get_address_on_vu_2(unsigned);
  char *get_address_on_vu_3(unsigned);
  unsigned get_cm_location(char *d);
  int get_subgrid_dimension(char *d,int dim);
  int get_offchip_position(char *d,int dim);
  int get_subgrid_axis_increment(char *d,int dim);
  int get_subgrid_size(char *d);
}

enum FArray {news = 'n', serial = 's', news_order = 0, serial_order = 2};

static int MyExp2(int k) {
  int tt, i;
  tt = 1; 
  for(i = 0; i < k; i++) 
    tt = tt*2;
  return tt;
}

//------- double data type -------------------------
FArrayDouble::FArrayDouble(int i)
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  axis_order[0] = news_order;

  d = (double*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,0);

  size = get_subgrid_size((char*)d)*4*sizeof(double);
  loc = get_cm_location((char*)d);

  p0 = (double*)get_address_on_vu_0(loc); 
  p1 = (double*)get_address_on_vu_1(loc);
  p2 = (double*)get_address_on_vu_2(loc);
  p3 = (double*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}

FArrayDouble::FArrayDouble(int i, int j)
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  
  axis_order[0] = news_order;
  axis_order[1] = news_order;

  d = (double*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,0);

  size = get_subgrid_size((char*)d)*4*sizeof(double);  
  loc = get_cm_location((char*)d);

  p0 = (double*)get_address_on_vu_0(loc); 
  p1 = (double*)get_address_on_vu_1(loc);
  p2 = (double*)get_address_on_vu_2(loc);
  p3 = (double*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayDouble::FArrayDouble(int i, int j, int k)
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  axis_order[0] = news_order;
  axis_order[1] = news_order;          
  axis_order[2] = news_order;

  d = (double*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,0);
  
  size = get_subgrid_size((char*)d)*4*sizeof(double);
  loc = get_cm_location((char*)d);

  p0 = (double*)get_address_on_vu_0(loc); 
  p1 = (double*)get_address_on_vu_1(loc);
  p2 = (double*)get_address_on_vu_2(loc);
  p3 = (double*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}

FArrayDouble::FArrayDouble(int i, char layout1) 
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  
  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    break;
  case serial:
    axis_order[0] = serial_order;
    printf("FArrayDouble: Serial 1D array not allowed.\n");
    break;
  default:
    printf("FArrayDouble: Wrong 1D layout.\n");
    break;
  }

  d = (double*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,0);

  size = get_subgrid_size((char*)d)*4*sizeof(double);
  loc = get_cm_location((char*)d);

  p0 = (double*)get_address_on_vu_0(loc); 
  p1 = (double*)get_address_on_vu_1(loc);
  p2 = (double*)get_address_on_vu_2(loc);
  p3 = (double*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}  

FArrayDouble::FArrayDouble(int i,int j,char layout1,char layout2) 
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  

  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      printf("FArrayDouble: Serial 2D array not allowed.\n");
      break;
    }
    break;
  default:
    printf("FArrayDouble: Wrong 2D layout.\n");
    break;
  }

  d = (double*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,0);
  
  size = get_subgrid_size((char*)d)*4*sizeof(double);
  loc = get_cm_location((char*)d);

  p0 = (double*)get_address_on_vu_0(loc); 
  p1 = (double*)get_address_on_vu_1(loc);
  p2 = (double*)get_address_on_vu_2(loc);
  p3 = (double*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
  
  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayDouble::FArrayDouble(int i,int j,int k,char layout1,char layout2,
			   char layout3) 
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  
  switch (layout1) {        
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:  
      axis_order[1] = news_order;          
      switch (layout3) {
      case news: 
	axis_order[2] = news_order;
	break;
      case serial:
	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
	axis_order[2] = news_order;
	break;
      case serial:
      	axis_order[2] = serial_order;
	break;
      }
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order;
	break;
      case serial: 
      	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order; 
	break;
      case serial:
      	axis_order[2] = serial_order; 
	printf("FArrayDouble: Serial 3D array not allowed.\n");
	break;
      }
      break;
    }
    break;
  }

  d = (double*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,0);

  size = get_subgrid_size((char*)d)*4*sizeof(double);
  loc = get_cm_location((char*)d);

  p0 = (double*)get_address_on_vu_0(loc); 
  p1 = (double*)get_address_on_vu_1(loc);
  p2 = (double*)get_address_on_vu_2(loc);
  p3 = (double*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}  

double& FArrayDouble::operator()(int i) 
{
  int node, subgrid_index;
  int i_phy, i_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;

  node = (int)(i_phy*MyExp2(i_off));
  subgrid_index = i_sub*i_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 1D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

double& FArrayDouble::operator()(int i, int j) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 2D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

double& FArrayDouble::operator()(int i, int j, int k) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;
  int k_phy,k_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;
  k_phy = k / k_len;
  k_sub = k % k_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off))
        +(int)(k_phy*MyExp2(k_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc
                + k_sub*k_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 3D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

FArrayDouble::~FArrayDouble()
{
  if(d != 0) deallocate_vu_array((char*)d);
}

// ------------ float data type --------------------
FArrayFloat::FArrayFloat(int i)
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  axis_order[0] = news_order;

  d = (float*)allocate_vu_array(extents,weights,axis_order,
				rank,num_of_elem,1);

  size = get_subgrid_size((char*)d)*4*sizeof(float);
  loc = get_cm_location((char*)d);

  p0 = (float*)get_address_on_vu_0(loc); 
  p1 = (float*)get_address_on_vu_1(loc);
  p2 = (float*)get_address_on_vu_2(loc);
  p3 = (float*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}

FArrayFloat::FArrayFloat(int i, int j)
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  
  axis_order[0] = news_order;
  axis_order[1] = news_order;

  d = (float*)allocate_vu_array(extents,weights,axis_order,
				rank,num_of_elem,1);

  size = get_subgrid_size((char*)d)*4*sizeof(float);  
  loc = get_cm_location((char*)d);

  p0 = (float*)get_address_on_vu_0(loc); 
  p1 = (float*)get_address_on_vu_1(loc);
  p2 = (float*)get_address_on_vu_2(loc);
  p3 = (float*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayFloat::FArrayFloat(int i, int j, int k)
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  axis_order[0] = news_order;
  axis_order[1] = news_order;          
  axis_order[2] = news_order;

  d = (float*)allocate_vu_array(extents,weights,axis_order,
				rank,num_of_elem,1);
  
  size = get_subgrid_size((char*)d)*4*sizeof(float);
  loc = get_cm_location((char*)d);

  p0 = (float*)get_address_on_vu_0(loc); 
  p1 = (float*)get_address_on_vu_1(loc);
  p2 = (float*)get_address_on_vu_2(loc);
  p3 = (float*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}

FArrayFloat::FArrayFloat(int i, char layout1) 
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  
  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    break;
  case serial:
    axis_order[0] = serial_order;
    printf("FArrayFloat: Serial 1D array not allowed.\n");
    break;
  default:
    printf("FArrayFloat: Wrong 1D layout.\n");
    break;
  }

  d = (float*)allocate_vu_array(extents,weights,axis_order,
				rank,num_of_elem,1);

  size = get_subgrid_size((char*)d)*4*sizeof(float);
  loc = get_cm_location((char*)d);

  p0 = (float*)get_address_on_vu_0(loc); 
  p1 = (float*)get_address_on_vu_1(loc);
  p2 = (float*)get_address_on_vu_2(loc);
  p3 = (float*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}  

FArrayFloat::FArrayFloat(int i,int j,char layout1,char layout2) 
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  

  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      printf("FArrayFloat: Serial 2D array not allowed.\n");
      break;
    }
    break;
  default:
    printf("FArrayFloat: Wrong 2D layout.\n");
    break;
  }

  d = (float*)allocate_vu_array(extents,weights,axis_order,
				rank,num_of_elem,1);
  
  size = get_subgrid_size((char*)d)*4*sizeof(float);
  loc = get_cm_location((char*)d);

  p0 = (float*)get_address_on_vu_0(loc); 
  p1 = (float*)get_address_on_vu_1(loc);
  p2 = (float*)get_address_on_vu_2(loc);
  p3 = (float*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
  
  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayFloat::FArrayFloat(int i,int j,int k,char layout1,char layout2,
			 char layout3) 
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  
  switch (layout1) {        
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:  
      axis_order[1] = news_order;          
      switch (layout3) {
      case news: 
	axis_order[2] = news_order;
	break;
      case serial:
	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
	axis_order[2] = news_order;
	break;
      case serial:
      	axis_order[2] = serial_order;
	break;
      }
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order;
	break;
      case serial: 
      	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order; 
	break;
      case serial:
      	axis_order[2] = serial_order; 
	printf("FArrayFloat: Serial 3D array not allowed.\n");
	break;
      }
      break;
    }
    break;
  }

  d = (float*)allocate_vu_array(extents,weights,axis_order,
				rank,num_of_elem,1);

  size = get_subgrid_size((char*)d)*4*sizeof(float);
  loc = get_cm_location((char*)d);

  p0 = (float*)get_address_on_vu_0(loc); 
  p1 = (float*)get_address_on_vu_1(loc);
  p2 = (float*)get_address_on_vu_2(loc);
  p3 = (float*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}  

float& FArrayFloat::operator()(int i) 
{
  int node, subgrid_index;
  int i_phy, i_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;

  node = (int)(i_phy*MyExp2(i_off));
  subgrid_index = i_sub*i_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 1D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

float& FArrayFloat::operator()(int i, int j) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 2D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

float& FArrayFloat::operator()(int i, int j, int k) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;
  int k_phy,k_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;
  k_phy = k / k_len;
  k_sub = k % k_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off))
        +(int)(k_phy*MyExp2(k_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc
                + k_sub*k_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 3D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

FArrayFloat::~FArrayFloat()
{
  if(d != 0) deallocate_vu_array((char*)d);
}

// ---------- complex double data type ----------------------
FArrayComplex::FArrayComplex(int i)
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  axis_order[0] = news_order;

  d = (Complex*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(Complex);
  loc = get_cm_location((char*)d);

  p0 = (Complex*)get_address_on_vu_0(loc); 
  p1 = (Complex*)get_address_on_vu_1(loc);
  p2 = (Complex*)get_address_on_vu_2(loc);
  p3 = (Complex*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}

FArrayComplex::FArrayComplex(int i, int j)
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  
  axis_order[0] = news_order;
  axis_order[1] = news_order;

  d = (Complex*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(Complex);  
  loc = get_cm_location((char*)d);

  p0 = (Complex*)get_address_on_vu_0(loc); 
  p1 = (Complex*)get_address_on_vu_1(loc);
  p2 = (Complex*)get_address_on_vu_2(loc);
  p3 = (Complex*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayComplex::FArrayComplex(int i, int j, int k)
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  axis_order[0] = news_order;
  axis_order[1] = news_order;          
  axis_order[2] = news_order;

  d = (Complex*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);
  
  size = get_subgrid_size((char*)d)*4*sizeof(Complex);
  loc = get_cm_location((char*)d);

  p0 = (Complex*)get_address_on_vu_0(loc); 
  p1 = (Complex*)get_address_on_vu_1(loc);
  p2 = (Complex*)get_address_on_vu_2(loc);
  p3 = (Complex*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}

FArrayComplex::FArrayComplex(int i, char layout1) 
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  
  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    break;
  case serial:
    axis_order[0] = serial_order;
    printf("FArrayComplex: Serial 1D array not allowed.\n");
    break;
  default:
    printf("FArrayComplex: Wrong 1D layout.\n");
    break;
  }

  d = (Complex*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(Complex);
  loc = get_cm_location((char*)d);

  p0 = (Complex*)get_address_on_vu_0(loc); 
  p1 = (Complex*)get_address_on_vu_1(loc);
  p2 = (Complex*)get_address_on_vu_2(loc);
  p3 = (Complex*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}  

FArrayComplex::FArrayComplex(int i,int j,char layout1,char layout2) 
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  

  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      printf("FArrayComplex: Serial 2D array not allowed.\n");
      break;
    }
    break;
  default:
    printf("FArrayComplex: Wrong 2D layout.\n");
    break;
  }

  d = (Complex*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);
  
  size = get_subgrid_size((char*)d)*4*sizeof(Complex);
  loc = get_cm_location((char*)d);

  p0 = (Complex*)get_address_on_vu_0(loc); 
  p1 = (Complex*)get_address_on_vu_1(loc);
  p2 = (Complex*)get_address_on_vu_2(loc);
  p3 = (Complex*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
  
  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayComplex::FArrayComplex(int i,int j,int k,char layout1,char layout2,
			   char layout3) 
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  
  switch (layout1) {        
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:  
      axis_order[1] = news_order;          
      switch (layout3) {
      case news: 
	axis_order[2] = news_order;
	break;
      case serial:
	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
	axis_order[2] = news_order;
	break;
      case serial:
      	axis_order[2] = serial_order;
	break;
      }
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order;
	break;
      case serial: 
      	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order; 
	break;
      case serial:
      	axis_order[2] = serial_order; 
	printf("FArrayComplex: Serial 3D array not allowed.\n");
	break;
      }
      break;
    }
    break;
  }

  d = (Complex*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(Complex);
  loc = get_cm_location((char*)d);

  p0 = (Complex*)get_address_on_vu_0(loc); 
  p1 = (Complex*)get_address_on_vu_1(loc);
  p2 = (Complex*)get_address_on_vu_2(loc);
  p3 = (Complex*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}  

Complex& FArrayComplex::operator()(int i) 
{
  int node, subgrid_index;
  int i_phy, i_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;

  node = (int)(i_phy*MyExp2(i_off));
  subgrid_index = i_sub*i_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 1D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

Complex& FArrayComplex::operator()(int i, int j) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 2D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

Complex& FArrayComplex::operator()(int i, int j, int k) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;
  int k_phy,k_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;
  k_phy = k / k_len;
  k_sub = k % k_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off))
        +(int)(k_phy*MyExp2(k_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc
                + k_sub*k_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 3D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

FArrayComplex::~FArrayComplex()
{
  if(d != 0) deallocate_vu_array((char*)d);
}

// ---------- complex float data type ----------------------
FArrayComplexFloat::FArrayComplexFloat(int i)
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  axis_order[0] = news_order;

  d = (ComplexFloat*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(ComplexFloat);
  loc = get_cm_location((char*)d);

  p0 = (ComplexFloat*)get_address_on_vu_0(loc); 
  p1 = (ComplexFloat*)get_address_on_vu_1(loc);
  p2 = (ComplexFloat*)get_address_on_vu_2(loc);
  p3 = (ComplexFloat*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}

FArrayComplexFloat::FArrayComplexFloat(int i, int j)
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  
  axis_order[0] = news_order;
  axis_order[1] = news_order;

  d = (ComplexFloat*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(ComplexFloat);  
  loc = get_cm_location((char*)d);

  p0 = (ComplexFloat*)get_address_on_vu_0(loc); 
  p1 = (ComplexFloat*)get_address_on_vu_1(loc);
  p2 = (ComplexFloat*)get_address_on_vu_2(loc);
  p3 = (ComplexFloat*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayComplexFloat::FArrayComplexFloat(int i, int j, int k)
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  axis_order[0] = news_order;
  axis_order[1] = news_order;          
  axis_order[2] = news_order;

  d = (ComplexFloat*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);
  
  size = get_subgrid_size((char*)d)*4*sizeof(ComplexFloat);
  loc = get_cm_location((char*)d);

  p0 = (ComplexFloat*)get_address_on_vu_0(loc); 
  p1 = (ComplexFloat*)get_address_on_vu_1(loc);
  p2 = (ComplexFloat*)get_address_on_vu_2(loc);
  p3 = (ComplexFloat*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}

FArrayComplexFloat::FArrayComplexFloat(int i, char layout1) 
{
  int num_of_elem,extents[1],weights[1],axis_order[1]; 
  unsigned loc;

  rank = 1;
  num_of_elem = i;
  extents[0] = i;
  weights[0] = 1;
  
  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    break;
  case serial:
    axis_order[0] = serial_order;
    printf("FArrayComplexFloat: Serial 1D array not allowed.\n");
    break;
  default:
    printf("FArrayComplexFloat: Wrong 1D layout.\n");
    break;
  }

  d = (ComplexFloat*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(ComplexFloat);
  loc = get_cm_location((char*)d);

  p0 = (ComplexFloat*)get_address_on_vu_0(loc); 
  p1 = (ComplexFloat*)get_address_on_vu_1(loc);
  p2 = (ComplexFloat*)get_address_on_vu_2(loc);
  p3 = (ComplexFloat*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
}  

FArrayComplexFloat::FArrayComplexFloat(int i,int j,char layout1,char layout2) 
{
  int num_of_elem,extents[2],weights[2],axis_order[2]; 
  unsigned loc;

  rank = 2;
  num_of_elem = i*j;
  extents[0] = i;
  extents[1] = j;
  weights[0] = 1;
  weights[1] = 1;  

  switch (layout1) {
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;
      break;
    case serial:
      axis_order[1] = serial_order;
      printf("FArrayComplexFloat: Serial 2D array not allowed.\n");
      break;
    }
    break;
  default:
    printf("FArrayComplexFloat: Wrong 2D layout.\n");
    break;
  }

  d = (ComplexFloat*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);
  
  size = get_subgrid_size((char*)d)*4*sizeof(ComplexFloat);
  loc = get_cm_location((char*)d);

  p0 = (ComplexFloat*)get_address_on_vu_0(loc); 
  p1 = (ComplexFloat*)get_address_on_vu_1(loc);
  p2 = (ComplexFloat*)get_address_on_vu_2(loc);
  p3 = (ComplexFloat*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);
  
  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);
}

FArrayComplexFloat::FArrayComplexFloat(int i,int j,int k,char layout1,char layout2,
			   char layout3) 
{
  int num_of_elem,extents[3],weights[3],axis_order[3]; 
  unsigned loc;

  rank = 3;
  num_of_elem = i*j*k;
  extents[0] = i;
  extents[1] = j;
  extents[2] = k;
  weights[0] = 1;
  weights[1] = 1;  
  weights[2] = 1;
  
  switch (layout1) {        
  case news:
    axis_order[0] = news_order;
    switch (layout2) {
    case news:  
      axis_order[1] = news_order;          
      switch (layout3) {
      case news: 
	axis_order[2] = news_order;
	break;
      case serial:
	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
	axis_order[2] = news_order;
	break;
      case serial:
      	axis_order[2] = serial_order;
	break;
      }
      break;
    }
    break;
  case serial:
    axis_order[0] = serial_order;
    switch (layout2) {
    case news:
      axis_order[1] = news_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order;
	break;
      case serial: 
      	axis_order[2] = serial_order;
	break;
      }
      break;
    case serial:
      axis_order[1] = serial_order;          
      switch (layout3) {
      case news:
      	axis_order[2] = news_order; 
	break;
      case serial:
      	axis_order[2] = serial_order; 
	printf("FArrayComplexFloat: Serial 3D array not allowed.\n");
	break;
      }
      break;
    }
    break;
  }

  d = (ComplexFloat*)allocate_vu_array(extents,weights,axis_order,
					rank,num_of_elem,3);

  size = get_subgrid_size((char*)d)*4*sizeof(ComplexFloat);
  loc = get_cm_location((char*)d);

  p0 = (ComplexFloat*)get_address_on_vu_0(loc); 
  p1 = (ComplexFloat*)get_address_on_vu_1(loc);
  p2 = (ComplexFloat*)get_address_on_vu_2(loc);
  p3 = (ComplexFloat*)get_address_on_vu_3(loc);

  i_len = get_subgrid_dimension((char*)d,0);
  i_inc = get_subgrid_axis_increment((char*)d,0);
  i_off = get_offchip_position((char*)d,0);

  j_len = get_subgrid_dimension((char*)d,1);
  j_inc = get_subgrid_axis_increment((char*)d,1);
  j_off = get_offchip_position((char*)d,1);

  k_len = get_subgrid_dimension((char*)d,2);
  k_inc = get_subgrid_axis_increment((char*)d,2);
  k_off = get_offchip_position((char*)d,2);
}  

ComplexFloat& FArrayComplexFloat::operator()(int i) 
{
  int node, subgrid_index;
  int i_phy, i_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;

  node = (int)(i_phy*MyExp2(i_off));
  subgrid_index = i_sub*i_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 1D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

ComplexFloat& FArrayComplexFloat::operator()(int i, int j) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 2D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

ComplexFloat& FArrayComplexFloat::operator()(int i, int j, int k) {
  int node,subgrid_index;
  int i_phy,i_sub;
  int j_phy,j_sub;
  int k_phy,k_sub;

  i_phy = i / i_len;
  i_sub = i % i_len;
  j_phy = j / j_len;
  j_sub = j % j_len;
  k_phy = k / k_len;
  k_sub = k % k_len;

  node = (int)(i_phy*MyExp2(i_off))
        +(int)(j_phy*MyExp2(j_off))
        +(int)(k_phy*MyExp2(k_off));

  subgrid_index = i_sub*i_inc
                + j_sub*j_inc
                + k_sub*k_inc;

  switch (node) {
  case 0:
    return *(p0+subgrid_index);
  case 1:
    return *(p1+subgrid_index);
  case 2:
    return *(p2+subgrid_index);
  case 3:
    return *(p3+subgrid_index);
  default:
    printf("Processor %d: Fatal error in accessing 3D FArray element!\n",
	   pcxx_NodeNumber());
    return *p0;
  }
}

FArrayComplexFloat::~FArrayComplexFloat()
{
  if(d != 0) deallocate_vu_array((char*)d);
}

#endif

/**************************************
  Non-CM5 implementation
 **************************************/
#ifndef CM5
//----- double data type --------
FArrayDouble::FArrayDouble(int i)
{
    int elem_count;
    elem_count = i;
    d = new double[elem_count];
    size = elem_count*sizeof(double);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayDouble::FArrayDouble(int i, int j)
{
    int elem_count;
    elem_count = i*j;
    d = new double[elem_count];
    size = elem_count*sizeof(double);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayDouble::FArrayDouble(int i, int j, int k)
{
    int elem_count;
    elem_count = i*j*k;
    d = new double[elem_count];
    size = elem_count*sizeof(double);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

FArrayDouble::FArrayDouble(int i, char)
{
    int elem_count;
    elem_count = i;
    d = new double[elem_count];
    size = elem_count*sizeof(double);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayDouble::FArrayDouble(int i, int j, char , char )
{ 
    int elem_count;
    elem_count = i*j;
    d = new double[elem_count];
    size = elem_count*sizeof(double);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayDouble::FArrayDouble(int i, int j, int k, char , char , 
			   char )
{
    int elem_count;
    elem_count = i*j*k;
    d = new double[elem_count];
    size = elem_count*sizeof(double);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

double& FArrayDouble::operator()(int i)
{
  return *(d + i);
}

double& FArrayDouble::operator()(int i, int j)
{
  return *(d + i + j*i_len);
}

double& FArrayDouble::operator()(int i, int j, int k)
{
  return *(d + i + j*i_len + k*i_len*j_len);
}

FArrayDouble::~FArrayDouble()
{
  if(d != 0) delete [] d;
}

//------- float data type -------------
FArrayFloat::FArrayFloat(int i)
{
    int elem_count;
    elem_count = i;
    d = new float[elem_count];
    size = elem_count*sizeof(float);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayFloat::FArrayFloat(int i, int j)
{
    int elem_count;
    elem_count = i*j;
    d = new float[elem_count];
    size = elem_count*sizeof(float);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayFloat::FArrayFloat(int i, int j, int k)
{
    int elem_count;
    elem_count = i*j*k;
    d = new float[elem_count];
    size = elem_count*sizeof(float);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

FArrayFloat::FArrayFloat(int i, char )
{
    int elem_count;
    elem_count = i;
    d = new float[elem_count];
    size = elem_count*sizeof(float);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayFloat::FArrayFloat(int i, int j, char , char )
{ 
    int elem_count;
    elem_count = i*j;
    d = new float[elem_count];
    size = elem_count*sizeof(float);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayFloat::FArrayFloat(int i, int j, int k, char , char , 
			   char )
{
    int elem_count;
    elem_count = i*j*k;
    d = new float[elem_count];
    size = elem_count*sizeof(float);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

float& FArrayFloat::operator()(int i)
{
  return *(d + i);
}

float& FArrayFloat::operator()(int i, int j)
{
  return *(d + i + j*i_len);
}

float& FArrayFloat::operator()(int i, int j, int k)
{
  return *(d + i + j*i_len + k*i_len*j_len);
}

FArrayFloat::~FArrayFloat()
{
  if(d != 0) delete [] d;
}

//----- complex double data type ----------
FArrayComplex::FArrayComplex(int i)
{
    int elem_count;
    elem_count = i;
    d = new Complex[elem_count];
    size = elem_count*sizeof(Complex);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayComplex::FArrayComplex(int i, int j)
{
    int elem_count;
    elem_count = i*j;
    d = new Complex[elem_count];
    size = elem_count*sizeof(Complex);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayComplex::FArrayComplex(int i, int j, int k)
{
    int elem_count;
    elem_count = i*j*k;
    d = new Complex[elem_count];
    size = elem_count*sizeof(Complex);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

FArrayComplex::FArrayComplex(int i, char )
{
    int elem_count;
    elem_count = i;
    d = new Complex[elem_count];
    size = elem_count*sizeof(Complex);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayComplex::FArrayComplex(int i, int j, char , char )
{ 
    int elem_count;
    elem_count = i*j;
    d = new Complex[elem_count];
    size = elem_count*sizeof(Complex);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayComplex::FArrayComplex(int i, int j, int k, char , char , 
			   char )
{
    int elem_count;
    elem_count = i*j*k;
    d = new Complex[elem_count];
    size = elem_count*sizeof(Complex);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

Complex& FArrayComplex::operator()(int i)
{
  return *(d + i);
}

Complex& FArrayComplex::operator()(int i, int j)
{
  return *(d + i + j*i_len);
}

Complex& FArrayComplex::operator()(int i, int j, int k)
{
  return *(d + i + j*i_len + k*i_len*j_len);
}

FArrayComplex::~FArrayComplex()
{
  if(d != 0) delete [] d;
}

//------- complex float data type -------------
FArrayComplexFloat::FArrayComplexFloat(int i)
{
    int elem_count;
    elem_count = i;
    d = new ComplexFloat[elem_count];
    size = elem_count*sizeof(ComplexFloat);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayComplexFloat::FArrayComplexFloat(int i, int j)
{
    int elem_count;
    elem_count = i*j;
    d = new ComplexFloat[elem_count];
    size = elem_count*sizeof(ComplexFloat);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayComplexFloat::FArrayComplexFloat(int i, int j, int k)
{
    int elem_count;
    elem_count = i*j*k;
    d = new ComplexFloat[elem_count];
    size = elem_count*sizeof(ComplexFloat);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

FArrayComplexFloat::FArrayComplexFloat(int i, char )
{
    int elem_count;
    elem_count = i;
    d = new ComplexFloat[elem_count];
    size = elem_count*sizeof(ComplexFloat);
    rank = 1; 
    i_len = i;
    j_len = 0;
    k_len = 0;
}

FArrayComplexFloat::FArrayComplexFloat(int i, int j, char , char )
{ 
    int elem_count;
    elem_count = i*j;
    d = new ComplexFloat[elem_count];
    size = elem_count*sizeof(ComplexFloat);
    rank = 2;
    i_len = i;
    j_len = j;
    k_len = 0;
}

FArrayComplexFloat::FArrayComplexFloat(int i, int j, int k, char , char , char)
{
    int elem_count;
    elem_count = i*j*k;
    d = new ComplexFloat[elem_count];
    size = elem_count*sizeof(ComplexFloat);
    rank = 3;
    i_len = i;
    j_len = j;
    k_len = k;
}

ComplexFloat& FArrayComplexFloat::operator()(int i)
{
  return *(d + i);
}

ComplexFloat& FArrayComplexFloat::operator()(int i, int j)
{
  return *(d + i + j*i_len);
}

ComplexFloat& FArrayComplexFloat::operator()(int i, int j, int k)
{
  return *(d + i + j*i_len + k*i_len*j_len);
}

FArrayComplexFloat::~FArrayComplexFloat()
{
  if(d != 0) delete [] d;
}

#endif

