/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************
  A special collection class for Fortran interface.  Used in 
  conjunction with FArray class.  Include this file after kernel.h
  is included.

  Written by Shelby Yang, Indiana University.  (9/15/93)
**********************************************************************/
extern "C" {
  int AskForFArray(int where, unsigned index, unsigned Coll_Id, int offset,
		   char *buffer, int size);
  void CopyFArray(char *dest,char *source,int rank,int size);
  void CopyFArrayElem(char* source, char *buffer,
		      int elem_size, int rank,
		      int i_len, int i_inc, int i_off, 
		      int j_len, int j_inc, int j_off,
		      int k_len, int k_inc, int k_off, 
		      int i, int j, int k);
  int AskForFArrayElem(int where,int index,int Coll_Id,int offset,char *buffer,
		       int elem_size, int rank,
		       int i_len, int i_inc, int i_off, 
		       int j_len, int j_inc, int j_off,
		       int k_len, int k_inc, int k_off, 
		       int i, int j, int k);
}

Collection Fortran : public SuperKernel {
 public:
  Fortran(Distribution *T, Align *A);

  // FArrayDouble related

  void GetFArray(int index1,int offset,FArrayDouble &buffer);
  void GetFArray(int index1,int index2,int offset,FArrayDouble &buffer);
  double GetFArrayElem(int index,int offset,FArrayDouble &which_farray,
                       int i);
  double GetFArrayElem(int index1,int offset,FArrayDouble &which_farray,
		       int i,int j);
  double GetFArrayElem(int index1,int offset,FArrayDouble &which_farray,
		       int i,int j,int k);
  double GetFArrayElem(int index1,int index2,int offset,
		       FArrayDouble &which_farray,int i);
  double GetFArrayElem(int index1,int index2,int offset,
		       FArrayDouble &which_farray,int i,int j);
  double GetFArrayElem(int index1,int index2,int offset,
		       FArrayDouble &which_farray,int i,int j,int k);

  // FArrayFloat related

  void GetFArray(int index1,int offset,FArrayFloat &buffer);
  void GetFArray(int index1,int index2,int offset,FArrayFloat &buffer);
  float GetFArrayElem(int index,int offset,FArrayFloat &which_farray,
                       int i);
  float GetFArrayElem(int index1,int offset,FArrayFloat &which_farray,
		       int i,int j);
  float GetFArrayElem(int index1,int offset,FArrayFloat &which_farray,
		       int i,int j,int k);
  float GetFArrayElem(int index1,int index2,int offset,
		       FArrayFloat &which_farray,int i);
  float GetFArrayElem(int index1,int index2,int offset,
		       FArrayFloat &which_farray,int i,int j);
  float GetFArrayElem(int index1,int index2,int offset,
		       FArrayFloat &which_farray,int i,int j,int k);

  // FArrayComplexFloat related

  void GetFArray(int index1,int offset,FArrayComplexFloat &buffer);
  void GetFArray(int index1,int index2,int offset,FArrayComplexFloat &buffer);
  ComplexFloat GetFArrayElem(int index,int offset,
			     FArrayComplexFloat &which_farray,int i);
  ComplexFloat GetFArrayElem(int index1,int offset, 
			     FArrayComplexFloat &which_farray,int i,int j);
  ComplexFloat GetFArrayElem(int index1,int offset, 
			     FArrayComplexFloat &which_farray,int i,int j, 
			     int k);
  ComplexFloat GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplexFloat &which_farray,int i);
  ComplexFloat GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplexFloat &which_farray,int i,int j);
  ComplexFloat GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplexFloat &which_farray,int i,int j, 
			     int k);

  // FArrayComplex related

  void GetFArray(int index1,int offset,FArrayComplex &buffer);
  void GetFArray(int index1,int index2,int offset,FArrayComplex &buffer);
  Complex GetFArrayElem(int index,int offset,
			     FArrayComplex &which_farray,int i);
  Complex GetFArrayElem(int index1,int offset, 
			     FArrayComplex &which_farray,int i,int j);
  Complex GetFArrayElem(int index1,int offset, 
			     FArrayComplex &which_farray,int i,int j, 
			     int k);
  Complex GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplex &which_farray,int i);
  Complex GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplex &which_farray,int i,int j);
  Complex GetFArrayElem(int index1,int index2,int offset,
			     FArrayComplex &which_farray,int i,int j, 
			     int k);
};  
/*be careful with semicolons; makes strange errors (not always syntax errors)*/

Fortran::Fortran(Distribution *T, Align *A) : SuperKernel(T, A) 
{}

//--------- FArrayDouble -------------------
void Fortran::GetFArray(int index1,int offset,FArrayDouble &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);
 
       if (Is_Local(index1))	 
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d, source, buffer.rank, buffer.size);
	 return ;
       }
       where = WhereIsData(index,CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d, 
		    buffer.size);
       return;
}

void Fortran::GetFArray(int index1,int index2,int offset,FArrayDouble &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1,index2);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);

       if (Is_Local(index1, index2))
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d,source,buffer.rank,buffer.size);
	 return ;
       }
       where = WhereIsData(index, CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d,
		    buffer.size);
       return;
}

double Fortran::GetFArrayElem(int index1,int offset,FArrayDouble &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(double), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(double*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(double), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int offset,FArrayDouble &which,int i,
			      int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(double), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(double*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(double), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int offset,FArrayDouble &which,int i,
			      int j, int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(double), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(double*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(double), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayDouble &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(double), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(double*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(double), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayDouble &which,int i,int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(double), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(double*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(double), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(double*)buffer;
}

double Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayDouble &which,int i,int j,int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(double), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(double*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(double), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(double*)buffer;
}


//--------- FArrayFloat -------------------
void Fortran::GetFArray(int index1,int offset,FArrayFloat &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);
 
       if (Is_Local(index1))	 
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d, source, buffer.rank, buffer.size);
	 return ;
       }
       where = WhereIsData(index,CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d, 
		    buffer.size);
       return;
}

void Fortran::GetFArray(int index1,int index2,int offset,FArrayFloat &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1,index2);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);

       if (Is_Local(index1, index2))
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d,source,buffer.rank,buffer.size);
	 return ;
       }
       where = WhereIsData(index, CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d,
		    buffer.size);
       return;
}

float Fortran::GetFArrayElem(int index1,int offset,FArrayFloat &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(float), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(float*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(float), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int offset,FArrayFloat &which,int i,
			      int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(float), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(float*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(float), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int offset,FArrayFloat &which,int i,
			      int j, int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(float), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(float*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(float), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayFloat &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(float), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(float*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(float), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayFloat &which,int i,int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(float), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(float*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(float), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(float*)buffer;
}

float Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayFloat &which,int i,int j,int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(float), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(float*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(float), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(float*)buffer;
}

//------- FArarrayComplexFloat ----------
void Fortran::GetFArray(int index1,int offset,FArrayComplexFloat &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);
 
       if (Is_Local(index1))	 
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d, source, buffer.rank, buffer.size);
	 return ;
       }
       where = WhereIsData(index,CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d, 
		    buffer.size);
       return;
}


void Fortran::GetFArray(int index1,int index2,int offset,FArrayComplexFloat &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1,index2);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);

       if (Is_Local(index1, index2))
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d,source,buffer.rank,buffer.size);
	 return ;
       }
       where = WhereIsData(index, CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d,
		    buffer.size);
       return;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int offset,FArrayComplexFloat &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(ComplexFloat), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(ComplexFloat*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(ComplexFloat), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int offset,FArrayComplexFloat &which,int i,
			      int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(ComplexFloat), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(ComplexFloat*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(ComplexFloat), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int offset,FArrayComplexFloat &which,int i,
			      int j, int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(ComplexFloat), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(ComplexFloat*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(ComplexFloat), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplexFloat &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(ComplexFloat), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(ComplexFloat*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(ComplexFloat), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplexFloat &which,int i,int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(ComplexFloat), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(ComplexFloat*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(ComplexFloat), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(ComplexFloat*)buffer;
}

ComplexFloat Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplexFloat &which,int i,int j,int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(ComplexFloat), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(ComplexFloat*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(ComplexFloat), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(ComplexFloat*)buffer;
}


//------- FArarrayComplex ----------
void Fortran::GetFArray(int index1,int offset,FArrayComplex &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);
 
       if (Is_Local(index1))	 
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d, source, buffer.rank, buffer.size);
	 return ;
       }
       where = WhereIsData(index,CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d, 
		    buffer.size);
       return;
}


void Fortran::GetFArray(int index1,int index2,int offset,FArrayComplex &buffer)
{
       // offset and size are in bytes. 
       int where, off;
       char *source;
       int index;

       index = ElemIndex(index1,index2);

       // convert offset to number of int's due to Francois' int table!
       off = offset/sizeof(int);

       if (Is_Local(index1, index2))
       {
	 source = (char*)*((long*)(CollectionInfo.Table[index] + off));
	 CopyFArray((char*)buffer.d,source,buffer.rank,buffer.size);
	 return ;
       }
       where = WhereIsData(index, CollectionNumber);
       AskForFArray(where,index,CollectionNumber,off,(char*)buffer.d,
		    buffer.size);
       return;
}

Complex Fortran::GetFArrayElem(int index1,int offset,FArrayComplex &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(Complex), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(Complex*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(Complex), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int offset,FArrayComplex &which,int i,
			      int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(Complex), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(Complex*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(Complex), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int offset,FArrayComplex &which,int i,
			      int j, int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(Complex), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(Complex*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(Complex), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplex &which,int i)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(Complex), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
      return *(Complex*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(Complex), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, 0, 0);
    return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplex &which,int i,int j)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(Complex), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
      return *(Complex*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(Complex), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, 0);
    return *(Complex*)buffer;
}

Complex Fortran::GetFArrayElem(int index1,int index2,int offset,
			      FArrayComplex &which,int i,int j,int k)
{
    // offset and size are in bytes. 
    int off, where;
    char *source;
    int index;
  
    index = ElemIndex(index1,index2);

    // convert offset to number of int's due to Francois' int table!
    off = offset/sizeof(int);
 
    if (Is_Local(index1,index2)) {
      source = (char*)*((long*)(CollectionInfo.Table[index] + off));
      CopyFArrayElem(source, (char*)buffer, sizeof(Complex), 
		     which.rank, 
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
      return *(Complex*)buffer;
    }
    where = WhereIsData(index, CollectionNumber);
    AskForFArrayElem(where, index, CollectionNumber, off, (char*)buffer,
		     sizeof(Complex), which.rank,
		     which.i_len, which.i_inc, which.i_off,
		     which.j_len, which.j_inc, which.j_off,
		     which.k_len, which.k_inc, which.k_off,
		     i, j, k);
    return *(Complex*)buffer;
}


