/* --------------------------------------------------
 *
 *  Darryl Brown
 *  
 *  March, 1995
 *
 *  distApi.h
 *  
 *  This file is the include file for the implementation of 
 *  the API for data TYPE transfer above the 
 *  transportation layer (cargo.c and transport.c).
 *
 * --------------------------------------------------
 */


#ifndef DIST_HDR
#define DIST_HDR
#ifdef BRKPNT_ON
#include "clServ.h"
#include "cargo.h"
#include "astrlib.h"

#ifdef __cplusplus 
extern "C" {
#endif

/*
 *  ------------------------------------------------------------
 *  Extern Declarations
 *  ------------------------------------------------------------
 */



/* 
 * --------------------------------------------------------------
 *  MACROS
 * --------------------------------------------------------------
 */

#define BRK_LIST_INSTANCES 1
#define BRK_LIST_TYPES 2
#define BRK_SET_CONTINUE 3
#define BRK_EXIT 4
#define BRK_GET_INSTANCE 5
#define BRK_GET_INSTANCE_TYPE 6
#define BRK_GET_FUNC_ID 7
#define BRK_CALL_USER_FUNC 8
#define BRK_GET_USER_FUNCS 9
#define BRK_CALL_USER_MEM_FUNC 10
#define BRK_GET_USER_MEM_FUNCS 11
#define BRK_CALL_USER_MEM_FUNC_OVER_ALL 12
#define BRK_GET_ALL_ELEMS 13
#define BRK_CALL_USER_MEM_FUNC_IN_PARALLEL 14
#define MAX_LN 100
#define BRK_EXEC_DEF 0

#define DA_MAX_LN 5000
#define CONNECT_ROLE "client"
#define LIST_OUTPUT 0
#define DEBUG 0

#ifndef NULL
#define NULL 0
#endif


/* toggle variable for whether or not to breakpoint */
extern int brk_BREAKPOINT_TOGGLE_VAR;

/* 
 * The variable that specifies whether or not access funcs
 * have been initialized.
 */
extern int DA_AF_Init;

/*
 * The variable that holds the connection number
 */
extern int brk_ConnectionNum;


/*
 * these variables will be set to identify the location
 * before call to generated barrier
 */
extern int brk_funcId;
extern char *brk_fileName;
extern int brk_lineNumber;


extern int DA_Continue;

/*
 * global structure for including pcxx_param,
 * allowing access to environment variables.
 */
extern pcxx_PARAM_t pcxx_param;

extern void brk_AccessFunctionTableInit();

extern connRoleEnum ConnectionRole;

/* 
 * --------------------------------------------------------------
 *  ACCESS FUNCTIONS
 * --------------------------------------------------------------
 */


/*
 *  char *DA_getTypes() -
 *       retrieve all the types for 
 *       the parallel data structures.
 *
 */
aPackage *DA_getTypes();

/*
 *  aPackage *DA_getVarType(strtokStruck *tokStr);
 *       retrieve the type of a particular
 *       variable.
 *
 */
aPackage *DA_getVarType(strtokStruct *tokStr);

/*
 *  aPackage *DA_getInstances();
 *       retrieve all the Instances of 
 *       the parallel data objects. 
 *
 */
aPackage *DA_getInstances(); 

/*
 *  aaPackage *DA_getFuncInfo();
 *       retrieve the function ID
 *       of the current breakpoint.
 *
 */
aPackage *DA_getFuncInfo(); 


/*
 *  void DA_callUserFunc(strtokStruct *tokStr);
 *      calls the access function for the given collection
 *      instance. 
 *
 */
void DA_callUserFunc(strtokStruct *tokStr);

/*
 *  void DA_getUserFuncs()
 *      send a list of the user defined functions
 *
 */
void DA_getUserFuncs();  


/*
 *  void DA_getUserMemFuncs();
 *      send a list of the user defined functions
 *
 */
void DA_getUserMemFuncs();



/*
 *  aPackage *DA_getCollElem(strtokStruct *tokStr)
 *       retrieve a specific collection
 *       or field of that collection.
 *
 */
void DA_AccessCollElem(strtokStruct *tokStr, int accessType); 


/*
 *  aPackage *DA_getRequest()
 *       retrieves the next request from the connection.
 *
 */
aPackage *DA_getRequest();


/*
 *  int DA_requestToInt(char *req)
 *       converts a character request to an integer constant.
 *
 */
int DA_requestToInt(char *req); 

/*
 *  int DA_getContinue(strtokStruct *tokStr)
 *       gets the integer value of the continue parameter
 *       passed in. 
 *
 */
int DA_getContinue(strtokStruct *tokStr);



/*
 *  int DA_beginConnection()
 *       Initiate connection with external application.
 *
 */
int DA_beginConnection();

/*
 *  void DA_ServeRequest()
 *       Serves the next request in the connection.
 *
 */
int DA_ServeRequest();


/*
 *  void DA_cleanup()
 *       Close connection.
 *
 */
void DA_cleanup();


/*
 *  int DA_isValidCollNum(int collnum)
 *       checks to see if collnum is valid 
 *       collection number.
 *
 */
int DA_isValidCollNum(int collnum);


/*----------------------------------------------------------;
*/
void brk_Init ();



/*
 *  void brk_BREAKPOINT_EXECUTIVE();
 *       Function called by parallel program to
 *       answer requests of external application.
 *
 */
void brk_BREAKPOINT_EXECUTIVE();

/*
 *  void brk_CHECK_PARALLEL();
 *       Function called by parallel program to
 *       see if the user wants to execute a func in
 *       parallel.
 *
 */
void brk_CHECK_PARALLEL();

#ifdef __cplusplus 
}
#endif

#else

/*----------------------------------------------------------
 * if breakpoint compiler switch is not on...;
 * get rid of function calls;
 */
#define brk_TOGGLE_BREAKPOINT(a)
#define brk_QUE_COLLECTION_INSTANCE(a,b,c) 
#define brk_ADD_COLLECTION_ENTRY(a,b,c) 
#define brk_DELETE_COLLECTION_ENTRY(a) 
#define brk_BREAKPOINT_EXECUTIVE()
#define brk_CHECK_PARALLEL()

#endif
#endif
