/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#define WHOLE_ELEMENT -1

extern void pcxx_Poll(); //defined here, so the instrumentor can find it

/* This code still gets INT aligned data, " /sizeof(int) " */
/* "size" is size in INTs, **NOT** BYTES */
/* "startof" is size INTs, **NOT** BYTES */
/* pcxx_MsgBlockSize is currently in **INTS** */
void Kernel::Fetch_Element(int elemIndex, int *buf, int startof, int size)
{
  int owner, lastmsgsize, numofblocks, i, totalSize;  /* in INTs */

  /* Find processor that owns the data */
  /* Right now the elements do not migrate at all, so we replace 
     WhereIsData call to reduce communication */
  /* owner = WhereIsData(elemIndex, CollectionNumber); */ 
  owner = CollectionInfo.Where[elemIndex];

  /* Find size of requested transfer */
  if (size == WHOLE_ELEMENT) 
    totalSize = (CollectionInfo.element_size/(int)sizeof(int));
  else
    totalSize = size;

  /* Is requested transfer too big for one message? */
  if (totalSize > pcxx_MsgBlockSize) {
    /* yes, blocking is required */

    lastmsgsize = totalSize % pcxx_MsgBlockSize;
    numofblocks = totalSize / pcxx_MsgBlockSize;

    /* Send "numofblock" messages of size pcxx_MsgBlockSize */
    for (i=0; i<numofblocks; i++) { 
      AskForDataBlock(owner, elemIndex, startof+i*pcxx_MsgBlockSize, 
		      CollectionNumber, pcxx_MsgBlockSize, 
		      buf+i*pcxx_MsgBlockSize); 
    }

    /* Send last remaining message (if any) */
    if (lastmsgsize > 0) {
      AskForDataBlock(owner, elemIndex, startof+numofblocks*pcxx_MsgBlockSize, 
		      CollectionNumber, lastmsgsize,
		      buf+numofblocks*pcxx_MsgBlockSize);
    }
  }
  else
    /* NO blocking is required, just fetch the whole buffer */
    AskForDataBlock(owner, elemIndex, startof, 
		    CollectionNumber, totalSize, buf);
}

int *Kernel::Get_Element(int index)
{
  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_Element(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS(CollectionInfo.element_size);
  if(Is_Local(index)) {
    if(!CollectionInfo.Table[ElemIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_Element()_1D: Error in loading local element.",FATAL);
      printf("Element index is %d, AllocatedProc is %d\n", 
	     index, AllocatedProc(index));
    }
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index)];
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index),buffer, 0,WHOLE_ELEMENT);
  return (int *) buffer;
}

int *Kernel::Get_Element(int index1, int index2)
{
  TRACE_2D_ACS(CollectionInfo.element_size);
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_Element()_2D: Error in loading local element.",FATAL);
      printf("Element index is (%d, %d), AllocatedProc is %d\n", 
	     index1, index2, AllocatedProc(index1, index2));
    }
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index1, index2)];
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index1, index2), buffer, 0, WHOLE_ELEMENT);
  return (int *) buffer;
}

int *Kernel::Get_ElementPart(int index,int startof, int size)
{
  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_ElementPart(Get_2DLowIndex(index), Get_2DHighIndex(index),
			   startof, size);
  }
  TRACE_1D_ACS(size);
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart()_1D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (CollectionInfo.Table[ElemIndex(index)] + startof);
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index),buffer,startof, size);
  return (int *) buffer;
}

int *Kernel::Get_ElementPart(int index1, int index2,int startof, int size)
{
  TRACE_2D_ACS(size);
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart()_2D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (CollectionInfo.Table[ElemIndex(index1, index2)] + startof);
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index1, index2),buffer,startof, size);
  return (int *) buffer;
}

int Kernel::Get_ElementPart_Int(int index, int startof)
{
  TRACE_1D_ACS(sizeof(int));
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Int()_1D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (int)*((int*) (CollectionInfo.Table[ElemIndex(index)] + startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index),buffer,startof,(int)sizeof(int));
  return (int)*((int*) buffer);
}

float Kernel::Get_ElementPart_Float(int index, int startof)
{
  TRACE_1D_ACS(sizeof(float));
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Float()_1D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (float)*((float*) (CollectionInfo.Table[ElemIndex(index)] + startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index),buffer,startof,(int)sizeof(float));
  return (float)*((float*) buffer);
}

double Kernel::Get_ElementPart_Double(int index, int startof)
{
  TRACE_1D_ACS(sizeof(double));
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Double()_1D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(CollectionInfo.Table[ElemIndex(index)]+startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index),buffer,startof,(int)sizeof(double));
  return (double)*((double*) buffer);
}

int Kernel::Get_ElementPart_Int(int index1, int index2, int startof)
{
  TRACE_2D_ACS(sizeof(int));
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Int()_2D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (int)*((int*)(CollectionInfo.Table[ElemIndex(index1, index2)] + 
			  startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index1, index2),buffer,startof,(int)sizeof(int));
  return (int)*((int *)buffer);
}

float Kernel::Get_ElementPart_Float(int index1, int index2, int startof)
{
  TRACE_2D_ACS(sizeof(float));
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Float()_2D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (float)*((float*)(CollectionInfo.Table[ElemIndex(index1, index2)] + 
			  startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index1, index2),buffer,startof,(int)sizeof(float));
  return  (float)*((float*)buffer);
}

double Kernel::Get_ElementPart_Double(int index1, int index2, int startof)
{
  TRACE_2D_ACS(sizeof(double));
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_ElementPart_Double()_2D: Error in loading local element.",
	      FATAL);
    }
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(CollectionInfo.Table[ElemIndex(index1, index2)] + 
			  startof));
  }
  PROFILE_REMOTE_ACS();
  Fetch_Element(ElemIndex(index1, index2),
              buffer,startof,(int)sizeof(double));
  return (double)*((double*)buffer);
}

int *Kernel::Get_CopyElem(int index)
{
  int *NBuf;

  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_CopyElem(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS(CollectionInfo.element_size);
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_CopyElem()_1D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index)];
  }
  
  PROFILE_REMOTE_ACS();
  NBuf = (int *) new double[CollectionInfo.element_size];
  Fetch_Element(ElemIndex(index), NBuf, 0, WHOLE_ELEMENT);
  return NBuf;
}

int *Kernel::Get_CopyElem(int index1, int index2)
{
  int *NBuf;

  TRACE_2D_ACS(CollectionInfo.element_size);
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)]) {
      printf("Processor %d : ", pcxx_NodeNumber());
      Message("Get_CopyElem()_2D: Error in loading local element.",FATAL);
    }
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index1, index2)];
  }

  PROFILE_REMOTE_ACS();
  NBuf = (int *) new double[CollectionInfo.element_size];
  Fetch_Element(ElemIndex(index1,index2),NBuf, 0,WHOLE_ELEMENT);
  return NBuf;
}

int Kernel :: Install_Collection(Distribution *, Align *A, int sizeelem)
{
  int i, j;
  CollectionNumber = GetNumberOfCollection();
  IncNumberOfCollection();

  brk_TOGGLE_BREAKPOINT(0);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_INSTL_COLL, CollectionNumber);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_1, CollectionInfo.number_dimension);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_2, CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_3, CollectionInfo.nb_element);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_4, CollectionInfo.dim1size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_5, CollectionInfo.dim2size);

  PROFILE_COLL_INIT(A->nbelement, CollectionInfo.number_dimension, sizeelem);

  ElementLink = new int* [A->nbelement];
  CollectionInfo.Table = new int * [A->nbelement];
  CollectionInfo.Where = new int[A->nbelement];
  //pass address of collection to lower level runtime system, KernelCom
  SetTableOfCollection(CollectionNumber,(char *) &CollectionInfo);
  buffer = (int *) new double[sizeelem/(int)sizeof(double)+4];
  if (A->nbelement == 0)
    Message("Install_Collection(): No element in collection.",FATAL);
  
  if (CollectionInfo.number_dimension == 1) {
    for(i = 0; i < CollectionInfo.dim1size; i++) {
      if(AllocatedProc(i) == pcxx_NodeNumber()) {
	CollectionInfo.Table[ElemIndex(i)] = (int*) 1;
	CollectionInfo.Where[ElemIndex(i)] = pcxx_NodeNumber(); 
	PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, ElemIndex(i));
      } else {
	CollectionInfo.Table[ElemIndex(i)] = (int *) 0;
	CollectionInfo.Where[ElemIndex(i)] = AllocatedProc(i);
      }
    }
  } 
  
  if (CollectionInfo.number_dimension == 2) {
    for (i=0; i < CollectionInfo.dim1size; i++) {
      for (j=0; j < CollectionInfo.dim2size; j++) {
	if (AllocatedProc(i, j) == pcxx_NodeNumber()) {
	  CollectionInfo.Table[ElemIndex(i,j)] = (int*) 1; 
	  CollectionInfo.Where[ElemIndex(i,j)] = pcxx_NodeNumber();
	  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, ElemIndex(i,j));
	} else {
	  CollectionInfo.Table[ElemIndex(i,j)] = (int *) 0;
	  CollectionInfo.Where[ElemIndex(i,j)] = AllocatedProc(i,j);
	}
      }
    }
  }
 
  if(CollectionInfo.number_dimension > 2) {
    Message("Install_Collection(): 3D implementation not available.",FATAL);
  }
  
  /* turn breakpointing on now that collection is initialized */
  brk_TOGGLE_BREAKPOINT(1);
  brk_ADD_COLLECTION_ENTRY(A->nbelement, sizeelem, this);
  return 1; 
}

void Kernel::~Kernel()
{
  delete [] buffer;
  delete [] CollectionInfo.Table;
  delete [] CollectionInfo.Where;
  SetTableOfCollection(CollectionNumber, (char *) 0);
}
