/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


void  Kernel ::Fetch_Element(int index, int *buf, int startof, int size)
{
  int cs;
  int *pt;
  
  pt = (int*)CollectionInfo.Table[index] + startof;
  
  if (!pt)
    Message("Element not found",WARNING);

  cs = size;
  if (size == -1)
    cs = CollectionInfo.element_size/sizeof(int)+4;
  for(int i = 0; i < cs; i++)
    buf[i] = CollectionInfo.Table[index][startof+i];
}

int *Kernel::Get_Element(int index)
{
  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_Element(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS(CollectionInfo.element_size);
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)])
      Message("Get_Element_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index)];
  } else { 
    PROFILE_REMOTE_ACS();
//    return (int *) CollectionInfo.Where[ElemIndex(index)];
    return (int *) CollectionInfo.Table[ElemIndex(index)];
  }
}

int *Kernel::Get_Element(int index1, int index2)
{
  TRACE_2D_ACS(CollectionInfo.element_size);
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)])
      Message("Get_Element_2D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index1, index2)];
  } else {
    PROFILE_REMOTE_ACS();
//    return (int *) CollectionInfo.Where[ElemIndex(index1, index2)];
    return CollectionInfo.Table[ElemIndex(index1, index2)];
  }
}

int *Kernel::Get_ElementPart(int index,int startof, int size)
{
  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_ElementPart(Get_2DLowIndex(index), Get_2DHighIndex(index),
			   startof, size);
  }
  TRACE_1D_ACS(size);
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)])
      Message("Get_ElementPart_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (CollectionInfo.Table[ElemIndex(index)] + startof);
  }else {
    PROFILE_REMOTE_ACS();
//    return ((int *) CollectionInfo.Where[ElemIndex(index)]) + startof ;
    return (int *) (CollectionInfo.Table[ElemIndex(index)] + startof);
  }
}

int *Kernel::Get_ElementPart(int index1, int index2, int startof, int size)
{
  TRACE_2D_ACS(size);
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)])
      Message("Get_ElementPart_2D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (CollectionInfo.Table[ElemIndex(index1, index2)] + startof);
  }else {
    PROFILE_REMOTE_ACS();
//    return ((int *) CollectionInfo.Where[ElemIndex(index1, index2)]) + startof ;
    return (CollectionInfo.Table[ElemIndex(index1, index2)] + startof);
  }
}

int Kernel::Get_ElementPart_Int(int index,int startof)
{
  TRACE_1D_ACS(sizeof(int));
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)])
      Message("Get_ElementPart_Int_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (int)*((int*)(CollectionInfo.Table[ElemIndex(index)] + startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (int)*((int*)(CollectionInfo.Where[ElemIndex(index)] + startof)) ;
    return (int)*((int*)(CollectionInfo.Table[ElemIndex(index)] + startof));
  }
}

float Kernel::Get_ElementPart_Float(int index,int startof)
{
  TRACE_1D_ACS(sizeof(float));
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)])
      Message("Get_ElementPart_Float_1D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (float)*((float*)(CollectionInfo.Table[ElemIndex(index)] + startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (float)*((float*)(CollectionInfo.Where[ElemIndex(index)] + startof)) ;
    return (float)*((float*)(CollectionInfo.Table[ElemIndex(index)] + startof));
  }
}

double Kernel::Get_ElementPart_Double(int index,int startof)
{
  TRACE_1D_ACS(sizeof(double));
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)])
      Message("Get_ElementPart_Double_1D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(CollectionInfo.Table[ElemIndex(index)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (double)*((double*)(CollectionInfo.Where[ElemIndex(index)]+startof)) ;
    return (double)*((double*)(CollectionInfo.Table[ElemIndex(index)]+startof));
  }
}

int Kernel::Get_ElementPart_Int(int index1, int index2, int startof)
{
  TRACE_2D_ACS(sizeof(int));
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)])
      Message("Get_ElementPart_Int_2D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return (int)*((int*)(CollectionInfo.Table[ElemIndex(index1,index2)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (int)*((int*)(CollectionInfo.Where[ElemIndex(index1,index2)]+startof));
    return (int)*((int*)(CollectionInfo.Table[ElemIndex(index1,index2)]+startof));
  }
}

float Kernel::Get_ElementPart_Float(int index1, int index2, int startof)
{
  TRACE_2D_ACS(sizeof(float));
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)])
      Message("Get_ElementPart_Float_2D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (float)*((float*)(CollectionInfo.Table[ElemIndex(index1,index2)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (float)*((float*)(CollectionInfo.Where[ElemIndex(index1,index2)]+startof));
    return (float)*((float*)(CollectionInfo.Table[ElemIndex(index1,index2)]+startof));
  }
}

double Kernel::Get_ElementPart_Double(int index1, int index2, int startof)
{
  TRACE_2D_ACS(sizeof(double));
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)])
      Message("Get_ElementPart_Double_2D: Error in loading local element.",
	      FATAL);
    PROFILE_LOCAL_ACS();
    return (double)*((double*)(CollectionInfo.Table[ElemIndex(index1,index2)]+startof));
  }else {
    PROFILE_REMOTE_ACS();
//    return (double)*((double*)(CollectionInfo.Where[ElemIndex(index1,index2)]+startof));
    return (double)*((double*)(CollectionInfo.Table[ElemIndex(index1,index2)]+startof));
  }
}

int *Kernel::Get_CopyElem(int index)
{
  int *NBuf;

  if (CollectionInfo.number_dimension == 2) { /* special case */
    return Get_CopyElem(Get_2DLowIndex(index), Get_2DHighIndex(index));
  }
  TRACE_1D_ACS(CollectionInfo.element_size);
  if (Is_Local(index)) {
    if (!CollectionInfo.Table[ElemIndex(index)])
      Message("Get_CopyElem_1D: Error in loading local element.",FATAL);
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index)];
  }
  
  PROFILE_REMOTE_ACS();
//  return CollectionInfo.Table[ElemIndex(index)];
  NBuf = new int[(CollectionInfo.element_size/(int)sizeof(int)) + 4];
  Fetch_Element(ElemIndex(index),NBuf, 0, -1);
  return NBuf;
}

int *Kernel::Get_CopyElem(int index1, int index2)
{
  int *NBuf;

  TRACE_2D_ACS(CollectionInfo.element_size);
  if (Is_Local(index1, index2)) {
    if (!CollectionInfo.Table[ElemIndex(index1, index2)])
      Message("Get_CopyElem_2D: Error in loading local element",FATAL);
    PROFILE_LOCAL_ACS();
    return CollectionInfo.Table[ElemIndex(index1, index2)];
  }
  
  PROFILE_REMOTE_ACS();
//  return CollectionInfo.Table[ElemIndex(index1, index2)];
  NBuf = new int[(CollectionInfo.element_size/(int)sizeof(int)) + 4];
  Fetch_Element(ElemIndex(index1,index2),NBuf, 0,-1);
  return NBuf;
}

int Kernel :: Install_Collection(Distribution *, Align *A, int sizeelem)
{
  int i, j;
  int flag;

  brk_TOGGLE_BREAKPOINT(0);
  /* initialize the number of collection */
  CollectionNumber = GetNumberOfCollection();
  IncNumberOfCollection();

  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_INSTL_COLL, CollectionNumber);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_1, CollectionInfo.number_dimension);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_2, CollectionInfo.element_size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_3, CollectionInfo.nb_element);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_4, CollectionInfo.dim1size);
  PCXX_EVENT(PCXX_EC_KERNEL, PCXX_COLL_INFO_5, CollectionInfo.dim2size);

  PROFILE_COLL_INIT(A->nbelement, CollectionInfo.number_dimension, sizeelem);

  if (A->nbelement == 0)
    Message("No Element in the collection",FATAL);

#ifdef _MASTER_SLAVE_

  int **p;
  flag = 1;
  if (pcxx_NodeNumber() == 0) {
    p = new int*[A->nbelement];
    flag = 0;
  }
  pcxx_BroadcastBytes(flag, sizeof(int**), &p); 
  ElementLink = (int**) p;

#else

  int *p;
  flag = 1;
  if(pcxx_NodeNumber() == 0) {
    p = (int*) tulip_SharedMalloc(A->nbelement*(int)sizeof(int *));
    flag = 0;
  }
  pcxx_BroadcastBytes(flag, sizeof(int*), &p); 
  ElementLink = (int**) p;

#endif

  CollectionInfo.Table = new int*[A->nbelement];
  CollectionInfo.Where = new int[A->nbelement];
  SetTableOfCollection(CollectionNumber,(char *) &CollectionInfo);
  /* allocate the buffer */
  buffer = new int[(sizeelem/(int)sizeof(int))  + 4];

  /* -- fill global info table -- */
  if (CollectionInfo.number_dimension == 1) {
    for (i=0; i < CollectionInfo.dim1size; i++) {
      PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, 
                ((AllocatedProc(i)<<24)+ElemIndex(i)));
      if (AllocatedProc(i) == pcxx_NodeNumber()) {
        CollectionInfo.Table[ElemIndex(i)] = (int*) 1;
        CollectionInfo.Where[ElemIndex(i)] = pcxx_NodeNumber();
        /* PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, ElemIndex(i)); */
      } else {
	CollectionInfo.Table[ElemIndex(i)] = (int*) 0;
        CollectionInfo.Where[ElemIndex(i)] = AllocatedProc(i);
      }
    }
  }
  
  if (CollectionInfo.number_dimension == 2) {
    for (i=0; i < CollectionInfo.dim1size; i++) {
      for (j=0; j < CollectionInfo.dim2size; j++) {
        PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, 
                   ((AllocatedProc(i,j)<<24)+ElemIndex(i,j)));
	if (AllocatedProc(i,j) == pcxx_NodeNumber()) {
          CollectionInfo.Table[ElemIndex(i,j)] = (int*) 1;
	  CollectionInfo.Where[ElemIndex(i,j)] = pcxx_NodeNumber();
          /* PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_INFO, ElemIndex(i)); */
        } else {
	  CollectionInfo.Table[ElemIndex(i,j)] = (int*) 0;
	  CollectionInfo.Where[ElemIndex(i,j)] = AllocatedProc(i,j);
	}
      }
    }
  }
  

  /* turn breakpointing on now that collection is initialized */
  brk_TOGGLE_BREAKPOINT(1);
  brk_ADD_COLLECTION_ENTRY(A->nbelement, sizeelem, this);
  return 1; 
}

void Kernel :: ~Kernel()
{
  SetTableOfCollection(CollectionNumber,(char *) 0);
  delete [] buffer;
  delete [] CollectionInfo.Table;
  delete [] CollectionInfo.Where;
}
