/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


# ifndef PROF_INITIALIZED
#   define PROF_INITIALIZED
#   include "profiler_defs.h"

#   ifdef PROFILING_ON

#     ifdef __cplusplus
        extern "C" void perror(const char *);
        extern "C" void exit(int);
#       ifndef __GNUC__
          /* extern "C" int sprintf(char * , const char * ,  ...); */
#       endif
#     endif

P_COLL_DESCR *p_coll_list[MaxCollection];
P_FUNC_DESCR *p_func_list;
char *p_collname, *p_elemname;
P_CQUEUE *p_varname = 0;

int p_curr_func = -1;

class Profiler {
    int parent;
    int ident;
    int numfunc;
    double start;
public:
    Profiler (int id);
    Profiler (int id, char countonly);
    Profiler (int id, int num);
    ~Profiler ();
};

Profiler :: Profiler (int id)
{
  parent  = p_curr_func;
  p_curr_func = ident = id;
  numfunc = 0;
  start   = pcxx_GetUSecD();
}

Profiler :: Profiler (int id, char countonly)
{
  parent  = p_curr_func;
  p_curr_func = ident = id;
  numfunc = -1;
}

Profiler :: Profiler (int id, int num)
{
  int i;

  p_func_list = new P_FUNC_DESCR[num];
  for (i=0; i<num; i++) {
      p_func_list[i].numcalls = 0;
      p_func_list[i].usec     = 0.0;
      p_func_list[i].cumusec  = 0.0;
  }
  parent  = -1;
  p_curr_func = ident = id;
  numfunc = num;
  start   = pcxx_GetUSecD();

  for (i=0; i<MaxCollection; i++) p_coll_list[i] = 0;
}

Profiler :: ~Profiler ()
{
  if ( numfunc < 0 ) {
    // count only destructor
    p_func_list[ident].numcalls++;
    p_curr_func = parent;
  }
  else {
    double duration = pcxx_GetUSecD() - start;
    p_func_list[ident].numcalls++;
    p_func_list[ident].usec    += duration;
    p_func_list[ident].cumusec += duration;

    if ( parent == -1 ) {
        // last exit: dump results
        PROFILE_DUMP(numfunc);
    }
    else {
        // child: restore current func
        p_curr_func = parent;

        // substract child time form parent's time
        p_func_list[parent].usec -= duration;
    }
  }
}

void PROFILE_COLL_NAMES (char *coll, char *elem, char *var) {
  P_CQUEUE *v;

  p_collname = coll;
  p_elemname = elem;

  if ( v = p_varname ) {
    while ( v->next ) v = v->next;
    v = v->next = new P_CQUEUE;
  }
  else {
    v = p_varname = new P_CQUEUE;
  }
  v->name = var;
  v->next = 0;
}
#   endif
# endif
