/* --------------------------------------------------
 *
 *  Darryl Brown
 *  
 *  October, 1994
 *
 *  typeApi.h
 *  
 *  This file is the include file for the implementation of 
 *  the API for data TYPE transfer above the 
 *  transportation layer (cargo.c and transport.c).
 *
 * --------------------------------------------------
 */


#ifndef TAPI_HDR
#define TAPI_HDR

#ifdef __cplusplus 
extern "C" {
#endif
/*
 *  ------------------------------------------------------------
 *  Macros
 *  ------------------------------------------------------------
 */

#define MAXCOMPS 500
#define MAXTYPES 500
#define MAXINST  500
#define MAXFUNCS 500
#define TA_MAX_LN 500
/*
 *  ------------------------------------------------------------
 *  Types
 *  ------------------------------------------------------------
 */


/* ------------------------- */

/* enumeration of all types */
typedef enum {
  t_notype, t_int, t_float, t_char, t_enum, t_class,
  t_struct, t_collection, t_derived_type, t_union,
  t_double, t_void, t_function
} TA_typeEnum; 

/* ------------------------- */

/* enumeration of descriptor types */
typedef enum {
  d_nodescriptor, d_volatile, d_register, d_long, d_short,
  d_signed, d_unsigned
} TA_descriptEnum;

/* ------------------------- */

/* dimensions struct */
typedef struct {
  /* number of dimensions.  If 0, then it is a pointer type */
  int numDims;
  int *sizes;
} TA_dimType;

/* ------------------------- */

/*  API object types */

/* ------------------------- */

/* name entry */
typedef struct TA_neStruct {
  char *name;
  struct TA_neStruct *next;
} TA_nameEntry ;

/* name list */
typedef struct {
  TA_nameEntry *first;
  TA_nameEntry *last;
  int length;
} TA_nameList;

/* name type pair */
typedef struct TA_ntStruct {
  char *name;
  TA_typeEnum thisType;
  int i;
} TA_nameTypePair;

/* ------------------------- */

/* struct for holding type descriptions. */
typedef struct {
  /* name of type (exists only in class, collection, 
     and struct types and typedef'ed typed) */
  char *name;
  /* descriptor that precedes type (if any, exists in simplest type
     such as int, float, etc) */
  TA_descriptEnum desc;
  /* reveals the basic type */  
  TA_typeEnum thisType;
  /* dimensions, if array type */
  TA_dimType *dims;
  /* note that is for pointer type too */
  int isPointer;
} TA_typeDesc;


/* ------------------------- */

/* structure for storing (linked list of) field(s) */
typedef struct TA_fldStruct {
  char *name;
  TA_typeDesc *type;
  struct TA_fldStruct *next;
} TA_field;

/* ------------------------- */

/* structure for storing TA_composite data structures such as classes 
   collections, and structures */
typedef struct {
  char *name;
  TA_nameList *parents;
  TA_field *fields;
  TA_typeEnum type;
} TA_composite;

/* structure for storing function name & ppointer pairs */
typedef struct {
  char *name;
  void (*funcPtr)();
  char *typeName;
} TA_UserFuncEntry;

/* structure for storing function name & ppointer pairs */
typedef struct {
  char *name;
  void (*funcPtr)(void *ptr, int elemNum, char *fname);
  char *typeName;
} TA_AccessFuncEntry;


/*
 *  ------------------------------------------------------------
 *  Extern Declarations
 *  ------------------------------------------------------------
 */


/*  array of MAXCOMPS TA_composite pointers */
extern TA_composite *TA_CompositeList[MAXCOMPS];

/*  array of MAXTYPES type pointers */
extern TA_typeDesc *TA_typeDescList[MAXTYPES];

/*  array of MAXINST of var name & type pairs */
extern TA_nameTypePair *TA_varTypeList[MAXINST];

/*  array of MAXINST of func name & type pairs */
extern TA_nameTypePair *TA_funcTypeList[MAXINST];

/* table of function names and function pointers */
extern TA_AccessFuncEntry *TA_AccessFuncTable[MAXFUNCS];

/* table of function names and function pointers */
extern TA_UserFuncEntry *TA_UserFuncTable[MAXFUNCS];

/* indices for the above lists */
extern int TA_CompListIdx, TA_TypeListIdx, TA_VarTypeIdx,
       TA_FuncTypeIdx, TA_AccessFuncTableIdx, 
       TA_UserFuncTableIdx;


/*
 *  ------------------------------------------------------------
 *  Access Function Declarations
 *  ------------------------------------------------------------
 */


/* 
 *  comp *TA_readComposite(char *buf) - returns TA_composite 
 *       representation of TA_composite in buf.
 *
 */
TA_composite *TA_readComposite(char *buf);

/*
 *
 *  char *TA_writeComposite(TA_composite *comp) - returns string 
 *       representation of TA_composite comp.
 *
 */
char *TA_writeComposite(TA_composite *comp);


/*
 *  comp *TA_findComposite(char *compName) returns a pointer to 
 *       TA_composite structure of type compName (or NULL if none
 *       exists).
 *
 */
TA_composite *TA_findComposite(char *compName);


/*       
 *  int TA_putComposite(TA_composite *comp) - adds comp to global 
 *       TA_composite list.
 *
 */
int TA_putComposite(TA_composite *comp);


/*
 *  int TA_putTypeDesc(TA_typeDesc *ty) - adds ty to global TA_typeDesc list.
 *
 */
int TA_putTypeDesc(TA_typeDesc *ty);



/*
 *  int TA_putVarType(char *name, char *type) - stores
 *       the name and type pair for later retrieval.
 *
 */
int TA_putVarType(char *name, char *type);

/*
 *  char *TA_getFuncType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 */
char *TA_getFuncType(char *name);

  
/*
 *  void TA_putFuncType(char *name, char *type) - stores
 *       the func name and type pair for later retrieval.
 *
 */
int TA_putFuncType(char *name, char *type);

  
/*
 *  int TA_putAccessFunc(char *name, void *funcPtr) - stores
 *       the func name and type pair for later retrieval.
 *
 */
int TA_putAccessFunc(char *name, void (*funcPtr)(void *,
						 int, char *));

/*
 *  int TA_putUserFunc(char *name, void *funcPtr) - stores
 *       the func name and type pair for later retrieval.
 *
 */
int TA_putUserFunc(char *name, void (*funcPtr)());

/*
 *  void TA_callUserFunc(char *name, int argc, char **argv) - 
 *       returns the string representation of the type
 *       that corresponds to name.
 *
 */
void TA_callUserFunc(char *name);

/*
 *  void TA_callAccessFunc(char *name, int argc, char **argv) - 
 *       returns the string representation of the type
 *       that corresponds to name.
 *
 */
void TA_callAccessFunc(char *name, void *collPtr, int elemNum,
		       char *fname);

/*
 *  char *TA_getVarType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 */
char *TA_getVarType(char *name);


/*
 *  TA_typeDesc *TA_findtypeDesc(char *typeName) returns a pointer to 
 *       TA_composite structure of type compName (or NULL if none
 *       exists).
 *
 */
TA_typeDesc *TA_findtypeDesc(char *typeName);


/*
 *  void TA_putType(char *buf) - adds a type string.
 *
 */
int TA_putType(char *buf);

/*
 *  char *TA_getType(char *name) - returns the string 
 *       representation of the type that corresponds
 *       to name.
 *
 */
char *TA_getType(char *name);
  

/*
 *  void TA_putNameType(char *name, char *type) - stores
 *       the name and type pair for later retrieval.
 *
 */
void TA_putNameType(char *name, char *type);

/*
 *  TA_typeDesc *TA_readTypeDesc(char *buf) - returns TA_typeDesc
 *       representation of type description in buf.
 *
 */
TA_typeDesc *TA_readTypeDesc(char *buf);


/*  
 *  char *TA_writeTypeDesc(TA_typeDesc *td) - returns string
 *       representation of type description td.
 *
 */
char *TA_writeTypeDesc(TA_typeDesc *td);



/*
 *  TA_nameEntry *TA_putName(char *name, TA_nameList *nl) - stores
 *       the name in a list.
 *
 */
TA_nameEntry *TA_putName(char *name, TA_nameList *nl);


/*
 *
 *  TA_nameList *TA_createNameList() - creates a TA_nameList
 *       structure, allocating memory.
 *
 */
TA_nameList *TA_createNameList();


/*
 *  TA_nameEntry *TA_pushName(char *name, TA_nameList *nl) - stores
 *       the name in a list.
 *
 */
TA_nameEntry *TA_pushName(char *name, TA_nameList *nl);


/*
 *  char *TA_popName(TA_nameList *nl) - pops 
 *       the top name in the list.
 *
 */
char *TA_popName(TA_nameList *nl);



/*
 *
 *  TA_dimType *TA_createDimType() - creates the TA_dimType
 *       representation of dimensions dt, allocating memory.
 *
 */
TA_dimType *TA_createDimType();

/*
 *
 *  void TA_addDim(TA_dimType *dt, int dim) - adds a dimension
 *       to TA_dimType dt.
 *
 */
void TA_addDim(TA_dimType *dt, int dim);



/*
 *
 *  TA_typeDesc *TA_createTypeDesc() - creates a TA_typeDesc
 *       structure, allocating memory.
 *
 */
TA_typeDesc *TA_createTypeDesc();


/*
 *  int TA_isComposite(TA_typeDesc te) 
 *       returns true if te is TA_composite type.
 */
int TA_isComposite(TA_typeEnum te);


/*
 *
 *  TA_composite TA_createComposite() - creates a TA_composite
 *       structure, allocating memory.
 *
 */
TA_composite *TA_createComposite();



/*
 *
 *  TA_field *TA_createField() - creates a TA_field
 *       structure, allocating memory.
 *
 */
TA_field *TA_createField();

/*
 *  char * TA_typeToFormat(TA_typeEnum ty) - changes TA_typeEnum
 *        representation into a string format used in
 *        printf statements (to print that type).
 *
 */

char * TA_typeToFormat(TA_typeEnum ty);



/*
 * char * TA_typeDescToFormat(TA_typeDesc *tyde) builds the
 *        TA_typeDesc representation into a string format 
 *        used in printf statements (to print that type).
 *
 */

char * TA_typeDescToFormat(TA_typeDesc *tyde);

/*
 *  char * TA_typeToStr (int ty) - changes TA_typeEnum into
 *        string representation  of that type.
 *  
 */

char * TA_typeToStr (int ty);


/*
 *  TA_typeEnum TA_strToType(char *s) - changes string to
 *       corresponding TA_typeEnum representation. 
 *  
 */
TA_typeEnum TA_strToType(char *s);
  

/*
 *  char *TA_writeUserDefMemFuncs() - returns string
 *       representation of all user defined member functions of
 *       element classes, each line separated by a newline.
 *
 */
char *TA_writeUserDefMemFuncs();

/*
 *  char *TA_writeUserDefFuncs() - returns string
 *       listing of all userDefFuncs, seperated by blanks
 *
 */
char *TA_writeUserDefFuncs();



/*  
 *  void TA_readTIFFile()
 *
 *    Reads in a Type Information File that will
 *    be read in by the parallel program during startup
 *    phase.  This type information will then be given
 *    to any applications that will be connecting to the
 *    parallel program for runtime access to the parallel
 *    data.
 *
 */
void TA_readTIFFile();
/* 
 * --------------------------------------------------------------
 *  Extern declarations
 * --------------------------------------------------------------
 */

#ifdef __cplusplus 
}
#endif

#endif
