/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * generic Master-Slave 
 * startup.h
 */
 
extern "C" void  MSNewProc(int);
extern "C" void  MSFinishProc(void);
extern "C" void  pcaw_Barrier();
extern "C" void  pcaw_Barrier_Uninstrumented();
extern "C" int   pcaw_inseq;

char *pcaw_cc = NULL;
int pcaw_cf = 0;

/* 
   NOTE: Until New_Main is called, execution is sequential. So tracing
   can be used because pcxx_NodeNumber will return 0. But after New_Main
   is called and until MSNewProc is completed, the parallel threads
   don't have ids. So putting instrumentation in between these two might
   be a problem. Right now, there is no instrumentation in pcaw_Barrier
   and MSNewProc's first stmt is to set the id. So everything is okay now.
   But one should watch out for problems here -- K7 (11/29/94)
*/

extern "C" void New_Main(int myid){
  if(myid==0){  // I am Main Thread
    pcaw_inseq--;
    pcaw_Barrier_Uninstrumented();
    MSNewProc(myid);
    pcaw_Barrier();
    pcaw_inseq++;

    ResetCollectionTraversalLevel ();

    Processor_Main();

    /* terminate threads */
    pcaw_inseq--;
    pcaw_cc = (char *)0;
    pcaw_Barrier();
    MSFinishProc();
    pcaw_Barrier_Uninstrumented();
    pcaw_inseq++;
  }
  else{
    int done = 0;

    pcaw_Barrier_Uninstrumented();
    MSNewProc(myid);
    pcaw_Barrier();

    while(!done){
      pcaw_Barrier();
      if(pcaw_cc) {
        WorkerSwitch(pcaw_cf, pcaw_cc);
        pcaw_Barrier();
      }
      else{
        done = 1;
        MSFinishProc();
        pcaw_Barrier_Uninstrumented();
      }
    }
  }
}


void pcaw_EnterPar(Kernel *coll, int selector)
{
  if(coll){
    ResetLocal(coll);
  }
  pcxx_inpar++;
  pcaw_inseq--;
  pcaw_cf = selector;
  pcaw_cc = (char *)coll;
  pcaw_Barrier();
  /* join in the work */
  WorkerSwitch(selector, (char *)coll);
}

void pcaw_SubsetEnterPar(Kernel *coll, int selector)
{
  pcaw_EnterPar(coll, selector);
}

void pcaw_ExitPar()
{
  pcaw_Barrier();  
  pcaw_inseq++;
  pcxx_inpar--;
  ResetCollectionTraversalLevel();
}

