/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/***************************************************************************
 *                pC++ Routine for managing access to data                 *
 *                PVM3 Version                                             *
 *                                                                         *
 *                Written by Pete Beckman (11/4/93)                        *
 *                based on SP1 (P. Beckman) version                        *
 *                based on paragon (D. Gannon) version                     *
 *                based on cm5 (F. Bodin)                                  *
 *                Last overhaul by S. Yang (2/11/94)                       *
 ***************************************************************************/

#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

#include <string.h>

#ifndef NeXT
#  include <unistd.h>
#endif

#include "passive.h"

extern pcxx_PARAM_t pcxx_param;

/**************************************************************************/
/**************************************************************************/
/* MACHINE DEPENDANT CODE */
/**************************************************************************/
/**************************************************************************/
#include "pvm3.h"

int MyTaskIds[256];

/**************************************************************************/
/**************************************************************************/
/* Machine Dependant Message Prototypes:
/* PVM Prototypes */

/*******************************************************************/
void pcxx_Abort()
{
  pvm_exit();
  exit(1);
}

/*******************************************************************/
/* Receive a buffer that has already arrived and is waiting */
/* Return the message type that was encoded in the message */
void MyRecv(int *type, int *from)
{
  int bufid, info, len;

  bufid = pvm_recv(DONTCARE, DONTCARE);  /* Get any message, Any! */
  if (bufid < 0)  Message("MyRecv: pvm_recv failed",FATAL);

  /* OK, pvm has a buffer for me, now I must unpack it! */
  info = pvm_upkint(type,1,1);                /* Message type */
  if (info < 0) Message("MyRecv: pvm_upint (0) failed",FATAL);
  info = pvm_upkint(from,1,1);   /* From Whom? */
  if (info < 0) Message("MyRecv: pvm_upint (1) failed",FATAL);
  info = pvm_upkint(&len,1,1);                /* Length of bytecoded msg */
  if (info < 0) Message("MyRecv: pvm_upint (2) failed",FATAL);

  /* EXCEPTION TO THE RULE... */
  /* If the type==EndOfBlockMsg, unpack DIRECTLY into BlockDataBuff */
  if (*type == EndOfBlockMsg || *type == BroadcastMsg) {
    info = pvm_upkbyte((char *) BlockDataBuff,len,1);
    if (info < 0) Message("MyRecv: pvm_upkbyte (1) failed",FATAL);
  }
  else {
    info = pvm_upkbyte((char *) ReceiveBuffer,len,1);
    if (info < 0) Message("MyRecv: pvm_upkbyte (2) failed",FATAL);
  }
}

/*******************************************************************/
/* Send out a buffer to a destination */
void MySend(int type, char * buffer, int len, int ToWhom)
{
  int info,bufid;

  if (VERBOSE>2) 
    printf("Processor %d sending message to %d (type:%d len:%d)\n",
	   MyNodeNumber,ToWhom,type,len);

  /* Clear send buffer */
  bufid = pvm_initsend( PvmDataDefault );
  if (bufid < 0) Message("MySend: pvm_initsend failed",FATAL);
  
  /* Pack in the data */
  info = pvm_pkint(&type,1,1);                /* Message type */
  if (info < 0) Message("MySend: pvm_pint (0) failed",FATAL);
  info = pvm_pkint(&MyNodeNumber,1,1);   /* From Whom? */
  if (info < 0) Message("MySend: pvm_pint (1) failed",FATAL);
  info = pvm_pkint(&len,1,1);                /* Length of bytecoded msg */
  if (info < 0) Message("MySend: pvm_pint (2) failed",FATAL);

  info = pvm_pkbyte(buffer,len,1);
  if (info < 0)   Message("MyRecv: pvm_pkbyte failed",FATAL);
  /* Send data */
  info = pvm_send(MyTaskIds[ToWhom], 0);  /* msgtag is 0, no tags... */  
  if (info < 0) Message("MySend: pvm_send failed",FATAL);
}

/* **********************************************************************/
void pcxx_MainBegin(int *argc, char *argv[])
{
  int mytid, me, i, x;
  int info;
  char *pwd;
  char *DelimiterPos;
  char basename[100];

  int nhost, narch;
  struct pvmhostinfo *hostp;
  
  pcxx_InitParam (argc, argv);
  pcxx_InitClocks();

/*
  int procstouse;
  sscanf(argv[1],"%d",&procstouse);
  printf("Using %d\n",procstouse);
*/

  /* enroll in pvm */
  mytid = pvm_mytid();

  info = pvm_config(&nhost, &narch, &hostp);
  if (info < 0) printf("Error: pvm_config failed \n");

/*
  if (procstouse > nhost) exit(1);
  nhost = procstouse;
*/

  /* find out if I am parent or child */
  MyTaskIds[0] = pvm_parent();
  if( MyTaskIds[0] < 0 ) {       /* then I am the parent */
    MyTaskIds[0] = mytid;
    me = 0;

    printf("There are %d hosts in this PVM configuration\n",nhost);
    printf("Main Thread Started : %20s:  MyTaskIds[%2d]=%8x\n"
	   ,hostp[0].hi_name,0,MyTaskIds[0]);

    /* Find out the name of this binary... */
    DelimiterPos = (char *) strrchr(argv[0],47);  /* find last "/" */
    if (DelimiterPos) {  /* There was a path, remove it for PVM calls */
      strcat(basename,DelimiterPos);
    }
    else /* there was no path, just use executable name */
      strcat(basename,argv[0]);

    /* start up copies of myself */
    for (x=1; x<nhost; x++)
      {
      i = pvm_spawn(basename, &(argv[1]), 1, 
		    hostp[x].hi_name, 1, &MyTaskIds[x]);
      if (i != 1) printf("Error, could not spawn task onto %s (error=%d)\n",
			 hostp[x].hi_name,i);

      printf("Spawned 'main()' onto %20s:  MyTaskIds[%2d]=%8x\n"
	     ,hostp[x].hi_name,x,MyTaskIds[x]);
    }

    /* multicast MyTaskIds array to children */
    pvm_initsend( PvmDataDefault );
    pvm_pkint(MyTaskIds, nhost, 1);
    printf("Broadcasting MyTaskIds data (mcast)...\n");
    for (x=1; x<=nhost-1; x++) {
      i = pvm_send(MyTaskIds[x], 0);
      if (i < 0)
	printf("ERROR pvm_send failed\n");
    }
  }

  else {    /* I am a child */
    /* receive MyTaskIds array */
    printf("Spawned task receiving MyTaskIds %x\n",&MyTaskIds[0]);
    i = pvm_recv(MyTaskIds[0], 0);
    pvm_upkint(MyTaskIds, nhost, 1);
    for( i=1; i<nhost ; i++ )
      if( mytid == MyTaskIds[i] ){ me = i; break; }
  }

  /* 'me' is now set for every processor, and tid[procnum] is the taskid */

  MyNodeNumber = me; MyTotalNodes = /*procstouse*/ nhost;

  /* we would like to do this earlier, but unfortuneately, we need
     MyNodeNumber for doing this */
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  if (me > MyTotalNodes-1) { pvm_exit(); exit(1); }

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  barrierFlag = 0;

  ResetCollectionTraversalLevel ();

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  /* Go to Processor_Main */

  }

/*******************************************************************/
void pcxx_MainEnd() {
  pvm_exit();
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose();
}

/*********************************************************************
 POLL FUNCTION
 *********************************************************************/
void pcxx_Poll()
{
  int info,type,from,i,j,k,b;

  PROFILE_ENTER(PROFILE_Poll);

  /* Has a message come in? */
  if (VERBOSE>4) printf("%d Polling()\n",MyNodeNumber);

  while ((info = pvm_probe(DONTCARE,DONTCARE)) != 0) {
    if (info < 0)
      Message("pcxx_Poll(): pvm_probe error",FATAL);

    /* Hey! get the data! */
    if (VERBOSE>2) printf("Processor %d received a message\n",
			MyNodeNumber);
    MyRecv(&type,&from);

    /* Sanity check, is the packed "from" the same as pvm's */
    b = pvm_bufinfo(info,&i,&j,&k);
    if (b < 0) Message("Poll: pvm_bufinfo failed",FATAL);

    if (MyTaskIds[from] != k)
      Message("Hey! 'from' and pvm_bufinfo don't match",FATAL);

    switch(type){
    case BroadcastMsg:
      broadcast_flag = 1;
      break;
    case ReduceMsg:
      reduction_flag++;
      PROFILE_EXIT(PROFILE_Poll);
      return;
    case HereIsMsg:
      here_is_data_handler(type, from);
      break;
    case WhereIsMsg:
      where_is_data_handler(type, from);
      break;
    case ReadBlockMsg:
      read_block_handler(type, from);
      break;
    case EndOfBlockMsg:
      end_of_block_handler(type, from);
      break;
    case ReadFArrayMsg:
      read_farray_handler(type, from);
      break;
    case ReadFArrayElemMsg:
      read_farray_elem_handler(type, from);
      break;
    case BarrierFanInMsg:
      BarrierFanIn(from);
      break;
    case BarrierFanOutMsg:
      /* There is nothing to "consume", just set flag, I'm done... */
      barrierFlag = DONE;
      break;
    case RemoteExecuteMsg:
      remote_execute_handler(type, from);
      break;
    case EndOfRemoteExecuteMsg:
      end_of_remote_handler(type, from);
      break;
    default:
      fprintf(stderr,"Flushing message %d on node %d\n",type, 
              pcxx_NodeNumber());
    }
  }
  PROFILE_EXIT(PROFILE_Poll);
  return;
}

