/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*/
/* sgilib : SGI master/slave run-time support library       		*/
/*==============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <ulocks.h>
#include <task.h>
#include <sys/sysmp.h>

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */

#define MAXBUFF 2048
#define MAX_PROC 128

extern pcxx_PARAM_t pcxx_param;
extern void New_Main();

char *BroadcastBuff = NULL;         /* a place to point to broadcast values */
char *ReduceBuff[MAX_PROC];         /* a place to store reduction values */

int   MyTotalNodes = -1;            /* number of processors */
int   pcaw_inseq   = 1;

/*
 @ SMRTSSetup () 
 */

int SMRTSSetup ()
{
  int avail;
  int i;

  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);
  
  /* set number of processors */
  
  avail = ((MAX_PROC > sysmp(MP_NAPROCS)) ? sysmp(MP_NAPROCS):MAX_PROC);
  if (pcxx_param.numproc > avail) {
    fprintf (stderr, "warning: only %d cells available\n", avail);
    pcxx_param.numproc = avail;
  }
  MyTotalNodes = pcxx_param.numproc;
  
  /* initialize */ 
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);
  m_set_procs(MyTotalNodes);

  /* allocate shared reduction buffer */
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) malloc(MAXBUFF);
  }
  
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return (0);
}

/*
 @ MSNewProc : more book-keeping for newly-created process
 */

void MSNewProc (void)
{
  int id=pcxx_NodeNumber();
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, id);
}


/*
 @ MSFinishProc : book-keeping for terminating process
 */

void MSFinishProc (void)
{
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose ();
}


/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork (argc, argv)
int argc;
char **argv;
{
  int i;

  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  /* uniproc version of timer and event-stuff */
  pcxx_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);

  New_Main();

  /* terminate the threads */
  m_kill_procs();
  
  /* finished */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return (0);
}


/*
 @ pcxx_NodeNumber : which processor I am
 */

int pcxx_NodeNumber()
{
  int my_id = 0;

  if(!pcaw_inseq) {
    my_id = m_get_myid();
  }
  return my_id;
}

/*
 @ IamtheMainThread 
 */

int IamtheMainThread()
{
  return pcaw_inseq;
}

/*
 barrier : wait at general barrier, must be called only in parallel sections
 */
void pcaw_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  m_sync();
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/* Can be used even in sequential sections */
void pcxx_Barrier()
{
  if(!pcaw_inseq)
    pcaw_Barrier();
}

/*
 @ pcxx_BroadcastBytes 
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive pcxx_BroadcastBytes are called */
  pcxx_Barrier();

  p = (char *) buffer;

  if (source == 0) BroadcastBuff = buffer;
  pcxx_Barrier();
  if (source != 0){
    c = BroadcastBuff;
    for (i=0; i<length; i++) p[i] = c[i];
  }
}

