/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1994                  */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * pcxx_event.c: simple SW monitor routines
 *
 * (c) 1994 Jerry Manic Saftware
 *
 * Version 3.0
 */

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <sys/types.h>
# include <fcntl.h>
# include <signal.h>

#   if !defined(__ksr__) || defined(UNIPROC)
#     define __private
#   endif

# define PCXX_EVENT_SRC
# define TRACING_ON
# include "pcxx_events.h"
# include "pcxx_timers.h"

extern time_t time PARAMS((time_t * t));

__private unsigned long int pcxx_ev_class = PCXX_EC_TRACER | PCXX_EC_TIMER;

# ifndef TRUE
# define FALSE 0
# define TRUE  1
# endif

/* -- event record buffer ------------------------------------ */
__private static PCXX_EV pcxx_buf[PCXX_BUFSIZE];

/* -- pointer to next free element of event record buffer ---- */
__private PCXX_EV *pcxx_ev_ptr = pcxx_buf;

/* -- pointer to last available element of event record buffer */
/* -- need one place for flush event => - 1 ------------------ */
__private PCXX_EV *pcxx_ev_max = pcxx_buf + PCXX_BUFSIZE - 1;

/* -- event trace file descriptor ---------------------------- */
__private static int pcxx_fd;

/* -- initialization status flags ---------------------------- */
__private static int pcxx_num_init[PCXX_MAXPROCS];
__private static int pcxx_not_first[PCXX_MAXPROCS];

/* -- write event to buffer only [without overflow check] ---- */
static void DECLFUN(pcxx_EventOnly, (ev, par), long int ev AND long int par)
{
  pcxx_ev_ptr->ev   = ev;
  pcxx_ev_ptr->ti   = pcxx_GetUSecL();
  pcxx_ev_ptr->par  = par;
# ifdef UNIPROC
  pcxx_ev_ptr->nid  = PCXX_MYNODE;
# endif
  pcxx_ev_ptr->tid  = PCXX_MYTHREAD;
  pcxx_ev_ptr++;
}

/* -- write event buffer to file ----------------------------- */
void DECLFUN_VOID(pcxx_EvFlush)
{
  static PCXX_EV flush_end = { PCXX_EV_FLUSH_EXIT, 0, 0, 0L };

  if ( pcxx_ev_ptr != pcxx_buf )
  {
    if ( pcxx_ev_class & PCXX_EC_TRACER )
      pcxx_EventOnly (PCXX_EV_FLUSH_ENTER, pcxx_ev_ptr - pcxx_buf);

    write (pcxx_fd, pcxx_buf, (pcxx_ev_ptr - pcxx_buf) * sizeof(PCXX_EV));
    if ( pcxx_ev_class & PCXX_EC_TRACER )
    {
# ifdef UNIPROC
      flush_end.nid = PCXX_MYNODE;
# endif
      flush_end.ti  = pcxx_GetUSecL();
      write (pcxx_fd, &flush_end, sizeof(PCXX_EV));
    }
    pcxx_ev_ptr = pcxx_buf;
  }
}

/* -- signal catching to flush event buffers ----------------- */
# ifndef NSIG
#   define NSIG 32
# endif
static SIGNAL_TYPE (*sighdlr[NSIG])(SIGNAL_ARG_TYPE);

static void DECLFUN(wrap_up, (sig), int sig)
{
  fprintf (stderr, "signal %d on %d - flushing event buffer...\n", sig, PCXX_MYNODE);
  pcxx_EvFlush ();
  fprintf (stderr, "done.\n");
  if ( sighdlr[sig] != SIG_IGN ) (* sighdlr)(sig);
  exit (1);
}

static void DECLFUN_VOID(init_wrap_up)
{
# ifdef SIGINT
  sighdlr[SIGINT ] = signal (SIGINT , wrap_up);
# endif
# ifdef SIGQUIT
  sighdlr[SIGQUIT] = signal (SIGQUIT, wrap_up);
# endif
# ifdef SIGILL
  sighdlr[SIGILL ] = signal (SIGILL , wrap_up);
# endif
# ifdef SIGFPE
  sighdlr[SIGFPE ] = signal (SIGFPE , wrap_up);
# endif
# ifdef SIGBUS
  sighdlr[SIGBUS ] = signal (SIGBUS , wrap_up);
# endif
# ifdef SIGTERM
  sighdlr[SIGTERM] = signal (SIGTERM, wrap_up);
# endif
# ifdef SIGABRT
  sighdlr[SIGABRT] = signal (SIGABRT, wrap_up);
# endif
# ifdef SIGSEGV
  sighdlr[SIGSEGV] = signal (SIGSEGV, wrap_up);
# endif
}

/* -- initialize SW monitor and open trace file(s) ----------- */
void DECLFUN(pcxx_EvInit, (name), char *name)
{
  char *ptr;
# ifndef UNIPROC
  char *ptr1;
# endif

  if ( pcxx_num_init[PCXX_MYNODE] == 0 )
  {
# ifdef UNIPROC
    ptr = name;
# else
    ptr = (char *) PCXX_MALLOC (strlen(name) + 1);
    strcpy (ptr, name);
    if ( ptr1 = strchr (ptr, '#') )
    {
      *ptr1++ = PCXX_MYNODE / 1000 + '0';
      *ptr1++ = PCXX_MYNODE % 1000 / 100 + '0';
      *ptr1++ = PCXX_MYNODE % 100 / 10 + '0';
      *ptr1   = PCXX_MYNODE % 10 + '0';
    }
    else
    {
      fprintf (stderr, "%s: trace file name does not contain '####'\n", name);
      exit (1);
    }
# endif

    if ( pcxx_not_first[PCXX_MYNODE] )
    {
      if ((pcxx_fd = open (ptr, O_WRONLY|O_APPEND, 0644)) < 0)
      {
        fprintf (stderr, "pcxx_EvInit[reopen]: ");
        perror (ptr);
        exit (1);
      }
    }
    else
    {
      pcxx_InitClocks();
      init_wrap_up ();

      if ((pcxx_fd = open (ptr, O_WRONLY|O_CREAT|O_TRUNC|O_APPEND, 0600)) < 0)
      {
        fprintf (stderr, "pcxx_EvInit[open]: ");
        perror (ptr);
        exit (1);
      }
      pcxx_ev_ptr = pcxx_buf;

# ifndef UNIPROC
      /* -- store nid for merge -- */
      pcxx_ev_ptr->nid = PCXX_MYNODE;
# endif
    }

    /* -- store in any case init event. it is needed for merge -- */
    pcxx_Event (PCXX_EV_INIT, pcxx_ev_class);

    if ( pcxx_ev_class & PCXX_EC_TRACER )
      pcxx_Event (PCXX_EV_WALL_CLOCK, time((time_t *)0));
  }
  pcxx_num_init[PCXX_MYNODE]++;
  pcxx_not_first[PCXX_MYNODE] = TRUE;
}

/* -- write event to buffer ---------------------------------- */
void DECLFUN(pcxx_Event, (ev, par), long int ev AND long int par)
{
  pcxx_ev_ptr->ev   = ev;
  pcxx_ev_ptr->ti   = pcxx_GetUSecL();
  pcxx_ev_ptr->par  = par;
# ifdef UNIPROC
  pcxx_ev_ptr->nid  = PCXX_MYNODE;
# endif
  pcxx_ev_ptr->tid  = PCXX_MYTHREAD;
  pcxx_ev_ptr++;

  if ( pcxx_ev_ptr >= pcxx_ev_max ) pcxx_EvFlush (); 
}

/* -- terminate SW tracing ----------------------------------- */
void DECLFUN_VOID(pcxx_EvClose)
{
  pcxx_num_init[PCXX_MYNODE]--;
  if ( pcxx_num_init[PCXX_MYNODE] == 0 )
  {
    if ( pcxx_ev_class & PCXX_EC_TRACER )
    {
      pcxx_Event (PCXX_EV_CLOSE, 0);
      pcxx_Event (PCXX_EV_WALL_CLOCK, time((time_t *)0));
    }
    pcxx_EvFlush ();
    close (pcxx_fd);
  }
}

/* -- write long event to buffer ----------------------------- */
# define CONT_EV_LEN (sizeof(PCXX_EV) - sizeof(long int))

void DECLFUN(pcxx_LongEvent, (ev, ln, par), long int ev AND int ln AND char *par)
{
  char *buf;
  int i, j, cev_no;

  cev_no = ln / CONT_EV_LEN + ((ln % CONT_EV_LEN) > 0);

  /* -- inlined pcxx_Event (ev, cev_no); ----------------------- */
  pcxx_ev_ptr->ev   = ev;
  pcxx_ev_ptr->ti   = pcxx_GetUSecL();
  pcxx_ev_ptr->par  = ln;
# ifdef UNIPROC
  pcxx_ev_ptr->nid  = PCXX_MYNODE;
# endif
  pcxx_ev_ptr->tid  = PCXX_MYTHREAD;
  pcxx_ev_ptr++;
  if ( pcxx_ev_ptr >= pcxx_ev_max ) pcxx_EvFlush (); 

  /* -- inlined pcxx_Event (ev, cev_no); ----------------------- */
  if ( ln )
  {
    for (i=0; i<cev_no; i++)
    {
      pcxx_ev_ptr->ev = PCXX_EV_CONT_EVENT;
      buf = ((char *) pcxx_ev_ptr) + sizeof(short unsigned int);
      for (j=0; j<CONT_EV_LEN; j++)
        *buf++ = (i*CONT_EV_LEN+j) < ln ? *par++ : '\0';
      pcxx_ev_ptr++;
      if ( pcxx_ev_ptr >= pcxx_ev_max ) pcxx_EvFlush (); 
    }
  }
}
