/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

# include <stdio.h>
# include "GlobalStruct.h"
# include "pcxx_events_def.h"
# include "pcxx_ansi.h"

#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif

# define DEF_NUMPROC   16

# ifdef _MASTER_SLAVE_
# define DEF_NUMCPUS    1
# define DEF_SHMEMSIZE  1000000
# define DEF_STACKSIZE  100000
# endif

# ifdef UNIPROC
#   define DEF_TRACEFILE "uniproc.trc"
# else
#   if defined(_SEQUENT_) || defined(sequent)
#     define DEF_TRACEFILE "symmetry.####.trc"
#   else
#     ifdef butterfly
#       define DEF_TRACEFILE "tc2000.####.trc"
#     else
#       ifdef __ksr__
#         define DEF_TRACEFILE "ksr1.####.trc"
#       else
#         ifdef __PARAGON__
#           define DEF_TRACEFILE "paragon.####.trc"
#         else
#           ifdef __SP1__
#             define DEF_TRACEFILE "sp1.####.trc"
#           else
#             ifdef CM5
#               define DEF_TRACEFILE "cm5.####.trc"
#             else
#               ifdef __PVM__
#                 define DEF_TRACEFILE "pvm.####.trc"
#               else
#                 ifdef __MPI__
#                   define DEF_TRACEFILE "mpi.####.trc"
#                 else
#                   ifdef __SGIMP__
#                     define DEF_TRACEFILE "sgi.####.trc"
#                   else
#                     ifdef _MASTER_SLAVE_
#                       define DEF_TRACEFILE "ms.trc"
#                     else
#                       define DEF_TRACEFILE "blurb.####.trc"
#                     endif
#                   endif
#                 endif
#               endif
#             endif
#           endif
#         endif
#       endif
#     endif
#   endif
# endif

# define GetNextArg(variable,str) \
  if ( (i+1) == *argc )	  \
    fprintf (stderr, "%s: missing argument for %s - ignored.\n", argv[0], str);\
  else			  \
  {			  \
    found = 1;		  \
    variable = argv[i+1]; \
  }

extern char *getenv();

pcxx_PARAM_t pcxx_param;

# ifdef TRACING_ON
/* -- parse event class descriptor string and set eventclass - */
static void DECLFUN(pcxx_SetAndParseEvclass, (prog, str), char *prog AND char *str)
{
  char *end;

  if ( str )
  {
    pcxx_param.eventclass = PCXX_EC_NULL;
    while ( str && *str )
    {
      if ( (end = strchr (str, '+')) != NULL ) *end = '\0';

      switch ( str[0] )
      {
        case 'A':
        case 'a': pcxx_param.eventclass |= PCXX_EC_ALL;
                  break;
        case 'B':
        case 'b': pcxx_param.eventclass |= PCXX_EC_BASIC;
                  break;
        case 'E':
        case 'e': pcxx_param.eventclass |= PCXX_EC_KERNEL;
                  pcxx_param.eventclass |= PCXX_EC_SMODE;
                  break;
        case 'K':
        case 'k': pcxx_param.eventclass |= PCXX_EC_KERNEL;
                  pcxx_param.eventclass |= PCXX_EC_SMODE_ON;
                  break;
        case 'P':
        case 'p': pcxx_param.eventclass |= PCXX_EC_PROFILER;
                  break;
        case 'R':
        case 'r': pcxx_param.eventclass |= PCXX_EC_RUNTIME;
                  break;
        case 'T':
        case 't': switch ( str[1] )
                  {
                    case 'I':
                    case 'i': pcxx_param.eventclass |= PCXX_EC_TIMER;
                              break;
                    case 'R':
                    case 'r': pcxx_param.eventclass |= PCXX_EC_TRACER;
                              break;
                    default : goto invalid;
                              /*break;*/
                  }
                  break;
        case 'U':
        case 'u': pcxx_param.eventclass |= PCXX_EC_USER;
                  break;
        case '1': pcxx_param.eventclass |= PCXX_EC_USER1;
                  break;
        case '2': pcxx_param.eventclass |= PCXX_EC_USER2;
                  break;
        case '3': pcxx_param.eventclass |= PCXX_EC_USER3;
                  break;
        case '4': pcxx_param.eventclass |= PCXX_EC_USER4;
                  break;
        invalid :
        default :
                  printf ("%s: invalid event class %s - ignored.\n",
                          prog, str);
                  break;
      }

      if ( (str = end) != NULL ) *str++ = '+';
    }
  }
  else
    pcxx_param.eventclass = PCXX_EC_PROFILER | PCXX_EC_BASIC | PCXX_EC_TIMER;

  pcxx_ev_class = pcxx_param.eventclass;
}
# endif

int DECLFUN(pcxx_InitParam, (argc, argv), int *argc AND char *argv[])
{
  char *procstr   = NULL;
  char *eventstr  = NULL;
  char *filestr   = NULL;
  char *socketStr = NULL;
  char *controlSocketStr = NULL;
  char *hostStr = NULL;
/* for master slave thread systems: */
  char *cpusstr = NULL;
  char *shmemsizestr = NULL;
  char *stacksizestr = NULL;

  int i, j, found;

  for (i=1; i<*argc; i++)
  {
    if ( strncmp (argv[i], "-pcxx_", 6) == 0 )
    {
      found = 0;

      if ( strcmp (argv[i], "-pcxx_NUMPROC") == 0 )
      {
        GetNextArg (procstr, "pcxx_NUMPROC");
#       if !defined(_SMRTS_) && !defined(_MASTER_SLAVE_)
        fprintf (stderr, "warning: numproc ignored on non SMRTS machines.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_EVENTCLASS") == 0 )
      {
        GetNextArg (eventstr, "pcxx_EVENTCLASS");
#       ifndef TRACING_ON
        fprintf (stderr, "warning: no tracing - event class ignored.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_TRACEFILE") == 0 )
      {
        GetNextArg (filestr, "pcxx_TRACEFILE");
#       ifndef TRACING_ON
        fprintf (stderr, "warning: no tracing - trace filename ignored.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_NUMCPUS") == 0 )
      {
        GetNextArg (cpusstr, "pcxx_NUMCPUS");
#       ifndef _MASTER_SLAVE_
        fprintf (stderr, "warning: numcpus ignored on non M/S versions.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_SHMEMSIZE") == 0 )
      {
        GetNextArg (shmemsizestr, "pcxx_SHMEMSIZE");
#       ifndef _MASTER_SLAVE_
        fprintf (stderr, "warning: shmemsize ignored on non M/S versions.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_STACKSIZE") == 0 )
      {
        GetNextArg (stacksizestr, "pcxx_STACKSIZE");
#       ifndef _MASTER_SLAVE_
        fprintf (stderr, "warning: stacksize ignored on non M/S versions.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_BRKSOCKET") == 0 )
      {
        GetNextArg (socketStr, "pcxx_BRKSOCKET");
#       ifndef BRKPNT_ON
        fprintf (stderr, 
		 "warning: breakpoint executive off - socket num ignored.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_BRKCTRLSOCKET") == 0 )
      {
        GetNextArg (controlSocketStr, "pcxx_BRKCTRLSOCKET");
#       ifndef BRKPNT_ON
        fprintf (stderr, 
		 "warning: breakpoint executive off - control socket num ignored.\n");
#       endif
      }
      else if ( strcmp (argv[i], "-pcxx_BRKHOST") == 0 )
      {
        GetNextArg (hostStr, "pcxx_BRKHOST");
#       ifndef BRKPNT_ON
        fprintf (stderr, 
		 "warning: breakpoint executive off - host address ignored.\n");
#       endif
      }

      if ( found )
      {
        *argc -= 2;
        for (j=i; j<*argc; j++) argv[j] = argv[j+2];
        i--;
      }
    }
  }

  if ( (procstr == NULL) && ((procstr = getenv ("pcxx_NUMPROC")) == NULL) )
    pcxx_param.numproc = DEF_NUMPROC;
  else if ( (pcxx_param.numproc = atoi(procstr)) <= 0 )
  {
    fprintf (stderr, "%s: invalid number of procs %s - ignored.\n",
             argv[0], procstr);
    pcxx_param.numproc = DEF_NUMPROC;
  }

# ifdef TRACING_ON
  if ( (filestr == NULL) && ((filestr = getenv ("pcxx_TRACEFILE")) == NULL) )
    pcxx_param.tracefile = DEF_TRACEFILE;
  else
    pcxx_param.tracefile = filestr;

  if ( eventstr == NULL ) eventstr = getenv ("pcxx_EVENTCLASS");
  pcxx_SetAndParseEvclass (argv[0], eventstr);
# endif

# ifdef _MASTER_SLAVE_
  if ( (cpusstr == NULL) && ((cpusstr = getenv ("pcxx_NUMCPUS")) == NULL) )
    pcxx_param.numcpus = DEF_NUMCPUS;
  else if ( (pcxx_param.numcpus = atoi(cpusstr)) <= 0 )
  {
    fprintf (stderr, "%s: invalid number of cpus %s - ignored.\n", argv[0],cpusstr);
    pcxx_param.numcpus = DEF_NUMCPUS;
  }
  if ( (shmemsizestr == NULL) && ((shmemsizestr = getenv ("pcxx_SHMEMSIZE")) == NULL) )
    pcxx_param.shmemsize = DEF_SHMEMSIZE;
  else if ( (pcxx_param.shmemsize = atoi(shmemsizestr)) <= 0 )
  {
    fprintf (stderr, "%s: invalid shmemsize %s - ignored.\n", argv[0],shmemsizestr);
    pcxx_param.shmemsize = DEF_SHMEMSIZE;
  }

  if ( (stacksizestr == NULL) && ((stacksizestr = getenv ("pcxx_STACKSIZE")) == NULL) )
    pcxx_param.stacksize = DEF_STACKSIZE;
  else if ( (pcxx_param.stacksize = atoi(stacksizestr)) <= 0 )
  {
    fprintf (stderr, "%s: invalid stacksize %s - ignored.\n", argv[0],stacksizestr);
    pcxx_param.stacksize = DEF_STACKSIZE;
  }
# endif

# ifdef BRKPNT_ON
  if ( (socketStr == NULL) && ((socketStr = getenv ("pcxx_BRKSOCKET")) == NULL))
    pcxx_param.SocketNum = 6871;
  else if ( (pcxx_param.SocketNum = atoi(socketStr)) <= 1024 )
  {
    fprintf (stderr, "%s: invalid socket number %s - ignored.\n",
             argv[0], socketStr);
    pcxx_param.SocketNum = 6871;
  }

  if ( (controlSocketStr == NULL) && ((controlSocketStr = getenv ("pcxx_BRKCTRLSOCKET")) == NULL))
    pcxx_param.ControlSocketNum = 6872;
  else if ( (pcxx_param.ControlSocketNum = atoi(controlSocketStr)) <= 1024 )
  {
    fprintf (stderr, "%s: invalid control socket number %s - ignored.\n",
             argv[0], socketStr);
    pcxx_param.ControlSocketNum = 6872;
  }

  if ( (hostStr == NULL) && ((hostStr = getenv ("pcxx_BRKHOST")) == NULL)) {
    /* localhost */
    pcxx_param.HostAddr = "127.0.0.1";
  } else 
    pcxx_param.HostAddr = hostStr;
  
# endif
  return 0;
}
