/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <task.h>

extern "C" void  MSNewProc(int);
extern "C" void  MSFinishProc(void);
extern "C" void  pcaw_Barrier(void);
extern "C" void  pcaw_Barrier_Uninstrumented(void);
extern "C" int   pcaw_inseq;
extern "C" int   SMRTSSetup();
extern "C" int   SMRTSFork(int, char**);
extern "C" int   pcxx_InitParam(int *, char**);

char *pcaw_cc = NULL;
int pcaw_cf = 0;

int pcaw_shmemsize;
int pcaw_numcpus;
int pcaw_nodes;
int pcaw_stacksize;

class Barrier: public object {
    public:
      Barrier(int i=0) { level = i; num = 0; }
      int  pending() { return level != num; }
      void wait();
      void height(int l) { level = l; }
    private:
      int num;
      int level;

};

void
Barrier::wait()
{
  num++;
  if (num == level) {
     num = 0;
     alert();
     return;
  } 
  thistask->sleep(this);
}

class pcxx_Thread : public task {
 public:
  int tid;
 public:
  pcxx_Thread(int i, int stacksize);
};

pcxx_Thread::pcxx_Thread(int i, int stacksize):
task("pcxx_Thread", DEFAULT_MODE, stacksize)
{
  tid = i;

  if (tid == 0) {			// Main thread
    pcaw_inseq--;
    pcaw_Barrier_Uninstrumented();
    MSNewProc(tid);
    pcaw_Barrier();
    pcaw_inseq++;

    ResetCollectionTraversalLevel ();

    Processor_Main();

    /* terminate threads */
    pcaw_inseq--;
    pcaw_cc = (char *)0;
    pcaw_Barrier();
    MSFinishProc();
    pcaw_Barrier_Uninstrumented();
    pcaw_inseq++;
  }
  else {				// Worker threads
    int done = 0;

    pcaw_Barrier_Uninstrumented();
    MSNewProc(tid);
    pcaw_Barrier();

    while(!done){
      pcaw_Barrier();
      if(pcaw_cc) {
        WorkerSwitch(pcaw_cf, pcaw_cc);
        pcaw_Barrier();
      }
      else{
        done = 1;
        MSFinishProc();
        pcaw_Barrier_Uninstrumented();
      }
    }
  }
  thistask->resultis(0);
}

Barrier AweBarrier;

extern "C" void rAwedezvous()
{
  AweBarrier.wait();
}

extern "C" void AweSetup(int nodes, int cpus, int shmemsize, int stacksize)
{
  pcaw_shmemsize = shmemsize;
  pcaw_numcpus = cpus;
  pcaw_nodes = nodes;
  pcaw_stacksize = stacksize;

  AweBarrier.height(nodes);
}

extern "C" int AweD()
{
  pcxx_Thread *thisThread = (pcxx_Thread *)thistask;

  return thisThread->tid;
}  

extern "C" void New_Main()
{  
  int i;
  pcxx_Thread **threads;

  /* startup the threads */
  threads = new pcxx_Thread*[pcaw_nodes];
  for(i=0; i<pcaw_nodes; i++){
    threads[i] = new pcxx_Thread(i, pcaw_stacksize);
  }
  for(i=0;i<pcaw_nodes;i++) {
    threads[i]->result();
  }
}

void pcaw_EnterPar(Kernel *coll, int selector)
{
  if(coll){
    ResetLocal(coll);
  }
  pcxx_inpar++;
  pcaw_inseq--;
  pcaw_cf = selector;
  pcaw_cc = (char *)coll;
  pcaw_Barrier();
  WorkerSwitch(selector, (char *)coll);
}

void pcaw_SubsetEnterPar(Kernel *coll, int selector)
{
  pcaw_EnterPar(coll, selector);
}

void pcaw_ExitPar()
{
  pcaw_Barrier();  
  pcaw_inseq++;
  pcxx_inpar--;
  ResetCollectionTraversalLevel();
}

main(int argc, char **argv)
{
  pcxx_InitParam (&argc, argv);
  SMRTSSetup ();
  SMRTSFork (argc, argv);
  thistask->resultis(0);
}
