/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 *      (C) Copyright 1990
 *
 *         The Regents of the University of California
 *                      All Rights Reserved.
 *
 *      This work was produced under the sponsorship of
 *      the U.S. Department of Energy.   The Government
 *      retains certains rights therein.
 *
 *
 *                                   Notice
 *      This computer code material was prepared as an  account  of  work
 *      sponsored  by  the  United States Government.  Neither the United
 *      States nor the United States Department of  Energy,  nor  any  of
 *      their employees, nor any of their contractors, subcontractors, or
 *      their employees, makes any warranty, express or implied,  or  as-
 *      sumes  any  legal  liability  or responsibility for the accuracy,
 *      completeness or usefulness of any information, apparatus, product
 *      or  process  disclosed,  or represents that its use would not in-
 *      fringe privately-owned rights.
 *
 */

#include 		<stdio.h>
#include                <ctype.h>
#include 		<sys/cluster.h>
#include 		<sys/kern_return.h>
#include 		<sys/wait.h>
#include		<sys/time.h>
#include		<sys/resource.h>
#include		<xmem.h>
#include		"pcplib.h"
#include                <sys/errno.h>
extern int errno;
#undef                  wait

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"

extern pcxx_PARAM_t pcxx_param;

/*
 *  Globals used in the main startup code
 */
static char *ProgramName;

/* 
 * Constants
 */
#define CMPTD "cmptd"

/*
 * if we are the gang scheduled version, include the header & map 
 */
#ifdef GSCHED
#include "gang.h"

#define NODEFLAG "PXP"
#define MAPFLAG "PXPNODES"
#define MAXNODEFLAG "max"
#define CLASSFLAG "PXPCLASS"
#define BENCHFLAG "bench"
#define BENCHMARKFLAG "benchmark"
#define INTERFLAG "interactive"
#define PRODFLAG  "production"
#define STANDFLAG "standby"
static int *parseMap();

#define ATTACH_RESENDABLE(x) ((x == gsBenchmarking) || (x == gsResourceInUse))
#define CREAT_RESENDABLE(x)  ((x == gsBenchmarking) || (x == gsResourceInUse))
#define BENCH_RESENDABLE(x)  ((x == gsBenchmarking) || (x == gsResourceInUse))
#define CLASS_RESENDABLE(x)  ((x == gsBenchmarking) || (x == gsResourceInUse))

static shared int       add_flag = 1;
#endif
#define PD_LENGTH 1024


/* globals */
static shared int	barrier_flag;      /*  used to delay main until class is set  */
static shared int	barrier_count = 0;
static shared int       (*add_sig)();         /*  pointer to parent sig handler   */      
static shared int       pd_array[PD_LENGTH];  /* array of children's pids */
static shared int       pid_that_terminated = 0;
static shared int       child_ret = 0;

extern void child_exit2();

/* macros */
#define cluster_type 	0

/*  
 * routine to allow child to say goodbye nicely, ie write out whatever files
 *  it needs to do via normal exit.
 */
static void child_exit()
{
    exit(0);
}

/*  
 * routine to the parent to exit and take the children with it
 */
static void gsExit()
{
    PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
    pcxx_EvClose ();
#ifdef GSCHED
    gsDetach();
#endif
    exit(0);
}

/*
 * die (status, children) - exit the program, killing off the children that
 *  do not leave in a timely manner.  Use the exit code "status".
 */
static void
die(status, children)
	int             status,
			children[];
{
    union wait      child_status;
    int             counter,
                    retryCount=0,
                    noExit = 1,
                    pidsLeft = 0;
    int		    myself = getpid();

    PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_KILL, 0);

    /*
     * complete the barrier protocol
     */
    sbarrier_shutdown();

    /*
     * Give the kids some time for finishing their work
     */
    sleep (3);

    /*
     * if we make it here, we have some stubborn kids that don't want 
     * to leave.  We need to send them all SIGTERM messages to get 
     * the point accross.
     */
    for (counter = 0; counter < MyTotalNodes; counter++) 
	if (children[counter] != 0 && children[counter] != myself)  {
	    (void) kill(children[counter], SIGTERM);
	    pidsLeft++;
	} /* if */

    PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
    pcxx_EvClose ();

#ifdef GSCHED
    gsDetach();
    add_flag = 0;	/* to allow Grandparent to die also */
#endif
    exit(status);
} /* die */

/*
 *  main heuristic baby - the real world starts here.
 */

SMRTSFork (argc, argv)
int argc;
char **argv;
{
    extern char	*getenv();
    int             aNode = 0,
                    cluster_id,
		    *pid_array, /*  array of children's pids   */
                    c_alloc,	/* num of nodes actually gotten */
                    exit_stat;	/* child's exit status */
    union wait      child_status;
    kern_return_t   success;	/* mach syscall success */
    char *string;
    int aFlag;  /* tmp counter */
    int anInt;  /* tmp counter */
    int   jobId; /* used to seed barrier init code */
#ifdef GSCHED
    gsPrioClass_t classFlag = pc_none;
    gsRetVal rc;
    int *nodeMap;   /* map for user selection of nodes */

    int   gangId;
    int   cluster;
    int   benchTime;
    char *pxpString,  /* env var or cmd line strings */
         *pxpClassString,
         *pxpNodeString;
    
#endif

    MyNodeNumber = PCXX_MAXPROCS - 1;

    pcxx_EvInit (pcxx_param.tracefile);
    PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, 0);
    ProgramName = argv[0];
    /*  zero out global copy of pid_array
       */
    {int i; for (i = 0; i < 1024; i++) {
       pd_array[i] = 0;
     }}
    /* 
     *   first a check to see if we are a pfp or pcp program
     */

    MyTotalNodes = pcxx_param.numproc;

    /* how many nodes? */
    if ((pxpString = getenv(NODEFLAG)) != NULL) {
	MyTotalNodes = atoi(pxpString);
    } else {
	if ((string = getenv("PCP")) != NULL) {
	    MyTotalNodes = atoi(string);
	} /* if */
    } /* if */
    if (MyTotalNodes > PD_LENGTH) {
	fprintf(stderr, "PXP too large. \n");
	exit(0);
    }

#ifdef GSCHED
    pxpClassString = getenv(CLASSFLAG);
    pxpNodeString = getenv(MAPFLAG);
    
    /* 
     * Which instance of the gang scheduler should we attach to?
     * No specification means we default to instance 0.
     */
    string = getenv("GSCHED");
    if (!string) {
	gangId = 0;
    } else {
	gangId = atoi(string);
    }
#endif

    /*
     * now parse for pxp flags:
     *   CMPTD    -cmptdesc : compact team descriptor limit.
     */
    _CMPTDESC = 257;
    for (aFlag = 1 ; aFlag < argc ; ) {
	int flag = 1; /* assume a flag for us */
	if (argv[aFlag][0] == '-') {	
	    if (strcasecmp(&argv[aFlag][1],CMPTD) == 0) {
		if ((aFlag+1) == argc) { /* ran out bore geting above string*/
		    fprintf(stderr, "%s: %s: no parameter, ignored\n", 
			    ProgramName, argv[aFlag]);
		    for (anInt = aFlag ; anInt < argc - 1 ; anInt++) {
			argv[anInt] = argv[anInt+1];
		    } /* for */
		    argc--;
		    aFlag++;
		} else {
		    int tmpInt;

		    tmpInt = atoi(argv[aFlag+1]);
		    if ((tmpInt < 0) || (tmpInt > _PXP_HARD_MAXTDESC)) {
			fprintf(stderr,"%s: %s illegal value (%s) - ignored\n",
				ProgramName, argv[aFlag], argv[aFlag+1]);
		    } else {
			_CMPTDESC = tmpInt;
		    } /* if */
		    /*
		     * now remove the flags from the command line.
		     */
		    for (anInt = aFlag ; anInt < argc - 2 ; anInt++) {
			argv[anInt] = argv[anInt+2];
		    } /* for */
		    argc-=2;
		    argv[argc] = (char *) NULL;
		} /* if */
	    } else {
		aFlag++;
	    } /* if */
	} else {
	    aFlag++;
	} /* if */
    } /* for */
#ifdef GSCHED

    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ATTACH_CALL, 0);
    rc = gsAttach(gangId, GS_REMOVE_ON_EXIT, &cluster);
    if (rc != gsSuccess) {
	if (ATTACH_RESENDABLE(rc)) {
	    unsigned int retries = 0;
	    gsErrorMsg(rc, &string);
	    fprintf(stderr, "WARNING: gsAttach: cannot attach to scheduler: %s - retrying\n", string);
	    
	    /*
	     * back off exponentially until we get a hit.
	     */
	    for (retries = 1 ; ATTACH_RESENDABLE(rc) ; retries <<= 1) {
		sleep(retries);
		rc = gsAttach(gangId, GS_REMOVE_ON_EXIT, &cluster);
	    } /* for */
	} /* if */
	if (rc != gsSuccess) {
	    gsErrorMsg(rc, &string);
	    fprintf(stderr, "%s: cannot attach to scheduler: %s\n",
		    ProgramName, string);
	    exit(1);
	} /* if */
    } /* if */
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ATTACH_RETURN, 0);
#endif
    for (aFlag = 1 ; aFlag < argc ; ) {
	int flag = 1; /* assume a flag for us */
	if (argv[aFlag][0] == '-') {	
	    if (strcasecmp(&argv[aFlag][1],NODEFLAG) == 0) {
		pxpString = argv[aFlag+1];
	    } else if (strcasecmp(&argv[aFlag][1],CLASSFLAG) == 0) {
		pxpClassString = argv[aFlag+1];
	    } else if (strcasecmp(&argv[aFlag][1],MAPFLAG) == 0) {
		pxpNodeString = argv[aFlag+1];
	    } else {
		flag = 0;
		aFlag++;
	    } /* if */
	    if (flag) {
		if ((aFlag+1) == argc) { /* ran out bore geting above string*/
		    fprintf(stderr, "%s: %s: no parameter, ignored\n", 
			    ProgramName, argv[aFlag]);
		    for (anInt = aFlag ; anInt < argc - 1 ; anInt++) {
			argv[anInt] = argv[anInt+1];
		    } /* for */
		    argc--;
		    aFlag++;
		} else {
		    for (anInt = aFlag ; anInt < argc - 2 ; anInt++) {
			argv[anInt] = argv[anInt+2];
		    } /* for */
		    argc-=2;
		    argv[argc] = (char *) NULL;
		} /* if */
	    } /* if */
	} else {
	    aFlag++;
	} /* if */
    } /* for */
    
    if (pxpString) {
	/* 
	 * found a PXP flag, find out the number with it
	 */
	if (strcasecmp(pxpString,MAXNODEFLAG) == 0) {
	    MyTotalNodes = maxNodes(); /* magic number to flag max*/
	} else if (isdigit(pxpString[0])) {
	    MyTotalNodes = atoi(pxpString);
	} else {
	    fprintf(stderr,"%s: %s: parameter %s cannot be parsed\n",
		    ProgramName, NODEFLAG, pxpString);
	    exit(1);
	} /* if*/
    } /* if */

    if (pxpClassString) {
#ifndef GSCHED
	fprintf(stderr,"%s: warning: commandline flag %s ignored\n",
		    ProgramName, pxpClassString);
#else
	/*
	 * we found a -PXPCLASS flag
	 *  Duplicate code exists for BENCHMARKFLAG && BENCHFLAG
	 *   to handle the longer string first (mainly to allow the time
	 *   to be specified as:
	 *      "bench 3400"   or
	 *      "benchmark 3400"
	 *
	 *  We need to handle the longest string (BENCHMARKFLAG) first.
	 */
	if (strncasecmp(pxpClassString,BENCHMARKFLAG,strlen(BENCHMARKFLAG))==0) {
	    classFlag = pc_benchmark;
	    /*
	     * if there is more to the string, it may be a guess at a time 
             */
	    if (strlen(&pxpClassString[strlen(BENCHMARKFLAG)])) {
		benchTime = atoi(&pxpClassString[strlen(BENCHMARKFLAG)]);
	    } else {
		benchTime = -1;
	    } /* if */
	} else if (strncasecmp(pxpClassString,BENCHFLAG, strlen(BENCHFLAG))==0) {
	    classFlag = pc_benchmark;
	    /*
	     * if there is more to the string, it may be a guess at a time 
             */
	    if (strlen(&pxpClassString[strlen(BENCHFLAG)])) {
		benchTime = atoi(&pxpClassString[strlen(BENCHFLAG)]);
	    } else {
		benchTime = -1;
	    } /* if */
	} else if (strcasecmp(pxpClassString,INTERFLAG) == 0) {
	    classFlag = pc_interactive;
	} else if (strcasecmp(pxpClassString,PRODFLAG) == 0) {
	    classFlag = pc_production;
	} else if (strcasecmp(pxpClassString,STANDFLAG) == 0) {
	    classFlag = pc_standby;
	} else {
	    fprintf(stderr,"%s: class flag %s unknown - ignored\n", 
		    ProgramName, pxpClassString);
	} /* if */
#endif
    } /* if */

    if (pxpNodeString) {
#ifndef GSCHED
	fprintf(stderr,"%s: warning: commandline flag %s ignored\n",
		    ProgramName, pxpNodeString);
#else
	if (MyTotalNodes == -1) {
	    fprintf(stderr,"%s: PXP set to max, mapping not allowed\n",
		    ProgramName);
	    exit(1);
	} /* if */
	/*
	 * we now are presented with a string of numbers
	 */
	nodeMap = parseMap(MyTotalNodes, pxpNodeString);
    } else { /* get default map */
	nodeMap = parseMap(MyTotalNodes, (int *) NULL);
    } /* if */
#endif
    
    /*
     * pid_array needs to handle 1 pid per proc.
     */
     pid_array = (int *) malloc(sizeof(int) * MyTotalNodes);

#ifndef GSCHED
    /* make the cluster */
    if ((success = cluster_create(MyTotalNodes, cluster_type, &cluster_id,
				  &c_alloc)) != KERN_SUCCESS) {
	switch(success) {
	    
	case KERN_PROTECTION_FAILURE:
	case KERN_INVALID_ARGUMENT:
	    mach_error("Internal error, invalid args to cluster_create", 
		       success);
	    break;
	    
	case KERN_INTERRUPTED:
	    mach_error("Interrupted", success);
	    break;
	    
	case KERN_RESOURCE_SHORTAGE:
	    mach_error("Insufficient resources", success);
	    break;
	    
	case KERN_NO_SPACE:
	    mach_error("Max clusters hit - cannot create", success);
	    break;
	    
	case KERN_FAILURE:
	    mach_error("Not enough free nodes", success);
	    break;
	    
	default:
	    mach_error("error in creating clusters", success);
	} /* switch */
	exit(success);
    }
    if (MyTotalNodes > 1) {
	signal(SIGTERM, child_exit);
    }

    PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
    PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

    /* start children */
    barrier_flag = 1;
    jobId = getpid();
    for (aNode = 0; aNode < MyTotalNodes; aNode++) {
	if ((success = fork_and_bind( aNode,
				     cluster_id, &pid_array[aNode]))
	    != KERN_SUCCESS) {
	    mach_error("fork_and_bind failure", success);
	    printf("Couldn't fork process %d\n", aNode);
	    die(1,pid_array);
	}
	if (pid_array[aNode] == 0) {
	    /* child process continues here */
	    /* set up team state */
	    MyNodeNumber = aNode;
	    _TINDEX = aNode;
	    _TDESC = 1;
	    _TSIZE = MyTotalNodes;
	    if (MyNodeNumber == 0) {
		_pxp_td_initl(_CMPTDESC);
		_pxp_td_initv();
	    } /* if master */
	    sbarrier_init(MyTotalNodes, NULL, jobId);
	    shmalloc(10);
	    while (barrier_flag) waitspin(1);
	    /* call the "Real" main */
            pcxx_EvInit (pcxx_param.tracefile);
	    New_Main (0);
	    exit(0);	/* this will cause parent to
			 * kill all remaining
			 * children */
	}
	/* parent continues here on next iteration of loop */
    }
#else
    
    /* Turn on LLNL-job-control */
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_JOBCNTL_CALL, 0);
    anInt = llnl_job_control(1);
    if (anInt == -1)
    {
	perror("llnl_job_control failed");
	exit(1);
    }
    
    if (MyTotalNodes > 1) {
	signal(SIGTERM, child_exit);
    }
    
    /*
     * no query (yet)
     */
    
    /*  allocate MyTotalNodes nodes:
       */
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_ALLOC_PROCS_CALL, 0);
    jobId = NEWJOBID;
    rc = gsAllocateProcs(&jobId, MyTotalNodes, nodeMap);
    if (rc != gsSuccess)    {
	gsErrorMsg(rc, &string);
	fprintf(stderr, "%s: cannot allocate nodes : %s\n",
		ProgramName, string);
	exit(1);
    } /* if (rc */

    PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);

    /*  create first node:
       */

    PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

    rc = gsCreatProcesses(jobId, MyTotalNodes); 
    if (rc != gsSuccess)    {
	while (CREAT_RESENDABLE(rc)) {
	    sleep(1);
	    rc = gsCreatProcesses(jobId, MyTotalNodes);
	} /* while */
    } /* if rc !=*/
    if (rc != gsSuccess)    {
	gsErrorMsg(rc, &string);
	fprintf(stderr, "%s: cannot obtain forking permission: %s\n",
		ProgramName, string);
	exit(1);
    } /* if rc != */

    barrier_flag = 1;    
    /* Fork the first node */
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_FIRST, 0);
    success = fork_and_bind(nodeMap[0], cluster,
			&pid_array[0]);
    if (success != KERN_SUCCESS) {
	    mach_error("fork_and_bind failure", success);
	    printf("Couldn't fork process %d\n", aNode);
	    die(1, pid_array);
    }
   /* set up team state of first node:
    */
    if (pid_array[0] == 0) {
	shmalloc(10);
	MyNodeNumber = 0;
	_TINDEX = 0;
	_TDESC = 1;
	_TSIZE = MyTotalNodes;
	_pxp_td_initl(_CMPTDESC);
	_pxp_td_initv();
	signal(SIGCHLD, child_exit2);
	/*  get address of segsegv signal handler
	    then reinstall it
	    */
	add_sig = signal(SIGSEGV, SIG_DFL);
	signal(SIGSEGV, add_sig);
	
	/* Fork the rest of the specified nodes */
        PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_REST, 0);
	for (aNode = 1 ; aNode < MyTotalNodes ; aNode++) {
	    success = fork_and_bind(nodeMap[aNode], cluster,
				    &pid_array[aNode]);
	    if (success != KERN_SUCCESS) {
	        mach_error("fork_and_bind failure", success);
	        printf("Couldn't fork process %d\n", aNode);
	        die(1, pid_array);
	    }
	    if (pid_array[aNode] == 0) {
		/* child process continues here */
		/* set up team state */
		MyNodeNumber  = aNode;
		_TINDEX = aNode;
		_TDESC  = 1;
		_TSIZE  = MyTotalNodes;
		signal(SIGSEGV, add_sig);  /*  use parent signal handler  */
		(void)atomadd32(&barrier_count, 1);
		while (barrier_count < MyTotalNodes) waitspin(1);
		sbarrier_init(MyTotalNodes, nodeMap, jobId);
		/*  wait till class is established before entering main
		 */
		while (barrier_flag)  waitspin(1);
                pcxx_EvInit (pcxx_param.tracefile);
		New_Main (0);
		exit(0);	/* this will cause parent to
				 * kill all remaining
				 * children */
	    }  /*  if pid_array[aNode] == 0   */
	    else {pd_array[aNode] = pid_array[aNode];}
	} /* for */
	
	
	add_flag = 0; /* done forking */
	/* call the "Real" main */
	(void)atomadd32(&barrier_count, 1);
	while (barrier_count < MyTotalNodes) waitspin(1);
	sbarrier_init(MyTotalNodes, nodeMap, jobId);
	/*  wait till class is established before entering main
	 */
	while (barrier_flag)   waitspin(1);
        pcxx_EvInit (pcxx_param.tracefile);
	New_Main (0);
	exit(0);	/* this will cause parent to
			 * kill all remaining
			 * children */
	
    }   /*  if Pid_array[0] == 0 for node 1   */
    
    /*  Grandparent continues here
     */
    pd_array[0] = pid_array[0];
    while (add_flag)  ; /* else wait for fork to complete */

    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_SETCLASS, 0);

    if ((rc = gsAddProcesses(jobId, MyTotalNodes, pd_array)) != gsSuccess) {
	gsErrorMsg(rc, &string);
	fprintf(stderr, "%s: cannot register nodes : %s\n", ProgramName,
		string);
	die(1, pd_array);
    } /* if */
#endif
    signal(SIGINT, gsExit);

#ifdef GSCHED
    if ( classFlag != pc_none) {
	if (classFlag != pc_benchmark) { /* setPclass needed */
	    rc = gsSetPClass(jobId, classFlag); /* creating one process */
	    if (rc != gsSuccess) {
		while (CLASS_RESENDABLE(rc)) {
		    rc = gsSetPClass(jobId, classFlag); 
		} /* if */
	    } /* if */
	} else {
	    rc = gsBenchMark(jobId, benchTime, GS_NOFLAG);
	} /* if */
	if ((rc != gsSuccess) && (rc != gsClassNotChanged))   {
	    gsErrorMsg(rc, &string);
	    fprintf(stderr, "%s: cannot obtain class desired: %s.\n",
		    ProgramName, string);
	    die(1, pd_array);
	} /* if */
    } /* if */
#endif
    /* 
     * send the kids out to play 
     *   - They won't really go until they are all ready (they wait on the
     *     condition (barrier_count < MyTotalNodes).
	   They must also wait to enter main until gsSetPClass is called
     */
    barrier_flag = 0;
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_FORK_COMPLETE, 0);
    
    /*
     * Children are now all off and running.  Parent waits until first
     * child exits, then kills all remaining children. 
     */
    if (child_ret) die(child_ret, pd_array);  
    do {
	if ((pid_that_terminated = wait(&child_status)) == -1) {
	    perror("unable to get child status");
	    die(1, pd_array);
	}
	if (WIFEXITED(child_status) ||
	    WIFSIGNALED(child_status)) { /* don't leave till someone exits */
	    break;
	} /* if */
    } while (1);  /* do forever */
    if (child_status.w_termsig == 0) {
	if (child_status.w_retcode) {
	    fprintf(stderr, "first child %d exited with status %d \n",
		    pid_that_terminated,
		    child_status.w_retcode);
	}
	die(child_status.w_retcode, pd_array);
    } else {
	psignal(child_status.w_termsig, "first child died on signal");
	die(1, pd_array);
    }
    /* NOTREACHED */
}

/*  function for parent to handle exits by actual children
    */
static void child_exit2()
{
    union wait      child_status;

    if ((pid_that_terminated = wait(&child_status)) == -1) {
	    perror("unable to get child status");
	    child_ret = 1;
	    exit(0);
    }
    if (child_status.w_termsig == 0) {
	if (child_status.w_retcode) {
	    fprintf(stderr, "child %d exited with status %d \n",
		    pid_that_terminated,
		    child_status.w_retcode);
	}
	child_ret = child_status.w_retcode;
	exit(0);
    } else {
	psignal(child_status.w_termsig, "child died on signal");
	child_ret = 1;
	exit(0);
    }
}


#ifdef GSCHED
/*
 * parseMap - parse the string into a cpu map for the program.  return
 *            an array of "count" numbers, or null if an error occurs
 */
static int *
#ifdef __STDC__
parseMap(int count, char *string)
#else
parseMap(count, string)
int count;
char *string;
#endif
{
    int anInt; /* tmp counter */
    char *index; /* index into string */
    int start, stop; /* tmp vars */
    int nodesFound = 0;

    int *map;
    map = (int *) malloc (sizeof(int) * count);

    for (anInt = 0 ; anInt < count ; anInt++) {
	map[anInt] = -1;
    } /* for */
    if (string == NULL) { /* no string, return this */
	return map;
    } /* if */
    start = -1;
    for (index = string ; *index ; ) { /* run down string */
	switch(*index) {
	case '0': case '1': case '2': case '3': case '4': 
	case '5': case '6': case '7': case '8': case '9':
	    if (start == -1) { /* if we ar looking for a start position */
		start = atoi(index);
		while(isdigit(*index)) index++;
	    } else { /* have start, looking for stop */
		stop = atoi(index);
		if (start == stop) {
		    fprintf(stderr, "%s: err in cpu range <%s>\n", ProgramName,
			    index);
		    exit(1);
		} /* if */
		if (start < stop) {
		    if ((stop - start) > (count - nodesFound)) {
			fprintf(stderr, "%s: error, asked for %d nodes, PXP = %d\n"
				,ProgramName, nodesFound + (stop -start), count);
			exit(1);
		    } /* if */
		    for (anInt = start ; anInt <= stop ; anInt++) {
			map[nodesFound++] = anInt;
		    } /* for */
		} else { /* stop < start */
		    if ((start - stop) > (count - nodesFound)) {
			fprintf(stderr, "%s: error, asked for %d nodes, PXP = %d\n"
				,ProgramName, nodesFound + (stop -start), count);
			exit(1);
		    } /* if */
		    for (anInt = start ; anInt >= stop ; anInt--) {
			map[nodesFound++] = anInt;
		    } /* for */
		} /* if */
		while(isdigit(*index)) index++;
		start = stop = -1;
	    } /* if */
	    break;

	case ',':
	case ' ':
	case 0: /* for end of string */
	    if (start != -1) {
		map[nodesFound++] = start;
		start = -1;
	    } /* if */
	    index++;
	    break;

	case '-':
	    if (start == -1) { /* error, range must have a start */
		fprintf(stderr, "%s: error, range with no start\n"
			,ProgramName);
		exit(1);
	    } /* if */
	    index++;
	    break;
	case '+':
	    index++;
	    if (*index) {
		fprintf(stderr, "%s: error, open range not last item in list\n"
			,ProgramName);
		exit(1);
	    } /* if */
	    if (start == -1) {
		fprintf(stderr, "%s: error, open range does not have start\n"
			,ProgramName);
		exit(1);
	    } /* if */
	    anInt = 0;
	    while (nodesFound < count) {
		map[nodesFound] = start + anInt;
		anInt++;
		nodesFound++;
	    } /* while */
	    start = -1;
	    break;
	default:
	    fprintf(stderr, "%s: error unknown character <%c>\n"
		    ,ProgramName, *index);
	    exit(1);
	} /* switch */
    } /* for */
    /*
     * if we make it here with start equal something, we have hit the end
     * of the string, and have to grab the last item.
     */
    if (start != -1) {
	map[nodesFound++] = start;
    } /* if */
    if (nodesFound != count) {
	fprintf(stderr,"%s: parsed %d nodes, needed %d\n", ProgramName, 
		nodesFound, count);
	exit(1);
    } /* if */
    for (anInt = 0 ; anInt < (count-1) ; anInt++) {
	for (start = anInt+1 ; start < count ; start ++) {
	    if (map[anInt] == map[start]) {
		fprintf(stderr,"%s: ERROR node %d selected twice\n", 
			ProgramName, map[start]);
		exit(1);
	    } /* if */
	} /* for */
    } /* for */
    return map;
} /* parseMap */

/*
 * maxNodes - get information from the gs telling how many nodes is "max"
 */
static int
maxNodes()
{
    gsRetVal rc;
    sysConfig_t *sysInfo;
    char *string;

    rc = gsSysConfig(&sysInfo);
    if (rc == gsSuccess) {
	return sysInfo->sc_cpuCount;
    } else {
	gsErrorMsg(rc, &string);
	fprintf(stderr, "%s: cannot obtain max nodes: %s\n",
			ProgramName, string);
	exit(1);
    } /* if */
    return -1;
} /* maxNodes */
#endif


