/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*==============================================================*n
/* bbnlib : BBN run-time support library			*/
/*==============================================================*/

#include <heap.h>

#include "GlobalStruct.h"
#include "pcxx_events_def.h"
#include "pcxx_timers.h"
#include "profiler_defs.h"

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */
#define MAXBUFF 4096
#define MAX_PROC 128

extern int MyTotalNodes;
extern int MyNodeNumber;
extern int _TSIZE;
extern int _TDESC;
extern int _TINDEX;
extern int _CMPTDESC;

extern char *heapmalloc();
extern char *shmalloc();
extern void pcxx_Barrier();

shared char BroadcastBuff[MAXBUFF];  /* a place to store broacast values */
char *ReduceBuff[MAX_PROC];   /* a place to store reduction values */

int BBN_Switch = 1;

SMRTSSetup ()
{
  int i;
  /* allocate shared reduction buffer */
  if(MyTotalNodes > MAX_PROC) {
    fprintf(stderr,"Error in sgilib.c: MAX_PROC < number of processors.\n");
    exit(0);
  }
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) shmalloc(MAXBUFF);
  }
}

int SMRTSNewProc(num)
int num;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);
  BBN_Switch = 1;
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}

void pcxx_Barrier()
{
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  BBN_Switch = 1 - BBN_Switch;
  sbarrier_half(_TDESC, _TINDEX, _TSIZE, MyNodeNumber, BBN_Switch);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}

/*
 @ pcxx_BrodcastBytes : sends messages from 0 to all other nodes
 */

void pcxx_BroadcastBytes(source, length, buffer)
int source;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive pcxx_BroadcastBytes are called */
  pcxx_Barrier();

  if ( length > MAXBUFF ) Message("Broadcast Too BIG!! ",FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (source == 0) for (i=0; i<length; i++) c[i] = p[i];
  pcxx_Barrier();
  if (source != 0) for (i=0; i<length; i++) p[i] = c[i];
}

void SMRTSDealloc(ptr)
void *ptr;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  heapfree(ptr);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}
