/*
 *	(C) Copyright 1991
 *
 *	   The Regents of the University of California
 *			All Rights Reserved.
 *
 *	This work was produced under the sponsorship of
 *	the U.S. Department of Energy.   The Government
 *	retains certains rights therein.
 *
 *
 *	                             Notice
 *	This computer code material was prepared as an  account  of  work
 *	sponsored  by  the  United States Government.  Neither the United
 *	States nor the United States Department of  Energy,  nor  any  of
 *	their employees, nor any of their contractors, subcontractors, or
 *	their employees, makes any warranty, express or implied,  or  as-
 *	sumes  any  legal  liability  or responsibility for the accuracy,
 *	completeness or usefulness of any information, apparatus, product
 *	or  process  disclosed,  or represents that its use would not in-
 *	fringe privately-owned rights.
 *
 */
/*  pxplib version of compact _TDESC routines   */

#include <stdio.h>
#include "pcplib.h"
#include <heap.h>

/*  compact _TDESC linked list structure:  */
typedef struct _TD_LST {
	int td;              /*  available _TDESC no.  */
	struct _TD_LST *next;
	} _TD_LST;
 
_shared _TD_LST *ptr_first = NULL;

_shared unsigned int list_lock = unlocked;


/*  team private structure to store team descriptors for sub-teams:   */
typedef struct _TD_STORE {
   int *ptd;      /*  pointer to actual storage array             */
   int size;      /*  size of actural storage array               */
   } _TD_STORE;

_shared _TD_STORE *_TDSTORE = NULL;


#define hpalloc(obj,n) ((obj *)heapmalloc(HEAP_UNCACHED|HEAP_INTERLEAVED,0, n*sizeof(obj)))

char filename[256];   /*  array for filenames                      */

/* int debug = 0;  */

/*****************************************************************/
/*  initialize linked list,  _TD_LST  PCP                        */
/*  called from main.c and td_chg                                */
/*****************************************************************/
void _pxp_td_initl(tstart)
int tstart;     /* lower boundary of compact _TDESC linked list  */
{
	register int i;
	int n;
	_TD_LST *ptr;

        if (_TSIZE != MyTotalNodes) 
	{
	   fprintf(stderr,
		" _pxp_td_initl:  Initialization occurring at incorrect level. \n");
	   exit(-1);
	}
	if (tstart > _PXP_HARD_MAXTDESC) 
	{
	   fprintf(stderr, 
		 "_pxp_td_initl:  _PXP_HARD_MAXTDESC limit to be exceeded.\n"); 
	   exit(-1);
	}

	/*  take care of first entry  */

        if ((ptr = hpalloc(_TD_LST, 1)) == NULL) 
	{
		fprintf(stderr, "_pxp_td_initl:  Problem in allocating _TD_LST. \n");
		exit(-1);
        }
	ptr_first  = ptr;
	ptr->td    = tstart;
	ptr->next  = NULL;

	n = _PXP_HARD_MAXTDESC - tstart+1;
	for (i = 1; i < n; i++) 
	{
           if ((ptr->next = hpalloc(_TD_LST, 1)) == NULL) 
	   {
		   fprintf(stderr, "_pxp_td_initl:  Problem in allocating _TD_LST. \n");
		   exit(-1);
            }
	  ptr       = ptr->next;
	  ptr->td   = tstart+i;
	  ptr->next = NULL;
	}
}


/*****************************************************************/
/*  print _TD_LST   PCP                                          */
/*****************************************************************/
void _pxp_td_print()
{
	_TD_LST *ptr;

        fprintf(stdout,"_pxp_td_print entered.  \n");
	ptr = ptr_first;
	while (ptr) 
	{
	   fprintf(stdout,"Available _TDESC:  %d \n", ptr->td);
           ptr = ptr->next;
	}
}

/*****************************************************************/
/*  print _TD_LST                                                */
/*****************************************************************/
void _pxp_td_print_()
{
	_TD_LST *ptr;

        fprintf(stdout,"_pxp_td_print entered.  \n");
	ptr = ptr_first;
	while (ptr) 
	{
	   fprintf(stdout,"Available _TDESC:  %d \n", ptr->td);
           ptr = ptr->next;
	}
}

/*****************************************************************/
/*  allocate teamprivate structure, _TDSTORE                     */
/*****************************************************************/
void _pxp_td_initv()
{
    register int i;



    if ((_TDSTORE =  hpalloc(_TD_STORE, _PXP_HARD_MAXTDESC)) == NULL)
      {
       fprintf(stderr, "_pxp_td_initv:  Problem in allocating _TDSTORE. \n");
       exit(-1);
      }
      for (i = 0; i < _PXP_HARD_MAXTDESC; i++)
      {
	if ((_TDSTORE[i].ptd = hpalloc(int, 10)) == NULL) 
           {
	      fprintf(stderr, "_pxp_td_initv:  Problem in allocating _TDSTORE.ptd. \n");
	      exit(-1);
           }
	_TDSTORE[i].size = 10;
      }

    return;
}



/*****************************************************************/
/*  get n _TDESC's from _TD_LST;                                 */
/*  store in teamprivate array, _TDSTORE    PCP                  */
/*****************************************************************/
void _pxp_td_get(n, j, lineno, file_name)
int n;           /*  number of _TDESC's to obtain                */
int j;           /*  calling processor _TDESC number             */
int lineno;      /* line no of original source file              */
char *file_name; /* name of original source file                 */
{
	_TD_LST *ptr, *sv_ptr;
	register int i;

	if (_TINDEX != 0) 
	{
	   fprintf(stderr,
	      "_pxp_td_get:  Incorrect team member requesting _TDESC, line %d file %s \n", 
				 lineno, file_name);
				 
	       exit(-1);
	}
	if (_TDSTORE[j].ptd == NULL)
	{
		fprintf(stderr,"_pxp_td_get:  No space for sub-team descriptors. \n");
		exit(-1);
	}

	if (n > _TDSTORE[j].size)
	{
		/*   allocate space for new _TDESC's  */
        
		free(_TDSTORE[j].ptd);
		if ((_TDSTORE[j].ptd = hpalloc(int, n)) == NULL) 
           	{
	      		fprintf(stderr, "_pxp_td_get:  Problem in allocating _TDSTORE.ptd. \n");
	      	        exit(-1);
           	}
		_TDSTORE[j].size = n;
	}
	lock(&list_lock);
	ptr = ptr_first;
	for (i = 0; i < n; i++) 
	{
	   if (ptr == NULL) 
	   {
	      fprintf(stderr, 
	          "_pxp_td_get:  Unable to get all _TDESCs. lineno %d file %s \n",
		      lineno, filename);
	      exit(-1);
	   }
	   *(_TDSTORE[j].ptd+i) = ptr->td;
	   /*
	   if (debug) printf("_pxp_td_get:  j, i, _TDESC:  %d %d %d \n",
				     j, i, *(_TDSTORE[j].ptd+i));
				     */
	   sv_ptr       = ptr->next;
	   (void) free(ptr);
	   ptr = sv_ptr;
	}
	ptr_first = ptr;
	unlock(&list_lock);
}
/*****************************************************************/
/*  print 1st 3 elts. of teamprivate array, _TDSTORE for given _TDESC */
/*****************************************************************/
void _pxp_td_stprnt_(tj)
int *tj;         /*  calling processor _TDESC number             */
{
	register int i, j;

	j   = *tj;
	for (i = 0; i < 3; i++) 
	{
	   printf("_pxp_td_stprnt:  j, i, _TDESC:  %d %d %d \n", 
		    j, i, *(_TDSTORE[j].ptd+i));
	}
}

/*****************************************************************/
/*  get n _TDESC's from _TD_LST;                                 */
/*  store in teamprivate array, _TDSTORE                         */
/*****************************************************************/
void _pxp_td_get_(en, tj, lineno, file_name, len)
int *en;         /*  number of _TDESC's to obtain                */
int *tj;         /*  calling processor _TDESC number             */
int *lineno;     /* line no of original source file              */
char *file_name; /* name of original source file                 */
int len;
{
	_TD_LST *ptr, *sv_ptr;
	register int i, j;
	int n;

        strncpy(filename, file_name, len+1);
	if (_TINDEX != 0) 
	{
	   fprintf(stderr,
	      "_pxp_td_get:  Incorrect team member requesting _TDESC, line %d file %s \n", 
				 *lineno, filename);
				 
	       exit(-1);
	}
	j   = *tj;
	n   = *en;
	if (_TDSTORE[j].ptd == NULL)
	{
		fprintf(stderr,"_pxp_td_get:  No space for sub-team descriptors. \n");
		exit(-1);
	}
	if (n > _TDSTORE[j].size)
	{
		/*   allocate space for new _TDESC's  */
        
		free(_TDSTORE[j].ptd);
		if ((_TDSTORE[j].ptd = hpalloc(int, n)) == NULL)
           	{
	      		fprintf(stderr, "_pxp_td_get:  Problem in allocating _TDSTORE.ptd. \n");
	      	exit(-1);
           	}
		_TDSTORE[j].size = n;
	}
	lock(&list_lock);
	ptr = ptr_first;
	for (i = 0; i < n; i++) 
	{
	   if (ptr == NULL) 
	   {
	      fprintf(stderr, 
	          "_pxp_td_get:  Unable to get all _TDESCs. lineno %d file %s \n",
		      *lineno, filename);
	      exit(-1);
	   }
	   *(_TDSTORE[j].ptd+i) = ptr->td;
	   /*
	   if (debug) printf("_pxp_td_get:  j, i, _TDESC:  %d %d %d \n", 
				      j, i, *(_TDSTORE[j].ptd+i));
				      */
	   sv_ptr       = ptr->next;
	   (void) free(ptr);
	   ptr = sv_ptr;
	}
	ptr_first = ptr;
	unlock(&list_lock);
}

/*****************************************************************/
/*  set _TDESC from teamprivate array   PCP                      */
/*****************************************************************/
int _pxp_td_set(i, j)
int i;      /*  task no                     */
int j;      /*  calling processor former _TDESC number     */
{
        register k;

        k = i - 1;

        /*
        if (debug) printf("_pxp_td_set:  j, k, new _TDESC:  %d %d %d \n", 
				 j, k, *(_TDSTORE[j].ptd+k));
				 */
	return(*(_TDSTORE[j].ptd+k));
}

/*****************************************************************/
/*  set _TDESC from teamprivate array                            */
/*****************************************************************/
int _pxp_td_set_(i, tj)
int *i;    /*  task no                            */
int *tj;   /*  calling processor former _TDESC number     */
{
        register k, j;

        k = *i - 1;
	j = *tj;
	/*
        if (debug) printf("_pxp_td_set:  j, k, new _TDESC:  %d %d %d \n", 
				 j, k, *(_TDSTORE[j].ptd+k));
				 */
	return(*(_TDSTORE[j].ptd+k));
}


/*****************************************************************/
/*  return _TSTORE entries to linked list, _TD_LST       PCP     */
/*****************************************************************/
void _pxp_td_ret(n, j, lineno, file_name)
int n;           /*  numbers of _TDESC's to be returned          */
int j;           /*  calling processor _TDESC number             */
int lineno;      /* line no of original source file              */
char *file_name; /* name of original source file                 */
{
     _TD_LST *ptr, *new_ptr, *sv_ptr;
     register int i;
     int tdrt;

     lock(&list_lock);
     ptr    = ptr_first;
     sv_ptr = NULL;
     for (i = 0; i < n; i++) 
	{
	tdrt   = *(_TDSTORE[j].ptd+i);
	/*
        if (debug) printf("_pxp_td_ret:  tdrt, MyNodeNumber, _TSIZE:  %d %d %d \n",
		       tdrt, MyNodeNumber, _TSIZE);
		       */
        if ((new_ptr = hpalloc(_TD_LST, 1)) == NULL) 
        {
	      fprintf( stderr, 
	       "_pxp_td_ret:  Problem in restoring _TD_LST. line %d file %s\n",
			       lineno, file_name);
	      exit(-1);
        }
	while (ptr && (ptr->td < tdrt)) 
	{
          sv_ptr = ptr;
          ptr    = ptr->next;
	}     /* while   */

        new_ptr->td   = tdrt;
        new_ptr->next = ptr;
	if (sv_ptr == NULL) 
	{
	   ptr_first = new_ptr;
	}
	else
	{
	   sv_ptr->next = new_ptr;
	}
	sv_ptr = new_ptr;
	ptr    = new_ptr->next;
     }   /*  for   */
     unlock(&list_lock);
}

/*****************************************************************/
/*  return _TSTORE entries to linked list, _TD_LST               */
/*****************************************************************/
void _pxp_td_ret_(en, tj, lineno, file_name, len)
int *en;         /*  numbers of _TDESC's to be returned          */
int *tj;         /*  calling processor _TDESC number             */
int *lineno;     /* line no of original source file              */
char *file_name; /* name of original source file                 */
int len;
{
     _TD_LST *ptr, *new_ptr, *sv_ptr;
     register int i;
     register int j;
     int n, tdrt;

     strncpy(filename, file_name, len+1);
     j      = *tj;
     n      = *en;
     lock(&list_lock);
     ptr    = ptr_first;
     sv_ptr = NULL;
     for (i = 0; i < n; i++) 
	{
	tdrt   = *(_TDSTORE[j].ptd+i);
	/*
        if (debug) printf("_pxp_td_ret:  tdrt, MyNodeNumber, _TSIZE:  %d %d %d \n",
		       tdrt, MyNodeNumber, _TSIZE);
		       */
        if ((new_ptr = hpalloc(_TD_LST, 1)) == NULL) 
        {
	      fprintf( stderr, 
	       "_pxp_td_ret:  Problem in restoring _TD_LST. line %d file %s\n",
			       *lineno, filename);
	      exit(-1);
        }
	while (ptr && (ptr->td < tdrt)) 
	{
          sv_ptr = ptr;
          ptr    = ptr->next;
	}     /* while   */

        new_ptr->td   = tdrt;
        new_ptr->next = ptr;
	if (sv_ptr == NULL) 
	{
	   ptr_first = new_ptr;
	}
	else
	{
	   sv_ptr->next = new_ptr;
	}
	sv_ptr = new_ptr;
	ptr    = new_ptr->next;
     }   /*  for   */
     unlock(&list_lock);
}

/*****************************************************************/
/*  change _TD_LST lower limit; called by each processor         */
/*****************************************************************/
void _pxp_td_chg(tstart, lineno, file_name)
int tstart;      /*  new lower boundary of _TDESC linked list    */
int lineno;      /* line no of original source file              */
char *file_name; /* name of original source file                 */
{
   _TD_LST *ptr, *sv_ptr;
   register int i;
   int n;

       
   if (_TSIZE != MyTotalNodes) 
   {
      fprintf(stderr,
	" _pxp_td_chg:  _CMPTDESC change occurring at incorrect level. \n");
      exit(-1);
   }
   _CMPTDESC = tstart; 

   if (_TINDEX == 0) 
   {
      if (tstart > _PXP_HARD_MAXTDESC) 
      {
          fprintf(stderr,
            "_pxp_td_chg:  _PXP_HARD_MAXTDESC limit to be exceeded. line no %d, file %s.\n", 
				    lineno, file_name);
          exit(-1);
      }

      ptr = ptr_first;
      while (ptr && (tstart > ptr->td)) 
      {         
          sv_ptr = ptr->next;
          (void) free(ptr);
          ptr       = sv_ptr;
          ptr_first = sv_ptr;
      }
      if (ptr_first == NULL)
      {
         /*  establish entirely new _TD_LST       */

         _pxp_td_initl(tstart);
      }
      else if (tstart == ptr->td)
      {
         ptr_first = ptr;
      }
      else   /*  tstart < ptr->td)   decrease lower bound of _TD_LIST   */
      {
         sv_ptr = ptr_first;
         n      = ptr->td - tstart;
         if ((ptr =  hpalloc(_TD_LST, 1)) == NULL)
         {
              fprintf(stderr, "_pxp_td_chg:  Problem in changing _TD_LST. \n");
              exit(-1);
         } 
         ptr_first  = ptr;
         ptr->td    = tstart;
         ptr->next  = NULL;
         for (i = 1; i < n; i++)
         {
            if ((ptr->next = hpalloc(_TD_LST, 1)) == NULL)
            {
                 fprintf(stderr, "_pxp_td_chg:  Problem in changing _TD_LST. \n");
                 exit(-1);
            } 
            ptr       = ptr->next;
            ptr->td   = tstart+i;
            ptr->next = sv_ptr;
         }  /*  for   */
      }   /*  else   */

   }  /*  if _TINDEX == 0   */
		

}
/*****************************************************************/
/*  change _TD_LST lower limit; called by each processor         */
/*****************************************************************/
void _pxp_td_chg_(tstart, lineno, file_name, len)
int *tstart;     /*  new lower boundary of _TDESC linked list    */
int *lineno;     /* line no of original source file              */
char *file_name; /* name of original source file                 */
int len;
{
   _TD_LST *ptr, *sv_ptr;
   register int i;
   int n;

       
   if (_TSIZE != MyTotalNodes) 
   {
      fprintf(stderr,
	" _pxp_td_chg:  _CMPTDESC change occurring at incorrect level. \n");
      exit(-1);
   }
   _CMPTDESC = *tstart; 

   if (_TINDEX == 0) 
   {
      strncpy(filename, file_name, len+1);
      if (*tstart > _PXP_HARD_MAXTDESC) 
      {
          fprintf(stderr,
            "_pxp_td_chg:  _PXP_HARD_MAXTDESC limit to be exceeded. line no %d, file %s.\n", 
				    *lineno, filename);
          exit(-1);
      }

      ptr = ptr_first;
      while (ptr && (*tstart > ptr->td)) 
      {         
          sv_ptr = ptr->next;
          (void) free(ptr);
          ptr       = sv_ptr;
          ptr_first = sv_ptr;
      }
      if (ptr_first == NULL)
      {
         /*  establish entirely new _TD_LST       */

         _pxp_td_initl(tstart);
      }
      else if (*tstart == ptr->td)
      {
         ptr_first = ptr;
      }
      else   /* *tstart < ptr->td)   decrease lower bound of _TD_LIST   */
      {
         sv_ptr = ptr_first;
         n      = ptr->td - (*tstart);
         if ((ptr = hpalloc(_TD_LST, 1)) == NULL)
         {
              fprintf(stderr, "_pxp_td_chg:  Problem in changing _TD_LST. \n");
              exit(-1);
         } 
         ptr_first  = ptr;
         ptr->td    = *tstart;
         ptr->next  = NULL;
         for (i = 1; i < n; i++)
         {
            if ((ptr->next =  hpalloc(_TD_LST, 1)) == NULL)
            {
                 fprintf(stderr, "_pxp_td_chg:  Problem in changing _TD_LST. \n");
                 exit(-1);
            } 
            ptr       = ptr->next;
            ptr->td   = (*tstart)+i;
            ptr->next = sv_ptr;
         }  /*  for   */
      }   /*  else   */

   }  /*  if _TINDEX == 0   */
		

}



