/*
 *	(C) Copyright 1990
 *
 *	   The Regents of the University of California
 *			All Rights Reserved.
 *
 *	This work was produced under the sponsorship of
 *	the U.S. Department of Energy.   The Government
 *	retains certains rights therein.
 *
 *
 *	                             Notice
 *	This computer code material was prepared as an  account  of  work
 *	sponsored  by  the  United States Government.  Neither the United
 *	States nor the United States Department of  Energy,  nor  any  of
 *	their employees, nor any of their contractors, subcontractors, or
 *	their employees, makes any warranty, express or implied,  or  as-
 *	sumes  any  legal  liability  or responsibility for the accuracy,
 *	completeness or usefulness of any information, apparatus, product
 *	or  process  disclosed,  or represents that its use would not in-
 *	fringe privately-owned rights.
 *
 */
#define NUM_EXITFUNC 32
/*
 *  exit and atexit code.
 *
 *    Under the pxp model of computation, a child node is killed as soon 
 *  as the parent detects that a sibling has taken an exception or exited.
 *  The children who are in exit at the time need to ignore the signal, 
 *  and continue through it.  Thus we set the sigmask to ignore SIGTERM, and 
 *  continue on through exit.
 */
#include <signal.h>

static void (*atExitRoutines[NUM_EXITFUNC])();

static int exitRoutineNum = 0;

int
#ifdef __ANSIC__
atexit(void (*aFunc)())
#else
atexit(aFunc)
void (*aFunc)();
#endif
{
    /*
     * ensure that there is room in the exit routine list, return -1 if not
     * register the function and return 0 otherwise
     */
    if (exitRoutineNum < NUM_EXITFUNC) {
	atExitRoutines[exitRoutineNum++] = aFunc;
	return(0);
    } else {
	return(-1); /* error, no room left */
    } /* if room left */
} /* atexit */

/*
 * exit routine - set the SIGTERM interrupt off, call the exit routines 
 *                registered with atexit, and then cleanup and _exit;
 */
int
#ifdef __ANSIC__
exit(int status)
#else
exit(status)
int status;
#endif
{
    void (*func)();
    int exitNum;

    /* 
     * first thing is turn of the SIGTERM exceptions.  We want to exit
     * with a little grace, and need the parent to leave us along as we do.
     */

    (void) sigblock(sigmask(SIGTERM)); /* turn off sigterms */
    /* 
     * remove the routine *before* calling it.  This will get rid of 
     *       the problem of recursion - when the routine trys to call
     *       exit itself.
     */
    for (exitNum = 0 ; exitNum < exitRoutineNum ; exitNum++) {
	func = atExitRoutines[exitNum];
	atExitRoutines[exitNum] = (void *) -1;
	if (func != (void *) -1) { /* not called already */
	    (*func)();
	} /* if */
    } /* for */
    _cleanup();   
    _exit(status);
} /* exit */
