/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/***************************************************************************
 *                HPC++ Routine for managing access to data                *
 *                Sequential Set Of routine                                *
 *                                                                         *
 *                F. Bodin, Indiana University                             *
 ***************************************************************************/
#include <stdarg.h>
#include <stdio.h>

/* these numbers have to be the same as they are in sgilib.c, ksrlib.c, ... */
#define MAXBUFF 4096
#define MAX_PROC 128
extern char *ReduceBuff[MAX_PROC];

#include <tulip.h>

/* Forward definition of activated function */
void SendData();
void UpdateMyData();

/**************************************************************************
 ********************** Kernel Request ************************************
 **************************************************************************/

main(int argc, char **argv)
{
  pcxx_InitParam (&argc, argv);
  SMRTSSetup ();
  SMRTSFork (argc, argv);
}

extern int  MyTotalNodes;
int tulip_NumContexts() 
{ 
  return(MyTotalNodes); 
}

extern int  MyNodeNumber;
int tulip_MyContext() 
{ 
  return(MyNodeNumber); 
}

void tulip_ReduceIntAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	((int*)ReduceBuff[self])[elem] += 
	  ((int*)ReduceBuff[self+span])[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceFloatAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	((float*)ReduceBuff[self])[elem] += 
	  ((float*)ReduceBuff[self+span])[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceDoubleAdd(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	((double*)ReduceBuff[self])[elem] += 
	  ((double*)ReduceBuff[self+span])[elem];
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceDoubleMax(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((double*)ReduceBuff[self+span])[elem] >
	   ((double*)ReduceBuff[self])[elem]) 
	  {
	    ((double*)ReduceBuff[self])[elem] = 
	      ((double*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceFloatMax(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((float*)ReduceBuff[self+span])[elem] >
	   ((float*)ReduceBuff[self])[elem]) 
	  {
	    ((float*)ReduceBuff[self])[elem] = 
	      ((float*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceIntMax(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((int*)ReduceBuff[self+span])[elem] >
	   ((int*)ReduceBuff[self])[elem]) 
	  {
	    ((int*)ReduceBuff[self])[elem] = 
	      ((int*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceDoubleMin(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(double);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((double*)ReduceBuff[self+span])[elem] <
	   ((double*)ReduceBuff[self])[elem]) 
	  {
	    ((double*)ReduceBuff[self])[elem] = 
	      ((double*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceFloatMin(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(float);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((float*)ReduceBuff[self+span])[elem] <
	   ((float*)ReduceBuff[self])[elem]) 
	  {
	    ((float*)ReduceBuff[self])[elem] = 
	      ((float*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}

void tulip_ReduceIntMin(void* buffer,int elem_count) 
{
  int self, span, flag, elem, size;

  size = elem_count*sizeof(int);
  if(size > MAXBUFF) 
    tulip_Message("Reduction buffer too small!  Increase buffer", TULIP_FATAL);
  self = tulip_MyContext();
  memcpy(ReduceBuff[self],(char*)buffer,size);
  tulip_Barrier();
  /* tree reduction */
  for(span = 1; span < tulip_NumContexts(); span*=2) {
    if(self % (2*span) == 0 && (self+span) < tulip_NumContexts()) {
      for(elem = 0; elem < elem_count; elem++) {
	if(((int*)ReduceBuff[self+span])[elem] <
	   ((int*)ReduceBuff[self])[elem]) 
	  {
	    ((int*)ReduceBuff[self])[elem] = 
	      ((int*)ReduceBuff[self+span])[elem];
	  }
      }
    }
    tulip_Barrier();
  }
  /* processor zero broadcasts reduction result */
  flag = 1;
  if(self == 0) {
    flag = 0;
    memcpy((char*)buffer,ReduceBuff[self],size);
  }
  tulip_BroadcastBytes(flag, size, buffer);
  /* barrier to prevent ReduceBuff from being overwritten when  */
  /* two consecutive tulip_ReduceIntAdd are called               */
  tulip_Barrier();
}




