/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*
 * pcxx_events_def.h: pC++ runtime system event definitions
 *
 * (c) 1994 Jerry Manic Saftware
 *
 * Version 2.1
 */

# ifndef __PCXX_EVENTS_DEF_H__
# define __PCXX_EVENTS_DEF_H__

# ifdef TRACING_ON
/* -- event classes ----------------------------- */
# define PCXX_EC_NULL       0x00000000

# define PCXX_EC_TRACER     0x00000001
# define PCXX_EC_TIMER      0x00000002

# define PCXX_EC_BASIC      0x00000010
# define PCXX_EC_RUNTIME    0x00000020
# define PCXX_EC_PROFILER   0x00000040
# define PCXX_EC_KERNEL     0x00000080

# define PCXX_EC_USER1      0x00000100
# define PCXX_EC_USER2      0x00000200
# define PCXX_EC_USER3      0x00000400
# define PCXX_EC_USER4      0x00000800
# define PCXX_EC_USER       0x00000F00

# define PCXX_EC_COLLACS    0x00001000
/* -- For Ariadne debugger - event based behavioural abstraction */
# define PCXX_EC_ARIADNE    0x0000F000

# define PCXX_EC_ALL        0xFFFFFFFF

/* These are like flags to the above classes: Setting the SMODE    */
/* makes  the classes switchable. Setting the SMODE_ON sets the    */
/* switch. Resetting the SMODE_ON resets the switch		   */
/* Example: pcxx_ev_class = 0x01000080 means that the KERNEL       */
/* events are recorded in a switchable mode with the switch        */
/* currently off.                                                  */
/* For hacking convenience, if the classes are specified without   */
/* the swithchable mode, the switches will be left on.             */

# define PCXX_EC_SMODE	    0x01000000
# define PCXX_EC_SMODE_ON   0x10000000

/* -- general events for all SMRTS machines ----- */
/* -- used in the runtime system ---------------- */
# define PCXX_SETUP_ENTER          61100     /* B */    /* num_proc   */
# define PCXX_SETUP_INIT           61101     /* R */
# define PCXX_SETUP_EXIT           61102     /* B */    /* MACH_Num_P */
# define PCXX_FORK_ENTER           61103     /* B */
# define PCXX_FORK_EXIT            61104     /* B */

# define PCXX_NEWPROC_ENTER        61105     /* R */
# define PCXX_NEWPROC_EXIT         61106     /* R */    /* MACH_My_ID */
# define PCXX_BARRIER_ENTER        61107     /* R */
# define PCXX_BARRIER_EXIT         61108     /* R */
# define PCXX_POLL_ENTER           61109     /* R */    /* ptr        */
# define PCXX_POLL_EXIT            61110     /* R */
# define PCXX_DEALLOC_ENTER        61111     /* R */
# define PCXX_DEALLOC_EXIT         61112     /* R */
# define PCXX_COLLACS_START        61113     /* C */    /* from_proc  */
# define PCXX_COLLACS_END          61114     /* C */    /* from_proc  */

# define PCXX_MAIN_ENTER           61120     /* B */
# define PCXX_MAIN_EXIT            61121     /* B */

# define PCXX_INSTL_COLL           61150     /* K */    /* coll_num   */
# define PCXX_COLL_INFO_1          61151     /* K */    /* num_dim    */
# define PCXX_COLL_INFO_2          61152     /* K */    /* sizeelem   */
# define PCXX_COLL_INFO_3          61153     /* K */    /* nb_elem    */
# define PCXX_COLL_INFO_4          61154     /* K */    /* dim1size   */
# define PCXX_COLL_INFO_5          61155     /* K */    /* dim2size   */
# define PCXX_ELEM_INFO            61156     /* K */    /* elem_num   */
# define PCXX_ELEM_ACCESS          61157     /* K */    /* elem_ident */
# define PCXX_ELEM_SIZE            61158     /* K */    /* size       */
# define PCXX_ACS_ON               61159     /* E */    /* func.    id*/
# define PCXX_ACS_OFF	           61160     /* E */    /* func. id   */

# define PCXX_SYNC_MARK            61170     /* K */
/* For ARIADNE */
# define PCXX_IPC		   61177     /* G */    /* type MRWU+oid */
# define PCXX_ARIADNE_PCXX_EV 	   61178     /* G */    /* other pcxx event */

/* -- used in the benchmarks -------------------- */
# define PCXX_USERCODE_ENTER       61122     /* B */

/* -- BBN TC2000 specific events ---------------- */
# ifdef butterfly
#    define PCXX_FORK_FIRST        61200     /* R */
#    define PCXX_FORK_REST         61201     /* R */
#    define PCXX_FORK_SETCLASS     61202     /* R */
#    define PCXX_FORK_COMPLETE     61203     /* R */
#    define PCXX_ATTACH_CALL       61204     /* R */
#    define PCXX_ATTACH_RETURN     61205     /* R */
#    define PCXX_JOBCNTL_CALL      61206     /* R */
#    define PCXX_ALLOC_PROCS_CALL  61207     /* R */
#    define PCXX_HEAPM_CALL        61208     /* R */
#    define PCXX_FORK_KILL         61209     /* R */
# endif

/* -- KSR1 specific events ---------------------- */
# ifdef __ksr__
#    define PCXX_FORK_FORK         61210     /* R */    /* process    */
#    define PCXX_FORK_JOIN         61211     /* R */    /* process    */
#    define PCXX_FORK_BIND         61212     /* R */    /* processor  */
#    define PCXX_IN_BARRIER        61213     /* - */    /* num        */
# endif

/* -- Sequent Symmetry specific events ---------- */
# if defined(_SEQUENT_) || defined(sequent)
#    define PCXX_FORK_KILL         61209     /* R */
# endif

# include "pcxx_events.h"

# ifdef _MASTER_SLAVE_

#   define TRACE_ACS_ON(id) \
    if (pcxx_ev_class & PCXX_EC_SMODE) { \
      pcxx_ev_class |= PCXX_EC_SMODE_ON; \
      pcxx_Event(PCXX_ACS_ON, id); \
     } 


#   define TRACE_ACS_OFF(id) \
    if (pcxx_ev_class & PCXX_EC_SMODE) { \
      pcxx_ev_class &= ~PCXX_EC_SMODE_ON; \
      pcxx_Event(PCXX_ACS_OFF, id); \
     }

# define TRACE_1D_ACS(size) \
  if ((pcxx_ev_class & (PCXX_EC_KERNEL | PCXX_EC_SMODE_ON)) == \
      (PCXX_EC_KERNEL | PCXX_EC_SMODE_ON)) \
    if (!Is_Local(index)) { \
      pcxx_Event(PCXX_ELEM_SIZE, size); \
      pcxx_Event(PCXX_ELEM_ACCESS, \
                (CollectionNumber<<24)+ElemIndex(index)); \
    }

# define TRACE_2D_ACS(size) \
  if ((pcxx_ev_class & (PCXX_EC_KERNEL | PCXX_EC_SMODE_ON)) == \
      (PCXX_EC_KERNEL | PCXX_EC_SMODE_ON)) \
    if (!Is_Local(index1,index2)) { \
      pcxx_Event(PCXX_ELEM_SIZE, size); \
      pcxx_Event(PCXX_ELEM_ACCESS, \
                (CollectionNumber<<24)+ElemIndex(index1,index2)); \
     }

# else

#   define TRACE_1D_ACS(size) \
    PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_ACCESS, \
    (CollectionNumber<<24)+ElemIndex(index))

#   define TRACE_2D_ACS(size) \
    PCXX_EVENT(PCXX_EC_KERNEL, PCXX_ELEM_ACCESS, \
    (CollectionNumber<<24)+ElemIndex(index1,index2))

# define TRACE_ACS_ON(id)
# define TRACE_ACS_OFF(id)

# endif

# else

# define TRACE_1D_ACS(size)
# define TRACE_2D_ACS(size)
# define TRACE_ACS_ON(id)
# define TRACE_ACS_OFF(id)

# endif
# endif
