/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/*                Parallel run-time library                          */
/*                                                                   */
/*                Generic MPI Version                                */
/*                                                                   */
/* Contributions to this module have been made by:                   */
/*                                                                   */
/*     Pete Beckman *** Francois Bodin *** Dennis Gannon             */
/*                Bernd Mohr *** Shelby Yang                         */
/*********************************************************************/

#include "tulip.h"
#include "profiler_defs.h"
#include "pcxx_ariadne.h" 

#ifdef SNEEZY
#include "pcxx_snz.h"
#endif

static void MyIRecv(int type, int mid, char *addr, int size, int from_whom);
#include "passive.h"

/*********************************************************************/
/* EXTERNAL, C-LINKAGE FUNCTION, THAT THE USER MUST PROVIDE: */
extern void tulip_RemoteActionHandler(int from_whom, char *buffer, int len);

extern pcxx_PARAM_t pcxx_param;

/*
#define DEBUG 
*/


/* defined instead in tulip/src/common.c */
/*
int aa_poll_count=0;
*/ /* poll_count for Ariadne */
/* For Ariadne */
#ifdef _AA_REPLAY
FILE *aa_fp; /* the replay file containing the seven fields per event */
char error_msg[256]; /* error string */
int aa_bufinfo; /* For MyRecv() */
int pcxx_ReplayFileInit(int myprocessor_id); /* For Ariadne replay file */
int pcxx_GetReplayMsgInfo(int *oid, int *rwtype, char *msgtag, int *trace_poll_cnt, int *bflag);
int pcxx_InitMsgtagArray(void); /* initialization */
int pcxx_GetBkFlag(int oid, int rwtype, char *msgtag, int stop_at_break_point);
int pcxx_CheckForBreakPoint(int rwtype, int msgtag_no);
int pcxx_IsPPRecv(int t_oid, char *t_msgtag, int *index); 
int aa_valid=0; /* whether the message is valid or stale -default invalid */
int ret; /* return value */
char mtag_string_array[256][64]; /* 64 chars in each string. 256 indices */
#endif

/*********************************************************************/
/* MACHINE DEPENDANT CODE */
/*********************************************************************/

/*********************************************************************/
/* MPI Message Prototypes: */
/*********************************************************************/

#include "mpi.h"

#define MEGABYTE 1048576
#define MY_COMM MPI_COMM_WORLD

#define MAX_PPRECV 512
MPI_Request PPRecv[MAX_PPRECV];
int NextPPRecv, FreePPRecv[MAX_PPRECV];
#ifdef _AA_REPLAY
int oidPPRecv[MAX_PPRECV], typePPRecv[MAX_PPRECV];
#endif
/* store the oid and real message type in the two arrays used to check if 
a recv is  a preposted receive or not */

/*********************************************************************/
static void initPPRecv()
{
  int i;
  
  /* Create a list of available message IDs */
  for (i=0; i<MAX_PPRECV; i++) {
    PPRecv[i]=MPI_REQUEST_NULL;
    FreePPRecv[i]=i;
  }
  NextPPRecv = 0;
}

/*********************************************************************/
void tulip_Abort() { 
  MPI_Abort(MY_COMM,1);
  exit(1); 
}

/*********************************************************************/
/* Receive a buffer that has already arrived, and just waiting */
/* This procedure NEEDS the type and size of the message to be received */
static void MyRecv(int type, int mid, char *addr, int size, int from_whom)
{ 
  MPI_Status status;

#ifdef _AA_REPLAY
  ret = pcxx_GetBkFlag(from_whom, PCXX_AA_MTAG_READ, mtag_string_array[type],0);
  if (ret == 1) pcxx_BreakPoint(); /* reached breakpoint before event */
  aa_valid = 0; /* invalidate record */
#endif /* _AA_REPLAY */ /* EVENT */
 

  MPI_Recv( (void *) addr, size, MPI_BYTE, from_whom, 
            mangle_msgtype(type,mid), MY_COMM, &status);
  /* check status */

#ifdef _AA_REPLAY 
  if (ret == 2) pcxx_BreakPoint(); /* reached breakpoint after event */
#endif /* _AA_REPLAY */  

  pcxx_AriadneTrace (PCXX_EC_ARIADNE, PCXX_IPC, MyNodeNumber, from_whom, 
		     PCXX_AA_MTAG_READ, type, aa_poll_count);
  
}

/*********************************************************************/
/* Post a recv for a buffer NOT YET SENT */
/* This procedure NEEDS the type and size of the message to be received */
static void MyIRecv(int type, int mid, char *addr, int size, int from_whom)
{ 
  int i;

  if (NextPPRecv >= MAX_PPRECV) {
    printf("%d: PrepostedRecv Overflow!\n",tulip_MyContext());
    tulip_Abort(); 
  }

  i = FreePPRecv[NextPPRecv]; ++NextPPRecv;

  MPI_Irecv( (void *) addr, size, MPI_BYTE, from_whom, 
            mangle_msgtype(type,mid), MY_COMM, &PPRecv[i]);
#ifdef _AA_REPLAY
  /* store the oid and the real message type in the two arrays */
  oidPPRecv[i] = from_whom;
  typePPRecv[i] = type;
#ifdef DEBUG
  printf("%d: Preposted RECV:  oid %d, msgtag %s\n",
	MyNodeNumber, from_whom, mtag_string_array[type]);
#endif /* DEBUG */
#endif /* _AA_REPLAY */
/* Don't instrument for Ariadne, do it when preposted recv is received 
in tulip_Poll() function */
/*
  pcxx_AriadneTrace (PCXX_EC_ARIADNE, PCXX_IPC, MyNodeNumber, from_whom, 
		     PCXX_AA_MTAG_READ, type, aa_poll_count);
*/
/*  printf("%d: Preposted RECV: %d, %d, %d\n",
	 tulip_MyContext(),mangle_msgtype(type,mid),size,from_whom);*/

}

/*********************************************************************/
/* Send a buffer to a destination */
static void MySend(int type, int mid, char * buffer, int len, int ToWhom)
{
#ifdef _AA_REPLAY
  ret = pcxx_GetBkFlag(ToWhom, PCXX_AA_MTAG_WRITE, mtag_string_array[type],0);
  if (ret == 1) pcxx_BreakPoint(); /* reached breakpoint before event */
  aa_valid = 0; /* invalidate record */
#endif /* _AA_REPLAY */

  MPI_Bsend( (void *) buffer, len, MPI_BYTE, ToWhom, 
	    mangle_msgtype(type,mid), MY_COMM);

#ifdef _AA_REPLAY 
  if (ret == 2) pcxx_BreakPoint(); /* reached breakpoint after event */
  aa_valid = 0; /* invalidate record */
#endif /* _AA_REPLAY */  
  
  pcxx_AriadneTrace (PCXX_EC_ARIADNE, PCXX_IPC, MyNodeNumber, ToWhom, 
		     PCXX_AA_MTAG_WRITE, type, aa_poll_count);
  
}

/*********************************************************************/
/* Send a buffer to a destination */
static void MySSend(int type, int mid, char * buffer, int len, int ToWhom)
{
#ifdef _AA_REPLAY
  ret = pcxx_GetBkFlag(ToWhom, PCXX_AA_MTAG_WRITE, mtag_string_array[type],0);
  if (ret == 1) pcxx_BreakPoint(); /* reached breakpoint before event */
  aa_valid = 0; /* invalidate record */
#endif /* _AA_REPLAY */
  MPI_Ssend( (void *) buffer, len, MPI_BYTE, ToWhom, 
	    mangle_msgtype(type,mid), MY_COMM);
 
#ifdef _AA_REPLAY 
  if (ret == 2) pcxx_BreakPoint(); /* reached breakpoint after event */
  aa_valid = 0; /* invalidate record */
#endif /* _AA_REPLAY */  
  pcxx_AriadneTrace (PCXX_EC_ARIADNE, PCXX_IPC, MyNodeNumber, ToWhom, 
		     PCXX_AA_MTAG_WRITE, type, aa_poll_count);
 
}

#ifdef SNEEZY
/* Sneezy MPI initialization code. 
 * This could/should be placed in a Sneezy library.
*/
void snz_init_mpi( int *argc, char ***argv )
{
  /* MPI calls main() again after the fork -- this avoids the circularity. */
  static int done = 0;

  char *getenv();
  int ac = *argc, i, portno = 5666;
  char **av = *argv;
  char *s;

  if (done) return;
  done = 1;

  /* PROBLEM: MyTotalNodes _must_ be known before server_init_agent() is
   * called: the sneezy proxy needs it to know how many contexts to listen
   * for.  But the value won't be known until later, after MPI_Init() has
   * run.  So we look for some environment variables and finally a
   * command line argument to get the value.
   */
  
  /* It appears that either the user defines this, or mpirun sets it.
     It should therefore always be there.  However, I'm going to leave
     the code for the other cases in until I can find a place where it
     is documented that mpirun actually sets this variable if not defined.
   */
  if ((s = getenv( "MPI_DEFAULT_NP" )) != 0) {
    MyTotalNodes = atoi( s );
    if (MyTotalNodes <= 0) {
      fprintf( stderr, "Sneezy error: MPI_DEFAULT_NP has illegal value %s\n",
	      s );
      exit ( 1 );
    }
  }
  else if ((s = getenv( "SNZ_DEFAULT_NP" )) != 0) {
    MyTotalNodes = atoi( s );
    if (MyTotalNodes <= 0) {
      fprintf( stderr, "Sneezy error: SNZ_DEFAULT_NP has illegal value %s\n",
	      s );
      exit ( 1 );
    }
  }
  else {
    /* FIXME: This should remove the arguments from the command line. */
    for ( i = 1 ; i < ac ; i++ ) {
      if (strcmp( av[i], "-SNZCNT" ) == 0) {
	if (i == ac-1) {
	  fprintf( stderr, "Usage: -SNZCNT node-count\n" );
	  exit( 1 );
	}
	MyTotalNodes = atoi( av[i+1] );
	break;
      }
    }
    if (i == ac) {
      fprintf( stderr, 
	      "Sneezy Error:\n"
	      "You must define how many processors you are using.\n"
	      "Use either the command line option\n"
	      "   -SNZCNT node-count\n"
	      "or define the environment variable SNZ_DEFAULT_NP\n"
	      "to the number of processors.  Alternatively, Sneezy will\n"
	      "use the environment variable MPI_DEFAULT_NP, if defined.\n" );
      exit( 1 );
    }
  }

  s = getenv( "SNZPORT" );
  if (s != 0) portno = atoi( s );

  snz_server_init_agent( portno, 0 );

  MyTotalNodes = 0;
}
#endif

/*********************************************************************/
static void My_MainBegin(int *argc, char ***argv)
{
  void *extra_buffer;

#ifdef SNEEZY
  snz_init_mpi( argc, argv );
#endif

  /* Pass args to MPI first */
  MPI_Init(argc,argv);

  /* HACK HACK HACK FIXME! */
  extra_buffer = (void *) malloc(MEGABYTE*4);
  if (!extra_buffer) {
    printf("Error My_MainBegin():  Could not allocate extra MPI buffer\n");
    tulip_Abort();
  }
  MPI_Buffer_attach(extra_buffer,MEGABYTE*4);

  /* Grab our args */
  pcxx_InitParam(argc, *argv);

  /* Initialize Timers/Clocks */
  tulip_InitClocks();

  MPI_Comm_size(MPI_COMM_WORLD,&MyTotalNodes);
  MPI_Comm_rank(MPI_COMM_WORLD,&MyNodeNumber);

  pcxx_EvInit(pcxx_param.tracefile);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  barrierFlag = 0;

  initMsgIDs();
  initPPRecv();
  
#ifdef _AA_REPLAY
  pcxx_ReplayFileInit(MyNodeNumber); /* For Ariadne replay file */
  pcxx_InitMsgtagArray(); /* put the strings in the array */
#endif



  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}

/*********************************************************************/
void My_MainEnd()
{
  /* close MPI */
  MPI_Finalize();

  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose();
}

/*********************************************************************/
/* tulip_Poll is changed for Replay */
#ifdef _AA_REPLAY

/* In replay we check the trace for the next communication - see the
poll count and check it with trace pollcount. Do the activity of read
at the appropriate time. As for checking for messages, when the poll
counts match, instead of a nonblocking MPI_Iprobe we use the blocking
MPI_Probe */

void tulip_Poll()
{
  int size, encoded_type, from_whom;
  int real_msg_type, mid ;
  MPI_Status status;
  int index, testany_flag;
  int t_oid, t_rwtype, t_poll_count, t_bflag;   
  char t_msgtag[64];

  PROFILE_ENTER(PROFILE_Poll);
  
  aa_poll_count ++;

/* Handle non-preposed messages here - read from the trace */
  while (1) { /* Keep doing this */
  ret = pcxx_GetReplayMsgInfo( &t_oid, &t_rwtype, t_msgtag, &t_poll_count, &t_bflag);
  if (ret < 0) /* EOF */
  {
#ifdef DEBUG
     printf("pcxx_GetReplayMsgInfo - EOF on trace file encountered\n");
#endif
     PROFILE_EXIT(PROFILE_Poll);
     return;
  }
  
/* if in the poll loop we encounter a non - read event M, W, U , then its
not the appropriate time for the read - process this and exit from the poll
without making the aa_valid = 0 - for reading next event from file */
  if ((t_rwtype != PCXX_AA_MTAG_READ) || (t_poll_count != aa_poll_count))
  /* if its not a read or if there's a mismatch or poll counts then */
  {  /* Check if the count is ok */
    /* pollcount is not to be checked for user events. It could be user
       defined parameter and should have nothing to do with replay. */
    if(t_rwtype == PCXX_AA_MTAG_USER) 
    {
#ifdef DEBUG
	printf("Poll:%d:ignoring pollcount %d for user event %s aa_poll_count %d\n",MyNodeNumber,t_poll_count,t_msgtag,aa_poll_count);
#endif
	return;
    }
    else 
    { /* its not a user event  - verify the poll count */
      if(aa_poll_count > t_poll_count) 
      {
         printf("ERROR pid %d, oid %d, rwtype %d msgtag %s. t_poll_count %d, aa_poll_count %d, t_bflag %d\n", 
  	MyNodeNumber, t_oid, t_rwtype, t_msgtag, t_poll_count, aa_poll_count, t_bflag);
  	printf("PANIC : REPLAY PROBLEM! aa_poll_count exceeds trace poll count\n");
          tulip_Abort();
      }
      else 
      { /* aa_poll_count  < t_poll_count  */
	return;
      /* Wrong message or shouldn't poll right now - go back */ 
      }
    } /* special handling of PCXX_user_event - ignoring poll count */

    PROFILE_EXIT(PROFILE_Poll);
    return;
  } 
  /* This is the correct time. Go on and recv the message */
  /* first check if the recv is a preposted recv */
  if (NextPPRecv > 0) /* there are preposted receives pending */
  { /* is this a preposted receive ? */
    if (pcxx_IsPPRecv(t_oid, t_msgtag,&index) == 1)
    {
      /* it is!  - returns index into the PPRecv request array */
#ifdef DEBUG
      printf("Going for PPRecv ");
#endif /* DEBUG */
      MPI_Wait(&PPRecv[index], &status); /* wait for it to complete*/
      /* after completion, process this */
      /* get info about message from header */
      encoded_type = status.MPI_TAG;
      from_whom    = status.MPI_SOURCE;
      MPI_Get_count(&status, MPI_BYTE, &size);
    
      /* Separate the MessageID (mid) from the message type */
      real_msg_type = mangle_gettype(encoded_type);
      mid           = mangle_getmid(encoded_type);
  
      tulip_SetACK(AckHandle[mid]);
    
      /* Decrement internal ACK count */
      --PendingSysACKs;
    
      /* recycle old MID */
      RecycleMsgID(mid);

      FreePPRecv[NextPPRecv-1]=index;
      --NextPPRecv;
      /* invalidate the event while reading from the trace file */
      aa_valid = 0; /* next fetch now gives a new record from the trace */
      /* Sanity check */
      if ((strcmp(mtag_string_array[real_msg_type], t_msgtag)== 0) && (t_oid == from_whom))
      {
#ifdef DEBUG
	printf("REPLAY SUCCESSFUL *** PPRECV! - t_oid %d, from_whom %d, msgtag %s , t_msgtag %s t_poll_count %d, rpl_poll_count %d\n", 
	t_oid, from_whom, mtag_string_array[real_msg_type], t_msgtag, t_poll_count, aa_poll_count);
#endif /* DEBUG */
	/* next go on to get the next message from the trace - so do continue */
	continue; /* for the while (1) loop */
      } 
      else 
      {
	printf("** REPLAY UNSUCCESSFUL - Failed PPRecv \n");
	printf("Processor %d, t_oid %d, from_whom %d, msgtag %s t_msgtag %s t_poll_count %d rpl_poll_count %d\n",
	MyNodeNumber, t_oid, from_whom, mtag_string_array[real_msg_type], t_msgtag, t_poll_count, aa_poll_count);
	tulip_Abort(); /* Stop */
      } /* replay sanity check */
    } /* if the message is not a preposted receive */
  } /* or if there are no preposted receives pending - then do the usual recv */
  MPI_Probe(t_oid, MPI_ANY_TAG, MY_COMM, &status); /* Block till a message arrives from the other pid */
  /* Here the MPI_ANY_TAG was used because we don't know the mangled type */
  /* Has a (non-preposted) message arrived? */
/* old code 
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MY_COMM, &iprobe_flag, &status);
  while (iprobe_flag) 
 */
  
    /* get info about message from header */
    encoded_type = status.MPI_TAG;
    from_whom    = status.MPI_SOURCE;
    MPI_Get_count(&status, MPI_BYTE, &size);
    
    /* Separate the MessageID (mid) from the message type */
    real_msg_type = mangle_gettype(encoded_type);
    mid           = mangle_getmid(encoded_type);
  /* Sanity check */
    if ((t_oid == from_whom) && (strcmp(t_msgtag, mtag_string_array[real_msg_type]) == 0))
    { 
#ifdef DEBUG
	printf("Poll : oid matches, msgtag matches %s %s\n", t_msgtag, mtag_string_array[real_msg_type]);
#endif /* DEBUG */
    }
    else
    {  
        printf("ERROR - sanity check in tulip_Poll() after MPI_Probe failed\n");
	printf("Error - Poll - t_oid %d, from_whom %d, t_msgtag %s, got %s - %d\n", t_oid, from_whom, t_msgtag, mtag_string_array[real_msg_type], real_msg_type);
	tulip_Abort();
    }

  
    /* Decode the message type */
    switch (real_msg_type) {
    case AskForDataMsg:
      fetch_block_handler(from_whom, mid);
      break;
    case FetchBlockMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfFetchMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;

    case PrefixStoreMsg:
      prefix_store_handler(from_whom, mid);
      break;
    case StoreBlockMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfStoreMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;

    case AckReplyMsg:
      ack_store_block_handler(from_whom, mid);
      break;
    case RemoteActionMsg:
      /* BUFFER COPY ALERT! */
      /* The buffer will need to be "pulled in" to pass to the handler */
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      tulip_RemoteActionHandler(from_whom, ReceiveBuffer, size);
      /* Ok, RemoteAction is done, now send ACK */
      MySend(AckReplyMsg, mid, SendBuffer, 0, from_whom);
      break;

    case BroadcastMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      broadcast_flag = 1;
      break; 
    case ReduceMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      reduction_flag++;
      PROFILE_EXIT(PROFILE_Poll);
      return;
    case BarrierFanInMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      BarrierFanIn(from_whom);
      break;
    case BarrierFanOutMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      barrierFlag = DONE;
      break;
    default:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      printf("Flushing message type %d on node %d\n",real_msg_type, 
	     tulip_MyContext());
      break;
    }
  }

}
#else /* _AA_REPLAY */

void tulip_Poll()
{
  int size, encoded_type, from_whom;
  int real_msg_type, mid, iprobe_flag;
  MPI_Status status;
  int index, testany_flag;

 

  PROFILE_ENTER(PROFILE_Poll);
  aa_poll_count ++; /* the no. of times poll is entered */

  /* Check for pre-posted recvs that might be filled... */
  if (NextPPRecv > 0) {
   MPI_Testany(MAX_PPRECV,PPRecv,&index,&testany_flag,&status);
   if (testany_flag) {
     /* printf("%d: PrepostedRecv Complete! %d\n",tulip_MyContext(),index);*/

    /* get info about message from header */
    encoded_type = status.MPI_TAG;
    from_whom    = status.MPI_SOURCE;
    MPI_Get_count(&status, MPI_BYTE, &size);
    
    /* Separate the MessageID (mid) from the message type */
    real_msg_type = mangle_gettype(encoded_type);
    mid           = mangle_getmid(encoded_type);
  
    tulip_SetACK(AckHandle[mid]);
    
    /* Decrement internal ACK count */
    --PendingSysACKs;
    
    /* recycle old MID */
    RecycleMsgID(mid);

    FreePPRecv[NextPPRecv-1]=index;
    --NextPPRecv;
/* Trace here for preposted receives received. 
   MyIRecv doesnot contain instrumentation */
    pcxx_AriadneTrace (PCXX_EC_ARIADNE, PCXX_IPC, MyNodeNumber, from_whom, 
		     PCXX_AA_MTAG_READ, real_msg_type, aa_poll_count);
   }
 }

  /* Has a (non-preposted) message arrived? */
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MY_COMM, &iprobe_flag, &status);
  while (iprobe_flag) {
  
    /* get info about message from header */
    encoded_type = status.MPI_TAG;
    from_whom    = status.MPI_SOURCE;
    MPI_Get_count(&status, MPI_BYTE, &size);
    
    /* Separate the MessageID (mid) from the message type */
    real_msg_type = mangle_gettype(encoded_type);
    mid           = mangle_getmid(encoded_type);
  
    /* Decode the message type */
    switch (real_msg_type) {
    case AskForDataMsg:
      fetch_block_handler(from_whom, mid);
      break;
    case FetchBlockMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfFetchMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;

    case PrefixStoreMsg:
      prefix_store_handler(from_whom, mid);
      break;
    case StoreBlockMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfStoreMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;

    case AckReplyMsg:
      ack_store_block_handler(from_whom, mid);
      break;
    case RemoteActionMsg:
      /* BUFFER COPY ALERT! */
      /* The buffer will need to be "pulled in" to pass to the handler */
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      tulip_RemoteActionHandler(from_whom, ReceiveBuffer, size);
      /* Ok, RemoteAction is done, now send ACK */
      MySend(AckReplyMsg, mid, SendBuffer, 0, from_whom);
      break;

    case BroadcastMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      broadcast_flag = 1;
      break; 
    case ReduceMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      reduction_flag++;
      PROFILE_EXIT(PROFILE_Poll);
      return;
    case BarrierFanInMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      BarrierFanIn(from_whom);
      break;
    case BarrierFanOutMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      barrierFlag = DONE;
      break;
    default:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      printf("Flushing message type %d on node %d\n",real_msg_type, 
	     tulip_MyContext());
      break;
    }
  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, MY_COMM, &iprobe_flag, &status);
  }

  PROFILE_EXIT(PROFILE_Poll);
  return;
}

#endif /* _AA_REPLAY */

/* Ariadne functions */

/* ARIADNE FUNCTIONS FOR REPLAY */

#ifdef _AA_REPLAY
int pcxx_ReplayFileInit(int myprocessor_id) /* For Ariadne replay file */
{
char filename[256];
  int pid, oid,  msize, bflag, iflag;
  char mtype[64]; /* For RWMU  event type or W_elim etc. */
  char mtag[64]; /* For message tag like PCXX_begin_barrier , AskForDataMsg */
  int ret;
	
	sprintf(filename,"pcxx_aa%d.trace", myprocessor_id);
	aa_fp = fopen(filename,"r+");
	if (aa_fp == (FILE *) NULL)
	{
		sprintf(error_msg, "Couldnot open file %s",filename);
		perror(error_msg);
		tulip_Abort();
	}

	return 1;
}

int pcxx_GetReplayMsgInfo(int *oid, int *rwtype, char *msgtag, int *trace_poll_cnt, int *bflag)
{
static int aa_pid, aa_oid, aa_rwtype, aa_trace_poll_cnt, aa_bflag, aa_iflag;
int ret;
static char aa_msgtag[64];
static char mtype[64];

	if(aa_valid == 0) /* the message in buffer is invalid read from file */
	{
		if ((ret = fscanf(aa_fp, "%d %d %s %s %d %d %d\n",
	        	&aa_pid, &aa_oid, mtype, aa_msgtag, &aa_trace_poll_cnt,
			&aa_bflag, &aa_iflag)) == -1)
		{
/*
			sprintf(error_msg,"ERROR fscanf on replay file fails");
			perror(error_msg);
*/
			return -1; /* EOF encountered */
/*
			tulip_Abort();
*/
		}

		/* Get the message type - Whether its a read, write etc. */
		switch (mtype[0])
		{
		case 'R' : 
			aa_rwtype = PCXX_AA_MTAG_READ;
			break;
		case 'W' : 
			aa_rwtype = PCXX_AA_MTAG_WRITE;
			break;
		case 'M' :
			aa_rwtype = PCXX_AA_MTAG_MCAST;
			break;
		case 'U' :
			aa_rwtype = PCXX_AA_MTAG_USER;
			break;
		default :
			aa_rwtype = -1;
			break;
		}
	} /* otherwise the static ints are returned */
	*oid = aa_oid;	
	*rwtype = aa_rwtype;
	strcpy(msgtag, aa_msgtag);
	*trace_poll_cnt = aa_trace_poll_cnt; 
        *bflag = aa_bflag;
   	aa_valid = 1; /* the message is valid. */
/*
	printf("%d %d %s %s %d %d %d\n", aa_pid, *oid, mtype, msgtag, 
		*trace_poll_cnt, aa_bflag, aa_iflag);
*/
	return 1;
}
	

int pcxx_GetBkFlag(int oid, int rwtype, char *msgtag, int stop_at_break_point)
{
int t_pid, t_oid, t_rwtype, t_poll_count, t_bflag; 
char t_msgtag[64]; /* These are for values obtained from the trace - t_ */

  pcxx_GetReplayMsgInfo( &t_oid, &t_rwtype, t_msgtag, &t_poll_count, &t_bflag);

  if ((oid == t_oid) && (rwtype == t_rwtype) && (strncmp(msgtag,t_msgtag,strlen(msgtag)) == 0)) 
  { 
 /* note : we used strncmp instead of strcmp because we might have
       PCXX_user_event_merge or something like that */
#ifdef DEBUG
          printf("REPLAY correct - pid %d oid %d rwtype %d msgtag %s tr.pollcnt %drpl_pollcnt %d  bflag %d\n",
                MyNodeNumber, oid, t_rwtype, t_msgtag, t_poll_count, aa_poll_count, t_bflag);
       	  printf("msgtag = %s, t_msgtag = %s\n", msgtag, t_msgtag);
#endif 
  }
  else
  {    
        sprintf(error_msg,"REPLAY UNSUCCESSFUL pid %d, oid %d t_oid %d, rw %d, t_rwtype %d, msgtag = %s t_msgtag %s, tr.pollcnt %d, rpl_pollcnt %d\n",
        MyNodeNumber, oid, t_oid, rwtype, t_rwtype, msgtag, t_msgtag, t_poll_count, aa_poll_count);
        printf("%s",error_msg);
        tulip_Abort();
  }
/* if the stop at break point flag is enabled, stop at the break point */
  if ((stop_at_break_point == 1) && (t_bflag > 0)) 
  { /* break point reached invalidate the event from the global buffer */
	pcxx_BreakPoint();
  }
/* else continue as it is */	
  return t_bflag; /* returns the 0 or 1 or 2 for no before and after event
	break point setting */

}

int pcxx_InitMsgtagArray(void)
{ /* we have a global array - mtag_string_array which holds all the tags and
     their names - it is initialized here */
     
    /* first the regular RW messages with pC++ types defined in passive.h */
    strcpy(mtag_string_array[ BarrierFanOutMsg ], "BarrierFanOutMsg");	
    strcpy(mtag_string_array[ BarrierFanInMsg ],  "BarrierFanInMsg");	
    strcpy(mtag_string_array[ BroadcastMsg ],     "BroadcastMsg");	
    strcpy(mtag_string_array[ ReduceMsg ],        "ReduceMsg");	
    strcpy(mtag_string_array[ AskForDataMsg ],    "AskForDataMsg");	
    strcpy(mtag_string_array[ FetchBlockMsg ],    "FetchBlockMsg");	
    strcpy(mtag_string_array[ EndOfFetchMsg ],    "EndOfFetchMsg");	
    strcpy(mtag_string_array[ PrefixStoreMsg ],   "PrefixStoreMsg");	
    strcpy(mtag_string_array[ StoreBlockMsg ],    "StoreBlockMsg");	
    strcpy(mtag_string_array[ EndOfStoreMsg ],    "EndOfStoreMsg");	
    strcpy(mtag_string_array[ RemoteActionMsg ],  "RemoteActionMsg");	
    strcpy(mtag_string_array[ AckReplyMsg ],      "AckReplyMsg");	
    strcpy(mtag_string_array[ BarrierFanInMsg ],  "BarrierFanInMsg");	

    /* next the pC++ Ariadne events defined in pcxx_ariadne.h */
    strcpy(mtag_string_array[ PCXX_error ],  
			     "PCXX_error");	
    strcpy(mtag_string_array[ PCXX_thread_terminate ],  
			     "PCXX_thread_terminate");	
    strcpy(mtag_string_array[ PCXX_create_collection ],  
		             "PCXX_create_collection");	
    strcpy(mtag_string_array[ PCXX_delete_collection ],  
			     "PCXX_delete_collection");	
    strcpy(mtag_string_array[ PCXX_begin_barrier ],  
			     "PCXX_begin_barrier");	
    strcpy(mtag_string_array[ PCXX_end_barrier ],  
		             "PCXX_end_barrier");	
    strcpy(mtag_string_array[ PCXX_begin_fetch ],  
			     "PCXX_begin_fetch");	
    strcpy(mtag_string_array[ PCXX_end_fetch ],  
			     "PCXX_end_fetch");	
    strcpy(mtag_string_array[ PCXX_begin_put ],  
			     "PCXX_begin_put");	
    strcpy(mtag_string_array[ PCXX_end_put ],  
			     "PCXX_end_put");	
    strcpy(mtag_string_array[ PCXX_begin_rpc ],  
			     "PCXX_begin_rpc");	
    strcpy(mtag_string_array[ PCXX_end_rpc ],  
			     "PCXX_end_rpc");	
    strcpy(mtag_string_array[ PCXX_begin_parallel ],  
			     "PCXX_begin_parallel");	
    strcpy(mtag_string_array[ PCXX_enter_element ],  
			     "PCXX_enter_element");	
    strcpy(mtag_string_array[ PCXX_exit_element ],  
			     "PCXX_exit_element");	
    strcpy(mtag_string_array[ PCXX_begin_service_rpc ],  
			     "PCXX_begin_service_rpc");	
    strcpy(mtag_string_array[ PCXX_end_service_rpc ],  
			     "PCXX_end_service_rpc");	
    strcpy(mtag_string_array[ PCXX_user_event ],  
			     "PCXX_user_event");	
    strcpy(mtag_string_array[ PCXX_begin_create_collection ],  
			     "PCXX_begin_create_collection");	
    strcpy(mtag_string_array[ PCXX_end_create_collection ],  
			     "PCXX_end_create_collection");	
    strcpy(mtag_string_array[ PCXX_begin_delete_collection ],  
			     "PCXX_begin_delete_collection");	
    strcpy(mtag_string_array[ PCXX_end_delete_collection ],  
			     "PCXX_end_delete_collection");	
    strcpy(mtag_string_array[ PCXX_end_parallel ], 
			     "PCXX_end_parallel");	

    return 1;
}

int pcxx_BreakPoint (void)
{ /* do nothing - break point is reached */
   printf("INSIDE pcxx_BreakPoint - process %d\n",MyNodeNumber);
#ifdef SNEEZY
   snz_user_event(PCXX_user_event, NULL, 0); 
	/* other 2 parameters pass data and length */
#endif
   return 1;
}

int pcxx_CheckForBreakPoint(int rwtype, int msgtag_no)
{
   pcxx_GetBkFlag(MyNodeNumber, rwtype, mtag_string_array[msgtag_no], 1);
   aa_valid = 0; /* invalidate the record */
   return 1;
}

int pcxx_IsPPRecv(int t_oid, char *t_msgtag, int *index) 
{ /* This function will check if a receive is a preposted recv or not */
  /* at the time of posting a preposted receive, we keep the oid and the type
   in arrays and we check these with the trace oids and msgtags to see if it is
   a PPRecv - return 1 for success or 0 for failure and return the index too. */
int i;
  
  for (i = 0; i < NextPPRecv; i++) /* scan the arrays */
  { 
    if (oidPPRecv[i] == t_oid) /* other process ids match now look for tags */
    {
      if (strcmp(mtag_string_array[typePPRecv[i]], t_msgtag) == 0)
      {
 	/* mtag_string_array[msgtag_no] gives the string. msgtag_no is the 
	real message type stored in typePPRecv[i] */
#ifdef DEBUG
	printf("pcxx_IsPPRecv - true - matches! t_oid = %d, oidPPRecv[i] %d, msgtag %s, t_msgtag %s\n", t_oid, oidPPRecv[i], 
	mtag_string_array[typePPRecv[i]], t_msgtag);
#endif /* DEBUG */
	
	/* success */
        *index = i; /* index into the PPRecv array */
        return 1; 
      }
           /* else do nothing - scan the entire array */
    }
  }
  /* didn't find anything */ /* this is not a preposted recv */
#ifdef DEBUG
  printf("not  a  PPRecv t_oid %d, t_msgtag %s\n", t_oid, t_msgtag);
#endif /* DEBUG */
  return 0;
}
#endif /* _AA_REPLAY */

#include "SPMD.h"

