/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/*                Parallel run-time library                          */
/*                                                                   */
/*                Paragon NX Version                                 */
/*                                                                   */
/* Contributions to this module have been made by:                   */
/*                                                                   */
/*     Pete Beckman *** Francois Bodin *** Dennis Gannon             */
/*                Bernd Mohr *** Shelby Yang                         */
/*********************************************************************/

#include "tulip.h"
#include "profiler_defs.h"

static void MyIRecv(int type, int mid, char *addr, int size, int from_whom);
#include "passive.h"

/*********************************************************************/
/* EXTERNAL, C-LINKAGE FUNCTION, THAT THE USER MUST PROVIDE: */
extern void tulip_RemoteActionHandler(int from_whom, char *buffer, int len);

extern pcxx_PARAM_t pcxx_param;

/*********************************************************************/
/* MACHINE DEPENDANT CODE */
/*********************************************************************/

/*********************************************************************/
/* NX Message Prototypes:
/*********************************************************************/
/*#include <nx.h>*/

#define MEGABYTE 1048576

#define MAX_PPRECV 256
long PPRecv[MAX_PPRECV];
int NextPPRecv;
long dummy[256][10];

/*********************************************************************/
static void initPPRecv()
{
  int i;
  
  /* Initialzie message IDs */
  for (i=0; i<MAX_PPRECV; i++)
    PPRecv[i] = (long) -1;

  NextPPRecv = 0;
}

/*********************************************************************/
void tulip_Abort() { 
  exit(1); 
}

/*********************************************************************/
/* Receive a buffer that has already arrived, and just waiting */
/* This procedure NEEDS the type and size of the message to be received */
static void MyRecv(int type, int mid, char *addr, int size, int from_whom)
  { crecv( (long)mangle_msgtype(type,mid), (char*) addr, (long)size); 
/*  printf("%d:     Blocking RECV: %6d, %3d, %3d, %3d\n",
	 tulip_MyContext(),size,type,mid,from_whom);*/

}
 
/*********************************************************************/
/* Post a recv for a buffer NOT YET SENT */
/* This procedure NEEDS the type and size of the message to be received */
static void MyIRecv(int type, int mid, char *addr, int size, int from_whom)
{ 
  int i,j=0;

  if (NextPPRecv >= MAX_PPRECV) {
    printf("%d: To many Preposted Recvs: waiting for msgs to catch up!\n",
	   tulip_MyContext());
    while (NextPPRecv > 32) {
      j++;
      if (j > 1000) {
	printf("%d: Outstanding PPRecvs: %d\n",tulip_MyContext(),NextPPRecv);
	j = 0;
      }
      tulip_Poll();
    }
  }

  i = NextPPRecv; NextPPRecv++;

  PPRecv[i] = irecvx( (long)mangle_msgtype(type,mid), (char*) addr,
		     (long)size, from_whom, -1, &(dummy[i][0]) );

  if (PPRecv[i] == (long) -1) {
    printf("%d: irecvx failed in MyIRecv()\n",tulip_MyContext());
    tulip_Abort();
  }

/*  printf("%d:    Preposted RECV: %6d, %3d, %3d, %3d\n",
	 tulip_MyContext(),size,type,mid,from_whom);*/

}

/*********************************************************************/
/* Send a buffer to a destination */
static void MySend(int type, int mid, char * buffer, int len, int ToWhom)
  { csend((long) mangle_msgtype(type,mid), 
        (long *) buffer, (long) len, (long) ToWhom, 0);
/*  printf("%d: non-blocking Send: %6d, %3d, %3d, %3d\n",
	 tulip_MyContext(),len,type,mid,ToWhom);*/
 }
 
/*********************************************************************/
/* Send a buffer to a destination */
static void MySSend(int type, int mid, char * buffer, int len, int ToWhom)
  { csend((long) mangle_msgtype(type,mid), 
        (long *) buffer, (long) len, (long) ToWhom, 0);
/*  printf("%d: non-blocking Send: %6d, %3d, %3d, %3d\n",
	 tulip_MyContext(),len,type,mid,ToWhom);*/
 }
 
/*********************************************************************/
static void My_MainBegin(int *argc, char ***argv)
{

  /* Grab our args */
  pcxx_InitParam(argc, *argv);

  /* Initialize Timers/Clocks */
  tulip_InitClocks();

  MyNodeNumber = mynode();
  MyTotalNodes = numnodes();

  pcxx_EvInit(pcxx_param.tracefile);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  barrierFlag = 0;

  initMsgIDs();
  initPPRecv();

  PCXX_EVENT(PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}

/*********************************************************************/
static void My_MainEnd()
{
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose();
}

/*********************************************************************/
void tulip_Poll()
{
  int size, encoded_type, from_whom;
  int real_msg_type, mid;
  int index, flag;

  PROFILE_ENTER(PROFILE_Poll);

  /* Has a system-buffered message arrived? */
  while (iprobe(-1) > 0) {
  
    /* get info about message from header */
    encoded_type = (int) infotype();
    from_whom    = (int) infonode();
    size         = (int) infocount();
    
    /* Separate the MessageID (mid) from the message type */
    real_msg_type = mangle_gettype(encoded_type);
    mid           = mangle_getmid(encoded_type);
  
    /* Decode the message type */
    switch (real_msg_type) {
    case AskForDataMsg:
      fetch_block_handler(from_whom, mid);
      break;
    case FetchBlockMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfFetchMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;

    case PrefixStoreMsg:
      prefix_store_handler(from_whom, mid);
      break;
    case StoreBlockMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfStoreMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;

    case AckReplyMsg:
      ack_store_block_handler(from_whom, mid);
      break;
    case RemoteActionMsg:
      /* BUFFER COPY ALERT! */
      /* The buffer will need to be "pulled in" to pass to the handler */
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      tulip_RemoteActionHandler(from_whom, ReceiveBuffer, size);
      /* Ok, RemoteAction is done, now send ACK */
      MySend(AckReplyMsg, mid, SendBuffer, 0, from_whom);
      break;

    case BroadcastMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      broadcast_flag = 1;
      break; 
    case ReduceMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      reduction_flag++;
      PROFILE_EXIT(PROFILE_Poll);
      return;
    case BarrierFanInMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      BarrierFanIn(from_whom);
      break;
    case BarrierFanOutMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      barrierFlag = DONE;
      break;
    default:
	/* A message arrived that I don't know what to do with, purge it... */
	MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
	printf("Flushing message type %d on node %d\n",real_msg_type, 
	       tulip_MyContext());
      break;
    }
  }

  /* Check for which pre-posted recvs was filled...  Unfortunately, we
     don't have a MPI style Testany, so we have to Cycle through all
     the preposed recvs to find the one that is done.  We don't have a
     linked list, we have a stack-ish vector, so we may have holes,
     which we ignore.  A hole is a -1 */

  index = 0;
  flag  = 0;
  while ( (flag == 0) && (index < NextPPRecv) ) {
    if  ( (PPRecv[index] != (long) -1) && (msgdone(PPRecv[index])) )
      flag = 1;
    else
      index++;
  }

  /* Ok, did we find one? */
  if (flag == 1) {

    /* OK, we have a pre-posted recv that has completed!! */

    /* get info about message from header */
    encoded_type = dummy[index][0];
    size         = dummy[index][1];
    from_whom    = dummy[index][2];
    
    /* Separate the MessageID (mid) from the message type */
    real_msg_type = mangle_gettype(encoded_type);
    mid           = mangle_getmid(encoded_type);

    tulip_SetACK(AckHandle[mid]);
    
    /* Decrement internal ACK count */
    --PendingSysACKs;
    
    /* recycle old MID */
    RecycleMsgID(mid);
    
    /* Ok, do a little compression on the PPRecv list. */
    /* Fill in the newly created hole. */
    PPRecv[index]        = PPRecv[NextPPRecv-1];
    PPRecv[NextPPRecv-1] = -1;               /* Mark slot empty */
    
    --NextPPRecv;
  }

  PROFILE_EXIT(PROFILE_Poll);
  return;
}

/*********************************************************************/
/* BROADCAST */
/*********************************************************************/
/* pcxx_BrodcastBytes sends messages to all other nodes from the node
with flag = 0.  The message is in the buffer and the length is known.
This function blocks until all processors call this function. */

void tulip_NativeBroadcastBytes(int flag, int length, void * buffer)
{
  int i;
  char *c, *p;

  /*BlockDataBuff = (char*)buffer;*/

  /* The following barrier is very important.  First, it prevents a receiver
     from receiving more than one broadcast during poll which may
     overwrite the receiving buffer.  Second, it prevents
     a receiver from receiving a broadcast when it is not ready to receive
     the broadcast, i.e., when it doest not have a buffer to save the
     broadcast.  Also, do not move the above assignment statement to anywhere
     else.            --- S.Y. (2/11/94)
   */
  tulip_Barrier();

  if (flag == 0) {
    MySend(BroadcastMsg, NOMID, buffer, length, -1);
  } else {
    while (broadcast_flag == 0) tulip_Poll();
    memcpy((char*)buffer, ReceiveBuffer, length);
  }
  broadcast_flag = 0;
}


/*********************************************************************/
/* REDUCTIONS */
/*********************************************************************/
/* tulip_NativeReduceDoubleAdd adds together corresponding buffer
elements and broadcasts the resulting buffer to all nodes.  The
original buffer content is overwritten. */

/*********************************************************************/
void tulip_NativeReduceDoubleAdd(void* buffer,int elem_count)
{
  if(elem_count*sizeof(double) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gdsum((double*)buffer,elem_count,(double*)ReceiveBuffer);  
}

/*********************************************************************/
void tulip_NativeReduceFloatAdd(void* buffer,int elem_count)
{
  if(elem_count*sizeof(float) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gssum((float*)buffer,elem_count,(float*)ReceiveBuffer);
}

/*********************************************************************/
void tulip_NativeReduceIntAdd(void* buffer,int elem_count)
{
  if(elem_count*sizeof(int) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gisum((int*)buffer,elem_count,(int*)ReceiveBuffer);
}

/*********************************************************************/
void tulip_NativeReduceDoubleMax(void* buffer,int elem_count)
{
  if(elem_count*sizeof(double) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gdhigh((double*)buffer,elem_count,(double*)ReceiveBuffer);  
}

/*********************************************************************/
void tulip_NativeReduceFloatMax(void* buffer,int elem_count)
{
  if(elem_count*sizeof(float) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gshigh((float*)buffer,elem_count,(float*)ReceiveBuffer);  
}

/*********************************************************************/
void tulip_NativeReduceIntMax(void* buffer,int elem_count)
{
  if(elem_count*sizeof(int) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gihigh((int*)buffer,elem_count,(int*)ReceiveBuffer);  
}

/*********************************************************************/
void tulip_NativeReduceDoubleMin(void* buffer,int elem_count)
{
  if(elem_count*sizeof(double) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gdlow((double*)buffer,elem_count,(double*)ReceiveBuffer);  
}

/*********************************************************************/
void tulip_NativeReduceFloatMin(void* buffer,int elem_count)
{
  if(elem_count*sizeof(float) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gslow((float*)buffer,elem_count,(float*)ReceiveBuffer);  
}

/*********************************************************************/
void tulip_NativeReduceIntMin(void* buffer,int elem_count)
{
  if(elem_count*sizeof(int) > MAXBUFSIZE) 
    tulip_Message("Reduction requires a too large buffer.", TULIP_FATAL);
  gilow((int*)buffer,elem_count,(int*)ReceiveBuffer);  
}

#include "SPMD.h"
