/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/*                Parallel run-time library                          */
/*                                                                   */
/*                PVM Version                                        */
/*                                                                   */
/* Contributions to this module have been made by:                   */
/*                                                                   */
/*     Pete Beckman *** Francois Bodin *** Dennis Gannon             */
/*                Bernd Mohr *** Shelby Yang                         */
/*********************************************************************/
 
#include "tulip.h"
#include "profiler_defs.h"

#include <string.h>

#ifndef NeXT
#  include <unistd.h>
#endif

#include "passive.h"

/*********************************************************************/
/* EXTERNAL, C-LINKAGE FUNCTION, THAT THE USER MUST PROVIDE: */
extern void tulip_RemoteActionHandler(int from_whom, char *buffer, int len);
 
extern pcxx_PARAM_t pcxx_param;

/**************************************************************************/
/* MACHINE DEPENDANT CODE */
/**************************************************************************/

/*********************************************************************/
/* Machine Dependant Message Prototypes:
/*********************************************************************/
#include "pvm3.h"
int MyTaskIds[PCXX_MAXPROCS];
#define ANY -1

/*******************************************************************/
void tulip_Abort()
{
  pvm_exit();
  exit(1);
}

/*********************************************************************/
/* Receive a buffer that has already arrived, and just waiting */
/* This procedure NEEDS the type and size of the message to be received */
static void MyRecv(int type, int mid, char *addr, int size, int from_whom)
{
  int bufid;

  /* grab incomming message */
  bufid = pvm_recv(MyTaskIds[from_whom],mangle_msgtype(type,mid));
  if (bufid < 0)  tulip_Message("MyRecv: pvm_recv failed",TULIP_FATAL);

/*printf("%d: receiving %d bytes from %d\n",MyNodeNumber,size,from_whom);*/

  /* OK, pvm has a buffer for me, now I must unpack it! */
  bufid = pvm_upkbyte(addr,size,1);
  if (bufid < 0) tulip_Message("MyRecv: pvm_upkbyte (2) failed",TULIP_FATAL);
}

/*******************************************************************/
/* Send out a buffer to a destination */
static void MySend(int type, int mid, char * buffer, int len, int ToWhom)
{
  int info,bufid,msgtag;

/*printf("%d: sending %d bytes to %d\n",MyNodeNumber,len,ToWhom);*/

  /* Clear send buffer, specify no out-going buffer copy */
  bufid = pvm_initsend( PvmDataRaw );  
  if (bufid < 0) tulip_Message("MySend: pvm_initsend failed",TULIP_FATAL);
  
  /* Pack data into send buffer */
  info = pvm_pkbyte(buffer,len,1);
  if (info < 0)   tulip_Message("MyRecv: pvm_pkbyte failed",TULIP_FATAL);

  /* Send data */
  info = pvm_send(MyTaskIds[ToWhom], mangle_msgtype(type,mid));
  if (info < 0) tulip_Message("MySend: pvm_send failed",TULIP_FATAL);
}

/*******************************************************************/
/* Send out a buffer to a destination */
static void MySSend(int type, int mid, char * buffer, int len, int ToWhom)
{
  int info,bufid,msgtag;

/*printf("%d: sending %d bytes to %d\n",MyNodeNumber,len,ToWhom);*/

  /* Clear send buffer, specify no out-going buffer copy */
  bufid = pvm_initsend( PvmDataRaw );  
  if (bufid < 0) tulip_Message("MySend: pvm_initsend failed",TULIP_FATAL);
  
  /* Pack data into send buffer */
  info = pvm_pkbyte(buffer,len,1);
  if (info < 0)   tulip_Message("MyRecv: pvm_pkbyte failed",TULIP_FATAL);

  /* Send data */
  info = pvm_send(MyTaskIds[ToWhom], mangle_msgtype(type,mid));
  if (info < 0) tulip_Message("MySend: pvm_send failed",TULIP_FATAL);
}

/* **********************************************************************/
static void My_MainBegin(int *argc, char ***argv)
{
  int mytid, me, i, x;
  int info;
  char *pwd;
  char *DelimiterPos;
  char basename[100];

  int nhost, narch;
  struct pvmhostinfo *hostp;
  
  pcxx_InitParam (argc, *argv);
  tulip_InitClocks();

  /* enroll in pvm */
  mytid = pvm_mytid();
  if (mytid < 0) {
    fprintf(stderr,"RTS: *********************************************\n");
    fprintf(stderr,"RTS: ERROR, Could not contact PVM daemon\n");
    fprintf(stderr,"RTS:        Are you sure you started pvmd?\n");
    fprintf(stderr,"RTS: *********************************************\n");
    tulip_Abort();
  }

  info = pvm_config(&nhost, &narch, &hostp);
  if (info < 0) printf("Error: pvm_config failed \n");

  /* find out if I am parent or child */
  MyTaskIds[0] = pvm_parent();

/* If we saved stdout messages, and wrote them to screen, we would use this */
/*
  info = pvm_setopt(PvmOutputTid,mytid);
  if (info < 0) tulip_Message("Error: pvm_setopt() failed",TULIP_FATAL);
  info = pvm_setopt(PvmOutputCode,999);
  if (info < 0) tulip_Message("Error: pvm_setopt() failed",TULIP_FATAL);
*/
  if( MyTaskIds[0] < 0 ) {       /* then I am the parent */
    MyTaskIds[0] = mytid;
    me = 0;

    /* Find out the name of this binary... */
    DelimiterPos = (char *) strrchr(*argv[0],47);  /* find last "/" */
    if (DelimiterPos) {  /* There was a path, remove it for PVM calls */
      strcat(basename,DelimiterPos);
    }
    else /* there was no path, just use executable name */
      strcat(basename,*argv[0]);

    if (nhost == 1) {
      printf("RTS: There is only one machine in this configuration,\n");
      printf("RTS:    %d processes will be spawned onto THIS host\n",
	     pcxx_param.numproc);

      i = pvm_spawn(basename, &(*argv[1]), PvmTaskDefault,
		    0, pcxx_param.numproc, &MyTaskIds[1]);
      if (i != pcxx_param.numproc) {
	printf("RTS: Error, could not spawn tasks onto %s (error=%d)\n",
			 hostp[0].hi_name,i);
	tulip_Abort();
      }
    }
    else
    if (pcxx_param.numproc > nhost) {
      printf("RTS: You requested %d processes.\n",pcxx_param.numproc);
      printf("RTS:    There are %d hosts in this PVM configuration.\n",nhost);
      printf("RTS:    PVM will allocate more than one process per host.\n");
      tulip_Abort();
    }
    else {
      printf("RTS: %d processes will be created\n",pcxx_param.numproc);
      printf("RTS:    There are %d hosts in this PVM configuration.\n",nhost);
      for (x=1; x<pcxx_param.numproc; x++) {
	i = pvm_spawn(basename, &(*argv[1]), 1, 
		      hostp[x].hi_name, 1, &MyTaskIds[x]);
	if (i != 1) printf("Error, could not spawn task onto %s (error=%d)\n",
			   hostp[x].hi_name,i);

	printf("Spawned 'main()' onto %20s:  MyTaskIds[%2d]=%8x\n"
	       ,hostp[x].hi_name,x,MyTaskIds[x]);
      }
    }

    /* multicast MyTaskIds array to children */
    pvm_initsend( PvmDataDefault );
    pvm_pkint(MyTaskIds, pcxx_param.numproc, 1);
    printf("Broadcasting MyTaskIds data (mcast)...\n");
printf("I Am MyTaskIds[%d]=%x\n",0,MyTaskIds[0]);
    for (x=1; x<=pcxx_param.numproc-1; x++) {
printf("Sending MyTaskIds[%d]=%x\n",x,MyTaskIds[x]);
      i = pvm_send(MyTaskIds[x], 0);
      if (i < 0)
	printf("ERROR pvm_send failed\n");
    }
  }

  else { /* I am a child */
    /* receive MyTaskIds array */
    printf("Spawned task receiving MyTaskIds %x\n",MyTaskIds[0]);
    i = pvm_recv(MyTaskIds[0], 0);
    pvm_upkint(MyTaskIds, pcxx_param.numproc, 1);
    for( i=1; i<pcxx_param.numproc ; i++ )
      if( mytid == MyTaskIds[i] ){ me = i; break; }
  }

/*
  info = pvm_setopt(PvmAutoErr,2);
  if (info < 0) tulip_Message("Error: pvm_setopt() failed",TULIP_FATAL);
*/

  /* 'me' is now set for every processor, and tid[procnum] is the taskid */

  MyNodeNumber = me; MyTotalNodes = pcxx_param.numproc;

  /* we would like to do this earlier, but unfortuneately, we need
     MyNodeNumber for doing this */
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  if (me > MyTotalNodes-1) { pvm_exit(); exit(1); }

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  barrierFlag = 0;

  initMsgIDs();

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
}

/*******************************************************************/
static void My_MainEnd() {
  pvm_exit();
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  pcxx_EvClose();
}

/*********************************************************************
 POLL FUNCTION
 *********************************************************************/
void tulip_Poll()
{
  int size, encoded_type, from_whom;
  int real_msg_type, mid, i, bufid;

  PROFILE_ENTER(PROFILE_Poll);

  /* Has a message arrived? */
  while ((bufid = pvm_probe(ANY,ANY)) != 0) {
    if (bufid < 0)
      tulip_Message("tulip_Poll(): pvm_probe error",TULIP_FATAL);

    /* get info about message from header */
    bufid = pvm_bufinfo(bufid,&size,&encoded_type,&from_whom);
    if (bufid < 0) tulip_Message("tulip_Poll(): pvm_bufinfo error",TULIP_FATAL);

    /* Note: from_whom currently the TID, not the processor number */
    /* Nasty linear search.... */
    for (i=0; i<PCXX_MAXPROCS; i++)
      if (MyTaskIds[i] == from_whom) { from_whom = i; break; }

    /* Separate the MessageID (mid) from the message type */
    real_msg_type = mangle_gettype(encoded_type);
    mid           = mangle_getmid(encoded_type);
  
    /* Decode the message type */
    switch (real_msg_type) {
    case AskForDataMsg:
      fetch_block_handler(from_whom, mid);
      break;
    case FetchBlockMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfFetchMsg:
      rcvd_fetch_block_handler(from_whom, mid, real_msg_type);
      break;

    case PrefixStoreMsg:
      prefix_store_handler(from_whom, mid);
      break;
    case StoreBlockMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;
    case EndOfStoreMsg:
      store_block_handler(from_whom, mid, real_msg_type);
      break;

    case AckReplyMsg:
      ack_store_block_handler(from_whom, mid);
      break;
    case RemoteActionMsg:
      /* BUFFER COPY ALERT! */
      /* The buffer will need to be "pulled in" to pass to the handler */
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      tulip_RemoteActionHandler(from_whom, ReceiveBuffer, size);
      /* Ok, RemoteAction is done, now send ACK */
      MySend(AckReplyMsg, mid, SendBuffer, 0, from_whom);
      break;

    case BroadcastMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      broadcast_flag = 1;
      break; 
    case ReduceMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      reduction_flag++;
      PROFILE_EXIT(PROFILE_Poll);
      return;
    case BarrierFanInMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      BarrierFanIn(from_whom);
      break;
    case BarrierFanOutMsg:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      barrierFlag = DONE;
      break;
    default:
      MyRecv(real_msg_type, mid, ReceiveBuffer, size, from_whom);
      printf("Flushing message type %d on node %d\n",real_msg_type, 
	     tulip_MyContext());
      break;
    }
  }

  PROFILE_EXIT(PROFILE_Poll);
  return;
}

