/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* Written by Suvas Vajracharya by modifying Pete Beckman sgilib.c version */
/* 9/15/95 */

/*==============================================================*/
/* awelib : AWESIME/DUDE run-time support library       	*/
/*==============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include "tulip.h"
#include "profiler_defs.h"

extern pcxx_PARAM_t pcxx_param;

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */
#define MAX_PROC 128

/*****************************/
/* Structs                   */
/*****************************/
struct ArgcArgv {
  int i;
  int argc;
  char **argv;
};

struct RA_params {
  char * buffer;
  int    length;
  tulip_ACK_t *ackhdl;
  int from_whom;
};
 
struct RA_params **MyRAQueue;

struct msg_params {
  char * address;
  int    length;
};
 
/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  */
#define MAXBUFF 4096

/*******************************************************************/
/* Shared data structures */
/*******************************************************************/
char *BroadcastBuff;           /* a place to store broacast values */
char *ReduceBuff[MAX_PROC];    /* a place to store reduction values */
struct RA_params ***GlobalRATable; /* after init, read-only */

int MyTotalNodes = -1;     /* number of processors */
int MyNodeNumber = -1;     /* who I am */

void * tulip_SharedMalloc(int size) {
    return (void *) malloc(size);
}

void tulip_SharedFree(void *p) {
   free(p);

}
int SMRTSSetup()
{
  int i;

  /* allocate processors */
  MyTotalNodes = pcxx_param.numproc;


/*  MyNodeNumber = MyTotalNodes - 1; */
  MyNodeNumber = 0;
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  BroadcastBuff = (char *) malloc(MAXBUFF);

  /* allocate shared locks and reduction buffer */
  if(MyTotalNodes > MAX_PROC) {
    fprintf(stderr,"Error in awelib.c: MAX_PROC < number of processors.\n");
    exit(0);
  }
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) malloc(MAXBUFF);
  }
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return(0);
}


/* tulip_BroadcastBytes  */
void tulip_BroadcastBytes(int flag, int length, void *buffer)
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive tulip_BroadcastBytes are called */
  tulip_Barrier();

  if ( length > MAXBUFF ) 
    tulip_Message("Broadcast buffer too small!  Increase buffer.",TULIP_FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (flag == 0) for (i=0; i<length; i++) c[i] = p[i];
  tulip_Barrier();
  if (flag != 0) for (i=0; i<length; i++) p[i] = c[i];
}

/*
 @ SMRTSDealloc : deallocate shared memory pointers
 */

void SMRTSDealloc(void *ptr)
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  free(ptr);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}


/*********************************************************************/
/* Check if a "handle" has completed (non-blocking) */
void  tulip_Abort()
{
  printf("Exiting!!!!! ERROR\n");
  exit(1);
}

/*********************************************************************/
int tulip_Get(char *dst, tulip_GlobalPointer_t src, int len, 
	      tulip_ACK_t *ackhandle)
{
  memcpy(dst, tulip_ToLocalPointer(&src), len);
  tulip_SetACK(ackhandle);
  return 0;
}  

/*********************************************************************/
int tulip_Put(tulip_GlobalPointer_t dst, char *src, int len, 
	      tulip_ACK_t *ackhandle)
{
  memcpy(tulip_ToLocalPointer(&dst), src, len);
  tulip_SetACK(ackhandle);
  return 0;
}  

/*********************************************************************/
/*********************************************************************/
/* Check for RemoteAction requests */
void tulip_Poll()
{

}
/*********************************************************************/
int tulip_RemoteAction(int remote_node, char *buffer, int len, 
		       tulip_ACK_t *ackhandle)
{

  /* RemoteAction to same uniproc node never happens?....???? */
  tulip_SetACK(ackhandle);
  return 0;

}

int SMRTSFork(int argc, char **argv)
{
  int k;
  struct ArgcArgv *acav;

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);
  tulip_ParallelMain(argc,argv);  /* Transfer control to compiler/user code */
  /* finished */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return(0);
}

/*
 @ SMRTSNewProc : more book-keeping for newly-created process
 */

int SMRTSNewProc(num)
  int num;
{
  /* record my ID */
  MyNodeNumber = num;
  tulip_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  return(MyNodeNumber);
}

/*
 @ barrier : wait at general barrier
 */

void pcxx_TestBarrier()
{
/*printf("%d: starting Barrier (%d)\n",MyNodeNumber,MyTotalNodes);*/

/*printf("%d: Ending Barrier (%d)\n",MyNodeNumber,MyTotalNodes);*/
}

/***********************************************************************/
/* Pete's super-fast barrier */
/***********************************************************************/
void tulip_Barrier()
{

}




void SMRTSWaitToEnd() {
  if ((int) MyNodeNumber != (MyTotalNodes-1))
    tulip_Barrier();
}

