/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* Written by Pete Beckman (by modifying Bernd Mohr's symmetry version) */
/* 6/22/94 */

/*==============================================================*/
/* sgilib : SGI run-time support library			*/
/*==============================================================*/

#include <sys/types.h>
#include <sys/prctl.h>
/*#include <ulocks.h>*/
#include <mutex.h>
#include <task.h>
#include <stdio.h>
#include <sys/sysmp.h>
#include <unistd.h>
 
#include "tulip.h"
#include "profiler_defs.h"

#define CACHE_LINE 1024

extern pcxx_PARAM_t pcxx_param;

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  MAX_PROC has to be >= MyTotalNodes */
#define MAX_PROC 128

/*****************************/
/* Structs                   */
/*****************************/
struct ArgcArgv {
  int i;
  int argc;
  char **argv;
};

struct RA_params {
  char * buffer;
  int    length;
  tulip_ACK_t *ackhdl;
  int from_whom;
};
 
struct RA_params **MyRAQueue;

struct msg_params {
  char * address;
  int    length;
};
 
void AlmostReady(void *);  /* Parallel function executed before calling user */

/* these numbers have to be the same as they are in KernelCom.c for 
   shared memory machines.  */
#define MAXBUFF 4096

static int tulip_SharedMemoryInitialized = 0;
int tulip_MsgBlockSize = -1;
int left_child, right_child, parent_id, num_children;
void Poll();

/*******************************************************************/
/* Shared data structures */
/*******************************************************************/
long *BarrierVector;
barrier_t *SGI_Gen_Barr;       /* general-purpose barrier */
barrier_t *SGI_Done_Barr;      /* Done with everything barrier */
char *BroadcastBuff;           /* a place to store broacast values */
char *ReduceBuff[MAX_PROC];    /* a place to store reduction values */
struct RA_params ***GlobalRATable; /* after init, read-only */

int *SharedHandle;      /* Name of "Arena" created by usinit() */
int MyTotalNodes = -1;     /* number of processors */
int MyNodeNumber = -1;     /* who I am */

void * tulip_SharedMalloc(int size) {
  if (tulip_SharedMemoryInitialized)
    return (void *) usmalloc(size, SharedHandle);
  else
    return (void *) malloc(size);
}

void tulip_SharedFree(void *p) {
  if (tulip_SharedMemoryInitialized)
    usfree(p,SharedHandle);
  else
    free(p);
}

/*
 @ SMRTSSetup() : set up to run on the SGI
 */

int SMRTSSetup()
{
  char fname[1024];
  int i;

  /* allocate processors */
  if (pcxx_param.numproc >= sysmp(MP_NAPROCS)){
    MyTotalNodes = sysmp(MP_NAPROCS);
  } else {
    MyTotalNodes = pcxx_param.numproc;
  }

  MyNodeNumber = MyTotalNodes - 1;
  pcxx_EvInit (pcxx_param.tracefile);
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_ENTER, pcxx_param.numproc);

  usconfig(CONF_INITUSERS,20);              /* Up to 20 processors */
  usconfig(CONF_INITSIZE,1024*1024*128);    /* 128 Meg of shared data */
  usconfig(CONF_ARENATYPE,US_SHAREDONLY);   /* unlink memmap after open */

  /* Open unique memmap filename */
  sprintf(fname,"/usr/tmp/pC++%d.memmap",getpid());
  SharedHandle = usinit(fname);

  /* initialize SGI barriers */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_SETUP_INIT, 0);
  SGI_Gen_Barr = new_barrier(SharedHandle);
  init_barrier(SGI_Gen_Barr);
  SGI_Done_Barr = new_barrier(SharedHandle);
  init_barrier(SGI_Done_Barr);

  tulip_SharedMemoryInitialized = 1;

  /* Allocate & init our homebrew polling Barrier area */
  BarrierVector = 
    usmalloc(sizeof(unsigned long)*CACHE_LINE*MAX_PROC,SharedHandle);
  for (i=0; i<CACHE_LINE*MAX_PROC; i++) BarrierVector[i]=0;

  /* Allocate Global list of RA queues */
  GlobalRATable =
    usmalloc(sizeof(unsigned long)*MAX_PROC,SharedHandle);
  for (i=0; i<MAX_PROC; i++) GlobalRATable[i]=0;

  /* allocate Shared Broadcast Buffer */
  BroadcastBuff = (char *) usmalloc(MAXBUFF,SharedHandle);

  /* allocate shared locks and reduction buffer */
  if(MyTotalNodes > MAX_PROC) {
    fprintf(stderr,"Error in sgilib.c: MAX_PROC < number of processors.\n");
    exit(0);
  }
  for(i = 0; i < MyTotalNodes; i++) {
    ReduceBuff[i] = (char*) usmalloc(MAXBUFF,SharedHandle);
  }

  PCXX_EVENT (PCXX_EC_BASIC, PCXX_SETUP_EXIT, MyTotalNodes);
  return(0);
}

void tulip_InternalBarrier(barrier_t *b)
{
/*printf("%d: starting Barrier (%d)\n",MyNodeNumber,MyTotalNodes);*/
  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);
  barrier(b,MyTotalNodes);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
/*printf("%d: Ending Barrier (%d)\n",MyNodeNumber,MyTotalNodes);*/
}

/*
 @ SMRTSFork : start parallel processing
 */

int SMRTSFork(int argc, char **argv)
{
  int k;
  struct ArgcArgv *acav;

  /* fork */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_ENTER, 0);

  for (k=0; k<MyTotalNodes-1; k++) {
    acav = (struct ArgcArgv *) tulip_SharedMalloc(sizeof(struct ArgcArgv));
    acav->i=k;
    acav->argc=argc;
    acav->argv=argv;
    sproc(AlmostReady, 0, (void *) acav);
  }

  /* Go do AlmostReady() myself! */
  acav = (struct ArgcArgv *) tulip_SharedMalloc(sizeof(struct ArgcArgv));
  acav->i=MyTotalNodes-1;
  acav->argc=argc;
  acav->argv=argv;
  AlmostReady( (void *) acav ); 
  tulip_InternalBarrier(SGI_Done_Barr);

  /* finished */
  PCXX_EVENT (PCXX_EC_BASIC, PCXX_FORK_EXIT, 0);
  pcxx_EvClose ();
  return(0);
}

/*
 @ SMRTSNewProc : more book-keeping for newly-created process
 */

int SMRTSNewProc(num)
  int num;
{
  /* record my ID */
  MyNodeNumber = num;
  tulip_InitClocks ();
  pcxx_EvInit (pcxx_param.tracefile);
  return(MyNodeNumber);
}

/***********************************************************************/
/* Pete's super-fast barrier */
/***********************************************************************/
void tulip_Barrier()
{
#define LNx 0
#define CNx 256
#define RNx 512

  PROFILE_ENTER(PROFILE_Barrier);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_ENTER, 0);

  /* Is a previous barrier still executing? */
  /* Hey! A show stopper! */
/*  while ( (BarrierVector[MyNodeNumber*CACHE_LINE+LNx] != 0) ||
          (BarrierVector[MyNodeNumber*CACHE_LINE+CNx] != 0) ||
          (BarrierVector[MyNodeNumber*CACHE_LINE+RNx] != 0) )
    printf("%d: Hey, slow down! %d %d %d\n",MyNodeNumber,
          (BarrierVector[MyNodeNumber*CACHE_LINE+LNx]),
          (BarrierVector[MyNodeNumber*CACHE_LINE+CNx]),
          (BarrierVector[MyNodeNumber*CACHE_LINE+RNx]));*/

  /* if there is only one processor, simply return.  S.Y. 8/9/95 */
  if(num_children == 0 && MyNodeNumber == 0) {
    PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
    PROFILE_EXIT(PROFILE_Barrier);
    return;
  }

  /* Start fan-in, leaves bubble up... */
  if (num_children == 0) {
    if (MyNodeNumber & 1) /* Odd numbered leaf */
      BarrierVector[parent_id*CACHE_LINE+LNx]=1;
    else 
      BarrierVector[parent_id*CACHE_LINE+RNx]=1;
  }

  /* I am not a leaf. Leaf me alone!!! */
  else
    {
      /* Wait for fan-in from children */
      if (num_children == 1)
	while  (BarrierVector[MyNodeNumber*CACHE_LINE+LNx] == 0)
	  tulip_Poll();
      else
	while ((BarrierVector[MyNodeNumber*CACHE_LINE+LNx] == 0) || 
	       (BarrierVector[MyNodeNumber*CACHE_LINE+RNx] == 0))
	  tulip_Poll();

      /* OK, I have received all my children */
      /* propagate fan-in to my parent */
      if (MyNodeNumber != 0) {
	if (MyNodeNumber & 1)  /* Odd numbered leaf */
	  BarrierVector[parent_id*CACHE_LINE+LNx]=1;
	else 
	  BarrierVector[parent_id*CACHE_LINE+RNx]=1;
      }
      else {
	/* Hey!! the root has received the fan-in message, start fan-out */
	/* reset my slots*/
	BarrierVector[MyNodeNumber*CACHE_LINE+LNx] =  0;
	BarrierVector[MyNodeNumber*CACHE_LINE+CNx] =  0;
	BarrierVector[MyNodeNumber*CACHE_LINE+RNx] =  0;

	/* Propagate to children */
	if (num_children >= 1) BarrierVector[ left_child*CACHE_LINE+CNx] = -1;
	if (num_children == 2) BarrierVector[right_child*CACHE_LINE+CNx] = -1;
        PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
        PROFILE_EXIT(PROFILE_Barrier);
	return;
      }
    }

  /* Ok, I have received any required fan-in msgs, and have propagated them */
  /* now wait of fan-out message */

  while (BarrierVector[MyNodeNumber*CACHE_LINE+CNx] != -1)
    tulip_Poll();
    
  /* Reset my variables */
  BarrierVector[MyNodeNumber*CACHE_LINE+LNx] =  0;
  BarrierVector[MyNodeNumber*CACHE_LINE+CNx] =  0;
  BarrierVector[MyNodeNumber*CACHE_LINE+RNx] =  0;

  /* Ok, I got the fan-out message, propagate it */
  if (num_children >= 1) BarrierVector[ left_child*CACHE_LINE+CNx] = -1;
  if (num_children == 2) BarrierVector[right_child*CACHE_LINE+CNx] = -1;
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_BARRIER_EXIT, 0);
  PROFILE_EXIT(PROFILE_Barrier);
}


/*
 @ tulip_BrodcastBytes : sends messages from 0 to all other nodes
 */

void tulip_BroadcastBytes(flag, length, buffer)
int flag;
int length;
void * buffer;
{
  int i;
  char *c, *p;

  /* This barrier is needed to prevent BroadcastBuff from being overwritten
     when two consecutive tulip_BroadcastBytes are called */
  tulip_Barrier();

  if ( length > MAXBUFF ) 
    tulip_Message("Broadcast buffer too small!  Increase buffer.",TULIP_FATAL);

  c = BroadcastBuff; p = (char *) buffer;

  if (flag == 0) for (i=0; i<length; i++) c[i] = p[i];
  tulip_Barrier();
  if (flag != 0) for (i=0; i<length; i++) p[i] = c[i];
}


/*
 @ SMRTSDealloc : deallocate shared memory pointers
 */

void SMRTSDealloc(ptr)
void *ptr;
{
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_ENTER, (long) ptr);
  usfree(ptr,SharedHandle);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_DEALLOC_EXIT, 0);
}

void SMRTSWaitToEnd() {
  if ((int) MyNodeNumber != (MyTotalNodes-1)) {
    tulip_InternalBarrier(SGI_Done_Barr);
  }
}

void AlmostReady(void *x)
{
  int i, num, argc;
  char **argv;
  struct ArgcArgv *acav;

  acav = (struct ArgcArgv *) x;
  num  = acav->i;
  argc = acav->argc;
  argv = acav->argv;

  SMRTSNewProc(num);
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_ENTER, 0);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_ENTER, 0);

  /* Allocate MY RemoteAction area */
  MyRAQueue = 
    usmalloc(sizeof(struct RA_params **)*CACHE_LINE*MAX_PROC,SharedHandle);
  for (i=0; i<CACHE_LINE*MAX_PROC; i++) MyRAQueue[i]=0;

  /* Register My RA area with Global Table */
  GlobalRATable[MyNodeNumber] = MyRAQueue;

  /* INITIALIZE BARRIER VARIABLES */
  left_child = (MyTotalNodes > 2*MyNodeNumber+1)? 
                                   2*MyNodeNumber+1: 0;
  right_child = (MyTotalNodes > 2*MyNodeNumber+2)? 
                                   2*MyNodeNumber+2: 0;
  if (left_child && right_child) num_children = 2;
  else if(left_child || right_child) num_children = 1;
  else num_children = 0;
  parent_id = (MyNodeNumber-1)/2;

  tulip_InternalBarrier(SGI_Gen_Barr);  /* sync up before Polling, etc */
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_NEWPROC_EXIT, MyNodeNumber);
  tulip_ParallelMain(argc,argv);  /* Transfer control to compiler/user code */
  PCXX_EVENT(PCXX_EC_BASIC, PCXX_MAIN_EXIT, 0);
  SMRTSWaitToEnd();
  pcxx_EvClose ();
}

/*********************************************************************/
/* Check if a "handle" has completed (non-blocking) */
void  tulip_Abort()
{
  printf("Exiting!!!!! ERROR\n");
  exit(1);
}

/*********************************************************************/
int tulip_Get(char *dst, tulip_GlobalPointer_t src, int len, 
	      tulip_ACK_t *ackhandle)
{
  /* Why call tulip_Get on a shared memory machine? */
  memcpy(dst, tulip_ToLocalPointer(&src), len);

  tulip_SetACK(ackhandle);

  return 0;
}  

/*********************************************************************/
int tulip_Put(tulip_GlobalPointer_t dst, char *src, int len, 
	      tulip_ACK_t *ackhandle)
{
  /* Why call tulip_Put on a shared memory machine? */
  memcpy(tulip_ToLocalPointer(&dst), src, len);

  tulip_SetACK(ackhandle);

  return 0;
}  

/*********************************************************************/
/*********************************************************************/
/* Check for RemoteAction requests */
void tulip_Poll()
{

int i;
struct RA_params *tmpparams;

  PROFILE_ENTER(PROFILE_Poll);
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_POLL_ENTER, 0);

/* Scan through RemoteAction area looking for work */
  for (i=0; i<=MyTotalNodes; i++) {
    if ( MyRAQueue[i*CACHE_LINE] !=  0 ) {
      /* Hey, found some work! */
      tmpparams = MyRAQueue[i*CACHE_LINE];

      tulip_RemoteActionHandler(tmpparams->from_whom, tmpparams->buffer,
			       tmpparams->length);

      /* OK, clean up! */

      /* free buffer */
      tulip_SharedFree(tmpparams->buffer);

      /* Send acknowledgement (through ACK handle) */
      tulip_SetACK(tmpparams->ackhdl);

      tulip_SharedFree(tmpparams);

      /* Clean out slot */
      MyRAQueue[i*CACHE_LINE] = 0;
    }
  }
  PCXX_EVENT (PCXX_EC_RUNTIME, PCXX_POLL_EXIT, 0);
  PROFILE_EXIT(PROFILE_Poll);
}

/*********************************************************************/
int tulip_RemoteAction(int remote_node, char *buffer, int len, 
		       tulip_ACK_t *ackhandle)
{

  struct RA_params **rmtslot;
  struct RA_params *tmpparams;
  char *tmpbuf;

  rmtslot = &( (GlobalRATable[remote_node])[MyNodeNumber*CACHE_LINE]);
 
  /* check to see if a slot is available */
  while (*rmtslot != 0) {
    /* no slot is available. poll and wait */
    tulip_Poll();
  }

  /* OK, the slot is free! */

  tulip_ClearACK(ackhandle);
 
  /* allocate buffer for remote action request */
  tmpbuf = (char *) tulip_SharedMalloc(len);
  memcpy(tmpbuf, buffer, len);

  /* allocate parameter structure */
  tmpparams = (struct RA_params *) tulip_SharedMalloc(sizeof(struct RA_params));

  /* install parameters */
  tmpparams->buffer    = tmpbuf;
  tmpparams->from_whom = MyNodeNumber;
  tmpparams->length    = len;
  tmpparams->ackhdl    = ackhandle;
 
  /* OK, now the pointer tmpparams is ready to pass to the remote node */

  /* fill remote slot */
  *rmtslot = tmpparams;
    
  return 0;
}
