/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "tulip.h"


static int WarningCount=0;

int aa_poll_count=0; /* For Ariadne polling - instead of tulip/mpi/src/KernelCom.c - because kernel.h accesses this for eg. trc-sgimp - so its a global */


/*********************************************************************/
void tulip_SetACK(tulip_ACK_t *ackhandle)
{
  if (ackhandle) *ackhandle = 1;
}

/*********************************************************************/
void tulip_ClearACK(tulip_ACK_t *ackhandle)
{
  if (ackhandle) *ackhandle = 0;
}

/*********************************************************************/
/* Return non-zero if the ACK is true */

int tulip_TestACK(tulip_ACK_t *ackhandle)
{
  /* On some machines, I may need a cache flush here... */
  if (ackhandle==0) return 1;
  return *ackhandle;
}

/*********************************************************************/
/* Wait on a message handle (from tulip_Get or tulip_Put).  This
   procedure is blocking */

void tulip_WaitACK(tulip_ACK_t *ackhandle)
{
    while (!(tulip_TestACK(ackhandle))) tulip_Poll();
}

/* --------------------------------------------------------------------
 * Create/Allocate <count> ACK handles.  Returns NULL on error
 * ----------------------------------------------------------------- */
tulip_ACK_t *tulip_ACKCreate(int count)
{ return tulip_SharedMalloc(sizeof(tulip_ACK_t)*count); }

/* --------------------------------------------------------------------
 * Destroy/Deallocate an ACK handle.
 * ----------------------------------------------------------------- */
void tulip_ACKDestroy(tulip_ACK_t *handle)
{ tulip_SharedFree(handle); }

/*********************************************************************/
/* Message is a general error reporting function that prints a message
   pointed to by "st".  The flag is either TULIP_WARNING or
   TULIP_FATAL.  On TULIP_FATAL it is necesary to abort the
   program. */
/*********************************************************************/

void tulip_Message(char * st, int flag)
{
  if (flag == TULIP_WARNING)
    {
      WarningCount++;
      fprintf(stderr,"Warning %s\n",st);
      return;
    }
  if (flag == TULIP_FATAL)
    {
      fprintf(stderr,"Fatal error %s\n",st);
      tulip_Abort();
    }
}

/*********************************************************************/
tulip_GlobalPointer_t *tulip_ToGlobalPointer(void * local_addr)
{
  void * p;
  p = tulip_SharedMalloc(sizeof(tulip_GlobalPointer_t));
  if (p == NULL) {
    tulip_Message("Error, SharedMalloc failed in ToGlobal",TULIP_FATAL);
  }
  return p;
}

/*********************************************************************/
void *tulip_ToLocalPointer(tulip_GlobalPointer_t *mygp)
{
  return mygp->LocalAddress;
}

/*********************************************************************/
int tulip_ToContext(tulip_GlobalPointer_t *mygp)
{
  return mygp->NodeID;
}

/*********************************************************************/
void tulip_SetLocalPointer(tulip_GlobalPointer_t *mygp, void *mylp)
{
  mygp->LocalAddress = mylp;
}

/*********************************************************************/

void tulip_SetContext(tulip_GlobalPointer_t *mygp, int mycntx)
{
  mygp->NodeID = mycntx;
}
