/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/*********************************************************************/
/*                      Pete Beckman  10/28/94                       */
/*      A communications and synchronization benchmark suite         */
/*********************************************************************/

/* #include "mpi.h" */
#include "tulip.h"
#include <stdlib.h>
#include <stdio.h>

#define BOOTSTRAP 23
#define PINGRPC   24
#define PONGRPC   25
#define ARRAY_LEN   65536

#define KILOBYTE     1024
#define MEGABYTE  1048576
#define MAXBUFFER   65536  /* large packet (nice power of 2) */
#define MINBUFFER       0   /* small packet */
#define LOOPS        5

/* Some simple macros to make the timing code easier to type */
#define ALL_BSTART(_a,_b)    tulip_Barrier(); \
                             _a=tulip_WallClock(); \
                             _b=tulip_UserClock();

#define ALL_BEND(_a,_b)      tulip_Barrier(); \
                             _a=tulip_WallClock()-_a; \
                             _b=tulip_UserClock()-_b;

#define ALL_START(_a,_b)     _a=tulip_WallClock(); \
                             _b=tulip_UserClock();

#define ALL_END(_a,_b)       _a=tulip_WallClock()-_a; \
                             _b=tulip_UserClock()-_b;

/***** Global Variables! *****/
tulip_GlobalPointer_t GlobalArrayTable[1024]; /* A map to each proc's array portion */
tulip_GlobalPointer_t GPSharedArray;
int PingPongFlag=0;
int *SharedArray;                 /* My array portion */
int *LocalArray;                 /* My array portion */
int *h;
int dummy[MAXBUFFER];
int BufferOffset;
int hlist[5000],force = 0;
double TW[20],TU[20];
double sum,x, y, s, u;     /* scalars */
double *A,*B,*C,*D;
int mplyer;
int BIG[MAXBUFFER];
char dummyspace[8];
tulip_ACK_t ackhandle;
int testlist[20];

/****************************************************************/
void GetRmtBlocking(tulip_GlobalPointer_t who, int loops, int length) {
  int i;
  for (i=0; i<loops; i++) {
    sum+=1.0;
    tulip_Get((char *) LocalArray, who, length, &ackhandle);
    tulip_WaitACK(&ackhandle);
  }
}

/****************************************************************/
void GetRmtEmpty(tulip_GlobalPointer_t who, int loops, int length) {
  int i;
  for (i=0; i<loops; i++) {
    sum+=1.0;
/*    tulip_Get((char *) LocalArray, who, length, &ackhandle); */
/*    tulip_WaitACK(&ackhandle); */
  }
}

/****************************************************************/
void GetRmtRndBlk(int loops, int length) {
  int i,x;
  for (i=0; i<loops; i++) {
    x = (rand()>>7) % tulip_NumContexts();
    sum+=(double) x;
/*printf("%d: fetching from %d\n",tulip_MyContext(),x);*/
    tulip_Get((char *) LocalArray, GlobalArrayTable[x], length, &ackhandle);
/*printf("%d: DONE from %d\n",tulip_MyContext(),x);*/
    tulip_WaitACK(&ackhandle);
  }
}

/****************************************************************/
void GetRmtRndEmpty(int loops, int length) {
  int i,x;
  for (i=0; i<loops; i++) {
    x = (rand()>>7) % tulip_NumContexts();
    sum+=(double) x;
/*  tulip_Get((char *) LocalArray, GlobalArrayTable[x], length, &ackhandle);*/
/*    tulip_WaitACK(&ackhandle);*/
  }
}

/****************************************************************/
void LateBarrier(int loops, int t) {
  int i,j;
  for (i=0; i<loops; i++) {
    sum += 0.00001;
    if (t == tulip_MyContext()) 
      for (j=0; j<1000; j++) sum += 0.00002;
tulip_Barrier();
  }
}

/****************************************************************/
void LateBarrierEmpty(int loops, int t) {
  int i,j;
  for (i=0; i<loops; i++) {
    sum += 0.00001;
    if (t == tulip_MyContext()) 
      for (j=0; j<1000; j++) sum += 0.00002;
/*    tulip_Barrier();*/
  }
}

/****************************************************************/
void MPI_msg(int loops, int t) {
  int i,j;
/*  MPI_Status status;*/
  for (i=0; i<loops; i++) {
    sum += 0.00001;
    if (tulip_MyContext() == 0) {
      sum += 0.00001;
/*      MPI_Ssend(BIG, t , MPI_BYTE, 1, 5, MPI_COMM_WORLD);*/
    }
    if (tulip_MyContext() == 1) {
      sum += 0.00002;
/*      MPI_Recv(dummy, t, MPI_BYTE, 0, 5, MPI_COMM_WORLD, &status);*/
    }
  }
}

/****************************************************************/
void MPI_msgEmpty(int loops, int t) {
  int i,j;
/*  MPI_Status status;*/
  for (i=0; i<loops; i++) {
    sum += 0.00001;
    if (tulip_MyContext() == 0) {
      sum += 0.00001;
/*      MPI_Ssend(BIG, t , MPI_BYTE, 1, 5, MPI_COMM_WORLD);*/
    }
    if (tulip_MyContext() == 1) {
      sum += 0.00002;
/*      MPI_Recv(dummy, t, MPI_BYTE, 0, 5, MPI_COMM_WORLD);*/
    }
  }
}

/****************************************************************/
void FS_Test(int loops, int t) {
  int i,j;

  for (i=0; i<loops; i++) {
    sum += 0.00001;
    tulip_Get((char *) LocalArray, GlobalArrayTable[1], t , &ackhandle);
    tulip_WaitACK(&ackhandle);
  }
}


/****************************************************************/
void FS_TestEmpty(int loops, int t) {
  int i,j;

  for (i=0; i<loops; i++) {
    sum += 0.00001;
/*    tulip_Get((char *) LocalArray, GlobalArrayTable[1], t , &ackhandle);
    tulip_WaitACK(&ackhandle);*/
  }
}


/****************************************************************/
void BarrierTest(int loops) {
  int i;
  for (i=0; i<loops; i++) {
    sum += 0.00001*i;
tulip_Barrier();
  }
}

/****************************************************************/
void BarrierTestEmpty(int loops) {
  int i;
  for (i=0; i<loops; i++) {
    sum += 0.00001*i;
/*    tulip_Barrier();*/
  }
}

/****************************************************************/
void IrregBarr(int loops, int inner) {
  int i,j;
  for (i=0; i<loops; i++) {
    for (j=0; j<inner; j++) sum += 0.00001;
/* MPI_Barrier(MPI_COMM_WORLD);*/
  }
}

/****************************************************************/
void IrregBarrEmpty(int loops, int inner) {
  int i,j;
  for (i=0; i<loops; i++) {
    for (j=0; j<inner; j++) sum += 0.00001;
/*    tulip_Barrier();*/
  }
}

#if 0
/****************************************************************/
void NativeBcastTest(int size) {
  int offset;
  char * buffer;

  tulip_NativeBroadcastBytes(tulip_MyContext(), size, buffer);
}

/****************************************************************/
void AlienBcastTest(int size) {
  int offset;
  char * buffer;

  tulip_NativeBroadcastBytes(tulip_MyContext(), size, buffer);
}

/****************************************************************/
void NativeReduceTest(int size) {
  tulip_NativeReduceDoubleAdd(0,size);
}

/****************************************************************/
void AlienReduceTest(int size) {
  tulip_AlienReduceDoubleAdd(0,size);
}
#endif




#undef LOOPS
#define LOOPS 10000
/****************************************************************/
void tests()
{
  int i,j,k,l,z,loops,d2,neighbor,d1;
  double ew, eu, t1w, t1u, t2w, t2u;
  double xfer, bandwidthw, bandwidthu, latencyw, latencyu;

#define BWINTERVAL 2048
#define BARRIERLOOPS 10000

  for (i=0; i<20; i++) testlist[i]=0;
  testlist[6]=1;

if (testlist[0]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 0: Maximum single node speeds for tulip_Get() \n");
printf("Node 0 fetches from Node 1, all other nodes wait patiently.\n");
printf("****************************************************************************\n");
#include "head1.h"
}

  if (tulip_NumContexts() == 1) neighbor=0;
  else neighbor = 1;

for (d1=MINBUFFER; d1<=MAXBUFFER-BWINTERVAL; d1+=BWINTERVAL) {

  d2=d1+BWINTERVAL;

  ALL_BSTART(ew,eu);
  if (tulip_MyContext() == 0)
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
  if (tulip_MyContext() == 0)
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

  /* Time Empty Get_Element Code */
  ALL_BSTART(ew,eu);
  if (tulip_MyContext() == 0)
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t2w,t2u);
  if (tulip_MyContext() == 0)
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(t2w,t2u); t2w-=ew; t2u-=eu;

#include "calc1.h"

}
}

if (testlist[1]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 1: Maximum single node speeds for tulip_Get() \n");
printf("Node 0 fetches from Node p, all other nodes wait patiently.\n");
printf("****************************************************************************\n");
#include "head1.h"
}

  neighbor = tulip_NumContexts() - 1;

for (d1=MINBUFFER; d1<=MAXBUFFER-BWINTERVAL; d1+=BWINTERVAL) {

  d2=d1+BWINTERVAL;

  ALL_BSTART(ew,eu);
  if (tulip_MyContext() == 0)
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
  if (tulip_MyContext() == 0)
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

  /* Time Empty Get_Element Code */
  ALL_BSTART(ew,eu);
  if (tulip_MyContext() == 0)
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t2w,t2u);
  if (tulip_MyContext() == 0)
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(t2w,t2u); t2w-=ew; t2u-=eu;

#include "calc1.h"
}
}

if (testlist[2]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 2: Effective HALF-HALF tulip_Get() test.\n");
printf("Even numbered nodes fetch elements from the odd numbered nodes.\n");
printf("Therefore, each node with either fetch or service requests,\n");
printf("but not both\n");
printf("****************************************************************************\n");
#include "head1.h"
}

  neighbor = (tulip_MyContext()+1) % tulip_NumContexts();
  /* Time Empty Get_Element Code */

for (d1=MINBUFFER; d1<=MAXBUFFER-BWINTERVAL; d1+=BWINTERVAL) {

  d2=d1+BWINTERVAL;

  ALL_BSTART(ew,eu);
  if (tulip_MyContext()%2)
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
  if (tulip_MyContext()%2)
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

  /* Time Empty Get_Element Code */
  ALL_BSTART(ew,eu);
  if (tulip_MyContext()%2)
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t2w,t2u);
  if (tulip_MyContext()%2)
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(t2w,t2u); t2w-=ew; t2u-=eu;

#include "calc1.h"
}
}

if (testlist[3]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 3: Effective ALL-ALL tulip_Get() test.\n");
printf("EVERY node will fetch a buffer from their neighbor (n=(i+1)%%p)\n");
printf("For distributed memory machines, this means the nodes must\n");
printf("request data and reply to requests from their neighbor.\n");
printf("****************************************************************************\n");
#include "head1.h"
}

  neighbor = (tulip_MyContext()+1) % tulip_NumContexts();
  /* Time Empty Get_Element Code */

for (d1=MINBUFFER; d1<=MAXBUFFER-BWINTERVAL; d1+=BWINTERVAL) {

  d2=d1+BWINTERVAL;

  ALL_BSTART(ew,eu);
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

  /* Time Empty Get_Element Code */
  ALL_BSTART(ew,eu);
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t2w,t2u);
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(t2w,t2u); t2w-=ew; t2u-=eu;

#include "calc1.h"
}
}

if (testlist[4]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 4: Effective ALL-ALL, random dest. tulip_Get() test.\n");
printf("EVERY node will fetch a buffer from another (selected randomly)\n");
printf("For distrubuted memory machines, this means the nodes must\n");
printf("request data and reply to requests from their neighbor.\n");
printf("****************************************************************************\n");
#include "head1.h"
}
  neighbor = (tulip_MyContext()+1) % tulip_NumContexts();
  /* Time Empty Get_Element Code */

for (d1=MINBUFFER; d1<=MAXBUFFER-BWINTERVAL; d1+=BWINTERVAL) {

  d2=d1+BWINTERVAL;

  ALL_BSTART(ew,eu);
    GetRmtRndEmpty(LOOPS, d1);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
    GetRmtRndBlk(LOOPS, d1);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

  /* Time Empty Get_Element Code */
  ALL_BSTART(ew,eu);
    GetRmtRndEmpty(LOOPS, d2);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t2w,t2u);
    GetRmtRndBlk(LOOPS, d2);
  ALL_BEND(t2w,t2u); t2w-=ew; t2u-=eu;

#include "calc1.h"
}
}

if (testlist[5]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 5: ALL-ONE (all fetch from one )\n");
printf("Even numbered nodes fetch elements from the odd numbered nodes.\n");
printf("Therefore, each node with either fetch or service requests,\n");
printf("but not both\n");
printf("****************************************************************************\n");
#include "head1.h"
}

  neighbor = 0;
  /* Time Empty Get_Element Code */

for (d1=MINBUFFER; d1<=MAXBUFFER-BWINTERVAL; d1+=BWINTERVAL) {

  d2=d1+BWINTERVAL;

  ALL_BSTART(ew,eu);
  if (tulip_MyContext() != 0) 
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
  if (tulip_MyContext() != 0) 
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d1);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

  /* Time Empty Get_Element Code */
  ALL_BSTART(ew,eu);
  if (tulip_MyContext() != 0) 
    GetRmtEmpty(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t2w,t2u);
  if (tulip_MyContext() != 0) 
    GetRmtBlocking(GlobalArrayTable[neighbor], LOOPS, d2);
  ALL_BEND(t2w,t2u); t2w-=ew; t2u-=eu;

#include "calc1.h"
}
}

if (testlist[6]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 6: Barrier (Synchronous Arrival)\n");
printf("****************************************************************************\n");
printf("============================================================================\n");
printf("     WALL CLOCK USER CLOCK\n");
printf(" PEs    Barrier    Barrier\n");
printf("           uSec       uSec\n");
printf("============================================================================\n");
}

  j = BARRIERLOOPS;

  ALL_BSTART(ew,eu);
    BarrierTestEmpty(j);
  ALL_BEND(ew,eu);

  /* Time REAL Test */
  ALL_BSTART(t1w,t1u);
    BarrierTest(j);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

if (tulip_MyContext() == 0) {
printf("%4d%11.2f%11.2f\n",
   tulip_NumContexts(),t1w/j*1000000,
   t1u/j*1000000);
}

tulip_Barrier();
}

if (testlist[7]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 7: Barrier (Staggered Arrival)\n");
printf("****************************************************************************\n");
printf("============================================================================\n");
printf("     WALL CLOCK USER CLOCK\n");
printf(" PEs    Barrier    Barrier\n");
printf("           uSec       uSec\n");
printf("============================================================================\n");
}

for (i=0; i<tulip_NumContexts(); i++) {

  j = ((tulip_MyContext()+i) % tulip_NumContexts()) * 100;

  ALL_BSTART(ew,eu);
    IrregBarrEmpty(BARRIERLOOPS,j);
  ALL_BEND(ew,eu);

  /* Time REAL Get_Element Code */
  ALL_BSTART(t1w,t1u);
    IrregBarr(BARRIERLOOPS,j);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

if (tulip_MyContext() == 0) {
printf("%4d%11.2f%11.2f\n",
   tulip_NumContexts(),t1w/(BARRIERLOOPS)*1000000,
   t1u/(BARRIERLOOPS)*1000000);
}
}

tulip_Barrier();
}

if (testlist[8]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 8: Barrier (Late Arrival)\n");
printf("****************************************************************************\n");
printf("============================================================================\n");
printf("     WALL CLOCK USER CLOCK\n");
printf(" PEs    Barrier    Barrier\n");
printf("           uSec       uSec\n");
printf("============================================================================\n");
}

for (i=-1; i<tulip_NumContexts(); i++) {

  ALL_BSTART(ew,eu);
    LateBarrierEmpty(BARRIERLOOPS, i);
  ALL_BEND(ew,eu);

  /* Time REAL Test */
  ALL_BSTART(t1w,t1u);
    LateBarrier(BARRIERLOOPS, i);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

if (tulip_MyContext() == 0) {
printf("%4d%11.2f%11.2f\n",
   tulip_NumContexts(),t1w/(BARRIERLOOPS)*1000000,
   t1u/(BARRIERLOOPS)*1000000);
}
}

tulip_Barrier();
}

if (testlist[9]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 9: MPI blocking send\n");
printf("****************************************************************************\n");
printf("============================================================================\n");
printf("     WALL CLOCK USER CLOCK\n");
printf(" Msg    Barrier    Barrier\n");
printf("           uSec       uSec\n");
printf("============================================================================\n");
}

for (i=32; i<=65536; i=i*2) {

  ALL_BSTART(ew,eu);
    MPI_msgEmpty(BARRIERLOOPS, i);
  ALL_BEND(ew,eu);

  /* Time REAL Test */
  ALL_BSTART(t1w,t1u);
    MPI_msg(BARRIERLOOPS, i);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

if (tulip_MyContext() == 0) {
printf("%6d%11.2f%11.2f\n",
   i,t1w/(BARRIERLOOPS)*1000000,
   t1u/(BARRIERLOOPS)*1000000);
}
}


tulip_Barrier();
}

if (testlist[10]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 10: One-sided Fetch Test\n");
printf("****************************************************************************\n");
printf("============================================================================\n");
printf("     WALL CLOCK USER CLOCK\n");
printf(" Msg    Barrier    Barrier\n");
printf("           uSec       uSec\n");
printf("============================================================================\n");
}

for (i=32; i<=65536; i=i*2) {

  ALL_BSTART(ew,eu);
  if (tulip_MyContext() == 0)
    FS_TestEmpty(BARRIERLOOPS, i);
  ALL_BEND(ew,eu);

  /* Time REAL Test */
  ALL_BSTART(t1w,t1u);
  if (tulip_MyContext() == 0)
    FS_Test(BARRIERLOOPS, i);
  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

if (tulip_MyContext() == 0) {
printf("%6d%11.2f%11.2f\n",
   i,t1w/(BARRIERLOOPS)*1000000,
   t1u/(BARRIERLOOPS)*1000000);
}
}


tulip_Barrier();
}

if (testlist[11]) {
if (tulip_MyContext() == 0){
printf("\n\n");
printf("****************************************************************************\n");
printf("#TEST 11: Ping Pong RPC Test\n");
printf("****************************************************************************\n");
printf("============================================================================\n");
printf("     WALL CLOCK USER CLOCK\n");
printf(" Msg    Barrier    Barrier\n");
printf("           uSec       uSec\n");
printf("============================================================================\n");
}

  ALL_BSTART(ew,eu);

if (tulip_MyContext() == 0) {
  for (i=0; i<BARRIERLOOPS; i++) {
    sum += 0.00001;
    dummy[0] = PINGRPC;
/*    tulip_RemoteAction(1,dummy,sizeof(long),0);*/
    /* Wait for PONG */
/*    while (PingPongFlag == 0) tulip_Poll();*/
    PingPongFlag = 0;
  }
}

  ALL_BEND(ew,eu);

  /* Time REAL Test */
  ALL_BSTART(t1w,t1u);

if (tulip_MyContext() == 0) {
  for (i=0; i<BARRIERLOOPS; i++) {
    sum += 0.00001;
    dummy[0] = PINGRPC;
    tulip_RemoteAction(1,(char *) dummy,sizeof(long),0);
    /* Wait for PONG */
    while (PingPongFlag == 0) tulip_Poll();
    PingPongFlag = 0;
  }
}

  ALL_BEND(t1w,t1u); t1w-=ew; t1u-=eu;

if (tulip_MyContext() == 0) {
printf("%6d%11.2f%11.2f\n",
   i,t1w/(BARRIERLOOPS)*1000000,
   t1u/(BARRIERLOOPS)*1000000);
}


tulip_Barrier();
}

#if 0

printf("****************************************************************************\n");
printf("#TEST x: This test determines the time for Broadcasts\n");
printf("In this test, a buffer at the NODE level is broadcast to each other\n");
printf("node.  After each node has executed pcxx_Broadcast, each node has an\n");
printf("identical copy of the buffer.  In these tests, node 0 initiates the\n");
printf("broadcast.  Both the Native and Alien functions are tested\n");
printf("****************************************************************************\n");
printf("          <-----WALL CLOCK-----><-----USER CLOCK----->\n");
printf("               Native      Alien     Native      Alien\n");
printf(" PEs Block  Broadcast  Broadcast  Broadcast  Broadcast\n");
printf("      size   MilliSec   MilliSec   MilliSec   MilliSec\n");
printf("============================================================================\n");

  for (j = MAXBUFFER; j>=4; j=j/4) {
    ALL_BSTART(t1w,t1u);
    for (i=0; i<=LOOPS*2; i++) X.NativeBcastTest(j); 
    ALL_END(t1w,t1u);

    ALL_BSTART(t2w,t2u);
    for (i=0; i<=LOOPS*2; i++) X.AlienBcastTest(j);
    ALL_END(t2w,t2u);
    printf("%4d%6d%11.2f%11.2f%11.2f%11.2f\n",
       tulip_NumContexts(),j,t1w/(LOOPS*2)*1000,t2w/(LOOPS*2)*1000,
                           t1u/(LOOPS*2)*1000,t2u/(LOOPS*2)*1000);
    }

printf("\n\n");

printf("****************************************************************************\n");
printf("#TEST 5: Measure the time for ReduceDoubleAdd\n");
printf("This test is very similar to the previous Broadcast test.  In this\n");
printf("test, each buffer (vector) is ADDED together, forming a global\n");
printf("(vector) sum.  The sum is then re-broadcast back to each node, so\n");
printf("after the reduction, each node has a copy of the global sum.  The\n");
printf("contents of the original contributing buffer are overwritten.  Both\n");
printf("Native and Alien functions are tested.\n");
printf("****************************************************************************\n");
printf("           <---------WALL CLOCK-------><---------USER CLOCK------->\n");
printf("              Native    Alien         Native    Alien\n");
printf(" PEs Vector   DblAdd   DblAdd Empty   DblAdd   DblAdd Empty\n");
printf("     length MilliSec MilliSec  Loop MilliSec MilliSec  Loop [check]\n");
printf("============================================================================\n");

double *dtemp;
dtemp = (double *) &(X.BIGNODE[0]);

/**********************************************************************/
x=1.1; y=1.2; u=1.3; s=1.4; 
/**********************************************************************/
  /* EMPTY LOOP TIMERS */
  ALL_START(t1w,t1u);
  for (i=0; i<=LOOPS*2; i++) {
    for (k=0; k<4096/sizeof(double); k++) dtemp[k] = 1.0;
    s = s + 4.0;
    x = x + 1.0;
    y = y + 2.0;
    u = u + 3.0;
  }
  ALL_END(t1w,t1u);
    
  ALL_START(ew,eu);
  for (i=0; i<=LOOPS*2; i++) {
    for (k=0; k<4096/sizeof(double); k++) dtemp[k] = 1.0;
    y = y + 2.0;
    u = u + 3.0;
  }
  ALL_END(ew,eu);

  ew = ew - (t1w - ew);   /* Time for empty loops */
  eu = eu - (t1u - eu);   /* Time for empty loops */
  /* Convert to milliseconds */
  ew = ew/(LOOPS*2)*1000;
  eu = eu/(LOOPS*2)*1000;

  if (s+x+y+u > 1000.0) force++;    /* Force -O to keep computation */
/**********************************************************************/
/**********************************************************************/

  for (j = 1; j<=4096/sizeof(double); j=j*4) {
    for (i=0; i<4096/sizeof(double); i++) dtemp[i] = 1.0; 

    ALL_BSTART(t1w,t1u);
    for (i=0; i<=LOOPS*2; i++) {
      for (k=0; k<4096/sizeof(double); k++) dtemp[k] = 1.0;
      X.NativeReduceTest(j);
    }
    ALL_END(t1w,t1u);
    y = dtemp[j-1];

    ALL_BSTART(t2w,t2u);
    for (i=0; i<=LOOPS*2; i++) {
      for (k=0; k<4096/sizeof(double); k++) dtemp[k] = 1.0;
      X.AlienReduceTest(j);
    }
    ALL_END(t2w,t2u);
    x = dtemp[j-1];

    printf("%4d%7d%9.2f%9.2f%6.1f%9.2f%9.2f%6.1f%6.1f%6.1f\n",
       tulip_NumContexts(),j,t1w/(LOOPS*2)*1000-ew,t2w/(LOOPS*2)*1000-ew,ew,
                           t1u/(LOOPS*2)*1000-eu,t2u/(LOOPS*2)*1000-eu,eu,
                           y,x);
    }
printf("\n\n");

/* Barrier times */
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
#define FLOPLOOP 100000    /* largest vector size */
#define MPLY 10            /* Millions of Flops to do per loop */
                           /* safe values: 1-100 */
mplyer = 10*MPLY;
A = (double *) malloc(sizeof(double)*(FLOPLOOP+100));
if (A == 0) { printf("Error, could not allocate A[]"); exit(1); }
B = (double *) malloc(sizeof(double)*(FLOPLOOP+100));
if (B == 0) { printf("Error, could not allocate B[]"); exit(1); }
C = (double *) malloc(sizeof(double)*(FLOPLOOP+100));
if (C == 0) { printf("Error, could not allocate C[]"); exit(1); }
D = (double *) malloc(sizeof(double)*(FLOPLOOP+100));
if (D == 0) { printf("Error, could not allocate D[]"); exit(1); }

printf("****************************************************************************\n");
printf("The next sections do some rudimentary MFLOP timing loops.  Some\n");
printf("important notes: 1) The timing loops are in BOTH WallClock and\n");
printf("UserClock seconds.  This may seem strange, but including both times\n");
printf("provides a more accurate measurement of the true speed.  If the OS is\n");
printf("poorly written, your programs will suffer, and the WallClock will run\n");
printf("slow.  This policy also lets you test the machine while it is loaded,\n");
printf("and examine the performance degradation curve.  If the WallClock\n");
printf("seconds and UserClock seconds are nearly identical, your machine is\n");
printf("doing very well!  2) For these tests, I've decided to subtract the\n");
printf("empty loop timings, which increases the MF/sec rating.  The empty loop\n");
printf("times are also printed so a reality check can be done.  All operations\n");
printf("use 'double' floats.  The inner loop (vector length) is varied to test\n");
printf("cache performance and pipe setup.\n");
printf("****************************************************************************\n");

/* Flop tests */
printf("****************************************************************************\n");
printf("#TEST 7A:\n");
printf("type A1) A[i]=B[i]+C[i]*D[i];\n");
printf("type A2) A[i]=  y +C[i]*D[i];\n");
printf("type A3) A[i]=B[i]+C[i]*  y;\n");
printf("****************************************************************************\n");
printf("            <---------WALL CLOCK--------><---------USER CLOCK-------->\n");
printf(" Outer Inner   Empty     A1     A2     A3   Empty     A1     A2     A3\n");
printf("  Loop  Loop    Time   MF/S   MF/S   MF/S    Time   MF/S   MF/S   MF/S\n");
printf("============================================================================\n");
#define TESTCASE1() A[i]=B[i]+C[i]*D[i]
#define TESTCASE2() A[i]=y+C[i]*D[i]
#define TESTCASE3() A[i]=B[i]+C[i]*y
#undef FLOPNUM
#define FLOPNUM 2
#include "flopt.h"

printf("****************************************************************************\n");
printf("#TEST 7B:\n");
printf("type B1) A[i]=A[i]+C[i]*D[i];\n");
printf("type B2) A[i]=  y +A[i]*D[i];\n");
printf("type B3) A[i]=A[i]+C[i]*  y;\n");
printf("****************************************************************************\n");
printf("            <---------WALL CLOCK--------><---------USER CLOCK-------->\n");
printf(" Outer Inner   Empty     B1     B2     B3   Empty     B1     B2     B3\n");
printf("  Loop  Loop    Time   MF/S   MF/S   MF/S    Time   MF/S   MF/S   MF/S\n");
printf("============================================================================\n");
#undef TESTCASE1
#undef TESTCASE2
#undef TESTCASE3
#define TESTCASE1() A[i]=A[i]+C[i]*D[i]
#define TESTCASE2() A[i]=y+A[i]*D[i]
#define TESTCASE3() A[i]=A[i]+C[i]*y
#undef FLOPNUM
#define FLOPNUM 2
#include "flopt.h"

printf("****************************************************************************\n");
printf("#TEST 7C:\n");
printf("type C1) x = x +C[i]*D[i];\n");
printf("type C2) x = y +  x *D[i];\n");
printf("type C3) x = x +C[i]*  y ;\n");
printf("****************************************************************************\n");
printf("            <---------WALL CLOCK--------><---------USER CLOCK-------->\n");
printf(" Outer Inner   Empty     C1     C2     C3   Empty     C1     C2     C3\n");
printf("  Loop  Loop    Time   MF/S   MF/S   MF/S    Time   MF/S   MF/S   MF/S\n");
printf("============================================================================\n");
#undef TESTCASE1
#undef TESTCASE2
#undef TESTCASE3
#define TESTCASE1() x=x+C[i]*D[i]
#define TESTCASE2() x=y+x*D[i]
#define TESTCASE3() x=x+C[i]*y
#undef FLOPNUM
#define FLOPNUM 2
#include "flopt.h"

printf("****************************************************************************\n");
printf("#TEST 7D:\n");
printf("type D1) x   =  x +C[i];\n");
printf("type D2) x   =  y *D[i];\n");
printf("type D3) x   =  x * y;\n");
printf("****************************************************************************\n");
printf("            <---------WALL CLOCK--------><---------USER CLOCK-------->\n");
printf(" Outer Inner   Empty     D1     D2     D3   Empty     D1     D2     D3\n");
printf("  Loop  Loop    Time   MF/S   MF/S   MF/S    Time   MF/S   MF/S   MF/S\n");
printf("============================================================================\n");
#undef TESTCASE1
#undef TESTCASE2
#undef TESTCASE3
#define TESTCASE1() x=x+C[i]
#define TESTCASE2() x=y*D[i]
#define TESTCASE3() x=x*y
#undef FLOPNUM
#define FLOPNUM 1
#include "flopt.h"

if (force > 100) printf(" \n"); else printf("\n");

tulip_Barrier();
printf("DONE\n");

#endif
}

/*********************************************************************/
void initArray(int *x, int n) {
  int i;

  for (i=0; i<ARRAY_LEN+1; i++)
    x[i]=n;
}

/*******************************************************************/
void BasicAcceptTests() {
  tulip_GlobalPointer_t p;
  int me,h,i,Rnbor,Lnbor;
  double ew, eu;

  me = tulip_MyContext();

  initArray(SharedArray,me);
  initArray(LocalArray,-1);

  /* Construct GlobalPtr to Neighbor's Array */
  Rnbor = (me + 1);  if (Rnbor == tulip_NumContexts()) Rnbor = 0;
  Lnbor = (me - 1);  if (Lnbor < 0) Lnbor = tulip_NumContexts()-1;

  p=GlobalArrayTable[Rnbor];
  initArray(SharedArray,me);
  initArray(LocalArray,-1);

  printf("%d: Beginning FETCH Test  -> %d\n",tulip_MyContext(),Rnbor);

  ALL_BSTART(ew,eu);
  for (i=0; i<LOOPS; i++) {
    /* Fetch my neighbor's shared array into my local array */
    tulip_Get((char *) LocalArray,p,sizeof(int)*ARRAY_LEN,&h);
    tulip_WaitACK(&h);
  }
  ALL_BEND(ew,eu);

  printf("%d:  %7.2f User, %7.2f Wall, %5.2f MB/sec\n",me,eu,ew
    ,LOOPS*ARRAY_LEN*sizeof(int)/1024/1024/ew);

  /* Checking result... */
  if ( (LocalArray[0] != Rnbor) || (LocalArray[ARRAY_LEN-1] != Rnbor)) {
    printf("%d: Error, SharedArray (%d) incorrect: %d,%d  (%d) \n"
     ,me,Rnbor,LocalArray[0],LocalArray[ARRAY_LEN-1],SharedArray[ARRAY_LEN]);
  }
  else
    printf("%d: SharedArray (%d) is correct %d,%d  (%d) \n"
     ,me,Rnbor,LocalArray[0],LocalArray[ARRAY_LEN-1],SharedArray[ARRAY_LEN]);

  tulip_Barrier();

  initArray(SharedArray,-1);
  initArray(LocalArray,me);

  printf("%d: Beginning STORE Test  -> %d\n",tulip_MyContext(),Rnbor);

  ALL_BSTART(ew,eu);
  for (i=0; i<LOOPS; i++) {
    /* Fill neighbor's shared array with my local array */
    tulip_Put(p,(char *) LocalArray,sizeof(int)*ARRAY_LEN,&h);
    tulip_WaitACK(&h);
  }
  ALL_BEND(ew,eu);

  printf("%d:  %7.2f User, %7.2f Wall, %5.2f MB/sec\n",me,eu,ew
    ,LOOPS*ARRAY_LEN*sizeof(int)/1024/1024/ew);

  /* Checking result... */
  if ( (SharedArray[0] != Lnbor) || (SharedArray[ARRAY_LEN-1] != Lnbor)) {
    printf("%d: Error, SharedArray (%d) incorrect: %d,%d  (%d) \n"
     ,me,Lnbor,SharedArray[0],SharedArray[ARRAY_LEN-1],LocalArray[ARRAY_LEN]);
  }
  else
    printf("%d: SharedArray (%d) is correct %d,%d  (%d) \n"
     ,me,Lnbor,SharedArray[0],SharedArray[ARRAY_LEN-1],LocalArray[ARRAY_LEN]);
}

/*********************************************************************/
/* HANDLER!  This handler is involked when a RemoteAction request is */
/* sent to this processor. */
/*********************************************************************/
void tulip_RemoteActionHandler(int from_whom, char *buffer, int len)
{
tulip_GlobalPointer_t * t;

  if ( *((long *) buffer) == BOOTSTRAP ) {
    t = (tulip_GlobalPointer_t *) (buffer + sizeof(long));
    GlobalArrayTable[from_whom]=*t;
/*    printf("%d: Received BootStrap data from %d\n"
      ,tulip_MyContext(),from_whom);*/
  }
  else
  if ( *((long *) buffer) == PINGRPC ) {
    if (tulip_MyContext() == 1) { /* Send a PONG back */
/*      printf("%d: Ping\n",tulip_MyContext());*/
      dummy[0] = PINGRPC;
      tulip_RemoteAction(0,(char *) dummy,sizeof(long),0);
    }
    else{
/*      printf("%d: Pong\n",tulip_MyContext());*/
      PingPongFlag=1;
    }
  }
  else
    printf("%d: Ignoring unknown RemoteAction from %d  (%x) \n",
             tulip_MyContext(),from_whom, *((long *) buffer) );
}

/*********************************************************************/
/* Set up information for where data is, and how it can be accessed */
/*********************************************************************/
void BootStrap()
{

tulip_GlobalPointer_t p;
int i;
char buffer[100];
char *tmpptr;

/* Allocate memory for SharedArray */
SharedArray = (int *) tulip_SharedMalloc(sizeof(int)*(ARRAY_LEN+1));
LocalArray  = (int *) tulip_SharedMalloc(sizeof(int)*(ARRAY_LEN+1));
h           = (int *) tulip_SharedMalloc(sizeof(int)*1024);

tulip_MsgBlockSize=-1;

/*printf("%d: Array byte length=(%x):%d  tulip_MsgBlockSize=%d\n"
  ,tulip_MyContext(),SharedArray,ARRAY_LEN*sizeof(int),tulip_MsgBlockSize);*/

if ( !(SharedArray) || !(LocalArray) ) {
  printf("%d: Error, could not allocate Local & Shared Arrays\n"
    ,tulip_MyContext());
  tulip_Abort();
}

/* Make Globalptr out of address */
tulip_SetLocalPointer(&p,(void *) SharedArray);
tulip_SetContext(&p,tulip_MyContext());

/* Create a RemoteAction */
tmpptr = buffer;
*((long *) tmpptr) = BOOTSTRAP;     /* Code for boot */
tmpptr = tmpptr + sizeof(long);
*((tulip_GlobalPointer_t *) tmpptr) = p;       /* Address for my portion of SharedArray */

/* Give every other processor a Global Pointer to my portion of SharedArray */
/* invoke a RemoteAction on every remote node, handing it buffer */

for (i=0; i<tulip_NumContexts(); i++)
  if (i != tulip_MyContext())
    tulip_RemoteAction(i,buffer,sizeof(long)+sizeof(tulip_GlobalPointer_t),&h[i]);

/* Wait for all my RemoteActions to complete */
for (i=0; i<tulip_NumContexts(); i++)
  if (i != tulip_MyContext()) {
    tulip_WaitACK(&h[i]);
  }

/* Wait for everyone to complete BootStrap */
tulip_Barrier();
}


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
int tulip_ParallelMain(int argc, char *argv[]) { 

  printf("%d: Beginning Bootstrap...\n",tulip_MyContext());

  /* BootStrap system, find other nodes, addresses, etc */
  BootStrap();  
  tulip_Barrier();

  /* My code */
/*  BasicAcceptTests(); */

  sum = 1.0;
  tests();
  printf("Final Sum: %f\n",sum);

  /* Proper Exit */
  return 0;
}
