/* This is a test of pcxx_* runtime library */

#include "tulip.h"
#include <stdio.h>

#define BOOTSTRAP 23
#define ARRAY_LEN 4096

#define ALL_BSTART(_a,_b) tulip_Barrier(); _a=tulip_WallClock(); \
                                          _b=tulip_UserClock();
#define ALL_BEND(_a,_b)   tulip_Barrier(); _a=tulip_WallClock()-_a; \
                                          _b=tulip_UserClock()-_b;

tulip_GlobalPointer_t GlobalArrayTable[1024]; /* A map to each proc's array portion */
int *SharedArray;                 /* My array portion */
int *LocalArray;                 /* My array portion */
int h[1024];

/*********************************************************************/
void initArray(int *x, int n) {
  int i;

  for (i=0; i<ARRAY_LEN+1; i++)
    x[i]=n;
}

/*******************************************************************/
void MyTestCode() {
  tulip_GlobalPointer_t p;
  int me,h,i,j;
  double ew, eu;

  me = tulip_MyContext();

  initArray(SharedArray,me);
  initArray(LocalArray,-1);

  /* Construct GlobalPtr to Neighbor's Array */
  j = (me + 1);  if (j == tulip_NumContexts()) j = 0;
  p=GlobalArrayTable[j];

  printf("%d: Beginning Test  -> %d\n",tulip_MyContext(),j);

  ALL_BSTART(ew,eu);
#define LOOPS 300
  for (i=0; i<LOOPS; i++) {
    /* Fill neighbor's shared array with my local array */
/*    tulip_Put(p,(char *) LocalArray,sizeof(int)*ARRAY_LEN,&h);*/
    tulip_Get((char *) LocalArray,p,sizeof(int)*ARRAY_LEN,&h);
    tulip_WaitACK(&h);
  }
  ALL_BEND(ew,eu);

  printf("%d:  %7.2f User, %7.2f Wall, %5.2f MB/sec\n",me,eu,ew
    ,LOOPS*ARRAY_LEN*sizeof(int)/1024/1024/ew);

  /* Checking result... */
/*  j = (me - 1); if (j < 0) j = tulip_NumContexts() - 1;*/

  if ( (LocalArray[0] != j) || (LocalArray[ARRAY_LEN-1] != j)) {
    printf("%d: Error, SharedArray (%d) incorrect: %d  %d\n"
      ,me,j,SharedArray[0],SharedArray[ARRAY_LEN-1]);
  }
  else
    printf("%d: SharedArray (%d) is correct %d,%d  (%d) \n"
     ,me,j,SharedArray[0],SharedArray[ARRAY_LEN-1],SharedArray[ARRAY_LEN]);

}

/*********************************************************************/
/* HANDLER!  This handler is involked when a RemoteAction request is */
/* sent to this processor. */
/*********************************************************************/
void tulip_RemoteActionHandler(int from_whom, char *buffer, int len)
{
tulip_GlobalPointer_t * t;
  if ( *((int *) buffer) == BOOTSTRAP ) {
    t = (tulip_GlobalPointer_t *) (buffer + sizeof(int));
    GlobalArrayTable[from_whom]=*t;
/*    printf("%d: Received BootStrap data from %d\n"
      ,tulip_MyContext(),from_whom);*/
  }
  else
    printf("%d: Ignoring unknown RemoteAction from %d  (%x) \n",
             tulip_MyContext(),from_whom, *((int *) buffer) );
}

/*********************************************************************/
/* Set up information for where data is, and how it can be accessed */
/*********************************************************************/
void BootStrap()
{

tulip_GlobalPointer_t p;
int i;
char buffer[100];
char *tmpptr;

/* Allocate memory for SharedArray */
SharedArray = (int *) malloc(sizeof(int)*(ARRAY_LEN+1));
LocalArray  = (int *) malloc(sizeof(int)*(ARRAY_LEN+1));

printf("%d: Array byte length=(%x):%d  tulip_MsgBlockSize=%d\n"
  ,tulip_MyContext(),SharedArray,ARRAY_LEN*sizeof(int),tulip_MsgBlockSize);

if ( !(SharedArray) || !(LocalArray) ) {
  printf("%d: Error, could not allocate Local & Shared Arrays\n"
    ,tulip_MyContext());
  tulip_Abort();
}

/* Make Globalptr out of address */
tulip_SetLocalPointer(&p,(void *) SharedArray);
tulip_SetContext(&p,tulip_MyContext());

/* Create a RemoteAction */
tmpptr = buffer;
*((int *) tmpptr) = BOOTSTRAP;     /* Code for boot */
tmpptr = tmpptr + sizeof(int);
*((tulip_GlobalPointer_t *) tmpptr) = p;       /* Address for my portion of SharedArray */

/* Give every other processor a Global Pointer to my portion of SharedArray */
/* invoke a RemoteAction on every remote node, handing it buffer */

for (i=0; i<tulip_NumContexts(); i++)
  if (i != tulip_MyContext())
    tulip_RemoteAction(i,buffer,sizeof(int)+sizeof(tulip_GlobalPointer_t),&h[i]);

/* Wait for all my RemoteActions to complete */
for (i=0; i<tulip_NumContexts(); i++)
  if (i != tulip_MyContext()) {
    tulip_WaitACK(&h[i]);
  }

/* Wait for everyone to complete BootStrap */
tulip_Barrier();
}


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
int tulip_ParallelMain(int argc, char *argv[]) { 

  printf("Beginning parall code\n");

  /* Parse Args */

  /* BootStrap system, find other nodes, addresses, etc */
  BootStrap();  

  /* My code */
  MyTestCode();  

  /* Proper Exit */
  return 0;
}

