#include <tulip.h>
#include "awe-extra.h"
#include <CpuMux.h>
#include <Barrier.h>
#include <MultiFifoMux.h>
#include <Iterate.h>

// Program written by Suvas Vajracharya from a template provided by 
// Pete Beckman.

extern CONT *contSwitch(void *pcaw_cc, int pcaw_cf);
extern void WorkerSwitch(void *, void *);
extern pcxx_PARAM_t pcxx_param;

/* These functions are defined in kernel-AWE.h */
extern void *kernel_getIterate(void *, int);
extern void *kernel_getIterBarrierPtr(void *);
extern void  kernel_setIterate(void *, int, void *);
extern void  kernel_setIterBarrierPtr(void *, void *);
extern  int  kernel_getHeight(void *);
extern void  kernel_storeElementNo(int,int);

extern  void Processor_Main(int , char **);

MultiFifoMux Cpu;
static Barrier AweBarrier;

// pcxx_collection is subclass of the Iterate class
void pcxx_collection::main() 
{
  //this is a hack to pass a parameter to the FirstLocal function that will
  // be called WorkerSwitch routine in the compiler generate file temp.C
  // kernel_storeElementNo stores the elementno now so that it can be 
  // picked up by FirstLocal later but using its cpu id as an index
  kernel_storeElementNo(elementno,CpuMux::cpuId()); 
  func(arg);
}

pcxx_collection::pcxx_collection(voidPtrFuncP _WorkerSwitch, void * _arg,
           INDEX _index, LOOP *_lp, int _elemno, IterateBarrier * _IterBarPtr) : Iterate(_index,_lp,_IterBarPtr) 
{
   arg = (struct params *)_arg;
   func = _WorkerSwitch;
   elementno = _elemno;
}

Iterate **pcxx_collection::generateGoals() 
{
  struct params *p = (struct params *)arg;

  // cont Switch returns a list of continuations which if they can
  // be enabled (i.e. there dependance count is zero after a decr) 
  // will provoke the runtime system to add them to the work queue.

  CONT *conts = contSwitch(p->c,p->s);
  if (conts == NULL) {
    barrierPtr->done(); 
    return NULL;
  }

  numGoals = conts[0].numGoals;
  if (goals == NULL)
    goals = new IteratePtr[numGoals];

  for (int i=0; i <numGoals; i++) {
    pcxx_collection *Iter = (pcxx_collection *)conts[i].Iterate;
    Iter->arg = conts[i].arg;
    goals[i++] = Iter;
  }
  return goals;
}

void MainControl::main()
{
    Processor_Main(argc,argv);
}
/* Initialize threads, uses argc/argv, and modifying them */
extern "C" void thread_Init(int argc, char **argv)
{
  AweBarrier.height(pcxx_param.numproc);
  CpuMux::add (new MainControl(argc,argv));
  Cpu.fireItUp(pcxx_param.numproc,pcxx_param.shmemsize);  
}

/* Shut down and wrap up the thread package */
extern "C" void thread_Term()
{
}

/* For some work onto a thread, passing a single void* argument */
extern "C" void thread_Fork( void (*function)(), void *arg )
{
  struct params *p = (struct params *)arg;
  INDEX index(0,0);
  IteratePtr IterPtr;

  IterateBarrier *BarPtr = (IterateBarrier *)kernel_getIterBarrierPtr(p->c);
  int nbelements = kernel_getHeight(p->c);

  if (BarPtr == NULL) {
    BarPtr = new IterateBarrier(kernel_getHeight(p->c));
    kernel_setIterBarrierPtr(p->c,(void *)BarPtr);
    for (int pcxx_0 = 0; pcxx_0 < nbelements; pcxx_0++) {
      IterPtr = (IteratePtr) kernel_getIterate(p->c,pcxx_0);
      if (IterPtr == NULL) {
	LOOP *loop = new LOOP(0,0,0,NULL,0);
	IterPtr = new pcxx_collection((voidPtrFuncP)function,p,index,loop,pcxx_0,BarPtr);
	IterPtr->last_proc = pcxx_0%CpuMux::Muxs;
	kernel_setIterate(p->c,pcxx_0,(void *)IterPtr);
      }
      CpuMux::AddToCpu(pcxx_0%CpuMux::Muxs,IterPtr);
    }
  }
  else {
    BarPtr->setHeight(nbelements);
    for (int pcxx_0 = 0; pcxx_0 < nbelements; pcxx_0++) {
      IterPtr = (IteratePtr) kernel_getIterate(p->c,pcxx_0);
      CpuMux::AddToCpu(IterPtr->whereIsTheData(),IterPtr);
    }
  }
  BarPtr->wait();
}

/* Am I somehow "distinguished" from the other threads? */
extern "C" void thread_IamDistinguished()
{}

extern "C" void thread_Barrier()
{
  AweBarrier.rendezvous();
}

extern "C" int thread_MyID() 
{
  return CpuMux::cpuId(); 
}

extern "C" int thread_NumThreads()
{
  return CpuMux::Muxs;
}
