/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/

/* REMOVE thread arguments from argc/argv */
/* Pete Beckman */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "thread_iface.h"

static char * ArgPrefix = { "-thread_" };

/* Declare struct to hold parameters */
/* This is what is exported from this program */
thread_PARAM_t thread_param;

static char * NumThreadsArgString = { "NUMTHREADS" } ;
static char * StackSizeArgString = { "STACKSIZE" } ;

/*********************************************************************/
/* Process parameters for the thread package */
void thread_InitParam(int *argc, char **argv)
{
  char thearg[1024];
  char *theargname;
  int i, j;
  char *envptr, *argptr;
  int temp1;
  int skip;
  int RealArgc;

  strcpy(thearg,ArgPrefix);
  theargname = &thearg[strlen(thearg)];
  theargname[0] = 0;

  RealArgc = *argc;

  /* Set default parameters */
  thread_param.numthreads = THREAD_NUMTHREADS;
  thread_param.stacksize = THREAD_STACKSIZE;

  /******************************/
  /* Set default for parameters */

  /* NumThreads */
  theargname[0] = 0;
  strcat(theargname,NumThreadsArgString);

  if ( (envptr = getenv(thearg+1)) != NULL ) {
    if ( (temp1 = atoi(envptr)) <= 1 )
      fprintf(stderr, "Warning: environment variable %s is bogus - ignored\n",
	       thearg+1);
    else thread_param.numthreads = temp1;
  }

  /* StackSize */
  theargname[0] = 0;
  strcat(theargname,StackSizeArgString);

  if ( (envptr = getenv(thearg+1)) != NULL ) {
    if ( (temp1 = atoi(envptr)) <= 1 )
      fprintf(stderr, "Warning: environment variable %s is bogus - ignored\n",
	       thearg+1);
    else thread_param.stacksize = temp1;
  }


  /******************************************************************/
  /* Cycle through all the arguments, looking for ones to pluck off */
  i = 1;
  while (i < RealArgc) {
    skip = 0;

    /* Is this argument for me? */
    if (!strncmp(argv[i], ArgPrefix, strlen(ArgPrefix))) {

      /* NumThreads */
      /***************************************************************/
      theargname[0] = 0;
      strcat(theargname,NumThreadsArgString);

      /* Is this the arg I'm looking for */
      if ( !(strcmp(argv[i], thearg)) ) {
	skip++;
	/* is there something after this to parse? */
	/* Assumes NON-NEGATIVE values */
	if ( (i+1 == RealArgc) || (argv[i+1][0] == '-')) {
	  /* at end of arg list, or no arg */
	  fprintf(stderr, 
		  "Warning: argument for %s missing, option ignored.\n", 
		  argv[i]);
	}
	else {
	  skip++;
	  /* Grab the parameter, is it valid? */
	  if ( (temp1 = atoi(argv[i+1])) < 1 ) {
	    fprintf(stderr, 
		    "Warning: arguments '%s %s' arg bogus, option ignored.\n", 
		    argv[i],argv[i+1]);
	  } 
	  else {
	    thread_param.numthreads = temp1;
	  }
	}
      }

      /* StackSize */
      /***************************************************************/
      theargname[0] = 0;
      strcat(theargname,StackSizeArgString);

      /* Is this the arg I'm looking for */
      if ( !(strcmp(argv[i], thearg)) ) {
	skip++;
	/* is there something after this to parse? */
	/* Assumes NON-NEGATIVE values */
	if ( (i+1 == RealArgc) || (argv[i+1][0] == '-')) {
	  /* at end of arg list, or no arg */
	  fprintf(stderr, 
		  "Warning: argument for %s missing, option ignored.\n", 
		  argv[i]);
	}
	else {
	  skip++;
	  /* Grab the parameter, is it valid? */
	  if ( (temp1 = atoi(argv[i+1])) < 1 ) {
	    fprintf(stderr, 
		    "Warning: arguments '%s %s' arg bogus, option ignored.\n", 
		    argv[i],argv[i+1]);
	  } 
	  else {
	    thread_param.stacksize = temp1;
	  }
	}
      }

      if (!skip) {
	fprintf(stderr, 
		"Error: parameter %s not understood, execution halted\n",
		argv[i]);
	exit(1);
      }

    }
    /* ONLY decrement the total args, if we remove some */
    *argc -= skip;
    /* Should I shift some args */
    if (skip) {
      for (j=i; j<RealArgc; j++) 
	argv[j] = argv[j+skip];
      RealArgc -= skip;
    }
    else 
      i++;
  }
}
