/*********************************************************************/
/*                  pC++/Sage++  Copyright (C) 1993,1995             */
/*  Indiana University  University of Oregon  University of Rennes   */
/*********************************************************************/


/* Insert Call to library on var references  ??? */

/**************************************************************************/
/* Rewritten by Pete Beckman on 11/19/92 */

/* Include files */
/********************/
#include <stdio.h>
#include "sage.h"
#include "macro.h"
#include "ext_lib.h"
#include "ext_low.h"
#include "sage.h"

extern int Warning_count;
extern int CommentOut;

FILE *outputfile;
int verboseMode = 0; /* Will there be lots of messages to console? */

/**********************************************************************/
void printUsage(char* execName)
{
  fprintf(stderr,"Usage: %s\n",execName);
  fprintf(stderr,"    [-help | -h] [-usage | -u]\n");
  fprintf(stderr,"    [-debug] [-verbose | -v] [-version]\n");
  fprintf(stderr,"    [-c] [-p]\n");
  fprintf(stderr,"    <filename>\n");
  fprintf(stderr,
     "\n'%s -help' prints description of switches, parameters\n",execName);
}

/**********************************************************************/
/* printHelp - prints description of switches and options

   Input:
   execName - pointer to char containing name of executable
   Output:
   to stdout console */
/**********************************************************************/
void printHelp(char* execName)
{
  printf("Usage: %s\n",execName);
  printf("    [-help | -h]  Print this help text\n");
  printf("    [-usage | -u] Print usage info (command line switches)\n");
  printf("    [-debug] Print useful debugging information\n");
  printf("    [-verbose | v] Turn on Verbose Mode \n");
  printf("    [-version]  Print version number of compiled program\n\n");
  printf("    [-c] print C++ comments\n");
  printf("    [-p] input is a project\n");
  printf("    <filename>\n\n");

  printf(SAGE_INFO);
}

/**********************************************************************/
/* main

   Input :
   command line switches and a filename, or stdin */
/**********************************************************************/
int main(int argc, char *argv[])
{
  char *finput;
  char *executableName = argv[0]; /* Save a pointer to the exec name */
  int projectName = 0;

  argv++;

  CommentOut = 0;  /* Default is to NOT print comments */

  while ((argc > 1) && (*argv)[0] == '-') {
    /* Begin standard flags */
    if ((!strcmp(argv[0],"-help")) || (!strcmp(argv[0],"-h"))) {
      printHelp(executableName);
      return 1;  /* error: non-zero exit condition */
    }
    else if ((!strcmp(argv[0],"-usage")) || (!strcmp(argv[0],"-u"))) {
      printUsage(executableName);
      return 1;  /* error: non-zero exit condition */
    }
    else if (!strcmp(argv[0],"-debug")) {
      printf("This program currently has no debugging information\n");
      return 1;  /* error: non-zero exit condition */
    }
    else if ((!strcmp(argv[0],"-verbose")) || (!strcmp(argv[0],"-v"))) {
      fprintf(stderr,"Verbose Mode on\n");
      verboseMode = 1;
    }
    else if (!strcmp(argv[0],"-version")) {
      printf("This is version %s of %s\n",VERSION_NUMBER,executableName);
      return 1;  /* error: non-zero exit condition */
    }

   
    /* begin more interesting flags*/
    else if (!strcmp(argv[0],"-c")) {
      CommentOut = 1;
    }
    else if (!strcmp(argv[0],"-p")) {
       projectName = 1;
    }
    else
      {
        fprintf (stderr, "%s: unrecognized option `%s'\n",
                 executableName, argv[0]);
        printUsage(executableName);
        return 1;  /* error: non-zero exit condition */
      }
    argc--;
    argv++;
  }

  outputfile = stdout;
  finput = *argv;

  if (projectName)
    /* A project name was given */ 
    if (*argv) {
      if (open_proj_toolbox(finput, finput) < 0) {
	fprintf(stderr,"ERROR: Cannot open project file %s\n",finput);
	return 1;
      }
      else {
	if (verboseMode) fprintf(stderr,"Opened project file %s\n",finput);
	Init_Tool_Box();
	if (verboseMode) fprintf(stderr,"Begining Unparse\n");
	UnparseProgram(outputfile); 
	if (Warning_count != 0)
	  fprintf(stderr,"You got %d Warning\n",Warning_count);
	return 0;
      }
    }
    else {
      fprintf(stderr,"ERROR: -p option, project file name must be given\n"); 
      printUsage(executableName);
      return 1; 
    }

  else {  /* A dep file specified */
    if (*argv) { /* filename was given */
      if (verboseMode) fprintf(stderr,"Attempting to open file %s\n",finput);
      if ((cur_proj = OpenProj("test", 1, argv))) {
	cur_file = (PTR_FILE) BLOB_VALUE (CUR_PROJ_FILE_CHAIN());
	pointer_on_file_proj = cur_file;
	Init_Tool_Box();
	Init_Unparser();
  if (Check_Lang_Fortran(cur_proj))
        {
	fprintf(outputfile,"%s",filter(Tool_Unparse_Bif(PROJ_FIRST_BIF ())));
        }
  else
	fprintf(outputfile,"%s",Tool_Unparse_Bif(PROJ_FIRST_BIF ()));

	return 0;
      } 
      else {
	fprintf(stderr,"ERROR: Cannot open input file %s\n",finput);
	return 1;
      }
      
    }
    else {
      fprintf(stderr,"ERROR: Input filename must be specified\n");
      printUsage(executableName);
      return 1;
    }
  }
}






