; t a s k i n g  . s c h
;
;   Copyright (C) 1995 The University of Oregon
;   Copyright (C) 1995 Lars Thomas Hansen <lth@cs.uoregon.edu>
;
;   Permission to use, copy, and/or distribute this software and its
;   documentation for any purpose and with or without fee is hereby granted,
;   provided that both the above copyright notice and this permission notice
;   appear in all copies and derived works.  
;
;   This software is provided ``as is'' without express or implied warranty.
;
; lth@cs.uoregon.edu / June 21, 1995
; $Id$
;
; Low-level task implementation: using call-with-current-continuation.
;
; This is the only part of the (Scheme-level) system which knows that
; tasks are implemented using continuations.
;
; BUGS and ANNOYANCES
; - 'Tasking/create-task' should prune stack before spawning, as the 
;   STk call/cc implementation re-captures all of the stack.
; - Call/cc does not work in well-defined ways across tasks and should
;   be redefined.

; Requires nothing.

; Capture the state of the current task in a data structure and pass that
; to p, which may or may not be run in a well-defined task (i.e., it
; may run in a type of supervisory mode.) Interrupts should be off.

(define (tasking/capture-task p)
  (call-with-current-continuation
   (lambda (k)
     (p (lambda () (k 'ignored))))))

; Given a task structure as returned by capture-task or create-task, 
; run that task.

(define (tasking/resume-task t)
  (t))

; Create a task structure on which to run p.

(define (tasking/create-task p)
  p)

; Destroy a task structure (and its task).

(define (tasking/destroy-task t)
  #t)

(provide "tasking.sch")

; eof
