/*************************************************************************/
/* 
pC++2dep BISON grammar for parsing pC++ programs (pC++ -> .dep)
 
MODIFIED BY:
       Pete Beckman & Francois Bodin  
       beginning 11/24/92
       Dennis Gannon
       beginning 6/1/93
       Beata Winnicka
       beginning 2/1/94
 
ORIGINAL PC++ GRAMMAR WRITTEN BY:
       Jenq-kuen Lee  Sep 28, 1987
 
DERIVED FROM:
       YACC parser for C syntax.
       Copyright (C) 1987 Free Software Foundation, Inc.

*/

/* "$Dummy" Variables have been replaced with the "" nil parameter (PHB) */

%{
/* Include files */
/********************/
#include "macro.h"        /* All the macros for node manipulation */
#include "vpc.h"          /* A file of #define constants */
#include "vextern.h"      /* declarations for external variables and procs */
#include "ext_low.h"
#include "ext_parser_lib.h"
#include "ext_vinit.h"
#include "ext_vhash.h"
#include "ext_utility.h"

/* We may use builtin alloca */
#include "compatible.h"
#ifdef SYS5
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef _NEEDALLOCAH_
#  include <alloca.h>
#endif

extern int yylex();    /* from vlex.c    */
extern void yyerror(); /* from verrors.c */
extern void warning(); /* from verrors.c */
extern int ExternLangDecl;   /* this is supposed to be in vextern!!!!!! PHB */
/*************************************************************************/
%}

%start program

%union { 
         int       token;
         char      charv;
         char      *charp;
         PTR_BFND   bfnode;
         PTR_LLND   ll_node;
         PTR_SYMB   symbol;
         PTR_TYPE   data_type;
         PTR_HASH   hash_entry;
         PTR_LABEL  label;        
         PTR_BLOB   blob_ptr;
       }

/*************************************************************************/
/* This group is used by the C/C++ language parser */
%token AUTO            DOUBLE          INT             STRUCT
%token BREAK           ELSE            LONG            SWITCH
%token CASE            ENUM            REGISTER        TYPEDEF
%token CHAR            EXTERN          RETURN          UNION
%token CONST           FLOAT           SHORT           UNSIGNED
%token CONTINUE        FOR             SIGNED          VOID
%token /*DEFAULT*/     GOTO            SIZEOF          VOLATILE
%token DO              IF              STATIC          WHILE

%token DEFAULT_TOKEN  /* used because DEFAULT already has a binding */

/* The following are used in C++ only. ANSI C would call these IDENTIFIERs */
%token NEW             DELETE
%token THIS
%token OPERATOR
%token CLASS
%token PUBLIC          PROTECTED       PRIVATE
%token VIRTUAL         FRIEND
%token INLINE          OVERLOAD
%token TRY CATCH THROW TEMPLATE

/* pC++ extentions to the language */
%token TECLASS COLLECTION METHODOFELEMENT

/* the following come from CC++ */
%token SYNC GLOBL PAR PERFER SPAWN ATOMIC
%token RESTRICT KSRPRIVATE  /* C++ extension */

/* from Roskind */
/*  There  are  really  several distinct sets of storage_classes. The 
sets vary depending on whether the declaration is at file scope, is a 
declaration within a struct/class, is within a function body, or in a 
function declaration/definition (prototype  parameter  declarations).  
They   are   grouped  here  to  simplify  the  grammar,  and  can  be 
semantically checked.  Note that this  approach  tends  to  ease  the 
syntactic restrictions in the grammar slightly, but allows for future 
language  development,  and tends to provide superior diagnostics and 
error recovery (i_e.: a syntax error does not disrupt the parse).
 
                File    File    Member  Member  Local   Local  Formal
                Var     Funct   Var     Funct   Var     Funct  Params
TYPEDEF         x       x       x       x       x       x
EXTERN          x       x                       x       x
STATIC          x       x       x       x       x
AUTO                                            x              x
REGISTER                                        x              x
FRIEND                                  x
OVERLOAD                x               x               x
INLINE                  x               x               x
VIRTUAL                                 x               x
*/


/*  Grammar Ambiguity noted by Jenq-Kuen Lee, Feb 26, 1992 
Totally y++.y contains 47 shift/reduce and 2 reduce/reduce conflicts

(* when error occurs, shift/reduce conflicts are okay *)
State 49 contains 1 shift/reduce conflict.   
State 97 contains 1 shift/reduce conflict.
State 102 contains 1 shift/reduce conflict.

(* declarator situations, shift gets the precedence *)
State 138 contains 2 shift/reduce conflicts.
State 140 contains 2 shift/reduce conflicts.
State 243 contains 2 shift/reduce conflicts.
State 244 contains 2 shift/reduce conflicts.
State 553 contains 2 shift/reduce conflicts.
State 554 contains 2 shift/reduce conflicts.
State 555 contains 2 shift/reduce conflicts.

(* constructor base init,  shift gets the precedence *)
State 200 contains 1 shift/reduce conflict.
State 333 contains 1 shift/reduce conflict.

(* reduce/reduce conflicts, we need to adjust grammar rules to solve
   the ambiguous. The following order is suggested. 
    notype_declarator_parm  ->  notype_declarator_parm '(' . expr ')'   
    notype_declarator_parm  ->  notype_declarator_parm '(' . parmlist ')'
    notype_declarator_parm  ->  notype_declarator_parm '(' . identifiers ')'
*)
State 202 contains 2 reduce/reduce conflicts.


(* similar to declarator situations only wihout declarator (absdcl),
   shift gets the precedence *)
State 276 contains 4 shift/reduce conflicts.
State 280 contains 4 shift/reduce conflicts.
State 348 contains 2 shift/reduce conflicts.
State 461 contains 2 shift/reduce conflicts.
State 462 contains 2 shift/reduce conflicts.
State 505 contains 2 shift/reduce conflicts.
State 506 contains 2 shift/reduce conflicts.
State 552 contains 2 shift/reduce conflicts.

(* initlist *)
State 420 contains 1 shift/reduce conflict.

(* others *)
State 218 contains 1 shift/reduce conflict.
State 277 contains 1 shift/reduce conflict.
State 281 contains 2 shift/reduce conflicts.
State 354 contains 2 shift/reduce conflicts.
State 535 contains 1 shift/reduce conflict.

*/


/*************************************************************************/
/* all identifiers that are not reserved words
   and are not declared typedefs in the current block */
%token IDENTIFIER
%token TFUNCNAME

/* all identifiers   that are declared typedefs in the current block.
   In some contexts, they are treated just like IDENTIFIER,
   but they can also serve as typespecs in declarations.  */
%token TYPENAME
%token SCOPED_TYPENAME /* BW typenames (enums) defined inside classes
/* reserved words that specify storage class.
   yylval contains an IDENTIFIER_NODE which indicates which one.  */

/* reserved words that specify type.
   yylval contains an IDENTIFIER_NODE which indicates which one.  */

/* reserved words that modify type: "const" or "volatile".
   yylval contains an IDENTIFIER_NODE which indicates which one.  */

%token FUTURE
/*character or numeric constants.
   yylval is the node for the constant.  */
%token CONSTANT

/* String constants in raw form.
   yylval is a STRING_CST node.  */
%token STRING

/* "...", used for functions with variable arglists.  */
%token ELLIPSIS

/* the reserved words */
%token ASM
%token  LOADEDOPR 

/* Define the operator tokens and their precedences.
   The value is an integer because, if used, it is the tree code
   to use in the expression made from the operator.  */

%left  <charv> ';'
%left IDENTIFIER TYPENAME SCOPED_TYPENAME TFUNCNAME
%left  <charv> SUPER_OP
%left  <charv> ','
%right <charv> '='
%right <token> ASSIGN 
%right <charv> '?' ':'
%left <charv> OROR
%left <charv> ANDAND
%left <charv> '|'
%left <charv> '^'
%left <charv> '&'
%left <token> EQCOMPARE
%left <token> ARITHCOMPARE  '>' '<' 
%left <charv> LSHIFT RSHIFT
%left <charv> '+' '-'
%left <token> '*' '/' '%'
%right <token> UNARY PLUSPLUS MINUSMINUS 
%left HYPERUNARY 
%left <token> POINTSAT '.'
%right <token> DOUBLEMARK

%type <token> unop  general_classname
%type <token> typespecfinal typed_declfinal
%type <token> typespec_reserved  component_decl  ovldcast 
%type <token> storage_class storage_class2 basic_type_name 
%type <token> type_modifier type_qualifier access_specifier ovldtypqual
%type <token> component_decl_list  nonempty_component_decl_list 
%type <token> declarator_list decl extern_lang_string
%type <token> PAR SPAWN PERFER CLASS UNION STRUCT TECLASS

%token EXTERN_UNKNOWN_LANG EXTERN_C_LANG

%type <hash_entry> IDENTIFIER TYPENAME  LOADEDOPR SCOPED_TYPENAME TFUNCNAME
%type <hash_entry> identifier tp_doublemark  

%type <bfnode> datadef

%type <data_type> typed_declspecs  maintype typed_declspecs_2  
%type <data_type> typename absdcl absdcl1 typemods
%type <data_type> typed_typespecs nonempty_typemods structsp
%type <data_type> declmods star_ops ovldname
%type <data_type> nested_name_spec 

%type <ll_node> nested_name_spec_1
%type <ll_node> CONSTANT STRING  primary expr_no_commas expr_noc2
%type <ll_node>  expr nonnull_exprlist newexp new_init new_placement
%type <ll_node> base_init member_init_list member_init exprlist
%type <ll_node> initdecls initdcl notype_initdecls string
%type <ll_node> init initlist vector_constant triplet
%type <ll_node> compound_constant vector_list
%type <ll_node> declarator inside components 
%type <ll_node> notype_components component_declarator
%type <ll_node> notype_component_declarator  future_declarator
%type <ll_node> notype_declarator  after_type_declarator notype_initdcl
%type <ll_node> single_v_expr array_expr_a array_expr_b expr_vector multiple_id
%type <ll_node> xexpr declarator1 notype_declarator1 scoped_identifier forinit
%type <ll_node> enumlist enumerator ellipsis
%type <ll_node> parmlist parms parm  identifiers  notype_declarator_parm  
%type <ll_node> assign_stmt TYPENAME_list parm_init tp_arglist
%type <ll_node> passing_identifiers passing_identifier_list
%type <ll_node> template_arg_list template_arg temp_arg_spec
%type <token> EXTERN INLINE FRIEND STATIC
/*%type <token> ref_pointer PHB future use? */


/* the declaration found for the last IDENTIFIER token read in.
   yylex must look this up to detect typedefs, which get token type TYPENAME,
   so it is left around in case the identifier is not a typedef but is
   used in a context which makes it a reference to a variable.  */


/* list of types and structure classes of the current declaration */
/* tree current_declspecs;  */

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

%{

/* There is still some stuff from vextern.h not declared here */

/* External Procedures from verror.c */ 
/********************/
extern void err_symb(); /* from verror.c */
extern int err(); /* from verrors.c */

/* Other External Procedures */ 
/********************/
extern char       *copys(); /* from ../lib/make_nodes.c */

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

/* Other External References */
/********************/
extern PTR_FILE   fi; /* from main.c */
extern struct data_type second_part; /* from ../h/symb.h */
/*********************************************************************/
/*********************************************************************/
/*********************************************************************/
/* Global Module Variables */
/********************/
char      *input_filename;
int       enum_id = 0;
static PTR_HASH Op_Name_Hash;  /* needed for operator overloading         */
static PTR_LLND Op_Params;     /* these used only in ovldcast and ovldname */
static int Op_Type_Qual;       /* for type qualifier of over loaded op func */
PTR_HASH Scope_Name;    /* BW */
int Scope_Flag;          /* BW */
int bif_level_decl_specs = 0;  /* BW */
int basic_block_flag = 0;  /* BW */
PTR_LLND member_templ_parms = NULL; /* dbg - for template parameters of */
                                    /* externally declared members of templates*/
%}     


%%
program:  /* empty */
            {}
	| extdefs
            {}
	;

/* the reason for the strange actions in this rule
 is so that notype_initdecls when reached via datadef
 can find a valid list of type and sc specs in $0. */

extdefs:
          extdef
            {}
	| extdefs extdef
            {}
	;

extdef:
          fndef
            { main_type_flag = TYPE_FOUR; }
	| datadef
            { fill_comment();
              main_type_flag = TYPE_FOUR;
            }
	| ASM '(' STRING ')' ';'
            {}
        | TEMPLATE  '<' parmlist '>'
            { PTR_SYMB symb1;
              symb1 = NULL;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               TEMPLATE_FUNDECL,symb1,LLNULL);
              BIF_LL1(pred_bfnd) = $3;
	      ResetSymbolTable(pred_bfnd, $3); 
             }
           extdef
            {
	      pred_bfnd = BIF_CP(pred_bfnd);
	    }

        ;

/* modified */
datadef:
         notype_initdecls ';'
            { PTR_TYPE ptr1;
              PTR_SYMB dummy_symb;

              ptr1 = GetAtomicType(T_INT);
              /* notype_initdecls returns expr_list ll_node */  
              addToDeclaration(ptr1,OFF,$1,&dummy_symb,NO);
            }
        | declmods notype_initdecls ';'
            { PTR_TYPE ptr1;
              PTR_SYMB dummy_symb;
   
              ptr1 = $1;
              if (ptr1)
                {
                  TYPE_DESCRIP_BASE_TYPE(ptr1) = GetAtomicType(DEFAULT);
                  /* notype_initdecls returns expr_list ll_node */
                  addToDeclaration(ptr1,OFF,$2,&dummy_symb,NO);
                }else
              /*      err("Internal Error"); */
                {
                 ptr1 = GetAtomicType(T_INT);
                 addToDeclaration(ptr1,OFF,$2,&dummy_symb,NO);
                }
            }
        | typed_declspecs initdecls ';'
            { PTR_SYMB dummy_symb;

              addToDeclaration($1,OFF,$2,&dummy_symb,NO);
            }
        | declmods ';'
            {}
        | typed_declspecs ';'
            { PTR_SYMB symb1,symb2,name;
              PTR_BFND bifd;
  
              bifd = lastBifInBlobList(pred_bfnd);
              if (bifd && BIF_BLOB1(bifd) && !(BIF_SYMB(bifd)))
                { 
                  symb1 =  getFirstFieldOfStruct(bifd);
                  for (name=symb1; name; name=getClassNextFieldOrMember(name))
                    {  symb2=find_type_symbol(SYMB_PARENT(name),
                                              VARIABLE_NAME,area());
                       if ((symb2) && (SYMB_SCOPE(symb2) == pred_bfnd))
                         err_symb("(10) redeclaration symbol ",SYMB_IDENT(symb2));
                       else   
                         SYMB_SCOPE(name) = pred_bfnd;
                     }
                }
            }
        | OVERLOAD declarator_list ';'
            { /* Needs future work  */ }

        | error ';'
            {}

        | error '}'
            {}

        | ';'
            {}
        ;


fndef:    typed_declspecs declarator1
            { enterFuncBody(pred_bfnd,$1,$2); }
          xdecls
            {
            if (pred_bfnd && BIF_SYMB(pred_bfnd))
              {
                SYMB_DOVAR(BIF_SYMB(pred_bfnd)) -= XDECL;
                setParamTypeToDefault(BIF_SYMB(pred_bfnd));
                lateAdjustFunctor(pred_bfnd);
              } else
                err("internal error 1");
            }
          compstmt
            { function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd); /*BIF_CP(pred_bfnd);*/
            }
        | typed_declspecs declarator1 error
            {}
        | declmods notype_declarator1
            {
              if($1 != NULL){ 
                TYPE_DESCRIP_BASE_TYPE($1) = GetAtomicType(DEFAULT);
                enterFuncBody(pred_bfnd,$1,$2);
                }
               else enterFuncBody(pred_bfnd,GetAtomicType(DEFAULT),$2);
            }
          xdecls
            { if (pred_bfnd && BIF_SYMB(pred_bfnd))
                {
                  SYMB_DOVAR(BIF_SYMB(pred_bfnd)) -= XDECL;
                  setParamTypeToDefault(BIF_SYMB(pred_bfnd)); 
                  lateAdjustFunctor(pred_bfnd);
                }
            }
          compstmt
            { function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | declmods  notype_declarator1 error
            {}

        | notype_declarator1
            { enterFuncBody(pred_bfnd,GetAtomicType(DEFAULT),$1); }
          xdecls
            { SYMB_DOVAR(BIF_SYMB(pred_bfnd)) -= XDECL;
              setParamTypeToDefault(BIF_SYMB(pred_bfnd));
              lateAdjustFunctor(pred_bfnd);
            }
          compstmt
            { function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | notype_declarator1 error
            {}

        | extern_lang_decl
        ;


extern_lang_decl:
          extern_lang_string '{' 
            { if (ExternLangDecl==0) /* Ignore nested 'extern "C"'s */
                makeExternC(&pred_bfnd,&cur_flag,node_of_struct($1)); 
              else dontmakeExtrn(); /* to do the line-queue fetch */
              ++ExternLangDecl;
            }
          extdefs '}'
            { if (ExternLangDecl==1) /* Ignore nested 'extern "C"'s */
                pred_bfnd = BIF_CP(pred_bfnd);
              --ExternLangDecl;
            }

        | extern_lang_string
            { if (ExternLangDecl==0) /* Ignore nested 'extern "C"'s */
                 makeExternC(&pred_bfnd,&cur_flag,node_of_struct($1)); 
              else dontmakeExtrn();
              ++ExternLangDecl;
            }
          extdef
            {  if (ExternLangDecl==1) /* Ignore nested 'extern "C"'s */
                pred_bfnd = BIF_CP(pred_bfnd);
              --ExternLangDecl;
            }

        | extern_lang_string '{' '}' 
            { }  /* Do nothing, throw it away! */
        ;       


extern_lang_string:
          EXTERN_UNKNOWN_LANG 
            { err("Unsupported external language specifier"); }
        | EXTERN_C_LANG 
            { $$ = EXTERN_C_LANG; }
        ;


declarator1:
          declarator
            { $$ = $1;
              installFunctionHedr(&pred_bfnd);
            }
        | declarator base_init
            { $$ = installLowLevelNode(BIT_NUMBER,$1,$2,SMNULL); }
        ;


notype_declarator1:
          notype_declarator
            { $$ = $1;
              installFunctionHedr(&pred_bfnd);
            }
        | notype_declarator base_init
            { $$ = installLowLevelNode(BIT_NUMBER,$1,$2,SMNULL); }
        ;


base_init: ':'  
            { installFunctionHedr(&pred_bfnd); }
          member_init_list  /*%prec SUPER_OP */
            { $$= $3;}
         ;


member_init_list:
          /* empty */
            { $$ = LLNULL;}
        | member_init
            { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL);}
        | member_init_list ',' member_init
            { PTR_LLND ll_node1;
              ll_node1 = Follow_Llnd($1,2);
              NODE_OPERAND1(ll_node1) = 
                   installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL); 
              $$=$1;
            }
        | member_init_list error
            {}

        ;


/* Non-virtual baseclass initializers can live on the temporary_obstack,
   since they are used only for the compilation of this function.  Virtual
   baseclass initializers, however, must be built on the permanent_obstack.*/

member_init:
          '(' exprlist ')'
            { $$= installLowLevelNode(CLASSINIT_OP,LLNULL,$2,SMNULL); }

        | identifier '(' exprlist ')'
           { PTR_SYMB symbptr;
             PTR_LLND pllnd;

             symbptr=fts_by_inherited_class($1,VARIABLE_NAME,area());
             if (!symbptr) err_symb("no proper member/class",SYMB_IDENT($1));
             else {
               pllnd = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symbptr);
               $$= installLowLevelNode(CLASSINIT_OP,pllnd,$3,SMNULL);
             }
           }
        | scoped_identifier identifier '(' exprlist ')'
           {}
        ;


scoped_identifier:
          DOUBLEMARK                    {}
        | identifier DOUBLEMARK         { Scope_Name = $1;}
        | tp_doublemark                 {}
        | scoped_identifier DOUBLEMARK  {}
        ;


declarator_list :
          declarator                     {}
        | declarator_list ',' declarator {}
        ;


identifier:
          IDENTIFIER   {}
        | TYPENAME     {}
/*      | PTYPENAME    {} */
        ;

unop:     '&'
            { $$ = ADDRESS_OP; }

        | '-'
            { $$ = MINUS_OP; }

        | '+'
            { $$ = UNARY_ADD_OP; }

        | PLUSPLUS
            { $$ = PLUSPLUS_OP; }

        | MINUSMINUS
            { $$ = MINUSMINUS_OP; }

        | '~'
            { $$ = BIT_COMPLEMENT_OP; }

        | '!'
            { $$ = NOT_OP; }

/*      | DOUBLEMARK
            {$$ = SCOPE_OP;}
*/
        ;


expr:     nonnull_exprlist
            { $$ = $1; }
        ;

exprlist:
          /* empty */
            { $$ = LLNULL; }
        | nonnull_exprlist
            { $$ = $1; }
        ;

template_arg_list:
	template_arg
            { 
             $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); 
             }
        | template_arg_list ',' template_arg
            { PTR_LLND ll_ptr;

              ll_ptr = Follow_Llnd($1,2);
              NODE_OPERAND1(ll_ptr) = 
                   installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
              $$=$1;
             }
         ;

template_arg:
       expr_noc2
          {$$ = $1; }
      | maintype
          {PTR_LLND tref;
	   PTR_TYPE nt, scnt;
           PTR_SYMB scsymb;
           tref = installLowLevelNode(TYPE_REF, LLNULL, LLNULL, SMNULL);
	   if($1 && isASymbNode(NODE_CODE($1))){
		nt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                TYPE_SYMB_DERIVE(nt) = (PTR_SYMB) $1;
                if (main_type_flag==TYPE_FIVE) {    /* BW */
 /*                 TYPE_SCOPE_SYMB_DERIVE(nt) */
                 scsymb  = find_scope_of_scoped_type_symbol(((PTR_SYMB) $1), Scope_Name);
                 scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                 TYPE_SYMB_DERIVE(scnt) = scsymb;
                 TYPE_SCOPE_SYMB_DERIVE(nt) = scnt;
                   main_type_flag = TYPE_TWO;
                   }
		NODE_TYPE(tref) = nt;
		}
	   else NODE_TYPE(tref) = $1;
           /* $$ = installLowLevelNode(EXPR_LIST,tref,LLNULL,SMNULL); */
	   $$ = tref;
	   }
        ;



/* modified */
nonnull_exprlist:
          expr_no_commas
            { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL);
              NODE_TYPE($$) = NODE_TYPE($1);
            }
        | nonnull_exprlist ',' expr_no_commas
            { PTR_LLND ll_ptr;

              ll_ptr = Follow_Llnd($1,2);
              NODE_OPERAND1(ll_ptr) = 
                   installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
              $$=$1;
              NODE_TYPE($$) = NODE_TYPE($3);
            }
        ;


/* modified */
vector_constant: 
          '['  ']'   %prec ','
            { PTR_SYMB symb;
              PTR_HASH entry;  
              int status;
              $$ = installLowLevelNode(VECTOR_CONST,LLNULL,LLNULL,SMNULL);
              primary_flag = VECTOR_CONST_APPEAR;
              /* Temporarily setting */
              entry=look_up_type("vconstant",&status);
              symb = find_object_symbol(entry);
              if (!symb) {
                symb=install_entry_2(entry,CLASS_NAME,OTHER);
              }
              NODE_TYPE($$) = (PTR_TYPE) newNode(T_DERIVED_TYPE);
              TYPE_SYMB_DERIVE(NODE_TYPE($$)) = symb;
            }
        | '['  vector_list  ']'   %prec ','
            { PTR_SYMB symb;
              PTR_HASH entry;  
              int status;
              $$ = installLowLevelNode(VECTOR_CONST,$2,LLNULL,SMNULL);
              primary_flag = VECTOR_CONST_APPEAR;
              entry=look_up_type("vconstant",&status);
              symb = find_object_symbol(entry);
              if (!symb) {
                symb=install_entry_2(entry,CLASS_NAME,OTHER);
              }
              NODE_TYPE($$) = (PTR_TYPE) newNode(T_DERIVED_TYPE);
              TYPE_SYMB_DERIVE(NODE_TYPE($$)) = symb;
            }
        ;


vector_list :
          single_v_expr
            { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); }
        | vector_list ',' single_v_expr
            {
              PTR_LLND ll_node1;
              ll_node1 = Follow_Llnd($1,2);
              NODE_OPERAND1(ll_node1) = 
                   installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
              $$=$1;
            }
        ;

/* modified */
single_v_expr :
          expr_no_commas
            { $$ = $1; }
        | triplet
            { $$ = $1; }
        | compound_constant
            { $$ = $1; }        
        ;


triplet :
          expr_no_commas ':' expr_no_commas ':' expr_no_commas   %prec '.'
            { PTR_LLND p1,p2;
              p1 = installLowLevelNode(DDOT,$1,$3,SMNULL);
              p2 = installLowLevelNode(DDOT,p1,$5,SMNULL);
              $$ = p2;
            }
        | expr_no_commas ':' expr_no_commas               %prec '.'
            { $$= installLowLevelNode(DDOT,$1,$3,SMNULL); }
        ;


compound_constant :
          expr_no_commas '#' expr_no_commas
            { $$= installLowLevelNode(COPY_NODE,$1,$3,SMNULL); }
        ;


/* modified */
array_expr_a : 
          /* empty */
            {}
        | expr_no_commas ':' expr_no_commas ':' expr_no_commas  %prec ','
            { PTR_LLND p1,p2;
              p1 = installLowLevelNode(DDOT,$1,$3,SMNULL);
              p2 = installLowLevelNode(DDOT,p1,$5,SMNULL);
              $$ = p2;
            }
        | expr_no_commas ':' expr_no_commas                     %prec ','
            { $$= installLowLevelNode(DDOT,$1,$3,SMNULL); }
        ;


/* modified */
array_expr_b :  expr_no_commas '#' expr_no_commas  
            {}
        ;


/* modified */
expr_vector : expr_no_commas   /* original is expr  */
            {}
        | array_expr_a
            {}
        ;
 
expr_no_commas:
/*       basic_type_name '(' expr ')'  /* my new rule. dbg */
/*            { $$=installLowLevelNode(CAST_OP,$3,LLNULL,SMNULL);
              NODE_TYPE($$) = map_rid_to_type($1);
            }
REMOVED by PHB, was causing '(' parmlist ')' problems */       
        THROW
            { $$=installLowLevelNode(THROW_OP,LLNULL,LLNULL,SMNULL);
            }
        
       | THROW expr_no_commas %prec UNARY
            { $$=installLowLevelNode(THROW_OP,$2,LLNULL,SMNULL);
            }
     /* the following will cause the same conflict that pete found */
     /* that is, int foo(A (*f)()) will not parse correctly */
       | TYPENAME '(' exprlist ')'  
            { PTR_SYMB symb2;
              PTR_TYPE type2;
	      PTR_LLND ll2;
	      if($3) ll2 = $3;
	      else ll2 = installLowLevelNode(EXPR_LIST,LLNULL,LLNULL,SMNULL);
	      $$=installLowLevelNode(CAST_OP,ll2,LLNULL,SMNULL);
              symb2= find_type_symbol($1,TYPE_NAME,area());
              type2= (PTR_TYPE) newNode(T_DERIVED_TYPE);
              TYPE_SYMB_DERIVE(type2) = (PTR_SYMB) symb2;
              NODE_TYPE($$) = type2;
            }
       | TYPENAME '(' exprlist ')' '.' identifier 
            { PTR_SYMB symb2;
              PTR_TYPE type2;
	      PTR_LLND ll1, ll2;
	      PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;
	      if($3) ll2 = $3;
	      else ll2 = installLowLevelNode(EXPR_LIST,LLNULL,LLNULL,SMNULL);
	      ll1=installLowLevelNode(CAST_OP,ll2,LLNULL,SMNULL);
              symb2= find_type_symbol($1,TYPE_NAME,area());
              type2= (PTR_TYPE) newNode(T_DERIVED_TYPE);
              TYPE_SYMB_DERIVE(type2) = (PTR_SYMB) symb2;
              NODE_TYPE(ll1) = type2;

              sw = checkFieldDecl(pred_bfnd,&right_symbol,ll1,$6,0,0);
              if (sw ==NO){ err("error member"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,ll1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType(ll1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
            }
       | primary
            { PTR_TYPE type;
              /* Need Another way to check this one */
	      char tmpbuf[50];
              if (exception_flag == ON) {
                sprintf(tmpbuf,"Undefined symbol '%s'",
                   HASH_IDENT(cur_id_entry));
		err(tmpbuf);
		exception_flag =OFF;
                }
              if (NODE_CODE($1) == FUNCTION_REF)
                {
                  type= (PTR_TYPE) newNode(T_POINTER);
                  TYPE_BASE(type) = NODE_TYPE(NODE_SYMB($1));
                  NODE_TYPE($1) = type;
                }
              $$=$1;
            }
        | '*' expr_no_commas   %prec UNARY
            { 
              $$ = installLowLevelNode(DEREF_OP,$2,LLNULL,SMNULL);
              if (!typeConversionOrObjectHandling($2,NULL,"*",$$,0))
              NODE_TYPE($$) = adjustDerefType(NODE_TYPE($2),DEREF_OP);
            }

        | unop expr_no_commas  %prec UNARY
            { $$=installLowLevelNode($1,$2,LLNULL,SMNULL);
              if (!typeConversionOrObjectHandling($2,NULL,
                                                  map_unop($1),$$,0))
                {
                  if (($1 == ADDRESS_OP)&&(is_variable_1($2)))        /* BW */
                    NODE_TYPE($$) = adjustDerefType(NODE_TYPE($2),
                                                    ADDRESS_OP);
                  else NODE_TYPE($$) = NODE_TYPE($2);
                }
            }
        | '(' typename ')' expr_no_commas  %prec UNARY
            { PTR_LLND ll2;
              ll2 = makeDeclarExp(SMNULL, $2);
              $$=installLowLevelNode(CAST_OP,$4,ll2,SMNULL);
              NODE_TYPE($$) = $2;
            }
        | SIZEOF expr_no_commas  %prec UNARY
            { $$=installLowLevelNode(SIZE_OP,$2,LLNULL,SMNULL);
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | SIZEOF '(' typename ')'  %prec HYPERUNARY
            { PTR_LLND ll_1;
              /* PTR_LLND ll2; */
              /* ll2 = makeDeclarExp(SMNULL, $3); */
              ll_1 = installLowLevelNode(CAST_OP,LLNULL,LLNULL,SMNULL);
              NODE_TYPE(ll_1) = $3;
              $$=installLowLevelNode(SIZE_OP,ll_1,LLNULL,SMNULL);
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }

        | DELETE  expr_no_commas  %prec UNARY
            { $$=installLowLevelNode(DELETE_OP,$2,LLNULL,SMNULL);
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }

        | DELETE vector_constant  expr_no_commas  %prec UNARY
            {
              $$=installLowLevelNode(DELETE_OP,$3, $2,SMNULL);
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }

        | expr_no_commas '+' expr_no_commas
            { $$=installLowLevelNode(ADD_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"+",$$,1);
            }
        | expr_no_commas '-' expr_no_commas
            { $$=installLowLevelNode(SUBT_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"-",$$,1);
            }
        | expr_no_commas '*' expr_no_commas
            { /* $$=installLowLevelNode(MULT_OP,$1,$3,SMNULL);*/
              if ($1)
                {
                  $$=newExpr(MULT_OP,NODE_TYPE($1),$1,$3);
                  typeConversionOrObjectHandling($1,$3,"*",$$,1);
                } else
                  err("expr null in binary *");
            }
        | expr_no_commas '/' expr_no_commas
            { $$=installLowLevelNode(DIV_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"/",$$,1);
            }
        | expr_no_commas '%' expr_no_commas
            { $$=installLowLevelNode(MOD_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"%",$$,1);
            }
        | expr_no_commas LSHIFT expr_no_commas
            { $$=installLowLevelNode(LSHIFT_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"<<",$$,0))
                NODE_TYPE($$) = NODE_TYPE($1);
            }
        | expr_no_commas RSHIFT expr_no_commas
            { $$=installLowLevelNode(RSHIFT_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,">>",$$,0))
                NODE_TYPE($$) = NODE_TYPE($1);
            }
        | expr_no_commas ARITHCOMPARE expr_no_commas
            { int op1;

              op1 = ($2 == ((int) LE_EXPR)) ? LE_OP : GE_OP;
              $$=installLowLevelNode(op1,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,map_op($2),$$,0))
                NODE_TYPE($$) = typeConversion(NODE_TYPE($1), 
                                               NODE_TYPE($3));
            }
        | expr_no_commas '<'  expr_no_commas  
            { $$=installLowLevelNode(LT_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"<",$$,0))
                NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas '>'  expr_no_commas   
            { $$=installLowLevelNode(GT_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,">",$$,0))
                NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas EQCOMPARE expr_no_commas
            { int op1;

              op1 = ($2 == ((int) NE_EXPR)) ? NE_OP : EQ_OP;
                   $$=installLowLevelNode(op1,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,map_op($2),$$,0))
                NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas '&' expr_no_commas
            { $$=installLowLevelNode(BITAND_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"&",$$,0))
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas '|' expr_no_commas
            { $$=installLowLevelNode(BITOR_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"|",$$,0))
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas '^' expr_no_commas
            { $$=installLowLevelNode(XOR_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"^",$$,1);
            }
        | expr_no_commas ANDAND expr_no_commas
            { $$=installLowLevelNode(AND_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"&&",$$,0))
                NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas OROR expr_no_commas
            { $$=installLowLevelNode(OR_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"||",$$,0))
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_no_commas '?' expr_no_commas ':' expr_no_commas  /* expr */
            { PTR_LLND ll_node1;

              ll_node1=installLowLevelNode(EXPR_IF_BODY,$3,$5,SMNULL);
              $$=installLowLevelNode(EXPR_IF,$1,ll_node1,SMNULL);
              NODE_TYPE($$) = typeConversion(NODE_TYPE($3),NODE_TYPE($5));
            }
        | expr_no_commas '=' expr_no_commas
            { $$=installLowLevelNode(ASSGN_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"=",$$,1))
                NODE_TYPE($$)  =NODE_TYPE($1);
            }
        | expr_no_commas ASSIGN expr_no_commas
            { int op1;

              op1 = map_assgn_op($2);
              $$=installLowLevelNode(op1,$1,$3,NULL);
              if (!typeConversionOrObjectHandling($1,$3,map_op($2),$$,1))
                NODE_TYPE($$)  =NODE_TYPE($1); 
/*              $$=installLowLevelNode(ASSGN_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"=",$$,1))
                NODE_TYPE($$)  =NODE_TYPE($1);
*/            }
        ;

/* expr_noc2 is just expr_no_commas without the >.  this is used
   only in the template T< expression > case.  we need a better
   solution that uses precedence in a clever way. dbg
*/

expr_noc2:
       basic_type_name '(' expr ')'  /* my new rule. dbg */
            { $$=installLowLevelNode(CAST_OP,LLNULL,$3,SMNULL);
              NODE_TYPE($$) = map_rid_to_type($1);
            }

       | TYPENAME '(' exprlist ')'  /* my new rule. dbg */
            { PTR_SYMB symb2;
              PTR_TYPE type2;
	      $$=installLowLevelNode(CAST_OP,$3,LLNULL,SMNULL);
              symb2= find_type_symbol($1,TYPE_NAME,area());
              type2= (PTR_TYPE) newNode(T_DERIVED_TYPE);
              TYPE_SYMB_DERIVE(type2) = (PTR_SYMB) symb2;
              NODE_TYPE($$) = type2;
            }

       | primary
            { PTR_TYPE type;
              /* Need Another way to check this one */
	      char tmpbuf[50];
              if (exception_flag == ON) {
                sprintf(tmpbuf,"Undefined symbol '%s'",
                   HASH_IDENT(cur_id_entry));
		err(tmpbuf);
		exception_flag =OFF;
                }
              if (NODE_CODE($1) == FUNCTION_REF)
                {
                  type= (PTR_TYPE) newNode(T_POINTER);
                  TYPE_BASE(type) = NODE_TYPE(NODE_SYMB($1));
                  NODE_TYPE($1) = type;
                }
              $$=$1;
            }
        | '*' expr_noc2   %prec UNARY
            { 
              $$ = installLowLevelNode(DEREF_OP,$2,LLNULL,SMNULL);
              if (!typeConversionOrObjectHandling($2,NULL,"*",$$,0))
              NODE_TYPE($$) = adjustDerefType(NODE_TYPE($2),DEREF_OP);
            }

        | unop expr_noc2  %prec UNARY
            { $$=installLowLevelNode($1,$2,LLNULL,SMNULL);
              if (!typeConversionOrObjectHandling($2,NULL,
                                                  map_unop($1),$$,0))
                {
                  if (($1 == ADDRESS_OP)&&(is_variable($2)))
                    NODE_TYPE($$) = adjustDerefType(NODE_TYPE($2),
                                                    ADDRESS_OP);
                  else NODE_TYPE($$) = NODE_TYPE($2);
                }
            }
        | '(' typename ')' expr_noc2  %prec UNARY
            { PTR_LLND ll2;
              ll2 = makeDeclarExp(SMNULL, $2);
              $$=installLowLevelNode(CAST_OP,$4,ll2,SMNULL);
              NODE_TYPE($$) = $2;
            }

        | SIZEOF expr_noc2  %prec UNARY
            { $$=installLowLevelNode(SIZE_OP,$2,LLNULL,SMNULL);
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | SIZEOF '(' typename ')'  %prec HYPERUNARY
            { PTR_LLND ll_1;
              /* PTR_LLND ll2;
              ll2 = makeDeclarExp(SMNULL, $3); */
              ll_1 = installLowLevelNode(CAST_OP,LLNULL,LLNULL,SMNULL);
              NODE_TYPE(ll_1) = $3;
              $$=installLowLevelNode(SIZE_OP,ll_1,LLNULL,SMNULL);
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_noc2 '+' expr_noc2
            { $$=installLowLevelNode(ADD_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"+",$$,1);
            }
        | expr_noc2 '-' expr_noc2
            { $$=installLowLevelNode(SUBT_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"-",$$,1);
            }
        | expr_noc2 '*' expr_noc2
            { /* $$=installLowLevelNode(MULT_OP,$1,$3,SMNULL);*/
              if ($1)
                {
                  $$=newExpr(MULT_OP,NODE_TYPE($1),$1,$3);
                  typeConversionOrObjectHandling($1,$3,"*",$$,1);
                } else
                  err("expr null in binary *");
            }
        | expr_noc2 '/' expr_noc2
            { $$=installLowLevelNode(DIV_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"/",$$,1);
            }
        | expr_noc2 '%' expr_noc2
            { $$=installLowLevelNode(MOD_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"%",$$,1);
            }
        | expr_noc2 LSHIFT expr_noc2
            { $$=installLowLevelNode(LSHIFT_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"<<",$$,0))
                NODE_TYPE($$) = NODE_TYPE($1);
            }
        | expr_noc2 RSHIFT expr_noc2
            { $$=installLowLevelNode(RSHIFT_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,">>",$$,0))
                NODE_TYPE($$) = NODE_TYPE($1);
            }
        | expr_noc2 ARITHCOMPARE expr_noc2
            { int op1;

              op1 = ($2 == ((int) LE_EXPR)) ? LE_OP : GE_OP;
              $$=installLowLevelNode(op1,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,map_op($2),$$,0))
                NODE_TYPE($$) = typeConversion(NODE_TYPE($1), 
                                               NODE_TYPE($3));
            }
        | expr_noc2 EQCOMPARE expr_noc2
            { int op1;

              op1 = ($2 == ((int) NE_EXPR)) ? NE_OP : EQ_OP;
                   $$=installLowLevelNode(op1,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,map_op($2),$$,0))
                NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_noc2 '&' expr_noc2
            { $$=installLowLevelNode(BITAND_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"&",$$,0))
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_noc2 '|' expr_noc2
            { $$=installLowLevelNode(BITOR_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"|",$$,0))
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_noc2 '^' expr_noc2
            { $$=installLowLevelNode(XOR_OP,$1,$3,SMNULL);
              typeConversionOrObjectHandling($1,$3,"^",$$,1);
            }
        | expr_noc2 ANDAND expr_noc2
            { $$=installLowLevelNode(AND_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"&&",$$,0))
                NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_noc2 OROR expr_noc2
            { $$=installLowLevelNode(OR_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"||",$$,0))
              NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | expr_noc2 '?' expr_noc2 ':' expr_noc2  /* expr */
            { PTR_LLND ll_node1;

              ll_node1=installLowLevelNode(EXPR_IF_BODY,$3,$5,SMNULL);
              $$=installLowLevelNode(EXPR_IF,$1,ll_node1,SMNULL);
              NODE_TYPE($$) = typeConversion(NODE_TYPE($3),NODE_TYPE($5));
            }
        | expr_noc2 '=' expr_noc2
            { $$=installLowLevelNode(ASSGN_OP,$1,$3,SMNULL);
              if (!typeConversionOrObjectHandling($1,$3,"=",$$,1))
                NODE_TYPE($$)  =NODE_TYPE($1);
            }
        | expr_noc2 ASSIGN expr_noc2
            { int op1;

              op1 = map_assgn_op($2);
              $$=installLowLevelNode(op1,$1,$3,NULL);
              if (!typeConversionOrObjectHandling($1,$3,map_op($2),$$,1))
                NODE_TYPE($$)  =NODE_TYPE($1); 
           }
        ;

/* modified */
primary:
          IDENTIFIER
            { $$= primaryIdentifierAction(&cur_id_entry,&exception_flag, $1); 
	     }

        | OPERATOR LOADEDOPR 
            { 
             $$= primaryIdentifierAction(&cur_id_entry,&exception_flag, $2);
            }  
        | THIS
            { PTR_TYPE type,type1,type2;
              PTR_BFND bf;
              PTR_SYMB symb;
              int status;

              $$ = installLowLevelNode(THIS_NODE,LLNULL,LLNULL,SMNULL);
              bf= find_current_class_bfnd(pred_bfnd,&status);
              if (!bf) err("not proper usage for this");
              symb = BIF_SYMB(bf);
              if (!symb) err("not proper usage for this");
              if (SYMB_CODE(symb)==COLLECTION_NAME) {
                type1 =  (PTR_TYPE) newNode(T_DERIVED_COLLECTION);
                TYPE_COLL_NAME(type1) = symb;
                type2= (PTR_TYPE) newNode(T_DERIVED_TYPE);
                TYPE_SYMB_DERIVE(type2) =  elementtype_symb;
                TYPE_COLL_BASE(type1) = type2;
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_DESCRIP_BASE_TYPE(type) = type1;
                NODE_TYPE($$) = type;
              }
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_DESCRIP_BASE_TYPE(type) = SYMB_TYPE(symb);
                NODE_TYPE($$) = type;
              }
            }

        | newexp
           { $$ = $1}

        | CONSTANT
            { $$ = $1;
              /* Temporarily setting */

              NODE_TYPE($$) = getTypeForConstantNode($1);
            }
        | string
            { PTR_TYPE type;
              $$ = $1;    
              /* Temporarily setting */
              type = (PTR_TYPE) newNode(T_POINTER);
              TYPE_BASE(type) = global_char;
              NODE_TYPE($$) = type;
            }
        | '(' expr ')'
            { primary_flag =EXPR_LR;
              $$ = $2;
            }

        | '(' error ')'
            {}

        | '(' compstmt ')'
            {}

        |   vector_constant       %prec '.'
            {}
        | TFUNCNAME '<' template_arg_list '>' '('  
            { PTR_SYMB symb;
              PTR_LLND ll;
              ll = primaryIdentifierAction(&cur_id_entry,&exception_flag, $1);
              if (exception_flag == ON)
                {
                  /* strange behavior for default function */
                  symb= install_entry_2(cur_id_entry,FUNCTION_NAME,OTHER);
                  SYMB_TYPE(symb) = GetAtomicType(T_INT);
                  exception_flag = OFF;
                  if (ll) {
                    NODE_OPERAND1(ll) = installLowLevelNode(
                                          FUNCTION_REF,LLNULL,LLNULL,symb);
                    $<ll_node>$ = ll;     
                    }                     
                  else  $<ll_node>$ = installLowLevelNode(
                                         FUNCTION_REF,LLNULL,LLNULL,symb);
                    NODE_TYPE($<ll_node>$) = SYMB_TYPE(symb);
                    }                    
                    else
                      $<ll_node>$ = ll;
                }     
            exprlist ')' 
            { PTR_LLND lnode_ptr ,llp;
              int      status;
              
/*            if (primary_flag & EXCEPTION_ON)
                     primary_flag -= EXCEPTION_ON; */
              llp = $<ll_node>6;
              status = OFF;
              if ((NODE_CODE(llp) == FUNC_CALL)
                  && (!NODE_OPERAND0(llp)))
                  {
                     lnode_ptr = llp;
                     status = FUNC_CALL;
		     NODE_OPERAND1(llp) = $3;
                   } 
              if ((!status) &&((NODE_CODE(llp) == RECORD_REF)||
                               (NODE_CODE(llp) == POINTST_OP)||
                               (NODE_CODE(llp) == SCOPE_OP)))
                {              
                  lnode_ptr = NODE_OPERAND1(llp);
                  if ((lnode_ptr)&&(NODE_CODE(lnode_ptr)== FUNCTION_REF))
                    { 
			NODE_CODE(lnode_ptr) = FUNC_CALL; 
			NODE_OPERAND1(llp) = $3;
			}
                  status = FUNC_CALL;
                }   
              if ((!status) &&(NODE_CODE(llp) == FUNCTION_REF))
                { NODE_CODE(llp) = FUNC_CALL;
		  NODE_OPERAND1(llp) = $3;
                  status = FUNC_CALL;
                  lnode_ptr = llp;
                } 
              if (!status) {
                   status = FUNCTION_OP;
                   lnode_ptr = llp;
                 } 

              switch (status) {    
              case FUNCTION_OP :
                            if(NODE_SYMB($<ll_node>6) &&
                               !(SYMB_TYPE(NODE_SYMB($<ll_node>6)))){
                                  PTR_SYMB dums = NODE_SYMB($<ll_node>6);
                                  $$ =installLowLevelNode(               
                                       FUNC_CALL,$7,$3, dums);       
                               }                                  
                               else{                              
                                 $$ =installLowLevelNode(
                                       FUNCTION_OP,$<ll_node>6,$7,SMNULL);
                                 NODE_TYPE($$) = NODE_TYPE($<ll_node>6);  
                                 }                                        
                                 break;                                 
              case FUNC_CALL :   NODE_OPERAND0(lnode_ptr)=$7;
                                 confirmOverloadedFunction(lnode_ptr,$7);
                                 $$ = $<ll_node>6;                       
                                 break;                                  
                default :        err("system error 10");
              }                                         
              typeConversionOrObjectHandling(llp,$7,"()",$$,0);
            }                                                  

        | primary  '('  
            { PTR_SYMB symb;

              if (exception_flag == ON)
                {
                  /* strange behavior for default function */
                  symb= install_entry_2(cur_id_entry,FUNCTION_NAME,OTHER);
                  SYMB_TYPE(symb) = GetAtomicType(T_INT);
                  exception_flag = OFF;
                  if ($1) {
                    NODE_OPERAND1($1) = installLowLevelNode(
                                          FUNCTION_REF,LLNULL,LLNULL,symb);
                    $<ll_node>$ = $1;
                    }
                  else  $<ll_node>$ = installLowLevelNode(
                                         FUNCTION_REF,LLNULL,LLNULL,symb);
                    NODE_TYPE($<ll_node>$) = SYMB_TYPE(symb);
                    }
                    else
                      $<ll_node>$ = $1;
                }

          exprlist ')'   %prec '.'
            { PTR_LLND lnode_ptr ,llp;
              int      status;

/*            if (primary_flag & EXCEPTION_ON) 
                     primary_flag -= EXCEPTION_ON; */
              llp = $<ll_node>3;
              status = OFF;
              if ((NODE_CODE(llp) == FUNC_CALL)
                  && (!NODE_OPERAND0(llp)))
                  { 
                     lnode_ptr = llp;
                     status = FUNC_CALL;
                   }
              if ((!status) &&((NODE_CODE(llp) == RECORD_REF)||
                               (NODE_CODE(llp) == POINTST_OP)||
                               (NODE_CODE(llp) == SCOPE_OP)))
                {
                  lnode_ptr = NODE_OPERAND1(llp);
                  if ((lnode_ptr)&&(NODE_CODE(lnode_ptr)== FUNCTION_REF))
                    { NODE_CODE(lnode_ptr) = FUNC_CALL; }
                  status = FUNC_CALL;
                }
              if ((!status) &&(NODE_CODE(llp) == FUNCTION_REF))
                { NODE_CODE(llp) = FUNC_CALL;
                  status = FUNC_CALL;
                  lnode_ptr = llp;
                }
              if (!status) {
                   status = FUNCTION_OP;
                   lnode_ptr = llp;
                 }
              switch (status) {
              case FUNCTION_OP : 
			    if(NODE_SYMB($<ll_node>3) &&
                               !(SYMB_TYPE(NODE_SYMB($<ll_node>3)))){
			          PTR_SYMB dums = NODE_SYMB($<ll_node>3);
				  $$ =installLowLevelNode(
                                       FUNC_CALL,$4,LLNULL, dums);
			       }
			       else{
			         $$ =installLowLevelNode(
                                       FUNCTION_OP,$<ll_node>3,$4,SMNULL);
                                 NODE_TYPE($$) = NODE_TYPE($<ll_node>3);
				 }
                                 break;
              case FUNC_CALL :   NODE_OPERAND0(lnode_ptr)=$4;
                                 NODE_CODE(lnode_ptr) = FUNC_CALL;   /* BW */
                                 confirmOverloadedFunction(lnode_ptr,$4);
                                 $$ = $<ll_node>3;
                                 break;
                default :        err("system error 10");
              }
              typeConversionOrObjectHandling(llp,$4,"()",$$,0);
            }

        | primary '[' expr_vector ']'   %prec '.'
            { int status;
              PTR_LLND ll_ptr,lp1;

              ll_ptr = checkArrayIdFormat($1,&status);
              if (!ll_ptr)
                {
                  err("syntax error");
                } else
                  {
                    switch (status) {
              case NO : err("syntax error");
                        break;
              case ARRAY_OP_NEED:
                        lp1 = installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
                        $$ = installLowLevelNode(ARRAY_OP,$1,lp1,SMNULL);
                        break;
              case ID_ONLY :
                        NODE_CODE(ll_ptr) = ARRAY_REF;
                        NODE_OPERAND0(ll_ptr) = 
                             installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
                        $$ = $1;
                        break;
              case RANGE_APPEAR :
                        NODE_OPERAND1(ll_ptr) = 
                             installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
                         $$ = $1; 
                        break;
              }
              
              if (!typeConversionOrObjectHandling($1,$3,"[]",$$,0))
                NODE_TYPE($$) = adjustDerefType(NODE_TYPE($1),DEREF_OP);               
                }
            }
        | primary '.' OPERATOR TYPENAME star_ops
	  {   PTR_SYMB symb2;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
              PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;
              symb2= find_type_symbol($4,TYPE_NAME,area());
              name = makeNameString(NULL, SYMB_IDENT(symb2), $5);
              entry = look_up_type(name,&type_stat);
              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,entry,0,0);
              if (sw ==NO){ err(" no such operator"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
             }
        | primary '.'  OPERATOR declmods TYPENAME star_ops
	  {   PTR_SYMB symb2;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
              PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;
              symb2= find_type_symbol($5,TYPE_NAME,area());
              name = makeNameString($4, SYMB_IDENT(symb2), $6);
              entry = look_up_type(name,&type_stat);
              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,entry,0,0);
              if (sw ==NO){ err(" no such operator"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
	  }  
        | primary '.'  OPERATOR basic_type_name star_ops
	  {  PTR_HASH entry;
              int type_stat;
              char *name;
              PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;
              name = makeNameString(NULL, map_rid_to_name($4), $5);
              entry = look_up_type(name,&type_stat);
              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,entry,0,0);
              if (sw ==NO){ err(" no such operator"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
	  }  
        | primary '.'  OPERATOR declmods basic_type_name star_ops
	  {  PTR_HASH entry;
              int type_stat;
              char *name;
              PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;
              name = makeNameString($4, map_rid_to_name($5), $6);
              entry = look_up_type(name,&type_stat);
              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,entry,0,0);
              if (sw ==NO){ err(" no such operator"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
	  }  
        | primary '.' OPERATOR LOADEDOPR 
            { PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;

              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,$4,0,0);
              if (sw ==NO){ err("error operator not overloaded"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
             }
        | primary '.' '~' identifier
	    { char tmp[256],*s1;
              int  status;
              PTR_HASH entry;
	      PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;
              s1 = tmp;
              *s1++ = '~';
              strcpy(s1,SYMB_IDENT($4));
              s1 = copys(tmp);
              entry = look_up_type(s1,&status);
              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,entry,0,0);
              if (sw ==NO){ err(" no such destructor"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
             }
	    
        | primary '.' identifier
            { PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;

              sw = checkFieldDecl(pred_bfnd,&right_symbol,$1,$3,0,0);
              if (sw ==NO){ err("error member"); $$ = LLNULL;}
              else 
                {
                  mytag  = map2(NODE_CODE(right_symbol));
                  l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                               LLNULL,right_symbol);
                  $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(right_symbol);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr2) = NODE_TYPE($$);
                }
             }

        | tp_doublemark identifier     /* BW */
            { 
              $$= installScopeOperator($1,$2,pred_bfnd);
            }

        | DOUBLEMARK identifier
            { PTR_SYMB symbptr;
              PTR_LLND ll;

              symbptr = find_global_symbol($2,VARIABLE_NAME);
              if (! symbptr){  /* default function expected */
                               exception_flag = ON;
                               cur_id_entry = $2;
                               ll= LLNULL;
                             }
              else {
                  switch (SYMB_CODE(symbptr)) {
                  case MEMBER_FUNC :
                  case FUNCTION_NAME:
                       /* primary_flag = EXCEPTION_ON;  */
                       /* above one incorrect while nested */
                       ll = installLowLevelNode(FUNCTION_REF,
                                                LLNULL,LLNULL,symbptr);
                       break;
                  case FIELD_NAME :
                  case VARIABLE_NAME:
                       ll = installLowLevelNode(VAR_REF,
                                                LLNULL,LLNULL,symbptr);
                       break;
                     }
	  
                }
	      /* moved some statements around (phb) */
              $$= installLowLevelNode(SCOPE_OP,LLNULL,ll,SMNULL);
	      NODE_TYPE($$) = adjustArrayType(symbptr);
              if (symbptr) NODE_TYPE($$) = SYMB_TYPE(symbptr);
              else         NODE_TYPE($$) = GetAtomicType(T_INT);
            }
        | primary POINTSAT '~' identifier
	    {char tmp[256],*s1;
              int  status;
              PTR_HASH entry;
	      PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;

              s1 = tmp;
              *s1++ = '~';
              strcpy(s1,SYMB_IDENT($4));
              s1 = copys(tmp);
              entry = look_up_type(s1,&status);
              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,entry);
              }
              else {                     
                if (!$1) err("undefined symbol");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1,entry,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("Internal Error 2");
                  }
            }
        | primary POINTSAT identifier
            { PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;

              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,$3);
              }
              else {                     
                if (!$1) err("undefined symbol");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1, $3,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("error in the -> expression; either right-hand-side is not a member or left-hand-side of -> is incorrect");
                  }
            }
        | primary POINTSAT OPERATOR LOADEDOPR 
            { PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;

              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,$4);
              }
              else {                     
                if (!$1) err("undefined symbol");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1, $4,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("Internal Error 4");
                  }
            }
        | primary POINTSAT OPERATOR TYPENAME star_ops
	  {   PTR_SYMB symb2;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
	      PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;
              symb2= find_type_symbol($4,TYPE_NAME,area());
              name = makeNameString(NULL, SYMB_IDENT(symb2), $5);
              entry = look_up_type(name,&type_stat);
              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,entry);
              }
              else {                     
                if (!$1) err("undefined type operator");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1,entry,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("Internal Error 5");
                  }
             }
        | primary POINTSAT  OPERATOR declmods TYPENAME star_ops
	  {   PTR_SYMB symb2;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
	      PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;
              symb2= find_type_symbol($5,TYPE_NAME,area());
              name = makeNameString($4, SYMB_IDENT(symb2), $6);
              entry = look_up_type(name,&type_stat);
              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,entry);
              }
              else {                     
                if (!$1) err("undefined type operator");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1,entry,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("Internal Error 6");
                  }
	  }  
        | primary POINTSAT  OPERATOR basic_type_name star_ops
	  {  PTR_HASH entry;
              int type_stat;
              char *name;
	      PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;
              name = makeNameString(NULL, map_rid_to_name($4), $5);
              entry = look_up_type(name,&type_stat);
              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,entry);
              }
              else {                     
                if (!$1) err("undefined type operator");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1,entry,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("Internal Error 7");
                  }
	  }  
        | primary POINTSAT  OPERATOR declmods basic_type_name star_ops
	  {  PTR_HASH entry;
              int type_stat;
              char *name;
	      PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1 = NULL;
              int      mytag;
              name = makeNameString($4, map_rid_to_name($5), $6);
              entry = look_up_type(name,&type_stat);
              if ($1 &&(NODE_CODE($1) == THIS_NODE)) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,entry);
              }
              else {                     
                if (!$1) err("undefined type operator");
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1,entry,0,1); 
              }
              if (symb_ptr1)
                {
                  mytag  = map2(SYMB_CODE(symb_ptr1));
                  l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
                  $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
                  NODE_TYPE($$) = adjustArrayType(symb_ptr1);
                  NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                  NODE_TYPE(l_ptr) = NODE_TYPE($$);
                } else
                  {
                    if ($1 && (NODE_CODE($1) == THIS_NODE))
                      err("Internal Err 'This' class/collection missing)");
                    err("Internal Error 8");
                  }
	  }  
        | primary POINTSAT '*' primary
            { 
                  $$ = installLowLevelNode(ARROWSTAR_OP,$1,$4,SMNULL);
                  NODE_TYPE($$) = NODE_TYPE($4); /* not true!! */
            }
        | primary '.' '*' primary
           {
                  $$ = installLowLevelNode(DOTSTAR_OP,$1,$4,SMNULL);
                  NODE_TYPE($$) = NODE_TYPE($4); /* not true!! */
           }                              
        | primary PLUSPLUS
            { char tmpbuf[50];
              $$ = installLowLevelNode(PLUSPLUS_OP,LLNULL,$1,SMNULL);
              if (exception_flag == ON) {
                sprintf(tmpbuf,"Undefined symbol '%s'",
                   HASH_IDENT(cur_id_entry));
		err(tmpbuf);
		exception_flag =OFF;
                }
              NODE_TYPE($$) = NODE_TYPE($1);
            }
        | primary MINUSMINUS
            { char tmpbuf[50];
	      $$ = installLowLevelNode(MINUSMINUS_OP,LLNULL,$1,SMNULL);
	      if (exception_flag == ON) {
                sprintf(tmpbuf,"Undefined symbol '%s'",
                   HASH_IDENT(cur_id_entry));
		err(tmpbuf);
		exception_flag =OFF;
                }
              NODE_TYPE($$) = NODE_TYPE($1);
            }
        ;

new_placement:
         '(' nonnull_exprlist ')'
	 { $$ = $2;}
;

new_init:
	{ $$ = LLNULL; } 
        | '(' exprlist ')'
         { $$ = $2; }
;

/* note:  the form of a NEW_OP is
   NEW_OP  ll-1 = the cast op.  ll-2 = the new_placement expression
  and for the CAST_OP we have
   CAST_OP ll_1 = the new_initializer or value to be cast.
           ll_2 = the type expression which is set only for those
                  types that need it.
*/
newexp:
         NEW  typename  %prec '='
            { PTR_LLND ll_1;
              PTR_TYPE type;
              /* printf("new type 1\n"); */
              ll_1 = installLowLevelNode(CAST_OP,LLNULL,LLNULL,SMNULL);
              NODE_TYPE(ll_1) = $2;
              $$=installLowLevelNode(NEW_OP,ll_1,LLNULL,SMNULL);
              if (TYPE_CODE(SYMB_TYPE(ll_1)) == T_ARRAY)
                  NODE_TYPE($$) = NODE_TYPE(ll_1);
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_BASE(type) = NODE_TYPE(ll_1);
                NODE_TYPE($$) = type;
              }
            }
         | NEW  maintype '(' exprlist ')' %prec '='
            { PTR_TYPE type;
              PTR_LLND ll_1,ll_2; 
              PTR_SYMB scsymb;
              PTR_TYPE scnt;       
              /* printf("new type 2\n"); */

              if ((SYMB_CODE($2) == TYPE_NAME) || 
                  is_symbol_an_object(SYMB_CODE($2)))
                {
                  type = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                  TYPE_SYMB_DERIVE(type) = (PTR_SYMB) $2;
                  if (main_type_flag==TYPE_FIVE) {    
               /*     TYPE_SCOPE_SYMB_DERIVE(type) */

                    scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) $2), Scope_Name);
                    scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                    TYPE_SYMB_DERIVE(scnt) = scsymb;
                    TYPE_SCOPE_SYMB_DERIVE(type) = scnt;

                    main_type_flag = TYPE_TWO;
                   }
                }
              else type = $2;
              ll_1 = makeDeclarExp(SMNULL, type);
              if($4) ll_2 = $4;
              else ll_2 = installLowLevelNode(EXPR_LIST, LLNULL, LLNULL, SMNULL);
              ll_1 = installLowLevelNode(CAST_OP,ll_2,LLNULL,SMNULL);
              NODE_TYPE(ll_1) = type;
              $$=installLowLevelNode(NEW_OP,ll_1,LLNULL,SMNULL);
              if (TYPE_CODE(SYMB_TYPE(ll_1)) == T_ARRAY)
                  NODE_TYPE($$) = NODE_TYPE(ll_1);
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_BASE(type) = NODE_TYPE(ll_1);
                NODE_TYPE($$) = type;
              }
            }
       |  NEW new_placement maintype '(' exprlist ')'
            { PTR_TYPE type;
              PTR_LLND ll_1;
              PTR_SYMB scsymb;
              PTR_TYPE scnt;
              /* printf("new type 2.b\n"); */

              if ((SYMB_CODE($3) == TYPE_NAME) ||
                  is_symbol_an_object(SYMB_CODE($3)))
                { 
                  type = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                  TYPE_SYMB_DERIVE(type) = (PTR_SYMB) $3;
                  if (main_type_flag==TYPE_FIVE) {
                  /*  TYPE_SCOPE_SYMB_DERIVE(type) */        
                    scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) $3), Scope_Name);
                    scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                    TYPE_SYMB_DERIVE(scnt) = scsymb;
                    TYPE_SCOPE_SYMB_DERIVE(type) = scnt;

                    main_type_flag = TYPE_TWO;                                        
                   }                                                                  
                }                             
              else type = $3;
              ll_1 = makeDeclarExp(SMNULL, type);
              ll_1 = installLowLevelNode(CAST_OP,$5,LLNULL,SMNULL);
              NODE_TYPE(ll_1) = type;                              
              $$=installLowLevelNode(NEW_OP,ll_1,$2,SMNULL);   
              if (TYPE_CODE(SYMB_TYPE(ll_1)) == T_ARRAY)
                  NODE_TYPE($$) = NODE_TYPE(ll_1);
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_BASE(type) = NODE_TYPE(ll_1);
                NODE_TYPE($$) = type;
              } 
            }   

       |  NEW new_placement typename new_init %prec '='
            { PTR_LLND ll_1;
              PTR_TYPE type;
              /* printf("new type 3\n"); */
              ll_1 = makeDeclarExp(SMNULL, $3);
              ll_1 = installLowLevelNode(CAST_OP,$4,ll_1,SMNULL);
              NODE_TYPE(ll_1) = $3;
              $$=installLowLevelNode(NEW_OP,ll_1,$2,SMNULL);
              if (TYPE_CODE(SYMB_TYPE(ll_1)) == T_ARRAY)
                  NODE_TYPE($$) = NODE_TYPE(ll_1);
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_BASE(type) = NODE_TYPE(ll_1);
                NODE_TYPE($$) = type;
              }
            }
        | NEW '(' typename ')' new_init %prec HYPERUNARY
            { PTR_TYPE type;
              PTR_LLND ll_1; 
              PTR_SYMB scsymb;
              PTR_TYPE scnt;       
              /* printf("new type 4\n"); */

              if ((SYMB_CODE($3) == TYPE_NAME) || 
                  is_symbol_an_object(SYMB_CODE($3)))
                {
                  type = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                  TYPE_SYMB_DERIVE(type) = (PTR_SYMB) $3;
                  if (main_type_flag==TYPE_FIVE) {    
                   /* TYPE_SCOPE_SYMB_DERIVE(type)*/ 
                    scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) $3), Scope_Name);
                    scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                    TYPE_SYMB_DERIVE(scnt) = scsymb;
                    TYPE_SCOPE_SYMB_DERIVE(type) = scnt;

                    main_type_flag = TYPE_TWO;
                   }
                }
              else type = $3;
              ll_1 = makeDeclarExp(SMNULL, type);
              ll_1 = installLowLevelNode(CAST_OP,$5,ll_1,SMNULL);
              NODE_TYPE(ll_1) = type;
              $$=installLowLevelNode(NEW_OP,ll_1,LLNULL,SMNULL);
              if (TYPE_CODE(SYMB_TYPE(ll_1)) == T_ARRAY)
                NODE_TYPE($$) = NODE_TYPE(ll_1);
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_BASE(type) = NODE_TYPE(ll_1);
                NODE_TYPE($$) = type;
              }
            }
        | NEW new_placement '(' typename ')' new_init %prec HYPERUNARY
            { PTR_TYPE type;
              PTR_LLND ll_1;
              PTR_SYMB scsymb;
              PTR_TYPE scnt;        
              /* printf("new type 5\n"); */

              if ((SYMB_CODE($4) == TYPE_NAME) || 
                  is_symbol_an_object(SYMB_CODE($4)))
                {
                  type = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                  TYPE_SYMB_DERIVE(type) = (PTR_SYMB) $4;
                  if (main_type_flag==TYPE_FIVE) {    
                   /* TYPE_SCOPE_SYMB_DERIVE(type) */ 
                    scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) $4), Scope_Name);
                    scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                    TYPE_SYMB_DERIVE(scnt) = scsymb;
                    TYPE_SCOPE_SYMB_DERIVE(type) = scnt;

                    main_type_flag = TYPE_TWO;
                   }
                }
              else type = $4;
              ll_1 = makeDeclarExp(SMNULL, type);
              ll_1 = installLowLevelNode(CAST_OP,$6,ll_1,SMNULL);
              NODE_TYPE(ll_1) = type;
              $$=installLowLevelNode(NEW_OP,ll_1,$2,SMNULL);
              if (TYPE_CODE(SYMB_TYPE(ll_1)) == T_ARRAY)
                NODE_TYPE($$) = NODE_TYPE(ll_1);
              else {
                type = (PTR_TYPE) newNode(T_POINTER);
                TYPE_BASE(type) = NODE_TYPE(ll_1);
                NODE_TYPE($$) = type;
              }
            }
;


/* In the old days,                                        
   it must appear precede expr for resolve precedence problem 

   But now we reoder it, so it now must appear after expr 
   for resolve precedence problem.
   A nonempty list of identifiers.
*/

identifiers:        
        IDENTIFIER
            { PTR_LLND ll;
              ll= installParameter(&new_cur_symb,$1,VARIABLE_NAME);
              $$= installLowLevelNode(EXPR_LIST,ll,LLNULL,SMNULL);
            }
        | identifiers ',' IDENTIFIER
            { PTR_LLND ll;
              ll = installParameter(&new_cur_symb,$3,VARIABLE_NAME);
              $$=chainParameterList($1,ll);
            }
        ;


/* Produces a STRING_CST with perhaps more STRING_CSTs chained onto it.  */
string:
          STRING
            { $$ = $1; }
        | string STRING
            {char * concstr = 
               (char *)calloc(1, (unsigned)(strlen(NODE_STR($1))+ 
                                            strlen(NODE_STR($2))+ 1));
             concstr=strcpy(concstr, NODE_STR($1));
             concstr=strcat(concstr, NODE_STR($2));
             NODE_STR($1)=concstr;
             $$=$1;
            }
        ;

xdecls:
        /* empty */
           {}
        | decls
           {}
        ;

decls:
        decl
           { main_type_flag = TYPE_FOUR; }
        | errstmt
           {}
        | decls decl
           { main_type_flag = TYPE_FOUR; }
        | decl errstmt
           {}
        ;


decl:
          typed_declspecs  initdecls ';'
            { PTR_SYMB dummy_symb;

              addToDeclaration($1,OFF,$2,&dummy_symb,NO);
              fill_comment();
            }

        | declmods    notype_initdecls ';'
           { PTR_TYPE ptr1;
             PTR_SYMB dummy_symb;

             if($1 != NULL){
               ptr1 = $1;
               TYPE_DESCRIP_BASE_TYPE(ptr1) = GetAtomicType(DEFAULT);
               /* notype_initdecls returns expr_list ll_node */
               addToDeclaration(ptr1,OFF,$2,&dummy_symb,NO);
               }
             else{
               ptr1 = GetAtomicType(DEFAULT);
               addToDeclaration(ptr1,OFF,$2,&dummy_symb,NO);
               }
             fill_comment();
           }

        | typed_declspecs ';'
           { PTR_SYMB symb1 /* ,symb2 */ , name;
             PTR_BFND bifd;

             bifd = lastBifInBlobList(pred_bfnd);
                 fill_comment();
             if (bifd && (BIF_BLOB1(bifd)) &&
                ((BIF_CODE(bifd) == ENUM_DECL)||(!(BIF_SYMB(bifd)))))
               {
                 symb1 =  getFirstFieldOfStruct(bifd);
                 for ( name =symb1; name; name=getClassNextFieldOrMember(name))
                   { /* symb2= */ find_type_symbol(SYMB_PARENT(name),
                                            VARIABLE_NAME,area());
                     /* if ((symb2) && (SYMB_SCOPE(symb2) == pred_bfnd))
                      err("(11) redeclaration symbol"); 
                     else */  SYMB_SCOPE(name) = pred_bfnd;
                   }
                }
          }

        | declmods ';'
            { fill_comment(); }

        /* dbg: i don't know why we have a rule for "private" here 
        | access_specifier typed_declspecs initdecls ';'
            { PTR_TYPE type_ptr2;
              PTR_BFND bf1;
              PTR_SYMB dummy_symb;

              if ($1 != RID_PRIVATE) err("keyword 'private' expected");

              addToDeclaration($2,$1,$3,&dummy_symb,NO);
              fill_comment();
            }
         
        | access_specifier declmods notype_initdecls ';'
            { PTR_TYPE ptr1;
              PTR_LLND lptr2;
              PTR_BFND bf1;
              PTR_SYMB dummy_symb;
  
              if ($1 != RID_PRIVATE) err("keyword 'private' expected");

              if($2 != NULL){
                ptr1 = $2;
                TYPE_DESCRIP_BASE_TYPE(ptr1) = GetAtomicType(DEFAULT);
                lptr2 = addToDeclaration(ptr1,$1,$3,&dummy_symb,NO);
                }
              else{
                ptr1 = GetAtomicType(DEFAULT);
                lptr2 = addToDeclaration(ptr1,$1,$3,&dummy_symb,NO);
                }
              fill_comment();
            }

        | access_specifier typed_declspecs ';'
            { PTR_SYMB symb1,symb2,name;
              PTR_BFND bifd;

              if ($1 != RID_PRIVATE) err("keyword 'private' expected");

              bifd = lastBifInBlobList(pred_bfnd);
              fill_comment();
              if ((BIF_BLOB1(bifd)) &&
                 ((BIF_CODE(bifd) == ENUM_DECL) || (!(BIF_SYMB(bifd)))))
                {
                symb1 =  getFirstFieldOfStruct(bifd);
                for ( name=symb1; name; name=getClassNextFieldOrMember(name) )
                  { symb2=find_type_symbol(SYMB_PARENT(name),
                                           VARIABLE_NAME,area());
                    if ((symb2) && (SYMB_SCOPE(symb2) == pred_bfnd))
                      err_symb("(12) redeclaration symbol ",SYMB_IDENT(symb2));
                    else   SYMB_SCOPE(name) = pred_bfnd;
                  }
                }
            }
        | access_specifier declmods ';'
            { if ($1 != RID_PRIVATE) err("keyword 'private' expected");
              fill_comment();
            }
        */
        ;

/* declspecs which contain at least one type specifier or typedef name.
   (Just `const' or `short' is not enough.)
   A typedef'd name following these is taken as a name to be declared.  */

typed_declspecs:
          maintype typed_declfinal
            {
              PTR_TYPE finalt;
              finalt = resolveFinal($1, &second_part);
              $$= (PTR_TYPE) typed_declspecs_without_declmods(finalt,NULL,0, Scope_Name);
            }

        | declmods maintype typed_declfinal
            {
              PTR_TYPE finalt;
              if($1 != NULL || bif_level_decl_specs){ 
                $$= (PTR_TYPE) typed_declspecs_with_declmods($2,$1,0,
                                                           &second_part, Scope_Name);
                }
              else { 
                finalt = resolveFinal($2, &second_part); 
                $$= (PTR_TYPE) typed_declspecs_without_declmods(finalt,NULL,0, Scope_Name);
              }
            }
        ;


typed_declspecs_2:
          maintype typed_declfinal
            {
/*              $$= (PTR_TYPE) typed_declspecs_without_declmods($1,$2,1); */

              $$= (PTR_TYPE) typed_declspecs_without_declmods($1,NULL,1, Scope_Name);
            }

        | declmods maintype typed_declfinal
            {
/*            $$= typedDeclspecsWithoutDeclmods \
               (pred_bfnd,&passed_bfnd,&new_cur_bfnd,$2,$1,1, &second_part);*/
             if($1 != NULL){
              $$= (PTR_TYPE) typed_declspecs_with_declmods($2,$1,1,
                                                           &second_part, Scope_Name);
              }
             else
              $$= (PTR_TYPE) typed_declspecs_without_declmods($2,NULL,1, Scope_Name);
             
            }
        ;


typed_declfinal:  
          /* empty */
            {}
        | typed_declfinal typespec_reserved
            { installDeclmod(&second_part,$2); }  
        | typed_declfinal storage_class
            { installDeclmod(&second_part,$2); }
        | typed_declfinal storage_class2     /* BW */
            { $$ = $1;}
        
        ;

/* List of just storage classes and type modifiers.
   A declaration can start with this, but then it cannot be used
   to redeclare a typedef-name (unless you use parentheses).  */

declmods:
          type_modifier
            { PTR_TYPE ptr;
              ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              $$ = installDeclmod(ptr, $1);
            }

        | storage_class
            { PTR_TYPE ptr;
              ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              $$ = installDeclmod(ptr, $1);
            }
        | storage_class2 { $$ = NULL;}      /* BW */
            

        /* This rule just duplicates type_modifier now, but it really should
           do something special with CONST and VOLATILE */
        | type_qualifier
            { PTR_TYPE ptr;
              ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              $$ = installDeclmod(ptr, $1);
            }

        | declmods type_modifier
            { $$ = installDeclmod($1,$2); }

        | declmods storage_class
            { $$ = installDeclmod($1,$2); }

        | declmods storage_class2      /* BW */
            { $$ = $1;}

        | declmods type_qualifier
            {$$ = installDeclmod($1,$2); }
        ;

/* Used instead of declspecs where storage classes are not allowed
   (typenames and structure components)
   Don't accept a typedef-name if anything but a modifier precedes it.  */

typed_typespecs:
          maintype typespecfinal

           { PTR_TYPE type_ptr1,type_ptr2;
             PTR_BFND bf2;
             PTR_SYMB gsym;
             PTR_SYMB scsymb;
             PTR_TYPE scnt;

             switch (main_type_flag) {
             case TYPE_ONE : type_ptr1=$1;
                             $$ = type_ptr1;
                             break;
             case TYPE_TWO : type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                             TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) $1;
                             type_ptr1 = type_ptr2;
                             $$ = type_ptr1;
                             break;
             case TYPE_FIVE : type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);    /* BW */
                             TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) $1;
                             /*TYPE_SCOPE_SYMB_DERIVE(type_ptr2) */ 
                             scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) $1), Scope_Name);
                             scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                             TYPE_SYMB_DERIVE(scnt) = scsymb;
                             TYPE_SCOPE_SYMB_DERIVE(type_ptr2) = scnt;

                             main_type_flag = TYPE_TWO;
                             type_ptr1 = type_ptr2;
                             $$ = type_ptr1;
                             break;
             case TYPE_THREE :
                             bf2 = lastBifInBlobList(pred_bfnd);
                             eliminate_lastlink();
                             if ((gsym= BIF_SYMB(bf2)))
                              { type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                                TYPE_SYMB_DERIVE(type_ptr2) = gsym;
                                type_ptr1 = type_ptr2;
                                $$ = type_ptr1;
                              }
                             else {
                                $$= $1;
                              }
                            /* passed_bfnd=new_cur_bfnd = bf2;     */
                            break;
             default :      err("System Crash");   }
           }

        | nonempty_typemods maintype typespecfinal
           { PTR_TYPE type_ptr1,type_ptr2;
             PTR_SYMB gsym;
             PTR_BFND bf2;
             PTR_SYMB scsymb;
             PTR_TYPE scnt;

             switch (main_type_flag) {
             case TYPE_ONE : type_ptr1 = combineDeclmod($1, &second_part);
                             type_ptr1 = $1;
                             TYPE_DESCRIP_BASE_TYPE(type_ptr1) = $2;
                             $$ = type_ptr1;
                             break;
             case TYPE_TWO : type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                             TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) $2;
                             type_ptr1 = combineDeclmod($1,&second_part);
                             TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
                             $$ = type_ptr1;
                             break;
             case TYPE_FIVE : type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                             TYPE_SYMB_DERIVE(type_ptr2) = (PTR_SYMB) $2;
                            /*  TYPE_SCOPE_SYMB_DERIVE(type_ptr2) */ 
                             scsymb  = find_scope_of_scoped_type_symbol(((PTR_SYMB) $2), Scope_Name);
                             scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                             TYPE_SYMB_DERIVE(scnt) = scsymb;
                             TYPE_SCOPE_SYMB_DERIVE(type_ptr2) = scnt;
 
                             main_type_flag = TYPE_TWO;
                             type_ptr1 = combineDeclmod($1,&second_part);
                             TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
                             $$ = type_ptr1;
                             break;
             case TYPE_THREE :
                             type_ptr1 = $1;
                             bf2 = lastBifInBlobList(pred_bfnd);
                             eliminate_lastlink();
                             if ((gsym = BIF_SYMB(bf2))) {
                               type_ptr2 = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                               TYPE_SYMB_DERIVE(type_ptr2) = gsym;
                               TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
                               $$ = type_ptr1;
                               }
                             else {
                               type_ptr2 = $2;
                               /* new_cur_type must be set inside structsp */
                               TYPE_DESCRIP_BASE_TYPE(type_ptr1) = type_ptr2;
                               $$=type_ptr1;
                               }
                            /* passed_bfnd=new_cur_bfnd = bf2;  */
                            break;
             default :      err("System Crash");   }
           }

        ;

typespecfinal:  
          /* empty */
            {}
        | typespecfinal typespec_reserved
            {}
        ;

/* Any kind of typespec.  

typespec: basic_type_name
            {}
        | type_modifier
            {}
        | structsp
            {}
        | TYPENAME
            {}
        ;
*/
/* A typespec that is not a modifier.
   It is never correct to use two of these in one declaration.
   But we try to parse it and report a "semantic" error message later.  */
tp_arglist:
        expr_noc2
          { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); }
        | parm
          { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); }
        | tp_arglist ',' expr_noc2
          {$$ = chainParameterList($1,$3); }
        | tp_arglist ',' parm
          { $$ = chainParameterList($1,$3); }
      ;
maintype: 

          basic_type_name
            { $$= map_rid_to_type($1);
              main_type_flag = TYPE_ONE;
            }


        | structsp
            { $$ = (PTR_TYPE) $1;
              main_type_flag = TYPE_THREE;
            }
      
        | nested_name_spec
            { $$ = $1 }

        | TYPENAME '<'  tp_arglist '>' 
            { PTR_SYMB symb1;
              PTR_TYPE type1;
	      PTR_LLND l;
              symb1= find_object_symbol($1);
              if (!symb1) err("undefined collection or template name");
              if (SYMB_TYPE(symb1) && TYPE_CODE(SYMB_TYPE(symb1)) == T_COLLECTION){
              	$$ = type1 =  (PTR_TYPE) newNode(T_DERIVED_COLLECTION);
              	TYPE_COLL_NAME(type1) = symb1;
		 /* extract first symbol typename */
                l = $3;
		fixTemplateArgs(l,1);
                while(l && NODE_CODE(l) != TYPE_REF) l = NODE_OPERAND0(l);
		if(!l) err("first collection argument must be a class name");
              	TYPE_COLL_BASE(type1) = NODE_TYPE(l);
		}
	      else{
                fixTemplateArgs($3,1); 
              	$$ = type1 =  (PTR_TYPE) newNode(T_DERIVED_TEMPLATE);
		TYPE_TEMPL_NAME(type1) = symb1;
		TYPE_TEMPL_ARGS(type1) = $3;
		}
              main_type_flag = TYPE_ONE;
            }

        ;

nested_name_spec :    /* BW */
         TYPENAME
            { PTR_SYMB s;
              /* PTR_TYPE type_ptr; */ 
              s= find_type_symbol($1,TYPE_NAME,area());
	      if(!s) s = find_scoped_type_symbol($1,Scope_Name);
               $$= (PTR_TYPE) s; 
	      /* printf("s = %d\n", s->variant); */
              main_type_flag = TYPE_TWO;
             /* type_ptr = (PTR_TYPE) newNode(T_DERIVED_TYPE);
              TYPE_SYMB_DERIVE(type_ptr) = s;
              $$ = type_ptr; */
            }

         | tp_doublemark SCOPED_TYPENAME
            { PTR_SYMB s;
              /* PTR_TYPE type_ptr; */ 
              s = find_scoped_type_symbol($2,$1);
              $$ = (PTR_TYPE) s; 
              main_type_flag = TYPE_FIVE;
              /*
              type_ptr = (PTR_TYPE) newNode(T_DERIVED_TYPE);  
              TYPE_SYMB_DERIVE(type_ptr) = s;
              TYPE_SCOPE_SYMB_DERIVE(type_ptr) 
                              = find_scope_of_scoped_type_symbol(((PTR_SYMB) $2), Scope_Name);
              main_type_flag = TYPE_TWO;
              $$ = type_ptr;
              */
            }
            ;

nested_name_spec_1 :    /* BW */
           TYPENAME temp_arg_spec
           { PTR_TYPE type_ptr;

             PTR_SYMB  symb, s;
             PTR_LLND  llnd;

             PTR_HASH entry;
             int status;

             s= find_type_symbol($1,TYPE_NAME,area());
	     if(!s) s = find_scoped_type_symbol($1,Scope_Name);
             main_type_flag = TYPE_TWO;

             if($2 != NULL){
               type_ptr = (PTR_TYPE) newNode(T_DERIVED_TEMPLATE);
               TYPE_TEMPL_NAME(type_ptr) = s;
               TYPE_TEMPL_ARGS(type_ptr) = $2;
               }
             else {
               type_ptr = (PTR_TYPE) newNode(T_DERIVED_TYPE);
               TYPE_SYMB_DERIVE(type_ptr) = s;
               }

             entry = look_up_type("",&status);
             symb=install_entry_2(entry,VARIABLE_NAME,OTHER);
             SYMB_TYPE(symb) = type_ptr;
             llnd = installLowLevelNode(VAR_REF,LLNULL,LLNULL, symb);
             NODE_TYPE(llnd) = type_ptr;
             $$ = llnd;
           }
        | TYPENAME temp_arg_spec DOUBLEMARK 
           { 
             Scope_Flag = 1;
             Scope_Name = $1;
           }

          SCOPED_TYPENAME temp_arg_spec
           {PTR_TYPE type_ptr;

             PTR_SYMB  symb, s;
             PTR_LLND  llnd;

             PTR_HASH entry;
             int status;
             PTR_SYMB scsymb;
             PTR_TYPE scnt;

             member_templ_parms = $2;    /* what does that do?  BW*/
             if($2 != NULL) fixTemplateArgs($2,1); /* this might break */
             s = find_scoped_type_symbol($5,$1);
             main_type_flag = TYPE_TWO;
             if($6 != NULL){
                type_ptr = (PTR_TYPE) newNode(T_DERIVED_TEMPLATE);  
                TYPE_TEMPL_NAME(type_ptr) = (PTR_SYMB) s;
                TYPE_TEMPL_ARGS(type_ptr) = $6;
                }
             else{
                type_ptr = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                TYPE_SYMB_DERIVE(type_ptr) = s;
               /*  TYPE_SCOPE_SYMB_DERIVE(type_ptr) */ 
                scsymb = find_scope_of_scoped_type_symbol(((PTR_SYMB) s), $1);
                scnt = (PTR_TYPE) newNode(T_DERIVED_TYPE);
                TYPE_SYMB_DERIVE(scnt) = scsymb;
                TYPE_SCOPE_SYMB_DERIVE(type_ptr) = scnt;

                }
             entry = look_up_type("",&status);
             symb=install_entry_2(entry,VARIABLE_NAME,OTHER);
             SYMB_TYPE(symb) = type_ptr;
             llnd = installLowLevelNode(VAR_REF,LLNULL,LLNULL, symb);
             NODE_TYPE(llnd) = type_ptr;
             $$ = llnd;
           }
           ;

/* A typespec that is a reserved word.
typespec_reserved: basic_type_name
            {}
        | type_modifier
            {}
        | structsp
            {}
        ;
*/

typespec_reserved: 
          type_modifier  
            { $$ = $1; }
         | type_qualifier
            { $$ = $1; }
        ;


initdecls:
          initdcl
            { PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              NODE_OPERAND0(exprlist_llnd) = $1;
              $$ = exprlist_llnd;
              main_type_flag = TYPE_CLEAN;
            }
        | initdecls ',' initdcl
            { PTR_LLND ll_node1;
              PTR_BFND bifd;
              PTR_LLND exprlist_llnd;
              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              ll_node1 = Follow_Llnd(exprlist_llnd,2);
              NODE_OPERAND0(ll_node1) = $3;
              $$ = exprlist_llnd;
              main_type_flag = TYPE_CLEAN; /* BW */
            }
        ;


notype_initdecls:
          notype_initdcl
            { PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              NODE_OPERAND0(exprlist_llnd) = $1;
              $$ = exprlist_llnd;
            }

        | notype_initdecls ',' initdcl
            { PTR_LLND ll_node1;
              PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              ll_node1 = Follow_Llnd(exprlist_llnd,2);
              NODE_OPERAND0(ll_node1) = $3;
              $$ = exprlist_llnd;
            }
        ;


initdcl:
          declarator  '=' init
/* Note how the declaration of the variable is in effect 
   while its init is parsed! */
            { $$ = installLowLevelNode(ASSGN_OP,$1,$3,SMNULL); }
        | declarator '=' error
            {}
        | declarator
            { $$=$1; }
        ;


notype_initdcl:
          notype_declarator '=' init
            { $$ = installLowLevelNode(ASSGN_OP,$1,$3,SMNULL); }

        | notype_declarator '=' error
            {}
        | notype_declarator
            { $$=$1; }
        ;



/* modified */
init:
          expr_no_commas
            { $$=$1; }

        | '{' initlist '}'
            { $$= installLowLevelNode(INIT_LIST,$2,LLNULL,SMNULL); }

        | '{' initlist ',' '}'
            { $$= installLowLevelNode(INIT_LIST,$2,LLNULL,SMNULL); }
        ;

/* This chain is built in reverse order,
   and put in forward order where initlist is used.  */
/* modified */
initlist:
          inside
            { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); }
        | initlist ',' inside
            { PTR_LLND ll_node1;

              ll_node1 = Follow_Llnd($1,2);
              NODE_OPERAND1(ll_node1) = 
                   installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
              $$=$1;
            }

        ;


/* modified */
inside: 
         init
           { $$=$1; }
        | array_expr_a
           { $$=$1; }
        | array_expr_b
           { $$=$1; }
       ;


/* Any kind of declarator (thus, all declarators allowed
   after an explicit typespec).  */

declarator:
          notype_declarator
            { PTR_SYMB symb;
              $$= $1;     
              symb = getCurrentProcName(pred_bfnd);
              redeclCheck(symb);
              if (SYMB_DECL(symb))  {
                checkExternFirstPart(symb);   
              }
            }
        | after_type_declarator
            { PTR_SYMB symb;
              $$= $1;     
              symb = getCurrentProcName(pred_bfnd);
              if (SYMB_DECL(symb))  {
                checkExternFirstPart(symb);
              }
            }
         ;

after_type_declarator:
          after_type_declarator '(' expr ')'      %prec '.'
            { PTR_LLND ll;

              ll = $1;
              if ((cur_flag-RULE_ID) && (cur_flag-RULE_ID-RULE_ARRAY))
                   err("syntax error 3");
              else {
                $$ = installLowLevelNode(CLASSINIT_OP,ll,$3,SMNULL);
                cur_flag = cur_flag | RULE_CLASSINIT;
              }
            }

        | after_type_declarator '(' parmlist ')'  %prec '.'
            { PTR_LLND ll; 
              PTR_SYMB symb;

              symb = getCurrentProcName(pred_bfnd);
              ll = $1;
              if ((NODE_CODE(ll) == VAR_REF) 
                  || (NODE_CODE(ll) == RECORD_REF)
                  || (NODE_CODE(ll) == FUNCTION_REF))                 
                setSymbNodeToFunction(symb);
              $$ = setFuncAfterDeclaratorAppear(
                             pred_bfnd,&cur_flag,ll,$3,symb);
            }

        | after_type_declarator '(' identifiers ')'  %prec '.'
            { PTR_LLND ll;
              PTR_SYMB symb;

              ll = $1;
              symb = getCurrentProcName(pred_bfnd);
              if ((NODE_CODE(ll) == VAR_REF) 
                  || (NODE_CODE(ll) == RECORD_REF) 
                  || (NODE_CODE(ll) == FUNCTION_REF))                 
                setSymbNodeToFunction(symb);
              $$ = setFuncAfterDeclaratorAppear(
                             pred_bfnd,&cur_flag,ll,$3,symb);
            }

        | after_type_declarator '(' error ')'  %prec '.'
            {}

        | '*' typemods after_type_declarator  %prec UNARY
            { PTR_TYPE btype, type /* ,type_1 */ ;
              PTR_LLND ll;
              PTR_SYMB symbptr1;

              ll=$3;
              $$ = installLowLevelNode(DEREF_OP,ll,LLNULL,SMNULL);
              btype = (PTR_TYPE) newNode(T_POINTER);
              if($2){
                 btype->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
                 NODE_TYPE($$) = btype;
                  }

              symbptr1 = getCurrentProcName(pred_bfnd);
              if (! SYMB_TYPE(symbptr1))
                  type = SYMB_TYPE(symbptr1)= btype;
              else  /* no type checking here */
                { type = lookForInternalBasetype(SYMB_TYPE(symbptr1));
                  switch (TYPE_CODE(type)) {
                    case T_ARRAY :
                         /*  type_1= */ TYPE_DECL_BASE(type)= btype;
                          break;
                    case T_FUNCTION:
                    case T_POINTER :
                    case T_REFERENCE:
                          /* type_1 = */ TYPE_BASE(type) =  btype;
                          break;
                    default : err("System Trouble 1");
                    }
		}
            }
        | '&' typemods after_type_declarator  %prec UNARY
            { PTR_TYPE btype, type /* ,type_1 */ ;
              PTR_LLND ll;
              PTR_SYMB symbptr1;

              ll=$3;
              $$ = installLowLevelNode(ADDRESS_OP,ll,LLNULL,SMNULL);
              symbptr1 = getCurrentProcName(pred_bfnd);
              btype = (PTR_TYPE) newNode(T_REFERENCE);
              if($2){
                 btype->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
                 NODE_TYPE($$) = btype;
                  }

              if (! SYMB_TYPE(symbptr1)) 
                   type = SYMB_TYPE(symbptr1) = btype;
              else  /* no type checking here */
                { type = lookForInternalBasetype(SYMB_TYPE(symbptr1));
                  switch (TYPE_CODE(type)) {
                  case T_ARRAY :
                       /* type_1 = */ TYPE_DECL_BASE(type) = btype;
                        break;
                  case T_FUNCTION:
                  case T_POINTER :
                  case T_REFERENCE:
                       /*  type_1 = */ TYPE_BASE(type) = btype;
                        break;
                  default : err("System Trouble 2");
                  }
                }
            }

        | after_type_declarator '[' expr ']'  %prec '.'
            { $$= addRangeToArray(&new_cur_type,$1,$3);
              cur_flag = cur_flag | RULE_ARRAY;
            }

        | after_type_declarator '[' ']'  %prec '.'
            { 
              $$= addRangeToArray(&new_cur_type,$1,NULL);
              cur_flag = cur_flag | RULE_ARRAY;
            }

        | TYPENAME 
            { $$=lazyInstall(pred_bfnd,$1,VARIABLE_NAME,NO); } 
        | '~' TYPENAME 
            { char tmp[256],*s1;
              int  status;
              PTR_HASH entry;

              s1 = tmp;
              *s1++ = '~';
              strcpy(s1,SYMB_IDENT($2));
              s1 = copys(tmp);
              entry = look_up_type(s1,&status);
              $$=lazyInstall(pred_bfnd,entry,VARIABLE_NAME,NO);
             }
        ;


/* A declarator allowed whether or not there has been
   an explicit typespec.  These cannot redeclare a typedef-name.  */
 
/* modified */
notype_declarator:

          notype_declarator '(' expr ')'      %prec '.'
            { PTR_LLND ll;

              ll = $1;
              if ((cur_flag - RULE_ID) && 
                  (cur_flag - RULE_ID - RULE_ARRAY))
                      err("syntax error 3");
              else {
                   $$ = installLowLevelNode(CLASSINIT_OP,ll,$3,SMNULL);
                   cur_flag = cur_flag | RULE_CLASSINIT;
                   }
            }

        | notype_declarator '(' parmlist ')'  %prec '.'
            { PTR_LLND ll; 
              PTR_SYMB symb;
 
              symb = getCurrentProcName(pred_bfnd);
              ll = $1;
              if ((NODE_CODE(ll) == VAR_REF)
                  || (NODE_CODE(ll) == RECORD_REF)
                  || (NODE_CODE(ll) == FUNCTION_REF))                 
                setSymbNodeToFunction(symb);
              $$ = setFuncAfterDeclaratorAppear(pred_bfnd,
                                                &cur_flag,ll,$3,symb);
            }

        /* Added (PHB) to accept type_qualifier, does not pass info to .dep */
        | notype_declarator '(' parmlist ')' type_qualifier %prec '.'
            { PTR_LLND ll; 
              PTR_SYMB symb;

              symb = getCurrentProcName(pred_bfnd);
              ll = $1;
              if ((NODE_CODE(ll) == VAR_REF)
                  || (NODE_CODE(ll) == RECORD_REF)
                  || (NODE_CODE(ll) == FUNCTION_REF))                 
                setSymbNodeToFunction(symb);
              SYMB_TEMPLATE_DUMMY8(symb) |=  64; /* the key for const */
              $$ = setFuncAfterDeclaratorAppear(pred_bfnd,
                                                &cur_flag,ll,$3,symb);
            }

        | notype_declarator '(' identifiers ')'  %prec '.'
            { PTR_LLND ll;
              PTR_SYMB symb;

              ll = $1;
              symb = getCurrentProcName(pred_bfnd);
              if ((NODE_CODE(ll) == VAR_REF)
                  || (NODE_CODE(ll) == RECORD_REF)
                  || (NODE_CODE(ll) == FUNCTION_REF))                 
                setSymbNodeToFunction(symb);
              $$ = setFuncAfterDeclaratorAppear(pred_bfnd,
                                                &cur_flag,ll,$3,symb);
            }

        | notype_declarator '(' error ')'  %prec '.'
            {}

        | '(' notype_declarator ')'
            { 
              cur_flag = CLEAN;
              cur_flag = RULE_LR;
              $$ = $2;
            }
        | '*'  typemods notype_declarator  %prec UNARY
            { PTR_TYPE type /* ,type_1 */ , btype;
              PTR_LLND ll, l1;
              PTR_SYMB symbptr1;

              ll=$3;
              l1 = installLowLevelNode(DEREF_OP,ll,LLNULL,SMNULL);
	      $$ = l1;
              symbptr1 = getCurrentProcName(pred_bfnd);
              btype = (PTR_TYPE) newNode(T_POINTER);
	      if($2){ 
		 btype->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
		 NODE_TYPE(l1) = btype; 
		  }
              if (symbptr1 && !SYMB_TYPE(symbptr1)) 
                 type = SYMB_TYPE(symbptr1) = btype;
              else  /* no type checking here */
                { type = lookForInternalBasetype(SYMB_TYPE(symbptr1));
                  switch (TYPE_CODE(type)) {
                  case T_ARRAY :
                        /* type_1 = */ TYPE_DECL_BASE(type) = btype;
                        break;
                  case T_FUNCTION:
                  case T_POINTER :
                  case T_REFERENCE:
                       /*  type_1 = */ TYPE_BASE(type) = btype;
                        break;
                  default: 
                     /* removed to allow member_pointers */
                     /* err("System Trouble 3"); */
                        break;
                  }
                }
            }
        | '&'  typemods notype_declarator  %prec UNARY
            { PTR_TYPE btype, type /* ,type_1 */;
              PTR_LLND ll;
              PTR_SYMB symbptr1;

               ll=$3;
               $$ = installLowLevelNode(ADDRESS_OP,ll,LLNULL,SMNULL);
               symbptr1 = getCurrentProcName(pred_bfnd);
               btype = (PTR_TYPE) newNode(T_REFERENCE);
               if($2){
                  btype->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
                  NODE_TYPE($$) = btype; 
                  }                   

               if (! SYMB_TYPE(symbptr1)) 
                   type = SYMB_TYPE(symbptr1) = btype;
               else  /* no type checking here */
                 { type = lookForInternalBasetype(SYMB_TYPE(symbptr1));
                   switch (TYPE_CODE(type)) {
                   case T_ARRAY :
                         /* type_1 = */ TYPE_DECL_BASE(type) = btype;
                         break;
                   case T_FUNCTION:
                   case T_POINTER :
                   case T_REFERENCE:
                         /* type_1 = */ TYPE_BASE(type) = btype;
                         break;
                   default :   /* err("System Trouble 4") ; */
                         break;
                   }
                 }
             }

        | notype_declarator '[' expr_no_commas ']'  %prec '.'
            { 
              $$=addRangeToArray(&new_cur_type,$1,$3);
              cur_flag = cur_flag | RULE_ARRAY;
            }
        | notype_declarator '[' ']'  %prec '.'
            { 
              $$=addRangeToArray(&new_cur_type,$1,NULL);
              cur_flag = cur_flag | RULE_ARRAY;
            }
     
        | IDENTIFIER
            { $$=lazyInstall(pred_bfnd,$1,VARIABLE_NAME,NO); }

        | '~' IDENTIFIER
            { char tmp[256],*s1;
              int  status;
              PTR_HASH entry;

              s1 = tmp;
              *s1++ = '~';
              strcpy(s1,SYMB_IDENT($2));
              s1 = copys(tmp);
              entry = look_up_type(s1,&status);
              $$=lazyInstall(pred_bfnd,entry,VARIABLE_NAME,NO);
             }

        | OPERATOR LOADEDOPR  
            { 
              $$ = lazyInstall(pred_bfnd,$2,FUNCTION_NAME,NO);
            }
        | multiple_id
            { $$=$1; }

        | OPERATOR NEW
            { int status;
              PTR_HASH entry;
              entry = look_up_type("new",&status);
              $$=lazyInstall(pred_bfnd,entry,FUNCTION_NAME,NO);
            }
        | NEW
            { int status;
              PTR_HASH entry;
              entry = look_up_type("new",&status);
              $$=lazyInstall(pred_bfnd,entry,FUNCTION_NAME,NO);
            }
        | OPERATOR DELETE
            { int status;
              PTR_HASH entry;
              entry = look_up_type("delete",&status);
              $$=lazyInstall(pred_bfnd,entry,FUNCTION_NAME,NO);
            }
        | DELETE
            { int status;
              PTR_HASH entry;
              entry = look_up_type("delete",&status);
              $$=lazyInstall(pred_bfnd,entry,FUNCTION_NAME,NO);
            }
        ;

/* '::'  DOUBLEMARK */

multiple_id :
          identifier DOUBLEMARK  identifier
            { PTR_SYMB symb,object;
	      char tmpbuf[50];
              Scope_Name = $1;
              object = find_object_symbol($1);
              if (!object) {
                sprintf(tmpbuf,"Undefined class/struct '%s'",
                   HASH_IDENT($1));
		err(tmpbuf);
	      }
	      if(object == BIF_SYMB(pred_bfnd)){
                  $$=lazyInstall(pred_bfnd,$3,VARIABLE_NAME,NO);
		  }
              else{
                /* note: to do static member function init B::a = 0
                   we have to change this.  However if you do, there
                   are lots of other problems.  dbg.
                 */
                 symb= veryWeakCheckTypeForField(object,$3,FUNCTION_ONLY); 
	         if (!symb) {
                   sprintf(tmpbuf,"Undefined field '%s'",
                      HASH_IDENT($3));
		    err(tmpbuf);
	          }
                 SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
                 SYMB_TYPE(symb)     = (PTR_TYPE) NULL; 
                 SYMB_DECL(symb)     = SYMB_CODE(symb);
                 $$ = derivedLazyInstall(symb);
		}
            }
         | identifier DOUBLEMARK '*'  notype_declarator
	     { PTR_TYPE btype, type;
               PTR_LLND ll,rhs, lhs;
               PTR_SYMB classname;
               PTR_SYMB symbptr1;
               ll = $4;
               Scope_Name = $1;
               classname = find_object_symbol($1);
               if(classname == NULL) err("could not find class name!");
               lhs = installLowLevelNode(VAR_REF,LLNULL, LLNULL, classname);
	       rhs = installLowLevelNode(DEREF_OP,ll, LLNULL, SMNULL);
               $$ = installLowLevelNode(SCOPE_OP,lhs,rhs,SMNULL);
               symbptr1 = getCurrentProcName(pred_bfnd);
               btype = (PTR_TYPE) newNode(T_MEMBER_POINTER);
	       TYPE_COLL_NAME(btype) = classname;
	       TYPE_COLL_BASE(btype) = NULL;
               if (! SYMB_TYPE(symbptr1)) 
                   type = SYMB_TYPE(symbptr1) = btype;
               else { 
	          type = lookForInternalBasetype(SYMB_TYPE(symbptr1));	      	
                  type = (TYPE_COLL_BASE(btype) = type);
		  SYMB_TYPE(symbptr1) = type;
                  NODE_TYPE($$) = type;
                  }
           }
        | identifier DOUBLEMARK  OPERATOR TYPENAME star_ops
	  {   PTR_SYMB symb, symb2, object;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
              Scope_Name = $1;
              symb2= find_type_symbol($4,TYPE_NAME,area());
              name = makeNameString(NULL, SYMB_IDENT(symb2), $5);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | identifier DOUBLEMARK  OPERATOR declmods TYPENAME star_ops
	  {   PTR_SYMB symb, symb2, object;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
              Scope_Name = $1;
              symb2= find_type_symbol($5,TYPE_NAME,area());
              name = makeNameString($4, SYMB_IDENT(symb2), $6);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | identifier DOUBLEMARK  OPERATOR basic_type_name star_ops
	  {  PTR_HASH entry;
              PTR_SYMB symb,object;
              int type_stat;
              char *name;
              Scope_Name = $1;
              name = makeNameString(NULL, map_rid_to_name($4), $5);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | identifier DOUBLEMARK  OPERATOR declmods basic_type_name star_ops
	  {  PTR_HASH entry;
              PTR_SYMB symb,object;
              int type_stat;
              char *name;
              Scope_Name = $1;
              name = makeNameString($4, map_rid_to_name($5), $6);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | identifier DOUBLEMARK  OPERATOR LOADEDOPR  
            { PTR_SYMB symb,object;

               Scope_Name = $1;
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,$4,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }

        | identifier DOUBLEMARK  OPERATOR NEW
            { int status;
              PTR_HASH entry;
              PTR_SYMB symb,object;

              Scope_Name = $1;
              entry = look_up_type("new",&status);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }

        | identifier DOUBLEMARK OPERATOR DELETE
            { int status;
              PTR_HASH entry;
              PTR_SYMB symb,object;

              Scope_Name = $1;
              entry = look_up_type("delete",&status);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }

        | identifier DOUBLEMARK '~' identifier
            { char tmp[256],*s1;
              int  status;
              PTR_HASH entry;
              PTR_SYMB symb,object;

              Scope_Name = $1;
              s1 = tmp;
              *s1++ = '~';
              strcpy(s1,SYMB_IDENT($4));
              s1 = copys(tmp);
              entry = look_up_type(s1,&status);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }
 
        |  tp_doublemark identifier
            { PTR_SYMB symb,object;
	      char tmpbuf[50];
              object = find_object_symbol($1);
              if (!object) {
               sprintf(tmpbuf,"Undefined class/struct  '%s'",
                    HASH_IDENT($1));
		err(tmpbuf);
	      }
	      if(object == BIF_SYMB(pred_bfnd)){
                  $$=lazyInstall(pred_bfnd,$2,VARIABLE_NAME,NO);
		  }
              else{
                /* note: to do static member function init B::a = 0
                   we have to change this.  However if you do, there
                   are lots of other problems.  dbg.
                 */
              symb= veryWeakCheckTypeForField(object,$2,FUNCTION_ONLY);  
	         if (!symb) {
                   sprintf(tmpbuf,"Undefined field '%s'",
                      HASH_IDENT($2));
		    err(tmpbuf);
	          }
                 SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
                 SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
                 SYMB_DECL(symb)     = SYMB_CODE(symb);
                 $$ = derivedLazyInstall(symb);
		}
            }
         | tp_doublemark '*'  notype_declarator
	     { PTR_TYPE btype, type;
               PTR_LLND ll,rhs, lhs;
               PTR_SYMB classname;
               PTR_SYMB symbptr1;
               ll = $3;
               classname = find_object_symbol($1);
               if(classname == NULL) err("could not find class name!");
               lhs = installLowLevelNode(VAR_REF,LLNULL, LLNULL, classname);
	       rhs = installLowLevelNode(DEREF_OP,ll, LLNULL, SMNULL);
               $$ = installLowLevelNode(SCOPE_OP,lhs,rhs,SMNULL);
               symbptr1 = getCurrentProcName(pred_bfnd);
               btype = (PTR_TYPE) newNode(T_MEMBER_POINTER);
	       TYPE_COLL_NAME(btype) = classname;
	       TYPE_COLL_BASE(btype) = NULL;
               if (! SYMB_TYPE(symbptr1)){ 
                   type = SYMB_TYPE(symbptr1) = btype;
		   }
               else { 
	          type = lookForInternalBasetype(SYMB_TYPE(symbptr1));	      	
                  type = (TYPE_COLL_BASE(btype) = type);
		  SYMB_TYPE(symbptr1) = type;
                  NODE_TYPE($$) = type;
                  }
           }
        | tp_doublemark OPERATOR TYPENAME star_ops
	  {   PTR_SYMB symb, symb2, object;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
              symb2= find_type_symbol($3,TYPE_NAME,area());
              name = makeNameString(NULL, SYMB_IDENT(symb2), $4);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | tp_doublemark OPERATOR declmods TYPENAME star_ops
	  {   PTR_SYMB symb, symb2, object;
	      PTR_HASH entry;
	      int type_stat;
              char *name;
              symb2= find_type_symbol($4,TYPE_NAME,area());
              name = makeNameString($3, SYMB_IDENT(symb2), $5);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | tp_doublemark  OPERATOR basic_type_name star_ops
	  {  PTR_HASH entry;
              PTR_SYMB symb,object;
              int type_stat;
              char *name;
              name = makeNameString(NULL, map_rid_to_name($3), $4);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | tp_doublemark OPERATOR declmods basic_type_name star_ops
	  {  PTR_HASH entry;
              PTR_SYMB symb,object;
              int type_stat;
              char *name;
              name = makeNameString($3, map_rid_to_name($4), $5);
              entry = look_up_type(name,&type_stat);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              if(!symb) err("unknown operator name");
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
	  }  
        | tp_doublemark  OPERATOR LOADEDOPR  
            { PTR_SYMB symb,object;

              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,$3,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }

        | tp_doublemark OPERATOR NEW
            { int status;
              PTR_HASH entry;
              PTR_SYMB symb,object;

              entry = look_up_type("new",&status);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }

        | tp_doublemark OPERATOR DELETE
            { int status;
              PTR_HASH entry;
              PTR_SYMB symb,object;

              entry = look_up_type("delete",&status);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }

        | tp_doublemark '~' identifier
            { char tmp[256],*s1;
              int  status;
              PTR_HASH entry;
              PTR_SYMB symb,object;

              s1 = tmp;
              *s1++ = '~';
              strcpy(s1,SYMB_IDENT($3));
              s1 = copys(tmp);
              entry = look_up_type(s1,&status);
              object = find_object_symbol($1);
              if (!object) err("unknown class/struct name");
              symb= veryWeakCheckTypeForField(object,entry,FUNCTION_ONLY);
              SYMB_UD_CHAIN(symb) = (PTR_BLOB) SYMB_TYPE(symb);
              SYMB_TYPE(symb)     = (PTR_TYPE) NULL;
              SYMB_DECL(symb)     = SYMB_CODE(symb);
              $$ = derivedLazyInstall(symb);
            }
        ;
/* BW */
tp_doublemark:
          TYPENAME DOUBLEMARK
            {  Scope_Name = $1;
               Scope_Flag = 1;
               member_templ_parms = NULL;
               $$ = $1;
            }
	  | TYPENAME '<' tp_arglist '>' DOUBLEMARK
            { PTR_SYMB symb1;
              symb1= find_object_symbol($1);
              if (!symb1) err("undefined collection or template name");
	      fixTemplateArgs($3,1);
              member_templ_parms = $3;
              Scope_Name = $1;
              Scope_Flag = 1;
	      $$ = $1;
            }
        ;

/* modified */
structsp: 
          UNION identifier
            { PTR_SYMB symb1;
              symb1 = find_type_symbol($2,UNION_NAME,area());
              if ((!symb1) ||
                 /* (!same_scope(SYMB_SCOPE(symb1),pred_bfnd)) */ /* BW */
                 (SYMB_SCOPE(symb1) != pred_bfnd))
                { symb1 = install_entry_2($2,UNION_NAME,YES); }
              else 
                { if (/* same_scope(SYMB_SCOPE(symb1),pred_bfnd) && */ /* BW */ 
                      (SYMB_TEMPLATE_DUMMY1(symb1) == BEEN_SEEN))
                   err("re_declaration symbol");
                }
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               UNION_DECL,symb1,LLNULL);
              }
          '{' component_decl_list '}'
            { 
              $$ = installStruct(&pred_bfnd,&new_cur_type,
                                        T_UNION,SMNULL);
            }
        | UNION 
            { char buffer[200];
              PTR_HASH entry;
              PTR_SYMB symb1;
              int status;
                 
	      /*sprintf(buffer,"",enum_id++);*/
	      buffer[0]=0;
              entry = look_up_type(buffer,&status);
              symb1 = install_entry_2(entry,UNION_NAME,YES);
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               UNION_DECL,symb1,LLNULL); 
            }

          '{' component_decl_list '}'
            { $$=installStruct(&pred_bfnd,&new_cur_type,T_UNION,SMNULL); }
        | UNION identifier
            { PTR_SYMB symb1;

              symb1 = find_type_symbol($2,UNION_NAME,area());
              if (!symb1) {
                 symb1 = install_entry_2($2,UNION_NAME,UP_TO_NODECL);
                 SYMB_TYPE(symb1) = (PTR_TYPE)NULL;
                 SYMB_TEMPLATE_DUMMY1(symb1) = NOT_SEEN;
                 }
              $$ = setBfndForStruct(pred_bfnd,&cur_flag,UNION_DECL,symb1);
            }
        /*  example of how a collection can be like a template 
        | COLLECTION '<' parmlist '>' CLASS identifier
            { PTR_SYMB symb1;
              symb1 = find_type_symbol($6,COLLECTION_NAME,area());
              if ((!symb1) ||(!same_scope(SYMB_SCOPE(symb1),pred_bfnd)))
                {
                  symb1 = install_entry_2($6,COLLECTION_NAME,
                                          YES);
                }
              else 
                { if (same_scope(SYMB_SCOPE(symb1),pred_bfnd)&& 
                      (SYMB_TEMPLATE_DUMMY1(symb1) == BEEN_SEEN))
                   err("re_declaration symbol");
                }
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               COLLECTION_DECL,symb1,LLNULL);
              BIF_LL3(pred_bfnd) = NULL; 
	      BIF_LL1(pred_bfnd) = $3;
            }
          '{' component_decl_list '}'
            { PTR_SYMB symb;
              PTR_TYPE type;
              PTR_BLOB blob;
              PTR_BFND bf1;

              type = installStruct(&pred_bfnd,&new_cur_type,
                                       T_COLLECTION,SMNULL);
              $$=type;
              bf1 =  lastBifInBlobList(pred_bfnd);
              BIF_PROPLIST(bf1) = (PTR_PLNK) NULL;
            }
        */
        | general_classname identifier ':' TYPENAME_list
            { PTR_SYMB symb1;
              symb1 = find_type_symbol($2,name_of_struct($1),area());
              if ((!symb1) || 
                  /* (!same_scope(SYMB_SCOPE(symb1),pred_bfnd))*/ /* BW */
                 (SYMB_SCOPE(symb1) != pred_bfnd))
                {
                symb1 = install_entry_2($2,name_of_struct($1),
                                           YES);
                }
              else 
                { if (/* same_scope(SYMB_SCOPE(symb1),pred_bfnd) && */ /* BW */ 
                     (SYMB_TEMPLATE_DUMMY1(symb1) == BEEN_SEEN))
                   err("re_declaration symbol");
                }
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               node_of_struct($1),symb1,$4);           
              BIF_LL2(pred_bfnd) = $4;
              BIF_LL3(pred_bfnd) = NULL;  /* was $3: OBSOLETE, I removed
                                             optionlist */
            }

          '{' component_decl_list  '}'
            { PTR_BFND bf1;
              PTR_TYPE type;

              type = installStruct(&pred_bfnd,&new_cur_type,
                                   type_of_struct($1),$4);
              $$ = type;
              bf1 = lastBifInBlobList(pred_bfnd);
              BIF_PROPLIST(bf1) = (PTR_PLNK) NULL;
            }

        | general_classname identifier 
            { PTR_SYMB symb1;
              symb1 = find_type_symbol($2,name_of_struct($1),area());
              if ((!symb1) ||
                /* (!same_scope(SYMB_SCOPE(symb1),pred_bfnd)) */ /* BW */
                   (SYMB_SCOPE(symb1) != pred_bfnd))
                {
                  symb1 = install_entry_2($2,name_of_struct($1),
                                          YES);
                }
              else 
                { if (/* same_scope(SYMB_SCOPE(symb1),pred_bfnd)&& */ /* BW */ 
                      (SYMB_TEMPLATE_DUMMY1(symb1) == BEEN_SEEN))
                   err("re_declaration symbol");
                }
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               node_of_struct($1),symb1,LLNULL);
              BIF_LL3(pred_bfnd) = NULL; /* was $3: OBSOLETE, I removed
                                           optionlist */
            }
          '{' component_decl_list '}'
            { PTR_TYPE type;
              PTR_BFND bf1;

              type = installStruct(&pred_bfnd,&new_cur_type,
                                       type_of_struct($1),SMNULL);
              $$=type;
              bf1 =  lastBifInBlobList(pred_bfnd);
              BIF_PROPLIST(bf1) = (PTR_PLNK) NULL;
            }

        | general_classname
            { char buffer[200];
              PTR_SYMB symb1;
              PTR_HASH entry;
              int status;
                 
              /*sprintf(buffer,"",enum_id++);*/
	      buffer[0]=0;
              entry = look_up_type(buffer,&status);
              symb1 = install_entry_2(entry,name_of_struct($1),
                                      YES);
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               node_of_struct($1),symb1,LLNULL); 
            }
          '{' component_decl_list  '}'
            { PTR_BFND bf1;

              $$ = installStruct(&pred_bfnd,&new_cur_type,
                                 type_of_struct($1),SMNULL);
              bf1 = lastBifInBlobList(pred_bfnd);
              BIF_PROPLIST(bf1) = (PTR_PLNK) NULL;
            }

        | general_classname identifier 
            { PTR_SYMB symb1;
                    
              symb1 = find_type_symbol($2,name_of_struct($1),area());
              if (!symb1)  
                 {  
                 symb1 = install_entry_2($2,name_of_struct($1),
                                         YES);
                 SYMB_TYPE(symb1) = (PTR_TYPE)NULL;
                 SYMB_TEMPLATE_DUMMY1(symb1) = NOT_SEEN;
                   }
              $$ = setBfndForStruct(pred_bfnd,&cur_flag,
                                    node_of_struct($1),symb1);
            }
        | ENUM identifier
            { PTR_SYMB symb1;
              symb1 = find_type_symbol($2,ENUM_NAME,area());
              if ((!symb1) ||
                  /* (!same_scope(SYMB_SCOPE(symb1),pred_bfnd)) */ /* BW */
                 (SYMB_SCOPE(symb1) != pred_bfnd))
                {
                  symb1 = install_entry_2($2,ENUM_NAME,YES); 
                                         /*dbg: was UP_TO_NODECL */
                }
              else 
                { if (/* same_scope(SYMB_SCOPE(symb1),pred_bfnd) &&  */ /* BW */
                     (SYMB_TEMPLATE_DUMMY1(symb1) == BEEN_SEEN))
                   err("re_declaration symbol");
                }
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               ENUM_DECL,symb1,LLNULL);
              }
            '{' enumlist '}'
               { 
               $$ = installStruct(&pred_bfnd,&new_cur_type,T_ENUM,SMNULL);
               }

        | ENUM 
            { char buffer[200];
              PTR_SYMB symb1;
              int status;
              PTR_HASH entry;
                 
              /*sprintf(buffer,"",enum_id++);*/
	      buffer[0]=0;
              entry = look_up_type(buffer,&status);
              symb1 = install_entry_2(entry,ENUM_NAME,YES);
              SYMB_TEMPLATE_DUMMY1(symb1) = BEEN_SEEN;
              innerLevelStruct(&pred_bfnd,&cur_flag,
                               ENUM_DECL,symb1,LLNULL); 
            }
          '{'  enumlist '}'
            { $$ = installStruct(&pred_bfnd,&new_cur_type,T_ENUM,SMNULL); }
  
        | ENUM identifier
            { PTR_SYMB symb1;
                 
              symb1 = find_type_symbol($2,ENUM_NAME,area());
              if (!symb1) {
                symb1 = install_entry_2($2,ENUM_NAME,YES);
                SYMB_TYPE(symb1) = (PTR_TYPE)NULL;
                SYMB_TEMPLATE_DUMMY1(symb1) = NOT_SEEN;
                }
              $$ = setBfndForStruct(pred_bfnd,&cur_flag,ENUM_DECL,symb1);
              }
        ;


general_classname:
          CLASS
            { $$ = CLASS; }
        | COLLECTION
            { $$ = COLLECTION;}
        | STRUCT
            { $$ = STRUCT; }
        | TECLASS
	    { $$ = TECLASS; }
        ;

temp_arg_spec:
       { $$ = NULL; }
       | '<'  template_arg_list '>' 
       { $$ = $2; }
        ;


TYPENAME_list:
          nested_name_spec_1 
            { 
              $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL);
            }

       | access_specifier nested_name_spec_1 
           { PTR_SYMB  symb;
	     PTR_TYPE  btype;

             $$ = installLowLevelNode(EXPR_LIST,$2,LLNULL,SMNULL);
             btype = (PTR_TYPE) newNode(T_DESCRIPT);
	     TYPE_BASE(btype) = NODE_TYPE($2);
             symb = NODE_SYMB($2);
             SYMB_TYPE(symb) = btype;
	     switch($1){
		case PUBLIC_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PUBLIC;
		    break;
		case PROTECTED_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PROTECTED;
		    break;
		case PRIVATE_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PRIVATE;
		    break;
		}
	     NODE_TYPE($2) =  btype;
           }

       | VIRTUAL access_specifier nested_name_spec_1 
           { PTR_SYMB  symb;
	     PTR_TYPE  btype;

             $$ = installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
             btype = (PTR_TYPE) newNode(T_DESCRIPT);
	     TYPE_BASE(btype) =  NODE_TYPE($3);
             symb = NODE_SYMB($3);
             SYMB_TYPE(symb) = btype;
	     switch($2){
		case PUBLIC_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PUBLIC | BIT_VIRTUAL;
		    break;
		case PROTECTED_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PROTECTED | BIT_VIRTUAL;
		    break;
		case PRIVATE_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PRIVATE | BIT_VIRTUAL;
		    break;
		}
	     NODE_TYPE($3) = btype;
           }

       | access_specifier VIRTUAL nested_name_spec_1 
           { PTR_SYMB  symb;
	     PTR_TYPE  btype;

             $$ = installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
             btype = (PTR_TYPE) newNode(T_DESCRIPT);
	     TYPE_BASE(btype) =  NODE_TYPE($3);
             symb = NODE_SYMB($3);
             SYMB_TYPE(symb) = btype;
	     switch($1){
		case PUBLIC_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PUBLIC | BIT_VIRTUAL;
		    break;
		case PROTECTED_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PROTECTED | BIT_VIRTUAL;
		    break;
		case PRIVATE_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PRIVATE | BIT_VIRTUAL;
		    break;
		}
	     NODE_TYPE($3) = btype;
           }

       | TYPENAME_list ',' nested_name_spec_1 
           { PTR_LLND  expr_list,ll_ptr;

             expr_list = installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
             ll_ptr = Follow_Llnd($1,2);
             NODE_OPERAND1(ll_ptr) = expr_list;
           }

       | TYPENAME_list ',' access_specifier nested_name_spec_1 
           { PTR_SYMB  symb;
             PTR_LLND  expr_list,ll_ptr;
	     PTR_TYPE btype;

             expr_list = installLowLevelNode(EXPR_LIST,$4,LLNULL,SMNULL);
             btype = (PTR_TYPE) newNode(T_DESCRIPT);
	     TYPE_BASE(btype) = NODE_TYPE($4);
             symb = NODE_SYMB($4);
             SYMB_TYPE(symb) = btype;
	     switch($3){
		case PUBLIC_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PUBLIC;
		    break;
		case PROTECTED_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PROTECTED;
		    break;
		case PRIVATE_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PRIVATE;
		    break;
		}
	     NODE_TYPE($4) = btype;
             ll_ptr = Follow_Llnd($1,2);
             NODE_OPERAND1(ll_ptr) = expr_list;
           }

       | TYPENAME_list ',' access_specifier VIRTUAL nested_name_spec_1 
           { PTR_SYMB  symb;
             PTR_LLND  expr_list,ll_ptr;
	     PTR_TYPE btype;

             expr_list = installLowLevelNode(EXPR_LIST,$5,LLNULL,SMNULL);
             btype = (PTR_TYPE) newNode(T_DESCRIPT);
	     TYPE_BASE(btype) = NODE_TYPE($5);
             symb = NODE_SYMB($5);
             SYMB_TYPE(symb) = btype;
	     switch($3){
		case PUBLIC_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PUBLIC | BIT_VIRTUAL;
		    break;
		case PROTECTED_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PROTECTED | BIT_VIRTUAL;
		    break;
		case PRIVATE_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PRIVATE | BIT_VIRTUAL;
		    break;
		}
	     NODE_TYPE($5) = btype;
             ll_ptr = Follow_Llnd($1,2);
             NODE_OPERAND1(ll_ptr) = expr_list;
           }

       | TYPENAME_list ',' VIRTUAL access_specifier nested_name_spec_1 
           { PTR_SYMB  symb;
             PTR_LLND  expr_list,ll_ptr;
	     PTR_TYPE btype;

             expr_list = installLowLevelNode(EXPR_LIST,$5,LLNULL,SMNULL);
             btype = (PTR_TYPE) newNode(T_DESCRIPT);
	     TYPE_BASE(btype) = NODE_TYPE($5);
             symb = NODE_SYMB($5);
             SYMB_TYPE(symb) = btype;
	     switch($4){
		case PUBLIC_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PUBLIC | BIT_VIRTUAL;
		    break;
		case PROTECTED_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PROTECTED | BIT_VIRTUAL;
		    break;
		case PRIVATE_FIELD:
                    btype->entry.Template.dummy5 |= BIT_PRIVATE | BIT_VIRTUAL;
		    break;
		}
	     NODE_TYPE($5) = btype;
             ll_ptr = Follow_Llnd($1,2);
             NODE_OPERAND1(ll_ptr) = expr_list;
           }
       ;


component_decl_list:   /* empty */
           { PTR_BFND bf1;
             bf1 = installBifNode(CONTROL_END,SMNULL,LLNULL,LLNULL,LLNULL);
             appendBfndToList(pred_bfnd,bf1);
            fill_comment();
           }
        | ';'
           { PTR_BFND bf1;
             bf1 = installBifNode(CONTROL_END,SMNULL,LLNULL,LLNULL,LLNULL);
             appendBfndToList(pred_bfnd,bf1);
            fill_comment();
           }
        | nonempty_component_decl_list
           {}   
        ;

/* modified */
nonempty_component_decl_list:
          component_decl
            { main_type_flag = TYPE_FOUR;
              first_symbol   = SMNULL;
            }
        | nonempty_component_decl_list  component_decl
            { main_type_flag = TYPE_FOUR;
              first_symbol   = SMNULL;
            }
        | nonempty_component_decl_list ';'
            { main_type_flag = TYPE_FOUR;
              first_symbol   = SMNULL;
            }   
        ;

/* modified */
component_decl:
          typed_declspecs components  ';'
            { PTR_SYMB symb1, dummy_symb;
              PTR_TYPE type1;
              type1 = (PTR_TYPE) $1;
              if(bif_level_decl_specs && type1 &&(TYPE_CODE(type1) == T_DERIVED_TYPE)){
                   /* this case is to handle the situation when */
                   /* you have a declartion in a class that is like */
                   /* friend a;  where a is another class */
		   /* this is not in the standard! */
		   warning("non-standard friend class-name; keword class required",0);
                   symb1 = TYPE_SYMB_DERIVE(type1);
                   setBfndForStruct(pred_bfnd,&cur_flag, CLASS_DECL,symb1);
		   }
              else addToDeclaration($1,OFF,$2,&dummy_symb,NO);
              fill_comment();
             }
        | typed_declspecs components '=' CONSTANT ';'
           {  PTR_SYMB dummy_symb;
              /* this case does pure functions.  */
              /*  we will use the SYMB_TEMPLATE_DUMMY8 field of   */
	      /*  the symbol to identify pure functions and const */
              /*  functions.  this is the local_size field:       */
              setPure($2);
              addToDeclaration($1,OFF,$2,&dummy_symb,NO);
              fill_comment();
            }
        /*
        | operhdr
            {}
         */
        | notype_component_declarator  ';'
            { PTR_BFND bifd;
              PTR_LLND exprlist_llnd;
              PTR_SYMB dummy_symb;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              NODE_OPERAND0(exprlist_llnd) = $1;
              addToDeclaration(GetAtomicType(DEFAULT),OFF,
                               exprlist_llnd,&dummy_symb,NO);
              fill_comment();        
            }

        | declmods   notype_components  ';'
            { PTR_SYMB dummy_symb;
              if($1 != NULL)
               {
                TYPE_DESCRIP_BASE_TYPE($1) = GetAtomicType(DEFAULT);
                addToDeclaration($1,OFF,$2,&dummy_symb,NO);
               }
              else
                addToDeclaration(GetAtomicType(DEFAULT),OFF,
                               $2,&dummy_symb,NO);
         
              fill_comment();
            }

        | typed_declspecs '('  parmlist ')' ';'
            { componentConstructorAction(&pred_bfnd,&cur_flag,
                                       $1,$3,CONSTRUCTOR,NO,NULL);
            }
        | '~' typed_declspecs '(' parmlist ')' ';'
            { componentConstructorAction(&pred_bfnd,&cur_flag,
                                          $2,$4,DESTRUCTOR,NO, NULL);
             }

        /* PHB, BW */
        | VIRTUAL '~' typed_declspecs '(' parmlist ')' ';' 
            { componentConstructorAction(&pred_bfnd,&cur_flag,
                                   $3,$5,VIRTUAL_DESTRUCTOR,NO, NULL);
             }
        | declmods '~' typed_declspecs '(' parmlist ')' ';' 
            { componentConstructorAction(&pred_bfnd,&cur_flag,
                                   $3,$5,DESTRUCTOR,NO, NULL);
             }
        | VIRTUAL '~' typed_declspecs '(' parmlist ')' '=' CONSTANT ';' 
            {PTR_LLND ll;
             ll = componentConstructorAction(&pred_bfnd,&cur_flag,
                                   $3,$5,VIRTUAL_DESTRUCTOR,NO, NULL);
             setPure(ll);
             }

        | '~' typed_declspecs '(' parmlist ')'
            { set_func_hedr2_with_typename($2,$4,1);
              lateAdjustFunctor(pred_bfnd);
            }

          compstmt
            {
              function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            }
        | typed_declspecs '(' parmlist ')' 
            { componentConstructorAction(&pred_bfnd,&cur_flag,
                                  $1,$3,CONSTRUCTOR,YES,NULL);
           } 
          base_init
            {  set_func_hedr3_with_typename();
              /* need something like  to back install init string  */
              lateAdjustFunctorWithInit(pred_bfnd, $6);
            }
          compstmt
            { function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            }                     

        | typed_declspecs  '('    parmlist ')'
            { set_func_hedr2_with_typename($1,$3,0);
              lateAdjustFunctor(pred_bfnd);
            }
         compstmt
            { function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            } 
        | ovldcast
            {}
        | fndef
            {}
        | error
            {}

        | access_specifier ':'
            { change_public_flag($1); }
        ;


ovldname:
     OPERATOR typed_declspecs star_ops '('    parmlist ')' ovldtypqual
	    {  
              PTR_HASH entry;
	      PTR_TYPE etype;
              int status;
	      char *name;
              if($3) etype = addToBaseTypeList($2, $3);
              else etype = $2;
              name = makeStringForType(etype);
              entry = look_up_type(name,&status);
              Op_Type_Qual  = 0;             
              Op_Name_Hash = entry;
              Op_Params = $5;
              $$ = etype;
	    }
     | storage_class OPERATOR typed_declspecs star_ops '('    parmlist ')' ovldtypqual
	    {  
              PTR_HASH entry;
	      PTR_TYPE etype, vir, ptr;
              int status;
	      char *name;
              if($4) etype = addToBaseTypeList($3, $4);
              else etype = $3;
              name = makeStringForType(etype);
              entry = look_up_type(name,&status);
              ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              vir = installDeclmod(ptr, $1);
              TYPE_DESCRIP_BASE_TYPE(vir) = etype;
              Op_Type_Qual  = $8;             
              Op_Name_Hash = entry;
              Op_Params = $6;
              $$ = vir;
	    }
     | storage_class2 OPERATOR typed_declspecs star_ops '('    parmlist ')' ovldtypqual
	    {  
              PTR_HASH entry;
	      PTR_TYPE etype /* , vir, ptr */ ;
              int status;
	      char *name;
              if($4) etype = addToBaseTypeList($3, $4);
              else etype = $3;
              name = makeStringForType(etype);
              entry = look_up_type(name,&status);
/*              ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              vir = installDeclmod(ptr, $1);
              TYPE_DESCRIP_BASE_TYPE(vir) = etype;        BW */
              Op_Type_Qual  = $8;             
              Op_Name_Hash = entry;
              Op_Params = $6;
              $$ = etype;
	    }
     ;


ovldtypqual:
       {$$ = 0;}
     | type_qualifier
       {$$ = $1;}
   ;

ovldcast:
        ovldname ';'
           {
             PTR_SYMB s;	      
             componentConstructorAction(&pred_bfnd,&cur_flag,
                                       $1, LLNULL,OVOPERATOR,NO, Op_Name_Hash);
            
             if(Op_Type_Qual == RID_CONST) {
                  s = getCurrentProcName(pred_bfnd); 
                  SYMB_TEMPLATE_DUMMY8(s) |=  64;  /* the key for const */
		  }
              fill_comment();
             }

        | ovldname 
            { PTR_SYMB s;
              s = set_func_cast_with_typename($1, Op_Params,0, Op_Name_Hash);
              if(Op_Type_Qual == RID_CONST) {
                  SYMB_TEMPLATE_DUMMY8(s) |=  64;  /* the key for const */
		  }
              lateAdjustFunctor(pred_bfnd);
            }
         compstmt
            { 
              function_flag = OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            }
    ;      

star_ops:
    {$$ = (PTR_TYPE) 0;}  
   | star_ops '*'
      { PTR_TYPE ptype, etype;
        ptype = (PTR_TYPE) newNode(T_POINTER);
        if($1) etype = addToBaseTypeList($1, ptype);
        else etype = ptype;
        $$ = ptype;  /* shouldn't this be etype??? BW */
       }
   ;


access_specifier:
          PUBLIC
            { $$ = PUBLIC_FIELD; }
        | PROTECTED
            { $$ = PROTECTED_FIELD; }
        | PRIVATE
            { $$ = PRIVATE_FIELD; }
        | METHODOFELEMENT                /* pC++ Extension */
            { $$ = ELEMENT_FIELD; }
        ;


storage_class:
          AUTO     {}
/*        | STATIC   {}  BW */
/*        | EXTERN   {}  BW */
        | TYPEDEF  {}
        | REGISTER {}
/*        | FRIEND   {}  BW */ /* C++, not ANSI C */
        | ATOMIC   {}
        | VIRTUAL  {} /* C++, not ANSI C */
/*        | INLINE   {}  BW */ /* C++, not ANSI C */
        ;

storage_class2:
          STATIC  {bif_level_decl_specs |= map_status ($1);}
        | EXTERN  {bif_level_decl_specs |= map_status ($1);}
        | FRIEND  {bif_level_decl_specs |= map_status ($1);}
        | INLINE  {bif_level_decl_specs |= map_status ($1);}
        ;


type_modifier:
          LONG     {}
        | SHORT    {}
        | SIGNED   {}
        | UNSIGNED {}
        | FUTURE   {} /* Unused pC++ extenstion */
        ;

type_qualifier:
          CONST    {}
	| SYNC	   {}
        | GLOBL    {}
        | VOLATILE {} /* C++, not ANSI C */
        | RESTRICT {}
        | KSRPRIVATE {}

basic_type_name:
          INT      {}
        | CHAR     {}
        | FLOAT    {}
        | DOUBLE   {}
        | VOID     {}
        ;

components:
          /* empty */
            { $$ = LLNULL; }
        | component_declarator
            { PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              NODE_OPERAND0(exprlist_llnd) = $1;
              $$ = exprlist_llnd;
            }
        | components ',' component_declarator
            { PTR_LLND ll_node1;
              PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              ll_node1 = Follow_Llnd(exprlist_llnd,2);
              NODE_OPERAND0(ll_node1) = $3;
              $$= exprlist_llnd;
            }
        ;


notype_components:
          /* empty */
            { $$ = LLNULL; }
        | notype_component_declarator
            { PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              NODE_OPERAND0(exprlist_llnd) = $1;
              $$ = exprlist_llnd;
            }
        | notype_components ',' component_declarator
            { PTR_LLND ll_node1;
              PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              ll_node1 = Follow_Llnd(exprlist_llnd,2);
              NODE_OPERAND0(ll_node1) = $3;
              $$= exprlist_llnd;
            }
        ;


component_declarator:
        declarator
            { $$ = $1; }
        | declarator ':' CONSTANT
            { PTR_SYMB symb;

              symb = getCurrentProcName(pred_bfnd);
              if (SYMB_CODE(symb) == FIELD_NAME)
                SYMB_RESTRICTED_BIT(symb) = $3;
              $$ = installLowLevelNode(BIT_NUMBER,$1,$3,SMNULL);
            }

        | ':' CONSTANT 
            { char buffer[40];
              int status,i;
              PTR_HASH pseudo_hash_entry;
              PTR_SYMB symb;
              PTR_LLND lp1;
              for (i=0; i<30; i++) buffer[i]= '\0';
              sprintf(buffer,"%s","");
              pseudo_hash_entry = look_up_type(buffer,&status);
              lp1 = lazyInstall(pred_bfnd, pseudo_hash_entry,
                                VARIABLE_NAME,NO);
              symb = getCurrentProcName(pred_bfnd);
              if (SYMB_CODE(symb) == FIELD_NAME)
                     SYMB_RESTRICTED_BIT(symb) = $2;
              $$ = installLowLevelNode(BIT_NUMBER,lp1,$2,SMNULL);
            }
        ;


notype_component_declarator:
          notype_declarator
            { $$ = $1; }
        | notype_declarator ':' CONSTANT 
            { PTR_SYMB symb;

              symb = getCurrentProcName(pred_bfnd);
              if (SYMB_CODE(symb) == FIELD_NAME)
                     SYMB_RESTRICTED_BIT(symb) = $3;
              $$ = installLowLevelNode(BIT_NUMBER,$1,$3,SMNULL);
            }

        | ':' CONSTANT 
            { char buffer[40];
              int status,i;
              PTR_HASH pseudo_hash_entry;
              PTR_SYMB symb;
              PTR_LLND lp1;

              for (i=0; i<30; i++) buffer[i]= '\0';
              sprintf(buffer,"%s%d","ZZZ_",num_symbs);
              pseudo_hash_entry = look_up_type(buffer,&status);
              lp1 = lazyInstall(pred_bfnd, pseudo_hash_entry,
                                VARIABLE_NAME,NO);
              symb = getCurrentProcName(pred_bfnd);
              if (SYMB_CODE(symb) == FIELD_NAME)
                     SYMB_RESTRICTED_BIT(symb) = $2;
              $$ = installLowLevelNode(BIT_NUMBER,lp1,$2,SMNULL);
            }
        ;


/* We chain the enumerators in reverse order.
   They are put in forward order where enumlist is used.
   (The order used to be significant, but no longer is so.
   However, we still maintain the order, just to be clean.)  */

enumlist:
          enumerator
            { PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              NODE_OPERAND0(exprlist_llnd) = $1;
              $$ = exprlist_llnd;
            }

        | enumlist ',' enumerator
            { PTR_LLND ll_node1;
              PTR_BFND bifd;
              PTR_LLND exprlist_llnd;

              bifd = lastBifInBlobList(pred_bfnd);
              exprlist_llnd = BIF_LL1(bifd);
              ll_node1 = Follow_Llnd(exprlist_llnd,2);
              NODE_OPERAND0(ll_node1) = $3;
              $$ = exprlist_llnd;
            }
        | enumlist ','
             {}
        ;


enumerator:
          identifier
            { PTR_TYPE ptr1;
              PTR_LLND llnd;
              PTR_SYMB dummy_symb;
              llnd = $$ = lazyInstall(pred_bfnd,$1,VARIABLE_NAME,YES);
              ptr1 = GetAtomicType(DEFAULT);
              /* notype_initdecls returns expr_list ll_node */
              addToDeclaration(ptr1,OFF,llnd,&dummy_symb,NO);
            }

        | identifier '=' expr_no_commas
            { PTR_LLND llnd;
              PTR_TYPE ptr1;
              PTR_SYMB dummy_symb;

              llnd = lazyInstall(pred_bfnd,$1,VARIABLE_NAME,YES);
              ptr1 = GetAtomicType(DEFAULT);
              /* notype_initdecls returns expr_list ll_node */
              addToDeclaration(ptr1,OFF,llnd,&dummy_symb,NO);
              $$= installLowLevelNode(ASSGN_OP,llnd,$3,SMNULL);
            }
        ;

typename:
          typed_typespecs absdcl
            { $$=addToBaseTypeList($1,$2); }
        | nonempty_typemods absdcl
            { PTR_TYPE new_type;
              new_type=addToBaseTypeList($1,$2);
              $$=addToBaseTypeList(GetAtomicType(DEFAULT),new_type);
            }
        ;
        
absdcl: /* an absolute declarator */
          /* empty */
           { $$ = (PTR_TYPE) NULL; }
        | absdcl1
           { $$ = $1; }
        ;

nonempty_typemods:
          type_modifier
           { PTR_TYPE ptr;
             ptr = (PTR_TYPE) newNode(T_DESCRIPT);
             $$ = installDeclmod(ptr, $1);
           }
        |  type_qualifier
           { PTR_TYPE ptr;
             ptr = (PTR_TYPE) newNode(T_DESCRIPT);
             $$ = installDeclmod(ptr, $1);
           }

        | nonempty_typemods type_modifier
           { $$ = installDeclmod($1,$2); }
        ;

typemods:
          /* empty */
            { $$ = (PTR_TYPE) NULL; }
        | typemods type_modifier
            { PTR_TYPE ptr;
              if (!$1) ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              else     ptr = $1;
              $$ = installDeclmod(ptr, $2);
            }
        | typemods type_qualifier
            { PTR_TYPE ptr;
              if (!$1) ptr = (PTR_TYPE) newNode(T_DESCRIPT);
              else     ptr = $1;
              $$ = installDeclmod(ptr, $2);
            }
        ;

absdcl1:  /* a nonempty absolute declarator */
          '(' absdcl1 ')'
            { $$ = $2; }

        | '*' typemods absdcl1 %prec SUPER_OP
            { PTR_TYPE type;
              type = (PTR_TYPE) newNode(T_POINTER);
              if($2) type->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
              $$ = addToBaseTypeList(type,$3);
            }

        | '&' typemods absdcl1 %prec SUPER_OP
            { PTR_TYPE type;
              type = (PTR_TYPE) newNode(T_REFERENCE);
              if($2) type->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
              $$ = addToBaseTypeList(type,$3);
            }

        | '*' typemods %prec SUPER_OP
            { 
	      PTR_TYPE type;
	      type = (PTR_TYPE) newNode(T_POINTER); 
              if($2) type->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
	      $$ = type;
	    }

        | '&' typemods  %prec SUPER_OP
	    {
	      PTR_TYPE type;
	      type = (PTR_TYPE) newNode(T_REFERENCE); 
              if($2) type->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
	      $$ = type;
	    }

        | absdcl1 '(' parmlist ')' %prec '.'
            { PTR_TYPE oldptr,temptr,type1;
              for (oldptr=temptr=$1;temptr;temptr = TYPE_BASE(temptr))
                oldptr = temptr;
              if (TYPE_CODE(oldptr) != T_POINTER) err("invalid typename");
              else {
                type1 = (PTR_TYPE) newNode(T_FUNCTION);
                TYPE_BASE(oldptr) = type1;
              }
              $$ = $1;
            }

        | absdcl1 '[' expr ']' %prec '.'
            { PTR_LLND p1;
              p1 = installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
              $$ = chainTypeWithArrayRange($1,p1);
            }

        | absdcl1 '[' ']' %prec '.'
            { PTR_LLND p1;
              p1 = installLowLevelNode(EXPR_LIST,LLNULL,LLNULL,SMNULL);
              $$ = chainTypeWithArrayRange($1,p1);
            }

        | '(' parmlist ')' %prec '.'
            { $$ = (PTR_TYPE) NULL; }

        | '[' expr ']' %prec '.'
            { PTR_TYPE type;
              PTR_LLND p1;
              type = (PTR_TYPE) newNode(T_ARRAY);
              p1 = installLowLevelNode(EXPR_LIST,$2,LLNULL,SMNULL);
              TYPE_DECL_RANGES(type) = p1;
              $$ =type;
             }

/*      |  new operand needed here  {} */

        | '[' ']'  %prec '.'
            { PTR_TYPE type;
              PTR_LLND p1;
              type = (PTR_TYPE) newNode(T_ARRAY);
              p1 = installLowLevelNode(EXPR_LIST,LLNULL,LLNULL,SMNULL);
              TYPE_DECL_RANGES(type) = p1;
              $$ =type;
            }
        ;

/* at least one statement, the first of which parses without error.  */
/* stmts is used only after decls, so an invalid first statement
   is actually regarded as an invalid decl and part of the decls.  */

stmts:
          stmt
            {}
        | stmts stmt
            {}
        | stmts errstmt
            {}
        ;

errstmt:
          error ';'
            {basic_block_flag = 0;}
        ;

pushlevel:  
          /* empty */
            { /* PTR_BLOB blob; */
              PTR_BFND bf1;

              /* blob = BIF_BLOB1(pred_bfnd); */
              if ( is_basic_block() || basic_block_flag)   
                { bf1 = installBifNode(BASIC_BLOCK,SMNULL,
                                       LLNULL,LLNULL,LLNULL);
                  appendBfndToList(pred_bfnd,bf1);
                  pred_bfnd = bf1;
                  basic_block_flag = 0;
                }
              else 
                { 
                 basic_block_flag = 1;
                }
 
             }
        ;

compstmt: 
          '{' '}'
            {}
        | '{' pushlevel stmts '}'
            { PTR_BFND bf1;
              if (BIF_CODE(pred_bfnd) == BASIC_BLOCK)
                { 
                  bf1 = installBifNode(CONTROL_END, SMNULL,
                                       LLNULL,LLNULL,LLNULL);
                  appendBfndToList(pred_bfnd,bf1); 
                  pred_bfnd = BIF_CP(pred_bfnd);
                }
            }
        | '{' pushlevel error '}'
            { if (BIF_CODE(pred_bfnd) == BASIC_BLOCK)
                    pred_bfnd = BIF_CP(pred_bfnd);
            }
        ;

simple_if:
          IF '(' expr ')'
            { PTR_BFND bf1;

              bf1 = installBifNode(IF_NODE,SMNULL,$3,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
             }        
          stmt
             { PTR_BFND bf1;
               bf1 = installBifNode(CONTROL_END,SMNULL,LLNULL,LLNULL,LLNULL);
               appendBfndToList(pred_bfnd,bf1);
               new_cur_bfnd = pred_bfnd;   /* for else using */
               pred_bfnd = BIF_CP(pred_bfnd);
             }
        ;


/* modified */
stmt:
          decl 
            { basic_block_flag = 0;
              main_type_flag = TYPE_FOUR; }

        | compstmt
            {basic_block_flag = 0;}

        | expr ';'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(EXPR_STMT_NODE,SMNULL,$1,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | simple_if ELSE 
            { PTR_BLOB blob;
              basic_block_flag = 0;
              if ( BIF_INDEX(pred_bfnd) == BRANCH_OFF)
                    blob = lastBlobInBlobList(pred_bfnd);
              else
                    blob = lastBlobInBlobList2(pred_bfnd);
              pred_bfnd = BLOB_VALUE(blob);
              BIF_INDEX(pred_bfnd) = BRANCH_ON;
            }
          stmt
            { PTR_BFND bf1;

              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              BIF_INDEX(pred_bfnd) = BRANCH_OFF;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | simple_if
            {basic_block_flag = 0;}

         /* the following is a first approximation to the try{} catch(){}
            statement.  we have parsed it as two statements and it should
            be only one.
          */
        | TRY
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(TRY_STAT,SMNULL,LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }
        | CATCH '(' parmlist ')'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(CATCH_STAT,SMNULL,LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
              BIF_LL1(pred_bfnd) = $3;
	      ResetSymbolTable(pred_bfnd, $3); 
            }
          stmt
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }
	
        | WHILE '(' expr ')'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(WHILE_NODE,SMNULL,$3,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | DO
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(DO_WHILE_NODE,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }

          stmt WHILE '(' expr ')' ';'
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              BIF_LL1(pred_bfnd) = $6;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | FOR '(' forinit ';' xexpr ';' xexpr ')'
            /* Don't let the tree nodes for $9 be discarded
               by clear_momentary during the parsing of the next stmt.  */
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(FOR_NODE,SMNULL,$3,$5,$7);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }
         /* note PARFOR is written as PERFER because of conflicts with */
         /* defines for PAR and FOR */
        | PERFER '(' forinit ';' xexpr ';' xexpr ')'
            /* Don't let the tree nodes for $9 be discarded
               by clear_momentary during the parsing of the next stmt.  */
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(PARFOR_NODE,SMNULL,$3,$5,$7);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }
        | PAR
            /* Don't let the tree nodes for $9 be discarded
               by clear_momentary during the parsing of the next stmt.  */
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(PAR_NODE,SMNULL,LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt
            { PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | SWITCH '(' expr ')'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(SWITCH_NODE,SMNULL,
                                   $3,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt
            {
              PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }

        | CASE expr ':'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(CASE_NODE,SMNULL,
                                   $2,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }
          stmt
            {}

        | DEFAULT_TOKEN ':'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(DEFAULT_NODE,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }
          stmt
            {}

        | BREAK ';'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(BREAK_NODE,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | CONTINUE ';'        
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(CONTINUE_NODE,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | RETURN ';'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(RETURN_NODE,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | RETURN expr ';'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(RETURN_NODE,SMNULL,
                                   $2,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | SPAWN expr ';'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(SPAWN_NODE,SMNULL,
                                   $2,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | ASM '(' STRING ')' ';'
            { PTR_BFND bf1;
              basic_block_flag = 0;
              bf1 = installBifNode(ASM_NODE,SMNULL,
                                   $3,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
            }

        | GOTO identifier ';'
            { PTR_LABEL temp_label;
              PTR_SYMB  temp_symb;
              PTR_BFND  bf1;
              basic_block_flag = 0;       
              bf1 = installBifNode(GOTO_NODE,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              temp_symb = find_type_symbol($2,LABEL_NAME,area());
              if (!temp_symb)
                { temp_label = InstallLabel($2);
                  LABEL_USED(temp_label)  = YES;
                }
              else
                { temp_label = find_label(temp_symb );
                  if (! temp_label) err("unexpected label");
                  else      LABEL_USED(temp_label)  = YES;
                }
              BIF_LABEL_USE(bf1) = temp_label;
            }    

        | identifier ':'
            {  PTR_LABEL temp_label;
               PTR_SYMB  temp_symb;
               PTR_BFND  bf1;
               basic_block_flag = 0;      
               bf1 = installBifNode(LABEL_STAT,SMNULL,
                                    LLNULL,LLNULL,LLNULL);
               appendBfndToList(pred_bfnd,bf1);
               temp_symb = find_type_symbol($1, LABEL_NAME,area());
               if (!temp_symb) 
                 {   temp_label = InstallLabel($1);
                     LABEL_DEFINED(temp_label) = YES;
                   }
               else
                 { temp_label = find_label(temp_symb );
                   if (temp_label)
                     {  if ((LABEL_DEFINED(temp_label)) == YES)
                          err("label redeclaration");
                        else   LABEL_DEFINED(temp_label) = YES;
                     }
                 }
               BIF_LABEL_USE(bf1) = temp_label;
               LABEL_BODY(temp_label) = bf1;
             } 
          stmt
             {}


        | FUTURE future_declarator '(' passing_identifier_list ')' 
            { PTR_BFND bf1;
              PTR_HASH entry;
              int status;
              PTR_SYMB symb;
              basic_block_flag = 0;
              bf1 = installBifNode(FUTURE_STMT,SMNULL,$2,$4,LLNULL);
              entry = look_up_type("$FUTURE",&status);
              symb  = install_entry_2(entry,VARIABLE_NAME,OTHER);
              BIF_SYMB(bf1) = symb;
              appendBfndToList(pred_bfnd,bf1);
              pred_bfnd = bf1;
            }
          stmt 
            {  
              PTR_BFND bf1;
              bf1 = installBifNode(CONTROL_END,SMNULL,
                                   LLNULL,LLNULL,LLNULL);
              appendBfndToList(pred_bfnd,bf1);
              new_cur_bfnd = pred_bfnd;
              pred_bfnd = BIF_CP(pred_bfnd);
            }


        | ';'
            {basic_block_flag = 0;}
        ;


passing_identifier_list:
          /* empty */
            { $$ = LLNULL;}     
        | passing_identifiers
            { $$ = $1;}     
        ;
 

passing_identifiers:
          IDENTIFIER
            { PTR_SYMB symbptr;
              PTR_LLND lptr;
              symbptr = find_type_symbol($1,VARIABLE_NAME,area());
              if (!symbptr) 
                symbptr=fts_by_inherited_class($1,VARIABLE_NAME,area());
              if (!symbptr)
                err("undefined symbol");
              else {
                lptr = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symbptr);
                $$ = installLowLevelNode(EXPR_LIST,lptr,LLNULL,SMNULL);
                }
            }

        | passing_identifiers ',' IDENTIFIER
            { PTR_SYMB symbptr;
              PTR_LLND lptr;
              PTR_LLND ll_node1;

              symbptr = find_type_symbol($3,VARIABLE_NAME,area());
              if (!symbptr) 
                symbptr=fts_by_inherited_class($3,VARIABLE_NAME,area());
              if (!symbptr)
                err("undefined symbol");
              else {
                lptr = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symbptr);
                lptr = installLowLevelNode(EXPR_LIST,lptr,LLNULL,SMNULL);
                ll_node1 = Follow_Llnd($1,2);
                NODE_OPERAND1(ll_node1) = lptr;
                }
            }
        ;


/* modified */
assign_stmt:
          IDENTIFIER  '='  vector_constant
            { PTR_SYMB symb;
              PTR_LLND ll_ptr1;
                 
              symb = find_type_symbol($1,VARIABLE_NAME,area()); 
              if (!symb) 
                symb = fts_by_inherited_class($1,VARIABLE_NAME,area());
              if (!symb) {
                err("undefined symbol");
                $$ = LLNULL;
                } 

              else {
                ll_ptr1 = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symb);
                $$ = installLowLevelNode(ASSGN_OP, ll_ptr1,$3,SMNULL);
                }
            }
       ;

xexpr:
          /* empty */
            { $$ = (PTR_LLND) LLNULL; }
        | expr
            { $$ = $1; }
        ;

forinit:
          xexpr 
             {$$ = $1; }
        |  parms 
            { $$ = installLowLevelNode(FORDECL_OP,$1,LLNULL,SMNULL); }
       ;

/* This is what appears inside the parens in a function declarator.
   Is value is represented in the format that grokdeclarator expects.  */
parmlist:  
          /* empty */
            { $$ = LLNULL; }
        | parms
            {}
        ;

ellipsis:
          ELLIPSIS
          /* A little unorthodox, but this rule creates 
             a variable called "..." as a parameter in the list (PHB) */
            { PTR_LLND ll;
              int status;
              PTR_HASH entry;
              PTR_SYMB symb;
              
              entry = look_up_type("...",&status);
              symb  = install_entry_2(entry,VARIABLE_NAME,OTHER);
              ll    = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symb);
              SYMB_TYPE(symb) = NODE_TYPE(ll) = NULL;
              $$ = ll;
            }
        ;

/* A nonempty list of parameter declarations or type names.  */
parms:        
          parm 
            { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); }
        | parms ',' parm
            { $$ = chainParameterList($1,$3); }
        | ellipsis  
            { $$ = installLowLevelNode(EXPR_LIST,$1,LLNULL,SMNULL); }
        | parms ',' ellipsis 
            { $$ = chainParameterList($1,$3); }
        | parms ellipsis 
            { $$ = chainParameterList($1,$2); }
        ;

parm:
        typed_declspecs_2 notype_declarator_parm parm_init
            { PTR_LLND ll;
	      /*
              if(bif_level_decl_specs != 0)
                  err("storage class specifiers invalid in parameter declarations");
	      */
              ll = giveLlSymbInDeclList($2);
              NODE_TYPE(ll) = addToDeclaration2($1,BIF_SYMB(ll));
              if ($3) $$ = installLowLevelNode(ASSGN_OP,$2,$3,SMNULL);
              else    $$ = $2;
            }
        | typed_declspecs_2 absdcl parm_init
            { PTR_LLND ll;
              int status;
              PTR_HASH entry;
              PTR_SYMB symb;
	      /*
              if(bif_level_decl_specs != 0)
                  err("storage class specifiers invalid in parameter declarations");
              */
              entry = look_up_type("",&status);
              symb  = install_entry_2(entry,VARIABLE_NAME,OTHER);
              /* the old way of handling blank parameters... wrong. dbg
	      if($2 && (NODE_CODE($2) == T_ARRAY)){
		  ll = TYPE_RANGES($2);
		  ll = installLowLevelNode(ARRAY_REF,ll,LLNULL,symb);
		  }
	      else ll = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symb);
              SYMB_TYPE(symb) = NODE_TYPE(ll) = addToBaseTypeList($1,$2);
              */
              if(!SYMB_TYPE(symb)) SYMB_TYPE(symb) = addToBaseTypeList($1,$2);
              ll = makeDeclarExp(symb, SYMB_TYPE(symb));
              if ($3) $$=  installLowLevelNode(ASSGN_OP,ll,$3,SMNULL);
              else    $$=  ll;
            }

        | declmods notype_declarator_parm parm_init
            { PTR_LLND ll;
              /*
              if($1 == NULL) 
                  err("storage class specifiers invalid in parameter declarations");
              */
	      ll = giveLlSymbInDeclList($2);
              NODE_TYPE(ll) =addToDeclaration2($1,BIF_SYMB(ll));
              if ($3) $$=  installLowLevelNode(ASSGN_OP,$2,$3,SMNULL);
              else    $$=  $2;
            }
        | declmods absdcl parm_init
            { PTR_TYPE  new_type;
              PTR_LLND ll; 
              int status;
              PTR_HASH entry;
              PTR_SYMB symb;
              /*
              if($1 == NULL) 
                   err("storage class specifiers invalid in parameter declarations");
              */
              entry = look_up_type("",&status);
              symb  = install_entry_2(entry,VARIABLE_NAME,OTHER);
          /* I've changed the code below to take arrays into account BW */
          /*    ll = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symb);   */
              new_type=addToBaseTypeList($1,$2);
         /*     NODE_TYPE(ll) = 
                       addToBaseTypeList(GetAtomicType(DEFAULT),new_type); */
              if(!SYMB_TYPE(symb)) SYMB_TYPE(symb) = /* NODE_TYPE(ll); */
                       addToBaseTypeList(GetAtomicType(DEFAULT),new_type);
              ll=makeDeclarExp(symb, SYMB_TYPE(symb));
              if ($3) $$=  installLowLevelNode(ASSGN_OP,ll,$3,SMNULL);
              else    $$=  ll;
            }
        ;


parm_init:
          /* empty */
            { $$= LLNULL; }
        | '=' init
            { $$ = $2;}
        ;


/* modified */
notype_declarator_parm:
          notype_declarator_parm '(' expr ')'      %prec '.'
            {  PTR_LLND ll;

               ll = $1;
               if ((cur_flag - RULE_ID) && 
                   (cur_flag - RULE_ID - RULE_ARRAY))
                       err("syntax error 3");
               else {
                   $$ = installLowLevelNode(CLASSINIT_OP,ll,$3,SMNULL);
                   cur_flag = cur_flag | RULE_CLASSINIT;
               }
            }

        | notype_declarator_parm '(' parmlist ')'  %prec '.'
            {
              $$ = handleComplicatedParameters($1,$3);
            }

        | notype_declarator_parm '(' identifiers ')'  %prec '.'
            { 
              $$ = handleComplicatedParameters($1,$3);
            }

        | notype_declarator_parm '(' error ')'  %prec '.'
            {}

        | '(' notype_declarator_parm ')'
            { 
              cur_flag = CLEAN;
              cur_flag = RULE_LR;
              $$ = $2;
            }

        | '*' typemods notype_declarator_parm  %prec UNARY
            { PTR_TYPE btype, type /* ,type_1 */ ;
              PTR_LLND ll,tmp;
              PTR_SYMB symbptr1;

              ll=$3;
              $$ = installLowLevelNode(DEREF_OP,ll,LLNULL,SMNULL);
              tmp = giveLlSymbInDeclList(ll);
              if (tmp)
              symbptr1  = NODE_SYMB(tmp);
              btype = (PTR_TYPE) newNode(T_POINTER);
              if($2){
                 btype->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
                 NODE_TYPE($$) = btype;
                  }

              if (! SYMB_TYPE(symbptr1)) 
                   type = SYMB_TYPE(symbptr1) = btype;
              else  /* no type checking here */
                { type = lookForInternalBasetype(SYMB_TYPE(symbptr1));
                  switch (TYPE_CODE(type)) {
                  case T_ARRAY :
                        /* type_1 = */ TYPE_DECL_BASE(type) = btype;
                        break;
                  case T_FUNCTION:
                  case T_POINTER :
                  case T_REFERENCE:
                        /* type_1 = */ TYPE_BASE(type) = btype;
                        break;
                  default : err("System Trouble 5");
                  }
                }
             }

        | tp_doublemark '*'  notype_declarator_parm
	     { PTR_TYPE btype, type;
               PTR_LLND ll,rhs, lhs, tmp;
               PTR_SYMB classname;
               PTR_SYMB symbptr1;
               ll = $3;
               classname = find_object_symbol($1);
               if(classname == NULL) err("could not find class name!");
               lhs = installLowLevelNode(VAR_REF,LLNULL, LLNULL, classname);
	       rhs = installLowLevelNode(DEREF_OP,ll, LLNULL, SMNULL);
               $$ = installLowLevelNode(SCOPE_OP,lhs,rhs,SMNULL);
               tmp = giveLlSymbInDeclList(ll);
               if (tmp)
                  symbptr1  = NODE_SYMB(tmp);
               btype = (PTR_TYPE) newNode(T_MEMBER_POINTER);
	       TYPE_COLL_NAME(btype) = classname;
	       TYPE_COLL_BASE(btype) = NULL;
               if (symbptr1 && (! SYMB_TYPE(symbptr1))) 
                   type = SYMB_TYPE(symbptr1) = btype;
               else { 
	          type = lookForInternalBasetype(SYMB_TYPE(symbptr1));	      	
                  type = (TYPE_COLL_BASE(btype) = type);
		  SYMB_TYPE(symbptr1) = type;
                  NODE_TYPE($$) = type;
                  }
           }

        | '&' typemods notype_declarator_parm  %prec UNARY
            { PTR_TYPE btype, type,type_1;
              PTR_LLND ll,tmp;
              PTR_SYMB symbptr1;

              ll=$3;
              $$ = installLowLevelNode(ADDRESS_OP,ll,LLNULL,SMNULL);
              tmp = giveLlSymbInDeclList(ll);
              if (tmp)
              symbptr1  = NODE_SYMB(tmp);
              btype = (PTR_TYPE) newNode(T_REFERENCE);
              if($2){
                 btype->entry.Template.dummy5 |= TYPE_LONG_SHORT($2);
                 NODE_TYPE($$) = btype;
                  }

              if (! SYMB_TYPE(symbptr1)) 
                  type = SYMB_TYPE(symbptr1) = btype;
              else  /* no type checking here */
                { type = lookForInternalBasetype(SYMB_TYPE(symbptr1));
                  switch (TYPE_CODE(type)) {
                  case T_ARRAY :
                       type_1 = TYPE_DECL_BASE(type) = btype;
                       break;
                  case T_FUNCTION:
                  case T_POINTER :
                  case T_REFERENCE:
                       type_1 = TYPE_BASE(type) = btype;
                       break;
                  default : err("System Trouble 6");
                  }
                }
            }

        | notype_declarator_parm '[' expr_no_commas ']'  %prec '.'
            { 
              $$=addRangeToArray(&new_cur_type,$1,$3);
              cur_flag = cur_flag | RULE_ARRAY;
            }

        | notype_declarator_parm '[' ']'  %prec '.'
            { 
              $$=addRangeToArray(&new_cur_type,$1,NULL);
              cur_flag = cur_flag | RULE_ARRAY;
            }

        | IDENTIFIER
            { $$ = installParameter(&new_cur_symb, $1, VARIABLE_NAME); }
        ;


future_declarator:
          future_declarator '[' expr_no_commas ']'  %prec '.'
            { int status;
              PTR_LLND ll_ptr,lp1;
 
              ll_ptr = checkArrayIdFormat($1,&status);
              switch (status) {
              case NO : err("syntax error");
                        break;
              case RANGE_APPEAR :
                        NODE_OPERAND1(ll_ptr) = 
                            installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
                        $$ = $1; 
                        break;
              case ARRAY_OP_NEED:
                lp1 = installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL); /*mod*/
                $$ = installLowLevelNode(ARRAY_OP,$1,lp1,SMNULL);
                break;
              case ID_ONLY :
              default:
                NODE_CODE(ll_ptr) = ARRAY_REF;
                NODE_OPERAND0(ll_ptr) = 
                   installLowLevelNode(EXPR_LIST,$3,LLNULL,SMNULL);
                $$ = $1;
                break;
              }
              NODE_TYPE($$) = adjustDerefType(NODE_TYPE($1),DEREF_OP);
            }

        | future_declarator '.' IDENTIFIER
            { PTR_LLND l_ptr2;
              int mytag;
              PTR_SYMB sw;

              sw= checkFieldDecl(pred_bfnd,&right_symbol, $1,$3,0,0);
              if (sw==NO){ err("error member"); $$ = LLNULL;}
              else {
                   mytag  = map2(NODE_CODE(right_symbol));
                   l_ptr2 = installLowLevelNode(mytag,LLNULL,
                                                LLNULL,right_symbol);
                   $$ = installLowLevelNode(RECORD_REF,$1,l_ptr2,SMNULL);
                   NODE_TYPE($$) = adjustArrayType(right_symbol);
                   NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
                   }
             }

        | future_declarator  POINTSAT  IDENTIFIER
            { PTR_LLND l_ptr;
              PTR_SYMB symb_ptr1;
              int      mytag;
 
              if (NODE_CODE($1) == THIS_NODE) {
                  symb_ptr1 = getFieldForThis(pred_bfnd,$3);
              }
              else {                     
                symb_ptr1= checkFieldDecl(pred_bfnd,NULL,$1, $3,0,1); 
              }
              mytag  = map2(SYMB_CODE(symb_ptr1));
              l_ptr = installLowLevelNode(mytag,LLNULL,LLNULL,symb_ptr1);
              $$ = installLowLevelNode(POINTST_OP,$1,l_ptr,SMNULL);
              NODE_TYPE($$) = adjustArrayType(symb_ptr1);
              NODE_TYPE($$) = adjustElementType($1,NODE_TYPE($$));
            }

        | IDENTIFIER
            { PTR_SYMB symbptr;
              symbptr = find_type_symbol($1,VARIABLE_NAME,area());
              if (!symbptr) 
                 symbptr=fts_by_inherited_class($1,VARIABLE_NAME,area());
              if (!symbptr) {
                 /* default function expected */
                  err("undefined symbol");
              }
              else {
                switch (SYMB_CODE(symbptr)) {
                case FIELD_NAME :
                case VARIABLE_NAME:
                default:
                  $$ = installLowLevelNode(VAR_REF,LLNULL,LLNULL,symbptr);
                  break;
                }
                NODE_TYPE($$) = adjustArrayType(symbptr);
              }
            }
        ;


/*
declarator_parm:
          notype_declarator_parm 
            {}
        | after_type_declarator_parm
            {}
        ;
*/

/* PHB Maybe this was intended for future use for references? */
/*ref_pointer:
         '*'
           { $$ = POINTST_OP; }
       | '&'
           { $$ = ADDRESS_OP; }
       ;
*/
