      program mg3_demo 
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c        'mg_demo' demonstrates the capabilities of a very simple multigrid
c        solver in computing a three dimensional potential field.  This is
c        a simplified multigrid solver in two important respects:
c
c                (1) it solves only a constant coefficient equation,
c                and that only on a uniform cubical grid,
c
c                (2) it solves only a single equation, representing
c                a scalar field rather than a vector field.
c
c        We chose it for its portability and simplicity, and expect that a
c        supercomputer which can run it effectively will also be able to
c        run more complex multigrid programs at least as well.
c
c              Eric Barszcz                        Paul Frederickson
c                                                      RIACS
c        NASA Ames Research Center            NASA Ames Research Center
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
      integer lm, nm, nv, nr, nit
      parameter( lm=5, nm=2+2**lm, nv=nm**3 )
      parameter( nr = (8*(nm**3+nm**2+5*nm-23+7*lm))/7 )
      parameter( nit=4 )
c
      double precision u(nr),v(nv),r(nr),a(0:3),c(0:3)
      double precision rnm2, rnmu, old2, oldu
      integer ir(lm), m1(lm), m2(lm), m3(lm)
      integer it, n, lt, lb
      common /fap/ ir,m1,m2,m3,lt,lb,a,c,u,v,r
c
      a(0) = -8.0D0/3.0D0 
      a(1) =  0.0D0 
      a(2) =  1.0D0/6.0D0 
      a(3) =  1.0D0/12.0D0
c
      c(0) =  -3.0D0/8.0D0
      c(1) =  1.0D0/32.0D0
      c(2) =  -1.0D0/64.0D0
      c(3) =   0.0D0
c
      lt = lm
      lb = 1
      n  = 2 + 2**lt
c
      write(6,7)n-2,n-2,n-2,nit
c
      call setup(n,n,n)
      call zero3(u,n,n,n)
      call zran3(v,n,n,n)
      call resid(u,v,r,n,a)
      call norm2u3(r,n,n,n,rnm2,rnmu,n,n,n)
           old2 = rnm2
           oldu = rnmu
c
      do 20 it=1,nit
        call mg3P(n,n,n)
        call resid(u,v,r,n,a)
 20   continue
c
        call norm2u3(r,n,n,n,rnm2,rnmu,n,n,n)
        write(6,*)
        write(6,*)' it    L2-norm(R)         unif-norm(R)'
        write(6,6) 0, old2, oldu
        write(6,6) nit, rnm2, rnmu
 6    format( i4, 2e19.12)
 7    format(/,' Kernel B:  Solving a Poisson problem on a ',i3,' by ',
     > i3,' by ',i3,' grid,',/,' using ',i1,' multigrid iterations.',/) 
c
      stop
      end
*
******  mg3P   implements a simple, constant coefficient version of the
******        MG algorithm FAPIN (fast approximate inverse)
*
      subroutine mg3P(n1,n2,n3)
c
      integer lm, nm, nv, nr
      parameter( lm=5, nm=2+2**lm, nv=nm**3 )
      parameter( nr = (8*(nm**3+nm**2+5*nm-23+7*lm))/7 )
c
      double precision u(nr),v(nv),r(nr)
      double precision a(0:3),c(0:3)
      integer ir(lm), m1(lm), m2(lm), m3(lm)
      integer lt, lb
      common /fap/ ir,m1,m2,m3,lt,lb,a,c,u,v,r
      integer n1, n2, n3, k, j
c
      do 50 k= lt, lb+1, -1
      j=k-1
      call rprj3(r(ir(k)),m1(k),m2(k),m3(k),r(ir(j)),m1(j),m2(j),m3(j))
 50   continue
c
      k =lb
      call zero3(u(ir(k)),m1(k),m2(k),m3(k))
      call psinv(r(ir(k)),u(ir(k)),m1(k),c)
      if( lt .eq. lb+1 )go to 200
      do 100 k =lb+1, lt-1
      j = k-1
      call zero3(u(ir(k)),m1(k),m2(k),m3(k))
      call interp(u(ir(j)),m1(j),m2(j),m3(j),u(ir(k)),m1(k),m2(k),m3(k))
      call resid(u(ir(k)),r(ir(k)),r(ir(k)),m1(k),a)
      call psinv(r(ir(k)),u(ir(k)),m1(k),c)
 100  continue
 200  continue
      j = lt - 1
      k =lt
      call interp(u(ir(j)),m1(j),m2(j),m3(j),u,n1,n2,n3)
      call resid(u,v,r,n1,a)
      call psinv(r,u,n1,c)
c
      return
      end
*
****** psinv applies an approximate inverse as smoother:  u = u + Cr
*
*      This simple implementation costs  27A + 4M per result, where
*      A and M denote the costs of Addition and Multiplication.  
*      By using several two-dimensional buffers one can reduce this
*      cost to  13A + 4M in the general case, or  11A + 3M when the
*      coefficient c(3) is zero.
*
      subroutine psinv(r,u,n,c)
      integer n
      double precision u(n,n,n),r(n,n,n),c(0:3)
      integer i3, i2, i1
c
      do 600 i3=2,n-1
      do 600 i2=2,n-1
      do 600 i1=2,n-1
 600  u(i1,i2,i3)=u(i1,i2,i3)
     >      +c(0)*( r(i1,  i2,  i3  ) )
     >      +c(1)*( r(i1-1,i2,  i3  ) + r(i1+1,i2,  i3  )
     >           +  r(i1,  i2-1,i3  ) + r(i1,  i2+1,i3  )
     >           +  r(i1,  i2,  i3-1) + r(i1,  i2,  i3+1) )
     >      +c(2)*( r(i1-1,i2-1,i3  ) + r(i1+1,i2-1,i3  )
     >           +  r(i1-1,i2+1,i3  ) + r(i1+1,i2+1,i3  )
     >           +  r(i1,  i2-1,i3-1) + r(i1,  i2+1,i3-1)
     >           +  r(i1,  i2-1,i3+1) + r(i1,  i2+1,i3+1)
     >           +  r(i1-1,i2,  i3-1) + r(i1-1,i2,  i3+1)
     >           +  r(i1+1,i2,  i3-1) + r(i1+1,i2,  i3+1) )
     >      +c(3)*( r(i1-1,i2-1,i3-1) + r(i1+1,i2-1,i3-1)
     >           +  r(i1-1,i2+1,i3-1) + r(i1+1,i2+1,i3-1)
     >           +  r(i1-1,i2-1,i3+1) + r(i1+1,i2-1,i3+1)
     >           +  r(i1-1,i2+1,i3+1) + r(i1+1,i2+1,i3+1) )
c
         call comm3(u,n,n,n)
c
      return
      end
*
******* resid computes the residual:  r = v - Au
*
*      This simple implementation costs  27A + 4M per result, where
*      A and M denote the costs of Addition (or Subtraction) and 
*      Multiplication, respectively.  By using several two-dimensional 
*      buffers one can reduce this cost to  13A + 4M in the general 
*      case, or  10A + 3M when the coefficient a(1) is zero.
*
      subroutine resid(u,v,r,n,a)
      integer n
      double precision u(n,n,n),v(n,n,n),r(n,n,n),a(0:3)
      integer i3, i2, i1
c
      do 600 i3=2,n-1
      do 600 i2=2,n-1
      do 600 i1=2,n-1
 600  r(i1,i2,i3)=v(i1,i2,i3)
     >      -a(0)*( u(i1,  i2,  i3  ) )
     >      -a(1)*( u(i1-1,i2,  i3  ) + u(i1+1,i2,  i3  )
     >                 +  u(i1,  i2-1,i3  ) + u(i1,  i2+1,i3  )
     >                 +  u(i1,  i2,  i3-1) + u(i1,  i2,  i3+1) )
     >      -a(2)*( u(i1-1,i2-1,i3  ) + u(i1+1,i2-1,i3  )
     >                 +  u(i1-1,i2+1,i3  ) + u(i1+1,i2+1,i3  )
     >                 +  u(i1,  i2-1,i3-1) + u(i1,  i2+1,i3-1)
     >                 +  u(i1,  i2-1,i3+1) + u(i1,  i2+1,i3+1)
     >                 +  u(i1-1,i2,  i3-1) + u(i1-1,i2,  i3+1)
     >                 +  u(i1+1,i2,  i3-1) + u(i1+1,i2,  i3+1) )
     >      -a(3)*( u(i1-1,i2-1,i3-1) + u(i1+1,i2-1,i3-1)
     >                 +  u(i1-1,i2+1,i3-1) + u(i1+1,i2+1,i3-1)
     >                 +  u(i1-1,i2-1,i3+1) + u(i1+1,i2-1,i3+1)
     >                 +  u(i1-1,i2+1,i3+1) + u(i1+1,i2+1,i3+1) )
c
         call comm3(r,n,n,n)
c
      return
      end
*
****** rprj3 projects onto the next coarser grid, 
****** using a trilinear Finite Element projection:  s = r' = P r
*
*      This simple implementation costs  27A + 4M per result, where
*      A and M denote the costs of Addition and Multiplication.  
*      By using several two-dimensional buffers one can reduce this
*      cost to  23A + 1M.  
*
      subroutine rprj3(r,m1k,m2k,m3k,s,m1j,m2j,m3j)
      integer m1k, m2k, m3k, m1j, m2j, m3j
      double precision r(m1k,m2k,m3k), s(m1j,m2j,m3j)
      integer j3, j2, j1, i3, i2, i1
c
      do 100 j3=2,m3j-1
        i3 = 2*j3-1
      do 100 j2=2,m2j-1
        i2 = 2*j2-1
      do 100 j1=2,m1j-1
        i1 = 2*j1-1
 100  s(j1,j2,j3) =
     >      0.5D0 * ( r(i1,  i2,  i3  ) )
     >   + 0.25D0 * ( r(i1-1,i2,  i3  ) + r(i1+1,i2,  i3  )
     >             +  r(i1,  i2-1,i3  ) + r(i1,  i2+1,i3  )
     >             +  r(i1,  i2,  i3-1) + r(i1,  i2,  i3+1) )
     >  + 0.125D0 * ( r(i1-1,i2-1,i3  ) + r(i1+1,i2-1,i3  )
     >             +  r(i1-1,i2+1,i3  ) + r(i1+1,i2+1,i3  )
     >             +  r(i1,  i2-1,i3-1) + r(i1,  i2+1,i3-1)
     >             +  r(i1,  i2-1,i3+1) + r(i1,  i2+1,i3+1)
     >             +  r(i1-1,i2,  i3-1) + r(i1-1,i2,  i3+1)
     >             +  r(i1+1,i2,  i3-1) + r(i1+1,i2,  i3+1) )
     > + 0.0625D0 * ( r(i1-1,i2-1,i3-1) + r(i1+1,i2-1,i3-1)
     >             +  r(i1-1,i2+1,i3-1) + r(i1+1,i2+1,i3-1)
     >             +  r(i1-1,i2-1,i3+1) + r(i1+1,i2-1,i3+1)
     >             +  r(i1-1,i2+1,i3+1) + r(i1+1,i2+1,i3+1) )
c
      call comm3(s,m1j,m2j,m3j)
c
      return
      end
*
****** interp adds the trilinear interpolation of the correction
****** from the coarser grid to the current approximation:  u = u + Qu'
*
*      Observe that this simple implementation costs  27A + 4M, where
*      A and M denote the costs of Addition and Multiplication.  
*      By using several two-dimensional buffers one can reduce this
*      cost to  (15/8)A + (7/8)M.
*
      subroutine interp(z,m1,m2,m3,u,n1,n2,n3)
      double precision z(m1,m2,m3),u(n1,n2,n3)
      integer m1, m2, m3, n1, n2, n3, mm1, mm2, mm3
      integer i3, i2, i1
c
      if( (m1-1)*2 .eq. n1 )then
         mm1 = m1-1
      else
         mm1 = m1
      endif
c
      if( (m2-1)*2 .eq. n2 )then
         mm2 = m2-1
      else
         mm2 = m2
      endif
c
      if( (m3-1)*2 .eq. n3 )then
         mm3 = m3-1
      else
         mm3 = m3
      endif
c
      do 400 i3=2,mm3
         do 200 i2=2,mm2
            do 100 i1=2,mm1
            u(2*i1-1,2*i2-1,2*i3-1)=u(2*i1-1,2*i2-1,2*i3-1)
     >      +z(i1,i2,i3)
 100        continue
            do 200 i1=2,mm1
            u(2*i1-2,2*i2-1,2*i3-1)=u(2*i1-2,2*i2-1,2*i3-1)
     >      +0.5D0*(z(i1-1,i2,i3)+z(i1,i2,i3))
 200     continue
         do 400 i2=2,mm2
            do 300 i1=2,mm1
            u(2*i1-1,2*i2-2,2*i3-1)=u(2*i1-1,2*i2-2,2*i3-1)
     >      +0.5D0*(z(i1,i2-1,i3)+z(i1,i2,i3))
 300        continue
               do 400 i1=2,mm1
            u(2*i1-2,2*i2-2,2*i3-1)=u(2*i1-2,2*i2-2,2*i3-1)
     >      +0.25D0*(z(i1-1,i2-1,i3)+z(i1-1,i2,i3)
     >            +z(i1,  i2-1,i3)+z(i1,  i2,i3))
 400  continue
c
      do 800 i3=2,mm3
         do 600 i2=2,mm2
            do 500 i1=2,mm1
            u(2*i1-1,2*i2-1,2*i3-2)=u(2*i1-1,2*i2-1,2*i3-2)
     >      +0.5D0*(z(i1,i2,i3-1)+z(i1,i2,i3))
 500        continue
            do 600 i1=2,mm1
            u(2*i1-2,2*i2-1,2*i3-2)=u(2*i1-2,2*i2-1,2*i3-2)
     >      +0.25D0*(z(i1-1,i2,i3-1)+z(i1,i2,i3-1)
     >            +z(i1-1,i2,i3  )+z(i1,i2,i3  ))
 600     continue
         do 800 i2=2,mm2
            do 700 i1=2,mm1
            u(2*i1-1,2*i2-2,2*i3-2)=u(2*i1-1,2*i2-2,2*i3-2)
     >      +0.25D0*(z(i1,i2-1,i3-1)+z(i1,i2,i3-1)
     >            +z(i1,i2-1,i3  )+z(i1,i2,i3  ))
 700        continue
               do 800 i1=2,mm1
            u(2*i1-2,2*i2-2,2*i3-2)=u(2*i1-2,2*i2-2,2*i3-2)
     >      +0.125D0*(z(i1-1,i2-1,i3-1)+z(i1-1,i2,i3-1)
     >             +z(i1,  i2-1,i3-1)+z(i1,  i2,i3-1)
     >             +z(i1-1,i2-1,i3  )+z(i1-1,i2,i3  )
     >             +z(i1,  i2-1,i3  )+z(i1,  i2,i3  ))
 800  continue
c
        call comm3(u,n1,n2,n3)
c
      return 
      end
*
****** comm3 communicates on all borders of an array
*
      subroutine comm3(u,n1,n2,n3)
      integer n1, n2, n3
      double precision u(n1,n2,n3)
      integer i3, i2, i1
c
      do 100 i3=2,n3-1
        do 100 i2=2,n2-1
          u( 1,i2,i3) = u(n1-1,i2,i3)
          u(n1,i2,i3) = u( 2,  i2,i3)
 100  continue
c
      do 200 i3=2,n3-1
        do 200 i1=1,n1
          u(i1, 1,i3) = u(i1,n2-1,i3)
          u(i1,n2,i3) = u(i1, 2,  i3)
 200  continue

      do 300 i2=1,n2
        do 300 i1=1,n1
          u(i1,i2, 1) = u(i1,i2,n3-1)
          u(i1,i2,n3) = u(i1,i2,  2 )
 300  continue
c
      return
      end
*
****** norm2u3 evaluates approximations to the L2 norm and the
****** uniform (or L-infinity or Chebyshev) norm, under the
****** assumption that the boundaries are periodic or zero.  Add the
****** boundaries in with half weight (quarter weight on the edges
****** and eigth weight at the corners) for inhomogeneous boundaries.
*
      subroutine norm2u3(r,n1,n2,n3,rnm2,rnmu,nx,ny,nz)
      integer n1, n2, n3, nx, ny, nz
      double precision rnm2, rnmu, r(n1,n2,n3)
      double precision s, a
      integer i3, i2, i1
c
      s=0.0D0
      rnmu = 0.0D0
      do 100 i3=2,n3-1
         do 100 i2=2,n2-1
            do 100 i1=2,n1-1
               s=s+r(i1,i2,i3)**2
               a=abs(r(i1,i2,i3))
               if(a.gt.rnmu)rnmu=a
 100  continue
      rnm2=sqrt(s/float((nx-2)*(ny-2)*(nz-2)))
c
      return
      end
*
****** setup   must be called before mg3P
*
      subroutine setup(n1,n2,n3)
      integer n1, n2, n3
c
      integer lm, nm, nv, nr
      parameter( lm=5, nm=2+2**lm, nv=nm**3 )
      parameter( nr = (8*(nm**3+nm**2+5*nm-23+7*lm))/7 )
c
      double precision u(nr),v(nv),r(nr)
      double precision a(0:3),c(0:3)
      integer ir(lm), m1(lm), m2(lm), m3(lm)
      integer lt, lb
      common /fap/ ir,m1,m2,m3,lt,lb,a,c,u,v,r
      integer l, k
c
        m1(lt) = n1
        m2(lt) = n2
        m3(lt) = n3
        do 100 k = lt-1, 1, -1
          m1(k) = 1 + m1(k+1)/2
          m2(k) = 1 + m2(k+1)/2
          m3(k) = 1 + m3(k+1)/2
 100     continue
            l = lt
c
            ir(l)=1
            do 200 k = l-1, 1, -1
            ir(k)=ir(k+1)+m1(k+1)*m2(k+1)*m3(k+1)
 200        continue
c
      return
      end
*
***
*
      subroutine zero3(z,n1,n2,n3)
      integer n1, n2, n3
      double precision z(n1,n2,n3)
      integer i1, i2, i3
c
      do 100 i3=1,n3
      do 100 i2=1,n2
      do 100 i1=1,n1
 100  z(i1,i2,i3)=0.0D0
c
      return
      end
*
****** zran3  loads +1 at ten randomly chosen points,
******        loads -1 at a different ten random points,
******        and zero elsewhere.
*
      subroutine zran3(z,n1,n2,n3)
      integer n1, n2, n3
      double precision z(n1,n2,n3)
      integer m, i1, i2, i3
      double precision x, a
      parameter( m = 10,  a = 5.D0 ** 13, x = 314159265.D0)
      double precision ten( m, 0:1 )
      integer n, i, j1( m, 0:1 ), j2( m, 0:1 ), j3( m, 0:1 )
c
      do 100 i3=2,n3-1
      do 100 i2=2,n2-1
        call vranlc( n1-2, x, a, z(2,i2,i3) )
 100  continue
c
	do 200 i=1,m
	  ten( i, 1 ) = 0.0D0
	  ten( i, 0 ) = 1.0D0
 200	continue
c
      do 300 i3=2,n3-1
      do 300 i2=2,n2-1
      do 300 i1=2,n1-1
	if( z(i1,i2,i3) .gt. ten(1,1) )then
	  ten(1,1) = z(i1,i2,i3) 
	  j1(1,1) = i1
	  j2(1,1) = i2
	  j3(1,1) = i3
	  call bubble( ten, j1, j2, j3, m, 1 )
        endif
        if( z(i1,i2,i3) .lt. ten( 1, 0) )then
          ten(1,0) = z(i1,i2,i3) 
	  j1(1,0) = i1
	  j2(1,0) = i2
	  j3(1,0) = i3
	  call bubble( ten, j1, j2, j3, m, 0 )
	endif
 300	continue
c 
        write(6,*)' negative charges at'
	write(6,7)(j1(i,0),j2(i,0),j3(i,0),i=1,m)
        write(6,*)' positive charges at'
	write(6,7)(j1(i,1),j2(i,1),j3(i,1),i=1,m)
 7	format(5(' (',i3,2(',',i3),')'))
c
      do 400 i3=1,n3
      do 400 i2=1,n2
      do 400 i1=1,n1
	z(i1,i2,i3) = 0.0D0
 400  continue
c
	do 500 i=1,m
	  z( j1(i,0), j2(i,0), j3(i,0) ) = -1.0D0
	  z( j1(i,1), j2(i,1), j3(i,1) ) = 1.0D0
 500	continue
c
        call comm3(z,n1,n2,n3)
c
      return 
      return
      end
*
***** bubble	does a bubble sort in direction dir
*
	subroutine bubble( ten, j1, j2, j3, m, ind )
	double precision ten( m, 0:1 )
        integer m, ind, j1( m, 0:1 ), j2( m, 0:1 ), j3( m, 0:1 )
	double precision dir, temp
	integer i, j_temp
c
	if( ind .eq. 1 )then
	  dir = 1.0D0
	else
	  dir = -1.0D0
	endif
c
	do 100 i=1,m-1
	  if( dir*ten(i,ind) .gt. dir*ten(i+1,ind) )then
	    temp = ten( i+1, ind )
	    ten( i+1, ind ) = ten( i, ind )
	    ten( i, ind ) = temp
c
	    j_temp	   = j1( i+1, ind )
	    j1( i+1, ind ) = j1( i,   ind )
	    j1( i,   ind ) = j_temp
c
	    j_temp	   = j2( i+1, ind )
	    j2( i+1, ind ) = j2( i,   ind )
	    j2( i,   ind ) = j_temp
c
	    j_temp	   = j3( i+1, ind )
	    j3( i+1, ind ) = j3( i,   ind )
	    j3( i,   ind ) = j_temp
c
	  else 
	    return
	  endif
c
 100	continue
c
	return
	end
*
***** randlc 
*

      DOUBLE PRECISION FUNCTION RANDLC (X, A)
C
C   This routine returns a uniform pseudorandom double precision number in the
C   range (0, 1) by using the linear congruential generator
C
C   x_{k+1} = a x_k  (mod 2^46)
C
C   where 0 < x_k < 2^46 and 0 < a < 2^46.  This scheme generates 2^44 numbers
C   before repeating.  The argument A is the same as 'a' in the above formula,
C   and X is the same as x_0.  A and X must be odd double precision integers
C   in the range (1, 2^46).  The returned value RANDLC is normalized to be
C   between 0 and 1, i.e. RANDLC = 2^(-46) x_1.  X is updated to contain
C   the new seed x_1, so that subsequent calls to RANDLC using the same
C   arguments will generate a continuous sequence.
C
C   This routine should produce the same results on any computer with at least
C   48 mantissa bits in double precision floating point data.  On Cray systems,
C   double precision should be disabled.
C
C   David H. Bailey    August 30, 1990
C
      IMPLICIT DOUBLE PRECISION (A-H, O-Z)
      PARAMETER (R23 = 2.D0 ** (-23), R46 = R23 * R23, T23 = 2.D0 ** 23,
     $  T46 = T23 * T23)
C
C   Break A into two parts such that A = 2^23 * A1 + A2 and set X = N.
C
      T1 = R23 * A
      A1 = AINT (T1)
      A2 = A - T23 * A1
C
C   Break X into two parts such that X = 2^23 * X1 + X2, compute
C   Y = A1 * X2 + A2 * X1  (mod 2^23), and then
C   X = 2^23 * Y + A2 * X2  (mod 2^46).
C
      T1 = R23 * X
      X1 = AINT (T1)
      X2 = X - T23 * X1
      T1 = A1 * X2 + A2 * X1
      T2 = AINT (R23 * T1)
      Y = T1 - T23 * T2
      T3 = T23 * Y + A2 * X2
      T4 = AINT (R46 * T3)
      X = T3 - T46 * T4
      RANDLC = R46 * X
C
      RETURN
      END
C
      SUBROUTINE VRANLC (N, X, A, Y)
C
C   This routine generates N uniform pseudorandom double precision numbers in
C   the range (0, 1) by using the linear congruential generator
C
C   x_{k+1} = a x_k  (mod 2^46)
C
C   where 0 < x_k < 2^46 and 0 < a < 2^46.  This scheme generates 2^44 numbers
C   before repeating.  The argument A is the same as 'a' in the above formula,
C   and X is the same as x_0.  A and X must be odd double precision integers
C   in the range (1, 2^46).  The N results are placed in Y and are normalized
C   to be between 0 and 1.  X is updated to contain the new seed, so that
C   subsequent calls to RANDLC using the same arguments will generate a
C   continuous sequence.
C
C   This routine generates the output sequence in batches of length NV, for
C   convenience on vector computers.  This routine should produce the same
C   results on any computer with at least 48 mantissa bits in double precision
C   floating point data.  On Cray systems, double precision should be disabled.
C
C   David H. Bailey    August 30, 1990
C
      integer n
      double precision x, a, y(n)
c
      double precision r23, r46, t23, t46
      integer nv
      PARAMETER (R23 = 2.D0 ** (-23), R46 = R23 * R23, T23 = 2.D0 ** 23,
     $  T46 = T23 * T23, NV = 64)
      double precision  xv(nv), t1, t2, t3, t4, an, a1, a2, x1, x2, yy
      integer n1, i, j
      double precision randlc
C
C   Compute the first NV elements of the sequence using RANDLC.
C
      T1 = X
      N1 = MIN (N, NV)
C
      DO 100 I = 1, N1
        XV(I) = T46 * RANDLC (T1, A)
 100  CONTINUE
C
C   It is not necessary to compute AN, A1 or A2 unless N is greater than NV.
C
      IF (N .GT. NV) THEN
C
C   Compute AN = AA ^ NV (mod 2^46) using successive calls to RANDLC.
C
        T1 = A
        T2 = R46 * A
C
        DO 110 I = 1, NV - 1
          T2 = RANDLC (T1, A)
 110    CONTINUE
C
        AN = T46 * T2
C
C   Break AN into two parts such that AN = 2^23 * A1 + A2.
C
        T1 = R23 * AN
        A1 = AINT (T1)
        A2 = AN - T23 * A1
      ENDIF
C
C   Compute N pseudorandom results in batches of size NV.
C
      DO 140 J = 0, N - 1, NV
        N1 = MIN (NV, N - J)
C
C   Compute up to NV results based on the current seed vector XV.
C
CDIR$ IVDEP
CDIR$ SHORTLOOP
        DO 120 I = 1, N1
          Y(I+J) = R46 * XV(I)
 120    CONTINUE
C
C   If this is the last pass through the 140 loop, it is not necessary to
C   update the XV vector.
C
        IF (J + N1 .EQ. N) GOTO 150
C
C   Update the XV vector by multiplying each element by AN (mod 2^46).
C
CDIR$ IVDEP
CDIR$ SHORTLOOP          
          DO 130 I = 1, NV
            T1 = R23 * XV(I)
            X1 = AINT (T1)
            X2 = XV(I) - T23 * X1
            T1 = A1 * X2 + A2 * X1
            T2 = AINT (R23 * T1)
            YY = T1 - T23 * T2
            T3 = T23 * YY + A2 * X2
            T4 = AINT (R46 * T3)
            XV(I) = T3 - T46 * T4
 130      CONTINUE
C
 140  CONTINUE
C
C   Save the last seed in X so that subsequent calls to VRANLC will generate
C   a continuous sequence.
C
 150  X = XV(N1)
      RETURN
      END

